/* Do not modify this file. Changes will be overwritten.                      */
/* Generated automatically by the ASN.1 to Wireshark dissector compiler       */
/* packet-rnsap.c                                                             */
/* asn2wrs.py -L -p rnsap -c ./rnsap.cnf -s ./packet-rnsap-template -D . -O ../.. RNSAP-CommonDataTypes.asn RNSAP-Constants.asn RNSAP-Containers.asn RNSAP-IEs.asn RNSAP-PDU-Contents.asn RNSAP-PDU-Descriptions.asn */

/* packet-rnsap.c
 * Routines for dissecting Universal Mobile Telecommunications System (UMTS);
 * UTRAN Iur interface Radio Network Subsystem
 * Application Part (RNSAP) signalling
 * (3GPP TS 25.423 version 6.7.0 Release 6) packet dissection
 * Copyright 2005 - 2006, Anders Broman <anders.broman@ericsson.com>
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Ref: 3GPP TS 25.423 version 6.7.0 Release 6
 */

#include "config.h"

#include <epan/packet.h>
#include <epan/asn1.h>
#include <epan/proto_data.h>

#include "packet-isup.h"
#include "packet-per.h"
#include "packet-ber.h"

#ifdef _MSC_VER
/* disable: "warning C4146: unary minus operator applied to unsigned type, result still unsigned" */
#pragma warning(disable:4146)
#endif

#define PNAME  "UTRAN Iur interface Radio Network Subsystem Application Part"
#define PSNAME "RNSAP"
#define PFNAME "rnsap"

#define SCCP_SSN_RNSAP 143

#define maxPrivateIEs                  65535
#define maxProtocolExtensions          65535
#define maxProtocolIEs                 65535
#define maxCellSIB11OrSIB12            32
#define maxCellSIB11ter_ext            48
#define maxCellSIB11ter_1              79
#define maxCellsMeas                   8
#define maxRateMatching                256
#define maxNoOfDSCHs                   10
#define maxNoOfDSCHsLCR                10
#define maxNoOfRB                      32
#define maxNoOfUSCHs                   10
#define maxNoOfUSCHsLCR                10
#define maxNrOfTFCs                    1024
#define maxNrOfTFs                     32
#define maxNrOfCCTrCHs                 16
#define maxNrOfCCTrCHsLCR              16
#define maxNrOfDCHs                    128
#define maxNrOfDL_Codes                8
#define maxNrOfDPCHs                   240
#define maxNrOfDPCHsPerRL_1            239
#define maxNrOfDPCHsLCR                240
#define maxNrOfDPCHsLCRPerRL_1         95
#define maxNrOfDPCHs768                480
#define maxNrOfDPCHs768PerRL_1         479
#define maxNrOfErrors                  256
#define maxNrOfMACcshSDU_Length        16
#define maxNrOfMBMSServices            128
#define maxNrOfActiveMBMSServices      256
#define maxNrOfPoints                  15
#define maxNrOfRLs                     16
#define maxNrOfRLSets                  maxNrOfRLs
#define maxNrOfRLSets_1                15
#define maxNrOfRLs_1                   15
#define maxNrOfRLs_2                   14
#define maxNrOfUEs                     4096
#define maxNrOfULTs                    15
#define maxNrOfULTsLCR                 6
#define maxNrOfDLTs                    15
#define maxNrOfDLTsLCR                 6
#define maxRNCinURA_1                  15
#define maxTTI_Count                   4
#define maxCTFC                        16777215
#define maxNrOfNeighbouringRNCs        10
#define maxNrOfFDDNeighboursPerRNC     256
#define maxNrOfGSMNeighboursPerRNC     256
#define maxNrOfTDDNeighboursPerRNC     256
#define maxNrOfFACHs                   8
#define maxNrOfLCRTDDNeighboursPerRNC  256
#define maxIBSEG                       16
#define maxNrOfSCCPCHs                 8
#define maxNrOfSCCPCHs768              16
#define maxTGPS                        6
#define maxNrOfTS                      15
#define maxNrOfLevels                  256
#define maxNrOfTsLCR                   6
#define maxNoSat                       16
#define maxNoGPSTypes                  8
#define maxNrOfMeasNCell               96
#define maxNrOfMeasNCell_1             95
#define maxResetContext                250
#define maxResetContextGroup           32
#define maxNrOfHARQProc                8
#define maxNrOfHSSCCHCodes             4
#define maxNrOfHSSICHs                 4
#define maxNrOfHSSICHs_1               3
#define maxNrOfMACdFlows               8
#define maxNrOfMACdFlows_1             7
#define maxNrOfMACdPDUSize             32
#define maxNrOfPDUIndexes              8
#define maxNrOfPDUIndexes_1            7
#define maxNrOfPrioQueues              8
#define maxNrOfPrioQueues_1            7
#define maxNrOfSNAs                    65536
#define maxNrOfSatAlmanac_maxNoSat     16
#define maxNrOfGERANSI                 8
#define maxNrOfInterfaces              16
#define maxNrofSigSeqERGHICH_1         39
#define maxNrOfCells                   65536
#define maxNrOfAddFreq                 8
#define maxNrOfCellsPerFreq            65536
#define maxNrOfEDCHMACdFlows_1         7
#define maxNrOfEDCH_HARQ_PO_QUANTSTEPs 6
#define maxNrOfEDPCCH_PO_QUANTSTEPs    8
#define maxNrOfEDCHHARQProcesses2msEDCH 8
#define maxNrOfBits_MACe_PDU_non_scheduled 19982
#define maxNrOfRefETFCIs               8
#define maxNrOfRefETFCI_PO_QUANTSTEPs  29
#define maxNrOfEDCHMACdFlows           8
#define maxNoOfLogicalChannels         16
#define maxNrOfRefBetas                8
#define maxNrOfEAGCHCodes              4
#define maxNrOfHS_DSCHTBSs             90
#define maxNrOfHS_DSCHTBSs_HS_SCCHless 4
#define maxHS_PDSCHCodeNrComp_1        15
#define maxNrOfEHICHCodes              4
#define maxGANSSSat                    64
#define maxNoGANSS                     8
#define maxSgnType                     8
#define maxNrOfBroadcastPLMNs          5
#define maxHSDPAFrequency              8
#define maxHSDPAFrequency_1            7
#define maxFrequencyinCell             12
#define maxFrequencyinCell_1           11
#define maxGANSSSatAlmanac             36
#define maxGANSSClockMod               4
#define maxNrOfEDCHRLs                 4
#define maxEARFCN                      65535
#define maxEARFCN_Extended             262143
#define maxNrOfEUTRANeighboursPerRNC   256
#define maxNrOfMCCHMessages            5
#define maxNrOfMBMSL3                  64
#define maxNrOfEDCHMACdFlowsLCR        256
#define maxNrOfEDCHMACdFlowsLCR_1      255
#define maxNrOfPreconfiguredNeighbours 256
#define maxNrOfHSDSCH_1                32
#define maxNrOfHSDSCH                  33
#define maxGANSS_1                     7
#define maxlengthMBMSconcatservlists   96
#define maxNoOfTBSs_Mapping_HS_DSCH_SPS 4
#define maxNoOfTBSs_Mapping_HS_DSCH_SPS_1 3
#define maxNoOfHS_DSCH_TBSsLCR         64
#define maxNoOfRepetition_Period_LCR   4
#define maxNoOfRepetitionPeriod_SPS_LCR_1 3
#define maxNoOf_HS_SICH_SPS            4
#define maxNoOf_HS_SICH_SPS_1          3
#define maxNoOfNon_HS_SCCH_Assosiated_HS_SICH 4
#define maxNrOfEDCH_1                  32
#define maxNrOfDCHMeasurementOccasionPatternSequence 6
#define maxNrOfULCarriersLCR_1         5
#define maxNrOfCellIds                 32
#define maxNrOfRAIs                    8
#define maxNrOfLAIs                    8
#define maxNrOfExtendedNeighbouringRNCs 64
#define maxNoOfGsmCell                 128
#define maxNrOfANRCells                256
#define maxFreqBandsTDD                16
#define maxSCPICHCell                  32
#define maxSCPICHCell_1                31
#define maxNoOfCommonRGCells           256
#define maxIGPInfo                     320
#define maxNrofConcatenatedDCH         3
#define maxMultipleFrequencyBands      8
#define maxMultipleFrequencyBandsEUTRA 8

/* enumerated values for ProcedureCode */
#define RNSAP_ID_COMMONTRANSPORTCHANNELRESOURCESINITIALISATION   0
#define RNSAP_ID_COMMONTRANSPORTCHANNELRESOURCESRELEASE   1
#define RNSAP_ID_COMPRESSEDMODECOMMAND   2
#define RNSAP_ID_DOWNLINKPOWERCONTROL   3
#define RNSAP_ID_DOWNLINKPOWERTIMESLOTCONTROL   4
#define RNSAP_ID_DOWNLINKSIGNALLINGTRANSFER   5
#define RNSAP_ID_ERRORINDICATION   6
#define RNSAP_ID_DEDICATEDMEASUREMENTFAILURE   7
#define RNSAP_ID_DEDICATEDMEASUREMENTINITIATION   8
#define RNSAP_ID_DEDICATEDMEASUREMENTREPORTING   9
#define RNSAP_ID_DEDICATEDMEASUREMENTTERMINATION  10
#define RNSAP_ID_PAGING  11
#define RNSAP_ID_PHYSICALCHANNELRECONFIGURATION  12
#define RNSAP_ID_PRIVATEMESSAGE  13
#define RNSAP_ID_RADIOLINKADDITION  14
#define RNSAP_ID_RADIOLINKCONGESTION  34
#define RNSAP_ID_RADIOLINKDELETION  15
#define RNSAP_ID_RADIOLINKFAILURE  16
#define RNSAP_ID_RADIOLINKPREEMPTION  17
#define RNSAP_ID_RADIOLINKRESTORATION  18
#define RNSAP_ID_RADIOLINKSETUP  19
#define RNSAP_ID_RELOCATIONCOMMIT  20
#define RNSAP_ID_SYNCHRONISEDRADIOLINKRECONFIGURATIONCANCELLATION  21
#define RNSAP_ID_SYNCHRONISEDRADIOLINKRECONFIGURATIONCOMMIT  22
#define RNSAP_ID_SYNCHRONISEDRADIOLINKRECONFIGURATIONPREPARATION  23
#define RNSAP_ID_UNSYNCHRONISEDRADIOLINKRECONFIGURATION  24
#define RNSAP_ID_UPLINKSIGNALLINGTRANSFER  25
#define RNSAP_ID_COMMONMEASUREMENTFAILURE  26
#define RNSAP_ID_COMMONMEASUREMENTINITIATION  27
#define RNSAP_ID_COMMONMEASUREMENTREPORTING  28
#define RNSAP_ID_COMMONMEASUREMENTTERMINATION  29
#define RNSAP_ID_INFORMATIONEXCHANGEFAILURE  30
#define RNSAP_ID_INFORMATIONEXCHANGEINITIATION  31
#define RNSAP_ID_INFORMATIONREPORTING  32
#define RNSAP_ID_INFORMATIONEXCHANGETERMINATION  33
#define RNSAP_ID_RESET  35
#define RNSAP_ID_RADIOLINKACTIVATION  36
#define RNSAP_ID_GERANUPLINKSIGNALLINGTRANSFER  37
#define RNSAP_ID_RADIOLINKPARAMETERUPDATE  38
#define RNSAP_ID_UEMEASUREMENTFAILURE  39
#define RNSAP_ID_UEMEASUREMENTINITIATION  40
#define RNSAP_ID_UEMEASUREMENTREPORTING  41
#define RNSAP_ID_UEMEASUREMENTTERMINATION  42
#define RNSAP_ID_IURDEACTIVATETRACE  43
#define RNSAP_ID_IURINVOKETRACE  44
#define RNSAP_ID_MBMSATTACH  45
#define RNSAP_ID_MBMSDETACH  46
#define RNSAP_ID_NOT_USED_47  47
#define RNSAP_ID_DIRECTINFORMATIONTRANSFER  48
#define RNSAP_ID_ENHANCEDRELOCATION  49
#define RNSAP_ID_ENHANCEDRELOCATIONCANCEL  50
#define RNSAP_ID_ENHANCEDRELOCATIONSIGNALLINGTRANSFER  51
#define RNSAP_ID_ENHANCEDRELOCATIONRELEASE  52
#define RNSAP_ID_MBSFNMCCHINFORMATION  53
#define RNSAP_ID_SECONDARYULFREQUENCYREPORTING  54
#define RNSAP_ID_SECONDARYULFREQUENCYUPDATE  55
#define RNSAP_ID_INFORMATIONTRANSFERCONTROL  56
#define RNSAP_ID_NOT_USED_57  57
#define RNSAP_ID_NOT_USED_58  58
#define RNSAP_ID_NOT_USED_59  59
#define RNSAP_ID_ENHANCEDRELOCATIONRESOURCEALLOCATION  60
#define RNSAP_ID_ENHANCEDRELOCATIONRESOURCERELEASE  61

typedef enum _DdMode_enum {
  tdd          =   0,
  fdd          =   1,
  common       =   2
} DdMode_enum;

typedef enum _ProtocolIE_ID_enum {
  id_AllowedQueuingTime =   4,
  id_Allowed_Rate_Information =  42,
  id_AntennaColocationIndicator = 309,
  id_BindingID =   5,
  id_C_ID      =   6,
  id_C_RNTI    =   7,
  id_Cell_Capacity_Class_Value = 303,
  id_CFN       =   8,
  id_CN_CS_DomainIdentifier =   9,
  id_CN_PS_DomainIdentifier =  10,
  id_Cause     =  11,
  id_CoverageIndicator = 310,
  id_CriticalityDiagnostics =  20,
  id_ContextInfoItem_Reset = 211,
  id_ContextGroupInfoItem_Reset = 515,
  id_D_RNTI    =  21,
  id_D_RNTI_ReleaseIndication =  22,
  id_DCHs_to_Add_FDD =  26,
  id_DCHs_to_Add_TDD =  27,
  id_DCH_DeleteList_RL_ReconfPrepFDD =  30,
  id_DCH_DeleteList_RL_ReconfPrepTDD =  31,
  id_DCH_DeleteList_RL_ReconfRqstFDD =  32,
  id_DCH_DeleteList_RL_ReconfRqstTDD =  33,
  id_DCH_FDD_Information =  34,
  id_DCH_TDD_Information =  35,
  id_FDD_DCHs_to_Modify =  39,
  id_TDD_DCHs_to_Modify =  40,
  id_DCH_InformationResponse =  43,
  id_DCH_Rate_InformationItem_RL_CongestInd =  38,
  id_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD =  44,
  id_DL_CCTrCH_InformationListIE_RL_ReconfReadyTDD =  45,
  id_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD =  46,
  id_DL_CCTrCH_InformationItem_RL_SetupRqstTDD =  47,
  id_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD =  48,
  id_DL_CCTrCH_InformationListIE_RL_AdditionRspTDD =  49,
  id_DL_CCTrCH_InformationListIE_RL_SetupRspTDD =  50,
  id_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD =  51,
  id_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD =  52,
  id_DL_CCTrCH_InformationList_RL_SetupRqstTDD =  53,
  id_FDD_DL_CodeInformation =  54,
  id_DL_DPCH_Information_RL_ReconfPrepFDD =  59,
  id_DL_DPCH_Information_RL_SetupRqstFDD =  60,
  id_DL_DPCH_Information_RL_ReconfRqstFDD =  61,
  id_DL_DPCH_InformationItem_PhyChReconfRqstTDD =  62,
  id_DL_DPCH_InformationItem_RL_AdditionRspTDD =  63,
  id_DL_DPCH_InformationItem_RL_SetupRspTDD =  64,
  id_DL_DPCH_TimingAdjustment = 278,
  id_DLReferencePower =  67,
  id_DLReferencePowerList_DL_PC_Rqst =  68,
  id_DL_ReferencePowerInformation_DL_PC_Rqst =  69,
  id_DPC_Mode  =  12,
  id_DRXCycleLengthCoefficient =  70,
  id_DedicatedMeasurementObjectType_DM_Fail_Ind = 470,
  id_DedicatedMeasurementObjectType_DM_Fail = 471,
  id_DedicatedMeasurementObjectType_DM_Rprt =  71,
  id_DedicatedMeasurementObjectType_DM_Rqst =  72,
  id_DedicatedMeasurementObjectType_DM_Rsp =  73,
  id_DedicatedMeasurementType =  74,
  id_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD =  82,
  id_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD =  83,
  id_Guaranteed_Rate_Information =  41,
  id_IMSI      =  84,
  id_HCS_Prio  = 311,
  id_L3_Information =  85,
  id_AdjustmentPeriod =  90,
  id_MaxAdjustmentStep =  91,
  id_MeasurementFilterCoefficient =  92,
  id_MessageStructure =  57,
  id_MeasurementID =  93,
  id_Neighbouring_GSM_CellInformation =  13,
  id_Neighbouring_UMTS_CellInformationItem =  95,
  id_Unknown_305 = 304,
  id_NRT_Load_Information_Value = 305,
  id_NRT_Load_Information_Value_IncrDecrThres = 306,
  id_PagingArea_PagingRqst = 102,
  id_FACH_FlowControlInformation = 103,
  id_PartialReportingIndicator = 472,
  id_Permanent_NAS_UE_Identity =  17,
  id_PowerAdjustmentType = 107,
  id_Unknown_108 = 108,
  id_RANAP_RelocationInformation = 109,
  id_RL_Information_PhyChReconfRqstFDD = 110,
  id_RL_Information_PhyChReconfRqstTDD = 111,
  id_RL_Information_RL_AdditionRqstFDD = 112,
  id_RL_Information_RL_AdditionRqstTDD = 113,
  id_RL_Information_RL_DeletionRqst = 114,
  id_RL_Information_RL_FailureInd = 115,
  id_RL_Information_RL_ReconfPrepFDD = 116,
  id_RL_Information_RL_RestoreInd = 117,
  id_RL_Information_RL_SetupRqstFDD = 118,
  id_RL_Information_RL_SetupRqstTDD = 119,
  id_RL_InformationItem_RL_CongestInd =  55,
  id_RL_InformationItem_DM_Rprt = 120,
  id_RL_InformationItem_DM_Rqst = 121,
  id_RL_InformationItem_DM_Rsp = 122,
  id_RL_InformationItem_RL_PreemptRequiredInd =   2,
  id_RL_InformationItem_RL_SetupRqstFDD = 123,
  id_RL_InformationList_RL_CongestInd =  56,
  id_RL_InformationList_RL_AdditionRqstFDD = 124,
  id_RL_InformationList_RL_DeletionRqst = 125,
  id_RL_InformationList_RL_PreemptRequiredInd =   1,
  id_RL_InformationList_RL_ReconfPrepFDD = 126,
  id_RL_InformationResponse_RL_AdditionRspTDD = 127,
  id_RL_InformationResponse_RL_ReconfReadyTDD = 128,
  id_RL_InformationResponse_RL_SetupRspTDD = 129,
  id_RL_InformationResponseItem_RL_AdditionRspFDD = 130,
  id_RL_InformationResponseItem_RL_ReconfReadyFDD = 131,
  id_RL_InformationResponseItem_RL_ReconfRspFDD = 132,
  id_RL_InformationResponseItem_RL_SetupRspFDD = 133,
  id_RL_InformationResponseList_RL_AdditionRspFDD = 134,
  id_RL_InformationResponseList_RL_ReconfReadyFDD = 135,
  id_RL_InformationResponseList_RL_ReconfRspFDD = 136,
  id_RL_InformationResponse_RL_ReconfRspTDD =  28,
  id_RL_InformationResponseList_RL_SetupRspFDD = 137,
  id_RL_ReconfigurationFailure_RL_ReconfFail = 141,
  id_RL_Set_InformationItem_DM_Rprt = 143,
  id_RL_Set_InformationItem_DM_Rqst = 144,
  id_RL_Set_InformationItem_DM_Rsp = 145,
  id_RL_Set_Information_RL_FailureInd = 146,
  id_RL_Set_Information_RL_RestoreInd = 147,
  id_RL_Set_Successful_InformationItem_DM_Fail = 473,
  id_RL_Set_Unsuccessful_InformationItem_DM_Fail = 474,
  id_RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind = 475,
  id_RL_Successful_InformationItem_DM_Fail = 476,
  id_RL_Unsuccessful_InformationItem_DM_Fail = 477,
  id_RL_Unsuccessful_InformationItem_DM_Fail_Ind = 478,
  id_ReportCharacteristics = 152,
  id_Reporting_Object_RL_FailureInd = 153,
  id_Reporing_Object_RL_RestoreInd = 154,
  id_RT_Load_Value = 307,
  id_RT_Load_Value_IncrDecrThres = 308,
  id_S_RNTI    = 155,
  id_ResetIndicator = 244,
  id_RNC_ID    = 245,
  id_SAI       = 156,
  id_SRNC_ID   = 157,
  id_SuccessfulRL_InformationResponse_RL_AdditionFailureFDD = 159,
  id_SuccessfulRL_InformationResponse_RL_SetupFailureFDD = 160,
  id_Unknown_161 = 161,
  id_Unknown_162 = 162,
  id_TransportBearerID = 163,
  id_TransportBearerRequestIndicator = 164,
  id_TransportLayerAddress = 165,
  id_TypeOfError = 140,
  id_UC_ID     = 166,
  id_UL_CCTrCH_AddInformation_RL_ReconfPrepTDD = 167,
  id_Unknown_168 = 168,
  id_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD = 169,
  id_Unknown_170 = 170,
  id_UL_CCTrCH_InformationItem_RL_SetupRqstTDD = 171,
  id_UL_CCTrCH_InformationList_RL_SetupRqstTDD = 172,
  id_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD = 173,
  id_UL_CCTrCH_InformationListIE_RL_AdditionRspTDD = 174,
  id_UL_CCTrCH_InformationListIE_RL_ReconfReadyTDD = 175,
  id_UL_CCTrCH_InformationListIE_RL_SetupRspTDD = 176,
  id_UL_DPCH_Information_RL_ReconfPrepFDD = 177,
  id_UL_DPCH_Information_RL_ReconfRqstFDD = 178,
  id_UL_DPCH_Information_RL_SetupRqstFDD = 179,
  id_UL_DPCH_InformationItem_PhyChReconfRqstTDD = 180,
  id_UL_DPCH_InformationItem_RL_AdditionRspTDD = 181,
  id_UL_DPCH_InformationItem_RL_SetupRspTDD = 182,
  id_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD = 183,
  id_UL_SIRTarget = 184,
  id_URA_Information = 185,
  id_Unknown_186 = 186,
  id_Unknown_187 = 187,
  id_UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD = 188,
  id_UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD = 189,
  id_UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD = 190,
  id_Unknown_191 = 191,
  id_Unknown_192 = 192,
  id_Active_Pattern_Sequence_Information = 193,
  id_AdjustmentRatio = 194,
  id_Unknown_195 = 195,
  id_Unknown_196 = 196,
  id_CauseLevel_RL_AdditionFailureFDD = 197,
  id_CauseLevel_RL_AdditionFailureTDD = 198,
  id_CauseLevel_RL_ReconfFailure = 199,
  id_CauseLevel_RL_SetupFailureFDD = 200,
  id_CauseLevel_RL_SetupFailureTDD = 201,
  id_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD = 205,
  id_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD = 206,
  id_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD = 207,
  id_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD = 208,
  id_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD = 209,
  id_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD = 210,
  id_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD = 212,
  id_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD = 213,
  id_DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD = 214,
  id_DSCHs_to_Add_TDD = 215,
  id_Unused_ProtocolIE_ID_216 = 216,
  id_DSCH_DeleteList_RL_ReconfPrepTDD = 217,
  id_Unused_ProtocolIE_ID_218 = 218,
  id_Unused_ProtocolIE_ID_219 = 219,
  id_DSCH_InformationListIE_RL_AdditionRspTDD = 220,
  id_DSCH_InformationListIEs_RL_SetupRspTDD = 221,
  id_DSCH_TDD_Information = 222,
  id_Unused_ProtocolIE_ID_223 = 223,
  id_Unused_ProtocolIE_ID_226 = 226,
  id_DSCH_ModifyList_RL_ReconfPrepTDD = 227,
  id_Unused_ProtocolIE_ID_228 = 228,
  id_Unused_ProtocolIE_ID_324 = 324,
  id_Unused_ProtocolIE_ID_229 = 229,
  id_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD = 230,
  id_Unknown_231 = 231,
  id_Unused_ProtocolIE_ID_29 =  29,
  id_Unused_ProtocolIE_ID_225 = 225,
  id_GA_Cell   = 232,
  id_GA_CellAdditionalShapes =   3,
  id_Unused_ProtocolIE_ID_246 = 246,
  id_Unknown_254 = 254,
  id_Transmission_Gap_Pattern_Sequence_Information = 255,
  id_UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD = 256,
  id_UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD = 257,
  id_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD = 258,
  id_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD = 259,
  id_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD = 260,
  id_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD = 261,
  id_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD = 262,
  id_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD = 263,
  id_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD = 264,
  id_UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD = 265,
  id_UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD = 266,
  id_USCHs_to_Add = 267,
  id_USCH_DeleteList_RL_ReconfPrepTDD = 268,
  id_USCH_InformationListIE_RL_AdditionRspTDD = 269,
  id_USCH_InformationListIEs_RL_SetupRspTDD = 270,
  id_USCH_Information = 271,
  id_USCH_ModifyList_RL_ReconfPrepTDD = 272,
  id_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD = 273,
  id_DL_Physical_Channel_Information_RL_SetupRqstTDD = 274,
  id_UL_Physical_Channel_Information_RL_SetupRqstTDD = 275,
  id_ClosedLoopMode1_SupportIndicator = 276,
  id_Unused_ProtocolIE_ID_277 = 277,
  id_STTD_SupportIndicator = 279,
  id_CFNReportingIndicator =  14,
  id_CNOriginatedPage_PagingRqst =  23,
  id_InnerLoopDLPCStatus =  24,
  id_PropagationDelay =  25,
  id_RxTimingDeviationForTA =  36,
  id_timeSlot_ISCP =  37,
  id_CCTrCH_InformationItem_RL_FailureInd =  15,
  id_CCTrCH_InformationItem_RL_RestoreInd =  16,
  id_CommonMeasurementAccuracy = 280,
  id_CommonMeasurementObjectType_CM_Rprt = 281,
  id_CommonMeasurementObjectType_CM_Rqst = 282,
  id_CommonMeasurementObjectType_CM_Rsp = 283,
  id_CommonMeasurementType = 284,
  id_CongestionCause =  18,
  id_SFN       = 285,
  id_SFNReportingIndicator = 286,
  id_InformationExchangeID = 287,
  id_InformationExchangeObjectType_InfEx_Rprt = 288,
  id_InformationExchangeObjectType_InfEx_Rqst = 289,
  id_InformationExchangeObjectType_InfEx_Rsp = 290,
  id_InformationReportCharacteristics = 291,
  id_InformationType = 292,
  id_neighbouring_LCR_TDD_CellInformation =  58,
  id_DL_Timeslot_ISCP_LCR_Information_RL_SetupRqstTDD =  65,
  id_RL_LCR_InformationResponse_RL_SetupRspTDD =  66,
  id_UL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD =  75,
  id_UL_DPCH_LCR_InformationItem_RL_SetupRspTDD =  76,
  id_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD =  77,
  id_DL_DPCH_LCR_InformationItem_RL_SetupRspTDD =  78,
  id_DSCH_LCR_InformationListIEs_RL_SetupRspTDD =  79,
  id_USCH_LCR_InformationListIEs_RL_SetupRspTDD =  80,
  id_DL_Timeslot_ISCP_LCR_Information_RL_AdditionRqstTDD =  81,
  id_RL_LCR_InformationResponse_RL_AdditionRspTDD =  86,
  id_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD =  87,
  id_UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD =  88,
  id_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD =  89,
  id_DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD =  94,
  id_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD =  96,
  id_USCH_LCR_InformationListIEs_RL_AdditionRspTDD =  97,
  id_UL_DPCH_LCR_InformationAddListIE_RL_ReconfReadyTDD =  98,
  id_Unknown_99 =  99,
  id_UL_Timeslot_LCR_InformationModifyList_RL_ReconfReadyTDD = 100,
  id_DL_DPCH_LCR_InformationAddListIE_RL_ReconfReadyTDD = 101,
  id_DL_Timeslot_LCR_InformationModifyList_RL_ReconfReadyTDD = 104,
  id_UL_Timeslot_LCR_InformationList_PhyChReconfRqstTDD = 105,
  id_DL_Timeslot_LCR_InformationList_PhyChReconfRqstTDD = 106,
  id_timeSlot_ISCP_LCR_List_DL_PC_Rqst_TDD = 138,
  id_TSTD_Support_Indicator_RL_SetupRqstTDD = 139,
  id_RestrictionStateIndicator = 142,
  id_Load_Value = 233,
  id_Load_Value_IncrDecrThres = 234,
  id_OnModification = 235,
  id_Received_Total_Wideband_Power_Value = 236,
  id_Received_Total_Wideband_Power_Value_IncrDecrThres = 237,
  id_SFNSFNMeasurementThresholdInformation = 238,
  id_Transmitted_Carrier_Power_Value = 239,
  id_Transmitted_Carrier_Power_Value_IncrDecrThres = 240,
  id_TUTRANGPSMeasurementThresholdInformation = 241,
  id_UL_Timeslot_ISCP_Value = 242,
  id_UL_Timeslot_ISCP_Value_IncrDecrThres = 243,
  id_Rx_Timing_Deviation_Value_LCR = 293,
  id_DPC_Mode_Change_SupportIndicator =  19,
  id_Unused_ProtocolIE_ID_247 = 247,
  id_Unknown_294 = 294,
  id_Unused_ProtocolIE_ID_295 = 295,
  id_PrimaryCCPCH_RSCP_RL_ReconfPrepTDD = 202,
  id_DL_TimeSlot_ISCP_Info_RL_ReconfPrepTDD = 203,
  id_DL_Timeslot_ISCP_LCR_Information_RL_ReconfPrepTDD = 204,
  id_DSCH_RNTI = 249,
  id_DL_PowerBalancing_Information = 296,
  id_DL_PowerBalancing_ActivationIndicator = 297,
  id_DL_PowerBalancing_UpdatedIndicator = 298,
  id_DL_ReferencePowerInformation = 299,
  id_Enhanced_PrimaryCPICH_EcNo = 224,
  id_IPDL_TDD_ParametersLCR = 252,
  id_CellCapabilityContainer_FDD = 300,
  id_CellCapabilityContainer_TDD = 301,
  id_CellCapabilityContainer_TDD_LCR = 302,
  id_RL_Specific_DCH_Info = 317,
  id_RL_ReconfigurationRequestFDD_RL_InformationList = 318,
  id_RL_ReconfigurationRequestFDD_RL_Information_IEs = 319,
  id_Unknown_320 = 320,
  id_RL_ReconfigurationRequestTDD_RL_Information = 321,
  id_Unknown_322 = 322,
  id_CommonTransportChannelResourcesInitialisationNotRequired = 250,
  id_DelayedActivation = 312,
  id_DelayedActivationList_RL_ActivationCmdFDD = 313,
  id_DelayedActivationInformation_RL_ActivationCmdFDD = 314,
  id_DelayedActivationList_RL_ActivationCmdTDD = 315,
  id_DelayedActivationInformation_RL_ActivationCmdTDD = 316,
  id_neighbouringTDDCellMeasurementInformationLCR = 251,
  id_UL_SIR_Target_CCTrCH_InformationItem_RL_SetupRspTDD = 150,
  id_UL_SIR_Target_CCTrCH_LCR_InformationItem_RL_SetupRspTDD = 151,
  id_PrimCCPCH_RSCP_DL_PC_RqstTDD = 451,
  id_HSDSCH_FDD_Information = 452,
  id_HSDSCH_FDD_Information_Response = 453,
  id_Unknown_454 = 454,
  id_Unknown_455 = 455,
  id_HSDSCH_FDD_Update_Information = 466,
  id_HSDSCH_Information_to_Modify = 456,
  id_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd = 516,
  id_HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd = 517,
  id_HSDSCH_RNTI = 457,
  id_HSDSCH_TDD_Information = 458,
  id_HSDSCH_TDD_Information_Response = 459,
  id_Unknown_460 = 460,
  id_Unknown_461 = 461,
  id_Unknown_462 = 462,
  id_HSDSCH_TDD_Update_Information = 467,
  id_HSPDSCH_RL_ID = 463,
  id_HSDSCH_MACdFlows_to_Add = 531,
  id_HSDSCH_MACdFlows_to_Delete = 532,
  id_Angle_Of_Arrival_Value_LCR = 148,
  id_Unknown_149 = 149,
  id_TrafficClass = 158,
  id_Unused_ProtocolIE_ID_248 = 248,
  id_Unused_ProtocolIE_ID_253 = 253,
  id_PDSCH_RL_ID = 323,
  id_TimeSlot_RL_SetupRspTDD = 325,
  id_Unknown_326 = 326,
  id_Unknown_327 = 327,
  id_Unknown_328 = 328,
  id_Unknown_329 = 329,
  id_Unknown_330 = 330,
  id_Unknown_331 = 331,
  id_Unknown_332 = 332,
  id_Unknown_333 = 333,
  id_Unknown_334 = 334,
  id_Unknown_335 = 335,
  id_Unknown_336 = 336,
  id_Unknown_337 = 337,
  id_Unknown_338 = 338,
  id_Unknown_339 = 339,
  id_Unknown_340 = 340,
  id_Unknown_341 = 341,
  id_Unknown_342 = 342,
  id_Unknown_343 = 343,
  id_Unknown_344 = 344,
  id_Unknown_345 = 345,
  id_Unknown_346 = 346,
  id_Unknown_347 = 347,
  id_Unknown_348 = 348,
  id_Unknown_349 = 349,
  id_Unknown_350 = 350,
  id_Unknown_351 = 351,
  id_Unknown_352 = 352,
  id_Unknown_353 = 353,
  id_Unknown_354 = 354,
  id_Unknown_355 = 355,
  id_Unknown_356 = 356,
  id_Unknown_357 = 357,
  id_Unknown_358 = 358,
  id_Unknown_359 = 359,
  id_Unknown_360 = 360,
  id_Unknown_361 = 361,
  id_Unknown_362 = 362,
  id_Unknown_363 = 363,
  id_Unknown_364 = 364,
  id_Unknown_365 = 365,
  id_Unknown_366 = 366,
  id_Unknown_367 = 367,
  id_Unknown_368 = 368,
  id_Unknown_369 = 369,
  id_Unknown_370 = 370,
  id_Unknown_371 = 371,
  id_Unknown_372 = 372,
  id_Unknown_373 = 373,
  id_Unknown_374 = 374,
  id_Unknown_375 = 375,
  id_Unknown_376 = 376,
  id_Unknown_377 = 377,
  id_Unknown_378 = 378,
  id_Unknown_379 = 379,
  id_Unknown_380 = 380,
  id_Unknown_381 = 381,
  id_Unknown_382 = 382,
  id_Unknown_383 = 383,
  id_Unknown_384 = 384,
  id_Unknown_385 = 385,
  id_Unknown_386 = 386,
  id_Unknown_387 = 387,
  id_Unknown_388 = 388,
  id_Unknown_389 = 389,
  id_Unknown_390 = 390,
  id_Unknown_391 = 391,
  id_Unknown_392 = 392,
  id_Unknown_393 = 393,
  id_Unknown_394 = 394,
  id_Unknown_395 = 395,
  id_Unknown_396 = 396,
  id_Unknown_397 = 397,
  id_Unknown_398 = 398,
  id_Unknown_399 = 399,
  id_Unknown_400 = 400,
  id_Unknown_401 = 401,
  id_Unknown_402 = 402,
  id_Unknown_403 = 403,
  id_Unknown_404 = 404,
  id_Unknown_405 = 405,
  id_Unknown_406 = 406,
  id_Unknown_407 = 407,
  id_Unknown_408 = 408,
  id_Unknown_409 = 409,
  id_Unknown_410 = 410,
  id_Unknown_411 = 411,
  id_Unknown_412 = 412,
  id_Unknown_413 = 413,
  id_Unknown_414 = 414,
  id_Unknown_415 = 415,
  id_Unknown_416 = 416,
  id_Unknown_417 = 417,
  id_Unknown_418 = 418,
  id_Unknown_419 = 419,
  id_Unknown_420 = 420,
  id_Unknown_421 = 421,
  id_Unknown_422 = 422,
  id_Unknown_423 = 423,
  id_Unknown_424 = 424,
  id_Unknown_425 = 425,
  id_Unknown_426 = 426,
  id_Unknown_427 = 427,
  id_Unknown_428 = 428,
  id_Unknown_429 = 429,
  id_Unknown_430 = 430,
  id_Unknown_431 = 431,
  id_Unknown_432 = 432,
  id_Unknown_433 = 433,
  id_Unknown_434 = 434,
  id_Unknown_435 = 435,
  id_Unknown_436 = 436,
  id_Unknown_437 = 437,
  id_Unknown_438 = 438,
  id_Unknown_439 = 439,
  id_Unknown_440 = 440,
  id_Unknown_441 = 441,
  id_Unknown_442 = 442,
  id_Unknown_443 = 443,
  id_Unknown_444 = 444,
  id_Unknown_445 = 445,
  id_Unknown_446 = 446,
  id_Unknown_447 = 447,
  id_Unknown_448 = 448,
  id_Unknown_449 = 449,
  id_Unknown_450 = 450,
  id_GERAN_Cell_Capability = 468,
  id_GERAN_Classmark = 469,
  id_DSCH_InitialWindowSize = 480,
  id_UL_Synchronisation_Parameters_LCR = 464,
  id_SNA_Information = 479,
  id_MAChs_ResetIndicator = 465,
  id_TDD_DL_DPCH_TimeSlotFormatModifyItem_LCR_RL_ReconfReadyTDD = 481,
  id_TDD_UL_DPCH_TimeSlotFormatModifyItem_LCR_RL_ReconfReadyTDD = 482,
  id_TDD_TPC_UplinkStepSize_LCR_RL_SetupRqstTDD = 483,
  id_UL_CCTrCH_InformationList_RL_AdditionRqstTDD = 484,
  id_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD = 485,
  id_DL_CCTrCH_InformationList_RL_AdditionRqstTDD = 486,
  id_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD = 487,
  id_TDD_TPC_UplinkStepSize_InformationAdd_LCR_RL_ReconfPrepTDD = 488,
  id_TDD_TPC_UplinkStepSize_InformationModify_LCR_RL_ReconfPrepTDD = 489,
  id_TDD_TPC_DownlinkStepSize_InformationAdd_RL_ReconfPrepTDD = 490,
  id_TDD_TPC_DownlinkStepSize_InformationModify_RL_ReconfPrepTDD = 491,
  id_UL_TimingAdvanceCtrl_LCR = 492,
  id_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD = 493,
  id_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD = 494,
  id_HS_SICH_Reception_Quality = 495,
  id_HS_SICH_Reception_Quality_Measurement_Value = 496,
  id_HSSICH_Info_DM_Rprt = 497,
  id_HSSICH_Info_DM_Rqst = 498,
  id_HSSICH_Info_DM = 499,
  id_CCTrCH_Maximum_DL_Power_RL_SetupRspTDD = 500,
  id_CCTrCH_Minimum_DL_Power_RL_SetupRspTDD = 501,
  id_CCTrCH_Maximum_DL_Power_RL_AdditionRspTDD = 502,
  id_CCTrCH_Minimum_DL_Power_RL_AdditionRspTDD = 503,
  id_CCTrCH_Maximum_DL_Power_RL_ReconfReadyTDD = 504,
  id_CCTrCH_Minimum_DL_Power_RL_ReconfReadyTDD = 505,
  id_Maximum_DL_Power_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD = 506,
  id_Minimum_DL_Power_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD = 507,
  id_DL_CCTrCH_InformationList_RL_ReconfRspTDD = 508,
  id_DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD = 509,
  id_Maximum_DL_Power_TimeslotLCR_InformationItem = 510,
  id_Minimum_DL_Power_TimeslotLCR_InformationItem = 511,
  id_TDD_Support_8PSK = 512,
  id_TDD_maxNrDLPhysicalchannels = 513,
  id_ExtendedGSMCellIndividualOffset = 514,
  id_RL_ParameterUpdateIndicationFDD_RL_InformationList = 518,
  id_Primary_CPICH_Usage_For_Channel_Estimation = 519,
  id_Secondary_CPICH_Information = 520,
  id_Secondary_CPICH_Information_Change = 521,
  id_Unused_ProtocolIE_ID_522 = 522,
  id_Unused_ProtocolIE_ID_523 = 523,
  id_RL_ParameterUpdateIndicationFDD_RL_Information_Item = 524,
  id_Phase_Reference_Update_Indicator = 525,
  id_Unidirectional_DCH_Indicator = 526,
  id_RL_Information_RL_ReconfPrepTDD = 527,
  id_Multiple_RL_InformationResponse_RL_ReconfReadyTDD = 528,
  id_RL_ReconfigurationResponseTDD_RL_Information = 529,
  id_Satellite_Almanac_Information_ExtItem = 530,
  id_HSDSCH_Information_to_Modify_Unsynchronised = 533,
  id_TnlQos    = 534,
  id_RTLoadValue = 535,
  id_NRTLoadInformationValue = 536,
  id_CellPortionID = 537,
  id_UpPTSInterferenceValue = 538,
  id_PrimaryCCPCH_RSCP_Delta = 539,
  id_UEMeasurementType = 540,
  id_UEMeasurementTimeslotInfoHCR = 541,
  id_UEMeasurementTimeslotInfoLCR = 542,
  id_UEMeasurementReportCharacteristics = 543,
  id_UEMeasurementParameterModAllow = 544,
  id_UEMeasurementValueInformation = 545,
  id_InterfacesToTraceItem = 546,
  id_ListOfInterfacesToTrace = 547,
  id_TraceDepth = 548,
  id_TraceRecordingSessionReference = 549,
  id_TraceReference = 550,
  id_UEIdentity = 551,
  id_NACC_Related_Data = 552,
  id_GSM_Cell_InfEx_Rqst = 553,
  id_MeasurementRecoveryBehavior = 554,
  id_MeasurementRecoveryReportingIndicator = 555,
  id_MeasurementRecoverySupportIndicator = 556,
  id_DL_DPCH_Power_Information_RL_ReconfPrepFDD = 557,
  id_F_DPCH_Information_RL_ReconfPrepFDD = 558,
  id_F_DPCH_Information_RL_SetupRqstFDD = 559,
  id_MBMS_Bearer_Service_List = 560,
  id_MBMS_Bearer_Service_List_InfEx_Rsp = 561,
  id_Active_MBMS_Bearer_ServiceFDD = 562,
  id_Active_MBMS_Bearer_ServiceTDD = 563,
  id_Old_URA_ID = 564,
  id_Unknown_565 = 565,
  id_Unknown_566 = 566,
  id_Unknown_567 = 567,
  id_UE_State  = 568,
  id_URA_ID    = 569,
  id_Unknown_570 = 570,
  id_HARQ_Preamble_Mode = 571,
  id_SynchronisationIndicator = 572,
  id_UL_DPDCHIndicatorEDCH = 573,
  id_EDPCH_Information = 574,
  id_RL_Specific_EDCH_Information = 575,
  id_EDCH_RL_Indication = 576,
  id_EDCH_FDD_Information = 577,
  id_EDCH_RLSet_Id = 578,
  id_Serving_EDCHRL_Id = 579,
  id_EDCH_FDD_DL_ControlChannelInformation = 580,
  id_EDCH_FDD_InformationResponse = 581,
  id_EDCH_MACdFlows_To_Add = 582,
  id_EDCH_FDD_Information_To_Modify = 583,
  id_EDCH_MACdFlows_To_Delete = 584,
  id_EDPCH_Information_RLReconfRequest_FDD = 585,
  id_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd = 586,
  id_EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd = 587,
  id_EDCH_MacdFlowSpecificInformationList_RL_CongestInd = 588,
  id_EDCH_MacdFlowSpecificInformationItem_RL_CongestInd = 589,
  id_MBMS_Bearer_Service_Full_Address = 590,
  id_Initial_DL_DPCH_TimingAdjustment = 591,
  id_Initial_DL_DPCH_TimingAdjustment_Allowed = 592,
  id_User_Plane_Congestion_Fields_Inclusion = 593,
  id_HARQ_Preamble_Mode_Activation_Indicator = 594,
  id_multiple_DedicatedMeasurementValueList_TDD_DM_Rsp = 595,
  id_multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp = 596,
  id_ProvidedInformation = 597,
  id_Active_MBMS_Bearer_ServiceFDD_PFL = 598,
  id_Active_MBMS_Bearer_ServiceTDD_PFL = 599,
  id_FrequencyBandIndicator = 600,
  id_Serving_cell_change_CFN = 601,
  id_HS_DSCH_serving_cell_change_information = 602,
  id_HS_DSCH_serving_cell_change_informationResponse = 603,
  id_E_DCH_Serving_cell_change_informationResponse = 604,
  id_secondary_LCR_CCPCH_Info_TDD = 605,
  id_E_DCH_FDD_Update_Information = 606,
  id_Inter_Frequency_Cell_List = 607,
  id_Inter_Frequency_Cell_Information = 608,
  id_multiple_HSSICHMeasurementValueList_TDD_DM_Rsp = 609,
  id_TDD_Support_PLCCH = 610,
  id_PLCCH_Information_UL_TimeslotLCR_Info = 611,
  id_PLCCH_Information_PhyChReconfRqstTDD = 612,
  id_TDD768_maxNrDLPhysicalchannelsTS = 613,
  id_RL_InformationResponse_RL_AdditionRspTDD768 = 614,
  id_UL_CCTrCH_InformationListIE_RL_AdditionRspTDD768 = 615,
  id_DL_CCTrCH_InformationListIE_RL_AdditionRspTDD768 = 616,
  id_UL_DPCH_InformationItem_RL_AdditionRspTDD768 = 617,
  id_DL_DPCH_InformationItem_RL_AdditionRspTDD768 = 618,
  id_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD768 = 619,
  id_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768 = 620,
  id_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD768 = 621,
  id_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768 = 622,
  id_secondary_CCPCH_Info_RL_ReconfReadyTDD768 = 623,
  id_hSSCCH_TDD_Specific_InfoList_Response768 = 624,
  id_hSPDSCH_TDD_Specific_InfoList_Response768 = 625,
  id_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768 = 626,
  id_UL_Timeslot_InformationList_PhyChReconfRqstTDD768 = 627,
  id_DL_Timeslot_InformationList_PhyChReconfRqstTDD768 = 628,
  id_CellCapabilityContainer_TDD768 = 629,
  id_multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp = 630,
  id_neighbouringTDDCellMeasurementInformation768 = 631,
  id_UEMeasurementTimeslotInfo768 = 632,
  id_Rx_Timing_Deviation_Value_768 = 633,
  id_UEMeasurementValueTransmittedPowerList768 = 634,
  id_UEMeasurementValueTimeslotISCPList768 = 635,
  id_RL_InformationResponse_RL_SetupRspTDD768 = 636,
  id_UL_CCTrCH_InformationListIE_RL_SetupRspTDD768 = 637,
  id_DL_CCTrCH_InformationListIE_RL_SetupRspTDD768 = 638,
  id_UL_DPCH_InformationItem_RL_SetupRspTDD768 = 639,
  id_DL_DPCH_InformationItem_RL_SetupRspTDD768 = 640,
  id_TDD768_minimumSpreadingFactor_UL = 641,
  id_TDD768_minimumSpreadingFactor_DL = 642,
  id_TDD768_maxNrDLPhysicalchannels = 643,
  id_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD = 644,
  id_DPCH_ID768_DM_Rsp = 645,
  id_DPCH_ID768_DM_Rqst = 646,
  id_DPCH_ID768_DM_Rprt = 647,
  id_EDPCH_Information_RLAdditionReq_FDD = 648,
  id_HSDSCH_Configured_Indicator = 649,
  id_RxTimingDeviationForTAext = 650,
  id_RxTimingDeviationForTA768 = 651,
  id_Rx_Timing_Deviation_Value_ext = 652,
  id_E_DCH_PowerOffset_for_SchedulingInfo = 653,
  id_TrCH_SrcStatisticsDescr = 654,
  id_E_DCH_Information = 655,
  id_E_DCH_Serving_RL_ID = 656,
  id_E_DCH_Information_Reconfig = 657,
  id_E_DCH_Information_Response = 658,
  id_E_DCH_768_Information = 659,
  id_E_DCH_768_Information_Reconfig = 660,
  id_E_DCH_768_Information_Response = 661,
  id_ExtendedPropagationDelay = 662,
  id_Extended_Round_Trip_Time_Value = 663,
  id_AlternativeFormatReportingIndicator = 664,
  id_DCH_Indicator_For_E_DCH_HSDPA_Operation = 665,
  id_E_RGCH_E_HICH_ChannelisationCodeValidityIndicator = 666,
  id_E_DCH_Minimum_Set_E_TFCIValidityIndicator = 667,
  id_Fast_Reconfiguration_Mode = 668,
  id_Fast_Reconfiguration_Permission = 669,
  id_Continuous_Packet_Connectivity_DTX_DRX_Information = 670,
  id_Continuous_Packet_Connectivity_HS_SCCH_Less_Information = 671,
  id_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response = 672,
  id_CPC_Information = 673,
  id_Unknown_674 = 674,
  id_MIMO_InformationResponse = 675,
  id_Unknown_676 = 676,
  id_E_DCH_LCR_Information = 677,
  id_E_DCH_LCR_Information_Reconfig = 678,
  id_E_DCH_LCR_Information_Response = 679,
  id_HS_PDSCH_Code_Change_Grant = 680,
  id_HS_PDSCH_Code_Change_Indicator = 681,
  id_Extended_SRNC_ID = 682,
  id_Extended_RNC_ID = 683,
  id_SixtyfourQAM_DL_SupportIndicator = 684,
  id_Enhanced_FACH_Support_Indicator = 685,
  id_Enhanced_FACH_Information_ResponseFDD = 686,
  id_Unknown_687 = 687,
  id_Unknown_688 = 688,
  id_Unknown_689 = 689,
  id_HSDSCH_MACdPDUSizeFormat = 690,
  id_MaximumMACdPDU_SizeExtended = 691,
  id_F_DPCH_SlotFormat = 692,
  id_F_DPCH_SlotFormatSupportRequest = 693,
  id_eDCH_MACdFlow_Retransmission_Timer_LCR = 694,
  id_Max_UE_DTX_Cycle = 695,
  id_Unknown_696 = 696,
  id_Unknown_697 = 697,
  id_Unknown_698 = 698,
  id_GANSS_Common_Data = 699,
  id_GANSS_Information = 700,
  id_GANSS_Generic_Data = 701,
  id_TUTRANGANSSMeasurementThresholdInformation = 702,
  id_TUTRANGANSSMeasurementValueInformation = 703,
  id_Unknown_704 = 704,
  id_Ext_Reference_E_TFCI_PO = 705,
  id_Ext_Max_Bits_MACe_PDU_non_scheduled = 706,
  id_HARQ_MemoryPartitioningInfoExtForMIMO = 707,
  id_MIMO_ActivationIndicator = 708,
  id_MIMO_Mode_Indicator = 709,
  id_MIMO_N_M_Ratio = 710,
  id_TransportBearerNotSetupIndicator = 711,
  id_TransportBearerNotRequestedIndicator = 712,
  id_PowerControlGAP = 713,
  id_UARFCNforNt = 714,
  id_LCRTDD_uplink_Physical_Channel_Capability = 715,
  id_number_Of_Supported_Carriers = 716,
  id_HSSICH_SIRTarget = 717,
  id_HSSICH_TPC_StepSize = 718,
  id_tSN_Length = 719,
  id_HS_SICH_ID_Extension = 720,
  id_HSSICH_Info_DM_Rqst_Extension = 721,
  id_multipleFreq_HSPDSCH_InformationList_ResponseTDDLCR = 722,
  id_multicarrier_number = 723,
  id_UPPCHPositionLCR = 724,
  id_UpPCH_InformationList_LCRTDD = 725,
  id_UpPCH_InformationItem_LCRTDD = 726,
  id_Multiple_PLMN_List = 727,
  id_UE_Capabilities_Info = 728,
  id_FrameOffset = 729,
  id_ChipOffset = 730,
  id_Enhanced_PCH_Capability = 731,
  id_SixteenQAM_UL_Operation_Indicator = 732,
  id_E_TFCI_Boost_Information = 733,
  id_SixtyfourQAM_UsageAllowedIndicator = 734,
  id_SixtyfourQAM_DL_UsageIndicator = 735,
  id_Default_Serving_Grant_in_DTX_Cycle2 = 736,
  id_E_DPDCH_PowerInterpolation = 737,
  id_Extended_E_DCH_LCRTDD_PhysicalLayerCategory = 738,
  id_E_DCH_MACdPDUSizeFormat = 739,
  id_Continuous_Packet_Connectivity_HS_SCCH_Less_Deactivate_Indicator = 740,
  id_E_DCH_DL_Control_Channel_Change_Information = 741,
  id_E_DCH_DL_Control_Channel_Grant_Information = 742,
  id_MaximumNumber_Of_Retransmission_For_SchedulingInfo_LCRTDD = 743,
  id_E_DCH_RetransmissionTimer_For_SchedulingInfo_LCRTDD = 744,
  id_E_PUCH_PowerControlGAP = 745,
  id_HSDSCH_TBSizeTableIndicator = 746,
  id_UE_with_enhanced_HS_SCCH_support_indicator = 747,
  id_DGANSS_Corrections_Req = 748,
  id_E_AGCH_Table_Choice = 749,
  id_RANAP_EnhancedRelocationInformationRequest = 750,
  id_RANAP_EnhancedRelocationInformationResponse = 751,
  id_Common_EDCH_MAC_d_Flow_Specific_InformationFDD = 752,
  id_Common_EDCH_Support_Indicator = 753,
  id_E_RNTI    = 754,
  id_Released_CN_Domain = 755,
  id_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst = 756,
  id_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp = 757,
  id_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt = 758,
  id_MBMS_Cell_InfEx_Rqst = 759,
  id_MBMS_Cell_InfEx_Rsp = 760,
  id_MBMS_Cell_InfEx_Rprt = 761,
  id_Counting_Information = 762,
  id_Transmission_Mode_Information = 763,
  id_MBMS_Neighbouring_Cell_Information = 764,
  id_MBMS_RLC_Sequence_Number_Information = 765,
  id_RLC_Sequence_Number = 766,
  id_Neighbouring_E_UTRA_CellInformation = 767,
  id_Unknown_768 = 768,
  id_MBSFN_Cluster_Identity = 769,
  id_MCCH_Configuration_Info = 770,
  id_MCCH_Message_List = 771,
  id_MBSFN_Scheduling_Transmission_Time_Interval_Info_List = 772,
  id_GANSS_Time_ID = 773,
  id_GANSS_AddIonoModelReq = 774,
  id_GANSS_EarthOrientParaReq = 775,
  id_GANSS_AddNavigationModelsReq = 776,
  id_GANSS_AddUTCModelsReq = 777,
  id_GANSS_AuxInfoReq = 778,
  id_GANSS_SBAS_ID = 779,
  id_GANSS_ID  = 780,
  id_GANSS_Additional_Ionospheric_Model = 781,
  id_GANSS_Earth_Orientation_Parameters = 782,
  id_GANSS_Additional_Time_Models = 783,
  id_GANSS_Additional_Navigation_Models = 784,
  id_GANSS_Additional_UTC_Models = 785,
  id_GANSS_Auxiliary_Information = 786,
  id_MinimumReducedE_DPDCH_GainFactor = 787,
  id_Enhanced_FACH_Information_ResponseLCR = 788,
  id_Common_EDCH_MAC_d_Flow_Specific_InformationLCR = 789,
  id_HSDSCH_PreconfigurationSetup = 790,
  id_HSDSCH_PreconfigurationInfo = 791,
  id_NoOfTargetCellHS_SCCH_Order = 792,
  id_EnhancedHSServingCC_Abort = 793,
  id_Additional_HS_Cell_Information_RL_Setup = 794,
  id_Additional_HS_Cell_Information_Response = 795,
  id_Additional_HS_Cell_Information_RL_Addition = 796,
  id_Additional_HS_Cell_Change_Information_Response = 797,
  id_Additional_HS_Cell_Information_RL_Reconf_Prep = 798,
  id_Additional_HS_Cell_Information_RL_Reconf_Req = 799,
  id_Additional_HS_Cell_RL_Reconf_Response = 800,
  id_Additional_HS_Cell_Information_RL_Param_Upd = 801,
  id_Secondary_Serving_Cell_List = 802,
  id_MultiCarrier_HSDSCH_Physical_Layer_Category = 803,
  id_IdleIntervalInformation = 804,
  id_NeedforIdleInterval = 805,
  id_IdleIntervalConfigurationIndicator = 806,
  id_ContinuousPacketConnectivity_DRX_InformationLCR = 807,
  id_ContinuousPacketConnectivity_DRX_Information_ResponseLCR = 808,
  id_E_AGCH_UE_Inactivity_Monitor_Threshold = 809,
  id_CPC_InformationLCR = 810,
  id_E_DCH_Semi_PersistentScheduling_Information_LCR = 811,
  id_HS_DSCH_Semi_PersistentScheduling_Information_LCR = 812,
  id_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR = 813,
  id_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR = 814,
  id_MIMO_SFMode_For_HSPDSCHDualStream = 815,
  id_MIMO_SFMode_Supported_For_HSPDSCHDualStream = 816,
  id_MIMO_ReferenceSignal_InformationListLCR = 817,
  id_GANSS_alm_keplerianNAVAlmanac = 818,
  id_GANSS_alm_keplerianReducedAlmanac = 819,
  id_GANSS_alm_keplerianMidiAlmanac = 820,
  id_GANSS_alm_keplerianGLONASS = 821,
  id_GANSS_alm_ecefSBASAlmanac = 822,
  id_DL_RLC_PDU_Size_Format = 823,
  id_MACes_Maximum_Bitrate_LCR = 824,
  id_Single_Stream_MIMO_ActivationIndicator = 825,
  id_Single_Stream_MIMO_Mode_Indicator = 826,
  id_Dual_Band_Secondary_Serving_Cell_List = 827,
  id_UE_AggregateMaximumBitRate = 828,
  id_power_offset_for_S_CPICH_for_MIMO = 829,
  id_power_offset_for_S_CPICH_for_MIMO_Request_Indicator = 830,
  id_UE_SupportIndicatorExtension = 831,
  id_Unknown_832 = 832,
  id_Unknown_833 = 833,
  id_Unknown_834 = 834,
  id_ActivationInformation = 835,
  id_CellPortionLCRID = 836,
  id_Additional_EDCH_Cell_Information_RL_Setup_Req = 837,
  id_Additional_EDCH_Cell_Information_Response = 838,
  id_Additional_EDCH_Cell_Information_RL_Add_Req = 839,
  id_Additional_EDCH_Cell_Information_Response_RLAdd = 840,
  id_Additional_EDCH_Cell_Information_RL_Reconf_Prep = 841,
  id_Additional_EDCH_Cell_Information_RL_Reconf_Req = 842,
  id_Additional_EDCH_Cell_Information_RL_Param_Upd = 843,
  id_Additional_EDCH_Preconfiguration_Information = 844,
  id_MulticellEDCH_Information = 845,
  id_Unknown_846 = 846,
  id_Unknown_847 = 847,
  id_Unknown_848 = 848,
  id_Unknown_849 = 849,
  id_Unknown_850 = 850,
  id_Unknown_851 = 851,
  id_Unknown_852 = 852,
  id_Unknown_853 = 853,
  id_Additional_EDCH_Cell_Information_ResponseRLReconf = 854,
  id_EDCH_Indicator = 855,
  id_DiversityMode = 856,
  id_TransmitDiversityIndicator = 857,
  id_NonCellSpecificTxDiversity = 858,
  id_CellCapabilityContainerExtension_FDD = 859,
  id_HSDSCH_Physical_Layer_Category = 860,
  id_E_RNTI_For_FACH = 861,
  id_H_RNTI_For_FACH = 862,
  id_RNTI_Allocation_Indicator = 863,
  id_UE_AggregateMaximumBitRate_Enforcement_Indicator = 864,
  id_DCH_MeasurementOccasion_Information = 865,
  id_DCH_MeasurementType_Indicator = 866,
  id_Out_of_Sychronization_Window = 867,
  id_MulticellEDCH_RL_SpecificInformation = 868,
  id_DGNSS_ValidityPeriod = 869,
  id_TS0_HS_PDSCH_Indication_LCR = 870,
  id_UE_TS0_CapabilityLCR = 871,
  id_Non_Serving_RL_Preconfig_Info = 872,
  id_Non_Serving_RL_Preconfig_Setup = 873,
  id_Non_Serving_RL_Preconfig_Removal = 874,
  id_Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup = 875,
  id_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList = 876,
  id_CellListValidityIndicator = 877,
  id_completeAlmanacProvided = 878,
  id_ganss_Delta_T = 879,
  id_OrdinalNumberOfFrequency = 880,
  id_Multicell_EDCH_Restriction = 882,
  id_CellCapabilityContainerExtension_TDD_LCR = 883,
  id_Multi_Carrier_EDCH_Setup = 884,
  id_Multi_Carrier_EDCH_Reconfigure = 885,
  id_Multi_Carrier_EDCH_Response = 886,
  id_SNPL_Carrier_Group_Indicator = 887,
  id_MU_MIMO_InformationLCR = 888,
  id_MU_MIMO_Indicator = 889,
  id_HS_SCCH_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext = 890,
  id_Usefulness_Of_Battery_Optimization = 891,
  id_Multi_Carrier_E_DCH_LCRTDD_PhysicalLayerCategory = 892,
  id_Measurement_Power_Offset = 893,
  id_MDT_Configuration = 894,
  id_Neighbouring_UMTS_CellInformation_Ext = 895,
  id_Neighbouring_UMTS_CellInformationExtensionItem = 896,
  id_Control_Type_InformationTransferControlReq = 897,
  id_UMTS_Cells_Info = 898,
  id_ANRReportIndication = 899,
  id_ANR_Cell_InfEx_Rqst = 900,
  id_ANR_Cell_InfEx_Rsp = 901,
  id_ANR_Cell_Information = 902,
  id_Trace_Collection_Entity_IP_Address = 904,
  id_Affected_HSDSCH_Serving_Cell_List = 905,
  id_UL_CLTD_Information = 906,
  id_UL_CLTD_Information_Reconf = 907,
  id_UL_CLTD_State_Update_Information = 908,
  id_Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order = 909,
  id_CPC_RecoveryReport = 910,
  id_FTPICH_Information = 911,
  id_FTPICH_Information_Reconf = 912,
  id_UE_RF_Band_CapabilityLCR = 913,
  id_Extended_S_RNTI = 914,
  id_ExtendedAffectedUEInformationForMBMS = 915,
  id_Extended_S_RNTI_Group = 916,
  id_FTPICH_Information_Response = 917,
  id_FTPICH_Reconfiguration_Information = 918,
  id_MIMO_withfourtransmitantennas_ActivationIndicator = 919,
  id_MIMO_withfourtransmitantennas_ModeIndicator = 920,
  id_power_offset_for_S_CPICH_for_MIMO_withfourtransmitantennas = 921,
  id_power_offset_for_S_CPICH_for_MIMO_withfourtransmitantennas_Request_Indicator = 922,
  id_DualStream_MIMO_withfourtransmitantennas_ActivationIndicator = 923,
  id_DualStream_MIMO_withfourtransmitantennas_ModeIndicator = 924,
  id_Additional_Associated_Secondary_CPICH = 925,
  id_UE_transmission_power_headroom = 926,
  id_Multiflow_Information = 927,
  id_Multiflow_Reconfiguration = 928,
  id_Multiflow_OrdinalNumberOfFrequency = 929,
  id_UL_MIMO_Information = 930,
  id_UL_MIMO_Reconfiguration = 931,
  id_UL_MIMO_DL_Control_Channel_Information = 932,
  id_SixtyfourQAM_UL_Operation_Indicator = 933,
  id_Common_E_RGCH_Cell_InfEx_Rqst = 935,
  id_Common_E_RGCH_Cell_InfEx_Rsp = 936,
  id_Common_E_RGCH_Cell_InfEx_Rprt = 937,
  id_Common_E_RGCH_Cell_Information = 938,
  id_PCI       = 939,
  id_TAC       = 940,
  id_PLMN_List = 941,
  id_PrecoderWeightSetRestriction = 942,
  id_EARFCN_FDD_Extended = 943,
  id_EARFCN_TDD_Extended = 944,
  id_Non_rectangular_resource_allocation_indicator = 945,
  id_Non_rectangular_resource_timeslot_set = 946,
  id_UE_Support_of_non_rectangular_resource_allocation = 947,
  id_DBDS_CorrectionsReq = 948,
  id_DBDS_Corrections = 949,
  id_BDS_IonosphericGridModelReq = 950,
  id_BDS_Ionospheric_Grid_Model = 951,
  id_GANSS_alm_keplerianBDSAlmanac = 952,
  id_Assisting_RepetitionFactors = 953,
  id_E_DCH_Decoupling_Indication = 954,
  id_UE_Measurement_Forwarding = 955,
  id_UPH_Filtering_Measurement_Forwarding_Request = 956,
  id_TTI_Update_Indicator = 957,
  id_CQI_Feedback_Cycle2_k = 958,
  id_CQI_Cycle_Switch_Timer = 959,
  id_UE_DRX_Cycle2 = 960,
  id_Inactivity_Threshold_for_UE_DRX_Cycle2 = 961,
  id_DTX_Information2 = 962,
  id_DCH_ENH_Information = 964,
  id_DCH_ENH_Information_Reconf = 965,
  id_Gainfactors_10ms_mode = 966,
  id_Radio_Links_without_DPCH_FDPCH_Indication = 967,
  id_UL_DPCCH2_Information = 968,
  id_UL_DPCCH2_Information_Reconf = 969,
  id_ImplicitGrantHandling = 970,
  id_MinimumTEBSthreshold = 971,
  id_MultipleFrequencyBandsIndicatorList_EUTRA = 972,
  id_MultipleFrequencyBandsIndicatorList = 973,
  id_ActivationDelay = 974,
  id_Fast_TTI_switching_Mode_synchronized = 975,
  id_Fast_TTI_switching_Mode_unsynchronized = 976,
  id_Fast_TTI_switching_Mode_Supported = 977,
  id_GANSS_BDSclockModel = 978,
  id_GANSS_NavModel_BDSKeplerianSet = 979,
  id_GANSS_UTCmodelSet4 = 980,
  id_Inter_Frequency_Cell_Information_SIB11ter = 981,
  id_Improved_Synchronized_Indicator = 982,
  id_Downlink_TPC_enhancements_Information = 983,
  id_Downlink_TPC_enhancements_Reconf = 984,
  id_TPC_slot_position = 985,
  id_DL_TBS    = 986,
  id_Selected_PLMN_Identity = 987,
  id_SpeechVersion = 2048,
  id_SourceID  = 2049,
  id_TargetID  = 2050,
  id_ClassmarkInformation2 = 2051,
  id_ClassmarkInformation3 = 2052,
  id_GSM_Cell_CM_Rqst = 2053,
  id_Extension_CommonMeasurementObjectType_CM_Rprt = 2054,
  id_Extension_CommonMeasurementObjectType_CM_Rqst = 2055,
  id_Extension_CommonMeasurementObjectType_CM_Rsp = 2056,
  id_Extension_FailureIndicationMeasurementList = 2057,
  id_Extension_FailureMeasurementList = 2058,
  id_Extension_TerminationMeasurementList = 2059,
  id_GsmCellList_CM_Rprt = 2060,
  id_GsmCellList_CM_Rqst = 2061,
  id_GsmCellList_CM_Rsp = 2062,
  id_LoadValue = 2063,
  id_EventH    = 2064
} ProtocolIE_ID_enum;

void proto_register_rnsap(void);
void proto_reg_handoff_rnsap(void);

typedef struct {
    guint32     ProcedureCode;
    guint32     ProtocolIE_ID;
    guint32     ddMode;
    const char *ProcedureID;
    const char *obj_id;
} rnsap_private_data_t;

static dissector_handle_t ranap_handle = NULL;
static dissector_handle_t rrc_dl_ccch_handle = NULL;
static dissector_handle_t rrc_ul_ccch_handle = NULL;

/* Initialize the protocol and registered fields */
static int proto_rnsap = -1;

static int hf_rnsap_transportLayerAddress_ipv4 = -1;
static int hf_rnsap_transportLayerAddress_ipv6 = -1;
static int hf_rnsap_transportLayerAddress_nsap = -1;
static int hf_rnsap_ActivationInformation_PDU = -1;  /* ActivationInformation */
static int hf_rnsap_Active_MBMS_Bearer_Service_ListFDD_PDU = -1;  /* Active_MBMS_Bearer_Service_ListFDD */
static int hf_rnsap_Active_MBMS_Bearer_Service_ListFDD_PFL_PDU = -1;  /* Active_MBMS_Bearer_Service_ListFDD_PFL */
static int hf_rnsap_Active_MBMS_Bearer_Service_ListTDD_PDU = -1;  /* Active_MBMS_Bearer_Service_ListTDD */
static int hf_rnsap_Active_MBMS_Bearer_Service_ListTDD_PFL_PDU = -1;  /* Active_MBMS_Bearer_Service_ListTDD_PFL */
static int hf_rnsap_Active_Pattern_Sequence_Information_PDU = -1;  /* Active_Pattern_Sequence_Information */
static int hf_rnsap_Additional_EDCH_Cell_Information_Response_RLAddList_PDU = -1;  /* Additional_EDCH_Cell_Information_Response_RLAddList */
static int hf_rnsap_Additional_EDCH_Setup_Info_PDU = -1;  /* Additional_EDCH_Setup_Info */
static int hf_rnsap_ANRReportIndication_PDU = -1;  /* ANRReportIndication */
static int hf_rnsap_DTX_Information2_PDU = -1;    /* DTX_Information2 */
static int hf_rnsap_Implicit_Grant_Handling_PDU = -1;  /* Implicit_Grant_Handling */
static int hf_rnsap_Minimum_TEBS_threshold_PDU = -1;  /* Minimum_TEBS_threshold */
static int hf_rnsap_Additional_EDCH_Cell_Information_Response_List_PDU = -1;  /* Additional_EDCH_Cell_Information_Response_List */
static int hf_rnsap_Additional_EDCH_Cell_Information_Response_RLReconf_List_PDU = -1;  /* Additional_EDCH_Cell_Information_Response_RLReconf_List */
static int hf_rnsap_AdjustmentPeriod_PDU = -1;    /* AdjustmentPeriod */
static int hf_rnsap_Allowed_Rate_Information_PDU = -1;  /* Allowed_Rate_Information */
static int hf_rnsap_AllowedQueuingTime_PDU = -1;  /* AllowedQueuingTime */
static int hf_rnsap_AlternativeFormatReportingIndicator_PDU = -1;  /* AlternativeFormatReportingIndicator */
static int hf_rnsap_Angle_Of_Arrival_Value_LCR_PDU = -1;  /* Angle_Of_Arrival_Value_LCR */
static int hf_rnsap_ANR_Cell_Information_PDU = -1;  /* ANR_Cell_Information */
static int hf_rnsap_AntennaColocationIndicator_PDU = -1;  /* AntennaColocationIndicator */
static int hf_rnsap_BDS_IonosphericGridModelReq_PDU = -1;  /* BDS_IonosphericGridModelReq */
static int hf_rnsap_BDS_Ionospheric_Grid_Model_PDU = -1;  /* BDS_Ionospheric_Grid_Model */
static int hf_rnsap_BindingID_PDU = -1;           /* BindingID */
static int hf_rnsap_Cause_PDU = -1;               /* Cause */
static int hf_rnsap_CellCapabilityContainer_FDD_PDU = -1;  /* CellCapabilityContainer_FDD */
static int hf_rnsap_CellCapabilityContainerExtension_FDD_PDU = -1;  /* CellCapabilityContainerExtension_FDD */
static int hf_rnsap_CellCapabilityContainer_TDD_PDU = -1;  /* CellCapabilityContainer_TDD */
static int hf_rnsap_CellCapabilityContainer_TDD_LCR_PDU = -1;  /* CellCapabilityContainer_TDD_LCR */
static int hf_rnsap_CellCapabilityContainerExtension_TDD_LCR_PDU = -1;  /* CellCapabilityContainerExtension_TDD_LCR */
static int hf_rnsap_CellCapabilityContainer_TDD768_PDU = -1;  /* CellCapabilityContainer_TDD768 */
static int hf_rnsap_C_ID_PDU = -1;                /* C_ID */
static int hf_rnsap_Cell_Capacity_Class_Value_PDU = -1;  /* Cell_Capacity_Class_Value */
static int hf_rnsap_CellListValidityIndicator_PDU = -1;  /* CellListValidityIndicator */
static int hf_rnsap_CellPortionID_PDU = -1;       /* CellPortionID */
static int hf_rnsap_CellPortionLCRID_PDU = -1;    /* CellPortionLCRID */
static int hf_rnsap_CFN_PDU = -1;                 /* CFN */
static int hf_rnsap_ChipOffset_PDU = -1;          /* ChipOffset */
static int hf_rnsap_ClassmarkInformation2_PDU = -1;  /* ClassmarkInformation2 */
static int hf_rnsap_ClassmarkInformation3_PDU = -1;  /* ClassmarkInformation3 */
static int hf_rnsap_ClosedLoopMode1_SupportIndicator_PDU = -1;  /* ClosedLoopMode1_SupportIndicator */
static int hf_rnsap_CommonMeasurementAccuracy_PDU = -1;  /* CommonMeasurementAccuracy */
static int hf_rnsap_CommonMeasurementType_PDU = -1;  /* CommonMeasurementType */
static int hf_rnsap_CongestionCause_PDU = -1;     /* CongestionCause */
static int hf_rnsap_CommonTransportChannelResourcesInitialisationNotRequired_PDU = -1;  /* CommonTransportChannelResourcesInitialisationNotRequired */
static int hf_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDD_PDU = -1;  /* Common_EDCH_MAC_d_Flow_Specific_InformationFDD */
static int hf_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationLCR_PDU = -1;  /* Common_EDCH_MAC_d_Flow_Specific_InformationLCR */
static int hf_rnsap_Common_EDCH_Support_Indicator_PDU = -1;  /* Common_EDCH_Support_Indicator */
static int hf_rnsap_Common_E_RGCH_Cell_Information_PDU = -1;  /* Common_E_RGCH_Cell_Information */
static int hf_rnsap_CompleteAlmanacProvided_PDU = -1;  /* CompleteAlmanacProvided */
static int hf_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information_PDU = -1;  /* Continuous_Packet_Connectivity_DTX_DRX_Information */
static int hf_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_PDU = -1;  /* Continuous_Packet_Connectivity_HS_SCCH_Less_Information */
static int hf_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response_PDU = -1;  /* Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response */
static int hf_rnsap_CoverageIndicator_PDU = -1;   /* CoverageIndicator */
static int hf_rnsap_CPC_Information_PDU = -1;     /* CPC_Information */
static int hf_rnsap_CPC_RecoveryReport_PDU = -1;  /* CPC_RecoveryReport */
static int hf_rnsap_Continuous_Packet_Connectivity_HS_SCCH_less_Deactivate_Indicator_PDU = -1;  /* Continuous_Packet_Connectivity_HS_SCCH_less_Deactivate_Indicator */
static int hf_rnsap_Counting_Information_PDU = -1;  /* Counting_Information */
static int hf_rnsap_CriticalityDiagnostics_PDU = -1;  /* CriticalityDiagnostics */
static int hf_rnsap_MessageStructure_PDU = -1;    /* MessageStructure */
static int hf_rnsap_CN_CS_DomainIdentifier_PDU = -1;  /* CN_CS_DomainIdentifier */
static int hf_rnsap_CN_PS_DomainIdentifier_PDU = -1;  /* CN_PS_DomainIdentifier */
static int hf_rnsap_CQI_Cycle_Switch_Timer_PDU = -1;  /* CQI_Cycle_Switch_Timer */
static int hf_rnsap_ControlGAP_PDU = -1;          /* ControlGAP */
static int hf_rnsap_CQI_Feedback_Cycle2_PDU = -1;  /* CQI_Feedback_Cycle2 */
static int hf_rnsap_C_RNTI_PDU = -1;              /* C_RNTI */
static int hf_rnsap_CPC_InformationLCR_PDU = -1;  /* CPC_InformationLCR */
static int hf_rnsap_ContinuousPacketConnectivity_DRX_InformationLCR_PDU = -1;  /* ContinuousPacketConnectivity_DRX_InformationLCR */
static int hf_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext_PDU = -1;  /* Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext */
static int hf_rnsap_E_AGCH_UE_Inactivity_Monitor_Threshold_PDU = -1;  /* E_AGCH_UE_Inactivity_Monitor_Threshold */
static int hf_rnsap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR_PDU = -1;  /* ContinuousPacketConnectivity_DRX_Information_ResponseLCR */
static int hf_rnsap_DBDS_CorrectionsReq_PDU = -1;  /* DBDS_CorrectionsReq */
static int hf_rnsap_DBDS_Corrections_PDU = -1;    /* DBDS_Corrections */
static int hf_rnsap_DCH_ENH_Information_PDU = -1;  /* DCH_ENH_Information */
static int hf_rnsap_DCH_ENH_Information_Reconf_PDU = -1;  /* DCH_ENH_Information_Reconf */
static int hf_rnsap_DCH_FDD_Information_PDU = -1;  /* DCH_FDD_Information */
static int hf_rnsap_DCH_MeasurementOccasion_Information_PDU = -1;  /* DCH_MeasurementOccasion_Information */
static int hf_rnsap_DCH_MeasurementType_Indicator_PDU = -1;  /* DCH_MeasurementType_Indicator */
static int hf_rnsap_DCH_Indicator_For_E_DCH_HSDPA_Operation_PDU = -1;  /* DCH_Indicator_For_E_DCH_HSDPA_Operation */
static int hf_rnsap_DCH_InformationResponse_PDU = -1;  /* DCH_InformationResponse */
static int hf_rnsap_DCH_TDD_Information_PDU = -1;  /* DCH_TDD_Information */
static int hf_rnsap_DedicatedMeasurementType_PDU = -1;  /* DedicatedMeasurementType */
static int hf_rnsap_DelayedActivation_PDU = -1;   /* DelayedActivation */
static int hf_rnsap_DGANSS_Corrections_Req_PDU = -1;  /* DGANSS_Corrections_Req */
static int hf_rnsap_DGNSS_ValidityPeriod_PDU = -1;  /* DGNSS_ValidityPeriod */
static int hf_rnsap_DiversityMode_PDU = -1;       /* DiversityMode */
static int hf_rnsap_DL_TBS_PDU = -1;              /* DL_TBS */
static int hf_rnsap_DL_DPCH_TimingAdjustment_PDU = -1;  /* DL_DPCH_TimingAdjustment */
static int hf_rnsap_DL_Power_PDU = -1;            /* DL_Power */
static int hf_rnsap_DL_PowerBalancing_Information_PDU = -1;  /* DL_PowerBalancing_Information */
static int hf_rnsap_DL_PowerBalancing_ActivationIndicator_PDU = -1;  /* DL_PowerBalancing_ActivationIndicator */
static int hf_rnsap_DL_PowerBalancing_UpdatedIndicator_PDU = -1;  /* DL_PowerBalancing_UpdatedIndicator */
static int hf_rnsap_DL_ReferencePowerInformation_PDU = -1;  /* DL_ReferencePowerInformation */
static int hf_rnsap_D_RNTI_PDU = -1;              /* D_RNTI */
static int hf_rnsap_D_RNTI_ReleaseIndication_PDU = -1;  /* D_RNTI_ReleaseIndication */
static int hf_rnsap_DL_TimeSlot_ISCP_Info_PDU = -1;  /* DL_TimeSlot_ISCP_Info */
static int hf_rnsap_DL_TimeSlot_ISCP_LCR_Information_PDU = -1;  /* DL_TimeSlot_ISCP_LCR_Information */
static int hf_rnsap_Downlink_TPC_enhancements_Information_PDU = -1;  /* Downlink_TPC_enhancements_Information */
static int hf_rnsap_Downlink_TPC_enhancements_Reconf_PDU = -1;  /* Downlink_TPC_enhancements_Reconf */
static int hf_rnsap_DPC_Mode_PDU = -1;            /* DPC_Mode */
static int hf_rnsap_DPC_Mode_Change_SupportIndicator_PDU = -1;  /* DPC_Mode_Change_SupportIndicator */
static int hf_rnsap_DPCH_ID768_PDU = -1;          /* DPCH_ID768 */
static int hf_rnsap_DRXCycleLengthCoefficient_PDU = -1;  /* DRXCycleLengthCoefficient */
static int hf_rnsap_DSCH_RNTI_PDU = -1;           /* DSCH_RNTI */
static int hf_rnsap_DSCH_InitialWindowSize_PDU = -1;  /* DSCH_InitialWindowSize */
static int hf_rnsap_DSCH_TDD_Information_PDU = -1;  /* DSCH_TDD_Information */
static int hf_rnsap_EARFCN_Extended_PDU = -1;     /* EARFCN_Extended */
static int hf_rnsap_EARFCN_FDD_Extended_PDU = -1;  /* EARFCN_FDD_Extended */
static int hf_rnsap_E_AGCH_Table_Choice_PDU = -1;  /* E_AGCH_Table_Choice */
static int hf_rnsap_E_DCH_Decoupling_Indication_PDU = -1;  /* E_DCH_Decoupling_Indication */
static int hf_rnsap_EDCH_FDD_DL_ControlChannelInformation_PDU = -1;  /* EDCH_FDD_DL_ControlChannelInformation */
static int hf_rnsap_E_RGCH_E_HICH_ChannelisationCodeValidityIndicator_PDU = -1;  /* E_RGCH_E_HICH_ChannelisationCodeValidityIndicator */
static int hf_rnsap_EDCH_FDD_Information_PDU = -1;  /* EDCH_FDD_Information */
static int hf_rnsap_UPH_Filtering_Measurement_Forwarding_Request_PDU = -1;  /* UPH_Filtering_Measurement_Forwarding_Request */
static int hf_rnsap_EDCH_FDD_InformationResponse_PDU = -1;  /* EDCH_FDD_InformationResponse */
static int hf_rnsap_EDCH_FDD_Information_To_Modify_PDU = -1;  /* EDCH_FDD_Information_To_Modify */
static int hf_rnsap_E_DCH_FDD_Update_Information_PDU = -1;  /* E_DCH_FDD_Update_Information */
static int hf_rnsap_E_DCH_DL_Control_Channel_Change_Information_PDU = -1;  /* E_DCH_DL_Control_Channel_Change_Information */
static int hf_rnsap_E_DCH_DL_Control_Channel_Grant_Information_PDU = -1;  /* E_DCH_DL_Control_Channel_Grant_Information */
static int hf_rnsap_E_DCH_PowerOffset_for_SchedulingInfo_PDU = -1;  /* E_DCH_PowerOffset_for_SchedulingInfo */
static int hf_rnsap_E_DCH_MACdPDUSizeFormat_PDU = -1;  /* E_DCH_MACdPDUSizeFormat */
static int hf_rnsap_EDCH_MACdFlows_Information_PDU = -1;  /* EDCH_MACdFlows_Information */
static int hf_rnsap_EDCH_MACdFlows_To_Delete_PDU = -1;  /* EDCH_MACdFlows_To_Delete */
static int hf_rnsap_EDCH_RL_Indication_PDU = -1;  /* EDCH_RL_Indication */
static int hf_rnsap_E_DCH_Serving_cell_change_informationResponse_PDU = -1;  /* E_DCH_Serving_cell_change_informationResponse */
static int hf_rnsap_EDPCH_Information_FDD_PDU = -1;  /* EDPCH_Information_FDD */
static int hf_rnsap_EDPCH_Information_RLReconfRequest_FDD_PDU = -1;  /* EDPCH_Information_RLReconfRequest_FDD */
static int hf_rnsap_E_DPDCH_PowerInterpolation_PDU = -1;  /* E_DPDCH_PowerInterpolation */
static int hf_rnsap_E_Serving_Grant_Value_PDU = -1;  /* E_Serving_Grant_Value */
static int hf_rnsap_EDCH_Serving_RL_PDU = -1;     /* EDCH_Serving_RL */
static int hf_rnsap_Enhanced_FACH_Information_ResponseFDD_PDU = -1;  /* Enhanced_FACH_Information_ResponseFDD */
static int hf_rnsap_Enhanced_FACH_Information_ResponseLCR_PDU = -1;  /* Enhanced_FACH_Information_ResponseLCR */
static int hf_rnsap_Enhanced_FACH_Support_Indicator_PDU = -1;  /* Enhanced_FACH_Support_Indicator */
static int hf_rnsap_EnhancedHSServingCC_Abort_PDU = -1;  /* EnhancedHSServingCC_Abort */
static int hf_rnsap_Enhanced_PCH_Capability_PDU = -1;  /* Enhanced_PCH_Capability */
static int hf_rnsap_E_RNTI_PDU = -1;              /* E_RNTI */
static int hf_rnsap_E_TFCI_Boost_Information_PDU = -1;  /* E_TFCI_Boost_Information */
static int hf_rnsap_E_DCH_Minimum_Set_E_TFCIValidityIndicator_PDU = -1;  /* E_DCH_Minimum_Set_E_TFCIValidityIndicator */
static int hf_rnsap_Enhanced_PrimaryCPICH_EcNo_PDU = -1;  /* Enhanced_PrimaryCPICH_EcNo */
static int hf_rnsap_EventH_PDU = -1;              /* EventH */
static int hf_rnsap_ExtendedGSMCellIndividualOffset_PDU = -1;  /* ExtendedGSMCellIndividualOffset */
static int hf_rnsap_E_DCH_Information_PDU = -1;   /* E_DCH_Information */
static int hf_rnsap_E_DCH_MACdFlow_Retransmission_Timer_LCR_PDU = -1;  /* E_DCH_MACdFlow_Retransmission_Timer_LCR */
static int hf_rnsap_E_DCH_Information_Reconfig_PDU = -1;  /* E_DCH_Information_Reconfig */
static int hf_rnsap_E_DCH_Information_Response_PDU = -1;  /* E_DCH_Information_Response */
static int hf_rnsap_ActivationDelay_PDU = -1;     /* ActivationDelay */
static int hf_rnsap_Fast_TTI_switching_Mode_synchronized_PDU = -1;  /* Fast_TTI_switching_Mode_synchronized */
static int hf_rnsap_Fast_TTI_switching_Mode_unsynchronized_PDU = -1;  /* Fast_TTI_switching_Mode_unsynchronized */
static int hf_rnsap_Fast_TTI_switching_Mode_Supported_PDU = -1;  /* Fast_TTI_switching_Mode_Supported */
static int hf_rnsap_E_DCH_768_Information_PDU = -1;  /* E_DCH_768_Information */
static int hf_rnsap_E_DCH_768_Information_Reconfig_PDU = -1;  /* E_DCH_768_Information_Reconfig */
static int hf_rnsap_E_DCH_768_Information_Response_PDU = -1;  /* E_DCH_768_Information_Response */
static int hf_rnsap_E_DCH_LCR_Information_PDU = -1;  /* E_DCH_LCR_Information */
static int hf_rnsap_Extended_E_DCH_LCRTDD_PhysicalLayerCategory_PDU = -1;  /* Extended_E_DCH_LCRTDD_PhysicalLayerCategory */
static int hf_rnsap_Multi_Carrier_E_DCH_LCRTDD_PhysicalLayerCategory_PDU = -1;  /* Multi_Carrier_E_DCH_LCRTDD_PhysicalLayerCategory */
static int hf_rnsap_E_DCH_LCR_Information_Reconfig_PDU = -1;  /* E_DCH_LCR_Information_Reconfig */
static int hf_rnsap_E_DCH_LCR_Information_Response_PDU = -1;  /* E_DCH_LCR_Information_Response */
static int hf_rnsap_Ext_Reference_E_TFCI_PO_PDU = -1;  /* Ext_Reference_E_TFCI_PO */
static int hf_rnsap_ExtendedPropagationDelay_PDU = -1;  /* ExtendedPropagationDelay */
static int hf_rnsap_Radio_Links_without_DPCH_FDPCH_Indication_PDU = -1;  /* Radio_Links_without_DPCH_FDPCH_Indication */
static int hf_rnsap_Extended_RNC_ID_PDU = -1;     /* Extended_RNC_ID */
static int hf_rnsap_Extended_RNTI_PDU = -1;       /* Extended_RNTI */
static int hf_rnsap_Extended_Round_Trip_Time_Value_PDU = -1;  /* Extended_Round_Trip_Time_Value */
static int hf_rnsap_Extended_S_RNTI_Group_PDU = -1;  /* Extended_S_RNTI_Group */
static int hf_rnsap_ExtendedAffectedUEInformationForMBMS_PDU = -1;  /* ExtendedAffectedUEInformationForMBMS */
static int hf_rnsap_Ext_Max_Bits_MACe_PDU_non_scheduled_PDU = -1;  /* Ext_Max_Bits_MACe_PDU_non_scheduled */
static int hf_rnsap_E_DCH_Semi_PersistentScheduling_Information_LCR_PDU = -1;  /* E_DCH_Semi_PersistentScheduling_Information_LCR */
static int hf_rnsap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU = -1;  /* E_DCH_Semi_PersistentScheduling_Information_ResponseLCR */
static int hf_rnsap_FACH_FlowControlInformation_PDU = -1;  /* FACH_FlowControlInformation */
static int hf_rnsap_Fast_Reconfiguration_Mode_PDU = -1;  /* Fast_Reconfiguration_Mode */
static int hf_rnsap_Fast_Reconfiguration_Permission_PDU = -1;  /* Fast_Reconfiguration_Permission */
static int hf_rnsap_FDD_DCHs_to_Modify_PDU = -1;  /* FDD_DCHs_to_Modify */
static int hf_rnsap_FDD_DL_CodeInformation_PDU = -1;  /* FDD_DL_CodeInformation */
static int hf_rnsap_F_DPCH_SlotFormat_PDU = -1;   /* F_DPCH_SlotFormat */
static int hf_rnsap_F_DPCH_SlotFormatSupportRequest_PDU = -1;  /* F_DPCH_SlotFormatSupportRequest */
static int hf_rnsap_FNReportingIndicator_PDU = -1;  /* FNReportingIndicator */
static int hf_rnsap_FTPICH_Information_PDU = -1;  /* FTPICH_Information */
static int hf_rnsap_FTPICH_Information_Reconf_PDU = -1;  /* FTPICH_Information_Reconf */
static int hf_rnsap_FTPICH_Reconfiguration_Information_PDU = -1;  /* FTPICH_Reconfiguration_Information */
static int hf_rnsap_FTPICH_Information_Response_PDU = -1;  /* FTPICH_Information_Response */
static int hf_rnsap_FrameOffset_PDU = -1;         /* FrameOffset */
static int hf_rnsap_FrequencyBandIndicator_PDU = -1;  /* FrequencyBandIndicator */
static int hf_rnsap_GA_Cell_PDU = -1;             /* GA_Cell */
static int hf_rnsap_GA_CellAdditionalShapes_PDU = -1;  /* GA_CellAdditionalShapes */
static int hf_rnsap_Gainfactors_10ms_mode_PDU = -1;  /* Gainfactors_10ms_mode */
static int hf_rnsap_GANSS_AddIonoModelReq_PDU = -1;  /* GANSS_AddIonoModelReq */
static int hf_rnsap_GANSS_AddNavigationModelsReq_PDU = -1;  /* GANSS_AddNavigationModelsReq */
static int hf_rnsap_GANSS_AddUTCModelsReq_PDU = -1;  /* GANSS_AddUTCModelsReq */
static int hf_rnsap_GANSS_Additional_Ionospheric_Model_PDU = -1;  /* GANSS_Additional_Ionospheric_Model */
static int hf_rnsap_GANSS_Additional_Navigation_Models_PDU = -1;  /* GANSS_Additional_Navigation_Models */
static int hf_rnsap_GANSS_Additional_Time_Models_PDU = -1;  /* GANSS_Additional_Time_Models */
static int hf_rnsap_GANSS_Additional_UTC_Models_PDU = -1;  /* GANSS_Additional_UTC_Models */
static int hf_rnsap_GANSS_alm_keplerianNAVAlmanac_PDU = -1;  /* GANSS_alm_keplerianNAVAlmanac */
static int hf_rnsap_GANSS_alm_KEPLERIANBDSALMANAC_PDU = -1;  /* GANSS_alm_KEPLERIANBDSALMANAC */
static int hf_rnsap_GANSS_alm_keplerianReducedAlmanac_PDU = -1;  /* GANSS_alm_keplerianReducedAlmanac */
static int hf_rnsap_GANSS_alm_keplerianMidiAlmanac_PDU = -1;  /* GANSS_alm_keplerianMidiAlmanac */
static int hf_rnsap_GANSS_alm_keplerianGLONASS_PDU = -1;  /* GANSS_alm_keplerianGLONASS */
static int hf_rnsap_GANSS_alm_ecefSBASAlmanac_PDU = -1;  /* GANSS_alm_ecefSBASAlmanac */
static int hf_rnsap_GANSS_Auxiliary_Information_PDU = -1;  /* GANSS_Auxiliary_Information */
static int hf_rnsap_GANSS_AuxInfoReq_PDU = -1;    /* GANSS_AuxInfoReq */
static int hf_rnsap_GANSS_BDSclockModel_PDU = -1;  /* GANSS_BDSclockModel */
static int hf_rnsap_GANSS_Common_Data_PDU = -1;   /* GANSS_Common_Data */
static int hf_rnsap_GANSS_Delta_T_PDU = -1;       /* GANSS_Delta_T */
static int hf_rnsap_GANSS_Earth_Orientation_Parameters_PDU = -1;  /* GANSS_Earth_Orientation_Parameters */
static int hf_rnsap_GANSS_EarthOrientParaReq_PDU = -1;  /* GANSS_EarthOrientParaReq */
static int hf_rnsap_GANSS_Generic_Data_PDU = -1;  /* GANSS_Generic_Data */
static int hf_rnsap_GANSS_ID_PDU = -1;            /* GANSS_ID */
static int hf_rnsap_GANSS_Information_PDU = -1;   /* GANSS_Information */
static int hf_rnsap_GANSS_NavModel_BDSKeplerianSet_PDU = -1;  /* GANSS_NavModel_BDSKeplerianSet */
static int hf_rnsap_GANSS_SBAS_ID_PDU = -1;       /* GANSS_SBAS_ID */
static int hf_rnsap_GANSS_Time_ID_PDU = -1;       /* GANSS_Time_ID */
static int hf_rnsap_GANSS_UTCmodelSet4_PDU = -1;  /* GANSS_UTCmodelSet4 */
static int hf_rnsap_GERAN_Cell_Capability_PDU = -1;  /* GERAN_Cell_Capability */
static int hf_rnsap_GERAN_Classmark_PDU = -1;     /* GERAN_Classmark */
static int hf_rnsap_Satellite_Almanac_Information_ExtItem_PDU = -1;  /* Satellite_Almanac_Information_ExtItem */
static int hf_rnsap_Guaranteed_Rate_Information_PDU = -1;  /* Guaranteed_Rate_Information */
static int hf_rnsap_HARQ_MemoryPartitioningInfoExtForMIMO_PDU = -1;  /* HARQ_MemoryPartitioningInfoExtForMIMO */
static int hf_rnsap_HARQ_Preamble_Mode_PDU = -1;  /* HARQ_Preamble_Mode */
static int hf_rnsap_HARQ_Preamble_Mode_Activation_Indicator_PDU = -1;  /* HARQ_Preamble_Mode_Activation_Indicator */
static int hf_rnsap_HCS_Prio_PDU = -1;            /* HCS_Prio */
static int hf_rnsap_HSDSCH_Configured_Indicator_PDU = -1;  /* HSDSCH_Configured_Indicator */
static int hf_rnsap_HSDSCH_FDD_Information_PDU = -1;  /* HSDSCH_FDD_Information */
static int hf_rnsap_HSDSCH_FDD_Information_Response_PDU = -1;  /* HSDSCH_FDD_Information_Response */
static int hf_rnsap_HSDSCH_Information_to_Modify_PDU = -1;  /* HSDSCH_Information_to_Modify */
static int hf_rnsap_HSDSCH_Information_to_Modify_Unsynchronised_PDU = -1;  /* HSDSCH_Information_to_Modify_Unsynchronised */
static int hf_rnsap_HSDSCH_MACdPDUSizeFormat_PDU = -1;  /* HSDSCH_MACdPDUSizeFormat */
static int hf_rnsap_HSDSCH_MACdFlows_Information_PDU = -1;  /* HSDSCH_MACdFlows_Information */
static int hf_rnsap_HSDSCH_MACdFlows_to_Delete_PDU = -1;  /* HSDSCH_MACdFlows_to_Delete */
static int hf_rnsap_HSDSCH_PreconfigurationInfo_PDU = -1;  /* HSDSCH_PreconfigurationInfo */
static int hf_rnsap_Additional_EDCH_Preconfiguration_Information_PDU = -1;  /* Additional_EDCH_Preconfiguration_Information */
static int hf_rnsap_HSDSCH_PreconfigurationSetup_PDU = -1;  /* HSDSCH_PreconfigurationSetup */
static int hf_rnsap_HSDSCH_RNTI_PDU = -1;         /* HSDSCH_RNTI */
static int hf_rnsap_HS_DSCH_serving_cell_change_information_PDU = -1;  /* HS_DSCH_serving_cell_change_information */
static int hf_rnsap_HS_DSCH_serving_cell_change_informationResponse_PDU = -1;  /* HS_DSCH_serving_cell_change_informationResponse */
static int hf_rnsap_HSDSCH_TBSizeTableIndicator_PDU = -1;  /* HSDSCH_TBSizeTableIndicator */
static int hf_rnsap_HSDSCH_TDD_Information_PDU = -1;  /* HSDSCH_TDD_Information */
static int hf_rnsap_HSDSCH_TDD_Information_Response_PDU = -1;  /* HSDSCH_TDD_Information_Response */
static int hf_rnsap_Multicarrier_Number_PDU = -1;  /* Multicarrier_Number */
static int hf_rnsap_HSPDSCH_TDD_Specific_InfoList_Response768_PDU = -1;  /* HSPDSCH_TDD_Specific_InfoList_Response768 */
static int hf_rnsap_HSSCCH_TDD_Specific_InfoList_Response768_PDU = -1;  /* HSSCCH_TDD_Specific_InfoList_Response768 */
static int hf_rnsap_HS_SICH_Reception_Quality_Value_PDU = -1;  /* HS_SICH_Reception_Quality_Value */
static int hf_rnsap_HS_SICH_Reception_Quality_Measurement_Value_PDU = -1;  /* HS_SICH_Reception_Quality_Measurement_Value */
static int hf_rnsap_HS_SICH_ID_PDU = -1;          /* HS_SICH_ID */
static int hf_rnsap_HS_SICH_ID_Extension_PDU = -1;  /* HS_SICH_ID_Extension */
static int hf_rnsap_HS_PDSCH_Code_Change_Indicator_PDU = -1;  /* HS_PDSCH_Code_Change_Indicator */
static int hf_rnsap_HS_PDSCH_Code_Change_Grant_PDU = -1;  /* HS_PDSCH_Code_Change_Grant */
static int hf_rnsap_HSDSCH_FDD_Update_Information_PDU = -1;  /* HSDSCH_FDD_Update_Information */
static int hf_rnsap_HSDSCH_TDD_Update_Information_PDU = -1;  /* HSDSCH_TDD_Update_Information */
static int hf_rnsap_MIMO_ReferenceSignal_InformationListLCR_PDU = -1;  /* MIMO_ReferenceSignal_InformationListLCR */
static int hf_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_LCR_PDU = -1;  /* HS_DSCH_Semi_PersistentScheduling_Information_LCR */
static int hf_rnsap_HSDSCH_Physical_Layer_Category_PDU = -1;  /* HSDSCH_Physical_Layer_Category */
static int hf_rnsap_TS0_HS_PDSCH_Indication_LCR_PDU = -1;  /* TS0_HS_PDSCH_Indication_LCR */
static int hf_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU = -1;  /* HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR */
static int hf_rnsap_IMSI_PDU = -1;                /* IMSI */
static int hf_rnsap_Improved_Synchronized_Indicator_PDU = -1;  /* Improved_Synchronized_Indicator */
static int hf_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_PDU = -1;  /* Inactivity_Threshold_for_UE_DRX_Cycle */
static int hf_rnsap_InformationExchangeID_PDU = -1;  /* InformationExchangeID */
static int hf_rnsap_InformationReportCharacteristics_PDU = -1;  /* InformationReportCharacteristics */
static int hf_rnsap_InformationType_PDU = -1;     /* InformationType */
static int hf_rnsap_Initial_DL_DPCH_TimingAdjustment_Allowed_PDU = -1;  /* Initial_DL_DPCH_TimingAdjustment_Allowed */
static int hf_rnsap_InnerLoopDLPCStatus_PDU = -1;  /* InnerLoopDLPCStatus */
static int hf_rnsap_Inter_Frequency_Cell_List_PDU = -1;  /* Inter_Frequency_Cell_List */
static int hf_rnsap_Inter_Frequency_Cell_Information_PDU = -1;  /* Inter_Frequency_Cell_Information */
static int hf_rnsap_Inter_Frequency_Cell_Information_SIB11ter_PDU = -1;  /* Inter_Frequency_Cell_Information_SIB11ter */
static int hf_rnsap_IPDL_TDD_ParametersLCR_PDU = -1;  /* IPDL_TDD_ParametersLCR */
static int hf_rnsap_IdleIntervalInformation_PDU = -1;  /* IdleIntervalInformation */
static int hf_rnsap_L3_Information_PDU = -1;      /* L3_Information */
static int hf_rnsap_Load_Value_IncrDecrThres_PDU = -1;  /* Load_Value_IncrDecrThres */
static int hf_rnsap_Load_Value_PDU = -1;          /* Load_Value */
static int hf_rnsap_LoadValue_PDU = -1;           /* LoadValue */
static int hf_rnsap_LCRTDD_Uplink_Physical_Channel_Capability_PDU = -1;  /* LCRTDD_Uplink_Physical_Channel_Capability */
static int hf_rnsap_MDT_Configuration_PDU = -1;   /* MDT_Configuration */
static int hf_rnsap_MAC_PDU_SizeExtended_PDU = -1;  /* MAC_PDU_SizeExtended */
static int hf_rnsap_MACes_Maximum_Bitrate_LCR_PDU = -1;  /* MACes_Maximum_Bitrate_LCR */
static int hf_rnsap_MaxNrDLPhysicalchannels768_PDU = -1;  /* MaxNrDLPhysicalchannels768 */
static int hf_rnsap_MaxNrDLPhysicalchannelsTS_PDU = -1;  /* MaxNrDLPhysicalchannelsTS */
static int hf_rnsap_MaxNrDLPhysicalchannelsTS768_PDU = -1;  /* MaxNrDLPhysicalchannelsTS768 */
static int hf_rnsap_MaxNr_Retransmissions_EDCH_PDU = -1;  /* MaxNr_Retransmissions_EDCH */
static int hf_rnsap_Max_UE_DTX_Cycle_PDU = -1;    /* Max_UE_DTX_Cycle */
static int hf_rnsap_MBMS_Bearer_Service_Full_Address_PDU = -1;  /* MBMS_Bearer_Service_Full_Address */
static int hf_rnsap_MBMS_Bearer_Service_List_PDU = -1;  /* MBMS_Bearer_Service_List */
static int hf_rnsap_MBMS_Neighbouring_Cell_Information_PDU = -1;  /* MBMS_Neighbouring_Cell_Information */
static int hf_rnsap_MBMS_RLC_Sequence_Number_Information_PDU = -1;  /* MBMS_RLC_Sequence_Number_Information */
static int hf_rnsap_MBSFN_Cluster_Identity_PDU = -1;  /* MBSFN_Cluster_Identity */
static int hf_rnsap_MCCH_Message_List_PDU = -1;   /* MCCH_Message_List */
static int hf_rnsap_MCCH_Configuration_Info_PDU = -1;  /* MCCH_Configuration_Info */
static int hf_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Info_List_PDU = -1;  /* MBSFN_Scheduling_Transmission_Time_Interval_Info_List */
static int hf_rnsap_MeasurementFilterCoefficient_PDU = -1;  /* MeasurementFilterCoefficient */
static int hf_rnsap_MeasurementID_PDU = -1;       /* MeasurementID */
static int hf_rnsap_Measurement_Power_Offset_PDU = -1;  /* Measurement_Power_Offset */
static int hf_rnsap_MinimumSpreadingFactor768_PDU = -1;  /* MinimumSpreadingFactor768 */
static int hf_rnsap_MaxAdjustmentStep_PDU = -1;   /* MaxAdjustmentStep */
static int hf_rnsap_MeasurementRecoveryBehavior_PDU = -1;  /* MeasurementRecoveryBehavior */
static int hf_rnsap_MeasurementRecoveryReportingIndicator_PDU = -1;  /* MeasurementRecoveryReportingIndicator */
static int hf_rnsap_MeasurementRecoverySupportIndicator_PDU = -1;  /* MeasurementRecoverySupportIndicator */
static int hf_rnsap_MIMO_ActivationIndicator_PDU = -1;  /* MIMO_ActivationIndicator */
static int hf_rnsap_MIMO_InformationResponse_PDU = -1;  /* MIMO_InformationResponse */
static int hf_rnsap_MIMO_Mode_Indicator_PDU = -1;  /* MIMO_Mode_Indicator */
static int hf_rnsap_MIMO_N_M_Ratio_PDU = -1;      /* MIMO_N_M_Ratio */
static int hf_rnsap_MIMO_withfourtransmitantennas_ActivationIndicator_PDU = -1;  /* MIMO_withfourtransmitantennas_ActivationIndicator */
static int hf_rnsap_MIMO_withfourtransmitantennas_ModeIndicator_PDU = -1;  /* MIMO_withfourtransmitantennas_ModeIndicator */
static int hf_rnsap_DualStream_MIMO_withfourtransmitantennas_ActivationIndicator_PDU = -1;  /* DualStream_MIMO_withfourtransmitantennas_ActivationIndicator */
static int hf_rnsap_DualStream_MIMO_withfourtransmitantennas_ModeIndicator_PDU = -1;  /* DualStream_MIMO_withfourtransmitantennas_ModeIndicator */
static int hf_rnsap_Additional_Associated_Secondary_CPICH_PDU = -1;  /* Additional_Associated_Secondary_CPICH */
static int hf_rnsap_MinimumReducedE_DPDCH_GainFactor_PDU = -1;  /* MinimumReducedE_DPDCH_GainFactor */
static int hf_rnsap_MulticellEDCH_InformationItemIEs_PDU = -1;  /* MulticellEDCH_InformationItemIEs */
static int hf_rnsap_MulticellEDCH_RL_SpecificInformationItemIEs_PDU = -1;  /* MulticellEDCH_RL_SpecificInformationItemIEs */
static int hf_rnsap_Multicell_EDCH_Restriction_PDU = -1;  /* Multicell_EDCH_Restriction */
static int hf_rnsap_Multiple_PLMN_List_PDU = -1;  /* Multiple_PLMN_List */
static int hf_rnsap_MAChs_ResetIndicator_PDU = -1;  /* MAChs_ResetIndicator */
static int hf_rnsap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_PDU = -1;  /* MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR */
static int hf_rnsap_MIMO_SFMode_For_HSPDSCHDualStream_PDU = -1;  /* MIMO_SFMode_For_HSPDSCHDualStream */
static int hf_rnsap_Multi_Carrier_EDCH_Info_PDU = -1;  /* Multi_Carrier_EDCH_Info */
static int hf_rnsap_SNPL_Carrier_Group_Indicator_PDU = -1;  /* SNPL_Carrier_Group_Indicator */
static int hf_rnsap_Multi_Carrier_EDCH_Reconfigure_PDU = -1;  /* Multi_Carrier_EDCH_Reconfigure */
static int hf_rnsap_Multi_Carrier_EDCH_Information_Response_PDU = -1;  /* Multi_Carrier_EDCH_Information_Response */
static int hf_rnsap_Multiflow_Reconfiguration_PDU = -1;  /* Multiflow_Reconfiguration */
static int hf_rnsap_Multiflow_Information_PDU = -1;  /* Multiflow_Information */
static int hf_rnsap_Multiflow_RepetitionFactors_PDU = -1;  /* Multiflow_RepetitionFactors */
static int hf_rnsap_Multiflow_OrdinalNumberOfFrequency_PDU = -1;  /* Multiflow_OrdinalNumberOfFrequency */
static int hf_rnsap_MU_MIMO_InformationLCR_PDU = -1;  /* MU_MIMO_InformationLCR */
static int hf_rnsap_MU_MIMO_Indicator_PDU = -1;   /* MU_MIMO_Indicator */
static int hf_rnsap_NACC_Related_Data_PDU = -1;   /* NACC_Related_Data */
static int hf_rnsap_Neighbouring_UMTS_CellInformation_Ext_PDU = -1;  /* Neighbouring_UMTS_CellInformation_Ext */
static int hf_rnsap_Neighbouring_UMTS_CellInformationExtensionItem_PDU = -1;  /* Neighbouring_UMTS_CellInformationExtensionItem */
static int hf_rnsap_Neighbouring_UMTS_CellInformationItem_PDU = -1;  /* Neighbouring_UMTS_CellInformationItem */
static int hf_rnsap_MultipleFrequencyBandsIndicatorList_PDU = -1;  /* MultipleFrequencyBandsIndicatorList */
static int hf_rnsap_Neighbouring_GSM_CellInformationIEs_PDU = -1;  /* Neighbouring_GSM_CellInformationIEs */
static int hf_rnsap_NeighbouringTDDCellMeasurementInformationLCR_PDU = -1;  /* NeighbouringTDDCellMeasurementInformationLCR */
static int hf_rnsap_NeighbouringTDDCellMeasurementInformation768_PDU = -1;  /* NeighbouringTDDCellMeasurementInformation768 */
static int hf_rnsap_Neighbouring_LCR_TDD_CellInformation_PDU = -1;  /* Neighbouring_LCR_TDD_CellInformation */
static int hf_rnsap_Neighbouring_E_UTRA_CellInformation_PDU = -1;  /* Neighbouring_E_UTRA_CellInformation */
static int hf_rnsap_MultipleFrequencyBandsIndicatorList_EUTRA_PDU = -1;  /* MultipleFrequencyBandsIndicatorList_EUTRA */
static int hf_rnsap_NonCellSpecificTxDiversity_PDU = -1;  /* NonCellSpecificTxDiversity */
static int hf_rnsap_Non_rectangular_resource_allocation_indicator_PDU = -1;  /* Non_rectangular_resource_allocation_indicator */
static int hf_rnsap_Non_rectangular_resource_timeslot_set_PDU = -1;  /* Non_rectangular_resource_timeslot_set */
static int hf_rnsap_NRT_Load_Information_Value_IncrDecrThres_PDU = -1;  /* NRT_Load_Information_Value_IncrDecrThres */
static int hf_rnsap_NRT_Load_Information_Value_PDU = -1;  /* NRT_Load_Information_Value */
static int hf_rnsap_NRTLoadInformationValue_PDU = -1;  /* NRTLoadInformationValue */
static int hf_rnsap_Number_Of_Supported_Carriers_PDU = -1;  /* Number_Of_Supported_Carriers */
static int hf_rnsap_NoOfTargetCellHS_SCCH_Order_PDU = -1;  /* NoOfTargetCellHS_SCCH_Order */
static int hf_rnsap_Non_Serving_RL_Preconfig_Setup_PDU = -1;  /* Non_Serving_RL_Preconfig_Setup */
static int hf_rnsap_Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup_PDU = -1;  /* Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup */
static int hf_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_PDU = -1;  /* Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList */
static int hf_rnsap_NeedforIdleInterval_PDU = -1;  /* NeedforIdleInterval */
static int hf_rnsap_OnModification_PDU = -1;      /* OnModification */
static int hf_rnsap_OrdinalNumberOfFrequency_PDU = -1;  /* OrdinalNumberOfFrequency */
static int hf_rnsap_Out_of_Sychronization_Window_PDU = -1;  /* Out_of_Sychronization_Window */
static int hf_rnsap_PartialReportingIndicator_PDU = -1;  /* PartialReportingIndicator */
static int hf_rnsap_PCI_PDU = -1;                 /* PCI */
static int hf_rnsap_Permanent_NAS_UE_Identity_PDU = -1;  /* Permanent_NAS_UE_Identity */
static int hf_rnsap_Phase_Reference_Update_Indicator_PDU = -1;  /* Phase_Reference_Update_Indicator */
static int hf_rnsap_PLMN_Identity_PDU = -1;       /* PLMN_Identity */
static int hf_rnsap_PLMN_List_PDU = -1;           /* PLMN_List */
static int hf_rnsap_PowerAdjustmentType_PDU = -1;  /* PowerAdjustmentType */
static int hf_rnsap_PowerOffsetForSecondaryCPICHforMIMO_PDU = -1;  /* PowerOffsetForSecondaryCPICHforMIMO */
static int hf_rnsap_PowerOffsetForSecondaryCPICHforMIMORequestIndicator_PDU = -1;  /* PowerOffsetForSecondaryCPICHforMIMORequestIndicator */
static int hf_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas_PDU = -1;  /* PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas */
static int hf_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennasRequestIndicator_PDU = -1;  /* PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennasRequestIndicator */
static int hf_rnsap_Precoder_Weight_Set_Restriction_PDU = -1;  /* Precoder_Weight_Set_Restriction */
static int hf_rnsap_Primary_CPICH_Usage_For_Channel_Estimation_PDU = -1;  /* Primary_CPICH_Usage_For_Channel_Estimation */
static int hf_rnsap_PrimaryCCPCH_RSCP_PDU = -1;   /* PrimaryCCPCH_RSCP */
static int hf_rnsap_PrimaryCCPCH_RSCP_Delta_PDU = -1;  /* PrimaryCCPCH_RSCP_Delta */
static int hf_rnsap_PropagationDelay_PDU = -1;    /* PropagationDelay */
static int hf_rnsap_ProvidedInformation_PDU = -1;  /* ProvidedInformation */
static int hf_rnsap_UpPCH_InformationList_LCRTDD_PDU = -1;  /* UpPCH_InformationList_LCRTDD */
static int hf_rnsap_UpPCH_InformationItem_LCRTDD_PDU = -1;  /* UpPCH_InformationItem_LCRTDD */
static int hf_rnsap_RANAP_EnhancedRelocationInformationRequest_PDU = -1;  /* RANAP_EnhancedRelocationInformationRequest */
static int hf_rnsap_RANAP_EnhancedRelocationInformationResponse_PDU = -1;  /* RANAP_EnhancedRelocationInformationResponse */
static int hf_rnsap_RANAP_RelocationInformation_PDU = -1;  /* RANAP_RelocationInformation */
static int hf_rnsap_Received_Total_Wideband_Power_Value_PDU = -1;  /* Received_Total_Wideband_Power_Value */
static int hf_rnsap_Received_Total_Wideband_Power_Value_IncrDecrThres_PDU = -1;  /* Received_Total_Wideband_Power_Value_IncrDecrThres */
static int hf_rnsap_Released_CN_Domain_PDU = -1;  /* Released_CN_Domain */
static int hf_rnsap_ReportCharacteristics_PDU = -1;  /* ReportCharacteristics */
static int hf_rnsap_RestrictionStateIndicator_PDU = -1;  /* RestrictionStateIndicator */
static int hf_rnsap_RL_ID_PDU = -1;               /* RL_ID */
static int hf_rnsap_RL_Set_ID_PDU = -1;           /* RL_Set_ID */
static int hf_rnsap_RL_Specific_DCH_Info_PDU = -1;  /* RL_Specific_DCH_Info */
static int hf_rnsap_RL_Specific_EDCH_Information_PDU = -1;  /* RL_Specific_EDCH_Information */
static int hf_rnsap_DL_RLC_PDU_Size_Format_PDU = -1;  /* DL_RLC_PDU_Size_Format */
static int hf_rnsap_RLC_Sequence_Number_PDU = -1;  /* RLC_Sequence_Number */
static int hf_rnsap_RNC_ID_PDU = -1;              /* RNC_ID */
static int hf_rnsap_RNTI_Allocation_Indicator_PDU = -1;  /* RNTI_Allocation_Indicator */
static int hf_rnsap_RT_Load_Value_IncrDecrThres_PDU = -1;  /* RT_Load_Value_IncrDecrThres */
static int hf_rnsap_RT_Load_Value_PDU = -1;       /* RT_Load_Value */
static int hf_rnsap_RTLoadValue_PDU = -1;         /* RTLoadValue */
static int hf_rnsap_RxTimingDeviationForTA_PDU = -1;  /* RxTimingDeviationForTA */
static int hf_rnsap_RxTimingDeviationForTAext_PDU = -1;  /* RxTimingDeviationForTAext */
static int hf_rnsap_RxTimingDeviationForTA768_PDU = -1;  /* RxTimingDeviationForTA768 */
static int hf_rnsap_Rx_Timing_Deviation_Value_ext_PDU = -1;  /* Rx_Timing_Deviation_Value_ext */
static int hf_rnsap_Rx_Timing_Deviation_Value_LCR_PDU = -1;  /* Rx_Timing_Deviation_Value_LCR */
static int hf_rnsap_Rx_Timing_Deviation_Value_768_PDU = -1;  /* Rx_Timing_Deviation_Value_768 */
static int hf_rnsap_SAI_PDU = -1;                 /* SAI */
static int hf_rnsap_ScaledAdjustmentRatio_PDU = -1;  /* ScaledAdjustmentRatio */
static int hf_rnsap_Secondary_CPICH_Information_PDU = -1;  /* Secondary_CPICH_Information */
static int hf_rnsap_Secondary_CPICH_Information_Change_PDU = -1;  /* Secondary_CPICH_Information_Change */
static int hf_rnsap_Secondary_LCR_CCPCH_Info_TDD_PDU = -1;  /* Secondary_LCR_CCPCH_Info_TDD */
static int hf_rnsap_Secondary_CCPCH_Info_TDD768_PDU = -1;  /* Secondary_CCPCH_Info_TDD768 */
static int hf_rnsap_Secondary_Serving_Cell_List_PDU = -1;  /* Secondary_Serving_Cell_List */
static int hf_rnsap_SFN_PDU = -1;                 /* SFN */
static int hf_rnsap_SFNSFNMeasurementThresholdInformation_PDU = -1;  /* SFNSFNMeasurementThresholdInformation */
static int hf_rnsap_Single_Stream_MIMO_ActivationIndicator_PDU = -1;  /* Single_Stream_MIMO_ActivationIndicator */
static int hf_rnsap_Single_Stream_MIMO_Mode_Indicator_PDU = -1;  /* Single_Stream_MIMO_Mode_Indicator */
static int hf_rnsap_SixteenQAM_UL_Operation_Indicator_PDU = -1;  /* SixteenQAM_UL_Operation_Indicator */
static int hf_rnsap_SixtyfourQAM_UL_Operation_Indicator_PDU = -1;  /* SixtyfourQAM_UL_Operation_Indicator */
static int hf_rnsap_SixtyfourQAM_UsageAllowedIndicator_PDU = -1;  /* SixtyfourQAM_UsageAllowedIndicator */
static int hf_rnsap_SixtyfourQAM_DL_SupportIndicator_PDU = -1;  /* SixtyfourQAM_DL_SupportIndicator */
static int hf_rnsap_SixtyfourQAM_DL_UsageIndicator_PDU = -1;  /* SixtyfourQAM_DL_UsageIndicator */
static int hf_rnsap_SNA_Information_PDU = -1;     /* SNA_Information */
static int hf_rnsap_SourceID_PDU = -1;            /* SourceID */
static int hf_rnsap_SpeechVersion_PDU = -1;       /* SpeechVersion */
static int hf_rnsap_S_RNTI_PDU = -1;              /* S_RNTI */
static int hf_rnsap_STTD_SupportIndicator_PDU = -1;  /* STTD_SupportIndicator */
static int hf_rnsap_Support_8PSK_PDU = -1;        /* Support_8PSK */
static int hf_rnsap_Support_PLCCH_PDU = -1;       /* Support_PLCCH */
static int hf_rnsap_Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order_PDU = -1;  /* Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order */
static int hf_rnsap_TAC_PDU = -1;                 /* TAC */
static int hf_rnsap_TargetID_PDU = -1;            /* TargetID */
static int hf_rnsap_TDD_DCHs_to_Modify_PDU = -1;  /* TDD_DCHs_to_Modify */
static int hf_rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR_PDU = -1;  /* TDD_DL_DPCH_TimeSlotFormat_LCR */
static int hf_rnsap_TDD_TPC_DownlinkStepSize_PDU = -1;  /* TDD_TPC_DownlinkStepSize */
static int hf_rnsap_TDD_TPC_UplinkStepSize_LCR_PDU = -1;  /* TDD_TPC_UplinkStepSize_LCR */
static int hf_rnsap_TDD_UL_DPCH_TimeSlotFormat_LCR_PDU = -1;  /* TDD_UL_DPCH_TimeSlotFormat_LCR */
static int hf_rnsap_TimeSlot_PDU = -1;            /* TimeSlot */
static int hf_rnsap_SynchronisationIndicator_PDU = -1;  /* SynchronisationIndicator */
static int hf_rnsap_TnlQos_PDU = -1;              /* TnlQos */
static int hf_rnsap_TraceDepth_PDU = -1;          /* TraceDepth */
static int hf_rnsap_TraceRecordingSessionReference_PDU = -1;  /* TraceRecordingSessionReference */
static int hf_rnsap_TraceReference_PDU = -1;      /* TraceReference */
static int hf_rnsap_TrafficClass_PDU = -1;        /* TrafficClass */
static int hf_rnsap_Transmission_Gap_Pattern_Sequence_Information_PDU = -1;  /* Transmission_Gap_Pattern_Sequence_Information */
static int hf_rnsap_Affected_HSDSCH_Serving_Cell_List_PDU = -1;  /* Affected_HSDSCH_Serving_Cell_List */
static int hf_rnsap_Transmission_Mode_Information_PDU = -1;  /* Transmission_Mode_Information */
static int hf_rnsap_TransmitDiversityIndicator_PDU = -1;  /* TransmitDiversityIndicator */
static int hf_rnsap_Transmitted_Carrier_Power_Value_PDU = -1;  /* Transmitted_Carrier_Power_Value */
static int hf_rnsap_Transmitted_Carrier_Power_Value_IncrDecrThres_PDU = -1;  /* Transmitted_Carrier_Power_Value_IncrDecrThres */
static int hf_rnsap_TUTRANGANSSMeasurementThresholdInformation_PDU = -1;  /* TUTRANGANSSMeasurementThresholdInformation */
static int hf_rnsap_TUTRANGANSSMeasurementValueInformation_PDU = -1;  /* TUTRANGANSSMeasurementValueInformation */
static int hf_rnsap_TUTRANGPSMeasurementThresholdInformation_PDU = -1;  /* TUTRANGPSMeasurementThresholdInformation */
static int hf_rnsap_TransportBearerID_PDU = -1;   /* TransportBearerID */
static int hf_rnsap_TransportBearerRequestIndicator_PDU = -1;  /* TransportBearerRequestIndicator */
static int hf_rnsap_TransportBearerNotRequestedIndicator_PDU = -1;  /* TransportBearerNotRequestedIndicator */
static int hf_rnsap_TransportBearerNotSetupIndicator_PDU = -1;  /* TransportBearerNotSetupIndicator */
static int hf_rnsap_TPC_slot_position_PDU = -1;   /* TPC_slot_position */
static int hf_rnsap_TransportLayerAddress_PDU = -1;  /* TransportLayerAddress */
static int hf_rnsap_TrCH_SrcStatisticsDescr_PDU = -1;  /* TrCH_SrcStatisticsDescr */
static int hf_rnsap_TSN_Length_PDU = -1;          /* TSN_Length */
static int hf_rnsap_TSTD_Support_Indicator_PDU = -1;  /* TSTD_Support_Indicator */
static int hf_rnsap_TTI_Update_Indicator_PDU = -1;  /* TTI_Update_Indicator */
static int hf_rnsap_TypeOfError_PDU = -1;         /* TypeOfError */
static int hf_rnsap_UARFCN_PDU = -1;              /* UARFCN */
static int hf_rnsap_UE_AggregateMaximumBitRate_PDU = -1;  /* UE_AggregateMaximumBitRate */
static int hf_rnsap_UE_AggregateMaximumBitRate_Enforcement_Indicator_PDU = -1;  /* UE_AggregateMaximumBitRate_Enforcement_Indicator */
static int hf_rnsap_UE_Capabilities_Info_PDU = -1;  /* UE_Capabilities_Info */
static int hf_rnsap_UE_Support_of_non_rectangular_resource_allocation_PDU = -1;  /* UE_Support_of_non_rectangular_resource_allocation */
static int hf_rnsap_UE_TS0_CapabilityLCR_PDU = -1;  /* UE_TS0_CapabilityLCR */
static int hf_rnsap_UE_RF_Band_CapabilityLCR_PDU = -1;  /* UE_RF_Band_CapabilityLCR */
static int hf_rnsap_LCRTDD_HSDSCH_Physical_Layer_Category_PDU = -1;  /* LCRTDD_HSDSCH_Physical_Layer_Category */
static int hf_rnsap_UE_DRX_Cycle_PDU = -1;        /* UE_DRX_Cycle */
static int hf_rnsap_UEIdentity_PDU = -1;          /* UEIdentity */
static int hf_rnsap_UEMeasurementParameterModAllow_PDU = -1;  /* UEMeasurementParameterModAllow */
static int hf_rnsap_UEMeasurementReportCharacteristics_PDU = -1;  /* UEMeasurementReportCharacteristics */
static int hf_rnsap_UEMeasurementTimeslotInfoHCR_PDU = -1;  /* UEMeasurementTimeslotInfoHCR */
static int hf_rnsap_UEMeasurementTimeslotInfoLCR_PDU = -1;  /* UEMeasurementTimeslotInfoLCR */
static int hf_rnsap_UEMeasurementTimeslotInfo768_PDU = -1;  /* UEMeasurementTimeslotInfo768 */
static int hf_rnsap_UEMeasurementType_PDU = -1;   /* UEMeasurementType */
static int hf_rnsap_UEMeasurementValueTransmittedPowerList768_PDU = -1;  /* UEMeasurementValueTransmittedPowerList768 */
static int hf_rnsap_UEMeasurementValueTimeslotISCPList768_PDU = -1;  /* UEMeasurementValueTimeslotISCPList768 */
static int hf_rnsap_UEMeasurementValueInformation_PDU = -1;  /* UEMeasurementValueInformation */
static int hf_rnsap_UE_SupportIndicatorExtension_PDU = -1;  /* UE_SupportIndicatorExtension */
static int hf_rnsap_UE_State_PDU = -1;            /* UE_State */
static int hf_rnsap_UE_transmission_power_headroom_Value_PDU = -1;  /* UE_transmission_power_headroom_Value */
static int hf_rnsap_UL_DPDCHIndicatorEDCH_PDU = -1;  /* UL_DPDCHIndicatorEDCH */
static int hf_rnsap_PLCCHinformation_PDU = -1;    /* PLCCHinformation */
static int hf_rnsap_UL_Timeslot_ISCP_Value_PDU = -1;  /* UL_Timeslot_ISCP_Value */
static int hf_rnsap_UL_Timeslot_ISCP_Value_IncrDecrThres_PDU = -1;  /* UL_Timeslot_ISCP_Value_IncrDecrThres */
static int hf_rnsap_UL_TimingAdvanceCtrl_LCR_PDU = -1;  /* UL_TimingAdvanceCtrl_LCR */
static int hf_rnsap_UL_SIR_PDU = -1;              /* UL_SIR */
static int hf_rnsap_UC_ID_PDU = -1;               /* UC_ID */
static int hf_rnsap_UL_Synchronisation_Parameters_LCR_PDU = -1;  /* UL_Synchronisation_Parameters_LCR */
static int hf_rnsap_UPPCHPositionLCR_PDU = -1;    /* UPPCHPositionLCR */
static int hf_rnsap_UpPTSInterferenceValue_PDU = -1;  /* UpPTSInterferenceValue */
static int hf_rnsap_Unidirectional_DCH_Indicator_PDU = -1;  /* Unidirectional_DCH_Indicator */
static int hf_rnsap_URA_ID_PDU = -1;              /* URA_ID */
static int hf_rnsap_URA_Information_PDU = -1;     /* URA_Information */
static int hf_rnsap_USCH_Information_PDU = -1;    /* USCH_Information */
static int hf_rnsap_Usefulness_Of_Battery_Optimization_PDU = -1;  /* Usefulness_Of_Battery_Optimization */
static int hf_rnsap_User_Plane_Congestion_Fields_Inclusion_PDU = -1;  /* User_Plane_Congestion_Fields_Inclusion */
static int hf_rnsap_UMTS_Cell_Info_PDU = -1;      /* UMTS_Cell_Info */
static int hf_rnsap_UL_CLTD_Information_PDU = -1;  /* UL_CLTD_Information */
static int hf_rnsap_UL_CLTD_Information_Reconf_PDU = -1;  /* UL_CLTD_Information_Reconf */
static int hf_rnsap_UL_CLTD_State_Update_Information_PDU = -1;  /* UL_CLTD_State_Update_Information */
static int hf_rnsap_UL_DPCCH2_Information_PDU = -1;  /* UL_DPCCH2_Information */
static int hf_rnsap_UL_DPCCH2_Information_Reconf_PDU = -1;  /* UL_DPCCH2_Information_Reconf */
static int hf_rnsap_UL_MIMO_Information_PDU = -1;  /* UL_MIMO_Information */
static int hf_rnsap_UL_MIMO_Reconfiguration_PDU = -1;  /* UL_MIMO_Reconfiguration */
static int hf_rnsap_UL_MIMO_DL_Control_Channel_Information_PDU = -1;  /* UL_MIMO_DL_Control_Channel_Information */
static int hf_rnsap_RadioLinkSetupRequestFDD_PDU = -1;  /* RadioLinkSetupRequestFDD */
static int hf_rnsap_UL_DPCH_Information_RL_SetupRqstFDD_PDU = -1;  /* UL_DPCH_Information_RL_SetupRqstFDD */
static int hf_rnsap_DL_DPCH_Information_RL_SetupRqstFDD_PDU = -1;  /* DL_DPCH_Information_RL_SetupRqstFDD */
static int hf_rnsap_RL_InformationList_RL_SetupRqstFDD_PDU = -1;  /* RL_InformationList_RL_SetupRqstFDD */
static int hf_rnsap_RL_InformationItem_RL_SetupRqstFDD_PDU = -1;  /* RL_InformationItem_RL_SetupRqstFDD */
static int hf_rnsap_Additional_HS_Cell_Information_RL_Setup_List_PDU = -1;  /* Additional_HS_Cell_Information_RL_Setup_List */
static int hf_rnsap_F_DPCH_Information_RL_SetupRqstFDD_PDU = -1;  /* F_DPCH_Information_RL_SetupRqstFDD */
static int hf_rnsap_RadioLinkSetupRequestTDD_PDU = -1;  /* RadioLinkSetupRequestTDD */
static int hf_rnsap_UL_Physical_Channel_Information_RL_SetupRqstTDD_PDU = -1;  /* UL_Physical_Channel_Information_RL_SetupRqstTDD */
static int hf_rnsap_DL_Physical_Channel_Information_RL_SetupRqstTDD_PDU = -1;  /* DL_Physical_Channel_Information_RL_SetupRqstTDD */
static int hf_rnsap_UL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationList_RL_SetupRqstTDD */
static int hf_rnsap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationItem_RL_SetupRqstTDD */
static int hf_rnsap_DL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationList_RL_SetupRqstTDD */
static int hf_rnsap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationItem_RL_SetupRqstTDD */
static int hf_rnsap_RL_Information_RL_SetupRqstTDD_PDU = -1;  /* RL_Information_RL_SetupRqstTDD */
static int hf_rnsap_RadioLinkSetupResponseFDD_PDU = -1;  /* RadioLinkSetupResponseFDD */
static int hf_rnsap_RL_InformationResponseList_RL_SetupRspFDD_PDU = -1;  /* RL_InformationResponseList_RL_SetupRspFDD */
static int hf_rnsap_RL_InformationResponseItem_RL_SetupRspFDD_PDU = -1;  /* RL_InformationResponseItem_RL_SetupRspFDD */
static int hf_rnsap_Additional_HS_Cell_Information_Response_List_PDU = -1;  /* Additional_HS_Cell_Information_Response_List */
static int hf_rnsap_RadioLinkSetupResponseTDD_PDU = -1;  /* RadioLinkSetupResponseTDD */
static int hf_rnsap_RL_InformationResponse_RL_SetupRspTDD_PDU = -1;  /* RL_InformationResponse_RL_SetupRspTDD */
static int hf_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD_PDU = -1;  /* UL_CCTrCHInformationListIE_RL_SetupRspTDD */
static int hf_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD_PDU = -1;  /* UL_DPCH_InformationItem_RL_SetupRspTDD */
static int hf_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD_PDU = -1;  /* DL_CCTrCHInformationListIE_RL_SetupRspTDD */
static int hf_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD_PDU = -1;  /* DL_DPCH_InformationItem_RL_SetupRspTDD */
static int hf_rnsap_DSCH_InformationListIEs_RL_SetupRspTDD_PDU = -1;  /* DSCH_InformationListIEs_RL_SetupRspTDD */
static int hf_rnsap_USCH_InformationListIEs_RL_SetupRspTDD_PDU = -1;  /* USCH_InformationListIEs_RL_SetupRspTDD */
static int hf_rnsap_RL_LCR_InformationResponse_RL_SetupRspTDD_PDU = -1;  /* RL_LCR_InformationResponse_RL_SetupRspTDD */
static int hf_rnsap_UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD_PDU = -1;  /* UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD */
static int hf_rnsap_UL_DPCH_LCR_InformationItem_RL_SetupRspTDD_PDU = -1;  /* UL_DPCH_LCR_InformationItem_RL_SetupRspTDD */
static int hf_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD_PDU = -1;  /* DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD */
static int hf_rnsap_DL_DPCH_LCR_InformationItem_RL_SetupRspTDD_PDU = -1;  /* DL_DPCH_LCR_InformationItem_RL_SetupRspTDD */
static int hf_rnsap_DSCH_LCR_InformationListIEs_RL_SetupRspTDD_PDU = -1;  /* DSCH_LCR_InformationListIEs_RL_SetupRspTDD */
static int hf_rnsap_USCH_LCR_InformationListIEs_RL_SetupRspTDD_PDU = -1;  /* USCH_LCR_InformationListIEs_RL_SetupRspTDD */
static int hf_rnsap_RL_InformationResponse_RL_SetupRspTDD768_PDU = -1;  /* RL_InformationResponse_RL_SetupRspTDD768 */
static int hf_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD768_PDU = -1;  /* UL_CCTrCHInformationListIE_RL_SetupRspTDD768 */
static int hf_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD768_PDU = -1;  /* UL_DPCH_InformationItem_RL_SetupRspTDD768 */
static int hf_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD768_PDU = -1;  /* DL_CCTrCHInformationListIE_RL_SetupRspTDD768 */
static int hf_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD768_PDU = -1;  /* DL_DPCH_InformationItem_RL_SetupRspTDD768 */
static int hf_rnsap_RadioLinkSetupFailureFDD_PDU = -1;  /* RadioLinkSetupFailureFDD */
static int hf_rnsap_CauseLevel_RL_SetupFailureFDD_PDU = -1;  /* CauseLevel_RL_SetupFailureFDD */
static int hf_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD_PDU = -1;  /* UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD */
static int hf_rnsap_SuccessfulRL_InformationResponse_RL_SetupFailureFDD_PDU = -1;  /* SuccessfulRL_InformationResponse_RL_SetupFailureFDD */
static int hf_rnsap_RadioLinkSetupFailureTDD_PDU = -1;  /* RadioLinkSetupFailureTDD */
static int hf_rnsap_CauseLevel_RL_SetupFailureTDD_PDU = -1;  /* CauseLevel_RL_SetupFailureTDD */
static int hf_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD_PDU = -1;  /* UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD */
static int hf_rnsap_RadioLinkAdditionRequestFDD_PDU = -1;  /* RadioLinkAdditionRequestFDD */
static int hf_rnsap_RL_InformationList_RL_AdditionRqstFDD_PDU = -1;  /* RL_InformationList_RL_AdditionRqstFDD */
static int hf_rnsap_RL_Information_RL_AdditionRqstFDD_PDU = -1;  /* RL_Information_RL_AdditionRqstFDD */
static int hf_rnsap_Additional_HS_Cell_Information_RL_Addition_List_PDU = -1;  /* Additional_HS_Cell_Information_RL_Addition_List */
static int hf_rnsap_Additional_EDCH_Cell_Information_RL_Add_Req_PDU = -1;  /* Additional_EDCH_Cell_Information_RL_Add_Req */
static int hf_rnsap_EDPCH_Information_RLAdditionReq_FDD_PDU = -1;  /* EDPCH_Information_RLAdditionReq_FDD */
static int hf_rnsap_RadioLinkAdditionRequestTDD_PDU = -1;  /* RadioLinkAdditionRequestTDD */
static int hf_rnsap_RL_Information_RL_AdditionRqstTDD_PDU = -1;  /* RL_Information_RL_AdditionRqstTDD */
static int hf_rnsap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationList_RL_AdditionRqstTDD */
static int hf_rnsap_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationItem_RL_AdditionRqstTDD */
static int hf_rnsap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationList_RL_AdditionRqstTDD */
static int hf_rnsap_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationItem_RL_AdditionRqstTDD */
static int hf_rnsap_RadioLinkAdditionResponseFDD_PDU = -1;  /* RadioLinkAdditionResponseFDD */
static int hf_rnsap_RL_InformationResponseList_RL_AdditionRspFDD_PDU = -1;  /* RL_InformationResponseList_RL_AdditionRspFDD */
static int hf_rnsap_RL_InformationResponseItem_RL_AdditionRspFDD_PDU = -1;  /* RL_InformationResponseItem_RL_AdditionRspFDD */
static int hf_rnsap_Additional_HS_Cell_Change_Information_Response_List_PDU = -1;  /* Additional_HS_Cell_Change_Information_Response_List */
static int hf_rnsap_RadioLinkAdditionResponseTDD_PDU = -1;  /* RadioLinkAdditionResponseTDD */
static int hf_rnsap_RL_InformationResponse_RL_AdditionRspTDD_PDU = -1;  /* RL_InformationResponse_RL_AdditionRspTDD */
static int hf_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD_PDU = -1;  /* UL_CCTrCHInformationListIE_RL_AdditionRspTDD */
static int hf_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD_PDU = -1;  /* UL_DPCH_InformationItem_RL_AdditionRspTDD */
static int hf_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD_PDU = -1;  /* DL_CCTrCHInformationListIE_RL_AdditionRspTDD */
static int hf_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD_PDU = -1;  /* DL_DPCH_InformationItem_RL_AdditionRspTDD */
static int hf_rnsap_DSCH_InformationListIE_RL_AdditionRspTDD_PDU = -1;  /* DSCH_InformationListIE_RL_AdditionRspTDD */
static int hf_rnsap_USCH_InformationListIE_RL_AdditionRspTDD_PDU = -1;  /* USCH_InformationListIE_RL_AdditionRspTDD */
static int hf_rnsap_RL_LCR_InformationResponse_RL_AdditionRspTDD_PDU = -1;  /* RL_LCR_InformationResponse_RL_AdditionRspTDD */
static int hf_rnsap_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_PDU = -1;  /* UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD */
static int hf_rnsap_UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_PDU = -1;  /* UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD */
static int hf_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_PDU = -1;  /* DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD */
static int hf_rnsap_DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_PDU = -1;  /* DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD */
static int hf_rnsap_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD_PDU = -1;  /* DSCH_LCR_InformationListIEs_RL_AdditionRspTDD */
static int hf_rnsap_USCH_LCR_InformationListIEs_RL_AdditionRspTDD_PDU = -1;  /* USCH_LCR_InformationListIEs_RL_AdditionRspTDD */
static int hf_rnsap_RL_InformationResponse_RL_AdditionRspTDD768_PDU = -1;  /* RL_InformationResponse_RL_AdditionRspTDD768 */
static int hf_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD768_PDU = -1;  /* UL_CCTrCHInformationListIE_RL_AdditionRspTDD768 */
static int hf_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD768_PDU = -1;  /* UL_DPCH_InformationItem_RL_AdditionRspTDD768 */
static int hf_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD768_PDU = -1;  /* DL_CCTrCHInformationListIE_RL_AdditionRspTDD768 */
static int hf_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD768_PDU = -1;  /* DL_DPCH_InformationItem_RL_AdditionRspTDD768 */
static int hf_rnsap_RadioLinkAdditionFailureFDD_PDU = -1;  /* RadioLinkAdditionFailureFDD */
static int hf_rnsap_CauseLevel_RL_AdditionFailureFDD_PDU = -1;  /* CauseLevel_RL_AdditionFailureFDD */
static int hf_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD_PDU = -1;  /* UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD */
static int hf_rnsap_SuccessfulRL_InformationResponse_RL_AdditionFailureFDD_PDU = -1;  /* SuccessfulRL_InformationResponse_RL_AdditionFailureFDD */
static int hf_rnsap_RadioLinkAdditionFailureTDD_PDU = -1;  /* RadioLinkAdditionFailureTDD */
static int hf_rnsap_CauseLevel_RL_AdditionFailureTDD_PDU = -1;  /* CauseLevel_RL_AdditionFailureTDD */
static int hf_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD_PDU = -1;  /* UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD */
static int hf_rnsap_RadioLinkDeletionRequest_PDU = -1;  /* RadioLinkDeletionRequest */
static int hf_rnsap_RL_InformationList_RL_DeletionRqst_PDU = -1;  /* RL_InformationList_RL_DeletionRqst */
static int hf_rnsap_RL_Information_RL_DeletionRqst_PDU = -1;  /* RL_Information_RL_DeletionRqst */
static int hf_rnsap_RadioLinkDeletionResponse_PDU = -1;  /* RadioLinkDeletionResponse */
static int hf_rnsap_RadioLinkReconfigurationPrepareFDD_PDU = -1;  /* RadioLinkReconfigurationPrepareFDD */
static int hf_rnsap_UL_DPCH_Information_RL_ReconfPrepFDD_PDU = -1;  /* UL_DPCH_Information_RL_ReconfPrepFDD */
static int hf_rnsap_DL_DPCH_Information_RL_ReconfPrepFDD_PDU = -1;  /* DL_DPCH_Information_RL_ReconfPrepFDD */
static int hf_rnsap_DL_DPCH_Power_Information_RL_ReconfPrepFDD_PDU = -1;  /* DL_DPCH_Power_Information_RL_ReconfPrepFDD */
static int hf_rnsap_DCH_DeleteList_RL_ReconfPrepFDD_PDU = -1;  /* DCH_DeleteList_RL_ReconfPrepFDD */
static int hf_rnsap_RL_InformationList_RL_ReconfPrepFDD_PDU = -1;  /* RL_InformationList_RL_ReconfPrepFDD */
static int hf_rnsap_RL_Information_RL_ReconfPrepFDD_PDU = -1;  /* RL_Information_RL_ReconfPrepFDD */
static int hf_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep_PDU = -1;  /* Additional_HS_Cell_Information_RL_Reconf_Prep */
static int hf_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Prep_PDU = -1;  /* Additional_EDCH_Cell_Information_RL_Reconf_Prep */
static int hf_rnsap_F_DPCH_Information_RL_ReconfPrepFDD_PDU = -1;  /* F_DPCH_Information_RL_ReconfPrepFDD */
static int hf_rnsap_RadioLinkReconfigurationPrepareTDD_PDU = -1;  /* RadioLinkReconfigurationPrepareTDD */
static int hf_rnsap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU = -1;  /* UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD */
static int hf_rnsap_UL_CCTrCH_AddInformation_RL_ReconfPrepTDD_PDU = -1;  /* UL_CCTrCH_AddInformation_RL_ReconfPrepTDD */
static int hf_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU = -1;  /* UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD */
static int hf_rnsap_UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD_PDU = -1;  /* UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD */
static int hf_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU = -1;  /* UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD */
static int hf_rnsap_UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD_PDU = -1;  /* UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD */
static int hf_rnsap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU = -1;  /* DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD */
static int hf_rnsap_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_PDU = -1;  /* DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD */
static int hf_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU = -1;  /* DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD */
static int hf_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_PDU = -1;  /* DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD */
static int hf_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU = -1;  /* DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD */
static int hf_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_PDU = -1;  /* DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD */
static int hf_rnsap_DCH_DeleteList_RL_ReconfPrepTDD_PDU = -1;  /* DCH_DeleteList_RL_ReconfPrepTDD */
static int hf_rnsap_DSCH_ModifyList_RL_ReconfPrepTDD_PDU = -1;  /* DSCH_ModifyList_RL_ReconfPrepTDD */
static int hf_rnsap_DSCH_DeleteList_RL_ReconfPrepTDD_PDU = -1;  /* DSCH_DeleteList_RL_ReconfPrepTDD */
static int hf_rnsap_USCH_ModifyList_RL_ReconfPrepTDD_PDU = -1;  /* USCH_ModifyList_RL_ReconfPrepTDD */
static int hf_rnsap_USCH_DeleteList_RL_ReconfPrepTDD_PDU = -1;  /* USCH_DeleteList_RL_ReconfPrepTDD */
static int hf_rnsap_RL_Information_RL_ReconfPrepTDD_PDU = -1;  /* RL_Information_RL_ReconfPrepTDD */
static int hf_rnsap_RadioLinkReconfigurationReadyFDD_PDU = -1;  /* RadioLinkReconfigurationReadyFDD */
static int hf_rnsap_RL_InformationResponseList_RL_ReconfReadyFDD_PDU = -1;  /* RL_InformationResponseList_RL_ReconfReadyFDD */
static int hf_rnsap_RL_InformationResponseItem_RL_ReconfReadyFDD_PDU = -1;  /* RL_InformationResponseItem_RL_ReconfReadyFDD */
static int hf_rnsap_Additional_HS_Cell_RL_Reconf_Response_PDU = -1;  /* Additional_HS_Cell_RL_Reconf_Response */
static int hf_rnsap_RadioLinkReconfigurationReadyTDD_PDU = -1;  /* RadioLinkReconfigurationReadyTDD */
static int hf_rnsap_UL_CCTrCHInformationListIE_RL_ReconfReadyTDD_PDU = -1;  /* UL_CCTrCHInformationListIE_RL_ReconfReadyTDD */
static int hf_rnsap_UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_PDU = -1;  /* UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD */
static int hf_rnsap_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_PDU = -1;  /* UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD */
static int hf_rnsap_UL_DPCH_InformationAddList_RL_ReconfReadyTDD768_PDU = -1;  /* UL_DPCH_InformationAddList_RL_ReconfReadyTDD768 */
static int hf_rnsap_UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_PDU = -1;  /* UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD */
static int hf_rnsap_UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_PDU = -1;  /* UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD */
static int hf_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_PDU = -1;  /* UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768 */
static int hf_rnsap_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_PDU = -1;  /* UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD */
static int hf_rnsap_DL_CCTrCHInformationListIE_RL_ReconfReadyTDD_PDU = -1;  /* DL_CCTrCHInformationListIE_RL_ReconfReadyTDD */
static int hf_rnsap_DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_PDU = -1;  /* DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD */
static int hf_rnsap_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_PDU = -1;  /* DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD */
static int hf_rnsap_DL_DPCH_InformationAddList_RL_ReconfReadyTDD768_PDU = -1;  /* DL_DPCH_InformationAddList_RL_ReconfReadyTDD768 */
static int hf_rnsap_DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_PDU = -1;  /* DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD */
static int hf_rnsap_DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_PDU = -1;  /* DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD */
static int hf_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_PDU = -1;  /* DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768 */
static int hf_rnsap_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_PDU = -1;  /* DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD */
static int hf_rnsap_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD_PDU = -1;  /* DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD */
static int hf_rnsap_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_PDU = -1;  /* DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD */
static int hf_rnsap_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_PDU = -1;  /* USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD */
static int hf_rnsap_Multiple_RL_InformationResponse_RL_ReconfReadyTDD_PDU = -1;  /* Multiple_RL_InformationResponse_RL_ReconfReadyTDD */
static int hf_rnsap_RadioLinkReconfigurationCommit_PDU = -1;  /* RadioLinkReconfigurationCommit */
static int hf_rnsap_RadioLinkReconfigurationFailure_PDU = -1;  /* RadioLinkReconfigurationFailure */
static int hf_rnsap_CauseLevel_RL_ReconfFailure_PDU = -1;  /* CauseLevel_RL_ReconfFailure */
static int hf_rnsap_RL_ReconfigurationFailure_RL_ReconfFail_PDU = -1;  /* RL_ReconfigurationFailure_RL_ReconfFail */
static int hf_rnsap_RadioLinkReconfigurationCancel_PDU = -1;  /* RadioLinkReconfigurationCancel */
static int hf_rnsap_RadioLinkReconfigurationRequestFDD_PDU = -1;  /* RadioLinkReconfigurationRequestFDD */
static int hf_rnsap_UL_DPCH_Information_RL_ReconfRqstFDD_PDU = -1;  /* UL_DPCH_Information_RL_ReconfRqstFDD */
static int hf_rnsap_DL_DPCH_Information_RL_ReconfRqstFDD_PDU = -1;  /* DL_DPCH_Information_RL_ReconfRqstFDD */
static int hf_rnsap_DCH_DeleteList_RL_ReconfRqstFDD_PDU = -1;  /* DCH_DeleteList_RL_ReconfRqstFDD */
static int hf_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req_PDU = -1;  /* Additional_HS_Cell_Information_RL_Reconf_Req */
static int hf_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Req_PDU = -1;  /* Additional_EDCH_Cell_Information_RL_Reconf_Req */
static int hf_rnsap_RL_ReconfigurationRequestFDD_RL_InformationList_PDU = -1;  /* RL_ReconfigurationRequestFDD_RL_InformationList */
static int hf_rnsap_RL_ReconfigurationRequestFDD_RL_Information_IEs_PDU = -1;  /* RL_ReconfigurationRequestFDD_RL_Information_IEs */
static int hf_rnsap_RadioLinkReconfigurationRequestTDD_PDU = -1;  /* RadioLinkReconfigurationRequestTDD */
static int hf_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD */
static int hf_rnsap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD */
static int hf_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD */
static int hf_rnsap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD */
static int hf_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD */
static int hf_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD */
static int hf_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD */
static int hf_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD */
static int hf_rnsap_DCH_DeleteList_RL_ReconfRqstTDD_PDU = -1;  /* DCH_DeleteList_RL_ReconfRqstTDD */
static int hf_rnsap_Multiple_RL_ReconfigurationRequestTDD_RL_Information_PDU = -1;  /* Multiple_RL_ReconfigurationRequestTDD_RL_Information */
static int hf_rnsap_RadioLinkReconfigurationResponseFDD_PDU = -1;  /* RadioLinkReconfigurationResponseFDD */
static int hf_rnsap_RL_InformationResponseList_RL_ReconfRspFDD_PDU = -1;  /* RL_InformationResponseList_RL_ReconfRspFDD */
static int hf_rnsap_RL_InformationResponseItem_RL_ReconfRspFDD_PDU = -1;  /* RL_InformationResponseItem_RL_ReconfRspFDD */
static int hf_rnsap_RadioLinkReconfigurationResponseTDD_PDU = -1;  /* RadioLinkReconfigurationResponseTDD */
static int hf_rnsap_RL_InformationResponse_RL_ReconfRspTDD_PDU = -1;  /* RL_InformationResponse_RL_ReconfRspTDD */
static int hf_rnsap_DL_CCTrCH_InformationList_RL_ReconfRspTDD_PDU = -1;  /* DL_CCTrCH_InformationList_RL_ReconfRspTDD */
static int hf_rnsap_DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD_PDU = -1;  /* DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD */
static int hf_rnsap_Multiple_RL_InformationResponse_RL_ReconfRspTDD_PDU = -1;  /* Multiple_RL_InformationResponse_RL_ReconfRspTDD */
static int hf_rnsap_RadioLinkFailureIndication_PDU = -1;  /* RadioLinkFailureIndication */
static int hf_rnsap_Reporting_Object_RL_FailureInd_PDU = -1;  /* Reporting_Object_RL_FailureInd */
static int hf_rnsap_RL_Information_RL_FailureInd_PDU = -1;  /* RL_Information_RL_FailureInd */
static int hf_rnsap_RL_Set_Information_RL_FailureInd_PDU = -1;  /* RL_Set_Information_RL_FailureInd */
static int hf_rnsap_CCTrCH_InformationItem_RL_FailureInd_PDU = -1;  /* CCTrCH_InformationItem_RL_FailureInd */
static int hf_rnsap_RadioLinkPreemptionRequiredIndication_PDU = -1;  /* RadioLinkPreemptionRequiredIndication */
static int hf_rnsap_RL_InformationList_RL_PreemptRequiredInd_PDU = -1;  /* RL_InformationList_RL_PreemptRequiredInd */
static int hf_rnsap_RL_InformationItem_RL_PreemptRequiredInd_PDU = -1;  /* RL_InformationItem_RL_PreemptRequiredInd */
static int hf_rnsap_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd_PDU = -1;  /* HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd */
static int hf_rnsap_HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd_PDU = -1;  /* HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd */
static int hf_rnsap_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd_PDU = -1;  /* EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd */
static int hf_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd_PDU = -1;  /* EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd */
static int hf_rnsap_RadioLinkRestoreIndication_PDU = -1;  /* RadioLinkRestoreIndication */
static int hf_rnsap_Reporting_Object_RL_RestoreInd_PDU = -1;  /* Reporting_Object_RL_RestoreInd */
static int hf_rnsap_RL_Information_RL_RestoreInd_PDU = -1;  /* RL_Information_RL_RestoreInd */
static int hf_rnsap_RL_Set_Information_RL_RestoreInd_PDU = -1;  /* RL_Set_Information_RL_RestoreInd */
static int hf_rnsap_CCTrCH_InformationItem_RL_RestoreInd_PDU = -1;  /* CCTrCH_InformationItem_RL_RestoreInd */
static int hf_rnsap_DL_PowerControlRequest_PDU = -1;  /* DL_PowerControlRequest */
static int hf_rnsap_DL_ReferencePowerInformationList_DL_PC_Rqst_PDU = -1;  /* DL_ReferencePowerInformationList_DL_PC_Rqst */
static int hf_rnsap_DL_ReferencePowerInformation_DL_PC_Rqst_PDU = -1;  /* DL_ReferencePowerInformation_DL_PC_Rqst */
static int hf_rnsap_DL_PowerTimeslotControlRequest_PDU = -1;  /* DL_PowerTimeslotControlRequest */
static int hf_rnsap_PhysicalChannelReconfigurationRequestFDD_PDU = -1;  /* PhysicalChannelReconfigurationRequestFDD */
static int hf_rnsap_RL_Information_PhyChReconfRqstFDD_PDU = -1;  /* RL_Information_PhyChReconfRqstFDD */
static int hf_rnsap_PhysicalChannelReconfigurationRequestTDD_PDU = -1;  /* PhysicalChannelReconfigurationRequestTDD */
static int hf_rnsap_RL_Information_PhyChReconfRqstTDD_PDU = -1;  /* RL_Information_PhyChReconfRqstTDD */
static int hf_rnsap_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_PDU = -1;  /* UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD */
static int hf_rnsap_UL_DPCH_InformationItem_PhyChReconfRqstTDD_PDU = -1;  /* UL_DPCH_InformationItem_PhyChReconfRqstTDD */
static int hf_rnsap_UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_PDU = -1;  /* UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD */
static int hf_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU = -1;  /* UL_Timeslot_InformationList_PhyChReconfRqstTDD768 */
static int hf_rnsap_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_PDU = -1;  /* DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD */
static int hf_rnsap_DL_DPCH_InformationItem_PhyChReconfRqstTDD_PDU = -1;  /* DL_DPCH_InformationItem_PhyChReconfRqstTDD */
static int hf_rnsap_DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_PDU = -1;  /* DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD */
static int hf_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU = -1;  /* DL_Timeslot_InformationList_PhyChReconfRqstTDD768 */
static int hf_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD_PDU = -1;  /* HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD */
static int hf_rnsap_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD_PDU = -1;  /* HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD */
static int hf_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU = -1;  /* HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768 */
static int hf_rnsap_PhysicalChannelReconfigurationCommand_PDU = -1;  /* PhysicalChannelReconfigurationCommand */
static int hf_rnsap_PhysicalChannelReconfigurationFailure_PDU = -1;  /* PhysicalChannelReconfigurationFailure */
static int hf_rnsap_RadioLinkCongestionIndication_PDU = -1;  /* RadioLinkCongestionIndication */
static int hf_rnsap_RL_InformationList_RL_CongestInd_PDU = -1;  /* RL_InformationList_RL_CongestInd */
static int hf_rnsap_RL_InformationItem_RL_CongestInd_PDU = -1;  /* RL_InformationItem_RL_CongestInd */
static int hf_rnsap_DCH_Rate_InformationItem_RL_CongestInd_PDU = -1;  /* DCH_Rate_InformationItem_RL_CongestInd */
static int hf_rnsap_EDCH_MacdFlowSpecificInformationList_RL_CongestInd_PDU = -1;  /* EDCH_MacdFlowSpecificInformationList_RL_CongestInd */
static int hf_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_CongestInd_PDU = -1;  /* EDCH_MacdFlowSpecificInformationItem_RL_CongestInd */
static int hf_rnsap_UplinkSignallingTransferIndicationFDD_PDU = -1;  /* UplinkSignallingTransferIndicationFDD */
static int hf_rnsap_UplinkSignallingTransferIndicationTDD_PDU = -1;  /* UplinkSignallingTransferIndicationTDD */
static int hf_rnsap_DownlinkSignallingTransferRequest_PDU = -1;  /* DownlinkSignallingTransferRequest */
static int hf_rnsap_RelocationCommit_PDU = -1;    /* RelocationCommit */
static int hf_rnsap_PagingRequest_PDU = -1;       /* PagingRequest */
static int hf_rnsap_PagingArea_PagingRqst_PDU = -1;  /* PagingArea_PagingRqst */
static int hf_rnsap_CNOriginatedPage_PagingRqst_PDU = -1;  /* CNOriginatedPage_PagingRqst */
static int hf_rnsap_DedicatedMeasurementInitiationRequest_PDU = -1;  /* DedicatedMeasurementInitiationRequest */
static int hf_rnsap_DedicatedMeasurementObjectType_DM_Rqst_PDU = -1;  /* DedicatedMeasurementObjectType_DM_Rqst */
static int hf_rnsap_RL_InformationItem_DM_Rqst_PDU = -1;  /* RL_InformationItem_DM_Rqst */
static int hf_rnsap_HSSICH_Info_DM_Rqst_PDU = -1;  /* HSSICH_Info_DM_Rqst */
static int hf_rnsap_HSSICH_Info_DM_Rqst_Extension_PDU = -1;  /* HSSICH_Info_DM_Rqst_Extension */
static int hf_rnsap_RL_Set_InformationItem_DM_Rqst_PDU = -1;  /* RL_Set_InformationItem_DM_Rqst */
static int hf_rnsap_DedicatedMeasurementInitiationResponse_PDU = -1;  /* DedicatedMeasurementInitiationResponse */
static int hf_rnsap_DedicatedMeasurementObjectType_DM_Rsp_PDU = -1;  /* DedicatedMeasurementObjectType_DM_Rsp */
static int hf_rnsap_RL_InformationItem_DM_Rsp_PDU = -1;  /* RL_InformationItem_DM_Rsp */
static int hf_rnsap_RL_Set_InformationItem_DM_Rsp_PDU = -1;  /* RL_Set_InformationItem_DM_Rsp */
static int hf_rnsap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_PDU = -1;  /* Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp */
static int hf_rnsap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_PDU = -1;  /* Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp */
static int hf_rnsap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_PDU = -1;  /* Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp */
static int hf_rnsap_Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp_PDU = -1;  /* Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp */
static int hf_rnsap_DedicatedMeasurementInitiationFailure_PDU = -1;  /* DedicatedMeasurementInitiationFailure */
static int hf_rnsap_DedicatedMeasurementObjectType_DM_Fail_PDU = -1;  /* DedicatedMeasurementObjectType_DM_Fail */
static int hf_rnsap_RL_Unsuccessful_InformationItem_DM_Fail_PDU = -1;  /* RL_Unsuccessful_InformationItem_DM_Fail */
static int hf_rnsap_RL_Successful_InformationItem_DM_Fail_PDU = -1;  /* RL_Successful_InformationItem_DM_Fail */
static int hf_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail_PDU = -1;  /* RL_Set_Unsuccessful_InformationItem_DM_Fail */
static int hf_rnsap_RL_Set_Successful_InformationItem_DM_Fail_PDU = -1;  /* RL_Set_Successful_InformationItem_DM_Fail */
static int hf_rnsap_DedicatedMeasurementReport_PDU = -1;  /* DedicatedMeasurementReport */
static int hf_rnsap_DedicatedMeasurementObjectType_DM_Rprt_PDU = -1;  /* DedicatedMeasurementObjectType_DM_Rprt */
static int hf_rnsap_RL_InformationItem_DM_Rprt_PDU = -1;  /* RL_InformationItem_DM_Rprt */
static int hf_rnsap_RL_Set_InformationItem_DM_Rprt_PDU = -1;  /* RL_Set_InformationItem_DM_Rprt */
static int hf_rnsap_DedicatedMeasurementTerminationRequest_PDU = -1;  /* DedicatedMeasurementTerminationRequest */
static int hf_rnsap_DedicatedMeasurementFailureIndication_PDU = -1;  /* DedicatedMeasurementFailureIndication */
static int hf_rnsap_DedicatedMeasurementObjectType_DM_Fail_Ind_PDU = -1;  /* DedicatedMeasurementObjectType_DM_Fail_Ind */
static int hf_rnsap_RL_Unsuccessful_InformationItem_DM_Fail_Ind_PDU = -1;  /* RL_Unsuccessful_InformationItem_DM_Fail_Ind */
static int hf_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind_PDU = -1;  /* RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind */
static int hf_rnsap_CommonTransportChannelResourcesReleaseRequest_PDU = -1;  /* CommonTransportChannelResourcesReleaseRequest */
static int hf_rnsap_CommonTransportChannelResourcesRequest_PDU = -1;  /* CommonTransportChannelResourcesRequest */
static int hf_rnsap_CommonTransportChannelResourcesResponseFDD_PDU = -1;  /* CommonTransportChannelResourcesResponseFDD */
static int hf_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD_PDU = -1;  /* FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD */
static int hf_rnsap_CommonTransportChannelResourcesResponseTDD_PDU = -1;  /* CommonTransportChannelResourcesResponseTDD */
static int hf_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD_PDU = -1;  /* FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD */
static int hf_rnsap_CommonTransportChannelResourcesFailure_PDU = -1;  /* CommonTransportChannelResourcesFailure */
static int hf_rnsap_CompressedModeCommand_PDU = -1;  /* CompressedModeCommand */
static int hf_rnsap_ErrorIndication_PDU = -1;     /* ErrorIndication */
static int hf_rnsap_CommonMeasurementInitiationRequest_PDU = -1;  /* CommonMeasurementInitiationRequest */
static int hf_rnsap_CommonMeasurementObjectType_CM_Rqst_PDU = -1;  /* CommonMeasurementObjectType_CM_Rqst */
static int hf_rnsap_GSM_Cell_CM_Rqst_PDU = -1;    /* GSM_Cell_CM_Rqst */
static int hf_rnsap_Extension_CommonMeasurementObjectType_CM_Rqst_PDU = -1;  /* Extension_CommonMeasurementObjectType_CM_Rqst */
static int hf_rnsap_GsmCellList_CM_Rqst_PDU = -1;  /* GsmCellList_CM_Rqst */
static int hf_rnsap_CommonMeasurementInitiationResponse_PDU = -1;  /* CommonMeasurementInitiationResponse */
static int hf_rnsap_CommonMeasurementObjectType_CM_Rsp_PDU = -1;  /* CommonMeasurementObjectType_CM_Rsp */
static int hf_rnsap_Extension_CommonMeasurementObjectType_CM_Rsp_PDU = -1;  /* Extension_CommonMeasurementObjectType_CM_Rsp */
static int hf_rnsap_GsmCellList_CM_Rsp_PDU = -1;  /* GsmCellList_CM_Rsp */
static int hf_rnsap_CommonMeasurementInitiationFailure_PDU = -1;  /* CommonMeasurementInitiationFailure */
static int hf_rnsap_Extension_FailureMeasurementList_PDU = -1;  /* Extension_FailureMeasurementList */
static int hf_rnsap_CommonMeasurementReport_PDU = -1;  /* CommonMeasurementReport */
static int hf_rnsap_CommonMeasurementObjectType_CM_Rprt_PDU = -1;  /* CommonMeasurementObjectType_CM_Rprt */
static int hf_rnsap_Extension_CommonMeasurementObjectType_CM_Rprt_PDU = -1;  /* Extension_CommonMeasurementObjectType_CM_Rprt */
static int hf_rnsap_GsmCellList_CM_Rprt_PDU = -1;  /* GsmCellList_CM_Rprt */
static int hf_rnsap_CommonMeasurementTerminationRequest_PDU = -1;  /* CommonMeasurementTerminationRequest */
static int hf_rnsap_Extension_TerminationMeasurementList_PDU = -1;  /* Extension_TerminationMeasurementList */
static int hf_rnsap_CommonMeasurementFailureIndication_PDU = -1;  /* CommonMeasurementFailureIndication */
static int hf_rnsap_Extension_FailureIndicationMeasurementList_PDU = -1;  /* Extension_FailureIndicationMeasurementList */
static int hf_rnsap_InformationExchangeInitiationRequest_PDU = -1;  /* InformationExchangeInitiationRequest */
static int hf_rnsap_InformationExchangeObjectType_InfEx_Rqst_PDU = -1;  /* InformationExchangeObjectType_InfEx_Rqst */
static int hf_rnsap_GSM_Cell_InfEx_Rqst_PDU = -1;  /* GSM_Cell_InfEx_Rqst */
static int hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst_PDU = -1;  /* MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst */
static int hf_rnsap_MBMS_Cell_InfEx_Rqst_PDU = -1;  /* MBMS_Cell_InfEx_Rqst */
static int hf_rnsap_ANR_Cell_InfEx_Rqst_PDU = -1;  /* ANR_Cell_InfEx_Rqst */
static int hf_rnsap_Common_E_RGCH_Cell_InfEx_Rqst_PDU = -1;  /* Common_E_RGCH_Cell_InfEx_Rqst */
static int hf_rnsap_InformationExchangeInitiationResponse_PDU = -1;  /* InformationExchangeInitiationResponse */
static int hf_rnsap_InformationExchangeObjectType_InfEx_Rsp_PDU = -1;  /* InformationExchangeObjectType_InfEx_Rsp */
static int hf_rnsap_MBMS_Bearer_Service_List_InfEx_Rsp_PDU = -1;  /* MBMS_Bearer_Service_List_InfEx_Rsp */
static int hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp_PDU = -1;  /* MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp */
static int hf_rnsap_MBMS_Cell_InfEx_Rsp_PDU = -1;  /* MBMS_Cell_InfEx_Rsp */
static int hf_rnsap_ANR_Cell_InfEx_Rsp_PDU = -1;  /* ANR_Cell_InfEx_Rsp */
static int hf_rnsap_Common_E_RGCH_Cell_InfEx_Rsp_PDU = -1;  /* Common_E_RGCH_Cell_InfEx_Rsp */
static int hf_rnsap_InformationExchangeInitiationFailure_PDU = -1;  /* InformationExchangeInitiationFailure */
static int hf_rnsap_InformationReport_PDU = -1;   /* InformationReport */
static int hf_rnsap_InformationExchangeObjectType_InfEx_Rprt_PDU = -1;  /* InformationExchangeObjectType_InfEx_Rprt */
static int hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt_PDU = -1;  /* MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt */
static int hf_rnsap_MBMS_Cell_InfEx_Rprt_PDU = -1;  /* MBMS_Cell_InfEx_Rprt */
static int hf_rnsap_Common_E_RGCH_Cell_InfEx_Rprt_PDU = -1;  /* Common_E_RGCH_Cell_InfEx_Rprt */
static int hf_rnsap_InformationExchangeTerminationRequest_PDU = -1;  /* InformationExchangeTerminationRequest */
static int hf_rnsap_InformationExchangeFailureIndication_PDU = -1;  /* InformationExchangeFailureIndication */
static int hf_rnsap_ResetRequest_PDU = -1;        /* ResetRequest */
static int hf_rnsap_ResetIndicator_PDU = -1;      /* ResetIndicator */
static int hf_rnsap_ContextInfoItem_Reset_PDU = -1;  /* ContextInfoItem_Reset */
static int hf_rnsap_ContextGroupInfoItem_Reset_PDU = -1;  /* ContextGroupInfoItem_Reset */
static int hf_rnsap_ResetResponse_PDU = -1;       /* ResetResponse */
static int hf_rnsap_RadioLinkActivationCommandFDD_PDU = -1;  /* RadioLinkActivationCommandFDD */
static int hf_rnsap_DelayedActivationInformationList_RL_ActivationCmdFDD_PDU = -1;  /* DelayedActivationInformationList_RL_ActivationCmdFDD */
static int hf_rnsap_DelayedActivationInformation_RL_ActivationCmdFDD_PDU = -1;  /* DelayedActivationInformation_RL_ActivationCmdFDD */
static int hf_rnsap_RadioLinkActivationCommandTDD_PDU = -1;  /* RadioLinkActivationCommandTDD */
static int hf_rnsap_DelayedActivationInformationList_RL_ActivationCmdTDD_PDU = -1;  /* DelayedActivationInformationList_RL_ActivationCmdTDD */
static int hf_rnsap_DelayedActivationInformation_RL_ActivationCmdTDD_PDU = -1;  /* DelayedActivationInformation_RL_ActivationCmdTDD */
static int hf_rnsap_GERANUplinkSignallingTransferIndication_PDU = -1;  /* GERANUplinkSignallingTransferIndication */
static int hf_rnsap_RadioLinkParameterUpdateIndicationFDD_PDU = -1;  /* RadioLinkParameterUpdateIndicationFDD */
static int hf_rnsap_RL_ParameterUpdateIndicationFDD_RL_InformationList_PDU = -1;  /* RL_ParameterUpdateIndicationFDD_RL_InformationList */
static int hf_rnsap_RL_ParameterUpdateIndicationFDD_RL_Information_Item_PDU = -1;  /* RL_ParameterUpdateIndicationFDD_RL_Information_Item */
static int hf_rnsap_UE_Measurement_Forwarding_PDU = -1;  /* UE_Measurement_Forwarding */
static int hf_rnsap_Additional_HS_Cell_Information_RL_Param_Upd_PDU = -1;  /* Additional_HS_Cell_Information_RL_Param_Upd */
static int hf_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd_PDU = -1;  /* Additional_EDCH_Cell_Information_RL_Param_Upd */
static int hf_rnsap_RadioLinkParameterUpdateIndicationTDD_PDU = -1;  /* RadioLinkParameterUpdateIndicationTDD */
static int hf_rnsap_UEMeasurementInitiationRequest_PDU = -1;  /* UEMeasurementInitiationRequest */
static int hf_rnsap_UEMeasurementInitiationResponse_PDU = -1;  /* UEMeasurementInitiationResponse */
static int hf_rnsap_UEMeasurementInitiationFailure_PDU = -1;  /* UEMeasurementInitiationFailure */
static int hf_rnsap_UEMeasurementReport_PDU = -1;  /* UEMeasurementReport */
static int hf_rnsap_UEMeasurementTerminationRequest_PDU = -1;  /* UEMeasurementTerminationRequest */
static int hf_rnsap_UEMeasurementFailureIndication_PDU = -1;  /* UEMeasurementFailureIndication */
static int hf_rnsap_IurInvokeTrace_PDU = -1;      /* IurInvokeTrace */
static int hf_rnsap_ListOfInterfacesToTrace_PDU = -1;  /* ListOfInterfacesToTrace */
static int hf_rnsap_InterfacesToTraceItem_PDU = -1;  /* InterfacesToTraceItem */
static int hf_rnsap_IurDeactivateTrace_PDU = -1;  /* IurDeactivateTrace */
static int hf_rnsap_MBMSAttachCommand_PDU = -1;   /* MBMSAttachCommand */
static int hf_rnsap_MBMSDetachCommand_PDU = -1;   /* MBMSDetachCommand */
static int hf_rnsap_DirectInformationTransfer_PDU = -1;  /* DirectInformationTransfer */
static int hf_rnsap_EnhancedRelocationRequest_PDU = -1;  /* EnhancedRelocationRequest */
static int hf_rnsap_EnhancedRelocationResponse_PDU = -1;  /* EnhancedRelocationResponse */
static int hf_rnsap_EnhancedRelocationFailure_PDU = -1;  /* EnhancedRelocationFailure */
static int hf_rnsap_EnhancedRelocationCancel_PDU = -1;  /* EnhancedRelocationCancel */
static int hf_rnsap_EnhancedRelocationSignallingTransfer_PDU = -1;  /* EnhancedRelocationSignallingTransfer */
static int hf_rnsap_EnhancedRelocationRelease_PDU = -1;  /* EnhancedRelocationRelease */
static int hf_rnsap_MBSFNMCCHInformation_PDU = -1;  /* MBSFNMCCHInformation */
static int hf_rnsap_SecondaryULFrequencyReport_PDU = -1;  /* SecondaryULFrequencyReport */
static int hf_rnsap_SecondaryULFrequencyUpdateIndication_PDU = -1;  /* SecondaryULFrequencyUpdateIndication */
static int hf_rnsap_EnhancedRelocationResourceRequest_PDU = -1;  /* EnhancedRelocationResourceRequest */
static int hf_rnsap_EnhancedRelocationResourceResponse_PDU = -1;  /* EnhancedRelocationResourceResponse */
static int hf_rnsap_EnhancedRelocationResourceFailure_PDU = -1;  /* EnhancedRelocationResourceFailure */
static int hf_rnsap_EnhancedRelocationResourceReleaseCommand_PDU = -1;  /* EnhancedRelocationResourceReleaseCommand */
static int hf_rnsap_EnhancedRelocationResourceReleaseComplete_PDU = -1;  /* EnhancedRelocationResourceReleaseComplete */
static int hf_rnsap_PrivateMessage_PDU = -1;      /* PrivateMessage */
static int hf_rnsap_InformationTransferControlRequest_PDU = -1;  /* InformationTransferControlRequest */
static int hf_rnsap_Control_Type_InformationTransferControlReq_PDU = -1;  /* Control_Type_InformationTransferControlReq */
static int hf_rnsap_RNSAP_PDU_PDU = -1;           /* RNSAP_PDU */
static int hf_rnsap_NULL_PDU = -1;                /* NULL */
static int hf_rnsap_local = -1;                   /* INTEGER_0_maxPrivateIEs */
static int hf_rnsap_global = -1;                  /* T_global */
static int hf_rnsap_procedureCode = -1;           /* ProcedureCode */
static int hf_rnsap_ddMode = -1;                  /* DdMode */
static int hf_rnsap_shortTransActionId = -1;      /* INTEGER_0_127 */
static int hf_rnsap_longTransActionId = -1;       /* INTEGER_0_32767 */
static int hf_rnsap_ProtocolIE_Container_item = -1;  /* ProtocolIE_Field */
static int hf_rnsap_id = -1;                      /* ProtocolIE_ID */
static int hf_rnsap_criticality = -1;             /* Criticality */
static int hf_rnsap_value = -1;                   /* ProtocolIE_Field_value */
static int hf_rnsap_ProtocolExtensionContainer_item = -1;  /* ProtocolExtensionField */
static int hf_rnsap_extensionValue = -1;          /* T_extensionValue */
static int hf_rnsap_PrivateIE_Container_item = -1;  /* PrivateIE_Field */
static int hf_rnsap_id_01 = -1;                   /* PrivateIE_ID */
static int hf_rnsap_value_01 = -1;                /* PrivateIE_Field_value */
static int hf_rnsap_ActivationInformation_item = -1;  /* ActivationInformationItem */
static int hf_rnsap_uU_ActivationState = -1;      /* Uu_ActivationState */
static int hf_rnsap_iE_Extensions = -1;           /* ProtocolExtensionContainer */
static int hf_rnsap_Active_MBMS_Bearer_Service_ListFDD_item = -1;  /* MBMS_Bearer_ServiceItemFDD */
static int hf_rnsap_Active_MBMS_Bearer_Service_ListFDD_PFL_item = -1;  /* MBMS_Bearer_ServiceItemFDD_PFL */
static int hf_rnsap_Active_MBMS_Bearer_Service_ListTDD_item = -1;  /* MBMS_Bearer_ServiceItemTDD */
static int hf_rnsap_Active_MBMS_Bearer_Service_ListTDD_PFL_item = -1;  /* MBMS_Bearer_ServiceItemTDD_PFL */
static int hf_rnsap_cMConfigurationChangeCFN = -1;  /* CFN */
static int hf_rnsap_transmission_Gap_Pattern_Sequence_Status = -1;  /* Transmission_Gap_Pattern_Sequence_Status_List */
static int hf_rnsap_Additional_EDCH_Cell_Information_Response_RLAddList_item = -1;  /* Additional_EDCH_Cell_Information_Response_RLAdd_ItemIEs */
static int hf_rnsap_additional_EDCH_FDD_Information_Response = -1;  /* Additional_EDCH_FDD_Information_Response_ItemIEs */
static int hf_rnsap_additional_EDCH_Serving_Cell_Change_Information_Response_RLAdd = -1;  /* E_DCH_Serving_cell_change_informationResponse */
static int hf_rnsap_multicell_EDCH_Transport_Bearer_Mode = -1;  /* Multicell_EDCH_Transport_Bearer_Mode */
static int hf_rnsap_additional_EDCH_Cell_Information_Setup = -1;  /* Additional_EDCH_Cell_Information_Setup */
static int hf_rnsap_pLMN_Identity = -1;           /* PLMN_Identity */
static int hf_rnsap_l3_Information = -1;          /* L3_Information */
static int hf_rnsap_Additional_EDCH_Cell_Information_Setup_item = -1;  /* Additional_EDCH_FDD_Setup_Cell_Information */
static int hf_rnsap_additional_EDCH_UL_DPCH_Information_Setup = -1;  /* Additional_EDCH_UL_DPCH_Information_Setup */
static int hf_rnsap_additional_EDCH_RL_Specific_Information_To_Setup = -1;  /* Additional_EDCH_RL_Specific_Information_To_Setup_List */
static int hf_rnsap_additional_EDCH_FDD_Information = -1;  /* Additional_EDCH_FDD_Information */
static int hf_rnsap_additional_EDCH_F_DPCH_Information_Setup = -1;  /* Additional_EDCH_F_DPCH_Information */
static int hf_rnsap_multicellEDCH_Information = -1;  /* MulticellEDCH_Information */
static int hf_rnsap_ul_ScramblingCode = -1;       /* UL_ScramblingCode */
static int hf_rnsap_ul_SIR_Target = -1;           /* UL_SIR */
static int hf_rnsap_fdd_TPC_DownlinkStepSize = -1;  /* FDD_TPC_DownlinkStepSize */
static int hf_rnsap_limitedPowerIncrease = -1;    /* LimitedPowerIncrease */
static int hf_rnsap_innerLoopDLPCStatus = -1;     /* InnerLoopDLPCStatus */
static int hf_rnsap_f_DPCH_SlotFormatSupportRequest = -1;  /* F_DPCH_SlotFormatSupportRequest */
static int hf_rnsap_Additional_EDCH_RL_Specific_Information_To_Setup_List_item = -1;  /* Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs */
static int hf_rnsap_eDCH_Additional_RL_ID = -1;   /* RL_ID */
static int hf_rnsap_c_ID = -1;                    /* C_ID */
static int hf_rnsap_firstRLS_indicator = -1;      /* FirstRLS_Indicator */
static int hf_rnsap_propagationDelay = -1;        /* PropagationDelay */
static int hf_rnsap_initialDL_transmissionPower = -1;  /* DL_Power */
static int hf_rnsap_primaryCPICH_EcNo = -1;       /* PrimaryCPICH_EcNo */
static int hf_rnsap_e_AGCH_PowerOffset = -1;      /* E_AGCH_PowerOffset */
static int hf_rnsap_e_RGCH_PowerOffset = -1;      /* E_RGCH_PowerOffset */
static int hf_rnsap_e_HICH_PowerOffset = -1;      /* E_HICH_PowerOffset */
static int hf_rnsap_additional_EDCH_MAC_d_Flows_Specific_Info_List = -1;  /* Additional_EDCH_MAC_d_Flows_Specific_Info_List */
static int hf_rnsap_multicellEDCH_RL_SpecificInformation = -1;  /* MulticellEDCH_RL_SpecificInformation */
static int hf_rnsap_Additional_EDCH_Cell_Information_To_Add_List_item = -1;  /* Additional_EDCH_Cell_Information_To_Add_ItemIEs */
static int hf_rnsap_additional_EDCH_UL_DPCH_Information_Setup_01 = -1;  /* Additional_EDCH_UL_DPCH_Information_Addition */
static int hf_rnsap_additional_EDCH_RL_Specific_Information_To_Add_List = -1;  /* Additional_EDCH_RL_Specific_Information_To_Add_List */
static int hf_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_List_item = -1;  /* Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs */
static int hf_rnsap_Additional_EDCH_RL_Specific_Information_To_Modify_List_item = -1;  /* Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs */
static int hf_rnsap_hARQ_Process_Allocation_Scheduled_2ms_EDCH = -1;  /* HARQ_Process_Allocation_2ms_EDCH */
static int hf_rnsap_e_DCH_Maximum_Bitrate = -1;   /* E_DCH_Maximum_Bitrate */
static int hf_rnsap_e_DCH_Processing_Overload_Level = -1;  /* E_DCH_Processing_Overload_Level */
static int hf_rnsap_e_DCH_Min_Set_E_TFCI = -1;    /* E_TFCI */
static int hf_rnsap_uE_DTX_Cycle1 = -1;           /* UE_DTX_Cycle1_2ms */
static int hf_rnsap_uE_DTX_Cycle2 = -1;           /* UE_DTX_Cycle2_ext_2ms */
static int hf_rnsap_inactivity_Threshold_for_UE_DTX_Cycle2 = -1;  /* Inactivity_Threshold_for_UE_DTX_Cycle2 */
static int hf_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_List_item = -1;  /* Additional_EDCH_MAC_d_Flows_Specific_Info */
static int hf_rnsap_e_DCH_MACdFlow_ID = -1;       /* EDCH_MACdFlow_ID */
static int hf_rnsap_bindingID = -1;               /* BindingID */
static int hf_rnsap_transportLayerAddress = -1;   /* TransportLayerAddress */
static int hf_rnsap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List_item = -1;  /* Additional_EDCH_MAC_d_Flows_Specific_Info_Response */
static int hf_rnsap_Additional_EDCH_Cell_Information_Response_List_item = -1;  /* Additional_EDCH_FDD_Information_Response_ItemIEs */
static int hf_rnsap_eDCH_Additional_RL_Specific_Information_Response = -1;  /* EDCH_Additional_RL_Specific_Information_Response_List */
static int hf_rnsap_EDCH_Additional_RL_Specific_Information_Response_List_item = -1;  /* EDCH_Additional_RL_Specific_Information_Response_ItemIEs */
static int hf_rnsap_received_total_wide_band_power = -1;  /* Received_total_wide_band_power */
static int hf_rnsap_dL_PowerBalancing_ActivationIndicator = -1;  /* DL_PowerBalancing_ActivationIndicator */
static int hf_rnsap_rL_Set_ID = -1;               /* RL_Set_ID */
static int hf_rnsap_e_DCH_RL_Set_ID = -1;         /* RL_Set_ID */
static int hf_rnsap_eDCH_FDD_DL_ControlChannelInformation = -1;  /* EDCH_FDD_DL_ControlChannelInformation */
static int hf_rnsap_dl_CodeInformation = -1;      /* FDD_DL_CodeInformation */
static int hf_rnsap_additional_EDCH_MAC_d_Flow_Specific_Information_Response_List = -1;  /* Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List */
static int hf_rnsap_maxUL_SIR = -1;               /* UL_SIR */
static int hf_rnsap_minUL_SIR = -1;               /* UL_SIR */
static int hf_rnsap_maximumAllowedULTxPower = -1;  /* MaximumAllowedULTxPower */
static int hf_rnsap_maximumDL_power = -1;         /* DL_Power */
static int hf_rnsap_minimumDL_power = -1;         /* DL_Power */
static int hf_rnsap_primaryScramblingCode = -1;   /* PrimaryScramblingCode */
static int hf_rnsap_uL_UARFCN = -1;               /* UARFCN */
static int hf_rnsap_dL_UARFCN = -1;               /* UARFCN */
static int hf_rnsap_primaryCPICH_Power = -1;      /* PrimaryCPICH_Power */
static int hf_rnsap_pC_Preamble = -1;             /* PC_Preamble */
static int hf_rnsap_primary_CPICH_Usage_For_Channel_Estimation = -1;  /* Primary_CPICH_Usage_For_Channel_Estimation */
static int hf_rnsap_secondary_CPICH_Information = -1;  /* Secondary_CPICH_Information */
static int hf_rnsap_f_DPCH_SlotFormat = -1;       /* F_DPCH_SlotFormat */
static int hf_rnsap_Additional_EDCH_Cell_Information_Response_RLReconf_List_item = -1;  /* Additional_EDCH_FDD_Information_Response_RLReconf_Items */
static int hf_rnsap_eDCH_Additional_RL_Specific_Modified_Information_Response = -1;  /* EDCH_Additional_RL_Specific_Modified_Information_Response_List */
static int hf_rnsap_EDCH_Additional_RL_Specific_Modified_Information_Response_List_item = -1;  /* EDCH_Additional_RL_Specific_Modified_Information_Response_ItemIEs */
static int hf_rnsap_dL_PowerBalancing_UpdatedIndicator = -1;  /* DL_PowerBalancing_UpdatedIndicator */
static int hf_rnsap_secondary_CPICH_Information_Change = -1;  /* Secondary_CPICH_Information_Change */
static int hf_rnsap_Additional_EDCH_Cell_Information_ConfigurationChange_List_item = -1;  /* Additional_EDCH_ConfigurationChange_Info_ItemIEs */
static int hf_rnsap_additional_EDCH_UL_DPCH_Information_Modify = -1;  /* Additional_EDCH_UL_DPCH_Information_Modify */
static int hf_rnsap_additional_EDCH_RL_Specific_Information_To_Add = -1;  /* Additional_EDCH_RL_Specific_Information_To_Add_List */
static int hf_rnsap_additional_EDCH_RL_Specific_Information_To_Modify = -1;  /* Additional_EDCH_RL_Specific_Information_To_Modify_List */
static int hf_rnsap_additional_EDCH_FDD_Information_To_Modify = -1;  /* Additional_EDCH_FDD_Information */
static int hf_rnsap_additional_EDCH_F_DPCH_Information_Modify = -1;  /* Additional_EDCH_F_DPCH_Information */
static int hf_rnsap_Additional_EDCH_Cell_Information_Removal_List_item = -1;  /* Additional_EDCH_Cell_Information_Removal_Info_ItemIEs */
static int hf_rnsap_rL_on_Secondary_UL_Frequency = -1;  /* RL_on_Secondary_UL_Frequency */
static int hf_rnsap_additional_EDCH_DL_Control_Channel_Change_Information = -1;  /* Additional_EDCH_DL_Control_Channel_Change_Information_List */
static int hf_rnsap_Additional_EDCH_DL_Control_Channel_Change_Information_List_item = -1;  /* Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs */
static int hf_rnsap_AdditionalPreferredFrequency_item = -1;  /* AdditionalPreferredFrequencyItem */
static int hf_rnsap_correspondingCells = -1;      /* CorrespondingCells */
static int hf_rnsap_AffectedUEInformationForMBMS_item = -1;  /* S_RNTI */
static int hf_rnsap_priorityLevel = -1;           /* PriorityLevel */
static int hf_rnsap_pre_emptionCapability = -1;   /* Pre_emptionCapability */
static int hf_rnsap_pre_emptionVulnerability = -1;  /* Pre_emptionVulnerability */
static int hf_rnsap_allowed_UL_Rate = -1;         /* Allowed_Rate */
static int hf_rnsap_allowed_DL_Rate = -1;         /* Allowed_Rate */
static int hf_rnsap_aOA_LCR = -1;                 /* AOA_LCR */
static int hf_rnsap_aOA_LCR_Accuracy_Class = -1;  /* AOA_LCR_Accuracy_Class */
static int hf_rnsap_rNC_ID = -1;                  /* RNC_ID */
static int hf_rnsap_lAC = -1;                     /* LAC */
static int hf_rnsap_rAC = -1;                     /* RAC */
static int hf_rnsap_aNR_FDD_CellInformation = -1;  /* ANR_FDD_CellInformation */
static int hf_rnsap_aNR_TDD_CellInformation = -1;  /* ANR_TDD_CellInformation */
static int hf_rnsap_aNR_LCR_TDD_CellInformation = -1;  /* ANR_LCR_TDD_CellInformation */
static int hf_rnsap_extended_RNC_ID = -1;         /* Extended_RNC_ID */
static int hf_rnsap_txDiversityIndicator = -1;    /* TxDiversityIndicator */
static int hf_rnsap_sTTD_SupportIndicator = -1;   /* STTD_SupportIndicator */
static int hf_rnsap_closedLoopMode1_SupportIndicator = -1;  /* ClosedLoopMode1_SupportIndicator */
static int hf_rnsap_restrictionStateIndicator = -1;  /* RestrictionStateIndicator */
static int hf_rnsap_dpc_ModeChangeSupportIndicator = -1;  /* DPC_Mode_Change_SupportIndicator */
static int hf_rnsap_cellCapabilityContainerFDD = -1;  /* CellCapabilityContainer_FDD */
static int hf_rnsap_sNA_Information = -1;         /* SNA_Information */
static int hf_rnsap_frequencyBandIndicator = -1;  /* FrequencyBandIndicator */
static int hf_rnsap_max_UE_DTX_Cycle = -1;        /* Max_UE_DTX_Cycle */
static int hf_rnsap_aNR_Multiple_PLMN_List = -1;  /* ANR_Multiple_PLMN_List */
static int hf_rnsap_secondaryServingCellList = -1;  /* Secondary_Serving_Cell_List */
static int hf_rnsap_dualBandSecondaryServingCellList = -1;  /* Secondary_Serving_Cell_List */
static int hf_rnsap_cellCapabilityContainerExtensionFDD = -1;  /* CellCapabilityContainerExtension_FDD */
static int hf_rnsap_syncCase = -1;                /* SyncCase */
static int hf_rnsap_timeSlot = -1;                /* TimeSlot */
static int hf_rnsap_sCH_TimeSlot = -1;            /* SCH_TimeSlot */
static int hf_rnsap_sCTD_Indicator = -1;          /* SCTD_Indicator */
static int hf_rnsap_dPCHConstantValue = -1;       /* DPCHConstantValue */
static int hf_rnsap_pCCPCH_Power = -1;            /* PCCPCH_Power */
static int hf_rnsap_cellCapabilityContainerTDD = -1;  /* CellCapabilityContainer_TDD */
static int hf_rnsap_cellCapabilityContainerTDD768 = -1;  /* CellCapabilityContainer_TDD768 */
static int hf_rnsap_cellCapabilityContainerTDD_LCR = -1;  /* CellCapabilityContainer_TDD_LCR */
static int hf_rnsap_cellCapabilityContainerExternsionTDD_LCR = -1;  /* CellCapabilityContainerExtension_TDD_LCR */
static int hf_rnsap_list_Of_PLMNs = -1;           /* List_Of_PLMNs */
static int hf_rnsap_badSatelliteInformation = -1;  /* T_badSatelliteInformation */
static int hf_rnsap_badSatelliteInformation_item = -1;  /* T_badSatelliteInformation_item */
static int hf_rnsap_badSAT_ID = -1;               /* SAT_ID */
static int hf_rnsap_BDS_IGPInfoList_item = -1;    /* BDS_IGPInfo */
static int hf_rnsap_bds_IGPNumber = -1;           /* INTEGER_1_320 */
static int hf_rnsap_bds_VerticalDelay = -1;       /* BIT_STRING_SIZE_9 */
static int hf_rnsap_bds_GIVEI = -1;               /* BIT_STRING_SIZE_4 */
static int hf_rnsap_bDS_RefTime = -1;             /* INTEGER_0_119 */
static int hf_rnsap_bds_IGPInfoList = -1;         /* BDS_IGPInfoList */
static int hf_rnsap_nCC = -1;                     /* NCC */
static int hf_rnsap_bCC = -1;                     /* BCC */
static int hf_rnsap_burstStart = -1;              /* INTEGER_0_15 */
static int hf_rnsap_burstLength = -1;             /* INTEGER_10_25 */
static int hf_rnsap_burstFreq = -1;               /* INTEGER_1_16 */
static int hf_rnsap_radioNetwork = -1;            /* CauseRadioNetwork */
static int hf_rnsap_transport = -1;               /* CauseTransport */
static int hf_rnsap_protocol = -1;                /* CauseProtocol */
static int hf_rnsap_misc = -1;                    /* CauseMisc */
static int hf_rnsap_cellIdList = -1;              /* CellIdList */
static int hf_rnsap_CellIdList_item = -1;         /* Cell_Id */
static int hf_rnsap_uplinkCellCapacityClassValue = -1;  /* INTEGER_1_100_ */
static int hf_rnsap_downlinkCellCapacityClassValue = -1;  /* INTEGER_1_100_ */
static int hf_rnsap_lAI = -1;                     /* LAI */
static int hf_rnsap_cI = -1;                      /* CI */
static int hf_rnsap_tUTRANGPSMeasurementAccuracyClass = -1;  /* TUTRANGPSAccuracyClass */
static int hf_rnsap_tUTRANGANSSMeasurementAccuracyClass = -1;  /* TUTRANGANSSAccuracyClass */
static int hf_rnsap_tUTRANGPSMeasurementValueInformation = -1;  /* TUTRANGPSMeasurementValueInformation */
static int hf_rnsap_sFNSFNMeasurementValueInformation = -1;  /* SFNSFNMeasurementValueInformation */
static int hf_rnsap_loadValue = -1;               /* LoadValue */
static int hf_rnsap_transmittedCarrierPowerValue = -1;  /* INTEGER_0_100 */
static int hf_rnsap_receivedTotalWideBandPowerValue = -1;  /* INTEGER_0_621 */
static int hf_rnsap_uplinkTimeslotISCPValue = -1;  /* UL_TimeslotISCP */
static int hf_rnsap_extension_CommonMeasurementValue = -1;  /* Extension_CommonMeasurementValue */
static int hf_rnsap_measurementAvailable = -1;    /* CommonMeasurementAvailable */
static int hf_rnsap_measurementnotAvailable = -1;  /* NULL */
static int hf_rnsap_commonMeasurementValue = -1;  /* CommonMeasurementValue */
static int hf_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDD_item = -1;  /* Common_EDCH_MAC_d_Flow_Specific_InformationFDDItem */
static int hf_rnsap_common_EDCH_MACdFlow_ID = -1;  /* EDCH_MACdFlow_ID */
static int hf_rnsap_maximum_Number_of_Retransmissions_For_E_DCH = -1;  /* MaxNr_Retransmissions_EDCH */
static int hf_rnsap_eDCH_HARQ_PO_FDD = -1;        /* E_DCH_HARQ_PO_FDD */
static int hf_rnsap_eDCH_MACdFlow_Multiplexing_List = -1;  /* E_DCH_MACdFlow_Multiplexing_List */
static int hf_rnsap_common_E_DCHLogicalChannelInformation = -1;  /* Common_E_DCH_LogicalChannelInformation */
static int hf_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationLCR_item = -1;  /* Common_EDCH_MAC_d_Flow_Specific_InformationItemLCR */
static int hf_rnsap_common_EDCH_MACdFlow_ID_LCR = -1;  /* EDCH_MACdFlow_ID_LCR */
static int hf_rnsap_eDCH_HARQ_PO_TDD = -1;        /* E_DCH_HARQ_PO_TDD */
static int hf_rnsap_Common_E_DCH_LogicalChannelInformation_item = -1;  /* Common_E_DCH_LogicalChannelInformationItem */
static int hf_rnsap_logicalChannelId = -1;        /* LogicalChannelID */
static int hf_rnsap_maximumMACdPDU_SizeExtended = -1;  /* MAC_PDU_SizeExtended */
static int hf_rnsap_deActivation = -1;            /* Common_E_RGCH_De_Activation */
static int hf_rnsap_common_E_RGCH_Cell_Info = -1;  /* Common_E_RGCH_Cell_Info */
static int hf_rnsap_eRGCH_ChannelisationCode = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_rnsap_eRGCH_SignatureSequence = -1;  /* ERGCH_SignatureSequence */
static int hf_rnsap_serving_Grant_Value = -1;     /* E_Serving_Grant_Value */
static int hf_rnsap_uE_DTX_DRX_Offset = -1;       /* UE_DTX_DRX_Offset */
static int hf_rnsap_enabling_Delay = -1;          /* Enabling_Delay */
static int hf_rnsap_dTX_Information = -1;         /* DTX_Information */
static int hf_rnsap_dRX_Information = -1;         /* DRX_Information */
static int hf_rnsap_dTX_Information_to_Modify = -1;  /* DTX_Information_to_Modify */
static int hf_rnsap_dRX_Information_to_Modify = -1;  /* DRX_Information_to_Modify */
static int hf_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_item = -1;  /* Continuous_Packet_Connectivity_HS_SCCH_Less_InformationItem */
static int hf_rnsap_transport_Block_Size_Index = -1;  /* Transport_Block_Size_Index */
static int hf_rnsap_hSPDSCH_Second_Code_Support = -1;  /* HSPDSCH_Second_Code_Support */
static int hf_rnsap_hSPDSCH_First_Code_Index = -1;  /* HSPDSCH_First_Code_Index */
static int hf_rnsap_hSPDSCH_Second_Code_Index = -1;  /* HSPDSCH_Second_Code_Index */
static int hf_rnsap_CorrespondingCells_item = -1;  /* C_ID */
static int hf_rnsap_continuous_Packet_Connectivity_DTX_DRX_Information = -1;  /* Continuous_Packet_Connectivity_DTX_DRX_Information */
static int hf_rnsap_continuous_Packet_Connectivity_DTX_DRX_Information_to_Modify = -1;  /* Continuous_Packet_Connectivity_DTX_DRX_Information_to_Modify */
static int hf_rnsap_continuous_Packet_Connectivity_HS_SCCH_Less_Information = -1;  /* Continuous_Packet_Connectivity_HS_SCCH_Less_Information */
static int hf_rnsap_Counting_Information_item = -1;  /* Counting_Information_List */
static int hf_rnsap_counting_Result = -1;         /* Counting_Result */
static int hf_rnsap_procedureID = -1;             /* ProcedureID */
static int hf_rnsap_triggeringMessage = -1;       /* TriggeringMessage */
static int hf_rnsap_procedureCriticality = -1;    /* Criticality */
static int hf_rnsap_transactionID = -1;           /* TransactionID */
static int hf_rnsap_iEsCriticalityDiagnostics = -1;  /* CriticalityDiagnostics_IE_List */
static int hf_rnsap_CriticalityDiagnostics_IE_List_item = -1;  /* CriticalityDiagnostics_IE_List_item */
static int hf_rnsap_iECriticality = -1;           /* Criticality */
static int hf_rnsap_iE_ID = -1;                   /* ProtocolIE_ID */
static int hf_rnsap_repetitionNumber = -1;        /* RepetitionNumber0 */
static int hf_rnsap_MessageStructure_item = -1;   /* MessageStructure_item */
static int hf_rnsap_repetitionNumber_01 = -1;     /* RepetitionNumber1 */
static int hf_rnsap_continuousPacketConnectivity_DRX_InformationLCR = -1;  /* ContinuousPacketConnectivity_DRX_InformationLCR */
static int hf_rnsap_continuousPacketConnectivity_DRX_Information_to_Modify_LCR = -1;  /* ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR */
static int hf_rnsap_hS_DSCH_Semi_PersistentScheduling_Information_LCR = -1;  /* HS_DSCH_Semi_PersistentScheduling_Information_LCR */
static int hf_rnsap_hS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR = -1;  /* HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR */
static int hf_rnsap_hS_DSCH_SPS_Deactivate_Indicator_LCR = -1;  /* NULL */
static int hf_rnsap_e_DCH_Semi_PersistentScheduling_Information_LCR = -1;  /* E_DCH_Semi_PersistentScheduling_Information_LCR */
static int hf_rnsap_e_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR = -1;  /* E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR */
static int hf_rnsap_e_DCH_SPS_Deactivate_Indicator_LCR = -1;  /* NULL */
static int hf_rnsap_hS_SCCH_DRX_Information_LCR = -1;  /* HS_SCCH_DRX_Information_LCR */
static int hf_rnsap_e_AGCH_DRX_Information_LCR = -1;  /* E_AGCH_DRX_Information_LCR */
static int hf_rnsap_hS_SCCH_UE_DRX_Cycle_LCR = -1;  /* UE_DRX_Cycle_LCR */
static int hf_rnsap_hS_SCCH_Inactivity_Threshold_for_UE_DRX_Cycle_LCR = -1;  /* Inactivity_Threshold_for_UE_DRX_Cycle_LCR */
static int hf_rnsap_hS_SCCH_UE_DRX_Offset_LCR = -1;  /* UE_DRX_Offset_LCR */
static int hf_rnsap_sameAsHS_SCCH = -1;           /* NULL */
static int hf_rnsap_e_AGCH_DRX_Parameters = -1;   /* E_AGCH_DRX_Parameters */
static int hf_rnsap_e_AGCH_UE_DRX_Cycle_LCR = -1;  /* UE_DRX_Cycle_LCR */
static int hf_rnsap_e_AGCH_UE_Inactivity_Monitor_Threshold = -1;  /* E_AGCH_UE_Inactivity_Monitor_Threshold */
static int hf_rnsap_e_AGCH_UE_DRX_Offset_LCR = -1;  /* UE_DRX_Offset_LCR */
static int hf_rnsap_dRX_Information_to_Modify_LCR = -1;  /* DRX_Information_to_Modify_LCR */
static int hf_rnsap_modify = -1;                  /* DRX_Information_to_Modify_Items_LCR */
static int hf_rnsap_deactivate = -1;              /* NULL */
static int hf_rnsap_hS_SCCH_DRX_Information_ResponseLCR = -1;  /* HS_SCCH_DRX_Information_ResponseLCR */
static int hf_rnsap_e_AGCH_DRX_Information_ResponseLCR = -1;  /* E_AGCH_DRX_Information_ResponseLCR */
static int hf_rnsap_e_AGCH_DRX_Parameters_Response = -1;  /* E_AGCH_DRX_Parameters_Response */
static int hf_rnsap_multiple_PLMN_List = -1;      /* Multiple_PLMN_List */
static int hf_rnsap_umts_cell_Info = -1;          /* UMTS_Cells_Info_List */
static int hf_rnsap_dGANSS_Signal_ID = -1;        /* BIT_STRING_SIZE_8 */
static int hf_rnsap_ie_Extensions = -1;           /* ProtocolExtensionContainer */
static int hf_rnsap_dBDS_InfoList = -1;           /* DBDS_InfoList */
static int hf_rnsap_dBDS_SignalID = -1;           /* GANSS_Signal_ID */
static int hf_rnsap_dBDS_SignalInfoList = -1;     /* DBDS_SignalInfoList */
static int hf_rnsap_DBDS_SignalInfoList_item = -1;  /* DBDS_SignalInfo */
static int hf_rnsap_svID = -1;                    /* INTEGER_0_63 */
static int hf_rnsap_bds_UDREI = -1;               /* INTEGER_0_15 */
static int hf_rnsap_bds_RURAI = -1;               /* INTEGER_0_15 */
static int hf_rnsap_bds_ECC_DeltaT = -1;          /* BIT_STRING_SIZE_13 */
static int hf_rnsap_DBDS_InfoList_item = -1;      /* DBDS_Info */
static int hf_rnsap_pO2_SRB = -1;                 /* PowerOffset */
static int hf_rnsap_dl_FET_Mode = -1;             /* DL_FET_Mode */
static int hf_rnsap_dCH_ENH_Concat = -1;          /* DCH_ENH_Concat */
static int hf_rnsap_pO_SRB = -1;                  /* PowerOffset */
static int hf_rnsap_DCH_ENH_Concat_item = -1;     /* DCH_ID */
static int hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH = -1;  /* Setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH */
static int hf_rnsap_DCH_FDD_Information_item = -1;  /* DCH_FDD_InformationItem */
static int hf_rnsap_payloadCRC_PresenceIndicator = -1;  /* PayloadCRC_PresenceIndicator */
static int hf_rnsap_ul_FP_Mode = -1;              /* UL_FP_Mode */
static int hf_rnsap_toAWS = -1;                   /* ToAWS */
static int hf_rnsap_toAWE = -1;                   /* ToAWE */
static int hf_rnsap_dCH_SpecificInformationList = -1;  /* DCH_Specific_FDD_InformationList */
static int hf_rnsap_DCH_MeasurementOccasion_Information_item = -1;  /* DchMeasurementOccasionInformation_Item */
static int hf_rnsap_pattern_Sequence_Identifier = -1;  /* Pattern_Sequence_Identifier */
static int hf_rnsap_status_Flag = -1;             /* Status_Flag */
static int hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters = -1;  /* Measurement_Occasion_Pattern_Sequence_parameters */
static int hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters_k = -1;  /* INTEGER_1_9 */
static int hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters_offset = -1;  /* INTEGER_0_511 */
static int hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters_M_Length = -1;  /* INTEGER_1_512 */
static int hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters_Timeslot_Bitmap = -1;  /* BIT_STRING_SIZE_7 */
static int hf_rnsap_DCH_Specific_FDD_InformationList_item = -1;  /* DCH_Specific_FDD_Item */
static int hf_rnsap_dCH_ID = -1;                  /* DCH_ID */
static int hf_rnsap_trCH_SrcStatisticsDescr = -1;  /* TrCH_SrcStatisticsDescr */
static int hf_rnsap_ul_transportFormatSet = -1;   /* TransportFormatSet */
static int hf_rnsap_dl_transportFormatSet = -1;   /* TransportFormatSet */
static int hf_rnsap_ul_BLER = -1;                 /* BLER */
static int hf_rnsap_dl_BLER = -1;                 /* BLER */
static int hf_rnsap_allocationRetentionPriority = -1;  /* AllocationRetentionPriority */
static int hf_rnsap_frameHandlingPriority = -1;   /* FrameHandlingPriority */
static int hf_rnsap_qE_Selector = -1;             /* QE_Selector */
static int hf_rnsap_dRACControl = -1;             /* DRACControl */
static int hf_rnsap_DCH_InformationResponse_item = -1;  /* DCH_InformationResponseItem */
static int hf_rnsap_DCH_TDD_Information_item = -1;  /* DCH_TDD_InformationItem */
static int hf_rnsap_dCH_SpecificInformationList_01 = -1;  /* DCH_Specific_TDD_InformationList */
static int hf_rnsap_DCH_Specific_TDD_InformationList_item = -1;  /* DCH_Specific_TDD_Item */
static int hf_rnsap_ul_cCTrCH_ID = -1;            /* CCTrCH_ID */
static int hf_rnsap_dl_cCTrCH_ID = -1;            /* CCTrCH_ID */
static int hf_rnsap_sIR_Value = -1;               /* SIR_Value */
static int hf_rnsap_sIR_ErrorValue = -1;          /* SIR_Error_Value */
static int hf_rnsap_transmittedCodePowerValue = -1;  /* Transmitted_Code_Power_Value */
static int hf_rnsap_rSCP = -1;                    /* RSCP_Value */
static int hf_rnsap_rxTimingDeviationValue = -1;  /* Rx_Timing_Deviation_Value */
static int hf_rnsap_roundTripTime = -1;           /* Round_Trip_Time_Value */
static int hf_rnsap_extension_DedicatedMeasurementValue = -1;  /* Extension_DedicatedMeasurementValue */
static int hf_rnsap_measurementAvailable_01 = -1;  /* DedicatedMeasurementAvailable */
static int hf_rnsap_measurementnotAvailable_01 = -1;  /* DedicatedMeasurementnotAvailable */
static int hf_rnsap_dedicatedmeasurementValue = -1;  /* DedicatedMeasurementValue */
static int hf_rnsap_cFN = -1;                     /* CFN */
static int hf_rnsap_cfn = -1;                     /* CFN */
static int hf_rnsap_separate_indication = -1;     /* NULL */
static int hf_rnsap_activate = -1;                /* Activate_Info */
static int hf_rnsap_deactivate_01 = -1;           /* Deactivate_Info */
static int hf_rnsap_activation_type = -1;         /* Execution_Type */
static int hf_rnsap_initial_dl_tx_power = -1;     /* DL_Power */
static int hf_rnsap_firstRLS_Indicator = -1;      /* FirstRLS_Indicator */
static int hf_rnsap_propagation_delay = -1;       /* PropagationDelay */
static int hf_rnsap_deactivation_type = -1;       /* Execution_Type */
static int hf_rnsap_synchronised = -1;            /* CFN */
static int hf_rnsap_unsynchronised = -1;          /* NULL */
static int hf_rnsap_dGANSS_ReferenceTime = -1;    /* INTEGER_0_119 */
static int hf_rnsap_dGANSS_Information = -1;      /* T_dGANSS_Information */
static int hf_rnsap_dGANSS_Information_item = -1;  /* T_dGANSS_Information_item */
static int hf_rnsap_gANSS_SignalId = -1;          /* GANSS_Signal_ID */
static int hf_rnsap_gANSS_StatusHealth = -1;      /* GANSS_StatusHealth */
static int hf_rnsap_dGANSS_SignalInformation = -1;  /* T_dGANSS_SignalInformation */
static int hf_rnsap_dGANSS_SignalInformation_item = -1;  /* T_dGANSS_SignalInformation_item */
static int hf_rnsap_satId = -1;                   /* INTEGER_0_63 */
static int hf_rnsap_gANSS_iod = -1;               /* BIT_STRING_SIZE_10 */
static int hf_rnsap_udre = -1;                    /* UDRE */
static int hf_rnsap_ganss_prc = -1;               /* INTEGER_M2047_2047 */
static int hf_rnsap_ganss_rrc = -1;               /* INTEGER_M127_127 */
static int hf_rnsap_pRCDeviation = -1;            /* PRCDeviation */
static int hf_rnsap_udreGrowthRate = -1;          /* UDREGrowthRate */
static int hf_rnsap_udreValidityTime = -1;        /* UDREValidityTime */
static int hf_rnsap_gPSTOW = -1;                  /* GPSTOW */
static int hf_rnsap_gPS_Status_Health = -1;       /* GPS_Status_Health */
static int hf_rnsap_satellite_DGPSCorrections_Information = -1;  /* T_satellite_DGPSCorrections_Information */
static int hf_rnsap_satellite_DGPSCorrections_Information_item = -1;  /* T_satellite_DGPSCorrections_Information_item */
static int hf_rnsap_sAT_ID = -1;                  /* SAT_ID */
static int hf_rnsap_iode_dgps = -1;               /* BIT_STRING_SIZE_8 */
static int hf_rnsap_uDRE = -1;                    /* UDRE */
static int hf_rnsap_pRC = -1;                     /* PRC */
static int hf_rnsap_range_Correction_Rate = -1;   /* Range_Correction_Rate */
static int hf_rnsap_hs_DSCH_Cell_List = -1;       /* HS_DSCH_Cell_List */
static int hf_rnsap_HS_DSCH_Cell_List_item = -1;  /* HS_DSCH_Cell */
static int hf_rnsap_hs_DSCH_Cell_TBS = -1;        /* DL_TBS_Value */
static int hf_rnsap_powerAdjustmentType = -1;     /* PowerAdjustmentType */
static int hf_rnsap_dLReferencePower = -1;        /* DL_Power */
static int hf_rnsap_dLReferencePowerList = -1;    /* DL_ReferencePowerInformationList */
static int hf_rnsap_maxAdjustmentStep = -1;       /* MaxAdjustmentStep */
static int hf_rnsap_adjustmentPeriod = -1;        /* AdjustmentPeriod */
static int hf_rnsap_adjustmentRatio = -1;         /* ScaledAdjustmentRatio */
static int hf_rnsap_DL_ReferencePowerInformationList_item = -1;  /* DL_ReferencePowerInformationItem */
static int hf_rnsap_rL_ID = -1;                   /* RL_ID */
static int hf_rnsap_dl_Reference_Power = -1;      /* DL_Power */
static int hf_rnsap_common_DL_ReferencePowerInformation = -1;  /* DL_Power */
static int hf_rnsap_individual_DL_ReferencePowerInformation = -1;  /* DL_ReferencePowerInformationList */
static int hf_rnsap_DL_Timeslot_Information_item = -1;  /* DL_Timeslot_InformationItem */
static int hf_rnsap_midambleShiftAndBurstType = -1;  /* MidambleShiftAndBurstType */
static int hf_rnsap_tFCI_Presence = -1;           /* TFCI_Presence */
static int hf_rnsap_dL_Code_Information = -1;     /* TDD_DL_Code_Information */
static int hf_rnsap_DL_TimeslotLCR_Information_item = -1;  /* DL_TimeslotLCR_InformationItem */
static int hf_rnsap_timeSlotLCR = -1;             /* TimeSlotLCR */
static int hf_rnsap_midambleShiftLCR = -1;        /* MidambleShiftLCR */
static int hf_rnsap_dL_Code_LCR_Information = -1;  /* TDD_DL_Code_LCR_Information */
static int hf_rnsap_DL_Timeslot_Information768_item = -1;  /* DL_Timeslot_InformationItem768 */
static int hf_rnsap_midambleShiftAndBurstType768 = -1;  /* MidambleShiftAndBurstType768 */
static int hf_rnsap_dL_Code_Information768 = -1;  /* TDD_DL_Code_Information768 */
static int hf_rnsap_DL_TimeSlot_ISCP_Info_item = -1;  /* DL_TimeSlot_ISCP_InfoItem */
static int hf_rnsap_dL_TimeslotISCP = -1;         /* DL_TimeslotISCP */
static int hf_rnsap_DL_TimeSlot_ISCP_LCR_Information_item = -1;  /* DL_TimeSlot_ISCP_LCR_InfoItem */
static int hf_rnsap_decimationFactorforPrimaryFrequency = -1;  /* DecimationFactor */
static int hf_rnsap_decimationFactorforSecFrequency = -1;  /* DecimationFactor */
static int hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements = -1;  /* Setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements */
static int hf_rnsap_uE_DRX_Cycle = -1;            /* UE_DRX_Cycle */
static int hf_rnsap_inactivity_Threshold_for_UE_DRX_Cycle = -1;  /* Inactivity_Threshold_for_UE_DRX_Cycle */
static int hf_rnsap_inactivity_Threshold_for_UE_Grant_Monitoring = -1;  /* Inactivity_Threshold_for_UE_Grant_Monitoring */
static int hf_rnsap_uE_DRX_Grant_Monitoring = -1;  /* UE_DRX_Grant_Monitoring */
static int hf_rnsap_modify_01 = -1;               /* DRX_Information_to_Modify_Items */
static int hf_rnsap_DSCH_FlowControlInformation_item = -1;  /* DSCH_FlowControlItem */
static int hf_rnsap_dSCH_SchedulingPriority = -1;  /* SchedulingPriorityIndicator */
static int hf_rnsap_mAC_c_sh_SDU_Lengths = -1;    /* MAC_c_sh_SDU_LengthList */
static int hf_rnsap_DSCH_TDD_Information_item = -1;  /* DSCH_TDD_InformationItem */
static int hf_rnsap_dSCH_ID = -1;                 /* DSCH_ID */
static int hf_rnsap_dl_ccTrCHID = -1;             /* CCTrCH_ID */
static int hf_rnsap_trChSourceStatisticsDescriptor = -1;  /* TrCH_SrcStatisticsDescr */
static int hf_rnsap_transportFormatSet = -1;      /* TransportFormatSet */
static int hf_rnsap_schedulingPriorityIndicator = -1;  /* SchedulingPriorityIndicator */
static int hf_rnsap_bLER = -1;                    /* BLER */
static int hf_rnsap_uE_DTX_Cycle1_2ms = -1;       /* UE_DTX_Cycle1_2ms */
static int hf_rnsap_uE_DTX_Cycle2_2ms = -1;       /* UE_DTX_Cycle2_2ms */
static int hf_rnsap_mAC_DTX_Cycle_2ms = -1;       /* MAC_DTX_Cycle_2ms */
static int hf_rnsap_uE_DTX_Cycle1_10ms = -1;      /* UE_DTX_Cycle1_10ms */
static int hf_rnsap_uE_DTX_Cycle2_10ms = -1;      /* UE_DTX_Cycle2_10ms */
static int hf_rnsap_mAC_DTX_Cycle_10ms = -1;      /* MAC_DTX_Cycle_10ms */
static int hf_rnsap_e_DCH_TTI_Length = -1;        /* E_DCH_TTI_Length */
static int hf_rnsap_uE_DTX_Long_Preamble = -1;    /* UE_DTX_Long_Preamble */
static int hf_rnsap_mAC_Inactivity_Threshold = -1;  /* MAC_Inactivity_Threshold */
static int hf_rnsap_cQI_DTX_Timer = -1;           /* CQI_DTX_Timer */
static int hf_rnsap_uE_DPCCH_burst1 = -1;         /* UE_DPCCH_burst1 */
static int hf_rnsap_uE_DPCCH_burst2 = -1;         /* UE_DPCCH_burst2 */
static int hf_rnsap_modify_02 = -1;               /* DTX_Information_to_Modify_Items */
static int hf_rnsap_e_DCH_TTI_Length_to_Modify = -1;  /* E_DCH_TTI_Length_to_Modify */
static int hf_rnsap_fDD = -1;                     /* EARFCN_FDD */
static int hf_rnsap_tDD = -1;                     /* EARFCN */
static int hf_rnsap_extension_EARFCN_Information = -1;  /* Extension_EARFCN_Information */
static int hf_rnsap_uL_EARFCN = -1;               /* EARFCN */
static int hf_rnsap_dL_EARFCN = -1;               /* EARFCN */
static int hf_rnsap_uL_EARFCN_01 = -1;            /* EARFCN_Extended */
static int hf_rnsap_dL_EARFCN_01 = -1;            /* EARFCN_Extended */
static int hf_rnsap_e_UTRAN_Cell_ID = -1;         /* BIT_STRING_SIZE_28 */
static int hf_rnsap_eAGCH_ERGCH_EHICH_FDD_ScramblingCode = -1;  /* DL_ScramblingCode */
static int hf_rnsap_eAGCH_ChannelisationCode = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_rnsap_primary_e_RNTI = -1;          /* E_RNTI */
static int hf_rnsap_secondary_e_RNTI = -1;        /* E_RNTI */
static int hf_rnsap_eRGCH_EHICH_ChannelisationCode = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_rnsap_eHICH_SignatureSequence = -1;  /* EHICH_SignatureSequence */
static int hf_rnsap_primary_Secondary_Grant_Selector = -1;  /* E_Primary_Secondary_Grant_Selector */
static int hf_rnsap_e_RGCH_Release_Indicator = -1;  /* E_RGCH_Release_Indicator */
static int hf_rnsap_eDCH_MACdFlows_Information = -1;  /* EDCH_MACdFlows_Information */
static int hf_rnsap_e_DCH_Reference_Power_Offset = -1;  /* E_DCH_Reference_Power_Offset */
static int hf_rnsap_eDCH_MACdFlow_Specific_InformationResponse = -1;  /* EDCH_MACdFlow_Specific_InformationResponse */
static int hf_rnsap_EDCH_MACdFlow_Specific_InformationResponse_item = -1;  /* EDCH_MACdFlow_Specific_InformationResponseItem */
static int hf_rnsap_eDCH_MACdFlow_ID = -1;        /* EDCH_MACdFlow_ID */
static int hf_rnsap_hARQ_Process_Allocation_NonSched_2ms_EDCH = -1;  /* HARQ_Process_Allocation_2ms_EDCH */
static int hf_rnsap_eDCH_MACdFlow_Specific_Information = -1;  /* EDCH_MACdFlow_Specific_InfoToModifyList */
static int hf_rnsap_mACeReset_Indicator = -1;     /* MACeReset_Indicator */
static int hf_rnsap_e_DCH_MACdFlow_Specific_UpdateInformation = -1;  /* E_DCH_MACdFlow_Specific_UpdateInformation */
static int hf_rnsap_E_DCH_MACdFlow_Specific_UpdateInformation_item = -1;  /* E_DCH_MACdFlow_Specific_UpdateInformation_Item */
static int hf_rnsap_E_DCH_DL_Control_Channel_Change_Information_item = -1;  /* E_DCH_DL_Control_Channel_Change_Information_Item */
static int hf_rnsap_e_DCH_RL_ID = -1;             /* RL_ID */
static int hf_rnsap_E_DCH_DL_Control_Channel_Grant_Information_item = -1;  /* E_DCH_DL_Control_Channel_Grant_Information_Item */
static int hf_rnsap_e_DCH_Non_Scheduled_Transmission_Grant = -1;  /* E_DCH_Non_Scheduled_Transmission_Grant_Items */
static int hf_rnsap_e_DCH_Scheduled_Transmission_Grant = -1;  /* NULL */
static int hf_rnsap_E_DCH_LogicalChannelInformation_item = -1;  /* E_DCH_LogicalChannelInformationItem */
static int hf_rnsap_schedulingInformation = -1;   /* SchedulingInformation */
static int hf_rnsap_mACes_GuaranteedBitRate = -1;  /* MACes_Guaranteed_Bitrate */
static int hf_rnsap_eDCH_DDI_Value = -1;          /* EDCH_DDI_Value */
static int hf_rnsap_mACd_PDU_Size_List = -1;      /* E_DCH_MACdPDU_SizeList */
static int hf_rnsap_E_DCH_MACdPDU_SizeList_item = -1;  /* E_DCH_MACdPDU_SizeListItem */
static int hf_rnsap_mACdPDU_Size = -1;            /* MACdPDU_Size */
static int hf_rnsap_E_DCH_LogicalChannelToModify_item = -1;  /* E_DCH_LogicalChannelToModifyItem */
static int hf_rnsap_mACd_PDU_Size_List_01 = -1;   /* E_DCH_MACdPDU_SizeToModifyList */
static int hf_rnsap_E_DCH_MACdPDU_SizeToModifyList_item = -1;  /* E_DCH_MACdPDU_SizeListItem */
static int hf_rnsap_E_DCH_LogicalChannelToDelete_item = -1;  /* E_DCH_LogicalChannelToDeleteItem */
static int hf_rnsap_eDCH_MACdFlow_Specific_Information_01 = -1;  /* EDCH_MACdFlow_Specific_InfoList */
static int hf_rnsap_EDCH_MACdFlow_Specific_InfoList_item = -1;  /* EDCH_MACdFlow_Specific_InfoItem */
static int hf_rnsap_tnlQoS = -1;                  /* TnlQos */
static int hf_rnsap_maxNr_Retransmissions_EDCH = -1;  /* MaxNr_Retransmissions_EDCH */
static int hf_rnsap_trafficClass = -1;            /* TrafficClass */
static int hf_rnsap_eDCH_Grant_Type_Information = -1;  /* E_DCH_Grant_Type_Information */
static int hf_rnsap_bundlingModeIndicator = -1;   /* BundlingModeIndicator */
static int hf_rnsap_eDCHLogicalChannelInformation = -1;  /* E_DCH_LogicalChannelInformation */
static int hf_rnsap_EDCH_MACdFlow_Specific_InfoToModifyList_item = -1;  /* EDCH_MACdFlow_Specific_InfoToModifyItem */
static int hf_rnsap_transportBearerRequestIndicator = -1;  /* TransportBearerRequestIndicator */
static int hf_rnsap_eDCH_LogicalChannelToAdd = -1;  /* E_DCH_LogicalChannelInformation */
static int hf_rnsap_eDCH_LogicalChannelToModify = -1;  /* E_DCH_LogicalChannelToModify */
static int hf_rnsap_eDCH_LogicalChannelToDelete = -1;  /* E_DCH_LogicalChannelToDelete */
static int hf_rnsap_EDCH_MACdFlows_To_Delete_item = -1;  /* EDCH_MACdFlows_To_Delete_Item */
static int hf_rnsap_maxBits_MACe_PDU_non_scheduled = -1;  /* Max_Bits_MACe_PDU_non_scheduled */
static int hf_rnsap_hARQ_Process_Allocation_NonSched_2ms = -1;  /* HARQ_Process_Allocation_2ms_EDCH */
static int hf_rnsap_e_DCH_serving_cell_outcome_choice = -1;  /* E_DCH_serving_cell_change_choice */
static int hf_rnsap_e_DCH_serving_cell_change_successful = -1;  /* E_DCH_serving_cell_change_successful */
static int hf_rnsap_e_DCH_serving_cell_change_unsuccessful = -1;  /* E_DCH_serving_cell_change_unsuccessful */
static int hf_rnsap_e_DCH_RL_InformationList_Rsp = -1;  /* E_DCH_RL_InformationList_Rsp */
static int hf_rnsap_E_DCH_RL_InformationList_Rsp_item = -1;  /* E_DCH_RL_InformationList_Rsp_Item */
static int hf_rnsap_e_DCH_reconfigured_RL_Id = -1;  /* RL_ID */
static int hf_rnsap_e_DCH_FDD_DL_Control_Channel_Info = -1;  /* EDCH_FDD_DL_ControlChannelInformation */
static int hf_rnsap_cause = -1;                   /* Cause */
static int hf_rnsap_two_ms = -1;                  /* DTX_Cycle_2ms_Items */
static int hf_rnsap_ten_ms = -1;                  /* DTX_Cycle_10ms_Items */
static int hf_rnsap_two_ms_01 = -1;               /* DTX_Cycle_2ms_to_Modify_Items */
static int hf_rnsap_ten_ms_01 = -1;               /* DTX_Cycle_10ms_to_Modify_Items */
static int hf_rnsap_maxSet_E_DPDCHs = -1;         /* Max_Set_E_DPDCHs */
static int hf_rnsap_punctureLimit = -1;           /* PunctureLimit */
static int hf_rnsap_e_TFCS_Information = -1;      /* E_TFCS_Information */
static int hf_rnsap_e_TTI = -1;                   /* E_TTI */
static int hf_rnsap_e_DPCCH_PO = -1;              /* E_DPCCH_PO */
static int hf_rnsap_e_RGCH_2_IndexStepThreshold = -1;  /* E_RGCH_2_IndexStepThreshold */
static int hf_rnsap_e_RGCH_3_IndexStepThreshold = -1;  /* E_RGCH_3_IndexStepThreshold */
static int hf_rnsap_hARQ_Info_for_E_DCH = -1;     /* HARQ_Info_for_E_DCH */
static int hf_rnsap_hSDSCH_Configured_Indicator = -1;  /* HSDSCH_Configured_Indicator */
static int hf_rnsap_e_DCH_Serving_RL_in_this_DRNS = -1;  /* EDCH_Serving_RL_in_this_DRNS */
static int hf_rnsap_e_DCH_Serving_RL_not_in_this_DRNS = -1;  /* NULL */
static int hf_rnsap_e_DCH_Serving_RL_Id = -1;     /* RL_ID */
static int hf_rnsap_common_HS_DSCH_RNTI_priorityQueueInfo_EnhancedFACH = -1;  /* PriorityQueue_InfoList_EnhancedFACH_PCH */
static int hf_rnsap_dedicated_HS_DSCH_RNTI_priorityQueueInfo_EnhancedFACH = -1;  /* PriorityQueue_InfoList_EnhancedFACH_PCH */
static int hf_rnsap_priorityQueueInfo_EnhancedPCH = -1;  /* PriorityQueue_InfoList_EnhancedFACH_PCH */
static int hf_rnsap_hSDSCH_Initial_Capacity_Allocation = -1;  /* HSDSCH_Initial_Capacity_Allocation */
static int hf_rnsap_hSDSCH_RNTI = -1;             /* HSDSCH_RNTI */
static int hf_rnsap_e_TFCI_BetaEC_Boost = -1;     /* E_TFCI_BetaEC_Boost */
static int hf_rnsap_uL_Delta_T2TP = -1;           /* UL_Delta_T2TP */
static int hf_rnsap_e_DCH_TFCI_Table_Index = -1;  /* E_DCH_TFCI_Table_Index */
static int hf_rnsap_reference_E_TFCI_Information = -1;  /* Reference_E_TFCI_Information */
static int hf_rnsap_measurementTreshold = -1;     /* MeasurementThreshold */
static int hf_rnsap_measurementHysteresisTime = -1;  /* MeasurementHysteresisTime */
static int hf_rnsap_measurementIncreaseDecreaseThreshold = -1;  /* MeasurementIncreaseDecreaseThreshold */
static int hf_rnsap_measurementChangeTime = -1;   /* MeasurementChangeTime */
static int hf_rnsap_measurementThreshold1 = -1;   /* MeasurementThreshold */
static int hf_rnsap_measurementThreshold2 = -1;   /* MeasurementThreshold */
static int hf_rnsap_reportPeriodicity = -1;       /* ReportPeriodicity */
static int hf_rnsap_measurementFluctuationRange = -1;  /* INTEGER_0_100 */
static int hf_rnsap_measurementQuantity = -1;     /* MeasurementQuantity */
static int hf_rnsap_threshold = -1;               /* INTEGER_M120_165 */
static int hf_rnsap_threshold_01 = -1;            /* INTEGER_M120_M25 */
static int hf_rnsap_e_PUCH_Information = -1;      /* E_PUCH_Information */
static int hf_rnsap_e_TFCS_Information_TDD = -1;  /* E_TFCS_Information_TDD */
static int hf_rnsap_e_DCH_MACdFlows_Information_TDD = -1;  /* E_DCH_MACdFlows_Information_TDD */
static int hf_rnsap_e_DCH_TDD_Information = -1;   /* E_DCH_TDD_Information */
static int hf_rnsap_minCR = -1;                   /* CodeRate */
static int hf_rnsap_maxCR = -1;                   /* CodeRate */
static int hf_rnsap_harqInfo = -1;                /* HARQ_Info_for_E_DCH */
static int hf_rnsap_n_E_UCCH = -1;                /* N_E_UCCH */
static int hf_rnsap_e_DCH_QPSK_RefBetaInfo = -1;  /* E_DCH_QPSK_RefBetaInfo */
static int hf_rnsap_e_DCH_sixteenQAM_RefBetaInfo = -1;  /* E_DCH_sixteenQAM_RefBetaInfo */
static int hf_rnsap_E_DCH_QPSK_RefBetaInfo_item = -1;  /* E_DCH_RefBeta_Item */
static int hf_rnsap_E_DCH_sixteenQAM_RefBetaInfo_item = -1;  /* E_DCH_RefBeta_Item */
static int hf_rnsap_refCodeRate = -1;             /* CodeRate_short */
static int hf_rnsap_refBeta = -1;                 /* RefBeta */
static int hf_rnsap_E_DCH_MACdFlows_Information_TDD_item = -1;  /* E_DCH_MACdFlow_InfoTDDItem */
static int hf_rnsap_tnlQos = -1;                  /* TnlQos */
static int hf_rnsap_eDCH_Grant_TypeTDD = -1;      /* E_DCH_Grant_TypeTDD */
static int hf_rnsap_e_DCH_TDD_Maximum_Bitrate = -1;  /* E_DCH_TDD_Maximum_Bitrate */
static int hf_rnsap_e_DCH_PowerOffset_for_SchedulingInfo = -1;  /* E_DCH_PowerOffset_for_SchedulingInfo */
static int hf_rnsap_e_DCH_MACdFlows_to_Add = -1;  /* E_DCH_MACdFlows_Information_TDD */
static int hf_rnsap_e_DCH_MACdFlows_to_Delete = -1;  /* EDCH_MACdFlows_To_Delete */
static int hf_rnsap_e_DCH_Non_Scheduled_Grant_Info = -1;  /* E_DCH_Non_Scheduled_Grant_Info */
static int hf_rnsap_e_DCH_TDD_Information_to_Modify = -1;  /* E_DCH_TDD_Information_to_Modify */
static int hf_rnsap_e_DCH_TDD_Information_to_Modify_List = -1;  /* E_DCH_TDD_Information_to_Modify_List */
static int hf_rnsap_E_DCH_TDD_Information_to_Modify_List_item = -1;  /* E_DCH_MACdFlow_ModifyTDDItem */
static int hf_rnsap_e_DCH_LogicalChannelToAdd = -1;  /* E_DCH_LogicalChannelInformation */
static int hf_rnsap_e_DCH_LogicalChannelToModify = -1;  /* E_DCH_LogicalChannelToModify */
static int hf_rnsap_e_DCH_LogicalChannelToDelete = -1;  /* E_DCH_LogicalChannelToDelete */
static int hf_rnsap_e_DCH_TDD_MACdFlow_Specific_InformationResp = -1;  /* E_DCH_TDD_MACdFlow_Specific_InformationResp */
static int hf_rnsap_e_AGCH_Specific_Information_ResponseTDD = -1;  /* E_AGCH_Specific_InformationRespListTDD */
static int hf_rnsap_e_HICH_Information_Response = -1;  /* E_HICH_InformationResp */
static int hf_rnsap_e_RNTI = -1;                  /* E_RNTI */
static int hf_rnsap_mode1 = -1;                   /* Mode1_Ind */
static int hf_rnsap_mode2 = -1;                   /* CFN */
static int hf_rnsap_mode1_01 = -1;                /* ActivationDelay */
static int hf_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp_item = -1;  /* E_DCH_TDD_MACdFlow_Specific_InformationResp_Item */
static int hf_rnsap_e_DCH_MacdFlow_Id = -1;       /* EDCH_MACdFlow_ID */
static int hf_rnsap_E_AGCH_Specific_InformationRespListTDD_item = -1;  /* E_AGCH_Specific_InformationResp_ItemTDD */
static int hf_rnsap_timeslot = -1;                /* TimeSlot */
static int hf_rnsap_tDD_ChannelisationCode = -1;  /* TDD_ChannelisationCode */
static int hf_rnsap_e_HICH_TimeOffset = -1;       /* E_HICH_TimeOffset */
static int hf_rnsap_timeslotResource = -1;        /* E_DCH_TimeslotResource */
static int hf_rnsap_powerResource = -1;           /* E_DCH_PowerResource */
static int hf_rnsap_repetitionPeriod = -1;        /* RepetitionPeriod */
static int hf_rnsap_repetitionLength = -1;        /* RepetitionLength */
static int hf_rnsap_tddE_PUCH_Offset = -1;        /* TddE_PUCH_Offset */
static int hf_rnsap_tdd_ChannelisationCode = -1;  /* TDD_ChannelisationCode */
static int hf_rnsap_e_DCH_TDD_Information768 = -1;  /* E_DCH_TDD_Information768 */
static int hf_rnsap_e_DCH_TDD_Maximum_Bitrate768 = -1;  /* E_DCH_TDD_Maximum_Bitrate768 */
static int hf_rnsap_e_DCH_Non_Scheduled_Grant_Info768 = -1;  /* E_DCH_Non_Scheduled_Grant_Info768 */
static int hf_rnsap_e_AGCH_Specific_Information_Response768TDD = -1;  /* E_AGCH_Specific_InformationRespList768TDD */
static int hf_rnsap_e_HICH_Information_Response768 = -1;  /* E_HICH_InformationResp768 */
static int hf_rnsap_E_AGCH_Specific_InformationRespList768TDD_item = -1;  /* E_AGCH_Specific_InformationResp_Item768TDD */
static int hf_rnsap_tDD_ChannelisationCode768 = -1;  /* TDD_ChannelisationCode768 */
static int hf_rnsap_tdd_ChannelisationCode768 = -1;  /* TDD_ChannelisationCode768 */
static int hf_rnsap_e_PUCH_LCR_Information = -1;  /* E_PUCH_LCR_Information */
static int hf_rnsap_e_DCH_LCR_TDD_Information = -1;  /* E_DCH_LCR_TDD_Information */
static int hf_rnsap_pRxdesBase = -1;              /* E_PUCH_PRXdesBase */
static int hf_rnsap_e_PUCH_TPC_Step_Size = -1;    /* TDD_TPC_UplinkStepSize_LCR */
static int hf_rnsap_n_E_UCCH_LCR = -1;            /* N_E_UCCH_LCR */
static int hf_rnsap_e_DCH_Physical_Layer_Category_LCR = -1;  /* E_DCH_Physical_Layer_Category_LCR */
static int hf_rnsap_e_AGCH_Specific_Information_Response_LCR_TDD = -1;  /* E_AGCH_Specific_InformationRespList_LCR_TDD */
static int hf_rnsap_e_HICH_Specific_Information_Response_LCR = -1;  /* E_HICH_Specific_InformationResp_LCR */
static int hf_rnsap_e_DCH_Non_Scheduled_Grant_Info_LCR = -1;  /* E_DCH_Non_Scheduled_Grant_Info_LCR */
static int hf_rnsap_E_AGCH_Specific_InformationRespList_LCR_TDD_item = -1;  /* E_AGCH_Specific_InformationResp_Item_LCR_TDD */
static int hf_rnsap_e_HICH_Scheduled_InformationResp_LCR = -1;  /* E_HICH_Scheduled_InformationRespList_LCR_TDD */
static int hf_rnsap_e_HICH_non_Scheduled_InformationResp_LCR = -1;  /* E_HICH_InformationResp_LCR */
static int hf_rnsap_e_HICH_TimeOffset_lcr = -1;   /* E_HICH_TimeOffset_LCR */
static int hf_rnsap_E_HICH_Scheduled_InformationRespList_LCR_TDD_item = -1;  /* E_HICH_Scheduled_InformationResp_Item_LCR_TDD */
static int hf_rnsap_e_HICH_EI = -1;               /* E_HICH_EI */
static int hf_rnsap_e_HICH_Scheduled_InformationResp_LCR_01 = -1;  /* E_HICH_InformationResp_LCR */
static int hf_rnsap_signatureSequenceGroupIndex = -1;  /* SignatureSequenceGroupIndex */
static int hf_rnsap_timeslotResource_LCR = -1;    /* E_DCH_TimeslotResource_LCR */
static int hf_rnsap_subframenumber = -1;          /* E_DCH_SubframeNumber_LCR */
static int hf_rnsap_Radio_Links_without_DPCH_FDPCH_Indication_item = -1;  /* Radio_Links_without_DPCH_FDPCH_Information */
static int hf_rnsap_radio_Links_without_DPCH_FDPCH_Operation_Indicator = -1;  /* T_radio_Links_without_DPCH_FDPCH_Operation_Indicator */
static int hf_rnsap_extended_sRNTI = -1;          /* Extended_RNTI */
static int hf_rnsap_extended_sRNTI_BitMaskIndex = -1;  /* T_extended_sRNTI_BitMaskIndex */
static int hf_rnsap_ExtendedAffectedUEInformationForMBMS_item = -1;  /* Extended_RNTI */
static int hf_rnsap_repetition_Period_List_LCR = -1;  /* Repetition_Period_List_LCR */
static int hf_rnsap_e_DCH_SPS_Indicator = -1;     /* E_DCH_SPS_Indicator */
static int hf_rnsap_e_DCH_SPS_Reservation_Indicator = -1;  /* SPS_Reservation_Indicator */
static int hf_rnsap_initial_E_DCH_SPS_resource = -1;  /* Initial_E_DCH_SPS_resource */
static int hf_rnsap_e_DCH_SPS_HICH_Information = -1;  /* E_DCH_SPS_HICH_Information */
static int hf_rnsap_timeslot_Resource_Related_Information = -1;  /* E_DCH_TimeslotResource_LCR */
static int hf_rnsap_repetitionPeriodIndex = -1;   /* RepetitionPeriodIndex */
static int hf_rnsap_subframeNumber = -1;          /* T_subframeNumber */
static int hf_rnsap_n_E_UCCHLCR = -1;             /* N_E_UCCH_LCR */
static int hf_rnsap_e_HICH_Configuration = -1;    /* E_HICH_Configuration */
static int hf_rnsap_same_As_Scheduled_E_HICH = -1;  /* Same_As_Scheduled_E_HICH */
static int hf_rnsap_explicit = -1;                /* E_HICH_InformationResp_ExplicitConfiguration_LCR */
static int hf_rnsap_FACH_FlowControlInformation_item = -1;  /* FACH_FlowControlInformationItem */
static int hf_rnsap_fACH_SchedulingPriority = -1;  /* SchedulingPriorityIndicator */
static int hf_rnsap_fACH_InitialWindowSize = -1;  /* FACH_InitialWindowSize */
static int hf_rnsap_FACH_InformationList_item = -1;  /* FACH_InformationItem */
static int hf_rnsap_FDD_DCHs_to_Modify_item = -1;  /* FDD_DCHs_to_ModifyItem */
static int hf_rnsap_dCH_SpecificInformationList_02 = -1;  /* FDD_DCHs_to_ModifySpecificInformationList */
static int hf_rnsap_FDD_DCHs_to_ModifySpecificInformationList_item = -1;  /* FDD_DCHs_to_ModifySpecificItem */
static int hf_rnsap_ul_TransportformatSet = -1;   /* TransportFormatSet */
static int hf_rnsap_dl_TransportformatSet = -1;   /* TransportFormatSet */
static int hf_rnsap_not_Used_dRACControl = -1;    /* NULL */
static int hf_rnsap_FDD_DL_CodeInformation_item = -1;  /* FDD_DL_CodeInformationItem */
static int hf_rnsap_dl_ScramblingCode = -1;       /* DL_ScramblingCode */
static int hf_rnsap_fDD_DL_ChannelisationCodeNumber = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_rnsap_transmission_Gap_Pattern_Sequence_ScramblingCode_Information = -1;  /* Transmission_Gap_Pattern_Sequence_ScramblingCode_Information */
static int hf_rnsap_fdd_dl_ChannelisationCodeNumber = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_rnsap_extended_E_DPCCH_PO = -1;     /* Extended_E_DPCCH_PO */
static int hf_rnsap_tDD_ChannelisationCodeLCR = -1;  /* TDD_ChannelisationCodeLCR */
static int hf_rnsap_wT = -1;                      /* INTEGER_1_4 */
static int hf_rnsap_fTPICH_Offset = -1;           /* FTPICH_Offset */
static int hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information = -1;  /* Setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information */
static int hf_rnsap_fTPICH_SlotFormat = -1;       /* FTPICH_SlotFormat */
static int hf_rnsap_fTPICH_ChannelisationCodenumber = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_rnsap_GA_Cell_item = -1;            /* GA_Cell_item */
static int hf_rnsap_cell_GAIgeographicalCoordinate = -1;  /* GeographicalCoordinate */
static int hf_rnsap_pointWithUncertainty = -1;    /* GA_PointWithUnCertainty */
static int hf_rnsap_pointWithUncertaintyEllipse = -1;  /* GA_PointWithUnCertaintyEllipse */
static int hf_rnsap_pointWithAltitude = -1;       /* GA_PointWithAltitude */
static int hf_rnsap_pointWithAltitudeAndUncertaintyEllipsoid = -1;  /* GA_PointWithAltitudeAndUncertaintyEllipsoid */
static int hf_rnsap_ellipsoidArc = -1;            /* GA_EllipsoidArc */
static int hf_rnsap_directionOfAltitude = -1;     /* T_directionOfAltitude */
static int hf_rnsap_altitude = -1;                /* INTEGER_0_32767 */
static int hf_rnsap_geographicalCoordinates = -1;  /* GeographicalCoordinate */
static int hf_rnsap_innerRadius = -1;             /* INTEGER_0_65535 */
static int hf_rnsap_uncertaintyRadius = -1;       /* INTEGER_0_127 */
static int hf_rnsap_offsetAngle = -1;             /* INTEGER_0_179 */
static int hf_rnsap_includedAngle = -1;           /* INTEGER_0_179 */
static int hf_rnsap_confidence = -1;              /* INTEGER_0_127 */
static int hf_rnsap_signalledGainFactors10ms = -1;  /* T_signalledGainFactors10ms */
static int hf_rnsap_gain10ms_betaC = -1;          /* BetaCD */
static int hf_rnsap_gain10ms_betaD = -1;          /* BetaCD */
static int hf_rnsap_gain10ms_refTFCNumber = -1;   /* RefTFCNumber */
static int hf_rnsap_computedGainFactors10ms = -1;  /* RefTFCNumber */
static int hf_rnsap_navClockModel = -1;           /* GANSS_NAVclockModel */
static int hf_rnsap_cnavClockModel = -1;          /* GANSS_CNAVclockModel */
static int hf_rnsap_glonassClockModel = -1;       /* GANSS_GLONASSclockModel */
static int hf_rnsap_sbasClockModel = -1;          /* GANSS_SBASclockModel */
static int hf_rnsap_extension_GANSS_AddClockModels = -1;  /* Extension_GANSS_AddClockModels */
static int hf_rnsap_navKeplerianSet = -1;         /* GANSS_NavModel_NAVKeplerianSet */
static int hf_rnsap_cnavKeplerianSet = -1;        /* GANSS_NavModel_CNAVKeplerianSet */
static int hf_rnsap_glonassECEF = -1;             /* GANSS_NavModel_GLONASSecef */
static int hf_rnsap_sbasECEF = -1;                /* GANSS_NavModel_SBASecef */
static int hf_rnsap_extension_GANSS_AddOrbitModels = -1;  /* Extension_GANSS_AddOrbitModels */
static int hf_rnsap_dataID = -1;                  /* BIT_STRING_SIZE_2 */
static int hf_rnsap_alpha_beta_parameters = -1;   /* GPS_Ionospheric_Model */
static int hf_rnsap_ganss_Transmission_Time = -1;  /* GANSS_Transmission_Time */
static int hf_rnsap_non_broadcastIndication = -1;  /* T_non_broadcastIndication */
static int hf_rnsap_ganssSatInfoNavList = -1;     /* Ganss_Sat_Info_AddNavList */
static int hf_rnsap_GANSS_Additional_Time_Models_item = -1;  /* GANSS_Time_Model */
static int hf_rnsap_utcModel1 = -1;               /* GANSS_UTCmodelSet1 */
static int hf_rnsap_utcModel2 = -1;               /* GANSS_UTCmodelSet2 */
static int hf_rnsap_utcModel3 = -1;               /* GANSS_UTCmodelSet3 */
static int hf_rnsap_extension_GANSS_Additional_UTC_Models = -1;  /* Extension_GANSS_Additional_UTC_Models */
static int hf_rnsap_ganss_wk_number = -1;         /* INTEGER_0_255 */
static int hf_rnsap_gANSS_AlmanacModel = -1;      /* T_gANSS_AlmanacModel */
static int hf_rnsap_gANSS_keplerianParameters = -1;  /* T_gANSS_keplerianParameters */
static int hf_rnsap_t_oa = -1;                    /* INTEGER_0_1023 */
static int hf_rnsap_iod_a = -1;                   /* INTEGER_0_15 */
static int hf_rnsap_gANSS_SatelliteInformationKP = -1;  /* GANSS_SatelliteInformationKP */
static int hf_rnsap_extension_GANSS_AlmanacModel = -1;  /* Extension_GANSS_AlmanacModel */
static int hf_rnsap_t_oa_01 = -1;                 /* INTEGER_0_255 */
static int hf_rnsap_sat_info_NAVkpList = -1;      /* GANSS_SAT_Info_Almanac_NAVkpList */
static int hf_rnsap_sat_info_BDSkpList = -1;      /* GANSS_SAT_Info_Almanac_BDSkpList */
static int hf_rnsap_sat_info_REDkpList = -1;      /* GANSS_SAT_Info_Almanac_REDkpList */
static int hf_rnsap_sat_info_MIDIkpList = -1;     /* GANSS_SAT_Info_Almanac_MIDIkpList */
static int hf_rnsap_sat_info_GLOkpList = -1;      /* GANSS_SAT_Info_Almanac_GLOkpList */
static int hf_rnsap_sat_info_SBASecefList = -1;   /* GANSS_SAT_Info_Almanac_SBASecefList */
static int hf_rnsap_ganssID1 = -1;                /* GANSS_AuxInfoGANSS_ID1 */
static int hf_rnsap_ganssID3 = -1;                /* GANSS_AuxInfoGANSS_ID3 */
static int hf_rnsap_GANSS_AuxInfoGANSS_ID1_item = -1;  /* GANSS_AuxInfoGANSS_ID1_item */
static int hf_rnsap_signalsAvailable = -1;        /* BIT_STRING_SIZE_8 */
static int hf_rnsap_GANSS_AuxInfoGANSS_ID3_item = -1;  /* GANSS_AuxInfoGANSS_ID3_item */
static int hf_rnsap_channelNumber = -1;           /* INTEGER_M7_13 */
static int hf_rnsap_bdsToc = -1;                  /* BIT_STRING_SIZE_17 */
static int hf_rnsap_bdsa0 = -1;                   /* BIT_STRING_SIZE_24 */
static int hf_rnsap_bdsa1 = -1;                   /* BIT_STRING_SIZE_22 */
static int hf_rnsap_bdsa2 = -1;                   /* BIT_STRING_SIZE_11 */
static int hf_rnsap_bdsTgd1 = -1;                 /* BIT_STRING_SIZE_10 */
static int hf_rnsap_bdsAODC = -1;                 /* BIT_STRING_SIZE_5 */
static int hf_rnsap_GANSS_Clock_Model_item = -1;  /* GANSS_Clock_Model_item */
static int hf_rnsap_t_oc = -1;                    /* BIT_STRING_SIZE_14 */
static int hf_rnsap_a_i2 = -1;                    /* BIT_STRING_SIZE_6 */
static int hf_rnsap_a_i1 = -1;                    /* BIT_STRING_SIZE_21 */
static int hf_rnsap_a_i0 = -1;                    /* BIT_STRING_SIZE_31 */
static int hf_rnsap_t_gd = -1;                    /* BIT_STRING_SIZE_10 */
static int hf_rnsap_sisa = -1;                    /* BIT_STRING_SIZE_8 */
static int hf_rnsap_model_id = -1;                /* INTEGER_0_1_ */
static int hf_rnsap_cnavToc = -1;                 /* BIT_STRING_SIZE_11 */
static int hf_rnsap_cnavTop = -1;                 /* BIT_STRING_SIZE_11 */
static int hf_rnsap_cnavURA0 = -1;                /* BIT_STRING_SIZE_5 */
static int hf_rnsap_cnavURA1 = -1;                /* BIT_STRING_SIZE_3 */
static int hf_rnsap_cnavURA2 = -1;                /* BIT_STRING_SIZE_3 */
static int hf_rnsap_cnavAf2 = -1;                 /* BIT_STRING_SIZE_10 */
static int hf_rnsap_cnavAf1 = -1;                 /* BIT_STRING_SIZE_20 */
static int hf_rnsap_cnavAf0 = -1;                 /* BIT_STRING_SIZE_26 */
static int hf_rnsap_cnavTgd = -1;                 /* BIT_STRING_SIZE_13 */
static int hf_rnsap_cnavISCl1cp = -1;             /* BIT_STRING_SIZE_13 */
static int hf_rnsap_cnavISCl1cd = -1;             /* BIT_STRING_SIZE_13 */
static int hf_rnsap_cnavISCl1ca = -1;             /* BIT_STRING_SIZE_13 */
static int hf_rnsap_cnavISCl2c = -1;              /* BIT_STRING_SIZE_13 */
static int hf_rnsap_cnavISCl5i5 = -1;             /* BIT_STRING_SIZE_13 */
static int hf_rnsap_cnavISCl5q5 = -1;             /* BIT_STRING_SIZE_13 */
static int hf_rnsap_ganss_Ionospheric_Model = -1;  /* GANSS_Ionospheric_Model */
static int hf_rnsap_ganss_Rx_Pos = -1;            /* GANSS_RX_Pos */
static int hf_rnsap_ionospheric_Model = -1;       /* BOOLEAN */
static int hf_rnsap_ganssTod = -1;                /* INTEGER_0_59_ */
static int hf_rnsap_dataBitAssistancelist = -1;   /* GANSS_DataBitAssistanceList */
static int hf_rnsap_GANSS_DataBitAssistanceList_item = -1;  /* GANSS_DataBitAssistanceItem */
static int hf_rnsap_dataBitAssistanceSgnList = -1;  /* GANSS_DataBitAssistanceSgnList */
static int hf_rnsap_GANSS_DataBitAssistanceSgnList_item = -1;  /* GANSS_DataBitAssistanceSgnItem */
static int hf_rnsap_ganss_SignalId = -1;          /* GANSS_Signal_ID */
static int hf_rnsap_ganssDataBits = -1;           /* BIT_STRING_SIZE_1_1024 */
static int hf_rnsap_ganssTod_01 = -1;             /* INTEGER_0_86399 */
static int hf_rnsap_ganss_Data_Bit_Assistance_ReqList = -1;  /* GANSS_Data_Bit_Assistance_ReqList */
static int hf_rnsap_ganss_DataBitInterval = -1;   /* INTEGER_0_15 */
static int hf_rnsap_ganss_SatelliteInfo = -1;     /* T_ganss_SatelliteInfo */
static int hf_rnsap_ganss_SatelliteInfo_item = -1;  /* INTEGER_0_63 */
static int hf_rnsap_b1 = -1;                      /* BIT_STRING_SIZE_11 */
static int hf_rnsap_b2 = -1;                      /* BIT_STRING_SIZE_10 */
static int hf_rnsap_teop = -1;                    /* BIT_STRING_SIZE_16 */
static int hf_rnsap_pmX = -1;                     /* BIT_STRING_SIZE_21 */
static int hf_rnsap_pmXdot = -1;                  /* BIT_STRING_SIZE_15 */
static int hf_rnsap_pmY = -1;                     /* BIT_STRING_SIZE_21 */
static int hf_rnsap_pmYdot = -1;                  /* BIT_STRING_SIZE_15 */
static int hf_rnsap_deltaUT1 = -1;                /* BIT_STRING_SIZE_31 */
static int hf_rnsap_deltaUT1dot = -1;             /* BIT_STRING_SIZE_19 */
static int hf_rnsap_GANSS_GenericDataInfoReqList_item = -1;  /* GANSS_GenericDataInfoReqItem */
static int hf_rnsap_ganss_Id = -1;                /* GANSS_ID */
static int hf_rnsap_ganss_Navigation_Model_And_Time_Recovery = -1;  /* BOOLEAN */
static int hf_rnsap_ganss_Time_Model_GNSS_GNSS = -1;  /* BIT_STRING_SIZE_9 */
static int hf_rnsap_ganss_UTC_Model = -1;         /* BOOLEAN */
static int hf_rnsap_ganss_Almanac = -1;           /* BOOLEAN */
static int hf_rnsap_ganss_Real_Time_Integrity = -1;  /* BOOLEAN */
static int hf_rnsap_ganss_Data_Bit_Assistance_Req = -1;  /* GANSS_Data_Bit_Assistance_ReqItem */
static int hf_rnsap_GANSS_Generic_Data_item = -1;  /* GANSS_Generic_DataItem */
static int hf_rnsap_dganss_Correction = -1;       /* DGANSSCorrections */
static int hf_rnsap_ganss_Navigation_Model_And_Time_Recovery_01 = -1;  /* GANSS_Navigation_Model_And_Time_Recovery */
static int hf_rnsap_ganss_Time_Model = -1;        /* GANSS_Time_Model */
static int hf_rnsap_ganss_UTC_TIME = -1;          /* GANSS_UTC_Model */
static int hf_rnsap_ganss_Almanac_01 = -1;        /* GANSS_Almanac */
static int hf_rnsap_ganss_Real_Time_Integrity_01 = -1;  /* GANSS_Real_Time_Integrity */
static int hf_rnsap_ganss_Data_Bit_Assistance = -1;  /* GANSS_Data_Bit_Assistance */
static int hf_rnsap_gloTau = -1;                  /* BIT_STRING_SIZE_22 */
static int hf_rnsap_gloGamma = -1;                /* BIT_STRING_SIZE_11 */
static int hf_rnsap_gloDeltaTau = -1;             /* BIT_STRING_SIZE_5 */
static int hf_rnsap_gANSS_CommonDataInfoReq = -1;  /* GANSS_CommonDataInfoReq */
static int hf_rnsap_gANSS_GenericDataInfoReqList = -1;  /* GANSS_GenericDataInfoReqList */
static int hf_rnsap_alpha_zero_ionos = -1;        /* BIT_STRING_SIZE_11 */
static int hf_rnsap_alpha_one_ionos = -1;         /* BIT_STRING_SIZE_11 */
static int hf_rnsap_alpha_two_ionos = -1;         /* BIT_STRING_SIZE_14 */
static int hf_rnsap_gANSS_IonosphereRegionalStormFlags = -1;  /* GANSS_IonosphereRegionalStormFlags */
static int hf_rnsap_storm_flag_one = -1;          /* BOOLEAN */
static int hf_rnsap_storm_flag_two = -1;          /* BOOLEAN */
static int hf_rnsap_storm_flag_three = -1;        /* BOOLEAN */
static int hf_rnsap_storm_flag_four = -1;         /* BOOLEAN */
static int hf_rnsap_storm_flag_five = -1;         /* BOOLEAN */
static int hf_rnsap_navToc = -1;                  /* BIT_STRING_SIZE_16 */
static int hf_rnsap_navaf2 = -1;                  /* BIT_STRING_SIZE_8 */
static int hf_rnsap_navaf1 = -1;                  /* BIT_STRING_SIZE_16 */
static int hf_rnsap_navaf0 = -1;                  /* BIT_STRING_SIZE_22 */
static int hf_rnsap_navTgd = -1;                  /* BIT_STRING_SIZE_8 */
static int hf_rnsap_non_broadcastIndication_01 = -1;  /* T_non_broadcastIndication_01 */
static int hf_rnsap_ganssSatInfoNav = -1;         /* GANSS_Sat_Info_Nav */
static int hf_rnsap_bdsURAI = -1;                 /* BIT_STRING_SIZE_4 */
static int hf_rnsap_bdsToe = -1;                  /* BIT_STRING_SIZE_17 */
static int hf_rnsap_bdsAPowerHalf = -1;           /* BIT_STRING_SIZE_32 */
static int hf_rnsap_bdsE = -1;                    /* BIT_STRING_SIZE_32 */
static int hf_rnsap_bdsW = -1;                    /* BIT_STRING_SIZE_32 */
static int hf_rnsap_bdsDeltaN = -1;               /* BIT_STRING_SIZE_16 */
static int hf_rnsap_bdsM0 = -1;                   /* BIT_STRING_SIZE_32 */
static int hf_rnsap_bdsOmega0 = -1;               /* BIT_STRING_SIZE_32 */
static int hf_rnsap_bdsOmegaDot = -1;             /* BIT_STRING_SIZE_24 */
static int hf_rnsap_bdsI0 = -1;                   /* BIT_STRING_SIZE_32 */
static int hf_rnsap_bdsIDot = -1;                 /* BIT_STRING_SIZE_14 */
static int hf_rnsap_bdsCuc = -1;                  /* BIT_STRING_SIZE_18 */
static int hf_rnsap_bdsCus = -1;                  /* BIT_STRING_SIZE_18 */
static int hf_rnsap_bdsCrc = -1;                  /* BIT_STRING_SIZE_18 */
static int hf_rnsap_bdsCrs = -1;                  /* BIT_STRING_SIZE_18 */
static int hf_rnsap_bdsCic = -1;                  /* BIT_STRING_SIZE_18 */
static int hf_rnsap_bdsCis = -1;                  /* BIT_STRING_SIZE_18 */
static int hf_rnsap_bdsAODE = -1;                 /* BIT_STRING_SIZE_5 */
static int hf_rnsap_cnavURAindex = -1;            /* BIT_STRING_SIZE_5 */
static int hf_rnsap_cnavDeltaA = -1;              /* BIT_STRING_SIZE_26 */
static int hf_rnsap_cnavAdot = -1;                /* BIT_STRING_SIZE_25 */
static int hf_rnsap_cnavDeltaNo = -1;             /* BIT_STRING_SIZE_17 */
static int hf_rnsap_cnavDeltaNoDot = -1;          /* BIT_STRING_SIZE_23 */
static int hf_rnsap_cnavMo = -1;                  /* BIT_STRING_SIZE_33 */
static int hf_rnsap_cnavE = -1;                   /* BIT_STRING_SIZE_33 */
static int hf_rnsap_cnavOmega = -1;               /* BIT_STRING_SIZE_33 */
static int hf_rnsap_cnavOMEGA0 = -1;              /* BIT_STRING_SIZE_33 */
static int hf_rnsap_cnavDeltaOmegaDot = -1;       /* BIT_STRING_SIZE_17 */
static int hf_rnsap_cnavIo = -1;                  /* BIT_STRING_SIZE_33 */
static int hf_rnsap_cnavIoDot = -1;               /* BIT_STRING_SIZE_15 */
static int hf_rnsap_cnavCis = -1;                 /* BIT_STRING_SIZE_16 */
static int hf_rnsap_cnavCic = -1;                 /* BIT_STRING_SIZE_16 */
static int hf_rnsap_cnavCrs = -1;                 /* BIT_STRING_SIZE_24 */
static int hf_rnsap_cnavCrc = -1;                 /* BIT_STRING_SIZE_24 */
static int hf_rnsap_cnavCus = -1;                 /* BIT_STRING_SIZE_21 */
static int hf_rnsap_cnavCuc = -1;                 /* BIT_STRING_SIZE_21 */
static int hf_rnsap_gloEn = -1;                   /* BIT_STRING_SIZE_5 */
static int hf_rnsap_gloP1 = -1;                   /* BIT_STRING_SIZE_2 */
static int hf_rnsap_gloP2 = -1;                   /* BIT_STRING_SIZE_1 */
static int hf_rnsap_gloM = -1;                    /* BIT_STRING_SIZE_2 */
static int hf_rnsap_gloX = -1;                    /* BIT_STRING_SIZE_27 */
static int hf_rnsap_gloXdot = -1;                 /* BIT_STRING_SIZE_24 */
static int hf_rnsap_gloXdotdot = -1;              /* BIT_STRING_SIZE_5 */
static int hf_rnsap_gloY = -1;                    /* BIT_STRING_SIZE_27 */
static int hf_rnsap_gloYdot = -1;                 /* BIT_STRING_SIZE_24 */
static int hf_rnsap_gloYdotdot = -1;              /* BIT_STRING_SIZE_5 */
static int hf_rnsap_gloZ = -1;                    /* BIT_STRING_SIZE_27 */
static int hf_rnsap_gloZdot = -1;                 /* BIT_STRING_SIZE_24 */
static int hf_rnsap_gloZdotdot = -1;              /* BIT_STRING_SIZE_5 */
static int hf_rnsap_navURA = -1;                  /* BIT_STRING_SIZE_4 */
static int hf_rnsap_navFitFlag = -1;              /* BIT_STRING_SIZE_1 */
static int hf_rnsap_navToe = -1;                  /* BIT_STRING_SIZE_16 */
static int hf_rnsap_navOmega = -1;                /* BIT_STRING_SIZE_32 */
static int hf_rnsap_navDeltaN = -1;               /* BIT_STRING_SIZE_16 */
static int hf_rnsap_navM0 = -1;                   /* BIT_STRING_SIZE_32 */
static int hf_rnsap_navOmegaADot = -1;            /* BIT_STRING_SIZE_24 */
static int hf_rnsap_navE = -1;                    /* BIT_STRING_SIZE_32 */
static int hf_rnsap_navIDot = -1;                 /* BIT_STRING_SIZE_14 */
static int hf_rnsap_navAPowerHalf = -1;           /* BIT_STRING_SIZE_32 */
static int hf_rnsap_navI0 = -1;                   /* BIT_STRING_SIZE_32 */
static int hf_rnsap_navOmegaA0 = -1;              /* BIT_STRING_SIZE_32 */
static int hf_rnsap_navCrs = -1;                  /* BIT_STRING_SIZE_16 */
static int hf_rnsap_navCis = -1;                  /* BIT_STRING_SIZE_16 */
static int hf_rnsap_navCus = -1;                  /* BIT_STRING_SIZE_16 */
static int hf_rnsap_navCrc = -1;                  /* BIT_STRING_SIZE_16 */
static int hf_rnsap_navCic = -1;                  /* BIT_STRING_SIZE_16 */
static int hf_rnsap_navCuc = -1;                  /* BIT_STRING_SIZE_16 */
static int hf_rnsap_sbasTo = -1;                  /* BIT_STRING_SIZE_13 */
static int hf_rnsap_sbasAccuracy = -1;            /* BIT_STRING_SIZE_4 */
static int hf_rnsap_sbasXg = -1;                  /* BIT_STRING_SIZE_30 */
static int hf_rnsap_sbasYg = -1;                  /* BIT_STRING_SIZE_30 */
static int hf_rnsap_sbasZg = -1;                  /* BIT_STRING_SIZE_25 */
static int hf_rnsap_sbasXgDot = -1;               /* BIT_STRING_SIZE_17 */
static int hf_rnsap_sbasYgDot = -1;               /* BIT_STRING_SIZE_17 */
static int hf_rnsap_sbasZgDot = -1;               /* BIT_STRING_SIZE_18 */
static int hf_rnsap_sbasXgDotDot = -1;            /* BIT_STRING_SIZE_10 */
static int hf_rnsap_sbagYgDotDot = -1;            /* BIT_STRING_SIZE_10 */
static int hf_rnsap_sbasZgDotDot = -1;            /* BIT_STRING_SIZE_10 */
static int hf_rnsap_gANSS_keplerianParameters_01 = -1;  /* T_gANSS_keplerianParameters_01 */
static int hf_rnsap_toe_nav = -1;                 /* BIT_STRING_SIZE_14 */
static int hf_rnsap_ganss_omega_nav = -1;         /* BIT_STRING_SIZE_32 */
static int hf_rnsap_delta_n_nav = -1;             /* BIT_STRING_SIZE_16 */
static int hf_rnsap_m_zero_nav = -1;              /* BIT_STRING_SIZE_32 */
static int hf_rnsap_omegadot_nav = -1;            /* BIT_STRING_SIZE_24 */
static int hf_rnsap_ganss_e_nav = -1;             /* BIT_STRING_SIZE_32 */
static int hf_rnsap_idot_nav = -1;                /* BIT_STRING_SIZE_14 */
static int hf_rnsap_a_sqrt_nav = -1;              /* BIT_STRING_SIZE_32 */
static int hf_rnsap_i_zero_nav = -1;              /* BIT_STRING_SIZE_32 */
static int hf_rnsap_omega_zero_nav = -1;          /* BIT_STRING_SIZE_32 */
static int hf_rnsap_c_rs_nav = -1;                /* BIT_STRING_SIZE_16 */
static int hf_rnsap_c_is_nav = -1;                /* BIT_STRING_SIZE_16 */
static int hf_rnsap_c_us_nav = -1;                /* BIT_STRING_SIZE_16 */
static int hf_rnsap_c_rc_nav = -1;                /* BIT_STRING_SIZE_16 */
static int hf_rnsap_c_ic_nav = -1;                /* BIT_STRING_SIZE_16 */
static int hf_rnsap_c_uc_nav = -1;                /* BIT_STRING_SIZE_16 */
static int hf_rnsap_GANSS_Real_Time_Integrity_item = -1;  /* GANSS_Real_Time_Integrity_item */
static int hf_rnsap_bad_ganss_satId = -1;         /* INTEGER_0_63 */
static int hf_rnsap_bad_ganss_signalId = -1;      /* BIT_STRING_SIZE_8 */
static int hf_rnsap_latitudeSign = -1;            /* T_latitudeSign */
static int hf_rnsap_degreesOfLatitude = -1;       /* INTEGER_0_2147483647 */
static int hf_rnsap_degreesOfLongitude = -1;      /* INTEGER_M2147483648_2147483647 */
static int hf_rnsap_directionOfAltitude_01 = -1;  /* T_directionOfAltitude_01 */
static int hf_rnsap_GANSS_SatelliteInformationKP_item = -1;  /* GANSS_SatelliteInformationKP_item */
static int hf_rnsap_ganss_e_alm = -1;             /* BIT_STRING_SIZE_11 */
static int hf_rnsap_ganss_delta_I_alm = -1;       /* BIT_STRING_SIZE_11 */
static int hf_rnsap_ganss_omegadot_alm = -1;      /* BIT_STRING_SIZE_11 */
static int hf_rnsap_ganss_svStatusINAV_alm = -1;  /* BIT_STRING_SIZE_4 */
static int hf_rnsap_ganss_svStatusFNAV_alm = -1;  /* BIT_STRING_SIZE_2 */
static int hf_rnsap_ganss_delta_a_sqrt_alm = -1;  /* BIT_STRING_SIZE_13 */
static int hf_rnsap_ganss_omegazero_alm = -1;     /* BIT_STRING_SIZE_16 */
static int hf_rnsap_ganss_m_zero_alm = -1;        /* BIT_STRING_SIZE_16 */
static int hf_rnsap_ganss_omega_alm = -1;         /* BIT_STRING_SIZE_16 */
static int hf_rnsap_ganss_af_zero_alm = -1;       /* BIT_STRING_SIZE_16 */
static int hf_rnsap_ganss_af_one_alm = -1;        /* BIT_STRING_SIZE_13 */
static int hf_rnsap_Ganss_Sat_Info_AddNavList_item = -1;  /* Ganss_Sat_Info_AddNavList_item */
static int hf_rnsap_svHealth = -1;                /* BIT_STRING_SIZE_9 */
static int hf_rnsap_iod = -1;                     /* BIT_STRING_SIZE_11 */
static int hf_rnsap_ganssAddClockModels = -1;     /* GANSS_AddClockModels */
static int hf_rnsap_ganssAddOrbitModels = -1;     /* GANSS_AddOrbitModels */
static int hf_rnsap_GANSS_SAT_Info_Almanac_BDSkpList_item = -1;  /* GANSS_SAT_Info_Almanac_BDS */
static int hf_rnsap_bdsAlmToa = -1;               /* BIT_STRING_SIZE_8 */
static int hf_rnsap_bdsAlmSqrtA = -1;             /* BIT_STRING_SIZE_24 */
static int hf_rnsap_bdsAlmE = -1;                 /* BIT_STRING_SIZE_17 */
static int hf_rnsap_bdsAlmW = -1;                 /* BIT_STRING_SIZE_24 */
static int hf_rnsap_bdsAlmM0 = -1;                /* BIT_STRING_SIZE_24 */
static int hf_rnsap_bdsAlmOmega0 = -1;            /* BIT_STRING_SIZE_24 */
static int hf_rnsap_bdsAlmOmegaDot = -1;          /* BIT_STRING_SIZE_17 */
static int hf_rnsap_bdsAlmDeltaI = -1;            /* BIT_STRING_SIZE_16 */
static int hf_rnsap_bdsAlmA0 = -1;                /* BIT_STRING_SIZE_11 */
static int hf_rnsap_bdsAlmA1 = -1;                /* BIT_STRING_SIZE_11 */
static int hf_rnsap_bdsSvHealth = -1;             /* BIT_STRING_SIZE_9 */
static int hf_rnsap_GANSS_SAT_Info_Almanac_GLOkpList_item = -1;  /* GANSS_SAT_Info_Almanac_GLOkpList_item */
static int hf_rnsap_gloAlmNA = -1;                /* BIT_STRING_SIZE_11 */
static int hf_rnsap_gloAlmnA = -1;                /* BIT_STRING_SIZE_5 */
static int hf_rnsap_gloAlmHA = -1;                /* BIT_STRING_SIZE_5 */
static int hf_rnsap_gloAlmLambdaA = -1;           /* BIT_STRING_SIZE_21 */
static int hf_rnsap_gloAlmTlambdaA = -1;          /* BIT_STRING_SIZE_21 */
static int hf_rnsap_gloAlmDeltaIA = -1;           /* BIT_STRING_SIZE_18 */
static int hf_rnsap_gloAkmDeltaTA = -1;           /* BIT_STRING_SIZE_22 */
static int hf_rnsap_gloAlmDeltaTdotA = -1;        /* BIT_STRING_SIZE_7 */
static int hf_rnsap_gloAlmEpsilonA = -1;          /* BIT_STRING_SIZE_15 */
static int hf_rnsap_gloAlmOmegaA = -1;            /* BIT_STRING_SIZE_16 */
static int hf_rnsap_gloAlmTauA = -1;              /* BIT_STRING_SIZE_10 */
static int hf_rnsap_gloAlmCA = -1;                /* BIT_STRING_SIZE_1 */
static int hf_rnsap_gloAlmMA = -1;                /* BIT_STRING_SIZE_2 */
static int hf_rnsap_GANSS_SAT_Info_Almanac_MIDIkpList_item = -1;  /* GANSS_SAT_Info_Almanac_MIDIkpList_item */
static int hf_rnsap_midiAlmE = -1;                /* BIT_STRING_SIZE_11 */
static int hf_rnsap_midiAlmDeltaI = -1;           /* BIT_STRING_SIZE_11 */
static int hf_rnsap_midiAlmOmegaDot = -1;         /* BIT_STRING_SIZE_11 */
static int hf_rnsap_midiAlmSqrtA = -1;            /* BIT_STRING_SIZE_17 */
static int hf_rnsap_midiAlmOmega0 = -1;           /* BIT_STRING_SIZE_16 */
static int hf_rnsap_midiAlmOmega = -1;            /* BIT_STRING_SIZE_16 */
static int hf_rnsap_midiAlmMo = -1;               /* BIT_STRING_SIZE_16 */
static int hf_rnsap_midiAlmaf0 = -1;              /* BIT_STRING_SIZE_11 */
static int hf_rnsap_midiAlmaf1 = -1;              /* BIT_STRING_SIZE_10 */
static int hf_rnsap_midiAlmL1Health = -1;         /* BIT_STRING_SIZE_1 */
static int hf_rnsap_midiAlmL2Health = -1;         /* BIT_STRING_SIZE_1 */
static int hf_rnsap_midiAlmL5Health = -1;         /* BIT_STRING_SIZE_1 */
static int hf_rnsap_GANSS_SAT_Info_Almanac_NAVkpList_item = -1;  /* GANSS_SAT_Info_Almanac_NAVkpList_item */
static int hf_rnsap_navAlmE = -1;                 /* BIT_STRING_SIZE_16 */
static int hf_rnsap_navAlmDeltaI = -1;            /* BIT_STRING_SIZE_16 */
static int hf_rnsap_navAlmOMEGADOT = -1;          /* BIT_STRING_SIZE_16 */
static int hf_rnsap_navAlmSVHealth = -1;          /* BIT_STRING_SIZE_8 */
static int hf_rnsap_navAlmSqrtA = -1;             /* BIT_STRING_SIZE_24 */
static int hf_rnsap_navAlmOMEGAo = -1;            /* BIT_STRING_SIZE_24 */
static int hf_rnsap_navAlmOmega = -1;             /* BIT_STRING_SIZE_24 */
static int hf_rnsap_navAlmMo = -1;                /* BIT_STRING_SIZE_24 */
static int hf_rnsap_navAlmaf0 = -1;               /* BIT_STRING_SIZE_11 */
static int hf_rnsap_navAlmaf1 = -1;               /* BIT_STRING_SIZE_11 */
static int hf_rnsap_GANSS_SAT_Info_Almanac_REDkpList_item = -1;  /* GANSS_SAT_Info_Almanac_REDkpList_item */
static int hf_rnsap_redAlmDeltaA = -1;            /* BIT_STRING_SIZE_8 */
static int hf_rnsap_redAlmOmega0 = -1;            /* BIT_STRING_SIZE_7 */
static int hf_rnsap_redAlmPhi0 = -1;              /* BIT_STRING_SIZE_7 */
static int hf_rnsap_redAlmL1Health = -1;          /* BIT_STRING_SIZE_1 */
static int hf_rnsap_redAlmL2Health = -1;          /* BIT_STRING_SIZE_1 */
static int hf_rnsap_redAlmL5Health = -1;          /* BIT_STRING_SIZE_1 */
static int hf_rnsap_GANSS_SAT_Info_Almanac_SBASecefList_item = -1;  /* GANSS_SAT_Info_Almanac_SBASecefList_item */
static int hf_rnsap_sbasAlmDataID = -1;           /* BIT_STRING_SIZE_2 */
static int hf_rnsap_sbasAlmHealth = -1;           /* BIT_STRING_SIZE_8 */
static int hf_rnsap_sbasAlmXg = -1;               /* BIT_STRING_SIZE_15 */
static int hf_rnsap_sbasAlmYg = -1;               /* BIT_STRING_SIZE_15 */
static int hf_rnsap_sbasAlmZg = -1;               /* BIT_STRING_SIZE_9 */
static int hf_rnsap_sbasAlmXgdot = -1;            /* BIT_STRING_SIZE_3 */
static int hf_rnsap_sbasAlmYgDot = -1;            /* BIT_STRING_SIZE_3 */
static int hf_rnsap_sbasAlmZgDot = -1;            /* BIT_STRING_SIZE_4 */
static int hf_rnsap_sbasAlmTo = -1;               /* BIT_STRING_SIZE_11 */
static int hf_rnsap_GANSS_Sat_Info_Nav_item = -1;  /* GANSS_Sat_Info_Nav_item */
static int hf_rnsap_svHealth_01 = -1;             /* BIT_STRING_SIZE_5 */
static int hf_rnsap_iod_01 = -1;                  /* BIT_STRING_SIZE_10 */
static int hf_rnsap_ganssClockModel = -1;         /* GANSS_Clock_Model */
static int hf_rnsap_ganssOrbitModel = -1;         /* GANSS_Orbit_Model */
static int hf_rnsap_sbasAgfo = -1;                /* BIT_STRING_SIZE_12 */
static int hf_rnsap_sbasAgf1 = -1;                /* BIT_STRING_SIZE_8 */
static int hf_rnsap_ganss_time_model_Ref_Time = -1;  /* INTEGER_0_37799 */
static int hf_rnsap_ganss_t_a0 = -1;              /* INTEGER_M2147483648_2147483647 */
static int hf_rnsap_ganss_t_a1 = -1;              /* INTEGER_M8388608_8388607 */
static int hf_rnsap_ganss_t_a2 = -1;              /* INTEGER_M64_63 */
static int hf_rnsap_gnss_to_id = -1;              /* T_gnss_to_id */
static int hf_rnsap_ganss_wk_number_01 = -1;      /* INTEGER_0_8191 */
static int hf_rnsap_ganssDay = -1;                /* INTEGER_0_8191 */
static int hf_rnsap_a_one_utc = -1;               /* BIT_STRING_SIZE_24 */
static int hf_rnsap_a_zero_utc = -1;              /* BIT_STRING_SIZE_32 */
static int hf_rnsap_t_ot_utc = -1;                /* BIT_STRING_SIZE_8 */
static int hf_rnsap_w_n_t_utc = -1;               /* BIT_STRING_SIZE_8 */
static int hf_rnsap_delta_t_ls_utc = -1;          /* BIT_STRING_SIZE_8 */
static int hf_rnsap_w_n_lsf_utc = -1;             /* BIT_STRING_SIZE_8 */
static int hf_rnsap_dn_utc = -1;                  /* BIT_STRING_SIZE_8 */
static int hf_rnsap_delta_t_lsf_utc = -1;         /* BIT_STRING_SIZE_8 */
static int hf_rnsap_utcA0 = -1;                   /* BIT_STRING_SIZE_16 */
static int hf_rnsap_utcA1 = -1;                   /* BIT_STRING_SIZE_13 */
static int hf_rnsap_utcA2 = -1;                   /* BIT_STRING_SIZE_7 */
static int hf_rnsap_utcDeltaTls = -1;             /* BIT_STRING_SIZE_8 */
static int hf_rnsap_utcTot = -1;                  /* BIT_STRING_SIZE_16 */
static int hf_rnsap_utcWNot = -1;                 /* BIT_STRING_SIZE_13 */
static int hf_rnsap_utcWNlsf = -1;                /* BIT_STRING_SIZE_8 */
static int hf_rnsap_utcDN = -1;                   /* BIT_STRING_SIZE_4 */
static int hf_rnsap_utcDeltaTlsf = -1;            /* BIT_STRING_SIZE_8 */
static int hf_rnsap_nA = -1;                      /* BIT_STRING_SIZE_11 */
static int hf_rnsap_tauC = -1;                    /* BIT_STRING_SIZE_32 */
static int hf_rnsap_deltaUT1_01 = -1;             /* GANSS_DeltaUT1 */
static int hf_rnsap_kp = -1;                      /* BIT_STRING_SIZE_2 */
static int hf_rnsap_utcA1wnt = -1;                /* BIT_STRING_SIZE_24 */
static int hf_rnsap_utcA0wnt = -1;                /* BIT_STRING_SIZE_32 */
static int hf_rnsap_utcTot_01 = -1;               /* BIT_STRING_SIZE_8 */
static int hf_rnsap_utcWNt = -1;                  /* BIT_STRING_SIZE_8 */
static int hf_rnsap_utcDN_01 = -1;                /* BIT_STRING_SIZE_8 */
static int hf_rnsap_utcStandardID = -1;           /* BIT_STRING_SIZE_3 */
static int hf_rnsap_utcA0_01 = -1;                /* BIT_STRING_SIZE_32 */
static int hf_rnsap_utcA1_01 = -1;                /* BIT_STRING_SIZE_24 */
static int hf_rnsap_altitudeAndDirection = -1;    /* GA_AltitudeAndDirection */
static int hf_rnsap_uncertaintyEllipse = -1;      /* GA_UncertaintyEllipse */
static int hf_rnsap_uncertaintyAltitude = -1;     /* INTEGER_0_127 */
static int hf_rnsap_uncertaintySemi_major = -1;   /* INTEGER_0_127 */
static int hf_rnsap_uncertaintySemi_minor = -1;   /* INTEGER_0_127 */
static int hf_rnsap_orientationOfMajorAxis = -1;  /* INTEGER_0_179 */
static int hf_rnsap_uncertaintyCode = -1;         /* INTEGER_0_127 */
static int hf_rnsap_geographicalCoordinate = -1;  /* GeographicalCoordinate */
static int hf_rnsap_latitudeSign_01 = -1;         /* T_latitudeSign_01 */
static int hf_rnsap_latitude = -1;                /* INTEGER_0_8388607 */
static int hf_rnsap_longitude = -1;               /* INTEGER_M8388608_8388607 */
static int hf_rnsap_sI = -1;                      /* GERAN_SystemInfo */
static int hf_rnsap_pSI = -1;                     /* GERAN_SystemInfo */
static int hf_rnsap_GERAN_SystemInfo_item = -1;   /* GERAN_SystemInfo_item */
static int hf_rnsap_gERAN_SI_block = -1;          /* OCTET_STRING_SIZE_1_23 */
static int hf_rnsap_wna_alm = -1;                 /* BIT_STRING_SIZE_8 */
static int hf_rnsap_satellite_Almanac_Information = -1;  /* T_satellite_Almanac_Information */
static int hf_rnsap_satellite_Almanac_Information_item = -1;  /* T_satellite_Almanac_Information_item */
static int hf_rnsap_dATA_ID = -1;                 /* DATA_ID */
static int hf_rnsap_gps_e_alm = -1;               /* BIT_STRING_SIZE_16 */
static int hf_rnsap_gps_toa_alm = -1;             /* BIT_STRING_SIZE_8 */
static int hf_rnsap_gps_delta_I_alm = -1;         /* BIT_STRING_SIZE_16 */
static int hf_rnsap_omegadot_alm = -1;            /* BIT_STRING_SIZE_16 */
static int hf_rnsap_svhealth_alm = -1;            /* BIT_STRING_SIZE_8 */
static int hf_rnsap_gps_a_sqrt_alm = -1;          /* BIT_STRING_SIZE_24 */
static int hf_rnsap_omegazero_alm = -1;           /* BIT_STRING_SIZE_24 */
static int hf_rnsap_m_zero_alm = -1;              /* BIT_STRING_SIZE_24 */
static int hf_rnsap_gps_omega_alm = -1;           /* BIT_STRING_SIZE_24 */
static int hf_rnsap_gps_af_zero_alm = -1;         /* BIT_STRING_SIZE_11 */
static int hf_rnsap_gps_af_one_alm = -1;          /* BIT_STRING_SIZE_11 */
static int hf_rnsap_sVGlobalHealth_alm = -1;      /* BIT_STRING_SIZE_364 */
static int hf_rnsap_Satellite_Almanac_Information_ExtItem_item = -1;  /* Satellite_Almanac_Information_ExtItem_item */
static int hf_rnsap_GPSInformation_item = -1;     /* GPSInformation_item */
static int hf_rnsap_gPSInformationItem = -1;      /* T_gPSInformationItem */
static int hf_rnsap_alpha_zero_ionos_01 = -1;     /* BIT_STRING_SIZE_8 */
static int hf_rnsap_alpha_one_ionos_01 = -1;      /* BIT_STRING_SIZE_8 */
static int hf_rnsap_alpha_two_ionos_01 = -1;      /* BIT_STRING_SIZE_8 */
static int hf_rnsap_alpha_three_ionos = -1;       /* BIT_STRING_SIZE_8 */
static int hf_rnsap_beta_zero_ionos = -1;         /* BIT_STRING_SIZE_8 */
static int hf_rnsap_beta_one_ionos = -1;          /* BIT_STRING_SIZE_8 */
static int hf_rnsap_beta_two_ionos = -1;          /* BIT_STRING_SIZE_8 */
static int hf_rnsap_beta_three_ionos = -1;        /* BIT_STRING_SIZE_8 */
static int hf_rnsap_GPS_NavigationModel_and_TimeRecovery_item = -1;  /* GPS_NavigationModel_and_TimeRecovery_item */
static int hf_rnsap_tx_tow_nav = -1;              /* INTEGER_0_1048575 */
static int hf_rnsap_tlm_message_nav = -1;         /* BIT_STRING_SIZE_14 */
static int hf_rnsap_tlm_revd_c_nav = -1;          /* BIT_STRING_SIZE_2 */
static int hf_rnsap_ho_word_nav = -1;             /* BIT_STRING_SIZE_22 */
static int hf_rnsap_w_n_nav = -1;                 /* BIT_STRING_SIZE_10 */
static int hf_rnsap_ca_or_p_on_l2_nav = -1;       /* BIT_STRING_SIZE_2 */
static int hf_rnsap_user_range_accuracy_index_nav = -1;  /* BIT_STRING_SIZE_4 */
static int hf_rnsap_sv_health_nav = -1;           /* BIT_STRING_SIZE_6 */
static int hf_rnsap_iodc_nav = -1;                /* BIT_STRING_SIZE_10 */
static int hf_rnsap_l2_p_dataflag_nav = -1;       /* BIT_STRING_SIZE_1 */
static int hf_rnsap_sf1_reserved_nav = -1;        /* BIT_STRING_SIZE_87 */
static int hf_rnsap_t_gd_nav = -1;                /* BIT_STRING_SIZE_8 */
static int hf_rnsap_t_oc_nav = -1;                /* BIT_STRING_SIZE_16 */
static int hf_rnsap_a_f_2_nav = -1;               /* BIT_STRING_SIZE_8 */
static int hf_rnsap_a_f_1_nav = -1;               /* BIT_STRING_SIZE_16 */
static int hf_rnsap_a_f_zero_nav = -1;            /* BIT_STRING_SIZE_22 */
static int hf_rnsap_gps_e_nav = -1;               /* BIT_STRING_SIZE_32 */
static int hf_rnsap_t_oe_nav = -1;                /* BIT_STRING_SIZE_16 */
static int hf_rnsap_fit_interval_flag_nav = -1;   /* BIT_STRING_SIZE_1 */
static int hf_rnsap_aodo_nav = -1;                /* BIT_STRING_SIZE_5 */
static int hf_rnsap_gps_omega_nav = -1;           /* BIT_STRING_SIZE_32 */
static int hf_rnsap_spare_zero_fill = -1;         /* BIT_STRING_SIZE_20 */
static int hf_rnsap_badSatellites = -1;           /* BadSatellites */
static int hf_rnsap_noBadSatellite = -1;          /* NULL */
static int hf_rnsap_guaranteed_UL_Rate = -1;      /* Guaranteed_Rate */
static int hf_rnsap_guaranteed_DL_Rate = -1;      /* Guaranteed_Rate */
static int hf_rnsap_implicit = -1;                /* HARQ_MemoryPartitioning_Implicit */
static int hf_rnsap_explicit_01 = -1;             /* HARQ_MemoryPartitioning_Explicit */
static int hf_rnsap_number_of_Processes = -1;     /* INTEGER_1_8_ */
static int hf_rnsap_hARQ_MemoryPartitioningList = -1;  /* HARQ_MemoryPartitioningList */
static int hf_rnsap_HARQ_MemoryPartitioningList_item = -1;  /* HARQ_MemoryPartitioningItem */
static int hf_rnsap_HARQ_MemoryPartitioningInfoExtForMIMO_item = -1;  /* HARQ_MemoryPartitioningItem */
static int hf_rnsap_process_Memory_Size = -1;     /* T_process_Memory_Size */
static int hf_rnsap_hSDSCH_MACdFlows_Information = -1;  /* HSDSCH_MACdFlows_Information */
static int hf_rnsap_uE_Capabilities_Info = -1;    /* UE_Capabilities_Info */
static int hf_rnsap_mAChs_Reordering_Buffer_Size_for_RLC_UM = -1;  /* MAChsReorderingBufferSize_for_RLC_UM */
static int hf_rnsap_cqiFeedback_CycleK = -1;      /* CQI_Feedback_Cycle */
static int hf_rnsap_cqiRepetitionFactor = -1;     /* CQI_RepetitionFactor */
static int hf_rnsap_ackNackRepetitionFactor = -1;  /* AckNack_RepetitionFactor */
static int hf_rnsap_cqiPowerOffset = -1;          /* CQI_Power_Offset */
static int hf_rnsap_ackPowerOffset = -1;          /* Ack_Power_Offset */
static int hf_rnsap_nackPowerOffset = -1;         /* Nack_Power_Offset */
static int hf_rnsap_hsscch_PowerOffset = -1;      /* HSSCCH_PowerOffset */
static int hf_rnsap_hSDSCH_MACdFlow_Specific_InfoList_Response = -1;  /* HSDSCH_MACdFlow_Specific_InfoList_Response */
static int hf_rnsap_hSSCCH_Specific_InfoList_Response = -1;  /* HSSCCH_FDD_Specific_InfoList_Response */
static int hf_rnsap_hSPDSCH_and_HSSCCH_ScramblingCode = -1;  /* DL_ScramblingCode */
static int hf_rnsap_measurement_Power_Offset = -1;  /* Measurement_Power_Offset */
static int hf_rnsap_hARQ_MemoryPartitioning = -1;  /* HARQ_MemoryPartitioning */
static int hf_rnsap_sixtyfourQAM_UsageAllowedIndicator = -1;  /* SixtyfourQAM_UsageAllowedIndicator */
static int hf_rnsap_sixtyfourQAM_DL_UsageIndicator = -1;  /* SixtyfourQAM_DL_UsageIndicator */
static int hf_rnsap_hSDSCH_TBSizeTableIndicator = -1;  /* HSDSCH_TBSizeTableIndicator */
static int hf_rnsap_hSSCCH_CodeChangeGrant = -1;  /* HSSCCH_Code_Change_Grant */
static int hf_rnsap_hsSCCHCodeChangeIndicator = -1;  /* HSSCCH_CodeChangeIndicator */
static int hf_rnsap_hS_PDSCH_Code_Change_Indicator = -1;  /* HS_PDSCH_Code_Change_Indicator */
static int hf_rnsap_hS_DSCH_Secondary_Serving_cell_choice = -1;  /* HS_DSCH_Secondary_Serving_cell_change_choice */
static int hf_rnsap_hS_Secondary_Serving_cell_change_successful = -1;  /* HS_Secondary_Serving_cell_change_successful */
static int hf_rnsap_hS_Secondary_Serving_cell_change_unsuccessful = -1;  /* HS_Secondary_Serving_cell_change_unsuccessful */
static int hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information_Response = -1;  /* HS_DSCH_FDD_Secondary_Serving_Information_Response */
static int hf_rnsap_hSDSCH_MACdFlow_Specific_InfoList_to_Modify = -1;  /* HSDSCH_MACdFlow_Specific_InfoList_to_Modify */
static int hf_rnsap_priorityQueue_Info_to_Modify = -1;  /* PriorityQueue_InfoList_to_Modify */
static int hf_rnsap_tDDAckNackPowerOffset = -1;   /* TDD_AckNack_Power_Offset */
static int hf_rnsap_priorityQueueInfotoModifyUnsynchronised = -1;  /* PriorityQueue_InfoList_to_Modify_Unsynchronised */
static int hf_rnsap_HSDSCH_MACdFlow_Specific_InfoList_item = -1;  /* HSDSCH_MACdFlow_Specific_InfoItem */
static int hf_rnsap_hSDSCH_MACdFlow_ID = -1;      /* HSDSCH_MACdFlow_ID */
static int hf_rnsap_HSDSCH_MACdFlow_Specific_InfoList_Response_item = -1;  /* HSDSCH_MACdFlow_Specific_InfoItem_Response */
static int hf_rnsap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify_item = -1;  /* HSDSCH_MACdFlow_Specific_InfoItem_to_Modify */
static int hf_rnsap_hSDSCH_MACdFlow_Specific_Info = -1;  /* HSDSCH_MACdFlow_Specific_InfoList */
static int hf_rnsap_priorityQueue_Info = -1;      /* PriorityQueue_InfoList */
static int hf_rnsap_HSDSCH_MACdFlows_to_Delete_item = -1;  /* HSDSCH_MACdFlows_to_Delete_Item */
static int hf_rnsap_hsDSCH_MACdFlow_ID = -1;      /* HSDSCH_MACdFlow_ID */
static int hf_rnsap_HSDSCH_Initial_Capacity_Allocation_item = -1;  /* HSDSCH_Initial_Capacity_AllocationItem */
static int hf_rnsap_maximum_MACdPDU_Size = -1;    /* MACdPDU_Size */
static int hf_rnsap_hSDSCH_InitialWindowSize = -1;  /* HSDSCH_InitialWindowSize */
static int hf_rnsap_setsOfHS_SCCH_Codes = -1;     /* SetsOfHS_SCCH_Codes */
static int hf_rnsap_hARQ_Preamble_Mode_Activation_Indicator = -1;  /* HARQ_Preamble_Mode_Activation_Indicator */
static int hf_rnsap_mIMO_N_M_Ratio = -1;          /* MIMO_InformationResponse */
static int hf_rnsap_continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response = -1;  /* Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response */
static int hf_rnsap_Additional_EDCH_Preconfiguration_Information_item = -1;  /* Additional_EDCH_Preconfiguration_Information_ItemIEs */
static int hf_rnsap_mAChsResetScheme = -1;        /* MAChsResetScheme */
static int hf_rnsap_hSDSCH_Physical_Layer_Category = -1;  /* INTEGER_1_64_ */
static int hf_rnsap_secondaryServingCells = -1;   /* SecondaryServingCells */
static int hf_rnsap_numPrimaryHS_SCCH_Codes = -1;  /* NumHS_SCCH_Codes */
static int hf_rnsap_hARQ_Preamble_Mode = -1;      /* HARQ_Preamble_Mode */
static int hf_rnsap_mIMO_ActivationIndicator = -1;  /* MIMO_ActivationIndicator */
static int hf_rnsap_hSDSCH_MACdPDUSizeFormat = -1;  /* HSDSCH_MACdPDUSizeFormat */
static int hf_rnsap_uE_with_enhanced_HS_SCCH_support_indicator = -1;  /* NULL */
static int hf_rnsap_HS_SCCH_PreconfiguredCodes_item = -1;  /* HS_SCCH_PreconfiguredCodesItem */
static int hf_rnsap_hS_SCCH_CodeNumber = -1;      /* HS_SCCH_CodeNumber */
static int hf_rnsap_hS_PDSCH_RLID = -1;           /* RL_ID */
static int hf_rnsap_hSDSCH_FDD_Information = -1;  /* HSDSCH_FDD_Information */
static int hf_rnsap_hS_DSCH_serving_cell_choice = -1;  /* HS_DSCH_serving_cell_change_choice */
static int hf_rnsap_hS_serving_cell_change_successful = -1;  /* HS_serving_cell_change_successful */
static int hf_rnsap_hS_serving_cell_change_unsuccessful = -1;  /* HS_serving_cell_change_unsuccessful */
static int hf_rnsap_hSDSCH_FDD_Information_Response = -1;  /* HSDSCH_FDD_Information_Response */
static int hf_rnsap_tDD_AckNack_Power_Offset = -1;  /* TDD_AckNack_Power_Offset */
static int hf_rnsap_hSSCCH_TDD_Specific_InfoList_Response = -1;  /* HSSCCH_TDD_Specific_InfoList_Response */
static int hf_rnsap_hSSCCH_TDD_Specific_InfoList_Response_LCR = -1;  /* HSSCCH_TDD_Specific_InfoList_Response_LCR */
static int hf_rnsap_hSPDSCH_TDD_Specific_InfoList_Response = -1;  /* HSPDSCH_TDD_Specific_InfoList_Response */
static int hf_rnsap_hSPDSCH_TDD_Specific_InfoList_Response_LCR = -1;  /* HSPDSCH_TDD_Specific_InfoList_Response_LCR */
static int hf_rnsap_HSPDSCH_TDD_Specific_InfoList_Response_item = -1;  /* HSPDSCH_TDD_Specific_InfoItem_Response */
static int hf_rnsap_HSPDSCH_TDD_Specific_InfoList_Response_LCR_item = -1;  /* HSPDSCH_TDD_Specific_InfoItem_Response_LCR */
static int hf_rnsap_timeslotLCR = -1;             /* TimeSlotLCR */
static int hf_rnsap_HSPDSCH_TDD_Specific_InfoList_Response768_item = -1;  /* HSPDSCH_TDD_Specific_InfoItem_Response768 */
static int hf_rnsap_HSSCCH_FDD_Specific_InfoList_Response_item = -1;  /* HSSCCH_FDD_Specific_InfoItem_Response */
static int hf_rnsap_code_Number = -1;             /* INTEGER_0_127 */
static int hf_rnsap_HSSCCH_TDD_Specific_InfoList_Response_item = -1;  /* HSSCCH_TDD_Specific_InfoItem_Response */
static int hf_rnsap_hSSICH_Info = -1;             /* HSSICH_Info */
static int hf_rnsap_HSSCCH_TDD_Specific_InfoList_Response_LCR_item = -1;  /* HSSCCH_TDD_Specific_InfoItem_Response_LCR */
static int hf_rnsap_first_TDD_ChannelisationCode = -1;  /* TDD_ChannelisationCode */
static int hf_rnsap_second_TDD_ChannelisationCode = -1;  /* TDD_ChannelisationCode */
static int hf_rnsap_hSSICH_InfoLCR = -1;          /* HSSICH_InfoLCR */
static int hf_rnsap_HSSCCH_TDD_Specific_InfoList_Response768_item = -1;  /* HSSCCH_TDD_Specific_InfoItem_Response768 */
static int hf_rnsap_hSSICH_Info768 = -1;          /* HSSICH_Info768 */
static int hf_rnsap_hsSICH_ID = -1;               /* HS_SICH_ID */
static int hf_rnsap_failed_HS_SICH = -1;          /* HS_SICH_failed */
static int hf_rnsap_missed_HS_SICH = -1;          /* HS_SICH_missed */
static int hf_rnsap_total_HS_SICH = -1;           /* HS_SICH_total */
static int hf_rnsap_MIMO_ReferenceSignal_InformationListLCR_item = -1;  /* HSSICH_ReferenceSignal_InformationLCR */
static int hf_rnsap_midambleConfigurationLCR = -1;  /* MidambleConfigurationLCR */
static int hf_rnsap_midambleShift = -1;           /* INTEGER_0_15 */
static int hf_rnsap_transport_Block_Size_List = -1;  /* Transport_Block_Size_List_LCR */
static int hf_rnsap_hS_DSCH_SPS_Reservation_Indicator = -1;  /* SPS_Reservation_Indicator */
static int hf_rnsap_hS_DSCH_SPS_Operation_Indicator = -1;  /* HS_DSCH_SPS_Operation_Indicator */
static int hf_rnsap_Transport_Block_Size_List_LCR_item = -1;  /* Transport_Block_Size_Item_LCR */
static int hf_rnsap_transport_Block_Size_maping_Index_LCR = -1;  /* Transport_Block_Size_maping_Index_LCR */
static int hf_rnsap_transport_Block_Size_Index_LCR = -1;  /* Transport_Block_Size_Index_LCR */
static int hf_rnsap_Repetition_Period_List_LCR_item = -1;  /* Repetition_Period_Item_LCR */
static int hf_rnsap_logicalChannellevel = -1;     /* LogicalChannellevel */
static int hf_rnsap_priorityQueuelevel = -1;      /* PriorityQueuelevel */
static int hf_rnsap_hS_SICH_InformationList_for_HS_DSCH_SPS = -1;  /* HS_SICH_InformationList_for_HS_DSCH_SPS */
static int hf_rnsap_initial_HS_PDSCH_SPS_Resource = -1;  /* Initial_HS_PDSCH_SPS_Resource */
static int hf_rnsap_buffer_Size_for_HS_DSCH_SPS = -1;  /* Process_Memory_Size */
static int hf_rnsap_number_of_Processes_for_HS_DSCH_SPS = -1;  /* Number_of_Processes_for_HS_DSCH_SPS */
static int hf_rnsap_HS_SICH_InformationList_for_HS_DSCH_SPS_item = -1;  /* HS_SICH_InformationItem_for_HS_DSCH_SPS */
static int hf_rnsap_hS_SICH_Mapping_Index = -1;   /* HS_SICH_Mapping_Index */
static int hf_rnsap_hS_SICH_Type = -1;            /* HS_SICH_Type */
static int hf_rnsap_hS_SCCH_Associated_HS_SICH = -1;  /* HS_SCCH_Associated_HS_SICH */
static int hf_rnsap_non_HS_SCCH_Associated_HS_SICH = -1;  /* Non_HS_SCCH_Associated_HS_SICH */
static int hf_rnsap_extended_HS_SICH_ID = -1;     /* HS_SICH_ID_Extension */
static int hf_rnsap_non_HS_SCCH_Aassociated_HS_SICH_ID = -1;  /* Non_HS_SCCH_Aassociated_HS_SICH_ID */
static int hf_rnsap_hS_PDSCH_Offset = -1;         /* TDD_PhysicalChannelOffset */
static int hf_rnsap_hS_PDSCH_Midamble_Configuation = -1;  /* MidambleShiftLCR */
static int hf_rnsap_timeslot_Resource_Related_Information_01 = -1;  /* HS_DSCH_TimeslotResourceLCR */
static int hf_rnsap_startCode = -1;               /* TDD_ChannelisationCode */
static int hf_rnsap_endCode = -1;                 /* TDD_ChannelisationCode */
static int hf_rnsap_transport_Block_Size_Index_01 = -1;  /* Transport_Block_Size_Index_LCR */
static int hf_rnsap_modulationType = -1;          /* ModulationSPS_LCR */
static int hf_rnsap_requestedDataValue = -1;      /* RequestedDataValue */
static int hf_rnsap_onDemand = -1;                /* NULL */
static int hf_rnsap_periodic = -1;                /* PeriodicInformation */
static int hf_rnsap_onModification = -1;          /* OnModificationInformation */
static int hf_rnsap_min = -1;                     /* INTEGER_1_60_ */
static int hf_rnsap_hour = -1;                    /* INTEGER_1_24_ */
static int hf_rnsap_dGPSThreshold = -1;           /* DGPSThreshold */
static int hf_rnsap_dGANSSThreshold = -1;         /* DGANSSThreshold */
static int hf_rnsap_informationTypeItem = -1;     /* T_informationTypeItem */
static int hf_rnsap_gPSInformation = -1;          /* GPSInformation */
static int hf_rnsap_iPDL_FDD_Parameters = -1;     /* IPDL_FDD_Parameters */
static int hf_rnsap_iPDL_TDD_Parameters = -1;     /* IPDL_TDD_Parameters */
static int hf_rnsap_extension_IPDLParameters = -1;  /* Extension_IPDLParameters */
static int hf_rnsap_Inter_Frequency_Cell_List_item = -1;  /* Inter_Frequency_Cell */
static int hf_rnsap_inter_Frequency_Cell_Information_SIB11 = -1;  /* Inter_Frequency_Cell_Information_SIB11 */
static int hf_rnsap_inter_Frequency_Cell_Information_SIB12 = -1;  /* Inter_Frequency_Cell_Information_SIB12 */
static int hf_rnsap_Inter_Frequency_Cell_Information_SIB11_item = -1;  /* Inter_Frequency_Cells_Information_SIB11_Per_Indication */
static int hf_rnsap_Inter_Frequency_Cell_Information_SIB12_item = -1;  /* Inter_Frequency_Cells_Information_SIB12_Per_Indication */
static int hf_rnsap_Inter_Frequency_Cell_Information_SIB11ter_item = -1;  /* Inter_Frequency_Cells_Information_SIB11ter_Per_Indication */
static int hf_rnsap_inter_Frequency_Cell_Indication_SIB11 = -1;  /* Inter_Frequency_Cell_Indication */
static int hf_rnsap_inter_Frequency_Cell_List_SIB11 = -1;  /* Inter_Frequency_Cell_SIB11_or_SIB12_List */
static int hf_rnsap_inter_Frequency_Cell_Indication_SIB12 = -1;  /* Inter_Frequency_Cell_Indication */
static int hf_rnsap_inter_Frequency_Cell_List_SIB12 = -1;  /* Inter_Frequency_Cell_SIB11_or_SIB12_List */
static int hf_rnsap_inter_Frequency_Cell_Indication_SIB11ter = -1;  /* Inter_Frequency_Cell_Indication */
static int hf_rnsap_inter_Frequency_Cell_List_SIB11ter = -1;  /* Inter_Frequency_Cell_SIB11ter_List */
static int hf_rnsap_Inter_Frequency_Cell_SIB11ter_List_item = -1;  /* Inter_Frequency_Cell_SIB11ter */
static int hf_rnsap_interFrequencyCellID = -1;    /* InterFrequencyCellID_Ext */
static int hf_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12_List_item = -1;  /* Inter_Frequency_Cell_SIB11_or_SIB12 */
static int hf_rnsap_interFrequencyCellID_01 = -1;  /* InterFrequencyCellID */
static int hf_rnsap_iPSpacingFDD = -1;            /* IPSpacingFDD */
static int hf_rnsap_iPLength = -1;                /* IPLength */
static int hf_rnsap_iPOffset = -1;                /* IPOffset */
static int hf_rnsap_seed = -1;                    /* Seed */
static int hf_rnsap_burstModeParameters = -1;     /* BurstModeParameters */
static int hf_rnsap_iPSpacingTDD = -1;            /* IPSpacingTDD */
static int hf_rnsap_iPStart = -1;                 /* IPStart */
static int hf_rnsap_iPSlot = -1;                  /* IPSlot */
static int hf_rnsap_iP_P_CCPCH = -1;              /* IP_P_CCPCH */
static int hf_rnsap_iPSub = -1;                   /* IPSub */
static int hf_rnsap_idleIntervalInfo_k = -1;      /* INTEGER_2_3 */
static int hf_rnsap_idleIntervalInfo_offset = -1;  /* INTEGER_0_7 */
static int hf_rnsap_laiList = -1;                 /* LAI_List */
static int hf_rnsap_LAI_List_item = -1;           /* LAI */
static int hf_rnsap_List_Of_PLMNs_item = -1;      /* PLMN_Identity */
static int hf_rnsap_uplinkLoadValue = -1;         /* INTEGER_0_100 */
static int hf_rnsap_downlinkLoadValue = -1;       /* INTEGER_0_100 */
static int hf_rnsap_maxTimeslotsPerSubFrame = -1;  /* INTEGER_1_6 */
static int hf_rnsap_maxPhysChPerTimeslot = -1;    /* T_maxPhysChPerTimeslot */
static int hf_rnsap_periodic_01 = -1;             /* MDT_Report_Parameters */
static int hf_rnsap_event1F = -1;                 /* Event1F_Parameters */
static int hf_rnsap_event1I = -1;                 /* Event1I_Parameters */
static int hf_rnsap_cellbased = -1;               /* CellBased */
static int hf_rnsap_labased = -1;                 /* LABased */
static int hf_rnsap_rabased = -1;                 /* RABased */
static int hf_rnsap_plmn_area_based = -1;         /* NULL */
static int hf_rnsap_mdtActivation = -1;           /* MDT_Activation */
static int hf_rnsap_mdtAreaScope = -1;            /* MDTAreaScope */
static int hf_rnsap_m1report = -1;                /* M1Report */
static int hf_rnsap_m2report = -1;                /* M2Report */
static int hf_rnsap_m3report = -1;                /* NULL */
static int hf_rnsap_reportInterval = -1;          /* ReportInterval */
static int hf_rnsap_reportAmount = -1;            /* ReportAmount */
static int hf_rnsap_MAC_c_sh_SDU_LengthList_item = -1;  /* MAC_c_sh_SDU_Length */
static int hf_rnsap_MACdPDU_Size_IndexList_item = -1;  /* MACdPDU_Size_IndexItem */
static int hf_rnsap_sID = -1;                     /* SID */
static int hf_rnsap_MACdPDU_Size_IndexList_to_Modify_item = -1;  /* MACdPDU_Size_IndexItem_to_Modify */
static int hf_rnsap_accessPointName = -1;         /* AccessPointName */
static int hf_rnsap_iPMulticastAddress = -1;      /* IPMulticastAddress */
static int hf_rnsap_MBMS_Bearer_Service_List_item = -1;  /* TMGI */
static int hf_rnsap_tmgi = -1;                    /* TMGI */
static int hf_rnsap_transmissionMode = -1;        /* TransmissionMode */
static int hf_rnsap_preferredFrequencyLayer = -1;  /* UARFCN */
static int hf_rnsap_tMGI = -1;                    /* TMGI */
static int hf_rnsap_pTM_Cell_List = -1;           /* PTMCellList */
static int hf_rnsap_pTP_Cell_List = -1;           /* PTPCellList */
static int hf_rnsap_not_Provided_Cell_List = -1;  /* NotProvidedCellList */
static int hf_rnsap_affectedUEInformationForMBMS = -1;  /* AffectedUEInformationForMBMS */
static int hf_rnsap_preferredFrequencyLayerInfo = -1;  /* PreferredFrequencyLayerInfo */
static int hf_rnsap_mBMS_ConcatenatedServiceList = -1;  /* MBMS_ConcatenatedServiceList */
static int hf_rnsap_l3_Information_1 = -1;        /* L3_Information */
static int hf_rnsap_l3_Information_2 = -1;        /* L3_Information */
static int hf_rnsap_MBMS_ConcatenatedServiceList_item = -1;  /* TMGI */
static int hf_rnsap_MBMS_RLC_Sequence_Number_Information_item = -1;  /* MBMS_RLC_Sequence_Number_Information_List */
static int hf_rnsap_mBMS_Bearer_Service_List_RLC = -1;  /* MBMS_Bearer_Service_List_RLC */
static int hf_rnsap_MBMS_Bearer_Service_List_RLC_item = -1;  /* MBMS_Bearer_Service_List_RLCinfo */
static int hf_rnsap_time_Stamp = -1;              /* Time_Stamp */
static int hf_rnsap_MCCH_Message_List_item = -1;  /* L3_Information */
static int hf_rnsap_secondaryCCPCHSystemInformationMBMS = -1;  /* Secondary_CCPCH_System_Information_MBMS */
static int hf_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Info_List_item = -1;  /* MBSFN_Scheduling_Transmission_Time_Interval_Item */
static int hf_rnsap_mbsfnSchedulingTransmissionTimeInterval = -1;  /* MbsfnSchedulingTransmissionTimeInterval */
static int hf_rnsap_sir = -1;                     /* SIR_Value_IncrDecrThres */
static int hf_rnsap_sir_error = -1;               /* SIR_Error_Value_IncrDecrThres */
static int hf_rnsap_transmitted_code_power = -1;  /* Transmitted_Code_Power_Value_IncrDecrThres */
static int hf_rnsap_rscp = -1;                    /* RSCP_Value_IncrDecrThres */
static int hf_rnsap_round_trip_time = -1;         /* Round_Trip_Time_IncrDecrThres */
static int hf_rnsap_extension_MeasurementIncreaseDecreaseThreshold = -1;  /* Extension_MeasurementIncreaseDecreaseThreshold */
static int hf_rnsap_sir_01 = -1;                  /* SIR_Value */
static int hf_rnsap_sir_error_01 = -1;            /* SIR_Error_Value */
static int hf_rnsap_transmitted_code_power_01 = -1;  /* Transmitted_Code_Power_Value */
static int hf_rnsap_rscp_01 = -1;                 /* RSCP_Value */
static int hf_rnsap_rx_timing_deviation = -1;     /* Rx_Timing_Deviation_Value */
static int hf_rnsap_round_trip_time_01 = -1;      /* Round_Trip_Time_Value */
static int hf_rnsap_extension_MeasurementThreshold = -1;  /* Extension_MeasurementThreshold */
static int hf_rnsap_type1 = -1;                   /* Type1 */
static int hf_rnsap_midambleConfigurationBurstType1And3 = -1;  /* MidambleConfigurationBurstType1And3 */
static int hf_rnsap_midambleAllocationMode = -1;  /* MidambleAllocationMode1 */
static int hf_rnsap_defaultMidamble = -1;         /* NULL */
static int hf_rnsap_commonMidamble = -1;          /* NULL */
static int hf_rnsap_ueSpecificMidamble = -1;      /* MidambleShiftLong */
static int hf_rnsap_type2 = -1;                   /* Type2 */
static int hf_rnsap_midambleConfigurationBurstType2 = -1;  /* MidambleConfigurationBurstType2 */
static int hf_rnsap_midambleAllocationMode_01 = -1;  /* MidambleAllocationMode2 */
static int hf_rnsap_ueSpecificMidamble_01 = -1;   /* MidambleShiftShort */
static int hf_rnsap_type3 = -1;                   /* Type3 */
static int hf_rnsap_midambleAllocationMode_02 = -1;  /* MidambleAllocationMode3 */
static int hf_rnsap_midambleAllocationMode_03 = -1;  /* MidambleAllocationMode */
static int hf_rnsap_midambleShift_01 = -1;        /* MidambleShiftLong */
static int hf_rnsap_type1_01 = -1;                /* Type7681 */
static int hf_rnsap_midambleAllocationMode_04 = -1;  /* MidambleAllocationMode7681 */
static int hf_rnsap_type2_01 = -1;                /* Type7682 */
static int hf_rnsap_midambleConfigurationBurstType2_768 = -1;  /* MidambleConfigurationBurstType2_768 */
static int hf_rnsap_midambleAllocationMode_05 = -1;  /* MidambleAllocationMode7682 */
static int hf_rnsap_ueSpecificMidamble_02 = -1;   /* MidambleShiftShort768 */
static int hf_rnsap_type3_01 = -1;                /* Type7683 */
static int hf_rnsap_midambleAllocationMode_06 = -1;  /* MidambleAllocationMode7683 */
static int hf_rnsap_mIMO_PilotConfiguration = -1;  /* MIMO_PilotConfiguration */
static int hf_rnsap_mIMO_N_M_Ratio_01 = -1;       /* MIMO_N_M_Ratio */
static int hf_rnsap_primary_and_secondary_CPICH = -1;  /* MIMO_S_CPICH_Channelisation_Code */
static int hf_rnsap_normal_and_diversity_primary_CPICH = -1;  /* NULL */
static int hf_rnsap_Additional_Associated_Secondary_CPICH_item = -1;  /* Additional_Secondary_CPICH */
static int hf_rnsap_additional_secondary_CPICH = -1;  /* MIMO_S_CPICH_Channelisation_Code */
static int hf_rnsap_additional_D_CPICH = -1;      /* MIMO_S_CPICH_Channelisation_Code */
static int hf_rnsap_addPriorityQueue = -1;        /* PriorityQueue_InfoItem_to_Add */
static int hf_rnsap_modifyPriorityQueue = -1;     /* PriorityQueue_InfoItem_to_Modify */
static int hf_rnsap_deletePriorityQueue = -1;     /* PriorityQueue_Id */
static int hf_rnsap_dL_PowerBalancing_Information = -1;  /* DL_PowerBalancing_Information */
static int hf_rnsap_minimumReducedE_DPDCH_GainFactor = -1;  /* MinimumReducedE_DPDCH_GainFactor */
static int hf_rnsap_secondary_UL_Frequency_Activation_State = -1;  /* Secondary_UL_Frequency_Activation_State */
static int hf_rnsap_extendedPropagationDelay = -1;  /* ExtendedPropagationDelay */
static int hf_rnsap_enhanced_PrimaryCPICH_EcNo = -1;  /* Enhanced_PrimaryCPICH_EcNo */
static int hf_rnsap_phase_Reference_Update_Indicator = -1;  /* Phase_Reference_Update_Indicator */
static int hf_rnsap_additional_e_DCH_DL_Control_Channel_Grant = -1;  /* NULL */
static int hf_rnsap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_item = -1;  /* MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR */
static int hf_rnsap_uARFCN = -1;                  /* UARFCN */
static int hf_rnsap_multicarrier_EDCH_Transport_Bearer_Mode = -1;  /* Multicarrier_EDCH_Transport_Bearer_Mode */
static int hf_rnsap_multi_carrier_EDCH_Information = -1;  /* Multi_Carrier_EDCH_Information */
static int hf_rnsap_Multi_Carrier_EDCH_Information_item = -1;  /* Multi_Carrier_EDCH_LCR_InformationItem */
static int hf_rnsap_uARFCNforNt = -1;             /* UARFCN */
static int hf_rnsap_sNPL_carrier_group_indicator = -1;  /* SNPL_Carrier_Group_Indicator */
static int hf_rnsap_multi_Carrier_EDCH_MACdFlows_Information_TDD = -1;  /* Multi_Carrier_EDCH_MACdFlows_Information_TDD */
static int hf_rnsap_Multi_Carrier_EDCH_MACdFlows_Information_TDD_item = -1;  /* Multi_Carrier_EDCH_MACdFlows_Specific_Info */
static int hf_rnsap_continue_setup_change_Of_Multi_Carrier_EDCH = -1;  /* Continue_Setup_Change_Multi_Carrier_EDCH */
static int hf_rnsap_continue = -1;                /* NULL */
static int hf_rnsap_setup = -1;                   /* Multi_Carrier_EDCH_Info */
static int hf_rnsap_change = -1;                  /* Multi_Carrier_EDCH_Change_Info */
static int hf_rnsap_multi_Carrier_EDCH_Information_Removal_List = -1;  /* Multi_Carrier_EDCH_Information_Removal_List */
static int hf_rnsap_Multi_Carrier_EDCH_Information_Removal_List_item = -1;  /* Multi_Carrier_EDCH_Information_Removal_Info_ItemIEs */
static int hf_rnsap_Multi_Carrier_EDCH_Information_Response_item = -1;  /* Multi_Carrier_EDCH_LCR_Information_ResponseItem */
static int hf_rnsap_setup_01 = -1;                /* Multiflow_Information */
static int hf_rnsap_configurationChange = -1;     /* Multiflow_Information_To_Modify */
static int hf_rnsap_stop = -1;                    /* Multiflow_Stop */
static int hf_rnsap_total_Number_of_HS_DSCH_Cells = -1;  /* INTEGER_2_32_ */
static int hf_rnsap_mode = -1;                    /* Multiflow_Mode */
static int hf_rnsap_mimo = -1;                    /* Multiflow_MIMO */
static int hf_rnsap_timing = -1;                  /* Multiflow_Timing */
static int hf_rnsap_max_Number_of_HS_SCCH_Sets_per_NodeB = -1;  /* INTEGER_1_16_ */
static int hf_rnsap_assisting_CQI_RepetitionFactor = -1;  /* CQI_RepetitionFactor */
static int hf_rnsap_assisting_AckNack_RepetitionFactor = -1;  /* AckNack_RepetitionFactor */
static int hf_rnsap_time_Reference = -1;          /* NULL */
static int hf_rnsap_non_Time_Reference = -1;      /* INTEGER_0_30_ */
static int hf_rnsap_mU_MIMO_Operation = -1;       /* MU_MIMO_Operation */
static int hf_rnsap_standalone_Midamble_Channel_Information = -1;  /* Standalone_Midamble_Channel_Information */
static int hf_rnsap_standalone_Midamble_Configuration = -1;  /* Standalone_Midamble_Configuration */
static int hf_rnsap_standalone_MidambleShift = -1;  /* Standalone_MidambleShift */
static int hf_rnsap_repetitionPeriod_01 = -1;     /* Standalone_Midamble_RepetitionPeriod */
static int hf_rnsap_offset = -1;                  /* Standalone_Midamble_Offset */
static int hf_rnsap_referenceBeta = -1;           /* ReferenceBeta */
static int hf_rnsap_mU_MIMO_Usage_Indicator = -1;  /* MU_MIMO_Usage_Indicator */
static int hf_rnsap_standalone_Midamble_Channel_Indicator = -1;  /* Standalone_Midamble_Channel_Indicator */
static int hf_rnsap_gERAN_SI_Type = -1;           /* GERAN_SI_Type */
static int hf_rnsap_Neighbouring_UMTS_CellInformation_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_Neighbouring_UMTS_CellInformation_Ext_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_cN_PS_DomainIdentifier = -1;  /* CN_PS_DomainIdentifier */
static int hf_rnsap_cN_CS_DomainIdentifier = -1;  /* CN_CS_DomainIdentifier */
static int hf_rnsap_neighbouring_FDD_CellInformation = -1;  /* Neighbouring_FDD_CellInformation */
static int hf_rnsap_neighbouring_TDD_CellInformation = -1;  /* Neighbouring_TDD_CellInformation */
static int hf_rnsap_neighbouring_LCR_TDD_CellInformation = -1;  /* Neighbouring_LCR_TDD_CellInformation */
static int hf_rnsap_Neighbouring_FDD_CellInformation_item = -1;  /* Neighbouring_FDD_CellInformationItem */
static int hf_rnsap_uARFCNforNu = -1;             /* UARFCN */
static int hf_rnsap_uARFCNforNd = -1;             /* UARFCN */
static int hf_rnsap_frameOffset = -1;             /* FrameOffset */
static int hf_rnsap_cellIndividualOffset = -1;    /* CellIndividualOffset */
static int hf_rnsap_not_used_closedLoopMode2_SupportIndicator = -1;  /* NULL */
static int hf_rnsap_MultipleFrequencyBandsIndicatorList_item = -1;  /* FrequencyBandIndicator */
static int hf_rnsap_uC_ID = -1;                   /* UC_ID */
static int hf_rnsap_Neighbouring_GSM_CellInformationIEs_item = -1;  /* Neighbouring_GSM_CellInformationItem */
static int hf_rnsap_cGI = -1;                     /* CGI */
static int hf_rnsap_bSIC = -1;                    /* BSIC */
static int hf_rnsap_band_Indicator = -1;          /* Band_Indicator */
static int hf_rnsap_bCCH_ARFCN = -1;              /* BCCH_ARFCN */
static int hf_rnsap_Neighbouring_TDD_CellInformation_item = -1;  /* Neighbouring_TDD_CellInformationItem */
static int hf_rnsap_cellParameterID = -1;         /* CellParameterID */
static int hf_rnsap_Neighbouring_LCR_TDD_CellInformation_item = -1;  /* Neighbouring_LCR_TDD_CellInformationItem */
static int hf_rnsap_Neighbouring_E_UTRA_CellInformation_item = -1;  /* Neighbouring_E_UTRA_CellInformationItem */
static int hf_rnsap_eCGI = -1;                    /* ECGI */
static int hf_rnsap_eARFCN_Information = -1;      /* EARFCN_Information */
static int hf_rnsap_MultipleFrequencyBandsIndicatorList_EUTRA_item = -1;  /* RadioFrequencyBandEUTRA */
static int hf_rnsap_NotProvidedCellList_item = -1;  /* MBMSChannelTypeCellList */
static int hf_rnsap_uplinkNRTLoadInformationValue = -1;  /* INTEGER_0_3 */
static int hf_rnsap_downlinkNRTLoadInformationValue = -1;  /* INTEGER_0_3 */
static int hf_rnsap_new_non_serving_RL_selection = -1;  /* New_non_serving_RL_setup_selection */
static int hf_rnsap_new_Serving_RL_in_DRNS = -1;  /* NULL */
static int hf_rnsap_new_Serving_RL_Not_in_DRNS = -1;  /* NULL */
static int hf_rnsap_new_Serving_RL_in_or_Not_in_DRNS = -1;  /* NULL */
static int hf_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_item = -1;  /* Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_item */
static int hf_rnsap_new_non_serving_RL_E_DCH_FDD_DL_Control_Channel_Information_A = -1;  /* EDCH_FDD_DL_ControlChannelInformation */
static int hf_rnsap_new_non_serving_RL_E_DCH_FDD_DL_Control_Channel_Information_B = -1;  /* EDCH_FDD_DL_ControlChannelInformation */
static int hf_rnsap_new_non_serving_RL_E_DCH_FDD_DL_Control_Channel_Information_C = -1;  /* EDCH_FDD_DL_ControlChannelInformation */
static int hf_rnsap_measurementThreshold = -1;    /* MeasurementThreshold */
static int hf_rnsap_informationThreshold = -1;    /* InformationThreshold */
static int hf_rnsap_PCH_InformationList_item = -1;  /* PCH_InformationItem */
static int hf_rnsap_informationReportPeriodicity = -1;  /* InformationReportPeriodicity */
static int hf_rnsap_imsi = -1;                    /* IMSI */
static int hf_rnsap_PLMN_List_item = -1;          /* PLMN_Identity */
static int hf_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas_item = -1;  /* Associated_Secondary_CPICH */
static int hf_rnsap_associated_secondary_CPICH_channelId = -1;  /* CommonPhysicalChannelID */
static int hf_rnsap_associated_secondary_CPICH = -1;  /* PowerOffsetForSCPICH_DCPICHforMIMOwithfourtransmitantennas */
static int hf_rnsap_associated_D_CPICH_channelId = -1;  /* CommonPhysicalChannelID */
static int hf_rnsap_associated_D_CPICH = -1;      /* PowerOffsetForSCPICH_DCPICHforMIMOwithfourtransmitantennas */
static int hf_rnsap_defaultPreferredFrequency = -1;  /* UARFCN */
static int hf_rnsap_additionalPreferredFrequency = -1;  /* AdditionalPreferredFrequency */
static int hf_rnsap_PriorityQueue_InfoList_item = -1;  /* PriorityQueue_InfoItem */
static int hf_rnsap_priorityQueue_Id = -1;        /* PriorityQueue_Id */
static int hf_rnsap_associatedHSDSCH_MACdFlow = -1;  /* HSDSCH_MACdFlow_ID */
static int hf_rnsap_t1 = -1;                      /* T1 */
static int hf_rnsap_discardTimer = -1;            /* DiscardTimer */
static int hf_rnsap_mAC_hsWindowSize = -1;        /* MAC_hsWindowSize */
static int hf_rnsap_mAChsGuaranteedBitRate = -1;  /* MAChsGuaranteedBitRate */
static int hf_rnsap_mACdPDU_Size_Index = -1;      /* MACdPDU_Size_IndexList */
static int hf_rnsap_rLC_Mode = -1;                /* RLC_Mode */
static int hf_rnsap_PriorityQueue_InfoList_EnhancedFACH_PCH_item = -1;  /* PriorityQueue_InfoItem_EnhancedFACH_PCH */
static int hf_rnsap_mAC_ehs_Reset_Timer = -1;     /* MAC_ehs_Reset_Timer */
static int hf_rnsap_maximum_MACdPDU_Size_01 = -1;  /* MAC_PDU_SizeExtended */
static int hf_rnsap_PriorityQueue_InfoList_to_Modify_item = -1;  /* ModifyPriorityQueue */
static int hf_rnsap_mACdPDU_Size_Index_to_Modify = -1;  /* MACdPDU_Size_IndexList_to_Modify */
static int hf_rnsap_PriorityQueue_InfoList_to_Modify_Unsynchronised_item = -1;  /* PriorityQueue_InfoItem_to_Modify_Unsynchronised */
static int hf_rnsap_priorityQueueId = -1;         /* PriorityQueue_Id */
static int hf_rnsap_mBMSChannelTypeInfo = -1;     /* MBMSChannelTypeInfo */
static int hf_rnsap_mBMSPreferredFreqLayerInfo = -1;  /* MBMSPreferredFreqLayerInfo */
static int hf_rnsap_UpPCH_InformationList_LCRTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_uPPCHPositionLCR = -1;        /* UPPCHPositionLCR */
static int hf_rnsap_PTMCellList_item = -1;        /* MBMSChannelTypeCellList */
static int hf_rnsap_PTPCellList_item = -1;        /* MBMSChannelTypeCellList */
static int hf_rnsap_raiList = -1;                 /* RAI_List */
static int hf_rnsap_RAI_List_item = -1;           /* RAI */
static int hf_rnsap_RB_Info_item = -1;            /* RB_Identity */
static int hf_rnsap_Reference_E_TFCI_Information_item = -1;  /* Reference_E_TFCI_Information_Item */
static int hf_rnsap_reference_E_TFCI = -1;        /* E_TFCI */
static int hf_rnsap_reference_E_TFCI_PO = -1;     /* Reference_E_TFCI_PO */
static int hf_rnsap_pSDomain = -1;                /* NULL */
static int hf_rnsap_cSDomain = -1;                /* NULL */
static int hf_rnsap_pS_CSDomain = -1;             /* NULL */
static int hf_rnsap_periodic_02 = -1;             /* Periodic */
static int hf_rnsap_eventA = -1;                  /* EventA */
static int hf_rnsap_eventB = -1;                  /* EventB */
static int hf_rnsap_eventC = -1;                  /* EventC */
static int hf_rnsap_eventD = -1;                  /* EventD */
static int hf_rnsap_eventE = -1;                  /* EventE */
static int hf_rnsap_eventF = -1;                  /* EventF */
static int hf_rnsap_extension_ReportCharacteristics = -1;  /* Extension_ReportCharacteristics */
static int hf_rnsap_ten_msec = -1;                /* INTEGER_1_6000_ */
static int hf_rnsap_gA_AccessPointPositionwithAltitude = -1;  /* GA_AccessPointPositionwithOptionalAltitude */
static int hf_rnsap_iPDLParameters = -1;          /* IPDLParameters */
static int hf_rnsap_dGPSCorrections = -1;         /* DGPSCorrections */
static int hf_rnsap_gPS_NavigationModel_and_TimeRecovery = -1;  /* GPS_NavigationModel_and_TimeRecovery */
static int hf_rnsap_gPS_Ionospheric_Model = -1;   /* GPS_Ionospheric_Model */
static int hf_rnsap_gPS_UTC_Model = -1;           /* GPS_UTC_Model */
static int hf_rnsap_gPS_Almanac = -1;             /* GPS_Almanac */
static int hf_rnsap_gPS_RealTime_Integrity = -1;  /* GPS_RealTime_Integrity */
static int hf_rnsap_gPS_RX_POS = -1;              /* GPS_RX_POS */
static int hf_rnsap_sFNSFN_GA_AccessPointPosition = -1;  /* GA_AccessPointPositionwithOptionalAltitude */
static int hf_rnsap_informationAvailable = -1;    /* InformationAvailable */
static int hf_rnsap_informationNotAvailable = -1;  /* InformationNotAvailable */
static int hf_rnsap_RL_Specific_DCH_Info_item = -1;  /* RL_Specific_DCH_Info_Item */
static int hf_rnsap_dCH_id = -1;                  /* DCH_ID */
static int hf_rnsap_rL_Specific_EDCH_Info = -1;   /* RL_Specific_EDCH_Info */
static int hf_rnsap_RL_Specific_EDCH_Info_item = -1;  /* RL_Specific_EDCH_InfoItem */
static int hf_rnsap_uplinkRTLoadValue = -1;       /* INTEGER_0_100 */
static int hf_rnsap_downlinkRTLoadValue = -1;     /* INTEGER_0_100 */
static int hf_rnsap_sAC = -1;                     /* SAC */
static int hf_rnsap_SecondaryServingCells_item = -1;  /* SecondaryServingCellsItem */
static int hf_rnsap_secondaryC_ID = -1;           /* C_ID */
static int hf_rnsap_numSecondaryHS_SCCH_Codes = -1;  /* NumHS_SCCH_Codes */
static int hf_rnsap_dl_TFCS = -1;                 /* TFCS */
static int hf_rnsap_tFCI_Coding = -1;             /* TFCI_Coding */
static int hf_rnsap_secondary_CCPCH_TDD_InformationList = -1;  /* Secondary_CCPCH_TDD_InformationList */
static int hf_rnsap_fACH_InformationList = -1;    /* FACH_InformationList */
static int hf_rnsap_pCH_InformationList = -1;     /* PCH_InformationList */
static int hf_rnsap_new_secondary_CPICH = -1;     /* Secondary_CPICH_Information */
static int hf_rnsap_secondary_CPICH_shall_not_be_used = -1;  /* NULL */
static int hf_rnsap_secondary_LCR_CCPCH_TDD_InformationList = -1;  /* Secondary_LCR_CCPCH_TDD_InformationList */
static int hf_rnsap_secondary_CCPCH_TDD_InformationList768 = -1;  /* Secondary_CCPCH_TDD_InformationList768 */
static int hf_rnsap_Secondary_CCPCH_TDD_InformationList_item = -1;  /* Secondary_CCPCH_TDD_InformationItem */
static int hf_rnsap_secondary_CCPCH_TDD_Code_Information = -1;  /* Secondary_CCPCH_TDD_Code_Information */
static int hf_rnsap_tDD_PhysicalChannelOffset = -1;  /* TDD_PhysicalChannelOffset */
static int hf_rnsap_Secondary_LCR_CCPCH_TDD_InformationList_item = -1;  /* Secondary_LCR_CCPCH_TDD_InformationItem */
static int hf_rnsap_secondary_LCR_CCPCH_TDD_Code_Information = -1;  /* Secondary_LCR_CCPCH_TDD_Code_Information */
static int hf_rnsap_Secondary_CCPCH_TDD_InformationList768_item = -1;  /* Secondary_CCPCH_TDD_InformationItem768 */
static int hf_rnsap_secondary_CCPCH_TDD_Code_Information768 = -1;  /* Secondary_CCPCH_TDD_Code_Information768 */
static int hf_rnsap_Secondary_CCPCH_TDD_Code_Information_item = -1;  /* Secondary_CCPCH_TDD_Code_InformationItem */
static int hf_rnsap_Secondary_LCR_CCPCH_TDD_Code_Information_item = -1;  /* Secondary_LCR_CCPCH_TDD_Code_InformationItem */
static int hf_rnsap_s_CCPCH_TimeSlotFormat_LCR = -1;  /* TDD_DL_DPCH_TimeSlotFormat_LCR */
static int hf_rnsap_Secondary_CCPCH_TDD_Code_Information768_item = -1;  /* Secondary_CCPCH_TDD_Code_InformationItem768 */
static int hf_rnsap_possible_Secondary_Serving_Cell_List = -1;  /* Possible_Secondary_Serving_Cell_List */
static int hf_rnsap_Possible_Secondary_Serving_Cell_List_item = -1;  /* Possible_Secondary_Serving_Cell */
static int hf_rnsap_SetsOfHS_SCCH_Codes_item = -1;  /* SetsOfHS_SCCH_CodesItem */
static int hf_rnsap_hS_SCCH_PreconfiguredCodes = -1;  /* HS_SCCH_PreconfiguredCodes */
static int hf_rnsap_sixtyfourQAM_DL_SupportIndicator = -1;  /* SixtyfourQAM_DL_SupportIndicator */
static int hf_rnsap_setup_02 = -1;                /* Additional_EDCH_Setup_Info */
static int hf_rnsap_configurationChange_01 = -1;  /* Additional_EDCH_Cell_Information_ConfigurationChange_List */
static int hf_rnsap_removal = -1;                 /* Additional_EDCH_Cell_Information_Removal_List */
static int hf_rnsap_setup_03 = -1;                /* Downlink_TPC_enhancements_Information */
static int hf_rnsap_configurationChange_02 = -1;  /* Downlink_TPC_enhancements_Information */
static int hf_rnsap_removal_01 = -1;              /* Downlink_TPC_enhancements_Information_Removal */
static int hf_rnsap_setup_04 = -1;                /* UL_CLTD_Information */
static int hf_rnsap_configurationChange_03 = -1;  /* UL_CLTD_Information_To_Modify */
static int hf_rnsap_removal_02 = -1;              /* UL_CLTD_Information_Removal */
static int hf_rnsap_setup_05 = -1;                /* FTPICH_Information */
static int hf_rnsap_configurationChange_04 = -1;  /* FTPICH_Information_To_Modify */
static int hf_rnsap_removal_03 = -1;              /* FTPICH_Information_Removal */
static int hf_rnsap_setup_06 = -1;                /* DCH_ENH_Information */
static int hf_rnsap_configurationChange_05 = -1;  /* DCH_ENH_Information_to_Modify */
static int hf_rnsap_removal_04 = -1;              /* DCH_ENH_Information_Removal */
static int hf_rnsap_setup_07 = -1;                /* UL_DPCCH2_Information */
static int hf_rnsap_configurationChange_06 = -1;  /* UL_DPCCH2_Information_To_Modify */
static int hf_rnsap_removal_05 = -1;              /* UL_DPCCH2_Information_Removal */
static int hf_rnsap_sFNSFNChangeLimit = -1;       /* SFNSFNChangeLimit */
static int hf_rnsap_predictedSFNSFNDeviationLimit = -1;  /* PredictedSFNSFNDeviationLimit */
static int hf_rnsap_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation = -1;  /* T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation */
static int hf_rnsap_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item = -1;  /* T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item */
static int hf_rnsap_sFNSFNValue = -1;             /* SFNSFNValue */
static int hf_rnsap_sFNSFNQuality = -1;           /* SFNSFNQuality */
static int hf_rnsap_sFNSFNDriftRate = -1;         /* SFNSFNDriftRate */
static int hf_rnsap_sFNSFNDriftRateQuality = -1;  /* SFNSFNDriftRateQuality */
static int hf_rnsap_sFNSFNTimeStampInformation = -1;  /* SFNSFNTimeStampInformation */
static int hf_rnsap_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation = -1;  /* T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation */
static int hf_rnsap_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item = -1;  /* T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item */
static int hf_rnsap_sFNSFNTimeStamp_FDD = -1;     /* SFN */
static int hf_rnsap_sFNSFNTimeStamp_TDD = -1;     /* SFNSFNTimeStamp_TDD */
static int hf_rnsap_sFN = -1;                     /* SFN */
static int hf_rnsap_sFNSFN_FDD = -1;              /* SFNSFN_FDD */
static int hf_rnsap_sFNSFN_TDD = -1;              /* SFNSFN_TDD */
static int hf_rnsap_sFNSFN_TDD768 = -1;           /* SFNSFN_TDD768 */
static int hf_rnsap_listOfSNAs = -1;              /* ListOfSNAs */
static int hf_rnsap_ListOfSNAs_item = -1;         /* SNACode */
static int hf_rnsap_sRNTI = -1;                   /* S_RNTI */
static int hf_rnsap_sRNTI_BitMaskIndex = -1;      /* T_sRNTI_BitMaskIndex */
static int hf_rnsap_n_INSYNC_IND = -1;            /* INTEGER_1_256 */
static int hf_rnsap_n_OUTSYNC_IND = -1;           /* INTEGER_1_256 */
static int hf_rnsap_t_RLFAILURE = -1;             /* INTEGER_0_255 */
static int hf_rnsap_maxSYNC_UL_transmissions = -1;  /* T_maxSYNC_UL_transmissions */
static int hf_rnsap_powerRampStep = -1;           /* INTEGER_0_3_ */
static int hf_rnsap_modulation = -1;              /* Modulation */
static int hf_rnsap_TDD_DCHs_to_Modify_item = -1;  /* TDD_DCHs_to_ModifyItem */
static int hf_rnsap_dCH_SpecificInformationList_03 = -1;  /* TDD_DCHs_to_ModifySpecificInformationList */
static int hf_rnsap_TDD_DCHs_to_ModifySpecificInformationList_item = -1;  /* TDD_DCHs_to_ModifySpecificItem */
static int hf_rnsap_ul_CCTrCH_ID = -1;            /* CCTrCH_ID */
static int hf_rnsap_dl_CCTrCH_ID = -1;            /* CCTrCH_ID */
static int hf_rnsap_TDD_DL_Code_Information_item = -1;  /* TDD_DL_Code_InformationItem */
static int hf_rnsap_dPCH_ID = -1;                 /* DPCH_ID */
static int hf_rnsap_TDD_DL_Code_LCR_Information_item = -1;  /* TDD_DL_Code_LCR_InformationItem */
static int hf_rnsap_tdd_ChannelisationCodeLCR = -1;  /* TDD_ChannelisationCodeLCR */
static int hf_rnsap_tdd_DL_DPCH_TimeSlotFormat_LCR = -1;  /* TDD_DL_DPCH_TimeSlotFormat_LCR */
static int hf_rnsap_TDD_DL_Code_Information768_item = -1;  /* TDD_DL_Code_InformationItem768 */
static int hf_rnsap_qPSK = -1;                    /* QPSK_DL_DPCH_TimeSlotFormatTDD_LCR */
static int hf_rnsap_eightPSK = -1;                /* EightPSK_DL_DPCH_TimeSlotFormatTDD_LCR */
static int hf_rnsap_initialOffset = -1;           /* INTEGER_0_255 */
static int hf_rnsap_noinitialOffset = -1;         /* INTEGER_0_63 */
static int hf_rnsap_TDD_UL_Code_Information_item = -1;  /* TDD_UL_Code_InformationItem */
static int hf_rnsap_TDD_UL_Code_LCR_Information_item = -1;  /* TDD_UL_Code_LCR_InformationItem */
static int hf_rnsap_tdd_UL_DPCH_TimeSlotFormat_LCR = -1;  /* TDD_UL_DPCH_TimeSlotFormat_LCR */
static int hf_rnsap_TDD_UL_Code_Information768_item = -1;  /* TDD_UL_Code_InformationItem768 */
static int hf_rnsap_qPSK_01 = -1;                 /* QPSK_UL_DPCH_TimeSlotFormatTDD_LCR */
static int hf_rnsap_eightPSK_01 = -1;             /* EightPSK_UL_DPCH_TimeSlotFormatTDD_LCR */
static int hf_rnsap_plmn_id = -1;                 /* PLMN_Identity */
static int hf_rnsap_service_id = -1;              /* Service_ID */
static int hf_rnsap_dsField = -1;                 /* DsField */
static int hf_rnsap_genericTrafficCategory = -1;  /* GenericTrafficCategory */
static int hf_rnsap_Transmission_Gap_Pattern_Sequence_Information_item = -1;  /* Transmission_Gap_Pattern_Sequence_Information_item */
static int hf_rnsap_tGPSID = -1;                  /* TGPSID */
static int hf_rnsap_tGSN = -1;                    /* TGSN */
static int hf_rnsap_tGL1 = -1;                    /* GapLength */
static int hf_rnsap_tGL2 = -1;                    /* GapLength */
static int hf_rnsap_tGD = -1;                     /* TGD */
static int hf_rnsap_tGPL1 = -1;                   /* GapDuration */
static int hf_rnsap_not_to_be_used_1 = -1;        /* GapDuration */
static int hf_rnsap_uL_DL_mode = -1;              /* UL_DL_mode */
static int hf_rnsap_downlink_Compressed_Mode_Method = -1;  /* Downlink_Compressed_Mode_Method */
static int hf_rnsap_uplink_Compressed_Mode_Method = -1;  /* Uplink_Compressed_Mode_Method */
static int hf_rnsap_dL_FrameType = -1;            /* DL_FrameType */
static int hf_rnsap_delta_SIR1 = -1;              /* DeltaSIR */
static int hf_rnsap_delta_SIR_after1 = -1;        /* DeltaSIR */
static int hf_rnsap_delta_SIR2 = -1;              /* DeltaSIR */
static int hf_rnsap_delta_SIR_after2 = -1;        /* DeltaSIR */
static int hf_rnsap_Transmission_Gap_Pattern_Sequence_Status_List_item = -1;  /* Transmission_Gap_Pattern_Sequence_Status_List_item */
static int hf_rnsap_tGPRC = -1;                   /* TGPRC */
static int hf_rnsap_tGCFN = -1;                   /* CFN */
static int hf_rnsap_Affected_HSDSCH_Serving_Cell_List_item = -1;  /* C_ID */
static int hf_rnsap_Transmission_Mode_Information_item = -1;  /* Transmission_Mode_Information_List */
static int hf_rnsap_mS = -1;                      /* INTEGER_0_16383 */
static int hf_rnsap_lS = -1;                      /* INTEGER_0_4294967295 */
static int hf_rnsap_tUTRANGANSSChangeLimit = -1;  /* INTEGER_1_256 */
static int hf_rnsap_predictedTUTRANGANSSDeviationLimit = -1;  /* INTEGER_1_256 */
static int hf_rnsap_tUTRANGANSS = -1;             /* TUTRANGANSS */
static int hf_rnsap_tUTRANGANSSQuality = -1;      /* INTEGER_0_255 */
static int hf_rnsap_tUTRANGANSSDriftRate = -1;    /* INTEGER_M50_50 */
static int hf_rnsap_tUTRANGANSSDriftRateQuality = -1;  /* INTEGER_0_50 */
static int hf_rnsap_ms_part = -1;                 /* INTEGER_0_16383 */
static int hf_rnsap_ls_part = -1;                 /* INTEGER_0_4294967295 */
static int hf_rnsap_tUTRANGPSChangeLimit = -1;    /* TUTRANGPSChangeLimit */
static int hf_rnsap_predictedTUTRANGPSDeviationLimit = -1;  /* PredictedTUTRANGPSDeviationLimit */
static int hf_rnsap_tUTRANGPS = -1;               /* TUTRANGPS */
static int hf_rnsap_tUTRANGPSQuality = -1;        /* TUTRANGPSQuality */
static int hf_rnsap_tUTRANGPSDriftRate = -1;      /* TUTRANGPSDriftRate */
static int hf_rnsap_tUTRANGPSDriftRateQuality = -1;  /* TUTRANGPSDriftRateQuality */
static int hf_rnsap_iEe_Extensions = -1;          /* ProtocolExtensionContainer */
static int hf_rnsap_signalledGainFactors = -1;    /* T_signalledGainFactors */
static int hf_rnsap_betaC = -1;                   /* BetaCD */
static int hf_rnsap_betaD = -1;                   /* BetaCD */
static int hf_rnsap_refTFCNumber = -1;            /* RefTFCNumber */
static int hf_rnsap_tFCSvalues = -1;              /* T_tFCSvalues */
static int hf_rnsap_no_Split_in_TFCI = -1;        /* TFCS_TFCSList */
static int hf_rnsap_not_Used_split_in_TFCI = -1;  /* NULL */
static int hf_rnsap_TFCS_TFCSList_item = -1;      /* TFCS_TFCSList_item */
static int hf_rnsap_cTFC = -1;                    /* TFCS_CTFC */
static int hf_rnsap_tFC_Beta = -1;                /* TransportFormatCombination_Beta */
static int hf_rnsap_ctfc2bit = -1;                /* INTEGER_0_3 */
static int hf_rnsap_ctfc4bit = -1;                /* INTEGER_0_15 */
static int hf_rnsap_ctfc6bit = -1;                /* INTEGER_0_63 */
static int hf_rnsap_ctfc8bit = -1;                /* INTEGER_0_255 */
static int hf_rnsap_ctfc12bit = -1;               /* INTEGER_0_4095 */
static int hf_rnsap_ctfc16bit = -1;               /* INTEGER_0_65535 */
static int hf_rnsap_ctfcmaxbit = -1;              /* INTEGER_0_maxCTFC */
static int hf_rnsap_slot_position = -1;           /* INTEGER_0_4 */
static int hf_rnsap_dynamicParts = -1;            /* TransportFormatSet_DynamicPartList */
static int hf_rnsap_semi_staticPart = -1;         /* TransportFormatSet_Semi_staticPart */
static int hf_rnsap_TransportFormatSet_DynamicPartList_item = -1;  /* TransportFormatSet_DynamicPartList_item */
static int hf_rnsap_nrOfTransportBlocks = -1;     /* NrOfTransportBlocks */
static int hf_rnsap_transportBlockSize = -1;      /* TransportBlockSize */
static int hf_rnsap_mode_01 = -1;                 /* TransportFormatSet_ModeDP */
static int hf_rnsap_tdd = -1;                     /* TDD_TransportFormatSet_ModeDP */
static int hf_rnsap_notApplicable = -1;           /* NULL */
static int hf_rnsap_transmissionTimeIntervalInformation = -1;  /* TransmissionTimeIntervalInformation */
static int hf_rnsap_TransmissionTimeIntervalInformation_item = -1;  /* TransmissionTimeIntervalInformation_item */
static int hf_rnsap_transmissionTimeInterval = -1;  /* TransmissionTimeIntervalDynamic */
static int hf_rnsap_transmissionTime = -1;        /* TransmissionTimeIntervalSemiStatic */
static int hf_rnsap_channelCoding = -1;           /* ChannelCodingType */
static int hf_rnsap_codingRate = -1;              /* CodingRate */
static int hf_rnsap_rateMatcingAttribute = -1;    /* RateMatchingAttribute */
static int hf_rnsap_cRC_Size = -1;                /* CRC_Size */
static int hf_rnsap_mode_02 = -1;                 /* TransportFormatSet_ModeSSP */
static int hf_rnsap_tdd_01 = -1;                  /* SecondInterleavingMode */
static int hf_rnsap_tTI_Update_CFN = -1;          /* CFN */
static int hf_rnsap_tTI_Update_Ind = -1;          /* TTI_Update_Ind */
static int hf_rnsap_uE_AggregateMaximumBitRateDownlink = -1;  /* UE_AggregateMaximumBitRateDownlink */
static int hf_rnsap_uE_AggregateMaximumBitRateUplink = -1;  /* UE_AggregateMaximumBitRateUplink */
static int hf_rnsap_UE_RF_Band_CapabilityLCR_item = -1;  /* Radio_Frequency_BandItem */
static int hf_rnsap_radio_Frequency_Band = -1;    /* Radio_Frequency_Band */
static int hf_rnsap_imei = -1;                    /* IMEI */
static int hf_rnsap_imeisv = -1;                  /* IMEISV */
static int hf_rnsap_periodic_03 = -1;             /* UEMeasurementReportCharacteristicsPeriodic */
static int hf_rnsap_event1h = -1;                 /* UEMeasurementReportCharacteristicsEvent1h */
static int hf_rnsap_event1i = -1;                 /* UEMeasurementReportCharacteristicsEvent1i */
static int hf_rnsap_event6a = -1;                 /* UEMeasurementReportCharacteristicsEvent6a */
static int hf_rnsap_event6b = -1;                 /* UEMeasurementReportCharacteristicsEvent6b */
static int hf_rnsap_event6c = -1;                 /* UEMeasurementReportCharacteristicsEvent6c */
static int hf_rnsap_event6d = -1;                 /* UEMeasurementReportCharacteristicsEvent6d */
static int hf_rnsap_extension_ReportCharacteristics_01 = -1;  /* UEMeasurementReportCharacteristics_Extension */
static int hf_rnsap_uEMeasurementTreshold = -1;   /* UEMeasurementThreshold */
static int hf_rnsap_uEMeasurementTimeToTrigger = -1;  /* UEMeasurementTimeToTrigger */
static int hf_rnsap_uEMeasurementHysteresisTime = -1;  /* UEMeasurementHysteresisTime */
static int hf_rnsap_amountofReporting = -1;       /* UEMeasurementReportCharacteristicsPeriodicAmountofReporting */
static int hf_rnsap_reportingInterval = -1;       /* UEMeasurementReportCharacteristicsPeriodicReportingInterval */
static int hf_rnsap_timeslotISCP = -1;            /* UEMeasurementThresholdDLTimeslotISCP */
static int hf_rnsap_uETransmitPower = -1;         /* UEMeasurementThresholdUETransmitPower */
static int hf_rnsap_extension_UEMeasurementThreshold = -1;  /* UEMeasurementThreshold_Extension */
static int hf_rnsap_UEMeasurementTimeslotInfoHCR_item = -1;  /* UEMeasurementTimeslotInfoHCR_IEs */
static int hf_rnsap_burstType = -1;               /* UEMeasurementTimeslotInfoHCRBurstType */
static int hf_rnsap_UEMeasurementTimeslotInfoLCR_item = -1;  /* UEMeasurementTimeslotInfoLCR_IEs */
static int hf_rnsap_UEMeasurementTimeslotInfo768_item = -1;  /* UEMeasurementTimeslotInfo768_IEs */
static int hf_rnsap_burstType_01 = -1;            /* UEMeasurementTimeslotInfo768BurstType */
static int hf_rnsap_uE_Transmitted_Power = -1;    /* UE_MeasurementValue_UE_Transmitted_Power */
static int hf_rnsap_primary_CCPCH_RSCP = -1;      /* UE_MeasurementValue_Primary_CCPCH_RSCP */
static int hf_rnsap_dL_Timeslot_ISCP = -1;        /* UE_MeasurementValue_DL_Timeslot_ISCP */
static int hf_rnsap_extension_UEMeasurementValue = -1;  /* UEMeasurementValue_Extension */
static int hf_rnsap_uEMeasurementTransmittedPowerListHCR = -1;  /* UEMeasurementValueTransmittedPowerListHCR */
static int hf_rnsap_uEMeasurementTransmittedPowerListLCR = -1;  /* UEMeasurementValueTransmittedPowerListLCR */
static int hf_rnsap_UEMeasurementValueTransmittedPowerListHCR_item = -1;  /* UEMeasurementValueTransmittedPowerListHCR_IEs */
static int hf_rnsap_uETransmitPower_01 = -1;      /* INTEGER_0_104 */
static int hf_rnsap_UEMeasurementValueTransmittedPowerListLCR_item = -1;  /* UEMeasurementValueTransmittedPowerListLCR_IEs */
static int hf_rnsap_UEMeasurementValueTransmittedPowerList768_item = -1;  /* UEMeasurementValueTransmittedPowerList768_IEs */
static int hf_rnsap_primaryCCPCH_RSCP = -1;       /* PrimaryCCPCH_RSCP */
static int hf_rnsap_primaryCCPCH_RSCP_Delta = -1;  /* PrimaryCCPCH_RSCP_Delta */
static int hf_rnsap_uEMeasurementTimeslotISCPListHCR = -1;  /* UEMeasurementValueTimeslotISCPListHCR */
static int hf_rnsap_uEMeasurementTimeslotISCPListLCR = -1;  /* UEMeasurementValueTimeslotISCPListLCR */
static int hf_rnsap_UEMeasurementValueTimeslotISCPListHCR_item = -1;  /* UEMeasurementValueTimeslotISCPListHCR_IEs */
static int hf_rnsap_UEMeasurementValueTimeslotISCPListLCR_item = -1;  /* UEMeasurementValueTimeslotISCPListLCR_IEs */
static int hf_rnsap_UEMeasurementValueTimeslotISCPList768_item = -1;  /* UEMeasurementValueTimeslotISCPList768_IEs */
static int hf_rnsap_measurementAvailable_02 = -1;  /* UEMeasurementValueInformationAvailable */
static int hf_rnsap_measurementnotAvailable_02 = -1;  /* UEMeasurementValueInformationnotAvailable */
static int hf_rnsap_uEmeasurementValue = -1;      /* UEMeasurementValue */
static int hf_rnsap_uPHFiltering_Value = -1;      /* UPHFiltering_Value */
static int hf_rnsap_extension_UE_Measurement_Value = -1;  /* Extension_UE_Measurement_Value */
static int hf_rnsap_cell_fach_pch = -1;           /* Cell_Fach_Pch_State */
static int hf_rnsap_ura_pch = -1;                 /* Ura_Pch_State */
static int hf_rnsap_d_RNTI = -1;                  /* D_RNTI */
static int hf_rnsap_srnc_id = -1;                 /* RNC_ID */
static int hf_rnsap_ura_id = -1;                  /* URA_ID */
static int hf_rnsap_UL_Timeslot_Information_item = -1;  /* UL_Timeslot_InformationItem */
static int hf_rnsap_uL_Code_Information = -1;     /* TDD_UL_Code_Information */
static int hf_rnsap_UL_TimeslotLCR_Information_item = -1;  /* UL_TimeslotLCR_InformationItem */
static int hf_rnsap_uL_Code_LCR_InformationList = -1;  /* TDD_UL_Code_LCR_Information */
static int hf_rnsap_sequenceNumber = -1;          /* PLCCHsequenceNumber */
static int hf_rnsap_UL_Timeslot_Information768_item = -1;  /* UL_Timeslot_InformationItem768 */
static int hf_rnsap_uL_Code_Information768 = -1;  /* TDD_UL_Code_Information768 */
static int hf_rnsap_UL_TimeSlot_ISCP_Info_item = -1;  /* UL_TimeSlot_ISCP_InfoItem */
static int hf_rnsap_uL_TimeslotISCP = -1;         /* UL_TimeslotISCP */
static int hf_rnsap_UL_TimeSlot_ISCP_LCR_Info_item = -1;  /* UL_TimeSlot_ISCP_LCR_InfoItem */
static int hf_rnsap_iSCP = -1;                    /* UL_Timeslot_ISCP_Value */
static int hf_rnsap_sync_UL_codes_bitmap = -1;    /* BIT_STRING_SIZE_8 */
static int hf_rnsap_fPACH_info = -1;              /* FPACH_Information */
static int hf_rnsap_prxUpPCHdes = -1;             /* INTEGER_M120_M58_ */
static int hf_rnsap_syncUL_procParameter = -1;    /* SYNC_UL_ProcParameters */
static int hf_rnsap_mMax = -1;                    /* INTEGER_1_32 */
static int hf_rnsap_ul_ScramblingCodeNumber = -1;  /* UL_ScramblingCodeNumber */
static int hf_rnsap_ul_ScramblingCodeLength = -1;  /* UL_ScramblingCodeLength */
static int hf_rnsap_uL_Synchronisation_StepSize = -1;  /* UL_Synchronisation_StepSize */
static int hf_rnsap_uL_Synchronisation_Frequency = -1;  /* UL_Synchronisation_Frequency */
static int hf_rnsap_uRA_ID = -1;                  /* URA_ID */
static int hf_rnsap_multipleURAsIndicator = -1;   /* MultipleURAsIndicator */
static int hf_rnsap_rNCsWithCellsInTheAccessedURA_List = -1;  /* RNCsWithCellsInTheAccessedURA_List */
static int hf_rnsap_RNCsWithCellsInTheAccessedURA_List_item = -1;  /* RNCsWithCellsInTheAccessedURA_Item */
static int hf_rnsap_USCH_Information_item = -1;   /* USCH_InformationItem */
static int hf_rnsap_uSCH_ID = -1;                 /* USCH_ID */
static int hf_rnsap_rb_Info = -1;                 /* RB_Info */
static int hf_rnsap_UMTS_Cells_Info_List_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_sDPCCH_PowerOffsetInformation = -1;  /* SDPCCH_PowerOffsetInformation */
static int hf_rnsap_uL_CLTD_Activation_Information = -1;  /* UL_CLTD_Activation_Information */
static int hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD = -1;  /* Setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD */
static int hf_rnsap_f_DPCH_Info = -1;             /* F_DPCH_Info */
static int hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2 = -1;  /* Setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2 */
static int hf_rnsap_f_DPCH_Info_01 = -1;          /* F_DPCH_Info_To_Modify */
static int hf_rnsap_e_roch_power_offset = -1;     /* E_ROCH_PowerOffset */
static int hf_rnsap_s_e_dpcch_power_offset = -1;  /* S_E_DPCCH_PowerOffset */
static int hf_rnsap_interstream_compensation = -1;  /* InterStream_Interference_Compensation */
static int hf_rnsap_minimum_E_TFCI_rank2 = -1;    /* INTEGER_0_127 */
static int hf_rnsap_setup_08 = -1;                /* UL_MIMO_Information */
static int hf_rnsap_configurationChange_07 = -1;  /* UL_MIMO_Information_To_Modify */
static int hf_rnsap_removal_06 = -1;              /* UL_MIMO_Removal */
static int hf_rnsap_e_roch_channelization_code = -1;  /* FDD_DL_ChannelisationCodeNumber */
static int hf_rnsap_s_e_rnti = -1;                /* E_RNTI */
static int hf_rnsap_s_signature_sequence = -1;    /* ERGCH_SignatureSequence */
static int hf_rnsap_s_e_roch_release_indicator = -1;  /* S_E_ROCH_Release_Indicator */
static int hf_rnsap_protocolIEs = -1;             /* ProtocolIE_Container */
static int hf_rnsap_protocolExtensions = -1;      /* ProtocolExtensionContainer */
static int hf_rnsap_minUL_ChannelisationCodeLength = -1;  /* MinUL_ChannelisationCodeLength */
static int hf_rnsap_maxNrOfUL_DPCHs = -1;         /* MaxNrOfUL_DPCHs */
static int hf_rnsap_ul_PunctureLimit = -1;        /* PunctureLimit */
static int hf_rnsap_ul_TFCS = -1;                 /* TFCS */
static int hf_rnsap_ul_DPCCH_SlotFormat = -1;     /* UL_DPCCH_SlotFormat */
static int hf_rnsap_ul_SIRTarget = -1;            /* UL_SIR */
static int hf_rnsap_diversityMode = -1;           /* DiversityMode */
static int hf_rnsap_not_Used_sSDT_CellIdLength = -1;  /* NULL */
static int hf_rnsap_not_Used_s_FieldLength = -1;  /* NULL */
static int hf_rnsap_tFCS = -1;                    /* TFCS */
static int hf_rnsap_dl_DPCH_SlotFormat = -1;      /* DL_DPCH_SlotFormat */
static int hf_rnsap_nrOfDLchannelisationcodes = -1;  /* NrOfDLchannelisationcodes */
static int hf_rnsap_tFCI_SignallingMode = -1;     /* TFCI_SignallingMode */
static int hf_rnsap_multiplexingPosition = -1;    /* MultiplexingPosition */
static int hf_rnsap_powerOffsetInformation = -1;  /* PowerOffsetInformation_RL_SetupRqstFDD */
static int hf_rnsap_fdd_dl_TPC_DownlinkStepSize = -1;  /* FDD_TPC_DownlinkStepSize */
static int hf_rnsap_po1_ForTFCI_Bits = -1;        /* PowerOffset */
static int hf_rnsap_po2_ForTPC_Bits = -1;         /* PowerOffset */
static int hf_rnsap_po3_ForPilotBits = -1;        /* PowerOffset */
static int hf_rnsap_RL_InformationList_RL_SetupRqstFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_chipOffset = -1;              /* ChipOffset */
static int hf_rnsap_diversityControlField = -1;   /* DiversityControlField */
static int hf_rnsap_dl_InitialTX_Power = -1;      /* DL_Power */
static int hf_rnsap_not_Used_sSDT_CellID = -1;    /* NULL */
static int hf_rnsap_transmitDiversityIndicator = -1;  /* TransmitDiversityIndicator */
static int hf_rnsap_Additional_HS_Cell_Information_RL_Setup_List_item = -1;  /* Additional_HS_Cell_Information_RL_Setup_ItemIEs */
static int hf_rnsap_hSPDSCH_RL_ID = -1;           /* RL_ID */
static int hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information = -1;  /* HS_DSCH_FDD_Secondary_Serving_Information */
static int hf_rnsap_powerOffsetInformation_01 = -1;  /* PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD */
static int hf_rnsap_maxNrTimeslots_UL = -1;       /* MaxNrTimeslots */
static int hf_rnsap_minimumSpreadingFactor_UL = -1;  /* MinimumSpreadingFactor */
static int hf_rnsap_maxNrULPhysicalchannels = -1;  /* MaxNrULPhysicalchannels */
static int hf_rnsap_maxNrTimeslots_DL = -1;       /* MaxNrTimeslots */
static int hf_rnsap_minimumSpreadingFactor_DL = -1;  /* MinimumSpreadingFactor */
static int hf_rnsap_maxNrDLPhysicalchannels = -1;  /* MaxNrDLPhysicalchannels */
static int hf_rnsap_UL_CCTrCH_InformationList_RL_SetupRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_cCTrCH_ID = -1;               /* CCTrCH_ID */
static int hf_rnsap_DL_CCTrCH_InformationList_RL_SetupRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_dl_PunctureLimit = -1;        /* PunctureLimit */
static int hf_rnsap_tdd_TPC_DownlinkStepSize = -1;  /* TDD_TPC_DownlinkStepSize */
static int hf_rnsap_cCTrCH_TPCList = -1;          /* CCTrCH_TPCList_RL_SetupRqstTDD */
static int hf_rnsap_CCTrCH_TPCList_RL_SetupRqstTDD_item = -1;  /* CCTrCH_TPCItem_RL_SetupRqstTDD */
static int hf_rnsap_specialBurstScheduling = -1;  /* SpecialBurstScheduling */
static int hf_rnsap_dL_TimeSlot_ISCP = -1;        /* DL_TimeSlot_ISCP_Info */
static int hf_rnsap_RL_InformationResponseList_RL_SetupRspFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_uRA_Information = -1;         /* URA_Information */
static int hf_rnsap_sAI = -1;                     /* SAI */
static int hf_rnsap_gA_Cell = -1;                 /* GA_Cell */
static int hf_rnsap_gA_AccessPointPosition = -1;  /* GA_AccessPointPosition */
static int hf_rnsap_not_Used_secondary_CCPCH_Info = -1;  /* NULL */
static int hf_rnsap_diversityIndication = -1;     /* DiversityIndication_RL_SetupRspFDD */
static int hf_rnsap_sSDT_SupportIndicator = -1;   /* SSDT_SupportIndicator */
static int hf_rnsap_closedlooptimingadjustmentmode = -1;  /* Closedlooptimingadjustmentmode */
static int hf_rnsap_maximumDLTxPower = -1;        /* DL_Power */
static int hf_rnsap_minimumDLTxPower = -1;        /* DL_Power */
static int hf_rnsap_not_Used_dSCHInformationResponse = -1;  /* NULL */
static int hf_rnsap_neighbouring_UMTS_CellInformation = -1;  /* Neighbouring_UMTS_CellInformation */
static int hf_rnsap_neighbouring_GSM_CellInformation = -1;  /* Neighbouring_GSM_CellInformation */
static int hf_rnsap_sRB_Delay = -1;               /* SRB_Delay */
static int hf_rnsap_combining = -1;               /* Combining_RL_SetupRspFDD */
static int hf_rnsap_nonCombiningOrFirstRL = -1;   /* NonCombiningOrFirstRL_RL_SetupRspFDD */
static int hf_rnsap_dCH_InformationResponse = -1;  /* DCH_InformationResponse */
static int hf_rnsap_Additional_HS_Cell_Information_Response_List_item = -1;  /* Additional_HS_Cell_Information_Response_ItemIEs */
static int hf_rnsap_ul_TimeSlot_ISCP_Info = -1;   /* UL_TimeSlot_ISCP_Info */
static int hf_rnsap_timingAdvanceApplied = -1;    /* TimingAdvanceApplied */
static int hf_rnsap_alphaValue = -1;              /* AlphaValue */
static int hf_rnsap_ul_PhysCH_SF_Variation = -1;  /* UL_PhysCH_SF_Variation */
static int hf_rnsap_synchronisationConfiguration = -1;  /* SynchronisationConfiguration */
static int hf_rnsap_secondary_CCPCH_Info_TDD = -1;  /* Secondary_CCPCH_Info_TDD */
static int hf_rnsap_ul_CCTrCHInformation = -1;    /* UL_CCTrCHInformationList_RL_SetupRspTDD */
static int hf_rnsap_dl_CCTrCHInformation = -1;    /* DL_CCTrCHInformationList_RL_SetupRspTDD */
static int hf_rnsap_dCH_InformationResponse_01 = -1;  /* DCH_InformationResponseList_RL_SetupRspTDD */
static int hf_rnsap_dsch_InformationResponse = -1;  /* DSCH_InformationResponse_RL_SetupRspTDD */
static int hf_rnsap_usch_InformationResponse = -1;  /* USCH_InformationResponse_RL_SetupRspTDD */
static int hf_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD_item = -1;  /* UL_CCTrCHInformationItem_RL_SetupRspTDD */
static int hf_rnsap_ul_DPCH_Information = -1;     /* UL_DPCH_InformationList_RL_SetupRspTDD */
static int hf_rnsap_tDD_DPCHOffset = -1;          /* TDD_DPCHOffset */
static int hf_rnsap_uL_Timeslot_Information = -1;  /* UL_Timeslot_Information */
static int hf_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD_item = -1;  /* DL_CCTrCHInformationItem_RL_SetupRspTDD */
static int hf_rnsap_dl_DPCH_Information = -1;     /* DL_DPCH_InformationList_RL_SetupRspTDD */
static int hf_rnsap_dL_Timeslot_Information = -1;  /* DL_Timeslot_Information */
static int hf_rnsap_DSCH_InformationListIEs_RL_SetupRspTDD_item = -1;  /* DSCHInformationItem_RL_SetupRspTDD */
static int hf_rnsap_dsch_ID = -1;                 /* DSCH_ID */
static int hf_rnsap_dSCH_FlowControlInformation = -1;  /* DSCH_FlowControlInformation */
static int hf_rnsap_transportFormatManagement = -1;  /* TransportFormatManagement */
static int hf_rnsap_USCH_InformationListIEs_RL_SetupRspTDD_item = -1;  /* USCHInformationItem_RL_SetupRspTDD */
static int hf_rnsap_usch_ID = -1;                 /* USCH_ID */
static int hf_rnsap_ul_TimeSlot_ISCP_LCR_Info = -1;  /* UL_TimeSlot_ISCP_LCR_Info */
static int hf_rnsap_secondary_LCR_CCPCH_Info_TDD = -1;  /* Secondary_LCR_CCPCH_Info_TDD */
static int hf_rnsap_ul_LCR_CCTrCHInformation = -1;  /* UL_LCR_CCTrCHInformationList_RL_SetupRspTDD */
static int hf_rnsap_dl_LCR_CCTrCHInformation = -1;  /* DL_LCR_CCTrCHInformationList_RL_SetupRspTDD */
static int hf_rnsap_dsch_LCR_InformationResponse = -1;  /* DSCH_LCR_InformationResponse_RL_SetupRspTDD */
static int hf_rnsap_usch_LCR_InformationResponse = -1;  /* USCH_LCR_InformationResponse_RL_SetupRspTDD */
static int hf_rnsap_UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD_item = -1;  /* UL_LCR_CCTrCHInformationItem_RL_SetupRspTDD */
static int hf_rnsap_ul_DPCH_LCR_Information = -1;  /* UL_DPCH_LCR_InformationList_RL_SetupRspTDD */
static int hf_rnsap_uL_TimeslotLCR_Information = -1;  /* UL_TimeslotLCR_Information */
static int hf_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD_item = -1;  /* DL_CCTrCH_LCR_InformationItem_RL_SetupRspTDD */
static int hf_rnsap_dl_DPCH_LCR_Information = -1;  /* DL_DPCH_LCR_InformationList_RL_SetupRspTDD */
static int hf_rnsap_dL_Timeslot_LCR_Information = -1;  /* DL_TimeslotLCR_Information */
static int hf_rnsap_tSTD_Indicator = -1;          /* TSTD_Indicator */
static int hf_rnsap_DSCH_LCR_InformationListIEs_RL_SetupRspTDD_item = -1;  /* DSCH_LCR_InformationItem_RL_SetupRspTDD */
static int hf_rnsap_USCH_LCR_InformationListIEs_RL_SetupRspTDD_item = -1;  /* USCH_LCR_InformationItem_RL_SetupRspTDD */
static int hf_rnsap_secondary_CCPCH_Info_TDD768 = -1;  /* Secondary_CCPCH_Info_TDD768 */
static int hf_rnsap_ul_CCTrCHInformation768 = -1;  /* UL_CCTrCHInformationList_RL_SetupRspTDD768 */
static int hf_rnsap_dl_CCTrCHInformation768 = -1;  /* DL_CCTrCHInformationList_RL_SetupRspTDD768 */
static int hf_rnsap_gA_CellAdditionalShapes = -1;  /* GA_CellAdditionalShapes */
static int hf_rnsap_hCS_Prio = -1;                /* HCS_Prio */
static int hf_rnsap_timeSlot_RL_SetupRspTDD = -1;  /* TimeSlot */
static int hf_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD768_item = -1;  /* UL_CCTrCHInformationItem_RL_SetupRspTDD768 */
static int hf_rnsap_ul_DPCH_Information768 = -1;  /* UL_DPCH_InformationList_RL_SetupRspTDD768 */
static int hf_rnsap_uL_SIR_Target_CCTrCH_InformationItem_RL_SetupRspTDD768 = -1;  /* UL_SIR */
static int hf_rnsap_uL_Timeslot_Information768 = -1;  /* UL_Timeslot_Information768 */
static int hf_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD768_item = -1;  /* DL_CCTrCHInformationItem_RL_SetupRspTDD768 */
static int hf_rnsap_dl_DPCH_Information768 = -1;  /* DL_DPCH_InformationList_RL_SetupRspTDD768 */
static int hf_rnsap_cCTrCH_Maximum_DL_Power = -1;  /* DL_Power */
static int hf_rnsap_cCTrCH_Minimum_DL_Power = -1;  /* DL_Power */
static int hf_rnsap_dL_Timeslot_Information768 = -1;  /* DL_Timeslot_Information768 */
static int hf_rnsap_generalCause = -1;            /* GeneralCauseList_RL_SetupFailureFDD */
static int hf_rnsap_rLSpecificCause = -1;         /* RLSpecificCauseList_RL_SetupFailureFDD */
static int hf_rnsap_unsuccessful_RL_InformationRespList_RL_SetupFailureFDD = -1;  /* UnsuccessfulRL_InformationResponseList_RL_SetupFailureFDD */
static int hf_rnsap_successful_RL_InformationRespList_RL_SetupFailureFDD = -1;  /* SuccessfulRL_InformationResponseList_RL_SetupFailureFDD */
static int hf_rnsap_UnsuccessfulRL_InformationResponseList_RL_SetupFailureFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_SuccessfulRL_InformationResponseList_RL_SetupFailureFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_diversityIndication_01 = -1;  /* DiversityIndication_RL_SetupFailureFDD */
static int hf_rnsap_not_Used_dSCH_InformationResponse_RL_SetupFailureFDD = -1;  /* NULL */
static int hf_rnsap_combining_01 = -1;            /* Combining_RL_SetupFailureFDD */
static int hf_rnsap_nonCombiningOrFirstRL_01 = -1;  /* NonCombiningOrFirstRL_RL_SetupFailureFDD */
static int hf_rnsap_generalCause_01 = -1;         /* GeneralCauseList_RL_SetupFailureTDD */
static int hf_rnsap_rLSpecificCause_01 = -1;      /* RLSpecificCauseList_RL_SetupFailureTDD */
static int hf_rnsap_unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD = -1;  /* Unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD */
static int hf_rnsap_RL_InformationList_RL_AdditionRqstFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_Additional_HS_Cell_Information_RL_Addition_List_item = -1;  /* Additional_HS_Cell_Information_RL_Addition_ItemIEs */
static int hf_rnsap_setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency = -1;  /* Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency */
static int hf_rnsap_addition = -1;                /* Additional_EDCH_Cell_Information_To_Add_List */
static int hf_rnsap_dL_TimeSlot_ISCP_Info = -1;   /* DL_TimeSlot_ISCP_Info */
static int hf_rnsap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_uplinkStepSizeLCR = -1;       /* TDD_TPC_UplinkStepSize_LCR */
static int hf_rnsap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_downlinkStepSize = -1;        /* TDD_TPC_DownlinkStepSize */
static int hf_rnsap_RL_InformationResponseList_RL_AdditionRspFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_dl_CodeInformation_01 = -1;   /* DL_CodeInformationList_RL_AdditionRspFDD */
static int hf_rnsap_diversityIndication_02 = -1;  /* DiversityIndication_RL_AdditionRspFDD */
static int hf_rnsap_combining_02 = -1;            /* Combining_RL_AdditionRspFDD */
static int hf_rnsap_nonCombining = -1;            /* NonCombining_RL_AdditionRspFDD */
static int hf_rnsap_Additional_HS_Cell_Change_Information_Response_List_item = -1;  /* Additional_HS_Cell_Change_Information_Response_ItemIEs */
static int hf_rnsap_hS_DSCH_Secondary_Serving_Cell_Change_Information_Response = -1;  /* HS_DSCH_Secondary_Serving_Cell_Change_Information_Response */
static int hf_rnsap_ul_CCTrCHInformation_01 = -1;  /* UL_CCTrCHInformationList_RL_AdditionRspTDD */
static int hf_rnsap_dl_CCTrCHInformation_01 = -1;  /* DL_CCTrCHInformationList_RL_AdditionRspTDD */
static int hf_rnsap_dCH_Information = -1;         /* DCH_Information_RL_AdditionRspTDD */
static int hf_rnsap_dSCH_InformationResponse = -1;  /* DSCH_InformationResponse_RL_AdditionRspTDD */
static int hf_rnsap_uSCH_InformationResponse = -1;  /* USCH_InformationResponse_RL_AdditionRspTDD */
static int hf_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD_item = -1;  /* UL_CCTrCHInformationItem_RL_AdditionRspTDD */
static int hf_rnsap_ul_DPCH_Information_01 = -1;  /* UL_DPCH_InformationList_RL_AdditionRspTDD */
static int hf_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD_item = -1;  /* DL_CCTrCHInformationItem_RL_AdditionRspTDD */
static int hf_rnsap_dl_DPCH_Information_01 = -1;  /* DL_DPCH_InformationList_RL_AdditionRspTDD */
static int hf_rnsap_diversityIndication_03 = -1;  /* DiversityIndication_RL_AdditionRspTDD */
static int hf_rnsap_combining_03 = -1;            /* Combining_RL_AdditionRspTDD */
static int hf_rnsap_nonCombining_01 = -1;         /* NonCombining_RL_AdditionRspTDD */
static int hf_rnsap_DSCH_InformationListIE_RL_AdditionRspTDD_item = -1;  /* DSCHInformationItem_RL_AdditionRspTDD */
static int hf_rnsap_diversityIndication_04 = -1;  /* DiversityIndication_RL_AdditionRspTDD2 */
static int hf_rnsap_USCH_InformationListIE_RL_AdditionRspTDD_item = -1;  /* USCHInformationItem_RL_AdditionRspTDD */
static int hf_rnsap_ul_CCTrCH_LCR_Information = -1;  /* UL_CCTrCH_LCR_InformationList_RL_AdditionRspTDD */
static int hf_rnsap_dl_CCTrCH_LCR_Information = -1;  /* DL_CCTrCH_LCR_InformationList_RL_AdditionRspTDD */
static int hf_rnsap_dCH_InformationResponse_02 = -1;  /* DCH_InformationResponseList_RL_AdditionRspTDD */
static int hf_rnsap_dsch_LCR_InformationResponse_01 = -1;  /* DSCH_LCR_InformationResponse_RL_AdditionRspTDD */
static int hf_rnsap_usch_LCR_InformationResponse_01 = -1;  /* USCH_LCR_InformationResponse_RL_AdditionRspTDD */
static int hf_rnsap_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_item = -1;  /* UL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD */
static int hf_rnsap_ul_DPCH_LCR_Information_01 = -1;  /* UL_DPCH_LCR_InformationList_RL_AdditionRspTDD */
static int hf_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_item = -1;  /* DL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD */
static int hf_rnsap_dl_DPCH_LCR_Information_01 = -1;  /* DL_DPCH_LCR_InformationList_RL_AdditionRspTDD */
static int hf_rnsap_dL_TimeslotLCR_Information = -1;  /* DL_TimeslotLCR_Information */
static int hf_rnsap_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD_item = -1;  /* DSCH_LCR_InformationItem_RL_AdditionRspTDD */
static int hf_rnsap_USCH_LCR_InformationListIEs_RL_AdditionRspTDD_item = -1;  /* USCH_LCR_InformationItem_RL_AdditionRspTDD */
static int hf_rnsap_ul_CCTrCHInformation768_01 = -1;  /* UL_CCTrCHInformationList_RL_AdditionRspTDD768 */
static int hf_rnsap_dl_CCTrCHInformation768_01 = -1;  /* DL_CCTrCHInformationList_RL_AdditionRspTDD768 */
static int hf_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD768_item = -1;  /* UL_CCTrCHInformationItem_RL_AdditionRspTDD768 */
static int hf_rnsap_ul_DPCH_Information768_01 = -1;  /* UL_DPCH_InformationList_RL_AdditionRspTDD768 */
static int hf_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD768_item = -1;  /* DL_CCTrCHInformationItem_RL_AdditionRspTDD768 */
static int hf_rnsap_dl_DPCH_Information768_01 = -1;  /* DL_DPCH_InformationList_RL_AdditionRspTDD768 */
static int hf_rnsap_generalCause_02 = -1;         /* GeneralCauseList_RL_AdditionFailureFDD */
static int hf_rnsap_rLSpecificCause_02 = -1;      /* RLSpecificCauseList_RL_AdditionFailureFDD */
static int hf_rnsap_unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD = -1;  /* UnsuccessfulRL_InformationResponseList_RL_AdditionFailureFDD */
static int hf_rnsap_successful_RL_InformationRespList_RL_AdditionFailureFDD = -1;  /* SuccessfulRL_InformationResponseList_RL_AdditionFailureFDD */
static int hf_rnsap_UnsuccessfulRL_InformationResponseList_RL_AdditionFailureFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_SuccessfulRL_InformationResponseList_RL_AdditionFailureFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_dl_CodeInformation_02 = -1;   /* DL_CodeInformationList_RL_AdditionFailureFDD */
static int hf_rnsap_diversityIndication_05 = -1;  /* DiversityIndication_RL_AdditionFailureFDD */
static int hf_rnsap_combining_04 = -1;            /* Combining_RL_AdditionFailureFDD */
static int hf_rnsap_nonCombining_02 = -1;         /* NonCombining_RL_AdditionFailureFDD */
static int hf_rnsap_generalCause_03 = -1;         /* GeneralCauseList_RL_AdditionFailureTDD */
static int hf_rnsap_rLSpecificCause_03 = -1;      /* RLSpecificCauseList_RL_AdditionFailureTDD */
static int hf_rnsap_unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD = -1;  /* Unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD */
static int hf_rnsap_RL_InformationList_RL_DeletionRqst_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_maxNrOfUL_DPDCHs = -1;        /* MaxNrOfUL_DPCHs */
static int hf_rnsap_not_Used_sSDT_CellIDLength = -1;  /* NULL */
static int hf_rnsap_powerOffsetInformation_02 = -1;  /* PowerOffsetInformation_RL_ReconfPrepFDD */
static int hf_rnsap_pO1_ForTFCI_Bits = -1;        /* PowerOffset */
static int hf_rnsap_pO2_ForTPC_Bits = -1;         /* PowerOffset */
static int hf_rnsap_pO3_ForPilotBits = -1;        /* PowerOffset */
static int hf_rnsap_DCH_DeleteList_RL_ReconfPrepFDD_item = -1;  /* DCH_DeleteItem_RL_ReconfPrepFDD */
static int hf_rnsap_RL_InformationList_RL_ReconfPrepFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_not_Used_sSDT_Indication = -1;  /* NULL */
static int hf_rnsap_not_Used_sSDT_CellIdentity = -1;  /* NULL */
static int hf_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep_item = -1;  /* Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs */
static int hf_rnsap_hS_DSCH_Secondary_Serving_Information_To_Modify = -1;  /* HS_DSCH_Secondary_Serving_Information_To_Modify */
static int hf_rnsap_hS_HS_DSCH_Secondary_Serving_Remove = -1;  /* HS_DSCH_Secondary_Serving_Remove */
static int hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency = -1;  /* Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency */
static int hf_rnsap_powerOffsetInformation_03 = -1;  /* PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD */
static int hf_rnsap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_cCTrCH_TPCList_01 = -1;       /* CCTrCH_TPCAddList_RL_ReconfPrepTDD */
static int hf_rnsap_CCTrCH_TPCAddList_RL_ReconfPrepTDD_item = -1;  /* CCTrCH_TPCAddItem_RL_ReconfPrepTDD */
static int hf_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_cCTrCH_TPCList_02 = -1;       /* CCTrCH_TPCModifyList_RL_ReconfPrepTDD */
static int hf_rnsap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD_item = -1;  /* CCTrCH_TPCModifyItem_RL_ReconfPrepTDD */
static int hf_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_DCH_DeleteList_RL_ReconfPrepTDD_item = -1;  /* DCH_DeleteItem_RL_ReconfPrepTDD */
static int hf_rnsap_DSCH_ModifyList_RL_ReconfPrepTDD_item = -1;  /* DSCH_ModifyItem_RL_ReconfPrepTDD */
static int hf_rnsap_DSCH_DeleteList_RL_ReconfPrepTDD_item = -1;  /* DSCH_DeleteItem_RL_ReconfPrepTDD */
static int hf_rnsap_USCH_ModifyList_RL_ReconfPrepTDD_item = -1;  /* USCH_ModifyItem_RL_ReconfPrepTDD */
static int hf_rnsap_ul_ccTrCHID = -1;             /* CCTrCH_ID */
static int hf_rnsap_USCH_DeleteList_RL_ReconfPrepTDD_item = -1;  /* USCH_DeleteItem_RL_ReconfPrepTDD */
static int hf_rnsap_RL_Information_RL_ReconfPrepTDD_item = -1;  /* RL_InformationIE_RL_ReconfPrepTDD */
static int hf_rnsap_rL_Specific_DCH_Info = -1;    /* RL_Specific_DCH_Info */
static int hf_rnsap_RL_InformationResponseList_RL_ReconfReadyFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_max_UL_SIR = -1;              /* UL_SIR */
static int hf_rnsap_min_UL_SIR = -1;              /* UL_SIR */
static int hf_rnsap_dl_CodeInformationList = -1;  /* DL_CodeInformationList_RL_ReconfReadyFDD */
static int hf_rnsap_dCHInformationResponse = -1;  /* DCH_InformationResponseList_RL_ReconfReadyFDD */
static int hf_rnsap_not_Used_dSCHsToBeAddedOrModified = -1;  /* NULL */
static int hf_rnsap_Additional_HS_Cell_RL_Reconf_Response_item = -1;  /* Additional_HS_Cell_RL_Reconf_Response_ItemIEs */
static int hf_rnsap_ul_CCTrCH_Information = -1;   /* UL_CCTrCH_InformationList_RL_ReconfReadyTDD */
static int hf_rnsap_dl_CCTrCH_Information = -1;   /* DL_CCTrCH_InformationList_RL_ReconfReadyTDD */
static int hf_rnsap_dCHInformationResponse_01 = -1;  /* DCH_InformationResponseList_RL_ReconfReadyTDD */
static int hf_rnsap_dSCHsToBeAddedOrModified = -1;  /* DSCHToBeAddedOrModified_RL_ReconfReadyTDD */
static int hf_rnsap_uSCHsToBeAddedOrModified = -1;  /* USCHToBeAddedOrModified_RL_ReconfReadyTDD */
static int hf_rnsap_UL_CCTrCHInformationListIE_RL_ReconfReadyTDD_item = -1;  /* UL_CCTrCH_InformationItem_RL_ReconfReadyTDD */
static int hf_rnsap_ul_DPCH_AddInformation = -1;  /* UL_DPCH_InformationAddList_RL_ReconfReadyTDD */
static int hf_rnsap_ul_DPCH_ModifyInformation = -1;  /* UL_DPCH_InformationModifyList_RL_ReconfReadyTDD */
static int hf_rnsap_ul_DPCH_DeleteInformation = -1;  /* UL_DPCH_InformationDeleteList_RL_ReconfReadyTDD */
static int hf_rnsap_uL_TimeslotLCR_Info = -1;     /* UL_TimeslotLCR_Information */
static int hf_rnsap_rxTimingDeviationForTA = -1;  /* RxTimingDeviationForTA */
static int hf_rnsap_rxTimingDeviationForTA768 = -1;  /* RxTimingDeviationForTA768 */
static int hf_rnsap_uL_Timeslot_InformationModifyList_RL_ReconfReadyTDD = -1;  /* UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD */
static int hf_rnsap_UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_item = -1;  /* UL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD */
static int hf_rnsap_tDD_uL_Code_LCR_Information = -1;  /* TDD_UL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD */
static int hf_rnsap_TDD_UL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD_item = -1;  /* TDD_UL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD */
static int hf_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD_item = -1;  /* UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD */
static int hf_rnsap_uL_Code_Information_01 = -1;  /* TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD */
static int hf_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD_item = -1;  /* TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD */
static int hf_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_item = -1;  /* UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768 */
static int hf_rnsap_uL_Code_Information768_01 = -1;  /* TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD768 */
static int hf_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD768_item = -1;  /* TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD768 */
static int hf_rnsap_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_item = -1;  /* UL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD */
static int hf_rnsap_DL_CCTrCHInformationListIE_RL_ReconfReadyTDD_item = -1;  /* DL_CCTrCH_InformationItem_RL_ReconfReadyTDD */
static int hf_rnsap_dl_DPCH_AddInformation = -1;  /* DL_DPCH_InformationAddList_RL_ReconfReadyTDD */
static int hf_rnsap_dl_DPCH_ModifyInformation = -1;  /* DL_DPCH_InformationModifyList_RL_ReconfReadyTDD */
static int hf_rnsap_dl_DPCH_DeleteInformation = -1;  /* DL_DPCH_InformationDeleteList_RL_ReconfReadyTDD */
static int hf_rnsap_dL_TimeslotLCR_Info = -1;     /* DL_TimeslotLCR_Information */
static int hf_rnsap_dL_Timeslot_InformationModifyList_RL_ReconfReadyTDD = -1;  /* DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD */
static int hf_rnsap_DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_item = -1;  /* DL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD */
static int hf_rnsap_tDD_dL_Code_LCR_Information = -1;  /* TDD_DL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD */
static int hf_rnsap_TDD_DL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD_item = -1;  /* TDD_DL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD */
static int hf_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD_item = -1;  /* DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD */
static int hf_rnsap_dL_Code_Information_01 = -1;  /* TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD */
static int hf_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD_item = -1;  /* TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD */
static int hf_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_item = -1;  /* DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768 */
static int hf_rnsap_dL_Code_Information768_01 = -1;  /* TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD768 */
static int hf_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD768_item = -1;  /* TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD768 */
static int hf_rnsap_dPCH_ID768 = -1;              /* DPCH_ID768 */
static int hf_rnsap_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_item = -1;  /* DL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD */
static int hf_rnsap_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD_item = -1;  /* DL_DPCH_InformationDeleteItem768_RL_ReconfReadyTDD */
static int hf_rnsap_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_item = -1;  /* DSCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD */
static int hf_rnsap_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_item = -1;  /* USCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD */
static int hf_rnsap_Multiple_RL_InformationResponse_RL_ReconfReadyTDD_item = -1;  /* RL_InformationResponse_RL_ReconfReadyTDD */
static int hf_rnsap_generalCause_04 = -1;         /* GeneralCauseList_RL_ReconfFailure */
static int hf_rnsap_rLSpecificCause_04 = -1;      /* RLSpecificCauseList_RL_ReconfFailure */
static int hf_rnsap_rL_ReconfigurationFailureList_RL_ReconfFailure = -1;  /* RL_ReconfigurationFailureList_RL_ReconfFailure */
static int hf_rnsap_RL_ReconfigurationFailureList_RL_ReconfFailure_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_DCH_DeleteList_RL_ReconfRqstFDD_item = -1;  /* DCH_DeleteItem_RL_ReconfRqstFDD */
static int hf_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req_item = -1;  /* Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs */
static int hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised = -1;  /* HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised */
static int hf_rnsap_hS_DSCH_Secondary_Serving_Remove = -1;  /* HS_DSCH_Secondary_Serving_Remove */
static int hf_rnsap_RL_ReconfigurationRequestFDD_RL_InformationList_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_DCH_DeleteList_RL_ReconfRqstTDD_item = -1;  /* DCH_DeleteItem_RL_ReconfRqstTDD */
static int hf_rnsap_Multiple_RL_ReconfigurationRequestTDD_RL_Information_item = -1;  /* RL_ReconfigurationRequestTDD_RL_Information */
static int hf_rnsap_RL_InformationResponseList_RL_ReconfRspFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_dCHsInformationResponseList = -1;  /* DCH_InformationResponseList_RL_ReconfRspFDD */
static int hf_rnsap_dL_CodeInformationList_RL_ReconfResp = -1;  /* DL_CodeInformationList_RL_ReconfRspFDD */
static int hf_rnsap_dCHsInformationResponseList_01 = -1;  /* DCH_InformationResponseList_RL_ReconfRspTDD */
static int hf_rnsap_DL_CCTrCH_InformationList_RL_ReconfRspTDD_item = -1;  /* DL_CCTrCH_InformationItem_RL_ReconfRspTDD */
static int hf_rnsap_dl_DPCH_ModifyInformation_LCR = -1;  /* DL_DPCH_InformationModifyList_LCR_RL_ReconfRspTDD */
static int hf_rnsap_dL_Timeslot_LCR_InformationModifyList_RL_ReconfRqstTDD = -1;  /* DL_Timeslot_LCR_InformationModifyList_RL_ReconfRspTDD */
static int hf_rnsap_DL_Timeslot_LCR_InformationModifyList_RL_ReconfRspTDD_item = -1;  /* DL_Timeslot_LCR_InformationModifyItem_RL_ReconfRspTDD */
static int hf_rnsap_maxPowerLCR = -1;             /* DL_Power */
static int hf_rnsap_minPowerLCR = -1;             /* DL_Power */
static int hf_rnsap_Multiple_RL_InformationResponse_RL_ReconfRspTDD_item = -1;  /* RL_InformationResponse_RL_ReconfRspTDD */
static int hf_rnsap_rL = -1;                      /* RL_RL_FailureInd */
static int hf_rnsap_rL_Set = -1;                  /* RL_Set_RL_FailureInd */
static int hf_rnsap_cCTrCH = -1;                  /* CCTrCH_RL_FailureInd */
static int hf_rnsap_rL_InformationList_RL_FailureInd = -1;  /* RL_InformationList_RL_FailureInd */
static int hf_rnsap_RL_InformationList_RL_FailureInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_rL_Set_InformationList_RL_FailureInd = -1;  /* RL_Set_InformationList_RL_FailureInd */
static int hf_rnsap_RL_Set_InformationList_RL_FailureInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_cCTrCH_InformationList_RL_FailureInd = -1;  /* CCTrCH_InformationList_RL_FailureInd */
static int hf_rnsap_CCTrCH_InformationList_RL_FailureInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_RL_InformationList_RL_PreemptRequiredInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_rL_01 = -1;                   /* RL_RL_RestoreInd */
static int hf_rnsap_rL_Set_01 = -1;               /* RL_Set_RL_RestoreInd */
static int hf_rnsap_cCTrCH_01 = -1;               /* CCTrCH_RL_RestoreInd */
static int hf_rnsap_rL_InformationList_RL_RestoreInd = -1;  /* RL_InformationList_RL_RestoreInd */
static int hf_rnsap_RL_InformationList_RL_RestoreInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_rL_Set_InformationList_RL_RestoreInd = -1;  /* RL_Set_InformationList_RL_RestoreInd */
static int hf_rnsap_RL_Set_InformationList_RL_RestoreInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_cCTrCH_InformationList_RL_RestoreInd = -1;  /* CCTrCH_InformationList_RL_RestoreInd */
static int hf_rnsap_CCTrCH_InformationList_RL_RestoreInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_DL_ReferencePowerInformationList_DL_PC_Rqst_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_dl_CodeInformation_03 = -1;   /* DL_CodeInformationList_PhyChReconfRqstFDD */
static int hf_rnsap_ul_CCTrCH_Information_01 = -1;  /* UL_CCTrCH_InformationList_PhyChReconfRqstTDD */
static int hf_rnsap_dl_CCTrCH_Information_01 = -1;  /* DL_CCTrCH_InformationList_PhyChReconfRqstTDD */
static int hf_rnsap_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_item = -1;  /* UL_CCTrCH_InformationItem_PhyChReconfRqstTDD */
static int hf_rnsap_ul_DPCH_Information_02 = -1;  /* UL_DPCH_InformationList_PhyChReconfRqstTDD */
static int hf_rnsap_uL_Timeslot_InformationList_PhyChReconfRqstTDD = -1;  /* UL_Timeslot_InformationList_PhyChReconfRqstTDD */
static int hf_rnsap_UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_item = -1;  /* UL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD */
static int hf_rnsap_uL_Code_LCR_Information = -1;  /* TDD_UL_Code_LCR_Information */
static int hf_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD_item = -1;  /* UL_Timeslot_InformationItem_PhyChReconfRqstTDD */
static int hf_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD768_item = -1;  /* UL_Timeslot_InformationItem_PhyChReconfRqstTDD768 */
static int hf_rnsap_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_item = -1;  /* DL_CCTrCH_InformationItem_PhyChReconfRqstTDD */
static int hf_rnsap_dl_DPCH_Information_02 = -1;  /* DL_DPCH_InformationList_PhyChReconfRqstTDD */
static int hf_rnsap_dL_Timeslot_InformationList_PhyChReconfRqstTDD = -1;  /* DL_Timeslot_InformationList_PhyChReconfRqstTDD */
static int hf_rnsap_DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_item = -1;  /* DL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD */
static int hf_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD_item = -1;  /* DL_Timeslot_InformationItem_PhyChReconfRqstTDD */
static int hf_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD768_item = -1;  /* DL_Timeslot_InformationItem_PhyChReconfRqstTDD768 */
static int hf_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD_item = -1;  /* HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD */
static int hf_rnsap_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD_item = -1;  /* HSPDSCH_Timeslot_InformationItemLCR_PhyChReconfRqstTDD */
static int hf_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768_item = -1;  /* HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD768 */
static int hf_rnsap_RL_InformationList_RL_CongestInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_dCH_Rate_Information = -1;    /* DCH_Rate_Information_RL_CongestInd */
static int hf_rnsap_DCH_Rate_Information_RL_CongestInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_allowed_Rate_Information = -1;  /* Allowed_Rate_Information */
static int hf_rnsap_EDCH_MacdFlowSpecificInformationList_RL_CongestInd_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_uRA = -1;                     /* URA_PagingRqst */
static int hf_rnsap_cell = -1;                    /* Cell_PagingRqst */
static int hf_rnsap_pagingCause = -1;             /* PagingCause */
static int hf_rnsap_cNDomainType = -1;            /* CNDomainType */
static int hf_rnsap_pagingRecordType = -1;        /* PagingRecordType */
static int hf_rnsap_rL_02 = -1;                   /* RL_DM_Rqst */
static int hf_rnsap_rLS = -1;                     /* RL_Set_DM_Rqst */
static int hf_rnsap_allRL = -1;                   /* All_RL_DM_Rqst */
static int hf_rnsap_allRLS = -1;                  /* All_RL_Set_DM_Rqst */
static int hf_rnsap_rL_InformationList_DM_Rqst = -1;  /* RL_InformationList_DM_Rqst */
static int hf_rnsap_RL_InformationList_DM_Rqst_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_HSSICH_Info_DM_Rqst_item = -1;  /* HS_SICH_ID */
static int hf_rnsap_HSSICH_Info_DM_Rqst_Extension_item = -1;  /* HS_SICH_ID_Extension */
static int hf_rnsap_rL_Set_InformationList_DM_Rqst = -1;  /* RL_Set_InformationList_DM_Rqst */
static int hf_rnsap_RL_Set_InformationList_DM_Rqst_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_rLs = -1;                     /* RL_DM_Rsp */
static int hf_rnsap_rLS_01 = -1;                  /* RL_Set_DM_Rsp */
static int hf_rnsap_allRL_01 = -1;                /* RL_DM_Rsp */
static int hf_rnsap_allRLS_01 = -1;               /* RL_Set_DM_Rsp */
static int hf_rnsap_rL_InformationList_DM_Rsp = -1;  /* RL_InformationList_DM_Rsp */
static int hf_rnsap_rL_Set_InformationList_DM_Rsp = -1;  /* RL_Set_InformationList_DM_Rsp */
static int hf_rnsap_RL_InformationList_DM_Rsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_dedicatedMeasurementValue = -1;  /* DedicatedMeasurementValue */
static int hf_rnsap_RL_Set_InformationList_DM_Rsp_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_item = -1;  /* Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp */
static int hf_rnsap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_item = -1;  /* Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp */
static int hf_rnsap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_item = -1;  /* Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp */
static int hf_rnsap_Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp_item = -1;  /* Multiple_DedicatedMeasurementValueItem_TDD768_DM_Rsp */
static int hf_rnsap_rL_03 = -1;                   /* RL_DM_Fail */
static int hf_rnsap_rLS_02 = -1;                  /* RL_Set_DM_Fail */
static int hf_rnsap_allRL_02 = -1;                /* RL_DM_Fail */
static int hf_rnsap_allRLS_02 = -1;               /* RL_Set_DM_Fail */
static int hf_rnsap_rL_unsuccessful_InformationRespList_DM_Fail = -1;  /* RL_Unsuccessful_InformationRespList_DM_Fail */
static int hf_rnsap_rL_successful_InformationRespList_DM_Fail = -1;  /* RL_Successful_InformationRespList_DM_Fail */
static int hf_rnsap_rL_Set_unsuccessful_InformationRespList_DM_Fail = -1;  /* RL_Set_Unsuccessful_InformationRespList_DM_Fail */
static int hf_rnsap_rL_Set_successful_InformationRespList_DM_Fail = -1;  /* RL_Set_Successful_InformationRespList_DM_Fail */
static int hf_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_individualcause = -1;         /* Cause */
static int hf_rnsap_RL_Successful_InformationRespList_DM_Fail_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_RL_Set_Successful_InformationRespList_DM_Fail_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_rLs_01 = -1;                  /* RL_DM_Rprt */
static int hf_rnsap_rLS_03 = -1;                  /* RL_Set_DM_Rprt */
static int hf_rnsap_allRL_03 = -1;                /* RL_DM_Rprt */
static int hf_rnsap_allRLS_03 = -1;               /* RL_Set_DM_Rprt */
static int hf_rnsap_rL_InformationList_DM_Rprt = -1;  /* RL_InformationList_DM_Rprt */
static int hf_rnsap_rL_Set_InformationList_DM_Rprt = -1;  /* RL_Set_InformationList_DM_Rprt */
static int hf_rnsap_RL_InformationList_DM_Rprt_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_dedicatedMeasurementValueInformation = -1;  /* DedicatedMeasurementValueInformation */
static int hf_rnsap_RL_Set_InformationList_DM_Rprt_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_rL_04 = -1;                   /* RL_DM_Fail_Ind */
static int hf_rnsap_rLS_04 = -1;                  /* RL_Set_DM_Fail_Ind */
static int hf_rnsap_allRL_04 = -1;                /* RL_DM_Fail_Ind */
static int hf_rnsap_allRLS_04 = -1;               /* RL_Set_DM_Fail_Ind */
static int hf_rnsap_rL_unsuccessful_InformationRespList_DM_Fail_Ind = -1;  /* RL_Unsuccessful_InformationRespList_DM_Fail_Ind */
static int hf_rnsap_rL_Set_unsuccessful_InformationRespList_DM_Fail_Ind = -1;  /* RL_Set_Unsuccessful_InformationRespList_DM_Fail_Ind */
static int hf_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail_Ind_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail_Ind_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_fACH_FlowControlInformation = -1;  /* FACH_FlowControlInformation_CTCH_ResourceRspFDD */
static int hf_rnsap_fACH_FlowControlInformation_01 = -1;  /* FACH_FlowControlInformation_CTCH_ResourceRspTDD */
static int hf_rnsap_cell_01 = -1;                 /* Cell_CM_Rqst */
static int hf_rnsap_additional_CommonMeasurementObjectType_CM_Rqst = -1;  /* Additional_CommonMeasurementObjectType_CM_Rqst */
static int hf_rnsap_neighbouringCellMeasurementInformation = -1;  /* NeighbouringCellMeasurementInfo */
static int hf_rnsap_NeighbouringCellMeasurementInfo_item = -1;  /* NeighbouringCellMeasurementInfo_item */
static int hf_rnsap_neighbouringFDDCellMeasurementInformation = -1;  /* NeighbouringFDDCellMeasurementInformation */
static int hf_rnsap_neighbouringTDDCellMeasurementInformation = -1;  /* NeighbouringTDDCellMeasurementInformation */
static int hf_rnsap_extension_neighbouringCellMeasurementInformation = -1;  /* Extension_neighbouringCellMeasurementInformation */
static int hf_rnsap_extension_neighbouringCellMeasurementInformation768 = -1;  /* Extension_neighbouringCellMeasurementInformation768 */
static int hf_rnsap_GsmCellList_CM_Rqst_item = -1;  /* GsmCellItem_CM_Rqst */
static int hf_rnsap_measurementID = -1;           /* MeasurementID */
static int hf_rnsap_gsmCell = -1;                 /* GSM_Cell_CM_Rqst */
static int hf_rnsap_cell_02 = -1;                 /* Cell_CM_Rsp */
static int hf_rnsap_GsmCellList_CM_Rsp_item = -1;  /* GsmCellItem_CM_Rsp */
static int hf_rnsap_Extension_FailureMeasurementList_item = -1;  /* Extension_FailureMeasurementItem */
static int hf_rnsap_cell_03 = -1;                 /* Cell_CM_Rprt */
static int hf_rnsap_commonMeasurementValueInformation = -1;  /* CommonMeasurementValueInformation */
static int hf_rnsap_GsmCellList_CM_Rprt_item = -1;  /* GsmCellItem_CM_Rprt */
static int hf_rnsap_Extension_TerminationMeasurementList_item = -1;  /* Extension_TerminationMeasurementItem */
static int hf_rnsap_Extension_FailureIndicationMeasurementList_item = -1;  /* Extension_FailureIndicationMeasurementItem */
static int hf_rnsap_cell_04 = -1;                 /* Cell_InfEx_Rqst */
static int hf_rnsap_extension_InformationExchangeObjectType_InfEx_Rqst = -1;  /* Extension_InformationExchangeObjectType_InfEx_Rqst */
static int hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst_item = -1;  /* MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rqst */
static int hf_rnsap_mBMS_Bearer_Service_List_InfEx_Rqst = -1;  /* MBMS_Bearer_Service_List_InfEx_Rqst */
static int hf_rnsap_MBMS_Bearer_Service_List_InfEx_Rqst_item = -1;  /* TMGI */
static int hf_rnsap_MBMS_Cell_InfEx_Rqst_item = -1;  /* C_ID */
static int hf_rnsap_ANR_Cell_InfEx_Rqst_item = -1;  /* C_ID */
static int hf_rnsap_Common_E_RGCH_Cell_InfEx_Rqst_item = -1;  /* C_ID */
static int hf_rnsap_cell_05 = -1;                 /* Cell_InfEx_Rsp */
static int hf_rnsap_extension_InformationExchangeObjectType_InfEx_Rsp = -1;  /* Extension_InformationExchangeObjectType_InfEx_Rsp */
static int hf_rnsap_MBMS_Bearer_Service_List_InfEx_Rsp_item = -1;  /* MBMS_Bearer_ServiceItemIEs_InfEx_Rsp */
static int hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp_item = -1;  /* MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rsp */
static int hf_rnsap_mBMS_Bearer_Service_List_InfEx_Rsp = -1;  /* MBMS_Bearer_Service_List_InfEx_Rsp */
static int hf_rnsap_MBMS_Cell_InfEx_Rsp_item = -1;  /* MBMS_Cell_Item_InfEx_Rsp */
static int hf_rnsap_ANR_Cell_InfEx_Rsp_item = -1;  /* ANR_Cell_ItemIEs_InfEx_Rsp */
static int hf_rnsap_Common_E_RGCH_Cell_InfEx_Rsp_item = -1;  /* Common_E_RGCH_Cell_ItemIEs_InfEx_Rsp */
static int hf_rnsap_cell_06 = -1;                 /* Cell_InfEx_Rprt */
static int hf_rnsap_extension_InformationExchangeObjectType_InfEx_Rprt = -1;  /* Extension_InformationExchangeObjectType_InfEx_Rprt */
static int hf_rnsap_requestedDataValueInformation = -1;  /* RequestedDataValueInformation */
static int hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt_item = -1;  /* MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rprt */
static int hf_rnsap_mBMS_Bearer_Service_List_InfEx_Rprt = -1;  /* MBMS_Bearer_Service_List_InfEx_Rprt */
static int hf_rnsap_MBMS_Bearer_Service_List_InfEx_Rprt_item = -1;  /* MBMS_Bearer_Service_List_Item_InfEx_Rprt */
static int hf_rnsap_MBMS_Cell_InfEx_Rprt_item = -1;  /* MBMS_Cell_Item_InfEx_Rprt */
static int hf_rnsap_Common_E_RGCH_Cell_InfEx_Rprt_item = -1;  /* Common_E_RGCH_Cell_Item_InfEx_Rprt */
static int hf_rnsap_context = -1;                 /* ContextList_Reset */
static int hf_rnsap_all_contexts = -1;            /* NULL */
static int hf_rnsap_contextGroup = -1;            /* ContextGroupList_Reset */
static int hf_rnsap_contextInfoList_Reset = -1;   /* ContextInfoList_Reset */
static int hf_rnsap_ContextInfoList_Reset_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_contextType_Reset = -1;       /* ContextType_Reset */
static int hf_rnsap_dRNTI = -1;                   /* D_RNTI */
static int hf_rnsap_extension_ContextType_Reset = -1;  /* Extension_ContextType_Reset */
static int hf_rnsap_contextGroupInfoList_Reset = -1;  /* ContextGroupInfoList_Reset */
static int hf_rnsap_ContextGroupInfoList_Reset_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_s_RNTI_Group = -1;            /* S_RNTI_Group */
static int hf_rnsap_DelayedActivationInformationList_RL_ActivationCmdFDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_delayed_activation_update = -1;  /* DelayedActivationUpdate */
static int hf_rnsap_DelayedActivationInformationList_RL_ActivationCmdTDD_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_RL_ParameterUpdateIndicationFDD_RL_InformationList_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_uE_Measurement_Value = -1;    /* UE_Measurement_Value */
static int hf_rnsap_Additional_HS_Cell_Information_RL_Param_Upd_item = -1;  /* Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs */
static int hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Update_Information = -1;  /* HS_DSCH_FDD_Secondary_Serving_Update_Information */
static int hf_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd_item = -1;  /* Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs */
static int hf_rnsap_additional_EDCH_FDD_Update_Information = -1;  /* Additional_EDCH_FDD_Update_Information */
static int hf_rnsap_ListOfInterfacesToTrace_item = -1;  /* ProtocolIE_Single_Container */
static int hf_rnsap_interface = -1;               /* T_interface */
static int hf_rnsap_privateIEs = -1;              /* PrivateIE_Container */
static int hf_rnsap_suspension = -1;              /* Suspension_Control_Type */
static int hf_rnsap_resume = -1;                  /* Resume_Control_Type */
static int hf_rnsap_controlled_Object_Scope = -1;  /* Controlled_Object_Scope */
static int hf_rnsap_initiatingMessage = -1;       /* InitiatingMessage */
static int hf_rnsap_successfulOutcome = -1;       /* SuccessfulOutcome */
static int hf_rnsap_unsuccessfulOutcome = -1;     /* UnsuccessfulOutcome */
static int hf_rnsap_outcome = -1;                 /* Outcome */
static int hf_rnsap_value_02 = -1;                /* InitiatingMessage_value */
static int hf_rnsap_value_03 = -1;                /* SuccessfulOutcome_value */
static int hf_rnsap_value_04 = -1;                /* UnsuccessfulOutcome_value */
static int hf_rnsap_value_05 = -1;                /* Outcome_value */

/* Initialize the subtree pointers */
static int ett_rnsap = -1;
static int ett_rnsap_transportLayerAddress = -1;
static int ett_rnsap_transportLayerAddress_nsap = -1;

static gint ett_rnsap_PrivateIE_ID = -1;
static gint ett_rnsap_ProcedureID = -1;
static gint ett_rnsap_TransactionID = -1;
static gint ett_rnsap_ProtocolIE_Container = -1;
static gint ett_rnsap_ProtocolIE_Field = -1;
static gint ett_rnsap_ProtocolExtensionContainer = -1;
static gint ett_rnsap_ProtocolExtensionField = -1;
static gint ett_rnsap_PrivateIE_Container = -1;
static gint ett_rnsap_PrivateIE_Field = -1;
static gint ett_rnsap_ActivationInformation = -1;
static gint ett_rnsap_ActivationInformationItem = -1;
static gint ett_rnsap_Active_MBMS_Bearer_Service_ListFDD = -1;
static gint ett_rnsap_Active_MBMS_Bearer_Service_ListFDD_PFL = -1;
static gint ett_rnsap_Active_MBMS_Bearer_Service_ListTDD = -1;
static gint ett_rnsap_Active_MBMS_Bearer_Service_ListTDD_PFL = -1;
static gint ett_rnsap_Active_Pattern_Sequence_Information = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_Response_RLAddList = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_Response_RLAdd_ItemIEs = -1;
static gint ett_rnsap_Additional_EDCH_Setup_Info = -1;
static gint ett_rnsap_ANRReportIndication = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_Setup = -1;
static gint ett_rnsap_Additional_EDCH_FDD_Setup_Cell_Information = -1;
static gint ett_rnsap_Additional_EDCH_UL_DPCH_Information_Setup = -1;
static gint ett_rnsap_Additional_EDCH_F_DPCH_Information = -1;
static gint ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Setup_List = -1;
static gint ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_To_Add_List = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_To_Add_ItemIEs = -1;
static gint ett_rnsap_Additional_EDCH_UL_DPCH_Information_Addition = -1;
static gint ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_List = -1;
static gint ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs = -1;
static gint ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Modify_List = -1;
static gint ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs = -1;
static gint ett_rnsap_Additional_EDCH_FDD_Information = -1;
static gint ett_rnsap_DTX_Information2 = -1;
static gint ett_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_List = -1;
static gint ett_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info = -1;
static gint ett_rnsap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List = -1;
static gint ett_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_Response = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_Response_List = -1;
static gint ett_rnsap_Additional_EDCH_FDD_Information_Response_ItemIEs = -1;
static gint ett_rnsap_EDCH_Additional_RL_Specific_Information_Response_List = -1;
static gint ett_rnsap_EDCH_Additional_RL_Specific_Information_Response_ItemIEs = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_Response_RLReconf_List = -1;
static gint ett_rnsap_Additional_EDCH_FDD_Information_Response_RLReconf_Items = -1;
static gint ett_rnsap_EDCH_Additional_RL_Specific_Modified_Information_Response_List = -1;
static gint ett_rnsap_EDCH_Additional_RL_Specific_Modified_Information_Response_ItemIEs = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_ConfigurationChange_List = -1;
static gint ett_rnsap_Additional_EDCH_ConfigurationChange_Info_ItemIEs = -1;
static gint ett_rnsap_Additional_EDCH_UL_DPCH_Information_Modify = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_Removal_List = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_Removal_Info_ItemIEs = -1;
static gint ett_rnsap_Additional_EDCH_FDD_Update_Information = -1;
static gint ett_rnsap_Additional_EDCH_DL_Control_Channel_Change_Information_List = -1;
static gint ett_rnsap_Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs = -1;
static gint ett_rnsap_AdditionalPreferredFrequency = -1;
static gint ett_rnsap_AdditionalPreferredFrequencyItem = -1;
static gint ett_rnsap_AffectedUEInformationForMBMS = -1;
static gint ett_rnsap_AllocationRetentionPriority = -1;
static gint ett_rnsap_Allowed_Rate_Information = -1;
static gint ett_rnsap_Angle_Of_Arrival_Value_LCR = -1;
static gint ett_rnsap_ANR_Cell_Information = -1;
static gint ett_rnsap_ANR_FDD_CellInformation = -1;
static gint ett_rnsap_ANR_TDD_CellInformation = -1;
static gint ett_rnsap_ANR_LCR_TDD_CellInformation = -1;
static gint ett_rnsap_ANR_Multiple_PLMN_List = -1;
static gint ett_rnsap_BadSatellites = -1;
static gint ett_rnsap_T_badSatelliteInformation = -1;
static gint ett_rnsap_T_badSatelliteInformation_item = -1;
static gint ett_rnsap_BDS_IGPInfoList = -1;
static gint ett_rnsap_BDS_IGPInfo = -1;
static gint ett_rnsap_BDS_Ionospheric_Grid_Model = -1;
static gint ett_rnsap_BSIC = -1;
static gint ett_rnsap_BurstModeParameters = -1;
static gint ett_rnsap_Cause = -1;
static gint ett_rnsap_CellBased = -1;
static gint ett_rnsap_CellIdList = -1;
static gint ett_rnsap_Cell_Capacity_Class_Value = -1;
static gint ett_rnsap_CGI = -1;
static gint ett_rnsap_CommonMeasurementAccuracy = -1;
static gint ett_rnsap_CommonMeasurementValue = -1;
static gint ett_rnsap_CommonMeasurementValueInformation = -1;
static gint ett_rnsap_CommonMeasurementAvailable = -1;
static gint ett_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDD = -1;
static gint ett_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDDItem = -1;
static gint ett_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationLCR = -1;
static gint ett_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationItemLCR = -1;
static gint ett_rnsap_Common_E_DCH_LogicalChannelInformation = -1;
static gint ett_rnsap_Common_E_DCH_LogicalChannelInformationItem = -1;
static gint ett_rnsap_Common_E_RGCH_Cell_Information = -1;
static gint ett_rnsap_Common_E_RGCH_Cell_Info = -1;
static gint ett_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information = -1;
static gint ett_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information_to_Modify = -1;
static gint ett_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information = -1;
static gint ett_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_InformationItem = -1;
static gint ett_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response = -1;
static gint ett_rnsap_CorrespondingCells = -1;
static gint ett_rnsap_CPC_Information = -1;
static gint ett_rnsap_Counting_Information = -1;
static gint ett_rnsap_Counting_Information_List = -1;
static gint ett_rnsap_CriticalityDiagnostics = -1;
static gint ett_rnsap_CriticalityDiagnostics_IE_List = -1;
static gint ett_rnsap_CriticalityDiagnostics_IE_List_item = -1;
static gint ett_rnsap_MessageStructure = -1;
static gint ett_rnsap_MessageStructure_item = -1;
static gint ett_rnsap_CN_CS_DomainIdentifier = -1;
static gint ett_rnsap_CN_PS_DomainIdentifier = -1;
static gint ett_rnsap_CPC_InformationLCR = -1;
static gint ett_rnsap_ContinuousPacketConnectivity_DRX_InformationLCR = -1;
static gint ett_rnsap_HS_SCCH_DRX_Information_LCR = -1;
static gint ett_rnsap_E_AGCH_DRX_Information_LCR = -1;
static gint ett_rnsap_E_AGCH_DRX_Parameters = -1;
static gint ett_rnsap_ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR = -1;
static gint ett_rnsap_DRX_Information_to_Modify_LCR = -1;
static gint ett_rnsap_DRX_Information_to_Modify_Items_LCR = -1;
static gint ett_rnsap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR = -1;
static gint ett_rnsap_HS_SCCH_DRX_Information_ResponseLCR = -1;
static gint ett_rnsap_E_AGCH_DRX_Information_ResponseLCR = -1;
static gint ett_rnsap_E_AGCH_DRX_Parameters_Response = -1;
static gint ett_rnsap_Controlled_Object_Scope = -1;
static gint ett_rnsap_DBDS_CorrectionsReq = -1;
static gint ett_rnsap_DBDS_Corrections = -1;
static gint ett_rnsap_DBDS_Info = -1;
static gint ett_rnsap_DBDS_SignalInfoList = -1;
static gint ett_rnsap_DBDS_SignalInfo = -1;
static gint ett_rnsap_DBDS_InfoList = -1;
static gint ett_rnsap_DCH_ENH_Information = -1;
static gint ett_rnsap_DCH_ENH_Information_to_Modify = -1;
static gint ett_rnsap_DCH_ENH_Concat = -1;
static gint ett_rnsap_DCH_ENH_Information_Reconf = -1;
static gint ett_rnsap_DCH_FDD_Information = -1;
static gint ett_rnsap_DCH_FDD_InformationItem = -1;
static gint ett_rnsap_DCH_MeasurementOccasion_Information = -1;
static gint ett_rnsap_DchMeasurementOccasionInformation_Item = -1;
static gint ett_rnsap_Measurement_Occasion_Pattern_Sequence_parameters = -1;
static gint ett_rnsap_DCH_Specific_FDD_InformationList = -1;
static gint ett_rnsap_DCH_Specific_FDD_Item = -1;
static gint ett_rnsap_DCH_InformationResponse = -1;
static gint ett_rnsap_DCH_InformationResponseItem = -1;
static gint ett_rnsap_DCH_TDD_Information = -1;
static gint ett_rnsap_DCH_TDD_InformationItem = -1;
static gint ett_rnsap_DCH_Specific_TDD_InformationList = -1;
static gint ett_rnsap_DCH_Specific_TDD_Item = -1;
static gint ett_rnsap_DedicatedMeasurementValue = -1;
static gint ett_rnsap_DedicatedMeasurementValueInformation = -1;
static gint ett_rnsap_DedicatedMeasurementAvailable = -1;
static gint ett_rnsap_DelayedActivation = -1;
static gint ett_rnsap_DelayedActivationUpdate = -1;
static gint ett_rnsap_Activate_Info = -1;
static gint ett_rnsap_Deactivate_Info = -1;
static gint ett_rnsap_Execution_Type = -1;
static gint ett_rnsap_DGANSSCorrections = -1;
static gint ett_rnsap_T_dGANSS_Information = -1;
static gint ett_rnsap_T_dGANSS_Information_item = -1;
static gint ett_rnsap_T_dGANSS_SignalInformation = -1;
static gint ett_rnsap_T_dGANSS_SignalInformation_item = -1;
static gint ett_rnsap_DGANSS_Corrections_Req = -1;
static gint ett_rnsap_DGANSSThreshold = -1;
static gint ett_rnsap_DGNSS_ValidityPeriod = -1;
static gint ett_rnsap_DGPSCorrections = -1;
static gint ett_rnsap_T_satellite_DGPSCorrections_Information = -1;
static gint ett_rnsap_T_satellite_DGPSCorrections_Information_item = -1;
static gint ett_rnsap_DGPSThreshold = -1;
static gint ett_rnsap_DL_TBS = -1;
static gint ett_rnsap_HS_DSCH_Cell_List = -1;
static gint ett_rnsap_HS_DSCH_Cell = -1;
static gint ett_rnsap_DL_PowerBalancing_Information = -1;
static gint ett_rnsap_DL_ReferencePowerInformationList = -1;
static gint ett_rnsap_DL_ReferencePowerInformationItem = -1;
static gint ett_rnsap_DL_ReferencePowerInformation = -1;
static gint ett_rnsap_DL_Timeslot_Information = -1;
static gint ett_rnsap_DL_Timeslot_InformationItem = -1;
static gint ett_rnsap_DL_TimeslotLCR_Information = -1;
static gint ett_rnsap_DL_TimeslotLCR_InformationItem = -1;
static gint ett_rnsap_DL_Timeslot_Information768 = -1;
static gint ett_rnsap_DL_Timeslot_InformationItem768 = -1;
static gint ett_rnsap_DL_TimeSlot_ISCP_Info = -1;
static gint ett_rnsap_DL_TimeSlot_ISCP_InfoItem = -1;
static gint ett_rnsap_DL_TimeSlot_ISCP_LCR_Information = -1;
static gint ett_rnsap_DL_TimeSlot_ISCP_LCR_InfoItem = -1;
static gint ett_rnsap_Downlink_TPC_enhancements_Information = -1;
static gint ett_rnsap_Downlink_TPC_enhancements_Reconf = -1;
static gint ett_rnsap_DRX_Information = -1;
static gint ett_rnsap_DRX_Information_to_Modify = -1;
static gint ett_rnsap_DRX_Information_to_Modify_Items = -1;
static gint ett_rnsap_DSCH_FlowControlInformation = -1;
static gint ett_rnsap_DSCH_FlowControlItem = -1;
static gint ett_rnsap_DSCH_TDD_Information = -1;
static gint ett_rnsap_DSCH_TDD_InformationItem = -1;
static gint ett_rnsap_DTX_Cycle_2ms_Items = -1;
static gint ett_rnsap_DTX_Cycle_2ms_to_Modify_Items = -1;
static gint ett_rnsap_DTX_Cycle_10ms_Items = -1;
static gint ett_rnsap_DTX_Cycle_10ms_to_Modify_Items = -1;
static gint ett_rnsap_DTX_Information = -1;
static gint ett_rnsap_DTX_Information_to_Modify = -1;
static gint ett_rnsap_DTX_Information_to_Modify_Items = -1;
static gint ett_rnsap_EARFCN_Information = -1;
static gint ett_rnsap_EARFCN_FDD = -1;
static gint ett_rnsap_EARFCN_FDD_Extended = -1;
static gint ett_rnsap_ECGI = -1;
static gint ett_rnsap_EDCH_FDD_DL_ControlChannelInformation = -1;
static gint ett_rnsap_EDCH_FDD_Information = -1;
static gint ett_rnsap_EDCH_FDD_InformationResponse = -1;
static gint ett_rnsap_EDCH_MACdFlow_Specific_InformationResponse = -1;
static gint ett_rnsap_EDCH_MACdFlow_Specific_InformationResponseItem = -1;
static gint ett_rnsap_EDCH_FDD_Information_To_Modify = -1;
static gint ett_rnsap_E_DCH_FDD_Update_Information = -1;
static gint ett_rnsap_E_DCH_MACdFlow_Specific_UpdateInformation = -1;
static gint ett_rnsap_E_DCH_MACdFlow_Specific_UpdateInformation_Item = -1;
static gint ett_rnsap_E_DCH_DL_Control_Channel_Change_Information = -1;
static gint ett_rnsap_E_DCH_DL_Control_Channel_Change_Information_Item = -1;
static gint ett_rnsap_E_DCH_DL_Control_Channel_Grant_Information = -1;
static gint ett_rnsap_E_DCH_DL_Control_Channel_Grant_Information_Item = -1;
static gint ett_rnsap_E_DCH_Grant_Type_Information = -1;
static gint ett_rnsap_E_DCH_LogicalChannelInformation = -1;
static gint ett_rnsap_E_DCH_LogicalChannelInformationItem = -1;
static gint ett_rnsap_E_DCH_MACdPDU_SizeList = -1;
static gint ett_rnsap_E_DCH_MACdPDU_SizeListItem = -1;
static gint ett_rnsap_E_DCH_LogicalChannelToModify = -1;
static gint ett_rnsap_E_DCH_LogicalChannelToModifyItem = -1;
static gint ett_rnsap_E_DCH_MACdPDU_SizeToModifyList = -1;
static gint ett_rnsap_E_DCH_LogicalChannelToDelete = -1;
static gint ett_rnsap_E_DCH_LogicalChannelToDeleteItem = -1;
static gint ett_rnsap_EDCH_MACdFlows_Information = -1;
static gint ett_rnsap_EDCH_MACdFlow_Specific_InfoList = -1;
static gint ett_rnsap_EDCH_MACdFlow_Specific_InfoItem = -1;
static gint ett_rnsap_EDCH_MACdFlow_Specific_InfoToModifyList = -1;
static gint ett_rnsap_EDCH_MACdFlow_Specific_InfoToModifyItem = -1;
static gint ett_rnsap_EDCH_MACdFlows_To_Delete = -1;
static gint ett_rnsap_EDCH_MACdFlows_To_Delete_Item = -1;
static gint ett_rnsap_E_DCH_Non_Scheduled_Transmission_Grant_Items = -1;
static gint ett_rnsap_E_DCH_Serving_cell_change_informationResponse = -1;
static gint ett_rnsap_E_DCH_serving_cell_change_choice = -1;
static gint ett_rnsap_E_DCH_serving_cell_change_successful = -1;
static gint ett_rnsap_E_DCH_RL_InformationList_Rsp = -1;
static gint ett_rnsap_E_DCH_RL_InformationList_Rsp_Item = -1;
static gint ett_rnsap_E_DCH_serving_cell_change_unsuccessful = -1;
static gint ett_rnsap_E_DCH_TTI_Length = -1;
static gint ett_rnsap_E_DCH_TTI_Length_to_Modify = -1;
static gint ett_rnsap_EDPCH_Information_FDD = -1;
static gint ett_rnsap_EDPCH_Information_RLReconfRequest_FDD = -1;
static gint ett_rnsap_EDCH_Serving_RL = -1;
static gint ett_rnsap_EDCH_Serving_RL_in_this_DRNS = -1;
static gint ett_rnsap_Enhanced_FACH_Information_ResponseFDD = -1;
static gint ett_rnsap_Enhanced_FACH_Information_ResponseLCR = -1;
static gint ett_rnsap_E_TFCI_Boost_Information = -1;
static gint ett_rnsap_E_TFCS_Information = -1;
static gint ett_rnsap_EventA = -1;
static gint ett_rnsap_EventB = -1;
static gint ett_rnsap_EventC = -1;
static gint ett_rnsap_EventD = -1;
static gint ett_rnsap_EventE = -1;
static gint ett_rnsap_EventF = -1;
static gint ett_rnsap_EventH = -1;
static gint ett_rnsap_Event1F_Parameters = -1;
static gint ett_rnsap_Event1I_Parameters = -1;
static gint ett_rnsap_E_DCH_Information = -1;
static gint ett_rnsap_E_PUCH_Information = -1;
static gint ett_rnsap_E_TFCS_Information_TDD = -1;
static gint ett_rnsap_E_DCH_QPSK_RefBetaInfo = -1;
static gint ett_rnsap_E_DCH_sixteenQAM_RefBetaInfo = -1;
static gint ett_rnsap_E_DCH_RefBeta_Item = -1;
static gint ett_rnsap_E_DCH_MACdFlows_Information_TDD = -1;
static gint ett_rnsap_E_DCH_MACdFlow_InfoTDDItem = -1;
static gint ett_rnsap_E_DCH_TDD_Information = -1;
static gint ett_rnsap_E_DCH_Information_Reconfig = -1;
static gint ett_rnsap_E_DCH_TDD_Information_to_Modify = -1;
static gint ett_rnsap_E_DCH_TDD_Information_to_Modify_List = -1;
static gint ett_rnsap_E_DCH_MACdFlow_ModifyTDDItem = -1;
static gint ett_rnsap_E_DCH_Information_Response = -1;
static gint ett_rnsap_Fast_TTI_switching_Mode_synchronized = -1;
static gint ett_rnsap_Fast_TTI_switching_Mode_unsynchronized = -1;
static gint ett_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp = -1;
static gint ett_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp_Item = -1;
static gint ett_rnsap_E_AGCH_Specific_InformationRespListTDD = -1;
static gint ett_rnsap_E_AGCH_Specific_InformationResp_ItemTDD = -1;
static gint ett_rnsap_E_HICH_InformationResp = -1;
static gint ett_rnsap_E_DCH_Non_Scheduled_Grant_Info = -1;
static gint ett_rnsap_E_DCH_768_Information = -1;
static gint ett_rnsap_E_DCH_TDD_Information768 = -1;
static gint ett_rnsap_E_DCH_768_Information_Reconfig = -1;
static gint ett_rnsap_E_DCH_768_Information_Response = -1;
static gint ett_rnsap_E_AGCH_Specific_InformationRespList768TDD = -1;
static gint ett_rnsap_E_AGCH_Specific_InformationResp_Item768TDD = -1;
static gint ett_rnsap_E_HICH_InformationResp768 = -1;
static gint ett_rnsap_E_DCH_Non_Scheduled_Grant_Info768 = -1;
static gint ett_rnsap_E_DCH_LCR_Information = -1;
static gint ett_rnsap_E_PUCH_LCR_Information = -1;
static gint ett_rnsap_E_DCH_LCR_TDD_Information = -1;
static gint ett_rnsap_E_DCH_LCR_Information_Reconfig = -1;
static gint ett_rnsap_E_DCH_LCR_Information_Response = -1;
static gint ett_rnsap_E_AGCH_Specific_InformationRespList_LCR_TDD = -1;
static gint ett_rnsap_E_AGCH_Specific_InformationResp_Item_LCR_TDD = -1;
static gint ett_rnsap_E_HICH_Specific_InformationResp_LCR = -1;
static gint ett_rnsap_E_HICH_Scheduled_InformationRespList_LCR_TDD = -1;
static gint ett_rnsap_E_HICH_Scheduled_InformationResp_Item_LCR_TDD = -1;
static gint ett_rnsap_E_HICH_InformationResp_LCR = -1;
static gint ett_rnsap_E_DCH_Non_Scheduled_Grant_Info_LCR = -1;
static gint ett_rnsap_Radio_Links_without_DPCH_FDPCH_Indication = -1;
static gint ett_rnsap_Radio_Links_without_DPCH_FDPCH_Information = -1;
static gint ett_rnsap_Extended_S_RNTI_Group = -1;
static gint ett_rnsap_ExtendedAffectedUEInformationForMBMS = -1;
static gint ett_rnsap_E_DCH_Semi_PersistentScheduling_Information_LCR = -1;
static gint ett_rnsap_E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR = -1;
static gint ett_rnsap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR = -1;
static gint ett_rnsap_Initial_E_DCH_SPS_resource = -1;
static gint ett_rnsap_E_DCH_SPS_HICH_Information = -1;
static gint ett_rnsap_E_HICH_Configuration = -1;
static gint ett_rnsap_Same_As_Scheduled_E_HICH = -1;
static gint ett_rnsap_E_HICH_InformationResp_ExplicitConfiguration_LCR = -1;
static gint ett_rnsap_FACH_FlowControlInformation = -1;
static gint ett_rnsap_FACH_FlowControlInformationItem = -1;
static gint ett_rnsap_FACH_InformationList = -1;
static gint ett_rnsap_FACH_InformationItem = -1;
static gint ett_rnsap_FDD_DCHs_to_Modify = -1;
static gint ett_rnsap_FDD_DCHs_to_ModifyItem = -1;
static gint ett_rnsap_FDD_DCHs_to_ModifySpecificInformationList = -1;
static gint ett_rnsap_FDD_DCHs_to_ModifySpecificItem = -1;
static gint ett_rnsap_FDD_DL_CodeInformation = -1;
static gint ett_rnsap_FDD_DL_CodeInformationItem = -1;
static gint ett_rnsap_F_DPCH_Info = -1;
static gint ett_rnsap_F_DPCH_Info_To_Modify = -1;
static gint ett_rnsap_FPACH_Information = -1;
static gint ett_rnsap_FTPICH_Information = -1;
static gint ett_rnsap_FTPICH_Information_To_Modify = -1;
static gint ett_rnsap_FTPICH_Information_Reconf = -1;
static gint ett_rnsap_FTPICH_Reconfiguration_Information = -1;
static gint ett_rnsap_FTPICH_Information_Response = -1;
static gint ett_rnsap_GA_Cell = -1;
static gint ett_rnsap_GA_Cell_item = -1;
static gint ett_rnsap_GA_CellAdditionalShapes = -1;
static gint ett_rnsap_GA_AltitudeAndDirection = -1;
static gint ett_rnsap_GA_EllipsoidArc = -1;
static gint ett_rnsap_Gainfactors_10ms_mode = -1;
static gint ett_rnsap_T_signalledGainFactors10ms = -1;
static gint ett_rnsap_GANSS_AddClockModels = -1;
static gint ett_rnsap_GANSS_AddOrbitModels = -1;
static gint ett_rnsap_GANSS_Additional_Ionospheric_Model = -1;
static gint ett_rnsap_GANSS_Additional_Navigation_Models = -1;
static gint ett_rnsap_GANSS_Additional_Time_Models = -1;
static gint ett_rnsap_GANSS_Additional_UTC_Models = -1;
static gint ett_rnsap_GANSS_Almanac = -1;
static gint ett_rnsap_T_gANSS_AlmanacModel = -1;
static gint ett_rnsap_T_gANSS_keplerianParameters = -1;
static gint ett_rnsap_GANSS_alm_keplerianNAVAlmanac = -1;
static gint ett_rnsap_GANSS_alm_KEPLERIANBDSALMANAC = -1;
static gint ett_rnsap_GANSS_alm_keplerianReducedAlmanac = -1;
static gint ett_rnsap_GANSS_alm_keplerianMidiAlmanac = -1;
static gint ett_rnsap_GANSS_alm_keplerianGLONASS = -1;
static gint ett_rnsap_GANSS_alm_ecefSBASAlmanac = -1;
static gint ett_rnsap_GANSS_Auxiliary_Information = -1;
static gint ett_rnsap_GANSS_AuxInfoGANSS_ID1 = -1;
static gint ett_rnsap_GANSS_AuxInfoGANSS_ID1_item = -1;
static gint ett_rnsap_GANSS_AuxInfoGANSS_ID3 = -1;
static gint ett_rnsap_GANSS_AuxInfoGANSS_ID3_item = -1;
static gint ett_rnsap_GANSS_BDSclockModel = -1;
static gint ett_rnsap_GANSS_Clock_Model = -1;
static gint ett_rnsap_GANSS_Clock_Model_item = -1;
static gint ett_rnsap_GANSS_CNAVclockModel = -1;
static gint ett_rnsap_GANSS_Common_Data = -1;
static gint ett_rnsap_GANSS_CommonDataInfoReq = -1;
static gint ett_rnsap_GANSS_Data_Bit_Assistance = -1;
static gint ett_rnsap_GANSS_DataBitAssistanceList = -1;
static gint ett_rnsap_GANSS_DataBitAssistanceItem = -1;
static gint ett_rnsap_GANSS_DataBitAssistanceSgnList = -1;
static gint ett_rnsap_GANSS_DataBitAssistanceSgnItem = -1;
static gint ett_rnsap_GANSS_Data_Bit_Assistance_ReqItem = -1;
static gint ett_rnsap_GANSS_Data_Bit_Assistance_ReqList = -1;
static gint ett_rnsap_T_ganss_SatelliteInfo = -1;
static gint ett_rnsap_GANSS_DeltaUT1 = -1;
static gint ett_rnsap_GANSS_Earth_Orientation_Parameters = -1;
static gint ett_rnsap_GANSS_GenericDataInfoReqList = -1;
static gint ett_rnsap_GANSS_GenericDataInfoReqItem = -1;
static gint ett_rnsap_GANSS_Generic_Data = -1;
static gint ett_rnsap_GANSS_Generic_DataItem = -1;
static gint ett_rnsap_GANSS_GLONASSclockModel = -1;
static gint ett_rnsap_GANSS_Information = -1;
static gint ett_rnsap_GANSS_Ionospheric_Model = -1;
static gint ett_rnsap_GANSS_IonosphereRegionalStormFlags = -1;
static gint ett_rnsap_GANSS_NAVclockModel = -1;
static gint ett_rnsap_GANSS_Navigation_Model_And_Time_Recovery = -1;
static gint ett_rnsap_GANSS_NavModel_BDSKeplerianSet = -1;
static gint ett_rnsap_GANSS_NavModel_CNAVKeplerianSet = -1;
static gint ett_rnsap_GANSS_NavModel_GLONASSecef = -1;
static gint ett_rnsap_GANSS_NavModel_NAVKeplerianSet = -1;
static gint ett_rnsap_GANSS_NavModel_SBASecef = -1;
static gint ett_rnsap_GANSS_Orbit_Model = -1;
static gint ett_rnsap_T_gANSS_keplerianParameters_01 = -1;
static gint ett_rnsap_GANSS_Real_Time_Integrity = -1;
static gint ett_rnsap_GANSS_Real_Time_Integrity_item = -1;
static gint ett_rnsap_GANSS_RX_Pos = -1;
static gint ett_rnsap_GANSS_SatelliteInformationKP = -1;
static gint ett_rnsap_GANSS_SatelliteInformationKP_item = -1;
static gint ett_rnsap_Ganss_Sat_Info_AddNavList = -1;
static gint ett_rnsap_Ganss_Sat_Info_AddNavList_item = -1;
static gint ett_rnsap_GANSS_SAT_Info_Almanac_BDSkpList = -1;
static gint ett_rnsap_GANSS_SAT_Info_Almanac_BDS = -1;
static gint ett_rnsap_GANSS_SAT_Info_Almanac_GLOkpList = -1;
static gint ett_rnsap_GANSS_SAT_Info_Almanac_GLOkpList_item = -1;
static gint ett_rnsap_GANSS_SAT_Info_Almanac_MIDIkpList = -1;
static gint ett_rnsap_GANSS_SAT_Info_Almanac_MIDIkpList_item = -1;
static gint ett_rnsap_GANSS_SAT_Info_Almanac_NAVkpList = -1;
static gint ett_rnsap_GANSS_SAT_Info_Almanac_NAVkpList_item = -1;
static gint ett_rnsap_GANSS_SAT_Info_Almanac_REDkpList = -1;
static gint ett_rnsap_GANSS_SAT_Info_Almanac_REDkpList_item = -1;
static gint ett_rnsap_GANSS_SAT_Info_Almanac_SBASecefList = -1;
static gint ett_rnsap_GANSS_SAT_Info_Almanac_SBASecefList_item = -1;
static gint ett_rnsap_GANSS_Sat_Info_Nav = -1;
static gint ett_rnsap_GANSS_Sat_Info_Nav_item = -1;
static gint ett_rnsap_GANSS_SBASclockModel = -1;
static gint ett_rnsap_GANSS_Time_Model = -1;
static gint ett_rnsap_GANSS_Transmission_Time = -1;
static gint ett_rnsap_GANSS_UTC_Model = -1;
static gint ett_rnsap_GANSS_UTCmodelSet1 = -1;
static gint ett_rnsap_GANSS_UTCmodelSet2 = -1;
static gint ett_rnsap_GANSS_UTCmodelSet3 = -1;
static gint ett_rnsap_GANSS_UTCmodelSet4 = -1;
static gint ett_rnsap_GA_PointWithAltitude = -1;
static gint ett_rnsap_GA_PointWithAltitudeAndUncertaintyEllipsoid = -1;
static gint ett_rnsap_GA_PointWithUnCertaintyEllipse = -1;
static gint ett_rnsap_GA_UncertaintyEllipse = -1;
static gint ett_rnsap_GA_PointWithUnCertainty = -1;
static gint ett_rnsap_GA_AccessPointPosition = -1;
static gint ett_rnsap_GeographicalCoordinate = -1;
static gint ett_rnsap_GERAN_SI_Type = -1;
static gint ett_rnsap_GERAN_SystemInfo = -1;
static gint ett_rnsap_GERAN_SystemInfo_item = -1;
static gint ett_rnsap_GPS_Almanac = -1;
static gint ett_rnsap_T_satellite_Almanac_Information = -1;
static gint ett_rnsap_T_satellite_Almanac_Information_item = -1;
static gint ett_rnsap_Satellite_Almanac_Information_ExtItem = -1;
static gint ett_rnsap_Satellite_Almanac_Information_ExtItem_item = -1;
static gint ett_rnsap_GPSInformation = -1;
static gint ett_rnsap_GPSInformation_item = -1;
static gint ett_rnsap_GPS_Ionospheric_Model = -1;
static gint ett_rnsap_GPS_NavigationModel_and_TimeRecovery = -1;
static gint ett_rnsap_GPS_NavigationModel_and_TimeRecovery_item = -1;
static gint ett_rnsap_GPS_RealTime_Integrity = -1;
static gint ett_rnsap_GPS_RX_POS = -1;
static gint ett_rnsap_GPS_UTC_Model = -1;
static gint ett_rnsap_Guaranteed_Rate_Information = -1;
static gint ett_rnsap_HARQ_MemoryPartitioning = -1;
static gint ett_rnsap_HARQ_MemoryPartitioning_Implicit = -1;
static gint ett_rnsap_HARQ_MemoryPartitioning_Explicit = -1;
static gint ett_rnsap_HARQ_MemoryPartitioningList = -1;
static gint ett_rnsap_HARQ_MemoryPartitioningInfoExtForMIMO = -1;
static gint ett_rnsap_HARQ_MemoryPartitioningItem = -1;
static gint ett_rnsap_HSDSCH_FDD_Information = -1;
static gint ett_rnsap_HSDSCH_FDD_Information_Response = -1;
static gint ett_rnsap_HS_DSCH_FDD_Secondary_Serving_Information = -1;
static gint ett_rnsap_HS_DSCH_FDD_Secondary_Serving_Information_Response = -1;
static gint ett_rnsap_HS_DSCH_Secondary_Serving_Information_To_Modify = -1;
static gint ett_rnsap_HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised = -1;
static gint ett_rnsap_HS_DSCH_FDD_Secondary_Serving_Update_Information = -1;
static gint ett_rnsap_HS_DSCH_Secondary_Serving_Cell_Change_Information_Response = -1;
static gint ett_rnsap_HS_DSCH_Secondary_Serving_cell_change_choice = -1;
static gint ett_rnsap_HS_Secondary_Serving_cell_change_successful = -1;
static gint ett_rnsap_HS_Secondary_Serving_cell_change_unsuccessful = -1;
static gint ett_rnsap_HSDSCH_Information_to_Modify = -1;
static gint ett_rnsap_HSDSCH_Information_to_Modify_Unsynchronised = -1;
static gint ett_rnsap_HSDSCH_MACdFlow_Specific_InfoList = -1;
static gint ett_rnsap_HSDSCH_MACdFlow_Specific_InfoItem = -1;
static gint ett_rnsap_HSDSCH_MACdFlow_Specific_InfoList_Response = -1;
static gint ett_rnsap_HSDSCH_MACdFlow_Specific_InfoItem_Response = -1;
static gint ett_rnsap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify = -1;
static gint ett_rnsap_HSDSCH_MACdFlow_Specific_InfoItem_to_Modify = -1;
static gint ett_rnsap_HSDSCH_MACdFlows_Information = -1;
static gint ett_rnsap_HSDSCH_MACdFlows_to_Delete = -1;
static gint ett_rnsap_HSDSCH_MACdFlows_to_Delete_Item = -1;
static gint ett_rnsap_HSDSCH_Initial_Capacity_Allocation = -1;
static gint ett_rnsap_HSDSCH_Initial_Capacity_AllocationItem = -1;
static gint ett_rnsap_HSDSCH_PreconfigurationInfo = -1;
static gint ett_rnsap_Additional_EDCH_Preconfiguration_Information = -1;
static gint ett_rnsap_Additional_EDCH_Preconfiguration_Information_ItemIEs = -1;
static gint ett_rnsap_HSDSCH_PreconfigurationSetup = -1;
static gint ett_rnsap_HS_SCCH_PreconfiguredCodes = -1;
static gint ett_rnsap_HS_SCCH_PreconfiguredCodesItem = -1;
static gint ett_rnsap_HS_DSCH_serving_cell_change_information = -1;
static gint ett_rnsap_HS_DSCH_serving_cell_change_informationResponse = -1;
static gint ett_rnsap_HS_DSCH_serving_cell_change_choice = -1;
static gint ett_rnsap_HS_serving_cell_change_successful = -1;
static gint ett_rnsap_HS_serving_cell_change_unsuccessful = -1;
static gint ett_rnsap_HSDSCH_TDD_Information = -1;
static gint ett_rnsap_HSDSCH_TDD_Information_Response = -1;
static gint ett_rnsap_HSPDSCH_TDD_Specific_InfoList_Response = -1;
static gint ett_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response = -1;
static gint ett_rnsap_HSPDSCH_TDD_Specific_InfoList_Response_LCR = -1;
static gint ett_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response_LCR = -1;
static gint ett_rnsap_HSPDSCH_TDD_Specific_InfoList_Response768 = -1;
static gint ett_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response768 = -1;
static gint ett_rnsap_HSSCCH_FDD_Specific_InfoList_Response = -1;
static gint ett_rnsap_HSSCCH_FDD_Specific_InfoItem_Response = -1;
static gint ett_rnsap_HSSCCH_TDD_Specific_InfoList_Response = -1;
static gint ett_rnsap_HSSCCH_TDD_Specific_InfoItem_Response = -1;
static gint ett_rnsap_HSSCCH_TDD_Specific_InfoList_Response_LCR = -1;
static gint ett_rnsap_HSSCCH_TDD_Specific_InfoItem_Response_LCR = -1;
static gint ett_rnsap_HSSCCH_TDD_Specific_InfoList_Response768 = -1;
static gint ett_rnsap_HSSCCH_TDD_Specific_InfoItem_Response768 = -1;
static gint ett_rnsap_HSSICH_Info = -1;
static gint ett_rnsap_HSSICH_InfoLCR = -1;
static gint ett_rnsap_HSSICH_Info768 = -1;
static gint ett_rnsap_HS_SICH_Reception_Quality_Value = -1;
static gint ett_rnsap_HSDSCH_FDD_Update_Information = -1;
static gint ett_rnsap_HSDSCH_TDD_Update_Information = -1;
static gint ett_rnsap_MIMO_ReferenceSignal_InformationListLCR = -1;
static gint ett_rnsap_HSSICH_ReferenceSignal_InformationLCR = -1;
static gint ett_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_LCR = -1;
static gint ett_rnsap_Transport_Block_Size_List_LCR = -1;
static gint ett_rnsap_Transport_Block_Size_Item_LCR = -1;
static gint ett_rnsap_Repetition_Period_List_LCR = -1;
static gint ett_rnsap_Repetition_Period_Item_LCR = -1;
static gint ett_rnsap_HS_DSCH_SPS_Operation_Indicator = -1;
static gint ett_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR = -1;
static gint ett_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR = -1;
static gint ett_rnsap_HS_SICH_InformationList_for_HS_DSCH_SPS = -1;
static gint ett_rnsap_HS_SICH_InformationItem_for_HS_DSCH_SPS = -1;
static gint ett_rnsap_HS_SICH_Type = -1;
static gint ett_rnsap_HS_SCCH_Associated_HS_SICH = -1;
static gint ett_rnsap_Non_HS_SCCH_Associated_HS_SICH = -1;
static gint ett_rnsap_Initial_HS_PDSCH_SPS_Resource = -1;
static gint ett_rnsap_InformationAvailable = -1;
static gint ett_rnsap_InformationReportCharacteristics = -1;
static gint ett_rnsap_InformationReportPeriodicity = -1;
static gint ett_rnsap_InformationThreshold = -1;
static gint ett_rnsap_InformationType = -1;
static gint ett_rnsap_IPDLParameters = -1;
static gint ett_rnsap_Inter_Frequency_Cell_List = -1;
static gint ett_rnsap_Inter_Frequency_Cell = -1;
static gint ett_rnsap_Inter_Frequency_Cell_Information = -1;
static gint ett_rnsap_Inter_Frequency_Cell_Information_SIB11 = -1;
static gint ett_rnsap_Inter_Frequency_Cell_Information_SIB12 = -1;
static gint ett_rnsap_Inter_Frequency_Cell_Information_SIB11ter = -1;
static gint ett_rnsap_Inter_Frequency_Cells_Information_SIB11_Per_Indication = -1;
static gint ett_rnsap_Inter_Frequency_Cells_Information_SIB12_Per_Indication = -1;
static gint ett_rnsap_Inter_Frequency_Cells_Information_SIB11ter_Per_Indication = -1;
static gint ett_rnsap_Inter_Frequency_Cell_SIB11ter_List = -1;
static gint ett_rnsap_Inter_Frequency_Cell_SIB11ter = -1;
static gint ett_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12_List = -1;
static gint ett_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12 = -1;
static gint ett_rnsap_IPDL_FDD_Parameters = -1;
static gint ett_rnsap_IPDL_TDD_Parameters = -1;
static gint ett_rnsap_IPDL_TDD_ParametersLCR = -1;
static gint ett_rnsap_IdleIntervalInformation = -1;
static gint ett_rnsap_LABased = -1;
static gint ett_rnsap_LAI_List = -1;
static gint ett_rnsap_LAI = -1;
static gint ett_rnsap_List_Of_PLMNs = -1;
static gint ett_rnsap_LoadValue = -1;
static gint ett_rnsap_LCRTDD_Uplink_Physical_Channel_Capability = -1;
static gint ett_rnsap_M1Report = -1;
static gint ett_rnsap_M2Report = -1;
static gint ett_rnsap_MDTAreaScope = -1;
static gint ett_rnsap_MDT_Configuration = -1;
static gint ett_rnsap_MDT_Report_Parameters = -1;
static gint ett_rnsap_MAC_c_sh_SDU_LengthList = -1;
static gint ett_rnsap_MACdPDU_Size_IndexList = -1;
static gint ett_rnsap_MACdPDU_Size_IndexItem = -1;
static gint ett_rnsap_MACdPDU_Size_IndexList_to_Modify = -1;
static gint ett_rnsap_MACdPDU_Size_IndexItem_to_Modify = -1;
static gint ett_rnsap_MBMS_Bearer_Service_Full_Address = -1;
static gint ett_rnsap_MBMS_Bearer_Service_List = -1;
static gint ett_rnsap_MBMS_Bearer_ServiceItemFDD = -1;
static gint ett_rnsap_MBMS_Bearer_ServiceItemFDD_PFL = -1;
static gint ett_rnsap_MBMS_Bearer_ServiceItemTDD = -1;
static gint ett_rnsap_MBMS_Bearer_ServiceItemTDD_PFL = -1;
static gint ett_rnsap_MBMSChannelTypeInfo = -1;
static gint ett_rnsap_MBMSChannelTypeCellList = -1;
static gint ett_rnsap_MBMSPreferredFreqLayerInfo = -1;
static gint ett_rnsap_MBMS_Neighbouring_Cell_Information = -1;
static gint ett_rnsap_MBMS_ConcatenatedServiceList = -1;
static gint ett_rnsap_MBMS_RLC_Sequence_Number_Information = -1;
static gint ett_rnsap_MBMS_RLC_Sequence_Number_Information_List = -1;
static gint ett_rnsap_MBMS_Bearer_Service_List_RLC = -1;
static gint ett_rnsap_MBMS_Bearer_Service_List_RLCinfo = -1;
static gint ett_rnsap_MCCH_Message_List = -1;
static gint ett_rnsap_MCCH_Configuration_Info = -1;
static gint ett_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Info_List = -1;
static gint ett_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Item = -1;
static gint ett_rnsap_MeasurementIncreaseDecreaseThreshold = -1;
static gint ett_rnsap_MeasurementThreshold = -1;
static gint ett_rnsap_MidambleShiftAndBurstType = -1;
static gint ett_rnsap_Type1 = -1;
static gint ett_rnsap_MidambleAllocationMode1 = -1;
static gint ett_rnsap_Type2 = -1;
static gint ett_rnsap_MidambleAllocationMode2 = -1;
static gint ett_rnsap_Type3 = -1;
static gint ett_rnsap_MidambleAllocationMode3 = -1;
static gint ett_rnsap_MidambleShiftLCR = -1;
static gint ett_rnsap_MidambleShiftAndBurstType768 = -1;
static gint ett_rnsap_Type7681 = -1;
static gint ett_rnsap_MidambleAllocationMode7681 = -1;
static gint ett_rnsap_Type7682 = -1;
static gint ett_rnsap_MidambleAllocationMode7682 = -1;
static gint ett_rnsap_Type7683 = -1;
static gint ett_rnsap_MidambleAllocationMode7683 = -1;
static gint ett_rnsap_MIMO_InformationResponse = -1;
static gint ett_rnsap_MIMO_PilotConfiguration = -1;
static gint ett_rnsap_Additional_Associated_Secondary_CPICH = -1;
static gint ett_rnsap_Additional_Secondary_CPICH = -1;
static gint ett_rnsap_ModifyPriorityQueue = -1;
static gint ett_rnsap_MulticellEDCH_InformationItemIEs = -1;
static gint ett_rnsap_MulticellEDCH_RL_SpecificInformationItemIEs = -1;
static gint ett_rnsap_Multiple_PLMN_List = -1;
static gint ett_rnsap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR = -1;
static gint ett_rnsap_MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR = -1;
static gint ett_rnsap_Multi_Carrier_EDCH_Info = -1;
static gint ett_rnsap_Multi_Carrier_EDCH_Information = -1;
static gint ett_rnsap_Multi_Carrier_EDCH_LCR_InformationItem = -1;
static gint ett_rnsap_Multi_Carrier_EDCH_MACdFlows_Information_TDD = -1;
static gint ett_rnsap_Multi_Carrier_EDCH_MACdFlows_Specific_Info = -1;
static gint ett_rnsap_Multi_Carrier_EDCH_Reconfigure = -1;
static gint ett_rnsap_Continue_Setup_Change_Multi_Carrier_EDCH = -1;
static gint ett_rnsap_Multi_Carrier_EDCH_Change_Info = -1;
static gint ett_rnsap_Multi_Carrier_EDCH_Information_Removal_List = -1;
static gint ett_rnsap_Multi_Carrier_EDCH_Information_Removal_Info_ItemIEs = -1;
static gint ett_rnsap_Multi_Carrier_EDCH_Information_Response = -1;
static gint ett_rnsap_Multi_Carrier_EDCH_LCR_Information_ResponseItem = -1;
static gint ett_rnsap_Multiflow_Reconfiguration = -1;
static gint ett_rnsap_Multiflow_Information = -1;
static gint ett_rnsap_Multiflow_Information_To_Modify = -1;
static gint ett_rnsap_Multiflow_RepetitionFactors = -1;
static gint ett_rnsap_Multiflow_Timing = -1;
static gint ett_rnsap_MU_MIMO_InformationLCR = -1;
static gint ett_rnsap_Standalone_Midamble_Channel_Information = -1;
static gint ett_rnsap_MU_MIMO_Indicator = -1;
static gint ett_rnsap_NACC_Related_Data = -1;
static gint ett_rnsap_Neighbouring_UMTS_CellInformation = -1;
static gint ett_rnsap_Neighbouring_UMTS_CellInformation_Ext = -1;
static gint ett_rnsap_Neighbouring_UMTS_CellInformationExtensionItem = -1;
static gint ett_rnsap_Neighbouring_UMTS_CellInformationItem = -1;
static gint ett_rnsap_Neighbouring_FDD_CellInformation = -1;
static gint ett_rnsap_Neighbouring_FDD_CellInformationItem = -1;
static gint ett_rnsap_MultipleFrequencyBandsIndicatorList = -1;
static gint ett_rnsap_NeighbouringFDDCellMeasurementInformation = -1;
static gint ett_rnsap_Neighbouring_GSM_CellInformationIEs = -1;
static gint ett_rnsap_Neighbouring_GSM_CellInformationItem = -1;
static gint ett_rnsap_Neighbouring_TDD_CellInformation = -1;
static gint ett_rnsap_Neighbouring_TDD_CellInformationItem = -1;
static gint ett_rnsap_NeighbouringTDDCellMeasurementInformation = -1;
static gint ett_rnsap_NeighbouringTDDCellMeasurementInformationLCR = -1;
static gint ett_rnsap_NeighbouringTDDCellMeasurementInformation768 = -1;
static gint ett_rnsap_Neighbouring_LCR_TDD_CellInformation = -1;
static gint ett_rnsap_Neighbouring_LCR_TDD_CellInformationItem = -1;
static gint ett_rnsap_Neighbouring_E_UTRA_CellInformation = -1;
static gint ett_rnsap_Neighbouring_E_UTRA_CellInformationItem = -1;
static gint ett_rnsap_MultipleFrequencyBandsIndicatorList_EUTRA = -1;
static gint ett_rnsap_NotProvidedCellList = -1;
static gint ett_rnsap_NRTLoadInformationValue = -1;
static gint ett_rnsap_Non_Serving_RL_Preconfig_Setup = -1;
static gint ett_rnsap_New_non_serving_RL_setup_selection = -1;
static gint ett_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList = -1;
static gint ett_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_item = -1;
static gint ett_rnsap_OnModification = -1;
static gint ett_rnsap_OnModificationInformation = -1;
static gint ett_rnsap_PCH_InformationList = -1;
static gint ett_rnsap_PCH_InformationItem = -1;
static gint ett_rnsap_Periodic = -1;
static gint ett_rnsap_PeriodicInformation = -1;
static gint ett_rnsap_Permanent_NAS_UE_Identity = -1;
static gint ett_rnsap_PLMN_List = -1;
static gint ett_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas = -1;
static gint ett_rnsap_Associated_Secondary_CPICH = -1;
static gint ett_rnsap_PreferredFrequencyLayerInfo = -1;
static gint ett_rnsap_PriorityQueue_InfoList = -1;
static gint ett_rnsap_PriorityQueue_InfoItem = -1;
static gint ett_rnsap_PriorityQueue_InfoList_EnhancedFACH_PCH = -1;
static gint ett_rnsap_PriorityQueue_InfoItem_EnhancedFACH_PCH = -1;
static gint ett_rnsap_PriorityQueue_InfoList_to_Modify = -1;
static gint ett_rnsap_PriorityQueue_InfoItem_to_Add = -1;
static gint ett_rnsap_PriorityQueue_InfoItem_to_Modify = -1;
static gint ett_rnsap_PriorityQueue_InfoList_to_Modify_Unsynchronised = -1;
static gint ett_rnsap_PriorityQueue_InfoItem_to_Modify_Unsynchronised = -1;
static gint ett_rnsap_ProvidedInformation = -1;
static gint ett_rnsap_UpPCH_InformationList_LCRTDD = -1;
static gint ett_rnsap_UpPCH_InformationItem_LCRTDD = -1;
static gint ett_rnsap_PTMCellList = -1;
static gint ett_rnsap_PTPCellList = -1;
static gint ett_rnsap_RABased = -1;
static gint ett_rnsap_RAI_List = -1;
static gint ett_rnsap_RAI = -1;
static gint ett_rnsap_RB_Info = -1;
static gint ett_rnsap_Reference_E_TFCI_Information = -1;
static gint ett_rnsap_Reference_E_TFCI_Information_Item = -1;
static gint ett_rnsap_Released_CN_Domain = -1;
static gint ett_rnsap_ReportCharacteristics = -1;
static gint ett_rnsap_ReportPeriodicity = -1;
static gint ett_rnsap_RequestedDataValue = -1;
static gint ett_rnsap_RequestedDataValueInformation = -1;
static gint ett_rnsap_RL_Specific_DCH_Info = -1;
static gint ett_rnsap_RL_Specific_DCH_Info_Item = -1;
static gint ett_rnsap_RL_Specific_EDCH_Information = -1;
static gint ett_rnsap_RL_Specific_EDCH_Info = -1;
static gint ett_rnsap_RL_Specific_EDCH_InfoItem = -1;
static gint ett_rnsap_RTLoadValue = -1;
static gint ett_rnsap_SAI = -1;
static gint ett_rnsap_SecondaryServingCells = -1;
static gint ett_rnsap_SecondaryServingCellsItem = -1;
static gint ett_rnsap_Secondary_CCPCH_Info_TDD = -1;
static gint ett_rnsap_Secondary_CPICH_Information = -1;
static gint ett_rnsap_Secondary_CPICH_Information_Change = -1;
static gint ett_rnsap_Secondary_LCR_CCPCH_Info_TDD = -1;
static gint ett_rnsap_Secondary_CCPCH_Info_TDD768 = -1;
static gint ett_rnsap_Secondary_CCPCH_TDD_InformationList = -1;
static gint ett_rnsap_Secondary_CCPCH_TDD_InformationItem = -1;
static gint ett_rnsap_Secondary_LCR_CCPCH_TDD_InformationList = -1;
static gint ett_rnsap_Secondary_LCR_CCPCH_TDD_InformationItem = -1;
static gint ett_rnsap_Secondary_CCPCH_TDD_InformationList768 = -1;
static gint ett_rnsap_Secondary_CCPCH_TDD_InformationItem768 = -1;
static gint ett_rnsap_Secondary_CCPCH_TDD_Code_Information = -1;
static gint ett_rnsap_Secondary_CCPCH_TDD_Code_InformationItem = -1;
static gint ett_rnsap_Secondary_LCR_CCPCH_TDD_Code_Information = -1;
static gint ett_rnsap_Secondary_LCR_CCPCH_TDD_Code_InformationItem = -1;
static gint ett_rnsap_Secondary_CCPCH_TDD_Code_Information768 = -1;
static gint ett_rnsap_Secondary_CCPCH_TDD_Code_InformationItem768 = -1;
static gint ett_rnsap_Secondary_Serving_Cell_List = -1;
static gint ett_rnsap_Possible_Secondary_Serving_Cell_List = -1;
static gint ett_rnsap_Possible_Secondary_Serving_Cell = -1;
static gint ett_rnsap_SetsOfHS_SCCH_Codes = -1;
static gint ett_rnsap_SetsOfHS_SCCH_CodesItem = -1;
static gint ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency = -1;
static gint ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements = -1;
static gint ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD = -1;
static gint ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information = -1;
static gint ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH = -1;
static gint ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2 = -1;
static gint ett_rnsap_GA_AccessPointPositionwithOptionalAltitude = -1;
static gint ett_rnsap_SFNSFNMeasurementThresholdInformation = -1;
static gint ett_rnsap_SFNSFNMeasurementValueInformation = -1;
static gint ett_rnsap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation = -1;
static gint ett_rnsap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item = -1;
static gint ett_rnsap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation = -1;
static gint ett_rnsap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item = -1;
static gint ett_rnsap_SFNSFNTimeStampInformation = -1;
static gint ett_rnsap_SFNSFNTimeStamp_TDD = -1;
static gint ett_rnsap_SFNSFNValue = -1;
static gint ett_rnsap_SNA_Information = -1;
static gint ett_rnsap_ListOfSNAs = -1;
static gint ett_rnsap_S_RNTI_Group = -1;
static gint ett_rnsap_SynchronisationConfiguration = -1;
static gint ett_rnsap_SYNC_UL_ProcParameters = -1;
static gint ett_rnsap_TDD_ChannelisationCodeLCR = -1;
static gint ett_rnsap_TDD_DCHs_to_Modify = -1;
static gint ett_rnsap_TDD_DCHs_to_ModifyItem = -1;
static gint ett_rnsap_TDD_DCHs_to_ModifySpecificInformationList = -1;
static gint ett_rnsap_TDD_DCHs_to_ModifySpecificItem = -1;
static gint ett_rnsap_TDD_DL_Code_Information = -1;
static gint ett_rnsap_TDD_DL_Code_InformationItem = -1;
static gint ett_rnsap_TDD_DL_Code_LCR_Information = -1;
static gint ett_rnsap_TDD_DL_Code_LCR_InformationItem = -1;
static gint ett_rnsap_TDD_DL_Code_Information768 = -1;
static gint ett_rnsap_TDD_DL_Code_InformationItem768 = -1;
static gint ett_rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR = -1;
static gint ett_rnsap_TDD_DPCHOffset = -1;
static gint ett_rnsap_TDD_UL_Code_Information = -1;
static gint ett_rnsap_TDD_UL_Code_InformationItem = -1;
static gint ett_rnsap_TDD_UL_Code_LCR_Information = -1;
static gint ett_rnsap_TDD_UL_Code_LCR_InformationItem = -1;
static gint ett_rnsap_TDD_UL_Code_Information768 = -1;
static gint ett_rnsap_TDD_UL_Code_InformationItem768 = -1;
static gint ett_rnsap_TDD_UL_DPCH_TimeSlotFormat_LCR = -1;
static gint ett_rnsap_TMGI = -1;
static gint ett_rnsap_TnlQos = -1;
static gint ett_rnsap_Transmission_Gap_Pattern_Sequence_Information = -1;
static gint ett_rnsap_Transmission_Gap_Pattern_Sequence_Information_item = -1;
static gint ett_rnsap_Transmission_Gap_Pattern_Sequence_Status_List = -1;
static gint ett_rnsap_Transmission_Gap_Pattern_Sequence_Status_List_item = -1;
static gint ett_rnsap_Affected_HSDSCH_Serving_Cell_List = -1;
static gint ett_rnsap_Transmission_Mode_Information = -1;
static gint ett_rnsap_Transmission_Mode_Information_List = -1;
static gint ett_rnsap_TUTRANGANSS = -1;
static gint ett_rnsap_TUTRANGANSSMeasurementThresholdInformation = -1;
static gint ett_rnsap_TUTRANGANSSMeasurementValueInformation = -1;
static gint ett_rnsap_TUTRANGPS = -1;
static gint ett_rnsap_TUTRANGPSMeasurementThresholdInformation = -1;
static gint ett_rnsap_TUTRANGPSMeasurementValueInformation = -1;
static gint ett_rnsap_TransportFormatCombination_Beta = -1;
static gint ett_rnsap_T_signalledGainFactors = -1;
static gint ett_rnsap_TFCS = -1;
static gint ett_rnsap_T_tFCSvalues = -1;
static gint ett_rnsap_TFCS_TFCSList = -1;
static gint ett_rnsap_TFCS_TFCSList_item = -1;
static gint ett_rnsap_TFCS_CTFC = -1;
static gint ett_rnsap_TPC_slot_position = -1;
static gint ett_rnsap_TransportFormatSet = -1;
static gint ett_rnsap_TransportFormatSet_DynamicPartList = -1;
static gint ett_rnsap_TransportFormatSet_DynamicPartList_item = -1;
static gint ett_rnsap_TransportFormatSet_ModeDP = -1;
static gint ett_rnsap_TDD_TransportFormatSet_ModeDP = -1;
static gint ett_rnsap_TransmissionTimeIntervalInformation = -1;
static gint ett_rnsap_TransmissionTimeIntervalInformation_item = -1;
static gint ett_rnsap_TransportFormatSet_Semi_staticPart = -1;
static gint ett_rnsap_TransportFormatSet_ModeSSP = -1;
static gint ett_rnsap_TTI_Update_Indicator = -1;
static gint ett_rnsap_UE_AggregateMaximumBitRate = -1;
static gint ett_rnsap_UE_Capabilities_Info = -1;
static gint ett_rnsap_UE_RF_Band_CapabilityLCR = -1;
static gint ett_rnsap_Radio_Frequency_BandItem = -1;
static gint ett_rnsap_UEIdentity = -1;
static gint ett_rnsap_UEMeasurementReportCharacteristics = -1;
static gint ett_rnsap_UEMeasurementReportCharacteristicsEvent1h = -1;
static gint ett_rnsap_UEMeasurementReportCharacteristicsEvent1i = -1;
static gint ett_rnsap_UEMeasurementReportCharacteristicsEvent6a = -1;
static gint ett_rnsap_UEMeasurementReportCharacteristicsEvent6b = -1;
static gint ett_rnsap_UEMeasurementReportCharacteristicsEvent6c = -1;
static gint ett_rnsap_UEMeasurementReportCharacteristicsEvent6d = -1;
static gint ett_rnsap_UEMeasurementReportCharacteristicsPeriodic = -1;
static gint ett_rnsap_UEMeasurementThreshold = -1;
static gint ett_rnsap_UEMeasurementTimeslotInfoHCR = -1;
static gint ett_rnsap_UEMeasurementTimeslotInfoHCR_IEs = -1;
static gint ett_rnsap_UEMeasurementTimeslotInfoLCR = -1;
static gint ett_rnsap_UEMeasurementTimeslotInfoLCR_IEs = -1;
static gint ett_rnsap_UEMeasurementTimeslotInfo768 = -1;
static gint ett_rnsap_UEMeasurementTimeslotInfo768_IEs = -1;
static gint ett_rnsap_UEMeasurementValue = -1;
static gint ett_rnsap_UE_MeasurementValue_UE_Transmitted_Power = -1;
static gint ett_rnsap_UEMeasurementValueTransmittedPowerListHCR = -1;
static gint ett_rnsap_UEMeasurementValueTransmittedPowerListHCR_IEs = -1;
static gint ett_rnsap_UEMeasurementValueTransmittedPowerListLCR = -1;
static gint ett_rnsap_UEMeasurementValueTransmittedPowerListLCR_IEs = -1;
static gint ett_rnsap_UEMeasurementValueTransmittedPowerList768 = -1;
static gint ett_rnsap_UEMeasurementValueTransmittedPowerList768_IEs = -1;
static gint ett_rnsap_UE_MeasurementValue_Primary_CCPCH_RSCP = -1;
static gint ett_rnsap_UE_MeasurementValue_DL_Timeslot_ISCP = -1;
static gint ett_rnsap_UEMeasurementValueTimeslotISCPListHCR = -1;
static gint ett_rnsap_UEMeasurementValueTimeslotISCPListHCR_IEs = -1;
static gint ett_rnsap_UEMeasurementValueTimeslotISCPListLCR = -1;
static gint ett_rnsap_UEMeasurementValueTimeslotISCPListLCR_IEs = -1;
static gint ett_rnsap_UEMeasurementValueTimeslotISCPList768 = -1;
static gint ett_rnsap_UEMeasurementValueTimeslotISCPList768_IEs = -1;
static gint ett_rnsap_UEMeasurementValueInformation = -1;
static gint ett_rnsap_UEMeasurementValueInformationAvailable = -1;
static gint ett_rnsap_UE_Measurement_Value = -1;
static gint ett_rnsap_UE_State = -1;
static gint ett_rnsap_Cell_Fach_Pch_State = -1;
static gint ett_rnsap_Ura_Pch_State = -1;
static gint ett_rnsap_UL_Timeslot_Information = -1;
static gint ett_rnsap_UL_Timeslot_InformationItem = -1;
static gint ett_rnsap_UL_TimeslotLCR_Information = -1;
static gint ett_rnsap_UL_TimeslotLCR_InformationItem = -1;
static gint ett_rnsap_PLCCHinformation = -1;
static gint ett_rnsap_UL_Timeslot_Information768 = -1;
static gint ett_rnsap_UL_Timeslot_InformationItem768 = -1;
static gint ett_rnsap_UL_TimeSlot_ISCP_Info = -1;
static gint ett_rnsap_UL_TimeSlot_ISCP_InfoItem = -1;
static gint ett_rnsap_UL_TimeSlot_ISCP_LCR_Info = -1;
static gint ett_rnsap_UL_TimeSlot_ISCP_LCR_InfoItem = -1;
static gint ett_rnsap_UL_TimingAdvanceCtrl_LCR = -1;
static gint ett_rnsap_UC_ID = -1;
static gint ett_rnsap_UL_ScramblingCode = -1;
static gint ett_rnsap_UL_Synchronisation_Parameters_LCR = -1;
static gint ett_rnsap_URA_Information = -1;
static gint ett_rnsap_RNCsWithCellsInTheAccessedURA_List = -1;
static gint ett_rnsap_RNCsWithCellsInTheAccessedURA_Item = -1;
static gint ett_rnsap_USCH_Information = -1;
static gint ett_rnsap_USCH_InformationItem = -1;
static gint ett_rnsap_UMTS_Cells_Info_List = -1;
static gint ett_rnsap_UMTS_Cell_Info = -1;
static gint ett_rnsap_UL_CLTD_Information = -1;
static gint ett_rnsap_UL_CLTD_Information_Reconf = -1;
static gint ett_rnsap_UL_CLTD_Information_To_Modify = -1;
static gint ett_rnsap_UL_DPCCH2_Information = -1;
static gint ett_rnsap_UL_DPCCH2_Information_Reconf = -1;
static gint ett_rnsap_UL_DPCCH2_Information_To_Modify = -1;
static gint ett_rnsap_UL_MIMO_Information = -1;
static gint ett_rnsap_UL_MIMO_Information_To_Modify = -1;
static gint ett_rnsap_UL_MIMO_Reconfiguration = -1;
static gint ett_rnsap_UL_MIMO_DL_Control_Channel_Information = -1;
static gint ett_rnsap_RadioLinkSetupRequestFDD = -1;
static gint ett_rnsap_UL_DPCH_Information_RL_SetupRqstFDD = -1;
static gint ett_rnsap_DL_DPCH_Information_RL_SetupRqstFDD = -1;
static gint ett_rnsap_PowerOffsetInformation_RL_SetupRqstFDD = -1;
static gint ett_rnsap_RL_InformationList_RL_SetupRqstFDD = -1;
static gint ett_rnsap_RL_InformationItem_RL_SetupRqstFDD = -1;
static gint ett_rnsap_Additional_HS_Cell_Information_RL_Setup_List = -1;
static gint ett_rnsap_Additional_HS_Cell_Information_RL_Setup_ItemIEs = -1;
static gint ett_rnsap_F_DPCH_Information_RL_SetupRqstFDD = -1;
static gint ett_rnsap_PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD = -1;
static gint ett_rnsap_RadioLinkSetupRequestTDD = -1;
static gint ett_rnsap_UL_Physical_Channel_Information_RL_SetupRqstTDD = -1;
static gint ett_rnsap_DL_Physical_Channel_Information_RL_SetupRqstTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationList_RL_SetupRqstTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationList_RL_SetupRqstTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD = -1;
static gint ett_rnsap_CCTrCH_TPCList_RL_SetupRqstTDD = -1;
static gint ett_rnsap_CCTrCH_TPCItem_RL_SetupRqstTDD = -1;
static gint ett_rnsap_RL_Information_RL_SetupRqstTDD = -1;
static gint ett_rnsap_RadioLinkSetupResponseFDD = -1;
static gint ett_rnsap_RL_InformationResponseList_RL_SetupRspFDD = -1;
static gint ett_rnsap_RL_InformationResponseItem_RL_SetupRspFDD = -1;
static gint ett_rnsap_DiversityIndication_RL_SetupRspFDD = -1;
static gint ett_rnsap_Combining_RL_SetupRspFDD = -1;
static gint ett_rnsap_NonCombiningOrFirstRL_RL_SetupRspFDD = -1;
static gint ett_rnsap_Additional_HS_Cell_Information_Response_List = -1;
static gint ett_rnsap_Additional_HS_Cell_Information_Response_ItemIEs = -1;
static gint ett_rnsap_RadioLinkSetupResponseTDD = -1;
static gint ett_rnsap_RL_InformationResponse_RL_SetupRspTDD = -1;
static gint ett_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD = -1;
static gint ett_rnsap_UL_CCTrCHInformationItem_RL_SetupRspTDD = -1;
static gint ett_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD = -1;
static gint ett_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD = -1;
static gint ett_rnsap_DL_CCTrCHInformationItem_RL_SetupRspTDD = -1;
static gint ett_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD = -1;
static gint ett_rnsap_DSCH_InformationListIEs_RL_SetupRspTDD = -1;
static gint ett_rnsap_DSCHInformationItem_RL_SetupRspTDD = -1;
static gint ett_rnsap_USCH_InformationListIEs_RL_SetupRspTDD = -1;
static gint ett_rnsap_USCHInformationItem_RL_SetupRspTDD = -1;
static gint ett_rnsap_RL_LCR_InformationResponse_RL_SetupRspTDD = -1;
static gint ett_rnsap_UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD = -1;
static gint ett_rnsap_UL_LCR_CCTrCHInformationItem_RL_SetupRspTDD = -1;
static gint ett_rnsap_UL_DPCH_LCR_InformationItem_RL_SetupRspTDD = -1;
static gint ett_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD = -1;
static gint ett_rnsap_DL_CCTrCH_LCR_InformationItem_RL_SetupRspTDD = -1;
static gint ett_rnsap_DL_DPCH_LCR_InformationItem_RL_SetupRspTDD = -1;
static gint ett_rnsap_DSCH_LCR_InformationListIEs_RL_SetupRspTDD = -1;
static gint ett_rnsap_DSCH_LCR_InformationItem_RL_SetupRspTDD = -1;
static gint ett_rnsap_USCH_LCR_InformationListIEs_RL_SetupRspTDD = -1;
static gint ett_rnsap_USCH_LCR_InformationItem_RL_SetupRspTDD = -1;
static gint ett_rnsap_RL_InformationResponse_RL_SetupRspTDD768 = -1;
static gint ett_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD768 = -1;
static gint ett_rnsap_UL_CCTrCHInformationItem_RL_SetupRspTDD768 = -1;
static gint ett_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD768 = -1;
static gint ett_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD768 = -1;
static gint ett_rnsap_DL_CCTrCHInformationItem_RL_SetupRspTDD768 = -1;
static gint ett_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD768 = -1;
static gint ett_rnsap_RadioLinkSetupFailureFDD = -1;
static gint ett_rnsap_CauseLevel_RL_SetupFailureFDD = -1;
static gint ett_rnsap_GeneralCauseList_RL_SetupFailureFDD = -1;
static gint ett_rnsap_RLSpecificCauseList_RL_SetupFailureFDD = -1;
static gint ett_rnsap_UnsuccessfulRL_InformationResponseList_RL_SetupFailureFDD = -1;
static gint ett_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD = -1;
static gint ett_rnsap_SuccessfulRL_InformationResponseList_RL_SetupFailureFDD = -1;
static gint ett_rnsap_SuccessfulRL_InformationResponse_RL_SetupFailureFDD = -1;
static gint ett_rnsap_DiversityIndication_RL_SetupFailureFDD = -1;
static gint ett_rnsap_Combining_RL_SetupFailureFDD = -1;
static gint ett_rnsap_NonCombiningOrFirstRL_RL_SetupFailureFDD = -1;
static gint ett_rnsap_RadioLinkSetupFailureTDD = -1;
static gint ett_rnsap_CauseLevel_RL_SetupFailureTDD = -1;
static gint ett_rnsap_GeneralCauseList_RL_SetupFailureTDD = -1;
static gint ett_rnsap_RLSpecificCauseList_RL_SetupFailureTDD = -1;
static gint ett_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD = -1;
static gint ett_rnsap_RadioLinkAdditionRequestFDD = -1;
static gint ett_rnsap_RL_InformationList_RL_AdditionRqstFDD = -1;
static gint ett_rnsap_RL_Information_RL_AdditionRqstFDD = -1;
static gint ett_rnsap_Additional_HS_Cell_Information_RL_Addition_List = -1;
static gint ett_rnsap_Additional_HS_Cell_Information_RL_Addition_ItemIEs = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_RL_Add_Req = -1;
static gint ett_rnsap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency = -1;
static gint ett_rnsap_EDPCH_Information_RLAdditionReq_FDD = -1;
static gint ett_rnsap_RadioLinkAdditionRequestTDD = -1;
static gint ett_rnsap_RL_Information_RL_AdditionRqstTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD = -1;
static gint ett_rnsap_RadioLinkAdditionResponseFDD = -1;
static gint ett_rnsap_RL_InformationResponseList_RL_AdditionRspFDD = -1;
static gint ett_rnsap_RL_InformationResponseItem_RL_AdditionRspFDD = -1;
static gint ett_rnsap_DiversityIndication_RL_AdditionRspFDD = -1;
static gint ett_rnsap_Combining_RL_AdditionRspFDD = -1;
static gint ett_rnsap_NonCombining_RL_AdditionRspFDD = -1;
static gint ett_rnsap_Additional_HS_Cell_Change_Information_Response_List = -1;
static gint ett_rnsap_Additional_HS_Cell_Change_Information_Response_ItemIEs = -1;
static gint ett_rnsap_RadioLinkAdditionResponseTDD = -1;
static gint ett_rnsap_RL_InformationResponse_RL_AdditionRspTDD = -1;
static gint ett_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD = -1;
static gint ett_rnsap_UL_CCTrCHInformationItem_RL_AdditionRspTDD = -1;
static gint ett_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD = -1;
static gint ett_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD = -1;
static gint ett_rnsap_DL_CCTrCHInformationItem_RL_AdditionRspTDD = -1;
static gint ett_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD = -1;
static gint ett_rnsap_DCH_Information_RL_AdditionRspTDD = -1;
static gint ett_rnsap_DiversityIndication_RL_AdditionRspTDD = -1;
static gint ett_rnsap_Combining_RL_AdditionRspTDD = -1;
static gint ett_rnsap_NonCombining_RL_AdditionRspTDD = -1;
static gint ett_rnsap_DSCH_InformationListIE_RL_AdditionRspTDD = -1;
static gint ett_rnsap_DSCHInformationItem_RL_AdditionRspTDD = -1;
static gint ett_rnsap_DiversityIndication_RL_AdditionRspTDD2 = -1;
static gint ett_rnsap_USCH_InformationListIE_RL_AdditionRspTDD = -1;
static gint ett_rnsap_USCHInformationItem_RL_AdditionRspTDD = -1;
static gint ett_rnsap_RL_LCR_InformationResponse_RL_AdditionRspTDD = -1;
static gint ett_rnsap_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD = -1;
static gint ett_rnsap_UL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD = -1;
static gint ett_rnsap_UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD = -1;
static gint ett_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD = -1;
static gint ett_rnsap_DL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD = -1;
static gint ett_rnsap_DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD = -1;
static gint ett_rnsap_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD = -1;
static gint ett_rnsap_DSCH_LCR_InformationItem_RL_AdditionRspTDD = -1;
static gint ett_rnsap_USCH_LCR_InformationListIEs_RL_AdditionRspTDD = -1;
static gint ett_rnsap_USCH_LCR_InformationItem_RL_AdditionRspTDD = -1;
static gint ett_rnsap_RL_InformationResponse_RL_AdditionRspTDD768 = -1;
static gint ett_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD768 = -1;
static gint ett_rnsap_UL_CCTrCHInformationItem_RL_AdditionRspTDD768 = -1;
static gint ett_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD768 = -1;
static gint ett_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD768 = -1;
static gint ett_rnsap_DL_CCTrCHInformationItem_RL_AdditionRspTDD768 = -1;
static gint ett_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD768 = -1;
static gint ett_rnsap_RadioLinkAdditionFailureFDD = -1;
static gint ett_rnsap_CauseLevel_RL_AdditionFailureFDD = -1;
static gint ett_rnsap_GeneralCauseList_RL_AdditionFailureFDD = -1;
static gint ett_rnsap_RLSpecificCauseList_RL_AdditionFailureFDD = -1;
static gint ett_rnsap_UnsuccessfulRL_InformationResponseList_RL_AdditionFailureFDD = -1;
static gint ett_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD = -1;
static gint ett_rnsap_SuccessfulRL_InformationResponseList_RL_AdditionFailureFDD = -1;
static gint ett_rnsap_SuccessfulRL_InformationResponse_RL_AdditionFailureFDD = -1;
static gint ett_rnsap_DiversityIndication_RL_AdditionFailureFDD = -1;
static gint ett_rnsap_Combining_RL_AdditionFailureFDD = -1;
static gint ett_rnsap_NonCombining_RL_AdditionFailureFDD = -1;
static gint ett_rnsap_RadioLinkAdditionFailureTDD = -1;
static gint ett_rnsap_CauseLevel_RL_AdditionFailureTDD = -1;
static gint ett_rnsap_GeneralCauseList_RL_AdditionFailureTDD = -1;
static gint ett_rnsap_RLSpecificCauseList_RL_AdditionFailureTDD = -1;
static gint ett_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD = -1;
static gint ett_rnsap_RadioLinkDeletionRequest = -1;
static gint ett_rnsap_RL_InformationList_RL_DeletionRqst = -1;
static gint ett_rnsap_RL_Information_RL_DeletionRqst = -1;
static gint ett_rnsap_RadioLinkDeletionResponse = -1;
static gint ett_rnsap_RadioLinkReconfigurationPrepareFDD = -1;
static gint ett_rnsap_UL_DPCH_Information_RL_ReconfPrepFDD = -1;
static gint ett_rnsap_DL_DPCH_Information_RL_ReconfPrepFDD = -1;
static gint ett_rnsap_DL_DPCH_Power_Information_RL_ReconfPrepFDD = -1;
static gint ett_rnsap_PowerOffsetInformation_RL_ReconfPrepFDD = -1;
static gint ett_rnsap_DCH_DeleteList_RL_ReconfPrepFDD = -1;
static gint ett_rnsap_DCH_DeleteItem_RL_ReconfPrepFDD = -1;
static gint ett_rnsap_RL_InformationList_RL_ReconfPrepFDD = -1;
static gint ett_rnsap_RL_Information_RL_ReconfPrepFDD = -1;
static gint ett_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep = -1;
static gint ett_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Prep = -1;
static gint ett_rnsap_F_DPCH_Information_RL_ReconfPrepFDD = -1;
static gint ett_rnsap_PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD = -1;
static gint ett_rnsap_RadioLinkReconfigurationPrepareTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_UL_CCTrCH_AddInformation_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_CCTrCH_TPCAddList_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_CCTrCH_TPCAddItem_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_CCTrCH_TPCModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_DCH_DeleteList_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_DCH_DeleteItem_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_DSCH_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_DSCH_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_DSCH_DeleteList_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_DSCH_DeleteItem_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_USCH_ModifyList_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_USCH_ModifyItem_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_USCH_DeleteList_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_USCH_DeleteItem_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_RL_Information_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_RL_InformationIE_RL_ReconfPrepTDD = -1;
static gint ett_rnsap_RadioLinkReconfigurationReadyFDD = -1;
static gint ett_rnsap_RL_InformationResponseList_RL_ReconfReadyFDD = -1;
static gint ett_rnsap_RL_InformationResponseItem_RL_ReconfReadyFDD = -1;
static gint ett_rnsap_Additional_HS_Cell_RL_Reconf_Response = -1;
static gint ett_rnsap_Additional_HS_Cell_RL_Reconf_Response_ItemIEs = -1;
static gint ett_rnsap_RadioLinkReconfigurationReadyTDD = -1;
static gint ett_rnsap_RL_InformationResponse_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_UL_CCTrCHInformationListIE_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_UL_DPCH_InformationAddList_RL_ReconfReadyTDD768 = -1;
static gint ett_rnsap_UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_UL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_TDD_UL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_TDD_UL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768 = -1;
static gint ett_rnsap_UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768 = -1;
static gint ett_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD768 = -1;
static gint ett_rnsap_TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD768 = -1;
static gint ett_rnsap_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_UL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DL_CCTrCHInformationListIE_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DL_DPCH_InformationAddList_RL_ReconfReadyTDD768 = -1;
static gint ett_rnsap_DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_TDD_DL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_TDD_DL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768 = -1;
static gint ett_rnsap_DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768 = -1;
static gint ett_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD768 = -1;
static gint ett_rnsap_TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD768 = -1;
static gint ett_rnsap_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DL_DPCH_InformationDeleteItem768_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_DSCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_USCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_Multiple_RL_InformationResponse_RL_ReconfReadyTDD = -1;
static gint ett_rnsap_RadioLinkReconfigurationCommit = -1;
static gint ett_rnsap_RadioLinkReconfigurationFailure = -1;
static gint ett_rnsap_CauseLevel_RL_ReconfFailure = -1;
static gint ett_rnsap_GeneralCauseList_RL_ReconfFailure = -1;
static gint ett_rnsap_RLSpecificCauseList_RL_ReconfFailure = -1;
static gint ett_rnsap_RL_ReconfigurationFailureList_RL_ReconfFailure = -1;
static gint ett_rnsap_RL_ReconfigurationFailure_RL_ReconfFail = -1;
static gint ett_rnsap_RadioLinkReconfigurationCancel = -1;
static gint ett_rnsap_RadioLinkReconfigurationRequestFDD = -1;
static gint ett_rnsap_UL_DPCH_Information_RL_ReconfRqstFDD = -1;
static gint ett_rnsap_DL_DPCH_Information_RL_ReconfRqstFDD = -1;
static gint ett_rnsap_DCH_DeleteList_RL_ReconfRqstFDD = -1;
static gint ett_rnsap_DCH_DeleteItem_RL_ReconfRqstFDD = -1;
static gint ett_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req = -1;
static gint ett_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Req = -1;
static gint ett_rnsap_RL_ReconfigurationRequestFDD_RL_InformationList = -1;
static gint ett_rnsap_RL_ReconfigurationRequestFDD_RL_Information_IEs = -1;
static gint ett_rnsap_RadioLinkReconfigurationRequestTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD = -1;
static gint ett_rnsap_DCH_DeleteList_RL_ReconfRqstTDD = -1;
static gint ett_rnsap_DCH_DeleteItem_RL_ReconfRqstTDD = -1;
static gint ett_rnsap_Multiple_RL_ReconfigurationRequestTDD_RL_Information = -1;
static gint ett_rnsap_RL_ReconfigurationRequestTDD_RL_Information = -1;
static gint ett_rnsap_RadioLinkReconfigurationResponseFDD = -1;
static gint ett_rnsap_RL_InformationResponseList_RL_ReconfRspFDD = -1;
static gint ett_rnsap_RL_InformationResponseItem_RL_ReconfRspFDD = -1;
static gint ett_rnsap_RadioLinkReconfigurationResponseTDD = -1;
static gint ett_rnsap_RL_InformationResponse_RL_ReconfRspTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationList_RL_ReconfRspTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationItem_RL_ReconfRspTDD = -1;
static gint ett_rnsap_DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD = -1;
static gint ett_rnsap_DL_Timeslot_LCR_InformationModifyList_RL_ReconfRspTDD = -1;
static gint ett_rnsap_DL_Timeslot_LCR_InformationModifyItem_RL_ReconfRspTDD = -1;
static gint ett_rnsap_Multiple_RL_InformationResponse_RL_ReconfRspTDD = -1;
static gint ett_rnsap_RadioLinkFailureIndication = -1;
static gint ett_rnsap_Reporting_Object_RL_FailureInd = -1;
static gint ett_rnsap_RL_RL_FailureInd = -1;
static gint ett_rnsap_RL_InformationList_RL_FailureInd = -1;
static gint ett_rnsap_RL_Information_RL_FailureInd = -1;
static gint ett_rnsap_RL_Set_RL_FailureInd = -1;
static gint ett_rnsap_RL_Set_InformationList_RL_FailureInd = -1;
static gint ett_rnsap_RL_Set_Information_RL_FailureInd = -1;
static gint ett_rnsap_CCTrCH_RL_FailureInd = -1;
static gint ett_rnsap_CCTrCH_InformationList_RL_FailureInd = -1;
static gint ett_rnsap_CCTrCH_InformationItem_RL_FailureInd = -1;
static gint ett_rnsap_RadioLinkPreemptionRequiredIndication = -1;
static gint ett_rnsap_RL_InformationList_RL_PreemptRequiredInd = -1;
static gint ett_rnsap_RL_InformationItem_RL_PreemptRequiredInd = -1;
static gint ett_rnsap_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd = -1;
static gint ett_rnsap_HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd = -1;
static gint ett_rnsap_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd = -1;
static gint ett_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd = -1;
static gint ett_rnsap_RadioLinkRestoreIndication = -1;
static gint ett_rnsap_Reporting_Object_RL_RestoreInd = -1;
static gint ett_rnsap_RL_RL_RestoreInd = -1;
static gint ett_rnsap_RL_InformationList_RL_RestoreInd = -1;
static gint ett_rnsap_RL_Information_RL_RestoreInd = -1;
static gint ett_rnsap_RL_Set_RL_RestoreInd = -1;
static gint ett_rnsap_RL_Set_InformationList_RL_RestoreInd = -1;
static gint ett_rnsap_RL_Set_Information_RL_RestoreInd = -1;
static gint ett_rnsap_CCTrCH_RL_RestoreInd = -1;
static gint ett_rnsap_CCTrCH_InformationList_RL_RestoreInd = -1;
static gint ett_rnsap_CCTrCH_InformationItem_RL_RestoreInd = -1;
static gint ett_rnsap_DL_PowerControlRequest = -1;
static gint ett_rnsap_DL_ReferencePowerInformationList_DL_PC_Rqst = -1;
static gint ett_rnsap_DL_ReferencePowerInformation_DL_PC_Rqst = -1;
static gint ett_rnsap_DL_PowerTimeslotControlRequest = -1;
static gint ett_rnsap_PhysicalChannelReconfigurationRequestFDD = -1;
static gint ett_rnsap_RL_Information_PhyChReconfRqstFDD = -1;
static gint ett_rnsap_PhysicalChannelReconfigurationRequestTDD = -1;
static gint ett_rnsap_RL_Information_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_UL_CCTrCH_InformationItem_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_UL_DPCH_InformationItem_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_UL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_UL_Timeslot_InformationItem_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD768 = -1;
static gint ett_rnsap_UL_Timeslot_InformationItem_PhyChReconfRqstTDD768 = -1;
static gint ett_rnsap_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_DL_CCTrCH_InformationItem_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_DL_DPCH_InformationItem_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_DL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_DL_Timeslot_InformationItem_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD768 = -1;
static gint ett_rnsap_DL_Timeslot_InformationItem_PhyChReconfRqstTDD768 = -1;
static gint ett_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_HSPDSCH_Timeslot_InformationItemLCR_PhyChReconfRqstTDD = -1;
static gint ett_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768 = -1;
static gint ett_rnsap_HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD768 = -1;
static gint ett_rnsap_PhysicalChannelReconfigurationCommand = -1;
static gint ett_rnsap_PhysicalChannelReconfigurationFailure = -1;
static gint ett_rnsap_RadioLinkCongestionIndication = -1;
static gint ett_rnsap_RL_InformationList_RL_CongestInd = -1;
static gint ett_rnsap_RL_InformationItem_RL_CongestInd = -1;
static gint ett_rnsap_DCH_Rate_Information_RL_CongestInd = -1;
static gint ett_rnsap_DCH_Rate_InformationItem_RL_CongestInd = -1;
static gint ett_rnsap_EDCH_MacdFlowSpecificInformationList_RL_CongestInd = -1;
static gint ett_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_CongestInd = -1;
static gint ett_rnsap_UplinkSignallingTransferIndicationFDD = -1;
static gint ett_rnsap_UplinkSignallingTransferIndicationTDD = -1;
static gint ett_rnsap_DownlinkSignallingTransferRequest = -1;
static gint ett_rnsap_RelocationCommit = -1;
static gint ett_rnsap_PagingRequest = -1;
static gint ett_rnsap_PagingArea_PagingRqst = -1;
static gint ett_rnsap_URA_PagingRqst = -1;
static gint ett_rnsap_Cell_PagingRqst = -1;
static gint ett_rnsap_CNOriginatedPage_PagingRqst = -1;
static gint ett_rnsap_DedicatedMeasurementInitiationRequest = -1;
static gint ett_rnsap_DedicatedMeasurementObjectType_DM_Rqst = -1;
static gint ett_rnsap_RL_DM_Rqst = -1;
static gint ett_rnsap_RL_InformationList_DM_Rqst = -1;
static gint ett_rnsap_RL_InformationItem_DM_Rqst = -1;
static gint ett_rnsap_HSSICH_Info_DM_Rqst = -1;
static gint ett_rnsap_HSSICH_Info_DM_Rqst_Extension = -1;
static gint ett_rnsap_RL_Set_DM_Rqst = -1;
static gint ett_rnsap_RL_Set_InformationList_DM_Rqst = -1;
static gint ett_rnsap_RL_Set_InformationItem_DM_Rqst = -1;
static gint ett_rnsap_DedicatedMeasurementInitiationResponse = -1;
static gint ett_rnsap_DedicatedMeasurementObjectType_DM_Rsp = -1;
static gint ett_rnsap_RL_DM_Rsp = -1;
static gint ett_rnsap_RL_Set_DM_Rsp = -1;
static gint ett_rnsap_RL_InformationList_DM_Rsp = -1;
static gint ett_rnsap_RL_InformationItem_DM_Rsp = -1;
static gint ett_rnsap_RL_Set_InformationList_DM_Rsp = -1;
static gint ett_rnsap_RL_Set_InformationItem_DM_Rsp = -1;
static gint ett_rnsap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp = -1;
static gint ett_rnsap_Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp = -1;
static gint ett_rnsap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp = -1;
static gint ett_rnsap_Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp = -1;
static gint ett_rnsap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp = -1;
static gint ett_rnsap_Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp = -1;
static gint ett_rnsap_Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp = -1;
static gint ett_rnsap_Multiple_DedicatedMeasurementValueItem_TDD768_DM_Rsp = -1;
static gint ett_rnsap_DedicatedMeasurementInitiationFailure = -1;
static gint ett_rnsap_DedicatedMeasurementObjectType_DM_Fail = -1;
static gint ett_rnsap_RL_DM_Fail = -1;
static gint ett_rnsap_RL_Set_DM_Fail = -1;
static gint ett_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail = -1;
static gint ett_rnsap_RL_Unsuccessful_InformationItem_DM_Fail = -1;
static gint ett_rnsap_RL_Successful_InformationRespList_DM_Fail = -1;
static gint ett_rnsap_RL_Successful_InformationItem_DM_Fail = -1;
static gint ett_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail = -1;
static gint ett_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail = -1;
static gint ett_rnsap_RL_Set_Successful_InformationRespList_DM_Fail = -1;
static gint ett_rnsap_RL_Set_Successful_InformationItem_DM_Fail = -1;
static gint ett_rnsap_DedicatedMeasurementReport = -1;
static gint ett_rnsap_DedicatedMeasurementObjectType_DM_Rprt = -1;
static gint ett_rnsap_RL_DM_Rprt = -1;
static gint ett_rnsap_RL_Set_DM_Rprt = -1;
static gint ett_rnsap_RL_InformationList_DM_Rprt = -1;
static gint ett_rnsap_RL_InformationItem_DM_Rprt = -1;
static gint ett_rnsap_RL_Set_InformationList_DM_Rprt = -1;
static gint ett_rnsap_RL_Set_InformationItem_DM_Rprt = -1;
static gint ett_rnsap_DedicatedMeasurementTerminationRequest = -1;
static gint ett_rnsap_DedicatedMeasurementFailureIndication = -1;
static gint ett_rnsap_DedicatedMeasurementObjectType_DM_Fail_Ind = -1;
static gint ett_rnsap_RL_DM_Fail_Ind = -1;
static gint ett_rnsap_RL_Set_DM_Fail_Ind = -1;
static gint ett_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail_Ind = -1;
static gint ett_rnsap_RL_Unsuccessful_InformationItem_DM_Fail_Ind = -1;
static gint ett_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail_Ind = -1;
static gint ett_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind = -1;
static gint ett_rnsap_CommonTransportChannelResourcesReleaseRequest = -1;
static gint ett_rnsap_CommonTransportChannelResourcesRequest = -1;
static gint ett_rnsap_CommonTransportChannelResourcesResponseFDD = -1;
static gint ett_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD = -1;
static gint ett_rnsap_CommonTransportChannelResourcesResponseTDD = -1;
static gint ett_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD = -1;
static gint ett_rnsap_CommonTransportChannelResourcesFailure = -1;
static gint ett_rnsap_CompressedModeCommand = -1;
static gint ett_rnsap_ErrorIndication = -1;
static gint ett_rnsap_CommonMeasurementInitiationRequest = -1;
static gint ett_rnsap_CommonMeasurementObjectType_CM_Rqst = -1;
static gint ett_rnsap_Cell_CM_Rqst = -1;
static gint ett_rnsap_NeighbouringCellMeasurementInfo = -1;
static gint ett_rnsap_NeighbouringCellMeasurementInfo_item = -1;
static gint ett_rnsap_GSM_Cell_CM_Rqst = -1;
static gint ett_rnsap_GsmCellList_CM_Rqst = -1;
static gint ett_rnsap_GsmCellItem_CM_Rqst = -1;
static gint ett_rnsap_CommonMeasurementInitiationResponse = -1;
static gint ett_rnsap_CommonMeasurementObjectType_CM_Rsp = -1;
static gint ett_rnsap_Cell_CM_Rsp = -1;
static gint ett_rnsap_GsmCellList_CM_Rsp = -1;
static gint ett_rnsap_GsmCellItem_CM_Rsp = -1;
static gint ett_rnsap_CommonMeasurementInitiationFailure = -1;
static gint ett_rnsap_Extension_FailureMeasurementList = -1;
static gint ett_rnsap_Extension_FailureMeasurementItem = -1;
static gint ett_rnsap_CommonMeasurementReport = -1;
static gint ett_rnsap_CommonMeasurementObjectType_CM_Rprt = -1;
static gint ett_rnsap_Cell_CM_Rprt = -1;
static gint ett_rnsap_GsmCellList_CM_Rprt = -1;
static gint ett_rnsap_GsmCellItem_CM_Rprt = -1;
static gint ett_rnsap_CommonMeasurementTerminationRequest = -1;
static gint ett_rnsap_Extension_TerminationMeasurementList = -1;
static gint ett_rnsap_Extension_TerminationMeasurementItem = -1;
static gint ett_rnsap_CommonMeasurementFailureIndication = -1;
static gint ett_rnsap_Extension_FailureIndicationMeasurementList = -1;
static gint ett_rnsap_Extension_FailureIndicationMeasurementItem = -1;
static gint ett_rnsap_InformationExchangeInitiationRequest = -1;
static gint ett_rnsap_InformationExchangeObjectType_InfEx_Rqst = -1;
static gint ett_rnsap_Cell_InfEx_Rqst = -1;
static gint ett_rnsap_GSM_Cell_InfEx_Rqst = -1;
static gint ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst = -1;
static gint ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rqst = -1;
static gint ett_rnsap_MBMS_Bearer_Service_List_InfEx_Rqst = -1;
static gint ett_rnsap_MBMS_Cell_InfEx_Rqst = -1;
static gint ett_rnsap_ANR_Cell_InfEx_Rqst = -1;
static gint ett_rnsap_Common_E_RGCH_Cell_InfEx_Rqst = -1;
static gint ett_rnsap_InformationExchangeInitiationResponse = -1;
static gint ett_rnsap_InformationExchangeObjectType_InfEx_Rsp = -1;
static gint ett_rnsap_Cell_InfEx_Rsp = -1;
static gint ett_rnsap_MBMS_Bearer_Service_List_InfEx_Rsp = -1;
static gint ett_rnsap_MBMS_Bearer_ServiceItemIEs_InfEx_Rsp = -1;
static gint ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp = -1;
static gint ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rsp = -1;
static gint ett_rnsap_MBMS_Cell_InfEx_Rsp = -1;
static gint ett_rnsap_MBMS_Cell_Item_InfEx_Rsp = -1;
static gint ett_rnsap_ANR_Cell_InfEx_Rsp = -1;
static gint ett_rnsap_ANR_Cell_ItemIEs_InfEx_Rsp = -1;
static gint ett_rnsap_Common_E_RGCH_Cell_InfEx_Rsp = -1;
static gint ett_rnsap_Common_E_RGCH_Cell_ItemIEs_InfEx_Rsp = -1;
static gint ett_rnsap_InformationExchangeInitiationFailure = -1;
static gint ett_rnsap_InformationReport = -1;
static gint ett_rnsap_InformationExchangeObjectType_InfEx_Rprt = -1;
static gint ett_rnsap_Cell_InfEx_Rprt = -1;
static gint ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt = -1;
static gint ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rprt = -1;
static gint ett_rnsap_MBMS_Bearer_Service_List_InfEx_Rprt = -1;
static gint ett_rnsap_MBMS_Bearer_Service_List_Item_InfEx_Rprt = -1;
static gint ett_rnsap_MBMS_Cell_InfEx_Rprt = -1;
static gint ett_rnsap_MBMS_Cell_Item_InfEx_Rprt = -1;
static gint ett_rnsap_Common_E_RGCH_Cell_InfEx_Rprt = -1;
static gint ett_rnsap_Common_E_RGCH_Cell_Item_InfEx_Rprt = -1;
static gint ett_rnsap_InformationExchangeTerminationRequest = -1;
static gint ett_rnsap_InformationExchangeFailureIndication = -1;
static gint ett_rnsap_ResetRequest = -1;
static gint ett_rnsap_ResetIndicator = -1;
static gint ett_rnsap_ContextList_Reset = -1;
static gint ett_rnsap_ContextInfoList_Reset = -1;
static gint ett_rnsap_ContextInfoItem_Reset = -1;
static gint ett_rnsap_ContextType_Reset = -1;
static gint ett_rnsap_ContextGroupList_Reset = -1;
static gint ett_rnsap_ContextGroupInfoList_Reset = -1;
static gint ett_rnsap_ContextGroupInfoItem_Reset = -1;
static gint ett_rnsap_ResetResponse = -1;
static gint ett_rnsap_RadioLinkActivationCommandFDD = -1;
static gint ett_rnsap_DelayedActivationInformationList_RL_ActivationCmdFDD = -1;
static gint ett_rnsap_DelayedActivationInformation_RL_ActivationCmdFDD = -1;
static gint ett_rnsap_RadioLinkActivationCommandTDD = -1;
static gint ett_rnsap_DelayedActivationInformationList_RL_ActivationCmdTDD = -1;
static gint ett_rnsap_DelayedActivationInformation_RL_ActivationCmdTDD = -1;
static gint ett_rnsap_GERANUplinkSignallingTransferIndication = -1;
static gint ett_rnsap_RadioLinkParameterUpdateIndicationFDD = -1;
static gint ett_rnsap_RL_ParameterUpdateIndicationFDD_RL_InformationList = -1;
static gint ett_rnsap_RL_ParameterUpdateIndicationFDD_RL_Information_Item = -1;
static gint ett_rnsap_UE_Measurement_Forwarding = -1;
static gint ett_rnsap_Additional_HS_Cell_Information_RL_Param_Upd = -1;
static gint ett_rnsap_Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd = -1;
static gint ett_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs = -1;
static gint ett_rnsap_RadioLinkParameterUpdateIndicationTDD = -1;
static gint ett_rnsap_UEMeasurementInitiationRequest = -1;
static gint ett_rnsap_UEMeasurementInitiationResponse = -1;
static gint ett_rnsap_UEMeasurementInitiationFailure = -1;
static gint ett_rnsap_UEMeasurementReport = -1;
static gint ett_rnsap_UEMeasurementTerminationRequest = -1;
static gint ett_rnsap_UEMeasurementFailureIndication = -1;
static gint ett_rnsap_IurInvokeTrace = -1;
static gint ett_rnsap_ListOfInterfacesToTrace = -1;
static gint ett_rnsap_InterfacesToTraceItem = -1;
static gint ett_rnsap_IurDeactivateTrace = -1;
static gint ett_rnsap_MBMSAttachCommand = -1;
static gint ett_rnsap_MBMSDetachCommand = -1;
static gint ett_rnsap_DirectInformationTransfer = -1;
static gint ett_rnsap_EnhancedRelocationRequest = -1;
static gint ett_rnsap_EnhancedRelocationResponse = -1;
static gint ett_rnsap_EnhancedRelocationFailure = -1;
static gint ett_rnsap_EnhancedRelocationCancel = -1;
static gint ett_rnsap_EnhancedRelocationSignallingTransfer = -1;
static gint ett_rnsap_EnhancedRelocationRelease = -1;
static gint ett_rnsap_MBSFNMCCHInformation = -1;
static gint ett_rnsap_SecondaryULFrequencyReport = -1;
static gint ett_rnsap_SecondaryULFrequencyUpdateIndication = -1;
static gint ett_rnsap_EnhancedRelocationResourceRequest = -1;
static gint ett_rnsap_EnhancedRelocationResourceResponse = -1;
static gint ett_rnsap_EnhancedRelocationResourceFailure = -1;
static gint ett_rnsap_EnhancedRelocationResourceReleaseCommand = -1;
static gint ett_rnsap_EnhancedRelocationResourceReleaseComplete = -1;
static gint ett_rnsap_PrivateMessage = -1;
static gint ett_rnsap_InformationTransferControlRequest = -1;
static gint ett_rnsap_Control_Type_InformationTransferControlReq = -1;
static gint ett_rnsap_Suspension_Control_Type = -1;
static gint ett_rnsap_Resume_Control_Type = -1;
static gint ett_rnsap_RNSAP_PDU = -1;
static gint ett_rnsap_InitiatingMessage = -1;
static gint ett_rnsap_SuccessfulOutcome = -1;
static gint ett_rnsap_UnsuccessfulOutcome = -1;
static gint ett_rnsap_Outcome = -1;


/* Dissector tables */
static dissector_table_t rnsap_ies_dissector_table;
static dissector_table_t rnsap_extension_dissector_table;
static dissector_table_t rnsap_proc_imsg_dissector_table;
static dissector_table_t rnsap_proc_sout_dissector_table;
static dissector_table_t rnsap_proc_uout_dissector_table;

static dissector_handle_t rnsap_handle;

static int dissect_ProtocolIEFieldValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *);
static int dissect_ProtocolExtensionFieldExtensionValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *);
static int dissect_PrivateIEFieldValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *);
static int dissect_InitiatingMessageValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *);
static int dissect_SuccessfulOutcomeValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *);
static int dissect_UnsuccessfulOutcomeValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *);

static rnsap_private_data_t *
rnsap_get_private_data(packet_info *pinfo)
{

    rnsap_private_data_t *pdata = (rnsap_private_data_t *)p_get_proto_data(pinfo->pool, pinfo, proto_rnsap, 0);
    if (!pdata) {
        pdata = wmem_new0(pinfo->pool, rnsap_private_data_t);
        pdata->ProcedureCode = 0xFFFF;
        pdata->ddMode = 0xFFFF;
        p_add_proto_data(pinfo->pool, pinfo, proto_rnsap, 0, pdata);
    }
    return pdata;
}


static const value_string rnsap_Criticality_vals[] = {
  {   0, "reject" },
  {   1, "ignore" },
  {   2, "notify" },
  { 0, NULL }
};


static int
dissect_rnsap_Criticality(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_maxPrivateIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxPrivateIEs, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_T_global(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_object_identifier_str(tvb, offset, actx, tree, hf_index, &rnsap_get_private_data(actx->pinfo)->obj_id);

  return offset;
}


static const value_string rnsap_PrivateIE_ID_vals[] = {
  {   0, "local" },
  {   1, "global" },
  { 0, NULL }
};

static const per_choice_t PrivateIE_ID_choice[] = {
  {   0, &hf_rnsap_local         , ASN1_NO_EXTENSIONS     , dissect_rnsap_INTEGER_0_maxPrivateIEs },
  {   1, &hf_rnsap_global        , ASN1_NO_EXTENSIONS     , dissect_rnsap_T_global },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_PrivateIE_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_PrivateIE_ID, PrivateIE_ID_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_ProcedureCode_vals[] = {
  { RNSAP_ID_COMMONTRANSPORTCHANNELRESOURCESINITIALISATION, "id-commonTransportChannelResourcesInitialisation" },
  { RNSAP_ID_COMMONTRANSPORTCHANNELRESOURCESRELEASE, "id-commonTransportChannelResourcesRelease" },
  { RNSAP_ID_COMPRESSEDMODECOMMAND, "id-compressedModeCommand" },
  { RNSAP_ID_DOWNLINKPOWERCONTROL, "id-downlinkPowerControl" },
  { RNSAP_ID_DOWNLINKPOWERTIMESLOTCONTROL, "id-downlinkPowerTimeslotControl" },
  { RNSAP_ID_DOWNLINKSIGNALLINGTRANSFER, "id-downlinkSignallingTransfer" },
  { RNSAP_ID_ERRORINDICATION, "id-errorIndication" },
  { RNSAP_ID_DEDICATEDMEASUREMENTFAILURE, "id-dedicatedMeasurementFailure" },
  { RNSAP_ID_DEDICATEDMEASUREMENTINITIATION, "id-dedicatedMeasurementInitiation" },
  { RNSAP_ID_DEDICATEDMEASUREMENTREPORTING, "id-dedicatedMeasurementReporting" },
  { RNSAP_ID_DEDICATEDMEASUREMENTTERMINATION, "id-dedicatedMeasurementTermination" },
  { RNSAP_ID_PAGING, "id-paging" },
  { RNSAP_ID_PHYSICALCHANNELRECONFIGURATION, "id-physicalChannelReconfiguration" },
  { RNSAP_ID_PRIVATEMESSAGE, "id-privateMessage" },
  { RNSAP_ID_RADIOLINKADDITION, "id-radioLinkAddition" },
  { RNSAP_ID_RADIOLINKDELETION, "id-radioLinkDeletion" },
  { RNSAP_ID_RADIOLINKFAILURE, "id-radioLinkFailure" },
  { RNSAP_ID_RADIOLINKPREEMPTION, "id-radioLinkPreemption" },
  { RNSAP_ID_RADIOLINKRESTORATION, "id-radioLinkRestoration" },
  { RNSAP_ID_RADIOLINKSETUP, "id-radioLinkSetup" },
  { RNSAP_ID_RELOCATIONCOMMIT, "id-relocationCommit" },
  { RNSAP_ID_SYNCHRONISEDRADIOLINKRECONFIGURATIONCANCELLATION, "id-synchronisedRadioLinkReconfigurationCancellation" },
  { RNSAP_ID_SYNCHRONISEDRADIOLINKRECONFIGURATIONCOMMIT, "id-synchronisedRadioLinkReconfigurationCommit" },
  { RNSAP_ID_SYNCHRONISEDRADIOLINKRECONFIGURATIONPREPARATION, "id-synchronisedRadioLinkReconfigurationPreparation" },
  { RNSAP_ID_UNSYNCHRONISEDRADIOLINKRECONFIGURATION, "id-unSynchronisedRadioLinkReconfiguration" },
  { RNSAP_ID_UPLINKSIGNALLINGTRANSFER, "id-uplinkSignallingTransfer" },
  { RNSAP_ID_COMMONMEASUREMENTFAILURE, "id-commonMeasurementFailure" },
  { RNSAP_ID_COMMONMEASUREMENTINITIATION, "id-commonMeasurementInitiation" },
  { RNSAP_ID_COMMONMEASUREMENTREPORTING, "id-commonMeasurementReporting" },
  { RNSAP_ID_COMMONMEASUREMENTTERMINATION, "id-commonMeasurementTermination" },
  { RNSAP_ID_INFORMATIONEXCHANGEFAILURE, "id-informationExchangeFailure" },
  { RNSAP_ID_INFORMATIONEXCHANGEINITIATION, "id-informationExchangeInitiation" },
  { RNSAP_ID_INFORMATIONREPORTING, "id-informationReporting" },
  { RNSAP_ID_INFORMATIONEXCHANGETERMINATION, "id-informationExchangeTermination" },
  { RNSAP_ID_RADIOLINKCONGESTION, "id-radioLinkCongestion" },
  { RNSAP_ID_RESET, "id-reset" },
  { RNSAP_ID_RADIOLINKACTIVATION, "id-radioLinkActivation" },
  { RNSAP_ID_GERANUPLINKSIGNALLINGTRANSFER, "id-gERANuplinkSignallingTransfer" },
  { RNSAP_ID_RADIOLINKPARAMETERUPDATE, "id-radioLinkParameterUpdate" },
  { RNSAP_ID_UEMEASUREMENTFAILURE, "id-uEMeasurementFailure" },
  { RNSAP_ID_UEMEASUREMENTINITIATION, "id-uEMeasurementInitiation" },
  { RNSAP_ID_UEMEASUREMENTREPORTING, "id-uEMeasurementReporting" },
  { RNSAP_ID_UEMEASUREMENTTERMINATION, "id-uEMeasurementTermination" },
  { RNSAP_ID_IURDEACTIVATETRACE, "id-iurDeactivateTrace" },
  { RNSAP_ID_IURINVOKETRACE, "id-iurInvokeTrace" },
  { RNSAP_ID_MBMSATTACH, "id-mBMSAttach" },
  { RNSAP_ID_MBMSDETACH, "id-mBMSDetach" },
  { RNSAP_ID_NOT_USED_47, "id-Not-used-47" },
  { RNSAP_ID_DIRECTINFORMATIONTRANSFER, "id-directInformationTransfer" },
  { RNSAP_ID_ENHANCEDRELOCATION, "id-enhancedRelocation" },
  { RNSAP_ID_ENHANCEDRELOCATIONCANCEL, "id-enhancedRelocationCancel" },
  { RNSAP_ID_ENHANCEDRELOCATIONSIGNALLINGTRANSFER, "id-enhancedRelocationSignallingTransfer" },
  { RNSAP_ID_ENHANCEDRELOCATIONRELEASE, "id-enhancedRelocationRelease" },
  { RNSAP_ID_MBSFNMCCHINFORMATION, "id-mBSFNMCCHInformation" },
  { RNSAP_ID_SECONDARYULFREQUENCYREPORTING, "id-secondaryULFrequencyReporting" },
  { RNSAP_ID_SECONDARYULFREQUENCYUPDATE, "id-secondaryULFrequencyUpdate" },
  { RNSAP_ID_INFORMATIONTRANSFERCONTROL, "id-informationTransferControl" },
  { RNSAP_ID_NOT_USED_57, "id-Not-used-57" },
  { RNSAP_ID_NOT_USED_58, "id-Not-used-58" },
  { RNSAP_ID_NOT_USED_59, "id-Not-used-59" },
  { RNSAP_ID_ENHANCEDRELOCATIONRESOURCEALLOCATION, "id-enhancedRelocationResourceAllocation" },
  { RNSAP_ID_ENHANCEDRELOCATIONRESOURCERELEASE, "id-enhancedRelocationResourceRelease" },
  { 0, NULL }
};

static value_string_ext rnsap_ProcedureCode_vals_ext = VALUE_STRING_EXT_INIT(rnsap_ProcedureCode_vals);


static int
dissect_rnsap_ProcedureCode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  rnsap_private_data_t *pdata = rnsap_get_private_data(actx->pinfo);
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, &pdata->ProcedureCode, FALSE);

   col_add_fstr(actx->pinfo->cinfo, COL_INFO, "%s ",
                val_to_str_ext_const(pdata->ProcedureCode, &rnsap_ProcedureCode_vals_ext,
                           "unknown message"));
  return offset;
}


static const value_string rnsap_DdMode_vals[] = {
  { tdd, "tdd" },
  { fdd, "fdd" },
  { common, "common" },
  { 0, NULL }
};


static int
dissect_rnsap_DdMode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, &rnsap_get_private_data(actx->pinfo)->ddMode, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t ProcedureID_sequence[] = {
  { &hf_rnsap_procedureCode , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProcedureCode },
  { &hf_rnsap_ddMode        , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DdMode },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ProcedureID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  rnsap_private_data_t *pdata = rnsap_get_private_data(actx->pinfo);
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ProcedureID, ProcedureID_sequence);

  pdata->ProcedureID = wmem_strdup_printf(actx->pinfo->pool, "%s/%s",
                                 val_to_str_ext(pdata->ProcedureCode, &rnsap_ProcedureCode_vals_ext, "unknown(%u)"),
                                 val_to_str(pdata->ddMode, rnsap_DdMode_vals, "unknown(%u)"));
  return offset;
}


static const value_string rnsap_ProtocolIE_ID_vals[] = {
  { id_RL_InformationList_RL_PreemptRequiredInd, "id-RL-InformationList-RL-PreemptRequiredInd" },
  { id_RL_InformationItem_RL_PreemptRequiredInd, "id-RL-InformationItem-RL-PreemptRequiredInd" },
  { id_GA_CellAdditionalShapes, "id-GA-CellAdditionalShapes" },
  { id_AllowedQueuingTime, "id-AllowedQueuingTime" },
  { id_BindingID, "id-BindingID" },
  { id_C_ID, "id-C-ID" },
  { id_C_RNTI, "id-C-RNTI" },
  { id_CFN, "id-CFN" },
  { id_CN_CS_DomainIdentifier, "id-CN-CS-DomainIdentifier" },
  { id_CN_PS_DomainIdentifier, "id-CN-PS-DomainIdentifier" },
  { id_Cause, "id-Cause" },
  { id_DPC_Mode, "id-DPC-Mode" },
  { id_Neighbouring_GSM_CellInformation, "id-Neighbouring-GSM-CellInformation" },
  { id_CFNReportingIndicator, "id-CFNReportingIndicator" },
  { id_CCTrCH_InformationItem_RL_FailureInd, "id-CCTrCH-InformationItem-RL-FailureInd" },
  { id_CCTrCH_InformationItem_RL_RestoreInd, "id-CCTrCH-InformationItem-RL-RestoreInd" },
  { id_Permanent_NAS_UE_Identity, "id-Permanent-NAS-UE-Identity" },
  { id_CongestionCause, "id-CongestionCause" },
  { id_DPC_Mode_Change_SupportIndicator, "id-DPC-Mode-Change-SupportIndicator" },
  { id_CriticalityDiagnostics, "id-CriticalityDiagnostics" },
  { id_D_RNTI, "id-D-RNTI" },
  { id_D_RNTI_ReleaseIndication, "id-D-RNTI-ReleaseIndication" },
  { id_CNOriginatedPage_PagingRqst, "id-CNOriginatedPage-PagingRqst" },
  { id_InnerLoopDLPCStatus, "id-InnerLoopDLPCStatus" },
  { id_PropagationDelay, "id-PropagationDelay" },
  { id_DCHs_to_Add_FDD, "id-DCHs-to-Add-FDD" },
  { id_DCHs_to_Add_TDD, "id-DCHs-to-Add-TDD" },
  { id_RL_InformationResponse_RL_ReconfRspTDD, "id-RL-InformationResponse-RL-ReconfRspTDD" },
  { id_Unused_ProtocolIE_ID_29, "id-Unused-ProtocolIE-ID-29" },
  { id_DCH_DeleteList_RL_ReconfPrepFDD, "id-DCH-DeleteList-RL-ReconfPrepFDD" },
  { id_DCH_DeleteList_RL_ReconfPrepTDD, "id-DCH-DeleteList-RL-ReconfPrepTDD" },
  { id_DCH_DeleteList_RL_ReconfRqstFDD, "id-DCH-DeleteList-RL-ReconfRqstFDD" },
  { id_DCH_DeleteList_RL_ReconfRqstTDD, "id-DCH-DeleteList-RL-ReconfRqstTDD" },
  { id_DCH_FDD_Information, "id-DCH-FDD-Information" },
  { id_DCH_TDD_Information, "id-DCH-TDD-Information" },
  { id_RxTimingDeviationForTA, "id-RxTimingDeviationForTA" },
  { id_timeSlot_ISCP, "id-timeSlot-ISCP" },
  { id_DCH_Rate_InformationItem_RL_CongestInd, "id-DCH-Rate-InformationItem-RL-CongestInd" },
  { id_FDD_DCHs_to_Modify, "id-FDD-DCHs-to-Modify" },
  { id_TDD_DCHs_to_Modify, "id-TDD-DCHs-to-Modify" },
  { id_Guaranteed_Rate_Information, "id-Guaranteed-Rate-Information" },
  { id_Allowed_Rate_Information, "id-Allowed-Rate-Information" },
  { id_DCH_InformationResponse, "id-DCH-InformationResponse" },
  { id_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD, "id-DL-CCTrCH-InformationAddItem-RL-ReconfPrepTDD" },
  { id_DL_CCTrCH_InformationListIE_RL_ReconfReadyTDD, "id-DL-CCTrCH-InformationListIE-RL-ReconfReadyTDD" },
  { id_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD, "id-DL-CCTrCH-InformationDeleteItem-RL-ReconfRqstTDD" },
  { id_DL_CCTrCH_InformationItem_RL_SetupRqstTDD, "id-DL-CCTrCH-InformationItem-RL-SetupRqstTDD" },
  { id_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD, "id-DL-CCTrCH-InformationListIE-PhyChReconfRqstTDD" },
  { id_DL_CCTrCH_InformationListIE_RL_AdditionRspTDD, "id-DL-CCTrCH-InformationListIE-RL-AdditionRspTDD" },
  { id_DL_CCTrCH_InformationListIE_RL_SetupRspTDD, "id-DL-CCTrCH-InformationListIE-RL-SetupRspTDD" },
  { id_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD, "id-DL-CCTrCH-InformationAddList-RL-ReconfPrepTDD" },
  { id_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD, "id-DL-CCTrCH-InformationDeleteList-RL-ReconfRqstTDD" },
  { id_DL_CCTrCH_InformationList_RL_SetupRqstTDD, "id-DL-CCTrCH-InformationList-RL-SetupRqstTDD" },
  { id_FDD_DL_CodeInformation, "id-FDD-DL-CodeInformation" },
  { id_RL_InformationItem_RL_CongestInd, "id-RL-InformationItem-RL-CongestInd" },
  { id_RL_InformationList_RL_CongestInd, "id-RL-InformationList-RL-CongestInd" },
  { id_MessageStructure, "id-MessageStructure" },
  { id_neighbouring_LCR_TDD_CellInformation, "id-neighbouring-LCR-TDD-CellInformation" },
  { id_DL_DPCH_Information_RL_ReconfPrepFDD, "id-DL-DPCH-Information-RL-ReconfPrepFDD" },
  { id_DL_DPCH_Information_RL_SetupRqstFDD, "id-DL-DPCH-Information-RL-SetupRqstFDD" },
  { id_DL_DPCH_Information_RL_ReconfRqstFDD, "id-DL-DPCH-Information-RL-ReconfRqstFDD" },
  { id_DL_DPCH_InformationItem_PhyChReconfRqstTDD, "id-DL-DPCH-InformationItem-PhyChReconfRqstTDD" },
  { id_DL_DPCH_InformationItem_RL_AdditionRspTDD, "id-DL-DPCH-InformationItem-RL-AdditionRspTDD" },
  { id_DL_DPCH_InformationItem_RL_SetupRspTDD, "id-DL-DPCH-InformationItem-RL-SetupRspTDD" },
  { id_DL_Timeslot_ISCP_LCR_Information_RL_SetupRqstTDD, "id-DL-Timeslot-ISCP-LCR-Information-RL-SetupRqstTDD" },
  { id_RL_LCR_InformationResponse_RL_SetupRspTDD, "id-RL-LCR-InformationResponse-RL-SetupRspTDD" },
  { id_DLReferencePower, "id-DLReferencePower" },
  { id_DLReferencePowerList_DL_PC_Rqst, "id-DLReferencePowerList-DL-PC-Rqst" },
  { id_DL_ReferencePowerInformation_DL_PC_Rqst, "id-DL-ReferencePowerInformation-DL-PC-Rqst" },
  { id_DRXCycleLengthCoefficient, "id-DRXCycleLengthCoefficient" },
  { id_DedicatedMeasurementObjectType_DM_Rprt, "id-DedicatedMeasurementObjectType-DM-Rprt" },
  { id_DedicatedMeasurementObjectType_DM_Rqst, "id-DedicatedMeasurementObjectType-DM-Rqst" },
  { id_DedicatedMeasurementObjectType_DM_Rsp, "id-DedicatedMeasurementObjectType-DM-Rsp" },
  { id_DedicatedMeasurementType, "id-DedicatedMeasurementType" },
  { id_UL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD, "id-UL-CCTrCH-LCR-InformationListIE-RL-SetupRspTDD" },
  { id_UL_DPCH_LCR_InformationItem_RL_SetupRspTDD, "id-UL-DPCH-LCR-InformationItem-RL-SetupRspTDD" },
  { id_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD, "id-DL-CCTrCH-LCR-InformationListIE-RL-SetupRspTDD" },
  { id_DL_DPCH_LCR_InformationItem_RL_SetupRspTDD, "id-DL-DPCH-LCR-InformationItem-RL-SetupRspTDD" },
  { id_DSCH_LCR_InformationListIEs_RL_SetupRspTDD, "id-DSCH-LCR-InformationListIEs-RL-SetupRspTDD" },
  { id_USCH_LCR_InformationListIEs_RL_SetupRspTDD, "id-USCH-LCR-InformationListIEs-RL-SetupRspTDD" },
  { id_DL_Timeslot_ISCP_LCR_Information_RL_AdditionRqstTDD, "id-DL-Timeslot-ISCP-LCR-Information-RL-AdditionRqstTDD" },
  { id_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD, "id-FACH-InfoForUESelectedS-CCPCH-CTCH-ResourceRspFDD" },
  { id_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD, "id-FACH-InfoForUESelectedS-CCPCH-CTCH-ResourceRspTDD" },
  { id_IMSI, "id-IMSI" },
  { id_L3_Information, "id-L3-Information" },
  { id_RL_LCR_InformationResponse_RL_AdditionRspTDD, "id-RL-LCR-InformationResponse-RL-AdditionRspTDD" },
  { id_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD, "id-UL-CCTrCH-LCR-InformationListIE-RL-AdditionRspTDD" },
  { id_UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD, "id-UL-DPCH-LCR-InformationItem-RL-AdditionRspTDD" },
  { id_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD, "id-DL-CCTrCH-LCR-InformationListIE-RL-AdditionRspTDD" },
  { id_AdjustmentPeriod, "id-AdjustmentPeriod" },
  { id_MaxAdjustmentStep, "id-MaxAdjustmentStep" },
  { id_MeasurementFilterCoefficient, "id-MeasurementFilterCoefficient" },
  { id_MeasurementID, "id-MeasurementID" },
  { id_DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD, "id-DL-DPCH-LCR-InformationItem-RL-AdditionRspTDD" },
  { id_Neighbouring_UMTS_CellInformationItem, "id-Neighbouring-UMTS-CellInformationItem" },
  { id_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD, "id-DSCH-LCR-InformationListIEs-RL-AdditionRspTDD" },
  { id_USCH_LCR_InformationListIEs_RL_AdditionRspTDD, "id-USCH-LCR-InformationListIEs-RL-AdditionRspTDD" },
  { id_UL_DPCH_LCR_InformationAddListIE_RL_ReconfReadyTDD, "id-UL-DPCH-LCR-InformationAddListIE-RL-ReconfReadyTDD" },
  { id_Unknown_99, "id-Unknown-99" },
  { id_UL_Timeslot_LCR_InformationModifyList_RL_ReconfReadyTDD, "id-UL-Timeslot-LCR-InformationModifyList-RL-ReconfReadyTDD" },
  { id_DL_DPCH_LCR_InformationAddListIE_RL_ReconfReadyTDD, "id-DL-DPCH-LCR-InformationAddListIE-RL-ReconfReadyTDD" },
  { id_PagingArea_PagingRqst, "id-PagingArea-PagingRqst" },
  { id_FACH_FlowControlInformation, "id-FACH-FlowControlInformation" },
  { id_DL_Timeslot_LCR_InformationModifyList_RL_ReconfReadyTDD, "id-DL-Timeslot-LCR-InformationModifyList-RL-ReconfReadyTDD" },
  { id_UL_Timeslot_LCR_InformationList_PhyChReconfRqstTDD, "id-UL-Timeslot-LCR-InformationList-PhyChReconfRqstTDD" },
  { id_DL_Timeslot_LCR_InformationList_PhyChReconfRqstTDD, "id-DL-Timeslot-LCR-InformationList-PhyChReconfRqstTDD" },
  { id_PowerAdjustmentType, "id-PowerAdjustmentType" },
  { id_Unknown_108, "id-Unknown-108" },
  { id_RANAP_RelocationInformation, "id-RANAP-RelocationInformation" },
  { id_RL_Information_PhyChReconfRqstFDD, "id-RL-Information-PhyChReconfRqstFDD" },
  { id_RL_Information_PhyChReconfRqstTDD, "id-RL-Information-PhyChReconfRqstTDD" },
  { id_RL_Information_RL_AdditionRqstFDD, "id-RL-Information-RL-AdditionRqstFDD" },
  { id_RL_Information_RL_AdditionRqstTDD, "id-RL-Information-RL-AdditionRqstTDD" },
  { id_RL_Information_RL_DeletionRqst, "id-RL-Information-RL-DeletionRqst" },
  { id_RL_Information_RL_FailureInd, "id-RL-Information-RL-FailureInd" },
  { id_RL_Information_RL_ReconfPrepFDD, "id-RL-Information-RL-ReconfPrepFDD" },
  { id_RL_Information_RL_RestoreInd, "id-RL-Information-RL-RestoreInd" },
  { id_RL_Information_RL_SetupRqstFDD, "id-RL-Information-RL-SetupRqstFDD" },
  { id_RL_Information_RL_SetupRqstTDD, "id-RL-Information-RL-SetupRqstTDD" },
  { id_RL_InformationItem_DM_Rprt, "id-RL-InformationItem-DM-Rprt" },
  { id_RL_InformationItem_DM_Rqst, "id-RL-InformationItem-DM-Rqst" },
  { id_RL_InformationItem_DM_Rsp, "id-RL-InformationItem-DM-Rsp" },
  { id_RL_InformationItem_RL_SetupRqstFDD, "id-RL-InformationItem-RL-SetupRqstFDD" },
  { id_RL_InformationList_RL_AdditionRqstFDD, "id-RL-InformationList-RL-AdditionRqstFDD" },
  { id_RL_InformationList_RL_DeletionRqst, "id-RL-InformationList-RL-DeletionRqst" },
  { id_RL_InformationList_RL_ReconfPrepFDD, "id-RL-InformationList-RL-ReconfPrepFDD" },
  { id_RL_InformationResponse_RL_AdditionRspTDD, "id-RL-InformationResponse-RL-AdditionRspTDD" },
  { id_RL_InformationResponse_RL_ReconfReadyTDD, "id-RL-InformationResponse-RL-ReconfReadyTDD" },
  { id_RL_InformationResponse_RL_SetupRspTDD, "id-RL-InformationResponse-RL-SetupRspTDD" },
  { id_RL_InformationResponseItem_RL_AdditionRspFDD, "id-RL-InformationResponseItem-RL-AdditionRspFDD" },
  { id_RL_InformationResponseItem_RL_ReconfReadyFDD, "id-RL-InformationResponseItem-RL-ReconfReadyFDD" },
  { id_RL_InformationResponseItem_RL_ReconfRspFDD, "id-RL-InformationResponseItem-RL-ReconfRspFDD" },
  { id_RL_InformationResponseItem_RL_SetupRspFDD, "id-RL-InformationResponseItem-RL-SetupRspFDD" },
  { id_RL_InformationResponseList_RL_AdditionRspFDD, "id-RL-InformationResponseList-RL-AdditionRspFDD" },
  { id_RL_InformationResponseList_RL_ReconfReadyFDD, "id-RL-InformationResponseList-RL-ReconfReadyFDD" },
  { id_RL_InformationResponseList_RL_ReconfRspFDD, "id-RL-InformationResponseList-RL-ReconfRspFDD" },
  { id_RL_InformationResponseList_RL_SetupRspFDD, "id-RL-InformationResponseList-RL-SetupRspFDD" },
  { id_timeSlot_ISCP_LCR_List_DL_PC_Rqst_TDD, "id-timeSlot-ISCP-LCR-List-DL-PC-Rqst-TDD" },
  { id_TSTD_Support_Indicator_RL_SetupRqstTDD, "id-TSTD-Support-Indicator-RL-SetupRqstTDD" },
  { id_TypeOfError, "id-TypeOfError" },
  { id_RL_ReconfigurationFailure_RL_ReconfFail, "id-RL-ReconfigurationFailure-RL-ReconfFail" },
  { id_RestrictionStateIndicator, "id-RestrictionStateIndicator" },
  { id_RL_Set_InformationItem_DM_Rprt, "id-RL-Set-InformationItem-DM-Rprt" },
  { id_RL_Set_InformationItem_DM_Rqst, "id-RL-Set-InformationItem-DM-Rqst" },
  { id_RL_Set_InformationItem_DM_Rsp, "id-RL-Set-InformationItem-DM-Rsp" },
  { id_RL_Set_Information_RL_FailureInd, "id-RL-Set-Information-RL-FailureInd" },
  { id_RL_Set_Information_RL_RestoreInd, "id-RL-Set-Information-RL-RestoreInd" },
  { id_Angle_Of_Arrival_Value_LCR, "id-Angle-Of-Arrival-Value-LCR" },
  { id_Unknown_149, "id-Unknown-149" },
  { id_UL_SIR_Target_CCTrCH_InformationItem_RL_SetupRspTDD, "id-UL-SIR-Target-CCTrCH-InformationItem-RL-SetupRspTDD" },
  { id_UL_SIR_Target_CCTrCH_LCR_InformationItem_RL_SetupRspTDD, "id-UL-SIR-Target-CCTrCH-LCR-InformationItem-RL-SetupRspTDD" },
  { id_ReportCharacteristics, "id-ReportCharacteristics" },
  { id_Reporting_Object_RL_FailureInd, "id-Reporting-Object-RL-FailureInd" },
  { id_Reporing_Object_RL_RestoreInd, "id-Reporing-Object-RL-RestoreInd" },
  { id_S_RNTI, "id-S-RNTI" },
  { id_SAI, "id-SAI" },
  { id_SRNC_ID, "id-SRNC-ID" },
  { id_TrafficClass, "id-TrafficClass" },
  { id_SuccessfulRL_InformationResponse_RL_AdditionFailureFDD, "id-SuccessfulRL-InformationResponse-RL-AdditionFailureFDD" },
  { id_SuccessfulRL_InformationResponse_RL_SetupFailureFDD, "id-SuccessfulRL-InformationResponse-RL-SetupFailureFDD" },
  { id_Unknown_161, "id-Unknown-161" },
  { id_Unknown_162, "id-Unknown-162" },
  { id_TransportBearerID, "id-TransportBearerID" },
  { id_TransportBearerRequestIndicator, "id-TransportBearerRequestIndicator" },
  { id_TransportLayerAddress, "id-TransportLayerAddress" },
  { id_UC_ID, "id-UC-ID" },
  { id_UL_CCTrCH_AddInformation_RL_ReconfPrepTDD, "id-UL-CCTrCH-AddInformation-RL-ReconfPrepTDD" },
  { id_Unknown_168, "id-Unknown-168" },
  { id_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD, "id-UL-CCTrCH-InformationAddList-RL-ReconfPrepTDD" },
  { id_Unknown_170, "id-Unknown-170" },
  { id_UL_CCTrCH_InformationItem_RL_SetupRqstTDD, "id-UL-CCTrCH-InformationItem-RL-SetupRqstTDD" },
  { id_UL_CCTrCH_InformationList_RL_SetupRqstTDD, "id-UL-CCTrCH-InformationList-RL-SetupRqstTDD" },
  { id_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD, "id-UL-CCTrCH-InformationListIE-PhyChReconfRqstTDD" },
  { id_UL_CCTrCH_InformationListIE_RL_AdditionRspTDD, "id-UL-CCTrCH-InformationListIE-RL-AdditionRspTDD" },
  { id_UL_CCTrCH_InformationListIE_RL_ReconfReadyTDD, "id-UL-CCTrCH-InformationListIE-RL-ReconfReadyTDD" },
  { id_UL_CCTrCH_InformationListIE_RL_SetupRspTDD, "id-UL-CCTrCH-InformationListIE-RL-SetupRspTDD" },
  { id_UL_DPCH_Information_RL_ReconfPrepFDD, "id-UL-DPCH-Information-RL-ReconfPrepFDD" },
  { id_UL_DPCH_Information_RL_ReconfRqstFDD, "id-UL-DPCH-Information-RL-ReconfRqstFDD" },
  { id_UL_DPCH_Information_RL_SetupRqstFDD, "id-UL-DPCH-Information-RL-SetupRqstFDD" },
  { id_UL_DPCH_InformationItem_PhyChReconfRqstTDD, "id-UL-DPCH-InformationItem-PhyChReconfRqstTDD" },
  { id_UL_DPCH_InformationItem_RL_AdditionRspTDD, "id-UL-DPCH-InformationItem-RL-AdditionRspTDD" },
  { id_UL_DPCH_InformationItem_RL_SetupRspTDD, "id-UL-DPCH-InformationItem-RL-SetupRspTDD" },
  { id_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD, "id-UL-DPCH-InformationAddListIE-RL-ReconfReadyTDD" },
  { id_UL_SIRTarget, "id-UL-SIRTarget" },
  { id_URA_Information, "id-URA-Information" },
  { id_Unknown_186, "id-Unknown-186" },
  { id_Unknown_187, "id-Unknown-187" },
  { id_UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD, "id-UnsuccessfulRL-InformationResponse-RL-AdditionFailureFDD" },
  { id_UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD, "id-UnsuccessfulRL-InformationResponse-RL-SetupFailureFDD" },
  { id_UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD, "id-UnsuccessfulRL-InformationResponse-RL-SetupFailureTDD" },
  { id_Unknown_191, "id-Unknown-191" },
  { id_Unknown_192, "id-Unknown-192" },
  { id_Active_Pattern_Sequence_Information, "id-Active-Pattern-Sequence-Information" },
  { id_AdjustmentRatio, "id-AdjustmentRatio" },
  { id_Unknown_195, "id-Unknown-195" },
  { id_Unknown_196, "id-Unknown-196" },
  { id_CauseLevel_RL_AdditionFailureFDD, "id-CauseLevel-RL-AdditionFailureFDD" },
  { id_CauseLevel_RL_AdditionFailureTDD, "id-CauseLevel-RL-AdditionFailureTDD" },
  { id_CauseLevel_RL_ReconfFailure, "id-CauseLevel-RL-ReconfFailure" },
  { id_CauseLevel_RL_SetupFailureFDD, "id-CauseLevel-RL-SetupFailureFDD" },
  { id_CauseLevel_RL_SetupFailureTDD, "id-CauseLevel-RL-SetupFailureTDD" },
  { id_PrimaryCCPCH_RSCP_RL_ReconfPrepTDD, "id-PrimaryCCPCH-RSCP-RL-ReconfPrepTDD" },
  { id_DL_TimeSlot_ISCP_Info_RL_ReconfPrepTDD, "id-DL-TimeSlot-ISCP-Info-RL-ReconfPrepTDD" },
  { id_DL_Timeslot_ISCP_LCR_Information_RL_ReconfPrepTDD, "id-DL-Timeslot-ISCP-LCR-Information-RL-ReconfPrepTDD" },
  { id_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD, "id-DL-CCTrCH-InformationDeleteItem-RL-ReconfPrepTDD" },
  { id_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD, "id-DL-CCTrCH-InformationModifyItem-RL-ReconfPrepTDD" },
  { id_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD, "id-DL-CCTrCH-InformationModifyItem-RL-ReconfRqstTDD" },
  { id_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD, "id-DL-CCTrCH-InformationDeleteList-RL-ReconfPrepTDD" },
  { id_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD, "id-DL-CCTrCH-InformationModifyList-RL-ReconfPrepTDD" },
  { id_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, "id-DL-CCTrCH-InformationModifyList-RL-ReconfRqstTDD" },
  { id_ContextInfoItem_Reset, "id-ContextInfoItem-Reset" },
  { id_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD, "id-DL-DPCH-InformationAddListIE-RL-ReconfReadyTDD" },
  { id_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD, "id-DL-DPCH-InformationDeleteListIE-RL-ReconfReadyTDD" },
  { id_DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD, "id-DL-DPCH-InformationModifyListIE-RL-ReconfReadyTDD" },
  { id_DSCHs_to_Add_TDD, "id-DSCHs-to-Add-TDD" },
  { id_Unused_ProtocolIE_ID_216, "id-Unused-ProtocolIE-ID-216" },
  { id_DSCH_DeleteList_RL_ReconfPrepTDD, "id-DSCH-DeleteList-RL-ReconfPrepTDD" },
  { id_Unused_ProtocolIE_ID_218, "id-Unused-ProtocolIE-ID-218" },
  { id_Unused_ProtocolIE_ID_219, "id-Unused-ProtocolIE-ID-219" },
  { id_DSCH_InformationListIE_RL_AdditionRspTDD, "id-DSCH-InformationListIE-RL-AdditionRspTDD" },
  { id_DSCH_InformationListIEs_RL_SetupRspTDD, "id-DSCH-InformationListIEs-RL-SetupRspTDD" },
  { id_DSCH_TDD_Information, "id-DSCH-TDD-Information" },
  { id_Unused_ProtocolIE_ID_223, "id-Unused-ProtocolIE-ID-223" },
  { id_Enhanced_PrimaryCPICH_EcNo, "id-Enhanced-PrimaryCPICH-EcNo" },
  { id_Unused_ProtocolIE_ID_225, "id-Unused-ProtocolIE-ID-225" },
  { id_Unused_ProtocolIE_ID_226, "id-Unused-ProtocolIE-ID-226" },
  { id_DSCH_ModifyList_RL_ReconfPrepTDD, "id-DSCH-ModifyList-RL-ReconfPrepTDD" },
  { id_Unused_ProtocolIE_ID_228, "id-Unused-ProtocolIE-ID-228" },
  { id_Unused_ProtocolIE_ID_229, "id-Unused-ProtocolIE-ID-229" },
  { id_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD, "id-DSCHToBeAddedOrModifiedList-RL-ReconfReadyTDD" },
  { id_Unknown_231, "id-Unknown-231" },
  { id_GA_Cell, "id-GA-Cell" },
  { id_Load_Value, "id-Load-Value" },
  { id_Load_Value_IncrDecrThres, "id-Load-Value-IncrDecrThres" },
  { id_OnModification, "id-OnModification" },
  { id_Received_Total_Wideband_Power_Value, "id-Received-Total-Wideband-Power-Value" },
  { id_Received_Total_Wideband_Power_Value_IncrDecrThres, "id-Received-Total-Wideband-Power-Value-IncrDecrThres" },
  { id_SFNSFNMeasurementThresholdInformation, "id-SFNSFNMeasurementThresholdInformation" },
  { id_Transmitted_Carrier_Power_Value, "id-Transmitted-Carrier-Power-Value" },
  { id_Transmitted_Carrier_Power_Value_IncrDecrThres, "id-Transmitted-Carrier-Power-Value-IncrDecrThres" },
  { id_TUTRANGPSMeasurementThresholdInformation, "id-TUTRANGPSMeasurementThresholdInformation" },
  { id_UL_Timeslot_ISCP_Value, "id-UL-Timeslot-ISCP-Value" },
  { id_UL_Timeslot_ISCP_Value_IncrDecrThres, "id-UL-Timeslot-ISCP-Value-IncrDecrThres" },
  { id_ResetIndicator, "id-ResetIndicator" },
  { id_RNC_ID, "id-RNC-ID" },
  { id_Unused_ProtocolIE_ID_246, "id-Unused-ProtocolIE-ID-246" },
  { id_Unused_ProtocolIE_ID_247, "id-Unused-ProtocolIE-ID-247" },
  { id_Unused_ProtocolIE_ID_248, "id-Unused-ProtocolIE-ID-248" },
  { id_DSCH_RNTI, "id-DSCH-RNTI" },
  { id_CommonTransportChannelResourcesInitialisationNotRequired, "id-CommonTransportChannelResourcesInitialisationNotRequired" },
  { id_neighbouringTDDCellMeasurementInformationLCR, "id-neighbouringTDDCellMeasurementInformationLCR" },
  { id_IPDL_TDD_ParametersLCR, "id-IPDL-TDD-ParametersLCR" },
  { id_Unused_ProtocolIE_ID_253, "id-Unused-ProtocolIE-ID-253" },
  { id_Unknown_254, "id-Unknown-254" },
  { id_Transmission_Gap_Pattern_Sequence_Information, "id-Transmission-Gap-Pattern-Sequence-Information" },
  { id_UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD, "id-UL-CCTrCH-DeleteInformation-RL-ReconfPrepTDD" },
  { id_UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD, "id-UL-CCTrCH-ModifyInformation-RL-ReconfPrepTDD" },
  { id_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD, "id-UL-CCTrCH-InformationModifyItem-RL-ReconfRqstTDD" },
  { id_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD, "id-UL-CCTrCH-InformationDeleteList-RL-ReconfPrepTDD" },
  { id_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD, "id-UL-CCTrCH-InformationModifyList-RL-ReconfPrepTDD" },
  { id_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, "id-UL-CCTrCH-InformationModifyList-RL-ReconfRqstTDD" },
  { id_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD, "id-UL-CCTrCH-InformationDeleteItem-RL-ReconfRqstTDD" },
  { id_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD, "id-UL-CCTrCH-InformationDeleteList-RL-ReconfRqstTDD" },
  { id_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD, "id-UL-DPCH-InformationDeleteListIE-RL-ReconfReadyTDD" },
  { id_UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD, "id-UL-DPCH-InformationModifyListIE-RL-ReconfReadyTDD" },
  { id_UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD, "id-UnsuccessfulRL-InformationResponse-RL-AdditionFailureTDD" },
  { id_USCHs_to_Add, "id-USCHs-to-Add" },
  { id_USCH_DeleteList_RL_ReconfPrepTDD, "id-USCH-DeleteList-RL-ReconfPrepTDD" },
  { id_USCH_InformationListIE_RL_AdditionRspTDD, "id-USCH-InformationListIE-RL-AdditionRspTDD" },
  { id_USCH_InformationListIEs_RL_SetupRspTDD, "id-USCH-InformationListIEs-RL-SetupRspTDD" },
  { id_USCH_Information, "id-USCH-Information" },
  { id_USCH_ModifyList_RL_ReconfPrepTDD, "id-USCH-ModifyList-RL-ReconfPrepTDD" },
  { id_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD, "id-USCHToBeAddedOrModifiedList-RL-ReconfReadyTDD" },
  { id_DL_Physical_Channel_Information_RL_SetupRqstTDD, "id-DL-Physical-Channel-Information-RL-SetupRqstTDD" },
  { id_UL_Physical_Channel_Information_RL_SetupRqstTDD, "id-UL-Physical-Channel-Information-RL-SetupRqstTDD" },
  { id_ClosedLoopMode1_SupportIndicator, "id-ClosedLoopMode1-SupportIndicator" },
  { id_Unused_ProtocolIE_ID_277, "id-Unused-ProtocolIE-ID-277" },
  { id_DL_DPCH_TimingAdjustment, "id-DL-DPCH-TimingAdjustment" },
  { id_STTD_SupportIndicator, "id-STTD-SupportIndicator" },
  { id_CommonMeasurementAccuracy, "id-CommonMeasurementAccuracy" },
  { id_CommonMeasurementObjectType_CM_Rprt, "id-CommonMeasurementObjectType-CM-Rprt" },
  { id_CommonMeasurementObjectType_CM_Rqst, "id-CommonMeasurementObjectType-CM-Rqst" },
  { id_CommonMeasurementObjectType_CM_Rsp, "id-CommonMeasurementObjectType-CM-Rsp" },
  { id_CommonMeasurementType, "id-CommonMeasurementType" },
  { id_SFN, "id-SFN" },
  { id_SFNReportingIndicator, "id-SFNReportingIndicator" },
  { id_InformationExchangeID, "id-InformationExchangeID" },
  { id_InformationExchangeObjectType_InfEx_Rprt, "id-InformationExchangeObjectType-InfEx-Rprt" },
  { id_InformationExchangeObjectType_InfEx_Rqst, "id-InformationExchangeObjectType-InfEx-Rqst" },
  { id_InformationExchangeObjectType_InfEx_Rsp, "id-InformationExchangeObjectType-InfEx-Rsp" },
  { id_InformationReportCharacteristics, "id-InformationReportCharacteristics" },
  { id_InformationType, "id-InformationType" },
  { id_Rx_Timing_Deviation_Value_LCR, "id-Rx-Timing-Deviation-Value-LCR" },
  { id_Unknown_294, "id-Unknown-294" },
  { id_Unused_ProtocolIE_ID_295, "id-Unused-ProtocolIE-ID-295" },
  { id_DL_PowerBalancing_Information, "id-DL-PowerBalancing-Information" },
  { id_DL_PowerBalancing_ActivationIndicator, "id-DL-PowerBalancing-ActivationIndicator" },
  { id_DL_PowerBalancing_UpdatedIndicator, "id-DL-PowerBalancing-UpdatedIndicator" },
  { id_DL_ReferencePowerInformation, "id-DL-ReferencePowerInformation" },
  { id_CellCapabilityContainer_FDD, "id-CellCapabilityContainer-FDD" },
  { id_CellCapabilityContainer_TDD, "id-CellCapabilityContainer-TDD" },
  { id_CellCapabilityContainer_TDD_LCR, "id-CellCapabilityContainer-TDD-LCR" },
  { id_Cell_Capacity_Class_Value, "id-Cell-Capacity-Class-Value" },
  { id_Unknown_305, "id-Unknown-305" },
  { id_NRT_Load_Information_Value, "id-NRT-Load-Information-Value" },
  { id_NRT_Load_Information_Value_IncrDecrThres, "id-NRT-Load-Information-Value-IncrDecrThres" },
  { id_RT_Load_Value, "id-RT-Load-Value" },
  { id_RT_Load_Value_IncrDecrThres, "id-RT-Load-Value-IncrDecrThres" },
  { id_AntennaColocationIndicator, "id-AntennaColocationIndicator" },
  { id_CoverageIndicator, "id-CoverageIndicator" },
  { id_HCS_Prio, "id-HCS-Prio" },
  { id_DelayedActivation, "id-DelayedActivation" },
  { id_DelayedActivationList_RL_ActivationCmdFDD, "id-DelayedActivationList-RL-ActivationCmdFDD" },
  { id_DelayedActivationInformation_RL_ActivationCmdFDD, "id-DelayedActivationInformation-RL-ActivationCmdFDD" },
  { id_DelayedActivationList_RL_ActivationCmdTDD, "id-DelayedActivationList-RL-ActivationCmdTDD" },
  { id_DelayedActivationInformation_RL_ActivationCmdTDD, "id-DelayedActivationInformation-RL-ActivationCmdTDD" },
  { id_RL_Specific_DCH_Info, "id-RL-Specific-DCH-Info" },
  { id_RL_ReconfigurationRequestFDD_RL_InformationList, "id-RL-ReconfigurationRequestFDD-RL-InformationList" },
  { id_RL_ReconfigurationRequestFDD_RL_Information_IEs, "id-RL-ReconfigurationRequestFDD-RL-Information-IEs" },
  { id_Unknown_320, "id-Unknown-320" },
  { id_RL_ReconfigurationRequestTDD_RL_Information, "id-RL-ReconfigurationRequestTDD-RL-Information" },
  { id_Unknown_322, "id-Unknown-322" },
  { id_PDSCH_RL_ID, "id-PDSCH-RL-ID" },
  { id_Unused_ProtocolIE_ID_324, "id-Unused-ProtocolIE-ID-324" },
  { id_TimeSlot_RL_SetupRspTDD, "id-TimeSlot-RL-SetupRspTDD" },
  { id_Unknown_326, "id-Unknown-326" },
  { id_Unknown_327, "id-Unknown-327" },
  { id_Unknown_328, "id-Unknown-328" },
  { id_Unknown_329, "id-Unknown-329" },
  { id_Unknown_330, "id-Unknown-330" },
  { id_Unknown_331, "id-Unknown-331" },
  { id_Unknown_332, "id-Unknown-332" },
  { id_Unknown_333, "id-Unknown-333" },
  { id_Unknown_334, "id-Unknown-334" },
  { id_Unknown_335, "id-Unknown-335" },
  { id_Unknown_336, "id-Unknown-336" },
  { id_Unknown_337, "id-Unknown-337" },
  { id_Unknown_338, "id-Unknown-338" },
  { id_Unknown_339, "id-Unknown-339" },
  { id_Unknown_340, "id-Unknown-340" },
  { id_Unknown_341, "id-Unknown-341" },
  { id_Unknown_342, "id-Unknown-342" },
  { id_Unknown_343, "id-Unknown-343" },
  { id_Unknown_344, "id-Unknown-344" },
  { id_Unknown_345, "id-Unknown-345" },
  { id_Unknown_346, "id-Unknown-346" },
  { id_Unknown_347, "id-Unknown-347" },
  { id_Unknown_348, "id-Unknown-348" },
  { id_Unknown_349, "id-Unknown-349" },
  { id_Unknown_350, "id-Unknown-350" },
  { id_Unknown_351, "id-Unknown-351" },
  { id_Unknown_352, "id-Unknown-352" },
  { id_Unknown_353, "id-Unknown-353" },
  { id_Unknown_354, "id-Unknown-354" },
  { id_Unknown_355, "id-Unknown-355" },
  { id_Unknown_356, "id-Unknown-356" },
  { id_Unknown_357, "id-Unknown-357" },
  { id_Unknown_358, "id-Unknown-358" },
  { id_Unknown_359, "id-Unknown-359" },
  { id_Unknown_360, "id-Unknown-360" },
  { id_Unknown_361, "id-Unknown-361" },
  { id_Unknown_362, "id-Unknown-362" },
  { id_Unknown_363, "id-Unknown-363" },
  { id_Unknown_364, "id-Unknown-364" },
  { id_Unknown_365, "id-Unknown-365" },
  { id_Unknown_366, "id-Unknown-366" },
  { id_Unknown_367, "id-Unknown-367" },
  { id_Unknown_368, "id-Unknown-368" },
  { id_Unknown_369, "id-Unknown-369" },
  { id_Unknown_370, "id-Unknown-370" },
  { id_Unknown_371, "id-Unknown-371" },
  { id_Unknown_372, "id-Unknown-372" },
  { id_Unknown_373, "id-Unknown-373" },
  { id_Unknown_374, "id-Unknown-374" },
  { id_Unknown_375, "id-Unknown-375" },
  { id_Unknown_376, "id-Unknown-376" },
  { id_Unknown_377, "id-Unknown-377" },
  { id_Unknown_378, "id-Unknown-378" },
  { id_Unknown_379, "id-Unknown-379" },
  { id_Unknown_380, "id-Unknown-380" },
  { id_Unknown_381, "id-Unknown-381" },
  { id_Unknown_382, "id-Unknown-382" },
  { id_Unknown_383, "id-Unknown-383" },
  { id_Unknown_384, "id-Unknown-384" },
  { id_Unknown_385, "id-Unknown-385" },
  { id_Unknown_386, "id-Unknown-386" },
  { id_Unknown_387, "id-Unknown-387" },
  { id_Unknown_388, "id-Unknown-388" },
  { id_Unknown_389, "id-Unknown-389" },
  { id_Unknown_390, "id-Unknown-390" },
  { id_Unknown_391, "id-Unknown-391" },
  { id_Unknown_392, "id-Unknown-392" },
  { id_Unknown_393, "id-Unknown-393" },
  { id_Unknown_394, "id-Unknown-394" },
  { id_Unknown_395, "id-Unknown-395" },
  { id_Unknown_396, "id-Unknown-396" },
  { id_Unknown_397, "id-Unknown-397" },
  { id_Unknown_398, "id-Unknown-398" },
  { id_Unknown_399, "id-Unknown-399" },
  { id_Unknown_400, "id-Unknown-400" },
  { id_Unknown_401, "id-Unknown-401" },
  { id_Unknown_402, "id-Unknown-402" },
  { id_Unknown_403, "id-Unknown-403" },
  { id_Unknown_404, "id-Unknown-404" },
  { id_Unknown_405, "id-Unknown-405" },
  { id_Unknown_406, "id-Unknown-406" },
  { id_Unknown_407, "id-Unknown-407" },
  { id_Unknown_408, "id-Unknown-408" },
  { id_Unknown_409, "id-Unknown-409" },
  { id_Unknown_410, "id-Unknown-410" },
  { id_Unknown_411, "id-Unknown-411" },
  { id_Unknown_412, "id-Unknown-412" },
  { id_Unknown_413, "id-Unknown-413" },
  { id_Unknown_414, "id-Unknown-414" },
  { id_Unknown_415, "id-Unknown-415" },
  { id_Unknown_416, "id-Unknown-416" },
  { id_Unknown_417, "id-Unknown-417" },
  { id_Unknown_418, "id-Unknown-418" },
  { id_Unknown_419, "id-Unknown-419" },
  { id_Unknown_420, "id-Unknown-420" },
  { id_Unknown_421, "id-Unknown-421" },
  { id_Unknown_422, "id-Unknown-422" },
  { id_Unknown_423, "id-Unknown-423" },
  { id_Unknown_424, "id-Unknown-424" },
  { id_Unknown_425, "id-Unknown-425" },
  { id_Unknown_426, "id-Unknown-426" },
  { id_Unknown_427, "id-Unknown-427" },
  { id_Unknown_428, "id-Unknown-428" },
  { id_Unknown_429, "id-Unknown-429" },
  { id_Unknown_430, "id-Unknown-430" },
  { id_Unknown_431, "id-Unknown-431" },
  { id_Unknown_432, "id-Unknown-432" },
  { id_Unknown_433, "id-Unknown-433" },
  { id_Unknown_434, "id-Unknown-434" },
  { id_Unknown_435, "id-Unknown-435" },
  { id_Unknown_436, "id-Unknown-436" },
  { id_Unknown_437, "id-Unknown-437" },
  { id_Unknown_438, "id-Unknown-438" },
  { id_Unknown_439, "id-Unknown-439" },
  { id_Unknown_440, "id-Unknown-440" },
  { id_Unknown_441, "id-Unknown-441" },
  { id_Unknown_442, "id-Unknown-442" },
  { id_Unknown_443, "id-Unknown-443" },
  { id_Unknown_444, "id-Unknown-444" },
  { id_Unknown_445, "id-Unknown-445" },
  { id_Unknown_446, "id-Unknown-446" },
  { id_Unknown_447, "id-Unknown-447" },
  { id_Unknown_448, "id-Unknown-448" },
  { id_Unknown_449, "id-Unknown-449" },
  { id_Unknown_450, "id-Unknown-450" },
  { id_PrimCCPCH_RSCP_DL_PC_RqstTDD, "id-PrimCCPCH-RSCP-DL-PC-RqstTDD" },
  { id_HSDSCH_FDD_Information, "id-HSDSCH-FDD-Information" },
  { id_HSDSCH_FDD_Information_Response, "id-HSDSCH-FDD-Information-Response" },
  { id_Unknown_454, "id-Unknown-454" },
  { id_Unknown_455, "id-Unknown-455" },
  { id_HSDSCH_Information_to_Modify, "id-HSDSCH-Information-to-Modify" },
  { id_HSDSCH_RNTI, "id-HSDSCH-RNTI" },
  { id_HSDSCH_TDD_Information, "id-HSDSCH-TDD-Information" },
  { id_HSDSCH_TDD_Information_Response, "id-HSDSCH-TDD-Information-Response" },
  { id_Unknown_460, "id-Unknown-460" },
  { id_Unknown_461, "id-Unknown-461" },
  { id_Unknown_462, "id-Unknown-462" },
  { id_HSPDSCH_RL_ID, "id-HSPDSCH-RL-ID" },
  { id_UL_Synchronisation_Parameters_LCR, "id-UL-Synchronisation-Parameters-LCR" },
  { id_MAChs_ResetIndicator, "id-MAChs-ResetIndicator" },
  { id_HSDSCH_FDD_Update_Information, "id-HSDSCH-FDD-Update-Information" },
  { id_HSDSCH_TDD_Update_Information, "id-HSDSCH-TDD-Update-Information" },
  { id_GERAN_Cell_Capability, "id-GERAN-Cell-Capability" },
  { id_GERAN_Classmark, "id-GERAN-Classmark" },
  { id_DedicatedMeasurementObjectType_DM_Fail_Ind, "id-DedicatedMeasurementObjectType-DM-Fail-Ind" },
  { id_DedicatedMeasurementObjectType_DM_Fail, "id-DedicatedMeasurementObjectType-DM-Fail" },
  { id_PartialReportingIndicator, "id-PartialReportingIndicator" },
  { id_RL_Set_Successful_InformationItem_DM_Fail, "id-RL-Set-Successful-InformationItem-DM-Fail" },
  { id_RL_Set_Unsuccessful_InformationItem_DM_Fail, "id-RL-Set-Unsuccessful-InformationItem-DM-Fail" },
  { id_RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind, "id-RL-Set-Unsuccessful-InformationItem-DM-Fail-Ind" },
  { id_RL_Successful_InformationItem_DM_Fail, "id-RL-Successful-InformationItem-DM-Fail" },
  { id_RL_Unsuccessful_InformationItem_DM_Fail, "id-RL-Unsuccessful-InformationItem-DM-Fail" },
  { id_RL_Unsuccessful_InformationItem_DM_Fail_Ind, "id-RL-Unsuccessful-InformationItem-DM-Fail-Ind" },
  { id_SNA_Information, "id-SNA-Information" },
  { id_DSCH_InitialWindowSize, "id-DSCH-InitialWindowSize" },
  { id_TDD_DL_DPCH_TimeSlotFormatModifyItem_LCR_RL_ReconfReadyTDD, "id-TDD-DL-DPCH-TimeSlotFormatModifyItem-LCR-RL-ReconfReadyTDD" },
  { id_TDD_UL_DPCH_TimeSlotFormatModifyItem_LCR_RL_ReconfReadyTDD, "id-TDD-UL-DPCH-TimeSlotFormatModifyItem-LCR-RL-ReconfReadyTDD" },
  { id_TDD_TPC_UplinkStepSize_LCR_RL_SetupRqstTDD, "id-TDD-TPC-UplinkStepSize-LCR-RL-SetupRqstTDD" },
  { id_UL_CCTrCH_InformationList_RL_AdditionRqstTDD, "id-UL-CCTrCH-InformationList-RL-AdditionRqstTDD" },
  { id_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD, "id-UL-CCTrCH-InformationItem-RL-AdditionRqstTDD" },
  { id_DL_CCTrCH_InformationList_RL_AdditionRqstTDD, "id-DL-CCTrCH-InformationList-RL-AdditionRqstTDD" },
  { id_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD, "id-DL-CCTrCH-InformationItem-RL-AdditionRqstTDD" },
  { id_TDD_TPC_UplinkStepSize_InformationAdd_LCR_RL_ReconfPrepTDD, "id-TDD-TPC-UplinkStepSize-InformationAdd-LCR-RL-ReconfPrepTDD" },
  { id_TDD_TPC_UplinkStepSize_InformationModify_LCR_RL_ReconfPrepTDD, "id-TDD-TPC-UplinkStepSize-InformationModify-LCR-RL-ReconfPrepTDD" },
  { id_TDD_TPC_DownlinkStepSize_InformationAdd_RL_ReconfPrepTDD, "id-TDD-TPC-DownlinkStepSize-InformationAdd-RL-ReconfPrepTDD" },
  { id_TDD_TPC_DownlinkStepSize_InformationModify_RL_ReconfPrepTDD, "id-TDD-TPC-DownlinkStepSize-InformationModify-RL-ReconfPrepTDD" },
  { id_UL_TimingAdvanceCtrl_LCR, "id-UL-TimingAdvanceCtrl-LCR" },
  { id_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD, "id-HSPDSCH-Timeslot-InformationList-PhyChReconfRqstTDD" },
  { id_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD, "id-HSPDSCH-Timeslot-InformationListLCR-PhyChReconfRqstTDD" },
  { id_HS_SICH_Reception_Quality, "id-HS-SICH-Reception-Quality" },
  { id_HS_SICH_Reception_Quality_Measurement_Value, "id-HS-SICH-Reception-Quality-Measurement-Value" },
  { id_HSSICH_Info_DM_Rprt, "id-HSSICH-Info-DM-Rprt" },
  { id_HSSICH_Info_DM_Rqst, "id-HSSICH-Info-DM-Rqst" },
  { id_HSSICH_Info_DM, "id-HSSICH-Info-DM" },
  { id_CCTrCH_Maximum_DL_Power_RL_SetupRspTDD, "id-CCTrCH-Maximum-DL-Power-RL-SetupRspTDD" },
  { id_CCTrCH_Minimum_DL_Power_RL_SetupRspTDD, "id-CCTrCH-Minimum-DL-Power-RL-SetupRspTDD" },
  { id_CCTrCH_Maximum_DL_Power_RL_AdditionRspTDD, "id-CCTrCH-Maximum-DL-Power-RL-AdditionRspTDD" },
  { id_CCTrCH_Minimum_DL_Power_RL_AdditionRspTDD, "id-CCTrCH-Minimum-DL-Power-RL-AdditionRspTDD" },
  { id_CCTrCH_Maximum_DL_Power_RL_ReconfReadyTDD, "id-CCTrCH-Maximum-DL-Power-RL-ReconfReadyTDD" },
  { id_CCTrCH_Minimum_DL_Power_RL_ReconfReadyTDD, "id-CCTrCH-Minimum-DL-Power-RL-ReconfReadyTDD" },
  { id_Maximum_DL_Power_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD, "id-Maximum-DL-Power-TimeslotLCR-InformationModifyItem-RL-ReconfReadyTDD" },
  { id_Minimum_DL_Power_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD, "id-Minimum-DL-Power-TimeslotLCR-InformationModifyItem-RL-ReconfReadyTDD" },
  { id_DL_CCTrCH_InformationList_RL_ReconfRspTDD, "id-DL-CCTrCH-InformationList-RL-ReconfRspTDD" },
  { id_DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD, "id-DL-DPCH-InformationModifyItem-LCR-RL-ReconfRspTDD" },
  { id_Maximum_DL_Power_TimeslotLCR_InformationItem, "id-Maximum-DL-Power-TimeslotLCR-InformationItem" },
  { id_Minimum_DL_Power_TimeslotLCR_InformationItem, "id-Minimum-DL-Power-TimeslotLCR-InformationItem" },
  { id_TDD_Support_8PSK, "id-TDD-Support-8PSK" },
  { id_TDD_maxNrDLPhysicalchannels, "id-TDD-maxNrDLPhysicalchannels" },
  { id_ExtendedGSMCellIndividualOffset, "id-ExtendedGSMCellIndividualOffset" },
  { id_ContextGroupInfoItem_Reset, "id-ContextGroupInfoItem-Reset" },
  { id_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd, "id-HSDSCHMacdFlowSpecificInformationList-RL-PreemptRequiredInd" },
  { id_HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd, "id-HSDSCHMacdFlowSpecificInformationItem-RL-PreemptRequiredInd" },
  { id_RL_ParameterUpdateIndicationFDD_RL_InformationList, "id-RL-ParameterUpdateIndicationFDD-RL-InformationList" },
  { id_Primary_CPICH_Usage_For_Channel_Estimation, "id-Primary-CPICH-Usage-For-Channel-Estimation" },
  { id_Secondary_CPICH_Information, "id-Secondary-CPICH-Information" },
  { id_Secondary_CPICH_Information_Change, "id-Secondary-CPICH-Information-Change" },
  { id_Unused_ProtocolIE_ID_522, "id-Unused-ProtocolIE-ID-522" },
  { id_Unused_ProtocolIE_ID_523, "id-Unused-ProtocolIE-ID-523" },
  { id_RL_ParameterUpdateIndicationFDD_RL_Information_Item, "id-RL-ParameterUpdateIndicationFDD-RL-Information-Item" },
  { id_Phase_Reference_Update_Indicator, "id-Phase-Reference-Update-Indicator" },
  { id_Unidirectional_DCH_Indicator, "id-Unidirectional-DCH-Indicator" },
  { id_RL_Information_RL_ReconfPrepTDD, "id-RL-Information-RL-ReconfPrepTDD" },
  { id_Multiple_RL_InformationResponse_RL_ReconfReadyTDD, "id-Multiple-RL-InformationResponse-RL-ReconfReadyTDD" },
  { id_RL_ReconfigurationResponseTDD_RL_Information, "id-RL-ReconfigurationResponseTDD-RL-Information" },
  { id_Satellite_Almanac_Information_ExtItem, "id-Satellite-Almanac-Information-ExtItem" },
  { id_HSDSCH_MACdFlows_to_Add, "id-HSDSCH-MACdFlows-to-Add" },
  { id_HSDSCH_MACdFlows_to_Delete, "id-HSDSCH-MACdFlows-to-Delete" },
  { id_HSDSCH_Information_to_Modify_Unsynchronised, "id-HSDSCH-Information-to-Modify-Unsynchronised" },
  { id_TnlQos, "id-TnlQos" },
  { id_RTLoadValue, "id-RTLoadValue" },
  { id_NRTLoadInformationValue, "id-NRTLoadInformationValue" },
  { id_CellPortionID, "id-CellPortionID" },
  { id_UpPTSInterferenceValue, "id-UpPTSInterferenceValue" },
  { id_PrimaryCCPCH_RSCP_Delta, "id-PrimaryCCPCH-RSCP-Delta" },
  { id_UEMeasurementType, "id-UEMeasurementType" },
  { id_UEMeasurementTimeslotInfoHCR, "id-UEMeasurementTimeslotInfoHCR" },
  { id_UEMeasurementTimeslotInfoLCR, "id-UEMeasurementTimeslotInfoLCR" },
  { id_UEMeasurementReportCharacteristics, "id-UEMeasurementReportCharacteristics" },
  { id_UEMeasurementParameterModAllow, "id-UEMeasurementParameterModAllow" },
  { id_UEMeasurementValueInformation, "id-UEMeasurementValueInformation" },
  { id_InterfacesToTraceItem, "id-InterfacesToTraceItem" },
  { id_ListOfInterfacesToTrace, "id-ListOfInterfacesToTrace" },
  { id_TraceDepth, "id-TraceDepth" },
  { id_TraceRecordingSessionReference, "id-TraceRecordingSessionReference" },
  { id_TraceReference, "id-TraceReference" },
  { id_UEIdentity, "id-UEIdentity" },
  { id_NACC_Related_Data, "id-NACC-Related-Data" },
  { id_GSM_Cell_InfEx_Rqst, "id-GSM-Cell-InfEx-Rqst" },
  { id_MeasurementRecoveryBehavior, "id-MeasurementRecoveryBehavior" },
  { id_MeasurementRecoveryReportingIndicator, "id-MeasurementRecoveryReportingIndicator" },
  { id_MeasurementRecoverySupportIndicator, "id-MeasurementRecoverySupportIndicator" },
  { id_DL_DPCH_Power_Information_RL_ReconfPrepFDD, "id-DL-DPCH-Power-Information-RL-ReconfPrepFDD" },
  { id_F_DPCH_Information_RL_ReconfPrepFDD, "id-F-DPCH-Information-RL-ReconfPrepFDD" },
  { id_F_DPCH_Information_RL_SetupRqstFDD, "id-F-DPCH-Information-RL-SetupRqstFDD" },
  { id_MBMS_Bearer_Service_List, "id-MBMS-Bearer-Service-List" },
  { id_MBMS_Bearer_Service_List_InfEx_Rsp, "id-MBMS-Bearer-Service-List-InfEx-Rsp" },
  { id_Active_MBMS_Bearer_ServiceFDD, "id-Active-MBMS-Bearer-ServiceFDD" },
  { id_Active_MBMS_Bearer_ServiceTDD, "id-Active-MBMS-Bearer-ServiceTDD" },
  { id_Old_URA_ID, "id-Old-URA-ID" },
  { id_Unknown_565, "id-Unknown-565" },
  { id_Unknown_566, "id-Unknown-566" },
  { id_Unknown_567, "id-Unknown-567" },
  { id_UE_State, "id-UE-State" },
  { id_URA_ID, "id-URA-ID" },
  { id_Unknown_570, "id-Unknown-570" },
  { id_HARQ_Preamble_Mode, "id-HARQ-Preamble-Mode" },
  { id_SynchronisationIndicator, "id-SynchronisationIndicator" },
  { id_UL_DPDCHIndicatorEDCH, "id-UL-DPDCHIndicatorEDCH" },
  { id_EDPCH_Information, "id-EDPCH-Information" },
  { id_RL_Specific_EDCH_Information, "id-RL-Specific-EDCH-Information" },
  { id_EDCH_RL_Indication, "id-EDCH-RL-Indication" },
  { id_EDCH_FDD_Information, "id-EDCH-FDD-Information" },
  { id_EDCH_RLSet_Id, "id-EDCH-RLSet-Id" },
  { id_Serving_EDCHRL_Id, "id-Serving-EDCHRL-Id" },
  { id_EDCH_FDD_DL_ControlChannelInformation, "id-EDCH-FDD-DL-ControlChannelInformation" },
  { id_EDCH_FDD_InformationResponse, "id-EDCH-FDD-InformationResponse" },
  { id_EDCH_MACdFlows_To_Add, "id-EDCH-MACdFlows-To-Add" },
  { id_EDCH_FDD_Information_To_Modify, "id-EDCH-FDD-Information-To-Modify" },
  { id_EDCH_MACdFlows_To_Delete, "id-EDCH-MACdFlows-To-Delete" },
  { id_EDPCH_Information_RLReconfRequest_FDD, "id-EDPCH-Information-RLReconfRequest-FDD" },
  { id_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd, "id-EDCH-MacdFlowSpecificInformationList-RL-PreemptRequiredInd" },
  { id_EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd, "id-EDCH-MacdFlowSpecificInformationItem-RL-PreemptRequiredInd" },
  { id_EDCH_MacdFlowSpecificInformationList_RL_CongestInd, "id-EDCH-MacdFlowSpecificInformationList-RL-CongestInd" },
  { id_EDCH_MacdFlowSpecificInformationItem_RL_CongestInd, "id-EDCH-MacdFlowSpecificInformationItem-RL-CongestInd" },
  { id_MBMS_Bearer_Service_Full_Address, "id-MBMS-Bearer-Service-Full-Address" },
  { id_Initial_DL_DPCH_TimingAdjustment, "id-Initial-DL-DPCH-TimingAdjustment" },
  { id_Initial_DL_DPCH_TimingAdjustment_Allowed, "id-Initial-DL-DPCH-TimingAdjustment-Allowed" },
  { id_User_Plane_Congestion_Fields_Inclusion, "id-User-Plane-Congestion-Fields-Inclusion" },
  { id_HARQ_Preamble_Mode_Activation_Indicator, "id-HARQ-Preamble-Mode-Activation-Indicator" },
  { id_multiple_DedicatedMeasurementValueList_TDD_DM_Rsp, "id-multiple-DedicatedMeasurementValueList-TDD-DM-Rsp" },
  { id_multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp, "id-multiple-DedicatedMeasurementValueList-LCR-TDD-DM-Rsp" },
  { id_ProvidedInformation, "id-ProvidedInformation" },
  { id_Active_MBMS_Bearer_ServiceFDD_PFL, "id-Active-MBMS-Bearer-ServiceFDD-PFL" },
  { id_Active_MBMS_Bearer_ServiceTDD_PFL, "id-Active-MBMS-Bearer-ServiceTDD-PFL" },
  { id_FrequencyBandIndicator, "id-FrequencyBandIndicator" },
  { id_Serving_cell_change_CFN, "id-Serving-cell-change-CFN" },
  { id_HS_DSCH_serving_cell_change_information, "id-HS-DSCH-serving-cell-change-information" },
  { id_HS_DSCH_serving_cell_change_informationResponse, "id-HS-DSCH-serving-cell-change-informationResponse" },
  { id_E_DCH_Serving_cell_change_informationResponse, "id-E-DCH-Serving-cell-change-informationResponse" },
  { id_secondary_LCR_CCPCH_Info_TDD, "id-secondary-LCR-CCPCH-Info-TDD" },
  { id_E_DCH_FDD_Update_Information, "id-E-DCH-FDD-Update-Information" },
  { id_Inter_Frequency_Cell_List, "id-Inter-Frequency-Cell-List" },
  { id_Inter_Frequency_Cell_Information, "id-Inter-Frequency-Cell-Information" },
  { id_multiple_HSSICHMeasurementValueList_TDD_DM_Rsp, "id-multiple-HSSICHMeasurementValueList-TDD-DM-Rsp" },
  { id_TDD_Support_PLCCH, "id-TDD-Support-PLCCH" },
  { id_PLCCH_Information_UL_TimeslotLCR_Info, "id-PLCCH-Information-UL-TimeslotLCR-Info" },
  { id_PLCCH_Information_PhyChReconfRqstTDD, "id-PLCCH-Information-PhyChReconfRqstTDD" },
  { id_TDD768_maxNrDLPhysicalchannelsTS, "id-TDD768-maxNrDLPhysicalchannelsTS" },
  { id_RL_InformationResponse_RL_AdditionRspTDD768, "id-RL-InformationResponse-RL-AdditionRspTDD768" },
  { id_UL_CCTrCH_InformationListIE_RL_AdditionRspTDD768, "id-UL-CCTrCH-InformationListIE-RL-AdditionRspTDD768" },
  { id_DL_CCTrCH_InformationListIE_RL_AdditionRspTDD768, "id-DL-CCTrCH-InformationListIE-RL-AdditionRspTDD768" },
  { id_UL_DPCH_InformationItem_RL_AdditionRspTDD768, "id-UL-DPCH-InformationItem-RL-AdditionRspTDD768" },
  { id_DL_DPCH_InformationItem_RL_AdditionRspTDD768, "id-DL-DPCH-InformationItem-RL-AdditionRspTDD768" },
  { id_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD768, "id-UL-DPCH-InformationAddListIE-RL-ReconfReadyTDD768" },
  { id_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768, "id-UL-Timeslot-InformationModifyList-RL-ReconfReadyTDD768" },
  { id_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD768, "id-DL-DPCH-InformationAddListIE-RL-ReconfReadyTDD768" },
  { id_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768, "id-DL-Timeslot-InformationModifyList-RL-ReconfReadyTDD768" },
  { id_secondary_CCPCH_Info_RL_ReconfReadyTDD768, "id-secondary-CCPCH-Info-RL-ReconfReadyTDD768" },
  { id_hSSCCH_TDD_Specific_InfoList_Response768, "id-hSSCCH-TDD-Specific-InfoList-Response768" },
  { id_hSPDSCH_TDD_Specific_InfoList_Response768, "id-hSPDSCH-TDD-Specific-InfoList-Response768" },
  { id_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768, "id-HSPDSCH-Timeslot-InformationList-PhyChReconfRqstTDD768" },
  { id_UL_Timeslot_InformationList_PhyChReconfRqstTDD768, "id-UL-Timeslot-InformationList-PhyChReconfRqstTDD768" },
  { id_DL_Timeslot_InformationList_PhyChReconfRqstTDD768, "id-DL-Timeslot-InformationList-PhyChReconfRqstTDD768" },
  { id_CellCapabilityContainer_TDD768, "id-CellCapabilityContainer-TDD768" },
  { id_multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp, "id-multiple-DedicatedMeasurementValueList-TDD768-DM-Rsp" },
  { id_neighbouringTDDCellMeasurementInformation768, "id-neighbouringTDDCellMeasurementInformation768" },
  { id_UEMeasurementTimeslotInfo768, "id-UEMeasurementTimeslotInfo768" },
  { id_Rx_Timing_Deviation_Value_768, "id-Rx-Timing-Deviation-Value-768" },
  { id_UEMeasurementValueTransmittedPowerList768, "id-UEMeasurementValueTransmittedPowerList768" },
  { id_UEMeasurementValueTimeslotISCPList768, "id-UEMeasurementValueTimeslotISCPList768" },
  { id_RL_InformationResponse_RL_SetupRspTDD768, "id-RL-InformationResponse-RL-SetupRspTDD768" },
  { id_UL_CCTrCH_InformationListIE_RL_SetupRspTDD768, "id-UL-CCTrCH-InformationListIE-RL-SetupRspTDD768" },
  { id_DL_CCTrCH_InformationListIE_RL_SetupRspTDD768, "id-DL-CCTrCH-InformationListIE-RL-SetupRspTDD768" },
  { id_UL_DPCH_InformationItem_RL_SetupRspTDD768, "id-UL-DPCH-InformationItem-RL-SetupRspTDD768" },
  { id_DL_DPCH_InformationItem_RL_SetupRspTDD768, "id-DL-DPCH-InformationItem-RL-SetupRspTDD768" },
  { id_TDD768_minimumSpreadingFactor_UL, "id-TDD768-minimumSpreadingFactor-UL" },
  { id_TDD768_minimumSpreadingFactor_DL, "id-TDD768-minimumSpreadingFactor-DL" },
  { id_TDD768_maxNrDLPhysicalchannels, "id-TDD768-maxNrDLPhysicalchannels" },
  { id_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD, "id-DL-DPCH-InformationDeleteList768-RL-ReconfReadyTDD" },
  { id_DPCH_ID768_DM_Rsp, "id-DPCH-ID768-DM-Rsp" },
  { id_DPCH_ID768_DM_Rqst, "id-DPCH-ID768-DM-Rqst" },
  { id_DPCH_ID768_DM_Rprt, "id-DPCH-ID768-DM-Rprt" },
  { id_EDPCH_Information_RLAdditionReq_FDD, "id-EDPCH-Information-RLAdditionReq-FDD" },
  { id_HSDSCH_Configured_Indicator, "id-HSDSCH-Configured-Indicator" },
  { id_RxTimingDeviationForTAext, "id-RxTimingDeviationForTAext" },
  { id_RxTimingDeviationForTA768, "id-RxTimingDeviationForTA768" },
  { id_Rx_Timing_Deviation_Value_ext, "id-Rx-Timing-Deviation-Value-ext" },
  { id_E_DCH_PowerOffset_for_SchedulingInfo, "id-E-DCH-PowerOffset-for-SchedulingInfo" },
  { id_TrCH_SrcStatisticsDescr, "id-TrCH-SrcStatisticsDescr" },
  { id_E_DCH_Information, "id-E-DCH-Information" },
  { id_E_DCH_Serving_RL_ID, "id-E-DCH-Serving-RL-ID" },
  { id_E_DCH_Information_Reconfig, "id-E-DCH-Information-Reconfig" },
  { id_E_DCH_Information_Response, "id-E-DCH-Information-Response" },
  { id_E_DCH_768_Information, "id-E-DCH-768-Information" },
  { id_E_DCH_768_Information_Reconfig, "id-E-DCH-768-Information-Reconfig" },
  { id_E_DCH_768_Information_Response, "id-E-DCH-768-Information-Response" },
  { id_ExtendedPropagationDelay, "id-ExtendedPropagationDelay" },
  { id_Extended_Round_Trip_Time_Value, "id-Extended-Round-Trip-Time-Value" },
  { id_AlternativeFormatReportingIndicator, "id-AlternativeFormatReportingIndicator" },
  { id_DCH_Indicator_For_E_DCH_HSDPA_Operation, "id-DCH-Indicator-For-E-DCH-HSDPA-Operation" },
  { id_E_RGCH_E_HICH_ChannelisationCodeValidityIndicator, "id-E-RGCH-E-HICH-ChannelisationCodeValidityIndicator" },
  { id_E_DCH_Minimum_Set_E_TFCIValidityIndicator, "id-E-DCH-Minimum-Set-E-TFCIValidityIndicator" },
  { id_Fast_Reconfiguration_Mode, "id-Fast-Reconfiguration-Mode" },
  { id_Fast_Reconfiguration_Permission, "id-Fast-Reconfiguration-Permission" },
  { id_Continuous_Packet_Connectivity_DTX_DRX_Information, "id-Continuous-Packet-Connectivity-DTX-DRX-Information" },
  { id_Continuous_Packet_Connectivity_HS_SCCH_Less_Information, "id-Continuous-Packet-Connectivity-HS-SCCH-Less-Information" },
  { id_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response, "id-Continuous-Packet-Connectivity-HS-SCCH-Less-Information-Response" },
  { id_CPC_Information, "id-CPC-Information" },
  { id_Unknown_674, "id-Unknown-674" },
  { id_MIMO_InformationResponse, "id-MIMO-InformationResponse" },
  { id_Unknown_676, "id-Unknown-676" },
  { id_E_DCH_LCR_Information, "id-E-DCH-LCR-Information" },
  { id_E_DCH_LCR_Information_Reconfig, "id-E-DCH-LCR-Information-Reconfig" },
  { id_E_DCH_LCR_Information_Response, "id-E-DCH-LCR-Information-Response" },
  { id_HS_PDSCH_Code_Change_Grant, "id-HS-PDSCH-Code-Change-Grant" },
  { id_HS_PDSCH_Code_Change_Indicator, "id-HS-PDSCH-Code-Change-Indicator" },
  { id_Extended_SRNC_ID, "id-Extended-SRNC-ID" },
  { id_Extended_RNC_ID, "id-Extended-RNC-ID" },
  { id_SixtyfourQAM_DL_SupportIndicator, "id-SixtyfourQAM-DL-SupportIndicator" },
  { id_Enhanced_FACH_Support_Indicator, "id-Enhanced-FACH-Support-Indicator" },
  { id_Enhanced_FACH_Information_ResponseFDD, "id-Enhanced-FACH-Information-ResponseFDD" },
  { id_Unknown_687, "id-Unknown-687" },
  { id_Unknown_688, "id-Unknown-688" },
  { id_Unknown_689, "id-Unknown-689" },
  { id_HSDSCH_MACdPDUSizeFormat, "id-HSDSCH-MACdPDUSizeFormat" },
  { id_MaximumMACdPDU_SizeExtended, "id-MaximumMACdPDU-SizeExtended" },
  { id_F_DPCH_SlotFormat, "id-F-DPCH-SlotFormat" },
  { id_F_DPCH_SlotFormatSupportRequest, "id-F-DPCH-SlotFormatSupportRequest" },
  { id_eDCH_MACdFlow_Retransmission_Timer_LCR, "id-eDCH-MACdFlow-Retransmission-Timer-LCR" },
  { id_Max_UE_DTX_Cycle, "id-Max-UE-DTX-Cycle" },
  { id_Unknown_696, "id-Unknown-696" },
  { id_Unknown_697, "id-Unknown-697" },
  { id_Unknown_698, "id-Unknown-698" },
  { id_GANSS_Common_Data, "id-GANSS-Common-Data" },
  { id_GANSS_Information, "id-GANSS-Information" },
  { id_GANSS_Generic_Data, "id-GANSS-Generic-Data" },
  { id_TUTRANGANSSMeasurementThresholdInformation, "id-TUTRANGANSSMeasurementThresholdInformation" },
  { id_TUTRANGANSSMeasurementValueInformation, "id-TUTRANGANSSMeasurementValueInformation" },
  { id_Unknown_704, "id-Unknown-704" },
  { id_Ext_Reference_E_TFCI_PO, "id-Ext-Reference-E-TFCI-PO" },
  { id_Ext_Max_Bits_MACe_PDU_non_scheduled, "id-Ext-Max-Bits-MACe-PDU-non-scheduled" },
  { id_HARQ_MemoryPartitioningInfoExtForMIMO, "id-HARQ-MemoryPartitioningInfoExtForMIMO" },
  { id_MIMO_ActivationIndicator, "id-MIMO-ActivationIndicator" },
  { id_MIMO_Mode_Indicator, "id-MIMO-Mode-Indicator" },
  { id_MIMO_N_M_Ratio, "id-MIMO-N-M-Ratio" },
  { id_TransportBearerNotSetupIndicator, "id-TransportBearerNotSetupIndicator" },
  { id_TransportBearerNotRequestedIndicator, "id-TransportBearerNotRequestedIndicator" },
  { id_PowerControlGAP, "id-PowerControlGAP" },
  { id_UARFCNforNt, "id-UARFCNforNt" },
  { id_LCRTDD_uplink_Physical_Channel_Capability, "id-LCRTDD-uplink-Physical-Channel-Capability" },
  { id_number_Of_Supported_Carriers, "id-number-Of-Supported-Carriers" },
  { id_HSSICH_SIRTarget, "id-HSSICH-SIRTarget" },
  { id_HSSICH_TPC_StepSize, "id-HSSICH-TPC-StepSize" },
  { id_tSN_Length, "id-tSN-Length" },
  { id_HS_SICH_ID_Extension, "id-HS-SICH-ID-Extension" },
  { id_HSSICH_Info_DM_Rqst_Extension, "id-HSSICH-Info-DM-Rqst-Extension" },
  { id_multipleFreq_HSPDSCH_InformationList_ResponseTDDLCR, "id-multipleFreq-HSPDSCH-InformationList-ResponseTDDLCR" },
  { id_multicarrier_number, "id-multicarrier-number" },
  { id_UPPCHPositionLCR, "id-UPPCHPositionLCR" },
  { id_UpPCH_InformationList_LCRTDD, "id-UpPCH-InformationList-LCRTDD" },
  { id_UpPCH_InformationItem_LCRTDD, "id-UpPCH-InformationItem-LCRTDD" },
  { id_Multiple_PLMN_List, "id-Multiple-PLMN-List" },
  { id_UE_Capabilities_Info, "id-UE-Capabilities-Info" },
  { id_FrameOffset, "id-FrameOffset" },
  { id_ChipOffset, "id-ChipOffset" },
  { id_Enhanced_PCH_Capability, "id-Enhanced-PCH-Capability" },
  { id_SixteenQAM_UL_Operation_Indicator, "id-SixteenQAM-UL-Operation-Indicator" },
  { id_E_TFCI_Boost_Information, "id-E-TFCI-Boost-Information" },
  { id_SixtyfourQAM_UsageAllowedIndicator, "id-SixtyfourQAM-UsageAllowedIndicator" },
  { id_SixtyfourQAM_DL_UsageIndicator, "id-SixtyfourQAM-DL-UsageIndicator" },
  { id_Default_Serving_Grant_in_DTX_Cycle2, "id-Default-Serving-Grant-in-DTX-Cycle2" },
  { id_E_DPDCH_PowerInterpolation, "id-E-DPDCH-PowerInterpolation" },
  { id_Extended_E_DCH_LCRTDD_PhysicalLayerCategory, "id-Extended-E-DCH-LCRTDD-PhysicalLayerCategory" },
  { id_E_DCH_MACdPDUSizeFormat, "id-E-DCH-MACdPDUSizeFormat" },
  { id_Continuous_Packet_Connectivity_HS_SCCH_Less_Deactivate_Indicator, "id-Continuous-Packet-Connectivity-HS-SCCH-Less-Deactivate-Indicator" },
  { id_E_DCH_DL_Control_Channel_Change_Information, "id-E-DCH-DL-Control-Channel-Change-Information" },
  { id_E_DCH_DL_Control_Channel_Grant_Information, "id-E-DCH-DL-Control-Channel-Grant-Information" },
  { id_MaximumNumber_Of_Retransmission_For_SchedulingInfo_LCRTDD, "id-MaximumNumber-Of-Retransmission-For-SchedulingInfo-LCRTDD" },
  { id_E_DCH_RetransmissionTimer_For_SchedulingInfo_LCRTDD, "id-E-DCH-RetransmissionTimer-For-SchedulingInfo-LCRTDD" },
  { id_E_PUCH_PowerControlGAP, "id-E-PUCH-PowerControlGAP" },
  { id_HSDSCH_TBSizeTableIndicator, "id-HSDSCH-TBSizeTableIndicator" },
  { id_UE_with_enhanced_HS_SCCH_support_indicator, "id-UE-with-enhanced-HS-SCCH-support-indicator" },
  { id_DGANSS_Corrections_Req, "id-DGANSS-Corrections-Req" },
  { id_E_AGCH_Table_Choice, "id-E-AGCH-Table-Choice" },
  { id_RANAP_EnhancedRelocationInformationRequest, "id-RANAP-EnhancedRelocationInformationRequest" },
  { id_RANAP_EnhancedRelocationInformationResponse, "id-RANAP-EnhancedRelocationInformationResponse" },
  { id_Common_EDCH_MAC_d_Flow_Specific_InformationFDD, "id-Common-EDCH-MAC-d-Flow-Specific-InformationFDD" },
  { id_Common_EDCH_Support_Indicator, "id-Common-EDCH-Support-Indicator" },
  { id_E_RNTI, "id-E-RNTI" },
  { id_Released_CN_Domain, "id-Released-CN-Domain" },
  { id_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst, "id-MBMS-Bearer-Service-in-MBMS-Cell-InfEx-Rqst" },
  { id_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp, "id-MBMS-Bearer-Service-in-MBMS-Cell-InfEx-Rsp" },
  { id_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt, "id-MBMS-Bearer-Service-in-MBMS-Cell-InfEx-Rprt" },
  { id_MBMS_Cell_InfEx_Rqst, "id-MBMS-Cell-InfEx-Rqst" },
  { id_MBMS_Cell_InfEx_Rsp, "id-MBMS-Cell-InfEx-Rsp" },
  { id_MBMS_Cell_InfEx_Rprt, "id-MBMS-Cell-InfEx-Rprt" },
  { id_Counting_Information, "id-Counting-Information" },
  { id_Transmission_Mode_Information, "id-Transmission-Mode-Information" },
  { id_MBMS_Neighbouring_Cell_Information, "id-MBMS-Neighbouring-Cell-Information" },
  { id_MBMS_RLC_Sequence_Number_Information, "id-MBMS-RLC-Sequence-Number-Information" },
  { id_RLC_Sequence_Number, "id-RLC-Sequence-Number" },
  { id_Neighbouring_E_UTRA_CellInformation, "id-Neighbouring-E-UTRA-CellInformation" },
  { id_Unknown_768, "id-Unknown-768" },
  { id_MBSFN_Cluster_Identity, "id-MBSFN-Cluster-Identity" },
  { id_MCCH_Configuration_Info, "id-MCCH-Configuration-Info" },
  { id_MCCH_Message_List, "id-MCCH-Message-List" },
  { id_MBSFN_Scheduling_Transmission_Time_Interval_Info_List, "id-MBSFN-Scheduling-Transmission-Time-Interval-Info-List" },
  { id_GANSS_Time_ID, "id-GANSS-Time-ID" },
  { id_GANSS_AddIonoModelReq, "id-GANSS-AddIonoModelReq" },
  { id_GANSS_EarthOrientParaReq, "id-GANSS-EarthOrientParaReq" },
  { id_GANSS_AddNavigationModelsReq, "id-GANSS-AddNavigationModelsReq" },
  { id_GANSS_AddUTCModelsReq, "id-GANSS-AddUTCModelsReq" },
  { id_GANSS_AuxInfoReq, "id-GANSS-AuxInfoReq" },
  { id_GANSS_SBAS_ID, "id-GANSS-SBAS-ID" },
  { id_GANSS_ID, "id-GANSS-ID" },
  { id_GANSS_Additional_Ionospheric_Model, "id-GANSS-Additional-Ionospheric-Model" },
  { id_GANSS_Earth_Orientation_Parameters, "id-GANSS-Earth-Orientation-Parameters" },
  { id_GANSS_Additional_Time_Models, "id-GANSS-Additional-Time-Models" },
  { id_GANSS_Additional_Navigation_Models, "id-GANSS-Additional-Navigation-Models" },
  { id_GANSS_Additional_UTC_Models, "id-GANSS-Additional-UTC-Models" },
  { id_GANSS_Auxiliary_Information, "id-GANSS-Auxiliary-Information" },
  { id_MinimumReducedE_DPDCH_GainFactor, "id-MinimumReducedE-DPDCH-GainFactor" },
  { id_Enhanced_FACH_Information_ResponseLCR, "id-Enhanced-FACH-Information-ResponseLCR" },
  { id_Common_EDCH_MAC_d_Flow_Specific_InformationLCR, "id-Common-EDCH-MAC-d-Flow-Specific-InformationLCR" },
  { id_HSDSCH_PreconfigurationSetup, "id-HSDSCH-PreconfigurationSetup" },
  { id_HSDSCH_PreconfigurationInfo, "id-HSDSCH-PreconfigurationInfo" },
  { id_NoOfTargetCellHS_SCCH_Order, "id-NoOfTargetCellHS-SCCH-Order" },
  { id_EnhancedHSServingCC_Abort, "id-EnhancedHSServingCC-Abort" },
  { id_Additional_HS_Cell_Information_RL_Setup, "id-Additional-HS-Cell-Information-RL-Setup" },
  { id_Additional_HS_Cell_Information_Response, "id-Additional-HS-Cell-Information-Response" },
  { id_Additional_HS_Cell_Information_RL_Addition, "id-Additional-HS-Cell-Information-RL-Addition" },
  { id_Additional_HS_Cell_Change_Information_Response, "id-Additional-HS-Cell-Change-Information-Response" },
  { id_Additional_HS_Cell_Information_RL_Reconf_Prep, "id-Additional-HS-Cell-Information-RL-Reconf-Prep" },
  { id_Additional_HS_Cell_Information_RL_Reconf_Req, "id-Additional-HS-Cell-Information-RL-Reconf-Req" },
  { id_Additional_HS_Cell_RL_Reconf_Response, "id-Additional-HS-Cell-RL-Reconf-Response" },
  { id_Additional_HS_Cell_Information_RL_Param_Upd, "id-Additional-HS-Cell-Information-RL-Param-Upd" },
  { id_Secondary_Serving_Cell_List, "id-Secondary-Serving-Cell-List" },
  { id_MultiCarrier_HSDSCH_Physical_Layer_Category, "id-MultiCarrier-HSDSCH-Physical-Layer-Category" },
  { id_IdleIntervalInformation, "id-IdleIntervalInformation" },
  { id_NeedforIdleInterval, "id-NeedforIdleInterval" },
  { id_IdleIntervalConfigurationIndicator, "id-IdleIntervalConfigurationIndicator" },
  { id_ContinuousPacketConnectivity_DRX_InformationLCR, "id-ContinuousPacketConnectivity-DRX-InformationLCR" },
  { id_ContinuousPacketConnectivity_DRX_Information_ResponseLCR, "id-ContinuousPacketConnectivity-DRX-Information-ResponseLCR" },
  { id_E_AGCH_UE_Inactivity_Monitor_Threshold, "id-E-AGCH-UE-Inactivity-Monitor-Threshold" },
  { id_CPC_InformationLCR, "id-CPC-InformationLCR" },
  { id_E_DCH_Semi_PersistentScheduling_Information_LCR, "id-E-DCH-Semi-PersistentScheduling-Information-LCR" },
  { id_HS_DSCH_Semi_PersistentScheduling_Information_LCR, "id-HS-DSCH-Semi-PersistentScheduling-Information-LCR" },
  { id_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR, "id-HS-DSCH-Semi-PersistentScheduling-Information-ResponseLCR" },
  { id_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR, "id-E-DCH-Semi-PersistentScheduling-Information-ResponseLCR" },
  { id_MIMO_SFMode_For_HSPDSCHDualStream, "id-MIMO-SFMode-For-HSPDSCHDualStream" },
  { id_MIMO_SFMode_Supported_For_HSPDSCHDualStream, "id-MIMO-SFMode-Supported-For-HSPDSCHDualStream" },
  { id_MIMO_ReferenceSignal_InformationListLCR, "id-MIMO-ReferenceSignal-InformationListLCR" },
  { id_GANSS_alm_keplerianNAVAlmanac, "id-GANSS-alm-keplerianNAVAlmanac" },
  { id_GANSS_alm_keplerianReducedAlmanac, "id-GANSS-alm-keplerianReducedAlmanac" },
  { id_GANSS_alm_keplerianMidiAlmanac, "id-GANSS-alm-keplerianMidiAlmanac" },
  { id_GANSS_alm_keplerianGLONASS, "id-GANSS-alm-keplerianGLONASS" },
  { id_GANSS_alm_ecefSBASAlmanac, "id-GANSS-alm-ecefSBASAlmanac" },
  { id_DL_RLC_PDU_Size_Format, "id-DL-RLC-PDU-Size-Format" },
  { id_MACes_Maximum_Bitrate_LCR, "id-MACes-Maximum-Bitrate-LCR" },
  { id_Single_Stream_MIMO_ActivationIndicator, "id-Single-Stream-MIMO-ActivationIndicator" },
  { id_Single_Stream_MIMO_Mode_Indicator, "id-Single-Stream-MIMO-Mode-Indicator" },
  { id_Dual_Band_Secondary_Serving_Cell_List, "id-Dual-Band-Secondary-Serving-Cell-List" },
  { id_UE_AggregateMaximumBitRate, "id-UE-AggregateMaximumBitRate" },
  { id_power_offset_for_S_CPICH_for_MIMO, "id-power-offset-for-S-CPICH-for-MIMO" },
  { id_power_offset_for_S_CPICH_for_MIMO_Request_Indicator, "id-power-offset-for-S-CPICH-for-MIMO-Request-Indicator" },
  { id_UE_SupportIndicatorExtension, "id-UE-SupportIndicatorExtension" },
  { id_Unknown_832, "id-Unknown-832" },
  { id_Unknown_833, "id-Unknown-833" },
  { id_Unknown_834, "id-Unknown-834" },
  { id_ActivationInformation, "id-ActivationInformation" },
  { id_CellPortionLCRID, "id-CellPortionLCRID" },
  { id_Additional_EDCH_Cell_Information_RL_Setup_Req, "id-Additional-EDCH-Cell-Information-RL-Setup-Req" },
  { id_Additional_EDCH_Cell_Information_Response, "id-Additional-EDCH-Cell-Information-Response" },
  { id_Additional_EDCH_Cell_Information_RL_Add_Req, "id-Additional-EDCH-Cell-Information-RL-Add-Req" },
  { id_Additional_EDCH_Cell_Information_Response_RLAdd, "id-Additional-EDCH-Cell-Information-Response-RLAdd" },
  { id_Additional_EDCH_Cell_Information_RL_Reconf_Prep, "id-Additional-EDCH-Cell-Information-RL-Reconf-Prep" },
  { id_Additional_EDCH_Cell_Information_RL_Reconf_Req, "id-Additional-EDCH-Cell-Information-RL-Reconf-Req" },
  { id_Additional_EDCH_Cell_Information_RL_Param_Upd, "id-Additional-EDCH-Cell-Information-RL-Param-Upd" },
  { id_Additional_EDCH_Preconfiguration_Information, "id-Additional-EDCH-Preconfiguration-Information" },
  { id_MulticellEDCH_Information, "id-MulticellEDCH-Information" },
  { id_Unknown_846, "id-Unknown-846" },
  { id_Unknown_847, "id-Unknown-847" },
  { id_Unknown_848, "id-Unknown-848" },
  { id_Unknown_849, "id-Unknown-849" },
  { id_Unknown_850, "id-Unknown-850" },
  { id_Unknown_851, "id-Unknown-851" },
  { id_Unknown_852, "id-Unknown-852" },
  { id_Unknown_853, "id-Unknown-853" },
  { id_Additional_EDCH_Cell_Information_ResponseRLReconf, "id-Additional-EDCH-Cell-Information-ResponseRLReconf" },
  { id_EDCH_Indicator, "id-EDCH-Indicator" },
  { id_DiversityMode, "id-DiversityMode" },
  { id_TransmitDiversityIndicator, "id-TransmitDiversityIndicator" },
  { id_NonCellSpecificTxDiversity, "id-NonCellSpecificTxDiversity" },
  { id_CellCapabilityContainerExtension_FDD, "id-CellCapabilityContainerExtension-FDD" },
  { id_HSDSCH_Physical_Layer_Category, "id-HSDSCH-Physical-Layer-Category" },
  { id_E_RNTI_For_FACH, "id-E-RNTI-For-FACH" },
  { id_H_RNTI_For_FACH, "id-H-RNTI-For-FACH" },
  { id_RNTI_Allocation_Indicator, "id-RNTI-Allocation-Indicator" },
  { id_UE_AggregateMaximumBitRate_Enforcement_Indicator, "id-UE-AggregateMaximumBitRate-Enforcement-Indicator" },
  { id_DCH_MeasurementOccasion_Information, "id-DCH-MeasurementOccasion-Information" },
  { id_DCH_MeasurementType_Indicator, "id-DCH-MeasurementType-Indicator" },
  { id_Out_of_Sychronization_Window, "id-Out-of-Sychronization-Window" },
  { id_MulticellEDCH_RL_SpecificInformation, "id-MulticellEDCH-RL-SpecificInformation" },
  { id_DGNSS_ValidityPeriod, "id-DGNSS-ValidityPeriod" },
  { id_TS0_HS_PDSCH_Indication_LCR, "id-TS0-HS-PDSCH-Indication-LCR" },
  { id_UE_TS0_CapabilityLCR, "id-UE-TS0-CapabilityLCR" },
  { id_Non_Serving_RL_Preconfig_Info, "id-Non-Serving-RL-Preconfig-Info" },
  { id_Non_Serving_RL_Preconfig_Setup, "id-Non-Serving-RL-Preconfig-Setup" },
  { id_Non_Serving_RL_Preconfig_Removal, "id-Non-Serving-RL-Preconfig-Removal" },
  { id_Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup, "id-Additional-E-DCH-Non-Serving-RL-Preconfiguration-Setup" },
  { id_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList, "id-Additional-E-DCH-New-non-serving-RL-E-DCH-FDD-DL-Control-Channel-InfoList" },
  { id_CellListValidityIndicator, "id-CellListValidityIndicator" },
  { id_completeAlmanacProvided, "id-completeAlmanacProvided" },
  { id_ganss_Delta_T, "id-ganss-Delta-T" },
  { id_OrdinalNumberOfFrequency, "id-OrdinalNumberOfFrequency" },
  { id_Multicell_EDCH_Restriction, "id-Multicell-EDCH-Restriction" },
  { id_CellCapabilityContainerExtension_TDD_LCR, "id-CellCapabilityContainerExtension-TDD-LCR" },
  { id_Multi_Carrier_EDCH_Setup, "id-Multi-Carrier-EDCH-Setup" },
  { id_Multi_Carrier_EDCH_Reconfigure, "id-Multi-Carrier-EDCH-Reconfigure" },
  { id_Multi_Carrier_EDCH_Response, "id-Multi-Carrier-EDCH-Response" },
  { id_SNPL_Carrier_Group_Indicator, "id-SNPL-Carrier-Group-Indicator" },
  { id_MU_MIMO_InformationLCR, "id-MU-MIMO-InformationLCR" },
  { id_MU_MIMO_Indicator, "id-MU-MIMO-Indicator" },
  { id_HS_SCCH_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext, "id-HS-SCCH-Inactivity-Threshold-for-UE-DRX-Cycle-LCR-Ext" },
  { id_Usefulness_Of_Battery_Optimization, "id-Usefulness-Of-Battery-Optimization" },
  { id_Multi_Carrier_E_DCH_LCRTDD_PhysicalLayerCategory, "id-Multi-Carrier-E-DCH-LCRTDD-PhysicalLayerCategory" },
  { id_Measurement_Power_Offset, "id-Measurement-Power-Offset" },
  { id_MDT_Configuration, "id-MDT-Configuration" },
  { id_Neighbouring_UMTS_CellInformation_Ext, "id-Neighbouring-UMTS-CellInformation-Ext" },
  { id_Neighbouring_UMTS_CellInformationExtensionItem, "id-Neighbouring-UMTS-CellInformationExtensionItem" },
  { id_Control_Type_InformationTransferControlReq, "id-Control-Type-InformationTransferControlReq" },
  { id_UMTS_Cells_Info, "id-UMTS-Cells-Info" },
  { id_ANRReportIndication, "id-ANRReportIndication" },
  { id_ANR_Cell_InfEx_Rqst, "id-ANR-Cell-InfEx-Rqst" },
  { id_ANR_Cell_InfEx_Rsp, "id-ANR-Cell-InfEx-Rsp" },
  { id_ANR_Cell_Information, "id-ANR-Cell-Information" },
  { id_Trace_Collection_Entity_IP_Address, "id-Trace-Collection-Entity-IP-Address" },
  { id_Affected_HSDSCH_Serving_Cell_List, "id-Affected-HSDSCH-Serving-Cell-List" },
  { id_UL_CLTD_Information, "id-UL-CLTD-Information" },
  { id_UL_CLTD_Information_Reconf, "id-UL-CLTD-Information-Reconf" },
  { id_UL_CLTD_State_Update_Information, "id-UL-CLTD-State-Update-Information" },
  { id_Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order, "id-Support-of-Dynamic-DTXDRX-Related-HS-SCCH-Order" },
  { id_CPC_RecoveryReport, "id-CPC-RecoveryReport" },
  { id_FTPICH_Information, "id-FTPICH-Information" },
  { id_FTPICH_Information_Reconf, "id-FTPICH-Information-Reconf" },
  { id_UE_RF_Band_CapabilityLCR, "id-UE-RF-Band-CapabilityLCR" },
  { id_Extended_S_RNTI, "id-Extended-S-RNTI" },
  { id_ExtendedAffectedUEInformationForMBMS, "id-ExtendedAffectedUEInformationForMBMS" },
  { id_Extended_S_RNTI_Group, "id-Extended-S-RNTI-Group" },
  { id_FTPICH_Information_Response, "id-FTPICH-Information-Response" },
  { id_FTPICH_Reconfiguration_Information, "id-FTPICH-Reconfiguration-Information" },
  { id_MIMO_withfourtransmitantennas_ActivationIndicator, "id-MIMO-withfourtransmitantennas-ActivationIndicator" },
  { id_MIMO_withfourtransmitantennas_ModeIndicator, "id-MIMO-withfourtransmitantennas-ModeIndicator" },
  { id_power_offset_for_S_CPICH_for_MIMO_withfourtransmitantennas, "id-power-offset-for-S-CPICH-for-MIMO-withfourtransmitantennas" },
  { id_power_offset_for_S_CPICH_for_MIMO_withfourtransmitantennas_Request_Indicator, "id-power-offset-for-S-CPICH-for-MIMO-withfourtransmitantennas-Request-Indicator" },
  { id_DualStream_MIMO_withfourtransmitantennas_ActivationIndicator, "id-DualStream-MIMO-withfourtransmitantennas-ActivationIndicator" },
  { id_DualStream_MIMO_withfourtransmitantennas_ModeIndicator, "id-DualStream-MIMO-withfourtransmitantennas-ModeIndicator" },
  { id_Additional_Associated_Secondary_CPICH, "id-Additional-Associated-Secondary-CPICH" },
  { id_UE_transmission_power_headroom, "id-UE-transmission-power-headroom" },
  { id_Multiflow_Information, "id-Multiflow-Information" },
  { id_Multiflow_Reconfiguration, "id-Multiflow-Reconfiguration" },
  { id_Multiflow_OrdinalNumberOfFrequency, "id-Multiflow-OrdinalNumberOfFrequency" },
  { id_UL_MIMO_Information, "id-UL-MIMO-Information" },
  { id_UL_MIMO_Reconfiguration, "id-UL-MIMO-Reconfiguration" },
  { id_UL_MIMO_DL_Control_Channel_Information, "id-UL-MIMO-DL-Control-Channel-Information" },
  { id_SixtyfourQAM_UL_Operation_Indicator, "id-SixtyfourQAM-UL-Operation-Indicator" },
  { id_Common_E_RGCH_Cell_InfEx_Rqst, "id-Common-E-RGCH-Cell-InfEx-Rqst" },
  { id_Common_E_RGCH_Cell_InfEx_Rsp, "id-Common-E-RGCH-Cell-InfEx-Rsp" },
  { id_Common_E_RGCH_Cell_InfEx_Rprt, "id-Common-E-RGCH-Cell-InfEx-Rprt" },
  { id_Common_E_RGCH_Cell_Information, "id-Common-E-RGCH-Cell-Information" },
  { id_PCI, "id-PCI" },
  { id_TAC, "id-TAC" },
  { id_PLMN_List, "id-PLMN-List" },
  { id_PrecoderWeightSetRestriction, "id-PrecoderWeightSetRestriction" },
  { id_EARFCN_FDD_Extended, "id-EARFCN-FDD-Extended" },
  { id_EARFCN_TDD_Extended, "id-EARFCN-TDD-Extended" },
  { id_Non_rectangular_resource_allocation_indicator, "id-Non-rectangular-resource-allocation-indicator" },
  { id_Non_rectangular_resource_timeslot_set, "id-Non-rectangular-resource-timeslot-set" },
  { id_UE_Support_of_non_rectangular_resource_allocation, "id-UE-Support-of-non-rectangular-resource-allocation" },
  { id_DBDS_CorrectionsReq, "id-DBDS-CorrectionsReq" },
  { id_DBDS_Corrections, "id-DBDS-Corrections" },
  { id_BDS_IonosphericGridModelReq, "id-BDS-IonosphericGridModelReq" },
  { id_BDS_Ionospheric_Grid_Model, "id-BDS-Ionospheric-Grid-Model" },
  { id_GANSS_alm_keplerianBDSAlmanac, "id-GANSS-alm-keplerianBDSAlmanac" },
  { id_Assisting_RepetitionFactors, "id-Assisting-RepetitionFactors" },
  { id_E_DCH_Decoupling_Indication, "id-E-DCH-Decoupling-Indication" },
  { id_UE_Measurement_Forwarding, "id-UE-Measurement-Forwarding" },
  { id_UPH_Filtering_Measurement_Forwarding_Request, "id-UPH-Filtering-Measurement-Forwarding-Request" },
  { id_TTI_Update_Indicator, "id-TTI-Update-Indicator" },
  { id_CQI_Feedback_Cycle2_k, "id-CQI-Feedback-Cycle2-k" },
  { id_CQI_Cycle_Switch_Timer, "id-CQI-Cycle-Switch-Timer" },
  { id_UE_DRX_Cycle2, "id-UE-DRX-Cycle2" },
  { id_Inactivity_Threshold_for_UE_DRX_Cycle2, "id-Inactivity-Threshold-for-UE-DRX-Cycle2" },
  { id_DTX_Information2, "id-DTX-Information2" },
  { id_DCH_ENH_Information, "id-DCH-ENH-Information" },
  { id_DCH_ENH_Information_Reconf, "id-DCH-ENH-Information-Reconf" },
  { id_Gainfactors_10ms_mode, "id-Gainfactors-10ms-mode" },
  { id_Radio_Links_without_DPCH_FDPCH_Indication, "id-Radio-Links-without-DPCH-FDPCH-Indication" },
  { id_UL_DPCCH2_Information, "id-UL-DPCCH2-Information" },
  { id_UL_DPCCH2_Information_Reconf, "id-UL-DPCCH2-Information-Reconf" },
  { id_ImplicitGrantHandling, "id-ImplicitGrantHandling" },
  { id_MinimumTEBSthreshold, "id-MinimumTEBSthreshold" },
  { id_MultipleFrequencyBandsIndicatorList_EUTRA, "id-MultipleFrequencyBandsIndicatorList-EUTRA" },
  { id_MultipleFrequencyBandsIndicatorList, "id-MultipleFrequencyBandsIndicatorList" },
  { id_ActivationDelay, "id-ActivationDelay" },
  { id_Fast_TTI_switching_Mode_synchronized, "id-Fast-TTI-switching-Mode-synchronized" },
  { id_Fast_TTI_switching_Mode_unsynchronized, "id-Fast-TTI-switching-Mode-unsynchronized" },
  { id_Fast_TTI_switching_Mode_Supported, "id-Fast-TTI-switching-Mode-Supported" },
  { id_GANSS_BDSclockModel, "id-GANSS-BDSclockModel" },
  { id_GANSS_NavModel_BDSKeplerianSet, "id-GANSS-NavModel-BDSKeplerianSet" },
  { id_GANSS_UTCmodelSet4, "id-GANSS-UTCmodelSet4" },
  { id_Inter_Frequency_Cell_Information_SIB11ter, "id-Inter-Frequency-Cell-Information-SIB11ter" },
  { id_Improved_Synchronized_Indicator, "id-Improved-Synchronized-Indicator" },
  { id_Downlink_TPC_enhancements_Information, "id-Downlink-TPC-enhancements-Information" },
  { id_Downlink_TPC_enhancements_Reconf, "id-Downlink-TPC-enhancements-Reconf" },
  { id_TPC_slot_position, "id-TPC-slot-position" },
  { id_DL_TBS, "id-DL-TBS" },
  { id_Selected_PLMN_Identity, "id-Selected-PLMN-Identity" },
  { id_SpeechVersion, "id-SpeechVersion" },
  { id_SourceID, "id-SourceID" },
  { id_TargetID, "id-TargetID" },
  { id_ClassmarkInformation2, "id-ClassmarkInformation2" },
  { id_ClassmarkInformation3, "id-ClassmarkInformation3" },
  { id_GSM_Cell_CM_Rqst, "id-GSM-Cell-CM-Rqst" },
  { id_Extension_CommonMeasurementObjectType_CM_Rprt, "id-Extension-CommonMeasurementObjectType-CM-Rprt" },
  { id_Extension_CommonMeasurementObjectType_CM_Rqst, "id-Extension-CommonMeasurementObjectType-CM-Rqst" },
  { id_Extension_CommonMeasurementObjectType_CM_Rsp, "id-Extension-CommonMeasurementObjectType-CM-Rsp" },
  { id_Extension_FailureIndicationMeasurementList, "id-Extension-FailureIndicationMeasurementList" },
  { id_Extension_FailureMeasurementList, "id-Extension-FailureMeasurementList" },
  { id_Extension_TerminationMeasurementList, "id-Extension-TerminationMeasurementList" },
  { id_GsmCellList_CM_Rprt, "id-GsmCellList-CM-Rprt" },
  { id_GsmCellList_CM_Rqst, "id-GsmCellList-CM-Rqst" },
  { id_GsmCellList_CM_Rsp, "id-GsmCellList-CM-Rsp" },
  { id_LoadValue, "id-LoadValue" },
  { id_EventH, "id-EventH" },
  { 0, NULL }
};

static value_string_ext rnsap_ProtocolIE_ID_vals_ext = VALUE_STRING_EXT_INIT(rnsap_ProtocolIE_ID_vals);


static int
dissect_rnsap_ProtocolIE_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxProtocolIEs, &rnsap_get_private_data(actx->pinfo)->ProtocolIE_ID, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_127(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_32767(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 32767U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_TransactionID_vals[] = {
  {   0, "shortTransActionId" },
  {   1, "longTransActionId" },
  { 0, NULL }
};

static const per_choice_t TransactionID_choice[] = {
  {   0, &hf_rnsap_shortTransActionId, ASN1_NO_EXTENSIONS     , dissect_rnsap_INTEGER_0_127 },
  {   1, &hf_rnsap_longTransActionId, ASN1_NO_EXTENSIONS     , dissect_rnsap_INTEGER_0_32767 },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_TransactionID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_TransactionID, TransactionID_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_TriggeringMessage_vals[] = {
  {   0, "initiating-message" },
  {   1, "successful-outcome" },
  {   2, "unsuccessful-outcome" },
  {   3, "outcome" },
  { 0, NULL }
};


static int
dissect_rnsap_TriggeringMessage(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_ProtocolIE_Field_value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type_pdu_new(tvb, offset, actx, tree, hf_index, dissect_ProtocolIEFieldValue);

  return offset;
}


static const per_sequence_t ProtocolIE_Field_sequence[] = {
  { &hf_rnsap_id            , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_ID },
  { &hf_rnsap_criticality   , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Criticality },
  { &hf_rnsap_value         , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Field_value },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ProtocolIE_Field(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ProtocolIE_Field, ProtocolIE_Field_sequence);

  return offset;
}


static const per_sequence_t ProtocolIE_Container_sequence_of[1] = {
  { &hf_rnsap_ProtocolIE_Container_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Field },
};

static int
dissect_rnsap_ProtocolIE_Container(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_ProtocolIE_Container, ProtocolIE_Container_sequence_of,
                                                  0, maxProtocolIEs, FALSE);

  return offset;
}



static int
dissect_rnsap_ProtocolIE_Single_Container(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Field(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_T_extensionValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type_pdu_new(tvb, offset, actx, tree, hf_index, dissect_ProtocolExtensionFieldExtensionValue);

  return offset;
}


static const per_sequence_t ProtocolExtensionField_sequence[] = {
  { &hf_rnsap_id            , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_ID },
  { &hf_rnsap_criticality   , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Criticality },
  { &hf_rnsap_extensionValue, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_T_extensionValue },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ProtocolExtensionField(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ProtocolExtensionField, ProtocolExtensionField_sequence);

  return offset;
}


static const per_sequence_t ProtocolExtensionContainer_sequence_of[1] = {
  { &hf_rnsap_ProtocolExtensionContainer_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolExtensionField },
};

static int
dissect_rnsap_ProtocolExtensionContainer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_ProtocolExtensionContainer, ProtocolExtensionContainer_sequence_of,
                                                  1, maxProtocolExtensions, FALSE);

  return offset;
}



static int
dissect_rnsap_PrivateIE_Field_value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type_pdu_new(tvb, offset, actx, tree, hf_index, dissect_PrivateIEFieldValue);

  return offset;
}


static const per_sequence_t PrivateIE_Field_sequence[] = {
  { &hf_rnsap_id_01         , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_PrivateIE_ID },
  { &hf_rnsap_criticality   , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Criticality },
  { &hf_rnsap_value_01      , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_PrivateIE_Field_value },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PrivateIE_Field(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PrivateIE_Field, PrivateIE_Field_sequence);

  return offset;
}


static const per_sequence_t PrivateIE_Container_sequence_of[1] = {
  { &hf_rnsap_PrivateIE_Container_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_PrivateIE_Field },
};

static int
dissect_rnsap_PrivateIE_Container(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_PrivateIE_Container, PrivateIE_Container_sequence_of,
                                                  1, maxPrivateIEs, FALSE);

  return offset;
}



static int
dissect_rnsap_AccessPointName(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       1, 255, FALSE, NULL);

  return offset;
}



static int
dissect_rnsap_AckNack_RepetitionFactor(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 4U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_Ack_Power_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_Uu_ActivationState_vals[] = {
  {   0, "activated" },
  {   1, "de-activated" },
  {   2, "changeRequest" },
  { 0, NULL }
};


static int
dissect_rnsap_Uu_ActivationState(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 1, NULL);

  return offset;
}


static const per_sequence_t ActivationInformationItem_sequence[] = {
  { &hf_rnsap_uU_ActivationState, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Uu_ActivationState },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ActivationInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ActivationInformationItem, ActivationInformationItem_sequence);

  return offset;
}


static const per_sequence_t ActivationInformation_sequence_of[1] = {
  { &hf_rnsap_ActivationInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ActivationInformationItem },
};

static int
dissect_rnsap_ActivationInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_ActivationInformation, ActivationInformation_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}



static int
dissect_rnsap_PLMN_Identity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       3, 3, FALSE, NULL);

  return offset;
}



static int
dissect_rnsap_Service_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       3, 3, FALSE, NULL);

  return offset;
}


static const per_sequence_t TMGI_sequence[] = {
  { &hf_rnsap_plmn_id       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PLMN_Identity },
  { &hf_rnsap_service_id    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Service_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TMGI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TMGI, TMGI_sequence);

  return offset;
}


static const value_string rnsap_TransmissionMode_vals[] = {
  {   0, "p-t-p" },
  {   1, "p-t-m" },
  {   2, "not-provided" },
  { 0, NULL }
};


static int
dissect_rnsap_TransmissionMode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t MBMS_Bearer_ServiceItemFDD_sequence[] = {
  { &hf_rnsap_tmgi          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TMGI },
  { &hf_rnsap_transmissionMode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransmissionMode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Bearer_ServiceItemFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Bearer_ServiceItemFDD, MBMS_Bearer_ServiceItemFDD_sequence);

  return offset;
}


static const per_sequence_t Active_MBMS_Bearer_Service_ListFDD_sequence_of[1] = {
  { &hf_rnsap_Active_MBMS_Bearer_Service_ListFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_ServiceItemFDD },
};

static int
dissect_rnsap_Active_MBMS_Bearer_Service_ListFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Active_MBMS_Bearer_Service_ListFDD, Active_MBMS_Bearer_Service_ListFDD_sequence_of,
                                                  1, maxNrOfActiveMBMSServices, FALSE);

  return offset;
}



static int
dissect_rnsap_UARFCN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16383U, NULL, TRUE);

  return offset;
}


static const per_sequence_t MBMS_Bearer_ServiceItemFDD_PFL_sequence[] = {
  { &hf_rnsap_tmgi          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TMGI },
  { &hf_rnsap_transmissionMode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransmissionMode },
  { &hf_rnsap_preferredFrequencyLayer, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Bearer_ServiceItemFDD_PFL(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Bearer_ServiceItemFDD_PFL, MBMS_Bearer_ServiceItemFDD_PFL_sequence);

  return offset;
}


static const per_sequence_t Active_MBMS_Bearer_Service_ListFDD_PFL_sequence_of[1] = {
  { &hf_rnsap_Active_MBMS_Bearer_Service_ListFDD_PFL_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_ServiceItemFDD_PFL },
};

static int
dissect_rnsap_Active_MBMS_Bearer_Service_ListFDD_PFL(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Active_MBMS_Bearer_Service_ListFDD_PFL, Active_MBMS_Bearer_Service_ListFDD_PFL_sequence_of,
                                                  1, maxNrOfActiveMBMSServices, FALSE);

  return offset;
}


static const per_sequence_t MBMS_Bearer_ServiceItemTDD_sequence[] = {
  { &hf_rnsap_tmgi          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TMGI },
  { &hf_rnsap_transmissionMode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransmissionMode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Bearer_ServiceItemTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Bearer_ServiceItemTDD, MBMS_Bearer_ServiceItemTDD_sequence);

  return offset;
}


static const per_sequence_t Active_MBMS_Bearer_Service_ListTDD_sequence_of[1] = {
  { &hf_rnsap_Active_MBMS_Bearer_Service_ListTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_ServiceItemTDD },
};

static int
dissect_rnsap_Active_MBMS_Bearer_Service_ListTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Active_MBMS_Bearer_Service_ListTDD, Active_MBMS_Bearer_Service_ListTDD_sequence_of,
                                                  1, maxNrOfActiveMBMSServices, FALSE);

  return offset;
}


static const per_sequence_t MBMS_Bearer_ServiceItemTDD_PFL_sequence[] = {
  { &hf_rnsap_tmgi          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TMGI },
  { &hf_rnsap_transmissionMode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransmissionMode },
  { &hf_rnsap_preferredFrequencyLayer, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Bearer_ServiceItemTDD_PFL(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Bearer_ServiceItemTDD_PFL, MBMS_Bearer_ServiceItemTDD_PFL_sequence);

  return offset;
}


static const per_sequence_t Active_MBMS_Bearer_Service_ListTDD_PFL_sequence_of[1] = {
  { &hf_rnsap_Active_MBMS_Bearer_Service_ListTDD_PFL_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_ServiceItemTDD_PFL },
};

static int
dissect_rnsap_Active_MBMS_Bearer_Service_ListTDD_PFL(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Active_MBMS_Bearer_Service_ListTDD_PFL, Active_MBMS_Bearer_Service_ListTDD_PFL_sequence_of,
                                                  1, maxNrOfActiveMBMSServices, FALSE);

  return offset;
}



static int
dissect_rnsap_CFN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_TGPSID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxTGPS, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_TGPRC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Transmission_Gap_Pattern_Sequence_Status_List_item_sequence[] = {
  { &hf_rnsap_tGPSID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TGPSID },
  { &hf_rnsap_tGPRC         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TGPRC },
  { &hf_rnsap_tGCFN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CFN },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Transmission_Gap_Pattern_Sequence_Status_List_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Transmission_Gap_Pattern_Sequence_Status_List_item, Transmission_Gap_Pattern_Sequence_Status_List_item_sequence);

  return offset;
}


static const per_sequence_t Transmission_Gap_Pattern_Sequence_Status_List_sequence_of[1] = {
  { &hf_rnsap_Transmission_Gap_Pattern_Sequence_Status_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Transmission_Gap_Pattern_Sequence_Status_List_item },
};

static int
dissect_rnsap_Transmission_Gap_Pattern_Sequence_Status_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Transmission_Gap_Pattern_Sequence_Status_List, Transmission_Gap_Pattern_Sequence_Status_List_sequence_of,
                                                  0, maxTGPS, FALSE);

  return offset;
}


static const per_sequence_t Active_Pattern_Sequence_Information_sequence[] = {
  { &hf_rnsap_cMConfigurationChangeCFN, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CFN },
  { &hf_rnsap_transmission_Gap_Pattern_Sequence_Status, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Transmission_Gap_Pattern_Sequence_Status_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Active_Pattern_Sequence_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Active_Pattern_Sequence_Information, Active_Pattern_Sequence_Information_sequence);

  return offset;
}



static int
dissect_rnsap_RL_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 31U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Received_total_wide_band_power(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 621U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_DL_PowerBalancing_ActivationIndicator_vals[] = {
  {   0, "dL-PowerBalancing-Activated" },
  { 0, NULL }
};


static int
dissect_rnsap_DL_PowerBalancing_ActivationIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_RL_Set_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 31U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_DL_ScramblingCode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_FDD_DL_ChannelisationCodeNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_E_RNTI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_ERGCH_SignatureSequence(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrofSigSeqERGHICH_1, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_EHICH_SignatureSequence(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrofSigSeqERGHICH_1, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_E_Serving_Grant_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 38U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_E_Primary_Secondary_Grant_Selector_vals[] = {
  {   0, "primary" },
  {   1, "secondary" },
  { 0, NULL }
};


static int
dissect_rnsap_E_Primary_Secondary_Grant_Selector(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_E_RGCH_Release_Indicator_vals[] = {
  {   0, "e-RGCHreleased" },
  { 0, NULL }
};


static int
dissect_rnsap_E_RGCH_Release_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t EDCH_FDD_DL_ControlChannelInformation_sequence[] = {
  { &hf_rnsap_eAGCH_ERGCH_EHICH_FDD_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_ScramblingCode },
  { &hf_rnsap_eAGCH_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_FDD_DL_ChannelisationCodeNumber },
  { &hf_rnsap_primary_e_RNTI, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_RNTI },
  { &hf_rnsap_secondary_e_RNTI, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_RNTI },
  { &hf_rnsap_eRGCH_EHICH_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DL_ChannelisationCodeNumber },
  { &hf_rnsap_eRGCH_SignatureSequence, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ERGCH_SignatureSequence },
  { &hf_rnsap_eHICH_SignatureSequence, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_EHICH_SignatureSequence },
  { &hf_rnsap_serving_Grant_Value, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_Serving_Grant_Value },
  { &hf_rnsap_primary_Secondary_Grant_Selector, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_Primary_Secondary_Grant_Selector },
  { &hf_rnsap_e_RGCH_Release_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_RGCH_Release_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_FDD_DL_ControlChannelInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_FDD_DL_ControlChannelInformation, EDCH_FDD_DL_ControlChannelInformation_sequence);

  return offset;
}


static const value_string rnsap_Transmission_Gap_Pattern_Sequence_ScramblingCode_Information_vals[] = {
  {   0, "code-change" },
  {   1, "nocode-change" },
  { 0, NULL }
};


static int
dissect_rnsap_Transmission_Gap_Pattern_Sequence_ScramblingCode_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t FDD_DL_CodeInformationItem_sequence[] = {
  { &hf_rnsap_dl_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_ScramblingCode },
  { &hf_rnsap_fDD_DL_ChannelisationCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DL_ChannelisationCodeNumber },
  { &hf_rnsap_transmission_Gap_Pattern_Sequence_ScramblingCode_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Transmission_Gap_Pattern_Sequence_ScramblingCode_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_FDD_DL_CodeInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_FDD_DL_CodeInformationItem, FDD_DL_CodeInformationItem_sequence);

  return offset;
}


static const per_sequence_t FDD_DL_CodeInformation_sequence_of[1] = {
  { &hf_rnsap_FDD_DL_CodeInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DL_CodeInformationItem },
};

static int
dissect_rnsap_FDD_DL_CodeInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_FDD_DL_CodeInformation, FDD_DL_CodeInformation_sequence_of,
                                                  1, maxNrOfDL_Codes, FALSE);

  return offset;
}



static int
dissect_rnsap_EDCH_MACdFlow_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfEDCHMACdFlows_1, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_BindingID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  tvbuff_t *parameter_tvb=NULL;
  guint16 binding_id_port;

  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       1, 4, TRUE, &parameter_tvb);


  if (!parameter_tvb)
    return offset;

  if(tvb_reported_length(parameter_tvb)>=2){
    binding_id_port = tvb_get_ntohs(parameter_tvb,0);
    proto_item_append_text(actx->created_item, " (%u)",binding_id_port);
  }


  return offset;
}



static int
dissect_rnsap_TransportLayerAddress(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  tvbuff_t *parameter_tvb = NULL;
  proto_item *item;
  proto_tree *subtree, *nsap_tree;
  guint8 *padded_nsap_bytes;
  tvbuff_t *nsap_tvb;
  gint tvb_len;

  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     1, 160, TRUE, NULL, 0, &parameter_tvb, NULL);

  if (!parameter_tvb)
    return offset;

  /* Get the length */
  tvb_len = tvb_reported_length(parameter_tvb);
  subtree = proto_item_add_subtree(actx->created_item, ett_rnsap_transportLayerAddress);
  if (tvb_len == 4){
    /* IPv4 */
    proto_tree_add_item(subtree, hf_rnsap_transportLayerAddress_ipv4, parameter_tvb, 0, tvb_len, ENC_BIG_ENDIAN);
  }
  if (tvb_len == 16){
    /* IPv6 */
    proto_tree_add_item(subtree, hf_rnsap_transportLayerAddress_ipv6, parameter_tvb, 0, tvb_len, ENC_NA);
  }
  if (tvb_len == 20 || tvb_len == 7){
    /* NSAP */
    if (tvb_len == 7){
      /* Unpadded IPv4 NSAP */
      /* Creating a new TVB with padding */
      padded_nsap_bytes = (guint8*) wmem_alloc0(actx->pinfo->pool, 20);
      tvb_memcpy(parameter_tvb, padded_nsap_bytes, 0, tvb_len);
      nsap_tvb = tvb_new_child_real_data(tvb, padded_nsap_bytes, 20, 20);
      add_new_data_source(actx->pinfo, nsap_tvb, "Padded NSAP Data");
    }else{
      /* Padded NSAP*/
      nsap_tvb = parameter_tvb;
    }
    item = proto_tree_add_item(subtree, hf_rnsap_transportLayerAddress_nsap, parameter_tvb, 0, tvb_len, ENC_NA);
    nsap_tree = proto_item_add_subtree(item, ett_rnsap_transportLayerAddress_nsap);
    dissect_nsap(nsap_tvb, 0, 20, nsap_tree);
  }


  return offset;
}


static const per_sequence_t Additional_EDCH_MAC_d_Flows_Specific_Info_Response_sequence[] = {
  { &hf_rnsap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_Response, Additional_EDCH_MAC_d_Flows_Specific_Info_Response_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_Response },
};

static int
dissect_rnsap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List, Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}



static int
dissect_rnsap_HARQ_Process_Allocation_2ms_EDCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     maxNrOfEDCHHARQProcesses2msEDCH, maxNrOfEDCHHARQProcesses2msEDCH, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_UL_SIR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -82, 173U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_MaximumAllowedULTxPower(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -50, 33U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_DL_Power(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -350, 150U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_PrimaryScramblingCode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_PrimaryCPICH_Power(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -100, 500U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_PC_Preamble(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_Primary_CPICH_Usage_For_Channel_Estimation_vals[] = {
  {   0, "primary-CPICH-may-be-used" },
  {   1, "primary-CPICH-shall-not-be-used" },
  { 0, NULL }
};


static int
dissect_rnsap_Primary_CPICH_Usage_For_Channel_Estimation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t Secondary_CPICH_Information_sequence[] = {
  { &hf_rnsap_dl_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_ScramblingCode },
  { &hf_rnsap_fDD_DL_ChannelisationCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DL_ChannelisationCodeNumber },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Secondary_CPICH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Secondary_CPICH_Information, Secondary_CPICH_Information_sequence);

  return offset;
}



static int
dissect_rnsap_F_DPCH_SlotFormat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 9U, NULL, FALSE);

  return offset;
}


static const per_sequence_t EDCH_Additional_RL_Specific_Information_Response_ItemIEs_sequence[] = {
  { &hf_rnsap_eDCH_Additional_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_received_total_wide_band_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Received_total_wide_band_power },
  { &hf_rnsap_dL_PowerBalancing_ActivationIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_PowerBalancing_ActivationIndicator },
  { &hf_rnsap_rL_Set_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_e_DCH_RL_Set_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_eDCH_FDD_DL_ControlChannelInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_FDD_DL_ControlChannelInformation },
  { &hf_rnsap_dl_CodeInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DL_CodeInformation },
  { &hf_rnsap_additional_EDCH_MAC_d_Flow_Specific_Information_Response_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List },
  { &hf_rnsap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_rnsap_maxUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_minUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_maximumAllowedULTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaximumAllowedULTxPower },
  { &hf_rnsap_maximumDL_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDL_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_primaryScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PrimaryScramblingCode },
  { &hf_rnsap_uL_UARFCN     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_dL_UARFCN     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_primaryCPICH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PrimaryCPICH_Power },
  { &hf_rnsap_pC_Preamble   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PC_Preamble },
  { &hf_rnsap_primary_CPICH_Usage_For_Channel_Estimation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Primary_CPICH_Usage_For_Channel_Estimation },
  { &hf_rnsap_secondary_CPICH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Secondary_CPICH_Information },
  { &hf_rnsap_f_DPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_F_DPCH_SlotFormat },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_Additional_RL_Specific_Information_Response_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_Additional_RL_Specific_Information_Response_ItemIEs, EDCH_Additional_RL_Specific_Information_Response_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t EDCH_Additional_RL_Specific_Information_Response_List_sequence_of[1] = {
  { &hf_rnsap_EDCH_Additional_RL_Specific_Information_Response_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_Additional_RL_Specific_Information_Response_ItemIEs },
};

static int
dissect_rnsap_EDCH_Additional_RL_Specific_Information_Response_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_EDCH_Additional_RL_Specific_Information_Response_List, EDCH_Additional_RL_Specific_Information_Response_List_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_FDD_Information_Response_ItemIEs_sequence[] = {
  { &hf_rnsap_eDCH_Additional_RL_Specific_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_Additional_RL_Specific_Information_Response_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_FDD_Information_Response_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_FDD_Information_Response_ItemIEs, Additional_EDCH_FDD_Information_Response_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t E_DCH_RL_InformationList_Rsp_Item_sequence[] = {
  { &hf_rnsap_e_DCH_reconfigured_RL_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_e_DCH_FDD_DL_Control_Channel_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_FDD_DL_ControlChannelInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_RL_InformationList_Rsp_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_RL_InformationList_Rsp_Item, E_DCH_RL_InformationList_Rsp_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_RL_InformationList_Rsp_sequence_of[1] = {
  { &hf_rnsap_E_DCH_RL_InformationList_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_RL_InformationList_Rsp_Item },
};

static int
dissect_rnsap_E_DCH_RL_InformationList_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_RL_InformationList_Rsp, E_DCH_RL_InformationList_Rsp_sequence_of,
                                                  0, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_serving_cell_change_successful_sequence[] = {
  { &hf_rnsap_e_DCH_RL_InformationList_Rsp, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_RL_InformationList_Rsp },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_serving_cell_change_successful(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_serving_cell_change_successful, E_DCH_serving_cell_change_successful_sequence);

  return offset;
}


static const value_string rnsap_CauseRadioNetwork_vals[] = {
  {   0, "unknown-C-ID" },
  {   1, "cell-not-available" },
  {   2, "power-level-not-supported" },
  {   3, "ul-scrambling-code-already-in-use" },
  {   4, "dl-radio-resources-not-available" },
  {   5, "ul-radio-resources-not-available" },
  {   6, "measurement-not-supported-for-the-object" },
  {   7, "combining-resources-not-available" },
  {   8, "combining-not-supported" },
  {   9, "reconfiguration-not-allowed" },
  {  10, "requested-configuration-not-supported" },
  {  11, "synchronisation-failure" },
  {  12, "requested-tx-diversity-mode-not-supported" },
  {  13, "measurement-temporaily-not-available" },
  {  14, "unspecified" },
  {  15, "invalid-CM-settings" },
  {  16, "reconfiguration-CFN-not-elapsed" },
  {  17, "number-of-DL-codes-not-supported" },
  {  18, "dedicated-transport-channel-type-not-supported" },
  {  19, "dl-shared-channel-type-not-supported" },
  {  20, "ul-shared-channel-type-not-supported" },
  {  21, "common-transport-channel-type-not-supported" },
  {  22, "ul-spreading-factor-not-supported" },
  {  23, "dl-spreading-factor-not-supported" },
  {  24, "cm-not-supported" },
  {  25, "transaction-not-supported-by-destination-node-b" },
  {  26, "rl-already-activated-or-alocated" },
  {  27, "number-of-UL-codes-not-supported" },
  {  28, "cell-reserved-for-operator-use" },
  {  29, "dpc-mode-change-not-supported" },
  {  30, "information-temporarily-not-available" },
  {  31, "information-provision-not-supported-for-the-object" },
  {  32, "power-balancing-status-not-compatible" },
  {  33, "delayed-activation-not-supported" },
  {  34, "rl-timing-adjustment-not-supported" },
  {  35, "unknown-RNTI" },
  {  36, "measurement-repetition-rate-not-compatible" },
  {  37, "ue-not-capable-of-support" },
  {  38, "f-dpch-not-supported" },
  {  39, "e-dch-not-supported" },
  {  40, "continuous-packet-connectivity-dtx-drx-operation-not-supported" },
  {  41, "continuous-packet-connectivity-hs-scch-less-operation-not-supported" },
  {  42, "mimo-not-supported" },
  {  43, "e-dch-tti2ms-not-supported" },
  {  44, "continuous-packet-connectivity-DTX-DRX-operation-not-available" },
  {  45, "continuous-packet-connectivity-UE-DTX-Cycle-not-available" },
  {  46, "mimo-not-available" },
  {  47, "sixteenQAM-UL-not-Supported" },
  {  48, "hSDSCH-MACdPDU-SizeFormatNotSupported" },
  {  49, "f-dpch-slot-format-operation-not-supported" },
  {  50, "e-DCH-MACdPDU-SizeFormat-not-available" },
  {  51, "e-DPCCH-Power-Boosting-not-supported" },
  {  52, "trelocprep-expiry" },
  {  53, "directed-retry" },
  {  54, "no-Iu-CS-UP-relocation" },
  {  55, "reduce-load-in-serving-cell" },
  {  56, "relocation-cancelled" },
  {  57, "relocation-desirable-for-radio-reasons" },
  {  58, "resource-optimisation-relocation" },
  {  59, "time-critical-relocation" },
  {  60, "traffic-load-in-the-target-cell-higher-than-in-the-source-cell" },
  {  61, "sixtyfourQAM-DL-and-MIMO-Combined-not-available" },
  {  62, "multi-Cell-operation-not-available" },
  {  63, "multi-Cell-operation-not-supported" },
  {  64, "semi-Persistent-scheduling-not-supported" },
  {  65, "continuous-Packet-Connectivity-DRX-not-supported" },
  {  66, "continuous-Packet-Connectivity-DRX-not-available" },
  {  67, "enhanced-relocation-not-supported" },
  {  68, "relocation-not-supported-due-to-PUESBINE-feature" },
  {  69, "relocation-failure-in-target-RNC" },
  {  70, "relocation-target-not-allowed" },
  {  71, "requested-ciphering-and-or-integrity-protection-algorithms-not-supported" },
  {  72, "sixtyfourQAM-DL-and-MIMO-Combined-not-supported" },
  {  73, "tx-diversity-for-mimo-on-DL-control-channels-not-available" },
  {  74, "single-Stream-MIMO-not-supported" },
  {  75, "single-Stream-MIMO-not-available" },
  {  76, "multi-Cell-operation-with-MIMO-not-supported" },
  {  77, "multi-Cell-operation-with-MIMO-not-available" },
  {  78, "multi-Cell-EDCH-operation-not-supported" },
  {  79, "multi-Cell-EDCH-operation-not-available" },
  {  80, "multi-Cell-operation-with-Single-Stream-MIMO-not-supported" },
  {  81, "multi-Cell-operation-with-Single-Stream-MIMO-not-available" },
  {  82, "cellSpecificTxDiversityHandlingForMultiCellOperationNotAvailable" },
  {  83, "cellSpecificTxDiversityHandlingForMultiCellOperationNotSupported" },
  {  84, "frequencySpecificCompressedModeNotAvailable" },
  {  85, "uL-CLTD-Operation-not-available" },
  {  86, "uL-CLTD-Operation-not-supported" },
  {  87, "mimo-withfourtransmitantennas-not-supported" },
  {  88, "mimo-withfourtransmitantennas-not-available" },
  {  89, "dualstream-mimo-withfourtransmitantennas-not-supported" },
  {  90, "dualstream-mimo-withfourtransmitantennas-not-available" },
  {  91, "multiflow-operation-not-supported" },
  {  92, "multiflow-operation-not-available" },
  {  93, "ul-SixtyfourQAM-not-available" },
  {  94, "ul-SixtyfourQAM-not-supported" },
  {  95, "ul-MIMO-Operation-not-available" },
  {  96, "uL-MIMO-Operation-not-supported" },
  {  97, "ul-MIMO-SixteenQAM-Operation-not-available" },
  {  98, "uL-MIMO-SixteenQAM-Operation-not-supported" },
  {  99, "ul-MIMO-SixtyfourQAM-Operation-not-available" },
  { 100, "uL-MIMO-SixtyfourQAM-Operation-not-supported" },
  { 101, "e-DCH-Decoupling-operation-not-available" },
  { 102, "e-DCH-Decoupling-operation-not-supported" },
  { 103, "basic-dch-enh-not-available" },
  { 104, "basic-dch-enh-not-supported" },
  { 105, "full-dch-enh-not-available" },
  { 106, "full-dch-enh-not-supported" },
  { 107, "radio-Links-without-DPCH-FDPCH-Indication-operation-not-available" },
  { 108, "radio-Links-without-DPCH-FDPCH-Indication-operation-not-supported" },
  { 109, "uL-DPCCH2-operation-not-available" },
  { 110, "uL-DPCCH2-operation-not-supported" },
  { 111, "downlink-TPC-enhancements-operation-not-available" },
  { 112, "downlink-TPC-enhancements-operation-not-supported" },
  { 0, NULL }
};

static value_string_ext rnsap_CauseRadioNetwork_vals_ext = VALUE_STRING_EXT_INIT(rnsap_CauseRadioNetwork_vals);


static int
dissect_rnsap_CauseRadioNetwork(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     27, NULL, TRUE, 86, NULL);

  return offset;
}


static const value_string rnsap_CauseTransport_vals[] = {
  {   0, "transport-resource-unavailable" },
  {   1, "unspecified" },
  { 0, NULL }
};


static int
dissect_rnsap_CauseTransport(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_CauseProtocol_vals[] = {
  {   0, "transfer-syntax-error" },
  {   1, "abstract-syntax-error-reject" },
  {   2, "abstract-syntax-error-ignore-and-notify" },
  {   3, "message-not-compatible-with-receiver-state" },
  {   4, "semantic-error" },
  {   5, "unspecified" },
  {   6, "abstract-syntax-error-falsely-constructed-message" },
  { 0, NULL }
};


static int
dissect_rnsap_CauseProtocol(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_CauseMisc_vals[] = {
  {   0, "control-processing-overload" },
  {   1, "hardware-failure" },
  {   2, "om-intervention" },
  {   3, "not-enough-user-plane-processing-resources" },
  {   4, "unspecified" },
  { 0, NULL }
};


static int
dissect_rnsap_CauseMisc(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_Cause_vals[] = {
  {   0, "radioNetwork" },
  {   1, "transport" },
  {   2, "protocol" },
  {   3, "misc" },
  { 0, NULL }
};

static const per_choice_t Cause_choice[] = {
  {   0, &hf_rnsap_radioNetwork  , ASN1_EXTENSION_ROOT    , dissect_rnsap_CauseRadioNetwork },
  {   1, &hf_rnsap_transport     , ASN1_EXTENSION_ROOT    , dissect_rnsap_CauseTransport },
  {   2, &hf_rnsap_protocol      , ASN1_EXTENSION_ROOT    , dissect_rnsap_CauseProtocol },
  {   3, &hf_rnsap_misc          , ASN1_EXTENSION_ROOT    , dissect_rnsap_CauseMisc },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Cause(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Cause, Cause_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t E_DCH_serving_cell_change_unsuccessful_sequence[] = {
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_serving_cell_change_unsuccessful(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_serving_cell_change_unsuccessful, E_DCH_serving_cell_change_unsuccessful_sequence);

  return offset;
}


static const value_string rnsap_E_DCH_serving_cell_change_choice_vals[] = {
  {   0, "e-DCH-serving-cell-change-successful" },
  {   1, "e-DCH-serving-cell-change-unsuccessful" },
  { 0, NULL }
};

static const per_choice_t E_DCH_serving_cell_change_choice_choice[] = {
  {   0, &hf_rnsap_e_DCH_serving_cell_change_successful, ASN1_EXTENSION_ROOT    , dissect_rnsap_E_DCH_serving_cell_change_successful },
  {   1, &hf_rnsap_e_DCH_serving_cell_change_unsuccessful, ASN1_EXTENSION_ROOT    , dissect_rnsap_E_DCH_serving_cell_change_unsuccessful },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_E_DCH_serving_cell_change_choice(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_E_DCH_serving_cell_change_choice, E_DCH_serving_cell_change_choice_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t E_DCH_Serving_cell_change_informationResponse_sequence[] = {
  { &hf_rnsap_e_DCH_serving_cell_outcome_choice, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_serving_cell_change_choice },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_Serving_cell_change_informationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_Serving_cell_change_informationResponse, E_DCH_Serving_cell_change_informationResponse_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Response_RLAdd_ItemIEs_sequence[] = {
  { &hf_rnsap_additional_EDCH_FDD_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_FDD_Information_Response_ItemIEs },
  { &hf_rnsap_additional_EDCH_Serving_Cell_Change_Information_Response_RLAdd, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Serving_cell_change_informationResponse },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_Response_RLAdd_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_Cell_Information_Response_RLAdd_ItemIEs, Additional_EDCH_Cell_Information_Response_RLAdd_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Response_RLAddList_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_Cell_Information_Response_RLAddList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_Cell_Information_Response_RLAdd_ItemIEs },
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_Response_RLAddList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_Cell_Information_Response_RLAddList, Additional_EDCH_Cell_Information_Response_RLAddList_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const value_string rnsap_Multicell_EDCH_Transport_Bearer_Mode_vals[] = {
  {   0, "separate-Iur-Transport-Bearer-Mode" },
  {   1, "uL-Flow-Multiplexing-Mode" },
  { 0, NULL }
};


static int
dissect_rnsap_Multicell_EDCH_Transport_Bearer_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_UL_ScramblingCodeNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16777215U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_UL_ScramblingCodeLength_vals[] = {
  {   0, "short" },
  {   1, "long" },
  { 0, NULL }
};


static int
dissect_rnsap_UL_ScramblingCodeLength(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t UL_ScramblingCode_sequence[] = {
  { &hf_rnsap_ul_ScramblingCodeNumber, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_ScramblingCodeNumber },
  { &hf_rnsap_ul_ScramblingCodeLength, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_ScramblingCodeLength },
  { &hf_rnsap_iE_Extensions , ASN1_NO_EXTENSIONS     , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_ScramblingCode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_ScramblingCode, UL_ScramblingCode_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_UL_DPCH_Information_Setup_sequence[] = {
  { &hf_rnsap_ul_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_ScramblingCode },
  { &hf_rnsap_ul_SIR_Target , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_UL_DPCH_Information_Setup(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_UL_DPCH_Information_Setup, Additional_EDCH_UL_DPCH_Information_Setup_sequence);

  return offset;
}



static int
dissect_rnsap_C_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_FirstRLS_Indicator_vals[] = {
  {   0, "first-RLS" },
  {   1, "not-first-RLS" },
  { 0, NULL }
};


static int
dissect_rnsap_FirstRLS_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_PropagationDelay(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_PrimaryCPICH_EcNo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -30, 30U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_E_AGCH_PowerOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_E_RGCH_PowerOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_E_HICH_PowerOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, TRUE);

  return offset;
}


static const per_sequence_t Additional_EDCH_MAC_d_Flows_Specific_Info_sequence[] = {
  { &hf_rnsap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info, Additional_EDCH_MAC_d_Flows_Specific_Info_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_MAC_d_Flows_Specific_Info_List_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info },
};

static int
dissect_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_List, Additional_EDCH_MAC_d_Flows_Specific_Info_List_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}



static int
dissect_rnsap_MulticellEDCH_RL_SpecificInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs_sequence[] = {
  { &hf_rnsap_eDCH_Additional_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_C_ID },
  { &hf_rnsap_firstRLS_indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FirstRLS_Indicator },
  { &hf_rnsap_propagationDelay, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PropagationDelay },
  { &hf_rnsap_initialDL_transmissionPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_primaryCPICH_EcNo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PrimaryCPICH_EcNo },
  { &hf_rnsap_e_AGCH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_AGCH_PowerOffset },
  { &hf_rnsap_e_RGCH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_RGCH_PowerOffset },
  { &hf_rnsap_e_HICH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_HICH_PowerOffset },
  { &hf_rnsap_additional_EDCH_MAC_d_Flows_Specific_Info_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_List },
  { &hf_rnsap_multicellEDCH_RL_SpecificInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MulticellEDCH_RL_SpecificInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs, Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_RL_Specific_Information_To_Setup_List_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_RL_Specific_Information_To_Setup_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs },
};

static int
dissect_rnsap_Additional_EDCH_RL_Specific_Information_To_Setup_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Setup_List, Additional_EDCH_RL_Specific_Information_To_Setup_List_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}



static int
dissect_rnsap_E_DCH_Maximum_Bitrate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 5742U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_E_DCH_Processing_Overload_Level(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 10U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_E_TFCI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_FDD_Information_sequence[] = {
  { &hf_rnsap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_rnsap_e_DCH_Maximum_Bitrate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Maximum_Bitrate },
  { &hf_rnsap_e_DCH_Processing_Overload_Level, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Processing_Overload_Level },
  { &hf_rnsap_e_DCH_Min_Set_E_TFCI, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_TFCI },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_FDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_FDD_Information, Additional_EDCH_FDD_Information_sequence);

  return offset;
}


static const value_string rnsap_FDD_TPC_DownlinkStepSize_vals[] = {
  {   0, "step-size0-5" },
  {   1, "step-size1" },
  {   2, "step-size1-5" },
  {   3, "step-size2" },
  { 0, NULL }
};


static int
dissect_rnsap_FDD_TPC_DownlinkStepSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_LimitedPowerIncrease_vals[] = {
  {   0, "used" },
  {   1, "not-used" },
  { 0, NULL }
};


static int
dissect_rnsap_LimitedPowerIncrease(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_InnerLoopDLPCStatus_vals[] = {
  {   0, "active" },
  {   1, "inactive" },
  { 0, NULL }
};


static int
dissect_rnsap_InnerLoopDLPCStatus(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_F_DPCH_SlotFormatSupportRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Additional_EDCH_F_DPCH_Information_sequence[] = {
  { &hf_rnsap_fdd_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_TPC_DownlinkStepSize },
  { &hf_rnsap_limitedPowerIncrease, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_LimitedPowerIncrease },
  { &hf_rnsap_innerLoopDLPCStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_InnerLoopDLPCStatus },
  { &hf_rnsap_f_DPCH_SlotFormatSupportRequest, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_F_DPCH_SlotFormatSupportRequest },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_F_DPCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_F_DPCH_Information, Additional_EDCH_F_DPCH_Information_sequence);

  return offset;
}



static int
dissect_rnsap_MulticellEDCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Additional_EDCH_FDD_Setup_Cell_Information_sequence[] = {
  { &hf_rnsap_additional_EDCH_UL_DPCH_Information_Setup, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_UL_DPCH_Information_Setup },
  { &hf_rnsap_additional_EDCH_RL_Specific_Information_To_Setup, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_RL_Specific_Information_To_Setup_List },
  { &hf_rnsap_additional_EDCH_FDD_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_FDD_Information },
  { &hf_rnsap_additional_EDCH_F_DPCH_Information_Setup, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_F_DPCH_Information },
  { &hf_rnsap_multicellEDCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MulticellEDCH_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_FDD_Setup_Cell_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_FDD_Setup_Cell_Information, Additional_EDCH_FDD_Setup_Cell_Information_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Setup_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_Cell_Information_Setup_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_FDD_Setup_Cell_Information },
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_Setup(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_Cell_Information_Setup, Additional_EDCH_Cell_Information_Setup_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_Setup_Info_sequence[] = {
  { &hf_rnsap_multicell_EDCH_Transport_Bearer_Mode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Multicell_EDCH_Transport_Bearer_Mode },
  { &hf_rnsap_additional_EDCH_Cell_Information_Setup, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_Cell_Information_Setup },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_Setup_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_Setup_Info, Additional_EDCH_Setup_Info_sequence);

  return offset;
}



static int
dissect_rnsap_L3_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
	tvbuff_t *parameter_tvb;
	dissector_handle_t parameter_handle = NULL;
	rnsap_private_data_t *pdata = rnsap_get_private_data(actx->pinfo);

  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     NO_BOUND, NO_BOUND, FALSE, NULL, 0, &parameter_tvb, NULL);


	if (!parameter_tvb)
		return offset;

	switch (pdata->ProcedureCode) {

	case RNSAP_ID_DOWNLINKSIGNALLINGTRANSFER:
		/* TODO: seperate into Iur and Iur-g cases: */
		/* For the Iur-g interface, L3 message is a GERAN-RRC message for which a dissector does not currently exist */
		/* For the Iur interface, L3 message is a UMTS RRC DL-CCCH message */
		parameter_handle = rrc_dl_ccch_handle;
		break;
	case RNSAP_ID_UPLINKSIGNALLINGTRANSFER:
		parameter_handle = rrc_ul_ccch_handle;
		break;
	default:
		break;
	}

    if (parameter_handle)
      call_dissector(parameter_handle, parameter_tvb, actx->pinfo, tree);


  return offset;
}


static const per_sequence_t ANRReportIndication_sequence[] = {
  { &hf_rnsap_pLMN_Identity , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PLMN_Identity },
  { &hf_rnsap_l3_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_L3_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ANRReportIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ANRReportIndication, ANRReportIndication_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_UL_DPCH_Information_Addition_sequence[] = {
  { &hf_rnsap_ul_SIR_Target , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_UL_DPCH_Information_Addition(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_UL_DPCH_Information_Addition, Additional_EDCH_UL_DPCH_Information_Addition_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs_sequence[] = {
  { &hf_rnsap_eDCH_Additional_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_primaryCPICH_EcNo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PrimaryCPICH_EcNo },
  { &hf_rnsap_e_AGCH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_AGCH_PowerOffset },
  { &hf_rnsap_e_RGCH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_RGCH_PowerOffset },
  { &hf_rnsap_e_HICH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_HICH_PowerOffset },
  { &hf_rnsap_additional_EDCH_MAC_d_Flows_Specific_Info_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_List },
  { &hf_rnsap_multicellEDCH_RL_SpecificInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MulticellEDCH_RL_SpecificInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs, Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_RL_Specific_Information_To_Add_List_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs },
};

static int
dissect_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_List, Additional_EDCH_RL_Specific_Information_To_Add_List_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_To_Add_ItemIEs_sequence[] = {
  { &hf_rnsap_additional_EDCH_UL_DPCH_Information_Setup_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_UL_DPCH_Information_Addition },
  { &hf_rnsap_additional_EDCH_RL_Specific_Information_To_Add_List, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_List },
  { &hf_rnsap_additional_EDCH_FDD_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_FDD_Information },
  { &hf_rnsap_multicellEDCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MulticellEDCH_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_To_Add_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_Cell_Information_To_Add_ItemIEs, Additional_EDCH_Cell_Information_To_Add_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_To_Add_List_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_Cell_Information_To_Add_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_Cell_Information_To_Add_ItemIEs },
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_To_Add_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_Cell_Information_To_Add_List, Additional_EDCH_Cell_Information_To_Add_List_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs_sequence[] = {
  { &hf_rnsap_eDCH_Additional_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_e_AGCH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_AGCH_PowerOffset },
  { &hf_rnsap_e_RGCH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_RGCH_PowerOffset },
  { &hf_rnsap_e_HICH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_HICH_PowerOffset },
  { &hf_rnsap_additional_EDCH_MAC_d_Flows_Specific_Info_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_List },
  { &hf_rnsap_multicellEDCH_RL_SpecificInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MulticellEDCH_RL_SpecificInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs, Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_RL_Specific_Information_To_Modify_List_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_RL_Specific_Information_To_Modify_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs },
};

static int
dissect_rnsap_Additional_EDCH_RL_Specific_Information_To_Modify_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Modify_List, Additional_EDCH_RL_Specific_Information_To_Modify_List_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}


static const value_string rnsap_UE_DTX_Cycle1_2ms_vals[] = {
  {   0, "v1" },
  {   1, "v4" },
  {   2, "v5" },
  {   3, "v8" },
  {   4, "v10" },
  {   5, "v16" },
  {   6, "v20" },
  { 0, NULL }
};


static int
dissect_rnsap_UE_DTX_Cycle1_2ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_UE_DTX_Cycle2_ext_2ms_vals[] = {
  {   0, "v4" },
  {   1, "v5" },
  {   2, "v8" },
  {   3, "v10" },
  {   4, "v16" },
  {   5, "v20" },
  {   6, "v32" },
  {   7, "v40" },
  {   8, "v64" },
  {   9, "v80" },
  {  10, "v128" },
  {  11, "v160" },
  {  12, "v256" },
  {  13, "v320" },
  {  14, "v512" },
  {  15, "v640" },
  {  16, "v1024" },
  {  17, "v1280" },
  { 0, NULL }
};


static int
dissect_rnsap_UE_DTX_Cycle2_ext_2ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     18, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_Inactivity_Threshold_for_UE_DTX_Cycle2_vals[] = {
  {   0, "v1" },
  {   1, "v4" },
  {   2, "v8" },
  {   3, "v16" },
  {   4, "v32" },
  {   5, "v64" },
  {   6, "v128" },
  {   7, "v256" },
  { 0, NULL }
};


static int
dissect_rnsap_Inactivity_Threshold_for_UE_DTX_Cycle2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DTX_Information2_sequence[] = {
  { &hf_rnsap_uE_DTX_Cycle1 , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DTX_Cycle1_2ms },
  { &hf_rnsap_uE_DTX_Cycle2 , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DTX_Cycle2_ext_2ms },
  { &hf_rnsap_inactivity_Threshold_for_UE_DTX_Cycle2, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Inactivity_Threshold_for_UE_DTX_Cycle2 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DTX_Information2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DTX_Information2, DTX_Information2_sequence);

  return offset;
}


static const value_string rnsap_Implicit_Grant_Handling_vals[] = {
  {   0, "true" },
  { 0, NULL }
};


static int
dissect_rnsap_Implicit_Grant_Handling(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_Minimum_TEBS_threshold_vals[] = {
  {   0, "v2" },
  {   1, "v4" },
  {   2, "v8" },
  {   3, "v16" },
  {   4, "v32" },
  {   5, "v64" },
  {   6, "v128" },
  {   7, "v256" },
  {   8, "v512" },
  {   9, "v1024" },
  {  10, "v2048" },
  {  11, "v4096" },
  {  12, "v8192" },
  {  13, "v16384" },
  {  14, "v32768" },
  {  15, "v65536" },
  {  16, "v131072" },
  {  17, "v262144" },
  {  18, "v524288" },
  {  19, "v1048576" },
  { 0, NULL }
};


static int
dissect_rnsap_Minimum_TEBS_threshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     20, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Response_List_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_Cell_Information_Response_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_FDD_Information_Response_ItemIEs },
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_Response_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_Cell_Information_Response_List, Additional_EDCH_Cell_Information_Response_List_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const value_string rnsap_DL_PowerBalancing_UpdatedIndicator_vals[] = {
  {   0, "dL-PowerBalancing-Updated" },
  { 0, NULL }
};


static int
dissect_rnsap_DL_PowerBalancing_UpdatedIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_NULL(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_Secondary_CPICH_Information_Change_vals[] = {
  {   0, "new-secondary-CPICH" },
  {   1, "secondary-CPICH-shall-not-be-used" },
  { 0, NULL }
};

static const per_choice_t Secondary_CPICH_Information_Change_choice[] = {
  {   0, &hf_rnsap_new_secondary_CPICH, ASN1_EXTENSION_ROOT    , dissect_rnsap_Secondary_CPICH_Information },
  {   1, &hf_rnsap_secondary_CPICH_shall_not_be_used, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Secondary_CPICH_Information_Change(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Secondary_CPICH_Information_Change, Secondary_CPICH_Information_Change_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t EDCH_Additional_RL_Specific_Modified_Information_Response_ItemIEs_sequence[] = {
  { &hf_rnsap_eDCH_Additional_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_dL_PowerBalancing_UpdatedIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_PowerBalancing_UpdatedIndicator },
  { &hf_rnsap_eDCH_FDD_DL_ControlChannelInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_EDCH_FDD_DL_ControlChannelInformation },
  { &hf_rnsap_additional_EDCH_MAC_d_Flow_Specific_Information_Response_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List },
  { &hf_rnsap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_rnsap_maxUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_minUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_maximumDL_power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDL_power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_primary_CPICH_Usage_For_Channel_Estimation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Primary_CPICH_Usage_For_Channel_Estimation },
  { &hf_rnsap_secondary_CPICH_Information_Change, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Secondary_CPICH_Information_Change },
  { &hf_rnsap_f_DPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_F_DPCH_SlotFormat },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_Additional_RL_Specific_Modified_Information_Response_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_Additional_RL_Specific_Modified_Information_Response_ItemIEs, EDCH_Additional_RL_Specific_Modified_Information_Response_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t EDCH_Additional_RL_Specific_Modified_Information_Response_List_sequence_of[1] = {
  { &hf_rnsap_EDCH_Additional_RL_Specific_Modified_Information_Response_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_Additional_RL_Specific_Modified_Information_Response_ItemIEs },
};

static int
dissect_rnsap_EDCH_Additional_RL_Specific_Modified_Information_Response_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_EDCH_Additional_RL_Specific_Modified_Information_Response_List, EDCH_Additional_RL_Specific_Modified_Information_Response_List_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_FDD_Information_Response_RLReconf_Items_sequence[] = {
  { &hf_rnsap_eDCH_Additional_RL_Specific_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_EDCH_Additional_RL_Specific_Information_Response_List },
  { &hf_rnsap_eDCH_Additional_RL_Specific_Modified_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_EDCH_Additional_RL_Specific_Modified_Information_Response_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_FDD_Information_Response_RLReconf_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_FDD_Information_Response_RLReconf_Items, Additional_EDCH_FDD_Information_Response_RLReconf_Items_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Response_RLReconf_List_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_Cell_Information_Response_RLReconf_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_FDD_Information_Response_RLReconf_Items },
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_Response_RLReconf_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_Cell_Information_Response_RLReconf_List, Additional_EDCH_Cell_Information_Response_RLReconf_List_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_UL_DPCH_Information_Modify_sequence[] = {
  { &hf_rnsap_ul_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_ScramblingCode },
  { &hf_rnsap_ul_SIR_Target , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_UL_DPCH_Information_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_UL_DPCH_Information_Modify, Additional_EDCH_UL_DPCH_Information_Modify_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_ConfigurationChange_Info_ItemIEs_sequence[] = {
  { &hf_rnsap_additional_EDCH_UL_DPCH_Information_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_UL_DPCH_Information_Modify },
  { &hf_rnsap_additional_EDCH_RL_Specific_Information_To_Add, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_List },
  { &hf_rnsap_additional_EDCH_RL_Specific_Information_To_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_RL_Specific_Information_To_Modify_List },
  { &hf_rnsap_additional_EDCH_FDD_Information_To_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_FDD_Information },
  { &hf_rnsap_additional_EDCH_F_DPCH_Information_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_F_DPCH_Information },
  { &hf_rnsap_multicellEDCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MulticellEDCH_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_ConfigurationChange_Info_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_ConfigurationChange_Info_ItemIEs, Additional_EDCH_ConfigurationChange_Info_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_ConfigurationChange_List_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_Cell_Information_ConfigurationChange_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_ConfigurationChange_Info_ItemIEs },
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_ConfigurationChange_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_Cell_Information_ConfigurationChange_List, Additional_EDCH_Cell_Information_ConfigurationChange_List_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const value_string rnsap_RL_on_Secondary_UL_Frequency_vals[] = {
  {   0, "remove" },
  { 0, NULL }
};


static int
dissect_rnsap_RL_on_Secondary_UL_Frequency(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Removal_Info_ItemIEs_sequence[] = {
  { &hf_rnsap_rL_on_Secondary_UL_Frequency, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_on_Secondary_UL_Frequency },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_Removal_Info_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_Cell_Information_Removal_Info_ItemIEs, Additional_EDCH_Cell_Information_Removal_Info_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_Removal_List_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_Cell_Information_Removal_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_Cell_Information_Removal_Info_ItemIEs },
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_Removal_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_Cell_Information_Removal_List, Additional_EDCH_Cell_Information_Removal_List_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs_sequence[] = {
  { &hf_rnsap_eDCH_Additional_RL_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs, Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_DL_Control_Channel_Change_Information_List_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_DL_Control_Channel_Change_Information_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs },
};

static int
dissect_rnsap_Additional_EDCH_DL_Control_Channel_Change_Information_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_DL_Control_Channel_Change_Information_List, Additional_EDCH_DL_Control_Channel_Change_Information_List_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_FDD_Update_Information_sequence[] = {
  { &hf_rnsap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_rnsap_additional_EDCH_DL_Control_Channel_Change_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Additional_EDCH_DL_Control_Channel_Change_Information_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_FDD_Update_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_FDD_Update_Information, Additional_EDCH_FDD_Update_Information_sequence);

  return offset;
}


static const per_sequence_t CorrespondingCells_sequence_of[1] = {
  { &hf_rnsap_CorrespondingCells_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
};

static int
dissect_rnsap_CorrespondingCells(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_CorrespondingCells, CorrespondingCells_sequence_of,
                                                  1, maxNrOfCellsPerFreq, FALSE);

  return offset;
}


static const per_sequence_t AdditionalPreferredFrequencyItem_sequence[] = {
  { &hf_rnsap_dL_UARFCN     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_correspondingCells, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CorrespondingCells },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_AdditionalPreferredFrequencyItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_AdditionalPreferredFrequencyItem, AdditionalPreferredFrequencyItem_sequence);

  return offset;
}


static const per_sequence_t AdditionalPreferredFrequency_sequence_of[1] = {
  { &hf_rnsap_AdditionalPreferredFrequency_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_AdditionalPreferredFrequencyItem },
};

static int
dissect_rnsap_AdditionalPreferredFrequency(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_AdditionalPreferredFrequency, AdditionalPreferredFrequency_sequence_of,
                                                  1, maxNrOfAddFreq, FALSE);

  return offset;
}



static int
dissect_rnsap_AdjustmentPeriod(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_S_RNTI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1048575U, NULL, FALSE);

  return offset;
}


static const per_sequence_t AffectedUEInformationForMBMS_sequence_of[1] = {
  { &hf_rnsap_AffectedUEInformationForMBMS_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_S_RNTI },
};

static int
dissect_rnsap_AffectedUEInformationForMBMS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_AffectedUEInformationForMBMS, AffectedUEInformationForMBMS_sequence_of,
                                                  1, maxNrOfUEs, FALSE);

  return offset;
}



static int
dissect_rnsap_PriorityLevel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_Pre_emptionCapability_vals[] = {
  {   0, "shall-not-trigger-pre-emption" },
  {   1, "may-trigger-pre-emption" },
  { 0, NULL }
};


static int
dissect_rnsap_Pre_emptionCapability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_Pre_emptionVulnerability_vals[] = {
  {   0, "not-pre-emptable" },
  {   1, "pre-emptable" },
  { 0, NULL }
};


static int
dissect_rnsap_Pre_emptionVulnerability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t AllocationRetentionPriority_sequence[] = {
  { &hf_rnsap_priorityLevel , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityLevel },
  { &hf_rnsap_pre_emptionCapability, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Pre_emptionCapability },
  { &hf_rnsap_pre_emptionVulnerability, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Pre_emptionVulnerability },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_AllocationRetentionPriority(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_AllocationRetentionPriority, AllocationRetentionPriority_sequence);

  return offset;
}



static int
dissect_rnsap_Allowed_Rate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNrOfTFs, NULL, FALSE);

  return offset;
}


static const per_sequence_t Allowed_Rate_Information_sequence[] = {
  { &hf_rnsap_allowed_UL_Rate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Allowed_Rate },
  { &hf_rnsap_allowed_DL_Rate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Allowed_Rate },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Allowed_Rate_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Allowed_Rate_Information, Allowed_Rate_Information_sequence);

  return offset;
}



static int
dissect_rnsap_AllowedQueuingTime(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 60U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_AlphaValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_AlternativeFormatReportingIndicator_vals[] = {
  {   0, "alternativeFormatAllowed" },
  { 0, NULL }
};


static int
dissect_rnsap_AlternativeFormatReportingIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_AOA_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 719U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_AOA_LCR_Accuracy_Class_vals[] = {
  {   0, "a" },
  {   1, "b" },
  {   2, "c" },
  {   3, "d" },
  {   4, "e" },
  {   5, "f" },
  {   6, "g" },
  {   7, "h" },
  { 0, NULL }
};


static int
dissect_rnsap_AOA_LCR_Accuracy_Class(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Angle_Of_Arrival_Value_LCR_sequence[] = {
  { &hf_rnsap_aOA_LCR       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AOA_LCR },
  { &hf_rnsap_aOA_LCR_Accuracy_Class, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AOA_LCR_Accuracy_Class },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Angle_Of_Arrival_Value_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Angle_Of_Arrival_Value_LCR, Angle_Of_Arrival_Value_LCR_sequence);

  return offset;
}



static int
dissect_rnsap_RNC_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4095U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_LAC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  tvbuff_t *parameter_tvb = NULL;
  offset = dissect_per_octet_string(tvb, offset, actx, tree, -1,
                                       2, 2, FALSE, &parameter_tvb);

  if (parameter_tvb) {
    actx->created_item = proto_tree_add_item(tree, hf_index, parameter_tvb, 0, 2, ENC_BIG_ENDIAN);
  }


  return offset;
}



static int
dissect_rnsap_RAC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  tvbuff_t *parameter_tvb = NULL;
  offset = dissect_per_octet_string(tvb, offset, actx, tree, -1,
                                       1, 1, FALSE, &parameter_tvb);

  if (parameter_tvb) {
    actx->created_item = proto_tree_add_item(tree, hf_index, parameter_tvb, 0, 1, ENC_BIG_ENDIAN);
  }


  return offset;
}


static const value_string rnsap_TxDiversityIndicator_vals[] = {
  {   0, "true" },
  {   1, "false" },
  { 0, NULL }
};


static int
dissect_rnsap_TxDiversityIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_STTD_SupportIndicator_vals[] = {
  {   0, "sTTD-Supported" },
  {   1, "sTTD-not-Supported" },
  { 0, NULL }
};


static int
dissect_rnsap_STTD_SupportIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_ClosedLoopMode1_SupportIndicator_vals[] = {
  {   0, "closedLoop-Mode1-Supported" },
  {   1, "closedLoop-Mode1-not-Supported" },
  { 0, NULL }
};


static int
dissect_rnsap_ClosedLoopMode1_SupportIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_RestrictionStateIndicator_vals[] = {
  {   0, "cellNotResevedForOperatorUse" },
  {   1, "cellResevedForOperatorUse" },
  { 0, NULL }
};


static int
dissect_rnsap_RestrictionStateIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_DPC_Mode_Change_SupportIndicator_vals[] = {
  {   0, "dPC-ModeChangeSupported" },
  { 0, NULL }
};


static int
dissect_rnsap_DPC_Mode_Change_SupportIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_CellCapabilityContainer_FDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     32, 32, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_SNACode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}


static const per_sequence_t ListOfSNAs_sequence_of[1] = {
  { &hf_rnsap_ListOfSNAs_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_SNACode },
};

static int
dissect_rnsap_ListOfSNAs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_ListOfSNAs, ListOfSNAs_sequence_of,
                                                  1, maxNrOfSNAs, FALSE);

  return offset;
}


static const per_sequence_t SNA_Information_sequence[] = {
  { &hf_rnsap_pLMN_Identity , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PLMN_Identity },
  { &hf_rnsap_listOfSNAs    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ListOfSNAs },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SNA_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SNA_Information, SNA_Information_sequence);

  return offset;
}


static const value_string rnsap_FrequencyBandIndicator_vals[] = {
  {   0, "bandI" },
  {   1, "bandII" },
  {   2, "bandIII" },
  {   3, "bandIV" },
  {   4, "bandV" },
  {   5, "bandVI" },
  {   6, "bandVII" },
  {   7, "bandVIII" },
  {   8, "bandIX" },
  {   9, "bandX" },
  {  10, "bandXI" },
  {  11, "bandXII" },
  {  12, "bandXIII" },
  {  13, "bandXIV" },
  {  14, "bandXV" },
  {  15, "bandXVI" },
  {  16, "bandXVII" },
  {  17, "bandXVIII" },
  {  18, "bandXIX" },
  {  19, "bandXX" },
  {  20, "bandXXI" },
  {  21, "bandXXII" },
  {  22, "reserved23" },
  {  23, "reserved24" },
  {  24, "bandXXV" },
  {  25, "bandXXVI" },
  {  26, "reserved27" },
  {  27, "reserved28" },
  {  28, "reserved29" },
  {  29, "reserved30" },
  {  30, "reserved31" },
  {  31, "bandXXXII" },
  { 0, NULL }
};

static value_string_ext rnsap_FrequencyBandIndicator_vals_ext = VALUE_STRING_EXT_INIT(rnsap_FrequencyBandIndicator_vals);


static int
dissect_rnsap_FrequencyBandIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     22, NULL, TRUE, 10, NULL);

  return offset;
}


static const value_string rnsap_Max_UE_DTX_Cycle_vals[] = {
  {   0, "v5" },
  {   1, "v10" },
  {   2, "v20" },
  {   3, "v40" },
  {   4, "v64" },
  {   5, "v80" },
  {   6, "v128" },
  {   7, "v160" },
  {   8, "v256" },
  {   9, "v320" },
  {  10, "v512" },
  {  11, "v640" },
  {  12, "v1024" },
  {  13, "v1280" },
  { 0, NULL }
};


static int
dissect_rnsap_Max_UE_DTX_Cycle(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 6, NULL);

  return offset;
}


static const per_sequence_t List_Of_PLMNs_sequence_of[1] = {
  { &hf_rnsap_List_Of_PLMNs_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_PLMN_Identity },
};

static int
dissect_rnsap_List_Of_PLMNs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_List_Of_PLMNs, List_Of_PLMNs_sequence_of,
                                                  1, maxNrOfBroadcastPLMNs, FALSE);

  return offset;
}


static const per_sequence_t ANR_Multiple_PLMN_List_sequence[] = {
  { &hf_rnsap_list_Of_PLMNs , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_List_Of_PLMNs },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ANR_Multiple_PLMN_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ANR_Multiple_PLMN_List, ANR_Multiple_PLMN_List_sequence);

  return offset;
}


static const per_sequence_t Possible_Secondary_Serving_Cell_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Possible_Secondary_Serving_Cell(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Possible_Secondary_Serving_Cell, Possible_Secondary_Serving_Cell_sequence);

  return offset;
}


static const per_sequence_t Possible_Secondary_Serving_Cell_List_sequence_of[1] = {
  { &hf_rnsap_Possible_Secondary_Serving_Cell_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Possible_Secondary_Serving_Cell },
};

static int
dissect_rnsap_Possible_Secondary_Serving_Cell_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Possible_Secondary_Serving_Cell_List, Possible_Secondary_Serving_Cell_List_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t Secondary_Serving_Cell_List_sequence[] = {
  { &hf_rnsap_possible_Secondary_Serving_Cell_List, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Possible_Secondary_Serving_Cell_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Secondary_Serving_Cell_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Secondary_Serving_Cell_List, Secondary_Serving_Cell_List_sequence);

  return offset;
}



static int
dissect_rnsap_CellCapabilityContainerExtension_FDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     128, 128, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t ANR_FDD_CellInformation_sequence[] = {
  { &hf_rnsap_primaryCPICH_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PrimaryCPICH_Power },
  { &hf_rnsap_txDiversityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TxDiversityIndicator },
  { &hf_rnsap_sTTD_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_STTD_SupportIndicator },
  { &hf_rnsap_closedLoopMode1_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ClosedLoopMode1_SupportIndicator },
  { &hf_rnsap_restrictionStateIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RestrictionStateIndicator },
  { &hf_rnsap_dpc_ModeChangeSupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DPC_Mode_Change_SupportIndicator },
  { &hf_rnsap_cellCapabilityContainerFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CellCapabilityContainer_FDD },
  { &hf_rnsap_sNA_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SNA_Information },
  { &hf_rnsap_frequencyBandIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_FrequencyBandIndicator },
  { &hf_rnsap_max_UE_DTX_Cycle, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Max_UE_DTX_Cycle },
  { &hf_rnsap_aNR_Multiple_PLMN_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ANR_Multiple_PLMN_List },
  { &hf_rnsap_secondaryServingCellList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Secondary_Serving_Cell_List },
  { &hf_rnsap_dualBandSecondaryServingCellList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Secondary_Serving_Cell_List },
  { &hf_rnsap_cellCapabilityContainerExtensionFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CellCapabilityContainerExtension_FDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ANR_FDD_CellInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ANR_FDD_CellInformation, ANR_FDD_CellInformation_sequence);

  return offset;
}



static int
dissect_rnsap_SyncCase(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 2U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_TimeSlot(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 14U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_SCH_TimeSlot(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 6U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_SCTD_Indicator_vals[] = {
  {   0, "active" },
  {   1, "inactive" },
  { 0, NULL }
};


static int
dissect_rnsap_SCTD_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_DPCHConstantValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -10, 10U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_PCCPCH_Power(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -150, 400U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_CellCapabilityContainer_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     32, 32, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_CellCapabilityContainer_TDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     32, 32, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t ANR_TDD_CellInformation_sequence[] = {
  { &hf_rnsap_syncCase      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SyncCase },
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TimeSlot },
  { &hf_rnsap_sCH_TimeSlot  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SCH_TimeSlot },
  { &hf_rnsap_sCTD_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SCTD_Indicator },
  { &hf_rnsap_dPCHConstantValue, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DPCHConstantValue },
  { &hf_rnsap_pCCPCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PCCPCH_Power },
  { &hf_rnsap_restrictionStateIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RestrictionStateIndicator },
  { &hf_rnsap_cellCapabilityContainerTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CellCapabilityContainer_TDD },
  { &hf_rnsap_cellCapabilityContainerTDD768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CellCapabilityContainer_TDD768 },
  { &hf_rnsap_sNA_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SNA_Information },
  { &hf_rnsap_aNR_Multiple_PLMN_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ANR_Multiple_PLMN_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ANR_TDD_CellInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ANR_TDD_CellInformation, ANR_TDD_CellInformation_sequence);

  return offset;
}



static int
dissect_rnsap_CellCapabilityContainer_TDD_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     32, 32, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_CellCapabilityContainerExtension_TDD_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     32, 32, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t ANR_LCR_TDD_CellInformation_sequence[] = {
  { &hf_rnsap_sCTD_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SCTD_Indicator },
  { &hf_rnsap_dPCHConstantValue, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DPCHConstantValue },
  { &hf_rnsap_pCCPCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PCCPCH_Power },
  { &hf_rnsap_restrictionStateIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RestrictionStateIndicator },
  { &hf_rnsap_cellCapabilityContainerTDD_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CellCapabilityContainer_TDD_LCR },
  { &hf_rnsap_sNA_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SNA_Information },
  { &hf_rnsap_aNR_Multiple_PLMN_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ANR_Multiple_PLMN_List },
  { &hf_rnsap_cellCapabilityContainerExternsionTDD_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CellCapabilityContainerExtension_TDD_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ANR_LCR_TDD_CellInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ANR_LCR_TDD_CellInformation, ANR_LCR_TDD_CellInformation_sequence);

  return offset;
}



static int
dissect_rnsap_Extended_RNC_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            4096U, 65535U, NULL, FALSE);

  return offset;
}


static const per_sequence_t ANR_Cell_Information_sequence[] = {
  { &hf_rnsap_rNC_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RNC_ID },
  { &hf_rnsap_lAC           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_LAC },
  { &hf_rnsap_rAC           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RAC },
  { &hf_rnsap_aNR_FDD_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ANR_FDD_CellInformation },
  { &hf_rnsap_aNR_TDD_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ANR_TDD_CellInformation },
  { &hf_rnsap_aNR_LCR_TDD_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ANR_LCR_TDD_CellInformation },
  { &hf_rnsap_extended_RNC_ID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Extended_RNC_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ANR_Cell_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ANR_Cell_Information, ANR_Cell_Information_sequence);

  return offset;
}


static const value_string rnsap_AntennaColocationIndicator_vals[] = {
  {   0, "co-located" },
  { 0, NULL }
};


static int
dissect_rnsap_AntennaColocationIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_SAT_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}


static const per_sequence_t T_badSatelliteInformation_item_sequence[] = {
  { &hf_rnsap_badSAT_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAT_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_T_badSatelliteInformation_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_T_badSatelliteInformation_item, T_badSatelliteInformation_item_sequence);

  return offset;
}


static const per_sequence_t T_badSatelliteInformation_sequence_of[1] = {
  { &hf_rnsap_badSatelliteInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_T_badSatelliteInformation_item },
};

static int
dissect_rnsap_T_badSatelliteInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_T_badSatelliteInformation, T_badSatelliteInformation_sequence_of,
                                                  1, maxNoSat, FALSE);

  return offset;
}


static const per_sequence_t BadSatellites_sequence[] = {
  { &hf_rnsap_badSatelliteInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_badSatelliteInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_BadSatellites(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_BadSatellites, BadSatellites_sequence);

  return offset;
}


static const value_string rnsap_Band_Indicator_vals[] = {
  {   0, "dcs1800Band" },
  {   1, "pcs1900Band" },
  { 0, NULL }
};


static int
dissect_rnsap_Band_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_BCC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     3, 3, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BCCH_ARFCN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1023U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_320(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 320U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_9(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     9, 9, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_4(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     4, 4, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t BDS_IGPInfo_sequence[] = {
  { &hf_rnsap_bds_IGPNumber , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_320 },
  { &hf_rnsap_bds_VerticalDelay, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_9 },
  { &hf_rnsap_bds_GIVEI     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_4 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_BDS_IGPInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_BDS_IGPInfo, BDS_IGPInfo_sequence);

  return offset;
}


static const per_sequence_t BDS_IGPInfoList_sequence_of[1] = {
  { &hf_rnsap_BDS_IGPInfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_BDS_IGPInfo },
};

static int
dissect_rnsap_BDS_IGPInfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_BDS_IGPInfoList, BDS_IGPInfoList_sequence_of,
                                                  1, maxIGPInfo, FALSE);

  return offset;
}


static const value_string rnsap_BDS_IonosphericGridModelReq_vals[] = {
  {   0, "requested" },
  { 0, NULL }
};


static int
dissect_rnsap_BDS_IonosphericGridModelReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_119(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 119U, NULL, FALSE);

  return offset;
}


static const per_sequence_t BDS_Ionospheric_Grid_Model_sequence[] = {
  { &hf_rnsap_bDS_RefTime   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_119 },
  { &hf_rnsap_bds_IGPInfoList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BDS_IGPInfoList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_BDS_Ionospheric_Grid_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_BDS_Ionospheric_Grid_Model, BDS_Ionospheric_Grid_Model_sequence);

  return offset;
}



static int
dissect_rnsap_BetaCD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_BLER(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -63, 0U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_NCC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     3, 3, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t BSIC_sequence[] = {
  { &hf_rnsap_nCC           , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_NCC },
  { &hf_rnsap_bCC           , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_BCC },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_BSIC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_BSIC, BSIC_sequence);

  return offset;
}


static const value_string rnsap_BundlingModeIndicator_vals[] = {
  {   0, "bundling" },
  {   1, "no-bundling" },
  { 0, NULL }
};


static int
dissect_rnsap_BundlingModeIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_15(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_10_25(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            10U, 25U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_16(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 16U, NULL, FALSE);

  return offset;
}


static const per_sequence_t BurstModeParameters_sequence[] = {
  { &hf_rnsap_burstStart    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_15 },
  { &hf_rnsap_burstLength   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_10_25 },
  { &hf_rnsap_burstFreq     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_16 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_BurstModeParameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_BurstModeParameters, BurstModeParameters_sequence);

  return offset;
}



static int
dissect_rnsap_Cell_Id(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 268435455U, NULL, FALSE);

  return offset;
}


static const per_sequence_t CellIdList_sequence_of[1] = {
  { &hf_rnsap_CellIdList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Cell_Id },
};

static int
dissect_rnsap_CellIdList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_CellIdList, CellIdList_sequence_of,
                                                  1, maxNrOfCellIds, FALSE);

  return offset;
}


static const per_sequence_t CellBased_sequence[] = {
  { &hf_rnsap_cellIdList    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CellIdList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CellBased(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CellBased, CellBased_sequence);

  return offset;
}



static int
dissect_rnsap_CCTrCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_100_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 100U, NULL, TRUE);

  return offset;
}


static const per_sequence_t Cell_Capacity_Class_Value_sequence[] = {
  { &hf_rnsap_uplinkCellCapacityClassValue, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_100_ },
  { &hf_rnsap_downlinkCellCapacityClassValue, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_100_ },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Cell_Capacity_Class_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Cell_Capacity_Class_Value, Cell_Capacity_Class_Value_sequence);

  return offset;
}



static int
dissect_rnsap_CellIndividualOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -20, 20U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_CellListValidityIndicator_vals[] = {
  {   0, "ignoreSecondaryServingCellList" },
  {   1, "ignoreDualBandSecondaryServingCellList" },
  {   2, "ignoreBoth" },
  { 0, NULL }
};


static int
dissect_rnsap_CellListValidityIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_CellParameterID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_CellPortionID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_CellPortionLCRID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, TRUE);

  return offset;
}


static const per_sequence_t LAI_sequence[] = {
  { &hf_rnsap_pLMN_Identity , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PLMN_Identity },
  { &hf_rnsap_lAC           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_LAC },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_LAI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_LAI, LAI_sequence);

  return offset;
}



static int
dissect_rnsap_CI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       2, 2, FALSE, NULL);

  return offset;
}


static const per_sequence_t CGI_sequence[] = {
  { &hf_rnsap_lAI           , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_LAI },
  { &hf_rnsap_cI            , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_CI },
  { &hf_rnsap_iE_Extensions , ASN1_NO_EXTENSIONS     , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CGI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CGI, CGI_sequence);

  return offset;
}


static const value_string rnsap_ChannelCodingType_vals[] = {
  {   0, "no-codingTDD" },
  {   1, "convolutional-coding" },
  {   2, "turbo-coding" },
  { 0, NULL }
};


static int
dissect_rnsap_ChannelCodingType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_ChipOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 38399U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_ClassmarkInformation2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       NO_BOUND, NO_BOUND, FALSE, NULL);

  return offset;
}



static int
dissect_rnsap_ClassmarkInformation3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       NO_BOUND, NO_BOUND, FALSE, NULL);

  return offset;
}


static const value_string rnsap_Closedlooptimingadjustmentmode_vals[] = {
  {   0, "adj-1-slot" },
  {   1, "adj-2-slot" },
  { 0, NULL }
};


static int
dissect_rnsap_Closedlooptimingadjustmentmode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_CodingRate_vals[] = {
  {   0, "half" },
  {   1, "third" },
  { 0, NULL }
};


static int
dissect_rnsap_CodingRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_TUTRANGPSAccuracyClass_vals[] = {
  {   0, "accuracy-class-A" },
  {   1, "accuracy-class-B" },
  {   2, "accuracy-class-C" },
  { 0, NULL }
};


static int
dissect_rnsap_TUTRANGPSAccuracyClass(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_TUTRANGANSSAccuracyClass_vals[] = {
  {   0, "ganssAccuracy-class-A" },
  {   1, "ganssAccuracy-class-B" },
  {   2, "ganssAccuracy-class-C" },
  { 0, NULL }
};


static int
dissect_rnsap_TUTRANGANSSAccuracyClass(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_CommonMeasurementAccuracy_vals[] = {
  {   0, "tUTRANGPSMeasurementAccuracyClass" },
  {   1, "tUTRANGANSSMeasurementAccuracyClass" },
  { 0, NULL }
};

static const per_choice_t CommonMeasurementAccuracy_choice[] = {
  {   0, &hf_rnsap_tUTRANGPSMeasurementAccuracyClass, ASN1_EXTENSION_ROOT    , dissect_rnsap_TUTRANGPSAccuracyClass },
  {   1, &hf_rnsap_tUTRANGANSSMeasurementAccuracyClass, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_TUTRANGANSSAccuracyClass },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_CommonMeasurementAccuracy(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_CommonMeasurementAccuracy, CommonMeasurementAccuracy_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_CommonMeasurementType_vals[] = {
  {   0, "uTRAN-GPS-timing-of-cell-frames-for-UE-Positioning" },
  {   1, "sFN-SFN-observerd-time-difference" },
  {   2, "load" },
  {   3, "transmitted-carrier-power" },
  {   4, "received-total-wide-band-power" },
  {   5, "uplink-timeslot-iscp" },
  {   6, "rT-load" },
  {   7, "nRT-load-Information" },
  {   8, "upPTSInterference" },
  {   9, "uTRAN-GANSS-timing-of-cell-frames-for-UE-Positioning" },
  { 0, NULL }
};


static int
dissect_rnsap_CommonMeasurementType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     6, NULL, TRUE, 4, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_16383(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16383U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_4294967295(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4294967295U, NULL, FALSE);

  return offset;
}


static const per_sequence_t TUTRANGPS_sequence[] = {
  { &hf_rnsap_ms_part       , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_16383 },
  { &hf_rnsap_ls_part       , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_4294967295 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TUTRANGPS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TUTRANGPS, TUTRANGPS_sequence);

  return offset;
}



static int
dissect_rnsap_TUTRANGPSQuality(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_TUTRANGPSDriftRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -50, 50U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_TUTRANGPSDriftRateQuality(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 50U, NULL, FALSE);

  return offset;
}


static const per_sequence_t TUTRANGPSMeasurementValueInformation_sequence[] = {
  { &hf_rnsap_tUTRANGPS     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TUTRANGPS },
  { &hf_rnsap_tUTRANGPSQuality, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TUTRANGPSQuality },
  { &hf_rnsap_tUTRANGPSDriftRate, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TUTRANGPSDriftRate },
  { &hf_rnsap_tUTRANGPSDriftRateQuality, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TUTRANGPSDriftRateQuality },
  { &hf_rnsap_iEe_Extensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TUTRANGPSMeasurementValueInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TUTRANGPSMeasurementValueInformation, TUTRANGPSMeasurementValueInformation_sequence);

  return offset;
}


static const per_sequence_t UC_ID_sequence[] = {
  { &hf_rnsap_rNC_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RNC_ID },
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UC_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UC_ID, UC_ID_sequence);

  return offset;
}



static int
dissect_rnsap_SFNSFN_FDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 614399U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_SFNSFN_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 40961U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_SFNSFN_TDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 81923U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_SFNSFNValue_vals[] = {
  {   0, "sFNSFN-FDD" },
  {   1, "sFNSFN-TDD" },
  {   2, "sFNSFN-TDD768" },
  { 0, NULL }
};

static const per_choice_t SFNSFNValue_choice[] = {
  {   0, &hf_rnsap_sFNSFN_FDD    , ASN1_EXTENSION_ROOT    , dissect_rnsap_SFNSFN_FDD },
  {   1, &hf_rnsap_sFNSFN_TDD    , ASN1_EXTENSION_ROOT    , dissect_rnsap_SFNSFN_TDD },
  {   2, &hf_rnsap_sFNSFN_TDD768 , ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_SFNSFN_TDD768 },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_SFNSFNValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_SFNSFNValue, SFNSFNValue_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_SFNSFNQuality(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_SFNSFNDriftRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -100, 100U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_SFNSFNDriftRateQuality(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_SFN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4095U, NULL, FALSE);

  return offset;
}


static const per_sequence_t SFNSFNTimeStamp_TDD_sequence[] = {
  { &hf_rnsap_sFN           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SFN },
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SFNSFNTimeStamp_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SFNSFNTimeStamp_TDD, SFNSFNTimeStamp_TDD_sequence);

  return offset;
}


static const value_string rnsap_SFNSFNTimeStampInformation_vals[] = {
  {   0, "sFNSFNTimeStamp-FDD" },
  {   1, "sFNSFNTimeStamp-TDD" },
  { 0, NULL }
};

static const per_choice_t SFNSFNTimeStampInformation_choice[] = {
  {   0, &hf_rnsap_sFNSFNTimeStamp_FDD, ASN1_EXTENSION_ROOT    , dissect_rnsap_SFN },
  {   1, &hf_rnsap_sFNSFNTimeStamp_TDD, ASN1_EXTENSION_ROOT    , dissect_rnsap_SFNSFNTimeStamp_TDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_SFNSFNTimeStampInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_SFNSFNTimeStampInformation, SFNSFNTimeStampInformation_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item_sequence[] = {
  { &hf_rnsap_uC_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UC_ID },
  { &hf_rnsap_sFNSFNValue   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SFNSFNValue },
  { &hf_rnsap_sFNSFNQuality , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SFNSFNQuality },
  { &hf_rnsap_sFNSFNDriftRate, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SFNSFNDriftRate },
  { &hf_rnsap_sFNSFNDriftRateQuality, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SFNSFNDriftRateQuality },
  { &hf_rnsap_sFNSFNTimeStampInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SFNSFNTimeStampInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item, T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item_sequence);

  return offset;
}


static const per_sequence_t T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_sequence_of[1] = {
  { &hf_rnsap_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item },
};

static int
dissect_rnsap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation, T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_sequence_of,
                                                  1, maxNrOfMeasNCell, FALSE);

  return offset;
}


static const per_sequence_t T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item_sequence[] = {
  { &hf_rnsap_uC_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UC_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item, T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item_sequence);

  return offset;
}


static const per_sequence_t T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_sequence_of[1] = {
  { &hf_rnsap_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item },
};

static int
dissect_rnsap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation, T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_sequence_of,
                                                  0, maxNrOfMeasNCell_1, FALSE);

  return offset;
}


static const per_sequence_t SFNSFNMeasurementValueInformation_sequence[] = {
  { &hf_rnsap_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation },
  { &hf_rnsap_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SFNSFNMeasurementValueInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SFNSFNMeasurementValueInformation, SFNSFNMeasurementValueInformation_sequence);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_100(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}


static const per_sequence_t LoadValue_sequence[] = {
  { &hf_rnsap_uplinkLoadValue, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_100 },
  { &hf_rnsap_downlinkLoadValue, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_100 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_LoadValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_LoadValue, LoadValue_sequence);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_621(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 621U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_UL_TimeslotISCP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Extension_CommonMeasurementValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_CommonMeasurementValue_vals[] = {
  {   0, "tUTRANGPSMeasurementValueInformation" },
  {   1, "sFNSFNMeasurementValueInformation" },
  {   2, "loadValue" },
  {   3, "transmittedCarrierPowerValue" },
  {   4, "receivedTotalWideBandPowerValue" },
  {   5, "uplinkTimeslotISCPValue" },
  {   6, "extension-CommonMeasurementValue" },
  { 0, NULL }
};

static const per_choice_t CommonMeasurementValue_choice[] = {
  {   0, &hf_rnsap_tUTRANGPSMeasurementValueInformation, ASN1_EXTENSION_ROOT    , dissect_rnsap_TUTRANGPSMeasurementValueInformation },
  {   1, &hf_rnsap_sFNSFNMeasurementValueInformation, ASN1_EXTENSION_ROOT    , dissect_rnsap_SFNSFNMeasurementValueInformation },
  {   2, &hf_rnsap_loadValue     , ASN1_EXTENSION_ROOT    , dissect_rnsap_LoadValue },
  {   3, &hf_rnsap_transmittedCarrierPowerValue, ASN1_EXTENSION_ROOT    , dissect_rnsap_INTEGER_0_100 },
  {   4, &hf_rnsap_receivedTotalWideBandPowerValue, ASN1_EXTENSION_ROOT    , dissect_rnsap_INTEGER_0_621 },
  {   5, &hf_rnsap_uplinkTimeslotISCPValue, ASN1_EXTENSION_ROOT    , dissect_rnsap_UL_TimeslotISCP },
  {   6, &hf_rnsap_extension_CommonMeasurementValue, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_CommonMeasurementValue },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_CommonMeasurementValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_CommonMeasurementValue, CommonMeasurementValue_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t CommonMeasurementAvailable_sequence[] = {
  { &hf_rnsap_commonMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CommonMeasurementValue },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CommonMeasurementAvailable(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CommonMeasurementAvailable, CommonMeasurementAvailable_sequence);

  return offset;
}


static const value_string rnsap_CommonMeasurementValueInformation_vals[] = {
  {   0, "measurementAvailable" },
  {   1, "measurementnotAvailable" },
  { 0, NULL }
};

static const per_choice_t CommonMeasurementValueInformation_choice[] = {
  {   0, &hf_rnsap_measurementAvailable, ASN1_NO_EXTENSIONS     , dissect_rnsap_CommonMeasurementAvailable },
  {   1, &hf_rnsap_measurementnotAvailable, ASN1_NO_EXTENSIONS     , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_CommonMeasurementValueInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_CommonMeasurementValueInformation, CommonMeasurementValueInformation_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_CongestionCause_vals[] = {
  {   0, "uTRAN-dynamic-resources" },
  {   1, "uTRAN-semistatic-resources" },
  { 0, NULL }
};


static int
dissect_rnsap_CongestionCause(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_CommonTransportChannelResourcesInitialisationNotRequired_vals[] = {
  {   0, "not-Required" },
  { 0, NULL }
};


static int
dissect_rnsap_CommonTransportChannelResourcesInitialisationNotRequired(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_MaxNr_Retransmissions_EDCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_E_DCH_HARQ_PO_FDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfEDCH_HARQ_PO_QUANTSTEPs, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_E_DCH_MACdFlow_Multiplexing_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     maxNrOfEDCHMACdFlows, maxNrOfEDCHMACdFlows, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_LogicalChannelID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_MAC_PDU_SizeExtended(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 1504U, NULL, TRUE);

  return offset;
}


static const per_sequence_t Common_E_DCH_LogicalChannelInformationItem_sequence[] = {
  { &hf_rnsap_logicalChannelId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_LogicalChannelID },
  { &hf_rnsap_maximumMACdPDU_SizeExtended, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAC_PDU_SizeExtended },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Common_E_DCH_LogicalChannelInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Common_E_DCH_LogicalChannelInformationItem, Common_E_DCH_LogicalChannelInformationItem_sequence);

  return offset;
}


static const per_sequence_t Common_E_DCH_LogicalChannelInformation_sequence_of[1] = {
  { &hf_rnsap_Common_E_DCH_LogicalChannelInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Common_E_DCH_LogicalChannelInformationItem },
};

static int
dissect_rnsap_Common_E_DCH_LogicalChannelInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Common_E_DCH_LogicalChannelInformation, Common_E_DCH_LogicalChannelInformation_sequence_of,
                                                  1, maxNoOfLogicalChannels, FALSE);

  return offset;
}


static const per_sequence_t Common_EDCH_MAC_d_Flow_Specific_InformationFDDItem_sequence[] = {
  { &hf_rnsap_common_EDCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_maximum_Number_of_Retransmissions_For_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaxNr_Retransmissions_EDCH },
  { &hf_rnsap_eDCH_HARQ_PO_FDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_HARQ_PO_FDD },
  { &hf_rnsap_eDCH_MACdFlow_Multiplexing_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_MACdFlow_Multiplexing_List },
  { &hf_rnsap_common_E_DCHLogicalChannelInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Common_E_DCH_LogicalChannelInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDDItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDDItem, Common_EDCH_MAC_d_Flow_Specific_InformationFDDItem_sequence);

  return offset;
}


static const per_sequence_t Common_EDCH_MAC_d_Flow_Specific_InformationFDD_sequence_of[1] = {
  { &hf_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDDItem },
};

static int
dissect_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDD, Common_EDCH_MAC_d_Flow_Specific_InformationFDD_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}



static int
dissect_rnsap_EDCH_MACdFlow_ID_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfEDCHMACdFlowsLCR_1, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_E_DCH_HARQ_PO_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 6U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Common_EDCH_MAC_d_Flow_Specific_InformationItemLCR_sequence[] = {
  { &hf_rnsap_common_EDCH_MACdFlow_ID_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID_LCR },
  { &hf_rnsap_maximum_Number_of_Retransmissions_For_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaxNr_Retransmissions_EDCH },
  { &hf_rnsap_eDCH_HARQ_PO_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_HARQ_PO_TDD },
  { &hf_rnsap_eDCH_MACdFlow_Multiplexing_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_MACdFlow_Multiplexing_List },
  { &hf_rnsap_common_E_DCHLogicalChannelInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Common_E_DCH_LogicalChannelInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationItemLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationItemLCR, Common_EDCH_MAC_d_Flow_Specific_InformationItemLCR_sequence);

  return offset;
}


static const per_sequence_t Common_EDCH_MAC_d_Flow_Specific_InformationLCR_sequence_of[1] = {
  { &hf_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationItemLCR },
};

static int
dissect_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationLCR, Common_EDCH_MAC_d_Flow_Specific_InformationLCR_sequence_of,
                                                  1, maxNrOfEDCHMACdFlowsLCR, FALSE);

  return offset;
}



static int
dissect_rnsap_Common_EDCH_Support_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_Common_E_RGCH_De_Activation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Common_E_RGCH_Cell_Info_sequence[] = {
  { &hf_rnsap_eRGCH_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DL_ChannelisationCodeNumber },
  { &hf_rnsap_eRGCH_SignatureSequence, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ERGCH_SignatureSequence },
  { &hf_rnsap_serving_Grant_Value, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_Serving_Grant_Value },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Common_E_RGCH_Cell_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Common_E_RGCH_Cell_Info, Common_E_RGCH_Cell_Info_sequence);

  return offset;
}


static const value_string rnsap_Common_E_RGCH_Cell_Information_vals[] = {
  {   0, "deActivation" },
  {   1, "common-E-RGCH-Cell-Info" },
  { 0, NULL }
};

static const per_choice_t Common_E_RGCH_Cell_Information_choice[] = {
  {   0, &hf_rnsap_deActivation  , ASN1_EXTENSION_ROOT    , dissect_rnsap_Common_E_RGCH_De_Activation },
  {   1, &hf_rnsap_common_E_RGCH_Cell_Info, ASN1_EXTENSION_ROOT    , dissect_rnsap_Common_E_RGCH_Cell_Info },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Common_E_RGCH_Cell_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Common_E_RGCH_Cell_Information, Common_E_RGCH_Cell_Information_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_CompleteAlmanacProvided(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}



static int
dissect_rnsap_UE_DTX_DRX_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 159U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_Enabling_Delay_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  {   2, "v2" },
  {   3, "v4" },
  {   4, "v8" },
  {   5, "v16" },
  {   6, "v32" },
  {   7, "v64" },
  {   8, "v128" },
  { 0, NULL }
};


static int
dissect_rnsap_Enabling_Delay(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     9, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_UE_DTX_Cycle2_2ms_vals[] = {
  {   0, "v4" },
  {   1, "v5" },
  {   2, "v8" },
  {   3, "v10" },
  {   4, "v16" },
  {   5, "v20" },
  {   6, "v32" },
  {   7, "v40" },
  {   8, "v64" },
  {   9, "v80" },
  {  10, "v128" },
  {  11, "v160" },
  { 0, NULL }
};


static int
dissect_rnsap_UE_DTX_Cycle2_2ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     12, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_MAC_DTX_Cycle_2ms_vals[] = {
  {   0, "v1" },
  {   1, "v4" },
  {   2, "v5" },
  {   3, "v8" },
  {   4, "v10" },
  {   5, "v16" },
  {   6, "v20" },
  { 0, NULL }
};


static int
dissect_rnsap_MAC_DTX_Cycle_2ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DTX_Cycle_2ms_Items_sequence[] = {
  { &hf_rnsap_uE_DTX_Cycle1_2ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DTX_Cycle1_2ms },
  { &hf_rnsap_uE_DTX_Cycle2_2ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DTX_Cycle2_2ms },
  { &hf_rnsap_mAC_DTX_Cycle_2ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAC_DTX_Cycle_2ms },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DTX_Cycle_2ms_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DTX_Cycle_2ms_Items, DTX_Cycle_2ms_Items_sequence);

  return offset;
}


static const value_string rnsap_UE_DTX_Cycle1_10ms_vals[] = {
  {   0, "v1" },
  {   1, "v5" },
  {   2, "v10" },
  {   3, "v20" },
  { 0, NULL }
};


static int
dissect_rnsap_UE_DTX_Cycle1_10ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_UE_DTX_Cycle2_10ms_vals[] = {
  {   0, "v5" },
  {   1, "v10" },
  {   2, "v20" },
  {   3, "v40" },
  {   4, "v80" },
  {   5, "v160" },
  { 0, NULL }
};


static int
dissect_rnsap_UE_DTX_Cycle2_10ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     6, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_MAC_DTX_Cycle_10ms_vals[] = {
  {   0, "v5" },
  {   1, "v10" },
  {   2, "v20" },
  { 0, NULL }
};


static int
dissect_rnsap_MAC_DTX_Cycle_10ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DTX_Cycle_10ms_Items_sequence[] = {
  { &hf_rnsap_uE_DTX_Cycle1_10ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DTX_Cycle1_10ms },
  { &hf_rnsap_uE_DTX_Cycle2_10ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DTX_Cycle2_10ms },
  { &hf_rnsap_mAC_DTX_Cycle_10ms, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAC_DTX_Cycle_10ms },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DTX_Cycle_10ms_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DTX_Cycle_10ms_Items, DTX_Cycle_10ms_Items_sequence);

  return offset;
}


static const value_string rnsap_E_DCH_TTI_Length_vals[] = {
  {   0, "two-ms" },
  {   1, "ten-ms" },
  { 0, NULL }
};

static const per_choice_t E_DCH_TTI_Length_choice[] = {
  {   0, &hf_rnsap_two_ms        , ASN1_EXTENSION_ROOT    , dissect_rnsap_DTX_Cycle_2ms_Items },
  {   1, &hf_rnsap_ten_ms        , ASN1_EXTENSION_ROOT    , dissect_rnsap_DTX_Cycle_10ms_Items },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_E_DCH_TTI_Length(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_E_DCH_TTI_Length, E_DCH_TTI_Length_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_UE_DTX_Long_Preamble_vals[] = {
  {   0, "v2" },
  {   1, "v4" },
  {   2, "v15" },
  { 0, NULL }
};


static int
dissect_rnsap_UE_DTX_Long_Preamble(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_MAC_Inactivity_Threshold_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v16" },
  {   5, "v32" },
  {   6, "v64" },
  {   7, "v128" },
  {   8, "v256" },
  {   9, "v512" },
  {  10, "infinity" },
  { 0, NULL }
};


static int
dissect_rnsap_MAC_Inactivity_Threshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     11, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_CQI_DTX_Timer_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  {   2, "v2" },
  {   3, "v4" },
  {   4, "v8" },
  {   5, "v16" },
  {   6, "v32" },
  {   7, "v64" },
  {   8, "v128" },
  {   9, "v256" },
  {  10, "v512" },
  {  11, "infinity" },
  { 0, NULL }
};


static int
dissect_rnsap_CQI_DTX_Timer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     12, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_UE_DPCCH_burst1_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v5" },
  { 0, NULL }
};


static int
dissect_rnsap_UE_DPCCH_burst1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_UE_DPCCH_burst2_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v5" },
  { 0, NULL }
};


static int
dissect_rnsap_UE_DPCCH_burst2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DTX_Information_sequence[] = {
  { &hf_rnsap_e_DCH_TTI_Length, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_TTI_Length },
  { &hf_rnsap_inactivity_Threshold_for_UE_DTX_Cycle2, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Inactivity_Threshold_for_UE_DTX_Cycle2 },
  { &hf_rnsap_uE_DTX_Long_Preamble, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DTX_Long_Preamble },
  { &hf_rnsap_mAC_Inactivity_Threshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAC_Inactivity_Threshold },
  { &hf_rnsap_cQI_DTX_Timer , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CQI_DTX_Timer },
  { &hf_rnsap_uE_DPCCH_burst1, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DPCCH_burst1 },
  { &hf_rnsap_uE_DPCCH_burst2, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DPCCH_burst2 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DTX_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DTX_Information, DTX_Information_sequence);

  return offset;
}


static const value_string rnsap_UE_DRX_Cycle_vals[] = {
  {   0, "v4" },
  {   1, "v5" },
  {   2, "v8" },
  {   3, "v10" },
  {   4, "v16" },
  {   5, "v20" },
  { 0, NULL }
};


static int
dissect_rnsap_UE_DRX_Cycle(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     6, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  {   2, "v2" },
  {   3, "v4" },
  {   4, "v8" },
  {   5, "v16" },
  {   6, "v32" },
  {   7, "v64" },
  {   8, "v128" },
  {   9, "v256" },
  {  10, "v512" },
  { 0, NULL }
};


static int
dissect_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     11, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_Inactivity_Threshold_for_UE_Grant_Monitoring_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  {   2, "v2" },
  {   3, "v4" },
  {   4, "v8" },
  {   5, "v16" },
  {   6, "v32" },
  {   7, "v64" },
  {   8, "v128" },
  {   9, "v256" },
  { 0, NULL }
};


static int
dissect_rnsap_Inactivity_Threshold_for_UE_Grant_Monitoring(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     10, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_UE_DRX_Grant_Monitoring(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t DRX_Information_sequence[] = {
  { &hf_rnsap_uE_DRX_Cycle  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DRX_Cycle },
  { &hf_rnsap_inactivity_Threshold_for_UE_DRX_Cycle, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle },
  { &hf_rnsap_inactivity_Threshold_for_UE_Grant_Monitoring, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Inactivity_Threshold_for_UE_Grant_Monitoring },
  { &hf_rnsap_uE_DRX_Grant_Monitoring, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DRX_Grant_Monitoring },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DRX_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DRX_Information, DRX_Information_sequence);

  return offset;
}


static const per_sequence_t Continuous_Packet_Connectivity_DTX_DRX_Information_sequence[] = {
  { &hf_rnsap_uE_DTX_DRX_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DTX_DRX_Offset },
  { &hf_rnsap_enabling_Delay, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Enabling_Delay },
  { &hf_rnsap_dTX_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DTX_Information },
  { &hf_rnsap_dRX_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DRX_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information, Continuous_Packet_Connectivity_DTX_DRX_Information_sequence);

  return offset;
}


static const per_sequence_t DTX_Cycle_2ms_to_Modify_Items_sequence[] = {
  { &hf_rnsap_uE_DTX_Cycle1_2ms, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DTX_Cycle1_2ms },
  { &hf_rnsap_uE_DTX_Cycle2_2ms, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DTX_Cycle2_2ms },
  { &hf_rnsap_mAC_DTX_Cycle_2ms, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MAC_DTX_Cycle_2ms },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DTX_Cycle_2ms_to_Modify_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DTX_Cycle_2ms_to_Modify_Items, DTX_Cycle_2ms_to_Modify_Items_sequence);

  return offset;
}


static const per_sequence_t DTX_Cycle_10ms_to_Modify_Items_sequence[] = {
  { &hf_rnsap_uE_DTX_Cycle1_10ms, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DTX_Cycle1_10ms },
  { &hf_rnsap_uE_DTX_Cycle2_10ms, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DTX_Cycle2_10ms },
  { &hf_rnsap_mAC_DTX_Cycle_10ms, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MAC_DTX_Cycle_10ms },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DTX_Cycle_10ms_to_Modify_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DTX_Cycle_10ms_to_Modify_Items, DTX_Cycle_10ms_to_Modify_Items_sequence);

  return offset;
}


static const value_string rnsap_E_DCH_TTI_Length_to_Modify_vals[] = {
  {   0, "two-ms" },
  {   1, "ten-ms" },
  { 0, NULL }
};

static const per_choice_t E_DCH_TTI_Length_to_Modify_choice[] = {
  {   0, &hf_rnsap_two_ms_01     , ASN1_EXTENSION_ROOT    , dissect_rnsap_DTX_Cycle_2ms_to_Modify_Items },
  {   1, &hf_rnsap_ten_ms_01     , ASN1_EXTENSION_ROOT    , dissect_rnsap_DTX_Cycle_10ms_to_Modify_Items },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_E_DCH_TTI_Length_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_E_DCH_TTI_Length_to_Modify, E_DCH_TTI_Length_to_Modify_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t DTX_Information_to_Modify_Items_sequence[] = {
  { &hf_rnsap_e_DCH_TTI_Length_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_TTI_Length_to_Modify },
  { &hf_rnsap_inactivity_Threshold_for_UE_DTX_Cycle2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Inactivity_Threshold_for_UE_DTX_Cycle2 },
  { &hf_rnsap_uE_DTX_Long_Preamble, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DTX_Long_Preamble },
  { &hf_rnsap_mAC_Inactivity_Threshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MAC_Inactivity_Threshold },
  { &hf_rnsap_cQI_DTX_Timer , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CQI_DTX_Timer },
  { &hf_rnsap_uE_DPCCH_burst1, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DPCCH_burst1 },
  { &hf_rnsap_uE_DPCCH_burst2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DPCCH_burst2 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DTX_Information_to_Modify_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DTX_Information_to_Modify_Items, DTX_Information_to_Modify_Items_sequence);

  return offset;
}


static const value_string rnsap_DTX_Information_to_Modify_vals[] = {
  {   0, "modify" },
  {   1, "deactivate" },
  { 0, NULL }
};

static const per_choice_t DTX_Information_to_Modify_choice[] = {
  {   0, &hf_rnsap_modify_02     , ASN1_EXTENSION_ROOT    , dissect_rnsap_DTX_Information_to_Modify_Items },
  {   1, &hf_rnsap_deactivate    , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DTX_Information_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DTX_Information_to_Modify, DTX_Information_to_Modify_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t DRX_Information_to_Modify_Items_sequence[] = {
  { &hf_rnsap_uE_DRX_Cycle  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DRX_Cycle },
  { &hf_rnsap_inactivity_Threshold_for_UE_DRX_Cycle, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle },
  { &hf_rnsap_inactivity_Threshold_for_UE_Grant_Monitoring, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Inactivity_Threshold_for_UE_Grant_Monitoring },
  { &hf_rnsap_uE_DRX_Grant_Monitoring, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DRX_Grant_Monitoring },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DRX_Information_to_Modify_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DRX_Information_to_Modify_Items, DRX_Information_to_Modify_Items_sequence);

  return offset;
}


static const value_string rnsap_DRX_Information_to_Modify_vals[] = {
  {   0, "modify" },
  {   1, "deactivate" },
  { 0, NULL }
};

static const per_choice_t DRX_Information_to_Modify_choice[] = {
  {   0, &hf_rnsap_modify_01     , ASN1_EXTENSION_ROOT    , dissect_rnsap_DRX_Information_to_Modify_Items },
  {   1, &hf_rnsap_deactivate    , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DRX_Information_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DRX_Information_to_Modify, DRX_Information_to_Modify_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Continuous_Packet_Connectivity_DTX_DRX_Information_to_Modify_sequence[] = {
  { &hf_rnsap_uE_DTX_DRX_Offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DTX_DRX_Offset },
  { &hf_rnsap_enabling_Delay, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Enabling_Delay },
  { &hf_rnsap_dTX_Information_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DTX_Information_to_Modify },
  { &hf_rnsap_dRX_Information_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DRX_Information_to_Modify },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information_to_Modify, Continuous_Packet_Connectivity_DTX_DRX_Information_to_Modify_sequence);

  return offset;
}



static int
dissect_rnsap_Transport_Block_Size_Index(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNrOfHS_DSCHTBSs, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_HSPDSCH_Second_Code_Support(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t Continuous_Packet_Connectivity_HS_SCCH_Less_InformationItem_sequence[] = {
  { &hf_rnsap_transport_Block_Size_Index, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Transport_Block_Size_Index },
  { &hf_rnsap_hSPDSCH_Second_Code_Support, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSPDSCH_Second_Code_Support },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_InformationItem, Continuous_Packet_Connectivity_HS_SCCH_Less_InformationItem_sequence);

  return offset;
}


static const per_sequence_t Continuous_Packet_Connectivity_HS_SCCH_Less_Information_sequence_of[1] = {
  { &hf_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_InformationItem },
};

static int
dissect_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information, Continuous_Packet_Connectivity_HS_SCCH_Less_Information_sequence_of,
                                                  1, maxNrOfHS_DSCHTBSs_HS_SCCHless, FALSE);

  return offset;
}



static int
dissect_rnsap_HSPDSCH_First_Code_Index(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxHS_PDSCHCodeNrComp_1, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_HSPDSCH_Second_Code_Index(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxHS_PDSCHCodeNrComp_1, NULL, FALSE);

  return offset;
}


static const per_sequence_t Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response_sequence[] = {
  { &hf_rnsap_hSPDSCH_First_Code_Index, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSPDSCH_First_Code_Index },
  { &hf_rnsap_hSPDSCH_Second_Code_Index, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSPDSCH_Second_Code_Index },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response, Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response_sequence);

  return offset;
}


static const value_string rnsap_CoverageIndicator_vals[] = {
  {   0, "overlap" },
  {   1, "covers" },
  {   2, "containedIn" },
  { 0, NULL }
};


static int
dissect_rnsap_CoverageIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t CPC_Information_sequence[] = {
  { &hf_rnsap_continuous_Packet_Connectivity_DTX_DRX_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information },
  { &hf_rnsap_continuous_Packet_Connectivity_DTX_DRX_Information_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information_to_Modify },
  { &hf_rnsap_continuous_Packet_Connectivity_HS_SCCH_Less_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CPC_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CPC_Information, CPC_Information_sequence);

  return offset;
}


static const value_string rnsap_CPC_RecoveryReport_vals[] = {
  {   0, "initiated" },
  { 0, NULL }
};


static int
dissect_rnsap_CPC_RecoveryReport(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_Continuous_Packet_Connectivity_HS_SCCH_less_Deactivate_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_Counting_Result(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Counting_Information_List_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_counting_Result, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Counting_Result },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Counting_Information_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Counting_Information_List, Counting_Information_List_sequence);

  return offset;
}


static const per_sequence_t Counting_Information_sequence_of[1] = {
  { &hf_rnsap_Counting_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Counting_Information_List },
};

static int
dissect_rnsap_Counting_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Counting_Information, Counting_Information_sequence_of,
                                                  1, maxNrOfFDDNeighboursPerRNC, TRUE);

  return offset;
}


static const value_string rnsap_CRC_Size_vals[] = {
  {   0, "v0" },
  {   1, "v8" },
  {   2, "v12" },
  {   3, "v16" },
  {   4, "v24" },
  { 0, NULL }
};


static int
dissect_rnsap_CRC_Size(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_RepetitionNumber0(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t CriticalityDiagnostics_IE_List_item_sequence[] = {
  { &hf_rnsap_iECriticality , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Criticality },
  { &hf_rnsap_iE_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_ID },
  { &hf_rnsap_repetitionNumber, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RepetitionNumber0 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CriticalityDiagnostics_IE_List_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CriticalityDiagnostics_IE_List_item, CriticalityDiagnostics_IE_List_item_sequence);

  return offset;
}


static const per_sequence_t CriticalityDiagnostics_IE_List_sequence_of[1] = {
  { &hf_rnsap_CriticalityDiagnostics_IE_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_CriticalityDiagnostics_IE_List_item },
};

static int
dissect_rnsap_CriticalityDiagnostics_IE_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_CriticalityDiagnostics_IE_List, CriticalityDiagnostics_IE_List_sequence_of,
                                                  1, maxNrOfErrors, FALSE);

  return offset;
}


static const per_sequence_t CriticalityDiagnostics_sequence[] = {
  { &hf_rnsap_procedureID   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProcedureID },
  { &hf_rnsap_triggeringMessage, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TriggeringMessage },
  { &hf_rnsap_procedureCriticality, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Criticality },
  { &hf_rnsap_transactionID , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransactionID },
  { &hf_rnsap_iEsCriticalityDiagnostics, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CriticalityDiagnostics_IE_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CriticalityDiagnostics(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CriticalityDiagnostics, CriticalityDiagnostics_sequence);

  return offset;
}



static int
dissect_rnsap_RepetitionNumber1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}


static const per_sequence_t MessageStructure_item_sequence[] = {
  { &hf_rnsap_iE_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_ID },
  { &hf_rnsap_repetitionNumber_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RepetitionNumber1 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MessageStructure_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MessageStructure_item, MessageStructure_item_sequence);

  return offset;
}


static const per_sequence_t MessageStructure_sequence_of[1] = {
  { &hf_rnsap_MessageStructure_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MessageStructure_item },
};

static int
dissect_rnsap_MessageStructure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MessageStructure, MessageStructure_sequence_of,
                                                  1, maxNrOfLevels, FALSE);

  return offset;
}


static const per_sequence_t CN_CS_DomainIdentifier_sequence[] = {
  { &hf_rnsap_pLMN_Identity , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_PLMN_Identity },
  { &hf_rnsap_lAC           , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_LAC },
  { &hf_rnsap_iE_Extensions , ASN1_NO_EXTENSIONS     , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CN_CS_DomainIdentifier(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CN_CS_DomainIdentifier, CN_CS_DomainIdentifier_sequence);

  return offset;
}


static const per_sequence_t CN_PS_DomainIdentifier_sequence[] = {
  { &hf_rnsap_pLMN_Identity , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_PLMN_Identity },
  { &hf_rnsap_lAC           , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_LAC },
  { &hf_rnsap_rAC           , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_RAC },
  { &hf_rnsap_iE_Extensions , ASN1_NO_EXTENSIONS     , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CN_PS_DomainIdentifier(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CN_PS_DomainIdentifier, CN_PS_DomainIdentifier_sequence);

  return offset;
}


static const value_string rnsap_CNDomainType_vals[] = {
  {   0, "cs-domain" },
  {   1, "ps-domain" },
  {   2, "i-care" },
  { 0, NULL }
};


static int
dissect_rnsap_CNDomainType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_CQI_Cycle_Switch_Timer_vals[] = {
  {   0, "v4" },
  {   1, "v8" },
  {   2, "v16" },
  {   3, "v32" },
  {   4, "v64" },
  {   5, "v128" },
  {   6, "v256" },
  {   7, "v512" },
  {   8, "infinity" },
  { 0, NULL }
};


static int
dissect_rnsap_CQI_Cycle_Switch_Timer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     9, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_ControlGAP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 255U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_CQI_Feedback_Cycle_vals[] = {
  {   0, "v0" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v10" },
  {   5, "v20" },
  {   6, "v40" },
  {   7, "v80" },
  {   8, "v160" },
  {   9, "v16" },
  {  10, "v32" },
  {  11, "v64" },
  { 0, NULL }
};


static int
dissect_rnsap_CQI_Feedback_Cycle(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     9, NULL, TRUE, 3, NULL);

  return offset;
}


static const value_string rnsap_CQI_Feedback_Cycle2_vals[] = {
  {   0, "v0" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v16" },
  {   5, "v10" },
  {   6, "v20" },
  {   7, "v32" },
  {   8, "v40" },
  {   9, "v64" },
  {  10, "v80" },
  {  11, "v160" },
  { 0, NULL }
};


static int
dissect_rnsap_CQI_Feedback_Cycle2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     12, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_CQI_Power_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_CQI_RepetitionFactor(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 4U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_C_RNTI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_CodeRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_CodeRate_short(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 10U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_UE_DRX_Cycle_LCR_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v16" },
  {   5, "v32" },
  {   6, "v64" },
  { 0, NULL }
};


static int
dissect_rnsap_UE_DRX_Cycle_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v16" },
  {   5, "v32" },
  {   6, "v64" },
  { 0, NULL }
};


static int
dissect_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_UE_DRX_Offset_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_SCCH_DRX_Information_LCR_sequence[] = {
  { &hf_rnsap_hS_SCCH_UE_DRX_Cycle_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DRX_Cycle_LCR },
  { &hf_rnsap_hS_SCCH_Inactivity_Threshold_for_UE_DRX_Cycle_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR },
  { &hf_rnsap_hS_SCCH_UE_DRX_Offset_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DRX_Offset_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_SCCH_DRX_Information_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_SCCH_DRX_Information_LCR, HS_SCCH_DRX_Information_LCR_sequence);

  return offset;
}


static const value_string rnsap_E_AGCH_UE_Inactivity_Monitor_Threshold_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  {   2, "v2" },
  {   3, "v4" },
  {   4, "v8" },
  {   5, "v16" },
  {   6, "v32" },
  {   7, "v64" },
  {   8, "v128" },
  {   9, "v256" },
  {  10, "v512" },
  {  11, "infinity" },
  { 0, NULL }
};


static int
dissect_rnsap_E_AGCH_UE_Inactivity_Monitor_Threshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     12, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t E_AGCH_DRX_Parameters_sequence[] = {
  { &hf_rnsap_e_AGCH_UE_DRX_Cycle_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DRX_Cycle_LCR },
  { &hf_rnsap_e_AGCH_UE_Inactivity_Monitor_Threshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_AGCH_UE_Inactivity_Monitor_Threshold },
  { &hf_rnsap_e_AGCH_UE_DRX_Offset_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_DRX_Offset_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_AGCH_DRX_Parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_AGCH_DRX_Parameters, E_AGCH_DRX_Parameters_sequence);

  return offset;
}


static const value_string rnsap_E_AGCH_DRX_Information_LCR_vals[] = {
  {   0, "sameAsHS-SCCH" },
  {   1, "e-AGCH-DRX-Parameters" },
  { 0, NULL }
};

static const per_choice_t E_AGCH_DRX_Information_LCR_choice[] = {
  {   0, &hf_rnsap_sameAsHS_SCCH , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_e_AGCH_DRX_Parameters, ASN1_EXTENSION_ROOT    , dissect_rnsap_E_AGCH_DRX_Parameters },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_E_AGCH_DRX_Information_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_E_AGCH_DRX_Information_LCR, E_AGCH_DRX_Information_LCR_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ContinuousPacketConnectivity_DRX_InformationLCR_sequence[] = {
  { &hf_rnsap_enabling_Delay, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Enabling_Delay },
  { &hf_rnsap_hS_SCCH_DRX_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SCCH_DRX_Information_LCR },
  { &hf_rnsap_e_AGCH_DRX_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_AGCH_DRX_Information_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ContinuousPacketConnectivity_DRX_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ContinuousPacketConnectivity_DRX_InformationLCR, ContinuousPacketConnectivity_DRX_InformationLCR_sequence);

  return offset;
}


static const per_sequence_t DRX_Information_to_Modify_Items_LCR_sequence[] = {
  { &hf_rnsap_hS_SCCH_DRX_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_SCCH_DRX_Information_LCR },
  { &hf_rnsap_e_AGCH_DRX_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_AGCH_DRX_Information_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DRX_Information_to_Modify_Items_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DRX_Information_to_Modify_Items_LCR, DRX_Information_to_Modify_Items_LCR_sequence);

  return offset;
}


static const value_string rnsap_DRX_Information_to_Modify_LCR_vals[] = {
  {   0, "modify" },
  {   1, "deactivate" },
  { 0, NULL }
};

static const per_choice_t DRX_Information_to_Modify_LCR_choice[] = {
  {   0, &hf_rnsap_modify        , ASN1_EXTENSION_ROOT    , dissect_rnsap_DRX_Information_to_Modify_Items_LCR },
  {   1, &hf_rnsap_deactivate    , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DRX_Information_to_Modify_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DRX_Information_to_Modify_LCR, DRX_Information_to_Modify_LCR_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR_sequence[] = {
  { &hf_rnsap_enabling_Delay, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Enabling_Delay },
  { &hf_rnsap_dRX_Information_to_Modify_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DRX_Information_to_Modify_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR, ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR_sequence);

  return offset;
}



static int
dissect_rnsap_Transport_Block_Size_maping_Index_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNoOfTBSs_Mapping_HS_DSCH_SPS_1, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Transport_Block_Size_Index_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNoOfHS_DSCH_TBSsLCR, NULL, FALSE);

  return offset;
}


static const per_sequence_t Transport_Block_Size_Item_LCR_sequence[] = {
  { &hf_rnsap_transport_Block_Size_maping_Index_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Transport_Block_Size_maping_Index_LCR },
  { &hf_rnsap_transport_Block_Size_Index_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Transport_Block_Size_Index_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Transport_Block_Size_Item_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Transport_Block_Size_Item_LCR, Transport_Block_Size_Item_LCR_sequence);

  return offset;
}


static const per_sequence_t Transport_Block_Size_List_LCR_sequence_of[1] = {
  { &hf_rnsap_Transport_Block_Size_List_LCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Transport_Block_Size_Item_LCR },
};

static int
dissect_rnsap_Transport_Block_Size_List_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Transport_Block_Size_List_LCR, Transport_Block_Size_List_LCR_sequence_of,
                                                  1, maxNoOfTBSs_Mapping_HS_DSCH_SPS, FALSE);

  return offset;
}



static int
dissect_rnsap_RepetitionPeriodIndex(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNoOfRepetitionPeriod_SPS_LCR_1, NULL, FALSE);

  return offset;
}


static const value_string rnsap_RepetitionPeriod_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v16" },
  {   5, "v32" },
  {   6, "v64" },
  { 0, NULL }
};


static int
dissect_rnsap_RepetitionPeriod(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_RepetitionLength(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 63U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Repetition_Period_Item_LCR_sequence[] = {
  { &hf_rnsap_repetitionPeriodIndex, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriodIndex },
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RepetitionLength },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Repetition_Period_Item_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Repetition_Period_Item_LCR, Repetition_Period_Item_LCR_sequence);

  return offset;
}


static const per_sequence_t Repetition_Period_List_LCR_sequence_of[1] = {
  { &hf_rnsap_Repetition_Period_List_LCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Repetition_Period_Item_LCR },
};

static int
dissect_rnsap_Repetition_Period_List_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Repetition_Period_List_LCR, Repetition_Period_List_LCR_sequence_of,
                                                  1, maxNoOfRepetition_Period_LCR, FALSE);

  return offset;
}


static const value_string rnsap_SPS_Reservation_Indicator_vals[] = {
  {   0, "reserve" },
  { 0, NULL }
};


static int
dissect_rnsap_SPS_Reservation_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_LogicalChannellevel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     16, 16, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_PriorityQueuelevel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     8, 8, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const value_string rnsap_HS_DSCH_SPS_Operation_Indicator_vals[] = {
  {   0, "logicalChannellevel" },
  {   1, "priorityQueuelevel" },
  { 0, NULL }
};

static const per_choice_t HS_DSCH_SPS_Operation_Indicator_choice[] = {
  {   0, &hf_rnsap_logicalChannellevel, ASN1_EXTENSION_ROOT    , dissect_rnsap_LogicalChannellevel },
  {   1, &hf_rnsap_priorityQueuelevel, ASN1_EXTENSION_ROOT    , dissect_rnsap_PriorityQueuelevel },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_SPS_Operation_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_HS_DSCH_SPS_Operation_Indicator, HS_DSCH_SPS_Operation_Indicator_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_Semi_PersistentScheduling_Information_LCR_sequence[] = {
  { &hf_rnsap_transport_Block_Size_List, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Transport_Block_Size_List_LCR },
  { &hf_rnsap_repetition_Period_List_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Repetition_Period_List_LCR },
  { &hf_rnsap_hS_DSCH_SPS_Reservation_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SPS_Reservation_Indicator },
  { &hf_rnsap_hS_DSCH_SPS_Operation_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_DSCH_SPS_Operation_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_LCR, HS_DSCH_Semi_PersistentScheduling_Information_LCR_sequence);

  return offset;
}


static const per_sequence_t HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR_sequence[] = {
  { &hf_rnsap_transport_Block_Size_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Transport_Block_Size_List_LCR },
  { &hf_rnsap_repetition_Period_List_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Repetition_Period_List_LCR },
  { &hf_rnsap_hS_DSCH_SPS_Reservation_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SPS_Reservation_Indicator },
  { &hf_rnsap_hS_DSCH_SPS_Operation_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_DSCH_SPS_Operation_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR, HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR_sequence);

  return offset;
}



static int
dissect_rnsap_E_DCH_SPS_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     16, 16, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t E_DCH_Semi_PersistentScheduling_Information_LCR_sequence[] = {
  { &hf_rnsap_repetition_Period_List_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Repetition_Period_List_LCR },
  { &hf_rnsap_e_DCH_SPS_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_SPS_Indicator },
  { &hf_rnsap_e_DCH_SPS_Reservation_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SPS_Reservation_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_Semi_PersistentScheduling_Information_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_Semi_PersistentScheduling_Information_LCR, E_DCH_Semi_PersistentScheduling_Information_LCR_sequence);

  return offset;
}


static const per_sequence_t E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR_sequence[] = {
  { &hf_rnsap_repetition_Period_List_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Repetition_Period_List_LCR },
  { &hf_rnsap_e_DCH_SPS_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_SPS_Indicator },
  { &hf_rnsap_e_DCH_SPS_Reservation_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SPS_Reservation_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR, E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR_sequence);

  return offset;
}


static const per_sequence_t CPC_InformationLCR_sequence[] = {
  { &hf_rnsap_continuousPacketConnectivity_DRX_InformationLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ContinuousPacketConnectivity_DRX_InformationLCR },
  { &hf_rnsap_continuousPacketConnectivity_DRX_Information_to_Modify_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR },
  { &hf_rnsap_hS_DSCH_Semi_PersistentScheduling_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_LCR },
  { &hf_rnsap_hS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR },
  { &hf_rnsap_hS_DSCH_SPS_Deactivate_Indicator_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_e_DCH_Semi_PersistentScheduling_Information_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Semi_PersistentScheduling_Information_LCR },
  { &hf_rnsap_e_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR },
  { &hf_rnsap_e_DCH_SPS_Deactivate_Indicator_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CPC_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CPC_InformationLCR, CPC_InformationLCR_sequence);

  return offset;
}


static const value_string rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext_vals[] = {
  {   0, "v128" },
  {   1, "v256" },
  {   2, "v512" },
  { 0, NULL }
};


static int
dissect_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t HS_SCCH_DRX_Information_ResponseLCR_sequence[] = {
  { &hf_rnsap_hS_SCCH_UE_DRX_Cycle_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DRX_Cycle_LCR },
  { &hf_rnsap_hS_SCCH_Inactivity_Threshold_for_UE_DRX_Cycle_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR },
  { &hf_rnsap_hS_SCCH_UE_DRX_Offset_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DRX_Offset_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_SCCH_DRX_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_SCCH_DRX_Information_ResponseLCR, HS_SCCH_DRX_Information_ResponseLCR_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_DRX_Parameters_Response_sequence[] = {
  { &hf_rnsap_e_AGCH_UE_DRX_Cycle_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DRX_Cycle_LCR },
  { &hf_rnsap_e_AGCH_UE_Inactivity_Monitor_Threshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_AGCH_UE_Inactivity_Monitor_Threshold },
  { &hf_rnsap_e_AGCH_UE_DRX_Offset_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_DRX_Offset_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_AGCH_DRX_Parameters_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_AGCH_DRX_Parameters_Response, E_AGCH_DRX_Parameters_Response_sequence);

  return offset;
}


static const value_string rnsap_E_AGCH_DRX_Information_ResponseLCR_vals[] = {
  {   0, "sameAsHS-SCCH" },
  {   1, "e-AGCH-DRX-Parameters-Response" },
  { 0, NULL }
};

static const per_choice_t E_AGCH_DRX_Information_ResponseLCR_choice[] = {
  {   0, &hf_rnsap_sameAsHS_SCCH , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_e_AGCH_DRX_Parameters_Response, ASN1_EXTENSION_ROOT    , dissect_rnsap_E_AGCH_DRX_Parameters_Response },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_E_AGCH_DRX_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_E_AGCH_DRX_Information_ResponseLCR, E_AGCH_DRX_Information_ResponseLCR_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ContinuousPacketConnectivity_DRX_Information_ResponseLCR_sequence[] = {
  { &hf_rnsap_enabling_Delay, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Enabling_Delay },
  { &hf_rnsap_hS_SCCH_DRX_Information_ResponseLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_SCCH_DRX_Information_ResponseLCR },
  { &hf_rnsap_e_AGCH_DRX_Information_ResponseLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_AGCH_DRX_Information_ResponseLCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR, ContinuousPacketConnectivity_DRX_Information_ResponseLCR_sequence);

  return offset;
}


static const per_sequence_t Multiple_PLMN_List_sequence[] = {
  { &hf_rnsap_pLMN_Identity , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PLMN_Identity },
  { &hf_rnsap_list_Of_PLMNs , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_List_Of_PLMNs },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multiple_PLMN_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multiple_PLMN_List, Multiple_PLMN_List_sequence);

  return offset;
}


static const per_sequence_t UMTS_Cells_Info_List_sequence_of[1] = {
  { &hf_rnsap_UMTS_Cells_Info_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_UMTS_Cells_Info_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UMTS_Cells_Info_List, UMTS_Cells_Info_List_sequence_of,
                                                  0, maxNrOfCells, FALSE);

  return offset;
}


static const per_sequence_t Controlled_Object_Scope_sequence[] = {
  { &hf_rnsap_rNC_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RNC_ID },
  { &hf_rnsap_extended_RNC_ID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Extended_RNC_ID },
  { &hf_rnsap_multiple_PLMN_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Multiple_PLMN_List },
  { &hf_rnsap_umts_cell_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UMTS_Cells_Info_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Controlled_Object_Scope(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Controlled_Object_Scope, Controlled_Object_Scope_sequence);

  return offset;
}



static int
dissect_rnsap_DATA_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 3U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_8(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     8, 8, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t DBDS_CorrectionsReq_sequence[] = {
  { &hf_rnsap_dGANSS_Signal_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DBDS_CorrectionsReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DBDS_CorrectionsReq, DBDS_CorrectionsReq_sequence);

  return offset;
}



static int
dissect_rnsap_GANSS_Signal_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_63(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_13(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     13, 13, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t DBDS_SignalInfo_sequence[] = {
  { &hf_rnsap_svID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_bds_UDREI     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_15 },
  { &hf_rnsap_bds_RURAI     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_15 },
  { &hf_rnsap_bds_ECC_DeltaT, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DBDS_SignalInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DBDS_SignalInfo, DBDS_SignalInfo_sequence);

  return offset;
}


static const per_sequence_t DBDS_SignalInfoList_sequence_of[1] = {
  { &hf_rnsap_DBDS_SignalInfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DBDS_SignalInfo },
};

static int
dissect_rnsap_DBDS_SignalInfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DBDS_SignalInfoList, DBDS_SignalInfoList_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const per_sequence_t DBDS_Info_sequence[] = {
  { &hf_rnsap_dBDS_SignalID , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_Signal_ID },
  { &hf_rnsap_dBDS_SignalInfoList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DBDS_SignalInfoList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DBDS_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DBDS_Info, DBDS_Info_sequence);

  return offset;
}


static const per_sequence_t DBDS_InfoList_sequence_of[1] = {
  { &hf_rnsap_DBDS_InfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DBDS_Info },
};

static int
dissect_rnsap_DBDS_InfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DBDS_InfoList, DBDS_InfoList_sequence_of,
                                                  1, maxSgnType, FALSE);

  return offset;
}


static const per_sequence_t DBDS_Corrections_sequence[] = {
  { &hf_rnsap_bDS_RefTime   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_119 },
  { &hf_rnsap_dBDS_InfoList , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DBDS_InfoList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DBDS_Corrections(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DBDS_Corrections, DBDS_Corrections_sequence);

  return offset;
}



static int
dissect_rnsap_PowerOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 24U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_DL_FET_Mode_vals[] = {
  {   0, "basic" },
  {   1, "full" },
  { 0, NULL }
};


static int
dissect_rnsap_DL_FET_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_DCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t DCH_ENH_Concat_sequence_of[1] = {
  { &hf_rnsap_DCH_ENH_Concat_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_ID },
};

static int
dissect_rnsap_DCH_ENH_Concat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DCH_ENH_Concat, DCH_ENH_Concat_sequence_of,
                                                  1, maxNrofConcatenatedDCH, FALSE);

  return offset;
}


static const per_sequence_t DCH_ENH_Information_sequence[] = {
  { &hf_rnsap_pO2_SRB       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerOffset },
  { &hf_rnsap_dl_FET_Mode   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_FET_Mode },
  { &hf_rnsap_dCH_ENH_Concat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DCH_ENH_Concat },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_ENH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_ENH_Information, DCH_ENH_Information_sequence);

  return offset;
}


static const per_sequence_t DCH_ENH_Information_to_Modify_sequence[] = {
  { &hf_rnsap_pO_SRB        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PowerOffset },
  { &hf_rnsap_dl_FET_Mode   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_FET_Mode },
  { &hf_rnsap_dCH_ENH_Concat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DCH_ENH_Concat },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_ENH_Information_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_ENH_Information_to_Modify, DCH_ENH_Information_to_Modify_sequence);

  return offset;
}


static const value_string rnsap_DCH_ENH_Information_Removal_vals[] = {
  {   0, "remove" },
  { 0, NULL }
};


static int
dissect_rnsap_DCH_ENH_Information_Removal(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH_vals[] = {
  {   0, "setup" },
  {   1, "configurationChange" },
  {   2, "removal" },
  { 0, NULL }
};

static const per_choice_t Setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH_choice[] = {
  {   0, &hf_rnsap_setup_06      , ASN1_EXTENSION_ROOT    , dissect_rnsap_DCH_ENH_Information },
  {   1, &hf_rnsap_configurationChange_05, ASN1_EXTENSION_ROOT    , dissect_rnsap_DCH_ENH_Information_to_Modify },
  {   2, &hf_rnsap_removal_04    , ASN1_EXTENSION_ROOT    , dissect_rnsap_DCH_ENH_Information_Removal },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH, Setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t DCH_ENH_Information_Reconf_sequence[] = {
  { &hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_ENH_Information_Reconf(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_ENH_Information_Reconf, DCH_ENH_Information_Reconf_sequence);

  return offset;
}


static const value_string rnsap_PayloadCRC_PresenceIndicator_vals[] = {
  {   0, "crc-included" },
  {   1, "crc-not-included" },
  { 0, NULL }
};


static int
dissect_rnsap_PayloadCRC_PresenceIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_UL_FP_Mode_vals[] = {
  {   0, "normal" },
  {   1, "silent" },
  { 0, NULL }
};


static int
dissect_rnsap_UL_FP_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_ToAWS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1279U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_ToAWE(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 2559U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_TrCH_SrcStatisticsDescr_vals[] = {
  {   0, "speech" },
  {   1, "rRC" },
  {   2, "unknown" },
  { 0, NULL }
};


static int
dissect_rnsap_TrCH_SrcStatisticsDescr(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_NrOfTransportBlocks(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 512U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_TransportBlockSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 5000U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_TransmissionTimeIntervalDynamic_vals[] = {
  {   0, "msec-10" },
  {   1, "msec-20" },
  {   2, "msec-40" },
  {   3, "msec-80" },
  { 0, NULL }
};


static int
dissect_rnsap_TransmissionTimeIntervalDynamic(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t TransmissionTimeIntervalInformation_item_sequence[] = {
  { &hf_rnsap_transmissionTimeInterval, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransmissionTimeIntervalDynamic },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TransmissionTimeIntervalInformation_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TransmissionTimeIntervalInformation_item, TransmissionTimeIntervalInformation_item_sequence);

  return offset;
}


static const per_sequence_t TransmissionTimeIntervalInformation_sequence_of[1] = {
  { &hf_rnsap_TransmissionTimeIntervalInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TransmissionTimeIntervalInformation_item },
};

static int
dissect_rnsap_TransmissionTimeIntervalInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TransmissionTimeIntervalInformation, TransmissionTimeIntervalInformation_sequence_of,
                                                  1, maxTTI_Count, FALSE);

  return offset;
}


static const per_sequence_t TDD_TransportFormatSet_ModeDP_sequence[] = {
  { &hf_rnsap_transmissionTimeIntervalInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransmissionTimeIntervalInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_TransportFormatSet_ModeDP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_TransportFormatSet_ModeDP, TDD_TransportFormatSet_ModeDP_sequence);

  return offset;
}


static const value_string rnsap_TransportFormatSet_ModeDP_vals[] = {
  {   0, "tdd" },
  {   1, "notApplicable" },
  { 0, NULL }
};

static const per_choice_t TransportFormatSet_ModeDP_choice[] = {
  {   0, &hf_rnsap_tdd           , ASN1_EXTENSION_ROOT    , dissect_rnsap_TDD_TransportFormatSet_ModeDP },
  {   1, &hf_rnsap_notApplicable , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_TransportFormatSet_ModeDP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_TransportFormatSet_ModeDP, TransportFormatSet_ModeDP_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TransportFormatSet_DynamicPartList_item_sequence[] = {
  { &hf_rnsap_nrOfTransportBlocks, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_NrOfTransportBlocks },
  { &hf_rnsap_transportBlockSize, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportBlockSize },
  { &hf_rnsap_mode_01       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatSet_ModeDP },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TransportFormatSet_DynamicPartList_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TransportFormatSet_DynamicPartList_item, TransportFormatSet_DynamicPartList_item_sequence);

  return offset;
}


static const per_sequence_t TransportFormatSet_DynamicPartList_sequence_of[1] = {
  { &hf_rnsap_TransportFormatSet_DynamicPartList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatSet_DynamicPartList_item },
};

static int
dissect_rnsap_TransportFormatSet_DynamicPartList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TransportFormatSet_DynamicPartList, TransportFormatSet_DynamicPartList_sequence_of,
                                                  1, maxNrOfTFs, FALSE);

  return offset;
}


static const value_string rnsap_TransmissionTimeIntervalSemiStatic_vals[] = {
  {   0, "msec-10" },
  {   1, "msec-20" },
  {   2, "msec-40" },
  {   3, "msec-80" },
  {   4, "dynamic" },
  { 0, NULL }
};


static int
dissect_rnsap_TransmissionTimeIntervalSemiStatic(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_RateMatchingAttribute(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxRateMatching, NULL, FALSE);

  return offset;
}


static const value_string rnsap_SecondInterleavingMode_vals[] = {
  {   0, "frame-related" },
  {   1, "timeslot-related" },
  { 0, NULL }
};


static int
dissect_rnsap_SecondInterleavingMode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_TransportFormatSet_ModeSSP_vals[] = {
  {   0, "tdd" },
  {   1, "notApplicable" },
  { 0, NULL }
};

static const per_choice_t TransportFormatSet_ModeSSP_choice[] = {
  {   0, &hf_rnsap_tdd_01        , ASN1_EXTENSION_ROOT    , dissect_rnsap_SecondInterleavingMode },
  {   1, &hf_rnsap_notApplicable , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_TransportFormatSet_ModeSSP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_TransportFormatSet_ModeSSP, TransportFormatSet_ModeSSP_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TransportFormatSet_Semi_staticPart_sequence[] = {
  { &hf_rnsap_transmissionTime, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransmissionTimeIntervalSemiStatic },
  { &hf_rnsap_channelCoding , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ChannelCodingType },
  { &hf_rnsap_codingRate    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CodingRate },
  { &hf_rnsap_rateMatcingAttribute, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RateMatchingAttribute },
  { &hf_rnsap_cRC_Size      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CRC_Size },
  { &hf_rnsap_mode_02       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatSet_ModeSSP },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TransportFormatSet_Semi_staticPart(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TransportFormatSet_Semi_staticPart, TransportFormatSet_Semi_staticPart_sequence);

  return offset;
}


static const per_sequence_t TransportFormatSet_sequence[] = {
  { &hf_rnsap_dynamicParts  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatSet_DynamicPartList },
  { &hf_rnsap_semi_staticPart, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatSet_Semi_staticPart },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TransportFormatSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TransportFormatSet, TransportFormatSet_sequence);

  return offset;
}


static const value_string rnsap_FrameHandlingPriority_vals[] = {
  {   0, "lowest" },
  {  15, "highest" },
  { 0, NULL }
};


static int
dissect_rnsap_FrameHandlingPriority(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_QE_Selector_vals[] = {
  {   0, "selected" },
  {   1, "non-selected" },
  { 0, NULL }
};


static int
dissect_rnsap_QE_Selector(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_DRACControl_vals[] = {
  {   0, "not-Used-requested" },
  {   1, "not-requested" },
  { 0, NULL }
};


static int
dissect_rnsap_DRACControl(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DCH_Specific_FDD_Item_sequence[] = {
  { &hf_rnsap_dCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_ID },
  { &hf_rnsap_trCH_SrcStatisticsDescr, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TrCH_SrcStatisticsDescr },
  { &hf_rnsap_ul_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_dl_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_ul_BLER       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BLER },
  { &hf_rnsap_dl_BLER       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BLER },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_frameHandlingPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FrameHandlingPriority },
  { &hf_rnsap_qE_Selector   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_QE_Selector },
  { &hf_rnsap_dRACControl   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DRACControl },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_Specific_FDD_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_Specific_FDD_Item, DCH_Specific_FDD_Item_sequence);

  return offset;
}


static const per_sequence_t DCH_Specific_FDD_InformationList_sequence_of[1] = {
  { &hf_rnsap_DCH_Specific_FDD_InformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_Specific_FDD_Item },
};

static int
dissect_rnsap_DCH_Specific_FDD_InformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DCH_Specific_FDD_InformationList, DCH_Specific_FDD_InformationList_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t DCH_FDD_InformationItem_sequence[] = {
  { &hf_rnsap_payloadCRC_PresenceIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PayloadCRC_PresenceIndicator },
  { &hf_rnsap_ul_FP_Mode    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_FP_Mode },
  { &hf_rnsap_toAWS         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ToAWS },
  { &hf_rnsap_toAWE         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ToAWE },
  { &hf_rnsap_dCH_SpecificInformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_Specific_FDD_InformationList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_FDD_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_FDD_InformationItem, DCH_FDD_InformationItem_sequence);

  return offset;
}


static const per_sequence_t DCH_FDD_Information_sequence_of[1] = {
  { &hf_rnsap_DCH_FDD_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_FDD_InformationItem },
};

static int
dissect_rnsap_DCH_FDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DCH_FDD_Information, DCH_FDD_Information_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}



static int
dissect_rnsap_Pattern_Sequence_Identifier(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNrOfDCHMeasurementOccasionPatternSequence, NULL, FALSE);

  return offset;
}


static const value_string rnsap_Status_Flag_vals[] = {
  {   0, "activate" },
  {   1, "deactivate" },
  { 0, NULL }
};


static int
dissect_rnsap_Status_Flag(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_9(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 9U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_511(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_512(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 512U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_7(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     7, 7, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t Measurement_Occasion_Pattern_Sequence_parameters_sequence[] = {
  { &hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters_k, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_9 },
  { &hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters_offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_511 },
  { &hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters_M_Length, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_512 },
  { &hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters_Timeslot_Bitmap, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_7 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Measurement_Occasion_Pattern_Sequence_parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Measurement_Occasion_Pattern_Sequence_parameters, Measurement_Occasion_Pattern_Sequence_parameters_sequence);

  return offset;
}


static const per_sequence_t DchMeasurementOccasionInformation_Item_sequence[] = {
  { &hf_rnsap_pattern_Sequence_Identifier, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Pattern_Sequence_Identifier },
  { &hf_rnsap_status_Flag   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Status_Flag },
  { &hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Measurement_Occasion_Pattern_Sequence_parameters },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DchMeasurementOccasionInformation_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DchMeasurementOccasionInformation_Item, DchMeasurementOccasionInformation_Item_sequence);

  return offset;
}


static const per_sequence_t DCH_MeasurementOccasion_Information_sequence_of[1] = {
  { &hf_rnsap_DCH_MeasurementOccasion_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DchMeasurementOccasionInformation_Item },
};

static int
dissect_rnsap_DCH_MeasurementOccasion_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DCH_MeasurementOccasion_Information, DCH_MeasurementOccasion_Information_sequence_of,
                                                  1, maxNrOfDCHMeasurementOccasionPatternSequence, FALSE);

  return offset;
}



static int
dissect_rnsap_DCH_MeasurementType_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     5, 5, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const value_string rnsap_DCH_Indicator_For_E_DCH_HSDPA_Operation_vals[] = {
  {   0, "dch-not-present" },
  { 0, NULL }
};


static int
dissect_rnsap_DCH_Indicator_For_E_DCH_HSDPA_Operation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DCH_InformationResponseItem_sequence[] = {
  { &hf_rnsap_dCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_ID },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_InformationResponseItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_InformationResponseItem, DCH_InformationResponseItem_sequence);

  return offset;
}


static const per_sequence_t DCH_InformationResponse_sequence_of[1] = {
  { &hf_rnsap_DCH_InformationResponse_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_InformationResponseItem },
};

static int
dissect_rnsap_DCH_InformationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DCH_InformationResponse, DCH_InformationResponse_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t DCH_Specific_TDD_Item_sequence[] = {
  { &hf_rnsap_dCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_ID },
  { &hf_rnsap_ul_cCTrCH_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_dl_cCTrCH_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_trCH_SrcStatisticsDescr, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TrCH_SrcStatisticsDescr },
  { &hf_rnsap_ul_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_dl_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_ul_BLER       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BLER },
  { &hf_rnsap_dl_BLER       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BLER },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_frameHandlingPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FrameHandlingPriority },
  { &hf_rnsap_qE_Selector   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_QE_Selector },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_Specific_TDD_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_Specific_TDD_Item, DCH_Specific_TDD_Item_sequence);

  return offset;
}


static const per_sequence_t DCH_Specific_TDD_InformationList_sequence_of[1] = {
  { &hf_rnsap_DCH_Specific_TDD_InformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_Specific_TDD_Item },
};

static int
dissect_rnsap_DCH_Specific_TDD_InformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DCH_Specific_TDD_InformationList, DCH_Specific_TDD_InformationList_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t DCH_TDD_InformationItem_sequence[] = {
  { &hf_rnsap_payloadCRC_PresenceIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PayloadCRC_PresenceIndicator },
  { &hf_rnsap_ul_FP_Mode    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_FP_Mode },
  { &hf_rnsap_toAWS         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ToAWS },
  { &hf_rnsap_toAWE         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ToAWE },
  { &hf_rnsap_dCH_SpecificInformationList_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_Specific_TDD_InformationList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_TDD_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_TDD_InformationItem, DCH_TDD_InformationItem_sequence);

  return offset;
}


static const per_sequence_t DCH_TDD_Information_sequence_of[1] = {
  { &hf_rnsap_DCH_TDD_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_TDD_InformationItem },
};

static int
dissect_rnsap_DCH_TDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DCH_TDD_Information, DCH_TDD_Information_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const value_string rnsap_DecimationFactor_vals[] = {
  {   0, "slots3" },
  {   1, "slots5" },
  { 0, NULL }
};


static int
dissect_rnsap_DecimationFactor(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_DedicatedMeasurementType_vals[] = {
  {   0, "sir" },
  {   1, "sir-error" },
  {   2, "transmitted-code-power" },
  {   3, "rSCP" },
  {   4, "rx-timing-deviation" },
  {   5, "round-trip-time" },
  {   6, "rx-timing-deviation-LCR" },
  {   7, "angle-Of-Arrival-LCR" },
  {   8, "hs-sich-quality" },
  {   9, "rx-timing-deviation-768" },
  {  10, "rx-timing-deviation-ext" },
  {  11, "ue-transmission-power-headroom" },
  {  12, "dl-TBS" },
  { 0, NULL }
};


static int
dissect_rnsap_DedicatedMeasurementType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     6, NULL, TRUE, 7, NULL);

  return offset;
}



static int
dissect_rnsap_SIR_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_SIR_Error_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 125U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Transmitted_Code_Power_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_RSCP_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Rx_Timing_Deviation_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8191U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Round_Trip_Time_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 32767U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Extension_DedicatedMeasurementValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_DedicatedMeasurementValue_vals[] = {
  {   0, "sIR-Value" },
  {   1, "sIR-ErrorValue" },
  {   2, "transmittedCodePowerValue" },
  {   3, "rSCP" },
  {   4, "rxTimingDeviationValue" },
  {   5, "roundTripTime" },
  {   6, "extension-DedicatedMeasurementValue" },
  { 0, NULL }
};

static const per_choice_t DedicatedMeasurementValue_choice[] = {
  {   0, &hf_rnsap_sIR_Value     , ASN1_EXTENSION_ROOT    , dissect_rnsap_SIR_Value },
  {   1, &hf_rnsap_sIR_ErrorValue, ASN1_EXTENSION_ROOT    , dissect_rnsap_SIR_Error_Value },
  {   2, &hf_rnsap_transmittedCodePowerValue, ASN1_EXTENSION_ROOT    , dissect_rnsap_Transmitted_Code_Power_Value },
  {   3, &hf_rnsap_rSCP          , ASN1_EXTENSION_ROOT    , dissect_rnsap_RSCP_Value },
  {   4, &hf_rnsap_rxTimingDeviationValue, ASN1_EXTENSION_ROOT    , dissect_rnsap_Rx_Timing_Deviation_Value },
  {   5, &hf_rnsap_roundTripTime , ASN1_EXTENSION_ROOT    , dissect_rnsap_Round_Trip_Time_Value },
  {   6, &hf_rnsap_extension_DedicatedMeasurementValue, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_DedicatedMeasurementValue },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DedicatedMeasurementValue, DedicatedMeasurementValue_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t DedicatedMeasurementAvailable_sequence[] = {
  { &hf_rnsap_dedicatedmeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DedicatedMeasurementValue },
  { &hf_rnsap_cFN           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CFN },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementAvailable(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DedicatedMeasurementAvailable, DedicatedMeasurementAvailable_sequence);

  return offset;
}



static int
dissect_rnsap_DedicatedMeasurementnotAvailable(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_DedicatedMeasurementValueInformation_vals[] = {
  {   0, "measurementAvailable" },
  {   1, "measurementnotAvailable" },
  { 0, NULL }
};

static const per_choice_t DedicatedMeasurementValueInformation_choice[] = {
  {   0, &hf_rnsap_measurementAvailable_01, ASN1_NO_EXTENSIONS     , dissect_rnsap_DedicatedMeasurementAvailable },
  {   1, &hf_rnsap_measurementnotAvailable_01, ASN1_NO_EXTENSIONS     , dissect_rnsap_DedicatedMeasurementnotAvailable },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementValueInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DedicatedMeasurementValueInformation, DedicatedMeasurementValueInformation_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_DelayedActivation_vals[] = {
  {   0, "cfn" },
  {   1, "separate-indication" },
  { 0, NULL }
};

static const per_choice_t DelayedActivation_choice[] = {
  {   0, &hf_rnsap_cfn           , ASN1_NO_EXTENSIONS     , dissect_rnsap_CFN },
  {   1, &hf_rnsap_separate_indication, ASN1_NO_EXTENSIONS     , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DelayedActivation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DelayedActivation, DelayedActivation_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_Execution_Type_vals[] = {
  {   0, "synchronised" },
  {   1, "unsynchronised" },
  { 0, NULL }
};

static const per_choice_t Execution_Type_choice[] = {
  {   0, &hf_rnsap_synchronised  , ASN1_NO_EXTENSIONS     , dissect_rnsap_CFN },
  {   1, &hf_rnsap_unsynchronised, ASN1_NO_EXTENSIONS     , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Execution_Type(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Execution_Type, Execution_Type_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Activate_Info_sequence[] = {
  { &hf_rnsap_activation_type, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Execution_Type },
  { &hf_rnsap_initial_dl_tx_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_firstRLS_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_FirstRLS_Indicator },
  { &hf_rnsap_propagation_delay, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PropagationDelay },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Activate_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Activate_Info, Activate_Info_sequence);

  return offset;
}


static const per_sequence_t Deactivate_Info_sequence[] = {
  { &hf_rnsap_deactivation_type, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Execution_Type },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Deactivate_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Deactivate_Info, Deactivate_Info_sequence);

  return offset;
}


static const value_string rnsap_DelayedActivationUpdate_vals[] = {
  {   0, "activate" },
  {   1, "deactivate" },
  { 0, NULL }
};

static const per_choice_t DelayedActivationUpdate_choice[] = {
  {   0, &hf_rnsap_activate      , ASN1_NO_EXTENSIONS     , dissect_rnsap_Activate_Info },
  {   1, &hf_rnsap_deactivate_01 , ASN1_NO_EXTENSIONS     , dissect_rnsap_Deactivate_Info },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DelayedActivationUpdate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DelayedActivationUpdate, DelayedActivationUpdate_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_DeltaSIR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 30U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_GANSS_StatusHealth_vals[] = {
  {   0, "udre-scale-1dot0" },
  {   1, "udre-scale-0dot75" },
  {   2, "udre-scale-0dot5" },
  {   3, "udre-scale-0dot3" },
  {   4, "udre-scale-0dot2" },
  {   5, "udre-scale-0dot1" },
  {   6, "no-data" },
  {   7, "invalid-data" },
  { 0, NULL }
};


static int
dissect_rnsap_GANSS_StatusHealth(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_10(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     10, 10, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const value_string rnsap_UDRE_vals[] = {
  {   0, "lessThan1" },
  {   1, "between1-and-4" },
  {   2, "between4-and-8" },
  {   3, "over8" },
  { 0, NULL }
};


static int
dissect_rnsap_UDRE(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_M2047_2047(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -2047, 2047U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_M127_127(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -127, 127U, NULL, FALSE);

  return offset;
}


static const per_sequence_t T_dGANSS_SignalInformation_item_sequence[] = {
  { &hf_rnsap_satId         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_gANSS_iod     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_10 },
  { &hf_rnsap_udre          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UDRE },
  { &hf_rnsap_ganss_prc     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_M2047_2047 },
  { &hf_rnsap_ganss_rrc     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_M127_127 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_T_dGANSS_SignalInformation_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_T_dGANSS_SignalInformation_item, T_dGANSS_SignalInformation_item_sequence);

  return offset;
}


static const per_sequence_t T_dGANSS_SignalInformation_sequence_of[1] = {
  { &hf_rnsap_dGANSS_SignalInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_T_dGANSS_SignalInformation_item },
};

static int
dissect_rnsap_T_dGANSS_SignalInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_T_dGANSS_SignalInformation, T_dGANSS_SignalInformation_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const per_sequence_t T_dGANSS_Information_item_sequence[] = {
  { &hf_rnsap_gANSS_SignalId, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_Signal_ID },
  { &hf_rnsap_gANSS_StatusHealth, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_StatusHealth },
  { &hf_rnsap_dGANSS_SignalInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_T_dGANSS_SignalInformation },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_T_dGANSS_Information_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_T_dGANSS_Information_item, T_dGANSS_Information_item_sequence);

  return offset;
}


static const per_sequence_t T_dGANSS_Information_sequence_of[1] = {
  { &hf_rnsap_dGANSS_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_T_dGANSS_Information_item },
};

static int
dissect_rnsap_T_dGANSS_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_T_dGANSS_Information, T_dGANSS_Information_sequence_of,
                                                  1, maxSgnType, FALSE);

  return offset;
}


static const per_sequence_t DGANSSCorrections_sequence[] = {
  { &hf_rnsap_dGANSS_ReferenceTime, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_119 },
  { &hf_rnsap_dGANSS_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_dGANSS_Information },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DGANSSCorrections(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DGANSSCorrections, DGANSSCorrections_sequence);

  return offset;
}


static const per_sequence_t DGANSS_Corrections_Req_sequence[] = {
  { &hf_rnsap_dGANSS_Signal_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DGANSS_Corrections_Req(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DGANSS_Corrections_Req, DGANSS_Corrections_Req_sequence);

  return offset;
}


static const value_string rnsap_PRCDeviation_vals[] = {
  {   0, "prcd1" },
  {   1, "prcd2" },
  {   2, "prcd5" },
  {   3, "prcd10" },
  { 0, NULL }
};


static int
dissect_rnsap_PRCDeviation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t DGANSSThreshold_sequence[] = {
  { &hf_rnsap_pRCDeviation  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PRCDeviation },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DGANSSThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DGANSSThreshold, DGANSSThreshold_sequence);

  return offset;
}


static const value_string rnsap_UDREGrowthRate_vals[] = {
  {   0, "growth-1-point-5" },
  {   1, "growth-2" },
  {   2, "growth-4" },
  {   3, "growth-6" },
  {   4, "growth-8" },
  {   5, "growth-10" },
  {   6, "growth-12" },
  {   7, "growth-16" },
  { 0, NULL }
};


static int
dissect_rnsap_UDREGrowthRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_UDREValidityTime_vals[] = {
  {   0, "val-20sec" },
  {   1, "val-40sec" },
  {   2, "val-80sec" },
  {   3, "val-160sec" },
  {   4, "val-320sec" },
  {   5, "val-640sec" },
  {   6, "val-1280sec" },
  {   7, "val-2560sec" },
  { 0, NULL }
};


static int
dissect_rnsap_UDREValidityTime(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DGNSS_ValidityPeriod_sequence[] = {
  { &hf_rnsap_udreGrowthRate, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UDREGrowthRate },
  { &hf_rnsap_udreValidityTime, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UDREValidityTime },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DGNSS_ValidityPeriod(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DGNSS_ValidityPeriod, DGNSS_ValidityPeriod_sequence);

  return offset;
}



static int
dissect_rnsap_GPSTOW(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 604799U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_GPS_Status_Health_vals[] = {
  {   0, "udre-1-0" },
  {   1, "udre-0-75" },
  {   2, "udre-0-5" },
  {   3, "udre-0-3" },
  {   4, "udre-0-1" },
  {   5, "no-data" },
  {   6, "invalid-data" },
  { 0, NULL }
};


static int
dissect_rnsap_GPS_Status_Health(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_PRC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -2047, 2047U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Range_Correction_Rate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -127, 127U, NULL, FALSE);

  return offset;
}


static const per_sequence_t T_satellite_DGPSCorrections_Information_item_sequence[] = {
  { &hf_rnsap_sAT_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAT_ID },
  { &hf_rnsap_iode_dgps     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_uDRE          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UDRE },
  { &hf_rnsap_pRC           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PRC },
  { &hf_rnsap_range_Correction_Rate, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Range_Correction_Rate },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_T_satellite_DGPSCorrections_Information_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_T_satellite_DGPSCorrections_Information_item, T_satellite_DGPSCorrections_Information_item_sequence);

  return offset;
}


static const per_sequence_t T_satellite_DGPSCorrections_Information_sequence_of[1] = {
  { &hf_rnsap_satellite_DGPSCorrections_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_T_satellite_DGPSCorrections_Information_item },
};

static int
dissect_rnsap_T_satellite_DGPSCorrections_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_T_satellite_DGPSCorrections_Information, T_satellite_DGPSCorrections_Information_sequence_of,
                                                  1, maxNoSat, FALSE);

  return offset;
}


static const per_sequence_t DGPSCorrections_sequence[] = {
  { &hf_rnsap_gPSTOW        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GPSTOW },
  { &hf_rnsap_gPS_Status_Health, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GPS_Status_Health },
  { &hf_rnsap_satellite_DGPSCorrections_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_satellite_DGPSCorrections_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DGPSCorrections(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DGPSCorrections, DGPSCorrections_sequence);

  return offset;
}


static const per_sequence_t DGPSThreshold_sequence[] = {
  { &hf_rnsap_pRCDeviation  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PRCDeviation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DGPSThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DGPSThreshold, DGPSThreshold_sequence);

  return offset;
}


static const value_string rnsap_DiscardTimer_vals[] = {
  {   0, "v20" },
  {   1, "v40" },
  {   2, "v60" },
  {   3, "v80" },
  {   4, "v100" },
  {   5, "v120" },
  {   6, "v140" },
  {   7, "v160" },
  {   8, "v180" },
  {   9, "v200" },
  {  10, "v250" },
  {  11, "v300" },
  {  12, "v400" },
  {  13, "v500" },
  {  14, "v750" },
  {  15, "v1000" },
  {  16, "v1250" },
  {  17, "v1500" },
  {  18, "v1750" },
  {  19, "v2000" },
  {  20, "v2500" },
  {  21, "v3000" },
  {  22, "v3500" },
  {  23, "v4000" },
  {  24, "v4500" },
  {  25, "v5000" },
  {  26, "v7500" },
  { 0, NULL }
};

static value_string_ext rnsap_DiscardTimer_vals_ext = VALUE_STRING_EXT_INIT(rnsap_DiscardTimer_vals);


static int
dissect_rnsap_DiscardTimer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     27, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_DiversityControlField_vals[] = {
  {   0, "may" },
  {   1, "must" },
  {   2, "must-not" },
  { 0, NULL }
};


static int
dissect_rnsap_DiversityControlField(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_DiversityMode_vals[] = {
  {   0, "none" },
  {   1, "sTTD" },
  {   2, "closedLoopMode1" },
  {   3, "not-used-closedLoopMode2" },
  { 0, NULL }
};


static int
dissect_rnsap_DiversityMode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_DL_TBS_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 160000U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_DSCH_Cell_sequence[] = {
  { &hf_rnsap_hs_DSCH_Cell_TBS, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TBS_Value },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_Cell(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_DSCH_Cell, HS_DSCH_Cell_sequence);

  return offset;
}


static const per_sequence_t HS_DSCH_Cell_List_sequence_of[1] = {
  { &hf_rnsap_HS_DSCH_Cell_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_DSCH_Cell },
};

static int
dissect_rnsap_HS_DSCH_Cell_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HS_DSCH_Cell_List, HS_DSCH_Cell_List_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t DL_TBS_sequence[] = {
  { &hf_rnsap_hs_DSCH_Cell_List, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_DSCH_Cell_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_TBS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_TBS, DL_TBS_sequence);

  return offset;
}



static int
dissect_rnsap_DL_DPCH_SlotFormat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_DL_DPCH_TimingAdjustment_vals[] = {
  {   0, "timing-advance" },
  {   1, "timing-delay" },
  { 0, NULL }
};


static int
dissect_rnsap_DL_DPCH_TimingAdjustment(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_PowerAdjustmentType_vals[] = {
  {   0, "none" },
  {   1, "common" },
  {   2, "individual" },
  { 0, NULL }
};


static int
dissect_rnsap_PowerAdjustmentType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t DL_ReferencePowerInformationItem_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_dl_Reference_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_ReferencePowerInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_ReferencePowerInformationItem, DL_ReferencePowerInformationItem_sequence);

  return offset;
}


static const per_sequence_t DL_ReferencePowerInformationList_sequence_of[1] = {
  { &hf_rnsap_DL_ReferencePowerInformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_ReferencePowerInformationItem },
};

static int
dissect_rnsap_DL_ReferencePowerInformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_ReferencePowerInformationList, DL_ReferencePowerInformationList_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}



static int
dissect_rnsap_MaxAdjustmentStep(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 10U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_ScaledAdjustmentRatio(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}


static const per_sequence_t DL_PowerBalancing_Information_sequence[] = {
  { &hf_rnsap_powerAdjustmentType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerAdjustmentType },
  { &hf_rnsap_dLReferencePower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_dLReferencePowerList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_ReferencePowerInformationList },
  { &hf_rnsap_maxAdjustmentStep, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MaxAdjustmentStep },
  { &hf_rnsap_adjustmentPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AdjustmentPeriod },
  { &hf_rnsap_adjustmentRatio, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ScaledAdjustmentRatio },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_PowerBalancing_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_PowerBalancing_Information, DL_PowerBalancing_Information_sequence);

  return offset;
}


static const per_sequence_t DL_ReferencePowerInformation_sequence[] = {
  { &hf_rnsap_common_DL_ReferencePowerInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_individual_DL_ReferencePowerInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_ReferencePowerInformationList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_ReferencePowerInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_ReferencePowerInformation, DL_ReferencePowerInformation_sequence);

  return offset;
}



static int
dissect_rnsap_D_RNTI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1048575U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_D_RNTI_ReleaseIndication_vals[] = {
  {   0, "release-D-RNTI" },
  {   1, "not-release-D-RNTI" },
  { 0, NULL }
};


static int
dissect_rnsap_D_RNTI_ReleaseIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_DL_FrameType_vals[] = {
  {   0, "typeA" },
  {   1, "typeB" },
  { 0, NULL }
};


static int
dissect_rnsap_DL_FrameType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_MidambleConfigurationBurstType1And3_vals[] = {
  {   0, "v4" },
  {   1, "v8" },
  {   2, "v16" },
  { 0, NULL }
};


static int
dissect_rnsap_MidambleConfigurationBurstType1And3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_MidambleShiftLong(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_MidambleAllocationMode1_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "commonMidamble" },
  {   2, "ueSpecificMidamble" },
  { 0, NULL }
};

static const per_choice_t MidambleAllocationMode1_choice[] = {
  {   0, &hf_rnsap_defaultMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_commonMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   2, &hf_rnsap_ueSpecificMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_MidambleShiftLong },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_MidambleAllocationMode1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_MidambleAllocationMode1, MidambleAllocationMode1_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Type1_sequence[] = {
  { &hf_rnsap_midambleConfigurationBurstType1And3, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleConfigurationBurstType1And3 },
  { &hf_rnsap_midambleAllocationMode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleAllocationMode1 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Type1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Type1, Type1_sequence);

  return offset;
}


static const value_string rnsap_MidambleConfigurationBurstType2_vals[] = {
  {   0, "v3" },
  {   1, "v6" },
  { 0, NULL }
};


static int
dissect_rnsap_MidambleConfigurationBurstType2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_MidambleShiftShort(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 5U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_MidambleAllocationMode2_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "commonMidamble" },
  {   2, "ueSpecificMidamble" },
  { 0, NULL }
};

static const per_choice_t MidambleAllocationMode2_choice[] = {
  {   0, &hf_rnsap_defaultMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_commonMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   2, &hf_rnsap_ueSpecificMidamble_01, ASN1_EXTENSION_ROOT    , dissect_rnsap_MidambleShiftShort },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_MidambleAllocationMode2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_MidambleAllocationMode2, MidambleAllocationMode2_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Type2_sequence[] = {
  { &hf_rnsap_midambleConfigurationBurstType2, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleConfigurationBurstType2 },
  { &hf_rnsap_midambleAllocationMode_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleAllocationMode2 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Type2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Type2, Type2_sequence);

  return offset;
}


static const value_string rnsap_MidambleAllocationMode3_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "ueSpecificMidamble" },
  { 0, NULL }
};

static const per_choice_t MidambleAllocationMode3_choice[] = {
  {   0, &hf_rnsap_defaultMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_ueSpecificMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_MidambleShiftLong },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_MidambleAllocationMode3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_MidambleAllocationMode3, MidambleAllocationMode3_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Type3_sequence[] = {
  { &hf_rnsap_midambleConfigurationBurstType1And3, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleConfigurationBurstType1And3 },
  { &hf_rnsap_midambleAllocationMode_02, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleAllocationMode3 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Type3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Type3, Type3_sequence);

  return offset;
}


static const value_string rnsap_MidambleShiftAndBurstType_vals[] = {
  {   0, "type1" },
  {   1, "type2" },
  {   2, "type3" },
  { 0, NULL }
};

static const per_choice_t MidambleShiftAndBurstType_choice[] = {
  {   0, &hf_rnsap_type1         , ASN1_EXTENSION_ROOT    , dissect_rnsap_Type1 },
  {   1, &hf_rnsap_type2         , ASN1_EXTENSION_ROOT    , dissect_rnsap_Type2 },
  {   2, &hf_rnsap_type3         , ASN1_EXTENSION_ROOT    , dissect_rnsap_Type3 },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_MidambleShiftAndBurstType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_MidambleShiftAndBurstType, MidambleShiftAndBurstType_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_TFCI_Presence_vals[] = {
  {   0, "present" },
  {   1, "not-present" },
  { 0, NULL }
};


static int
dissect_rnsap_TFCI_Presence(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_DPCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 239U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_TDD_ChannelisationCode_vals[] = {
  {   0, "chCode1div1" },
  {   1, "chCode2div1" },
  {   2, "chCode2div2" },
  {   3, "chCode4div1" },
  {   4, "chCode4div2" },
  {   5, "chCode4div3" },
  {   6, "chCode4div4" },
  {   7, "chCode8div1" },
  {   8, "chCode8div2" },
  {   9, "chCode8div3" },
  {  10, "chCode8div4" },
  {  11, "chCode8div5" },
  {  12, "chCode8div6" },
  {  13, "chCode8div7" },
  {  14, "chCode8div8" },
  {  15, "chCode16div1" },
  {  16, "chCode16div2" },
  {  17, "chCode16div3" },
  {  18, "chCode16div4" },
  {  19, "chCode16div5" },
  {  20, "chCode16div6" },
  {  21, "chCode16div7" },
  {  22, "chCode16div8" },
  {  23, "chCode16div9" },
  {  24, "chCode16div10" },
  {  25, "chCode16div11" },
  {  26, "chCode16div12" },
  {  27, "chCode16div13" },
  {  28, "chCode16div14" },
  {  29, "chCode16div15" },
  {  30, "chCode16div16" },
  { 0, NULL }
};

static value_string_ext rnsap_TDD_ChannelisationCode_vals_ext = VALUE_STRING_EXT_INIT(rnsap_TDD_ChannelisationCode_vals);


static int
dissect_rnsap_TDD_ChannelisationCode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     31, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t TDD_DL_Code_InformationItem_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_DL_Code_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_DL_Code_InformationItem, TDD_DL_Code_InformationItem_sequence);

  return offset;
}


static const per_sequence_t TDD_DL_Code_Information_sequence_of[1] = {
  { &hf_rnsap_TDD_DL_Code_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DL_Code_InformationItem },
};

static int
dissect_rnsap_TDD_DL_Code_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_DL_Code_Information, TDD_DL_Code_Information_sequence_of,
                                                  1, maxNrOfDPCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationItem_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_dL_Code_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DL_Code_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_Timeslot_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_Timeslot_InformationItem, DL_Timeslot_InformationItem_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_Information_sequence_of[1] = {
  { &hf_rnsap_DL_Timeslot_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Timeslot_InformationItem },
};

static int
dissect_rnsap_DL_Timeslot_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_Timeslot_Information, DL_Timeslot_Information_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}



static int
dissect_rnsap_TimeSlotLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 6U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_MidambleAllocationMode_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "commonMidamble" },
  {   2, "uESpecificMidamble" },
  { 0, NULL }
};


static int
dissect_rnsap_MidambleAllocationMode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_MidambleConfigurationLCR_vals[] = {
  {   0, "v2" },
  {   1, "v4" },
  {   2, "v6" },
  {   3, "v8" },
  {   4, "v10" },
  {   5, "v12" },
  {   6, "v14" },
  {   7, "v16" },
  { 0, NULL }
};


static int
dissect_rnsap_MidambleConfigurationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t MidambleShiftLCR_sequence[] = {
  { &hf_rnsap_midambleAllocationMode_03, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleAllocationMode },
  { &hf_rnsap_midambleShift_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftLong },
  { &hf_rnsap_midambleConfigurationLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleConfigurationLCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MidambleShiftLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MidambleShiftLCR, MidambleShiftLCR_sequence);

  return offset;
}


static const value_string rnsap_Modulation_vals[] = {
  {   0, "qPSK" },
  {   1, "eightPSK" },
  { 0, NULL }
};


static int
dissect_rnsap_Modulation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t TDD_ChannelisationCodeLCR_sequence[] = {
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_modulation    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Modulation },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_ChannelisationCodeLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_ChannelisationCodeLCR, TDD_ChannelisationCodeLCR_sequence);

  return offset;
}



static int
dissect_rnsap_QPSK_DL_DPCH_TimeSlotFormatTDD_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 24U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_EightPSK_DL_DPCH_TimeSlotFormatTDD_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 24U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR_vals[] = {
  {   0, "qPSK" },
  {   1, "eightPSK" },
  { 0, NULL }
};

static const per_choice_t TDD_DL_DPCH_TimeSlotFormat_LCR_choice[] = {
  {   0, &hf_rnsap_qPSK          , ASN1_EXTENSION_ROOT    , dissect_rnsap_QPSK_DL_DPCH_TimeSlotFormatTDD_LCR },
  {   1, &hf_rnsap_eightPSK      , ASN1_EXTENSION_ROOT    , dissect_rnsap_EightPSK_DL_DPCH_TimeSlotFormatTDD_LCR },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR, TDD_DL_DPCH_TimeSlotFormat_LCR_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TDD_DL_Code_LCR_InformationItem_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCodeLCR },
  { &hf_rnsap_tdd_DL_DPCH_TimeSlotFormat_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_DL_Code_LCR_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_DL_Code_LCR_InformationItem, TDD_DL_Code_LCR_InformationItem_sequence);

  return offset;
}


static const per_sequence_t TDD_DL_Code_LCR_Information_sequence_of[1] = {
  { &hf_rnsap_TDD_DL_Code_LCR_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DL_Code_LCR_InformationItem },
};

static int
dissect_rnsap_TDD_DL_Code_LCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_DL_Code_LCR_Information, TDD_DL_Code_LCR_Information_sequence_of,
                                                  1, maxNrOfDPCHsLCR, FALSE);

  return offset;
}


static const per_sequence_t DL_TimeslotLCR_InformationItem_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_dL_Code_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DL_Code_LCR_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_TimeslotLCR_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_TimeslotLCR_InformationItem, DL_TimeslotLCR_InformationItem_sequence);

  return offset;
}


static const per_sequence_t DL_TimeslotLCR_Information_sequence_of[1] = {
  { &hf_rnsap_DL_TimeslotLCR_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TimeslotLCR_InformationItem },
};

static int
dissect_rnsap_DL_TimeslotLCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_TimeslotLCR_Information, DL_TimeslotLCR_Information_sequence_of,
                                                  1, maxNrOfDLTsLCR, FALSE);

  return offset;
}


static const value_string rnsap_MidambleAllocationMode7681_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "commonMidamble" },
  {   2, "ueSpecificMidamble" },
  { 0, NULL }
};

static const per_choice_t MidambleAllocationMode7681_choice[] = {
  {   0, &hf_rnsap_defaultMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_commonMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   2, &hf_rnsap_ueSpecificMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_MidambleShiftLong },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_MidambleAllocationMode7681(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_MidambleAllocationMode7681, MidambleAllocationMode7681_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Type7681_sequence[] = {
  { &hf_rnsap_midambleConfigurationBurstType1And3, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleConfigurationBurstType1And3 },
  { &hf_rnsap_midambleAllocationMode_04, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleAllocationMode7681 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Type7681(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Type7681, Type7681_sequence);

  return offset;
}


static const value_string rnsap_MidambleConfigurationBurstType2_768_vals[] = {
  {   0, "v4" },
  {   1, "v8" },
  { 0, NULL }
};


static int
dissect_rnsap_MidambleConfigurationBurstType2_768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_MidambleShiftShort768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_MidambleAllocationMode7682_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "commonMidamble" },
  {   2, "ueSpecificMidamble" },
  { 0, NULL }
};

static const per_choice_t MidambleAllocationMode7682_choice[] = {
  {   0, &hf_rnsap_defaultMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_commonMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   2, &hf_rnsap_ueSpecificMidamble_02, ASN1_EXTENSION_ROOT    , dissect_rnsap_MidambleShiftShort768 },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_MidambleAllocationMode7682(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_MidambleAllocationMode7682, MidambleAllocationMode7682_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Type7682_sequence[] = {
  { &hf_rnsap_midambleConfigurationBurstType2_768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleConfigurationBurstType2_768 },
  { &hf_rnsap_midambleAllocationMode_05, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleAllocationMode7682 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Type7682(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Type7682, Type7682_sequence);

  return offset;
}


static const value_string rnsap_MidambleAllocationMode7683_vals[] = {
  {   0, "defaultMidamble" },
  {   1, "ueSpecificMidamble" },
  { 0, NULL }
};

static const per_choice_t MidambleAllocationMode7683_choice[] = {
  {   0, &hf_rnsap_defaultMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_ueSpecificMidamble, ASN1_EXTENSION_ROOT    , dissect_rnsap_MidambleShiftLong },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_MidambleAllocationMode7683(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_MidambleAllocationMode7683, MidambleAllocationMode7683_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Type7683_sequence[] = {
  { &hf_rnsap_midambleConfigurationBurstType1And3, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleConfigurationBurstType1And3 },
  { &hf_rnsap_midambleAllocationMode_06, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleAllocationMode7683 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Type7683(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Type7683, Type7683_sequence);

  return offset;
}


static const value_string rnsap_MidambleShiftAndBurstType768_vals[] = {
  {   0, "type1" },
  {   1, "type2" },
  {   2, "type3" },
  { 0, NULL }
};

static const per_choice_t MidambleShiftAndBurstType768_choice[] = {
  {   0, &hf_rnsap_type1_01      , ASN1_EXTENSION_ROOT    , dissect_rnsap_Type7681 },
  {   1, &hf_rnsap_type2_01      , ASN1_EXTENSION_ROOT    , dissect_rnsap_Type7682 },
  {   2, &hf_rnsap_type3_01      , ASN1_EXTENSION_ROOT    , dissect_rnsap_Type7683 },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_MidambleShiftAndBurstType768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_MidambleShiftAndBurstType768, MidambleShiftAndBurstType768_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_TDD_ChannelisationCode768_vals[] = {
  {   0, "chCode1div1" },
  {   1, "chCode2div1" },
  {   2, "chCode2div2" },
  {   3, "chCode4div1" },
  {   4, "chCode4div2" },
  {   5, "chCode4div3" },
  {   6, "chCode4div4" },
  {   7, "chCode8div1" },
  {   8, "chCode8div2" },
  {   9, "chCode8div3" },
  {  10, "chCode8div4" },
  {  11, "chCode8div5" },
  {  12, "chCode8div6" },
  {  13, "chCode8div7" },
  {  14, "chCode8div8" },
  {  15, "chCode16div1" },
  {  16, "chCode16div2" },
  {  17, "chCode16div3" },
  {  18, "chCode16div4" },
  {  19, "chCode16div5" },
  {  20, "chCode16div6" },
  {  21, "chCode16div7" },
  {  22, "chCode16div8" },
  {  23, "chCode16div9" },
  {  24, "chCode16div10" },
  {  25, "chCode16div11" },
  {  26, "chCode16div12" },
  {  27, "chCode16div13" },
  {  28, "chCode16div14" },
  {  29, "chCode16div15" },
  {  30, "chCode16div16" },
  {  31, "chCode32div1" },
  {  32, "chCode32div2" },
  {  33, "chCode32div3" },
  {  34, "chCode32div4" },
  {  35, "chCode32div5" },
  {  36, "chCode32div6" },
  {  37, "chCode32div7" },
  {  38, "chCode32div8" },
  {  39, "chCode32div9" },
  {  40, "chCode32div10" },
  {  41, "chCode32div11" },
  {  42, "chCode32div12" },
  {  43, "chCode32div13" },
  {  44, "chCode32div14" },
  {  45, "chCode32div15" },
  {  46, "chCode32div16" },
  {  47, "chCode32div17" },
  {  48, "chCode32div18" },
  {  49, "chCode32div19" },
  {  50, "chCode32div20" },
  {  51, "chCode32div21" },
  {  52, "chCode32div22" },
  {  53, "chCode32div23" },
  {  54, "chCode32div24" },
  {  55, "chCode32div25" },
  {  56, "chCode32div26" },
  {  57, "chCode32div27" },
  {  58, "chCode32div28" },
  {  59, "chCode32div29" },
  {  60, "chCode32div30" },
  {  61, "chCode32div31" },
  {  62, "chCode32div32" },
  { 0, NULL }
};

static value_string_ext rnsap_TDD_ChannelisationCode768_vals_ext = VALUE_STRING_EXT_INIT(rnsap_TDD_ChannelisationCode768_vals);


static int
dissect_rnsap_TDD_ChannelisationCode768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     63, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t TDD_DL_Code_InformationItem768_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_tDD_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_DL_Code_InformationItem768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_DL_Code_InformationItem768, TDD_DL_Code_InformationItem768_sequence);

  return offset;
}


static const per_sequence_t TDD_DL_Code_Information768_sequence_of[1] = {
  { &hf_rnsap_TDD_DL_Code_Information768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DL_Code_InformationItem768 },
};

static int
dissect_rnsap_TDD_DL_Code_Information768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_DL_Code_Information768, TDD_DL_Code_Information768_sequence_of,
                                                  1, maxNrOfDPCHs768, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationItem768_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_dL_Code_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DL_Code_Information768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_Timeslot_InformationItem768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_Timeslot_InformationItem768, DL_Timeslot_InformationItem768_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_Information768_sequence_of[1] = {
  { &hf_rnsap_DL_Timeslot_Information768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Timeslot_InformationItem768 },
};

static int
dissect_rnsap_DL_Timeslot_Information768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_Timeslot_Information768, DL_Timeslot_Information768_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}



static int
dissect_rnsap_DL_TimeslotISCP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 91U, NULL, FALSE);

  return offset;
}


static const per_sequence_t DL_TimeSlot_ISCP_InfoItem_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_dL_TimeslotISCP, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TimeslotISCP },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_TimeSlot_ISCP_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_TimeSlot_ISCP_InfoItem, DL_TimeSlot_ISCP_InfoItem_sequence);

  return offset;
}


static const per_sequence_t DL_TimeSlot_ISCP_Info_sequence_of[1] = {
  { &hf_rnsap_DL_TimeSlot_ISCP_Info_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TimeSlot_ISCP_InfoItem },
};

static int
dissect_rnsap_DL_TimeSlot_ISCP_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_TimeSlot_ISCP_Info, DL_TimeSlot_ISCP_Info_sequence_of,
                                                  1, maxNrOfDLTs, FALSE);

  return offset;
}


static const per_sequence_t DL_TimeSlot_ISCP_LCR_InfoItem_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_dL_TimeslotISCP, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TimeslotISCP },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_TimeSlot_ISCP_LCR_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_TimeSlot_ISCP_LCR_InfoItem, DL_TimeSlot_ISCP_LCR_InfoItem_sequence);

  return offset;
}


static const per_sequence_t DL_TimeSlot_ISCP_LCR_Information_sequence_of[1] = {
  { &hf_rnsap_DL_TimeSlot_ISCP_LCR_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TimeSlot_ISCP_LCR_InfoItem },
};

static int
dissect_rnsap_DL_TimeSlot_ISCP_LCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_TimeSlot_ISCP_LCR_Information, DL_TimeSlot_ISCP_LCR_Information_sequence_of,
                                                  1, maxNrOfDLTsLCR, FALSE);

  return offset;
}


static const value_string rnsap_Downlink_Compressed_Mode_Method_vals[] = {
  {   0, "not-Used-puncturing" },
  {   1, "sFdiv2" },
  {   2, "higher-layer-scheduling" },
  { 0, NULL }
};


static int
dissect_rnsap_Downlink_Compressed_Mode_Method(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Downlink_TPC_enhancements_Information_sequence[] = {
  { &hf_rnsap_decimationFactorforPrimaryFrequency, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DecimationFactor },
  { &hf_rnsap_decimationFactorforSecFrequency, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DecimationFactor },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Downlink_TPC_enhancements_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Downlink_TPC_enhancements_Information, Downlink_TPC_enhancements_Information_sequence);

  return offset;
}


static const value_string rnsap_Downlink_TPC_enhancements_Information_Removal_vals[] = {
  {   0, "remove" },
  { 0, NULL }
};


static int
dissect_rnsap_Downlink_TPC_enhancements_Information_Removal(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements_vals[] = {
  {   0, "setup" },
  {   1, "configurationChange" },
  {   2, "removal" },
  { 0, NULL }
};

static const per_choice_t Setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements_choice[] = {
  {   0, &hf_rnsap_setup_03      , ASN1_EXTENSION_ROOT    , dissect_rnsap_Downlink_TPC_enhancements_Information },
  {   1, &hf_rnsap_configurationChange_02, ASN1_EXTENSION_ROOT    , dissect_rnsap_Downlink_TPC_enhancements_Information },
  {   2, &hf_rnsap_removal_01    , ASN1_EXTENSION_ROOT    , dissect_rnsap_Downlink_TPC_enhancements_Information_Removal },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements, Setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Downlink_TPC_enhancements_Reconf_sequence[] = {
  { &hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Downlink_TPC_enhancements_Reconf(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Downlink_TPC_enhancements_Reconf, Downlink_TPC_enhancements_Reconf_sequence);

  return offset;
}


static const value_string rnsap_DPC_Mode_vals[] = {
  {   0, "mode0" },
  {   1, "mode1" },
  { 0, NULL }
};


static int
dissect_rnsap_DPC_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_DPCH_ID768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 479U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_DRXCycleLengthCoefficient(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            3U, 9U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_DSCH_RNTI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_SchedulingPriorityIndicator_vals[] = {
  {   0, "lowest" },
  {  15, "highest" },
  { 0, NULL }
};


static int
dissect_rnsap_SchedulingPriorityIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_MAC_c_sh_SDU_Length(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 5000U, NULL, FALSE);

  return offset;
}


static const per_sequence_t MAC_c_sh_SDU_LengthList_sequence_of[1] = {
  { &hf_rnsap_MAC_c_sh_SDU_LengthList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MAC_c_sh_SDU_Length },
};

static int
dissect_rnsap_MAC_c_sh_SDU_LengthList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MAC_c_sh_SDU_LengthList, MAC_c_sh_SDU_LengthList_sequence_of,
                                                  1, maxNrOfMACcshSDU_Length, FALSE);

  return offset;
}


static const per_sequence_t DSCH_FlowControlItem_sequence[] = {
  { &hf_rnsap_dSCH_SchedulingPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_mAC_c_sh_SDU_Lengths, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAC_c_sh_SDU_LengthList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DSCH_FlowControlItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DSCH_FlowControlItem, DSCH_FlowControlItem_sequence);

  return offset;
}


static const per_sequence_t DSCH_FlowControlInformation_sequence_of[1] = {
  { &hf_rnsap_DSCH_FlowControlInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_FlowControlItem },
};

static int
dissect_rnsap_DSCH_FlowControlInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DSCH_FlowControlInformation, DSCH_FlowControlInformation_sequence_of,
                                                  1, 16, FALSE);

  return offset;
}



static int
dissect_rnsap_DSCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_DSCH_InitialWindowSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t DSCH_TDD_InformationItem_sequence[] = {
  { &hf_rnsap_dSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_ID },
  { &hf_rnsap_dl_ccTrCHID   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_trChSourceStatisticsDescriptor, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TrCH_SrcStatisticsDescr },
  { &hf_rnsap_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_bLER          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BLER },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DSCH_TDD_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DSCH_TDD_InformationItem, DSCH_TDD_InformationItem_sequence);

  return offset;
}


static const per_sequence_t DSCH_TDD_Information_sequence_of[1] = {
  { &hf_rnsap_DSCH_TDD_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_TDD_InformationItem },
};

static int
dissect_rnsap_DSCH_TDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DSCH_TDD_Information, DSCH_TDD_Information_sequence_of,
                                                  1, maxNoOfDSCHs, FALSE);

  return offset;
}



static int
dissect_rnsap_DsField(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     8, 8, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_EARFCN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxEARFCN, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_EARFCN_Extended(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxEARFCN_Extended, NULL, TRUE);

  return offset;
}


static const per_sequence_t EARFCN_FDD_sequence[] = {
  { &hf_rnsap_uL_EARFCN     , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_EARFCN },
  { &hf_rnsap_dL_EARFCN     , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_EARFCN },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EARFCN_FDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EARFCN_FDD, EARFCN_FDD_sequence);

  return offset;
}



static int
dissect_rnsap_Extension_EARFCN_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_EARFCN_Information_vals[] = {
  {   0, "fDD" },
  {   1, "tDD" },
  {   2, "extension-EARFCN-Information" },
  { 0, NULL }
};

static const per_choice_t EARFCN_Information_choice[] = {
  {   0, &hf_rnsap_fDD           , ASN1_EXTENSION_ROOT    , dissect_rnsap_EARFCN_FDD },
  {   1, &hf_rnsap_tDD           , ASN1_EXTENSION_ROOT    , dissect_rnsap_EARFCN },
  {   2, &hf_rnsap_extension_EARFCN_Information, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_EARFCN_Information },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_EARFCN_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_EARFCN_Information, EARFCN_Information_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t EARFCN_FDD_Extended_sequence[] = {
  { &hf_rnsap_uL_EARFCN_01  , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_EARFCN_Extended },
  { &hf_rnsap_dL_EARFCN_01  , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_EARFCN_Extended },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EARFCN_FDD_Extended(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EARFCN_FDD_Extended, EARFCN_FDD_Extended_sequence);

  return offset;
}


static const value_string rnsap_E_AGCH_Table_Choice_vals[] = {
  {   0, "table16B" },
  {   1, "table16B-1" },
  { 0, NULL }
};


static int
dissect_rnsap_E_AGCH_Table_Choice(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_28(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     28, 28, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t ECGI_sequence[] = {
  { &hf_rnsap_pLMN_Identity , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PLMN_Identity },
  { &hf_rnsap_e_UTRAN_Cell_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_28 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ECGI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ECGI, ECGI_sequence);

  return offset;
}



static int
dissect_rnsap_EDCH_DDI_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 62U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_E_DCH_Decoupling_Indication_vals[] = {
  {   0, "serving-E-DCH-cell-only" },
  {   1, "serving-HS-DSCH-cell-only" },
  { 0, NULL }
};


static int
dissect_rnsap_E_DCH_Decoupling_Indication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_E_RGCH_E_HICH_ChannelisationCodeValidityIndicator_vals[] = {
  {   0, "e-RGCH-E-HICH-Channelisation-Code-response-not-valid" },
  { 0, NULL }
};


static int
dissect_rnsap_E_RGCH_E_HICH_ChannelisationCodeValidityIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_GenericTrafficCategory(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     8, 8, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const value_string rnsap_TnlQos_vals[] = {
  {   0, "dsField" },
  {   1, "genericTrafficCategory" },
  { 0, NULL }
};

static const per_choice_t TnlQos_choice[] = {
  {   0, &hf_rnsap_dsField       , ASN1_EXTENSION_ROOT    , dissect_rnsap_DsField },
  {   1, &hf_rnsap_genericTrafficCategory, ASN1_EXTENSION_ROOT    , dissect_rnsap_GenericTrafficCategory },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_TnlQos(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_TnlQos, TnlQos_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_TrafficClass_vals[] = {
  {   0, "conversational" },
  {   1, "streaming" },
  {   2, "interactive" },
  {   3, "background" },
  { 0, NULL }
};


static int
dissect_rnsap_TrafficClass(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_Max_Bits_MACe_PDU_non_scheduled(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNrOfBits_MACe_PDU_non_scheduled, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_Non_Scheduled_Transmission_Grant_Items_sequence[] = {
  { &hf_rnsap_maxBits_MACe_PDU_non_scheduled, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Max_Bits_MACe_PDU_non_scheduled },
  { &hf_rnsap_hARQ_Process_Allocation_NonSched_2ms, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_Non_Scheduled_Transmission_Grant_Items(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_Non_Scheduled_Transmission_Grant_Items, E_DCH_Non_Scheduled_Transmission_Grant_Items_sequence);

  return offset;
}


static const value_string rnsap_E_DCH_Grant_Type_Information_vals[] = {
  {   0, "e-DCH-Non-Scheduled-Transmission-Grant" },
  {   1, "e-DCH-Scheduled-Transmission-Grant" },
  { 0, NULL }
};

static const per_choice_t E_DCH_Grant_Type_Information_choice[] = {
  {   0, &hf_rnsap_e_DCH_Non_Scheduled_Transmission_Grant, ASN1_EXTENSION_ROOT    , dissect_rnsap_E_DCH_Non_Scheduled_Transmission_Grant_Items },
  {   1, &hf_rnsap_e_DCH_Scheduled_Transmission_Grant, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_E_DCH_Grant_Type_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_E_DCH_Grant_Type_Information, E_DCH_Grant_Type_Information_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_SchedulingInformation_vals[] = {
  {   0, "included" },
  {   1, "not-included" },
  { 0, NULL }
};


static int
dissect_rnsap_SchedulingInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_MACes_Guaranteed_Bitrate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16777215U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_MACdPDU_Size(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 5000U, NULL, TRUE);

  return offset;
}


static const per_sequence_t E_DCH_MACdPDU_SizeListItem_sequence[] = {
  { &hf_rnsap_mACdPDU_Size  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MACdPDU_Size },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_MACdPDU_SizeListItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_MACdPDU_SizeListItem, E_DCH_MACdPDU_SizeListItem_sequence);

  return offset;
}


static const per_sequence_t E_DCH_MACdPDU_SizeList_sequence_of[1] = {
  { &hf_rnsap_E_DCH_MACdPDU_SizeList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_MACdPDU_SizeListItem },
};

static int
dissect_rnsap_E_DCH_MACdPDU_SizeList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_MACdPDU_SizeList, E_DCH_MACdPDU_SizeList_sequence_of,
                                                  1, maxNrOfMACdPDUSize, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_LogicalChannelInformationItem_sequence[] = {
  { &hf_rnsap_logicalChannelId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_LogicalChannelID },
  { &hf_rnsap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_schedulingInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SchedulingInformation },
  { &hf_rnsap_mACes_GuaranteedBitRate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MACes_Guaranteed_Bitrate },
  { &hf_rnsap_eDCH_DDI_Value, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_DDI_Value },
  { &hf_rnsap_mACd_PDU_Size_List, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_MACdPDU_SizeList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_LogicalChannelInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_LogicalChannelInformationItem, E_DCH_LogicalChannelInformationItem_sequence);

  return offset;
}


static const per_sequence_t E_DCH_LogicalChannelInformation_sequence_of[1] = {
  { &hf_rnsap_E_DCH_LogicalChannelInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_LogicalChannelInformationItem },
};

static int
dissect_rnsap_E_DCH_LogicalChannelInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_LogicalChannelInformation, E_DCH_LogicalChannelInformation_sequence_of,
                                                  1, maxNoOfLogicalChannels, FALSE);

  return offset;
}


static const per_sequence_t EDCH_MACdFlow_Specific_InfoItem_sequence[] = {
  { &hf_rnsap_eDCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_tnlQoS        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TnlQos },
  { &hf_rnsap_payloadCRC_PresenceIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PayloadCRC_PresenceIndicator },
  { &hf_rnsap_maxNr_Retransmissions_EDCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaxNr_Retransmissions_EDCH },
  { &hf_rnsap_trafficClass  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TrafficClass },
  { &hf_rnsap_eDCH_HARQ_PO_FDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_HARQ_PO_FDD },
  { &hf_rnsap_eDCH_MACdFlow_Multiplexing_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_MACdFlow_Multiplexing_List },
  { &hf_rnsap_eDCH_Grant_Type_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Grant_Type_Information },
  { &hf_rnsap_bundlingModeIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BundlingModeIndicator },
  { &hf_rnsap_eDCHLogicalChannelInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_LogicalChannelInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_MACdFlow_Specific_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_MACdFlow_Specific_InfoItem, EDCH_MACdFlow_Specific_InfoItem_sequence);

  return offset;
}


static const per_sequence_t EDCH_MACdFlow_Specific_InfoList_sequence_of[1] = {
  { &hf_rnsap_EDCH_MACdFlow_Specific_InfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_Specific_InfoItem },
};

static int
dissect_rnsap_EDCH_MACdFlow_Specific_InfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_EDCH_MACdFlow_Specific_InfoList, EDCH_MACdFlow_Specific_InfoList_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t EDCH_MACdFlows_Information_sequence[] = {
  { &hf_rnsap_eDCH_MACdFlow_Specific_Information_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_Specific_InfoList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_MACdFlows_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_MACdFlows_Information, EDCH_MACdFlows_Information_sequence);

  return offset;
}



static int
dissect_rnsap_E_DCH_Reference_Power_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfEDCH_HARQ_PO_QUANTSTEPs, NULL, FALSE);

  return offset;
}


static const per_sequence_t EDCH_FDD_Information_sequence[] = {
  { &hf_rnsap_eDCH_MACdFlows_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlows_Information },
  { &hf_rnsap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_rnsap_e_DCH_Maximum_Bitrate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Maximum_Bitrate },
  { &hf_rnsap_e_DCH_Processing_Overload_Level, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Processing_Overload_Level },
  { &hf_rnsap_e_DCH_Reference_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Reference_Power_Offset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_FDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_FDD_Information, EDCH_FDD_Information_sequence);

  return offset;
}


static const value_string rnsap_UPH_Filtering_Measurement_Forwarding_Request_vals[] = {
  {   0, "requested" },
  {   1, "notRequested" },
  { 0, NULL }
};


static int
dissect_rnsap_UPH_Filtering_Measurement_Forwarding_Request(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t EDCH_MACdFlow_Specific_InformationResponseItem_sequence[] = {
  { &hf_rnsap_eDCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_hARQ_Process_Allocation_NonSched_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_MACdFlow_Specific_InformationResponseItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_MACdFlow_Specific_InformationResponseItem, EDCH_MACdFlow_Specific_InformationResponseItem_sequence);

  return offset;
}


static const per_sequence_t EDCH_MACdFlow_Specific_InformationResponse_sequence_of[1] = {
  { &hf_rnsap_EDCH_MACdFlow_Specific_InformationResponse_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_Specific_InformationResponseItem },
};

static int
dissect_rnsap_EDCH_MACdFlow_Specific_InformationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_EDCH_MACdFlow_Specific_InformationResponse, EDCH_MACdFlow_Specific_InformationResponse_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t EDCH_FDD_InformationResponse_sequence[] = {
  { &hf_rnsap_eDCH_MACdFlow_Specific_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_Specific_InformationResponse },
  { &hf_rnsap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_FDD_InformationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_FDD_InformationResponse, EDCH_FDD_InformationResponse_sequence);

  return offset;
}


static const value_string rnsap_TransportBearerRequestIndicator_vals[] = {
  {   0, "bearer-requested" },
  {   1, "bearer-not-requested" },
  { 0, NULL }
};


static int
dissect_rnsap_TransportBearerRequestIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t E_DCH_MACdPDU_SizeToModifyList_sequence_of[1] = {
  { &hf_rnsap_E_DCH_MACdPDU_SizeToModifyList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_MACdPDU_SizeListItem },
};

static int
dissect_rnsap_E_DCH_MACdPDU_SizeToModifyList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_MACdPDU_SizeToModifyList, E_DCH_MACdPDU_SizeToModifyList_sequence_of,
                                                  0, maxNrOfMACdPDUSize, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_LogicalChannelToModifyItem_sequence[] = {
  { &hf_rnsap_logicalChannelId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_LogicalChannelID },
  { &hf_rnsap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_schedulingInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SchedulingInformation },
  { &hf_rnsap_mACes_GuaranteedBitRate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MACes_Guaranteed_Bitrate },
  { &hf_rnsap_eDCH_DDI_Value, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_EDCH_DDI_Value },
  { &hf_rnsap_mACd_PDU_Size_List_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_MACdPDU_SizeToModifyList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_LogicalChannelToModifyItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_LogicalChannelToModifyItem, E_DCH_LogicalChannelToModifyItem_sequence);

  return offset;
}


static const per_sequence_t E_DCH_LogicalChannelToModify_sequence_of[1] = {
  { &hf_rnsap_E_DCH_LogicalChannelToModify_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_LogicalChannelToModifyItem },
};

static int
dissect_rnsap_E_DCH_LogicalChannelToModify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_LogicalChannelToModify, E_DCH_LogicalChannelToModify_sequence_of,
                                                  1, maxNoOfLogicalChannels, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_LogicalChannelToDeleteItem_sequence[] = {
  { &hf_rnsap_logicalChannelId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_LogicalChannelID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_LogicalChannelToDeleteItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_LogicalChannelToDeleteItem, E_DCH_LogicalChannelToDeleteItem_sequence);

  return offset;
}


static const per_sequence_t E_DCH_LogicalChannelToDelete_sequence_of[1] = {
  { &hf_rnsap_E_DCH_LogicalChannelToDelete_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_LogicalChannelToDeleteItem },
};

static int
dissect_rnsap_E_DCH_LogicalChannelToDelete(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_LogicalChannelToDelete, E_DCH_LogicalChannelToDelete_sequence_of,
                                                  1, maxNoOfLogicalChannels, FALSE);

  return offset;
}


static const per_sequence_t EDCH_MACdFlow_Specific_InfoToModifyItem_sequence[] = {
  { &hf_rnsap_eDCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportBearerRequestIndicator },
  { &hf_rnsap_tnlQoS        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TnlQos },
  { &hf_rnsap_maxNr_Retransmissions_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MaxNr_Retransmissions_EDCH },
  { &hf_rnsap_trafficClass  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TrafficClass },
  { &hf_rnsap_eDCH_HARQ_PO_FDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_HARQ_PO_FDD },
  { &hf_rnsap_eDCH_MACdFlow_Multiplexing_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_MACdFlow_Multiplexing_List },
  { &hf_rnsap_eDCH_Grant_Type_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Grant_Type_Information },
  { &hf_rnsap_bundlingModeIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BundlingModeIndicator },
  { &hf_rnsap_eDCH_LogicalChannelToAdd, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_LogicalChannelInformation },
  { &hf_rnsap_eDCH_LogicalChannelToModify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_LogicalChannelToModify },
  { &hf_rnsap_eDCH_LogicalChannelToDelete, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_LogicalChannelToDelete },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_MACdFlow_Specific_InfoToModifyItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_MACdFlow_Specific_InfoToModifyItem, EDCH_MACdFlow_Specific_InfoToModifyItem_sequence);

  return offset;
}


static const per_sequence_t EDCH_MACdFlow_Specific_InfoToModifyList_sequence_of[1] = {
  { &hf_rnsap_EDCH_MACdFlow_Specific_InfoToModifyList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_Specific_InfoToModifyItem },
};

static int
dissect_rnsap_EDCH_MACdFlow_Specific_InfoToModifyList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_EDCH_MACdFlow_Specific_InfoToModifyList, EDCH_MACdFlow_Specific_InfoToModifyList_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const value_string rnsap_MACeReset_Indicator_vals[] = {
  {   0, "mACeReset" },
  { 0, NULL }
};


static int
dissect_rnsap_MACeReset_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t EDCH_FDD_Information_To_Modify_sequence[] = {
  { &hf_rnsap_eDCH_MACdFlow_Specific_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_Specific_InfoToModifyList },
  { &hf_rnsap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_rnsap_e_DCH_Maximum_Bitrate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Maximum_Bitrate },
  { &hf_rnsap_e_DCH_Processing_Overload_Level, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Processing_Overload_Level },
  { &hf_rnsap_e_DCH_Reference_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Reference_Power_Offset },
  { &hf_rnsap_mACeReset_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MACeReset_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_FDD_Information_To_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_FDD_Information_To_Modify, EDCH_FDD_Information_To_Modify_sequence);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_Specific_UpdateInformation_Item_sequence[] = {
  { &hf_rnsap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_hARQ_Process_Allocation_NonSched_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_MACdFlow_Specific_UpdateInformation_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_MACdFlow_Specific_UpdateInformation_Item, E_DCH_MACdFlow_Specific_UpdateInformation_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_Specific_UpdateInformation_sequence_of[1] = {
  { &hf_rnsap_E_DCH_MACdFlow_Specific_UpdateInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_MACdFlow_Specific_UpdateInformation_Item },
};

static int
dissect_rnsap_E_DCH_MACdFlow_Specific_UpdateInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_MACdFlow_Specific_UpdateInformation, E_DCH_MACdFlow_Specific_UpdateInformation_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_FDD_Update_Information_sequence[] = {
  { &hf_rnsap_e_DCH_MACdFlow_Specific_UpdateInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_MACdFlow_Specific_UpdateInformation },
  { &hf_rnsap_hARQ_Process_Allocation_Scheduled_2ms_EDCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Process_Allocation_2ms_EDCH },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_FDD_Update_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_FDD_Update_Information, E_DCH_FDD_Update_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCH_DL_Control_Channel_Change_Information_Item_sequence[] = {
  { &hf_rnsap_e_DCH_RL_ID   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_DL_Control_Channel_Change_Information_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_DL_Control_Channel_Change_Information_Item, E_DCH_DL_Control_Channel_Change_Information_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_DL_Control_Channel_Change_Information_sequence_of[1] = {
  { &hf_rnsap_E_DCH_DL_Control_Channel_Change_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_DL_Control_Channel_Change_Information_Item },
};

static int
dissect_rnsap_E_DCH_DL_Control_Channel_Change_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_DL_Control_Channel_Change_Information, E_DCH_DL_Control_Channel_Change_Information_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_DL_Control_Channel_Grant_Information_Item_sequence[] = {
  { &hf_rnsap_e_DCH_RL_ID   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_DL_Control_Channel_Grant_Information_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_DL_Control_Channel_Grant_Information_Item, E_DCH_DL_Control_Channel_Grant_Information_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_DL_Control_Channel_Grant_Information_sequence_of[1] = {
  { &hf_rnsap_E_DCH_DL_Control_Channel_Grant_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_DL_Control_Channel_Grant_Information_Item },
};

static int
dissect_rnsap_E_DCH_DL_Control_Channel_Grant_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_DL_Control_Channel_Grant_Information, E_DCH_DL_Control_Channel_Grant_Information_sequence_of,
                                                  1, maxNrOfEDCHRLs, FALSE);

  return offset;
}



static int
dissect_rnsap_E_DCH_PowerOffset_for_SchedulingInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfEDCH_HARQ_PO_QUANTSTEPs, NULL, FALSE);

  return offset;
}


static const value_string rnsap_E_DCH_MACdPDUSizeFormat_vals[] = {
  {   0, "fixedMACdPDU-Size" },
  {   1, "flexibleMACdPDU-Size" },
  { 0, NULL }
};


static int
dissect_rnsap_E_DCH_MACdPDUSizeFormat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t EDCH_MACdFlows_To_Delete_Item_sequence[] = {
  { &hf_rnsap_eDCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_MACdFlows_To_Delete_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_MACdFlows_To_Delete_Item, EDCH_MACdFlows_To_Delete_Item_sequence);

  return offset;
}


static const per_sequence_t EDCH_MACdFlows_To_Delete_sequence_of[1] = {
  { &hf_rnsap_EDCH_MACdFlows_To_Delete_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlows_To_Delete_Item },
};

static int
dissect_rnsap_EDCH_MACdFlows_To_Delete(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_EDCH_MACdFlows_To_Delete, EDCH_MACdFlows_To_Delete_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const value_string rnsap_EDCH_RL_Indication_vals[] = {
  {   0, "eDCH" },
  {   1, "non-EDCH" },
  { 0, NULL }
};


static int
dissect_rnsap_EDCH_RL_Indication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_E_DCH_TFCI_Table_Index(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_Max_Set_E_DPDCHs_vals[] = {
  {   0, "vN256" },
  {   1, "vN128" },
  {   2, "vN64" },
  {   3, "vN32" },
  {   4, "vN16" },
  {   5, "vN8" },
  {   6, "vN4" },
  {   7, "v2xN4" },
  {   8, "v2xN2" },
  {   9, "v2xN2plus2xN4" },
  {  10, "v2xM2plus2xM4" },
  { 0, NULL }
};


static int
dissect_rnsap_Max_Set_E_DPDCHs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     10, NULL, TRUE, 1, NULL);

  return offset;
}



static int
dissect_rnsap_PunctureLimit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Reference_E_TFCI_PO(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfRefETFCI_PO_QUANTSTEPs, NULL, FALSE);

  return offset;
}


static const per_sequence_t Reference_E_TFCI_Information_Item_sequence[] = {
  { &hf_rnsap_reference_E_TFCI, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_TFCI },
  { &hf_rnsap_reference_E_TFCI_PO, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Reference_E_TFCI_PO },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Reference_E_TFCI_Information_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Reference_E_TFCI_Information_Item, Reference_E_TFCI_Information_Item_sequence);

  return offset;
}


static const per_sequence_t Reference_E_TFCI_Information_sequence_of[1] = {
  { &hf_rnsap_Reference_E_TFCI_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Reference_E_TFCI_Information_Item },
};

static int
dissect_rnsap_Reference_E_TFCI_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Reference_E_TFCI_Information, Reference_E_TFCI_Information_sequence_of,
                                                  1, maxNrOfRefETFCIs, FALSE);

  return offset;
}


static const per_sequence_t E_TFCS_Information_sequence[] = {
  { &hf_rnsap_e_DCH_TFCI_Table_Index, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_TFCI_Table_Index },
  { &hf_rnsap_e_DCH_Min_Set_E_TFCI, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_TFCI },
  { &hf_rnsap_reference_E_TFCI_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Reference_E_TFCI_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_TFCS_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_TFCS_Information, E_TFCS_Information_sequence);

  return offset;
}


static const value_string rnsap_E_TTI_vals[] = {
  {   0, "tti10" },
  {   1, "tti2" },
  { 0, NULL }
};


static int
dissect_rnsap_E_TTI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_E_DPCCH_PO(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfEDPCCH_PO_QUANTSTEPs, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_E_RGCH_2_IndexStepThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 37U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_E_RGCH_3_IndexStepThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 37U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_HARQ_Info_for_E_DCH_vals[] = {
  {   0, "rv0" },
  {   1, "rvtable" },
  { 0, NULL }
};


static int
dissect_rnsap_HARQ_Info_for_E_DCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_HSDSCH_Configured_Indicator_vals[] = {
  {   0, "configured-HS-DSCH" },
  {   1, "no-configured-HS-DSCH" },
  { 0, NULL }
};


static int
dissect_rnsap_HSDSCH_Configured_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t EDPCH_Information_FDD_sequence[] = {
  { &hf_rnsap_maxSet_E_DPDCHs, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Max_Set_E_DPDCHs },
  { &hf_rnsap_punctureLimit , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PunctureLimit },
  { &hf_rnsap_e_TFCS_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_TFCS_Information },
  { &hf_rnsap_e_TTI         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_TTI },
  { &hf_rnsap_e_DPCCH_PO    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DPCCH_PO },
  { &hf_rnsap_e_RGCH_2_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_RGCH_2_IndexStepThreshold },
  { &hf_rnsap_e_RGCH_3_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_RGCH_3_IndexStepThreshold },
  { &hf_rnsap_hARQ_Info_for_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HARQ_Info_for_E_DCH },
  { &hf_rnsap_hSDSCH_Configured_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_Configured_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDPCH_Information_FDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDPCH_Information_FDD, EDPCH_Information_FDD_sequence);

  return offset;
}


static const per_sequence_t EDPCH_Information_RLReconfRequest_FDD_sequence[] = {
  { &hf_rnsap_maxSet_E_DPDCHs, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Max_Set_E_DPDCHs },
  { &hf_rnsap_punctureLimit , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PunctureLimit },
  { &hf_rnsap_e_TFCS_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_TFCS_Information },
  { &hf_rnsap_e_TTI         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_TTI },
  { &hf_rnsap_e_DPCCH_PO    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DPCCH_PO },
  { &hf_rnsap_e_RGCH_2_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_RGCH_2_IndexStepThreshold },
  { &hf_rnsap_e_RGCH_3_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_RGCH_3_IndexStepThreshold },
  { &hf_rnsap_hARQ_Info_for_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Info_for_E_DCH },
  { &hf_rnsap_hSDSCH_Configured_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSDSCH_Configured_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDPCH_Information_RLReconfRequest_FDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDPCH_Information_RLReconfRequest_FDD, EDPCH_Information_RLReconfRequest_FDD_sequence);

  return offset;
}



static int
dissect_rnsap_Extended_E_DPCCH_PO(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            9U, 15U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_E_DPDCH_PowerInterpolation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t EDCH_Serving_RL_in_this_DRNS_sequence[] = {
  { &hf_rnsap_e_DCH_Serving_RL_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_Serving_RL_in_this_DRNS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_Serving_RL_in_this_DRNS, EDCH_Serving_RL_in_this_DRNS_sequence);

  return offset;
}


static const value_string rnsap_EDCH_Serving_RL_vals[] = {
  {   0, "e-DCH-Serving-RL-in-this-DRNS" },
  {   1, "e-DCH-Serving-RL-not-in-this-DRNS" },
  { 0, NULL }
};

static const per_choice_t EDCH_Serving_RL_choice[] = {
  {   0, &hf_rnsap_e_DCH_Serving_RL_in_this_DRNS, ASN1_EXTENSION_ROOT    , dissect_rnsap_EDCH_Serving_RL_in_this_DRNS },
  {   1, &hf_rnsap_e_DCH_Serving_RL_not_in_this_DRNS, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_EDCH_Serving_RL(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_EDCH_Serving_RL, EDCH_Serving_RL_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_PriorityQueue_Id(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfPrioQueues_1, NULL, FALSE);

  return offset;
}


static const value_string rnsap_T1_vals[] = {
  {   0, "v10" },
  {   1, "v20" },
  {   2, "v30" },
  {   3, "v40" },
  {   4, "v50" },
  {   5, "v60" },
  {   6, "v70" },
  {   7, "v80" },
  {   8, "v90" },
  {   9, "v100" },
  {  10, "v120" },
  {  11, "v140" },
  {  12, "v160" },
  {  13, "v200" },
  {  14, "v300" },
  {  15, "v400" },
  { 0, NULL }
};


static int
dissect_rnsap_T1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     16, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_MAC_ehs_Reset_Timer_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v3" },
  {   3, "v4" },
  { 0, NULL }
};


static int
dissect_rnsap_MAC_ehs_Reset_Timer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_MAC_hsWindowSize_vals[] = {
  {   0, "v4" },
  {   1, "v6" },
  {   2, "v8" },
  {   3, "v12" },
  {   4, "v16" },
  {   5, "v24" },
  {   6, "v32" },
  {   7, "v64" },
  {   8, "v128" },
  {   9, "v256" },
  { 0, NULL }
};


static int
dissect_rnsap_MAC_hsWindowSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, TRUE, 3, NULL);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoItem_EnhancedFACH_PCH_sequence[] = {
  { &hf_rnsap_priorityQueue_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityQueue_Id },
  { &hf_rnsap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_t1            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T1 },
  { &hf_rnsap_mAC_ehs_Reset_Timer, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAC_ehs_Reset_Timer },
  { &hf_rnsap_discardTimer  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DiscardTimer },
  { &hf_rnsap_mAC_hsWindowSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAC_hsWindowSize },
  { &hf_rnsap_maximum_MACdPDU_Size_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAC_PDU_SizeExtended },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PriorityQueue_InfoItem_EnhancedFACH_PCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PriorityQueue_InfoItem_EnhancedFACH_PCH, PriorityQueue_InfoItem_EnhancedFACH_PCH_sequence);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoList_EnhancedFACH_PCH_sequence_of[1] = {
  { &hf_rnsap_PriorityQueue_InfoList_EnhancedFACH_PCH_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityQueue_InfoItem_EnhancedFACH_PCH },
};

static int
dissect_rnsap_PriorityQueue_InfoList_EnhancedFACH_PCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_PriorityQueue_InfoList_EnhancedFACH_PCH, PriorityQueue_InfoList_EnhancedFACH_PCH_sequence_of,
                                                  1, maxNrOfPrioQueues, FALSE);

  return offset;
}



static int
dissect_rnsap_HSDSCH_InitialWindowSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_Initial_Capacity_AllocationItem_sequence[] = {
  { &hf_rnsap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_maximum_MACdPDU_Size, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MACdPDU_Size },
  { &hf_rnsap_hSDSCH_InitialWindowSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_InitialWindowSize },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_Initial_Capacity_AllocationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_Initial_Capacity_AllocationItem, HSDSCH_Initial_Capacity_AllocationItem_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_Initial_Capacity_Allocation_sequence_of[1] = {
  { &hf_rnsap_HSDSCH_Initial_Capacity_Allocation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_Initial_Capacity_AllocationItem },
};

static int
dissect_rnsap_HSDSCH_Initial_Capacity_Allocation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSDSCH_Initial_Capacity_Allocation, HSDSCH_Initial_Capacity_Allocation_sequence_of,
                                                  1, maxNrOfPrioQueues, FALSE);

  return offset;
}



static int
dissect_rnsap_HSDSCH_RNTI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Enhanced_FACH_Information_ResponseFDD_sequence[] = {
  { &hf_rnsap_common_HS_DSCH_RNTI_priorityQueueInfo_EnhancedFACH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityQueue_InfoList_EnhancedFACH_PCH },
  { &hf_rnsap_dedicated_HS_DSCH_RNTI_priorityQueueInfo_EnhancedFACH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityQueue_InfoList_EnhancedFACH_PCH },
  { &hf_rnsap_priorityQueueInfo_EnhancedPCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PriorityQueue_InfoList_EnhancedFACH_PCH },
  { &hf_rnsap_hSDSCH_Initial_Capacity_Allocation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_Initial_Capacity_Allocation },
  { &hf_rnsap_hSDSCH_RNTI   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSDSCH_RNTI },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Enhanced_FACH_Information_ResponseFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Enhanced_FACH_Information_ResponseFDD, Enhanced_FACH_Information_ResponseFDD_sequence);

  return offset;
}


static const per_sequence_t Enhanced_FACH_Information_ResponseLCR_sequence[] = {
  { &hf_rnsap_common_HS_DSCH_RNTI_priorityQueueInfo_EnhancedFACH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityQueue_InfoList_EnhancedFACH_PCH },
  { &hf_rnsap_dedicated_HS_DSCH_RNTI_priorityQueueInfo_EnhancedFACH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityQueue_InfoList_EnhancedFACH_PCH },
  { &hf_rnsap_priorityQueueInfo_EnhancedPCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PriorityQueue_InfoList_EnhancedFACH_PCH },
  { &hf_rnsap_hSDSCH_Initial_Capacity_Allocation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_Initial_Capacity_Allocation },
  { &hf_rnsap_hSDSCH_RNTI   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSDSCH_RNTI },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Enhanced_FACH_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Enhanced_FACH_Information_ResponseLCR, Enhanced_FACH_Information_ResponseLCR_sequence);

  return offset;
}



static int
dissect_rnsap_Enhanced_FACH_Support_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_EnhancedHSServingCC_Abort_vals[] = {
  {   0, "abortEnhancedHSServingCC" },
  { 0, NULL }
};


static int
dissect_rnsap_EnhancedHSServingCC_Abort(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_Enhanced_PCH_Capability_vals[] = {
  {   0, "enhanced-pch-capable" },
  {   1, "enhanced-pch-not-capable" },
  { 0, NULL }
};


static int
dissect_rnsap_Enhanced_PCH_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_E_ROCH_PowerOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_E_TFCI_BetaEC_Boost(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_UL_Delta_T2TP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 6U, NULL, TRUE);

  return offset;
}


static const per_sequence_t E_TFCI_Boost_Information_sequence[] = {
  { &hf_rnsap_e_TFCI_BetaEC_Boost, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_TFCI_BetaEC_Boost },
  { &hf_rnsap_uL_Delta_T2TP , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_Delta_T2TP },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_TFCI_Boost_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_TFCI_Boost_Information, E_TFCI_Boost_Information_sequence);

  return offset;
}


static const value_string rnsap_E_DCH_Minimum_Set_E_TFCIValidityIndicator_vals[] = {
  {   0, "e-DCH-Minimum-Set-E-TFCI-response-not-valid" },
  { 0, NULL }
};


static int
dissect_rnsap_E_DCH_Minimum_Set_E_TFCIValidityIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_Enhanced_PrimaryCPICH_EcNo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 49U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Extension_MeasurementThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_MeasurementThreshold_vals[] = {
  {   0, "sir" },
  {   1, "sir-error" },
  {   2, "transmitted-code-power" },
  {   3, "rscp" },
  {   4, "rx-timing-deviation" },
  {   5, "round-trip-time" },
  {   6, "extension-MeasurementThreshold" },
  { 0, NULL }
};

static const per_choice_t MeasurementThreshold_choice[] = {
  {   0, &hf_rnsap_sir_01        , ASN1_EXTENSION_ROOT    , dissect_rnsap_SIR_Value },
  {   1, &hf_rnsap_sir_error_01  , ASN1_EXTENSION_ROOT    , dissect_rnsap_SIR_Error_Value },
  {   2, &hf_rnsap_transmitted_code_power_01, ASN1_EXTENSION_ROOT    , dissect_rnsap_Transmitted_Code_Power_Value },
  {   3, &hf_rnsap_rscp_01       , ASN1_EXTENSION_ROOT    , dissect_rnsap_RSCP_Value },
  {   4, &hf_rnsap_rx_timing_deviation, ASN1_EXTENSION_ROOT    , dissect_rnsap_Rx_Timing_Deviation_Value },
  {   5, &hf_rnsap_round_trip_time_01, ASN1_EXTENSION_ROOT    , dissect_rnsap_Round_Trip_Time_Value },
  {   6, &hf_rnsap_extension_MeasurementThreshold, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_MeasurementThreshold },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_MeasurementThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_MeasurementThreshold, MeasurementThreshold_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_MeasurementHysteresisTime(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 6000U, NULL, TRUE);

  return offset;
}


static const per_sequence_t EventA_sequence[] = {
  { &hf_rnsap_measurementTreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementThreshold },
  { &hf_rnsap_measurementHysteresisTime, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MeasurementHysteresisTime },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EventA(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EventA, EventA_sequence);

  return offset;
}


static const per_sequence_t EventB_sequence[] = {
  { &hf_rnsap_measurementTreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementThreshold },
  { &hf_rnsap_measurementHysteresisTime, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MeasurementHysteresisTime },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EventB(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EventB, EventB_sequence);

  return offset;
}



static int
dissect_rnsap_SIR_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 62U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_SIR_Error_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 124U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Transmitted_Code_Power_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 112U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_RSCP_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 126U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Round_Trip_Time_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 32766U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Extension_MeasurementIncreaseDecreaseThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_MeasurementIncreaseDecreaseThreshold_vals[] = {
  {   0, "sir" },
  {   1, "sir-error" },
  {   2, "transmitted-code-power" },
  {   3, "rscp" },
  {   4, "round-trip-time" },
  {   5, "extension-MeasurementIncreaseDecreaseThreshold" },
  { 0, NULL }
};

static const per_choice_t MeasurementIncreaseDecreaseThreshold_choice[] = {
  {   0, &hf_rnsap_sir           , ASN1_EXTENSION_ROOT    , dissect_rnsap_SIR_Value_IncrDecrThres },
  {   1, &hf_rnsap_sir_error     , ASN1_EXTENSION_ROOT    , dissect_rnsap_SIR_Error_Value_IncrDecrThres },
  {   2, &hf_rnsap_transmitted_code_power, ASN1_EXTENSION_ROOT    , dissect_rnsap_Transmitted_Code_Power_Value_IncrDecrThres },
  {   3, &hf_rnsap_rscp          , ASN1_EXTENSION_ROOT    , dissect_rnsap_RSCP_Value_IncrDecrThres },
  {   4, &hf_rnsap_round_trip_time, ASN1_EXTENSION_ROOT    , dissect_rnsap_Round_Trip_Time_IncrDecrThres },
  {   5, &hf_rnsap_extension_MeasurementIncreaseDecreaseThreshold, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_MeasurementIncreaseDecreaseThreshold },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_MeasurementIncreaseDecreaseThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_MeasurementIncreaseDecreaseThreshold, MeasurementIncreaseDecreaseThreshold_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_MeasurementChangeTime(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 6000U, NULL, TRUE);

  return offset;
}


static const per_sequence_t EventC_sequence[] = {
  { &hf_rnsap_measurementIncreaseDecreaseThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementIncreaseDecreaseThreshold },
  { &hf_rnsap_measurementChangeTime, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementChangeTime },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EventC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EventC, EventC_sequence);

  return offset;
}


static const per_sequence_t EventD_sequence[] = {
  { &hf_rnsap_measurementIncreaseDecreaseThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementIncreaseDecreaseThreshold },
  { &hf_rnsap_measurementChangeTime, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementChangeTime },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EventD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EventD, EventD_sequence);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_6000_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 6000U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_60_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 60U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_ReportPeriodicity_vals[] = {
  {   0, "ten-msec" },
  {   1, "min" },
  { 0, NULL }
};

static const per_choice_t ReportPeriodicity_choice[] = {
  {   0, &hf_rnsap_ten_msec      , ASN1_EXTENSION_ROOT    , dissect_rnsap_INTEGER_1_6000_ },
  {   1, &hf_rnsap_min           , ASN1_EXTENSION_ROOT    , dissect_rnsap_INTEGER_1_60_ },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_ReportPeriodicity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_ReportPeriodicity, ReportPeriodicity_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t EventE_sequence[] = {
  { &hf_rnsap_measurementThreshold1, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementThreshold },
  { &hf_rnsap_measurementThreshold2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MeasurementThreshold },
  { &hf_rnsap_measurementHysteresisTime, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MeasurementHysteresisTime },
  { &hf_rnsap_reportPeriodicity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ReportPeriodicity },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EventE(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EventE, EventE_sequence);

  return offset;
}


static const per_sequence_t EventF_sequence[] = {
  { &hf_rnsap_measurementThreshold1, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementThreshold },
  { &hf_rnsap_measurementThreshold2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MeasurementThreshold },
  { &hf_rnsap_measurementHysteresisTime, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MeasurementHysteresisTime },
  { &hf_rnsap_reportPeriodicity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ReportPeriodicity },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EventF(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EventF, EventF_sequence);

  return offset;
}


static const per_sequence_t EventH_sequence[] = {
  { &hf_rnsap_measurementThreshold1, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementThreshold },
  { &hf_rnsap_measurementThreshold2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MeasurementThreshold },
  { &hf_rnsap_measurementHysteresisTime, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MeasurementHysteresisTime },
  { &hf_rnsap_reportPeriodicity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ReportPeriodicity },
  { &hf_rnsap_measurementFluctuationRange, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_0_100 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EventH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EventH, EventH_sequence);

  return offset;
}


static const value_string rnsap_MeasurementQuantity_vals[] = {
  {   0, "cpichEcNo" },
  {   1, "cpichRSCP" },
  {   2, "pathloss" },
  { 0, NULL }
};


static int
dissect_rnsap_MeasurementQuantity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_M120_165(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -120, 165U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Event1F_Parameters_sequence[] = {
  { &hf_rnsap_measurementQuantity, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementQuantity },
  { &hf_rnsap_threshold     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_M120_165 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Event1F_Parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Event1F_Parameters, Event1F_Parameters_sequence);

  return offset;
}



static int
dissect_rnsap_INTEGER_M120_M25(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -120, -25, NULL, FALSE);

  return offset;
}


static const per_sequence_t Event1I_Parameters_sequence[] = {
  { &hf_rnsap_threshold_01  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_M120_M25 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Event1I_Parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Event1I_Parameters, Event1I_Parameters_sequence);

  return offset;
}



static int
dissect_rnsap_ExtendedGSMCellIndividualOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -50, 50U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_N_E_UCCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 12U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_PUCH_Information_sequence[] = {
  { &hf_rnsap_minCR         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CodeRate },
  { &hf_rnsap_maxCR         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CodeRate },
  { &hf_rnsap_harqInfo      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HARQ_Info_for_E_DCH },
  { &hf_rnsap_n_E_UCCH      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_N_E_UCCH },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_PUCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_PUCH_Information, E_PUCH_Information_sequence);

  return offset;
}



static int
dissect_rnsap_RefBeta(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -15, 16U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_RefBeta_Item_sequence[] = {
  { &hf_rnsap_refCodeRate   , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_CodeRate_short },
  { &hf_rnsap_refBeta       , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_RefBeta },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_RefBeta_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_RefBeta_Item, E_DCH_RefBeta_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_QPSK_RefBetaInfo_sequence_of[1] = {
  { &hf_rnsap_E_DCH_QPSK_RefBetaInfo_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_RefBeta_Item },
};

static int
dissect_rnsap_E_DCH_QPSK_RefBetaInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_QPSK_RefBetaInfo, E_DCH_QPSK_RefBetaInfo_sequence_of,
                                                  1, maxNrOfRefBetas, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_sixteenQAM_RefBetaInfo_sequence_of[1] = {
  { &hf_rnsap_E_DCH_sixteenQAM_RefBetaInfo_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_RefBeta_Item },
};

static int
dissect_rnsap_E_DCH_sixteenQAM_RefBetaInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_sixteenQAM_RefBetaInfo, E_DCH_sixteenQAM_RefBetaInfo_sequence_of,
                                                  1, maxNrOfRefBetas, FALSE);

  return offset;
}


static const per_sequence_t E_TFCS_Information_TDD_sequence[] = {
  { &hf_rnsap_e_DCH_QPSK_RefBetaInfo, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_QPSK_RefBetaInfo },
  { &hf_rnsap_e_DCH_sixteenQAM_RefBetaInfo, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_sixteenQAM_RefBetaInfo },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_TFCS_Information_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_TFCS_Information_TDD, E_TFCS_Information_TDD_sequence);

  return offset;
}


static const value_string rnsap_E_DCH_Grant_TypeTDD_vals[] = {
  {   0, "scheduled" },
  {   1, "non-scheduled" },
  { 0, NULL }
};


static int
dissect_rnsap_E_DCH_Grant_TypeTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_InfoTDDItem_sequence[] = {
  { &hf_rnsap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_tnlQos        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TnlQos },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_payloadCRC_PresenceIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PayloadCRC_PresenceIndicator },
  { &hf_rnsap_maximum_Number_of_Retransmissions_For_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaxNr_Retransmissions_EDCH },
  { &hf_rnsap_eDCH_HARQ_PO_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_HARQ_PO_TDD },
  { &hf_rnsap_eDCH_MACdFlow_Multiplexing_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_MACdFlow_Multiplexing_List },
  { &hf_rnsap_eDCH_Grant_TypeTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_Grant_TypeTDD },
  { &hf_rnsap_eDCHLogicalChannelInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_LogicalChannelInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_MACdFlow_InfoTDDItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_MACdFlow_InfoTDDItem, E_DCH_MACdFlow_InfoTDDItem_sequence);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlows_Information_TDD_sequence_of[1] = {
  { &hf_rnsap_E_DCH_MACdFlows_Information_TDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_MACdFlow_InfoTDDItem },
};

static int
dissect_rnsap_E_DCH_MACdFlows_Information_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_MACdFlows_Information_TDD, E_DCH_MACdFlows_Information_TDD_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}



static int
dissect_rnsap_E_DCH_TDD_Maximum_Bitrate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 9201U, NULL, TRUE);

  return offset;
}


static const per_sequence_t E_DCH_TDD_Information_sequence[] = {
  { &hf_rnsap_e_DCH_TDD_Maximum_Bitrate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_TDD_Maximum_Bitrate },
  { &hf_rnsap_e_DCH_Processing_Overload_Level, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Processing_Overload_Level },
  { &hf_rnsap_e_DCH_PowerOffset_for_SchedulingInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_PowerOffset_for_SchedulingInfo },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_TDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_TDD_Information, E_DCH_TDD_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCH_Information_sequence[] = {
  { &hf_rnsap_e_PUCH_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_PUCH_Information },
  { &hf_rnsap_e_TFCS_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_TFCS_Information_TDD },
  { &hf_rnsap_e_DCH_MACdFlows_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_MACdFlows_Information_TDD },
  { &hf_rnsap_e_DCH_TDD_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_TDD_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_Information, E_DCH_Information_sequence);

  return offset;
}


static const value_string rnsap_E_DCH_MACdFlow_Retransmission_Timer_LCR_vals[] = {
  {   0, "ms10" },
  {   1, "ms15" },
  {   2, "ms20" },
  {   3, "ms25" },
  {   4, "ms30" },
  {   5, "ms35" },
  {   6, "ms40" },
  {   7, "ms45" },
  {   8, "ms50" },
  {   9, "ms55" },
  {  10, "ms60" },
  {  11, "ms65" },
  {  12, "ms70" },
  {  13, "ms75" },
  {  14, "ms80" },
  {  15, "ms85" },
  {  16, "ms90" },
  {  17, "ms95" },
  {  18, "ms100" },
  {  19, "ms110" },
  {  20, "ms120" },
  {  21, "ms140" },
  {  22, "ms160" },
  {  23, "ms200" },
  {  24, "ms240" },
  {  25, "ms280" },
  {  26, "ms320" },
  {  27, "ms400" },
  {  28, "ms480" },
  {  29, "ms560" },
  { 0, NULL }
};

static value_string_ext rnsap_E_DCH_MACdFlow_Retransmission_Timer_LCR_vals_ext = VALUE_STRING_EXT_INIT(rnsap_E_DCH_MACdFlow_Retransmission_Timer_LCR_vals);


static int
dissect_rnsap_E_DCH_MACdFlow_Retransmission_Timer_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     30, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_E_DCH_TimeslotResource(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     13, 13, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_E_DCH_PowerResource(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 32U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_TddE_PUCH_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_Non_Scheduled_Grant_Info_sequence[] = {
  { &hf_rnsap_timeslotResource, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_TimeslotResource },
  { &hf_rnsap_powerResource , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_PowerResource },
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tddE_PUCH_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TddE_PUCH_Offset },
  { &hf_rnsap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_Non_Scheduled_Grant_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_Non_Scheduled_Grant_Info, E_DCH_Non_Scheduled_Grant_Info_sequence);

  return offset;
}


static const per_sequence_t E_DCH_MACdFlow_ModifyTDDItem_sequence[] = {
  { &hf_rnsap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportBearerRequestIndicator },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_tnlQos        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TnlQos },
  { &hf_rnsap_maximum_Number_of_Retransmissions_For_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MaxNr_Retransmissions_EDCH },
  { &hf_rnsap_eDCH_HARQ_PO_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_HARQ_PO_TDD },
  { &hf_rnsap_eDCH_MACdFlow_Multiplexing_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_MACdFlow_Multiplexing_List },
  { &hf_rnsap_eDCH_Grant_TypeTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Grant_TypeTDD },
  { &hf_rnsap_e_DCH_LogicalChannelToAdd, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_LogicalChannelInformation },
  { &hf_rnsap_e_DCH_LogicalChannelToModify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_LogicalChannelToModify },
  { &hf_rnsap_e_DCH_LogicalChannelToDelete, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_LogicalChannelToDelete },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_MACdFlow_ModifyTDDItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_MACdFlow_ModifyTDDItem, E_DCH_MACdFlow_ModifyTDDItem_sequence);

  return offset;
}


static const per_sequence_t E_DCH_TDD_Information_to_Modify_List_sequence_of[1] = {
  { &hf_rnsap_E_DCH_TDD_Information_to_Modify_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_MACdFlow_ModifyTDDItem },
};

static int
dissect_rnsap_E_DCH_TDD_Information_to_Modify_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_TDD_Information_to_Modify_List, E_DCH_TDD_Information_to_Modify_List_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_TDD_Information_to_Modify_sequence[] = {
  { &hf_rnsap_e_DCH_TDD_Information_to_Modify_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_TDD_Information_to_Modify_List },
  { &hf_rnsap_mACeReset_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MACeReset_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_TDD_Information_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_TDD_Information_to_Modify, E_DCH_TDD_Information_to_Modify_sequence);

  return offset;
}


static const per_sequence_t E_DCH_Information_Reconfig_sequence[] = {
  { &hf_rnsap_e_PUCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_PUCH_Information },
  { &hf_rnsap_e_TFCS_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_TFCS_Information_TDD },
  { &hf_rnsap_e_DCH_MACdFlows_to_Add, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_MACdFlows_Information_TDD },
  { &hf_rnsap_e_DCH_MACdFlows_to_Delete, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_EDCH_MACdFlows_To_Delete },
  { &hf_rnsap_e_DCH_Non_Scheduled_Grant_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Non_Scheduled_Grant_Info },
  { &hf_rnsap_e_DCH_TDD_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_TDD_Information },
  { &hf_rnsap_e_DCH_TDD_Information_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_TDD_Information_to_Modify },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_Information_Reconfig(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_Information_Reconfig, E_DCH_Information_Reconfig_sequence);

  return offset;
}


static const per_sequence_t E_DCH_TDD_MACdFlow_Specific_InformationResp_Item_sequence[] = {
  { &hf_rnsap_e_DCH_MacdFlow_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp_Item, E_DCH_TDD_MACdFlow_Specific_InformationResp_Item_sequence);

  return offset;
}


static const per_sequence_t E_DCH_TDD_MACdFlow_Specific_InformationResp_sequence_of[1] = {
  { &hf_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp_Item },
};

static int
dissect_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp, E_DCH_TDD_MACdFlow_Specific_InformationResp_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t E_AGCH_Specific_InformationResp_ItemTDD_sequence[] = {
  { &hf_rnsap_timeslot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_AGCH_Specific_InformationResp_ItemTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_AGCH_Specific_InformationResp_ItemTDD, E_AGCH_Specific_InformationResp_ItemTDD_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_Specific_InformationRespListTDD_sequence_of[1] = {
  { &hf_rnsap_E_AGCH_Specific_InformationRespListTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_AGCH_Specific_InformationResp_ItemTDD },
};

static int
dissect_rnsap_E_AGCH_Specific_InformationRespListTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_AGCH_Specific_InformationRespListTDD, E_AGCH_Specific_InformationRespListTDD_sequence_of,
                                                  1, maxNrOfEAGCHCodes, FALSE);

  return offset;
}



static int
dissect_rnsap_E_HICH_TimeOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            4U, 44U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_HICH_InformationResp_sequence[] = {
  { &hf_rnsap_timeslot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_e_HICH_TimeOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_HICH_TimeOffset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_HICH_InformationResp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_HICH_InformationResp, E_HICH_InformationResp_sequence);

  return offset;
}


static const per_sequence_t E_DCH_Information_Response_sequence[] = {
  { &hf_rnsap_e_DCH_TDD_MACdFlow_Specific_InformationResp, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp },
  { &hf_rnsap_e_AGCH_Specific_Information_ResponseTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_AGCH_Specific_InformationRespListTDD },
  { &hf_rnsap_e_HICH_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_HICH_InformationResp },
  { &hf_rnsap_e_DCH_Non_Scheduled_Grant_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Non_Scheduled_Grant_Info },
  { &hf_rnsap_e_RNTI        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_RNTI },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_Information_Response, E_DCH_Information_Response_sequence);

  return offset;
}


static const value_string rnsap_ActivationDelay_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  {   2, "v2" },
  {   3, "v3" },
  {   4, "v4" },
  {   5, "v5" },
  { 0, NULL }
};


static int
dissect_rnsap_ActivationDelay(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     6, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_Mode1_Ind(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_Fast_TTI_switching_Mode_synchronized_vals[] = {
  {   0, "mode1" },
  {   1, "mode2" },
  { 0, NULL }
};

static const per_choice_t Fast_TTI_switching_Mode_synchronized_choice[] = {
  {   0, &hf_rnsap_mode1         , ASN1_NO_EXTENSIONS     , dissect_rnsap_Mode1_Ind },
  {   1, &hf_rnsap_mode2         , ASN1_NO_EXTENSIONS     , dissect_rnsap_CFN },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Fast_TTI_switching_Mode_synchronized(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Fast_TTI_switching_Mode_synchronized, Fast_TTI_switching_Mode_synchronized_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_Fast_TTI_switching_Mode_unsynchronized_vals[] = {
  {   0, "mode1" },
  {   1, "mode2" },
  { 0, NULL }
};

static const per_choice_t Fast_TTI_switching_Mode_unsynchronized_choice[] = {
  {   0, &hf_rnsap_mode1_01      , ASN1_NO_EXTENSIONS     , dissect_rnsap_ActivationDelay },
  {   1, &hf_rnsap_mode2         , ASN1_NO_EXTENSIONS     , dissect_rnsap_CFN },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Fast_TTI_switching_Mode_unsynchronized(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Fast_TTI_switching_Mode_unsynchronized, Fast_TTI_switching_Mode_unsynchronized_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_Fast_TTI_switching_Mode_Supported_vals[] = {
  {   0, "mode1" },
  {   1, "mode2" },
  { 0, NULL }
};


static int
dissect_rnsap_Fast_TTI_switching_Mode_Supported(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_E_DCH_TDD_Maximum_Bitrate768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 17713U, NULL, TRUE);

  return offset;
}


static const per_sequence_t E_DCH_TDD_Information768_sequence[] = {
  { &hf_rnsap_e_DCH_TDD_Maximum_Bitrate768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_TDD_Maximum_Bitrate768 },
  { &hf_rnsap_e_DCH_Processing_Overload_Level, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Processing_Overload_Level },
  { &hf_rnsap_e_DCH_PowerOffset_for_SchedulingInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_PowerOffset_for_SchedulingInfo },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_TDD_Information768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_TDD_Information768, E_DCH_TDD_Information768_sequence);

  return offset;
}


static const per_sequence_t E_DCH_768_Information_sequence[] = {
  { &hf_rnsap_e_PUCH_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_PUCH_Information },
  { &hf_rnsap_e_TFCS_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_TFCS_Information_TDD },
  { &hf_rnsap_e_DCH_MACdFlows_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_MACdFlows_Information_TDD },
  { &hf_rnsap_e_DCH_TDD_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_TDD_Information768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_768_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_768_Information, E_DCH_768_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCH_Non_Scheduled_Grant_Info768_sequence[] = {
  { &hf_rnsap_timeslotResource, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_TimeslotResource },
  { &hf_rnsap_powerResource , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_PowerResource },
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tddE_PUCH_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TddE_PUCH_Offset },
  { &hf_rnsap_tdd_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_Non_Scheduled_Grant_Info768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_Non_Scheduled_Grant_Info768, E_DCH_Non_Scheduled_Grant_Info768_sequence);

  return offset;
}


static const per_sequence_t E_DCH_768_Information_Reconfig_sequence[] = {
  { &hf_rnsap_e_PUCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_PUCH_Information },
  { &hf_rnsap_e_TFCS_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_TFCS_Information_TDD },
  { &hf_rnsap_e_DCH_MACdFlows_to_Add, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_MACdFlows_Information_TDD },
  { &hf_rnsap_e_DCH_MACdFlows_to_Delete, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_EDCH_MACdFlows_To_Delete },
  { &hf_rnsap_e_DCH_Non_Scheduled_Grant_Info768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Non_Scheduled_Grant_Info768 },
  { &hf_rnsap_e_DCH_TDD_Information768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_TDD_Information768 },
  { &hf_rnsap_e_DCH_TDD_Information_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_TDD_Information_to_Modify },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_768_Information_Reconfig(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_768_Information_Reconfig, E_DCH_768_Information_Reconfig_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_Specific_InformationResp_Item768TDD_sequence[] = {
  { &hf_rnsap_timeslot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_tDD_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_AGCH_Specific_InformationResp_Item768TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_AGCH_Specific_InformationResp_Item768TDD, E_AGCH_Specific_InformationResp_Item768TDD_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_Specific_InformationRespList768TDD_sequence_of[1] = {
  { &hf_rnsap_E_AGCH_Specific_InformationRespList768TDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_AGCH_Specific_InformationResp_Item768TDD },
};

static int
dissect_rnsap_E_AGCH_Specific_InformationRespList768TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_AGCH_Specific_InformationRespList768TDD, E_AGCH_Specific_InformationRespList768TDD_sequence_of,
                                                  1, maxNrOfEAGCHCodes, FALSE);

  return offset;
}


static const per_sequence_t E_HICH_InformationResp768_sequence[] = {
  { &hf_rnsap_timeslot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_tDD_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode768 },
  { &hf_rnsap_e_HICH_TimeOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_HICH_TimeOffset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_HICH_InformationResp768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_HICH_InformationResp768, E_HICH_InformationResp768_sequence);

  return offset;
}


static const per_sequence_t E_DCH_768_Information_Response_sequence[] = {
  { &hf_rnsap_e_DCH_TDD_MACdFlow_Specific_InformationResp, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp },
  { &hf_rnsap_e_AGCH_Specific_Information_Response768TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_AGCH_Specific_InformationRespList768TDD },
  { &hf_rnsap_e_HICH_Information_Response768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_HICH_InformationResp768 },
  { &hf_rnsap_e_DCH_Non_Scheduled_Grant_Info768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Non_Scheduled_Grant_Info768 },
  { &hf_rnsap_e_RNTI        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_RNTI },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_768_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_768_Information_Response, E_DCH_768_Information_Response_sequence);

  return offset;
}



static int
dissect_rnsap_E_PUCH_PRXdesBase(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -112, -50, NULL, FALSE);

  return offset;
}


static const value_string rnsap_TDD_TPC_UplinkStepSize_LCR_vals[] = {
  {   0, "step-size1" },
  {   1, "step-size2" },
  {   2, "step-size3" },
  { 0, NULL }
};


static int
dissect_rnsap_TDD_TPC_UplinkStepSize_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_N_E_UCCH_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 8U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_PUCH_LCR_Information_sequence[] = {
  { &hf_rnsap_minCR         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CodeRate },
  { &hf_rnsap_maxCR         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CodeRate },
  { &hf_rnsap_harqInfo      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HARQ_Info_for_E_DCH },
  { &hf_rnsap_pRxdesBase    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_PUCH_PRXdesBase },
  { &hf_rnsap_e_PUCH_TPC_Step_Size, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_TPC_UplinkStepSize_LCR },
  { &hf_rnsap_n_E_UCCH_LCR  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_N_E_UCCH_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_PUCH_LCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_PUCH_LCR_Information, E_PUCH_LCR_Information_sequence);

  return offset;
}



static int
dissect_rnsap_E_DCH_Physical_Layer_Category_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 5U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_DCH_LCR_TDD_Information_sequence[] = {
  { &hf_rnsap_e_DCH_Physical_Layer_Category_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Physical_Layer_Category_LCR },
  { &hf_rnsap_e_DCH_Processing_Overload_Level, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Processing_Overload_Level },
  { &hf_rnsap_e_DCH_PowerOffset_for_SchedulingInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_PowerOffset_for_SchedulingInfo },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_LCR_TDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_LCR_TDD_Information, E_DCH_LCR_TDD_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCH_LCR_Information_sequence[] = {
  { &hf_rnsap_e_PUCH_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_PUCH_LCR_Information },
  { &hf_rnsap_e_TFCS_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_TFCS_Information_TDD },
  { &hf_rnsap_e_DCH_MACdFlows_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_MACdFlows_Information_TDD },
  { &hf_rnsap_e_DCH_LCR_TDD_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_LCR_TDD_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_LCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_LCR_Information, E_DCH_LCR_Information_sequence);

  return offset;
}



static int
dissect_rnsap_Extended_E_DCH_LCRTDD_PhysicalLayerCategory(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            6U, 6U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_Multi_Carrier_E_DCH_LCRTDD_PhysicalLayerCategory(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 8U, NULL, TRUE);

  return offset;
}


static const per_sequence_t E_DCH_LCR_Information_Reconfig_sequence[] = {
  { &hf_rnsap_e_PUCH_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_PUCH_LCR_Information },
  { &hf_rnsap_e_TFCS_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_TFCS_Information_TDD },
  { &hf_rnsap_e_DCH_MACdFlows_to_Add, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_MACdFlows_Information_TDD },
  { &hf_rnsap_e_DCH_MACdFlows_to_Delete, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_EDCH_MACdFlows_To_Delete },
  { &hf_rnsap_e_DCH_LCR_TDD_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_LCR_TDD_Information },
  { &hf_rnsap_e_DCH_TDD_Information_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_TDD_Information_to_Modify },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_LCR_Information_Reconfig(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_LCR_Information_Reconfig, E_DCH_LCR_Information_Reconfig_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_Specific_InformationResp_Item_LCR_TDD_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_AGCH_Specific_InformationResp_Item_LCR_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_AGCH_Specific_InformationResp_Item_LCR_TDD, E_AGCH_Specific_InformationResp_Item_LCR_TDD_sequence);

  return offset;
}


static const per_sequence_t E_AGCH_Specific_InformationRespList_LCR_TDD_sequence_of[1] = {
  { &hf_rnsap_E_AGCH_Specific_InformationRespList_LCR_TDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_AGCH_Specific_InformationResp_Item_LCR_TDD },
};

static int
dissect_rnsap_E_AGCH_Specific_InformationRespList_LCR_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_AGCH_Specific_InformationRespList_LCR_TDD, E_AGCH_Specific_InformationRespList_LCR_TDD_sequence_of,
                                                  1, maxNrOfEAGCHCodes, FALSE);

  return offset;
}



static int
dissect_rnsap_E_HICH_EI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 3U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_SignatureSequenceGroupIndex(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 19U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_HICH_InformationResp_LCR_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_signatureSequenceGroupIndex, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SignatureSequenceGroupIndex },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_HICH_InformationResp_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_HICH_InformationResp_LCR, E_HICH_InformationResp_LCR_sequence);

  return offset;
}


static const per_sequence_t E_HICH_Scheduled_InformationResp_Item_LCR_TDD_sequence[] = {
  { &hf_rnsap_e_HICH_EI     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_HICH_EI },
  { &hf_rnsap_e_HICH_Scheduled_InformationResp_LCR_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_HICH_InformationResp_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_HICH_Scheduled_InformationResp_Item_LCR_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_HICH_Scheduled_InformationResp_Item_LCR_TDD, E_HICH_Scheduled_InformationResp_Item_LCR_TDD_sequence);

  return offset;
}


static const per_sequence_t E_HICH_Scheduled_InformationRespList_LCR_TDD_sequence_of[1] = {
  { &hf_rnsap_E_HICH_Scheduled_InformationRespList_LCR_TDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_E_HICH_Scheduled_InformationResp_Item_LCR_TDD },
};

static int
dissect_rnsap_E_HICH_Scheduled_InformationRespList_LCR_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_E_HICH_Scheduled_InformationRespList_LCR_TDD, E_HICH_Scheduled_InformationRespList_LCR_TDD_sequence_of,
                                                  1, maxNrOfEHICHCodes, FALSE);

  return offset;
}



static int
dissect_rnsap_E_HICH_TimeOffset_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            4U, 15U, NULL, FALSE);

  return offset;
}


static const per_sequence_t E_HICH_Specific_InformationResp_LCR_sequence[] = {
  { &hf_rnsap_e_HICH_Scheduled_InformationResp_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_HICH_Scheduled_InformationRespList_LCR_TDD },
  { &hf_rnsap_e_HICH_non_Scheduled_InformationResp_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_HICH_InformationResp_LCR },
  { &hf_rnsap_e_HICH_TimeOffset_lcr, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_HICH_TimeOffset_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_HICH_Specific_InformationResp_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_HICH_Specific_InformationResp_LCR, E_HICH_Specific_InformationResp_LCR_sequence);

  return offset;
}



static int
dissect_rnsap_E_DCH_TimeslotResource_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     5, 5, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const value_string rnsap_E_DCH_SubframeNumber_LCR_vals[] = {
  {   0, "s0" },
  {   1, "s1" },
  { 0, NULL }
};


static int
dissect_rnsap_E_DCH_SubframeNumber_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t E_DCH_Non_Scheduled_Grant_Info_LCR_sequence[] = {
  { &hf_rnsap_timeslotResource_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_TimeslotResource_LCR },
  { &hf_rnsap_powerResource , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_PowerResource },
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_subframenumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_SubframeNumber_LCR },
  { &hf_rnsap_tddE_PUCH_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TddE_PUCH_Offset },
  { &hf_rnsap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_Non_Scheduled_Grant_Info_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_Non_Scheduled_Grant_Info_LCR, E_DCH_Non_Scheduled_Grant_Info_LCR_sequence);

  return offset;
}


static const per_sequence_t E_DCH_LCR_Information_Response_sequence[] = {
  { &hf_rnsap_e_DCH_TDD_MACdFlow_Specific_InformationResp, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp },
  { &hf_rnsap_e_AGCH_Specific_Information_Response_LCR_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_AGCH_Specific_InformationRespList_LCR_TDD },
  { &hf_rnsap_e_HICH_Specific_Information_Response_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_HICH_Specific_InformationResp_LCR },
  { &hf_rnsap_e_DCH_Non_Scheduled_Grant_Info_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_Non_Scheduled_Grant_Info_LCR },
  { &hf_rnsap_e_RNTI        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_RNTI },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_LCR_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_LCR_Information_Response, E_DCH_LCR_Information_Response_sequence);

  return offset;
}



static int
dissect_rnsap_Ext_Reference_E_TFCI_PO(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            30U, 31U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_ExtendedPropagationDelay(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            255U, 1023U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_T_radio_Links_without_DPCH_FDPCH_Operation_Indicator_vals[] = {
  {   0, "true" },
  { 0, NULL }
};


static int
dissect_rnsap_T_radio_Links_without_DPCH_FDPCH_Operation_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t Radio_Links_without_DPCH_FDPCH_Information_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_radio_Links_without_DPCH_FDPCH_Operation_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_radio_Links_without_DPCH_FDPCH_Operation_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Radio_Links_without_DPCH_FDPCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Radio_Links_without_DPCH_FDPCH_Information, Radio_Links_without_DPCH_FDPCH_Information_sequence);

  return offset;
}


static const per_sequence_t Radio_Links_without_DPCH_FDPCH_Indication_sequence_of[1] = {
  { &hf_rnsap_Radio_Links_without_DPCH_FDPCH_Indication_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Radio_Links_without_DPCH_FDPCH_Information },
};

static int
dissect_rnsap_Radio_Links_without_DPCH_FDPCH_Indication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Radio_Links_without_DPCH_FDPCH_Indication, Radio_Links_without_DPCH_FDPCH_Indication_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}



static int
dissect_rnsap_Extended_RNTI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1048576U, 4194303U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_Extended_Round_Trip_Time_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            32767U, 103041U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_T_extended_sRNTI_BitMaskIndex_vals[] = {
  {   0, "b1" },
  {   1, "b2" },
  {   2, "b3" },
  {   3, "b4" },
  {   4, "b5" },
  {   5, "b6" },
  {   6, "b7" },
  {   7, "b8" },
  {   8, "b9" },
  {   9, "b10" },
  {  10, "b11" },
  {  11, "b12" },
  {  12, "b13" },
  {  13, "b14" },
  {  14, "b15" },
  {  15, "b16" },
  {  16, "b17" },
  {  17, "b18" },
  {  18, "b19" },
  {  19, "b20" },
  {  20, "b21" },
  { 0, NULL }
};


static int
dissect_rnsap_T_extended_sRNTI_BitMaskIndex(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     21, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Extended_S_RNTI_Group_sequence[] = {
  { &hf_rnsap_extended_sRNTI, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Extended_RNTI },
  { &hf_rnsap_extended_sRNTI_BitMaskIndex, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_T_extended_sRNTI_BitMaskIndex },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Extended_S_RNTI_Group(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Extended_S_RNTI_Group, Extended_S_RNTI_Group_sequence);

  return offset;
}


static const per_sequence_t ExtendedAffectedUEInformationForMBMS_sequence_of[1] = {
  { &hf_rnsap_ExtendedAffectedUEInformationForMBMS_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Extended_RNTI },
};

static int
dissect_rnsap_ExtendedAffectedUEInformationForMBMS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_ExtendedAffectedUEInformationForMBMS, ExtendedAffectedUEInformationForMBMS_sequence_of,
                                                  0, maxNrOfUEs, FALSE);

  return offset;
}



static int
dissect_rnsap_Ext_Max_Bits_MACe_PDU_non_scheduled(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            19983U, 22978U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_T_subframeNumber_vals[] = {
  {   0, "v0" },
  {   1, "v1" },
  { 0, NULL }
};


static int
dissect_rnsap_T_subframeNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t Initial_E_DCH_SPS_resource_sequence[] = {
  { &hf_rnsap_timeslot_Resource_Related_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_TimeslotResource_LCR },
  { &hf_rnsap_powerResource , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DCH_PowerResource },
  { &hf_rnsap_repetitionPeriodIndex, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriodIndex },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_subframeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_subframeNumber },
  { &hf_rnsap_tddE_PUCH_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TddE_PUCH_Offset },
  { &hf_rnsap_tdd_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_n_E_UCCHLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_N_E_UCCH_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Initial_E_DCH_SPS_resource(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Initial_E_DCH_SPS_resource, Initial_E_DCH_SPS_resource_sequence);

  return offset;
}


static const per_sequence_t Same_As_Scheduled_E_HICH_sequence[] = {
  { &hf_rnsap_e_HICH_EI     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_HICH_EI },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Same_As_Scheduled_E_HICH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Same_As_Scheduled_E_HICH, Same_As_Scheduled_E_HICH_sequence);

  return offset;
}


static const per_sequence_t E_HICH_InformationResp_ExplicitConfiguration_LCR_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_HICH_InformationResp_ExplicitConfiguration_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_HICH_InformationResp_ExplicitConfiguration_LCR, E_HICH_InformationResp_ExplicitConfiguration_LCR_sequence);

  return offset;
}


static const value_string rnsap_E_HICH_Configuration_vals[] = {
  {   0, "same-As-Scheduled-E-HICH" },
  {   1, "explicit" },
  { 0, NULL }
};

static const per_choice_t E_HICH_Configuration_choice[] = {
  {   0, &hf_rnsap_same_As_Scheduled_E_HICH, ASN1_EXTENSION_ROOT    , dissect_rnsap_Same_As_Scheduled_E_HICH },
  {   1, &hf_rnsap_explicit      , ASN1_EXTENSION_ROOT    , dissect_rnsap_E_HICH_InformationResp_ExplicitConfiguration_LCR },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_E_HICH_Configuration(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_E_HICH_Configuration, E_HICH_Configuration_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t E_DCH_SPS_HICH_Information_sequence[] = {
  { &hf_rnsap_e_HICH_Configuration, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_HICH_Configuration },
  { &hf_rnsap_signatureSequenceGroupIndex, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SignatureSequenceGroupIndex },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_SPS_HICH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_SPS_HICH_Information, E_DCH_SPS_HICH_Information_sequence);

  return offset;
}


static const per_sequence_t E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_sequence[] = {
  { &hf_rnsap_initial_E_DCH_SPS_resource, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Initial_E_DCH_SPS_resource },
  { &hf_rnsap_e_DCH_SPS_HICH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_SPS_HICH_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR, E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_sequence);

  return offset;
}


static const value_string rnsap_FACH_InitialWindowSize_vals[] = {
  { 255, "unlimited" },
  { 0, NULL }
};


static int
dissect_rnsap_FACH_InitialWindowSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t FACH_FlowControlInformationItem_sequence[] = {
  { &hf_rnsap_fACH_SchedulingPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_mAC_c_sh_SDU_Lengths, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAC_c_sh_SDU_LengthList },
  { &hf_rnsap_fACH_InitialWindowSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FACH_InitialWindowSize },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_FACH_FlowControlInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_FACH_FlowControlInformationItem, FACH_FlowControlInformationItem_sequence);

  return offset;
}


static const per_sequence_t FACH_FlowControlInformation_sequence_of[1] = {
  { &hf_rnsap_FACH_FlowControlInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_FACH_FlowControlInformationItem },
};

static int
dissect_rnsap_FACH_FlowControlInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_FACH_FlowControlInformation, FACH_FlowControlInformation_sequence_of,
                                                  1, 16, FALSE);

  return offset;
}


static const per_sequence_t FACH_InformationItem_sequence[] = {
  { &hf_rnsap_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_FACH_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_FACH_InformationItem, FACH_InformationItem_sequence);

  return offset;
}


static const per_sequence_t FACH_InformationList_sequence_of[1] = {
  { &hf_rnsap_FACH_InformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_FACH_InformationItem },
};

static int
dissect_rnsap_FACH_InformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_FACH_InformationList, FACH_InformationList_sequence_of,
                                                  0, maxNrOfFACHs, FALSE);

  return offset;
}


static const value_string rnsap_Fast_Reconfiguration_Mode_vals[] = {
  {   0, "fast" },
  { 0, NULL }
};


static int
dissect_rnsap_Fast_Reconfiguration_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_Fast_Reconfiguration_Permission_vals[] = {
  {   0, "allowed" },
  { 0, NULL }
};


static int
dissect_rnsap_Fast_Reconfiguration_Permission(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t FDD_DCHs_to_ModifySpecificItem_sequence[] = {
  { &hf_rnsap_dCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_ID },
  { &hf_rnsap_ul_TransportformatSet, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_dl_TransportformatSet, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_frameHandlingPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_FrameHandlingPriority },
  { &hf_rnsap_not_Used_dRACControl, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_FDD_DCHs_to_ModifySpecificItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_FDD_DCHs_to_ModifySpecificItem, FDD_DCHs_to_ModifySpecificItem_sequence);

  return offset;
}


static const per_sequence_t FDD_DCHs_to_ModifySpecificInformationList_sequence_of[1] = {
  { &hf_rnsap_FDD_DCHs_to_ModifySpecificInformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DCHs_to_ModifySpecificItem },
};

static int
dissect_rnsap_FDD_DCHs_to_ModifySpecificInformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_FDD_DCHs_to_ModifySpecificInformationList, FDD_DCHs_to_ModifySpecificInformationList_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t FDD_DCHs_to_ModifyItem_sequence[] = {
  { &hf_rnsap_ul_FP_Mode    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_FP_Mode },
  { &hf_rnsap_toAWS         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ToAWS },
  { &hf_rnsap_toAWE         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ToAWE },
  { &hf_rnsap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportBearerRequestIndicator },
  { &hf_rnsap_dCH_SpecificInformationList_02, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DCHs_to_ModifySpecificInformationList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_FDD_DCHs_to_ModifyItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_FDD_DCHs_to_ModifyItem, FDD_DCHs_to_ModifyItem_sequence);

  return offset;
}


static const per_sequence_t FDD_DCHs_to_Modify_sequence_of[1] = {
  { &hf_rnsap_FDD_DCHs_to_Modify_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DCHs_to_ModifyItem },
};

static int
dissect_rnsap_FDD_DCHs_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_FDD_DCHs_to_Modify, FDD_DCHs_to_Modify_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t F_DPCH_Info_sequence[] = {
  { &hf_rnsap_f_DPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_F_DPCH_SlotFormat },
  { &hf_rnsap_fdd_dl_ChannelisationCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DL_ChannelisationCodeNumber },
  { &hf_rnsap_extended_E_DPCCH_PO, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Extended_E_DPCCH_PO },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_F_DPCH_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_F_DPCH_Info, F_DPCH_Info_sequence);

  return offset;
}


static const per_sequence_t F_DPCH_Info_To_Modify_sequence[] = {
  { &hf_rnsap_f_DPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_F_DPCH_SlotFormat },
  { &hf_rnsap_fdd_dl_ChannelisationCodeNumber, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_FDD_DL_ChannelisationCodeNumber },
  { &hf_rnsap_extended_E_DPCCH_PO, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Extended_E_DPCCH_PO },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_F_DPCH_Info_To_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_F_DPCH_Info_To_Modify, F_DPCH_Info_To_Modify_sequence);

  return offset;
}


static const value_string rnsap_FNReportingIndicator_vals[] = {
  {   0, "fN-reporting-required" },
  {   1, "fN-reporting-not-required" },
  { 0, NULL }
};


static int
dissect_rnsap_FNReportingIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_4(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 4U, NULL, FALSE);

  return offset;
}


static const per_sequence_t FPACH_Information_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_tDD_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCodeLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_wT            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_4 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_FPACH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_FPACH_Information, FPACH_Information_sequence);

  return offset;
}



static int
dissect_rnsap_FTPICH_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 149U, NULL, FALSE);

  return offset;
}


static const per_sequence_t FTPICH_Information_sequence[] = {
  { &hf_rnsap_fTPICH_Offset , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FTPICH_Offset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_FTPICH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_FTPICH_Information, FTPICH_Information_sequence);

  return offset;
}



static int
dissect_rnsap_FTPICH_SlotFormat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 9U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_FTPICH_Information_Removal_vals[] = {
  {   0, "remove" },
  { 0, NULL }
};


static int
dissect_rnsap_FTPICH_Information_Removal(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t FTPICH_Information_To_Modify_sequence[] = {
  { &hf_rnsap_fTPICH_Offset , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_FTPICH_Offset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_FTPICH_Information_To_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_FTPICH_Information_To_Modify, FTPICH_Information_To_Modify_sequence);

  return offset;
}


static const value_string rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information_vals[] = {
  {   0, "setup" },
  {   1, "configurationChange" },
  {   2, "removal" },
  { 0, NULL }
};

static const per_choice_t Setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information_choice[] = {
  {   0, &hf_rnsap_setup_05      , ASN1_EXTENSION_ROOT    , dissect_rnsap_FTPICH_Information },
  {   1, &hf_rnsap_configurationChange_04, ASN1_EXTENSION_ROOT    , dissect_rnsap_FTPICH_Information_To_Modify },
  {   2, &hf_rnsap_removal_03    , ASN1_EXTENSION_ROOT    , dissect_rnsap_FTPICH_Information_Removal },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information, Setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t FTPICH_Information_Reconf_sequence[] = {
  { &hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_FTPICH_Information_Reconf(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_FTPICH_Information_Reconf, FTPICH_Information_Reconf_sequence);

  return offset;
}


static const per_sequence_t FTPICH_Reconfiguration_Information_sequence[] = {
  { &hf_rnsap_fTPICH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_FTPICH_SlotFormat },
  { &hf_rnsap_fTPICH_ChannelisationCodenumber, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_FDD_DL_ChannelisationCodeNumber },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_FTPICH_Reconfiguration_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_FTPICH_Reconfiguration_Information, FTPICH_Reconfiguration_Information_sequence);

  return offset;
}


static const per_sequence_t FTPICH_Information_Response_sequence[] = {
  { &hf_rnsap_fTPICH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FTPICH_SlotFormat },
  { &hf_rnsap_fTPICH_ChannelisationCodenumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DL_ChannelisationCodeNumber },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_FTPICH_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_FTPICH_Information_Response, FTPICH_Information_Response_sequence);

  return offset;
}



static int
dissect_rnsap_FrameOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_GapLength(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 14U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_GapDuration(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 144U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_T_latitudeSign_01_vals[] = {
  {   0, "north" },
  {   1, "south" },
  { 0, NULL }
};


static int
dissect_rnsap_T_latitudeSign_01(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_8388607(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8388607U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_M8388608_8388607(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -8388608, 8388607U, NULL, FALSE);

  return offset;
}


static const per_sequence_t GeographicalCoordinate_sequence[] = {
  { &hf_rnsap_latitudeSign_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_latitudeSign_01 },
  { &hf_rnsap_latitude      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_8388607 },
  { &hf_rnsap_longitude     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_M8388608_8388607 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GeographicalCoordinate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GeographicalCoordinate, GeographicalCoordinate_sequence);

  return offset;
}


static const per_sequence_t GA_Cell_item_sequence[] = {
  { &hf_rnsap_cell_GAIgeographicalCoordinate, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GeographicalCoordinate },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GA_Cell_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GA_Cell_item, GA_Cell_item_sequence);

  return offset;
}


static const per_sequence_t GA_Cell_sequence_of[1] = {
  { &hf_rnsap_GA_Cell_item  , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GA_Cell_item },
};

static int
dissect_rnsap_GA_Cell(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GA_Cell, GA_Cell_sequence_of,
                                                  1, maxNrOfPoints, FALSE);

  return offset;
}


static const per_sequence_t GA_PointWithUnCertainty_sequence[] = {
  { &hf_rnsap_geographicalCoordinates, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GeographicalCoordinate },
  { &hf_rnsap_uncertaintyCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_127 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GA_PointWithUnCertainty(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GA_PointWithUnCertainty, GA_PointWithUnCertainty_sequence);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_179(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 179U, NULL, FALSE);

  return offset;
}


static const per_sequence_t GA_UncertaintyEllipse_sequence[] = {
  { &hf_rnsap_uncertaintySemi_major, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_127 },
  { &hf_rnsap_uncertaintySemi_minor, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_127 },
  { &hf_rnsap_orientationOfMajorAxis, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_179 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GA_UncertaintyEllipse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GA_UncertaintyEllipse, GA_UncertaintyEllipse_sequence);

  return offset;
}


static const per_sequence_t GA_PointWithUnCertaintyEllipse_sequence[] = {
  { &hf_rnsap_geographicalCoordinates, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GeographicalCoordinate },
  { &hf_rnsap_uncertaintyEllipse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GA_UncertaintyEllipse },
  { &hf_rnsap_confidence    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_127 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GA_PointWithUnCertaintyEllipse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GA_PointWithUnCertaintyEllipse, GA_PointWithUnCertaintyEllipse_sequence);

  return offset;
}


static const value_string rnsap_T_directionOfAltitude_vals[] = {
  {   0, "height" },
  {   1, "depth" },
  { 0, NULL }
};


static int
dissect_rnsap_T_directionOfAltitude(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t GA_AltitudeAndDirection_sequence[] = {
  { &hf_rnsap_directionOfAltitude, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_directionOfAltitude },
  { &hf_rnsap_altitude      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_32767 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GA_AltitudeAndDirection(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GA_AltitudeAndDirection, GA_AltitudeAndDirection_sequence);

  return offset;
}


static const per_sequence_t GA_PointWithAltitude_sequence[] = {
  { &hf_rnsap_geographicalCoordinates, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GeographicalCoordinate },
  { &hf_rnsap_altitudeAndDirection, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GA_AltitudeAndDirection },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GA_PointWithAltitude(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GA_PointWithAltitude, GA_PointWithAltitude_sequence);

  return offset;
}


static const per_sequence_t GA_PointWithAltitudeAndUncertaintyEllipsoid_sequence[] = {
  { &hf_rnsap_geographicalCoordinates, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GeographicalCoordinate },
  { &hf_rnsap_altitudeAndDirection, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GA_AltitudeAndDirection },
  { &hf_rnsap_uncertaintyEllipse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GA_UncertaintyEllipse },
  { &hf_rnsap_uncertaintyAltitude, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_127 },
  { &hf_rnsap_confidence    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_127 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GA_PointWithAltitudeAndUncertaintyEllipsoid(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GA_PointWithAltitudeAndUncertaintyEllipsoid, GA_PointWithAltitudeAndUncertaintyEllipsoid_sequence);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_65535(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}


static const per_sequence_t GA_EllipsoidArc_sequence[] = {
  { &hf_rnsap_geographicalCoordinates, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GeographicalCoordinate },
  { &hf_rnsap_innerRadius   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_65535 },
  { &hf_rnsap_uncertaintyRadius, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_127 },
  { &hf_rnsap_offsetAngle   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_179 },
  { &hf_rnsap_includedAngle , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_179 },
  { &hf_rnsap_confidence    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_127 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GA_EllipsoidArc(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GA_EllipsoidArc, GA_EllipsoidArc_sequence);

  return offset;
}


static const value_string rnsap_GA_CellAdditionalShapes_vals[] = {
  {   0, "pointWithUncertainty" },
  {   1, "pointWithUncertaintyEllipse" },
  {   2, "pointWithAltitude" },
  {   3, "pointWithAltitudeAndUncertaintyEllipsoid" },
  {   4, "ellipsoidArc" },
  { 0, NULL }
};

static const per_choice_t GA_CellAdditionalShapes_choice[] = {
  {   0, &hf_rnsap_pointWithUncertainty, ASN1_EXTENSION_ROOT    , dissect_rnsap_GA_PointWithUnCertainty },
  {   1, &hf_rnsap_pointWithUncertaintyEllipse, ASN1_EXTENSION_ROOT    , dissect_rnsap_GA_PointWithUnCertaintyEllipse },
  {   2, &hf_rnsap_pointWithAltitude, ASN1_EXTENSION_ROOT    , dissect_rnsap_GA_PointWithAltitude },
  {   3, &hf_rnsap_pointWithAltitudeAndUncertaintyEllipsoid, ASN1_EXTENSION_ROOT    , dissect_rnsap_GA_PointWithAltitudeAndUncertaintyEllipsoid },
  {   4, &hf_rnsap_ellipsoidArc  , ASN1_EXTENSION_ROOT    , dissect_rnsap_GA_EllipsoidArc },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_GA_CellAdditionalShapes(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_GA_CellAdditionalShapes, GA_CellAdditionalShapes_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_RefTFCNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}


static const per_sequence_t T_signalledGainFactors10ms_sequence[] = {
  { &hf_rnsap_gain10ms_betaC, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BetaCD },
  { &hf_rnsap_gain10ms_betaD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BetaCD },
  { &hf_rnsap_gain10ms_refTFCNumber, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RefTFCNumber },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_T_signalledGainFactors10ms(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_T_signalledGainFactors10ms, T_signalledGainFactors10ms_sequence);

  return offset;
}


static const value_string rnsap_Gainfactors_10ms_mode_vals[] = {
  {   0, "signalledGainFactors10ms" },
  {   1, "computedGainFactors10ms" },
  { 0, NULL }
};

static const per_choice_t Gainfactors_10ms_mode_choice[] = {
  {   0, &hf_rnsap_signalledGainFactors10ms, ASN1_EXTENSION_ROOT    , dissect_rnsap_T_signalledGainFactors10ms },
  {   1, &hf_rnsap_computedGainFactors10ms, ASN1_EXTENSION_ROOT    , dissect_rnsap_RefTFCNumber },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Gainfactors_10ms_mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Gainfactors_10ms_mode, Gainfactors_10ms_mode_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_16(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     16, 16, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_22(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     22, 22, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_NAVclockModel_sequence[] = {
  { &hf_rnsap_navToc        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_navaf2        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_navaf1        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_navaf0        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_22 },
  { &hf_rnsap_navTgd        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_NAVclockModel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_NAVclockModel, GANSS_NAVclockModel_sequence);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_11(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     11, 11, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_5(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     5, 5, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     3, 3, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_20(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     20, 20, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_26(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     26, 26, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_CNAVclockModel_sequence[] = {
  { &hf_rnsap_cnavToc       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_cnavTop       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_cnavURA0      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_5 },
  { &hf_rnsap_cnavURA1      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_3 },
  { &hf_rnsap_cnavURA2      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_3 },
  { &hf_rnsap_cnavAf2       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_10 },
  { &hf_rnsap_cnavAf1       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_20 },
  { &hf_rnsap_cnavAf0       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_26 },
  { &hf_rnsap_cnavTgd       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_cnavISCl1cp   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_cnavISCl1cd   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_cnavISCl1ca   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_cnavISCl2c    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_cnavISCl5i5   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_cnavISCl5q5   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_CNAVclockModel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_CNAVclockModel, GANSS_CNAVclockModel_sequence);

  return offset;
}


static const per_sequence_t GANSS_GLONASSclockModel_sequence[] = {
  { &hf_rnsap_gloTau        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_22 },
  { &hf_rnsap_gloGamma      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_gloDeltaTau   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_5 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_GLONASSclockModel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_GLONASSclockModel, GANSS_GLONASSclockModel_sequence);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_12(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     12, 12, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_SBASclockModel_sequence[] = {
  { &hf_rnsap_sbasTo        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_sbasAgfo      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_12 },
  { &hf_rnsap_sbasAgf1      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_SBASclockModel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_SBASclockModel, GANSS_SBASclockModel_sequence);

  return offset;
}



static int
dissect_rnsap_Extension_GANSS_AddClockModels(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_GANSS_AddClockModels_vals[] = {
  {   0, "navClockModel" },
  {   1, "cnavClockModel" },
  {   2, "glonassClockModel" },
  {   3, "sbasClockModel" },
  {   4, "extension-GANSS-AddClockModels" },
  { 0, NULL }
};

static const per_choice_t GANSS_AddClockModels_choice[] = {
  {   0, &hf_rnsap_navClockModel , ASN1_EXTENSION_ROOT    , dissect_rnsap_GANSS_NAVclockModel },
  {   1, &hf_rnsap_cnavClockModel, ASN1_EXTENSION_ROOT    , dissect_rnsap_GANSS_CNAVclockModel },
  {   2, &hf_rnsap_glonassClockModel, ASN1_EXTENSION_ROOT    , dissect_rnsap_GANSS_GLONASSclockModel },
  {   3, &hf_rnsap_sbasClockModel, ASN1_EXTENSION_ROOT    , dissect_rnsap_GANSS_SBASclockModel },
  {   4, &hf_rnsap_extension_GANSS_AddClockModels, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_GANSS_AddClockModels },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_GANSS_AddClockModels(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_GANSS_AddClockModels, GANSS_AddClockModels_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_GANSS_AddIonoModelReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     2, 2, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_GANSS_AddNavigationModelsReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     1, 1, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_32(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     32, 32, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_24(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     24, 24, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_14(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     14, 14, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_NavModel_NAVKeplerianSet_sequence[] = {
  { &hf_rnsap_navURA        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_4 },
  { &hf_rnsap_navFitFlag    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_1 },
  { &hf_rnsap_navToe        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_navOmega      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_navDeltaN     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_navM0         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_navOmegaADot  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_navE          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_navIDot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_14 },
  { &hf_rnsap_navAPowerHalf , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_navI0         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_navOmegaA0    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_navCrs        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_navCis        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_navCus        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_navCrc        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_navCic        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_navCuc        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_NavModel_NAVKeplerianSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_NavModel_NAVKeplerianSet, GANSS_NavModel_NAVKeplerianSet_sequence);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_25(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     25, 25, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_17(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     17, 17, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_23(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     23, 23, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_33(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     33, 33, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_15(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     15, 15, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_21(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     21, 21, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_NavModel_CNAVKeplerianSet_sequence[] = {
  { &hf_rnsap_cnavTop       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_cnavURAindex  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_5 },
  { &hf_rnsap_cnavDeltaA    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_26 },
  { &hf_rnsap_cnavAdot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_25 },
  { &hf_rnsap_cnavDeltaNo   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_17 },
  { &hf_rnsap_cnavDeltaNoDot, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_23 },
  { &hf_rnsap_cnavMo        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_33 },
  { &hf_rnsap_cnavE         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_33 },
  { &hf_rnsap_cnavOmega     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_33 },
  { &hf_rnsap_cnavOMEGA0    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_33 },
  { &hf_rnsap_cnavDeltaOmegaDot, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_17 },
  { &hf_rnsap_cnavIo        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_33 },
  { &hf_rnsap_cnavIoDot     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_15 },
  { &hf_rnsap_cnavCis       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_cnavCic       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_cnavCrs       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_cnavCrc       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_cnavCus       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_21 },
  { &hf_rnsap_cnavCuc       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_21 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_NavModel_CNAVKeplerianSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_NavModel_CNAVKeplerianSet, GANSS_NavModel_CNAVKeplerianSet_sequence);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     2, 2, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_27(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     27, 27, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_NavModel_GLONASSecef_sequence[] = {
  { &hf_rnsap_gloEn         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_5 },
  { &hf_rnsap_gloP1         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_2 },
  { &hf_rnsap_gloP2         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_1 },
  { &hf_rnsap_gloM          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_2 },
  { &hf_rnsap_gloX          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_27 },
  { &hf_rnsap_gloXdot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_gloXdotdot    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_5 },
  { &hf_rnsap_gloY          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_27 },
  { &hf_rnsap_gloYdot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_gloYdotdot    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_5 },
  { &hf_rnsap_gloZ          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_27 },
  { &hf_rnsap_gloZdot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_gloZdotdot    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_5 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_NavModel_GLONASSecef(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_NavModel_GLONASSecef, GANSS_NavModel_GLONASSecef_sequence);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_30(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     30, 30, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_18(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     18, 18, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_NavModel_SBASecef_sequence[] = {
  { &hf_rnsap_sbasTo        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_sbasAccuracy  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_4 },
  { &hf_rnsap_sbasXg        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_30 },
  { &hf_rnsap_sbasYg        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_30 },
  { &hf_rnsap_sbasZg        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_25 },
  { &hf_rnsap_sbasXgDot     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_17 },
  { &hf_rnsap_sbasYgDot     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_17 },
  { &hf_rnsap_sbasZgDot     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_18 },
  { &hf_rnsap_sbasXgDotDot  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_10 },
  { &hf_rnsap_sbagYgDotDot  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_10 },
  { &hf_rnsap_sbasZgDotDot  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_10 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_NavModel_SBASecef(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_NavModel_SBASecef, GANSS_NavModel_SBASecef_sequence);

  return offset;
}



static int
dissect_rnsap_Extension_GANSS_AddOrbitModels(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_GANSS_AddOrbitModels_vals[] = {
  {   0, "navKeplerianSet" },
  {   1, "cnavKeplerianSet" },
  {   2, "glonassECEF" },
  {   3, "sbasECEF" },
  {   4, "extension-GANSS-AddOrbitModels" },
  { 0, NULL }
};

static const per_choice_t GANSS_AddOrbitModels_choice[] = {
  {   0, &hf_rnsap_navKeplerianSet, ASN1_EXTENSION_ROOT    , dissect_rnsap_GANSS_NavModel_NAVKeplerianSet },
  {   1, &hf_rnsap_cnavKeplerianSet, ASN1_EXTENSION_ROOT    , dissect_rnsap_GANSS_NavModel_CNAVKeplerianSet },
  {   2, &hf_rnsap_glonassECEF   , ASN1_EXTENSION_ROOT    , dissect_rnsap_GANSS_NavModel_GLONASSecef },
  {   3, &hf_rnsap_sbasECEF      , ASN1_EXTENSION_ROOT    , dissect_rnsap_GANSS_NavModel_SBASecef },
  {   4, &hf_rnsap_extension_GANSS_AddOrbitModels, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_GANSS_AddOrbitModels },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_GANSS_AddOrbitModels(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_GANSS_AddOrbitModels, GANSS_AddOrbitModels_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_GANSS_AddUTCModelsReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t GPS_Ionospheric_Model_sequence[] = {
  { &hf_rnsap_alpha_zero_ionos_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_alpha_one_ionos_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_alpha_two_ionos_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_alpha_three_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_beta_zero_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_beta_one_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_beta_two_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_beta_three_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GPS_Ionospheric_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GPS_Ionospheric_Model, GPS_Ionospheric_Model_sequence);

  return offset;
}


static const per_sequence_t GANSS_Additional_Ionospheric_Model_sequence[] = {
  { &hf_rnsap_dataID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_2 },
  { &hf_rnsap_alpha_beta_parameters, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GPS_Ionospheric_Model },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Additional_Ionospheric_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Additional_Ionospheric_Model, GANSS_Additional_Ionospheric_Model_sequence);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_8191(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8191U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_86399(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 86399U, NULL, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Transmission_Time_sequence[] = {
  { &hf_rnsap_ganssDay      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_0_8191 },
  { &hf_rnsap_ganssTod_01   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_86399 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Transmission_Time(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Transmission_Time, GANSS_Transmission_Time_sequence);

  return offset;
}


static const value_string rnsap_T_non_broadcastIndication_vals[] = {
  {   0, "true" },
  { 0, NULL }
};


static int
dissect_rnsap_T_non_broadcastIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t Ganss_Sat_Info_AddNavList_item_sequence[] = {
  { &hf_rnsap_satId         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_svHealth      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_9 },
  { &hf_rnsap_iod           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_ganssAddClockModels, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_AddClockModels },
  { &hf_rnsap_ganssAddOrbitModels, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_AddOrbitModels },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Ganss_Sat_Info_AddNavList_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Ganss_Sat_Info_AddNavList_item, Ganss_Sat_Info_AddNavList_item_sequence);

  return offset;
}


static const per_sequence_t Ganss_Sat_Info_AddNavList_sequence_of[1] = {
  { &hf_rnsap_Ganss_Sat_Info_AddNavList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Ganss_Sat_Info_AddNavList_item },
};

static int
dissect_rnsap_Ganss_Sat_Info_AddNavList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Ganss_Sat_Info_AddNavList, Ganss_Sat_Info_AddNavList_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Additional_Navigation_Models_sequence[] = {
  { &hf_rnsap_ganss_Transmission_Time, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_Transmission_Time },
  { &hf_rnsap_non_broadcastIndication, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_T_non_broadcastIndication },
  { &hf_rnsap_ganssSatInfoNavList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Ganss_Sat_Info_AddNavList },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Additional_Navigation_Models(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Additional_Navigation_Models, GANSS_Additional_Navigation_Models_sequence);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_37799(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 37799U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_M2147483648_2147483647(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            G_MININT32, 2147483647U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_M64_63(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -64, 63U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_T_gnss_to_id_vals[] = {
  {   0, "gps" },
  {   1, "galileo" },
  {   2, "qzss" },
  {   3, "glonass" },
  {   4, "bds" },
  { 0, NULL }
};


static int
dissect_rnsap_T_gnss_to_id(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 4, NULL);

  return offset;
}


static const per_sequence_t GANSS_Time_Model_sequence[] = {
  { &hf_rnsap_ganss_time_model_Ref_Time, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_37799 },
  { &hf_rnsap_ganss_t_a0    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_M2147483648_2147483647 },
  { &hf_rnsap_ganss_t_a1    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_M8388608_8388607 },
  { &hf_rnsap_ganss_t_a2    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_M64_63 },
  { &hf_rnsap_gnss_to_id    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_gnss_to_id },
  { &hf_rnsap_ganss_wk_number_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_0_8191 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Time_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Time_Model, GANSS_Time_Model_sequence);

  return offset;
}


static const per_sequence_t GANSS_Additional_Time_Models_sequence_of[1] = {
  { &hf_rnsap_GANSS_Additional_Time_Models_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_Time_Model },
};

static int
dissect_rnsap_GANSS_Additional_Time_Models(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_Additional_Time_Models, GANSS_Additional_Time_Models_sequence_of,
                                                  1, maxGANSS_1, FALSE);

  return offset;
}


static const per_sequence_t GANSS_UTCmodelSet1_sequence[] = {
  { &hf_rnsap_utcA0         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_utcA1         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_utcA2         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_7 },
  { &hf_rnsap_utcDeltaTls   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_utcTot        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_utcWNot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_utcWNlsf      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_utcDN         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_4 },
  { &hf_rnsap_utcDeltaTlsf  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_UTCmodelSet1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_UTCmodelSet1, GANSS_UTCmodelSet1_sequence);

  return offset;
}


static const per_sequence_t GANSS_DeltaUT1_sequence[] = {
  { &hf_rnsap_b1            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_b2            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_10 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_DeltaUT1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_DeltaUT1, GANSS_DeltaUT1_sequence);

  return offset;
}


static const per_sequence_t GANSS_UTCmodelSet2_sequence[] = {
  { &hf_rnsap_nA            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_tauC          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_deltaUT1_01   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_DeltaUT1 },
  { &hf_rnsap_kp            , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_2 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_UTCmodelSet2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_UTCmodelSet2, GANSS_UTCmodelSet2_sequence);

  return offset;
}


static const per_sequence_t GANSS_UTCmodelSet3_sequence[] = {
  { &hf_rnsap_utcA1wnt      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_utcA0wnt      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_utcTot_01     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_utcWNt        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_utcDeltaTls   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_utcWNlsf      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_utcDN_01      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_utcDeltaTlsf  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_utcStandardID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_3 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_UTCmodelSet3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_UTCmodelSet3, GANSS_UTCmodelSet3_sequence);

  return offset;
}



static int
dissect_rnsap_Extension_GANSS_Additional_UTC_Models(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_GANSS_Additional_UTC_Models_vals[] = {
  {   0, "utcModel1" },
  {   1, "utcModel2" },
  {   2, "utcModel3" },
  {   3, "extension-GANSS-Additional-UTC-Models" },
  { 0, NULL }
};

static const per_choice_t GANSS_Additional_UTC_Models_choice[] = {
  {   0, &hf_rnsap_utcModel1     , ASN1_EXTENSION_ROOT    , dissect_rnsap_GANSS_UTCmodelSet1 },
  {   1, &hf_rnsap_utcModel2     , ASN1_EXTENSION_ROOT    , dissect_rnsap_GANSS_UTCmodelSet2 },
  {   2, &hf_rnsap_utcModel3     , ASN1_EXTENSION_ROOT    , dissect_rnsap_GANSS_UTCmodelSet3 },
  {   3, &hf_rnsap_extension_GANSS_Additional_UTC_Models, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_GANSS_Additional_UTC_Models },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_GANSS_Additional_UTC_Models(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_GANSS_Additional_UTC_Models, GANSS_Additional_UTC_Models_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_255(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_1023(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1023U, NULL, FALSE);

  return offset;
}


static const per_sequence_t GANSS_SatelliteInformationKP_item_sequence[] = {
  { &hf_rnsap_satId         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_ganss_e_alm   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_ganss_delta_I_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_ganss_omegadot_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_ganss_svStatusINAV_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_4 },
  { &hf_rnsap_ganss_svStatusFNAV_alm, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_2 },
  { &hf_rnsap_ganss_delta_a_sqrt_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_ganss_omegazero_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_ganss_m_zero_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_ganss_omega_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_ganss_af_zero_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_ganss_af_one_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_13 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_SatelliteInformationKP_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_SatelliteInformationKP_item, GANSS_SatelliteInformationKP_item_sequence);

  return offset;
}


static const per_sequence_t GANSS_SatelliteInformationKP_sequence_of[1] = {
  { &hf_rnsap_GANSS_SatelliteInformationKP_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SatelliteInformationKP_item },
};

static int
dissect_rnsap_GANSS_SatelliteInformationKP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_SatelliteInformationKP, GANSS_SatelliteInformationKP_sequence_of,
                                                  1, maxGANSSSatAlmanac, FALSE);

  return offset;
}


static const per_sequence_t T_gANSS_keplerianParameters_sequence[] = {
  { &hf_rnsap_t_oa          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_1023 },
  { &hf_rnsap_iod_a         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_15 },
  { &hf_rnsap_gANSS_SatelliteInformationKP, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SatelliteInformationKP },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_T_gANSS_keplerianParameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_T_gANSS_keplerianParameters, T_gANSS_keplerianParameters_sequence);

  return offset;
}



static int
dissect_rnsap_Extension_GANSS_AlmanacModel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_T_gANSS_AlmanacModel_vals[] = {
  {   0, "gANSS-keplerianParameters" },
  {   1, "extension-GANSS-AlmanacModel" },
  { 0, NULL }
};

static const per_choice_t T_gANSS_AlmanacModel_choice[] = {
  {   0, &hf_rnsap_gANSS_keplerianParameters, ASN1_EXTENSION_ROOT    , dissect_rnsap_T_gANSS_keplerianParameters },
  {   1, &hf_rnsap_extension_GANSS_AlmanacModel, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_GANSS_AlmanacModel },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_T_gANSS_AlmanacModel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_T_gANSS_AlmanacModel, T_gANSS_AlmanacModel_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t GANSS_Almanac_sequence[] = {
  { &hf_rnsap_ganss_wk_number, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_255 },
  { &hf_rnsap_gANSS_AlmanacModel, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_gANSS_AlmanacModel },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Almanac(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Almanac, GANSS_Almanac_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_NAVkpList_item_sequence[] = {
  { &hf_rnsap_svID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_navAlmE       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_navAlmDeltaI  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_navAlmOMEGADOT, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_navAlmSVHealth, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_navAlmSqrtA   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_navAlmOMEGAo  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_navAlmOmega   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_navAlmMo      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_navAlmaf0     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_navAlmaf1     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_SAT_Info_Almanac_NAVkpList_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_SAT_Info_Almanac_NAVkpList_item, GANSS_SAT_Info_Almanac_NAVkpList_item_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_NAVkpList_sequence_of[1] = {
  { &hf_rnsap_GANSS_SAT_Info_Almanac_NAVkpList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SAT_Info_Almanac_NAVkpList_item },
};

static int
dissect_rnsap_GANSS_SAT_Info_Almanac_NAVkpList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_SAT_Info_Almanac_NAVkpList, GANSS_SAT_Info_Almanac_NAVkpList_sequence_of,
                                                  1, maxGANSSSatAlmanac, FALSE);

  return offset;
}


static const per_sequence_t GANSS_alm_keplerianNAVAlmanac_sequence[] = {
  { &hf_rnsap_t_oa_01       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_255 },
  { &hf_rnsap_sat_info_NAVkpList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SAT_Info_Almanac_NAVkpList },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_alm_keplerianNAVAlmanac(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_alm_keplerianNAVAlmanac, GANSS_alm_keplerianNAVAlmanac_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_BDS_sequence[] = {
  { &hf_rnsap_satId         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_bdsAlmToa     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_bdsAlmSqrtA   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_bdsAlmE       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_17 },
  { &hf_rnsap_bdsAlmW       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_bdsAlmM0      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_bdsAlmOmega0  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_bdsAlmOmegaDot, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_17 },
  { &hf_rnsap_bdsAlmDeltaI  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_bdsAlmA0      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_bdsAlmA1      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_bdsSvHealth   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_9 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_SAT_Info_Almanac_BDS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_SAT_Info_Almanac_BDS, GANSS_SAT_Info_Almanac_BDS_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_BDSkpList_sequence_of[1] = {
  { &hf_rnsap_GANSS_SAT_Info_Almanac_BDSkpList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SAT_Info_Almanac_BDS },
};

static int
dissect_rnsap_GANSS_SAT_Info_Almanac_BDSkpList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_SAT_Info_Almanac_BDSkpList, GANSS_SAT_Info_Almanac_BDSkpList_sequence_of,
                                                  1, maxGANSSSatAlmanac, FALSE);

  return offset;
}


static const per_sequence_t GANSS_alm_KEPLERIANBDSALMANAC_sequence[] = {
  { &hf_rnsap_sat_info_BDSkpList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SAT_Info_Almanac_BDSkpList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_alm_KEPLERIANBDSALMANAC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_alm_KEPLERIANBDSALMANAC, GANSS_alm_KEPLERIANBDSALMANAC_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_REDkpList_item_sequence[] = {
  { &hf_rnsap_svID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_redAlmDeltaA  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_redAlmOmega0  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_7 },
  { &hf_rnsap_redAlmPhi0    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_7 },
  { &hf_rnsap_redAlmL1Health, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_1 },
  { &hf_rnsap_redAlmL2Health, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_1 },
  { &hf_rnsap_redAlmL5Health, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_1 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_SAT_Info_Almanac_REDkpList_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_SAT_Info_Almanac_REDkpList_item, GANSS_SAT_Info_Almanac_REDkpList_item_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_REDkpList_sequence_of[1] = {
  { &hf_rnsap_GANSS_SAT_Info_Almanac_REDkpList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SAT_Info_Almanac_REDkpList_item },
};

static int
dissect_rnsap_GANSS_SAT_Info_Almanac_REDkpList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_SAT_Info_Almanac_REDkpList, GANSS_SAT_Info_Almanac_REDkpList_sequence_of,
                                                  1, maxGANSSSatAlmanac, FALSE);

  return offset;
}


static const per_sequence_t GANSS_alm_keplerianReducedAlmanac_sequence[] = {
  { &hf_rnsap_t_oa_01       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_255 },
  { &hf_rnsap_sat_info_REDkpList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SAT_Info_Almanac_REDkpList },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_alm_keplerianReducedAlmanac(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_alm_keplerianReducedAlmanac, GANSS_alm_keplerianReducedAlmanac_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_MIDIkpList_item_sequence[] = {
  { &hf_rnsap_svID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_midiAlmE      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_midiAlmDeltaI , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_midiAlmOmegaDot, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_midiAlmSqrtA  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_17 },
  { &hf_rnsap_midiAlmOmega0 , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_midiAlmOmega  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_midiAlmMo     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_midiAlmaf0    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_midiAlmaf1    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_10 },
  { &hf_rnsap_midiAlmL1Health, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_1 },
  { &hf_rnsap_midiAlmL2Health, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_1 },
  { &hf_rnsap_midiAlmL5Health, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_1 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_SAT_Info_Almanac_MIDIkpList_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_SAT_Info_Almanac_MIDIkpList_item, GANSS_SAT_Info_Almanac_MIDIkpList_item_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_MIDIkpList_sequence_of[1] = {
  { &hf_rnsap_GANSS_SAT_Info_Almanac_MIDIkpList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SAT_Info_Almanac_MIDIkpList_item },
};

static int
dissect_rnsap_GANSS_SAT_Info_Almanac_MIDIkpList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_SAT_Info_Almanac_MIDIkpList, GANSS_SAT_Info_Almanac_MIDIkpList_sequence_of,
                                                  1, maxGANSSSatAlmanac, FALSE);

  return offset;
}


static const per_sequence_t GANSS_alm_keplerianMidiAlmanac_sequence[] = {
  { &hf_rnsap_t_oa_01       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_255 },
  { &hf_rnsap_sat_info_MIDIkpList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SAT_Info_Almanac_MIDIkpList },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_alm_keplerianMidiAlmanac(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_alm_keplerianMidiAlmanac, GANSS_alm_keplerianMidiAlmanac_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_GLOkpList_item_sequence[] = {
  { &hf_rnsap_gloAlmNA      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_gloAlmnA      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_5 },
  { &hf_rnsap_gloAlmHA      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_5 },
  { &hf_rnsap_gloAlmLambdaA , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_21 },
  { &hf_rnsap_gloAlmTlambdaA, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_21 },
  { &hf_rnsap_gloAlmDeltaIA , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_18 },
  { &hf_rnsap_gloAkmDeltaTA , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_22 },
  { &hf_rnsap_gloAlmDeltaTdotA, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_7 },
  { &hf_rnsap_gloAlmEpsilonA, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_15 },
  { &hf_rnsap_gloAlmOmegaA  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_gloAlmTauA    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_10 },
  { &hf_rnsap_gloAlmCA      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_1 },
  { &hf_rnsap_gloAlmMA      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_2 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_SAT_Info_Almanac_GLOkpList_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_SAT_Info_Almanac_GLOkpList_item, GANSS_SAT_Info_Almanac_GLOkpList_item_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_GLOkpList_sequence_of[1] = {
  { &hf_rnsap_GANSS_SAT_Info_Almanac_GLOkpList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SAT_Info_Almanac_GLOkpList_item },
};

static int
dissect_rnsap_GANSS_SAT_Info_Almanac_GLOkpList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_SAT_Info_Almanac_GLOkpList, GANSS_SAT_Info_Almanac_GLOkpList_sequence_of,
                                                  1, maxGANSSSatAlmanac, FALSE);

  return offset;
}


static const per_sequence_t GANSS_alm_keplerianGLONASS_sequence[] = {
  { &hf_rnsap_sat_info_GLOkpList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SAT_Info_Almanac_GLOkpList },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_alm_keplerianGLONASS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_alm_keplerianGLONASS, GANSS_alm_keplerianGLONASS_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_SBASecefList_item_sequence[] = {
  { &hf_rnsap_sbasAlmDataID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_2 },
  { &hf_rnsap_svID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_sbasAlmHealth , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_sbasAlmXg     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_15 },
  { &hf_rnsap_sbasAlmYg     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_15 },
  { &hf_rnsap_sbasAlmZg     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_9 },
  { &hf_rnsap_sbasAlmXgdot  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_3 },
  { &hf_rnsap_sbasAlmYgDot  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_3 },
  { &hf_rnsap_sbasAlmZgDot  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_4 },
  { &hf_rnsap_sbasAlmTo     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_SAT_Info_Almanac_SBASecefList_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_SAT_Info_Almanac_SBASecefList_item, GANSS_SAT_Info_Almanac_SBASecefList_item_sequence);

  return offset;
}


static const per_sequence_t GANSS_SAT_Info_Almanac_SBASecefList_sequence_of[1] = {
  { &hf_rnsap_GANSS_SAT_Info_Almanac_SBASecefList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SAT_Info_Almanac_SBASecefList_item },
};

static int
dissect_rnsap_GANSS_SAT_Info_Almanac_SBASecefList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_SAT_Info_Almanac_SBASecefList, GANSS_SAT_Info_Almanac_SBASecefList_sequence_of,
                                                  1, maxGANSSSatAlmanac, FALSE);

  return offset;
}


static const per_sequence_t GANSS_alm_ecefSBASAlmanac_sequence[] = {
  { &hf_rnsap_sat_info_SBASecefList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_SAT_Info_Almanac_SBASecefList },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_alm_ecefSBASAlmanac(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_alm_ecefSBASAlmanac, GANSS_alm_ecefSBASAlmanac_sequence);

  return offset;
}


static const per_sequence_t GANSS_AuxInfoGANSS_ID1_item_sequence[] = {
  { &hf_rnsap_svID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_signalsAvailable, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_AuxInfoGANSS_ID1_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_AuxInfoGANSS_ID1_item, GANSS_AuxInfoGANSS_ID1_item_sequence);

  return offset;
}


static const per_sequence_t GANSS_AuxInfoGANSS_ID1_sequence_of[1] = {
  { &hf_rnsap_GANSS_AuxInfoGANSS_ID1_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_AuxInfoGANSS_ID1_item },
};

static int
dissect_rnsap_GANSS_AuxInfoGANSS_ID1(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_AuxInfoGANSS_ID1, GANSS_AuxInfoGANSS_ID1_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_M7_13(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -7, 13U, NULL, FALSE);

  return offset;
}


static const per_sequence_t GANSS_AuxInfoGANSS_ID3_item_sequence[] = {
  { &hf_rnsap_svID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_signalsAvailable, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_channelNumber , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_M7_13 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_AuxInfoGANSS_ID3_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_AuxInfoGANSS_ID3_item, GANSS_AuxInfoGANSS_ID3_item_sequence);

  return offset;
}


static const per_sequence_t GANSS_AuxInfoGANSS_ID3_sequence_of[1] = {
  { &hf_rnsap_GANSS_AuxInfoGANSS_ID3_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_AuxInfoGANSS_ID3_item },
};

static int
dissect_rnsap_GANSS_AuxInfoGANSS_ID3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_AuxInfoGANSS_ID3, GANSS_AuxInfoGANSS_ID3_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const value_string rnsap_GANSS_Auxiliary_Information_vals[] = {
  {   0, "ganssID1" },
  {   1, "ganssID3" },
  { 0, NULL }
};

static const per_choice_t GANSS_Auxiliary_Information_choice[] = {
  {   0, &hf_rnsap_ganssID1      , ASN1_EXTENSION_ROOT    , dissect_rnsap_GANSS_AuxInfoGANSS_ID1 },
  {   1, &hf_rnsap_ganssID3      , ASN1_EXTENSION_ROOT    , dissect_rnsap_GANSS_AuxInfoGANSS_ID3 },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_GANSS_Auxiliary_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_GANSS_Auxiliary_Information, GANSS_Auxiliary_Information_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_GANSS_AuxInfoReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t GANSS_BDSclockModel_sequence[] = {
  { &hf_rnsap_bdsToc        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_17 },
  { &hf_rnsap_bdsa0         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_bdsa1         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_22 },
  { &hf_rnsap_bdsa2         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_bdsTgd1       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_10 },
  { &hf_rnsap_bdsAODC       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_5 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_BDSclockModel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_BDSclockModel, GANSS_BDSclockModel_sequence);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_6(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     6, 6, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_31(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     31, 31, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_1_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1U, NULL, TRUE);

  return offset;
}


static const per_sequence_t GANSS_Clock_Model_item_sequence[] = {
  { &hf_rnsap_t_oc          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_14 },
  { &hf_rnsap_a_i2          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_6 },
  { &hf_rnsap_a_i1          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_21 },
  { &hf_rnsap_a_i0          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_31 },
  { &hf_rnsap_t_gd          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_10 },
  { &hf_rnsap_sisa          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_model_id      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_0_1_ },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Clock_Model_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Clock_Model_item, GANSS_Clock_Model_item_sequence);

  return offset;
}


static const per_sequence_t GANSS_Clock_Model_sequence_of[1] = {
  { &hf_rnsap_GANSS_Clock_Model_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_Clock_Model_item },
};

static int
dissect_rnsap_GANSS_Clock_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_Clock_Model, GANSS_Clock_Model_sequence_of,
                                                  1, maxGANSSClockMod, FALSE);

  return offset;
}



static int
dissect_rnsap_BOOLEAN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t GANSS_IonosphereRegionalStormFlags_sequence[] = {
  { &hf_rnsap_storm_flag_one, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BOOLEAN },
  { &hf_rnsap_storm_flag_two, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BOOLEAN },
  { &hf_rnsap_storm_flag_three, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BOOLEAN },
  { &hf_rnsap_storm_flag_four, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BOOLEAN },
  { &hf_rnsap_storm_flag_five, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BOOLEAN },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_IonosphereRegionalStormFlags(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_IonosphereRegionalStormFlags, GANSS_IonosphereRegionalStormFlags_sequence);

  return offset;
}


static const per_sequence_t GANSS_Ionospheric_Model_sequence[] = {
  { &hf_rnsap_alpha_zero_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_alpha_one_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_alpha_two_ionos, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_14 },
  { &hf_rnsap_gANSS_IonosphereRegionalStormFlags, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_IonosphereRegionalStormFlags },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Ionospheric_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Ionospheric_Model, GANSS_Ionospheric_Model_sequence);

  return offset;
}


static const value_string rnsap_T_latitudeSign_vals[] = {
  {   0, "north" },
  {   1, "south" },
  { 0, NULL }
};


static int
dissect_rnsap_T_latitudeSign(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_2147483647(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 2147483647U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_T_directionOfAltitude_01_vals[] = {
  {   0, "height" },
  {   1, "depth" },
  { 0, NULL }
};


static int
dissect_rnsap_T_directionOfAltitude_01(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t GANSS_RX_Pos_sequence[] = {
  { &hf_rnsap_latitudeSign  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_latitudeSign },
  { &hf_rnsap_degreesOfLatitude, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_2147483647 },
  { &hf_rnsap_degreesOfLongitude, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_M2147483648_2147483647 },
  { &hf_rnsap_directionOfAltitude_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_directionOfAltitude_01 },
  { &hf_rnsap_altitude      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_32767 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_RX_Pos(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_RX_Pos, GANSS_RX_Pos_sequence);

  return offset;
}


static const per_sequence_t GANSS_Common_Data_sequence[] = {
  { &hf_rnsap_ganss_Ionospheric_Model, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_Ionospheric_Model },
  { &hf_rnsap_ganss_Rx_Pos  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_RX_Pos },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Common_Data(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Common_Data, GANSS_Common_Data_sequence);

  return offset;
}


static const per_sequence_t GANSS_CommonDataInfoReq_sequence[] = {
  { &hf_rnsap_ionospheric_Model, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BOOLEAN },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_CommonDataInfoReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_CommonDataInfoReq, GANSS_CommonDataInfoReq_sequence);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_59_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 59U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_1_1024(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     1, 1024, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_DataBitAssistanceSgnItem_sequence[] = {
  { &hf_rnsap_ganss_SignalId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_Signal_ID },
  { &hf_rnsap_ganssDataBits , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_1_1024 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_DataBitAssistanceSgnItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_DataBitAssistanceSgnItem, GANSS_DataBitAssistanceSgnItem_sequence);

  return offset;
}


static const per_sequence_t GANSS_DataBitAssistanceSgnList_sequence_of[1] = {
  { &hf_rnsap_GANSS_DataBitAssistanceSgnList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_DataBitAssistanceSgnItem },
};

static int
dissect_rnsap_GANSS_DataBitAssistanceSgnList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_DataBitAssistanceSgnList, GANSS_DataBitAssistanceSgnList_sequence_of,
                                                  1, maxSgnType, FALSE);

  return offset;
}


static const per_sequence_t GANSS_DataBitAssistanceItem_sequence[] = {
  { &hf_rnsap_satId         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_dataBitAssistanceSgnList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_DataBitAssistanceSgnList },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_DataBitAssistanceItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_DataBitAssistanceItem, GANSS_DataBitAssistanceItem_sequence);

  return offset;
}


static const per_sequence_t GANSS_DataBitAssistanceList_sequence_of[1] = {
  { &hf_rnsap_GANSS_DataBitAssistanceList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_DataBitAssistanceItem },
};

static int
dissect_rnsap_GANSS_DataBitAssistanceList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_DataBitAssistanceList, GANSS_DataBitAssistanceList_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Data_Bit_Assistance_sequence[] = {
  { &hf_rnsap_ganssTod      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_59_ },
  { &hf_rnsap_dataBitAssistancelist, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_DataBitAssistanceList },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Data_Bit_Assistance(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Data_Bit_Assistance, GANSS_Data_Bit_Assistance_sequence);

  return offset;
}


static const per_sequence_t T_ganss_SatelliteInfo_sequence_of[1] = {
  { &hf_rnsap_ganss_SatelliteInfo_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
};

static int
dissect_rnsap_T_ganss_SatelliteInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_T_ganss_SatelliteInfo, T_ganss_SatelliteInfo_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Data_Bit_Assistance_ReqList_sequence[] = {
  { &hf_rnsap_dGANSS_Signal_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_ganss_DataBitInterval, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_15 },
  { &hf_rnsap_ganss_SatelliteInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_T_ganss_SatelliteInfo },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Data_Bit_Assistance_ReqList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Data_Bit_Assistance_ReqList, GANSS_Data_Bit_Assistance_ReqList_sequence);

  return offset;
}


static const per_sequence_t GANSS_Data_Bit_Assistance_ReqItem_sequence[] = {
  { &hf_rnsap_ganssTod_01   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_86399 },
  { &hf_rnsap_ganss_Data_Bit_Assistance_ReqList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_Data_Bit_Assistance_ReqList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Data_Bit_Assistance_ReqItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Data_Bit_Assistance_ReqItem, GANSS_Data_Bit_Assistance_ReqItem_sequence);

  return offset;
}



static int
dissect_rnsap_GANSS_Delta_T(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -128, 127U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_19(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     19, 19, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GANSS_Earth_Orientation_Parameters_sequence[] = {
  { &hf_rnsap_teop          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_pmX           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_21 },
  { &hf_rnsap_pmXdot        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_15 },
  { &hf_rnsap_pmY           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_21 },
  { &hf_rnsap_pmYdot        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_15 },
  { &hf_rnsap_deltaUT1      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_31 },
  { &hf_rnsap_deltaUT1dot   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_19 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Earth_Orientation_Parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Earth_Orientation_Parameters, GANSS_Earth_Orientation_Parameters_sequence);

  return offset;
}



static int
dissect_rnsap_GANSS_EarthOrientParaReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}



static int
dissect_rnsap_GANSS_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, TRUE);

  return offset;
}


static const per_sequence_t GANSS_GenericDataInfoReqItem_sequence[] = {
  { &hf_rnsap_ganss_Id      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_ID },
  { &hf_rnsap_ganss_Navigation_Model_And_Time_Recovery, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BOOLEAN },
  { &hf_rnsap_ganss_Time_Model_GNSS_GNSS, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_9 },
  { &hf_rnsap_ganss_UTC_Model, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BOOLEAN },
  { &hf_rnsap_ganss_Almanac , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BOOLEAN },
  { &hf_rnsap_ganss_Real_Time_Integrity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BOOLEAN },
  { &hf_rnsap_ganss_Data_Bit_Assistance_Req, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_Data_Bit_Assistance_ReqItem },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_GenericDataInfoReqItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_GenericDataInfoReqItem, GANSS_GenericDataInfoReqItem_sequence);

  return offset;
}


static const per_sequence_t GANSS_GenericDataInfoReqList_sequence_of[1] = {
  { &hf_rnsap_GANSS_GenericDataInfoReqList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_GenericDataInfoReqItem },
};

static int
dissect_rnsap_GANSS_GenericDataInfoReqList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_GenericDataInfoReqList, GANSS_GenericDataInfoReqList_sequence_of,
                                                  1, maxNoGANSS, FALSE);

  return offset;
}


static const value_string rnsap_T_non_broadcastIndication_01_vals[] = {
  {   0, "true" },
  { 0, NULL }
};


static int
dissect_rnsap_T_non_broadcastIndication_01(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t T_gANSS_keplerianParameters_01_sequence[] = {
  { &hf_rnsap_toe_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_14 },
  { &hf_rnsap_ganss_omega_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_delta_n_nav   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_m_zero_nav    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_omegadot_nav  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_ganss_e_nav   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_idot_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_14 },
  { &hf_rnsap_a_sqrt_nav    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_i_zero_nav    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_omega_zero_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_c_rs_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_c_is_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_c_us_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_c_rc_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_c_ic_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_c_uc_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_T_gANSS_keplerianParameters_01(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_T_gANSS_keplerianParameters_01, T_gANSS_keplerianParameters_01_sequence);

  return offset;
}


static const value_string rnsap_GANSS_Orbit_Model_vals[] = {
  {   0, "gANSS-keplerianParameters" },
  { 0, NULL }
};

static const per_choice_t GANSS_Orbit_Model_choice[] = {
  {   0, &hf_rnsap_gANSS_keplerianParameters_01, ASN1_EXTENSION_ROOT    , dissect_rnsap_T_gANSS_keplerianParameters_01 },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_GANSS_Orbit_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_GANSS_Orbit_Model, GANSS_Orbit_Model_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t GANSS_Sat_Info_Nav_item_sequence[] = {
  { &hf_rnsap_satId         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_svHealth_01   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_5 },
  { &hf_rnsap_iod_01        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_10 },
  { &hf_rnsap_ganssClockModel, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_Clock_Model },
  { &hf_rnsap_ganssOrbitModel, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_Orbit_Model },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Sat_Info_Nav_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Sat_Info_Nav_item, GANSS_Sat_Info_Nav_item_sequence);

  return offset;
}


static const per_sequence_t GANSS_Sat_Info_Nav_sequence_of[1] = {
  { &hf_rnsap_GANSS_Sat_Info_Nav_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_Sat_Info_Nav_item },
};

static int
dissect_rnsap_GANSS_Sat_Info_Nav(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_Sat_Info_Nav, GANSS_Sat_Info_Nav_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Navigation_Model_And_Time_Recovery_sequence[] = {
  { &hf_rnsap_ganss_Transmission_Time, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_Transmission_Time },
  { &hf_rnsap_non_broadcastIndication_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_T_non_broadcastIndication_01 },
  { &hf_rnsap_ganssSatInfoNav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_Sat_Info_Nav },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Navigation_Model_And_Time_Recovery(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Navigation_Model_And_Time_Recovery, GANSS_Navigation_Model_And_Time_Recovery_sequence);

  return offset;
}


static const per_sequence_t GANSS_UTC_Model_sequence[] = {
  { &hf_rnsap_a_one_utc     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_a_zero_utc    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_t_ot_utc      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_w_n_t_utc     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_delta_t_ls_utc, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_w_n_lsf_utc   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_dn_utc        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_delta_t_lsf_utc, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_UTC_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_UTC_Model, GANSS_UTC_Model_sequence);

  return offset;
}


static const per_sequence_t GANSS_Real_Time_Integrity_item_sequence[] = {
  { &hf_rnsap_bad_ganss_satId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_63 },
  { &hf_rnsap_bad_ganss_signalId, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Real_Time_Integrity_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Real_Time_Integrity_item, GANSS_Real_Time_Integrity_item_sequence);

  return offset;
}


static const per_sequence_t GANSS_Real_Time_Integrity_sequence_of[1] = {
  { &hf_rnsap_GANSS_Real_Time_Integrity_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_Real_Time_Integrity_item },
};

static int
dissect_rnsap_GANSS_Real_Time_Integrity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_Real_Time_Integrity, GANSS_Real_Time_Integrity_sequence_of,
                                                  1, maxGANSSSat, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Generic_DataItem_sequence[] = {
  { &hf_rnsap_ganss_Id      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_ID },
  { &hf_rnsap_dganss_Correction, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DGANSSCorrections },
  { &hf_rnsap_ganss_Navigation_Model_And_Time_Recovery_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_Navigation_Model_And_Time_Recovery },
  { &hf_rnsap_ganss_Time_Model, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_Time_Model },
  { &hf_rnsap_ganss_UTC_TIME, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_UTC_Model },
  { &hf_rnsap_ganss_Almanac_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_Almanac },
  { &hf_rnsap_ganss_Real_Time_Integrity_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_Real_Time_Integrity },
  { &hf_rnsap_ganss_Data_Bit_Assistance, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_Data_Bit_Assistance },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Generic_DataItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Generic_DataItem, GANSS_Generic_DataItem_sequence);

  return offset;
}


static const per_sequence_t GANSS_Generic_Data_sequence_of[1] = {
  { &hf_rnsap_GANSS_Generic_Data_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GANSS_Generic_DataItem },
};

static int
dissect_rnsap_GANSS_Generic_Data(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GANSS_Generic_Data, GANSS_Generic_Data_sequence_of,
                                                  1, maxNoGANSS, FALSE);

  return offset;
}


static const per_sequence_t GANSS_Information_sequence[] = {
  { &hf_rnsap_gANSS_CommonDataInfoReq, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_CommonDataInfoReq },
  { &hf_rnsap_gANSS_GenericDataInfoReqList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GANSS_GenericDataInfoReqList },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_Information, GANSS_Information_sequence);

  return offset;
}


static const per_sequence_t GANSS_NavModel_BDSKeplerianSet_sequence[] = {
  { &hf_rnsap_bdsURAI       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_4 },
  { &hf_rnsap_bdsToe        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_17 },
  { &hf_rnsap_bdsAPowerHalf , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_bdsE          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_bdsW          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_bdsDeltaN     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_bdsM0         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_bdsOmega0     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_bdsOmegaDot   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_bdsI0         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_bdsIDot       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_14 },
  { &hf_rnsap_bdsCuc        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_18 },
  { &hf_rnsap_bdsCus        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_18 },
  { &hf_rnsap_bdsCrc        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_18 },
  { &hf_rnsap_bdsCrs        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_18 },
  { &hf_rnsap_bdsCic        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_18 },
  { &hf_rnsap_bdsCis        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_18 },
  { &hf_rnsap_bdsAODE       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_5 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_NavModel_BDSKeplerianSet(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_NavModel_BDSKeplerianSet, GANSS_NavModel_BDSKeplerianSet_sequence);

  return offset;
}


static const value_string rnsap_GANSS_SBAS_ID_vals[] = {
  {   0, "waas" },
  {   1, "egnos" },
  {   2, "msas" },
  {   3, "gagan" },
  { 0, NULL }
};


static int
dissect_rnsap_GANSS_SBAS_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_GANSS_Time_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, TRUE);

  return offset;
}


static const per_sequence_t GANSS_UTCmodelSet4_sequence[] = {
  { &hf_rnsap_utcA0_01      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_utcA1_01      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_utcDeltaTls   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_utcWNlsf      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_utcDN_01      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_utcDeltaTlsf  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GANSS_UTCmodelSet4(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GANSS_UTCmodelSet4, GANSS_UTCmodelSet4_sequence);

  return offset;
}


static const per_sequence_t GA_AccessPointPosition_sequence[] = {
  { &hf_rnsap_geographicalCoordinate, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GeographicalCoordinate },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GA_AccessPointPosition(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GA_AccessPointPosition, GA_AccessPointPosition_sequence);

  return offset;
}



static int
dissect_rnsap_GERAN_Cell_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     16, 16, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_GERAN_Classmark(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       NO_BOUND, NO_BOUND, FALSE, NULL);

  return offset;
}



static int
dissect_rnsap_OCTET_STRING_SIZE_1_23(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       1, 23, FALSE, NULL);

  return offset;
}


static const per_sequence_t GERAN_SystemInfo_item_sequence[] = {
  { &hf_rnsap_gERAN_SI_block, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_OCTET_STRING_SIZE_1_23 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GERAN_SystemInfo_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GERAN_SystemInfo_item, GERAN_SystemInfo_item_sequence);

  return offset;
}


static const per_sequence_t GERAN_SystemInfo_sequence_of[1] = {
  { &hf_rnsap_GERAN_SystemInfo_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GERAN_SystemInfo_item },
};

static int
dissect_rnsap_GERAN_SystemInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GERAN_SystemInfo, GERAN_SystemInfo_sequence_of,
                                                  1, maxNrOfGERANSI, FALSE);

  return offset;
}


static const value_string rnsap_GERAN_SI_Type_vals[] = {
  {   0, "sI" },
  {   1, "pSI" },
  { 0, NULL }
};

static const per_choice_t GERAN_SI_Type_choice[] = {
  {   0, &hf_rnsap_sI            , ASN1_EXTENSION_ROOT    , dissect_rnsap_GERAN_SystemInfo },
  {   1, &hf_rnsap_pSI           , ASN1_EXTENSION_ROOT    , dissect_rnsap_GERAN_SystemInfo },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_GERAN_SI_Type(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_GERAN_SI_Type, GERAN_SI_Type_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t T_satellite_Almanac_Information_item_sequence[] = {
  { &hf_rnsap_dATA_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DATA_ID },
  { &hf_rnsap_sAT_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAT_ID },
  { &hf_rnsap_gps_e_alm     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_gps_toa_alm   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_gps_delta_I_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_omegadot_alm  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_svhealth_alm  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_gps_a_sqrt_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_omegazero_alm , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_m_zero_alm    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_gps_omega_alm , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_gps_af_zero_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_gps_af_one_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_T_satellite_Almanac_Information_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_T_satellite_Almanac_Information_item, T_satellite_Almanac_Information_item_sequence);

  return offset;
}


static const per_sequence_t T_satellite_Almanac_Information_sequence_of[1] = {
  { &hf_rnsap_satellite_Almanac_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_T_satellite_Almanac_Information_item },
};

static int
dissect_rnsap_T_satellite_Almanac_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_T_satellite_Almanac_Information, T_satellite_Almanac_Information_sequence_of,
                                                  1, maxNoSat, FALSE);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_364(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     364, 364, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GPS_Almanac_sequence[] = {
  { &hf_rnsap_wna_alm       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_satellite_Almanac_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_satellite_Almanac_Information },
  { &hf_rnsap_sVGlobalHealth_alm, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BIT_STRING_SIZE_364 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GPS_Almanac(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GPS_Almanac, GPS_Almanac_sequence);

  return offset;
}


static const per_sequence_t Satellite_Almanac_Information_ExtItem_item_sequence[] = {
  { &hf_rnsap_dATA_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DATA_ID },
  { &hf_rnsap_sAT_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAT_ID },
  { &hf_rnsap_gps_e_alm     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_gps_toa_alm   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_gps_delta_I_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_omegadot_alm  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_svhealth_alm  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_gps_a_sqrt_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_omegazero_alm , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_m_zero_alm    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_gps_omega_alm , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_gps_af_zero_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_gps_af_one_alm, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_11 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Satellite_Almanac_Information_ExtItem_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Satellite_Almanac_Information_ExtItem_item, Satellite_Almanac_Information_ExtItem_item_sequence);

  return offset;
}


static const per_sequence_t Satellite_Almanac_Information_ExtItem_sequence_of[1] = {
  { &hf_rnsap_Satellite_Almanac_Information_ExtItem_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Satellite_Almanac_Information_ExtItem_item },
};

static int
dissect_rnsap_Satellite_Almanac_Information_ExtItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Satellite_Almanac_Information_ExtItem, Satellite_Almanac_Information_ExtItem_sequence_of,
                                                  1, maxNrOfSatAlmanac_maxNoSat, FALSE);

  return offset;
}


static const value_string rnsap_T_gPSInformationItem_vals[] = {
  {   0, "gPS-NavigationModel-and-TimeRecovery" },
  {   1, "gPS-Ionospheric-Model" },
  {   2, "gPS-UTC-Model" },
  {   3, "gPS-Almanac" },
  {   4, "gPS-RealTime-Integrity" },
  { 0, NULL }
};


static int
dissect_rnsap_T_gPSInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t GPSInformation_item_sequence[] = {
  { &hf_rnsap_gPSInformationItem, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_gPSInformationItem },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GPSInformation_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GPSInformation_item, GPSInformation_item_sequence);

  return offset;
}


static const per_sequence_t GPSInformation_sequence_of[1] = {
  { &hf_rnsap_GPSInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GPSInformation_item },
};

static int
dissect_rnsap_GPSInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GPSInformation, GPSInformation_sequence_of,
                                                  1, maxNoGPSTypes, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_1048575(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1048575U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_BIT_STRING_SIZE_87(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     87, 87, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t GPS_NavigationModel_and_TimeRecovery_item_sequence[] = {
  { &hf_rnsap_tx_tow_nav    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_1048575 },
  { &hf_rnsap_sAT_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAT_ID },
  { &hf_rnsap_tlm_message_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_14 },
  { &hf_rnsap_tlm_revd_c_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_2 },
  { &hf_rnsap_ho_word_nav   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_22 },
  { &hf_rnsap_w_n_nav       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_10 },
  { &hf_rnsap_ca_or_p_on_l2_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_2 },
  { &hf_rnsap_user_range_accuracy_index_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_4 },
  { &hf_rnsap_sv_health_nav , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_6 },
  { &hf_rnsap_iodc_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_10 },
  { &hf_rnsap_l2_p_dataflag_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_1 },
  { &hf_rnsap_sf1_reserved_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_87 },
  { &hf_rnsap_t_gd_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_t_oc_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_a_f_2_nav     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_a_f_1_nav     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_a_f_zero_nav  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_22 },
  { &hf_rnsap_c_rs_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_delta_n_nav   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_m_zero_nav    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_c_uc_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_gps_e_nav     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_c_us_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_a_sqrt_nav    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_t_oe_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_fit_interval_flag_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_1 },
  { &hf_rnsap_aodo_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_5 },
  { &hf_rnsap_c_ic_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_omega_zero_nav, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_c_is_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_i_zero_nav    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_c_rc_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_16 },
  { &hf_rnsap_gps_omega_nav , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_omegadot_nav  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_idot_nav      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_14 },
  { &hf_rnsap_spare_zero_fill, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_20 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GPS_NavigationModel_and_TimeRecovery_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GPS_NavigationModel_and_TimeRecovery_item, GPS_NavigationModel_and_TimeRecovery_item_sequence);

  return offset;
}


static const per_sequence_t GPS_NavigationModel_and_TimeRecovery_sequence_of[1] = {
  { &hf_rnsap_GPS_NavigationModel_and_TimeRecovery_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GPS_NavigationModel_and_TimeRecovery_item },
};

static int
dissect_rnsap_GPS_NavigationModel_and_TimeRecovery(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GPS_NavigationModel_and_TimeRecovery, GPS_NavigationModel_and_TimeRecovery_sequence_of,
                                                  1, maxNoSat, FALSE);

  return offset;
}


static const value_string rnsap_GPS_RealTime_Integrity_vals[] = {
  {   0, "badSatellites" },
  {   1, "noBadSatellite" },
  { 0, NULL }
};

static const per_choice_t GPS_RealTime_Integrity_choice[] = {
  {   0, &hf_rnsap_badSatellites , ASN1_NO_EXTENSIONS     , dissect_rnsap_BadSatellites },
  {   1, &hf_rnsap_noBadSatellite, ASN1_NO_EXTENSIONS     , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_GPS_RealTime_Integrity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_GPS_RealTime_Integrity, GPS_RealTime_Integrity_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t GPS_RX_POS_sequence[] = {
  { &hf_rnsap_geographicalCoordinate, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GeographicalCoordinate },
  { &hf_rnsap_altitudeAndDirection, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GA_AltitudeAndDirection },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GPS_RX_POS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GPS_RX_POS, GPS_RX_POS_sequence);

  return offset;
}


static const per_sequence_t GPS_UTC_Model_sequence[] = {
  { &hf_rnsap_a_one_utc     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_24 },
  { &hf_rnsap_a_zero_utc    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_32 },
  { &hf_rnsap_t_ot_utc      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_delta_t_ls_utc, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_w_n_t_utc     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_w_n_lsf_utc   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_dn_utc        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_delta_t_lsf_utc, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GPS_UTC_Model(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GPS_UTC_Model, GPS_UTC_Model_sequence);

  return offset;
}



static int
dissect_rnsap_Guaranteed_Rate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNrOfTFs, NULL, FALSE);

  return offset;
}


static const per_sequence_t Guaranteed_Rate_Information_sequence[] = {
  { &hf_rnsap_guaranteed_UL_Rate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Guaranteed_Rate },
  { &hf_rnsap_guaranteed_DL_Rate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Guaranteed_Rate },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Guaranteed_Rate_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Guaranteed_Rate_Information, Guaranteed_Rate_Information_sequence);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_8_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 8U, NULL, TRUE);

  return offset;
}


static const per_sequence_t HARQ_MemoryPartitioning_Implicit_sequence[] = {
  { &hf_rnsap_number_of_Processes, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_8_ },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HARQ_MemoryPartitioning_Implicit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HARQ_MemoryPartitioning_Implicit, HARQ_MemoryPartitioning_Implicit_sequence);

  return offset;
}


static const value_string rnsap_T_process_Memory_Size_vals[] = {
  {   0, "hms800" },
  {   1, "hms1600" },
  {   2, "hms2400" },
  {   3, "hms3200" },
  {   4, "hms4000" },
  {   5, "hms4800" },
  {   6, "hms5600" },
  {   7, "hms6400" },
  {   8, "hms7200" },
  {   9, "hms8000" },
  {  10, "hms8800" },
  {  11, "hms9600" },
  {  12, "hms10400" },
  {  13, "hms11200" },
  {  14, "hms12000" },
  {  15, "hms12800" },
  {  16, "hms13600" },
  {  17, "hms14400" },
  {  18, "hms15200" },
  {  19, "hms16000" },
  {  20, "hms17600" },
  {  21, "hms19200" },
  {  22, "hms20800" },
  {  23, "hms22400" },
  {  24, "hms24000" },
  {  25, "hms25600" },
  {  26, "hms27200" },
  {  27, "hms28800" },
  {  28, "hms30400" },
  {  29, "hms32000" },
  {  30, "hms36000" },
  {  31, "hms40000" },
  {  32, "hms44000" },
  {  33, "hms48000" },
  {  34, "hms52000" },
  {  35, "hms56000" },
  {  36, "hms60000" },
  {  37, "hms64000" },
  {  38, "hms68000" },
  {  39, "hms72000" },
  {  40, "hms76000" },
  {  41, "hms80000" },
  {  42, "hms88000" },
  {  43, "hms96000" },
  {  44, "hms104000" },
  {  45, "hms112000" },
  {  46, "hms120000" },
  {  47, "hms128000" },
  {  48, "hms136000" },
  {  49, "hms144000" },
  {  50, "hms152000" },
  {  51, "hms160000" },
  {  52, "hms176000" },
  {  53, "hms192000" },
  {  54, "hms208000" },
  {  55, "hms224000" },
  {  56, "hms240000" },
  {  57, "hms256000" },
  {  58, "hms272000" },
  {  59, "hms288000" },
  {  60, "hms304000" },
  { 0, NULL }
};


static int
dissect_rnsap_T_process_Memory_Size(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     61, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t HARQ_MemoryPartitioningItem_sequence[] = {
  { &hf_rnsap_process_Memory_Size, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_process_Memory_Size },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HARQ_MemoryPartitioningItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HARQ_MemoryPartitioningItem, HARQ_MemoryPartitioningItem_sequence);

  return offset;
}


static const per_sequence_t HARQ_MemoryPartitioningList_sequence_of[1] = {
  { &hf_rnsap_HARQ_MemoryPartitioningList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HARQ_MemoryPartitioningItem },
};

static int
dissect_rnsap_HARQ_MemoryPartitioningList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HARQ_MemoryPartitioningList, HARQ_MemoryPartitioningList_sequence_of,
                                                  1, maxNrOfHARQProc, FALSE);

  return offset;
}


static const per_sequence_t HARQ_MemoryPartitioning_Explicit_sequence[] = {
  { &hf_rnsap_hARQ_MemoryPartitioningList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HARQ_MemoryPartitioningList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HARQ_MemoryPartitioning_Explicit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HARQ_MemoryPartitioning_Explicit, HARQ_MemoryPartitioning_Explicit_sequence);

  return offset;
}


static const value_string rnsap_HARQ_MemoryPartitioning_vals[] = {
  {   0, "implicit" },
  {   1, "explicit" },
  { 0, NULL }
};

static const per_choice_t HARQ_MemoryPartitioning_choice[] = {
  {   0, &hf_rnsap_implicit      , ASN1_EXTENSION_ROOT    , dissect_rnsap_HARQ_MemoryPartitioning_Implicit },
  {   1, &hf_rnsap_explicit_01   , ASN1_EXTENSION_ROOT    , dissect_rnsap_HARQ_MemoryPartitioning_Explicit },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_HARQ_MemoryPartitioning(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_HARQ_MemoryPartitioning, HARQ_MemoryPartitioning_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t HARQ_MemoryPartitioningInfoExtForMIMO_sequence_of[1] = {
  { &hf_rnsap_HARQ_MemoryPartitioningInfoExtForMIMO_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HARQ_MemoryPartitioningItem },
};

static int
dissect_rnsap_HARQ_MemoryPartitioningInfoExtForMIMO(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HARQ_MemoryPartitioningInfoExtForMIMO, HARQ_MemoryPartitioningInfoExtForMIMO_sequence_of,
                                                  4, 8, FALSE);

  return offset;
}


static const value_string rnsap_HARQ_Preamble_Mode_vals[] = {
  {   0, "mode0" },
  {   1, "mode1" },
  { 0, NULL }
};


static int
dissect_rnsap_HARQ_Preamble_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_HARQ_Preamble_Mode_Activation_Indicator_vals[] = {
  {   0, "harqPreambleModeSupported" },
  { 0, NULL }
};


static int
dissect_rnsap_HARQ_Preamble_Mode_Activation_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_HCS_Prio(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_HSDSCH_MACdFlow_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfMACdFlows_1, NULL, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlow_Specific_InfoItem_sequence[] = {
  { &hf_rnsap_hSDSCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlow_ID },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_trafficClass  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TrafficClass },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_MACdFlow_Specific_InfoItem, HSDSCH_MACdFlow_Specific_InfoItem_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlow_Specific_InfoList_sequence_of[1] = {
  { &hf_rnsap_HSDSCH_MACdFlow_Specific_InfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoItem },
};

static int
dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSDSCH_MACdFlow_Specific_InfoList, HSDSCH_MACdFlow_Specific_InfoList_sequence_of,
                                                  1, maxNrOfMACdFlows, FALSE);

  return offset;
}



static int
dissect_rnsap_MAChsGuaranteedBitRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 16777215U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_SID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNrOfPDUIndexes_1, NULL, FALSE);

  return offset;
}


static const per_sequence_t MACdPDU_Size_IndexItem_sequence[] = {
  { &hf_rnsap_sID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SID },
  { &hf_rnsap_mACdPDU_Size  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MACdPDU_Size },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MACdPDU_Size_IndexItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MACdPDU_Size_IndexItem, MACdPDU_Size_IndexItem_sequence);

  return offset;
}


static const per_sequence_t MACdPDU_Size_IndexList_sequence_of[1] = {
  { &hf_rnsap_MACdPDU_Size_IndexList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MACdPDU_Size_IndexItem },
};

static int
dissect_rnsap_MACdPDU_Size_IndexList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MACdPDU_Size_IndexList, MACdPDU_Size_IndexList_sequence_of,
                                                  1, maxNrOfPDUIndexes, FALSE);

  return offset;
}


static const value_string rnsap_RLC_Mode_vals[] = {
  {   0, "rLC-AM" },
  {   1, "rLC-UM" },
  { 0, NULL }
};


static int
dissect_rnsap_RLC_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoItem_sequence[] = {
  { &hf_rnsap_priorityQueue_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityQueue_Id },
  { &hf_rnsap_associatedHSDSCH_MACdFlow, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlow_ID },
  { &hf_rnsap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_t1            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T1 },
  { &hf_rnsap_discardTimer  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DiscardTimer },
  { &hf_rnsap_mAC_hsWindowSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAC_hsWindowSize },
  { &hf_rnsap_mAChsGuaranteedBitRate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MAChsGuaranteedBitRate },
  { &hf_rnsap_mACdPDU_Size_Index, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MACdPDU_Size_IndexList },
  { &hf_rnsap_rLC_Mode      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RLC_Mode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PriorityQueue_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PriorityQueue_InfoItem, PriorityQueue_InfoItem_sequence);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoList_sequence_of[1] = {
  { &hf_rnsap_PriorityQueue_InfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityQueue_InfoItem },
};

static int
dissect_rnsap_PriorityQueue_InfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_PriorityQueue_InfoList, PriorityQueue_InfoList_sequence_of,
                                                  1, maxNrOfPrioQueues, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlows_Information_sequence[] = {
  { &hf_rnsap_hSDSCH_MACdFlow_Specific_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoList },
  { &hf_rnsap_priorityQueue_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityQueue_InfoList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_MACdFlows_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_MACdFlows_Information, HSDSCH_MACdFlows_Information_sequence);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_64_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 64U, NULL, TRUE);

  return offset;
}


static const per_sequence_t UE_Capabilities_Info_sequence[] = {
  { &hf_rnsap_hSDSCH_Physical_Layer_Category, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_64_ },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UE_Capabilities_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UE_Capabilities_Info, UE_Capabilities_Info_sequence);

  return offset;
}



static int
dissect_rnsap_MAChsReorderingBufferSize_for_RLC_UM(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 300U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_Nack_Power_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 8U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_HSSCCH_PowerOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_FDD_Information_sequence[] = {
  { &hf_rnsap_hSDSCH_MACdFlows_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlows_Information },
  { &hf_rnsap_uE_Capabilities_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_Capabilities_Info },
  { &hf_rnsap_mAChs_Reordering_Buffer_Size_for_RLC_UM, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAChsReorderingBufferSize_for_RLC_UM },
  { &hf_rnsap_cqiFeedback_CycleK, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CQI_Feedback_Cycle },
  { &hf_rnsap_cqiRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CQI_RepetitionFactor },
  { &hf_rnsap_ackNackRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AckNack_RepetitionFactor },
  { &hf_rnsap_cqiPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CQI_Power_Offset },
  { &hf_rnsap_ackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Ack_Power_Offset },
  { &hf_rnsap_nackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Nack_Power_Offset },
  { &hf_rnsap_hsscch_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_PowerOffset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_FDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_FDD_Information, HSDSCH_FDD_Information_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlow_Specific_InfoItem_Response_sequence[] = {
  { &hf_rnsap_hSDSCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlow_ID },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_hSDSCH_Initial_Capacity_Allocation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSDSCH_Initial_Capacity_Allocation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoItem_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_MACdFlow_Specific_InfoItem_Response, HSDSCH_MACdFlow_Specific_InfoItem_Response_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlow_Specific_InfoList_Response_sequence_of[1] = {
  { &hf_rnsap_HSDSCH_MACdFlow_Specific_InfoList_Response_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoItem_Response },
};

static int
dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoList_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSDSCH_MACdFlow_Specific_InfoList_Response, HSDSCH_MACdFlow_Specific_InfoList_Response_sequence_of,
                                                  0, maxNrOfMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t HSSCCH_FDD_Specific_InfoItem_Response_sequence[] = {
  { &hf_rnsap_code_Number   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_127 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSSCCH_FDD_Specific_InfoItem_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSSCCH_FDD_Specific_InfoItem_Response, HSSCCH_FDD_Specific_InfoItem_Response_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_FDD_Specific_InfoList_Response_sequence_of[1] = {
  { &hf_rnsap_HSSCCH_FDD_Specific_InfoList_Response_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSSCCH_FDD_Specific_InfoItem_Response },
};

static int
dissect_rnsap_HSSCCH_FDD_Specific_InfoList_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSSCCH_FDD_Specific_InfoList_Response, HSSCCH_FDD_Specific_InfoList_Response_sequence_of,
                                                  0, maxNrOfHSSCCHCodes, FALSE);

  return offset;
}



static int
dissect_rnsap_Measurement_Power_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -12, 26U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_FDD_Information_Response_sequence[] = {
  { &hf_rnsap_hSDSCH_MACdFlow_Specific_InfoList_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoList_Response },
  { &hf_rnsap_hSSCCH_Specific_InfoList_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_FDD_Specific_InfoList_Response },
  { &hf_rnsap_hSPDSCH_and_HSSCCH_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_ScramblingCode },
  { &hf_rnsap_measurement_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Measurement_Power_Offset },
  { &hf_rnsap_hARQ_MemoryPartitioning, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_MemoryPartitioning },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_FDD_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_FDD_Information_Response, HSDSCH_FDD_Information_Response_sequence);

  return offset;
}


static const value_string rnsap_SixtyfourQAM_UsageAllowedIndicator_vals[] = {
  {   0, "allowed" },
  {   1, "not-allowed" },
  { 0, NULL }
};


static int
dissect_rnsap_SixtyfourQAM_UsageAllowedIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_FDD_Secondary_Serving_Information_sequence[] = {
  { &hf_rnsap_hsscch_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_PowerOffset },
  { &hf_rnsap_sixtyfourQAM_UsageAllowedIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SixtyfourQAM_UsageAllowedIndicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_FDD_Secondary_Serving_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_DSCH_FDD_Secondary_Serving_Information, HS_DSCH_FDD_Secondary_Serving_Information_sequence);

  return offset;
}


static const value_string rnsap_SixtyfourQAM_DL_UsageIndicator_vals[] = {
  {   0, "sixtyfourQAM-DL-used" },
  {   1, "sixtyfourQAM-DL-not-used" },
  { 0, NULL }
};


static int
dissect_rnsap_SixtyfourQAM_DL_UsageIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_HSDSCH_TBSizeTableIndicator_vals[] = {
  {   0, "octet-aligned" },
  { 0, NULL }
};


static int
dissect_rnsap_HSDSCH_TBSizeTableIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_FDD_Secondary_Serving_Information_Response_sequence[] = {
  { &hf_rnsap_hSSCCH_Specific_InfoList_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_FDD_Specific_InfoList_Response },
  { &hf_rnsap_hSPDSCH_and_HSSCCH_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_ScramblingCode },
  { &hf_rnsap_measurement_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Measurement_Power_Offset },
  { &hf_rnsap_sixtyfourQAM_DL_UsageIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SixtyfourQAM_DL_UsageIndicator },
  { &hf_rnsap_hSDSCH_TBSizeTableIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSDSCH_TBSizeTableIndicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_FDD_Secondary_Serving_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_DSCH_FDD_Secondary_Serving_Information_Response, HS_DSCH_FDD_Secondary_Serving_Information_Response_sequence);

  return offset;
}


static const value_string rnsap_HSSCCH_Code_Change_Grant_vals[] = {
  {   0, "changeGranted" },
  { 0, NULL }
};


static int
dissect_rnsap_HSSCCH_Code_Change_Grant(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_Secondary_Serving_Information_To_Modify_sequence[] = {
  { &hf_rnsap_hsscch_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_PowerOffset },
  { &hf_rnsap_hSSCCH_CodeChangeGrant, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_Code_Change_Grant },
  { &hf_rnsap_sixtyfourQAM_UsageAllowedIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SixtyfourQAM_UsageAllowedIndicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_Secondary_Serving_Information_To_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_DSCH_Secondary_Serving_Information_To_Modify, HS_DSCH_Secondary_Serving_Information_To_Modify_sequence);

  return offset;
}


static const per_sequence_t HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised_sequence[] = {
  { &hf_rnsap_hsscch_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_PowerOffset },
  { &hf_rnsap_sixtyfourQAM_UsageAllowedIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SixtyfourQAM_UsageAllowedIndicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised, HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised_sequence);

  return offset;
}


static const value_string rnsap_HSSCCH_CodeChangeIndicator_vals[] = {
  {   0, "hsSCCHCodeChangeNeeded" },
  { 0, NULL }
};


static int
dissect_rnsap_HSSCCH_CodeChangeIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_HS_PDSCH_Code_Change_Indicator_vals[] = {
  {   0, "hsPDSCHCodeChangeNeeded" },
  { 0, NULL }
};


static int
dissect_rnsap_HS_PDSCH_Code_Change_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_FDD_Secondary_Serving_Update_Information_sequence[] = {
  { &hf_rnsap_hsSCCHCodeChangeIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_CodeChangeIndicator },
  { &hf_rnsap_hS_PDSCH_Code_Change_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_PDSCH_Code_Change_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_FDD_Secondary_Serving_Update_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_DSCH_FDD_Secondary_Serving_Update_Information, HS_DSCH_FDD_Secondary_Serving_Update_Information_sequence);

  return offset;
}


static const per_sequence_t HS_Secondary_Serving_cell_change_successful_sequence[] = {
  { &hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_DSCH_FDD_Secondary_Serving_Information_Response },
  { &hf_rnsap_hSDSCH_RNTI   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_RNTI },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_Secondary_Serving_cell_change_successful(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_Secondary_Serving_cell_change_successful, HS_Secondary_Serving_cell_change_successful_sequence);

  return offset;
}


static const per_sequence_t HS_Secondary_Serving_cell_change_unsuccessful_sequence[] = {
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_Secondary_Serving_cell_change_unsuccessful(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_Secondary_Serving_cell_change_unsuccessful, HS_Secondary_Serving_cell_change_unsuccessful_sequence);

  return offset;
}


static const value_string rnsap_HS_DSCH_Secondary_Serving_cell_change_choice_vals[] = {
  {   0, "hS-Secondary-Serving-cell-change-successful" },
  {   1, "hS-Secondary-Serving-cell-change-unsuccessful" },
  { 0, NULL }
};

static const per_choice_t HS_DSCH_Secondary_Serving_cell_change_choice_choice[] = {
  {   0, &hf_rnsap_hS_Secondary_Serving_cell_change_successful, ASN1_EXTENSION_ROOT    , dissect_rnsap_HS_Secondary_Serving_cell_change_successful },
  {   1, &hf_rnsap_hS_Secondary_Serving_cell_change_unsuccessful, ASN1_EXTENSION_ROOT    , dissect_rnsap_HS_Secondary_Serving_cell_change_unsuccessful },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_Secondary_Serving_cell_change_choice(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_HS_DSCH_Secondary_Serving_cell_change_choice, HS_DSCH_Secondary_Serving_cell_change_choice_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_Secondary_Serving_Cell_Change_Information_Response_sequence[] = {
  { &hf_rnsap_hS_DSCH_Secondary_Serving_cell_choice, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_DSCH_Secondary_Serving_cell_change_choice },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_Secondary_Serving_Cell_Change_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_DSCH_Secondary_Serving_Cell_Change_Information_Response, HS_DSCH_Secondary_Serving_Cell_Change_Information_Response_sequence);

  return offset;
}



static int
dissect_rnsap_HS_DSCH_Secondary_Serving_Remove(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlow_Specific_InfoItem_to_Modify_sequence[] = {
  { &hf_rnsap_hSDSCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlow_ID },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportBearerRequestIndicator },
  { &hf_rnsap_trafficClass  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TrafficClass },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoItem_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_MACdFlow_Specific_InfoItem_to_Modify, HSDSCH_MACdFlow_Specific_InfoItem_to_Modify_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlow_Specific_InfoList_to_Modify_sequence_of[1] = {
  { &hf_rnsap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoItem_to_Modify },
};

static int
dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify, HSDSCH_MACdFlow_Specific_InfoList_to_Modify_sequence_of,
                                                  1, maxNrOfMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoItem_to_Add_sequence[] = {
  { &hf_rnsap_priorityQueue_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityQueue_Id },
  { &hf_rnsap_associatedHSDSCH_MACdFlow, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlow_ID },
  { &hf_rnsap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_t1            , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T1 },
  { &hf_rnsap_discardTimer  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DiscardTimer },
  { &hf_rnsap_mAC_hsWindowSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAC_hsWindowSize },
  { &hf_rnsap_mAChsGuaranteedBitRate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MAChsGuaranteedBitRate },
  { &hf_rnsap_mACdPDU_Size_Index, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MACdPDU_Size_IndexList },
  { &hf_rnsap_rLC_Mode      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RLC_Mode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PriorityQueue_InfoItem_to_Add(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PriorityQueue_InfoItem_to_Add, PriorityQueue_InfoItem_to_Add_sequence);

  return offset;
}


static const per_sequence_t MACdPDU_Size_IndexItem_to_Modify_sequence[] = {
  { &hf_rnsap_sID           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SID },
  { &hf_rnsap_mACdPDU_Size  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MACdPDU_Size },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MACdPDU_Size_IndexItem_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MACdPDU_Size_IndexItem_to_Modify, MACdPDU_Size_IndexItem_to_Modify_sequence);

  return offset;
}


static const per_sequence_t MACdPDU_Size_IndexList_to_Modify_sequence_of[1] = {
  { &hf_rnsap_MACdPDU_Size_IndexList_to_Modify_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MACdPDU_Size_IndexItem_to_Modify },
};

static int
dissect_rnsap_MACdPDU_Size_IndexList_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MACdPDU_Size_IndexList_to_Modify, MACdPDU_Size_IndexList_to_Modify_sequence_of,
                                                  1, maxNrOfPDUIndexes, FALSE);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoItem_to_Modify_sequence[] = {
  { &hf_rnsap_priorityQueue_Id, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityQueue_Id },
  { &hf_rnsap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_t1            , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_T1 },
  { &hf_rnsap_discardTimer  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DiscardTimer },
  { &hf_rnsap_mAC_hsWindowSize, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MAC_hsWindowSize },
  { &hf_rnsap_mAChsGuaranteedBitRate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MAChsGuaranteedBitRate },
  { &hf_rnsap_mACdPDU_Size_Index_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MACdPDU_Size_IndexList_to_Modify },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PriorityQueue_InfoItem_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PriorityQueue_InfoItem_to_Modify, PriorityQueue_InfoItem_to_Modify_sequence);

  return offset;
}


static const value_string rnsap_ModifyPriorityQueue_vals[] = {
  {   0, "addPriorityQueue" },
  {   1, "modifyPriorityQueue" },
  {   2, "deletePriorityQueue" },
  { 0, NULL }
};

static const per_choice_t ModifyPriorityQueue_choice[] = {
  {   0, &hf_rnsap_addPriorityQueue, ASN1_EXTENSION_ROOT    , dissect_rnsap_PriorityQueue_InfoItem_to_Add },
  {   1, &hf_rnsap_modifyPriorityQueue, ASN1_EXTENSION_ROOT    , dissect_rnsap_PriorityQueue_InfoItem_to_Modify },
  {   2, &hf_rnsap_deletePriorityQueue, ASN1_EXTENSION_ROOT    , dissect_rnsap_PriorityQueue_Id },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_ModifyPriorityQueue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_ModifyPriorityQueue, ModifyPriorityQueue_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoList_to_Modify_sequence_of[1] = {
  { &hf_rnsap_PriorityQueue_InfoList_to_Modify_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ModifyPriorityQueue },
};

static int
dissect_rnsap_PriorityQueue_InfoList_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_PriorityQueue_InfoList_to_Modify, PriorityQueue_InfoList_to_Modify_sequence_of,
                                                  1, maxNrOfPrioQueues, FALSE);

  return offset;
}



static int
dissect_rnsap_TDD_AckNack_Power_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -7, 8U, NULL, TRUE);

  return offset;
}


static const per_sequence_t HSDSCH_Information_to_Modify_sequence[] = {
  { &hf_rnsap_hSDSCH_MACdFlow_Specific_InfoList_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify },
  { &hf_rnsap_priorityQueue_Info_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PriorityQueue_InfoList_to_Modify },
  { &hf_rnsap_mAChs_Reordering_Buffer_Size_for_RLC_UM, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MAChsReorderingBufferSize_for_RLC_UM },
  { &hf_rnsap_cqiFeedback_CycleK, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CQI_Feedback_Cycle },
  { &hf_rnsap_cqiRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CQI_RepetitionFactor },
  { &hf_rnsap_ackNackRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AckNack_RepetitionFactor },
  { &hf_rnsap_cqiPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CQI_Power_Offset },
  { &hf_rnsap_ackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Ack_Power_Offset },
  { &hf_rnsap_nackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Nack_Power_Offset },
  { &hf_rnsap_hsscch_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_PowerOffset },
  { &hf_rnsap_hSSCCH_CodeChangeGrant, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_Code_Change_Grant },
  { &hf_rnsap_tDDAckNackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_AckNack_Power_Offset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_Information_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_Information_to_Modify, HSDSCH_Information_to_Modify_sequence);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoItem_to_Modify_Unsynchronised_sequence[] = {
  { &hf_rnsap_priorityQueueId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityQueue_Id },
  { &hf_rnsap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_discardTimer  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DiscardTimer },
  { &hf_rnsap_mAChsGuaranteedBitRate, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MAChsGuaranteedBitRate },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PriorityQueue_InfoItem_to_Modify_Unsynchronised(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PriorityQueue_InfoItem_to_Modify_Unsynchronised, PriorityQueue_InfoItem_to_Modify_Unsynchronised_sequence);

  return offset;
}


static const per_sequence_t PriorityQueue_InfoList_to_Modify_Unsynchronised_sequence_of[1] = {
  { &hf_rnsap_PriorityQueue_InfoList_to_Modify_Unsynchronised_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_PriorityQueue_InfoItem_to_Modify_Unsynchronised },
};

static int
dissect_rnsap_PriorityQueue_InfoList_to_Modify_Unsynchronised(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_PriorityQueue_InfoList_to_Modify_Unsynchronised, PriorityQueue_InfoList_to_Modify_Unsynchronised_sequence_of,
                                                  0, maxNrOfPrioQueues, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_Information_to_Modify_Unsynchronised_sequence[] = {
  { &hf_rnsap_hSDSCH_MACdFlow_Specific_InfoList_to_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify },
  { &hf_rnsap_priorityQueueInfotoModifyUnsynchronised, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PriorityQueue_InfoList_to_Modify_Unsynchronised },
  { &hf_rnsap_cqiPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CQI_Power_Offset },
  { &hf_rnsap_ackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Ack_Power_Offset },
  { &hf_rnsap_nackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Nack_Power_Offset },
  { &hf_rnsap_hsscch_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_PowerOffset },
  { &hf_rnsap_tDDAckNackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_AckNack_Power_Offset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_Information_to_Modify_Unsynchronised(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_Information_to_Modify_Unsynchronised, HSDSCH_Information_to_Modify_Unsynchronised_sequence);

  return offset;
}


static const value_string rnsap_HSDSCH_MACdPDUSizeFormat_vals[] = {
  {   0, "indexedMACdPDU-Size" },
  {   1, "flexibleMACdPDU-Size" },
  { 0, NULL }
};


static int
dissect_rnsap_HSDSCH_MACdPDUSizeFormat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlows_to_Delete_Item_sequence[] = {
  { &hf_rnsap_hsDSCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlow_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_MACdFlows_to_Delete_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_MACdFlows_to_Delete_Item, HSDSCH_MACdFlows_to_Delete_Item_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_MACdFlows_to_Delete_sequence_of[1] = {
  { &hf_rnsap_HSDSCH_MACdFlows_to_Delete_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlows_to_Delete_Item },
};

static int
dissect_rnsap_HSDSCH_MACdFlows_to_Delete(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSDSCH_MACdFlows_to_Delete, HSDSCH_MACdFlows_to_Delete_sequence_of,
                                                  1, maxNrOfMACdFlows, FALSE);

  return offset;
}



static int
dissect_rnsap_HS_SCCH_CodeNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_SCCH_PreconfiguredCodesItem_sequence[] = {
  { &hf_rnsap_hS_SCCH_CodeNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SCCH_CodeNumber },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_SCCH_PreconfiguredCodesItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_SCCH_PreconfiguredCodesItem, HS_SCCH_PreconfiguredCodesItem_sequence);

  return offset;
}


static const per_sequence_t HS_SCCH_PreconfiguredCodes_sequence_of[1] = {
  { &hf_rnsap_HS_SCCH_PreconfiguredCodes_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SCCH_PreconfiguredCodesItem },
};

static int
dissect_rnsap_HS_SCCH_PreconfiguredCodes(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HS_SCCH_PreconfiguredCodes, HS_SCCH_PreconfiguredCodes_sequence_of,
                                                  1, maxNrOfHSSCCHCodes, FALSE);

  return offset;
}


static const value_string rnsap_SixtyfourQAM_DL_SupportIndicator_vals[] = {
  {   0, "sixtyfourQAM-DL-supported" },
  {   1, "sixtyfourQAM-DL-not-supported" },
  { 0, NULL }
};


static int
dissect_rnsap_SixtyfourQAM_DL_SupportIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t SetsOfHS_SCCH_CodesItem_sequence[] = {
  { &hf_rnsap_hS_SCCH_PreconfiguredCodes, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SCCH_PreconfiguredCodes },
  { &hf_rnsap_hSDSCH_RNTI   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_RNTI },
  { &hf_rnsap_hSPDSCH_and_HSSCCH_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_ScramblingCode },
  { &hf_rnsap_sixtyfourQAM_DL_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SixtyfourQAM_DL_SupportIndicator },
  { &hf_rnsap_sixtyfourQAM_DL_UsageIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SixtyfourQAM_DL_UsageIndicator },
  { &hf_rnsap_hSDSCH_TBSizeTableIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSDSCH_TBSizeTableIndicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SetsOfHS_SCCH_CodesItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SetsOfHS_SCCH_CodesItem, SetsOfHS_SCCH_CodesItem_sequence);

  return offset;
}


static const per_sequence_t SetsOfHS_SCCH_Codes_sequence_of[1] = {
  { &hf_rnsap_SetsOfHS_SCCH_Codes_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_SetsOfHS_SCCH_CodesItem },
};

static int
dissect_rnsap_SetsOfHS_SCCH_Codes(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_SetsOfHS_SCCH_Codes, SetsOfHS_SCCH_Codes_sequence_of,
                                                  1, maxNrOfHSDSCH, FALSE);

  return offset;
}



static int
dissect_rnsap_MIMO_S_CPICH_Channelisation_Code(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_MIMO_PilotConfiguration_vals[] = {
  {   0, "primary-and-secondary-CPICH" },
  {   1, "normal-and-diversity-primary-CPICH" },
  { 0, NULL }
};

static const per_choice_t MIMO_PilotConfiguration_choice[] = {
  {   0, &hf_rnsap_primary_and_secondary_CPICH, ASN1_EXTENSION_ROOT    , dissect_rnsap_MIMO_S_CPICH_Channelisation_Code },
  {   1, &hf_rnsap_normal_and_diversity_primary_CPICH, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_MIMO_PilotConfiguration(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_MIMO_PilotConfiguration, MIMO_PilotConfiguration_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_MIMO_N_M_Ratio_vals[] = {
  {   0, "v1-2" },
  {   1, "v2-3" },
  {   2, "v3-4" },
  {   3, "v4-5" },
  {   4, "v5-6" },
  {   5, "v6-7" },
  {   6, "v7-8" },
  {   7, "v8-9" },
  {   8, "v9-10" },
  {   9, "v1-1" },
  { 0, NULL }
};


static int
dissect_rnsap_MIMO_N_M_Ratio(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     10, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t MIMO_InformationResponse_sequence[] = {
  { &hf_rnsap_mIMO_PilotConfiguration, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MIMO_PilotConfiguration },
  { &hf_rnsap_mIMO_N_M_Ratio_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MIMO_N_M_Ratio },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MIMO_InformationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MIMO_InformationResponse, MIMO_InformationResponse_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_PreconfigurationInfo_sequence[] = {
  { &hf_rnsap_setsOfHS_SCCH_Codes, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SetsOfHS_SCCH_Codes },
  { &hf_rnsap_hARQ_MemoryPartitioning, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HARQ_MemoryPartitioning },
  { &hf_rnsap_eDCH_FDD_DL_ControlChannelInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_EDCH_FDD_DL_ControlChannelInformation },
  { &hf_rnsap_hARQ_Preamble_Mode_Activation_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Preamble_Mode_Activation_Indicator },
  { &hf_rnsap_mIMO_N_M_Ratio, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MIMO_InformationResponse },
  { &hf_rnsap_continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_PreconfigurationInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_PreconfigurationInfo, HSDSCH_PreconfigurationInfo_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Preconfiguration_Information_ItemIEs_sequence[] = {
  { &hf_rnsap_eDCH_FDD_DL_ControlChannelInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_FDD_DL_ControlChannelInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_Preconfiguration_Information_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_Preconfiguration_Information_ItemIEs, Additional_EDCH_Preconfiguration_Information_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Preconfiguration_Information_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_Preconfiguration_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_Preconfiguration_Information_ItemIEs },
};

static int
dissect_rnsap_Additional_EDCH_Preconfiguration_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_Preconfiguration_Information, Additional_EDCH_Preconfiguration_Information_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const value_string rnsap_MAChsResetScheme_vals[] = {
  {   0, "always" },
  {   1, "interNodeB-change" },
  { 0, NULL }
};


static int
dissect_rnsap_MAChsResetScheme(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_NumHS_SCCH_Codes(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxNrOfHSSCCHCodes, NULL, FALSE);

  return offset;
}


static const per_sequence_t SecondaryServingCellsItem_sequence[] = {
  { &hf_rnsap_secondaryC_ID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_numSecondaryHS_SCCH_Codes, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NumHS_SCCH_Codes },
  { &hf_rnsap_sixtyfourQAM_UsageAllowedIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SixtyfourQAM_UsageAllowedIndicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SecondaryServingCellsItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SecondaryServingCellsItem, SecondaryServingCellsItem_sequence);

  return offset;
}


static const per_sequence_t SecondaryServingCells_sequence_of[1] = {
  { &hf_rnsap_SecondaryServingCells_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_SecondaryServingCellsItem },
};

static int
dissect_rnsap_SecondaryServingCells(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_SecondaryServingCells, SecondaryServingCells_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}



static int
dissect_rnsap_MIMO_ActivationIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t HSDSCH_PreconfigurationSetup_sequence[] = {
  { &hf_rnsap_mAChsResetScheme, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAChsResetScheme },
  { &hf_rnsap_hSDSCH_Physical_Layer_Category, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_64_ },
  { &hf_rnsap_mAChs_Reordering_Buffer_Size_for_RLC_UM, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAChsReorderingBufferSize_for_RLC_UM },
  { &hf_rnsap_secondaryServingCells, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SecondaryServingCells },
  { &hf_rnsap_numPrimaryHS_SCCH_Codes, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NumHS_SCCH_Codes },
  { &hf_rnsap_hARQ_Preamble_Mode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_Preamble_Mode },
  { &hf_rnsap_mIMO_ActivationIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MIMO_ActivationIndicator },
  { &hf_rnsap_hSDSCH_MACdPDUSizeFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSDSCH_MACdPDUSizeFormat },
  { &hf_rnsap_sixtyfourQAM_UsageAllowedIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SixtyfourQAM_UsageAllowedIndicator },
  { &hf_rnsap_uE_with_enhanced_HS_SCCH_support_indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_continuous_Packet_Connectivity_HS_SCCH_Less_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_PreconfigurationSetup(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_PreconfigurationSetup, HSDSCH_PreconfigurationSetup_sequence);

  return offset;
}


static const per_sequence_t HS_DSCH_serving_cell_change_information_sequence[] = {
  { &hf_rnsap_hS_PDSCH_RLID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_hSDSCH_FDD_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSDSCH_FDD_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_serving_cell_change_information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_DSCH_serving_cell_change_information, HS_DSCH_serving_cell_change_information_sequence);

  return offset;
}


static const per_sequence_t HS_serving_cell_change_successful_sequence[] = {
  { &hf_rnsap_hSDSCH_FDD_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_FDD_Information_Response },
  { &hf_rnsap_hSDSCH_RNTI   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_RNTI },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_serving_cell_change_successful(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_serving_cell_change_successful, HS_serving_cell_change_successful_sequence);

  return offset;
}


static const per_sequence_t HS_serving_cell_change_unsuccessful_sequence[] = {
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_serving_cell_change_unsuccessful(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_serving_cell_change_unsuccessful, HS_serving_cell_change_unsuccessful_sequence);

  return offset;
}


static const value_string rnsap_HS_DSCH_serving_cell_change_choice_vals[] = {
  {   0, "hS-serving-cell-change-successful" },
  {   1, "hS-serving-cell-change-unsuccessful" },
  { 0, NULL }
};

static const per_choice_t HS_DSCH_serving_cell_change_choice_choice[] = {
  {   0, &hf_rnsap_hS_serving_cell_change_successful, ASN1_EXTENSION_ROOT    , dissect_rnsap_HS_serving_cell_change_successful },
  {   1, &hf_rnsap_hS_serving_cell_change_unsuccessful, ASN1_EXTENSION_ROOT    , dissect_rnsap_HS_serving_cell_change_unsuccessful },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_serving_cell_change_choice(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_HS_DSCH_serving_cell_change_choice, HS_DSCH_serving_cell_change_choice_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t HS_DSCH_serving_cell_change_informationResponse_sequence[] = {
  { &hf_rnsap_hS_DSCH_serving_cell_choice, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_DSCH_serving_cell_change_choice },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_serving_cell_change_informationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_DSCH_serving_cell_change_informationResponse, HS_DSCH_serving_cell_change_informationResponse_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_TDD_Information_sequence[] = {
  { &hf_rnsap_hSDSCH_MACdFlows_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlows_Information },
  { &hf_rnsap_uE_Capabilities_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_Capabilities_Info },
  { &hf_rnsap_mAChs_Reordering_Buffer_Size_for_RLC_UM, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MAChsReorderingBufferSize_for_RLC_UM },
  { &hf_rnsap_tDD_AckNack_Power_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_AckNack_Power_Offset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_TDD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_TDD_Information, HSDSCH_TDD_Information_sequence);

  return offset;
}



static int
dissect_rnsap_HS_SICH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 31U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HSSICH_Info_sequence[] = {
  { &hf_rnsap_hsSICH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_ID },
  { &hf_rnsap_timeslot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSSICH_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSSICH_Info, HSSICH_Info_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_TDD_Specific_InfoItem_Response_sequence[] = {
  { &hf_rnsap_timeslot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_hSSICH_Info   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSSICH_Info },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSSCCH_TDD_Specific_InfoItem_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSSCCH_TDD_Specific_InfoItem_Response, HSSCCH_TDD_Specific_InfoItem_Response_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_TDD_Specific_InfoList_Response_sequence_of[1] = {
  { &hf_rnsap_HSSCCH_TDD_Specific_InfoList_Response_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSSCCH_TDD_Specific_InfoItem_Response },
};

static int
dissect_rnsap_HSSCCH_TDD_Specific_InfoList_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSSCCH_TDD_Specific_InfoList_Response, HSSCCH_TDD_Specific_InfoList_Response_sequence_of,
                                                  0, maxNrOfHSSCCHCodes, FALSE);

  return offset;
}


static const per_sequence_t HSSICH_InfoLCR_sequence[] = {
  { &hf_rnsap_hsSICH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_ID },
  { &hf_rnsap_timeslotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSSICH_InfoLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSSICH_InfoLCR, HSSICH_InfoLCR_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_TDD_Specific_InfoItem_Response_LCR_sequence[] = {
  { &hf_rnsap_timeslotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_first_TDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_second_TDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_hSSICH_InfoLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSSICH_InfoLCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSSCCH_TDD_Specific_InfoItem_Response_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSSCCH_TDD_Specific_InfoItem_Response_LCR, HSSCCH_TDD_Specific_InfoItem_Response_LCR_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_TDD_Specific_InfoList_Response_LCR_sequence_of[1] = {
  { &hf_rnsap_HSSCCH_TDD_Specific_InfoList_Response_LCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSSCCH_TDD_Specific_InfoItem_Response_LCR },
};

static int
dissect_rnsap_HSSCCH_TDD_Specific_InfoList_Response_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSSCCH_TDD_Specific_InfoList_Response_LCR, HSSCCH_TDD_Specific_InfoList_Response_LCR_sequence_of,
                                                  0, maxNrOfHSSCCHCodes, FALSE);

  return offset;
}


static const per_sequence_t HSPDSCH_TDD_Specific_InfoItem_Response_sequence[] = {
  { &hf_rnsap_timeslot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response, HSPDSCH_TDD_Specific_InfoItem_Response_sequence);

  return offset;
}


static const per_sequence_t HSPDSCH_TDD_Specific_InfoList_Response_sequence_of[1] = {
  { &hf_rnsap_HSPDSCH_TDD_Specific_InfoList_Response_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response },
};

static int
dissect_rnsap_HSPDSCH_TDD_Specific_InfoList_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSPDSCH_TDD_Specific_InfoList_Response, HSPDSCH_TDD_Specific_InfoList_Response_sequence_of,
                                                  0, maxNrOfDLTs, FALSE);

  return offset;
}


static const per_sequence_t HSPDSCH_TDD_Specific_InfoItem_Response_LCR_sequence[] = {
  { &hf_rnsap_timeslotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response_LCR, HSPDSCH_TDD_Specific_InfoItem_Response_LCR_sequence);

  return offset;
}


static const per_sequence_t HSPDSCH_TDD_Specific_InfoList_Response_LCR_sequence_of[1] = {
  { &hf_rnsap_HSPDSCH_TDD_Specific_InfoList_Response_LCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response_LCR },
};

static int
dissect_rnsap_HSPDSCH_TDD_Specific_InfoList_Response_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSPDSCH_TDD_Specific_InfoList_Response_LCR, HSPDSCH_TDD_Specific_InfoList_Response_LCR_sequence_of,
                                                  1, maxNrOfDLTsLCR, FALSE);

  return offset;
}


static const per_sequence_t HSDSCH_TDD_Information_Response_sequence[] = {
  { &hf_rnsap_hSDSCH_MACdFlow_Specific_InfoList_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSDSCH_MACdFlow_Specific_InfoList_Response },
  { &hf_rnsap_hSSCCH_TDD_Specific_InfoList_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_TDD_Specific_InfoList_Response },
  { &hf_rnsap_hSSCCH_TDD_Specific_InfoList_Response_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_TDD_Specific_InfoList_Response_LCR },
  { &hf_rnsap_hSPDSCH_TDD_Specific_InfoList_Response, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSPDSCH_TDD_Specific_InfoList_Response },
  { &hf_rnsap_hSPDSCH_TDD_Specific_InfoList_Response_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSPDSCH_TDD_Specific_InfoList_Response_LCR },
  { &hf_rnsap_hARQ_MemoryPartitioning, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_MemoryPartitioning },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_TDD_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_TDD_Information_Response, HSDSCH_TDD_Information_Response_sequence);

  return offset;
}



static int
dissect_rnsap_Multicarrier_Number(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, maxHSDPAFrequency, NULL, FALSE);

  return offset;
}


static const per_sequence_t HSPDSCH_TDD_Specific_InfoItem_Response768_sequence[] = {
  { &hf_rnsap_timeslot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response768, HSPDSCH_TDD_Specific_InfoItem_Response768_sequence);

  return offset;
}


static const per_sequence_t HSPDSCH_TDD_Specific_InfoList_Response768_sequence_of[1] = {
  { &hf_rnsap_HSPDSCH_TDD_Specific_InfoList_Response768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response768 },
};

static int
dissect_rnsap_HSPDSCH_TDD_Specific_InfoList_Response768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSPDSCH_TDD_Specific_InfoList_Response768, HSPDSCH_TDD_Specific_InfoList_Response768_sequence_of,
                                                  0, maxNrOfDLTs, FALSE);

  return offset;
}


static const per_sequence_t HSSICH_Info768_sequence[] = {
  { &hf_rnsap_hsSICH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_ID },
  { &hf_rnsap_timeslot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_tDD_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSSICH_Info768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSSICH_Info768, HSSICH_Info768_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_TDD_Specific_InfoItem_Response768_sequence[] = {
  { &hf_rnsap_timeslot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_tDD_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode768 },
  { &hf_rnsap_hSSICH_Info768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSSICH_Info768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSSCCH_TDD_Specific_InfoItem_Response768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSSCCH_TDD_Specific_InfoItem_Response768, HSSCCH_TDD_Specific_InfoItem_Response768_sequence);

  return offset;
}


static const per_sequence_t HSSCCH_TDD_Specific_InfoList_Response768_sequence_of[1] = {
  { &hf_rnsap_HSSCCH_TDD_Specific_InfoList_Response768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSSCCH_TDD_Specific_InfoItem_Response768 },
};

static int
dissect_rnsap_HSSCCH_TDD_Specific_InfoList_Response768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSSCCH_TDD_Specific_InfoList_Response768, HSSCCH_TDD_Specific_InfoList_Response768_sequence_of,
                                                  0, maxNrOfHSSCCHCodes, FALSE);

  return offset;
}



static int
dissect_rnsap_HS_SICH_failed(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 20U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_HS_SICH_missed(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 20U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_HS_SICH_total(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 20U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_SICH_Reception_Quality_Value_sequence[] = {
  { &hf_rnsap_failed_HS_SICH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_failed },
  { &hf_rnsap_missed_HS_SICH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_missed },
  { &hf_rnsap_total_HS_SICH , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_total },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_SICH_Reception_Quality_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_SICH_Reception_Quality_Value, HS_SICH_Reception_Quality_Value_sequence);

  return offset;
}



static int
dissect_rnsap_HS_SICH_Reception_Quality_Measurement_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 20U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_HS_SICH_ID_Extension(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            32U, 255U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_HS_PDSCH_Code_Change_Grant_vals[] = {
  {   0, "changeGranted" },
  { 0, NULL }
};


static int
dissect_rnsap_HS_PDSCH_Code_Change_Grant(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t HSDSCH_FDD_Update_Information_sequence[] = {
  { &hf_rnsap_hsSCCHCodeChangeIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_CodeChangeIndicator },
  { &hf_rnsap_cqiFeedback_CycleK, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CQI_Feedback_Cycle },
  { &hf_rnsap_cqiRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CQI_RepetitionFactor },
  { &hf_rnsap_ackNackRepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AckNack_RepetitionFactor },
  { &hf_rnsap_cqiPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CQI_Power_Offset },
  { &hf_rnsap_ackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Ack_Power_Offset },
  { &hf_rnsap_nackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Nack_Power_Offset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_FDD_Update_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_FDD_Update_Information, HSDSCH_FDD_Update_Information_sequence);

  return offset;
}


static const per_sequence_t HSDSCH_TDD_Update_Information_sequence[] = {
  { &hf_rnsap_hsSCCHCodeChangeIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_CodeChangeIndicator },
  { &hf_rnsap_tDDAckNackPowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_AckNack_Power_Offset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCH_TDD_Update_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCH_TDD_Update_Information, HSDSCH_TDD_Update_Information_sequence);

  return offset;
}


static const per_sequence_t HSSICH_ReferenceSignal_InformationLCR_sequence[] = {
  { &hf_rnsap_midambleConfigurationLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleConfigurationLCR },
  { &hf_rnsap_midambleShift , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_15 },
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSSICH_ReferenceSignal_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSSICH_ReferenceSignal_InformationLCR, HSSICH_ReferenceSignal_InformationLCR_sequence);

  return offset;
}


static const per_sequence_t MIMO_ReferenceSignal_InformationListLCR_sequence_of[1] = {
  { &hf_rnsap_MIMO_ReferenceSignal_InformationListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSSICH_ReferenceSignal_InformationLCR },
};

static int
dissect_rnsap_MIMO_ReferenceSignal_InformationListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MIMO_ReferenceSignal_InformationListLCR, MIMO_ReferenceSignal_InformationListLCR_sequence_of,
                                                  1, maxNrOfHSSCCHCodes, FALSE);

  return offset;
}



static int
dissect_rnsap_HSDSCH_Physical_Layer_Category(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 64U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_TS0_HS_PDSCH_Indication_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_HS_SICH_Mapping_Index(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxNoOf_HS_SICH_SPS_1, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_SCCH_Associated_HS_SICH_sequence[] = {
  { &hf_rnsap_hsSICH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_ID },
  { &hf_rnsap_extended_HS_SICH_ID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_SICH_ID_Extension },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_SCCH_Associated_HS_SICH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_SCCH_Associated_HS_SICH, HS_SCCH_Associated_HS_SICH_sequence);

  return offset;
}



static int
dissect_rnsap_Non_HS_SCCH_Aassociated_HS_SICH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Non_HS_SCCH_Associated_HS_SICH_sequence[] = {
  { &hf_rnsap_non_HS_SCCH_Aassociated_HS_SICH_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Non_HS_SCCH_Aassociated_HS_SICH_ID },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Non_HS_SCCH_Associated_HS_SICH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Non_HS_SCCH_Associated_HS_SICH, Non_HS_SCCH_Associated_HS_SICH_sequence);

  return offset;
}


static const value_string rnsap_HS_SICH_Type_vals[] = {
  {   0, "hS-SCCH-Associated-HS-SICH" },
  {   1, "non-HS-SCCH-Associated-HS-SICH" },
  { 0, NULL }
};

static const per_choice_t HS_SICH_Type_choice[] = {
  {   0, &hf_rnsap_hS_SCCH_Associated_HS_SICH, ASN1_EXTENSION_ROOT    , dissect_rnsap_HS_SCCH_Associated_HS_SICH },
  {   1, &hf_rnsap_non_HS_SCCH_Associated_HS_SICH, ASN1_EXTENSION_ROOT    , dissect_rnsap_Non_HS_SCCH_Associated_HS_SICH },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_HS_SICH_Type(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_HS_SICH_Type, HS_SICH_Type_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t HS_SICH_InformationItem_for_HS_DSCH_SPS_sequence[] = {
  { &hf_rnsap_hS_SICH_Mapping_Index, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_SICH_Mapping_Index },
  { &hf_rnsap_hS_SICH_Type  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_Type },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_SICH_InformationItem_for_HS_DSCH_SPS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_SICH_InformationItem_for_HS_DSCH_SPS, HS_SICH_InformationItem_for_HS_DSCH_SPS_sequence);

  return offset;
}


static const per_sequence_t HS_SICH_InformationList_for_HS_DSCH_SPS_sequence_of[1] = {
  { &hf_rnsap_HS_SICH_InformationList_for_HS_DSCH_SPS_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_InformationItem_for_HS_DSCH_SPS },
};

static int
dissect_rnsap_HS_SICH_InformationList_for_HS_DSCH_SPS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HS_SICH_InformationList_for_HS_DSCH_SPS, HS_SICH_InformationList_for_HS_DSCH_SPS_sequence_of,
                                                  1, maxNoOf_HS_SICH_SPS, FALSE);

  return offset;
}



static int
dissect_rnsap_TDD_PhysicalChannelOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_HS_DSCH_TimeslotResourceLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     5, 5, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const value_string rnsap_ModulationSPS_LCR_vals[] = {
  {   0, "qPSK" },
  {   1, "sixteenQAM" },
  { 0, NULL }
};


static int
dissect_rnsap_ModulationSPS_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Initial_HS_PDSCH_SPS_Resource_sequence[] = {
  { &hf_rnsap_repetitionPeriodIndex, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriodIndex },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RepetitionLength },
  { &hf_rnsap_hS_PDSCH_Offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_PhysicalChannelOffset },
  { &hf_rnsap_hS_PDSCH_Midamble_Configuation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_timeslot_Resource_Related_Information_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_DSCH_TimeslotResourceLCR },
  { &hf_rnsap_startCode     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_endCode       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_transport_Block_Size_Index_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Transport_Block_Size_Index_LCR },
  { &hf_rnsap_modulationType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ModulationSPS_LCR },
  { &hf_rnsap_hS_SICH_Mapping_Index, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_Mapping_Index },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Initial_HS_PDSCH_SPS_Resource(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Initial_HS_PDSCH_SPS_Resource, Initial_HS_PDSCH_SPS_Resource_sequence);

  return offset;
}


static const value_string rnsap_Process_Memory_Size_vals[] = {
  {   0, "hms800" },
  {   1, "hms1600" },
  {   2, "hms2400" },
  {   3, "hms3200" },
  {   4, "hms4000" },
  {   5, "hms4800" },
  {   6, "hms5600" },
  {   7, "hms6400" },
  {   8, "hms7200" },
  {   9, "hms8000" },
  {  10, "hms8800" },
  {  11, "hms9600" },
  {  12, "hms10400" },
  {  13, "hms11200" },
  {  14, "hms12000" },
  {  15, "hms12800" },
  {  16, "hms13600" },
  {  17, "hms14400" },
  {  18, "hms15200" },
  {  19, "hms16000" },
  {  20, "hms17600" },
  {  21, "hms19200" },
  {  22, "hms20800" },
  {  23, "hms22400" },
  {  24, "hms24000" },
  {  25, "hms25600" },
  {  26, "hms27200" },
  {  27, "hms28800" },
  {  28, "hms30400" },
  {  29, "hms32000" },
  {  30, "hms36000" },
  {  31, "hms40000" },
  {  32, "hms44000" },
  {  33, "hms48000" },
  {  34, "hms52000" },
  {  35, "hms56000" },
  {  36, "hms60000" },
  {  37, "hms64000" },
  {  38, "hms68000" },
  {  39, "hms72000" },
  {  40, "hms76000" },
  {  41, "hms80000" },
  {  42, "hms88000" },
  {  43, "hms96000" },
  {  44, "hms104000" },
  {  45, "hms112000" },
  {  46, "hms120000" },
  {  47, "hms128000" },
  {  48, "hms136000" },
  {  49, "hms144000" },
  {  50, "hms152000" },
  {  51, "hms160000" },
  {  52, "hms176000" },
  {  53, "hms192000" },
  {  54, "hms208000" },
  {  55, "hms224000" },
  {  56, "hms240000" },
  {  57, "hms256000" },
  {  58, "hms272000" },
  {  59, "hms288000" },
  {  60, "hms304000" },
  { 0, NULL }
};

static value_string_ext rnsap_Process_Memory_Size_vals_ext = VALUE_STRING_EXT_INIT(rnsap_Process_Memory_Size_vals);


static int
dissect_rnsap_Process_Memory_Size(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     61, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_Number_of_Processes_for_HS_DSCH_SPS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 16U, NULL, FALSE);

  return offset;
}


static const per_sequence_t HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_sequence[] = {
  { &hf_rnsap_hS_SICH_InformationList_for_HS_DSCH_SPS, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_InformationList_for_HS_DSCH_SPS },
  { &hf_rnsap_initial_HS_PDSCH_SPS_Resource, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Initial_HS_PDSCH_SPS_Resource },
  { &hf_rnsap_buffer_Size_for_HS_DSCH_SPS, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Process_Memory_Size },
  { &hf_rnsap_number_of_Processes_for_HS_DSCH_SPS, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Number_of_Processes_for_HS_DSCH_SPS },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR, HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_sequence);

  return offset;
}



static int
dissect_rnsap_IMEI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       8, 8, FALSE, NULL);

  return offset;
}



static int
dissect_rnsap_IMEISV(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       8, 8, FALSE, NULL);

  return offset;
}



static int
dissect_rnsap_IMSI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       3, 8, FALSE, NULL);

  return offset;
}


static const value_string rnsap_Improved_Synchronized_Indicator_vals[] = {
  {   0, "true" },
  { 0, NULL }
};


static int
dissect_rnsap_Improved_Synchronized_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t GA_AccessPointPositionwithOptionalAltitude_sequence[] = {
  { &hf_rnsap_geographicalCoordinate, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GeographicalCoordinate },
  { &hf_rnsap_altitudeAndDirection, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_AltitudeAndDirection },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GA_AccessPointPositionwithOptionalAltitude(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GA_AccessPointPositionwithOptionalAltitude, GA_AccessPointPositionwithOptionalAltitude_sequence);

  return offset;
}


static const value_string rnsap_IPSpacingFDD_vals[] = {
  {   0, "ipsF5" },
  {   1, "ipsF7" },
  {   2, "ipsF10" },
  {   3, "ipsF15" },
  {   4, "ipsF20" },
  {   5, "ipsF30" },
  {   6, "ipsF40" },
  {   7, "ipsF50" },
  { 0, NULL }
};


static int
dissect_rnsap_IPSpacingFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_IPLength_vals[] = {
  {   0, "ipl5" },
  {   1, "ipl10" },
  { 0, NULL }
};


static int
dissect_rnsap_IPLength(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_IPOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 9U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Seed(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}


static const per_sequence_t IPDL_FDD_Parameters_sequence[] = {
  { &hf_rnsap_iPSpacingFDD  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_IPSpacingFDD },
  { &hf_rnsap_iPLength      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_IPLength },
  { &hf_rnsap_iPOffset      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_IPOffset },
  { &hf_rnsap_seed          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Seed },
  { &hf_rnsap_burstModeParameters, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BurstModeParameters },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_IPDL_FDD_Parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_IPDL_FDD_Parameters, IPDL_FDD_Parameters_sequence);

  return offset;
}


static const value_string rnsap_IPSpacingTDD_vals[] = {
  {   0, "ipsT30" },
  {   1, "ipsT40" },
  {   2, "ipsT50" },
  {   3, "ipsT70" },
  {   4, "ipsT100" },
  { 0, NULL }
};


static int
dissect_rnsap_IPSpacingTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_IPStart(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4095U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_IPSlot(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 14U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_IP_P_CCPCH_vals[] = {
  {   0, "switchOff-1-Frame" },
  {   1, "switchOff-2-Frames" },
  { 0, NULL }
};


static int
dissect_rnsap_IP_P_CCPCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t IPDL_TDD_Parameters_sequence[] = {
  { &hf_rnsap_iPSpacingTDD  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_IPSpacingTDD },
  { &hf_rnsap_iPStart       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_IPStart },
  { &hf_rnsap_iPSlot        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_IPSlot },
  { &hf_rnsap_iP_P_CCPCH    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_IP_P_CCPCH },
  { &hf_rnsap_burstModeParameters, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BurstModeParameters },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_IPDL_TDD_Parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_IPDL_TDD_Parameters, IPDL_TDD_Parameters_sequence);

  return offset;
}



static int
dissect_rnsap_Extension_IPDLParameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_IPDLParameters_vals[] = {
  {   0, "iPDL-FDD-Parameters" },
  {   1, "iPDL-TDD-Parameters" },
  {   2, "extension-IPDLParameters" },
  { 0, NULL }
};

static const per_choice_t IPDLParameters_choice[] = {
  {   0, &hf_rnsap_iPDL_FDD_Parameters, ASN1_EXTENSION_ROOT    , dissect_rnsap_IPDL_FDD_Parameters },
  {   1, &hf_rnsap_iPDL_TDD_Parameters, ASN1_EXTENSION_ROOT    , dissect_rnsap_IPDL_TDD_Parameters },
  {   2, &hf_rnsap_extension_IPDLParameters, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_IPDLParameters },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_IPDLParameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_IPDLParameters, IPDLParameters_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RequestedDataValue_sequence[] = {
  { &hf_rnsap_gA_AccessPointPositionwithAltitude, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_AccessPointPositionwithOptionalAltitude },
  { &hf_rnsap_iPDLParameters, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_IPDLParameters },
  { &hf_rnsap_dGPSCorrections, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DGPSCorrections },
  { &hf_rnsap_gPS_NavigationModel_and_TimeRecovery, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GPS_NavigationModel_and_TimeRecovery },
  { &hf_rnsap_gPS_Ionospheric_Model, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GPS_Ionospheric_Model },
  { &hf_rnsap_gPS_UTC_Model , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GPS_UTC_Model },
  { &hf_rnsap_gPS_Almanac   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GPS_Almanac },
  { &hf_rnsap_gPS_RealTime_Integrity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GPS_RealTime_Integrity },
  { &hf_rnsap_gPS_RX_POS    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GPS_RX_POS },
  { &hf_rnsap_sFNSFN_GA_AccessPointPosition, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_AccessPointPositionwithOptionalAltitude },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RequestedDataValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RequestedDataValue, RequestedDataValue_sequence);

  return offset;
}


static const per_sequence_t InformationAvailable_sequence[] = {
  { &hf_rnsap_requestedDataValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RequestedDataValue },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_InformationAvailable(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_InformationAvailable, InformationAvailable_sequence);

  return offset;
}



static int
dissect_rnsap_InformationExchangeID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1048575U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_InformationNotAvailable(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_24_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 24U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_InformationReportPeriodicity_vals[] = {
  {   0, "min" },
  {   1, "hour" },
  { 0, NULL }
};

static const per_choice_t InformationReportPeriodicity_choice[] = {
  {   0, &hf_rnsap_min           , ASN1_EXTENSION_ROOT    , dissect_rnsap_INTEGER_1_60_ },
  {   1, &hf_rnsap_hour          , ASN1_EXTENSION_ROOT    , dissect_rnsap_INTEGER_1_24_ },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_InformationReportPeriodicity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_InformationReportPeriodicity, InformationReportPeriodicity_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t PeriodicInformation_sequence[] = {
  { &hf_rnsap_informationReportPeriodicity, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_InformationReportPeriodicity },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PeriodicInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PeriodicInformation, PeriodicInformation_sequence);

  return offset;
}


static const value_string rnsap_InformationThreshold_vals[] = {
  {   0, "dGPSThreshold" },
  {   1, "dGANSSThreshold" },
  { 0, NULL }
};

static const per_choice_t InformationThreshold_choice[] = {
  {   0, &hf_rnsap_dGPSThreshold , ASN1_EXTENSION_ROOT    , dissect_rnsap_DGPSThreshold },
  {   1, &hf_rnsap_dGANSSThreshold, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_DGANSSThreshold },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_InformationThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_InformationThreshold, InformationThreshold_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t OnModificationInformation_sequence[] = {
  { &hf_rnsap_informationThreshold, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_InformationThreshold },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_OnModificationInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_OnModificationInformation, OnModificationInformation_sequence);

  return offset;
}


static const value_string rnsap_InformationReportCharacteristics_vals[] = {
  {   0, "onDemand" },
  {   1, "periodic" },
  {   2, "onModification" },
  { 0, NULL }
};

static const per_choice_t InformationReportCharacteristics_choice[] = {
  {   0, &hf_rnsap_onDemand      , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_periodic      , ASN1_EXTENSION_ROOT    , dissect_rnsap_PeriodicInformation },
  {   2, &hf_rnsap_onModification, ASN1_EXTENSION_ROOT    , dissect_rnsap_OnModificationInformation },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_InformationReportCharacteristics(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_InformationReportCharacteristics, InformationReportCharacteristics_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_T_informationTypeItem_vals[] = {
  {   0, "gA-AccessPointPositionwithAltitude" },
  {   1, "gA-AccessPointPosition" },
  {   2, "iPDLParameters" },
  {   3, "gPSInformation" },
  {   4, "dGPSCorrections" },
  {   5, "gPS-RX-POS" },
  {   6, "sFNSFN-GA-AccessPointPosition" },
  {   7, "cell-Capacity-Class" },
  {   8, "nACC-Related-Data" },
  {   9, "mBMSBearerServiceFullAddress" },
  {  10, "interFrequencyCellInformation" },
  {  11, "gANSSInformation" },
  {  12, "dGANSSCorrections" },
  {  13, "gANSS-RX-Pos" },
  {  14, "mBMS-Counting-Information" },
  {  15, "mBMS-Transmission-Mode" },
  {  16, "mBMS-Neighbouring-Cell-Information" },
  {  17, "mBMS-RLC-Sequence-Number" },
  {  18, "aNR-Cell-Information" },
  {  19, "cOmmonERGCH-Cell-Information" },
  { 0, NULL }
};


static int
dissect_rnsap_T_informationTypeItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, TRUE, 13, NULL);

  return offset;
}


static const per_sequence_t InformationType_sequence[] = {
  { &hf_rnsap_informationTypeItem, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_informationTypeItem },
  { &hf_rnsap_gPSInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GPSInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_InformationType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_InformationType, InformationType_sequence);

  return offset;
}


static const value_string rnsap_Initial_DL_DPCH_TimingAdjustment_Allowed_vals[] = {
  {   0, "initial-DL-DPCH-TimingAdjustment-Allowed" },
  { 0, NULL }
};


static int
dissect_rnsap_Initial_DL_DPCH_TimingAdjustment_Allowed(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t Inter_Frequency_Cell_sequence[] = {
  { &hf_rnsap_dL_UARFCN     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_uL_UARFCN     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_primaryScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PrimaryScramblingCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Inter_Frequency_Cell(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Inter_Frequency_Cell, Inter_Frequency_Cell_sequence);

  return offset;
}


static const per_sequence_t Inter_Frequency_Cell_List_sequence_of[1] = {
  { &hf_rnsap_Inter_Frequency_Cell_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cell },
};

static int
dissect_rnsap_Inter_Frequency_Cell_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Inter_Frequency_Cell_List, Inter_Frequency_Cell_List_sequence_of,
                                                  0, maxCellsMeas, FALSE);

  return offset;
}



static int
dissect_rnsap_Inter_Frequency_Cell_Indication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_InterFrequencyCellID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 31U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Inter_Frequency_Cell_SIB11_or_SIB12_sequence[] = {
  { &hf_rnsap_interFrequencyCellID_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_InterFrequencyCellID },
  { &hf_rnsap_dL_UARFCN     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_uL_UARFCN     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_primaryScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PrimaryScramblingCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12, Inter_Frequency_Cell_SIB11_or_SIB12_sequence);

  return offset;
}


static const per_sequence_t Inter_Frequency_Cell_SIB11_or_SIB12_List_sequence_of[1] = {
  { &hf_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12 },
};

static int
dissect_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12_List, Inter_Frequency_Cell_SIB11_or_SIB12_List_sequence_of,
                                                  0, maxCellSIB11OrSIB12, FALSE);

  return offset;
}


static const per_sequence_t Inter_Frequency_Cells_Information_SIB11_Per_Indication_sequence[] = {
  { &hf_rnsap_inter_Frequency_Cell_Indication_SIB11, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cell_Indication },
  { &hf_rnsap_inter_Frequency_Cell_List_SIB11, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Inter_Frequency_Cells_Information_SIB11_Per_Indication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Inter_Frequency_Cells_Information_SIB11_Per_Indication, Inter_Frequency_Cells_Information_SIB11_Per_Indication_sequence);

  return offset;
}


static const per_sequence_t Inter_Frequency_Cell_Information_SIB11_sequence_of[1] = {
  { &hf_rnsap_Inter_Frequency_Cell_Information_SIB11_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cells_Information_SIB11_Per_Indication },
};

static int
dissect_rnsap_Inter_Frequency_Cell_Information_SIB11(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Inter_Frequency_Cell_Information_SIB11, Inter_Frequency_Cell_Information_SIB11_sequence_of,
                                                  0, 2, FALSE);

  return offset;
}


static const per_sequence_t Inter_Frequency_Cells_Information_SIB12_Per_Indication_sequence[] = {
  { &hf_rnsap_inter_Frequency_Cell_Indication_SIB12, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cell_Indication },
  { &hf_rnsap_inter_Frequency_Cell_List_SIB12, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Inter_Frequency_Cells_Information_SIB12_Per_Indication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Inter_Frequency_Cells_Information_SIB12_Per_Indication, Inter_Frequency_Cells_Information_SIB12_Per_Indication_sequence);

  return offset;
}


static const per_sequence_t Inter_Frequency_Cell_Information_SIB12_sequence_of[1] = {
  { &hf_rnsap_Inter_Frequency_Cell_Information_SIB12_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cells_Information_SIB12_Per_Indication },
};

static int
dissect_rnsap_Inter_Frequency_Cell_Information_SIB12(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Inter_Frequency_Cell_Information_SIB12, Inter_Frequency_Cell_Information_SIB12_sequence_of,
                                                  0, 2, FALSE);

  return offset;
}


static const per_sequence_t Inter_Frequency_Cell_Information_sequence[] = {
  { &hf_rnsap_inter_Frequency_Cell_Information_SIB11, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cell_Information_SIB11 },
  { &hf_rnsap_inter_Frequency_Cell_Information_SIB12, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cell_Information_SIB12 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Inter_Frequency_Cell_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Inter_Frequency_Cell_Information, Inter_Frequency_Cell_Information_sequence);

  return offset;
}



static int
dissect_rnsap_InterFrequencyCellID_Ext(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxCellSIB11ter_1, NULL, FALSE);

  return offset;
}


static const per_sequence_t Inter_Frequency_Cell_SIB11ter_sequence[] = {
  { &hf_rnsap_interFrequencyCellID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_InterFrequencyCellID_Ext },
  { &hf_rnsap_dL_UARFCN     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_uL_UARFCN     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_primaryScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PrimaryScramblingCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Inter_Frequency_Cell_SIB11ter(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Inter_Frequency_Cell_SIB11ter, Inter_Frequency_Cell_SIB11ter_sequence);

  return offset;
}


static const per_sequence_t Inter_Frequency_Cell_SIB11ter_List_sequence_of[1] = {
  { &hf_rnsap_Inter_Frequency_Cell_SIB11ter_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cell_SIB11ter },
};

static int
dissect_rnsap_Inter_Frequency_Cell_SIB11ter_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Inter_Frequency_Cell_SIB11ter_List, Inter_Frequency_Cell_SIB11ter_List_sequence_of,
                                                  1, maxCellSIB11ter_ext, FALSE);

  return offset;
}


static const per_sequence_t Inter_Frequency_Cells_Information_SIB11ter_Per_Indication_sequence[] = {
  { &hf_rnsap_inter_Frequency_Cell_Indication_SIB11ter, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cell_Indication },
  { &hf_rnsap_inter_Frequency_Cell_List_SIB11ter, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cell_SIB11ter_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Inter_Frequency_Cells_Information_SIB11ter_Per_Indication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Inter_Frequency_Cells_Information_SIB11ter_Per_Indication, Inter_Frequency_Cells_Information_SIB11ter_Per_Indication_sequence);

  return offset;
}


static const per_sequence_t Inter_Frequency_Cell_Information_SIB11ter_sequence_of[1] = {
  { &hf_rnsap_Inter_Frequency_Cell_Information_SIB11ter_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Inter_Frequency_Cells_Information_SIB11ter_Per_Indication },
};

static int
dissect_rnsap_Inter_Frequency_Cell_Information_SIB11ter(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Inter_Frequency_Cell_Information_SIB11ter, Inter_Frequency_Cell_Information_SIB11ter_sequence_of,
                                                  1, 2, FALSE);

  return offset;
}



static int
dissect_rnsap_InterStream_Interference_Compensation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_IPSub_vals[] = {
  {   0, "first" },
  {   1, "second" },
  {   2, "both" },
  { 0, NULL }
};


static int
dissect_rnsap_IPSub(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t IPDL_TDD_ParametersLCR_sequence[] = {
  { &hf_rnsap_iPSpacingTDD  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_IPSpacingTDD },
  { &hf_rnsap_iPStart       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_IPStart },
  { &hf_rnsap_iPSub         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_IPSub },
  { &hf_rnsap_burstModeParameters, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BurstModeParameters },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_IPDL_TDD_ParametersLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_IPDL_TDD_ParametersLCR, IPDL_TDD_ParametersLCR_sequence);

  return offset;
}



static int
dissect_rnsap_IPMulticastAddress(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       4, 16, FALSE, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_2_3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            2U, 3U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_7(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, FALSE);

  return offset;
}


static const per_sequence_t IdleIntervalInformation_sequence[] = {
  { &hf_rnsap_idleIntervalInfo_k, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_2_3 },
  { &hf_rnsap_idleIntervalInfo_offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_7 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_IdleIntervalInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_IdleIntervalInformation, IdleIntervalInformation_sequence);

  return offset;
}


static const per_sequence_t LAI_List_sequence_of[1] = {
  { &hf_rnsap_LAI_List_item , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_LAI },
};

static int
dissect_rnsap_LAI_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_LAI_List, LAI_List_sequence_of,
                                                  1, maxNrOfLAIs, FALSE);

  return offset;
}


static const per_sequence_t LABased_sequence[] = {
  { &hf_rnsap_laiList       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_LAI_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_LABased(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_LABased, LABased_sequence);

  return offset;
}



static int
dissect_rnsap_Load_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Load_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_6(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 6U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_T_maxPhysChPerTimeslot_vals[] = {
  {   0, "ts1" },
  {   1, "ts2" },
  {   2, "ts3" },
  {   3, "ts4" },
  { 0, NULL }
};


static int
dissect_rnsap_T_maxPhysChPerTimeslot(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t LCRTDD_Uplink_Physical_Channel_Capability_sequence[] = {
  { &hf_rnsap_maxTimeslotsPerSubFrame, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_6 },
  { &hf_rnsap_maxPhysChPerTimeslot, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_maxPhysChPerTimeslot },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_LCRTDD_Uplink_Physical_Channel_Capability(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_LCRTDD_Uplink_Physical_Channel_Capability, LCRTDD_Uplink_Physical_Channel_Capability_sequence);

  return offset;
}


static const value_string rnsap_ReportInterval_vals[] = {
  {   0, "ms250" },
  {   1, "ms500" },
  {   2, "ms1000" },
  {   3, "ms2000" },
  {   4, "ms3000" },
  {   5, "ms4000" },
  {   6, "ms6000" },
  {   7, "ms12000" },
  {   8, "ms16000" },
  {   9, "ms20000" },
  {  10, "ms24000" },
  {  11, "ms32000" },
  {  12, "ms64000" },
  { 0, NULL }
};


static int
dissect_rnsap_ReportInterval(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     13, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_ReportAmount_vals[] = {
  {   0, "n1" },
  {   1, "n2" },
  {   2, "n4" },
  {   3, "n8" },
  {   4, "n16" },
  {   5, "n32" },
  {   6, "n64" },
  {   7, "infinity" },
  { 0, NULL }
};


static int
dissect_rnsap_ReportAmount(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t MDT_Report_Parameters_sequence[] = {
  { &hf_rnsap_reportInterval, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ReportInterval },
  { &hf_rnsap_reportAmount  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ReportAmount },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MDT_Report_Parameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MDT_Report_Parameters, MDT_Report_Parameters_sequence);

  return offset;
}


static const value_string rnsap_M1Report_vals[] = {
  {   0, "periodic" },
  {   1, "event1F" },
  { 0, NULL }
};

static const per_choice_t M1Report_choice[] = {
  {   0, &hf_rnsap_periodic_01   , ASN1_EXTENSION_ROOT    , dissect_rnsap_MDT_Report_Parameters },
  {   1, &hf_rnsap_event1F       , ASN1_EXTENSION_ROOT    , dissect_rnsap_Event1F_Parameters },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_M1Report(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_M1Report, M1Report_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_M2Report_vals[] = {
  {   0, "periodic" },
  {   1, "event1I" },
  { 0, NULL }
};

static const per_choice_t M2Report_choice[] = {
  {   0, &hf_rnsap_periodic_01   , ASN1_EXTENSION_ROOT    , dissect_rnsap_MDT_Report_Parameters },
  {   1, &hf_rnsap_event1I       , ASN1_EXTENSION_ROOT    , dissect_rnsap_Event1I_Parameters },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_M2Report(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_M2Report, M2Report_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_MDT_Activation_vals[] = {
  {   0, "mdt-only" },
  {   1, "mdt-and-trace" },
  { 0, NULL }
};


static int
dissect_rnsap_MDT_Activation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t RAI_sequence[] = {
  { &hf_rnsap_pLMN_Identity , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_PLMN_Identity },
  { &hf_rnsap_lAC           , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_LAC },
  { &hf_rnsap_rAC           , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_RAC },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RAI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RAI, RAI_sequence);

  return offset;
}


static const per_sequence_t RAI_List_sequence_of[1] = {
  { &hf_rnsap_RAI_List_item , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_RAI },
};

static int
dissect_rnsap_RAI_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RAI_List, RAI_List_sequence_of,
                                                  1, maxNrOfRAIs, FALSE);

  return offset;
}


static const per_sequence_t RABased_sequence[] = {
  { &hf_rnsap_raiList       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RAI_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RABased(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RABased, RABased_sequence);

  return offset;
}


static const value_string rnsap_MDTAreaScope_vals[] = {
  {   0, "cellbased" },
  {   1, "labased" },
  {   2, "rabased" },
  {   3, "plmn-area-based" },
  { 0, NULL }
};

static const per_choice_t MDTAreaScope_choice[] = {
  {   0, &hf_rnsap_cellbased     , ASN1_EXTENSION_ROOT    , dissect_rnsap_CellBased },
  {   1, &hf_rnsap_labased       , ASN1_EXTENSION_ROOT    , dissect_rnsap_LABased },
  {   2, &hf_rnsap_rabased       , ASN1_EXTENSION_ROOT    , dissect_rnsap_RABased },
  {   3, &hf_rnsap_plmn_area_based, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_MDTAreaScope(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_MDTAreaScope, MDTAreaScope_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t MDT_Configuration_sequence[] = {
  { &hf_rnsap_mdtActivation , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MDT_Activation },
  { &hf_rnsap_mdtAreaScope  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MDTAreaScope },
  { &hf_rnsap_m1report      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_M1Report },
  { &hf_rnsap_m2report      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_M2Report },
  { &hf_rnsap_m3report      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MDT_Configuration(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MDT_Configuration, MDT_Configuration_sequence);

  return offset;
}



static int
dissect_rnsap_MaxNrOfUL_DPCHs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 6U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_MACes_Maximum_Bitrate_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 256000000U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_MaxNrDLPhysicalchannels(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 224U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_MaxNrDLPhysicalchannels768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 448U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_MaxNrDLPhysicalchannelsTS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 16U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_MaxNrDLPhysicalchannelsTS768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 32U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_MaxNrTimeslots(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 14U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_MaxNrULPhysicalchannels(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 2U, NULL, FALSE);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_Full_Address_sequence[] = {
  { &hf_rnsap_accessPointName, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AccessPointName },
  { &hf_rnsap_iPMulticastAddress, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_IPMulticastAddress },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Bearer_Service_Full_Address(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Bearer_Service_Full_Address, MBMS_Bearer_Service_Full_Address_sequence);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_List_sequence_of[1] = {
  { &hf_rnsap_MBMS_Bearer_Service_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TMGI },
};

static int
dissect_rnsap_MBMS_Bearer_Service_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBMS_Bearer_Service_List, MBMS_Bearer_Service_List_sequence_of,
                                                  1, maxNrOfMBMSServices, FALSE);

  return offset;
}


static const per_sequence_t MBMSChannelTypeCellList_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_affectedUEInformationForMBMS, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AffectedUEInformationForMBMS },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMSChannelTypeCellList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMSChannelTypeCellList, MBMSChannelTypeCellList_sequence);

  return offset;
}


static const per_sequence_t PTMCellList_sequence_of[1] = {
  { &hf_rnsap_PTMCellList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMSChannelTypeCellList },
};

static int
dissect_rnsap_PTMCellList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_PTMCellList, PTMCellList_sequence_of,
                                                  1, maxNrOfCells, FALSE);

  return offset;
}


static const per_sequence_t PTPCellList_sequence_of[1] = {
  { &hf_rnsap_PTPCellList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMSChannelTypeCellList },
};

static int
dissect_rnsap_PTPCellList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_PTPCellList, PTPCellList_sequence_of,
                                                  1, maxNrOfCells, FALSE);

  return offset;
}


static const per_sequence_t NotProvidedCellList_sequence_of[1] = {
  { &hf_rnsap_NotProvidedCellList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMSChannelTypeCellList },
};

static int
dissect_rnsap_NotProvidedCellList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_NotProvidedCellList, NotProvidedCellList_sequence_of,
                                                  1, maxNrOfCells, FALSE);

  return offset;
}


static const per_sequence_t MBMSChannelTypeInfo_sequence[] = {
  { &hf_rnsap_tMGI          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TMGI },
  { &hf_rnsap_pTM_Cell_List , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PTMCellList },
  { &hf_rnsap_pTP_Cell_List , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PTPCellList },
  { &hf_rnsap_not_Provided_Cell_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NotProvidedCellList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMSChannelTypeInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMSChannelTypeInfo, MBMSChannelTypeInfo_sequence);

  return offset;
}


static const per_sequence_t PreferredFrequencyLayerInfo_sequence[] = {
  { &hf_rnsap_defaultPreferredFrequency, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_additionalPreferredFrequency, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AdditionalPreferredFrequency },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PreferredFrequencyLayerInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PreferredFrequencyLayerInfo, PreferredFrequencyLayerInfo_sequence);

  return offset;
}


static const per_sequence_t MBMSPreferredFreqLayerInfo_sequence[] = {
  { &hf_rnsap_tMGI          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TMGI },
  { &hf_rnsap_preferredFrequencyLayerInfo, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PreferredFrequencyLayerInfo },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMSPreferredFreqLayerInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMSPreferredFreqLayerInfo, MBMSPreferredFreqLayerInfo_sequence);

  return offset;
}


static const per_sequence_t MBMS_ConcatenatedServiceList_sequence_of[1] = {
  { &hf_rnsap_MBMS_ConcatenatedServiceList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TMGI },
};

static int
dissect_rnsap_MBMS_ConcatenatedServiceList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBMS_ConcatenatedServiceList, MBMS_ConcatenatedServiceList_sequence_of,
                                                  1, maxlengthMBMSconcatservlists, FALSE);

  return offset;
}


static const per_sequence_t MBMS_Neighbouring_Cell_Information_sequence[] = {
  { &hf_rnsap_mBMS_ConcatenatedServiceList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_ConcatenatedServiceList },
  { &hf_rnsap_l3_Information_1, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_L3_Information },
  { &hf_rnsap_l3_Information_2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_L3_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Neighbouring_Cell_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Neighbouring_Cell_Information, MBMS_Neighbouring_Cell_Information_sequence);

  return offset;
}



static int
dissect_rnsap_Time_Stamp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 9999U, NULL, FALSE);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_List_RLCinfo_sequence[] = {
  { &hf_rnsap_tmgi          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TMGI },
  { &hf_rnsap_time_Stamp    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Time_Stamp },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Bearer_Service_List_RLCinfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Bearer_Service_List_RLCinfo, MBMS_Bearer_Service_List_RLCinfo_sequence);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_List_RLC_sequence_of[1] = {
  { &hf_rnsap_MBMS_Bearer_Service_List_RLC_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_Service_List_RLCinfo },
};

static int
dissect_rnsap_MBMS_Bearer_Service_List_RLC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBMS_Bearer_Service_List_RLC, MBMS_Bearer_Service_List_RLC_sequence_of,
                                                  1, maxNrOfMBMSServices, FALSE);

  return offset;
}


static const per_sequence_t MBMS_RLC_Sequence_Number_Information_List_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_mBMS_Bearer_Service_List_RLC, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_Service_List_RLC },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_RLC_Sequence_Number_Information_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_RLC_Sequence_Number_Information_List, MBMS_RLC_Sequence_Number_Information_List_sequence);

  return offset;
}


static const per_sequence_t MBMS_RLC_Sequence_Number_Information_sequence_of[1] = {
  { &hf_rnsap_MBMS_RLC_Sequence_Number_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_RLC_Sequence_Number_Information_List },
};

static int
dissect_rnsap_MBMS_RLC_Sequence_Number_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBMS_RLC_Sequence_Number_Information, MBMS_RLC_Sequence_Number_Information_sequence_of,
                                                  1, maxNrOfCells, FALSE);

  return offset;
}



static int
dissect_rnsap_MBSFN_Cluster_Identity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}


static const per_sequence_t MCCH_Message_List_sequence_of[1] = {
  { &hf_rnsap_MCCH_Message_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_L3_Information },
};

static int
dissect_rnsap_MCCH_Message_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MCCH_Message_List, MCCH_Message_List_sequence_of,
                                                  1, maxNrOfMCCHMessages, FALSE);

  return offset;
}



static int
dissect_rnsap_Secondary_CCPCH_System_Information_MBMS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     NO_BOUND, NO_BOUND, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t MCCH_Configuration_Info_sequence[] = {
  { &hf_rnsap_secondaryCCPCHSystemInformationMBMS, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Secondary_CCPCH_System_Information_MBMS },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MCCH_Configuration_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MCCH_Configuration_Info, MCCH_Configuration_Info_sequence);

  return offset;
}


static const value_string rnsap_MbsfnSchedulingTransmissionTimeInterval_vals[] = {
  {   0, "tti4" },
  {   1, "tti8" },
  {   2, "tti16" },
  {   3, "tti32" },
  {   4, "tti64" },
  {   5, "tti128" },
  {   6, "tti256" },
  { 0, NULL }
};


static int
dissect_rnsap_MbsfnSchedulingTransmissionTimeInterval(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t MBSFN_Scheduling_Transmission_Time_Interval_Item_sequence[] = {
  { &hf_rnsap_tMGI          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TMGI },
  { &hf_rnsap_mbsfnSchedulingTransmissionTimeInterval, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MbsfnSchedulingTransmissionTimeInterval },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Item, MBSFN_Scheduling_Transmission_Time_Interval_Item_sequence);

  return offset;
}


static const per_sequence_t MBSFN_Scheduling_Transmission_Time_Interval_Info_List_sequence_of[1] = {
  { &hf_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Info_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Item },
};

static int
dissect_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Info_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Info_List, MBSFN_Scheduling_Transmission_Time_Interval_Info_List_sequence_of,
                                                  1, maxNrOfMBMSL3, FALSE);

  return offset;
}


static const value_string rnsap_MeasurementFilterCoefficient_vals[] = {
  {   0, "k0" },
  {   1, "k1" },
  {   2, "k2" },
  {   3, "k3" },
  {   4, "k4" },
  {   5, "k5" },
  {   6, "k6" },
  {   7, "k7" },
  {   8, "k8" },
  {   9, "k9" },
  {  10, "k11" },
  {  11, "k13" },
  {  12, "k15" },
  {  13, "k17" },
  {  14, "k19" },
  { 0, NULL }
};

static value_string_ext rnsap_MeasurementFilterCoefficient_vals_ext = VALUE_STRING_EXT_INIT(rnsap_MeasurementFilterCoefficient_vals);


static int
dissect_rnsap_MeasurementFilterCoefficient(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     15, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_MeasurementID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1048575U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_MinimumSpreadingFactor(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 16U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_MinimumSpreadingFactor768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 32U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_MultipleURAsIndicator_vals[] = {
  {   0, "multiple-URAs-exist" },
  {   1, "single-URA-exists" },
  { 0, NULL }
};


static int
dissect_rnsap_MultipleURAsIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_MeasurementRecoveryBehavior(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_MeasurementRecoveryReportingIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_MeasurementRecoverySupportIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_MIMO_Mode_Indicator_vals[] = {
  {   0, "activate" },
  {   1, "deactivate" },
  { 0, NULL }
};


static int
dissect_rnsap_MIMO_Mode_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_MIMO_withfourtransmitantennas_ActivationIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_MIMO_withfourtransmitantennas_ModeIndicator_vals[] = {
  {   0, "activate" },
  {   1, "deactivate" },
  { 0, NULL }
};


static int
dissect_rnsap_MIMO_withfourtransmitantennas_ModeIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_DualStream_MIMO_withfourtransmitantennas_ActivationIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_DualStream_MIMO_withfourtransmitantennas_ModeIndicator_vals[] = {
  {   0, "activate" },
  {   1, "deactivate" },
  { 0, NULL }
};


static int
dissect_rnsap_DualStream_MIMO_withfourtransmitantennas_ModeIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t Additional_Secondary_CPICH_sequence[] = {
  { &hf_rnsap_additional_secondary_CPICH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MIMO_S_CPICH_Channelisation_Code },
  { &hf_rnsap_additional_D_CPICH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MIMO_S_CPICH_Channelisation_Code },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_Secondary_CPICH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_Secondary_CPICH, Additional_Secondary_CPICH_sequence);

  return offset;
}


static const per_sequence_t Additional_Associated_Secondary_CPICH_sequence_of[1] = {
  { &hf_rnsap_Additional_Associated_Secondary_CPICH_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_Secondary_CPICH },
};

static int
dissect_rnsap_Additional_Associated_Secondary_CPICH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_Associated_Secondary_CPICH, Additional_Associated_Secondary_CPICH_sequence_of,
                                                  1, maxSCPICHCell_1, FALSE);

  return offset;
}


static const value_string rnsap_MinUL_ChannelisationCodeLength_vals[] = {
  {   0, "v4" },
  {   1, "v8" },
  {   2, "v16" },
  {   3, "v32" },
  {   4, "v64" },
  {   5, "v128" },
  {   6, "v256" },
  { 0, NULL }
};


static int
dissect_rnsap_MinUL_ChannelisationCodeLength(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_MinimumReducedE_DPDCH_GainFactor_vals[] = {
  {   0, "m8-15" },
  {   1, "m11-15" },
  {   2, "m15-15" },
  {   3, "m21-15" },
  {   4, "m30-15" },
  {   5, "m42-15" },
  {   6, "m60-15" },
  {   7, "m84-15" },
  { 0, NULL }
};


static int
dissect_rnsap_MinimumReducedE_DPDCH_GainFactor(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_Secondary_UL_Frequency_Activation_State_vals[] = {
  {   0, "activated" },
  {   1, "deactivated" },
  { 0, NULL }
};


static int
dissect_rnsap_Secondary_UL_Frequency_Activation_State(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t MulticellEDCH_InformationItemIEs_sequence[] = {
  { &hf_rnsap_dL_PowerBalancing_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_PowerBalancing_Information },
  { &hf_rnsap_minimumReducedE_DPDCH_GainFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MinimumReducedE_DPDCH_GainFactor },
  { &hf_rnsap_secondary_UL_Frequency_Activation_State, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Secondary_UL_Frequency_Activation_State },
  { &hf_rnsap_f_DPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_F_DPCH_SlotFormat },
  { &hf_rnsap_common_DL_ReferencePowerInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MulticellEDCH_InformationItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MulticellEDCH_InformationItemIEs, MulticellEDCH_InformationItemIEs_sequence);

  return offset;
}


static const value_string rnsap_Phase_Reference_Update_Indicator_vals[] = {
  {   0, "phase-reference-needs-to-be-changed" },
  { 0, NULL }
};


static int
dissect_rnsap_Phase_Reference_Update_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t MulticellEDCH_RL_SpecificInformationItemIEs_sequence[] = {
  { &hf_rnsap_extendedPropagationDelay, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ExtendedPropagationDelay },
  { &hf_rnsap_enhanced_PrimaryCPICH_EcNo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Enhanced_PrimaryCPICH_EcNo },
  { &hf_rnsap_dl_Reference_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_phase_Reference_Update_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Phase_Reference_Update_Indicator },
  { &hf_rnsap_additional_e_DCH_DL_Control_Channel_Grant, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MulticellEDCH_RL_SpecificInformationItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MulticellEDCH_RL_SpecificInformationItemIEs, MulticellEDCH_RL_SpecificInformationItemIEs_sequence);

  return offset;
}



static int
dissect_rnsap_Multicell_EDCH_Restriction(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_boolean(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const value_string rnsap_MultiplexingPosition_vals[] = {
  {   0, "fixed" },
  {   1, "flexible" },
  { 0, NULL }
};


static int
dissect_rnsap_MultiplexingPosition(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_MAChs_ResetIndicator_vals[] = {
  {   0, "mAChs-NotReset" },
  { 0, NULL }
};


static int
dissect_rnsap_MAChs_ResetIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR_sequence[] = {
  { &hf_rnsap_hSSCCH_TDD_Specific_InfoList_Response_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HSSCCH_TDD_Specific_InfoList_Response_LCR },
  { &hf_rnsap_hARQ_MemoryPartitioning, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HARQ_MemoryPartitioning },
  { &hf_rnsap_uARFCN        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR, MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR_sequence);

  return offset;
}


static const per_sequence_t MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_sequence_of[1] = {
  { &hf_rnsap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR },
};

static int
dissect_rnsap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR, MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_sequence_of,
                                                  1, maxHSDPAFrequency_1, FALSE);

  return offset;
}


static const value_string rnsap_MIMO_SFMode_For_HSPDSCHDualStream_vals[] = {
  {   0, "sF1" },
  {   1, "sF1SF16" },
  { 0, NULL }
};


static int
dissect_rnsap_MIMO_SFMode_For_HSPDSCHDualStream(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_Multicarrier_EDCH_Transport_Bearer_Mode_vals[] = {
  {   0, "separate-Iur-Transport-Bearer-Mode" },
  {   1, "shared-Iur-Transport-Bearer-Mode" },
  { 0, NULL }
};


static int
dissect_rnsap_Multicarrier_EDCH_Transport_Bearer_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_SNPL_Carrier_Group_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 3U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Multi_Carrier_EDCH_MACdFlows_Specific_Info_sequence[] = {
  { &hf_rnsap_e_DCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multi_Carrier_EDCH_MACdFlows_Specific_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multi_Carrier_EDCH_MACdFlows_Specific_Info, Multi_Carrier_EDCH_MACdFlows_Specific_Info_sequence);

  return offset;
}


static const per_sequence_t Multi_Carrier_EDCH_MACdFlows_Information_TDD_sequence_of[1] = {
  { &hf_rnsap_Multi_Carrier_EDCH_MACdFlows_Information_TDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Multi_Carrier_EDCH_MACdFlows_Specific_Info },
};

static int
dissect_rnsap_Multi_Carrier_EDCH_MACdFlows_Information_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Multi_Carrier_EDCH_MACdFlows_Information_TDD, Multi_Carrier_EDCH_MACdFlows_Information_TDD_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t Multi_Carrier_EDCH_LCR_InformationItem_sequence[] = {
  { &hf_rnsap_uARFCNforNt   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_sNPL_carrier_group_indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SNPL_Carrier_Group_Indicator },
  { &hf_rnsap_pRxdesBase    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_PUCH_PRXdesBase },
  { &hf_rnsap_multi_Carrier_EDCH_MACdFlows_Information_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Multi_Carrier_EDCH_MACdFlows_Information_TDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multi_Carrier_EDCH_LCR_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multi_Carrier_EDCH_LCR_InformationItem, Multi_Carrier_EDCH_LCR_InformationItem_sequence);

  return offset;
}


static const per_sequence_t Multi_Carrier_EDCH_Information_sequence_of[1] = {
  { &hf_rnsap_Multi_Carrier_EDCH_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Multi_Carrier_EDCH_LCR_InformationItem },
};

static int
dissect_rnsap_Multi_Carrier_EDCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Multi_Carrier_EDCH_Information, Multi_Carrier_EDCH_Information_sequence_of,
                                                  1, maxNrOfULCarriersLCR_1, FALSE);

  return offset;
}


static const per_sequence_t Multi_Carrier_EDCH_Info_sequence[] = {
  { &hf_rnsap_multicarrier_EDCH_Transport_Bearer_Mode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Multicarrier_EDCH_Transport_Bearer_Mode },
  { &hf_rnsap_multi_carrier_EDCH_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Multi_Carrier_EDCH_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multi_Carrier_EDCH_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multi_Carrier_EDCH_Info, Multi_Carrier_EDCH_Info_sequence);

  return offset;
}


static const per_sequence_t Multi_Carrier_EDCH_Information_Removal_Info_ItemIEs_sequence[] = {
  { &hf_rnsap_uARFCNforNt   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multi_Carrier_EDCH_Information_Removal_Info_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multi_Carrier_EDCH_Information_Removal_Info_ItemIEs, Multi_Carrier_EDCH_Information_Removal_Info_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Multi_Carrier_EDCH_Information_Removal_List_sequence_of[1] = {
  { &hf_rnsap_Multi_Carrier_EDCH_Information_Removal_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Multi_Carrier_EDCH_Information_Removal_Info_ItemIEs },
};

static int
dissect_rnsap_Multi_Carrier_EDCH_Information_Removal_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Multi_Carrier_EDCH_Information_Removal_List, Multi_Carrier_EDCH_Information_Removal_List_sequence_of,
                                                  1, maxNrOfULCarriersLCR_1, FALSE);

  return offset;
}


static const per_sequence_t Multi_Carrier_EDCH_Change_Info_sequence[] = {
  { &hf_rnsap_multicarrier_EDCH_Transport_Bearer_Mode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Multicarrier_EDCH_Transport_Bearer_Mode },
  { &hf_rnsap_multi_carrier_EDCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Multi_Carrier_EDCH_Information },
  { &hf_rnsap_multi_Carrier_EDCH_Information_Removal_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Multi_Carrier_EDCH_Information_Removal_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multi_Carrier_EDCH_Change_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multi_Carrier_EDCH_Change_Info, Multi_Carrier_EDCH_Change_Info_sequence);

  return offset;
}


static const value_string rnsap_Continue_Setup_Change_Multi_Carrier_EDCH_vals[] = {
  {   0, "continue" },
  {   1, "setup" },
  {   2, "change" },
  { 0, NULL }
};

static const per_choice_t Continue_Setup_Change_Multi_Carrier_EDCH_choice[] = {
  {   0, &hf_rnsap_continue      , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_setup         , ASN1_EXTENSION_ROOT    , dissect_rnsap_Multi_Carrier_EDCH_Info },
  {   2, &hf_rnsap_change        , ASN1_EXTENSION_ROOT    , dissect_rnsap_Multi_Carrier_EDCH_Change_Info },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Continue_Setup_Change_Multi_Carrier_EDCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Continue_Setup_Change_Multi_Carrier_EDCH, Continue_Setup_Change_Multi_Carrier_EDCH_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Multi_Carrier_EDCH_Reconfigure_sequence[] = {
  { &hf_rnsap_continue_setup_change_Of_Multi_Carrier_EDCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Continue_Setup_Change_Multi_Carrier_EDCH },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multi_Carrier_EDCH_Reconfigure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multi_Carrier_EDCH_Reconfigure, Multi_Carrier_EDCH_Reconfigure_sequence);

  return offset;
}


static const per_sequence_t Multi_Carrier_EDCH_LCR_Information_ResponseItem_sequence[] = {
  { &hf_rnsap_uARFCNforNt   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_e_DCH_TDD_MACdFlow_Specific_InformationResp, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp },
  { &hf_rnsap_e_AGCH_Specific_Information_Response_LCR_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_AGCH_Specific_InformationRespList_LCR_TDD },
  { &hf_rnsap_e_HICH_Scheduled_InformationResp_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_HICH_Scheduled_InformationRespList_LCR_TDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multi_Carrier_EDCH_LCR_Information_ResponseItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multi_Carrier_EDCH_LCR_Information_ResponseItem, Multi_Carrier_EDCH_LCR_Information_ResponseItem_sequence);

  return offset;
}


static const per_sequence_t Multi_Carrier_EDCH_Information_Response_sequence_of[1] = {
  { &hf_rnsap_Multi_Carrier_EDCH_Information_Response_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Multi_Carrier_EDCH_LCR_Information_ResponseItem },
};

static int
dissect_rnsap_Multi_Carrier_EDCH_Information_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Multi_Carrier_EDCH_Information_Response, Multi_Carrier_EDCH_Information_Response_sequence_of,
                                                  1, maxNrOfULCarriersLCR_1, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_2_32_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            2U, 32U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_Multiflow_Mode_vals[] = {
  {   0, "primary" },
  {   1, "assisting" },
  { 0, NULL }
};


static int
dissect_rnsap_Multiflow_Mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_Multiflow_MIMO_vals[] = {
  {   0, "on" },
  {   1, "off" },
  { 0, NULL }
};


static int
dissect_rnsap_Multiflow_MIMO(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_30_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 30U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_Multiflow_Timing_vals[] = {
  {   0, "time-Reference" },
  {   1, "non-Time-Reference" },
  { 0, NULL }
};

static const per_choice_t Multiflow_Timing_choice[] = {
  {   0, &hf_rnsap_time_Reference, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_non_Time_Reference, ASN1_EXTENSION_ROOT    , dissect_rnsap_INTEGER_0_30_ },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Multiflow_Timing(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Multiflow_Timing, Multiflow_Timing_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_16_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 16U, NULL, TRUE);

  return offset;
}


static const per_sequence_t Multiflow_Information_sequence[] = {
  { &hf_rnsap_total_Number_of_HS_DSCH_Cells, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_2_32_ },
  { &hf_rnsap_mode          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Multiflow_Mode },
  { &hf_rnsap_mimo          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Multiflow_MIMO },
  { &hf_rnsap_timing        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Multiflow_Timing },
  { &hf_rnsap_max_Number_of_HS_SCCH_Sets_per_NodeB, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_1_16_ },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multiflow_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multiflow_Information, Multiflow_Information_sequence);

  return offset;
}


static const per_sequence_t Multiflow_Information_To_Modify_sequence[] = {
  { &hf_rnsap_total_Number_of_HS_DSCH_Cells, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_2_32_ },
  { &hf_rnsap_mode          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Multiflow_Mode },
  { &hf_rnsap_mimo          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Multiflow_MIMO },
  { &hf_rnsap_timing        , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Multiflow_Timing },
  { &hf_rnsap_max_Number_of_HS_SCCH_Sets_per_NodeB, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_1_16_ },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multiflow_Information_To_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multiflow_Information_To_Modify, Multiflow_Information_To_Modify_sequence);

  return offset;
}


static const value_string rnsap_Multiflow_Stop_vals[] = {
  {   0, "stop" },
  { 0, NULL }
};


static int
dissect_rnsap_Multiflow_Stop(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_Multiflow_Reconfiguration_vals[] = {
  {   0, "setup" },
  {   1, "configurationChange" },
  {   2, "stop" },
  { 0, NULL }
};

static const per_choice_t Multiflow_Reconfiguration_choice[] = {
  {   0, &hf_rnsap_setup_01      , ASN1_EXTENSION_ROOT    , dissect_rnsap_Multiflow_Information },
  {   1, &hf_rnsap_configurationChange, ASN1_EXTENSION_ROOT    , dissect_rnsap_Multiflow_Information_To_Modify },
  {   2, &hf_rnsap_stop          , ASN1_EXTENSION_ROOT    , dissect_rnsap_Multiflow_Stop },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Multiflow_Reconfiguration(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Multiflow_Reconfiguration, Multiflow_Reconfiguration_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Multiflow_RepetitionFactors_sequence[] = {
  { &hf_rnsap_assisting_CQI_RepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CQI_RepetitionFactor },
  { &hf_rnsap_assisting_AckNack_RepetitionFactor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AckNack_RepetitionFactor },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multiflow_RepetitionFactors(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multiflow_RepetitionFactors, Multiflow_RepetitionFactors_sequence);

  return offset;
}



static int
dissect_rnsap_Multiflow_OrdinalNumberOfFrequency(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 32U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_MU_MIMO_Operation_vals[] = {
  {   0, "mU-MIMO-Used" },
  {   1, "mU-MIMO-Not-Used" },
  { 0, NULL }
};


static int
dissect_rnsap_MU_MIMO_Operation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_Standalone_Midamble_Configuration_vals[] = {
  {   0, "v2" },
  {   1, "v4" },
  {   2, "v6" },
  {   3, "v8" },
  {   4, "v10" },
  {   5, "v12" },
  {   6, "v14" },
  {   7, "v16" },
  { 0, NULL }
};


static int
dissect_rnsap_Standalone_Midamble_Configuration(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_Standalone_MidambleShift(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_Standalone_Midamble_RepetitionPeriod_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  {   4, "v16" },
  {   5, "v32" },
  {   6, "v64" },
  { 0, NULL }
};


static int
dissect_rnsap_Standalone_Midamble_RepetitionPeriod(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     7, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_Standalone_Midamble_Offset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 63U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_ReferenceBeta(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -15, 16U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Standalone_Midamble_Channel_Information_sequence[] = {
  { &hf_rnsap_standalone_Midamble_Configuration, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Standalone_Midamble_Configuration },
  { &hf_rnsap_standalone_MidambleShift, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Standalone_MidambleShift },
  { &hf_rnsap_timeslotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_repetitionPeriod_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Standalone_Midamble_RepetitionPeriod },
  { &hf_rnsap_offset        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Standalone_Midamble_Offset },
  { &hf_rnsap_referenceBeta , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ReferenceBeta },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Standalone_Midamble_Channel_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Standalone_Midamble_Channel_Information, Standalone_Midamble_Channel_Information_sequence);

  return offset;
}


static const per_sequence_t MU_MIMO_InformationLCR_sequence[] = {
  { &hf_rnsap_mU_MIMO_Operation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MU_MIMO_Operation },
  { &hf_rnsap_standalone_Midamble_Channel_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Standalone_Midamble_Channel_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MU_MIMO_InformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MU_MIMO_InformationLCR, MU_MIMO_InformationLCR_sequence);

  return offset;
}


static const value_string rnsap_MU_MIMO_Usage_Indicator_vals[] = {
  {   0, "ul-Only" },
  {   1, "dl-Only" },
  {   2, "ul-and-dl" },
  { 0, NULL }
};


static int
dissect_rnsap_MU_MIMO_Usage_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_Standalone_Midamble_Channel_Indicator_vals[] = {
  {   0, "used" },
  {   1, "not-used" },
  { 0, NULL }
};


static int
dissect_rnsap_Standalone_Midamble_Channel_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t MU_MIMO_Indicator_sequence[] = {
  { &hf_rnsap_mU_MIMO_Usage_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MU_MIMO_Usage_Indicator },
  { &hf_rnsap_standalone_Midamble_Channel_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Standalone_Midamble_Channel_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MU_MIMO_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MU_MIMO_Indicator, MU_MIMO_Indicator_sequence);

  return offset;
}


static const per_sequence_t NACC_Related_Data_sequence[] = {
  { &hf_rnsap_gERAN_SI_Type , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GERAN_SI_Type },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_NACC_Related_Data(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_NACC_Related_Data, NACC_Related_Data_sequence);

  return offset;
}


static const per_sequence_t Neighbouring_UMTS_CellInformation_sequence_of[1] = {
  { &hf_rnsap_Neighbouring_UMTS_CellInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_Neighbouring_UMTS_CellInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Neighbouring_UMTS_CellInformation, Neighbouring_UMTS_CellInformation_sequence_of,
                                                  1, maxNrOfNeighbouringRNCs, FALSE);

  return offset;
}


static const per_sequence_t Neighbouring_UMTS_CellInformation_Ext_sequence_of[1] = {
  { &hf_rnsap_Neighbouring_UMTS_CellInformation_Ext_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_Neighbouring_UMTS_CellInformation_Ext(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Neighbouring_UMTS_CellInformation_Ext, Neighbouring_UMTS_CellInformation_Ext_sequence_of,
                                                  1, maxNrOfExtendedNeighbouringRNCs, FALSE);

  return offset;
}


static const per_sequence_t Neighbouring_FDD_CellInformationItem_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_uARFCNforNu   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_uARFCNforNd   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_frameOffset   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_FrameOffset },
  { &hf_rnsap_primaryScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PrimaryScramblingCode },
  { &hf_rnsap_primaryCPICH_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PrimaryCPICH_Power },
  { &hf_rnsap_cellIndividualOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CellIndividualOffset },
  { &hf_rnsap_txDiversityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TxDiversityIndicator },
  { &hf_rnsap_sTTD_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_STTD_SupportIndicator },
  { &hf_rnsap_closedLoopMode1_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ClosedLoopMode1_SupportIndicator },
  { &hf_rnsap_not_used_closedLoopMode2_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Neighbouring_FDD_CellInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Neighbouring_FDD_CellInformationItem, Neighbouring_FDD_CellInformationItem_sequence);

  return offset;
}


static const per_sequence_t Neighbouring_FDD_CellInformation_sequence_of[1] = {
  { &hf_rnsap_Neighbouring_FDD_CellInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Neighbouring_FDD_CellInformationItem },
};

static int
dissect_rnsap_Neighbouring_FDD_CellInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Neighbouring_FDD_CellInformation, Neighbouring_FDD_CellInformation_sequence_of,
                                                  1, maxNrOfFDDNeighboursPerRNC, TRUE);

  return offset;
}


static const per_sequence_t Neighbouring_TDD_CellInformationItem_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_uARFCNforNt   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_frameOffset   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_FrameOffset },
  { &hf_rnsap_cellParameterID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CellParameterID },
  { &hf_rnsap_syncCase      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SyncCase },
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TimeSlot },
  { &hf_rnsap_sCH_TimeSlot  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SCH_TimeSlot },
  { &hf_rnsap_sCTD_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SCTD_Indicator },
  { &hf_rnsap_cellIndividualOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CellIndividualOffset },
  { &hf_rnsap_dPCHConstantValue, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DPCHConstantValue },
  { &hf_rnsap_pCCPCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PCCPCH_Power },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Neighbouring_TDD_CellInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Neighbouring_TDD_CellInformationItem, Neighbouring_TDD_CellInformationItem_sequence);

  return offset;
}


static const per_sequence_t Neighbouring_TDD_CellInformation_sequence_of[1] = {
  { &hf_rnsap_Neighbouring_TDD_CellInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Neighbouring_TDD_CellInformationItem },
};

static int
dissect_rnsap_Neighbouring_TDD_CellInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Neighbouring_TDD_CellInformation, Neighbouring_TDD_CellInformation_sequence_of,
                                                  1, maxNrOfTDDNeighboursPerRNC, TRUE);

  return offset;
}


static const per_sequence_t Neighbouring_LCR_TDD_CellInformationItem_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_uARFCNforNt   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_frameOffset   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_FrameOffset },
  { &hf_rnsap_cellParameterID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CellParameterID },
  { &hf_rnsap_sCTD_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SCTD_Indicator },
  { &hf_rnsap_cellIndividualOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CellIndividualOffset },
  { &hf_rnsap_dPCHConstantValue, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DPCHConstantValue },
  { &hf_rnsap_pCCPCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PCCPCH_Power },
  { &hf_rnsap_restrictionStateIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RestrictionStateIndicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Neighbouring_LCR_TDD_CellInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Neighbouring_LCR_TDD_CellInformationItem, Neighbouring_LCR_TDD_CellInformationItem_sequence);

  return offset;
}


static const per_sequence_t Neighbouring_LCR_TDD_CellInformation_sequence_of[1] = {
  { &hf_rnsap_Neighbouring_LCR_TDD_CellInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Neighbouring_LCR_TDD_CellInformationItem },
};

static int
dissect_rnsap_Neighbouring_LCR_TDD_CellInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Neighbouring_LCR_TDD_CellInformation, Neighbouring_LCR_TDD_CellInformation_sequence_of,
                                                  1, maxNrOfLCRTDDNeighboursPerRNC, TRUE);

  return offset;
}


static const per_sequence_t Neighbouring_UMTS_CellInformationExtensionItem_sequence[] = {
  { &hf_rnsap_rNC_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RNC_ID },
  { &hf_rnsap_cN_PS_DomainIdentifier, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CN_PS_DomainIdentifier },
  { &hf_rnsap_cN_CS_DomainIdentifier, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CN_CS_DomainIdentifier },
  { &hf_rnsap_neighbouring_FDD_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_FDD_CellInformation },
  { &hf_rnsap_neighbouring_TDD_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_TDD_CellInformation },
  { &hf_rnsap_neighbouring_LCR_TDD_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_LCR_TDD_CellInformation },
  { &hf_rnsap_extended_RNC_ID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Extended_RNC_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Neighbouring_UMTS_CellInformationExtensionItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Neighbouring_UMTS_CellInformationExtensionItem, Neighbouring_UMTS_CellInformationExtensionItem_sequence);

  return offset;
}


static const per_sequence_t Neighbouring_UMTS_CellInformationItem_sequence[] = {
  { &hf_rnsap_rNC_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RNC_ID },
  { &hf_rnsap_cN_PS_DomainIdentifier, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CN_PS_DomainIdentifier },
  { &hf_rnsap_cN_CS_DomainIdentifier, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CN_CS_DomainIdentifier },
  { &hf_rnsap_neighbouring_FDD_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_FDD_CellInformation },
  { &hf_rnsap_neighbouring_TDD_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_TDD_CellInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Neighbouring_UMTS_CellInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Neighbouring_UMTS_CellInformationItem, Neighbouring_UMTS_CellInformationItem_sequence);

  return offset;
}


static const per_sequence_t MultipleFrequencyBandsIndicatorList_sequence_of[1] = {
  { &hf_rnsap_MultipleFrequencyBandsIndicatorList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_FrequencyBandIndicator },
};

static int
dissect_rnsap_MultipleFrequencyBandsIndicatorList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MultipleFrequencyBandsIndicatorList, MultipleFrequencyBandsIndicatorList_sequence_of,
                                                  0, maxMultipleFrequencyBands, FALSE);

  return offset;
}


static const per_sequence_t NeighbouringFDDCellMeasurementInformation_sequence[] = {
  { &hf_rnsap_uC_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UC_ID },
  { &hf_rnsap_uARFCN        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_primaryScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PrimaryScramblingCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_NeighbouringFDDCellMeasurementInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_NeighbouringFDDCellMeasurementInformation, NeighbouringFDDCellMeasurementInformation_sequence);

  return offset;
}



static int
dissect_rnsap_Neighbouring_GSM_CellInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Neighbouring_GSM_CellInformationItem_sequence[] = {
  { &hf_rnsap_cGI           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CGI },
  { &hf_rnsap_cellIndividualOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CellIndividualOffset },
  { &hf_rnsap_bSIC          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BSIC },
  { &hf_rnsap_band_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Band_Indicator },
  { &hf_rnsap_bCCH_ARFCN    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BCCH_ARFCN },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Neighbouring_GSM_CellInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Neighbouring_GSM_CellInformationItem, Neighbouring_GSM_CellInformationItem_sequence);

  return offset;
}


static const per_sequence_t Neighbouring_GSM_CellInformationIEs_sequence_of[1] = {
  { &hf_rnsap_Neighbouring_GSM_CellInformationIEs_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Neighbouring_GSM_CellInformationItem },
};

static int
dissect_rnsap_Neighbouring_GSM_CellInformationIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Neighbouring_GSM_CellInformationIEs, Neighbouring_GSM_CellInformationIEs_sequence_of,
                                                  1, maxNrOfGSMNeighboursPerRNC, TRUE);

  return offset;
}


static const per_sequence_t NeighbouringTDDCellMeasurementInformation_sequence[] = {
  { &hf_rnsap_uC_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UC_ID },
  { &hf_rnsap_uARFCN        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_cellParameterID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CellParameterID },
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_NeighbouringTDDCellMeasurementInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_NeighbouringTDDCellMeasurementInformation, NeighbouringTDDCellMeasurementInformation_sequence);

  return offset;
}


static const per_sequence_t NeighbouringTDDCellMeasurementInformationLCR_sequence[] = {
  { &hf_rnsap_uC_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UC_ID },
  { &hf_rnsap_uARFCN        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_cellParameterID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CellParameterID },
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_NeighbouringTDDCellMeasurementInformationLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_NeighbouringTDDCellMeasurementInformationLCR, NeighbouringTDDCellMeasurementInformationLCR_sequence);

  return offset;
}


static const per_sequence_t NeighbouringTDDCellMeasurementInformation768_sequence[] = {
  { &hf_rnsap_uC_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UC_ID },
  { &hf_rnsap_uARFCN        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UARFCN },
  { &hf_rnsap_cellParameterID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CellParameterID },
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_NeighbouringTDDCellMeasurementInformation768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_NeighbouringTDDCellMeasurementInformation768, NeighbouringTDDCellMeasurementInformation768_sequence);

  return offset;
}


static const per_sequence_t Neighbouring_E_UTRA_CellInformationItem_sequence[] = {
  { &hf_rnsap_eCGI          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ECGI },
  { &hf_rnsap_eARFCN_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EARFCN_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Neighbouring_E_UTRA_CellInformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Neighbouring_E_UTRA_CellInformationItem, Neighbouring_E_UTRA_CellInformationItem_sequence);

  return offset;
}


static const per_sequence_t Neighbouring_E_UTRA_CellInformation_sequence_of[1] = {
  { &hf_rnsap_Neighbouring_E_UTRA_CellInformation_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Neighbouring_E_UTRA_CellInformationItem },
};

static int
dissect_rnsap_Neighbouring_E_UTRA_CellInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Neighbouring_E_UTRA_CellInformation, Neighbouring_E_UTRA_CellInformation_sequence_of,
                                                  1, maxNrOfEUTRANeighboursPerRNC, TRUE);

  return offset;
}



static int
dissect_rnsap_RadioFrequencyBandEUTRA(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}


static const per_sequence_t MultipleFrequencyBandsIndicatorList_EUTRA_sequence_of[1] = {
  { &hf_rnsap_MultipleFrequencyBandsIndicatorList_EUTRA_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_RadioFrequencyBandEUTRA },
};

static int
dissect_rnsap_MultipleFrequencyBandsIndicatorList_EUTRA(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MultipleFrequencyBandsIndicatorList_EUTRA, MultipleFrequencyBandsIndicatorList_EUTRA_sequence_of,
                                                  0, maxMultipleFrequencyBandsEUTRA, FALSE);

  return offset;
}


static const value_string rnsap_NonCellSpecificTxDiversity_vals[] = {
  {   0, "txDiversity" },
  { 0, NULL }
};


static int
dissect_rnsap_NonCellSpecificTxDiversity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_Non_rectangular_resource_allocation_indicator_vals[] = {
  {   0, "activate" },
  { 0, NULL }
};


static int
dissect_rnsap_Non_rectangular_resource_allocation_indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_Non_rectangular_resource_timeslot_set(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     7, 7, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_NrOfDLchannelisationcodes(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 8U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_NRT_Load_Information_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 3U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_NRT_Load_Information_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 3U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_3(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 3U, NULL, FALSE);

  return offset;
}


static const per_sequence_t NRTLoadInformationValue_sequence[] = {
  { &hf_rnsap_uplinkNRTLoadInformationValue, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_3 },
  { &hf_rnsap_downlinkNRTLoadInformationValue, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_3 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_NRTLoadInformationValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_NRTLoadInformationValue, NRTLoadInformationValue_sequence);

  return offset;
}


static const value_string rnsap_Number_Of_Supported_Carriers_vals[] = {
  {   0, "one-one-carrier" },
  {   1, "one-three-carrier" },
  {   2, "three-three-carrier" },
  {   3, "one-six-carrier" },
  {   4, "three-six-carrier" },
  {   5, "six-six-carrier" },
  {   6, "one-two-carrier-discontiguous" },
  {   7, "two-two-carrier-discontiguous" },
  {   8, "one-two-carrier-contiguous" },
  {   9, "two-two-carrier-contiguous" },
  { 0, NULL }
};


static int
dissect_rnsap_Number_Of_Supported_Carriers(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     6, NULL, TRUE, 4, NULL);

  return offset;
}



static int
dissect_rnsap_NoOfTargetCellHS_SCCH_Order(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 30U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_New_non_serving_RL_setup_selection_vals[] = {
  {   0, "new-Serving-RL-in-DRNS" },
  {   1, "new-Serving-RL-Not-in-DRNS" },
  {   2, "new-Serving-RL-in-or-Not-in-DRNS" },
  { 0, NULL }
};

static const per_choice_t New_non_serving_RL_setup_selection_choice[] = {
  {   0, &hf_rnsap_new_Serving_RL_in_DRNS, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_new_Serving_RL_Not_in_DRNS, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   2, &hf_rnsap_new_Serving_RL_in_or_Not_in_DRNS, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_New_non_serving_RL_setup_selection(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_New_non_serving_RL_setup_selection, New_non_serving_RL_setup_selection_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Non_Serving_RL_Preconfig_Setup_sequence[] = {
  { &hf_rnsap_new_non_serving_RL_selection, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_New_non_serving_RL_setup_selection },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Non_Serving_RL_Preconfig_Setup(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Non_Serving_RL_Preconfig_Setup, Non_Serving_RL_Preconfig_Setup_sequence);

  return offset;
}



static int
dissect_rnsap_Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_item_sequence[] = {
  { &hf_rnsap_new_non_serving_RL_E_DCH_FDD_DL_Control_Channel_Information_A, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_EDCH_FDD_DL_ControlChannelInformation },
  { &hf_rnsap_new_non_serving_RL_E_DCH_FDD_DL_Control_Channel_Information_B, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_EDCH_FDD_DL_ControlChannelInformation },
  { &hf_rnsap_new_non_serving_RL_E_DCH_FDD_DL_Control_Channel_Information_C, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_EDCH_FDD_DL_ControlChannelInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_item, Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_item_sequence);

  return offset;
}


static const per_sequence_t Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_sequence_of[1] = {
  { &hf_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_item },
};

static int
dissect_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList, Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const value_string rnsap_NeedforIdleInterval_vals[] = {
  {   0, "true" },
  {   1, "false" },
  { 0, NULL }
};


static int
dissect_rnsap_NeedforIdleInterval(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t OnModification_sequence[] = {
  { &hf_rnsap_measurementThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementThreshold },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_OnModification(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_OnModification, OnModification_sequence);

  return offset;
}



static int
dissect_rnsap_OrdinalNumberOfFrequency(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 32U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_Out_of_Sychronization_Window_vals[] = {
  {   0, "ms40" },
  {   1, "ms80" },
  {   2, "ms160" },
  {   3, "ms320" },
  {   4, "ms640" },
  { 0, NULL }
};


static int
dissect_rnsap_Out_of_Sychronization_Window(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_PagingCause_vals[] = {
  {   0, "terminating-conversational-call" },
  {   1, "terminating-streaming-call" },
  {   2, "terminating-interactive-call" },
  {   3, "terminating-background-call" },
  {   4, "terminating-low-priority-signalling" },
  {   5, "terminating-high-priority-signalling" },
  {   6, "terminating-cause-unknown" },
  { 0, NULL }
};


static int
dissect_rnsap_PagingCause(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 2, NULL);

  return offset;
}


static const value_string rnsap_PagingRecordType_vals[] = {
  {   0, "imsi-gsm-map" },
  {   1, "tmsi-gsm-map" },
  {   2, "p-tmsi-gsm-map" },
  {   3, "imsi-ds-41" },
  {   4, "tmsi-ds-41" },
  { 0, NULL }
};


static int
dissect_rnsap_PagingRecordType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     5, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_PartialReportingIndicator_vals[] = {
  {   0, "partial-reporting-allowed" },
  { 0, NULL }
};


static int
dissect_rnsap_PartialReportingIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t PCH_InformationItem_sequence[] = {
  { &hf_rnsap_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PCH_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PCH_InformationItem, PCH_InformationItem_sequence);

  return offset;
}


static const per_sequence_t PCH_InformationList_sequence_of[1] = {
  { &hf_rnsap_PCH_InformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_PCH_InformationItem },
};

static int
dissect_rnsap_PCH_InformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_PCH_InformationList, PCH_InformationList_sequence_of,
                                                  0, 1, FALSE);

  return offset;
}



static int
dissect_rnsap_PCI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 503U, NULL, TRUE);

  return offset;
}


static const per_sequence_t Periodic_sequence[] = {
  { &hf_rnsap_reportPeriodicity, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ReportPeriodicity },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Periodic(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Periodic, Periodic_sequence);

  return offset;
}


static const value_string rnsap_Permanent_NAS_UE_Identity_vals[] = {
  {   0, "imsi" },
  { 0, NULL }
};

static const per_choice_t Permanent_NAS_UE_Identity_choice[] = {
  {   0, &hf_rnsap_imsi          , ASN1_EXTENSION_ROOT    , dissect_rnsap_IMSI },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Permanent_NAS_UE_Identity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Permanent_NAS_UE_Identity, Permanent_NAS_UE_Identity_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_PLCCHsequenceNumber(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 14U, NULL, FALSE);

  return offset;
}


static const per_sequence_t PLMN_List_sequence_of[1] = {
  { &hf_rnsap_PLMN_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_PLMN_Identity },
};

static int
dissect_rnsap_PLMN_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_PLMN_List, PLMN_List_sequence_of,
                                                  0, maxNrOfBroadcastPLMNs, FALSE);

  return offset;
}



static int
dissect_rnsap_PowerOffsetForSecondaryCPICHforMIMO(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -6, 0U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_PowerOffsetForSecondaryCPICHforMIMORequestIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_CommonPhysicalChannelID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_PowerOffsetForSCPICH_DCPICHforMIMOwithfourtransmitantennas(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -12, 0U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Associated_Secondary_CPICH_sequence[] = {
  { &hf_rnsap_associated_secondary_CPICH_channelId, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CommonPhysicalChannelID },
  { &hf_rnsap_associated_secondary_CPICH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PowerOffsetForSCPICH_DCPICHforMIMOwithfourtransmitantennas },
  { &hf_rnsap_associated_D_CPICH_channelId, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CommonPhysicalChannelID },
  { &hf_rnsap_associated_D_CPICH, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PowerOffsetForSCPICH_DCPICHforMIMOwithfourtransmitantennas },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Associated_Secondary_CPICH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Associated_Secondary_CPICH, Associated_Secondary_CPICH_sequence);

  return offset;
}


static const per_sequence_t PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas_sequence_of[1] = {
  { &hf_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Associated_Secondary_CPICH },
};

static int
dissect_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas, PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas_sequence_of,
                                                  1, maxSCPICHCell, FALSE);

  return offset;
}



static int
dissect_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennasRequestIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_Precoder_Weight_Set_Restriction(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     64, 64, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_PredictedSFNSFNDeviationLimit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_PredictedTUTRANGPSDeviationLimit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_PrimaryCCPCH_RSCP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 91U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_PrimaryCCPCH_RSCP_Delta(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -5, -1, NULL, TRUE);

  return offset;
}


static const per_sequence_t ProvidedInformation_sequence[] = {
  { &hf_rnsap_mBMSChannelTypeInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MBMSChannelTypeInfo },
  { &hf_rnsap_mBMSPreferredFreqLayerInfo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MBMSPreferredFreqLayerInfo },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ProvidedInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ProvidedInformation, ProvidedInformation_sequence);

  return offset;
}


static const per_sequence_t UpPCH_InformationList_LCRTDD_sequence_of[1] = {
  { &hf_rnsap_UpPCH_InformationList_LCRTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_UpPCH_InformationList_LCRTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UpPCH_InformationList_LCRTDD, UpPCH_InformationList_LCRTDD_sequence_of,
                                                  0, maxFrequencyinCell, FALSE);

  return offset;
}



static int
dissect_rnsap_UPPCHPositionLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}


static const per_sequence_t UpPCH_InformationItem_LCRTDD_sequence[] = {
  { &hf_rnsap_uARFCNforNt   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_uPPCHPositionLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UPPCHPositionLCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UpPCH_InformationItem_LCRTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UpPCH_InformationItem_LCRTDD, UpPCH_InformationItem_LCRTDD_sequence);

  return offset;
}



static int
dissect_rnsap_RANAP_EnhancedRelocationInformationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     NO_BOUND, NO_BOUND, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_RANAP_EnhancedRelocationInformationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     NO_BOUND, NO_BOUND, FALSE, NULL, 0, NULL, NULL);

  return offset;
}



static int
dissect_rnsap_RANAP_RelocationInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  tvbuff_t *parameter_tvb=NULL;

  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     NO_BOUND, NO_BOUND, FALSE, NULL, 0, &parameter_tvb, NULL);


  if (!parameter_tvb)
    return offset;

  /* Calling RANAP dissector but preventing changes to the protocol column */
  col_set_writable(actx->pinfo->cinfo, COL_PROTOCOL, FALSE);
  call_dissector_only(ranap_handle, parameter_tvb, actx->pinfo, tree, NULL);
  col_set_writable(actx->pinfo->cinfo, COL_PROTOCOL, TRUE);


  return offset;
}



static int
dissect_rnsap_RB_Identity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 31U, NULL, FALSE);

  return offset;
}


static const per_sequence_t RB_Info_sequence_of[1] = {
  { &hf_rnsap_RB_Info_item  , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_RB_Identity },
};

static int
dissect_rnsap_RB_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RB_Info, RB_Info_sequence_of,
                                                  1, maxNoOfRB, FALSE);

  return offset;
}



static int
dissect_rnsap_Received_Total_Wideband_Power_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_Received_total_wide_band_power(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_Received_Total_Wideband_Power_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 620U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_Released_CN_Domain_vals[] = {
  {   0, "pSDomain" },
  {   1, "cSDomain" },
  {   2, "pS-CSDomain" },
  { 0, NULL }
};

static const per_choice_t Released_CN_Domain_choice[] = {
  {   0, &hf_rnsap_pSDomain      , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_cSDomain      , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   2, &hf_rnsap_pS_CSDomain   , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Released_CN_Domain(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Released_CN_Domain, Released_CN_Domain_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_Extension_ReportCharacteristics(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_ReportCharacteristics_vals[] = {
  {   0, "onDemand" },
  {   1, "periodic" },
  {   2, "eventA" },
  {   3, "eventB" },
  {   4, "eventC" },
  {   5, "eventD" },
  {   6, "eventE" },
  {   7, "eventF" },
  {   8, "extension-ReportCharacteristics" },
  { 0, NULL }
};

static const per_choice_t ReportCharacteristics_choice[] = {
  {   0, &hf_rnsap_onDemand      , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   1, &hf_rnsap_periodic_02   , ASN1_EXTENSION_ROOT    , dissect_rnsap_Periodic },
  {   2, &hf_rnsap_eventA        , ASN1_EXTENSION_ROOT    , dissect_rnsap_EventA },
  {   3, &hf_rnsap_eventB        , ASN1_EXTENSION_ROOT    , dissect_rnsap_EventB },
  {   4, &hf_rnsap_eventC        , ASN1_EXTENSION_ROOT    , dissect_rnsap_EventC },
  {   5, &hf_rnsap_eventD        , ASN1_EXTENSION_ROOT    , dissect_rnsap_EventD },
  {   6, &hf_rnsap_eventE        , ASN1_EXTENSION_ROOT    , dissect_rnsap_EventE },
  {   7, &hf_rnsap_eventF        , ASN1_EXTENSION_ROOT    , dissect_rnsap_EventF },
  {   8, &hf_rnsap_extension_ReportCharacteristics, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_ReportCharacteristics },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_ReportCharacteristics(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_ReportCharacteristics, ReportCharacteristics_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_RequestedDataValueInformation_vals[] = {
  {   0, "informationAvailable" },
  {   1, "informationNotAvailable" },
  { 0, NULL }
};

static const per_choice_t RequestedDataValueInformation_choice[] = {
  {   0, &hf_rnsap_informationAvailable, ASN1_NO_EXTENSIONS     , dissect_rnsap_InformationAvailable },
  {   1, &hf_rnsap_informationNotAvailable, ASN1_NO_EXTENSIONS     , dissect_rnsap_InformationNotAvailable },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_RequestedDataValueInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_RequestedDataValueInformation, RequestedDataValueInformation_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_Specific_DCH_Info_Item_sequence[] = {
  { &hf_rnsap_dCH_id        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_ID },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Specific_DCH_Info_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Specific_DCH_Info_Item, RL_Specific_DCH_Info_Item_sequence);

  return offset;
}


static const per_sequence_t RL_Specific_DCH_Info_sequence_of[1] = {
  { &hf_rnsap_RL_Specific_DCH_Info_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Specific_DCH_Info_Item },
};

static int
dissect_rnsap_RL_Specific_DCH_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Specific_DCH_Info, RL_Specific_DCH_Info_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t RL_Specific_EDCH_InfoItem_sequence[] = {
  { &hf_rnsap_eDCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Specific_EDCH_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Specific_EDCH_InfoItem, RL_Specific_EDCH_InfoItem_sequence);

  return offset;
}


static const per_sequence_t RL_Specific_EDCH_Info_sequence_of[1] = {
  { &hf_rnsap_RL_Specific_EDCH_Info_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Specific_EDCH_InfoItem },
};

static int
dissect_rnsap_RL_Specific_EDCH_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Specific_EDCH_Info, RL_Specific_EDCH_Info_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t RL_Specific_EDCH_Information_sequence[] = {
  { &hf_rnsap_rL_Specific_EDCH_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Specific_EDCH_Info },
  { &hf_rnsap_e_AGCH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_AGCH_PowerOffset },
  { &hf_rnsap_e_RGCH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_RGCH_PowerOffset },
  { &hf_rnsap_e_HICH_PowerOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_HICH_PowerOffset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Specific_EDCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Specific_EDCH_Information, RL_Specific_EDCH_Information_sequence);

  return offset;
}


static const value_string rnsap_DL_RLC_PDU_Size_Format_vals[] = {
  {   0, "fixed-RLC-PDU-Size" },
  {   1, "flexible-RLC-PDU-Size" },
  { 0, NULL }
};


static int
dissect_rnsap_DL_RLC_PDU_Size_Format(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_RLC_Sequence_Number(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_RNTI_Allocation_Indicator_vals[] = {
  {   0, "true" },
  { 0, NULL }
};


static int
dissect_rnsap_RNTI_Allocation_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_RT_Load_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_RT_Load_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}


static const per_sequence_t RTLoadValue_sequence[] = {
  { &hf_rnsap_uplinkRTLoadValue, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_100 },
  { &hf_rnsap_downlinkRTLoadValue, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_100 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RTLoadValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RTLoadValue, RTLoadValue_sequence);

  return offset;
}



static int
dissect_rnsap_RxTimingDeviationForTA(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_RxTimingDeviationForTAext(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_RxTimingDeviationForTA768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 1023U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Rx_Timing_Deviation_Value_ext(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 32767U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Rx_Timing_Deviation_Value_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 511U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Rx_Timing_Deviation_Value_768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_SAC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  tvbuff_t *parameter_tvb = NULL;
  offset = dissect_per_octet_string(tvb, offset, actx, tree, -1,
                                       2, 2, FALSE, &parameter_tvb);

  if (parameter_tvb) {
    actx->created_item = proto_tree_add_item(tree, hf_index, parameter_tvb, 0, 2, ENC_BIG_ENDIAN);
  }


  return offset;
}


static const per_sequence_t SAI_sequence[] = {
  { &hf_rnsap_pLMN_Identity , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_PLMN_Identity },
  { &hf_rnsap_lAC           , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_LAC },
  { &hf_rnsap_sAC           , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_SAC },
  { &hf_rnsap_iE_Extensions , ASN1_NO_EXTENSIONS     , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SAI(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SAI, SAI_sequence);

  return offset;
}



static int
dissect_rnsap_SDPCCH_PowerOffsetInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 6U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_4095(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4095U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_maxCTFC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, maxCTFC, NULL, FALSE);

  return offset;
}


static const value_string rnsap_TFCS_CTFC_vals[] = {
  {   0, "ctfc2bit" },
  {   1, "ctfc4bit" },
  {   2, "ctfc6bit" },
  {   3, "ctfc8bit" },
  {   4, "ctfc12bit" },
  {   5, "ctfc16bit" },
  {   6, "ctfcmaxbit" },
  { 0, NULL }
};

static const per_choice_t TFCS_CTFC_choice[] = {
  {   0, &hf_rnsap_ctfc2bit      , ASN1_NO_EXTENSIONS     , dissect_rnsap_INTEGER_0_3 },
  {   1, &hf_rnsap_ctfc4bit      , ASN1_NO_EXTENSIONS     , dissect_rnsap_INTEGER_0_15 },
  {   2, &hf_rnsap_ctfc6bit      , ASN1_NO_EXTENSIONS     , dissect_rnsap_INTEGER_0_63 },
  {   3, &hf_rnsap_ctfc8bit      , ASN1_NO_EXTENSIONS     , dissect_rnsap_INTEGER_0_255 },
  {   4, &hf_rnsap_ctfc12bit     , ASN1_NO_EXTENSIONS     , dissect_rnsap_INTEGER_0_4095 },
  {   5, &hf_rnsap_ctfc16bit     , ASN1_NO_EXTENSIONS     , dissect_rnsap_INTEGER_0_65535 },
  {   6, &hf_rnsap_ctfcmaxbit    , ASN1_NO_EXTENSIONS     , dissect_rnsap_INTEGER_0_maxCTFC },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_TFCS_CTFC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_TFCS_CTFC, TFCS_CTFC_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t T_signalledGainFactors_sequence[] = {
  { &hf_rnsap_betaC         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BetaCD },
  { &hf_rnsap_betaD         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BetaCD },
  { &hf_rnsap_refTFCNumber  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RefTFCNumber },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_T_signalledGainFactors(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_T_signalledGainFactors, T_signalledGainFactors_sequence);

  return offset;
}


static const value_string rnsap_TransportFormatCombination_Beta_vals[] = {
  {   0, "signalledGainFactors" },
  {   1, "refTFCNumber" },
  { 0, NULL }
};

static const per_choice_t TransportFormatCombination_Beta_choice[] = {
  {   0, &hf_rnsap_signalledGainFactors, ASN1_EXTENSION_ROOT    , dissect_rnsap_T_signalledGainFactors },
  {   1, &hf_rnsap_refTFCNumber  , ASN1_EXTENSION_ROOT    , dissect_rnsap_RefTFCNumber },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_TransportFormatCombination_Beta(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_TransportFormatCombination_Beta, TransportFormatCombination_Beta_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TFCS_TFCSList_item_sequence[] = {
  { &hf_rnsap_cTFC          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCS_CTFC },
  { &hf_rnsap_tFC_Beta      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportFormatCombination_Beta },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TFCS_TFCSList_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TFCS_TFCSList_item, TFCS_TFCSList_item_sequence);

  return offset;
}


static const per_sequence_t TFCS_TFCSList_sequence_of[1] = {
  { &hf_rnsap_TFCS_TFCSList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCS_TFCSList_item },
};

static int
dissect_rnsap_TFCS_TFCSList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TFCS_TFCSList, TFCS_TFCSList_sequence_of,
                                                  1, maxNrOfTFCs, FALSE);

  return offset;
}


static const value_string rnsap_T_tFCSvalues_vals[] = {
  {   0, "no-Split-in-TFCI" },
  {   1, "not-Used-split-in-TFCI" },
  { 0, NULL }
};

static const per_choice_t T_tFCSvalues_choice[] = {
  {   0, &hf_rnsap_no_Split_in_TFCI, ASN1_EXTENSION_ROOT    , dissect_rnsap_TFCS_TFCSList },
  {   1, &hf_rnsap_not_Used_split_in_TFCI, ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_T_tFCSvalues(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_T_tFCSvalues, T_tFCSvalues_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TFCS_sequence[] = {
  { &hf_rnsap_tFCSvalues    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_tFCSvalues },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TFCS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TFCS, TFCS_sequence);

  return offset;
}


static const value_string rnsap_TFCI_Coding_vals[] = {
  {   0, "v4" },
  {   1, "v8" },
  {   2, "v16" },
  {   3, "v32" },
  { 0, NULL }
};


static int
dissect_rnsap_TFCI_Coding(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_TDD_Code_InformationItem_sequence[] = {
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Secondary_CCPCH_TDD_Code_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Secondary_CCPCH_TDD_Code_InformationItem, Secondary_CCPCH_TDD_Code_InformationItem_sequence);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_TDD_Code_Information_sequence_of[1] = {
  { &hf_rnsap_Secondary_CCPCH_TDD_Code_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Secondary_CCPCH_TDD_Code_InformationItem },
};

static int
dissect_rnsap_Secondary_CCPCH_TDD_Code_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Secondary_CCPCH_TDD_Code_Information, Secondary_CCPCH_TDD_Code_Information_sequence_of,
                                                  1, maxNrOfSCCPCHs, FALSE);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_TDD_InformationItem_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_secondary_CCPCH_TDD_Code_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Secondary_CCPCH_TDD_Code_Information },
  { &hf_rnsap_tDD_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_PhysicalChannelOffset },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Secondary_CCPCH_TDD_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Secondary_CCPCH_TDD_InformationItem, Secondary_CCPCH_TDD_InformationItem_sequence);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_TDD_InformationList_sequence_of[1] = {
  { &hf_rnsap_Secondary_CCPCH_TDD_InformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Secondary_CCPCH_TDD_InformationItem },
};

static int
dissect_rnsap_Secondary_CCPCH_TDD_InformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Secondary_CCPCH_TDD_InformationList, Secondary_CCPCH_TDD_InformationList_sequence_of,
                                                  0, maxNrOfSCCPCHs, FALSE);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_Info_TDD_sequence[] = {
  { &hf_rnsap_dl_TFCS       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCS },
  { &hf_rnsap_tFCI_Coding   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Coding },
  { &hf_rnsap_secondary_CCPCH_TDD_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Secondary_CCPCH_TDD_InformationList },
  { &hf_rnsap_fACH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FACH_InformationList },
  { &hf_rnsap_pCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PCH_InformationList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Secondary_CCPCH_Info_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Secondary_CCPCH_Info_TDD, Secondary_CCPCH_Info_TDD_sequence);

  return offset;
}


static const per_sequence_t Secondary_LCR_CCPCH_TDD_Code_InformationItem_sequence[] = {
  { &hf_rnsap_tDD_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCodeLCR },
  { &hf_rnsap_s_CCPCH_TimeSlotFormat_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Secondary_LCR_CCPCH_TDD_Code_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Secondary_LCR_CCPCH_TDD_Code_InformationItem, Secondary_LCR_CCPCH_TDD_Code_InformationItem_sequence);

  return offset;
}


static const per_sequence_t Secondary_LCR_CCPCH_TDD_Code_Information_sequence_of[1] = {
  { &hf_rnsap_Secondary_LCR_CCPCH_TDD_Code_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Secondary_LCR_CCPCH_TDD_Code_InformationItem },
};

static int
dissect_rnsap_Secondary_LCR_CCPCH_TDD_Code_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Secondary_LCR_CCPCH_TDD_Code_Information, Secondary_LCR_CCPCH_TDD_Code_Information_sequence_of,
                                                  1, maxNrOfSCCPCHs, FALSE);

  return offset;
}


static const per_sequence_t Secondary_LCR_CCPCH_TDD_InformationItem_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_secondary_LCR_CCPCH_TDD_Code_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Secondary_LCR_CCPCH_TDD_Code_Information },
  { &hf_rnsap_tDD_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_PhysicalChannelOffset },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Secondary_LCR_CCPCH_TDD_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Secondary_LCR_CCPCH_TDD_InformationItem, Secondary_LCR_CCPCH_TDD_InformationItem_sequence);

  return offset;
}


static const per_sequence_t Secondary_LCR_CCPCH_TDD_InformationList_sequence_of[1] = {
  { &hf_rnsap_Secondary_LCR_CCPCH_TDD_InformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Secondary_LCR_CCPCH_TDD_InformationItem },
};

static int
dissect_rnsap_Secondary_LCR_CCPCH_TDD_InformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Secondary_LCR_CCPCH_TDD_InformationList, Secondary_LCR_CCPCH_TDD_InformationList_sequence_of,
                                                  0, maxNrOfSCCPCHs, FALSE);

  return offset;
}


static const per_sequence_t Secondary_LCR_CCPCH_Info_TDD_sequence[] = {
  { &hf_rnsap_dl_TFCS       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCS },
  { &hf_rnsap_tFCI_Coding   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Coding },
  { &hf_rnsap_secondary_LCR_CCPCH_TDD_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Secondary_LCR_CCPCH_TDD_InformationList },
  { &hf_rnsap_fACH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FACH_InformationList },
  { &hf_rnsap_pCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PCH_InformationList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Secondary_LCR_CCPCH_Info_TDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Secondary_LCR_CCPCH_Info_TDD, Secondary_LCR_CCPCH_Info_TDD_sequence);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_TDD_Code_InformationItem768_sequence[] = {
  { &hf_rnsap_tDD_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Secondary_CCPCH_TDD_Code_InformationItem768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Secondary_CCPCH_TDD_Code_InformationItem768, Secondary_CCPCH_TDD_Code_InformationItem768_sequence);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_TDD_Code_Information768_sequence_of[1] = {
  { &hf_rnsap_Secondary_CCPCH_TDD_Code_Information768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Secondary_CCPCH_TDD_Code_InformationItem768 },
};

static int
dissect_rnsap_Secondary_CCPCH_TDD_Code_Information768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Secondary_CCPCH_TDD_Code_Information768, Secondary_CCPCH_TDD_Code_Information768_sequence_of,
                                                  1, maxNrOfSCCPCHs768, FALSE);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_TDD_InformationItem768_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_secondary_CCPCH_TDD_Code_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Secondary_CCPCH_TDD_Code_Information768 },
  { &hf_rnsap_tDD_PhysicalChannelOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_PhysicalChannelOffset },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Secondary_CCPCH_TDD_InformationItem768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Secondary_CCPCH_TDD_InformationItem768, Secondary_CCPCH_TDD_InformationItem768_sequence);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_TDD_InformationList768_sequence_of[1] = {
  { &hf_rnsap_Secondary_CCPCH_TDD_InformationList768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Secondary_CCPCH_TDD_InformationItem768 },
};

static int
dissect_rnsap_Secondary_CCPCH_TDD_InformationList768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Secondary_CCPCH_TDD_InformationList768, Secondary_CCPCH_TDD_InformationList768_sequence_of,
                                                  0, maxNrOfSCCPCHs768, FALSE);

  return offset;
}


static const per_sequence_t Secondary_CCPCH_Info_TDD768_sequence[] = {
  { &hf_rnsap_dl_TFCS       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCS },
  { &hf_rnsap_tFCI_Coding   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Coding },
  { &hf_rnsap_secondary_CCPCH_TDD_InformationList768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Secondary_CCPCH_TDD_InformationList768 },
  { &hf_rnsap_fACH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FACH_InformationList },
  { &hf_rnsap_pCH_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PCH_InformationList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Secondary_CCPCH_Info_TDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Secondary_CCPCH_Info_TDD768, Secondary_CCPCH_Info_TDD768_sequence);

  return offset;
}



static int
dissect_rnsap_S_E_DPCCH_PowerOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 17U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_S_E_ROCH_Release_Indicator_vals[] = {
  {   0, "s-E-ROCHreleased" },
  { 0, NULL }
};


static int
dissect_rnsap_S_E_ROCH_Release_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency_vals[] = {
  {   0, "setup" },
  {   1, "configurationChange" },
  {   2, "removal" },
  { 0, NULL }
};

static const per_choice_t Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency_choice[] = {
  {   0, &hf_rnsap_setup_02      , ASN1_EXTENSION_ROOT    , dissect_rnsap_Additional_EDCH_Setup_Info },
  {   1, &hf_rnsap_configurationChange_01, ASN1_EXTENSION_ROOT    , dissect_rnsap_Additional_EDCH_Cell_Information_ConfigurationChange_List },
  {   2, &hf_rnsap_removal       , ASN1_EXTENSION_ROOT    , dissect_rnsap_Additional_EDCH_Cell_Information_Removal_List },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency, Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_UL_CLTD_Activation_Information_vals[] = {
  {   0, "activated" },
  {   1, "de-activated" },
  { 0, NULL }
};


static int
dissect_rnsap_UL_CLTD_Activation_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t UL_CLTD_Information_sequence[] = {
  { &hf_rnsap_sDPCCH_PowerOffsetInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SDPCCH_PowerOffsetInformation },
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_C_ID },
  { &hf_rnsap_uL_CLTD_Activation_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_CLTD_Activation_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CLTD_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CLTD_Information, UL_CLTD_Information_sequence);

  return offset;
}


static const per_sequence_t UL_CLTD_Information_To_Modify_sequence[] = {
  { &hf_rnsap_sDPCCH_PowerOffsetInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SDPCCH_PowerOffsetInformation },
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_C_ID },
  { &hf_rnsap_uL_CLTD_Activation_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_CLTD_Activation_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CLTD_Information_To_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CLTD_Information_To_Modify, UL_CLTD_Information_To_Modify_sequence);

  return offset;
}


static const value_string rnsap_UL_CLTD_Information_Removal_vals[] = {
  {   0, "remove" },
  { 0, NULL }
};


static int
dissect_rnsap_UL_CLTD_Information_Removal(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD_vals[] = {
  {   0, "setup" },
  {   1, "configurationChange" },
  {   2, "removal" },
  { 0, NULL }
};

static const per_choice_t Setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD_choice[] = {
  {   0, &hf_rnsap_setup_04      , ASN1_EXTENSION_ROOT    , dissect_rnsap_UL_CLTD_Information },
  {   1, &hf_rnsap_configurationChange_03, ASN1_EXTENSION_ROOT    , dissect_rnsap_UL_CLTD_Information_To_Modify },
  {   2, &hf_rnsap_removal_02    , ASN1_EXTENSION_ROOT    , dissect_rnsap_UL_CLTD_Information_Removal },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD, Setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t UL_DPCCH2_Information_sequence[] = {
  { &hf_rnsap_f_DPCH_Info   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_F_DPCH_Info },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCCH2_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCCH2_Information, UL_DPCCH2_Information_sequence);

  return offset;
}


static const per_sequence_t UL_DPCCH2_Information_To_Modify_sequence[] = {
  { &hf_rnsap_f_DPCH_Info_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_F_DPCH_Info_To_Modify },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCCH2_Information_To_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCCH2_Information_To_Modify, UL_DPCCH2_Information_To_Modify_sequence);

  return offset;
}


static const value_string rnsap_UL_DPCCH2_Information_Removal_vals[] = {
  {   0, "remove" },
  { 0, NULL }
};


static int
dissect_rnsap_UL_DPCCH2_Information_Removal(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2_vals[] = {
  {   0, "setup" },
  {   1, "configurationChange" },
  {   2, "removal" },
  { 0, NULL }
};

static const per_choice_t Setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2_choice[] = {
  {   0, &hf_rnsap_setup_07      , ASN1_EXTENSION_ROOT    , dissect_rnsap_UL_DPCCH2_Information },
  {   1, &hf_rnsap_configurationChange_06, ASN1_EXTENSION_ROOT    , dissect_rnsap_UL_DPCCH2_Information_To_Modify },
  {   2, &hf_rnsap_removal_05    , ASN1_EXTENSION_ROOT    , dissect_rnsap_UL_DPCCH2_Information_Removal },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2, Setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_SFNSFNChangeLimit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}


static const per_sequence_t SFNSFNMeasurementThresholdInformation_sequence[] = {
  { &hf_rnsap_sFNSFNChangeLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SFNSFNChangeLimit },
  { &hf_rnsap_predictedSFNSFNDeviationLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PredictedSFNSFNDeviationLimit },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SFNSFNMeasurementThresholdInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SFNSFNMeasurementThresholdInformation, SFNSFNMeasurementThresholdInformation_sequence);

  return offset;
}



static int
dissect_rnsap_Single_Stream_MIMO_ActivationIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_Single_Stream_MIMO_Mode_Indicator_vals[] = {
  {   0, "activate" },
  {   1, "deactivate" },
  { 0, NULL }
};


static int
dissect_rnsap_Single_Stream_MIMO_Mode_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_SixteenQAM_UL_Operation_Indicator_vals[] = {
  {   0, "activate" },
  {   1, "deactivate" },
  { 0, NULL }
};


static int
dissect_rnsap_SixteenQAM_UL_Operation_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_SixtyfourQAM_UL_Operation_Indicator_vals[] = {
  {   0, "activate" },
  {   1, "deactivate" },
  { 0, NULL }
};


static int
dissect_rnsap_SixtyfourQAM_UL_Operation_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_SourceID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_SAI(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_SpecialBurstScheduling(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_SpeechVersion(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     4, 4, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const value_string rnsap_T_sRNTI_BitMaskIndex_vals[] = {
  {   0, "b1" },
  {   1, "b2" },
  {   2, "b3" },
  {   3, "b4" },
  {   4, "b5" },
  {   5, "b6" },
  {   6, "b7" },
  {   7, "b8" },
  {   8, "b9" },
  {   9, "b10" },
  {  10, "b11" },
  {  11, "b12" },
  {  12, "b13" },
  {  13, "b14" },
  {  14, "b15" },
  {  15, "b16" },
  {  16, "b17" },
  {  17, "b18" },
  {  18, "b19" },
  { 0, NULL }
};


static int
dissect_rnsap_T_sRNTI_BitMaskIndex(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     19, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t S_RNTI_Group_sequence[] = {
  { &hf_rnsap_sRNTI         , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_S_RNTI },
  { &hf_rnsap_sRNTI_BitMaskIndex, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_T_sRNTI_BitMaskIndex },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_S_RNTI_Group(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_S_RNTI_Group, S_RNTI_Group_sequence);

  return offset;
}



static int
dissect_rnsap_SRB_Delay(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 7U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_SSDT_SupportIndicator_vals[] = {
  {   0, "not-Used-sSDT-supported" },
  {   1, "sSDT-not-supported" },
  { 0, NULL }
};


static int
dissect_rnsap_SSDT_SupportIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_Support_8PSK_vals[] = {
  {   0, "v8PSK-Supported" },
  { 0, NULL }
};


static int
dissect_rnsap_Support_8PSK(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_Support_PLCCH_vals[] = {
  {   0, "vPLCCH-Supported" },
  { 0, NULL }
};


static int
dissect_rnsap_Support_PLCCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order_vals[] = {
  {   0, "supported" },
  {   1, "not-supported" },
  { 0, NULL }
};


static int
dissect_rnsap_Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_256(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}


static const per_sequence_t SynchronisationConfiguration_sequence[] = {
  { &hf_rnsap_n_INSYNC_IND  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_256 },
  { &hf_rnsap_n_OUTSYNC_IND , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_256 },
  { &hf_rnsap_t_RLFAILURE   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_255 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SynchronisationConfiguration(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SynchronisationConfiguration, SynchronisationConfiguration_sequence);

  return offset;
}


static const value_string rnsap_T_maxSYNC_UL_transmissions_vals[] = {
  {   0, "v1" },
  {   1, "v2" },
  {   2, "v4" },
  {   3, "v8" },
  { 0, NULL }
};


static int
dissect_rnsap_T_maxSYNC_UL_transmissions(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     4, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_3_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 3U, NULL, TRUE);

  return offset;
}


static const per_sequence_t SYNC_UL_ProcParameters_sequence[] = {
  { &hf_rnsap_maxSYNC_UL_transmissions, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_maxSYNC_UL_transmissions },
  { &hf_rnsap_powerRampStep , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_3_ },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SYNC_UL_ProcParameters(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SYNC_UL_ProcParameters, SYNC_UL_ProcParameters_sequence);

  return offset;
}



static int
dissect_rnsap_TAC(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  tvbuff_t *parameter_tvb = NULL;
  offset = dissect_per_octet_string(tvb, offset, actx, tree, -1,
                                       2, 2, FALSE, &parameter_tvb);

  if (parameter_tvb) {
    actx->created_item = proto_tree_add_item(tree, hf_index, parameter_tvb, 0, 2, ENC_BIG_ENDIAN);
  }


  return offset;
}



static int
dissect_rnsap_TargetID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_CGI(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t TDD_DCHs_to_ModifySpecificItem_sequence[] = {
  { &hf_rnsap_dCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_ID },
  { &hf_rnsap_ul_CCTrCH_ID  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_dl_CCTrCH_ID  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_ul_TransportformatSet, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_dl_TransportformatSet, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_frameHandlingPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_FrameHandlingPriority },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_DCHs_to_ModifySpecificItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_DCHs_to_ModifySpecificItem, TDD_DCHs_to_ModifySpecificItem_sequence);

  return offset;
}


static const per_sequence_t TDD_DCHs_to_ModifySpecificInformationList_sequence_of[1] = {
  { &hf_rnsap_TDD_DCHs_to_ModifySpecificInformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DCHs_to_ModifySpecificItem },
};

static int
dissect_rnsap_TDD_DCHs_to_ModifySpecificInformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_DCHs_to_ModifySpecificInformationList, TDD_DCHs_to_ModifySpecificInformationList_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t TDD_DCHs_to_ModifyItem_sequence[] = {
  { &hf_rnsap_ul_FP_Mode    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_FP_Mode },
  { &hf_rnsap_toAWS         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ToAWS },
  { &hf_rnsap_toAWE         , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ToAWE },
  { &hf_rnsap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportBearerRequestIndicator },
  { &hf_rnsap_dCH_SpecificInformationList_03, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DCHs_to_ModifySpecificInformationList },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_DCHs_to_ModifyItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_DCHs_to_ModifyItem, TDD_DCHs_to_ModifyItem_sequence);

  return offset;
}


static const per_sequence_t TDD_DCHs_to_Modify_sequence_of[1] = {
  { &hf_rnsap_TDD_DCHs_to_Modify_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DCHs_to_ModifyItem },
};

static int
dissect_rnsap_TDD_DCHs_to_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_DCHs_to_Modify, TDD_DCHs_to_Modify_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const value_string rnsap_TDD_DPCHOffset_vals[] = {
  {   0, "initialOffset" },
  {   1, "noinitialOffset" },
  { 0, NULL }
};

static const per_choice_t TDD_DPCHOffset_choice[] = {
  {   0, &hf_rnsap_initialOffset , ASN1_NO_EXTENSIONS     , dissect_rnsap_INTEGER_0_255 },
  {   1, &hf_rnsap_noinitialOffset, ASN1_NO_EXTENSIONS     , dissect_rnsap_INTEGER_0_63 },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_TDD_DPCHOffset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_TDD_DPCHOffset, TDD_DPCHOffset_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_TDD_TPC_DownlinkStepSize_vals[] = {
  {   0, "step-size1" },
  {   1, "step-size2" },
  {   2, "step-size3" },
  { 0, NULL }
};


static int
dissect_rnsap_TDD_TPC_DownlinkStepSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t TDD_UL_Code_InformationItem_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_UL_Code_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_UL_Code_InformationItem, TDD_UL_Code_InformationItem_sequence);

  return offset;
}


static const per_sequence_t TDD_UL_Code_Information_sequence_of[1] = {
  { &hf_rnsap_TDD_UL_Code_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_UL_Code_InformationItem },
};

static int
dissect_rnsap_TDD_UL_Code_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_UL_Code_Information, TDD_UL_Code_Information_sequence_of,
                                                  1, maxNrOfDPCHs, FALSE);

  return offset;
}



static int
dissect_rnsap_QPSK_UL_DPCH_TimeSlotFormatTDD_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 69U, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_EightPSK_UL_DPCH_TimeSlotFormatTDD_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 24U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_TDD_UL_DPCH_TimeSlotFormat_LCR_vals[] = {
  {   0, "qPSK" },
  {   1, "eightPSK" },
  { 0, NULL }
};

static const per_choice_t TDD_UL_DPCH_TimeSlotFormat_LCR_choice[] = {
  {   0, &hf_rnsap_qPSK_01       , ASN1_EXTENSION_ROOT    , dissect_rnsap_QPSK_UL_DPCH_TimeSlotFormatTDD_LCR },
  {   1, &hf_rnsap_eightPSK_01   , ASN1_EXTENSION_ROOT    , dissect_rnsap_EightPSK_UL_DPCH_TimeSlotFormatTDD_LCR },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_TDD_UL_DPCH_TimeSlotFormat_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_TDD_UL_DPCH_TimeSlotFormat_LCR, TDD_UL_DPCH_TimeSlotFormat_LCR_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t TDD_UL_Code_LCR_InformationItem_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_tdd_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCodeLCR },
  { &hf_rnsap_tdd_UL_DPCH_TimeSlotFormat_LCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_UL_DPCH_TimeSlotFormat_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_UL_Code_LCR_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_UL_Code_LCR_InformationItem, TDD_UL_Code_LCR_InformationItem_sequence);

  return offset;
}


static const per_sequence_t TDD_UL_Code_LCR_Information_sequence_of[1] = {
  { &hf_rnsap_TDD_UL_Code_LCR_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_UL_Code_LCR_InformationItem },
};

static int
dissect_rnsap_TDD_UL_Code_LCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_UL_Code_LCR_Information, TDD_UL_Code_LCR_Information_sequence_of,
                                                  1, maxNrOfDPCHsLCR, FALSE);

  return offset;
}


static const per_sequence_t TDD_UL_Code_InformationItem768_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_tDD_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_UL_Code_InformationItem768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_UL_Code_InformationItem768, TDD_UL_Code_InformationItem768_sequence);

  return offset;
}


static const per_sequence_t TDD_UL_Code_Information768_sequence_of[1] = {
  { &hf_rnsap_TDD_UL_Code_Information768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_UL_Code_InformationItem768 },
};

static int
dissect_rnsap_TDD_UL_Code_Information768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_UL_Code_Information768, TDD_UL_Code_Information768_sequence_of,
                                                  1, maxNrOfDPCHs768, FALSE);

  return offset;
}


static const value_string rnsap_TFCI_SignallingMode_vals[] = {
  {   0, "normal" },
  {   1, "not-Used-split" },
  { 0, NULL }
};


static int
dissect_rnsap_TFCI_SignallingMode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_TGD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 269U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_TGSN(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 14U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_TimingAdvanceApplied_vals[] = {
  {   0, "yes" },
  {   1, "no" },
  { 0, NULL }
};


static int
dissect_rnsap_TimingAdvanceApplied(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_SynchronisationIndicator_vals[] = {
  {   0, "timingMaintainedSynchronisation" },
  { 0, NULL }
};


static int
dissect_rnsap_SynchronisationIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_TraceDepth_vals[] = {
  {   0, "minimum" },
  {   1, "medium" },
  {   2, "maximum" },
  { 0, NULL }
};


static int
dissect_rnsap_TraceDepth(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_TraceRecordingSessionReference(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_TraceReference(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index,
                                       2, 3, FALSE, NULL);

  return offset;
}


static const value_string rnsap_UL_DL_mode_vals[] = {
  {   0, "ul-only" },
  {   1, "dl-only" },
  {   2, "both-ul-and-dl" },
  { 0, NULL }
};


static int
dissect_rnsap_UL_DL_mode(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_Uplink_Compressed_Mode_Method_vals[] = {
  {   0, "sFdiv2" },
  {   1, "higher-layer-scheduling" },
  { 0, NULL }
};


static int
dissect_rnsap_Uplink_Compressed_Mode_Method(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Transmission_Gap_Pattern_Sequence_Information_item_sequence[] = {
  { &hf_rnsap_tGPSID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TGPSID },
  { &hf_rnsap_tGSN          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TGSN },
  { &hf_rnsap_tGL1          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GapLength },
  { &hf_rnsap_tGL2          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GapLength },
  { &hf_rnsap_tGD           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TGD },
  { &hf_rnsap_tGPL1         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GapDuration },
  { &hf_rnsap_not_to_be_used_1, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GapDuration },
  { &hf_rnsap_uL_DL_mode    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_DL_mode },
  { &hf_rnsap_downlink_Compressed_Mode_Method, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Downlink_Compressed_Mode_Method },
  { &hf_rnsap_uplink_Compressed_Mode_Method, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Uplink_Compressed_Mode_Method },
  { &hf_rnsap_dL_FrameType  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_FrameType },
  { &hf_rnsap_delta_SIR1    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DeltaSIR },
  { &hf_rnsap_delta_SIR_after1, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DeltaSIR },
  { &hf_rnsap_delta_SIR2    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DeltaSIR },
  { &hf_rnsap_delta_SIR_after2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DeltaSIR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Transmission_Gap_Pattern_Sequence_Information_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Transmission_Gap_Pattern_Sequence_Information_item, Transmission_Gap_Pattern_Sequence_Information_item_sequence);

  return offset;
}


static const per_sequence_t Transmission_Gap_Pattern_Sequence_Information_sequence_of[1] = {
  { &hf_rnsap_Transmission_Gap_Pattern_Sequence_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Transmission_Gap_Pattern_Sequence_Information_item },
};

static int
dissect_rnsap_Transmission_Gap_Pattern_Sequence_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Transmission_Gap_Pattern_Sequence_Information, Transmission_Gap_Pattern_Sequence_Information_sequence_of,
                                                  1, maxTGPS, FALSE);

  return offset;
}


static const per_sequence_t Affected_HSDSCH_Serving_Cell_List_sequence_of[1] = {
  { &hf_rnsap_Affected_HSDSCH_Serving_Cell_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
};

static int
dissect_rnsap_Affected_HSDSCH_Serving_Cell_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Affected_HSDSCH_Serving_Cell_List, Affected_HSDSCH_Serving_Cell_List_sequence_of,
                                                  0, maxNrOfHSDSCH, FALSE);

  return offset;
}


static const per_sequence_t Transmission_Mode_Information_List_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_transmissionMode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransmissionMode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Transmission_Mode_Information_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Transmission_Mode_Information_List, Transmission_Mode_Information_List_sequence);

  return offset;
}


static const per_sequence_t Transmission_Mode_Information_sequence_of[1] = {
  { &hf_rnsap_Transmission_Mode_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Transmission_Mode_Information_List },
};

static int
dissect_rnsap_Transmission_Mode_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Transmission_Mode_Information, Transmission_Mode_Information_sequence_of,
                                                  1, maxNrOfFDDNeighboursPerRNC, TRUE);

  return offset;
}


static const value_string rnsap_TransmitDiversityIndicator_vals[] = {
  {   0, "active" },
  {   1, "inactive" },
  { 0, NULL }
};


static int
dissect_rnsap_TransmitDiversityIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_Transmitted_Carrier_Power_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Transmitted_Carrier_Power_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 100U, NULL, FALSE);

  return offset;
}


static const per_sequence_t TUTRANGANSS_sequence[] = {
  { &hf_rnsap_mS            , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_16383 },
  { &hf_rnsap_lS            , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_4294967295 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TUTRANGANSS(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TUTRANGANSS, TUTRANGANSS_sequence);

  return offset;
}


static const per_sequence_t TUTRANGANSSMeasurementThresholdInformation_sequence[] = {
  { &hf_rnsap_tUTRANGANSSChangeLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_1_256 },
  { &hf_rnsap_predictedTUTRANGANSSDeviationLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_1_256 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TUTRANGANSSMeasurementThresholdInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TUTRANGANSSMeasurementThresholdInformation, TUTRANGANSSMeasurementThresholdInformation_sequence);

  return offset;
}



static int
dissect_rnsap_INTEGER_M50_50(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -50, 50U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_50(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 50U, NULL, FALSE);

  return offset;
}


static const per_sequence_t TUTRANGANSSMeasurementValueInformation_sequence[] = {
  { &hf_rnsap_tUTRANGANSS   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TUTRANGANSS },
  { &hf_rnsap_tUTRANGANSSQuality, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_0_255 },
  { &hf_rnsap_tUTRANGANSSDriftRate, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_M50_50 },
  { &hf_rnsap_tUTRANGANSSDriftRateQuality, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_0_50 },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TUTRANGANSSMeasurementValueInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TUTRANGANSSMeasurementValueInformation, TUTRANGANSSMeasurementValueInformation_sequence);

  return offset;
}



static int
dissect_rnsap_TUTRANGPSChangeLimit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 256U, NULL, FALSE);

  return offset;
}


static const per_sequence_t TUTRANGPSMeasurementThresholdInformation_sequence[] = {
  { &hf_rnsap_tUTRANGPSChangeLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TUTRANGPSChangeLimit },
  { &hf_rnsap_predictedTUTRANGPSDeviationLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PredictedTUTRANGPSDeviationLimit },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TUTRANGPSMeasurementThresholdInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TUTRANGPSMeasurementThresholdInformation, TUTRANGPSMeasurementThresholdInformation_sequence);

  return offset;
}



static int
dissect_rnsap_TransportBearerID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4095U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_TransportBearerNotRequestedIndicator_vals[] = {
  {   0, "transport-bearer-shall-not-be-established" },
  {   1, "transport-bearer-may-not-be-established" },
  { 0, NULL }
};


static int
dissect_rnsap_TransportBearerNotRequestedIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_TransportBearerNotSetupIndicator_vals[] = {
  {   0, "transport-bearer-not-setup" },
  { 0, NULL }
};


static int
dissect_rnsap_TransportBearerNotSetupIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_4(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 4U, NULL, FALSE);

  return offset;
}


static const per_sequence_t TPC_slot_position_sequence[] = {
  { &hf_rnsap_slot_position , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_0_4 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TPC_slot_position(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TPC_slot_position, TPC_slot_position_sequence);

  return offset;
}


static const value_string rnsap_TransportFormatManagement_vals[] = {
  {   0, "cell-based" },
  {   1, "ue-based" },
  { 0, NULL }
};


static int
dissect_rnsap_TransportFormatManagement(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_TSN_Length_vals[] = {
  {   0, "tsn-6bits" },
  {   1, "tsn-9bits" },
  { 0, NULL }
};


static int
dissect_rnsap_TSN_Length(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_TSTD_Indicator_vals[] = {
  {   0, "active" },
  {   1, "inactive" },
  { 0, NULL }
};


static int
dissect_rnsap_TSTD_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_TSTD_Support_Indicator_vals[] = {
  {   0, "tSTD-supported" },
  {   1, "tSTD-not-supported" },
  { 0, NULL }
};


static int
dissect_rnsap_TSTD_Support_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_TTI_Update_Ind(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_TTI_Update_Indicator_vals[] = {
  {   0, "tTI-Update-CFN" },
  {   1, "tTI-Update-Ind" },
  { 0, NULL }
};

static const per_choice_t TTI_Update_Indicator_choice[] = {
  {   0, &hf_rnsap_tTI_Update_CFN, ASN1_NO_EXTENSIONS     , dissect_rnsap_CFN },
  {   1, &hf_rnsap_tTI_Update_Ind, ASN1_NO_EXTENSIONS     , dissect_rnsap_TTI_Update_Ind },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_TTI_Update_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_TTI_Update_Indicator, TTI_Update_Indicator_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_TypeOfError_vals[] = {
  {   0, "not-understood" },
  {   1, "missing" },
  { 0, NULL }
};


static int
dissect_rnsap_TypeOfError(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_UE_AggregateMaximumBitRateDownlink(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 1000000000U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_UE_AggregateMaximumBitRateUplink(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 1000000000U, NULL, FALSE);

  return offset;
}


static const per_sequence_t UE_AggregateMaximumBitRate_sequence[] = {
  { &hf_rnsap_uE_AggregateMaximumBitRateDownlink, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_AggregateMaximumBitRateDownlink },
  { &hf_rnsap_uE_AggregateMaximumBitRateUplink, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UE_AggregateMaximumBitRateUplink },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UE_AggregateMaximumBitRate(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UE_AggregateMaximumBitRate, UE_AggregateMaximumBitRate_sequence);

  return offset;
}



static int
dissect_rnsap_UE_AggregateMaximumBitRate_Enforcement_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_UE_Support_of_non_rectangular_resource_allocation_vals[] = {
  {   0, "support" },
  { 0, NULL }
};


static int
dissect_rnsap_UE_Support_of_non_rectangular_resource_allocation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_UE_TS0_CapabilityLCR_vals[] = {
  {   0, "tS0-Capable" },
  {   1, "tS0-Not-Capable" },
  { 0, NULL }
};


static int
dissect_rnsap_UE_TS0_CapabilityLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_Radio_Frequency_Band_vals[] = {
  {   0, "a" },
  {   1, "b" },
  {   2, "c" },
  {   3, "d" },
  {   4, "e" },
  {   5, "f" },
  {   6, "g" },
  {   7, "h" },
  {   8, "i" },
  {   9, "j" },
  {  10, "k" },
  {  11, "l" },
  {  12, "m" },
  {  13, "n" },
  {  14, "o" },
  {  15, "p" },
  { 0, NULL }
};


static int
dissect_rnsap_Radio_Frequency_Band(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     16, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t Radio_Frequency_BandItem_sequence[] = {
  { &hf_rnsap_radio_Frequency_Band, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Radio_Frequency_Band },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Radio_Frequency_BandItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Radio_Frequency_BandItem, Radio_Frequency_BandItem_sequence);

  return offset;
}


static const per_sequence_t UE_RF_Band_CapabilityLCR_sequence_of[1] = {
  { &hf_rnsap_UE_RF_Band_CapabilityLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Radio_Frequency_BandItem },
};

static int
dissect_rnsap_UE_RF_Band_CapabilityLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UE_RF_Band_CapabilityLCR, UE_RF_Band_CapabilityLCR_sequence_of,
                                                  1, maxFreqBandsTDD, FALSE);

  return offset;
}



static int
dissect_rnsap_LCRTDD_HSDSCH_Physical_Layer_Category(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 64U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_UEIdentity_vals[] = {
  {   0, "imsi" },
  {   1, "imei" },
  {   2, "imeisv" },
  { 0, NULL }
};

static const per_choice_t UEIdentity_choice[] = {
  {   0, &hf_rnsap_imsi          , ASN1_EXTENSION_ROOT    , dissect_rnsap_IMSI },
  {   1, &hf_rnsap_imei          , ASN1_EXTENSION_ROOT    , dissect_rnsap_IMEI },
  {   2, &hf_rnsap_imeisv        , ASN1_EXTENSION_ROOT    , dissect_rnsap_IMEISV },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_UEIdentity(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_UEIdentity, UEIdentity_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_UEMeasurementHysteresisTime(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 15U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_UEMeasurementParameterModAllow_vals[] = {
  {   0, "parameterModificationAllowed" },
  { 0, NULL }
};


static int
dissect_rnsap_UEMeasurementParameterModAllow(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_UEMeasurementReportCharacteristicsPeriodicAmountofReporting_vals[] = {
  {   0, "r1" },
  {   1, "r2" },
  {   2, "r4" },
  {   3, "r8" },
  {   4, "r16" },
  {   5, "r32" },
  {   6, "r64" },
  {   7, "rInfinity" },
  { 0, NULL }
};


static int
dissect_rnsap_UEMeasurementReportCharacteristicsPeriodicAmountofReporting(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     8, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_UEMeasurementReportCharacteristicsPeriodicReportingInterval_vals[] = {
  {   0, "r250" },
  {   1, "r500" },
  {   2, "r1000" },
  {   3, "r2000" },
  {   4, "r3000" },
  {   5, "r4000" },
  {   6, "r6000" },
  {   7, "r8000" },
  {   8, "r12000" },
  {   9, "r16000" },
  {  10, "r20000" },
  {  11, "r24000" },
  {  12, "r28000" },
  {  13, "r32000" },
  {  14, "r64000" },
  { 0, NULL }
};


static int
dissect_rnsap_UEMeasurementReportCharacteristicsPeriodicReportingInterval(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     15, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t UEMeasurementReportCharacteristicsPeriodic_sequence[] = {
  { &hf_rnsap_amountofReporting, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementReportCharacteristicsPeriodicAmountofReporting },
  { &hf_rnsap_reportingInterval, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementReportCharacteristicsPeriodicReportingInterval },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementReportCharacteristicsPeriodic(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementReportCharacteristicsPeriodic, UEMeasurementReportCharacteristicsPeriodic_sequence);

  return offset;
}



static int
dissect_rnsap_UEMeasurementThresholdDLTimeslotISCP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -115, -25, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_UEMeasurementThresholdUETransmitPower(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -50, 33U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_UEMeasurementThreshold_Extension(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_UEMeasurementThreshold_vals[] = {
  {   0, "timeslotISCP" },
  {   1, "uETransmitPower" },
  {   2, "extension-UEMeasurementThreshold" },
  { 0, NULL }
};

static const per_choice_t UEMeasurementThreshold_choice[] = {
  {   0, &hf_rnsap_timeslotISCP  , ASN1_EXTENSION_ROOT    , dissect_rnsap_UEMeasurementThresholdDLTimeslotISCP },
  {   1, &hf_rnsap_uETransmitPower, ASN1_EXTENSION_ROOT    , dissect_rnsap_UEMeasurementThresholdUETransmitPower },
  {   2, &hf_rnsap_extension_UEMeasurementThreshold, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_UEMeasurementThreshold_Extension },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementThreshold(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_UEMeasurementThreshold, UEMeasurementThreshold_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_UEMeasurementTimeToTrigger_vals[] = {
  {   0, "r0" },
  {   1, "r10" },
  {   2, "r20" },
  {   3, "r40" },
  {   4, "r60" },
  {   5, "r80" },
  {   6, "r100" },
  {   7, "r120" },
  {   8, "r160" },
  {   9, "r200" },
  {  10, "r240" },
  {  11, "r320" },
  {  12, "r640" },
  {  13, "r1280" },
  {  14, "r2560" },
  {  15, "r5000" },
  { 0, NULL }
};


static int
dissect_rnsap_UEMeasurementTimeToTrigger(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     16, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t UEMeasurementReportCharacteristicsEvent1h_sequence[] = {
  { &hf_rnsap_uEMeasurementTreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementThreshold },
  { &hf_rnsap_uEMeasurementTimeToTrigger, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementTimeToTrigger },
  { &hf_rnsap_uEMeasurementHysteresisTime, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementHysteresisTime },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementReportCharacteristicsEvent1h(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementReportCharacteristicsEvent1h, UEMeasurementReportCharacteristicsEvent1h_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementReportCharacteristicsEvent1i_sequence[] = {
  { &hf_rnsap_uEMeasurementTreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementThreshold },
  { &hf_rnsap_uEMeasurementTimeToTrigger, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementTimeToTrigger },
  { &hf_rnsap_uEMeasurementHysteresisTime, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementHysteresisTime },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementReportCharacteristicsEvent1i(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementReportCharacteristicsEvent1i, UEMeasurementReportCharacteristicsEvent1i_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementReportCharacteristicsEvent6a_sequence[] = {
  { &hf_rnsap_uEMeasurementTreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementThreshold },
  { &hf_rnsap_uEMeasurementTimeToTrigger, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementTimeToTrigger },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementReportCharacteristicsEvent6a(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementReportCharacteristicsEvent6a, UEMeasurementReportCharacteristicsEvent6a_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementReportCharacteristicsEvent6b_sequence[] = {
  { &hf_rnsap_uEMeasurementTreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementThreshold },
  { &hf_rnsap_uEMeasurementTimeToTrigger, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementTimeToTrigger },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementReportCharacteristicsEvent6b(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementReportCharacteristicsEvent6b, UEMeasurementReportCharacteristicsEvent6b_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementReportCharacteristicsEvent6c_sequence[] = {
  { &hf_rnsap_uEMeasurementTimeToTrigger, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementTimeToTrigger },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementReportCharacteristicsEvent6c(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementReportCharacteristicsEvent6c, UEMeasurementReportCharacteristicsEvent6c_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementReportCharacteristicsEvent6d_sequence[] = {
  { &hf_rnsap_uEMeasurementTimeToTrigger, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementTimeToTrigger },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementReportCharacteristicsEvent6d(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementReportCharacteristicsEvent6d, UEMeasurementReportCharacteristicsEvent6d_sequence);

  return offset;
}



static int
dissect_rnsap_UEMeasurementReportCharacteristics_Extension(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_UEMeasurementReportCharacteristics_vals[] = {
  {   0, "periodic" },
  {   1, "event1h" },
  {   2, "event1i" },
  {   3, "event6a" },
  {   4, "event6b" },
  {   5, "event6c" },
  {   6, "event6d" },
  {   7, "extension-ReportCharacteristics" },
  { 0, NULL }
};

static const per_choice_t UEMeasurementReportCharacteristics_choice[] = {
  {   0, &hf_rnsap_periodic_03   , ASN1_EXTENSION_ROOT    , dissect_rnsap_UEMeasurementReportCharacteristicsPeriodic },
  {   1, &hf_rnsap_event1h       , ASN1_EXTENSION_ROOT    , dissect_rnsap_UEMeasurementReportCharacteristicsEvent1h },
  {   2, &hf_rnsap_event1i       , ASN1_EXTENSION_ROOT    , dissect_rnsap_UEMeasurementReportCharacteristicsEvent1i },
  {   3, &hf_rnsap_event6a       , ASN1_EXTENSION_ROOT    , dissect_rnsap_UEMeasurementReportCharacteristicsEvent6a },
  {   4, &hf_rnsap_event6b       , ASN1_EXTENSION_ROOT    , dissect_rnsap_UEMeasurementReportCharacteristicsEvent6b },
  {   5, &hf_rnsap_event6c       , ASN1_EXTENSION_ROOT    , dissect_rnsap_UEMeasurementReportCharacteristicsEvent6c },
  {   6, &hf_rnsap_event6d       , ASN1_EXTENSION_ROOT    , dissect_rnsap_UEMeasurementReportCharacteristicsEvent6d },
  {   7, &hf_rnsap_extension_ReportCharacteristics_01, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_UEMeasurementReportCharacteristics_Extension },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementReportCharacteristics(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_UEMeasurementReportCharacteristics, UEMeasurementReportCharacteristics_choice,
                                 NULL);

  return offset;
}


static const value_string rnsap_UEMeasurementTimeslotInfoHCRBurstType_vals[] = {
  {   0, "type1" },
  {   1, "type2" },
  {   2, "type3" },
  { 0, NULL }
};


static int
dissect_rnsap_UEMeasurementTimeslotInfoHCRBurstType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t UEMeasurementTimeslotInfoHCR_IEs_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_burstType     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementTimeslotInfoHCRBurstType },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementTimeslotInfoHCR_IEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementTimeslotInfoHCR_IEs, UEMeasurementTimeslotInfoHCR_IEs_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementTimeslotInfoHCR_sequence_of[1] = {
  { &hf_rnsap_UEMeasurementTimeslotInfoHCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementTimeslotInfoHCR_IEs },
};

static int
dissect_rnsap_UEMeasurementTimeslotInfoHCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UEMeasurementTimeslotInfoHCR, UEMeasurementTimeslotInfoHCR_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t UEMeasurementTimeslotInfoLCR_IEs_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementTimeslotInfoLCR_IEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementTimeslotInfoLCR_IEs, UEMeasurementTimeslotInfoLCR_IEs_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementTimeslotInfoLCR_sequence_of[1] = {
  { &hf_rnsap_UEMeasurementTimeslotInfoLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementTimeslotInfoLCR_IEs },
};

static int
dissect_rnsap_UEMeasurementTimeslotInfoLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UEMeasurementTimeslotInfoLCR, UEMeasurementTimeslotInfoLCR_sequence_of,
                                                  1, maxNrOfTsLCR, FALSE);

  return offset;
}


static const value_string rnsap_UEMeasurementTimeslotInfo768BurstType_vals[] = {
  {   0, "type1" },
  {   1, "type2" },
  {   2, "type3" },
  { 0, NULL }
};


static int
dissect_rnsap_UEMeasurementTimeslotInfo768BurstType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t UEMeasurementTimeslotInfo768_IEs_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_burstType_01  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementTimeslotInfo768BurstType },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementTimeslotInfo768_IEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementTimeslotInfo768_IEs, UEMeasurementTimeslotInfo768_IEs_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementTimeslotInfo768_sequence_of[1] = {
  { &hf_rnsap_UEMeasurementTimeslotInfo768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementTimeslotInfo768_IEs },
};

static int
dissect_rnsap_UEMeasurementTimeslotInfo768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UEMeasurementTimeslotInfo768, UEMeasurementTimeslotInfo768_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const value_string rnsap_UEMeasurementType_vals[] = {
  {   0, "primary-CCPCH-RSCP" },
  {   1, "dL-Timeslot-ISCP" },
  {   2, "uE-Transmitted-power" },
  { 0, NULL }
};


static int
dissect_rnsap_UEMeasurementType(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     3, NULL, TRUE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_INTEGER_0_104(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 104U, NULL, FALSE);

  return offset;
}


static const per_sequence_t UEMeasurementValueTransmittedPowerListHCR_IEs_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_uETransmitPower_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_104 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementValueTransmittedPowerListHCR_IEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementValueTransmittedPowerListHCR_IEs, UEMeasurementValueTransmittedPowerListHCR_IEs_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementValueTransmittedPowerListHCR_sequence_of[1] = {
  { &hf_rnsap_UEMeasurementValueTransmittedPowerListHCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementValueTransmittedPowerListHCR_IEs },
};

static int
dissect_rnsap_UEMeasurementValueTransmittedPowerListHCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UEMeasurementValueTransmittedPowerListHCR, UEMeasurementValueTransmittedPowerListHCR_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t UEMeasurementValueTransmittedPowerListLCR_IEs_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_uETransmitPower_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_104 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementValueTransmittedPowerListLCR_IEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementValueTransmittedPowerListLCR_IEs, UEMeasurementValueTransmittedPowerListLCR_IEs_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementValueTransmittedPowerListLCR_sequence_of[1] = {
  { &hf_rnsap_UEMeasurementValueTransmittedPowerListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementValueTransmittedPowerListLCR_IEs },
};

static int
dissect_rnsap_UEMeasurementValueTransmittedPowerListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UEMeasurementValueTransmittedPowerListLCR, UEMeasurementValueTransmittedPowerListLCR_sequence_of,
                                                  1, maxNrOfTsLCR, FALSE);

  return offset;
}


static const per_sequence_t UE_MeasurementValue_UE_Transmitted_Power_sequence[] = {
  { &hf_rnsap_uEMeasurementTransmittedPowerListHCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UEMeasurementValueTransmittedPowerListHCR },
  { &hf_rnsap_uEMeasurementTransmittedPowerListLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UEMeasurementValueTransmittedPowerListLCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UE_MeasurementValue_UE_Transmitted_Power(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UE_MeasurementValue_UE_Transmitted_Power, UE_MeasurementValue_UE_Transmitted_Power_sequence);

  return offset;
}


static const per_sequence_t UE_MeasurementValue_Primary_CCPCH_RSCP_sequence[] = {
  { &hf_rnsap_primaryCCPCH_RSCP, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PrimaryCCPCH_RSCP },
  { &hf_rnsap_primaryCCPCH_RSCP_Delta, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PrimaryCCPCH_RSCP_Delta },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UE_MeasurementValue_Primary_CCPCH_RSCP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UE_MeasurementValue_Primary_CCPCH_RSCP, UE_MeasurementValue_Primary_CCPCH_RSCP_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementValueTimeslotISCPListHCR_IEs_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_dL_TimeslotISCP, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TimeslotISCP },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementValueTimeslotISCPListHCR_IEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementValueTimeslotISCPListHCR_IEs, UEMeasurementValueTimeslotISCPListHCR_IEs_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementValueTimeslotISCPListHCR_sequence_of[1] = {
  { &hf_rnsap_UEMeasurementValueTimeslotISCPListHCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementValueTimeslotISCPListHCR_IEs },
};

static int
dissect_rnsap_UEMeasurementValueTimeslotISCPListHCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UEMeasurementValueTimeslotISCPListHCR, UEMeasurementValueTimeslotISCPListHCR_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t UEMeasurementValueTimeslotISCPListLCR_IEs_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_dL_TimeslotISCP, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TimeslotISCP },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementValueTimeslotISCPListLCR_IEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementValueTimeslotISCPListLCR_IEs, UEMeasurementValueTimeslotISCPListLCR_IEs_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementValueTimeslotISCPListLCR_sequence_of[1] = {
  { &hf_rnsap_UEMeasurementValueTimeslotISCPListLCR_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementValueTimeslotISCPListLCR_IEs },
};

static int
dissect_rnsap_UEMeasurementValueTimeslotISCPListLCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UEMeasurementValueTimeslotISCPListLCR, UEMeasurementValueTimeslotISCPListLCR_sequence_of,
                                                  1, maxNrOfTsLCR, FALSE);

  return offset;
}


static const per_sequence_t UE_MeasurementValue_DL_Timeslot_ISCP_sequence[] = {
  { &hf_rnsap_uEMeasurementTimeslotISCPListHCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UEMeasurementValueTimeslotISCPListHCR },
  { &hf_rnsap_uEMeasurementTimeslotISCPListLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UEMeasurementValueTimeslotISCPListLCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UE_MeasurementValue_DL_Timeslot_ISCP(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UE_MeasurementValue_DL_Timeslot_ISCP, UE_MeasurementValue_DL_Timeslot_ISCP_sequence);

  return offset;
}



static int
dissect_rnsap_UEMeasurementValue_Extension(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_UEMeasurementValue_vals[] = {
  {   0, "uE-Transmitted-Power" },
  {   1, "primary-CCPCH-RSCP" },
  {   2, "dL-Timeslot-ISCP" },
  {   3, "extension-UEMeasurementValue" },
  { 0, NULL }
};

static const per_choice_t UEMeasurementValue_choice[] = {
  {   0, &hf_rnsap_uE_Transmitted_Power, ASN1_EXTENSION_ROOT    , dissect_rnsap_UE_MeasurementValue_UE_Transmitted_Power },
  {   1, &hf_rnsap_primary_CCPCH_RSCP, ASN1_EXTENSION_ROOT    , dissect_rnsap_UE_MeasurementValue_Primary_CCPCH_RSCP },
  {   2, &hf_rnsap_dL_Timeslot_ISCP, ASN1_EXTENSION_ROOT    , dissect_rnsap_UE_MeasurementValue_DL_Timeslot_ISCP },
  {   3, &hf_rnsap_extension_UEMeasurementValue, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_UEMeasurementValue_Extension },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_UEMeasurementValue, UEMeasurementValue_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t UEMeasurementValueTransmittedPowerList768_IEs_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_uETransmitPower_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_104 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementValueTransmittedPowerList768_IEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementValueTransmittedPowerList768_IEs, UEMeasurementValueTransmittedPowerList768_IEs_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementValueTransmittedPowerList768_sequence_of[1] = {
  { &hf_rnsap_UEMeasurementValueTransmittedPowerList768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementValueTransmittedPowerList768_IEs },
};

static int
dissect_rnsap_UEMeasurementValueTransmittedPowerList768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UEMeasurementValueTransmittedPowerList768, UEMeasurementValueTransmittedPowerList768_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t UEMeasurementValueTimeslotISCPList768_IEs_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_dL_TimeslotISCP, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TimeslotISCP },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementValueTimeslotISCPList768_IEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementValueTimeslotISCPList768_IEs, UEMeasurementValueTimeslotISCPList768_IEs_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementValueTimeslotISCPList768_sequence_of[1] = {
  { &hf_rnsap_UEMeasurementValueTimeslotISCPList768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementValueTimeslotISCPList768_IEs },
};

static int
dissect_rnsap_UEMeasurementValueTimeslotISCPList768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UEMeasurementValueTimeslotISCPList768, UEMeasurementValueTimeslotISCPList768_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t UEMeasurementValueInformationAvailable_sequence[] = {
  { &hf_rnsap_uEmeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UEMeasurementValue },
  { &hf_rnsap_ie_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementValueInformationAvailable(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementValueInformationAvailable, UEMeasurementValueInformationAvailable_sequence);

  return offset;
}



static int
dissect_rnsap_UEMeasurementValueInformationnotAvailable(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_UEMeasurementValueInformation_vals[] = {
  {   0, "measurementAvailable" },
  {   1, "measurementnotAvailable" },
  { 0, NULL }
};

static const per_choice_t UEMeasurementValueInformation_choice[] = {
  {   0, &hf_rnsap_measurementAvailable_02, ASN1_NO_EXTENSIONS     , dissect_rnsap_UEMeasurementValueInformationAvailable },
  {   1, &hf_rnsap_measurementnotAvailable_02, ASN1_NO_EXTENSIONS     , dissect_rnsap_UEMeasurementValueInformationnotAvailable },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementValueInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_UEMeasurementValueInformation, UEMeasurementValueInformation_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_UPHFiltering_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 32U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_Extension_UE_Measurement_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_UE_Measurement_Value_vals[] = {
  {   0, "uPHFiltering-Value" },
  {   1, "extension-UE-Measurement-Value" },
  { 0, NULL }
};

static const per_choice_t UE_Measurement_Value_choice[] = {
  {   0, &hf_rnsap_uPHFiltering_Value, ASN1_NO_EXTENSIONS     , dissect_rnsap_UPHFiltering_Value },
  {   1, &hf_rnsap_extension_UE_Measurement_Value, ASN1_NO_EXTENSIONS     , dissect_rnsap_Extension_UE_Measurement_Value },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_UE_Measurement_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_UE_Measurement_Value, UE_Measurement_Value_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_UE_SupportIndicatorExtension(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_bit_string(tvb, offset, actx, tree, hf_index,
                                     32, 32, FALSE, NULL, 0, NULL, NULL);

  return offset;
}


static const per_sequence_t Cell_Fach_Pch_State_sequence[] = {
  { &hf_rnsap_d_RNTI        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_D_RNTI },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Cell_Fach_Pch_State(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Cell_Fach_Pch_State, Cell_Fach_Pch_State_sequence);

  return offset;
}



static int
dissect_rnsap_URA_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 65535U, NULL, FALSE);

  return offset;
}


static const per_sequence_t Ura_Pch_State_sequence[] = {
  { &hf_rnsap_srnc_id       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RNC_ID },
  { &hf_rnsap_ura_id        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_URA_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Ura_Pch_State(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Ura_Pch_State, Ura_Pch_State_sequence);

  return offset;
}


static const value_string rnsap_UE_State_vals[] = {
  {   0, "cell-fach-pch" },
  {   1, "ura-pch" },
  { 0, NULL }
};

static const per_choice_t UE_State_choice[] = {
  {   0, &hf_rnsap_cell_fach_pch , ASN1_EXTENSION_ROOT    , dissect_rnsap_Cell_Fach_Pch_State },
  {   1, &hf_rnsap_ura_pch       , ASN1_EXTENSION_ROOT    , dissect_rnsap_Ura_Pch_State },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_UE_State(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_UE_State, UE_State_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_UE_transmission_power_headroom_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 31U, NULL, FALSE);

  return offset;
}


static const value_string rnsap_UL_DPDCHIndicatorEDCH_vals[] = {
  {   0, "uL-DPDCH-present" },
  {   1, "uL-DPDCH-not-present" },
  { 0, NULL }
};


static int
dissect_rnsap_UL_DPDCHIndicatorEDCH(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationItem_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_uL_Code_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_UL_Code_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_Timeslot_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_Timeslot_InformationItem, UL_Timeslot_InformationItem_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_Information_sequence_of[1] = {
  { &hf_rnsap_UL_Timeslot_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Timeslot_InformationItem },
};

static int
dissect_rnsap_UL_Timeslot_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_Timeslot_Information, UL_Timeslot_Information_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t UL_TimeslotLCR_InformationItem_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_uL_Code_LCR_InformationList, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_UL_Code_LCR_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_TimeslotLCR_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_TimeslotLCR_InformationItem, UL_TimeslotLCR_InformationItem_sequence);

  return offset;
}


static const per_sequence_t UL_TimeslotLCR_Information_sequence_of[1] = {
  { &hf_rnsap_UL_TimeslotLCR_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeslotLCR_InformationItem },
};

static int
dissect_rnsap_UL_TimeslotLCR_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_TimeslotLCR_Information, UL_TimeslotLCR_Information_sequence_of,
                                                  1, maxNrOfULTsLCR, FALSE);

  return offset;
}


static const per_sequence_t PLCCHinformation_sequence[] = {
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_sequenceNumber, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PLCCHsequenceNumber },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PLCCHinformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PLCCHinformation, PLCCHinformation_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationItem768_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_uL_Code_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_UL_Code_Information768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_Timeslot_InformationItem768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_Timeslot_InformationItem768, UL_Timeslot_InformationItem768_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_Information768_sequence_of[1] = {
  { &hf_rnsap_UL_Timeslot_Information768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Timeslot_InformationItem768 },
};

static int
dissect_rnsap_UL_Timeslot_Information768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_Timeslot_Information768, UL_Timeslot_Information768_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t UL_TimeSlot_ISCP_InfoItem_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_uL_TimeslotISCP, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeslotISCP },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_TimeSlot_ISCP_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_TimeSlot_ISCP_InfoItem, UL_TimeSlot_ISCP_InfoItem_sequence);

  return offset;
}


static const per_sequence_t UL_TimeSlot_ISCP_Info_sequence_of[1] = {
  { &hf_rnsap_UL_TimeSlot_ISCP_Info_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeSlot_ISCP_InfoItem },
};

static int
dissect_rnsap_UL_TimeSlot_ISCP_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_TimeSlot_ISCP_Info, UL_TimeSlot_ISCP_Info_sequence_of,
                                                  1, maxNrOfULTs, FALSE);

  return offset;
}



static int
dissect_rnsap_UL_Timeslot_ISCP_Value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_UL_TimeslotISCP(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t UL_TimeSlot_ISCP_LCR_InfoItem_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_iSCP          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Timeslot_ISCP_Value },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_TimeSlot_ISCP_LCR_InfoItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_TimeSlot_ISCP_LCR_InfoItem, UL_TimeSlot_ISCP_LCR_InfoItem_sequence);

  return offset;
}


static const per_sequence_t UL_TimeSlot_ISCP_LCR_Info_sequence_of[1] = {
  { &hf_rnsap_UL_TimeSlot_ISCP_LCR_Info_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeSlot_ISCP_LCR_InfoItem },
};

static int
dissect_rnsap_UL_TimeSlot_ISCP_LCR_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_TimeSlot_ISCP_LCR_Info, UL_TimeSlot_ISCP_LCR_Info_sequence_of,
                                                  1, maxNrOfULTsLCR, FALSE);

  return offset;
}



static int
dissect_rnsap_UL_Timeslot_ISCP_Value_IncrDecrThres(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 126U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_INTEGER_M120_M58_(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            -120, -58, NULL, TRUE);

  return offset;
}



static int
dissect_rnsap_INTEGER_1_32(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 32U, NULL, FALSE);

  return offset;
}


static const per_sequence_t UL_TimingAdvanceCtrl_LCR_sequence[] = {
  { &hf_rnsap_sync_UL_codes_bitmap, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_BIT_STRING_SIZE_8 },
  { &hf_rnsap_fPACH_info    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FPACH_Information },
  { &hf_rnsap_prxUpPCHdes   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_M120_M58_ },
  { &hf_rnsap_syncUL_procParameter, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SYNC_UL_ProcParameters },
  { &hf_rnsap_mMax          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_1_32 },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_TimingAdvanceCtrl_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_TimingAdvanceCtrl_LCR, UL_TimingAdvanceCtrl_LCR_sequence);

  return offset;
}



static int
dissect_rnsap_UL_DPCCH_SlotFormat(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 5U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_UL_PhysCH_SF_Variation_vals[] = {
  {   0, "sf-variation-supported" },
  {   1, "sf-variation-not-supported" },
  { 0, NULL }
};


static int
dissect_rnsap_UL_PhysCH_SF_Variation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}



static int
dissect_rnsap_UL_Synchronisation_StepSize(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 8U, NULL, FALSE);

  return offset;
}



static int
dissect_rnsap_UL_Synchronisation_Frequency(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            1U, 8U, NULL, FALSE);

  return offset;
}


static const per_sequence_t UL_Synchronisation_Parameters_LCR_sequence[] = {
  { &hf_rnsap_uL_Synchronisation_StepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Synchronisation_StepSize },
  { &hf_rnsap_uL_Synchronisation_Frequency, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Synchronisation_Frequency },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_Synchronisation_Parameters_LCR(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_Synchronisation_Parameters_LCR, UL_Synchronisation_Parameters_LCR_sequence);

  return offset;
}



static int
dissect_rnsap_UpPTSInterferenceValue(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 127U, NULL, TRUE);

  return offset;
}


static const value_string rnsap_Unidirectional_DCH_Indicator_vals[] = {
  {   0, "downlink-DCH-only" },
  {   1, "uplink-DCH-only" },
  { 0, NULL }
};


static int
dissect_rnsap_Unidirectional_DCH_Indicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t RNCsWithCellsInTheAccessedURA_Item_sequence[] = {
  { &hf_rnsap_rNC_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RNC_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RNCsWithCellsInTheAccessedURA_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RNCsWithCellsInTheAccessedURA_Item, RNCsWithCellsInTheAccessedURA_Item_sequence);

  return offset;
}


static const per_sequence_t RNCsWithCellsInTheAccessedURA_List_sequence_of[1] = {
  { &hf_rnsap_RNCsWithCellsInTheAccessedURA_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_RNCsWithCellsInTheAccessedURA_Item },
};

static int
dissect_rnsap_RNCsWithCellsInTheAccessedURA_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RNCsWithCellsInTheAccessedURA_List, RNCsWithCellsInTheAccessedURA_List_sequence_of,
                                                  1, maxRNCinURA_1, FALSE);

  return offset;
}


static const per_sequence_t URA_Information_sequence[] = {
  { &hf_rnsap_uRA_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_URA_ID },
  { &hf_rnsap_multipleURAsIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MultipleURAsIndicator },
  { &hf_rnsap_rNCsWithCellsInTheAccessedURA_List, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RNCsWithCellsInTheAccessedURA_List },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_URA_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_URA_Information, URA_Information_sequence);

  return offset;
}



static int
dissect_rnsap_USCH_ID(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_integer(tvb, offset, actx, tree, hf_index,
                                                            0U, 255U, NULL, FALSE);

  return offset;
}


static const per_sequence_t USCH_InformationItem_sequence[] = {
  { &hf_rnsap_uSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_USCH_ID },
  { &hf_rnsap_ul_CCTrCH_ID  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_trChSourceStatisticsDescriptor, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TrCH_SrcStatisticsDescr },
  { &hf_rnsap_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_rb_Info       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RB_Info },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_USCH_InformationItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_USCH_InformationItem, USCH_InformationItem_sequence);

  return offset;
}


static const per_sequence_t USCH_Information_sequence_of[1] = {
  { &hf_rnsap_USCH_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_USCH_InformationItem },
};

static int
dissect_rnsap_USCH_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_USCH_Information, USCH_Information_sequence_of,
                                                  1, maxNoOfUSCHs, FALSE);

  return offset;
}


static const value_string rnsap_Usefulness_Of_Battery_Optimization_vals[] = {
  {   0, "can-benefit" },
  {   1, "cannot-benefit" },
  { 0, NULL }
};


static int
dissect_rnsap_Usefulness_Of_Battery_Optimization(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, FALSE, 0, NULL);

  return offset;
}


static const value_string rnsap_User_Plane_Congestion_Fields_Inclusion_vals[] = {
  {   0, "shall-be-included" },
  { 0, NULL }
};


static int
dissect_rnsap_User_Plane_Congestion_Fields_Inclusion(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, FALSE, 0, NULL);

  return offset;
}


static const per_sequence_t UMTS_Cell_Info_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UMTS_Cell_Info(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UMTS_Cell_Info, UMTS_Cell_Info_sequence);

  return offset;
}


static const per_sequence_t UL_CLTD_Information_Reconf_sequence[] = {
  { &hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CLTD_Information_Reconf(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CLTD_Information_Reconf, UL_CLTD_Information_Reconf_sequence);

  return offset;
}


static const value_string rnsap_UL_CLTD_State_Update_Information_vals[] = {
  {   0, "activate" },
  {   1, "de-activate" },
  { 0, NULL }
};


static int
dissect_rnsap_UL_CLTD_State_Update_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t UL_DPCCH2_Information_Reconf_sequence[] = {
  { &hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCCH2_Information_Reconf(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCCH2_Information_Reconf, UL_DPCCH2_Information_Reconf_sequence);

  return offset;
}


static const per_sequence_t UL_MIMO_Information_sequence[] = {
  { &hf_rnsap_e_roch_power_offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_ROCH_PowerOffset },
  { &hf_rnsap_s_e_dpcch_power_offset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_S_E_DPCCH_PowerOffset },
  { &hf_rnsap_interstream_compensation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_InterStream_Interference_Compensation },
  { &hf_rnsap_minimum_E_TFCI_rank2, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_INTEGER_0_127 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_MIMO_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_MIMO_Information, UL_MIMO_Information_sequence);

  return offset;
}


static const per_sequence_t UL_MIMO_Information_To_Modify_sequence[] = {
  { &hf_rnsap_e_roch_power_offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_E_ROCH_PowerOffset },
  { &hf_rnsap_s_e_dpcch_power_offset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_S_E_DPCCH_PowerOffset },
  { &hf_rnsap_interstream_compensation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_InterStream_Interference_Compensation },
  { &hf_rnsap_minimum_E_TFCI_rank2, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_INTEGER_0_127 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_MIMO_Information_To_Modify(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_MIMO_Information_To_Modify, UL_MIMO_Information_To_Modify_sequence);

  return offset;
}


static const value_string rnsap_UL_MIMO_Removal_vals[] = {
  {   0, "remove" },
  { 0, NULL }
};


static int
dissect_rnsap_UL_MIMO_Removal(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     1, NULL, TRUE, 0, NULL);

  return offset;
}


static const value_string rnsap_UL_MIMO_Reconfiguration_vals[] = {
  {   0, "setup" },
  {   1, "configurationChange" },
  {   2, "removal" },
  { 0, NULL }
};

static const per_choice_t UL_MIMO_Reconfiguration_choice[] = {
  {   0, &hf_rnsap_setup_08      , ASN1_NO_EXTENSIONS     , dissect_rnsap_UL_MIMO_Information },
  {   1, &hf_rnsap_configurationChange_07, ASN1_NO_EXTENSIONS     , dissect_rnsap_UL_MIMO_Information_To_Modify },
  {   2, &hf_rnsap_removal_06    , ASN1_NO_EXTENSIONS     , dissect_rnsap_UL_MIMO_Removal },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_UL_MIMO_Reconfiguration(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_UL_MIMO_Reconfiguration, UL_MIMO_Reconfiguration_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t UL_MIMO_DL_Control_Channel_Information_sequence[] = {
  { &hf_rnsap_e_roch_channelization_code, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DL_ChannelisationCodeNumber },
  { &hf_rnsap_s_e_rnti      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_RNTI },
  { &hf_rnsap_s_signature_sequence, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ERGCH_SignatureSequence },
  { &hf_rnsap_s_e_roch_release_indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_S_E_ROCH_Release_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_MIMO_DL_Control_Channel_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_MIMO_DL_Control_Channel_Information, UL_MIMO_DL_Control_Channel_Information_sequence);

  return offset;
}


static const per_sequence_t RadioLinkSetupRequestFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkSetupRequestFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkSetupRequestFDD, RadioLinkSetupRequestFDD_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_Information_RL_SetupRqstFDD_sequence[] = {
  { &hf_rnsap_ul_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_ScramblingCode },
  { &hf_rnsap_minUL_ChannelisationCodeLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MinUL_ChannelisationCodeLength },
  { &hf_rnsap_maxNrOfUL_DPCHs, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MaxNrOfUL_DPCHs },
  { &hf_rnsap_ul_PunctureLimit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PunctureLimit },
  { &hf_rnsap_ul_TFCS       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCS },
  { &hf_rnsap_ul_DPCCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_DPCCH_SlotFormat },
  { &hf_rnsap_ul_SIRTarget  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_diversityMode , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DiversityMode },
  { &hf_rnsap_not_Used_sSDT_CellIdLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_not_Used_s_FieldLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_Information_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_Information_RL_SetupRqstFDD, UL_DPCH_Information_RL_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t PowerOffsetInformation_RL_SetupRqstFDD_sequence[] = {
  { &hf_rnsap_po1_ForTFCI_Bits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerOffset },
  { &hf_rnsap_po2_ForTPC_Bits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerOffset },
  { &hf_rnsap_po3_ForPilotBits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerOffset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PowerOffsetInformation_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PowerOffsetInformation_RL_SetupRqstFDD, PowerOffsetInformation_RL_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_Information_RL_SetupRqstFDD_sequence[] = {
  { &hf_rnsap_tFCS          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCS },
  { &hf_rnsap_dl_DPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_DPCH_SlotFormat },
  { &hf_rnsap_nrOfDLchannelisationcodes, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_NrOfDLchannelisationcodes },
  { &hf_rnsap_tFCI_SignallingMode, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_SignallingMode },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_multiplexingPosition, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MultiplexingPosition },
  { &hf_rnsap_powerOffsetInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerOffsetInformation_RL_SetupRqstFDD },
  { &hf_rnsap_fdd_dl_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_TPC_DownlinkStepSize },
  { &hf_rnsap_limitedPowerIncrease, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_LimitedPowerIncrease },
  { &hf_rnsap_innerLoopDLPCStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_InnerLoopDLPCStatus },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_Information_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_Information_RL_SetupRqstFDD, DL_DPCH_Information_RL_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_SetupRqstFDD_sequence_of[1] = {
  { &hf_rnsap_RL_InformationList_RL_SetupRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationList_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationList_RL_SetupRqstFDD, RL_InformationList_RL_SetupRqstFDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationItem_RL_SetupRqstFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_firstRLS_indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FirstRLS_Indicator },
  { &hf_rnsap_frameOffset   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FrameOffset },
  { &hf_rnsap_chipOffset    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ChipOffset },
  { &hf_rnsap_propagationDelay, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PropagationDelay },
  { &hf_rnsap_diversityControlField, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DiversityControlField },
  { &hf_rnsap_dl_InitialTX_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_primaryCPICH_EcNo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PrimaryCPICH_EcNo },
  { &hf_rnsap_not_Used_sSDT_CellID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_transmitDiversityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransmitDiversityIndicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationItem_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationItem_RL_SetupRqstFDD, RL_InformationItem_RL_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Setup_ItemIEs_sequence[] = {
  { &hf_rnsap_hSPDSCH_RL_ID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_DSCH_FDD_Secondary_Serving_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_HS_Cell_Information_RL_Setup_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_HS_Cell_Information_RL_Setup_ItemIEs, Additional_HS_Cell_Information_RL_Setup_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Setup_List_sequence_of[1] = {
  { &hf_rnsap_Additional_HS_Cell_Information_RL_Setup_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_HS_Cell_Information_RL_Setup_ItemIEs },
};

static int
dissect_rnsap_Additional_HS_Cell_Information_RL_Setup_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_HS_Cell_Information_RL_Setup_List, Additional_HS_Cell_Information_RL_Setup_List_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD_sequence[] = {
  { &hf_rnsap_po2_ForTPC_Bits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerOffset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD, PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t F_DPCH_Information_RL_SetupRqstFDD_sequence[] = {
  { &hf_rnsap_powerOffsetInformation_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD },
  { &hf_rnsap_fdd_dl_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_TPC_DownlinkStepSize },
  { &hf_rnsap_limitedPowerIncrease, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_LimitedPowerIncrease },
  { &hf_rnsap_innerLoopDLPCStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_InnerLoopDLPCStatus },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_F_DPCH_Information_RL_SetupRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_F_DPCH_Information_RL_SetupRqstFDD, F_DPCH_Information_RL_SetupRqstFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkSetupRequestTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkSetupRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkSetupRequestTDD, RadioLinkSetupRequestTDD_sequence);

  return offset;
}


static const per_sequence_t UL_Physical_Channel_Information_RL_SetupRqstTDD_sequence[] = {
  { &hf_rnsap_maxNrTimeslots_UL, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaxNrTimeslots },
  { &hf_rnsap_minimumSpreadingFactor_UL, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MinimumSpreadingFactor },
  { &hf_rnsap_maxNrULPhysicalchannels, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaxNrULPhysicalchannels },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_Physical_Channel_Information_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_Physical_Channel_Information_RL_SetupRqstTDD, UL_Physical_Channel_Information_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Physical_Channel_Information_RL_SetupRqstTDD_sequence[] = {
  { &hf_rnsap_maxNrTimeslots_DL, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaxNrTimeslots },
  { &hf_rnsap_minimumSpreadingFactor_DL, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MinimumSpreadingFactor },
  { &hf_rnsap_maxNrDLPhysicalchannels, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaxNrDLPhysicalchannels },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_Physical_Channel_Information_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_Physical_Channel_Information_RL_SetupRqstTDD, DL_Physical_Channel_Information_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationList_RL_SetupRqstTDD_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCH_InformationList_RL_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_UL_CCTrCH_InformationList_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCH_InformationList_RL_SetupRqstTDD, UL_CCTrCH_InformationList_RL_SetupRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationItem_RL_SetupRqstTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_ul_TFCS       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCS },
  { &hf_rnsap_tFCI_Coding   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Coding },
  { &hf_rnsap_ul_PunctureLimit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PunctureLimit },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD, UL_CCTrCH_InformationItem_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationList_RL_SetupRqstTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCH_InformationList_RL_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_DL_CCTrCH_InformationList_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCH_InformationList_RL_SetupRqstTDD, DL_CCTrCH_InformationList_RL_SetupRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t CCTrCH_TPCItem_RL_SetupRqstTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CCTrCH_TPCItem_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CCTrCH_TPCItem_RL_SetupRqstTDD, CCTrCH_TPCItem_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_TPCList_RL_SetupRqstTDD_sequence_of[1] = {
  { &hf_rnsap_CCTrCH_TPCList_RL_SetupRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_TPCItem_RL_SetupRqstTDD },
};

static int
dissect_rnsap_CCTrCH_TPCList_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_CCTrCH_TPCList_RL_SetupRqstTDD, CCTrCH_TPCList_RL_SetupRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationItem_RL_SetupRqstTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_dl_TFCS       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCS },
  { &hf_rnsap_tFCI_Coding   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Coding },
  { &hf_rnsap_dl_PunctureLimit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PunctureLimit },
  { &hf_rnsap_tdd_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_TPC_DownlinkStepSize },
  { &hf_rnsap_cCTrCH_TPCList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CCTrCH_TPCList_RL_SetupRqstTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD, DL_CCTrCH_InformationItem_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t RL_Information_RL_SetupRqstTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_frameOffset   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FrameOffset },
  { &hf_rnsap_specialBurstScheduling, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SpecialBurstScheduling },
  { &hf_rnsap_primaryCCPCH_RSCP, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PrimaryCCPCH_RSCP },
  { &hf_rnsap_dL_TimeSlot_ISCP, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_TimeSlot_ISCP_Info },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Information_RL_SetupRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Information_RL_SetupRqstTDD, RL_Information_RL_SetupRqstTDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkSetupResponseFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkSetupResponseFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkSetupResponseFDD, RadioLinkSetupResponseFDD_sequence);

  return offset;
}


static const per_sequence_t RL_InformationResponseList_RL_SetupRspFDD_sequence_of[1] = {
  { &hf_rnsap_RL_InformationResponseList_RL_SetupRspFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationResponseList_RL_SetupRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationResponseList_RL_SetupRspFDD, RL_InformationResponseList_RL_SetupRspFDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t Combining_RL_SetupRspFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Combining_RL_SetupRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Combining_RL_SetupRspFDD, Combining_RL_SetupRspFDD_sequence);

  return offset;
}


static const per_sequence_t NonCombiningOrFirstRL_RL_SetupRspFDD_sequence[] = {
  { &hf_rnsap_dCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_InformationResponse },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_NonCombiningOrFirstRL_RL_SetupRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_NonCombiningOrFirstRL_RL_SetupRspFDD, NonCombiningOrFirstRL_RL_SetupRspFDD_sequence);

  return offset;
}


static const value_string rnsap_DiversityIndication_RL_SetupRspFDD_vals[] = {
  {   0, "combining" },
  {   1, "nonCombiningOrFirstRL" },
  { 0, NULL }
};

static const per_choice_t DiversityIndication_RL_SetupRspFDD_choice[] = {
  {   0, &hf_rnsap_combining     , ASN1_NO_EXTENSIONS     , dissect_rnsap_Combining_RL_SetupRspFDD },
  {   1, &hf_rnsap_nonCombiningOrFirstRL, ASN1_NO_EXTENSIONS     , dissect_rnsap_NonCombiningOrFirstRL_RL_SetupRspFDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DiversityIndication_RL_SetupRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DiversityIndication_RL_SetupRspFDD, DiversityIndication_RL_SetupRspFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_InformationResponseItem_RL_SetupRspFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_rL_Set_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_uRA_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_URA_Information },
  { &hf_rnsap_sAI           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAI },
  { &hf_rnsap_gA_Cell       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_Cell },
  { &hf_rnsap_gA_AccessPointPosition, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_AccessPointPosition },
  { &hf_rnsap_received_total_wide_band_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Received_total_wide_band_power },
  { &hf_rnsap_not_Used_secondary_CCPCH_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_dl_CodeInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DL_CodeInformation },
  { &hf_rnsap_diversityIndication, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DiversityIndication_RL_SetupRspFDD },
  { &hf_rnsap_sSDT_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SSDT_SupportIndicator },
  { &hf_rnsap_maxUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_minUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_closedlooptimingadjustmentmode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Closedlooptimingadjustmentmode },
  { &hf_rnsap_maximumAllowedULTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaximumAllowedULTxPower },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_primaryScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PrimaryScramblingCode },
  { &hf_rnsap_uL_UARFCN     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_dL_UARFCN     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_primaryCPICH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PrimaryCPICH_Power },
  { &hf_rnsap_not_Used_dSCHInformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_neighbouring_UMTS_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_UMTS_CellInformation },
  { &hf_rnsap_neighbouring_GSM_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_GSM_CellInformation },
  { &hf_rnsap_pC_Preamble   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PC_Preamble },
  { &hf_rnsap_sRB_Delay     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SRB_Delay },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationResponseItem_RL_SetupRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationResponseItem_RL_SetupRspFDD, RL_InformationResponseItem_RL_SetupRspFDD_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_Response_ItemIEs_sequence[] = {
  { &hf_rnsap_hSPDSCH_RL_ID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_hSDSCH_RNTI   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_RNTI },
  { &hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_DSCH_FDD_Secondary_Serving_Information_Response },
  { &hf_rnsap_sixtyfourQAM_DL_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SixtyfourQAM_DL_SupportIndicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_HS_Cell_Information_Response_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_HS_Cell_Information_Response_ItemIEs, Additional_HS_Cell_Information_Response_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_Response_List_sequence_of[1] = {
  { &hf_rnsap_Additional_HS_Cell_Information_Response_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_HS_Cell_Information_Response_ItemIEs },
};

static int
dissect_rnsap_Additional_HS_Cell_Information_Response_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_HS_Cell_Information_Response_List, Additional_HS_Cell_Information_Response_List_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t RadioLinkSetupResponseTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkSetupResponseTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkSetupResponseTDD, RadioLinkSetupResponseTDD_sequence);

  return offset;
}



static int
dissect_rnsap_UL_CCTrCHInformationList_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DL_CCTrCHInformationList_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DCH_InformationResponseList_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DSCH_InformationResponse_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_USCH_InformationResponse_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_InformationResponse_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_uRA_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_URA_Information },
  { &hf_rnsap_sAI           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAI },
  { &hf_rnsap_gA_Cell       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_Cell },
  { &hf_rnsap_gA_AccessPointPosition, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_AccessPointPosition },
  { &hf_rnsap_ul_TimeSlot_ISCP_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeSlot_ISCP_Info },
  { &hf_rnsap_maxUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_minUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_maximumAllowedULTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaximumAllowedULTxPower },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_uARFCNforNt   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_cellParameterID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CellParameterID },
  { &hf_rnsap_syncCase      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SyncCase },
  { &hf_rnsap_sCH_TimeSlot  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SCH_TimeSlot },
  { &hf_rnsap_sCTD_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SCTD_Indicator },
  { &hf_rnsap_pCCPCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PCCPCH_Power },
  { &hf_rnsap_timingAdvanceApplied, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimingAdvanceApplied },
  { &hf_rnsap_alphaValue    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AlphaValue },
  { &hf_rnsap_ul_PhysCH_SF_Variation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_PhysCH_SF_Variation },
  { &hf_rnsap_synchronisationConfiguration, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SynchronisationConfiguration },
  { &hf_rnsap_secondary_CCPCH_Info_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Secondary_CCPCH_Info_TDD },
  { &hf_rnsap_ul_CCTrCHInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_CCTrCHInformationList_RL_SetupRspTDD },
  { &hf_rnsap_dl_CCTrCHInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_CCTrCHInformationList_RL_SetupRspTDD },
  { &hf_rnsap_dCH_InformationResponse_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DCH_InformationResponseList_RL_SetupRspTDD },
  { &hf_rnsap_dsch_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DSCH_InformationResponse_RL_SetupRspTDD },
  { &hf_rnsap_usch_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_USCH_InformationResponse_RL_SetupRspTDD },
  { &hf_rnsap_neighbouring_UMTS_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_UMTS_CellInformation },
  { &hf_rnsap_neighbouring_GSM_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_GSM_CellInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationResponse_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationResponse_RL_SetupRspTDD, RL_InformationResponse_RL_SetupRspTDD_sequence);

  return offset;
}



static int
dissect_rnsap_UL_DPCH_InformationList_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t UL_CCTrCHInformationItem_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_ul_DPCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_DPCH_InformationList_RL_SetupRspTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCHInformationItem_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCHInformationItem_RL_SetupRspTDD, UL_CCTrCHInformationItem_RL_SetupRspTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCHInformationListIE_RL_SetupRspTDD_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_CCTrCHInformationItem_RL_SetupRspTDD },
};

static int
dissect_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD, UL_CCTrCHInformationListIE_RL_SetupRspTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationItem_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_uL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Timeslot_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD, UL_DPCH_InformationItem_RL_SetupRspTDD_sequence);

  return offset;
}



static int
dissect_rnsap_DL_DPCH_InformationList_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DL_CCTrCHInformationItem_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_dl_DPCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_DPCH_InformationList_RL_SetupRspTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCHInformationItem_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCHInformationItem_RL_SetupRspTDD, DL_CCTrCHInformationItem_RL_SetupRspTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCHInformationListIE_RL_SetupRspTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_CCTrCHInformationItem_RL_SetupRspTDD },
};

static int
dissect_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD, DL_CCTrCHInformationListIE_RL_SetupRspTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationItem_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_dL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Timeslot_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD, DL_DPCH_InformationItem_RL_SetupRspTDD_sequence);

  return offset;
}


static const per_sequence_t DSCHInformationItem_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_dsch_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_ID },
  { &hf_rnsap_dSCH_FlowControlInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_FlowControlInformation },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_transportFormatManagement, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatManagement },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DSCHInformationItem_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DSCHInformationItem_RL_SetupRspTDD, DSCHInformationItem_RL_SetupRspTDD_sequence);

  return offset;
}


static const per_sequence_t DSCH_InformationListIEs_RL_SetupRspTDD_sequence_of[1] = {
  { &hf_rnsap_DSCH_InformationListIEs_RL_SetupRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCHInformationItem_RL_SetupRspTDD },
};

static int
dissect_rnsap_DSCH_InformationListIEs_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DSCH_InformationListIEs_RL_SetupRspTDD, DSCH_InformationListIEs_RL_SetupRspTDD_sequence_of,
                                                  0, maxNoOfDSCHs, FALSE);

  return offset;
}


static const per_sequence_t USCHInformationItem_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_usch_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_USCH_ID },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_transportFormatManagement, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatManagement },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_USCHInformationItem_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_USCHInformationItem_RL_SetupRspTDD, USCHInformationItem_RL_SetupRspTDD_sequence);

  return offset;
}


static const per_sequence_t USCH_InformationListIEs_RL_SetupRspTDD_sequence_of[1] = {
  { &hf_rnsap_USCH_InformationListIEs_RL_SetupRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_USCHInformationItem_RL_SetupRspTDD },
};

static int
dissect_rnsap_USCH_InformationListIEs_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_USCH_InformationListIEs_RL_SetupRspTDD, USCH_InformationListIEs_RL_SetupRspTDD_sequence_of,
                                                  0, maxNoOfUSCHs, FALSE);

  return offset;
}



static int
dissect_rnsap_UL_LCR_CCTrCHInformationList_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DL_LCR_CCTrCHInformationList_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DSCH_LCR_InformationResponse_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_USCH_LCR_InformationResponse_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_LCR_InformationResponse_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_uRA_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_URA_Information },
  { &hf_rnsap_sAI           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAI },
  { &hf_rnsap_gA_Cell       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_Cell },
  { &hf_rnsap_gA_AccessPointPosition, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_AccessPointPosition },
  { &hf_rnsap_ul_TimeSlot_ISCP_LCR_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeSlot_ISCP_LCR_Info },
  { &hf_rnsap_maxUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_minUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_maximumAllowedULTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaximumAllowedULTxPower },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_uARFCNforNt   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_cellParameterID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CellParameterID },
  { &hf_rnsap_sCTD_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SCTD_Indicator },
  { &hf_rnsap_pCCPCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PCCPCH_Power },
  { &hf_rnsap_alphaValue    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AlphaValue },
  { &hf_rnsap_ul_PhysCH_SF_Variation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_PhysCH_SF_Variation },
  { &hf_rnsap_synchronisationConfiguration, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SynchronisationConfiguration },
  { &hf_rnsap_secondary_LCR_CCPCH_Info_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Secondary_LCR_CCPCH_Info_TDD },
  { &hf_rnsap_ul_LCR_CCTrCHInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_LCR_CCTrCHInformationList_RL_SetupRspTDD },
  { &hf_rnsap_dl_LCR_CCTrCHInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_LCR_CCTrCHInformationList_RL_SetupRspTDD },
  { &hf_rnsap_dCH_InformationResponse_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DCH_InformationResponseList_RL_SetupRspTDD },
  { &hf_rnsap_dsch_LCR_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DSCH_LCR_InformationResponse_RL_SetupRspTDD },
  { &hf_rnsap_usch_LCR_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_USCH_LCR_InformationResponse_RL_SetupRspTDD },
  { &hf_rnsap_neighbouring_UMTS_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_UMTS_CellInformation },
  { &hf_rnsap_neighbouring_GSM_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_GSM_CellInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_LCR_InformationResponse_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_LCR_InformationResponse_RL_SetupRspTDD, RL_LCR_InformationResponse_RL_SetupRspTDD_sequence);

  return offset;
}



static int
dissect_rnsap_UL_DPCH_LCR_InformationList_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t UL_LCR_CCTrCHInformationItem_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_ul_DPCH_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_DPCH_LCR_InformationList_RL_SetupRspTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_LCR_CCTrCHInformationItem_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_LCR_CCTrCHInformationItem_RL_SetupRspTDD, UL_LCR_CCTrCHInformationItem_RL_SetupRspTDD_sequence);

  return offset;
}


static const per_sequence_t UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD_sequence_of[1] = {
  { &hf_rnsap_UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_LCR_CCTrCHInformationItem_RL_SetupRspTDD },
};

static int
dissect_rnsap_UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD, UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD_sequence_of,
                                                  1, maxNrOfCCTrCHsLCR, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_LCR_InformationItem_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_uL_TimeslotLCR_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeslotLCR_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_LCR_InformationItem_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_LCR_InformationItem_RL_SetupRspTDD, UL_DPCH_LCR_InformationItem_RL_SetupRspTDD_sequence);

  return offset;
}



static int
dissect_rnsap_DL_DPCH_LCR_InformationList_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DL_CCTrCH_LCR_InformationItem_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_dl_DPCH_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_DPCH_LCR_InformationList_RL_SetupRspTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCH_LCR_InformationItem_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCH_LCR_InformationItem_RL_SetupRspTDD, DL_CCTrCH_LCR_InformationItem_RL_SetupRspTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_CCTrCH_LCR_InformationItem_RL_SetupRspTDD },
};

static int
dissect_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD, DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD_sequence_of,
                                                  1, maxNrOfCCTrCHsLCR, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_LCR_InformationItem_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_dL_Timeslot_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TimeslotLCR_Information },
  { &hf_rnsap_tSTD_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TSTD_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_LCR_InformationItem_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_LCR_InformationItem_RL_SetupRspTDD, DL_DPCH_LCR_InformationItem_RL_SetupRspTDD_sequence);

  return offset;
}


static const per_sequence_t DSCH_LCR_InformationItem_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_dsch_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_ID },
  { &hf_rnsap_dSCH_FlowControlInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_FlowControlInformation },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_transportFormatManagement, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatManagement },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DSCH_LCR_InformationItem_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DSCH_LCR_InformationItem_RL_SetupRspTDD, DSCH_LCR_InformationItem_RL_SetupRspTDD_sequence);

  return offset;
}


static const per_sequence_t DSCH_LCR_InformationListIEs_RL_SetupRspTDD_sequence_of[1] = {
  { &hf_rnsap_DSCH_LCR_InformationListIEs_RL_SetupRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_LCR_InformationItem_RL_SetupRspTDD },
};

static int
dissect_rnsap_DSCH_LCR_InformationListIEs_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DSCH_LCR_InformationListIEs_RL_SetupRspTDD, DSCH_LCR_InformationListIEs_RL_SetupRspTDD_sequence_of,
                                                  0, maxNoOfDSCHsLCR, FALSE);

  return offset;
}


static const per_sequence_t USCH_LCR_InformationItem_RL_SetupRspTDD_sequence[] = {
  { &hf_rnsap_usch_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_USCH_ID },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_transportFormatManagement, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatManagement },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_USCH_LCR_InformationItem_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_USCH_LCR_InformationItem_RL_SetupRspTDD, USCH_LCR_InformationItem_RL_SetupRspTDD_sequence);

  return offset;
}


static const per_sequence_t USCH_LCR_InformationListIEs_RL_SetupRspTDD_sequence_of[1] = {
  { &hf_rnsap_USCH_LCR_InformationListIEs_RL_SetupRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_USCH_LCR_InformationItem_RL_SetupRspTDD },
};

static int
dissect_rnsap_USCH_LCR_InformationListIEs_RL_SetupRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_USCH_LCR_InformationListIEs_RL_SetupRspTDD, USCH_LCR_InformationListIEs_RL_SetupRspTDD_sequence_of,
                                                  0, maxNoOfUSCHsLCR, FALSE);

  return offset;
}



static int
dissect_rnsap_UL_CCTrCHInformationList_RL_SetupRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DL_CCTrCHInformationList_RL_SetupRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_InformationResponse_RL_SetupRspTDD768_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_uRA_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_URA_Information },
  { &hf_rnsap_sAI           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAI },
  { &hf_rnsap_gA_Cell       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_Cell },
  { &hf_rnsap_gA_AccessPointPosition, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_AccessPointPosition },
  { &hf_rnsap_ul_TimeSlot_ISCP_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeSlot_ISCP_Info },
  { &hf_rnsap_maxUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_minUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_maximumAllowedULTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaximumAllowedULTxPower },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_uARFCNforNt   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_cellParameterID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CellParameterID },
  { &hf_rnsap_syncCase      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SyncCase },
  { &hf_rnsap_sCH_TimeSlot  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SCH_TimeSlot },
  { &hf_rnsap_sCTD_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SCTD_Indicator },
  { &hf_rnsap_pCCPCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PCCPCH_Power },
  { &hf_rnsap_timingAdvanceApplied, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimingAdvanceApplied },
  { &hf_rnsap_alphaValue    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AlphaValue },
  { &hf_rnsap_ul_PhysCH_SF_Variation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_PhysCH_SF_Variation },
  { &hf_rnsap_synchronisationConfiguration, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SynchronisationConfiguration },
  { &hf_rnsap_secondary_CCPCH_Info_TDD768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Secondary_CCPCH_Info_TDD768 },
  { &hf_rnsap_ul_CCTrCHInformation768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_CCTrCHInformationList_RL_SetupRspTDD768 },
  { &hf_rnsap_dl_CCTrCHInformation768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_CCTrCHInformationList_RL_SetupRspTDD768 },
  { &hf_rnsap_dCH_InformationResponse_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DCH_InformationResponseList_RL_SetupRspTDD },
  { &hf_rnsap_dsch_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DSCH_InformationResponse_RL_SetupRspTDD },
  { &hf_rnsap_usch_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_USCH_InformationResponse_RL_SetupRspTDD },
  { &hf_rnsap_neighbouring_UMTS_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_UMTS_CellInformation },
  { &hf_rnsap_neighbouring_GSM_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_GSM_CellInformation },
  { &hf_rnsap_gA_CellAdditionalShapes, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_CellAdditionalShapes },
  { &hf_rnsap_hCS_Prio      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HCS_Prio },
  { &hf_rnsap_timeSlot_RL_SetupRspTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TimeSlot },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationResponse_RL_SetupRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationResponse_RL_SetupRspTDD768, RL_InformationResponse_RL_SetupRspTDD768_sequence);

  return offset;
}



static int
dissect_rnsap_UL_DPCH_InformationList_RL_SetupRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t UL_CCTrCHInformationItem_RL_SetupRspTDD768_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_ul_DPCH_Information768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_DPCH_InformationList_RL_SetupRspTDD768 },
  { &hf_rnsap_uL_SIR_Target_CCTrCH_InformationItem_RL_SetupRspTDD768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCHInformationItem_RL_SetupRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCHInformationItem_RL_SetupRspTDD768, UL_CCTrCHInformationItem_RL_SetupRspTDD768_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCHInformationListIE_RL_SetupRspTDD768_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_CCTrCHInformationItem_RL_SetupRspTDD768 },
};

static int
dissect_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD768, UL_CCTrCHInformationListIE_RL_SetupRspTDD768_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationItem_RL_SetupRspTDD768_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_uL_Timeslot_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Timeslot_Information768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD768, UL_DPCH_InformationItem_RL_SetupRspTDD768_sequence);

  return offset;
}



static int
dissect_rnsap_DL_DPCH_InformationList_RL_SetupRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DL_CCTrCHInformationItem_RL_SetupRspTDD768_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_dl_DPCH_Information768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_DPCH_InformationList_RL_SetupRspTDD768 },
  { &hf_rnsap_cCTrCH_Maximum_DL_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_cCTrCH_Minimum_DL_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCHInformationItem_RL_SetupRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCHInformationItem_RL_SetupRspTDD768, DL_CCTrCHInformationItem_RL_SetupRspTDD768_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCHInformationListIE_RL_SetupRspTDD768_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_CCTrCHInformationItem_RL_SetupRspTDD768 },
};

static int
dissect_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD768, DL_CCTrCHInformationListIE_RL_SetupRspTDD768_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationItem_RL_SetupRspTDD768_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_dL_Timeslot_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Timeslot_Information768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD768, DL_DPCH_InformationItem_RL_SetupRspTDD768_sequence);

  return offset;
}


static const per_sequence_t RadioLinkSetupFailureFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkSetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkSetupFailureFDD, RadioLinkSetupFailureFDD_sequence);

  return offset;
}


static const per_sequence_t GeneralCauseList_RL_SetupFailureFDD_sequence[] = {
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GeneralCauseList_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GeneralCauseList_RL_SetupFailureFDD, GeneralCauseList_RL_SetupFailureFDD_sequence);

  return offset;
}


static const per_sequence_t UnsuccessfulRL_InformationResponseList_RL_SetupFailureFDD_sequence_of[1] = {
  { &hf_rnsap_UnsuccessfulRL_InformationResponseList_RL_SetupFailureFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_UnsuccessfulRL_InformationResponseList_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UnsuccessfulRL_InformationResponseList_RL_SetupFailureFDD, UnsuccessfulRL_InformationResponseList_RL_SetupFailureFDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t SuccessfulRL_InformationResponseList_RL_SetupFailureFDD_sequence_of[1] = {
  { &hf_rnsap_SuccessfulRL_InformationResponseList_RL_SetupFailureFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_SuccessfulRL_InformationResponseList_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_SuccessfulRL_InformationResponseList_RL_SetupFailureFDD, SuccessfulRL_InformationResponseList_RL_SetupFailureFDD_sequence_of,
                                                  0, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t RLSpecificCauseList_RL_SetupFailureFDD_sequence[] = {
  { &hf_rnsap_unsuccessful_RL_InformationRespList_RL_SetupFailureFDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UnsuccessfulRL_InformationResponseList_RL_SetupFailureFDD },
  { &hf_rnsap_successful_RL_InformationRespList_RL_SetupFailureFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SuccessfulRL_InformationResponseList_RL_SetupFailureFDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RLSpecificCauseList_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RLSpecificCauseList_RL_SetupFailureFDD, RLSpecificCauseList_RL_SetupFailureFDD_sequence);

  return offset;
}


static const value_string rnsap_CauseLevel_RL_SetupFailureFDD_vals[] = {
  {   0, "generalCause" },
  {   1, "rLSpecificCause" },
  { 0, NULL }
};

static const per_choice_t CauseLevel_RL_SetupFailureFDD_choice[] = {
  {   0, &hf_rnsap_generalCause  , ASN1_EXTENSION_ROOT    , dissect_rnsap_GeneralCauseList_RL_SetupFailureFDD },
  {   1, &hf_rnsap_rLSpecificCause, ASN1_EXTENSION_ROOT    , dissect_rnsap_RLSpecificCauseList_RL_SetupFailureFDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_CauseLevel_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_CauseLevel_RL_SetupFailureFDD, CauseLevel_RL_SetupFailureFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD, UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD_sequence);

  return offset;
}


static const per_sequence_t Combining_RL_SetupFailureFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Combining_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Combining_RL_SetupFailureFDD, Combining_RL_SetupFailureFDD_sequence);

  return offset;
}


static const per_sequence_t NonCombiningOrFirstRL_RL_SetupFailureFDD_sequence[] = {
  { &hf_rnsap_dCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_InformationResponse },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_NonCombiningOrFirstRL_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_NonCombiningOrFirstRL_RL_SetupFailureFDD, NonCombiningOrFirstRL_RL_SetupFailureFDD_sequence);

  return offset;
}


static const value_string rnsap_DiversityIndication_RL_SetupFailureFDD_vals[] = {
  {   0, "combining" },
  {   1, "nonCombiningOrFirstRL" },
  { 0, NULL }
};

static const per_choice_t DiversityIndication_RL_SetupFailureFDD_choice[] = {
  {   0, &hf_rnsap_combining_01  , ASN1_NO_EXTENSIONS     , dissect_rnsap_Combining_RL_SetupFailureFDD },
  {   1, &hf_rnsap_nonCombiningOrFirstRL_01, ASN1_NO_EXTENSIONS     , dissect_rnsap_NonCombiningOrFirstRL_RL_SetupFailureFDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DiversityIndication_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DiversityIndication_RL_SetupFailureFDD, DiversityIndication_RL_SetupFailureFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t SuccessfulRL_InformationResponse_RL_SetupFailureFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_rL_Set_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_uRA_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_URA_Information },
  { &hf_rnsap_sAI           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAI },
  { &hf_rnsap_gA_Cell       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_Cell },
  { &hf_rnsap_gA_AccessPointPosition, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_AccessPointPosition },
  { &hf_rnsap_received_total_wide_band_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Received_total_wide_band_power },
  { &hf_rnsap_not_Used_secondary_CCPCH_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_dl_CodeInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_DL_CodeInformation },
  { &hf_rnsap_diversityIndication_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DiversityIndication_RL_SetupFailureFDD },
  { &hf_rnsap_sSDT_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SSDT_SupportIndicator },
  { &hf_rnsap_maxUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_minUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_closedlooptimingadjustmentmode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Closedlooptimingadjustmentmode },
  { &hf_rnsap_maximumAllowedULTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaximumAllowedULTxPower },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_primaryCPICH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PrimaryCPICH_Power },
  { &hf_rnsap_primaryScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PrimaryScramblingCode },
  { &hf_rnsap_uL_UARFCN     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_dL_UARFCN     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UARFCN },
  { &hf_rnsap_not_Used_dSCH_InformationResponse_RL_SetupFailureFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_neighbouring_UMTS_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_UMTS_CellInformation },
  { &hf_rnsap_neighbouring_GSM_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_GSM_CellInformation },
  { &hf_rnsap_pC_Preamble   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PC_Preamble },
  { &hf_rnsap_sRB_Delay     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SRB_Delay },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SuccessfulRL_InformationResponse_RL_SetupFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SuccessfulRL_InformationResponse_RL_SetupFailureFDD, SuccessfulRL_InformationResponse_RL_SetupFailureFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkSetupFailureTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkSetupFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkSetupFailureTDD, RadioLinkSetupFailureTDD_sequence);

  return offset;
}


static const per_sequence_t GeneralCauseList_RL_SetupFailureTDD_sequence[] = {
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GeneralCauseList_RL_SetupFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GeneralCauseList_RL_SetupFailureTDD, GeneralCauseList_RL_SetupFailureTDD_sequence);

  return offset;
}



static int
dissect_rnsap_Unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RLSpecificCauseList_RL_SetupFailureTDD_sequence[] = {
  { &hf_rnsap_unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RLSpecificCauseList_RL_SetupFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RLSpecificCauseList_RL_SetupFailureTDD, RLSpecificCauseList_RL_SetupFailureTDD_sequence);

  return offset;
}


static const value_string rnsap_CauseLevel_RL_SetupFailureTDD_vals[] = {
  {   0, "generalCause" },
  {   1, "rLSpecificCause" },
  { 0, NULL }
};

static const per_choice_t CauseLevel_RL_SetupFailureTDD_choice[] = {
  {   0, &hf_rnsap_generalCause_01, ASN1_EXTENSION_ROOT    , dissect_rnsap_GeneralCauseList_RL_SetupFailureTDD },
  {   1, &hf_rnsap_rLSpecificCause_01, ASN1_EXTENSION_ROOT    , dissect_rnsap_RLSpecificCauseList_RL_SetupFailureTDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_CauseLevel_RL_SetupFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_CauseLevel_RL_SetupFailureTDD, CauseLevel_RL_SetupFailureTDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD, UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkAdditionRequestFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkAdditionRequestFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkAdditionRequestFDD, RadioLinkAdditionRequestFDD_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_AdditionRqstFDD_sequence_of[1] = {
  { &hf_rnsap_RL_InformationList_RL_AdditionRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationList_RL_AdditionRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationList_RL_AdditionRqstFDD, RL_InformationList_RL_AdditionRqstFDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t RL_Information_RL_AdditionRqstFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_frameOffset   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FrameOffset },
  { &hf_rnsap_chipOffset    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ChipOffset },
  { &hf_rnsap_diversityControlField, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DiversityControlField },
  { &hf_rnsap_primaryCPICH_EcNo, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PrimaryCPICH_EcNo },
  { &hf_rnsap_not_Used_sSDT_CellID, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_transmitDiversityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransmitDiversityIndicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Information_RL_AdditionRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Information_RL_AdditionRqstFDD, RL_Information_RL_AdditionRqstFDD_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Addition_ItemIEs_sequence[] = {
  { &hf_rnsap_hSPDSCH_RL_ID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_DSCH_FDD_Secondary_Serving_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_HS_Cell_Information_RL_Addition_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_HS_Cell_Information_RL_Addition_ItemIEs, Additional_HS_Cell_Information_RL_Addition_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Addition_List_sequence_of[1] = {
  { &hf_rnsap_Additional_HS_Cell_Information_RL_Addition_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_HS_Cell_Information_RL_Addition_ItemIEs },
};

static int
dissect_rnsap_Additional_HS_Cell_Information_RL_Addition_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_HS_Cell_Information_RL_Addition_List, Additional_HS_Cell_Information_RL_Addition_List_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const value_string rnsap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency_vals[] = {
  {   0, "setup" },
  {   1, "addition" },
  { 0, NULL }
};

static const per_choice_t Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency_choice[] = {
  {   0, &hf_rnsap_setup_02      , ASN1_EXTENSION_ROOT    , dissect_rnsap_Additional_EDCH_Setup_Info },
  {   1, &hf_rnsap_addition      , ASN1_EXTENSION_ROOT    , dissect_rnsap_Additional_EDCH_Cell_Information_To_Add_List },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency, Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_RL_Add_Req_sequence[] = {
  { &hf_rnsap_setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_RL_Add_Req(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_Cell_Information_RL_Add_Req, Additional_EDCH_Cell_Information_RL_Add_Req_sequence);

  return offset;
}


static const per_sequence_t EDPCH_Information_RLAdditionReq_FDD_sequence[] = {
  { &hf_rnsap_maxSet_E_DPDCHs, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Max_Set_E_DPDCHs },
  { &hf_rnsap_ul_PunctureLimit, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PunctureLimit },
  { &hf_rnsap_e_TFCS_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_TFCS_Information },
  { &hf_rnsap_e_TTI         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_TTI },
  { &hf_rnsap_e_DPCCH_PO    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_DPCCH_PO },
  { &hf_rnsap_e_RGCH_2_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_RGCH_2_IndexStepThreshold },
  { &hf_rnsap_e_RGCH_3_IndexStepThreshold, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_E_RGCH_3_IndexStepThreshold },
  { &hf_rnsap_hARQ_Info_for_E_DCH, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HARQ_Info_for_E_DCH },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDPCH_Information_RLAdditionReq_FDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDPCH_Information_RLAdditionReq_FDD, EDPCH_Information_RLAdditionReq_FDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkAdditionRequestTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkAdditionRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkAdditionRequestTDD, RadioLinkAdditionRequestTDD_sequence);

  return offset;
}


static const per_sequence_t RL_Information_RL_AdditionRqstTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_frameOffset   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FrameOffset },
  { &hf_rnsap_diversityControlField, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DiversityControlField },
  { &hf_rnsap_primaryCCPCH_RSCP, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PrimaryCCPCH_RSCP },
  { &hf_rnsap_dL_TimeSlot_ISCP_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_TimeSlot_ISCP_Info },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Information_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Information_RL_AdditionRqstTDD, RL_Information_RL_AdditionRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationList_RL_AdditionRqstTDD_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD, UL_CCTrCH_InformationList_RL_AdditionRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationItem_RL_AdditionRqstTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_uplinkStepSizeLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_TPC_UplinkStepSize_LCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD, UL_CCTrCH_InformationItem_RL_AdditionRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationList_RL_AdditionRqstTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD, DL_CCTrCH_InformationList_RL_AdditionRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationItem_RL_AdditionRqstTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_downlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_TPC_DownlinkStepSize },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD, DL_CCTrCH_InformationItem_RL_AdditionRqstTDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkAdditionResponseFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkAdditionResponseFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkAdditionResponseFDD, RadioLinkAdditionResponseFDD_sequence);

  return offset;
}


static const per_sequence_t RL_InformationResponseList_RL_AdditionRspFDD_sequence_of[1] = {
  { &hf_rnsap_RL_InformationResponseList_RL_AdditionRspFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationResponseList_RL_AdditionRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationResponseList_RL_AdditionRspFDD, RL_InformationResponseList_RL_AdditionRspFDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}



static int
dissect_rnsap_DL_CodeInformationList_RL_AdditionRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Combining_RL_AdditionRspFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Combining_RL_AdditionRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Combining_RL_AdditionRspFDD, Combining_RL_AdditionRspFDD_sequence);

  return offset;
}


static const per_sequence_t NonCombining_RL_AdditionRspFDD_sequence[] = {
  { &hf_rnsap_dCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_InformationResponse },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_NonCombining_RL_AdditionRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_NonCombining_RL_AdditionRspFDD, NonCombining_RL_AdditionRspFDD_sequence);

  return offset;
}


static const value_string rnsap_DiversityIndication_RL_AdditionRspFDD_vals[] = {
  {   0, "combining" },
  {   1, "nonCombining" },
  { 0, NULL }
};

static const per_choice_t DiversityIndication_RL_AdditionRspFDD_choice[] = {
  {   0, &hf_rnsap_combining_02  , ASN1_NO_EXTENSIONS     , dissect_rnsap_Combining_RL_AdditionRspFDD },
  {   1, &hf_rnsap_nonCombining  , ASN1_NO_EXTENSIONS     , dissect_rnsap_NonCombining_RL_AdditionRspFDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DiversityIndication_RL_AdditionRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DiversityIndication_RL_AdditionRspFDD, DiversityIndication_RL_AdditionRspFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_InformationResponseItem_RL_AdditionRspFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_rL_Set_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_uRA_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_URA_Information },
  { &hf_rnsap_sAI           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAI },
  { &hf_rnsap_gA_Cell       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_Cell },
  { &hf_rnsap_gA_AccessPointPosition, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_AccessPointPosition },
  { &hf_rnsap_received_total_wide_band_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Received_total_wide_band_power },
  { &hf_rnsap_not_Used_secondary_CCPCH_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_dl_CodeInformation_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_CodeInformationList_RL_AdditionRspFDD },
  { &hf_rnsap_diversityIndication_02, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DiversityIndication_RL_AdditionRspFDD },
  { &hf_rnsap_sSDT_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SSDT_SupportIndicator },
  { &hf_rnsap_minUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_maxUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_closedlooptimingadjustmentmode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Closedlooptimingadjustmentmode },
  { &hf_rnsap_maximumAllowedULTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaximumAllowedULTxPower },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_neighbouring_UMTS_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_UMTS_CellInformation },
  { &hf_rnsap_neighbouring_GSM_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_GSM_CellInformation },
  { &hf_rnsap_pC_Preamble   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PC_Preamble },
  { &hf_rnsap_sRB_Delay     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SRB_Delay },
  { &hf_rnsap_primaryCPICH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PrimaryCPICH_Power },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationResponseItem_RL_AdditionRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationResponseItem_RL_AdditionRspFDD, RL_InformationResponseItem_RL_AdditionRspFDD_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Change_Information_Response_ItemIEs_sequence[] = {
  { &hf_rnsap_hSPDSCH_RL_ID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_hSDSCH_RNTI   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_RNTI },
  { &hf_rnsap_hS_DSCH_Secondary_Serving_Cell_Change_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_DSCH_Secondary_Serving_Cell_Change_Information_Response },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_HS_Cell_Change_Information_Response_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_HS_Cell_Change_Information_Response_ItemIEs, Additional_HS_Cell_Change_Information_Response_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Change_Information_Response_List_sequence_of[1] = {
  { &hf_rnsap_Additional_HS_Cell_Change_Information_Response_List_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_HS_Cell_Change_Information_Response_ItemIEs },
};

static int
dissect_rnsap_Additional_HS_Cell_Change_Information_Response_List(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_HS_Cell_Change_Information_Response_List, Additional_HS_Cell_Change_Information_Response_List_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t RadioLinkAdditionResponseTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkAdditionResponseTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkAdditionResponseTDD, RadioLinkAdditionResponseTDD_sequence);

  return offset;
}



static int
dissect_rnsap_UL_CCTrCHInformationList_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DL_CCTrCHInformationList_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Combining_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Combining_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Combining_RL_AdditionRspTDD, Combining_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t NonCombining_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_dCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_InformationResponse },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_NonCombining_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_NonCombining_RL_AdditionRspTDD, NonCombining_RL_AdditionRspTDD_sequence);

  return offset;
}


static const value_string rnsap_DiversityIndication_RL_AdditionRspTDD_vals[] = {
  {   0, "combining" },
  {   1, "nonCombining" },
  { 0, NULL }
};

static const per_choice_t DiversityIndication_RL_AdditionRspTDD_choice[] = {
  {   0, &hf_rnsap_combining_03  , ASN1_NO_EXTENSIONS     , dissect_rnsap_Combining_RL_AdditionRspTDD },
  {   1, &hf_rnsap_nonCombining_01, ASN1_NO_EXTENSIONS     , dissect_rnsap_NonCombining_RL_AdditionRspTDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DiversityIndication_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DiversityIndication_RL_AdditionRspTDD, DiversityIndication_RL_AdditionRspTDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t DCH_Information_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_diversityIndication_03, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DiversityIndication_RL_AdditionRspTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_Information_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_Information_RL_AdditionRspTDD, DCH_Information_RL_AdditionRspTDD_sequence);

  return offset;
}



static int
dissect_rnsap_DSCH_InformationResponse_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_USCH_InformationResponse_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_InformationResponse_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_uRA_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_URA_Information },
  { &hf_rnsap_sAI           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAI },
  { &hf_rnsap_gA_Cell       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_Cell },
  { &hf_rnsap_gA_AccessPointPosition, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_AccessPointPosition },
  { &hf_rnsap_ul_TimeSlot_ISCP_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeSlot_ISCP_Info },
  { &hf_rnsap_minUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_maxUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_maximumAllowedULTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaximumAllowedULTxPower },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_pCCPCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PCCPCH_Power },
  { &hf_rnsap_timingAdvanceApplied, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimingAdvanceApplied },
  { &hf_rnsap_alphaValue    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AlphaValue },
  { &hf_rnsap_ul_PhysCH_SF_Variation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_PhysCH_SF_Variation },
  { &hf_rnsap_synchronisationConfiguration, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SynchronisationConfiguration },
  { &hf_rnsap_secondary_CCPCH_Info_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Secondary_CCPCH_Info_TDD },
  { &hf_rnsap_ul_CCTrCHInformation_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_CCTrCHInformationList_RL_AdditionRspTDD },
  { &hf_rnsap_dl_CCTrCHInformation_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_CCTrCHInformationList_RL_AdditionRspTDD },
  { &hf_rnsap_dCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DCH_Information_RL_AdditionRspTDD },
  { &hf_rnsap_dSCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DSCH_InformationResponse_RL_AdditionRspTDD },
  { &hf_rnsap_uSCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_USCH_InformationResponse_RL_AdditionRspTDD },
  { &hf_rnsap_neighbouring_UMTS_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_UMTS_CellInformation },
  { &hf_rnsap_neighbouring_GSM_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_GSM_CellInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationResponse_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationResponse_RL_AdditionRspTDD, RL_InformationResponse_RL_AdditionRspTDD_sequence);

  return offset;
}



static int
dissect_rnsap_UL_DPCH_InformationList_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t UL_CCTrCHInformationItem_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_ul_DPCH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_DPCH_InformationList_RL_AdditionRspTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCHInformationItem_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCHInformationItem_RL_AdditionRspTDD, UL_CCTrCHInformationItem_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCHInformationListIE_RL_AdditionRspTDD_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_CCTrCHInformationItem_RL_AdditionRspTDD },
};

static int
dissect_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD, UL_CCTrCHInformationListIE_RL_AdditionRspTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationItem_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_uL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Timeslot_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD, UL_DPCH_InformationItem_RL_AdditionRspTDD_sequence);

  return offset;
}



static int
dissect_rnsap_DL_DPCH_InformationList_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DL_CCTrCHInformationItem_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_dl_DPCH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_DPCH_InformationList_RL_AdditionRspTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCHInformationItem_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCHInformationItem_RL_AdditionRspTDD, DL_CCTrCHInformationItem_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCHInformationListIE_RL_AdditionRspTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_CCTrCHInformationItem_RL_AdditionRspTDD },
};

static int
dissect_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD, DL_CCTrCHInformationListIE_RL_AdditionRspTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationItem_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_dL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Timeslot_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD, DL_DPCH_InformationItem_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t DiversityIndication_RL_AdditionRspTDD2_sequence[] = {
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DiversityIndication_RL_AdditionRspTDD2(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DiversityIndication_RL_AdditionRspTDD2, DiversityIndication_RL_AdditionRspTDD2_sequence);

  return offset;
}


static const per_sequence_t DSCHInformationItem_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_dsch_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_ID },
  { &hf_rnsap_transportFormatManagement, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatManagement },
  { &hf_rnsap_dSCH_FlowControlInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_FlowControlInformation },
  { &hf_rnsap_diversityIndication_04, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DiversityIndication_RL_AdditionRspTDD2 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DSCHInformationItem_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DSCHInformationItem_RL_AdditionRspTDD, DSCHInformationItem_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t DSCH_InformationListIE_RL_AdditionRspTDD_sequence_of[1] = {
  { &hf_rnsap_DSCH_InformationListIE_RL_AdditionRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCHInformationItem_RL_AdditionRspTDD },
};

static int
dissect_rnsap_DSCH_InformationListIE_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DSCH_InformationListIE_RL_AdditionRspTDD, DSCH_InformationListIE_RL_AdditionRspTDD_sequence_of,
                                                  0, maxNoOfDSCHs, FALSE);

  return offset;
}


static const per_sequence_t USCHInformationItem_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_uSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_USCH_ID },
  { &hf_rnsap_transportFormatManagement, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatManagement },
  { &hf_rnsap_diversityIndication_04, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DiversityIndication_RL_AdditionRspTDD2 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_USCHInformationItem_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_USCHInformationItem_RL_AdditionRspTDD, USCHInformationItem_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t USCH_InformationListIE_RL_AdditionRspTDD_sequence_of[1] = {
  { &hf_rnsap_USCH_InformationListIE_RL_AdditionRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_USCHInformationItem_RL_AdditionRspTDD },
};

static int
dissect_rnsap_USCH_InformationListIE_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_USCH_InformationListIE_RL_AdditionRspTDD, USCH_InformationListIE_RL_AdditionRspTDD_sequence_of,
                                                  0, maxNoOfUSCHs, FALSE);

  return offset;
}



static int
dissect_rnsap_UL_CCTrCH_LCR_InformationList_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DL_CCTrCH_LCR_InformationList_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DCH_InformationResponseList_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DSCH_LCR_InformationResponse_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_USCH_LCR_InformationResponse_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_LCR_InformationResponse_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_uRA_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_URA_Information },
  { &hf_rnsap_sAI           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAI },
  { &hf_rnsap_gA_Cell       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_Cell },
  { &hf_rnsap_gA_AccessPointPosition, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_AccessPointPosition },
  { &hf_rnsap_ul_TimeSlot_ISCP_LCR_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeSlot_ISCP_LCR_Info },
  { &hf_rnsap_maxUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_minUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_pCCPCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PCCPCH_Power },
  { &hf_rnsap_maximumAllowedULTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaximumAllowedULTxPower },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_alphaValue    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AlphaValue },
  { &hf_rnsap_ul_PhysCH_SF_Variation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_PhysCH_SF_Variation },
  { &hf_rnsap_synchronisationConfiguration, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SynchronisationConfiguration },
  { &hf_rnsap_secondary_LCR_CCPCH_Info_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Secondary_LCR_CCPCH_Info_TDD },
  { &hf_rnsap_ul_CCTrCH_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_CCTrCH_LCR_InformationList_RL_AdditionRspTDD },
  { &hf_rnsap_dl_CCTrCH_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_CCTrCH_LCR_InformationList_RL_AdditionRspTDD },
  { &hf_rnsap_dCH_InformationResponse_02, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DCH_InformationResponseList_RL_AdditionRspTDD },
  { &hf_rnsap_dsch_LCR_InformationResponse_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DSCH_LCR_InformationResponse_RL_AdditionRspTDD },
  { &hf_rnsap_usch_LCR_InformationResponse_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_USCH_LCR_InformationResponse_RL_AdditionRspTDD },
  { &hf_rnsap_neighbouring_UMTS_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_UMTS_CellInformation },
  { &hf_rnsap_neighbouring_GSM_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_GSM_CellInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_LCR_InformationResponse_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_LCR_InformationResponse_RL_AdditionRspTDD, RL_LCR_InformationResponse_RL_AdditionRspTDD_sequence);

  return offset;
}



static int
dissect_rnsap_UL_DPCH_LCR_InformationList_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t UL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_ul_DPCH_LCR_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_DPCH_LCR_InformationList_RL_AdditionRspTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD, UL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD },
};

static int
dissect_rnsap_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD, UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_sequence_of,
                                                  1, maxNrOfCCTrCHsLCR, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_uL_TimeslotLCR_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeslotLCR_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD, UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_sequence);

  return offset;
}



static int
dissect_rnsap_DL_DPCH_LCR_InformationList_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_dl_DPCH_LCR_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_DPCH_LCR_InformationList_RL_AdditionRspTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD, DL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD },
};

static int
dissect_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD, DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_sequence_of,
                                                  1, maxNrOfCCTrCHsLCR, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_dL_TimeslotLCR_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TimeslotLCR_Information },
  { &hf_rnsap_tSTD_Indicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TSTD_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD, DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t DSCH_LCR_InformationItem_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_dsch_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_ID },
  { &hf_rnsap_dSCH_FlowControlInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_FlowControlInformation },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_transportFormatManagement, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatManagement },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DSCH_LCR_InformationItem_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DSCH_LCR_InformationItem_RL_AdditionRspTDD, DSCH_LCR_InformationItem_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t DSCH_LCR_InformationListIEs_RL_AdditionRspTDD_sequence_of[1] = {
  { &hf_rnsap_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_LCR_InformationItem_RL_AdditionRspTDD },
};

static int
dissect_rnsap_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD, DSCH_LCR_InformationListIEs_RL_AdditionRspTDD_sequence_of,
                                                  0, maxNoOfDSCHsLCR, FALSE);

  return offset;
}


static const per_sequence_t USCH_LCR_InformationItem_RL_AdditionRspTDD_sequence[] = {
  { &hf_rnsap_usch_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_USCH_ID },
  { &hf_rnsap_transportFormatManagement, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatManagement },
  { &hf_rnsap_diversityIndication_04, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DiversityIndication_RL_AdditionRspTDD2 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_USCH_LCR_InformationItem_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_USCH_LCR_InformationItem_RL_AdditionRspTDD, USCH_LCR_InformationItem_RL_AdditionRspTDD_sequence);

  return offset;
}


static const per_sequence_t USCH_LCR_InformationListIEs_RL_AdditionRspTDD_sequence_of[1] = {
  { &hf_rnsap_USCH_LCR_InformationListIEs_RL_AdditionRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_USCH_LCR_InformationItem_RL_AdditionRspTDD },
};

static int
dissect_rnsap_USCH_LCR_InformationListIEs_RL_AdditionRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_USCH_LCR_InformationListIEs_RL_AdditionRspTDD, USCH_LCR_InformationListIEs_RL_AdditionRspTDD_sequence_of,
                                                  0, maxNoOfUSCHsLCR, FALSE);

  return offset;
}



static int
dissect_rnsap_UL_CCTrCHInformationList_RL_AdditionRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DL_CCTrCHInformationList_RL_AdditionRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_InformationResponse_RL_AdditionRspTDD768_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_uRA_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_URA_Information },
  { &hf_rnsap_sAI           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAI },
  { &hf_rnsap_gA_Cell       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_Cell },
  { &hf_rnsap_gA_AccessPointPosition, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_AccessPointPosition },
  { &hf_rnsap_ul_TimeSlot_ISCP_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeSlot_ISCP_Info },
  { &hf_rnsap_minUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_maxUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_maximumAllowedULTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaximumAllowedULTxPower },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_pCCPCH_Power  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PCCPCH_Power },
  { &hf_rnsap_timingAdvanceApplied, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimingAdvanceApplied },
  { &hf_rnsap_alphaValue    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_AlphaValue },
  { &hf_rnsap_ul_PhysCH_SF_Variation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_PhysCH_SF_Variation },
  { &hf_rnsap_synchronisationConfiguration, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SynchronisationConfiguration },
  { &hf_rnsap_secondary_CCPCH_Info_TDD768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Secondary_CCPCH_Info_TDD768 },
  { &hf_rnsap_ul_CCTrCHInformation768_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_CCTrCHInformationList_RL_AdditionRspTDD768 },
  { &hf_rnsap_dl_CCTrCHInformation768_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_CCTrCHInformationList_RL_AdditionRspTDD768 },
  { &hf_rnsap_dCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DCH_Information_RL_AdditionRspTDD },
  { &hf_rnsap_dSCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DSCH_InformationResponse_RL_AdditionRspTDD },
  { &hf_rnsap_uSCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_USCH_InformationResponse_RL_AdditionRspTDD },
  { &hf_rnsap_neighbouring_UMTS_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_UMTS_CellInformation },
  { &hf_rnsap_neighbouring_GSM_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_GSM_CellInformation },
  { &hf_rnsap_gA_CellAdditionalShapes, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_CellAdditionalShapes },
  { &hf_rnsap_hCS_Prio      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HCS_Prio },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationResponse_RL_AdditionRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationResponse_RL_AdditionRspTDD768, RL_InformationResponse_RL_AdditionRspTDD768_sequence);

  return offset;
}



static int
dissect_rnsap_UL_DPCH_InformationList_RL_AdditionRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t UL_CCTrCHInformationItem_RL_AdditionRspTDD768_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_ul_DPCH_Information768_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_DPCH_InformationList_RL_AdditionRspTDD768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCHInformationItem_RL_AdditionRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCHInformationItem_RL_AdditionRspTDD768, UL_CCTrCHInformationItem_RL_AdditionRspTDD768_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCHInformationListIE_RL_AdditionRspTDD768_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_CCTrCHInformationItem_RL_AdditionRspTDD768 },
};

static int
dissect_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD768, UL_CCTrCHInformationListIE_RL_AdditionRspTDD768_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationItem_RL_AdditionRspTDD768_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_uL_Timeslot_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Timeslot_Information768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD768, UL_DPCH_InformationItem_RL_AdditionRspTDD768_sequence);

  return offset;
}



static int
dissect_rnsap_DL_DPCH_InformationList_RL_AdditionRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DL_CCTrCHInformationItem_RL_AdditionRspTDD768_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_dl_DPCH_Information768_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_DPCH_InformationList_RL_AdditionRspTDD768 },
  { &hf_rnsap_cCTrCH_Maximum_DL_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_cCTrCH_Minimum_DL_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCHInformationItem_RL_AdditionRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCHInformationItem_RL_AdditionRspTDD768, DL_CCTrCHInformationItem_RL_AdditionRspTDD768_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCHInformationListIE_RL_AdditionRspTDD768_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_CCTrCHInformationItem_RL_AdditionRspTDD768 },
};

static int
dissect_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD768, DL_CCTrCHInformationListIE_RL_AdditionRspTDD768_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationItem_RL_AdditionRspTDD768_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_dL_Timeslot_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Timeslot_Information768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD768, DL_DPCH_InformationItem_RL_AdditionRspTDD768_sequence);

  return offset;
}


static const per_sequence_t RadioLinkAdditionFailureFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkAdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkAdditionFailureFDD, RadioLinkAdditionFailureFDD_sequence);

  return offset;
}


static const per_sequence_t GeneralCauseList_RL_AdditionFailureFDD_sequence[] = {
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GeneralCauseList_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GeneralCauseList_RL_AdditionFailureFDD, GeneralCauseList_RL_AdditionFailureFDD_sequence);

  return offset;
}


static const per_sequence_t UnsuccessfulRL_InformationResponseList_RL_AdditionFailureFDD_sequence_of[1] = {
  { &hf_rnsap_UnsuccessfulRL_InformationResponseList_RL_AdditionFailureFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_UnsuccessfulRL_InformationResponseList_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UnsuccessfulRL_InformationResponseList_RL_AdditionFailureFDD, UnsuccessfulRL_InformationResponseList_RL_AdditionFailureFDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t SuccessfulRL_InformationResponseList_RL_AdditionFailureFDD_sequence_of[1] = {
  { &hf_rnsap_SuccessfulRL_InformationResponseList_RL_AdditionFailureFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_SuccessfulRL_InformationResponseList_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_SuccessfulRL_InformationResponseList_RL_AdditionFailureFDD, SuccessfulRL_InformationResponseList_RL_AdditionFailureFDD_sequence_of,
                                                  0, maxNrOfRLs_2, FALSE);

  return offset;
}


static const per_sequence_t RLSpecificCauseList_RL_AdditionFailureFDD_sequence[] = {
  { &hf_rnsap_unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UnsuccessfulRL_InformationResponseList_RL_AdditionFailureFDD },
  { &hf_rnsap_successful_RL_InformationRespList_RL_AdditionFailureFDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SuccessfulRL_InformationResponseList_RL_AdditionFailureFDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RLSpecificCauseList_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RLSpecificCauseList_RL_AdditionFailureFDD, RLSpecificCauseList_RL_AdditionFailureFDD_sequence);

  return offset;
}


static const value_string rnsap_CauseLevel_RL_AdditionFailureFDD_vals[] = {
  {   0, "generalCause" },
  {   1, "rLSpecificCause" },
  { 0, NULL }
};

static const per_choice_t CauseLevel_RL_AdditionFailureFDD_choice[] = {
  {   0, &hf_rnsap_generalCause_02, ASN1_EXTENSION_ROOT    , dissect_rnsap_GeneralCauseList_RL_AdditionFailureFDD },
  {   1, &hf_rnsap_rLSpecificCause_02, ASN1_EXTENSION_ROOT    , dissect_rnsap_RLSpecificCauseList_RL_AdditionFailureFDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_CauseLevel_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_CauseLevel_RL_AdditionFailureFDD, CauseLevel_RL_AdditionFailureFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD, UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD_sequence);

  return offset;
}



static int
dissect_rnsap_DL_CodeInformationList_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t Combining_RL_AdditionFailureFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Combining_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Combining_RL_AdditionFailureFDD, Combining_RL_AdditionFailureFDD_sequence);

  return offset;
}


static const per_sequence_t NonCombining_RL_AdditionFailureFDD_sequence[] = {
  { &hf_rnsap_dCH_InformationResponse, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_InformationResponse },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_NonCombining_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_NonCombining_RL_AdditionFailureFDD, NonCombining_RL_AdditionFailureFDD_sequence);

  return offset;
}


static const value_string rnsap_DiversityIndication_RL_AdditionFailureFDD_vals[] = {
  {   0, "combining" },
  {   1, "nonCombining" },
  { 0, NULL }
};

static const per_choice_t DiversityIndication_RL_AdditionFailureFDD_choice[] = {
  {   0, &hf_rnsap_combining_04  , ASN1_NO_EXTENSIONS     , dissect_rnsap_Combining_RL_AdditionFailureFDD },
  {   1, &hf_rnsap_nonCombining_02, ASN1_NO_EXTENSIONS     , dissect_rnsap_NonCombining_RL_AdditionFailureFDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DiversityIndication_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DiversityIndication_RL_AdditionFailureFDD, DiversityIndication_RL_AdditionFailureFDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t SuccessfulRL_InformationResponse_RL_AdditionFailureFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_rL_Set_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_uRA_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_URA_Information },
  { &hf_rnsap_sAI           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SAI },
  { &hf_rnsap_gA_Cell       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_Cell },
  { &hf_rnsap_gA_AccessPointPosition, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_GA_AccessPointPosition },
  { &hf_rnsap_received_total_wide_band_power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Received_total_wide_band_power },
  { &hf_rnsap_not_Used_secondary_CCPCH_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_dl_CodeInformation_02, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_CodeInformationList_RL_AdditionFailureFDD },
  { &hf_rnsap_diversityIndication_05, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DiversityIndication_RL_AdditionFailureFDD },
  { &hf_rnsap_sSDT_SupportIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SSDT_SupportIndicator },
  { &hf_rnsap_minUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_maxUL_SIR     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_SIR },
  { &hf_rnsap_closedlooptimingadjustmentmode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Closedlooptimingadjustmentmode },
  { &hf_rnsap_maximumAllowedULTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MaximumAllowedULTxPower },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_neighbouring_UMTS_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_UMTS_CellInformation },
  { &hf_rnsap_neighbouring_GSM_CellInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Neighbouring_GSM_CellInformation },
  { &hf_rnsap_primaryCPICH_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PrimaryCPICH_Power },
  { &hf_rnsap_pC_Preamble   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PC_Preamble },
  { &hf_rnsap_sRB_Delay     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_SRB_Delay },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SuccessfulRL_InformationResponse_RL_AdditionFailureFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SuccessfulRL_InformationResponse_RL_AdditionFailureFDD, SuccessfulRL_InformationResponse_RL_AdditionFailureFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkAdditionFailureTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkAdditionFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkAdditionFailureTDD, RadioLinkAdditionFailureTDD_sequence);

  return offset;
}


static const per_sequence_t GeneralCauseList_RL_AdditionFailureTDD_sequence[] = {
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GeneralCauseList_RL_AdditionFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GeneralCauseList_RL_AdditionFailureTDD, GeneralCauseList_RL_AdditionFailureTDD_sequence);

  return offset;
}



static int
dissect_rnsap_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RLSpecificCauseList_RL_AdditionFailureTDD_sequence[] = {
  { &hf_rnsap_unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RLSpecificCauseList_RL_AdditionFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RLSpecificCauseList_RL_AdditionFailureTDD, RLSpecificCauseList_RL_AdditionFailureTDD_sequence);

  return offset;
}


static const value_string rnsap_CauseLevel_RL_AdditionFailureTDD_vals[] = {
  {   0, "generalCause" },
  {   1, "rLSpecificCause" },
  { 0, NULL }
};

static const per_choice_t CauseLevel_RL_AdditionFailureTDD_choice[] = {
  {   0, &hf_rnsap_generalCause_03, ASN1_EXTENSION_ROOT    , dissect_rnsap_GeneralCauseList_RL_AdditionFailureTDD },
  {   1, &hf_rnsap_rLSpecificCause_03, ASN1_EXTENSION_ROOT    , dissect_rnsap_RLSpecificCauseList_RL_AdditionFailureTDD },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_CauseLevel_RL_AdditionFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_CauseLevel_RL_AdditionFailureTDD, CauseLevel_RL_AdditionFailureTDD_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD, UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkDeletionRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkDeletionRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkDeletionRequest, RadioLinkDeletionRequest_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_DeletionRqst_sequence_of[1] = {
  { &hf_rnsap_RL_InformationList_RL_DeletionRqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationList_RL_DeletionRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationList_RL_DeletionRqst, RL_InformationList_RL_DeletionRqst_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_Information_RL_DeletionRqst_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Information_RL_DeletionRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Information_RL_DeletionRqst, RL_Information_RL_DeletionRqst_sequence);

  return offset;
}


static const per_sequence_t RadioLinkDeletionResponse_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkDeletionResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkDeletionResponse, RadioLinkDeletionResponse_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationPrepareFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkReconfigurationPrepareFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkReconfigurationPrepareFDD, RadioLinkReconfigurationPrepareFDD_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_Information_RL_ReconfPrepFDD_sequence[] = {
  { &hf_rnsap_ul_ScramblingCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_ScramblingCode },
  { &hf_rnsap_ul_SIRTarget  , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_minUL_ChannelisationCodeLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MinUL_ChannelisationCodeLength },
  { &hf_rnsap_maxNrOfUL_DPDCHs, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MaxNrOfUL_DPCHs },
  { &hf_rnsap_ul_PunctureLimit, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PunctureLimit },
  { &hf_rnsap_tFCS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCS },
  { &hf_rnsap_ul_DPCCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_DPCCH_SlotFormat },
  { &hf_rnsap_diversityMode , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DiversityMode },
  { &hf_rnsap_not_Used_sSDT_CellIDLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_not_Used_s_FieldLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_Information_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_Information_RL_ReconfPrepFDD, UL_DPCH_Information_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_Information_RL_ReconfPrepFDD_sequence[] = {
  { &hf_rnsap_tFCS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCS },
  { &hf_rnsap_dl_DPCH_SlotFormat, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_DPCH_SlotFormat },
  { &hf_rnsap_nrOfDLchannelisationcodes, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NrOfDLchannelisationcodes },
  { &hf_rnsap_tFCI_SignallingMode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_SignallingMode },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_multiplexingPosition, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MultiplexingPosition },
  { &hf_rnsap_limitedPowerIncrease, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_LimitedPowerIncrease },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_Information_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_Information_RL_ReconfPrepFDD, DL_DPCH_Information_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t PowerOffsetInformation_RL_ReconfPrepFDD_sequence[] = {
  { &hf_rnsap_pO1_ForTFCI_Bits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerOffset },
  { &hf_rnsap_pO2_ForTPC_Bits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerOffset },
  { &hf_rnsap_pO3_ForPilotBits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerOffset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PowerOffsetInformation_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PowerOffsetInformation_RL_ReconfPrepFDD, PowerOffsetInformation_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_Power_Information_RL_ReconfPrepFDD_sequence[] = {
  { &hf_rnsap_powerOffsetInformation_02, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerOffsetInformation_RL_ReconfPrepFDD },
  { &hf_rnsap_fdd_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_TPC_DownlinkStepSize },
  { &hf_rnsap_innerLoopDLPCStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_InnerLoopDLPCStatus },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_Power_Information_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_Power_Information_RL_ReconfPrepFDD, DL_DPCH_Power_Information_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteItem_RL_ReconfPrepFDD_sequence[] = {
  { &hf_rnsap_dCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_DeleteItem_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_DeleteItem_RL_ReconfPrepFDD, DCH_DeleteItem_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteList_RL_ReconfPrepFDD_sequence_of[1] = {
  { &hf_rnsap_DCH_DeleteList_RL_ReconfPrepFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_DeleteItem_RL_ReconfPrepFDD },
};

static int
dissect_rnsap_DCH_DeleteList_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DCH_DeleteList_RL_ReconfPrepFDD, DCH_DeleteList_RL_ReconfPrepFDD_sequence_of,
                                                  0, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_ReconfPrepFDD_sequence_of[1] = {
  { &hf_rnsap_RL_InformationList_RL_ReconfPrepFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationList_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationList_RL_ReconfPrepFDD, RL_InformationList_RL_ReconfPrepFDD_sequence_of,
                                                  0, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_Information_RL_ReconfPrepFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_not_Used_sSDT_Indication, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_not_Used_sSDT_CellIdentity, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_transmitDiversityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransmitDiversityIndicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Information_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Information_RL_ReconfPrepFDD, RL_Information_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs_sequence[] = {
  { &hf_rnsap_hSPDSCH_RL_ID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_C_ID },
  { &hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_DSCH_FDD_Secondary_Serving_Information },
  { &hf_rnsap_hS_DSCH_Secondary_Serving_Information_To_Modify, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_DSCH_Secondary_Serving_Information_To_Modify },
  { &hf_rnsap_hS_HS_DSCH_Secondary_Serving_Remove, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_DSCH_Secondary_Serving_Remove },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs, Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Reconf_Prep_sequence_of[1] = {
  { &hf_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs },
};

static int
dissect_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep, Additional_HS_Cell_Information_RL_Reconf_Prep_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_RL_Reconf_Prep_sequence[] = {
  { &hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Prep(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Prep, Additional_EDCH_Cell_Information_RL_Reconf_Prep_sequence);

  return offset;
}


static const per_sequence_t PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD_sequence[] = {
  { &hf_rnsap_po2_ForTPC_Bits, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerOffset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD, PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t F_DPCH_Information_RL_ReconfPrepFDD_sequence[] = {
  { &hf_rnsap_powerOffsetInformation_03, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD },
  { &hf_rnsap_fdd_dl_TPC_DownlinkStepSize, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FDD_TPC_DownlinkStepSize },
  { &hf_rnsap_limitedPowerIncrease, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_LimitedPowerIncrease },
  { &hf_rnsap_innerLoopDLPCStatus, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_InnerLoopDLPCStatus },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_F_DPCH_Information_RL_ReconfPrepFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_F_DPCH_Information_RL_ReconfPrepFDD, F_DPCH_Information_RL_ReconfPrepFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationPrepareTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkReconfigurationPrepareTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkReconfigurationPrepareTDD, RadioLinkReconfigurationPrepareTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD, UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_sequence_of,
                                                  0, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_CCTrCH_AddInformation_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_tFCS          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCS },
  { &hf_rnsap_tFCI_Coding   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Coding },
  { &hf_rnsap_punctureLimit , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PunctureLimit },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCH_AddInformation_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCH_AddInformation_RL_ReconfPrepTDD, UL_CCTrCH_AddInformation_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD, UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  0, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_tFCS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCS },
  { &hf_rnsap_tFCI_Coding   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Coding },
  { &hf_rnsap_punctureLimit , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PunctureLimit },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD, UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD, UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_sequence_of,
                                                  0, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD, UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD, DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_sequence_of,
                                                  0, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t CCTrCH_TPCAddItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CCTrCH_TPCAddItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CCTrCH_TPCAddItem_RL_ReconfPrepTDD, CCTrCH_TPCAddItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_TPCAddList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_CCTrCH_TPCAddList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_TPCAddItem_RL_ReconfPrepTDD },
};

static int
dissect_rnsap_CCTrCH_TPCAddList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_CCTrCH_TPCAddList_RL_ReconfPrepTDD, CCTrCH_TPCAddList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_tFCS          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCS },
  { &hf_rnsap_tFCI_Coding   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TFCI_Coding },
  { &hf_rnsap_punctureLimit , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PunctureLimit },
  { &hf_rnsap_cCTrCH_TPCList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CCTrCH_TPCAddList_RL_ReconfPrepTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD, DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD, DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  0, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t CCTrCH_TPCModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CCTrCH_TPCModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CCTrCH_TPCModifyItem_RL_ReconfPrepTDD, CCTrCH_TPCModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_TPCModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_TPCModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_rnsap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD, CCTrCH_TPCModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_tFCS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCS },
  { &hf_rnsap_tFCI_Coding   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Coding },
  { &hf_rnsap_punctureLimit , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_PunctureLimit },
  { &hf_rnsap_cCTrCH_TPCList_02, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD, DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD, DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_sequence_of,
                                                  0, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD, DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_dCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_DeleteItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_DeleteItem_RL_ReconfPrepTDD, DCH_DeleteItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_DCH_DeleteList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_DeleteItem_RL_ReconfPrepTDD },
};

static int
dissect_rnsap_DCH_DeleteList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DCH_DeleteList_RL_ReconfPrepTDD, DCH_DeleteList_RL_ReconfPrepTDD_sequence_of,
                                                  0, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t DSCH_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_dSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_ID },
  { &hf_rnsap_dl_ccTrCHID   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_trChSourceStatisticsDescriptor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TrCH_SrcStatisticsDescr },
  { &hf_rnsap_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_bLER          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BLER },
  { &hf_rnsap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportBearerRequestIndicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DSCH_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DSCH_ModifyItem_RL_ReconfPrepTDD, DSCH_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DSCH_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_DSCH_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_rnsap_DSCH_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DSCH_ModifyList_RL_ReconfPrepTDD, DSCH_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  0, maxNoOfDSCHs, FALSE);

  return offset;
}


static const per_sequence_t DSCH_DeleteItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_dSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DSCH_DeleteItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DSCH_DeleteItem_RL_ReconfPrepTDD, DSCH_DeleteItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t DSCH_DeleteList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_DSCH_DeleteList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_DeleteItem_RL_ReconfPrepTDD },
};

static int
dissect_rnsap_DSCH_DeleteList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DSCH_DeleteList_RL_ReconfPrepTDD, DSCH_DeleteList_RL_ReconfPrepTDD_sequence_of,
                                                  0, maxNoOfDSCHs, FALSE);

  return offset;
}


static const per_sequence_t USCH_ModifyItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_uSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_USCH_ID },
  { &hf_rnsap_ul_ccTrCHID   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_trChSourceStatisticsDescriptor, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TrCH_SrcStatisticsDescr },
  { &hf_rnsap_transportFormatSet, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportFormatSet },
  { &hf_rnsap_allocationRetentionPriority, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_AllocationRetentionPriority },
  { &hf_rnsap_schedulingPriorityIndicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_SchedulingPriorityIndicator },
  { &hf_rnsap_bLER          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BLER },
  { &hf_rnsap_transportBearerRequestIndicator, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportBearerRequestIndicator },
  { &hf_rnsap_rb_Info       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RB_Info },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_USCH_ModifyItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_USCH_ModifyItem_RL_ReconfPrepTDD, USCH_ModifyItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t USCH_ModifyList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_USCH_ModifyList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_USCH_ModifyItem_RL_ReconfPrepTDD },
};

static int
dissect_rnsap_USCH_ModifyList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_USCH_ModifyList_RL_ReconfPrepTDD, USCH_ModifyList_RL_ReconfPrepTDD_sequence_of,
                                                  0, maxNoOfUSCHs, FALSE);

  return offset;
}


static const per_sequence_t USCH_DeleteItem_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_uSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_USCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_USCH_DeleteItem_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_USCH_DeleteItem_RL_ReconfPrepTDD, USCH_DeleteItem_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t USCH_DeleteList_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_USCH_DeleteList_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_USCH_DeleteItem_RL_ReconfPrepTDD },
};

static int
dissect_rnsap_USCH_DeleteList_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_USCH_DeleteList_RL_ReconfPrepTDD, USCH_DeleteList_RL_ReconfPrepTDD_sequence_of,
                                                  0, maxNoOfUSCHs, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationIE_RL_ReconfPrepTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_rL_Specific_DCH_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RL_Specific_DCH_Info },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationIE_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationIE_RL_ReconfPrepTDD, RL_InformationIE_RL_ReconfPrepTDD_sequence);

  return offset;
}


static const per_sequence_t RL_Information_RL_ReconfPrepTDD_sequence_of[1] = {
  { &hf_rnsap_RL_Information_RL_ReconfPrepTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_InformationIE_RL_ReconfPrepTDD },
};

static int
dissect_rnsap_RL_Information_RL_ReconfPrepTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Information_RL_ReconfPrepTDD, RL_Information_RL_ReconfPrepTDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationReadyFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkReconfigurationReadyFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkReconfigurationReadyFDD, RadioLinkReconfigurationReadyFDD_sequence);

  return offset;
}


static const per_sequence_t RL_InformationResponseList_RL_ReconfReadyFDD_sequence_of[1] = {
  { &hf_rnsap_RL_InformationResponseList_RL_ReconfReadyFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationResponseList_RL_ReconfReadyFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationResponseList_RL_ReconfReadyFDD, RL_InformationResponseList_RL_ReconfReadyFDD_sequence_of,
                                                  0, maxNrOfRLs, FALSE);

  return offset;
}



static int
dissect_rnsap_DL_CodeInformationList_RL_ReconfReadyFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DCH_InformationResponseList_RL_ReconfReadyFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_InformationResponseItem_RL_ReconfReadyFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_max_UL_SIR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_min_UL_SIR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_not_Used_secondary_CCPCH_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_dl_CodeInformationList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_CodeInformationList_RL_ReconfReadyFDD },
  { &hf_rnsap_dCHInformationResponse, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DCH_InformationResponseList_RL_ReconfReadyFDD },
  { &hf_rnsap_not_Used_dSCHsToBeAddedOrModified, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationResponseItem_RL_ReconfReadyFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationResponseItem_RL_ReconfReadyFDD, RL_InformationResponseItem_RL_ReconfReadyFDD_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_RL_Reconf_Response_ItemIEs_sequence[] = {
  { &hf_rnsap_hSPDSCH_RL_ID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_hSDSCH_RNTI   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_RNTI },
  { &hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information_Response, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_DSCH_FDD_Secondary_Serving_Information_Response },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_HS_Cell_RL_Reconf_Response_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_HS_Cell_RL_Reconf_Response_ItemIEs, Additional_HS_Cell_RL_Reconf_Response_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_RL_Reconf_Response_sequence_of[1] = {
  { &hf_rnsap_Additional_HS_Cell_RL_Reconf_Response_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_HS_Cell_RL_Reconf_Response_ItemIEs },
};

static int
dissect_rnsap_Additional_HS_Cell_RL_Reconf_Response(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_HS_Cell_RL_Reconf_Response, Additional_HS_Cell_RL_Reconf_Response_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationReadyTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkReconfigurationReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkReconfigurationReadyTDD, RadioLinkReconfigurationReadyTDD_sequence);

  return offset;
}



static int
dissect_rnsap_UL_CCTrCH_InformationList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DL_CCTrCH_InformationList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DCH_InformationResponseList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DSCHToBeAddedOrModified_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_USCHToBeAddedOrModified_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_InformationResponse_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_max_UL_SIR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_min_UL_SIR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_secondary_CCPCH_Info_TDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Secondary_CCPCH_Info_TDD },
  { &hf_rnsap_ul_CCTrCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_CCTrCH_InformationList_RL_ReconfReadyTDD },
  { &hf_rnsap_dl_CCTrCH_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_CCTrCH_InformationList_RL_ReconfReadyTDD },
  { &hf_rnsap_dCHInformationResponse_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DCH_InformationResponseList_RL_ReconfReadyTDD },
  { &hf_rnsap_dSCHsToBeAddedOrModified, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DSCHToBeAddedOrModified_RL_ReconfReadyTDD },
  { &hf_rnsap_uSCHsToBeAddedOrModified, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_USCHToBeAddedOrModified_RL_ReconfReadyTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationResponse_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationResponse_RL_ReconfReadyTDD, RL_InformationResponse_RL_ReconfReadyTDD_sequence);

  return offset;
}



static int
dissect_rnsap_UL_DPCH_InformationAddList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_UL_DPCH_InformationModifyList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_UL_DPCH_InformationDeleteList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_ul_DPCH_AddInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_DPCH_InformationAddList_RL_ReconfReadyTDD },
  { &hf_rnsap_ul_DPCH_ModifyInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_DPCH_InformationModifyList_RL_ReconfReadyTDD },
  { &hf_rnsap_ul_DPCH_DeleteInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_DPCH_InformationDeleteList_RL_ReconfReadyTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCH_InformationItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCH_InformationItem_RL_ReconfReadyTDD, UL_CCTrCH_InformationItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCHInformationListIE_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCHInformationListIE_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_CCTrCH_InformationItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_UL_CCTrCHInformationListIE_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCHInformationListIE_RL_ReconfReadyTDD, UL_CCTrCHInformationListIE_RL_ReconfReadyTDD_sequence_of,
                                                  0, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_uL_TimeslotLCR_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeslotLCR_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD, UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_rxTimingDeviationForTA, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RxTimingDeviationForTA },
  { &hf_rnsap_uL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Timeslot_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD, UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationAddList_RL_ReconfReadyTDD768_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_rxTimingDeviationForTA768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RxTimingDeviationForTA768 },
  { &hf_rnsap_uL_Timeslot_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Timeslot_Information768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_InformationAddList_RL_ReconfReadyTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_InformationAddList_RL_ReconfReadyTDD768, UL_DPCH_InformationAddList_RL_ReconfReadyTDD768_sequence);

  return offset;
}


static const per_sequence_t TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD, TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD, TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD_sequence_of,
                                                  1, maxNrOfDPCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_uL_Code_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD, UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD, UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_uL_Timeslot_InformationModifyList_RL_ReconfReadyTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD, UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t TDD_UL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_tDD_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_ChannelisationCodeLCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_UL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_UL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD, TDD_UL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t TDD_UL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_TDD_UL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_UL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_TDD_UL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_UL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD, TDD_UL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD_sequence_of,
                                                  1, maxNrOfDPCHsLCR, FALSE);

  return offset;
}


static const per_sequence_t UL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_tDD_uL_Code_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_UL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD, UL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD, UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_sequence_of,
                                                  1, maxNrOfTsLCR, FALSE);

  return offset;
}


static const per_sequence_t TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD768_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_tDD_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_ChannelisationCode768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD768, TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD768_sequence);

  return offset;
}


static const per_sequence_t TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD768_sequence_of[1] = {
  { &hf_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD768 },
};

static int
dissect_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD768, TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD768_sequence_of,
                                                  1, maxNrOfDPCHs768, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_uL_Code_Information768_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768, UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_sequence_of[1] = {
  { &hf_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768 },
};

static int
dissect_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768, UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD, UL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD, UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_sequence_of,
                                                  0, maxNrOfDPCHs, FALSE);

  return offset;
}



static int
dissect_rnsap_DL_DPCH_InformationAddList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DL_DPCH_InformationModifyList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DL_DPCH_InformationDeleteList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_dl_DPCH_AddInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_DPCH_InformationAddList_RL_ReconfReadyTDD },
  { &hf_rnsap_dl_DPCH_ModifyInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_DPCH_InformationModifyList_RL_ReconfReadyTDD },
  { &hf_rnsap_dl_DPCH_DeleteInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_DPCH_InformationDeleteList_RL_ReconfReadyTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCH_InformationItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCH_InformationItem_RL_ReconfReadyTDD, DL_CCTrCH_InformationItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCHInformationListIE_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCHInformationListIE_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_CCTrCH_InformationItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_DL_CCTrCHInformationListIE_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCHInformationListIE_RL_ReconfReadyTDD, DL_CCTrCHInformationListIE_RL_ReconfReadyTDD_sequence_of,
                                                  0, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_dL_TimeslotLCR_Info, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TimeslotLCR_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD, DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_dL_Timeslot_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Timeslot_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD, DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationAddList_RL_ReconfReadyTDD768_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_dL_Timeslot_Information768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Timeslot_Information768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_InformationAddList_RL_ReconfReadyTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_InformationAddList_RL_ReconfReadyTDD768, DL_DPCH_InformationAddList_RL_ReconfReadyTDD768_sequence);

  return offset;
}


static const per_sequence_t TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_tDD_ChannelisationCode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_ChannelisationCode },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD, TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD, TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD_sequence_of,
                                                  1, maxNrOfDPCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_dL_Code_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD, DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD, DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_dL_Timeslot_InformationModifyList_RL_ReconfReadyTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD, DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t TDD_DL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_tDD_ChannelisationCodeLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_ChannelisationCodeLCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_DL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_DL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD, TDD_DL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t TDD_DL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_TDD_DL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_TDD_DL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_DL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD, TDD_DL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD_sequence_of,
                                                  1, maxNrOfDPCHsLCR, FALSE);

  return offset;
}


static const per_sequence_t DL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_tDD_dL_Code_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_DL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD, DL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD, DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_sequence_of,
                                                  1, maxNrOfTsLCR, FALSE);

  return offset;
}


static const per_sequence_t TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD768_sequence[] = {
  { &hf_rnsap_dPCH_ID768    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID768 },
  { &hf_rnsap_tDD_ChannelisationCode768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_ChannelisationCode768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD768, TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD768_sequence);

  return offset;
}


static const per_sequence_t TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD768_sequence_of[1] = {
  { &hf_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD768 },
};

static int
dissect_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD768, TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD768_sequence_of,
                                                  1, maxNrOfDPCHs768, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_dL_Code_Information768_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768, DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_sequence_of[1] = {
  { &hf_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768 },
};

static int
dissect_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768, DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD, DL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD, DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_sequence_of,
                                                  0, maxNrOfDPCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationDeleteItem768_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_dPCH_ID768    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_InformationDeleteItem768_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_InformationDeleteItem768_RL_ReconfReadyTDD, DL_DPCH_InformationDeleteItem768_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_DPCH_InformationDeleteItem768_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD, DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD_sequence_of,
                                                  0, maxNrOfDPCHs768, FALSE);

  return offset;
}


static const per_sequence_t DSCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_dsch_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_ID },
  { &hf_rnsap_transportFormatManagement, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatManagement },
  { &hf_rnsap_dSCH_FlowControlInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCH_FlowControlInformation },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DSCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DSCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD, DSCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DSCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD, DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_sequence_of,
                                                  0, maxNoOfDSCHs, FALSE);

  return offset;
}


static const per_sequence_t USCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD_sequence[] = {
  { &hf_rnsap_uSCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_USCH_ID },
  { &hf_rnsap_transportFormatManagement, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TransportFormatManagement },
  { &hf_rnsap_bindingID     , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_BindingID },
  { &hf_rnsap_transportLayerAddress, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TransportLayerAddress },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_USCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_USCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD, USCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD_sequence);

  return offset;
}


static const per_sequence_t USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_USCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD, USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_sequence_of,
                                                  0, maxNoOfUSCHs, FALSE);

  return offset;
}


static const per_sequence_t Multiple_RL_InformationResponse_RL_ReconfReadyTDD_sequence_of[1] = {
  { &hf_rnsap_Multiple_RL_InformationResponse_RL_ReconfReadyTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_InformationResponse_RL_ReconfReadyTDD },
};

static int
dissect_rnsap_Multiple_RL_InformationResponse_RL_ReconfReadyTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Multiple_RL_InformationResponse_RL_ReconfReadyTDD, Multiple_RL_InformationResponse_RL_ReconfReadyTDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationCommit_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkReconfigurationCommit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkReconfigurationCommit, RadioLinkReconfigurationCommit_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationFailure_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkReconfigurationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkReconfigurationFailure, RadioLinkReconfigurationFailure_sequence);

  return offset;
}


static const per_sequence_t GeneralCauseList_RL_ReconfFailure_sequence[] = {
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GeneralCauseList_RL_ReconfFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GeneralCauseList_RL_ReconfFailure, GeneralCauseList_RL_ReconfFailure_sequence);

  return offset;
}


static const per_sequence_t RL_ReconfigurationFailureList_RL_ReconfFailure_sequence_of[1] = {
  { &hf_rnsap_RL_ReconfigurationFailureList_RL_ReconfFailure_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_ReconfigurationFailureList_RL_ReconfFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_ReconfigurationFailureList_RL_ReconfFailure, RL_ReconfigurationFailureList_RL_ReconfFailure_sequence_of,
                                                  0, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RLSpecificCauseList_RL_ReconfFailure_sequence[] = {
  { &hf_rnsap_rL_ReconfigurationFailureList_RL_ReconfFailure, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RL_ReconfigurationFailureList_RL_ReconfFailure },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RLSpecificCauseList_RL_ReconfFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RLSpecificCauseList_RL_ReconfFailure, RLSpecificCauseList_RL_ReconfFailure_sequence);

  return offset;
}


static const value_string rnsap_CauseLevel_RL_ReconfFailure_vals[] = {
  {   0, "generalCause" },
  {   1, "rLSpecificCause" },
  { 0, NULL }
};

static const per_choice_t CauseLevel_RL_ReconfFailure_choice[] = {
  {   0, &hf_rnsap_generalCause_04, ASN1_EXTENSION_ROOT    , dissect_rnsap_GeneralCauseList_RL_ReconfFailure },
  {   1, &hf_rnsap_rLSpecificCause_04, ASN1_EXTENSION_ROOT    , dissect_rnsap_RLSpecificCauseList_RL_ReconfFailure },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_CauseLevel_RL_ReconfFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_CauseLevel_RL_ReconfFailure, CauseLevel_RL_ReconfFailure_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_ReconfigurationFailure_RL_ReconfFail_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_ReconfigurationFailure_RL_ReconfFail(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_ReconfigurationFailure_RL_ReconfFail, RL_ReconfigurationFailure_RL_ReconfFail_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationCancel_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkReconfigurationCancel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkReconfigurationCancel, RadioLinkReconfigurationCancel_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationRequestFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkReconfigurationRequestFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkReconfigurationRequestFDD, RadioLinkReconfigurationRequestFDD_sequence);

  return offset;
}


static const per_sequence_t UL_DPCH_Information_RL_ReconfRqstFDD_sequence[] = {
  { &hf_rnsap_tFCS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCS },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_Information_RL_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_Information_RL_ReconfRqstFDD, UL_DPCH_Information_RL_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t DL_DPCH_Information_RL_ReconfRqstFDD_sequence[] = {
  { &hf_rnsap_tFCS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCS },
  { &hf_rnsap_tFCI_SignallingMode, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_SignallingMode },
  { &hf_rnsap_limitedPowerIncrease, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_LimitedPowerIncrease },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_Information_RL_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_Information_RL_ReconfRqstFDD, DL_DPCH_Information_RL_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteItem_RL_ReconfRqstFDD_sequence[] = {
  { &hf_rnsap_dCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_DeleteItem_RL_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_DeleteItem_RL_ReconfRqstFDD, DCH_DeleteItem_RL_ReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteList_RL_ReconfRqstFDD_sequence_of[1] = {
  { &hf_rnsap_DCH_DeleteList_RL_ReconfRqstFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_DeleteItem_RL_ReconfRqstFDD },
};

static int
dissect_rnsap_DCH_DeleteList_RL_ReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DCH_DeleteList_RL_ReconfRqstFDD, DCH_DeleteList_RL_ReconfRqstFDD_sequence_of,
                                                  0, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs_sequence[] = {
  { &hf_rnsap_hSPDSCH_RL_ID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_C_ID },
  { &hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_DSCH_FDD_Secondary_Serving_Information },
  { &hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised },
  { &hf_rnsap_hS_DSCH_Secondary_Serving_Remove, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_HS_DSCH_Secondary_Serving_Remove },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs, Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Reconf_Req_sequence_of[1] = {
  { &hf_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs },
};

static int
dissect_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req, Additional_HS_Cell_Information_RL_Reconf_Req_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_RL_Reconf_Req_sequence[] = {
  { &hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Req(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Req, Additional_EDCH_Cell_Information_RL_Reconf_Req_sequence);

  return offset;
}


static const per_sequence_t RL_ReconfigurationRequestFDD_RL_InformationList_sequence_of[1] = {
  { &hf_rnsap_RL_ReconfigurationRequestFDD_RL_InformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_ReconfigurationRequestFDD_RL_InformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_ReconfigurationRequestFDD_RL_InformationList, RL_ReconfigurationRequestFDD_RL_InformationList_sequence_of,
                                                  0, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_ReconfigurationRequestFDD_RL_Information_IEs_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_rL_Specific_DCH_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RL_Specific_DCH_Info },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_ReconfigurationRequestFDD_RL_Information_IEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_ReconfigurationRequestFDD_RL_Information_IEs, RL_ReconfigurationRequestFDD_RL_Information_IEs_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationRequestTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkReconfigurationRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkReconfigurationRequestTDD, RadioLinkReconfigurationRequestTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_sequence_of,
                                                  0, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_tFCS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCS },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD, UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD, UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_sequence_of,
                                                  0, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD, UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_sequence_of,
                                                  0, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_tFCS          , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCS },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD, DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD, DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_sequence_of,
                                                  0, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD, DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteItem_RL_ReconfRqstTDD_sequence[] = {
  { &hf_rnsap_dCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_DeleteItem_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_DeleteItem_RL_ReconfRqstTDD, DCH_DeleteItem_RL_ReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DCH_DeleteList_RL_ReconfRqstTDD_sequence_of[1] = {
  { &hf_rnsap_DCH_DeleteList_RL_ReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_DeleteItem_RL_ReconfRqstTDD },
};

static int
dissect_rnsap_DCH_DeleteList_RL_ReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DCH_DeleteList_RL_ReconfRqstTDD, DCH_DeleteList_RL_ReconfRqstTDD_sequence_of,
                                                  0, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t RL_ReconfigurationRequestTDD_RL_Information_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_rL_Specific_DCH_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RL_Specific_DCH_Info },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_ReconfigurationRequestTDD_RL_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_ReconfigurationRequestTDD_RL_Information, RL_ReconfigurationRequestTDD_RL_Information_sequence);

  return offset;
}


static const per_sequence_t Multiple_RL_ReconfigurationRequestTDD_RL_Information_sequence_of[1] = {
  { &hf_rnsap_Multiple_RL_ReconfigurationRequestTDD_RL_Information_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ReconfigurationRequestTDD_RL_Information },
};

static int
dissect_rnsap_Multiple_RL_ReconfigurationRequestTDD_RL_Information(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Multiple_RL_ReconfigurationRequestTDD_RL_Information, Multiple_RL_ReconfigurationRequestTDD_RL_Information_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationResponseFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkReconfigurationResponseFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkReconfigurationResponseFDD, RadioLinkReconfigurationResponseFDD_sequence);

  return offset;
}


static const per_sequence_t RL_InformationResponseList_RL_ReconfRspFDD_sequence_of[1] = {
  { &hf_rnsap_RL_InformationResponseList_RL_ReconfRspFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationResponseList_RL_ReconfRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationResponseList_RL_ReconfRspFDD, RL_InformationResponseList_RL_ReconfRspFDD_sequence_of,
                                                  0, maxNrOfRLs, FALSE);

  return offset;
}



static int
dissect_rnsap_DCH_InformationResponseList_RL_ReconfRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DL_CodeInformationList_RL_ReconfRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_InformationResponseItem_RL_ReconfRspFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_max_UL_SIR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_min_UL_SIR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_not_Used_secondary_CCPCH_Info, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NULL },
  { &hf_rnsap_dCHsInformationResponseList, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DCH_InformationResponseList_RL_ReconfRspFDD },
  { &hf_rnsap_dL_CodeInformationList_RL_ReconfResp, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_CodeInformationList_RL_ReconfRspFDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationResponseItem_RL_ReconfRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationResponseItem_RL_ReconfRspFDD, RL_InformationResponseItem_RL_ReconfRspFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkReconfigurationResponseTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkReconfigurationResponseTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkReconfigurationResponseTDD, RadioLinkReconfigurationResponseTDD_sequence);

  return offset;
}



static int
dissect_rnsap_DCH_InformationResponseList_RL_ReconfRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_InformationResponse_RL_ReconfRspTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_max_UL_SIR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_min_UL_SIR    , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_SIR },
  { &hf_rnsap_maximumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_minimumDLTxPower, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_dCHsInformationResponseList_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DCH_InformationResponseList_RL_ReconfRspTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationResponse_RL_ReconfRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationResponse_RL_ReconfRspTDD, RL_InformationResponse_RL_ReconfRspTDD_sequence);

  return offset;
}



static int
dissect_rnsap_DL_DPCH_InformationModifyList_LCR_RL_ReconfRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationItem_RL_ReconfRspTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_dl_DPCH_ModifyInformation_LCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_DPCH_InformationModifyList_LCR_RL_ReconfRspTDD },
  { &hf_rnsap_cCTrCH_Maximum_DL_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_cCTrCH_Minimum_DL_Power, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCH_InformationItem_RL_ReconfRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCH_InformationItem_RL_ReconfRspTDD, DL_CCTrCH_InformationItem_RL_ReconfRspTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationList_RL_ReconfRspTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCH_InformationList_RL_ReconfRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_CCTrCH_InformationItem_RL_ReconfRspTDD },
};

static int
dissect_rnsap_DL_CCTrCH_InformationList_RL_ReconfRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCH_InformationList_RL_ReconfRspTDD, DL_CCTrCH_InformationList_RL_ReconfRspTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_LCR_InformationModifyItem_RL_ReconfRspTDD_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_maxPowerLCR   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_minPowerLCR   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Power },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_Timeslot_LCR_InformationModifyItem_RL_ReconfRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_Timeslot_LCR_InformationModifyItem_RL_ReconfRspTDD, DL_Timeslot_LCR_InformationModifyItem_RL_ReconfRspTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_LCR_InformationModifyList_RL_ReconfRspTDD_sequence_of[1] = {
  { &hf_rnsap_DL_Timeslot_LCR_InformationModifyList_RL_ReconfRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Timeslot_LCR_InformationModifyItem_RL_ReconfRspTDD },
};

static int
dissect_rnsap_DL_Timeslot_LCR_InformationModifyList_RL_ReconfRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_Timeslot_LCR_InformationModifyList_RL_ReconfRspTDD, DL_Timeslot_LCR_InformationModifyList_RL_ReconfRspTDD_sequence_of,
                                                  1, maxNrOfDLTsLCR, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD_sequence[] = {
  { &hf_rnsap_dL_Timeslot_LCR_InformationModifyList_RL_ReconfRqstTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Timeslot_LCR_InformationModifyList_RL_ReconfRspTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD, DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD_sequence);

  return offset;
}


static const per_sequence_t Multiple_RL_InformationResponse_RL_ReconfRspTDD_sequence_of[1] = {
  { &hf_rnsap_Multiple_RL_InformationResponse_RL_ReconfRspTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_InformationResponse_RL_ReconfRspTDD },
};

static int
dissect_rnsap_Multiple_RL_InformationResponse_RL_ReconfRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Multiple_RL_InformationResponse_RL_ReconfRspTDD, Multiple_RL_InformationResponse_RL_ReconfRspTDD_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t RadioLinkFailureIndication_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkFailureIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkFailureIndication, RadioLinkFailureIndication_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_FailureInd_sequence_of[1] = {
  { &hf_rnsap_RL_InformationList_RL_FailureInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationList_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationList_RL_FailureInd, RL_InformationList_RL_FailureInd_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_RL_FailureInd_sequence[] = {
  { &hf_rnsap_rL_InformationList_RL_FailureInd, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_InformationList_RL_FailureInd },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_RL_FailureInd, RL_RL_FailureInd_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationList_RL_FailureInd_sequence_of[1] = {
  { &hf_rnsap_RL_Set_InformationList_RL_FailureInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_Set_InformationList_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Set_InformationList_RL_FailureInd, RL_Set_InformationList_RL_FailureInd_sequence_of,
                                                  1, maxNrOfRLSets, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_RL_FailureInd_sequence[] = {
  { &hf_rnsap_rL_Set_InformationList_RL_FailureInd, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_InformationList_RL_FailureInd },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_RL_FailureInd, RL_Set_RL_FailureInd_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_InformationList_RL_FailureInd_sequence_of[1] = {
  { &hf_rnsap_CCTrCH_InformationList_RL_FailureInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_CCTrCH_InformationList_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_CCTrCH_InformationList_RL_FailureInd, CCTrCH_InformationList_RL_FailureInd_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t CCTrCH_RL_FailureInd_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_cCTrCH_InformationList_RL_FailureInd, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_InformationList_RL_FailureInd },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CCTrCH_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CCTrCH_RL_FailureInd, CCTrCH_RL_FailureInd_sequence);

  return offset;
}


static const value_string rnsap_Reporting_Object_RL_FailureInd_vals[] = {
  {   0, "rL" },
  {   1, "rL-Set" },
  {   2, "cCTrCH" },
  { 0, NULL }
};

static const per_choice_t Reporting_Object_RL_FailureInd_choice[] = {
  {   0, &hf_rnsap_rL            , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_RL_FailureInd },
  {   1, &hf_rnsap_rL_Set        , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_Set_RL_FailureInd },
  {   2, &hf_rnsap_cCTrCH        , ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_CCTrCH_RL_FailureInd },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Reporting_Object_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Reporting_Object_RL_FailureInd, Reporting_Object_RL_FailureInd_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_Information_RL_FailureInd_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Information_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Information_RL_FailureInd, RL_Information_RL_FailureInd_sequence);

  return offset;
}


static const per_sequence_t RL_Set_Information_RL_FailureInd_sequence[] = {
  { &hf_rnsap_rL_Set_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_Information_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_Information_RL_FailureInd, RL_Set_Information_RL_FailureInd_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_InformationItem_RL_FailureInd_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CCTrCH_InformationItem_RL_FailureInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CCTrCH_InformationItem_RL_FailureInd, CCTrCH_InformationItem_RL_FailureInd_sequence);

  return offset;
}


static const per_sequence_t RadioLinkPreemptionRequiredIndication_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkPreemptionRequiredIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkPreemptionRequiredIndication, RadioLinkPreemptionRequiredIndication_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_PreemptRequiredInd_sequence_of[1] = {
  { &hf_rnsap_RL_InformationList_RL_PreemptRequiredInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationList_RL_PreemptRequiredInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationList_RL_PreemptRequiredInd, RL_InformationList_RL_PreemptRequiredInd_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationItem_RL_PreemptRequiredInd_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationItem_RL_PreemptRequiredInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationItem_RL_PreemptRequiredInd, RL_InformationItem_RL_PreemptRequiredInd_sequence);

  return offset;
}


static const per_sequence_t HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd_sequence_of[1] = {
  { &hf_rnsap_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd, HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd_sequence_of,
                                                  1, maxNrOfMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd_sequence[] = {
  { &hf_rnsap_hSDSCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HSDSCH_MACdFlow_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd, HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd_sequence);

  return offset;
}


static const per_sequence_t EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd_sequence_of[1] = {
  { &hf_rnsap_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd, EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd_sequence[] = {
  { &hf_rnsap_eDCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd, EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd_sequence);

  return offset;
}


static const per_sequence_t RadioLinkRestoreIndication_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkRestoreIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkRestoreIndication, RadioLinkRestoreIndication_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_RestoreInd_sequence_of[1] = {
  { &hf_rnsap_RL_InformationList_RL_RestoreInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationList_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationList_RL_RestoreInd, RL_InformationList_RL_RestoreInd_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_RL_RestoreInd_sequence[] = {
  { &hf_rnsap_rL_InformationList_RL_RestoreInd, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_InformationList_RL_RestoreInd },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_RL_RestoreInd, RL_RL_RestoreInd_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationList_RL_RestoreInd_sequence_of[1] = {
  { &hf_rnsap_RL_Set_InformationList_RL_RestoreInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_Set_InformationList_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Set_InformationList_RL_RestoreInd, RL_Set_InformationList_RL_RestoreInd_sequence_of,
                                                  1, maxNrOfRLSets, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_RL_RestoreInd_sequence[] = {
  { &hf_rnsap_rL_Set_InformationList_RL_RestoreInd, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_InformationList_RL_RestoreInd },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_RL_RestoreInd, RL_Set_RL_RestoreInd_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_InformationList_RL_RestoreInd_sequence_of[1] = {
  { &hf_rnsap_CCTrCH_InformationList_RL_RestoreInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_CCTrCH_InformationList_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_CCTrCH_InformationList_RL_RestoreInd, CCTrCH_InformationList_RL_RestoreInd_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t CCTrCH_RL_RestoreInd_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_cCTrCH_InformationList_RL_RestoreInd, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_InformationList_RL_RestoreInd },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CCTrCH_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CCTrCH_RL_RestoreInd, CCTrCH_RL_RestoreInd_sequence);

  return offset;
}


static const value_string rnsap_Reporting_Object_RL_RestoreInd_vals[] = {
  {   0, "rL" },
  {   1, "rL-Set" },
  {   2, "cCTrCH" },
  { 0, NULL }
};

static const per_choice_t Reporting_Object_RL_RestoreInd_choice[] = {
  {   0, &hf_rnsap_rL_01         , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_RL_RestoreInd },
  {   1, &hf_rnsap_rL_Set_01     , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_Set_RL_RestoreInd },
  {   2, &hf_rnsap_cCTrCH_01     , ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_CCTrCH_RL_RestoreInd },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Reporting_Object_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Reporting_Object_RL_RestoreInd, Reporting_Object_RL_RestoreInd_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_Information_RL_RestoreInd_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Information_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Information_RL_RestoreInd, RL_Information_RL_RestoreInd_sequence);

  return offset;
}


static const per_sequence_t RL_Set_Information_RL_RestoreInd_sequence[] = {
  { &hf_rnsap_rL_Set_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_Information_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_Information_RL_RestoreInd, RL_Set_Information_RL_RestoreInd_sequence);

  return offset;
}


static const per_sequence_t CCTrCH_InformationItem_RL_RestoreInd_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CCTrCH_InformationItem_RL_RestoreInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CCTrCH_InformationItem_RL_RestoreInd, CCTrCH_InformationItem_RL_RestoreInd_sequence);

  return offset;
}


static const per_sequence_t DL_PowerControlRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_PowerControlRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_PowerControlRequest, DL_PowerControlRequest_sequence);

  return offset;
}


static const per_sequence_t DL_ReferencePowerInformationList_DL_PC_Rqst_sequence_of[1] = {
  { &hf_rnsap_DL_ReferencePowerInformationList_DL_PC_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_DL_ReferencePowerInformationList_DL_PC_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_ReferencePowerInformationList_DL_PC_Rqst, DL_ReferencePowerInformationList_DL_PC_Rqst_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t DL_ReferencePowerInformation_DL_PC_Rqst_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_dl_Reference_Power, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Power },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_ReferencePowerInformation_DL_PC_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_ReferencePowerInformation_DL_PC_Rqst, DL_ReferencePowerInformation_DL_PC_Rqst_sequence);

  return offset;
}


static const per_sequence_t DL_PowerTimeslotControlRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_PowerTimeslotControlRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_PowerTimeslotControlRequest, DL_PowerTimeslotControlRequest_sequence);

  return offset;
}


static const per_sequence_t PhysicalChannelReconfigurationRequestFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PhysicalChannelReconfigurationRequestFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PhysicalChannelReconfigurationRequestFDD, PhysicalChannelReconfigurationRequestFDD_sequence);

  return offset;
}



static int
dissect_rnsap_DL_CodeInformationList_PhyChReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_Information_PhyChReconfRqstFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_dl_CodeInformation_03, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_CodeInformationList_PhyChReconfRqstFDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Information_PhyChReconfRqstFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Information_PhyChReconfRqstFDD, RL_Information_PhyChReconfRqstFDD_sequence);

  return offset;
}


static const per_sequence_t PhysicalChannelReconfigurationRequestTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PhysicalChannelReconfigurationRequestTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PhysicalChannelReconfigurationRequestTDD, PhysicalChannelReconfigurationRequestTDD_sequence);

  return offset;
}



static int
dissect_rnsap_UL_CCTrCH_InformationList_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_DL_CCTrCH_InformationList_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t RL_Information_PhyChReconfRqstTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_ul_CCTrCH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_CCTrCH_InformationList_PhyChReconfRqstTDD },
  { &hf_rnsap_dl_CCTrCH_Information_01, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_CCTrCH_InformationList_PhyChReconfRqstTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Information_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Information_PhyChReconfRqstTDD, RL_Information_PhyChReconfRqstTDD_sequence);

  return offset;
}



static int
dissect_rnsap_UL_DPCH_InformationList_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationItem_PhyChReconfRqstTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_ul_DPCH_Information_02, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_DPCH_InformationList_PhyChReconfRqstTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_CCTrCH_InformationItem_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_CCTrCH_InformationItem_PhyChReconfRqstTDD, UL_CCTrCH_InformationItem_PhyChReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_sequence_of[1] = {
  { &hf_rnsap_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_CCTrCH_InformationItem_PhyChReconfRqstTDD },
};

static int
dissect_rnsap_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD, UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationItem_PhyChReconfRqstTDD_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_uL_Code_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_UL_Code_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_Timeslot_InformationItem_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_Timeslot_InformationItem_PhyChReconfRqstTDD, UL_Timeslot_InformationItem_PhyChReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationList_PhyChReconfRqstTDD_sequence_of[1] = {
  { &hf_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Timeslot_InformationItem_PhyChReconfRqstTDD },
};

static int
dissect_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD, UL_Timeslot_InformationList_PhyChReconfRqstTDD_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t UL_DPCH_InformationItem_PhyChReconfRqstTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_uL_Timeslot_InformationList_PhyChReconfRqstTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_DPCH_InformationItem_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_DPCH_InformationItem_PhyChReconfRqstTDD, UL_DPCH_InformationItem_PhyChReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_uL_Code_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_UL_Code_LCR_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD, UL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_sequence_of[1] = {
  { &hf_rnsap_UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD },
};

static int
dissect_rnsap_UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD, UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_sequence_of,
                                                  1, maxNrOfTsLCR, FALSE);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationItem_PhyChReconfRqstTDD768_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_uL_Code_Information768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_UL_Code_Information768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UL_Timeslot_InformationItem_PhyChReconfRqstTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UL_Timeslot_InformationItem_PhyChReconfRqstTDD768, UL_Timeslot_InformationItem_PhyChReconfRqstTDD768_sequence);

  return offset;
}


static const per_sequence_t UL_Timeslot_InformationList_PhyChReconfRqstTDD768_sequence_of[1] = {
  { &hf_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UL_Timeslot_InformationItem_PhyChReconfRqstTDD768 },
};

static int
dissect_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD768, UL_Timeslot_InformationList_PhyChReconfRqstTDD768_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}



static int
dissect_rnsap_DL_DPCH_InformationList_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationItem_PhyChReconfRqstTDD_sequence[] = {
  { &hf_rnsap_cCTrCH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CCTrCH_ID },
  { &hf_rnsap_dl_DPCH_Information_02, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_DPCH_InformationList_PhyChReconfRqstTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_CCTrCH_InformationItem_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_CCTrCH_InformationItem_PhyChReconfRqstTDD, DL_CCTrCH_InformationItem_PhyChReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_sequence_of[1] = {
  { &hf_rnsap_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_CCTrCH_InformationItem_PhyChReconfRqstTDD },
};

static int
dissect_rnsap_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD, DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_sequence_of,
                                                  1, maxNrOfCCTrCHs, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationItem_PhyChReconfRqstTDD_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_dL_Code_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_DL_Code_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_Timeslot_InformationItem_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_Timeslot_InformationItem_PhyChReconfRqstTDD, DL_Timeslot_InformationItem_PhyChReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationList_PhyChReconfRqstTDD_sequence_of[1] = {
  { &hf_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Timeslot_InformationItem_PhyChReconfRqstTDD },
};

static int
dissect_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD, DL_Timeslot_InformationList_PhyChReconfRqstTDD_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t DL_DPCH_InformationItem_PhyChReconfRqstTDD_sequence[] = {
  { &hf_rnsap_repetitionPeriod, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RepetitionPeriod },
  { &hf_rnsap_repetitionLength, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RepetitionLength },
  { &hf_rnsap_tDD_DPCHOffset, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_DPCHOffset },
  { &hf_rnsap_dL_Timeslot_InformationList_PhyChReconfRqstTDD, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_DPCH_InformationItem_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_DPCH_InformationItem_PhyChReconfRqstTDD, DL_DPCH_InformationItem_PhyChReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD_sequence[] = {
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_dL_Code_LCR_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_DL_Code_LCR_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD, DL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_sequence_of[1] = {
  { &hf_rnsap_DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD },
};

static int
dissect_rnsap_DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD, DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_sequence_of,
                                                  1, maxNrOfTsLCR, FALSE);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationItem_PhyChReconfRqstTDD768_sequence[] = {
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_tFCI_Presence , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TFCI_Presence },
  { &hf_rnsap_dL_Code_Information768, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TDD_DL_Code_Information768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DL_Timeslot_InformationItem_PhyChReconfRqstTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DL_Timeslot_InformationItem_PhyChReconfRqstTDD768, DL_Timeslot_InformationItem_PhyChReconfRqstTDD768_sequence);

  return offset;
}


static const per_sequence_t DL_Timeslot_InformationList_PhyChReconfRqstTDD768_sequence_of[1] = {
  { &hf_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_DL_Timeslot_InformationItem_PhyChReconfRqstTDD768 },
};

static int
dissect_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD768, DL_Timeslot_InformationList_PhyChReconfRqstTDD768_sequence_of,
                                                  1, maxNrOfTS, FALSE);

  return offset;
}


static const per_sequence_t HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD_sequence[] = {
  { &hf_rnsap_timeslot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD, HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD_sequence_of[1] = {
  { &hf_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD },
};

static int
dissect_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD, HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD_sequence_of,
                                                  1, maxNrOfDLTs, FALSE);

  return offset;
}


static const per_sequence_t HSPDSCH_Timeslot_InformationItemLCR_PhyChReconfRqstTDD_sequence[] = {
  { &hf_rnsap_timeslotLCR   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_midambleShiftLCR, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftLCR },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSPDSCH_Timeslot_InformationItemLCR_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSPDSCH_Timeslot_InformationItemLCR_PhyChReconfRqstTDD, HSPDSCH_Timeslot_InformationItemLCR_PhyChReconfRqstTDD_sequence);

  return offset;
}


static const per_sequence_t HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD_sequence_of[1] = {
  { &hf_rnsap_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSPDSCH_Timeslot_InformationItemLCR_PhyChReconfRqstTDD },
};

static int
dissect_rnsap_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD, HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD_sequence_of,
                                                  1, maxNrOfDLTsLCR, FALSE);

  return offset;
}


static const per_sequence_t HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD768_sequence[] = {
  { &hf_rnsap_timeslot      , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TimeSlot },
  { &hf_rnsap_midambleShiftAndBurstType768, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MidambleShiftAndBurstType768 },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD768, HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD768_sequence);

  return offset;
}


static const per_sequence_t HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768_sequence_of[1] = {
  { &hf_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD768 },
};

static int
dissect_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768, HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768_sequence_of,
                                                  1, maxNrOfDLTs, FALSE);

  return offset;
}


static const per_sequence_t PhysicalChannelReconfigurationCommand_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PhysicalChannelReconfigurationCommand(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PhysicalChannelReconfigurationCommand, PhysicalChannelReconfigurationCommand_sequence);

  return offset;
}


static const per_sequence_t PhysicalChannelReconfigurationFailure_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PhysicalChannelReconfigurationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PhysicalChannelReconfigurationFailure, PhysicalChannelReconfigurationFailure_sequence);

  return offset;
}


static const per_sequence_t RadioLinkCongestionIndication_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkCongestionIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkCongestionIndication, RadioLinkCongestionIndication_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_RL_CongestInd_sequence_of[1] = {
  { &hf_rnsap_RL_InformationList_RL_CongestInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationList_RL_CongestInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationList_RL_CongestInd, RL_InformationList_RL_CongestInd_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t DCH_Rate_Information_RL_CongestInd_sequence_of[1] = {
  { &hf_rnsap_DCH_Rate_Information_RL_CongestInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_DCH_Rate_Information_RL_CongestInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DCH_Rate_Information_RL_CongestInd, DCH_Rate_Information_RL_CongestInd_sequence_of,
                                                  1, maxNrOfDCHs, FALSE);

  return offset;
}


static const per_sequence_t RL_InformationItem_RL_CongestInd_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_dCH_Rate_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_Rate_Information_RL_CongestInd },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationItem_RL_CongestInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationItem_RL_CongestInd, RL_InformationItem_RL_CongestInd_sequence);

  return offset;
}


static const per_sequence_t DCH_Rate_InformationItem_RL_CongestInd_sequence[] = {
  { &hf_rnsap_dCH_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DCH_ID },
  { &hf_rnsap_allowed_Rate_Information, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Allowed_Rate_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DCH_Rate_InformationItem_RL_CongestInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DCH_Rate_InformationItem_RL_CongestInd, DCH_Rate_InformationItem_RL_CongestInd_sequence);

  return offset;
}


static const per_sequence_t EDCH_MacdFlowSpecificInformationList_RL_CongestInd_sequence_of[1] = {
  { &hf_rnsap_EDCH_MacdFlowSpecificInformationList_RL_CongestInd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_EDCH_MacdFlowSpecificInformationList_RL_CongestInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_EDCH_MacdFlowSpecificInformationList_RL_CongestInd, EDCH_MacdFlowSpecificInformationList_RL_CongestInd_sequence_of,
                                                  1, maxNrOfEDCHMACdFlows, FALSE);

  return offset;
}


static const per_sequence_t EDCH_MacdFlowSpecificInformationItem_RL_CongestInd_sequence[] = {
  { &hf_rnsap_eDCH_MACdFlow_ID, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_EDCH_MACdFlow_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_CongestInd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_CongestInd, EDCH_MacdFlowSpecificInformationItem_RL_CongestInd_sequence);

  return offset;
}


static const per_sequence_t UplinkSignallingTransferIndicationFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UplinkSignallingTransferIndicationFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UplinkSignallingTransferIndicationFDD, UplinkSignallingTransferIndicationFDD_sequence);

  return offset;
}


static const per_sequence_t UplinkSignallingTransferIndicationTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UplinkSignallingTransferIndicationTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UplinkSignallingTransferIndicationTDD, UplinkSignallingTransferIndicationTDD_sequence);

  return offset;
}


static const per_sequence_t DownlinkSignallingTransferRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DownlinkSignallingTransferRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DownlinkSignallingTransferRequest, DownlinkSignallingTransferRequest_sequence);

  return offset;
}


static const per_sequence_t RelocationCommit_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RelocationCommit(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RelocationCommit, RelocationCommit_sequence);

  return offset;
}


static const per_sequence_t PagingRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PagingRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PagingRequest, PagingRequest_sequence);

  return offset;
}


static const per_sequence_t URA_PagingRqst_sequence[] = {
  { &hf_rnsap_uRA_ID        , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_URA_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_URA_PagingRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_URA_PagingRqst, URA_PagingRqst_sequence);

  return offset;
}


static const per_sequence_t Cell_PagingRqst_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Cell_PagingRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Cell_PagingRqst, Cell_PagingRqst_sequence);

  return offset;
}


static const value_string rnsap_PagingArea_PagingRqst_vals[] = {
  {   0, "uRA" },
  {   1, "cell" },
  { 0, NULL }
};

static const per_choice_t PagingArea_PagingRqst_choice[] = {
  {   0, &hf_rnsap_uRA           , ASN1_EXTENSION_ROOT    , dissect_rnsap_URA_PagingRqst },
  {   1, &hf_rnsap_cell          , ASN1_EXTENSION_ROOT    , dissect_rnsap_Cell_PagingRqst },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_PagingArea_PagingRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_PagingArea_PagingRqst, PagingArea_PagingRqst_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t CNOriginatedPage_PagingRqst_sequence[] = {
  { &hf_rnsap_pagingCause   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PagingCause },
  { &hf_rnsap_cNDomainType  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CNDomainType },
  { &hf_rnsap_pagingRecordType, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PagingRecordType },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CNOriginatedPage_PagingRqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CNOriginatedPage_PagingRqst, CNOriginatedPage_PagingRqst_sequence);

  return offset;
}


static const per_sequence_t DedicatedMeasurementInitiationRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementInitiationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DedicatedMeasurementInitiationRequest, DedicatedMeasurementInitiationRequest_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_DM_Rqst_sequence_of[1] = {
  { &hf_rnsap_RL_InformationList_DM_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationList_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationList_DM_Rqst, RL_InformationList_DM_Rqst_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_DM_Rqst_sequence[] = {
  { &hf_rnsap_rL_InformationList_DM_Rqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_InformationList_DM_Rqst },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_DM_Rqst, RL_DM_Rqst_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationList_DM_Rqst_sequence_of[1] = {
  { &hf_rnsap_RL_Set_InformationList_DM_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_Set_InformationList_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Set_InformationList_DM_Rqst, RL_Set_InformationList_DM_Rqst_sequence_of,
                                                  1, maxNrOfRLSets, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_DM_Rqst_sequence[] = {
  { &hf_rnsap_rL_Set_InformationList_DM_Rqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_InformationList_DM_Rqst },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_DM_Rqst, RL_Set_DM_Rqst_sequence);

  return offset;
}



static int
dissect_rnsap_All_RL_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_All_RL_Set_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_null(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_DedicatedMeasurementObjectType_DM_Rqst_vals[] = {
  {   0, "rL" },
  {   1, "rLS" },
  {   2, "allRL" },
  {   3, "allRLS" },
  { 0, NULL }
};

static const per_choice_t DedicatedMeasurementObjectType_DM_Rqst_choice[] = {
  {   0, &hf_rnsap_rL_02         , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_DM_Rqst },
  {   1, &hf_rnsap_rLS           , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_Set_DM_Rqst },
  {   2, &hf_rnsap_allRL         , ASN1_EXTENSION_ROOT    , dissect_rnsap_All_RL_DM_Rqst },
  {   3, &hf_rnsap_allRLS        , ASN1_EXTENSION_ROOT    , dissect_rnsap_All_RL_Set_DM_Rqst },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementObjectType_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DedicatedMeasurementObjectType_DM_Rqst, DedicatedMeasurementObjectType_DM_Rqst_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_InformationItem_DM_Rqst_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DPCH_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationItem_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationItem_DM_Rqst, RL_InformationItem_DM_Rqst_sequence);

  return offset;
}


static const per_sequence_t HSSICH_Info_DM_Rqst_sequence_of[1] = {
  { &hf_rnsap_HSSICH_Info_DM_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_ID },
};

static int
dissect_rnsap_HSSICH_Info_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSSICH_Info_DM_Rqst, HSSICH_Info_DM_Rqst_sequence_of,
                                                  1, maxNrOfHSSICHs, FALSE);

  return offset;
}


static const per_sequence_t HSSICH_Info_DM_Rqst_Extension_sequence_of[1] = {
  { &hf_rnsap_HSSICH_Info_DM_Rqst_Extension_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_ID_Extension },
};

static int
dissect_rnsap_HSSICH_Info_DM_Rqst_Extension(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_HSSICH_Info_DM_Rqst_Extension, HSSICH_Info_DM_Rqst_Extension_sequence_of,
                                                  1, maxNrOfHSSICHs, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_InformationItem_DM_Rqst_sequence[] = {
  { &hf_rnsap_rL_Set_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_InformationItem_DM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_InformationItem_DM_Rqst, RL_Set_InformationItem_DM_Rqst_sequence);

  return offset;
}


static const per_sequence_t DedicatedMeasurementInitiationResponse_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementInitiationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DedicatedMeasurementInitiationResponse, DedicatedMeasurementInitiationResponse_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_DM_Rsp_sequence_of[1] = {
  { &hf_rnsap_RL_InformationList_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationList_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationList_DM_Rsp, RL_InformationList_DM_Rsp_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_DM_Rsp_sequence[] = {
  { &hf_rnsap_rL_InformationList_DM_Rsp, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_InformationList_DM_Rsp },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_DM_Rsp, RL_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationList_DM_Rsp_sequence_of[1] = {
  { &hf_rnsap_RL_Set_InformationList_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_Set_InformationList_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Set_InformationList_DM_Rsp, RL_Set_InformationList_DM_Rsp_sequence_of,
                                                  1, maxNrOfRLSets, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_DM_Rsp_sequence[] = {
  { &hf_rnsap_rL_Set_InformationList_DM_Rsp, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_InformationList_DM_Rsp },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_DM_Rsp, RL_Set_DM_Rsp_sequence);

  return offset;
}


static const value_string rnsap_DedicatedMeasurementObjectType_DM_Rsp_vals[] = {
  {   0, "rLs" },
  {   1, "rLS" },
  {   2, "allRL" },
  {   3, "allRLS" },
  { 0, NULL }
};

static const per_choice_t DedicatedMeasurementObjectType_DM_Rsp_choice[] = {
  {   0, &hf_rnsap_rLs           , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_DM_Rsp },
  {   1, &hf_rnsap_rLS_01        , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_Set_DM_Rsp },
  {   2, &hf_rnsap_allRL_01      , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_DM_Rsp },
  {   3, &hf_rnsap_allRLS_01     , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_Set_DM_Rsp },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementObjectType_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DedicatedMeasurementObjectType_DM_Rsp, DedicatedMeasurementObjectType_DM_Rsp_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_InformationItem_DM_Rsp_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DPCH_ID },
  { &hf_rnsap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DedicatedMeasurementValue },
  { &hf_rnsap_cFN           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CFN },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationItem_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationItem_DM_Rsp, RL_InformationItem_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationItem_DM_Rsp_sequence[] = {
  { &hf_rnsap_rL_Set_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DedicatedMeasurementValue },
  { &hf_rnsap_cFN           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CFN },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_InformationItem_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_InformationItem_DM_Rsp, RL_Set_InformationItem_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DedicatedMeasurementValue },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp, Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_sequence_of[1] = {
  { &hf_rnsap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp },
};

static int
dissect_rnsap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp, Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_sequence_of,
                                                  1, maxNrOfDPCHsPerRL_1, FALSE);

  return offset;
}


static const per_sequence_t Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp_sequence[] = {
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID },
  { &hf_rnsap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DedicatedMeasurementValue },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp, Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_sequence_of[1] = {
  { &hf_rnsap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp },
};

static int
dissect_rnsap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp, Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_sequence_of,
                                                  1, maxNrOfDPCHsLCRPerRL_1, FALSE);

  return offset;
}


static const per_sequence_t Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp_sequence[] = {
  { &hf_rnsap_hsSICH_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_SICH_ID },
  { &hf_rnsap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DedicatedMeasurementValue },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp, Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_sequence_of[1] = {
  { &hf_rnsap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp },
};

static int
dissect_rnsap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp, Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_sequence_of,
                                                  1, maxNrOfHSSICHs_1, FALSE);

  return offset;
}


static const per_sequence_t Multiple_DedicatedMeasurementValueItem_TDD768_DM_Rsp_sequence[] = {
  { &hf_rnsap_dPCH_ID768    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DPCH_ID768 },
  { &hf_rnsap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DedicatedMeasurementValue },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Multiple_DedicatedMeasurementValueItem_TDD768_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Multiple_DedicatedMeasurementValueItem_TDD768_DM_Rsp, Multiple_DedicatedMeasurementValueItem_TDD768_DM_Rsp_sequence);

  return offset;
}


static const per_sequence_t Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp_sequence_of[1] = {
  { &hf_rnsap_Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Multiple_DedicatedMeasurementValueItem_TDD768_DM_Rsp },
};

static int
dissect_rnsap_Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp, Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp_sequence_of,
                                                  1, maxNrOfDPCHs768PerRL_1, FALSE);

  return offset;
}


static const per_sequence_t DedicatedMeasurementInitiationFailure_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementInitiationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DedicatedMeasurementInitiationFailure, DedicatedMeasurementInitiationFailure_sequence);

  return offset;
}


static const per_sequence_t RL_Unsuccessful_InformationRespList_DM_Fail_sequence_of[1] = {
  { &hf_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail, RL_Unsuccessful_InformationRespList_DM_Fail_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_Successful_InformationRespList_DM_Fail_sequence_of[1] = {
  { &hf_rnsap_RL_Successful_InformationRespList_DM_Fail_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_Successful_InformationRespList_DM_Fail(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Successful_InformationRespList_DM_Fail, RL_Successful_InformationRespList_DM_Fail_sequence_of,
                                                  1, maxNrOfRLs_1, FALSE);

  return offset;
}


static const per_sequence_t RL_DM_Fail_sequence[] = {
  { &hf_rnsap_rL_unsuccessful_InformationRespList_DM_Fail, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail },
  { &hf_rnsap_rL_successful_InformationRespList_DM_Fail, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RL_Successful_InformationRespList_DM_Fail },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_DM_Fail(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_DM_Fail, RL_DM_Fail_sequence);

  return offset;
}


static const per_sequence_t RL_Set_Unsuccessful_InformationRespList_DM_Fail_sequence_of[1] = {
  { &hf_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail, RL_Set_Unsuccessful_InformationRespList_DM_Fail_sequence_of,
                                                  1, maxNrOfRLSets, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_Successful_InformationRespList_DM_Fail_sequence_of[1] = {
  { &hf_rnsap_RL_Set_Successful_InformationRespList_DM_Fail_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_Set_Successful_InformationRespList_DM_Fail(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Set_Successful_InformationRespList_DM_Fail, RL_Set_Successful_InformationRespList_DM_Fail_sequence_of,
                                                  1, maxNrOfRLSets_1, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_DM_Fail_sequence[] = {
  { &hf_rnsap_rL_Set_unsuccessful_InformationRespList_DM_Fail, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail },
  { &hf_rnsap_rL_Set_successful_InformationRespList_DM_Fail, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_RL_Set_Successful_InformationRespList_DM_Fail },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_DM_Fail(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_DM_Fail, RL_Set_DM_Fail_sequence);

  return offset;
}


static const value_string rnsap_DedicatedMeasurementObjectType_DM_Fail_vals[] = {
  {   0, "rL" },
  {   1, "rLS" },
  {   2, "allRL" },
  {   3, "allRLS" },
  { 0, NULL }
};

static const per_choice_t DedicatedMeasurementObjectType_DM_Fail_choice[] = {
  {   0, &hf_rnsap_rL_03         , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_DM_Fail },
  {   1, &hf_rnsap_rLS_02        , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_Set_DM_Fail },
  {   2, &hf_rnsap_allRL_02      , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_DM_Fail },
  {   3, &hf_rnsap_allRLS_02     , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_Set_DM_Fail },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementObjectType_DM_Fail(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DedicatedMeasurementObjectType_DM_Fail, DedicatedMeasurementObjectType_DM_Fail_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_Unsuccessful_InformationItem_DM_Fail_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_individualcause, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Unsuccessful_InformationItem_DM_Fail(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Unsuccessful_InformationItem_DM_Fail, RL_Unsuccessful_InformationItem_DM_Fail_sequence);

  return offset;
}


static const per_sequence_t RL_Successful_InformationItem_DM_Fail_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DPCH_ID },
  { &hf_rnsap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DedicatedMeasurementValue },
  { &hf_rnsap_cFN           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CFN },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Successful_InformationItem_DM_Fail(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Successful_InformationItem_DM_Fail, RL_Successful_InformationItem_DM_Fail_sequence);

  return offset;
}


static const per_sequence_t RL_Set_Unsuccessful_InformationItem_DM_Fail_sequence[] = {
  { &hf_rnsap_rL_Set_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_individualcause, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail, RL_Set_Unsuccessful_InformationItem_DM_Fail_sequence);

  return offset;
}


static const per_sequence_t RL_Set_Successful_InformationItem_DM_Fail_sequence[] = {
  { &hf_rnsap_rL_Set_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_dedicatedMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DedicatedMeasurementValue },
  { &hf_rnsap_cFN           , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CFN },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_Successful_InformationItem_DM_Fail(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_Successful_InformationItem_DM_Fail, RL_Set_Successful_InformationItem_DM_Fail_sequence);

  return offset;
}


static const per_sequence_t DedicatedMeasurementReport_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementReport(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DedicatedMeasurementReport, DedicatedMeasurementReport_sequence);

  return offset;
}


static const per_sequence_t RL_InformationList_DM_Rprt_sequence_of[1] = {
  { &hf_rnsap_RL_InformationList_DM_Rprt_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_InformationList_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_InformationList_DM_Rprt, RL_InformationList_DM_Rprt_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_DM_Rprt_sequence[] = {
  { &hf_rnsap_rL_InformationList_DM_Rprt, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_InformationList_DM_Rprt },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_DM_Rprt, RL_DM_Rprt_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationList_DM_Rprt_sequence_of[1] = {
  { &hf_rnsap_RL_Set_InformationList_DM_Rprt_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_Set_InformationList_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Set_InformationList_DM_Rprt, RL_Set_InformationList_DM_Rprt_sequence_of,
                                                  1, maxNrOfRLSets, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_DM_Rprt_sequence[] = {
  { &hf_rnsap_rL_Set_InformationList_DM_Rprt, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_InformationList_DM_Rprt },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_DM_Rprt, RL_Set_DM_Rprt_sequence);

  return offset;
}


static const value_string rnsap_DedicatedMeasurementObjectType_DM_Rprt_vals[] = {
  {   0, "rLs" },
  {   1, "rLS" },
  {   2, "allRL" },
  {   3, "allRLS" },
  { 0, NULL }
};

static const per_choice_t DedicatedMeasurementObjectType_DM_Rprt_choice[] = {
  {   0, &hf_rnsap_rLs_01        , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_DM_Rprt },
  {   1, &hf_rnsap_rLS_03        , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_Set_DM_Rprt },
  {   2, &hf_rnsap_allRL_03      , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_DM_Rprt },
  {   3, &hf_rnsap_allRLS_03     , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_Set_DM_Rprt },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementObjectType_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DedicatedMeasurementObjectType_DM_Rprt, DedicatedMeasurementObjectType_DM_Rprt_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_InformationItem_DM_Rprt_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_dPCH_ID       , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_DPCH_ID },
  { &hf_rnsap_dedicatedMeasurementValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DedicatedMeasurementValueInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_InformationItem_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_InformationItem_DM_Rprt, RL_InformationItem_DM_Rprt_sequence);

  return offset;
}


static const per_sequence_t RL_Set_InformationItem_DM_Rprt_sequence[] = {
  { &hf_rnsap_rL_Set_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_dedicatedMeasurementValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DedicatedMeasurementValueInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_InformationItem_DM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_InformationItem_DM_Rprt, RL_Set_InformationItem_DM_Rprt_sequence);

  return offset;
}


static const per_sequence_t DedicatedMeasurementTerminationRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementTerminationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DedicatedMeasurementTerminationRequest, DedicatedMeasurementTerminationRequest_sequence);

  return offset;
}


static const per_sequence_t DedicatedMeasurementFailureIndication_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementFailureIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DedicatedMeasurementFailureIndication, DedicatedMeasurementFailureIndication_sequence);

  return offset;
}


static const per_sequence_t RL_Unsuccessful_InformationRespList_DM_Fail_Ind_sequence_of[1] = {
  { &hf_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail_Ind_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail_Ind(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail_Ind, RL_Unsuccessful_InformationRespList_DM_Fail_Ind_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_DM_Fail_Ind_sequence[] = {
  { &hf_rnsap_rL_unsuccessful_InformationRespList_DM_Fail_Ind, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail_Ind },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_DM_Fail_Ind(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_DM_Fail_Ind, RL_DM_Fail_Ind_sequence);

  return offset;
}


static const per_sequence_t RL_Set_Unsuccessful_InformationRespList_DM_Fail_Ind_sequence_of[1] = {
  { &hf_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail_Ind_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail_Ind(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail_Ind, RL_Set_Unsuccessful_InformationRespList_DM_Fail_Ind_sequence_of,
                                                  1, maxNrOfRLSets, FALSE);

  return offset;
}


static const per_sequence_t RL_Set_DM_Fail_Ind_sequence[] = {
  { &hf_rnsap_rL_Set_unsuccessful_InformationRespList_DM_Fail_Ind, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail_Ind },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_DM_Fail_Ind(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_DM_Fail_Ind, RL_Set_DM_Fail_Ind_sequence);

  return offset;
}


static const value_string rnsap_DedicatedMeasurementObjectType_DM_Fail_Ind_vals[] = {
  {   0, "rL" },
  {   1, "rLS" },
  {   2, "allRL" },
  {   3, "allRLS" },
  { 0, NULL }
};

static const per_choice_t DedicatedMeasurementObjectType_DM_Fail_Ind_choice[] = {
  {   0, &hf_rnsap_rL_04         , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_DM_Fail_Ind },
  {   1, &hf_rnsap_rLS_04        , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_Set_DM_Fail_Ind },
  {   2, &hf_rnsap_allRL_04      , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_DM_Fail_Ind },
  {   3, &hf_rnsap_allRLS_04     , ASN1_EXTENSION_ROOT    , dissect_rnsap_RL_Set_DM_Fail_Ind },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_DedicatedMeasurementObjectType_DM_Fail_Ind(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_DedicatedMeasurementObjectType_DM_Fail_Ind, DedicatedMeasurementObjectType_DM_Fail_Ind_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t RL_Unsuccessful_InformationItem_DM_Fail_Ind_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_individualcause, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Unsuccessful_InformationItem_DM_Fail_Ind(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Unsuccessful_InformationItem_DM_Fail_Ind, RL_Unsuccessful_InformationItem_DM_Fail_Ind_sequence);

  return offset;
}


static const per_sequence_t RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind_sequence[] = {
  { &hf_rnsap_rL_Set_ID     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_Set_ID },
  { &hf_rnsap_individualcause, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind, RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind_sequence);

  return offset;
}


static const per_sequence_t CommonTransportChannelResourcesReleaseRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CommonTransportChannelResourcesReleaseRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CommonTransportChannelResourcesReleaseRequest, CommonTransportChannelResourcesReleaseRequest_sequence);

  return offset;
}


static const per_sequence_t CommonTransportChannelResourcesRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CommonTransportChannelResourcesRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CommonTransportChannelResourcesRequest, CommonTransportChannelResourcesRequest_sequence);

  return offset;
}


static const per_sequence_t CommonTransportChannelResourcesResponseFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CommonTransportChannelResourcesResponseFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CommonTransportChannelResourcesResponseFDD, CommonTransportChannelResourcesResponseFDD_sequence);

  return offset;
}



static int
dissect_rnsap_FACH_FlowControlInformation_CTCH_ResourceRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD_sequence[] = {
  { &hf_rnsap_fACH_FlowControlInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FACH_FlowControlInformation_CTCH_ResourceRspFDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD, FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD_sequence);

  return offset;
}


static const per_sequence_t CommonTransportChannelResourcesResponseTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CommonTransportChannelResourcesResponseTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CommonTransportChannelResourcesResponseTDD, CommonTransportChannelResourcesResponseTDD_sequence);

  return offset;
}



static int
dissect_rnsap_FACH_FlowControlInformation_CTCH_ResourceRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD_sequence[] = {
  { &hf_rnsap_fACH_FlowControlInformation_01, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_FACH_FlowControlInformation_CTCH_ResourceRspTDD },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD, FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD_sequence);

  return offset;
}


static const per_sequence_t CommonTransportChannelResourcesFailure_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CommonTransportChannelResourcesFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CommonTransportChannelResourcesFailure, CommonTransportChannelResourcesFailure_sequence);

  return offset;
}


static const per_sequence_t CompressedModeCommand_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CompressedModeCommand(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CompressedModeCommand, CompressedModeCommand_sequence);

  return offset;
}


static const per_sequence_t ErrorIndication_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ErrorIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ErrorIndication, ErrorIndication_sequence);

  return offset;
}


static const per_sequence_t CommonMeasurementInitiationRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CommonMeasurementInitiationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CommonMeasurementInitiationRequest, CommonMeasurementInitiationRequest_sequence);

  return offset;
}



static int
dissect_rnsap_Extension_neighbouringCellMeasurementInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}



static int
dissect_rnsap_Extension_neighbouringCellMeasurementInformation768(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_NeighbouringCellMeasurementInfo_item_vals[] = {
  {   0, "neighbouringFDDCellMeasurementInformation" },
  {   1, "neighbouringTDDCellMeasurementInformation" },
  {   2, "extension-neighbouringCellMeasurementInformation" },
  {   3, "extension-neighbouringCellMeasurementInformation768" },
  { 0, NULL }
};

static const per_choice_t NeighbouringCellMeasurementInfo_item_choice[] = {
  {   0, &hf_rnsap_neighbouringFDDCellMeasurementInformation, ASN1_EXTENSION_ROOT    , dissect_rnsap_NeighbouringFDDCellMeasurementInformation },
  {   1, &hf_rnsap_neighbouringTDDCellMeasurementInformation, ASN1_EXTENSION_ROOT    , dissect_rnsap_NeighbouringTDDCellMeasurementInformation },
  {   2, &hf_rnsap_extension_neighbouringCellMeasurementInformation, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_neighbouringCellMeasurementInformation },
  {   3, &hf_rnsap_extension_neighbouringCellMeasurementInformation768, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_neighbouringCellMeasurementInformation768 },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_NeighbouringCellMeasurementInfo_item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_NeighbouringCellMeasurementInfo_item, NeighbouringCellMeasurementInfo_item_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t NeighbouringCellMeasurementInfo_sequence_of[1] = {
  { &hf_rnsap_NeighbouringCellMeasurementInfo_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_NeighbouringCellMeasurementInfo_item },
};

static int
dissect_rnsap_NeighbouringCellMeasurementInfo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_NeighbouringCellMeasurementInfo, NeighbouringCellMeasurementInfo_sequence_of,
                                                  1, maxNrOfMeasNCell, FALSE);

  return offset;
}


static const per_sequence_t Cell_CM_Rqst_sequence[] = {
  { &hf_rnsap_uC_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UC_ID },
  { &hf_rnsap_timeSlot      , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TimeSlot },
  { &hf_rnsap_timeSlotLCR   , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_TimeSlotLCR },
  { &hf_rnsap_neighbouringCellMeasurementInformation, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_NeighbouringCellMeasurementInfo },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Cell_CM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Cell_CM_Rqst, Cell_CM_Rqst_sequence);

  return offset;
}



static int
dissect_rnsap_Additional_CommonMeasurementObjectType_CM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_CommonMeasurementObjectType_CM_Rqst_vals[] = {
  {   0, "cell" },
  {   1, "additional-CommonMeasurementObjectType-CM-Rqst" },
  { 0, NULL }
};

static const per_choice_t CommonMeasurementObjectType_CM_Rqst_choice[] = {
  {   0, &hf_rnsap_cell_01       , ASN1_EXTENSION_ROOT    , dissect_rnsap_Cell_CM_Rqst },
  {   1, &hf_rnsap_additional_CommonMeasurementObjectType_CM_Rqst, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Additional_CommonMeasurementObjectType_CM_Rqst },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_CommonMeasurementObjectType_CM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_CommonMeasurementObjectType_CM_Rqst, CommonMeasurementObjectType_CM_Rqst_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t GSM_Cell_CM_Rqst_sequence[] = {
  { &hf_rnsap_cGI           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CGI },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GSM_Cell_CM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GSM_Cell_CM_Rqst, GSM_Cell_CM_Rqst_sequence);

  return offset;
}



static int
dissect_rnsap_Extension_CommonMeasurementObjectType_CM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t GsmCellItem_CM_Rqst_sequence[] = {
  { &hf_rnsap_measurementID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementID },
  { &hf_rnsap_gsmCell       , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_GSM_Cell_CM_Rqst },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GsmCellItem_CM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GsmCellItem_CM_Rqst, GsmCellItem_CM_Rqst_sequence);

  return offset;
}


static const per_sequence_t GsmCellList_CM_Rqst_sequence_of[1] = {
  { &hf_rnsap_GsmCellList_CM_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GsmCellItem_CM_Rqst },
};

static int
dissect_rnsap_GsmCellList_CM_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GsmCellList_CM_Rqst, GsmCellList_CM_Rqst_sequence_of,
                                                  1, maxNoOfGsmCell, FALSE);

  return offset;
}


static const per_sequence_t CommonMeasurementInitiationResponse_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CommonMeasurementInitiationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CommonMeasurementInitiationResponse, CommonMeasurementInitiationResponse_sequence);

  return offset;
}


static const per_sequence_t Cell_CM_Rsp_sequence[] = {
  { &hf_rnsap_commonMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CommonMeasurementValue },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Cell_CM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Cell_CM_Rsp, Cell_CM_Rsp_sequence);

  return offset;
}


static const value_string rnsap_CommonMeasurementObjectType_CM_Rsp_vals[] = {
  {   0, "cell" },
  { 0, NULL }
};

static const per_choice_t CommonMeasurementObjectType_CM_Rsp_choice[] = {
  {   0, &hf_rnsap_cell_02       , ASN1_EXTENSION_ROOT    , dissect_rnsap_Cell_CM_Rsp },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_CommonMeasurementObjectType_CM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_CommonMeasurementObjectType_CM_Rsp, CommonMeasurementObjectType_CM_Rsp_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_Extension_CommonMeasurementObjectType_CM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t GsmCellItem_CM_Rsp_sequence[] = {
  { &hf_rnsap_measurementID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementID },
  { &hf_rnsap_commonMeasurementValue, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_CommonMeasurementValue },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GsmCellItem_CM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GsmCellItem_CM_Rsp, GsmCellItem_CM_Rsp_sequence);

  return offset;
}


static const per_sequence_t GsmCellList_CM_Rsp_sequence_of[1] = {
  { &hf_rnsap_GsmCellList_CM_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GsmCellItem_CM_Rsp },
};

static int
dissect_rnsap_GsmCellList_CM_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GsmCellList_CM_Rsp, GsmCellList_CM_Rsp_sequence_of,
                                                  1, maxNoOfGsmCell, FALSE);

  return offset;
}


static const per_sequence_t CommonMeasurementInitiationFailure_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CommonMeasurementInitiationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CommonMeasurementInitiationFailure, CommonMeasurementInitiationFailure_sequence);

  return offset;
}


static const per_sequence_t Extension_FailureMeasurementItem_sequence[] = {
  { &hf_rnsap_measurementID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementID },
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Extension_FailureMeasurementItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Extension_FailureMeasurementItem, Extension_FailureMeasurementItem_sequence);

  return offset;
}


static const per_sequence_t Extension_FailureMeasurementList_sequence_of[1] = {
  { &hf_rnsap_Extension_FailureMeasurementList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Extension_FailureMeasurementItem },
};

static int
dissect_rnsap_Extension_FailureMeasurementList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Extension_FailureMeasurementList, Extension_FailureMeasurementList_sequence_of,
                                                  1, maxNoOfGsmCell, FALSE);

  return offset;
}


static const per_sequence_t CommonMeasurementReport_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CommonMeasurementReport(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CommonMeasurementReport, CommonMeasurementReport_sequence);

  return offset;
}


static const per_sequence_t Cell_CM_Rprt_sequence[] = {
  { &hf_rnsap_commonMeasurementValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CommonMeasurementValueInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Cell_CM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Cell_CM_Rprt, Cell_CM_Rprt_sequence);

  return offset;
}


static const value_string rnsap_CommonMeasurementObjectType_CM_Rprt_vals[] = {
  {   0, "cell" },
  { 0, NULL }
};

static const per_choice_t CommonMeasurementObjectType_CM_Rprt_choice[] = {
  {   0, &hf_rnsap_cell_03       , ASN1_EXTENSION_ROOT    , dissect_rnsap_Cell_CM_Rprt },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_CommonMeasurementObjectType_CM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_CommonMeasurementObjectType_CM_Rprt, CommonMeasurementObjectType_CM_Rprt_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_Extension_CommonMeasurementObjectType_CM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const per_sequence_t GsmCellItem_CM_Rprt_sequence[] = {
  { &hf_rnsap_measurementID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementID },
  { &hf_rnsap_commonMeasurementValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CommonMeasurementValueInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GsmCellItem_CM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GsmCellItem_CM_Rprt, GsmCellItem_CM_Rprt_sequence);

  return offset;
}


static const per_sequence_t GsmCellList_CM_Rprt_sequence_of[1] = {
  { &hf_rnsap_GsmCellList_CM_Rprt_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_GsmCellItem_CM_Rprt },
};

static int
dissect_rnsap_GsmCellList_CM_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_GsmCellList_CM_Rprt, GsmCellList_CM_Rprt_sequence_of,
                                                  1, maxNoOfGsmCell, FALSE);

  return offset;
}


static const per_sequence_t CommonMeasurementTerminationRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CommonMeasurementTerminationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CommonMeasurementTerminationRequest, CommonMeasurementTerminationRequest_sequence);

  return offset;
}


static const per_sequence_t Extension_TerminationMeasurementItem_sequence[] = {
  { &hf_rnsap_measurementID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Extension_TerminationMeasurementItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Extension_TerminationMeasurementItem, Extension_TerminationMeasurementItem_sequence);

  return offset;
}


static const per_sequence_t Extension_TerminationMeasurementList_sequence_of[1] = {
  { &hf_rnsap_Extension_TerminationMeasurementList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Extension_TerminationMeasurementItem },
};

static int
dissect_rnsap_Extension_TerminationMeasurementList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Extension_TerminationMeasurementList, Extension_TerminationMeasurementList_sequence_of,
                                                  1, maxNoOfGsmCell, FALSE);

  return offset;
}


static const per_sequence_t CommonMeasurementFailureIndication_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_CommonMeasurementFailureIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_CommonMeasurementFailureIndication, CommonMeasurementFailureIndication_sequence);

  return offset;
}


static const per_sequence_t Extension_FailureIndicationMeasurementItem_sequence[] = {
  { &hf_rnsap_measurementID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementID },
  { &hf_rnsap_cause         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Cause },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Extension_FailureIndicationMeasurementItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Extension_FailureIndicationMeasurementItem, Extension_FailureIndicationMeasurementItem_sequence);

  return offset;
}


static const per_sequence_t Extension_FailureIndicationMeasurementList_sequence_of[1] = {
  { &hf_rnsap_Extension_FailureIndicationMeasurementList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Extension_FailureIndicationMeasurementItem },
};

static int
dissect_rnsap_Extension_FailureIndicationMeasurementList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Extension_FailureIndicationMeasurementList, Extension_FailureIndicationMeasurementList_sequence_of,
                                                  1, maxNoOfGsmCell, FALSE);

  return offset;
}


static const per_sequence_t InformationExchangeInitiationRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_InformationExchangeInitiationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_InformationExchangeInitiationRequest, InformationExchangeInitiationRequest_sequence);

  return offset;
}


static const per_sequence_t Cell_InfEx_Rqst_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Cell_InfEx_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Cell_InfEx_Rqst, Cell_InfEx_Rqst_sequence);

  return offset;
}



static int
dissect_rnsap_Extension_InformationExchangeObjectType_InfEx_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_InformationExchangeObjectType_InfEx_Rqst_vals[] = {
  {   0, "cell" },
  {   1, "extension-InformationExchangeObjectType-InfEx-Rqst" },
  { 0, NULL }
};

static const per_choice_t InformationExchangeObjectType_InfEx_Rqst_choice[] = {
  {   0, &hf_rnsap_cell_04       , ASN1_EXTENSION_ROOT    , dissect_rnsap_Cell_InfEx_Rqst },
  {   1, &hf_rnsap_extension_InformationExchangeObjectType_InfEx_Rqst, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_InformationExchangeObjectType_InfEx_Rqst },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_InformationExchangeObjectType_InfEx_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_InformationExchangeObjectType_InfEx_Rqst, InformationExchangeObjectType_InfEx_Rqst_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t GSM_Cell_InfEx_Rqst_sequence[] = {
  { &hf_rnsap_cGI           , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_CGI },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GSM_Cell_InfEx_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GSM_Cell_InfEx_Rqst, GSM_Cell_InfEx_Rqst_sequence);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_List_InfEx_Rqst_sequence_of[1] = {
  { &hf_rnsap_MBMS_Bearer_Service_List_InfEx_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TMGI },
};

static int
dissect_rnsap_MBMS_Bearer_Service_List_InfEx_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBMS_Bearer_Service_List_InfEx_Rqst, MBMS_Bearer_Service_List_InfEx_Rqst_sequence_of,
                                                  1, maxNrOfMBMSServices, FALSE);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rqst_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_mBMS_Bearer_Service_List_InfEx_Rqst, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_Service_List_InfEx_Rqst },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rqst, MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rqst_sequence);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst_sequence_of[1] = {
  { &hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rqst },
};

static int
dissect_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst, MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst_sequence_of,
                                                  1, maxNrOfCells, FALSE);

  return offset;
}


static const per_sequence_t MBMS_Cell_InfEx_Rqst_sequence_of[1] = {
  { &hf_rnsap_MBMS_Cell_InfEx_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
};

static int
dissect_rnsap_MBMS_Cell_InfEx_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBMS_Cell_InfEx_Rqst, MBMS_Cell_InfEx_Rqst_sequence_of,
                                                  1, maxNrOfCells, FALSE);

  return offset;
}


static const per_sequence_t ANR_Cell_InfEx_Rqst_sequence_of[1] = {
  { &hf_rnsap_ANR_Cell_InfEx_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
};

static int
dissect_rnsap_ANR_Cell_InfEx_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_ANR_Cell_InfEx_Rqst, ANR_Cell_InfEx_Rqst_sequence_of,
                                                  1, maxNrOfANRCells, FALSE);

  return offset;
}


static const per_sequence_t Common_E_RGCH_Cell_InfEx_Rqst_sequence_of[1] = {
  { &hf_rnsap_Common_E_RGCH_Cell_InfEx_Rqst_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
};

static int
dissect_rnsap_Common_E_RGCH_Cell_InfEx_Rqst(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Common_E_RGCH_Cell_InfEx_Rqst, Common_E_RGCH_Cell_InfEx_Rqst_sequence_of,
                                                  1, maxNoOfCommonRGCells, FALSE);

  return offset;
}


static const per_sequence_t InformationExchangeInitiationResponse_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_InformationExchangeInitiationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_InformationExchangeInitiationResponse, InformationExchangeInitiationResponse_sequence);

  return offset;
}


static const per_sequence_t Cell_InfEx_Rsp_sequence[] = {
  { &hf_rnsap_requestedDataValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RequestedDataValue },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Cell_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Cell_InfEx_Rsp, Cell_InfEx_Rsp_sequence);

  return offset;
}



static int
dissect_rnsap_Extension_InformationExchangeObjectType_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_InformationExchangeObjectType_InfEx_Rsp_vals[] = {
  {   0, "cell" },
  {   1, "extension-InformationExchangeObjectType-InfEx-Rsp" },
  { 0, NULL }
};

static const per_choice_t InformationExchangeObjectType_InfEx_Rsp_choice[] = {
  {   0, &hf_rnsap_cell_05       , ASN1_EXTENSION_ROOT    , dissect_rnsap_Cell_InfEx_Rsp },
  {   1, &hf_rnsap_extension_InformationExchangeObjectType_InfEx_Rsp, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_InformationExchangeObjectType_InfEx_Rsp },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_InformationExchangeObjectType_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_InformationExchangeObjectType_InfEx_Rsp, InformationExchangeObjectType_InfEx_Rsp_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t MBMS_Bearer_ServiceItemIEs_InfEx_Rsp_sequence[] = {
  { &hf_rnsap_tmgi          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TMGI },
  { &hf_rnsap_requestedDataValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RequestedDataValue },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Bearer_ServiceItemIEs_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Bearer_ServiceItemIEs_InfEx_Rsp, MBMS_Bearer_ServiceItemIEs_InfEx_Rsp_sequence);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_List_InfEx_Rsp_sequence_of[1] = {
  { &hf_rnsap_MBMS_Bearer_Service_List_InfEx_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_ServiceItemIEs_InfEx_Rsp },
};

static int
dissect_rnsap_MBMS_Bearer_Service_List_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBMS_Bearer_Service_List_InfEx_Rsp, MBMS_Bearer_Service_List_InfEx_Rsp_sequence_of,
                                                  1, maxNrOfMBMSServices, FALSE);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rsp_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_mBMS_Bearer_Service_List_InfEx_Rsp, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_Service_List_InfEx_Rsp },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rsp, MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rsp_sequence);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp_sequence_of[1] = {
  { &hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rsp },
};

static int
dissect_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp, MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp_sequence_of,
                                                  1, maxNrOfCells, FALSE);

  return offset;
}


static const per_sequence_t MBMS_Cell_Item_InfEx_Rsp_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_requestedDataValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RequestedDataValue },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Cell_Item_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Cell_Item_InfEx_Rsp, MBMS_Cell_Item_InfEx_Rsp_sequence);

  return offset;
}


static const per_sequence_t MBMS_Cell_InfEx_Rsp_sequence_of[1] = {
  { &hf_rnsap_MBMS_Cell_InfEx_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Cell_Item_InfEx_Rsp },
};

static int
dissect_rnsap_MBMS_Cell_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBMS_Cell_InfEx_Rsp, MBMS_Cell_InfEx_Rsp_sequence_of,
                                                  1, maxNrOfCells, FALSE);

  return offset;
}


static const per_sequence_t ANR_Cell_ItemIEs_InfEx_Rsp_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_requestedDataValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RequestedDataValue },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ANR_Cell_ItemIEs_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ANR_Cell_ItemIEs_InfEx_Rsp, ANR_Cell_ItemIEs_InfEx_Rsp_sequence);

  return offset;
}


static const per_sequence_t ANR_Cell_InfEx_Rsp_sequence_of[1] = {
  { &hf_rnsap_ANR_Cell_InfEx_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ANR_Cell_ItemIEs_InfEx_Rsp },
};

static int
dissect_rnsap_ANR_Cell_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_ANR_Cell_InfEx_Rsp, ANR_Cell_InfEx_Rsp_sequence_of,
                                                  1, maxNrOfANRCells, FALSE);

  return offset;
}


static const per_sequence_t Common_E_RGCH_Cell_ItemIEs_InfEx_Rsp_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_requestedDataValue, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RequestedDataValue },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Common_E_RGCH_Cell_ItemIEs_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Common_E_RGCH_Cell_ItemIEs_InfEx_Rsp, Common_E_RGCH_Cell_ItemIEs_InfEx_Rsp_sequence);

  return offset;
}


static const per_sequence_t Common_E_RGCH_Cell_InfEx_Rsp_sequence_of[1] = {
  { &hf_rnsap_Common_E_RGCH_Cell_InfEx_Rsp_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Common_E_RGCH_Cell_ItemIEs_InfEx_Rsp },
};

static int
dissect_rnsap_Common_E_RGCH_Cell_InfEx_Rsp(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Common_E_RGCH_Cell_InfEx_Rsp, Common_E_RGCH_Cell_InfEx_Rsp_sequence_of,
                                                  1, maxNoOfCommonRGCells, FALSE);

  return offset;
}


static const per_sequence_t InformationExchangeInitiationFailure_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_InformationExchangeInitiationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_InformationExchangeInitiationFailure, InformationExchangeInitiationFailure_sequence);

  return offset;
}


static const per_sequence_t InformationReport_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_InformationReport(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_InformationReport, InformationReport_sequence);

  return offset;
}


static const per_sequence_t Cell_InfEx_Rprt_sequence[] = {
  { &hf_rnsap_requestedDataValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RequestedDataValueInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Cell_InfEx_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Cell_InfEx_Rprt, Cell_InfEx_Rprt_sequence);

  return offset;
}



static int
dissect_rnsap_Extension_InformationExchangeObjectType_InfEx_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_InformationExchangeObjectType_InfEx_Rprt_vals[] = {
  {   0, "cell" },
  {   1, "extension-InformationExchangeObjectType-InfEx-Rprt" },
  { 0, NULL }
};

static const per_choice_t InformationExchangeObjectType_InfEx_Rprt_choice[] = {
  {   0, &hf_rnsap_cell_06       , ASN1_EXTENSION_ROOT    , dissect_rnsap_Cell_InfEx_Rprt },
  {   1, &hf_rnsap_extension_InformationExchangeObjectType_InfEx_Rprt, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_InformationExchangeObjectType_InfEx_Rprt },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_InformationExchangeObjectType_InfEx_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_InformationExchangeObjectType_InfEx_Rprt, InformationExchangeObjectType_InfEx_Rprt_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_List_Item_InfEx_Rprt_sequence[] = {
  { &hf_rnsap_tmgi          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_TMGI },
  { &hf_rnsap_requestedDataValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RequestedDataValueInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Bearer_Service_List_Item_InfEx_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Bearer_Service_List_Item_InfEx_Rprt, MBMS_Bearer_Service_List_Item_InfEx_Rprt_sequence);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_List_InfEx_Rprt_sequence_of[1] = {
  { &hf_rnsap_MBMS_Bearer_Service_List_InfEx_Rprt_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_Service_List_Item_InfEx_Rprt },
};

static int
dissect_rnsap_MBMS_Bearer_Service_List_InfEx_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBMS_Bearer_Service_List_InfEx_Rprt, MBMS_Bearer_Service_List_InfEx_Rprt_sequence_of,
                                                  1, maxNrOfMBMSServices, FALSE);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rprt_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_mBMS_Bearer_Service_List_InfEx_Rprt, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_Service_List_InfEx_Rprt },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rprt, MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rprt_sequence);

  return offset;
}


static const per_sequence_t MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt_sequence_of[1] = {
  { &hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rprt },
};

static int
dissect_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt, MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt_sequence_of,
                                                  1, maxNrOfCells, FALSE);

  return offset;
}


static const per_sequence_t MBMS_Cell_Item_InfEx_Rprt_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_requestedDataValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RequestedDataValueInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMS_Cell_Item_InfEx_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMS_Cell_Item_InfEx_Rprt, MBMS_Cell_Item_InfEx_Rprt_sequence);

  return offset;
}


static const per_sequence_t MBMS_Cell_InfEx_Rprt_sequence_of[1] = {
  { &hf_rnsap_MBMS_Cell_InfEx_Rprt_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_MBMS_Cell_Item_InfEx_Rprt },
};

static int
dissect_rnsap_MBMS_Cell_InfEx_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_MBMS_Cell_InfEx_Rprt, MBMS_Cell_InfEx_Rprt_sequence_of,
                                                  1, maxNrOfCells, FALSE);

  return offset;
}


static const per_sequence_t Common_E_RGCH_Cell_Item_InfEx_Rprt_sequence[] = {
  { &hf_rnsap_c_ID          , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_C_ID },
  { &hf_rnsap_requestedDataValueInformation, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RequestedDataValueInformation },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Common_E_RGCH_Cell_Item_InfEx_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Common_E_RGCH_Cell_Item_InfEx_Rprt, Common_E_RGCH_Cell_Item_InfEx_Rprt_sequence);

  return offset;
}


static const per_sequence_t Common_E_RGCH_Cell_InfEx_Rprt_sequence_of[1] = {
  { &hf_rnsap_Common_E_RGCH_Cell_InfEx_Rprt_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Common_E_RGCH_Cell_Item_InfEx_Rprt },
};

static int
dissect_rnsap_Common_E_RGCH_Cell_InfEx_Rprt(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Common_E_RGCH_Cell_InfEx_Rprt, Common_E_RGCH_Cell_InfEx_Rprt_sequence_of,
                                                  1, maxNoOfCommonRGCells, FALSE);

  return offset;
}


static const per_sequence_t InformationExchangeTerminationRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_InformationExchangeTerminationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_InformationExchangeTerminationRequest, InformationExchangeTerminationRequest_sequence);

  return offset;
}


static const per_sequence_t InformationExchangeFailureIndication_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_InformationExchangeFailureIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_InformationExchangeFailureIndication, InformationExchangeFailureIndication_sequence);

  return offset;
}


static const per_sequence_t ResetRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ResetRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ResetRequest, ResetRequest_sequence);

  return offset;
}


static const per_sequence_t ContextInfoList_Reset_sequence_of[1] = {
  { &hf_rnsap_ContextInfoList_Reset_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_ContextInfoList_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_ContextInfoList_Reset, ContextInfoList_Reset_sequence_of,
                                                  1, maxResetContext, FALSE);

  return offset;
}


static const per_sequence_t ContextList_Reset_sequence[] = {
  { &hf_rnsap_contextInfoList_Reset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ContextInfoList_Reset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ContextList_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ContextList_Reset, ContextList_Reset_sequence);

  return offset;
}


static const per_sequence_t ContextGroupInfoList_Reset_sequence_of[1] = {
  { &hf_rnsap_ContextGroupInfoList_Reset_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_ContextGroupInfoList_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_ContextGroupInfoList_Reset, ContextGroupInfoList_Reset_sequence_of,
                                                  1, maxResetContextGroup, FALSE);

  return offset;
}


static const per_sequence_t ContextGroupList_Reset_sequence[] = {
  { &hf_rnsap_contextGroupInfoList_Reset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ContextGroupInfoList_Reset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ContextGroupList_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ContextGroupList_Reset, ContextGroupList_Reset_sequence);

  return offset;
}


static const value_string rnsap_ResetIndicator_vals[] = {
  {   0, "context" },
  {   1, "all-contexts" },
  {   2, "contextGroup" },
  { 0, NULL }
};

static const per_choice_t ResetIndicator_choice[] = {
  {   0, &hf_rnsap_context       , ASN1_EXTENSION_ROOT    , dissect_rnsap_ContextList_Reset },
  {   1, &hf_rnsap_all_contexts  , ASN1_EXTENSION_ROOT    , dissect_rnsap_NULL },
  {   2, &hf_rnsap_contextGroup  , ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_ContextGroupList_Reset },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_ResetIndicator(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_ResetIndicator, ResetIndicator_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_Extension_ContextType_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_rnsap_ProtocolIE_Single_Container(tvb, offset, actx, tree, hf_index);

  return offset;
}


static const value_string rnsap_ContextType_Reset_vals[] = {
  {   0, "sRNTI" },
  {   1, "dRNTI" },
  {   2, "extension-ContextType-Reset" },
  { 0, NULL }
};

static const per_choice_t ContextType_Reset_choice[] = {
  {   0, &hf_rnsap_sRNTI         , ASN1_EXTENSION_ROOT    , dissect_rnsap_S_RNTI },
  {   1, &hf_rnsap_dRNTI         , ASN1_EXTENSION_ROOT    , dissect_rnsap_D_RNTI },
  {   2, &hf_rnsap_extension_ContextType_Reset, ASN1_NOT_EXTENSION_ROOT, dissect_rnsap_Extension_ContextType_Reset },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_ContextType_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_ContextType_Reset, ContextType_Reset_choice,
                                 NULL);

  return offset;
}


static const per_sequence_t ContextInfoItem_Reset_sequence[] = {
  { &hf_rnsap_contextType_Reset, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ContextType_Reset },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ContextInfoItem_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ContextInfoItem_Reset, ContextInfoItem_Reset_sequence);

  return offset;
}


static const per_sequence_t ContextGroupInfoItem_Reset_sequence[] = {
  { &hf_rnsap_s_RNTI_Group  , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_S_RNTI_Group },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ContextGroupInfoItem_Reset(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ContextGroupInfoItem_Reset, ContextGroupInfoItem_Reset_sequence);

  return offset;
}


static const per_sequence_t ResetResponse_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_ResetResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_ResetResponse, ResetResponse_sequence);

  return offset;
}


static const per_sequence_t RadioLinkActivationCommandFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkActivationCommandFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkActivationCommandFDD, RadioLinkActivationCommandFDD_sequence);

  return offset;
}


static const per_sequence_t DelayedActivationInformationList_RL_ActivationCmdFDD_sequence_of[1] = {
  { &hf_rnsap_DelayedActivationInformationList_RL_ActivationCmdFDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_DelayedActivationInformationList_RL_ActivationCmdFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DelayedActivationInformationList_RL_ActivationCmdFDD, DelayedActivationInformationList_RL_ActivationCmdFDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t DelayedActivationInformation_RL_ActivationCmdFDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_delayed_activation_update, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DelayedActivationUpdate },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DelayedActivationInformation_RL_ActivationCmdFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DelayedActivationInformation_RL_ActivationCmdFDD, DelayedActivationInformation_RL_ActivationCmdFDD_sequence);

  return offset;
}


static const per_sequence_t RadioLinkActivationCommandTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkActivationCommandTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkActivationCommandTDD, RadioLinkActivationCommandTDD_sequence);

  return offset;
}


static const per_sequence_t DelayedActivationInformationList_RL_ActivationCmdTDD_sequence_of[1] = {
  { &hf_rnsap_DelayedActivationInformationList_RL_ActivationCmdTDD_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_DelayedActivationInformationList_RL_ActivationCmdTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_DelayedActivationInformationList_RL_ActivationCmdTDD, DelayedActivationInformationList_RL_ActivationCmdTDD_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t DelayedActivationInformation_RL_ActivationCmdTDD_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_delayed_activation_update, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_DelayedActivationUpdate },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DelayedActivationInformation_RL_ActivationCmdTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DelayedActivationInformation_RL_ActivationCmdTDD, DelayedActivationInformation_RL_ActivationCmdTDD_sequence);

  return offset;
}


static const per_sequence_t GERANUplinkSignallingTransferIndication_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_GERANUplinkSignallingTransferIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_GERANUplinkSignallingTransferIndication, GERANUplinkSignallingTransferIndication_sequence);

  return offset;
}


static const per_sequence_t RadioLinkParameterUpdateIndicationFDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkParameterUpdateIndicationFDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkParameterUpdateIndicationFDD, RadioLinkParameterUpdateIndicationFDD_sequence);

  return offset;
}


static const per_sequence_t RL_ParameterUpdateIndicationFDD_RL_InformationList_sequence_of[1] = {
  { &hf_rnsap_RL_ParameterUpdateIndicationFDD_RL_InformationList_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_RL_ParameterUpdateIndicationFDD_RL_InformationList(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_RL_ParameterUpdateIndicationFDD_RL_InformationList, RL_ParameterUpdateIndicationFDD_RL_InformationList_sequence_of,
                                                  1, maxNrOfRLs, FALSE);

  return offset;
}


static const per_sequence_t RL_ParameterUpdateIndicationFDD_RL_Information_Item_sequence[] = {
  { &hf_rnsap_rL_ID         , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_phase_Reference_Update_Indicator, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_Phase_Reference_Update_Indicator },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RL_ParameterUpdateIndicationFDD_RL_Information_Item(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RL_ParameterUpdateIndicationFDD_RL_Information_Item, RL_ParameterUpdateIndicationFDD_RL_Information_Item_sequence);

  return offset;
}


static const per_sequence_t UE_Measurement_Forwarding_sequence[] = {
  { &hf_rnsap_measurementID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_MeasurementID },
  { &hf_rnsap_uE_Measurement_Value, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_UE_Measurement_Value },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UE_Measurement_Forwarding(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UE_Measurement_Forwarding, UE_Measurement_Forwarding_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs_sequence[] = {
  { &hf_rnsap_hSPDSCH_RL_ID , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_RL_ID },
  { &hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Update_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_HS_DSCH_FDD_Secondary_Serving_Update_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs, Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_HS_Cell_Information_RL_Param_Upd_sequence_of[1] = {
  { &hf_rnsap_Additional_HS_Cell_Information_RL_Param_Upd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs },
};

static int
dissect_rnsap_Additional_HS_Cell_Information_RL_Param_Upd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_HS_Cell_Information_RL_Param_Upd, Additional_HS_Cell_Information_RL_Param_Upd_sequence_of,
                                                  1, maxNrOfHSDSCH_1, FALSE);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs_sequence[] = {
  { &hf_rnsap_additional_EDCH_FDD_Update_Information, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_FDD_Update_Information },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs, Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs_sequence);

  return offset;
}


static const per_sequence_t Additional_EDCH_Cell_Information_RL_Param_Upd_sequence_of[1] = {
  { &hf_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs },
};

static int
dissect_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd, Additional_EDCH_Cell_Information_RL_Param_Upd_sequence_of,
                                                  1, maxNrOfEDCH_1, FALSE);

  return offset;
}


static const per_sequence_t RadioLinkParameterUpdateIndicationTDD_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_RadioLinkParameterUpdateIndicationTDD(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_RadioLinkParameterUpdateIndicationTDD, RadioLinkParameterUpdateIndicationTDD_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementInitiationRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementInitiationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementInitiationRequest, UEMeasurementInitiationRequest_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementInitiationResponse_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementInitiationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementInitiationResponse, UEMeasurementInitiationResponse_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementInitiationFailure_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementInitiationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementInitiationFailure, UEMeasurementInitiationFailure_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementReport_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementReport(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementReport, UEMeasurementReport_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementTerminationRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementTerminationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementTerminationRequest, UEMeasurementTerminationRequest_sequence);

  return offset;
}


static const per_sequence_t UEMeasurementFailureIndication_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UEMeasurementFailureIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UEMeasurementFailureIndication, UEMeasurementFailureIndication_sequence);

  return offset;
}


static const per_sequence_t IurInvokeTrace_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_IurInvokeTrace(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_IurInvokeTrace, IurInvokeTrace_sequence);

  return offset;
}


static const per_sequence_t ListOfInterfacesToTrace_sequence_of[1] = {
  { &hf_rnsap_ListOfInterfacesToTrace_item, ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Single_Container },
};

static int
dissect_rnsap_ListOfInterfacesToTrace(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_constrained_sequence_of(tvb, offset, actx, tree, hf_index,
                                                  ett_rnsap_ListOfInterfacesToTrace, ListOfInterfacesToTrace_sequence_of,
                                                  1, maxNrOfInterfaces, FALSE);

  return offset;
}


static const value_string rnsap_T_interface_vals[] = {
  {   0, "iub" },
  {   1, "iur" },
  { 0, NULL }
};


static int
dissect_rnsap_T_interface(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_enumerated(tvb, offset, actx, tree, hf_index,
                                     2, NULL, TRUE, 0, NULL);

  return offset;
}


static const per_sequence_t InterfacesToTraceItem_sequence[] = {
  { &hf_rnsap_interface     , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_T_interface },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_InterfacesToTraceItem(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_InterfacesToTraceItem, InterfacesToTraceItem_sequence);

  return offset;
}


static const per_sequence_t IurDeactivateTrace_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_IurDeactivateTrace(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_IurDeactivateTrace, IurDeactivateTrace_sequence);

  return offset;
}


static const per_sequence_t MBMSAttachCommand_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMSAttachCommand(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMSAttachCommand, MBMSAttachCommand_sequence);

  return offset;
}


static const per_sequence_t MBMSDetachCommand_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBMSDetachCommand(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBMSDetachCommand, MBMSDetachCommand_sequence);

  return offset;
}


static const per_sequence_t DirectInformationTransfer_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_DirectInformationTransfer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_DirectInformationTransfer, DirectInformationTransfer_sequence);

  return offset;
}


static const per_sequence_t EnhancedRelocationRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EnhancedRelocationRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EnhancedRelocationRequest, EnhancedRelocationRequest_sequence);

  return offset;
}


static const per_sequence_t EnhancedRelocationResponse_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EnhancedRelocationResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EnhancedRelocationResponse, EnhancedRelocationResponse_sequence);

  return offset;
}


static const per_sequence_t EnhancedRelocationFailure_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EnhancedRelocationFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EnhancedRelocationFailure, EnhancedRelocationFailure_sequence);

  return offset;
}


static const per_sequence_t EnhancedRelocationCancel_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EnhancedRelocationCancel(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EnhancedRelocationCancel, EnhancedRelocationCancel_sequence);

  return offset;
}


static const per_sequence_t EnhancedRelocationSignallingTransfer_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EnhancedRelocationSignallingTransfer(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EnhancedRelocationSignallingTransfer, EnhancedRelocationSignallingTransfer_sequence);

  return offset;
}


static const per_sequence_t EnhancedRelocationRelease_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EnhancedRelocationRelease(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EnhancedRelocationRelease, EnhancedRelocationRelease_sequence);

  return offset;
}


static const per_sequence_t MBSFNMCCHInformation_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_MBSFNMCCHInformation(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_MBSFNMCCHInformation, MBSFNMCCHInformation_sequence);

  return offset;
}


static const per_sequence_t SecondaryULFrequencyReport_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SecondaryULFrequencyReport(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SecondaryULFrequencyReport, SecondaryULFrequencyReport_sequence);

  return offset;
}


static const per_sequence_t SecondaryULFrequencyUpdateIndication_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SecondaryULFrequencyUpdateIndication(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SecondaryULFrequencyUpdateIndication, SecondaryULFrequencyUpdateIndication_sequence);

  return offset;
}


static const per_sequence_t EnhancedRelocationResourceRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EnhancedRelocationResourceRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EnhancedRelocationResourceRequest, EnhancedRelocationResourceRequest_sequence);

  return offset;
}


static const per_sequence_t EnhancedRelocationResourceResponse_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EnhancedRelocationResourceResponse(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EnhancedRelocationResourceResponse, EnhancedRelocationResourceResponse_sequence);

  return offset;
}


static const per_sequence_t EnhancedRelocationResourceFailure_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EnhancedRelocationResourceFailure(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EnhancedRelocationResourceFailure, EnhancedRelocationResourceFailure_sequence);

  return offset;
}


static const per_sequence_t EnhancedRelocationResourceReleaseCommand_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EnhancedRelocationResourceReleaseCommand(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EnhancedRelocationResourceReleaseCommand, EnhancedRelocationResourceReleaseCommand_sequence);

  return offset;
}


static const per_sequence_t EnhancedRelocationResourceReleaseComplete_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_EnhancedRelocationResourceReleaseComplete(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_EnhancedRelocationResourceReleaseComplete, EnhancedRelocationResourceReleaseComplete_sequence);

  return offset;
}


static const per_sequence_t PrivateMessage_sequence[] = {
  { &hf_rnsap_privateIEs    , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_PrivateIE_Container },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_PrivateMessage(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_PrivateMessage, PrivateMessage_sequence);

  return offset;
}


static const per_sequence_t InformationTransferControlRequest_sequence[] = {
  { &hf_rnsap_protocolIEs   , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolIE_Container },
  { &hf_rnsap_protocolExtensions, ASN1_EXTENSION_ROOT    , ASN1_OPTIONAL    , dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_InformationTransferControlRequest(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_InformationTransferControlRequest, InformationTransferControlRequest_sequence);

  return offset;
}


static const per_sequence_t Suspension_Control_Type_sequence[] = {
  { &hf_rnsap_controlled_Object_Scope, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Controlled_Object_Scope },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Suspension_Control_Type(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Suspension_Control_Type, Suspension_Control_Type_sequence);

  return offset;
}


static const per_sequence_t Resume_Control_Type_sequence[] = {
  { &hf_rnsap_controlled_Object_Scope, ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_Controlled_Object_Scope },
  { &hf_rnsap_iE_Extensions , ASN1_EXTENSION_ROOT    , ASN1_NOT_OPTIONAL, dissect_rnsap_ProtocolExtensionContainer },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Resume_Control_Type(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Resume_Control_Type, Resume_Control_Type_sequence);

  return offset;
}


static const value_string rnsap_Control_Type_InformationTransferControlReq_vals[] = {
  {   0, "suspension" },
  {   1, "resume" },
  { 0, NULL }
};

static const per_choice_t Control_Type_InformationTransferControlReq_choice[] = {
  {   0, &hf_rnsap_suspension    , ASN1_EXTENSION_ROOT    , dissect_rnsap_Suspension_Control_Type },
  {   1, &hf_rnsap_resume        , ASN1_EXTENSION_ROOT    , dissect_rnsap_Resume_Control_Type },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_Control_Type_InformationTransferControlReq(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_Control_Type_InformationTransferControlReq, Control_Type_InformationTransferControlReq_choice,
                                 NULL);

  return offset;
}



static int
dissect_rnsap_InitiatingMessage_value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type_pdu_new(tvb, offset, actx, tree, hf_index, dissect_InitiatingMessageValue);

  return offset;
}


static const per_sequence_t InitiatingMessage_sequence[] = {
  { &hf_rnsap_procedureID   , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProcedureID },
  { &hf_rnsap_criticality   , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Criticality },
  { &hf_rnsap_transactionID , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TransactionID },
  { &hf_rnsap_value_02      , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_InitiatingMessage_value },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_InitiatingMessage(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_InitiatingMessage, InitiatingMessage_sequence);

  return offset;
}



static int
dissect_rnsap_SuccessfulOutcome_value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type_pdu_new(tvb, offset, actx, tree, hf_index, dissect_SuccessfulOutcomeValue);

  return offset;
}


static const per_sequence_t SuccessfulOutcome_sequence[] = {
  { &hf_rnsap_procedureID   , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProcedureID },
  { &hf_rnsap_criticality   , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Criticality },
  { &hf_rnsap_transactionID , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TransactionID },
  { &hf_rnsap_value_03      , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_SuccessfulOutcome_value },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_SuccessfulOutcome(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_SuccessfulOutcome, SuccessfulOutcome_sequence);

  return offset;
}



static int
dissect_rnsap_UnsuccessfulOutcome_value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type_pdu_new(tvb, offset, actx, tree, hf_index, dissect_UnsuccessfulOutcomeValue);

  return offset;
}


static const per_sequence_t UnsuccessfulOutcome_sequence[] = {
  { &hf_rnsap_procedureID   , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProcedureID },
  { &hf_rnsap_criticality   , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Criticality },
  { &hf_rnsap_transactionID , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TransactionID },
  { &hf_rnsap_value_04      , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_UnsuccessfulOutcome_value },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_UnsuccessfulOutcome(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_UnsuccessfulOutcome, UnsuccessfulOutcome_sequence);

  return offset;
}



static int
dissect_rnsap_Outcome_value(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_open_type(tvb, offset, actx, tree, hf_index, NULL);

  return offset;
}


static const per_sequence_t Outcome_sequence[] = {
  { &hf_rnsap_procedureID   , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_ProcedureID },
  { &hf_rnsap_criticality   , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Criticality },
  { &hf_rnsap_transactionID , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_TransactionID },
  { &hf_rnsap_value_05      , ASN1_NO_EXTENSIONS     , ASN1_NOT_OPTIONAL, dissect_rnsap_Outcome_value },
  { NULL, 0, 0, NULL }
};

static int
dissect_rnsap_Outcome(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_sequence(tvb, offset, actx, tree, hf_index,
                                   ett_rnsap_Outcome, Outcome_sequence);

  return offset;
}


static const value_string rnsap_RNSAP_PDU_vals[] = {
  {   0, "initiatingMessage" },
  {   1, "successfulOutcome" },
  {   2, "unsuccessfulOutcome" },
  {   3, "outcome" },
  { 0, NULL }
};

static const per_choice_t RNSAP_PDU_choice[] = {
  {   0, &hf_rnsap_initiatingMessage, ASN1_EXTENSION_ROOT    , dissect_rnsap_InitiatingMessage },
  {   1, &hf_rnsap_successfulOutcome, ASN1_EXTENSION_ROOT    , dissect_rnsap_SuccessfulOutcome },
  {   2, &hf_rnsap_unsuccessfulOutcome, ASN1_EXTENSION_ROOT    , dissect_rnsap_UnsuccessfulOutcome },
  {   3, &hf_rnsap_outcome       , ASN1_EXTENSION_ROOT    , dissect_rnsap_Outcome },
  { 0, NULL, 0, NULL }
};

static int
dissect_rnsap_RNSAP_PDU(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  offset = dissect_per_choice(tvb, offset, actx, tree, hf_index,
                                 ett_rnsap_RNSAP_PDU, RNSAP_PDU_choice,
                                 NULL);

  return offset;
}

/*--- PDUs ---*/

static int dissect_ActivationInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ActivationInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_ActivationInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Active_MBMS_Bearer_Service_ListFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Active_MBMS_Bearer_Service_ListFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_Active_MBMS_Bearer_Service_ListFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Active_MBMS_Bearer_Service_ListFDD_PFL_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Active_MBMS_Bearer_Service_ListFDD_PFL(tvb, offset, &asn1_ctx, tree, hf_rnsap_Active_MBMS_Bearer_Service_ListFDD_PFL_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Active_MBMS_Bearer_Service_ListTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Active_MBMS_Bearer_Service_ListTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_Active_MBMS_Bearer_Service_ListTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Active_MBMS_Bearer_Service_ListTDD_PFL_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Active_MBMS_Bearer_Service_ListTDD_PFL(tvb, offset, &asn1_ctx, tree, hf_rnsap_Active_MBMS_Bearer_Service_ListTDD_PFL_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Active_Pattern_Sequence_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Active_Pattern_Sequence_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Active_Pattern_Sequence_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_Response_RLAddList_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_EDCH_Cell_Information_Response_RLAddList(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_EDCH_Cell_Information_Response_RLAddList_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Setup_Info_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_EDCH_Setup_Info(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_EDCH_Setup_Info_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ANRReportIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ANRReportIndication(tvb, offset, &asn1_ctx, tree, hf_rnsap_ANRReportIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DTX_Information2_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DTX_Information2(tvb, offset, &asn1_ctx, tree, hf_rnsap_DTX_Information2_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Implicit_Grant_Handling_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Implicit_Grant_Handling(tvb, offset, &asn1_ctx, tree, hf_rnsap_Implicit_Grant_Handling_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Minimum_TEBS_threshold_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Minimum_TEBS_threshold(tvb, offset, &asn1_ctx, tree, hf_rnsap_Minimum_TEBS_threshold_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_Response_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_EDCH_Cell_Information_Response_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_EDCH_Cell_Information_Response_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_Response_RLReconf_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_EDCH_Cell_Information_Response_RLReconf_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_EDCH_Cell_Information_Response_RLReconf_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_AdjustmentPeriod_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_AdjustmentPeriod(tvb, offset, &asn1_ctx, tree, hf_rnsap_AdjustmentPeriod_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Allowed_Rate_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Allowed_Rate_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Allowed_Rate_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_AllowedQueuingTime_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_AllowedQueuingTime(tvb, offset, &asn1_ctx, tree, hf_rnsap_AllowedQueuingTime_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_AlternativeFormatReportingIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_AlternativeFormatReportingIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_AlternativeFormatReportingIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Angle_Of_Arrival_Value_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Angle_Of_Arrival_Value_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_Angle_Of_Arrival_Value_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ANR_Cell_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ANR_Cell_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_ANR_Cell_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_AntennaColocationIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_AntennaColocationIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_AntennaColocationIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_BDS_IonosphericGridModelReq_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_BDS_IonosphericGridModelReq(tvb, offset, &asn1_ctx, tree, hf_rnsap_BDS_IonosphericGridModelReq_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_BDS_Ionospheric_Grid_Model_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_BDS_Ionospheric_Grid_Model(tvb, offset, &asn1_ctx, tree, hf_rnsap_BDS_Ionospheric_Grid_Model_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_BindingID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_BindingID(tvb, offset, &asn1_ctx, tree, hf_rnsap_BindingID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Cause_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Cause(tvb, offset, &asn1_ctx, tree, hf_rnsap_Cause_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellCapabilityContainer_FDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CellCapabilityContainer_FDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_CellCapabilityContainer_FDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellCapabilityContainerExtension_FDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CellCapabilityContainerExtension_FDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_CellCapabilityContainerExtension_FDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellCapabilityContainer_TDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CellCapabilityContainer_TDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_CellCapabilityContainer_TDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellCapabilityContainer_TDD_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CellCapabilityContainer_TDD_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_CellCapabilityContainer_TDD_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellCapabilityContainerExtension_TDD_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CellCapabilityContainerExtension_TDD_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_CellCapabilityContainerExtension_TDD_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellCapabilityContainer_TDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CellCapabilityContainer_TDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_CellCapabilityContainer_TDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_C_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_C_ID(tvb, offset, &asn1_ctx, tree, hf_rnsap_C_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Cell_Capacity_Class_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Cell_Capacity_Class_Value(tvb, offset, &asn1_ctx, tree, hf_rnsap_Cell_Capacity_Class_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellListValidityIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CellListValidityIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_CellListValidityIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellPortionID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CellPortionID(tvb, offset, &asn1_ctx, tree, hf_rnsap_CellPortionID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CellPortionLCRID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CellPortionLCRID(tvb, offset, &asn1_ctx, tree, hf_rnsap_CellPortionLCRID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CFN_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CFN(tvb, offset, &asn1_ctx, tree, hf_rnsap_CFN_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ChipOffset_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ChipOffset(tvb, offset, &asn1_ctx, tree, hf_rnsap_ChipOffset_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ClassmarkInformation2_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ClassmarkInformation2(tvb, offset, &asn1_ctx, tree, hf_rnsap_ClassmarkInformation2_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ClassmarkInformation3_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ClassmarkInformation3(tvb, offset, &asn1_ctx, tree, hf_rnsap_ClassmarkInformation3_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ClosedLoopMode1_SupportIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ClosedLoopMode1_SupportIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_ClosedLoopMode1_SupportIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementAccuracy_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonMeasurementAccuracy(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonMeasurementAccuracy_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementType_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonMeasurementType(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonMeasurementType_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CongestionCause_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CongestionCause(tvb, offset, &asn1_ctx, tree, hf_rnsap_CongestionCause_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelResourcesInitialisationNotRequired_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonTransportChannelResourcesInitialisationNotRequired(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonTransportChannelResourcesInitialisationNotRequired_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_EDCH_MAC_d_Flow_Specific_InformationFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_EDCH_MAC_d_Flow_Specific_InformationLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_EDCH_Support_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Common_EDCH_Support_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_Common_EDCH_Support_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_E_RGCH_Cell_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Common_E_RGCH_Cell_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Common_E_RGCH_Cell_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CompleteAlmanacProvided_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CompleteAlmanacProvided(tvb, offset, &asn1_ctx, tree, hf_rnsap_CompleteAlmanacProvided_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Continuous_Packet_Connectivity_DTX_DRX_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response(tvb, offset, &asn1_ctx, tree, hf_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CoverageIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CoverageIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_CoverageIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CPC_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CPC_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_CPC_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CPC_RecoveryReport_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CPC_RecoveryReport(tvb, offset, &asn1_ctx, tree, hf_rnsap_CPC_RecoveryReport_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Continuous_Packet_Connectivity_HS_SCCH_less_Deactivate_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Continuous_Packet_Connectivity_HS_SCCH_less_Deactivate_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_Continuous_Packet_Connectivity_HS_SCCH_less_Deactivate_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Counting_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Counting_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Counting_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CriticalityDiagnostics_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CriticalityDiagnostics(tvb, offset, &asn1_ctx, tree, hf_rnsap_CriticalityDiagnostics_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MessageStructure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MessageStructure(tvb, offset, &asn1_ctx, tree, hf_rnsap_MessageStructure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CN_CS_DomainIdentifier_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CN_CS_DomainIdentifier(tvb, offset, &asn1_ctx, tree, hf_rnsap_CN_CS_DomainIdentifier_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CN_PS_DomainIdentifier_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CN_PS_DomainIdentifier(tvb, offset, &asn1_ctx, tree, hf_rnsap_CN_PS_DomainIdentifier_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CQI_Cycle_Switch_Timer_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CQI_Cycle_Switch_Timer(tvb, offset, &asn1_ctx, tree, hf_rnsap_CQI_Cycle_Switch_Timer_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ControlGAP_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ControlGAP(tvb, offset, &asn1_ctx, tree, hf_rnsap_ControlGAP_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CQI_Feedback_Cycle2_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CQI_Feedback_Cycle2(tvb, offset, &asn1_ctx, tree, hf_rnsap_CQI_Feedback_Cycle2_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_C_RNTI_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_C_RNTI(tvb, offset, &asn1_ctx, tree, hf_rnsap_C_RNTI_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CPC_InformationLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CPC_InformationLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_CPC_InformationLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ContinuousPacketConnectivity_DRX_InformationLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ContinuousPacketConnectivity_DRX_InformationLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_ContinuousPacketConnectivity_DRX_InformationLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext(tvb, offset, &asn1_ctx, tree, hf_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_AGCH_UE_Inactivity_Monitor_Threshold_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_AGCH_UE_Inactivity_Monitor_Threshold(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_AGCH_UE_Inactivity_Monitor_Threshold_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ContinuousPacketConnectivity_DRX_Information_ResponseLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DBDS_CorrectionsReq_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DBDS_CorrectionsReq(tvb, offset, &asn1_ctx, tree, hf_rnsap_DBDS_CorrectionsReq_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DBDS_Corrections_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DBDS_Corrections(tvb, offset, &asn1_ctx, tree, hf_rnsap_DBDS_Corrections_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_ENH_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DCH_ENH_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_DCH_ENH_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_ENH_Information_Reconf_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DCH_ENH_Information_Reconf(tvb, offset, &asn1_ctx, tree, hf_rnsap_DCH_ENH_Information_Reconf_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_FDD_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DCH_FDD_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_DCH_FDD_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_MeasurementOccasion_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DCH_MeasurementOccasion_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_DCH_MeasurementOccasion_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_MeasurementType_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DCH_MeasurementType_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_DCH_MeasurementType_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_Indicator_For_E_DCH_HSDPA_Operation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DCH_Indicator_For_E_DCH_HSDPA_Operation(tvb, offset, &asn1_ctx, tree, hf_rnsap_DCH_Indicator_For_E_DCH_HSDPA_Operation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_InformationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DCH_InformationResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_DCH_InformationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_TDD_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DCH_TDD_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_DCH_TDD_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementType_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DedicatedMeasurementType(tvb, offset, &asn1_ctx, tree, hf_rnsap_DedicatedMeasurementType_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DelayedActivation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DelayedActivation(tvb, offset, &asn1_ctx, tree, hf_rnsap_DelayedActivation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DGANSS_Corrections_Req_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DGANSS_Corrections_Req(tvb, offset, &asn1_ctx, tree, hf_rnsap_DGANSS_Corrections_Req_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DGNSS_ValidityPeriod_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DGNSS_ValidityPeriod(tvb, offset, &asn1_ctx, tree, hf_rnsap_DGNSS_ValidityPeriod_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DiversityMode_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DiversityMode(tvb, offset, &asn1_ctx, tree, hf_rnsap_DiversityMode_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_TBS_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_TBS(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_TBS_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_TimingAdjustment_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_TimingAdjustment(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_TimingAdjustment_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_Power_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_Power(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_Power_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_PowerBalancing_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_PowerBalancing_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_PowerBalancing_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_PowerBalancing_ActivationIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_PowerBalancing_ActivationIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_PowerBalancing_ActivationIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_PowerBalancing_UpdatedIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_PowerBalancing_UpdatedIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_PowerBalancing_UpdatedIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_ReferencePowerInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_ReferencePowerInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_ReferencePowerInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_D_RNTI_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_D_RNTI(tvb, offset, &asn1_ctx, tree, hf_rnsap_D_RNTI_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_D_RNTI_ReleaseIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_D_RNTI_ReleaseIndication(tvb, offset, &asn1_ctx, tree, hf_rnsap_D_RNTI_ReleaseIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_TimeSlot_ISCP_Info_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_TimeSlot_ISCP_Info(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_TimeSlot_ISCP_Info_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_TimeSlot_ISCP_LCR_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_TimeSlot_ISCP_LCR_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_TimeSlot_ISCP_LCR_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Downlink_TPC_enhancements_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Downlink_TPC_enhancements_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Downlink_TPC_enhancements_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Downlink_TPC_enhancements_Reconf_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Downlink_TPC_enhancements_Reconf(tvb, offset, &asn1_ctx, tree, hf_rnsap_Downlink_TPC_enhancements_Reconf_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DPC_Mode_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DPC_Mode(tvb, offset, &asn1_ctx, tree, hf_rnsap_DPC_Mode_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DPC_Mode_Change_SupportIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DPC_Mode_Change_SupportIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_DPC_Mode_Change_SupportIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DPCH_ID768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DPCH_ID768(tvb, offset, &asn1_ctx, tree, hf_rnsap_DPCH_ID768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DRXCycleLengthCoefficient_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DRXCycleLengthCoefficient(tvb, offset, &asn1_ctx, tree, hf_rnsap_DRXCycleLengthCoefficient_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_RNTI_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DSCH_RNTI(tvb, offset, &asn1_ctx, tree, hf_rnsap_DSCH_RNTI_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_InitialWindowSize_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DSCH_InitialWindowSize(tvb, offset, &asn1_ctx, tree, hf_rnsap_DSCH_InitialWindowSize_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_TDD_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DSCH_TDD_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_DSCH_TDD_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EARFCN_Extended_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EARFCN_Extended(tvb, offset, &asn1_ctx, tree, hf_rnsap_EARFCN_Extended_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EARFCN_FDD_Extended_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EARFCN_FDD_Extended(tvb, offset, &asn1_ctx, tree, hf_rnsap_EARFCN_FDD_Extended_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_AGCH_Table_Choice_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_AGCH_Table_Choice(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_AGCH_Table_Choice_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Decoupling_Indication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_Decoupling_Indication(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_Decoupling_Indication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_FDD_DL_ControlChannelInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDCH_FDD_DL_ControlChannelInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDCH_FDD_DL_ControlChannelInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_RGCH_E_HICH_ChannelisationCodeValidityIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_RGCH_E_HICH_ChannelisationCodeValidityIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_RGCH_E_HICH_ChannelisationCodeValidityIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_FDD_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDCH_FDD_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDCH_FDD_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UPH_Filtering_Measurement_Forwarding_Request_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UPH_Filtering_Measurement_Forwarding_Request(tvb, offset, &asn1_ctx, tree, hf_rnsap_UPH_Filtering_Measurement_Forwarding_Request_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_FDD_InformationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDCH_FDD_InformationResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDCH_FDD_InformationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_FDD_Information_To_Modify_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDCH_FDD_Information_To_Modify(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDCH_FDD_Information_To_Modify_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_FDD_Update_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_FDD_Update_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_FDD_Update_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_DL_Control_Channel_Change_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_DL_Control_Channel_Change_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_DL_Control_Channel_Change_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_DL_Control_Channel_Grant_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_DL_Control_Channel_Grant_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_DL_Control_Channel_Grant_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_PowerOffset_for_SchedulingInfo_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_PowerOffset_for_SchedulingInfo(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_PowerOffset_for_SchedulingInfo_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_MACdPDUSizeFormat_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_MACdPDUSizeFormat(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_MACdPDUSizeFormat_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_MACdFlows_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDCH_MACdFlows_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDCH_MACdFlows_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_MACdFlows_To_Delete_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDCH_MACdFlows_To_Delete(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDCH_MACdFlows_To_Delete_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_RL_Indication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDCH_RL_Indication(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDCH_RL_Indication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Serving_cell_change_informationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_Serving_cell_change_informationResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_Serving_cell_change_informationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDPCH_Information_FDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDPCH_Information_FDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDPCH_Information_FDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDPCH_Information_RLReconfRequest_FDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDPCH_Information_RLReconfRequest_FDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDPCH_Information_RLReconfRequest_FDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DPDCH_PowerInterpolation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DPDCH_PowerInterpolation(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DPDCH_PowerInterpolation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_Serving_Grant_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_Serving_Grant_Value(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_Serving_Grant_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_Serving_RL_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDCH_Serving_RL(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDCH_Serving_RL_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Enhanced_FACH_Information_ResponseFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Enhanced_FACH_Information_ResponseFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_Enhanced_FACH_Information_ResponseFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Enhanced_FACH_Information_ResponseLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Enhanced_FACH_Information_ResponseLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_Enhanced_FACH_Information_ResponseLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Enhanced_FACH_Support_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Enhanced_FACH_Support_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_Enhanced_FACH_Support_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EnhancedHSServingCC_Abort_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EnhancedHSServingCC_Abort(tvb, offset, &asn1_ctx, tree, hf_rnsap_EnhancedHSServingCC_Abort_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Enhanced_PCH_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Enhanced_PCH_Capability(tvb, offset, &asn1_ctx, tree, hf_rnsap_Enhanced_PCH_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_RNTI_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_RNTI(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_RNTI_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_TFCI_Boost_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_TFCI_Boost_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_TFCI_Boost_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Minimum_Set_E_TFCIValidityIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_Minimum_Set_E_TFCIValidityIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_Minimum_Set_E_TFCIValidityIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Enhanced_PrimaryCPICH_EcNo_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Enhanced_PrimaryCPICH_EcNo(tvb, offset, &asn1_ctx, tree, hf_rnsap_Enhanced_PrimaryCPICH_EcNo_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EventH_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EventH(tvb, offset, &asn1_ctx, tree, hf_rnsap_EventH_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ExtendedGSMCellIndividualOffset_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ExtendedGSMCellIndividualOffset(tvb, offset, &asn1_ctx, tree, hf_rnsap_ExtendedGSMCellIndividualOffset_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_MACdFlow_Retransmission_Timer_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_MACdFlow_Retransmission_Timer_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_MACdFlow_Retransmission_Timer_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Information_Reconfig_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_Information_Reconfig(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_Information_Reconfig_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Information_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_Information_Response(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_Information_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ActivationDelay_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ActivationDelay(tvb, offset, &asn1_ctx, tree, hf_rnsap_ActivationDelay_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Fast_TTI_switching_Mode_synchronized_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Fast_TTI_switching_Mode_synchronized(tvb, offset, &asn1_ctx, tree, hf_rnsap_Fast_TTI_switching_Mode_synchronized_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Fast_TTI_switching_Mode_unsynchronized_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Fast_TTI_switching_Mode_unsynchronized(tvb, offset, &asn1_ctx, tree, hf_rnsap_Fast_TTI_switching_Mode_unsynchronized_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Fast_TTI_switching_Mode_Supported_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Fast_TTI_switching_Mode_Supported(tvb, offset, &asn1_ctx, tree, hf_rnsap_Fast_TTI_switching_Mode_Supported_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_768_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_768_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_768_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_768_Information_Reconfig_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_768_Information_Reconfig(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_768_Information_Reconfig_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_768_Information_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_768_Information_Response(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_768_Information_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_LCR_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_LCR_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_LCR_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extended_E_DCH_LCRTDD_PhysicalLayerCategory_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Extended_E_DCH_LCRTDD_PhysicalLayerCategory(tvb, offset, &asn1_ctx, tree, hf_rnsap_Extended_E_DCH_LCRTDD_PhysicalLayerCategory_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multi_Carrier_E_DCH_LCRTDD_PhysicalLayerCategory_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multi_Carrier_E_DCH_LCRTDD_PhysicalLayerCategory(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multi_Carrier_E_DCH_LCRTDD_PhysicalLayerCategory_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_LCR_Information_Reconfig_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_LCR_Information_Reconfig(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_LCR_Information_Reconfig_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_LCR_Information_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_LCR_Information_Response(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_LCR_Information_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Ext_Reference_E_TFCI_PO_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Ext_Reference_E_TFCI_PO(tvb, offset, &asn1_ctx, tree, hf_rnsap_Ext_Reference_E_TFCI_PO_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ExtendedPropagationDelay_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ExtendedPropagationDelay(tvb, offset, &asn1_ctx, tree, hf_rnsap_ExtendedPropagationDelay_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Radio_Links_without_DPCH_FDPCH_Indication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Radio_Links_without_DPCH_FDPCH_Indication(tvb, offset, &asn1_ctx, tree, hf_rnsap_Radio_Links_without_DPCH_FDPCH_Indication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extended_RNC_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Extended_RNC_ID(tvb, offset, &asn1_ctx, tree, hf_rnsap_Extended_RNC_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extended_RNTI_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Extended_RNTI(tvb, offset, &asn1_ctx, tree, hf_rnsap_Extended_RNTI_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extended_Round_Trip_Time_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Extended_Round_Trip_Time_Value(tvb, offset, &asn1_ctx, tree, hf_rnsap_Extended_Round_Trip_Time_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extended_S_RNTI_Group_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Extended_S_RNTI_Group(tvb, offset, &asn1_ctx, tree, hf_rnsap_Extended_S_RNTI_Group_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ExtendedAffectedUEInformationForMBMS_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ExtendedAffectedUEInformationForMBMS(tvb, offset, &asn1_ctx, tree, hf_rnsap_ExtendedAffectedUEInformationForMBMS_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Ext_Max_Bits_MACe_PDU_non_scheduled_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Ext_Max_Bits_MACe_PDU_non_scheduled(tvb, offset, &asn1_ctx, tree, hf_rnsap_Ext_Max_Bits_MACe_PDU_non_scheduled_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Semi_PersistentScheduling_Information_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_Semi_PersistentScheduling_Information_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_Semi_PersistentScheduling_Information_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FACH_FlowControlInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_FACH_FlowControlInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_FACH_FlowControlInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Fast_Reconfiguration_Mode_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Fast_Reconfiguration_Mode(tvb, offset, &asn1_ctx, tree, hf_rnsap_Fast_Reconfiguration_Mode_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Fast_Reconfiguration_Permission_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Fast_Reconfiguration_Permission(tvb, offset, &asn1_ctx, tree, hf_rnsap_Fast_Reconfiguration_Permission_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FDD_DCHs_to_Modify_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_FDD_DCHs_to_Modify(tvb, offset, &asn1_ctx, tree, hf_rnsap_FDD_DCHs_to_Modify_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FDD_DL_CodeInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_FDD_DL_CodeInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_FDD_DL_CodeInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_F_DPCH_SlotFormat_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_F_DPCH_SlotFormat(tvb, offset, &asn1_ctx, tree, hf_rnsap_F_DPCH_SlotFormat_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_F_DPCH_SlotFormatSupportRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_F_DPCH_SlotFormatSupportRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_F_DPCH_SlotFormatSupportRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FNReportingIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_FNReportingIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_FNReportingIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FTPICH_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_FTPICH_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_FTPICH_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FTPICH_Information_Reconf_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_FTPICH_Information_Reconf(tvb, offset, &asn1_ctx, tree, hf_rnsap_FTPICH_Information_Reconf_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FTPICH_Reconfiguration_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_FTPICH_Reconfiguration_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_FTPICH_Reconfiguration_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FTPICH_Information_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_FTPICH_Information_Response(tvb, offset, &asn1_ctx, tree, hf_rnsap_FTPICH_Information_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FrameOffset_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_FrameOffset(tvb, offset, &asn1_ctx, tree, hf_rnsap_FrameOffset_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FrequencyBandIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_FrequencyBandIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_FrequencyBandIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GA_Cell_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GA_Cell(tvb, offset, &asn1_ctx, tree, hf_rnsap_GA_Cell_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GA_CellAdditionalShapes_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GA_CellAdditionalShapes(tvb, offset, &asn1_ctx, tree, hf_rnsap_GA_CellAdditionalShapes_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Gainfactors_10ms_mode_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Gainfactors_10ms_mode(tvb, offset, &asn1_ctx, tree, hf_rnsap_Gainfactors_10ms_mode_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_AddIonoModelReq_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_AddIonoModelReq(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_AddIonoModelReq_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_AddNavigationModelsReq_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_AddNavigationModelsReq(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_AddNavigationModelsReq_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_AddUTCModelsReq_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_AddUTCModelsReq(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_AddUTCModelsReq_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Additional_Ionospheric_Model_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_Additional_Ionospheric_Model(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_Additional_Ionospheric_Model_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Additional_Navigation_Models_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_Additional_Navigation_Models(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_Additional_Navigation_Models_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Additional_Time_Models_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_Additional_Time_Models(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_Additional_Time_Models_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Additional_UTC_Models_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_Additional_UTC_Models(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_Additional_UTC_Models_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_alm_keplerianNAVAlmanac_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_alm_keplerianNAVAlmanac(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_alm_keplerianNAVAlmanac_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_alm_KEPLERIANBDSALMANAC_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_alm_KEPLERIANBDSALMANAC(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_alm_KEPLERIANBDSALMANAC_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_alm_keplerianReducedAlmanac_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_alm_keplerianReducedAlmanac(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_alm_keplerianReducedAlmanac_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_alm_keplerianMidiAlmanac_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_alm_keplerianMidiAlmanac(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_alm_keplerianMidiAlmanac_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_alm_keplerianGLONASS_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_alm_keplerianGLONASS(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_alm_keplerianGLONASS_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_alm_ecefSBASAlmanac_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_alm_ecefSBASAlmanac(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_alm_ecefSBASAlmanac_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Auxiliary_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_Auxiliary_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_Auxiliary_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_AuxInfoReq_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_AuxInfoReq(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_AuxInfoReq_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_BDSclockModel_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_BDSclockModel(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_BDSclockModel_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Common_Data_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_Common_Data(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_Common_Data_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Delta_T_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_Delta_T(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_Delta_T_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Earth_Orientation_Parameters_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_Earth_Orientation_Parameters(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_Earth_Orientation_Parameters_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_EarthOrientParaReq_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_EarthOrientParaReq(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_EarthOrientParaReq_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Generic_Data_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_Generic_Data(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_Generic_Data_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_ID(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_NavModel_BDSKeplerianSet_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_NavModel_BDSKeplerianSet(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_NavModel_BDSKeplerianSet_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_SBAS_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_SBAS_ID(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_SBAS_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_Time_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_Time_ID(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_Time_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GANSS_UTCmodelSet4_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GANSS_UTCmodelSet4(tvb, offset, &asn1_ctx, tree, hf_rnsap_GANSS_UTCmodelSet4_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GERAN_Cell_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GERAN_Cell_Capability(tvb, offset, &asn1_ctx, tree, hf_rnsap_GERAN_Cell_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GERAN_Classmark_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GERAN_Classmark(tvb, offset, &asn1_ctx, tree, hf_rnsap_GERAN_Classmark_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Satellite_Almanac_Information_ExtItem_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Satellite_Almanac_Information_ExtItem(tvb, offset, &asn1_ctx, tree, hf_rnsap_Satellite_Almanac_Information_ExtItem_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Guaranteed_Rate_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Guaranteed_Rate_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Guaranteed_Rate_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HARQ_MemoryPartitioningInfoExtForMIMO_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HARQ_MemoryPartitioningInfoExtForMIMO(tvb, offset, &asn1_ctx, tree, hf_rnsap_HARQ_MemoryPartitioningInfoExtForMIMO_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HARQ_Preamble_Mode_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HARQ_Preamble_Mode(tvb, offset, &asn1_ctx, tree, hf_rnsap_HARQ_Preamble_Mode_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HARQ_Preamble_Mode_Activation_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HARQ_Preamble_Mode_Activation_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_HARQ_Preamble_Mode_Activation_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HCS_Prio_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HCS_Prio(tvb, offset, &asn1_ctx, tree, hf_rnsap_HCS_Prio_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Configured_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_Configured_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_Configured_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_FDD_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_FDD_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_FDD_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_FDD_Information_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_FDD_Information_Response(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_FDD_Information_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Information_to_Modify_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_Information_to_Modify(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_Information_to_Modify_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Information_to_Modify_Unsynchronised_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_Information_to_Modify_Unsynchronised(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_Information_to_Modify_Unsynchronised_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_MACdPDUSizeFormat_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_MACdPDUSizeFormat(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_MACdPDUSizeFormat_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_MACdFlows_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_MACdFlows_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_MACdFlows_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_MACdFlows_to_Delete_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_MACdFlows_to_Delete(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_MACdFlows_to_Delete_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_PreconfigurationInfo_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_PreconfigurationInfo(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_PreconfigurationInfo_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Preconfiguration_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_EDCH_Preconfiguration_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_EDCH_Preconfiguration_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_PreconfigurationSetup_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_PreconfigurationSetup(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_PreconfigurationSetup_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_RNTI_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_RNTI(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_RNTI_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCH_serving_cell_change_information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HS_DSCH_serving_cell_change_information(tvb, offset, &asn1_ctx, tree, hf_rnsap_HS_DSCH_serving_cell_change_information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCH_serving_cell_change_informationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HS_DSCH_serving_cell_change_informationResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_HS_DSCH_serving_cell_change_informationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_TBSizeTableIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_TBSizeTableIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_TBSizeTableIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_TDD_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_TDD_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_TDD_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_TDD_Information_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_TDD_Information_Response(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_TDD_Information_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multicarrier_Number_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multicarrier_Number(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multicarrier_Number_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSPDSCH_TDD_Specific_InfoList_Response768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSPDSCH_TDD_Specific_InfoList_Response768(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSPDSCH_TDD_Specific_InfoList_Response768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSSCCH_TDD_Specific_InfoList_Response768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSSCCH_TDD_Specific_InfoList_Response768(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSSCCH_TDD_Specific_InfoList_Response768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SICH_Reception_Quality_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HS_SICH_Reception_Quality_Value(tvb, offset, &asn1_ctx, tree, hf_rnsap_HS_SICH_Reception_Quality_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SICH_Reception_Quality_Measurement_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HS_SICH_Reception_Quality_Measurement_Value(tvb, offset, &asn1_ctx, tree, hf_rnsap_HS_SICH_Reception_Quality_Measurement_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SICH_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HS_SICH_ID(tvb, offset, &asn1_ctx, tree, hf_rnsap_HS_SICH_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_SICH_ID_Extension_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HS_SICH_ID_Extension(tvb, offset, &asn1_ctx, tree, hf_rnsap_HS_SICH_ID_Extension_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_PDSCH_Code_Change_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HS_PDSCH_Code_Change_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_HS_PDSCH_Code_Change_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_PDSCH_Code_Change_Grant_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HS_PDSCH_Code_Change_Grant(tvb, offset, &asn1_ctx, tree, hf_rnsap_HS_PDSCH_Code_Change_Grant_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_FDD_Update_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_FDD_Update_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_FDD_Update_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_TDD_Update_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_TDD_Update_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_TDD_Update_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_ReferenceSignal_InformationListLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MIMO_ReferenceSignal_InformationListLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_MIMO_ReferenceSignal_InformationListLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCH_Semi_PersistentScheduling_Information_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCH_Physical_Layer_Category_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCH_Physical_Layer_Category(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCH_Physical_Layer_Category_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TS0_HS_PDSCH_Indication_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TS0_HS_PDSCH_Indication_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_TS0_HS_PDSCH_Indication_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IMSI_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_IMSI(tvb, offset, &asn1_ctx, tree, hf_rnsap_IMSI_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Improved_Synchronized_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Improved_Synchronized_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_Improved_Synchronized_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Inactivity_Threshold_for_UE_DRX_Cycle_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle(tvb, offset, &asn1_ctx, tree, hf_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InformationExchangeID(tvb, offset, &asn1_ctx, tree, hf_rnsap_InformationExchangeID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationReportCharacteristics_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InformationReportCharacteristics(tvb, offset, &asn1_ctx, tree, hf_rnsap_InformationReportCharacteristics_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationType_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InformationType(tvb, offset, &asn1_ctx, tree, hf_rnsap_InformationType_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Initial_DL_DPCH_TimingAdjustment_Allowed_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Initial_DL_DPCH_TimingAdjustment_Allowed(tvb, offset, &asn1_ctx, tree, hf_rnsap_Initial_DL_DPCH_TimingAdjustment_Allowed_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InnerLoopDLPCStatus_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InnerLoopDLPCStatus(tvb, offset, &asn1_ctx, tree, hf_rnsap_InnerLoopDLPCStatus_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Inter_Frequency_Cell_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Inter_Frequency_Cell_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_Inter_Frequency_Cell_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Inter_Frequency_Cell_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Inter_Frequency_Cell_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Inter_Frequency_Cell_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Inter_Frequency_Cell_Information_SIB11ter_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Inter_Frequency_Cell_Information_SIB11ter(tvb, offset, &asn1_ctx, tree, hf_rnsap_Inter_Frequency_Cell_Information_SIB11ter_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IPDL_TDD_ParametersLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_IPDL_TDD_ParametersLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_IPDL_TDD_ParametersLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IdleIntervalInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_IdleIntervalInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_IdleIntervalInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_L3_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_L3_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_L3_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Load_Value_IncrDecrThres_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Load_Value_IncrDecrThres(tvb, offset, &asn1_ctx, tree, hf_rnsap_Load_Value_IncrDecrThres_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Load_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Load_Value(tvb, offset, &asn1_ctx, tree, hf_rnsap_Load_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_LoadValue_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_LoadValue(tvb, offset, &asn1_ctx, tree, hf_rnsap_LoadValue_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_LCRTDD_Uplink_Physical_Channel_Capability_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_LCRTDD_Uplink_Physical_Channel_Capability(tvb, offset, &asn1_ctx, tree, hf_rnsap_LCRTDD_Uplink_Physical_Channel_Capability_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MDT_Configuration_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MDT_Configuration(tvb, offset, &asn1_ctx, tree, hf_rnsap_MDT_Configuration_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MAC_PDU_SizeExtended_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MAC_PDU_SizeExtended(tvb, offset, &asn1_ctx, tree, hf_rnsap_MAC_PDU_SizeExtended_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MACes_Maximum_Bitrate_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MACes_Maximum_Bitrate_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_MACes_Maximum_Bitrate_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MaxNrDLPhysicalchannels768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MaxNrDLPhysicalchannels768(tvb, offset, &asn1_ctx, tree, hf_rnsap_MaxNrDLPhysicalchannels768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MaxNrDLPhysicalchannelsTS_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MaxNrDLPhysicalchannelsTS(tvb, offset, &asn1_ctx, tree, hf_rnsap_MaxNrDLPhysicalchannelsTS_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MaxNrDLPhysicalchannelsTS768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MaxNrDLPhysicalchannelsTS768(tvb, offset, &asn1_ctx, tree, hf_rnsap_MaxNrDLPhysicalchannelsTS768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MaxNr_Retransmissions_EDCH_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MaxNr_Retransmissions_EDCH(tvb, offset, &asn1_ctx, tree, hf_rnsap_MaxNr_Retransmissions_EDCH_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Max_UE_DTX_Cycle_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Max_UE_DTX_Cycle(tvb, offset, &asn1_ctx, tree, hf_rnsap_Max_UE_DTX_Cycle_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMS_Bearer_Service_Full_Address_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBMS_Bearer_Service_Full_Address(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBMS_Bearer_Service_Full_Address_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMS_Bearer_Service_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBMS_Bearer_Service_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBMS_Bearer_Service_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMS_Neighbouring_Cell_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBMS_Neighbouring_Cell_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBMS_Neighbouring_Cell_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMS_RLC_Sequence_Number_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBMS_RLC_Sequence_Number_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBMS_RLC_Sequence_Number_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBSFN_Cluster_Identity_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBSFN_Cluster_Identity(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBSFN_Cluster_Identity_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MCCH_Message_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MCCH_Message_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_MCCH_Message_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MCCH_Configuration_Info_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MCCH_Configuration_Info(tvb, offset, &asn1_ctx, tree, hf_rnsap_MCCH_Configuration_Info_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBSFN_Scheduling_Transmission_Time_Interval_Info_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Info_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Info_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MeasurementFilterCoefficient_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MeasurementFilterCoefficient(tvb, offset, &asn1_ctx, tree, hf_rnsap_MeasurementFilterCoefficient_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MeasurementID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MeasurementID(tvb, offset, &asn1_ctx, tree, hf_rnsap_MeasurementID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Measurement_Power_Offset_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Measurement_Power_Offset(tvb, offset, &asn1_ctx, tree, hf_rnsap_Measurement_Power_Offset_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MinimumSpreadingFactor768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MinimumSpreadingFactor768(tvb, offset, &asn1_ctx, tree, hf_rnsap_MinimumSpreadingFactor768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MaxAdjustmentStep_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MaxAdjustmentStep(tvb, offset, &asn1_ctx, tree, hf_rnsap_MaxAdjustmentStep_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MeasurementRecoveryBehavior_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MeasurementRecoveryBehavior(tvb, offset, &asn1_ctx, tree, hf_rnsap_MeasurementRecoveryBehavior_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MeasurementRecoveryReportingIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MeasurementRecoveryReportingIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_MeasurementRecoveryReportingIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MeasurementRecoverySupportIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MeasurementRecoverySupportIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_MeasurementRecoverySupportIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_ActivationIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MIMO_ActivationIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_MIMO_ActivationIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_InformationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MIMO_InformationResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_MIMO_InformationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_Mode_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MIMO_Mode_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_MIMO_Mode_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_N_M_Ratio_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MIMO_N_M_Ratio(tvb, offset, &asn1_ctx, tree, hf_rnsap_MIMO_N_M_Ratio_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_withfourtransmitantennas_ActivationIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MIMO_withfourtransmitantennas_ActivationIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_MIMO_withfourtransmitantennas_ActivationIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_withfourtransmitantennas_ModeIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MIMO_withfourtransmitantennas_ModeIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_MIMO_withfourtransmitantennas_ModeIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DualStream_MIMO_withfourtransmitantennas_ActivationIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DualStream_MIMO_withfourtransmitantennas_ActivationIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_DualStream_MIMO_withfourtransmitantennas_ActivationIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DualStream_MIMO_withfourtransmitantennas_ModeIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DualStream_MIMO_withfourtransmitantennas_ModeIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_DualStream_MIMO_withfourtransmitantennas_ModeIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_Associated_Secondary_CPICH_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_Associated_Secondary_CPICH(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_Associated_Secondary_CPICH_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MinimumReducedE_DPDCH_GainFactor_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MinimumReducedE_DPDCH_GainFactor(tvb, offset, &asn1_ctx, tree, hf_rnsap_MinimumReducedE_DPDCH_GainFactor_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MulticellEDCH_InformationItemIEs_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MulticellEDCH_InformationItemIEs(tvb, offset, &asn1_ctx, tree, hf_rnsap_MulticellEDCH_InformationItemIEs_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MulticellEDCH_RL_SpecificInformationItemIEs_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MulticellEDCH_RL_SpecificInformationItemIEs(tvb, offset, &asn1_ctx, tree, hf_rnsap_MulticellEDCH_RL_SpecificInformationItemIEs_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multicell_EDCH_Restriction_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multicell_EDCH_Restriction(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multicell_EDCH_Restriction_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_PLMN_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multiple_PLMN_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multiple_PLMN_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MAChs_ResetIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MAChs_ResetIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_MAChs_ResetIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MIMO_SFMode_For_HSPDSCHDualStream_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MIMO_SFMode_For_HSPDSCHDualStream(tvb, offset, &asn1_ctx, tree, hf_rnsap_MIMO_SFMode_For_HSPDSCHDualStream_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multi_Carrier_EDCH_Info_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multi_Carrier_EDCH_Info(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multi_Carrier_EDCH_Info_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SNPL_Carrier_Group_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SNPL_Carrier_Group_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_SNPL_Carrier_Group_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multi_Carrier_EDCH_Reconfigure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multi_Carrier_EDCH_Reconfigure(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multi_Carrier_EDCH_Reconfigure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multi_Carrier_EDCH_Information_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multi_Carrier_EDCH_Information_Response(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multi_Carrier_EDCH_Information_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiflow_Reconfiguration_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multiflow_Reconfiguration(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multiflow_Reconfiguration_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiflow_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multiflow_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multiflow_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiflow_RepetitionFactors_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multiflow_RepetitionFactors(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multiflow_RepetitionFactors_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiflow_OrdinalNumberOfFrequency_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multiflow_OrdinalNumberOfFrequency(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multiflow_OrdinalNumberOfFrequency_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MU_MIMO_InformationLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MU_MIMO_InformationLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_MU_MIMO_InformationLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MU_MIMO_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MU_MIMO_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_MU_MIMO_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NACC_Related_Data_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_NACC_Related_Data(tvb, offset, &asn1_ctx, tree, hf_rnsap_NACC_Related_Data_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Neighbouring_UMTS_CellInformation_Ext_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Neighbouring_UMTS_CellInformation_Ext(tvb, offset, &asn1_ctx, tree, hf_rnsap_Neighbouring_UMTS_CellInformation_Ext_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Neighbouring_UMTS_CellInformationExtensionItem_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Neighbouring_UMTS_CellInformationExtensionItem(tvb, offset, &asn1_ctx, tree, hf_rnsap_Neighbouring_UMTS_CellInformationExtensionItem_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Neighbouring_UMTS_CellInformationItem_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Neighbouring_UMTS_CellInformationItem(tvb, offset, &asn1_ctx, tree, hf_rnsap_Neighbouring_UMTS_CellInformationItem_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFrequencyBandsIndicatorList_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MultipleFrequencyBandsIndicatorList(tvb, offset, &asn1_ctx, tree, hf_rnsap_MultipleFrequencyBandsIndicatorList_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Neighbouring_GSM_CellInformationIEs_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Neighbouring_GSM_CellInformationIEs(tvb, offset, &asn1_ctx, tree, hf_rnsap_Neighbouring_GSM_CellInformationIEs_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NeighbouringTDDCellMeasurementInformationLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_NeighbouringTDDCellMeasurementInformationLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_NeighbouringTDDCellMeasurementInformationLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NeighbouringTDDCellMeasurementInformation768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_NeighbouringTDDCellMeasurementInformation768(tvb, offset, &asn1_ctx, tree, hf_rnsap_NeighbouringTDDCellMeasurementInformation768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Neighbouring_LCR_TDD_CellInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Neighbouring_LCR_TDD_CellInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_Neighbouring_LCR_TDD_CellInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Neighbouring_E_UTRA_CellInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Neighbouring_E_UTRA_CellInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_Neighbouring_E_UTRA_CellInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MultipleFrequencyBandsIndicatorList_EUTRA_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MultipleFrequencyBandsIndicatorList_EUTRA(tvb, offset, &asn1_ctx, tree, hf_rnsap_MultipleFrequencyBandsIndicatorList_EUTRA_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NonCellSpecificTxDiversity_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_NonCellSpecificTxDiversity(tvb, offset, &asn1_ctx, tree, hf_rnsap_NonCellSpecificTxDiversity_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Non_rectangular_resource_allocation_indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Non_rectangular_resource_allocation_indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_Non_rectangular_resource_allocation_indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Non_rectangular_resource_timeslot_set_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Non_rectangular_resource_timeslot_set(tvb, offset, &asn1_ctx, tree, hf_rnsap_Non_rectangular_resource_timeslot_set_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NRT_Load_Information_Value_IncrDecrThres_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_NRT_Load_Information_Value_IncrDecrThres(tvb, offset, &asn1_ctx, tree, hf_rnsap_NRT_Load_Information_Value_IncrDecrThres_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NRT_Load_Information_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_NRT_Load_Information_Value(tvb, offset, &asn1_ctx, tree, hf_rnsap_NRT_Load_Information_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NRTLoadInformationValue_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_NRTLoadInformationValue(tvb, offset, &asn1_ctx, tree, hf_rnsap_NRTLoadInformationValue_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Number_Of_Supported_Carriers_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Number_Of_Supported_Carriers(tvb, offset, &asn1_ctx, tree, hf_rnsap_Number_Of_Supported_Carriers_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NoOfTargetCellHS_SCCH_Order_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_NoOfTargetCellHS_SCCH_Order(tvb, offset, &asn1_ctx, tree, hf_rnsap_NoOfTargetCellHS_SCCH_Order_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Non_Serving_RL_Preconfig_Setup_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Non_Serving_RL_Preconfig_Setup(tvb, offset, &asn1_ctx, tree, hf_rnsap_Non_Serving_RL_Preconfig_Setup_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NeedforIdleInterval_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_NeedforIdleInterval(tvb, offset, &asn1_ctx, tree, hf_rnsap_NeedforIdleInterval_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_OnModification_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_OnModification(tvb, offset, &asn1_ctx, tree, hf_rnsap_OnModification_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_OrdinalNumberOfFrequency_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_OrdinalNumberOfFrequency(tvb, offset, &asn1_ctx, tree, hf_rnsap_OrdinalNumberOfFrequency_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Out_of_Sychronization_Window_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Out_of_Sychronization_Window(tvb, offset, &asn1_ctx, tree, hf_rnsap_Out_of_Sychronization_Window_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PartialReportingIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PartialReportingIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_PartialReportingIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PCI_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PCI(tvb, offset, &asn1_ctx, tree, hf_rnsap_PCI_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Permanent_NAS_UE_Identity_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Permanent_NAS_UE_Identity(tvb, offset, &asn1_ctx, tree, hf_rnsap_Permanent_NAS_UE_Identity_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Phase_Reference_Update_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Phase_Reference_Update_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_Phase_Reference_Update_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PLMN_Identity_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PLMN_Identity(tvb, offset, &asn1_ctx, tree, hf_rnsap_PLMN_Identity_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PLMN_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PLMN_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_PLMN_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PowerAdjustmentType_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PowerAdjustmentType(tvb, offset, &asn1_ctx, tree, hf_rnsap_PowerAdjustmentType_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PowerOffsetForSecondaryCPICHforMIMO_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PowerOffsetForSecondaryCPICHforMIMO(tvb, offset, &asn1_ctx, tree, hf_rnsap_PowerOffsetForSecondaryCPICHforMIMO_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PowerOffsetForSecondaryCPICHforMIMORequestIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PowerOffsetForSecondaryCPICHforMIMORequestIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_PowerOffsetForSecondaryCPICHforMIMORequestIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas(tvb, offset, &asn1_ctx, tree, hf_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennasRequestIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennasRequestIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennasRequestIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Precoder_Weight_Set_Restriction_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Precoder_Weight_Set_Restriction(tvb, offset, &asn1_ctx, tree, hf_rnsap_Precoder_Weight_Set_Restriction_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Primary_CPICH_Usage_For_Channel_Estimation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Primary_CPICH_Usage_For_Channel_Estimation(tvb, offset, &asn1_ctx, tree, hf_rnsap_Primary_CPICH_Usage_For_Channel_Estimation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrimaryCCPCH_RSCP_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PrimaryCCPCH_RSCP(tvb, offset, &asn1_ctx, tree, hf_rnsap_PrimaryCCPCH_RSCP_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrimaryCCPCH_RSCP_Delta_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PrimaryCCPCH_RSCP_Delta(tvb, offset, &asn1_ctx, tree, hf_rnsap_PrimaryCCPCH_RSCP_Delta_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PropagationDelay_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PropagationDelay(tvb, offset, &asn1_ctx, tree, hf_rnsap_PropagationDelay_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ProvidedInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ProvidedInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_ProvidedInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UpPCH_InformationList_LCRTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UpPCH_InformationList_LCRTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UpPCH_InformationList_LCRTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UpPCH_InformationItem_LCRTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UpPCH_InformationItem_LCRTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UpPCH_InformationItem_LCRTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RANAP_EnhancedRelocationInformationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RANAP_EnhancedRelocationInformationRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_RANAP_EnhancedRelocationInformationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RANAP_EnhancedRelocationInformationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RANAP_EnhancedRelocationInformationResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_RANAP_EnhancedRelocationInformationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RANAP_RelocationInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RANAP_RelocationInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_RANAP_RelocationInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Received_Total_Wideband_Power_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Received_Total_Wideband_Power_Value(tvb, offset, &asn1_ctx, tree, hf_rnsap_Received_Total_Wideband_Power_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Received_Total_Wideband_Power_Value_IncrDecrThres_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Received_Total_Wideband_Power_Value_IncrDecrThres(tvb, offset, &asn1_ctx, tree, hf_rnsap_Received_Total_Wideband_Power_Value_IncrDecrThres_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Released_CN_Domain_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Released_CN_Domain(tvb, offset, &asn1_ctx, tree, hf_rnsap_Released_CN_Domain_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ReportCharacteristics_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ReportCharacteristics(tvb, offset, &asn1_ctx, tree, hf_rnsap_ReportCharacteristics_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RestrictionStateIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RestrictionStateIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_RestrictionStateIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_ID(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Set_ID(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Set_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Specific_DCH_Info_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Specific_DCH_Info(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Specific_DCH_Info_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Specific_EDCH_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Specific_EDCH_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Specific_EDCH_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_RLC_PDU_Size_Format_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_RLC_PDU_Size_Format(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_RLC_PDU_Size_Format_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RLC_Sequence_Number_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RLC_Sequence_Number(tvb, offset, &asn1_ctx, tree, hf_rnsap_RLC_Sequence_Number_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RNC_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RNC_ID(tvb, offset, &asn1_ctx, tree, hf_rnsap_RNC_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RNTI_Allocation_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RNTI_Allocation_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_RNTI_Allocation_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RT_Load_Value_IncrDecrThres_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RT_Load_Value_IncrDecrThres(tvb, offset, &asn1_ctx, tree, hf_rnsap_RT_Load_Value_IncrDecrThres_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RT_Load_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RT_Load_Value(tvb, offset, &asn1_ctx, tree, hf_rnsap_RT_Load_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RTLoadValue_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RTLoadValue(tvb, offset, &asn1_ctx, tree, hf_rnsap_RTLoadValue_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RxTimingDeviationForTA_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RxTimingDeviationForTA(tvb, offset, &asn1_ctx, tree, hf_rnsap_RxTimingDeviationForTA_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RxTimingDeviationForTAext_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RxTimingDeviationForTAext(tvb, offset, &asn1_ctx, tree, hf_rnsap_RxTimingDeviationForTAext_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RxTimingDeviationForTA768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RxTimingDeviationForTA768(tvb, offset, &asn1_ctx, tree, hf_rnsap_RxTimingDeviationForTA768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Rx_Timing_Deviation_Value_ext_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Rx_Timing_Deviation_Value_ext(tvb, offset, &asn1_ctx, tree, hf_rnsap_Rx_Timing_Deviation_Value_ext_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Rx_Timing_Deviation_Value_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Rx_Timing_Deviation_Value_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_Rx_Timing_Deviation_Value_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Rx_Timing_Deviation_Value_768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Rx_Timing_Deviation_Value_768(tvb, offset, &asn1_ctx, tree, hf_rnsap_Rx_Timing_Deviation_Value_768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SAI_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SAI(tvb, offset, &asn1_ctx, tree, hf_rnsap_SAI_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ScaledAdjustmentRatio_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ScaledAdjustmentRatio(tvb, offset, &asn1_ctx, tree, hf_rnsap_ScaledAdjustmentRatio_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CPICH_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Secondary_CPICH_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Secondary_CPICH_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CPICH_Information_Change_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Secondary_CPICH_Information_Change(tvb, offset, &asn1_ctx, tree, hf_rnsap_Secondary_CPICH_Information_Change_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_LCR_CCPCH_Info_TDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Secondary_LCR_CCPCH_Info_TDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_Secondary_LCR_CCPCH_Info_TDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_CCPCH_Info_TDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Secondary_CCPCH_Info_TDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_Secondary_CCPCH_Info_TDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Secondary_Serving_Cell_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Secondary_Serving_Cell_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_Secondary_Serving_Cell_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SFN_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SFN(tvb, offset, &asn1_ctx, tree, hf_rnsap_SFN_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SFNSFNMeasurementThresholdInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SFNSFNMeasurementThresholdInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_SFNSFNMeasurementThresholdInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Single_Stream_MIMO_ActivationIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Single_Stream_MIMO_ActivationIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_Single_Stream_MIMO_ActivationIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Single_Stream_MIMO_Mode_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Single_Stream_MIMO_Mode_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_Single_Stream_MIMO_Mode_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SixteenQAM_UL_Operation_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SixteenQAM_UL_Operation_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_SixteenQAM_UL_Operation_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SixtyfourQAM_UL_Operation_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SixtyfourQAM_UL_Operation_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_SixtyfourQAM_UL_Operation_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SixtyfourQAM_UsageAllowedIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SixtyfourQAM_UsageAllowedIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_SixtyfourQAM_UsageAllowedIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SixtyfourQAM_DL_SupportIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SixtyfourQAM_DL_SupportIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_SixtyfourQAM_DL_SupportIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SixtyfourQAM_DL_UsageIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SixtyfourQAM_DL_UsageIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_SixtyfourQAM_DL_UsageIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SNA_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SNA_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_SNA_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SourceID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SourceID(tvb, offset, &asn1_ctx, tree, hf_rnsap_SourceID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SpeechVersion_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SpeechVersion(tvb, offset, &asn1_ctx, tree, hf_rnsap_SpeechVersion_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_S_RNTI_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_S_RNTI(tvb, offset, &asn1_ctx, tree, hf_rnsap_S_RNTI_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_STTD_SupportIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_STTD_SupportIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_STTD_SupportIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Support_8PSK_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Support_8PSK(tvb, offset, &asn1_ctx, tree, hf_rnsap_Support_8PSK_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Support_PLCCH_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Support_PLCCH(tvb, offset, &asn1_ctx, tree, hf_rnsap_Support_PLCCH_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order(tvb, offset, &asn1_ctx, tree, hf_rnsap_Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TAC_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TAC(tvb, offset, &asn1_ctx, tree, hf_rnsap_TAC_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TargetID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TargetID(tvb, offset, &asn1_ctx, tree, hf_rnsap_TargetID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TDD_DCHs_to_Modify_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TDD_DCHs_to_Modify(tvb, offset, &asn1_ctx, tree, hf_rnsap_TDD_DCHs_to_Modify_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TDD_DL_DPCH_TimeSlotFormat_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TDD_TPC_DownlinkStepSize_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TDD_TPC_DownlinkStepSize(tvb, offset, &asn1_ctx, tree, hf_rnsap_TDD_TPC_DownlinkStepSize_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TDD_TPC_UplinkStepSize_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TDD_TPC_UplinkStepSize_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_TDD_TPC_UplinkStepSize_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TDD_UL_DPCH_TimeSlotFormat_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TDD_UL_DPCH_TimeSlotFormat_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_TDD_UL_DPCH_TimeSlotFormat_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TimeSlot_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TimeSlot(tvb, offset, &asn1_ctx, tree, hf_rnsap_TimeSlot_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SynchronisationIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SynchronisationIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_SynchronisationIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TnlQos_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TnlQos(tvb, offset, &asn1_ctx, tree, hf_rnsap_TnlQos_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TraceDepth_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TraceDepth(tvb, offset, &asn1_ctx, tree, hf_rnsap_TraceDepth_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TraceRecordingSessionReference_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TraceRecordingSessionReference(tvb, offset, &asn1_ctx, tree, hf_rnsap_TraceRecordingSessionReference_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TraceReference_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TraceReference(tvb, offset, &asn1_ctx, tree, hf_rnsap_TraceReference_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TrafficClass_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TrafficClass(tvb, offset, &asn1_ctx, tree, hf_rnsap_TrafficClass_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Transmission_Gap_Pattern_Sequence_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Transmission_Gap_Pattern_Sequence_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Transmission_Gap_Pattern_Sequence_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Affected_HSDSCH_Serving_Cell_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Affected_HSDSCH_Serving_Cell_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_Affected_HSDSCH_Serving_Cell_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Transmission_Mode_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Transmission_Mode_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Transmission_Mode_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransmitDiversityIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TransmitDiversityIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_TransmitDiversityIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Transmitted_Carrier_Power_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Transmitted_Carrier_Power_Value(tvb, offset, &asn1_ctx, tree, hf_rnsap_Transmitted_Carrier_Power_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Transmitted_Carrier_Power_Value_IncrDecrThres_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Transmitted_Carrier_Power_Value_IncrDecrThres(tvb, offset, &asn1_ctx, tree, hf_rnsap_Transmitted_Carrier_Power_Value_IncrDecrThres_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TUTRANGANSSMeasurementThresholdInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TUTRANGANSSMeasurementThresholdInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_TUTRANGANSSMeasurementThresholdInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TUTRANGANSSMeasurementValueInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TUTRANGANSSMeasurementValueInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_TUTRANGANSSMeasurementValueInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TUTRANGPSMeasurementThresholdInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TUTRANGPSMeasurementThresholdInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_TUTRANGPSMeasurementThresholdInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransportBearerID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TransportBearerID(tvb, offset, &asn1_ctx, tree, hf_rnsap_TransportBearerID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransportBearerRequestIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TransportBearerRequestIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_TransportBearerRequestIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransportBearerNotRequestedIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TransportBearerNotRequestedIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_TransportBearerNotRequestedIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransportBearerNotSetupIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TransportBearerNotSetupIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_TransportBearerNotSetupIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TPC_slot_position_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TPC_slot_position(tvb, offset, &asn1_ctx, tree, hf_rnsap_TPC_slot_position_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TransportLayerAddress_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TransportLayerAddress(tvb, offset, &asn1_ctx, tree, hf_rnsap_TransportLayerAddress_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TrCH_SrcStatisticsDescr_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TrCH_SrcStatisticsDescr(tvb, offset, &asn1_ctx, tree, hf_rnsap_TrCH_SrcStatisticsDescr_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TSN_Length_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TSN_Length(tvb, offset, &asn1_ctx, tree, hf_rnsap_TSN_Length_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TSTD_Support_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TSTD_Support_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_TSTD_Support_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TTI_Update_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TTI_Update_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_TTI_Update_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_TypeOfError_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_TypeOfError(tvb, offset, &asn1_ctx, tree, hf_rnsap_TypeOfError_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UARFCN_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UARFCN(tvb, offset, &asn1_ctx, tree, hf_rnsap_UARFCN_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_AggregateMaximumBitRate_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UE_AggregateMaximumBitRate(tvb, offset, &asn1_ctx, tree, hf_rnsap_UE_AggregateMaximumBitRate_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_AggregateMaximumBitRate_Enforcement_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UE_AggregateMaximumBitRate_Enforcement_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_UE_AggregateMaximumBitRate_Enforcement_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_Capabilities_Info_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UE_Capabilities_Info(tvb, offset, &asn1_ctx, tree, hf_rnsap_UE_Capabilities_Info_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_Support_of_non_rectangular_resource_allocation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UE_Support_of_non_rectangular_resource_allocation(tvb, offset, &asn1_ctx, tree, hf_rnsap_UE_Support_of_non_rectangular_resource_allocation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_TS0_CapabilityLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UE_TS0_CapabilityLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_UE_TS0_CapabilityLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_RF_Band_CapabilityLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UE_RF_Band_CapabilityLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_UE_RF_Band_CapabilityLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_LCRTDD_HSDSCH_Physical_Layer_Category_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_LCRTDD_HSDSCH_Physical_Layer_Category(tvb, offset, &asn1_ctx, tree, hf_rnsap_LCRTDD_HSDSCH_Physical_Layer_Category_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_DRX_Cycle_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UE_DRX_Cycle(tvb, offset, &asn1_ctx, tree, hf_rnsap_UE_DRX_Cycle_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEIdentity_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEIdentity(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEIdentity_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementParameterModAllow_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementParameterModAllow(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementParameterModAllow_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementReportCharacteristics_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementReportCharacteristics(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementReportCharacteristics_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementTimeslotInfoHCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementTimeslotInfoHCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementTimeslotInfoHCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementTimeslotInfoLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementTimeslotInfoLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementTimeslotInfoLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementTimeslotInfo768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementTimeslotInfo768(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementTimeslotInfo768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementType_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementType(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementType_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementValueTransmittedPowerList768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementValueTransmittedPowerList768(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementValueTransmittedPowerList768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementValueTimeslotISCPList768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementValueTimeslotISCPList768(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementValueTimeslotISCPList768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementValueInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementValueInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementValueInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_SupportIndicatorExtension_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UE_SupportIndicatorExtension(tvb, offset, &asn1_ctx, tree, hf_rnsap_UE_SupportIndicatorExtension_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_State_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UE_State(tvb, offset, &asn1_ctx, tree, hf_rnsap_UE_State_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_transmission_power_headroom_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UE_transmission_power_headroom_Value(tvb, offset, &asn1_ctx, tree, hf_rnsap_UE_transmission_power_headroom_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPDCHIndicatorEDCH_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPDCHIndicatorEDCH(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPDCHIndicatorEDCH_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PLCCHinformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PLCCHinformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_PLCCHinformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_Timeslot_ISCP_Value_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_Timeslot_ISCP_Value(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_Timeslot_ISCP_Value_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_Timeslot_ISCP_Value_IncrDecrThres_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_Timeslot_ISCP_Value_IncrDecrThres(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_Timeslot_ISCP_Value_IncrDecrThres_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_TimingAdvanceCtrl_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_TimingAdvanceCtrl_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_TimingAdvanceCtrl_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_SIR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_SIR(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_SIR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UC_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UC_ID(tvb, offset, &asn1_ctx, tree, hf_rnsap_UC_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_Synchronisation_Parameters_LCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_Synchronisation_Parameters_LCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_Synchronisation_Parameters_LCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UPPCHPositionLCR_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UPPCHPositionLCR(tvb, offset, &asn1_ctx, tree, hf_rnsap_UPPCHPositionLCR_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UpPTSInterferenceValue_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UpPTSInterferenceValue(tvb, offset, &asn1_ctx, tree, hf_rnsap_UpPTSInterferenceValue_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Unidirectional_DCH_Indicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Unidirectional_DCH_Indicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_Unidirectional_DCH_Indicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_URA_ID_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_URA_ID(tvb, offset, &asn1_ctx, tree, hf_rnsap_URA_ID_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_URA_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_URA_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_URA_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_USCH_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_USCH_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_USCH_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Usefulness_Of_Battery_Optimization_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Usefulness_Of_Battery_Optimization(tvb, offset, &asn1_ctx, tree, hf_rnsap_Usefulness_Of_Battery_Optimization_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_User_Plane_Congestion_Fields_Inclusion_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_User_Plane_Congestion_Fields_Inclusion(tvb, offset, &asn1_ctx, tree, hf_rnsap_User_Plane_Congestion_Fields_Inclusion_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UMTS_Cell_Info_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UMTS_Cell_Info(tvb, offset, &asn1_ctx, tree, hf_rnsap_UMTS_Cell_Info_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CLTD_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CLTD_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CLTD_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CLTD_Information_Reconf_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CLTD_Information_Reconf(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CLTD_Information_Reconf_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CLTD_State_Update_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CLTD_State_Update_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CLTD_State_Update_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCCH2_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCCH2_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCCH2_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCCH2_Information_Reconf_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCCH2_Information_Reconf(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCCH2_Information_Reconf_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_MIMO_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_MIMO_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_MIMO_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_MIMO_Reconfiguration_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_MIMO_Reconfiguration(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_MIMO_Reconfiguration_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_MIMO_DL_Control_Channel_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_MIMO_DL_Control_Channel_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_MIMO_DL_Control_Channel_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkSetupRequestFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkSetupRequestFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkSetupRequestFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_Information_RL_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_Information_RL_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_Information_RL_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_Information_RL_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_Information_RL_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_Information_RL_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationList_RL_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationList_RL_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationList_RL_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_RL_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationItem_RL_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationItem_RL_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Information_RL_Setup_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_HS_Cell_Information_RL_Setup_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_HS_Cell_Information_RL_Setup_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_F_DPCH_Information_RL_SetupRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_F_DPCH_Information_RL_SetupRqstFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_F_DPCH_Information_RL_SetupRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkSetupRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkSetupRequestTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkSetupRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_Physical_Channel_Information_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_Physical_Channel_Information_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_Physical_Channel_Information_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_Physical_Channel_Information_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_Physical_Channel_Information_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_Physical_Channel_Information_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_InformationList_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationList_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_RL_SetupRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Information_RL_SetupRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Information_RL_SetupRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkSetupResponseFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkSetupResponseFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkSetupResponseFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseList_RL_SetupRspFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationResponseList_RL_SetupRspFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationResponseList_RL_SetupRspFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseItem_RL_SetupRspFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationResponseItem_RL_SetupRspFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationResponseItem_RL_SetupRspFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Information_Response_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_HS_Cell_Information_Response_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_HS_Cell_Information_Response_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkSetupResponseTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkSetupResponseTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkSetupResponseTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponse_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationResponse_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationResponse_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCHInformationListIE_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationItem_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCHInformationListIE_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationItem_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_InformationListIEs_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DSCH_InformationListIEs_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DSCH_InformationListIEs_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_USCH_InformationListIEs_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_USCH_InformationListIEs_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_USCH_InformationListIEs_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_LCR_InformationResponse_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_LCR_InformationResponse_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_LCR_InformationResponse_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_LCR_InformationItem_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_LCR_InformationItem_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_LCR_InformationItem_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_LCR_InformationItem_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_LCR_InformationItem_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_LCR_InformationItem_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_LCR_InformationListIEs_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DSCH_LCR_InformationListIEs_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DSCH_LCR_InformationListIEs_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_USCH_LCR_InformationListIEs_RL_SetupRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_USCH_LCR_InformationListIEs_RL_SetupRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_USCH_LCR_InformationListIEs_RL_SetupRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponse_RL_SetupRspTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationResponse_RL_SetupRspTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationResponse_RL_SetupRspTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCHInformationListIE_RL_SetupRspTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationItem_RL_SetupRspTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCHInformationListIE_RL_SetupRspTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationItem_RL_SetupRspTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkSetupFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkSetupFailureFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkSetupFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CauseLevel_RL_SetupFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CauseLevel_RL_SetupFailureFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_CauseLevel_RL_SetupFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SuccessfulRL_InformationResponse_RL_SetupFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SuccessfulRL_InformationResponse_RL_SetupFailureFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_SuccessfulRL_InformationResponse_RL_SetupFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkSetupFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkSetupFailureTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkSetupFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CauseLevel_RL_SetupFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CauseLevel_RL_SetupFailureTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_CauseLevel_RL_SetupFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkAdditionRequestFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkAdditionRequestFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkAdditionRequestFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationList_RL_AdditionRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationList_RL_AdditionRqstFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationList_RL_AdditionRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_RL_AdditionRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Information_RL_AdditionRqstFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Information_RL_AdditionRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Information_RL_Addition_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_HS_Cell_Information_RL_Addition_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_HS_Cell_Information_RL_Addition_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_RL_Add_Req_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_EDCH_Cell_Information_RL_Add_Req(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_EDCH_Cell_Information_RL_Add_Req_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDPCH_Information_RLAdditionReq_FDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDPCH_Information_RLAdditionReq_FDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDPCH_Information_RLAdditionReq_FDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkAdditionRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkAdditionRequestTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkAdditionRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Information_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Information_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkAdditionResponseFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkAdditionResponseFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkAdditionResponseFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseList_RL_AdditionRspFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationResponseList_RL_AdditionRspFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationResponseList_RL_AdditionRspFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseItem_RL_AdditionRspFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationResponseItem_RL_AdditionRspFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationResponseItem_RL_AdditionRspFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Change_Information_Response_List_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_HS_Cell_Change_Information_Response_List(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_HS_Cell_Change_Information_Response_List_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkAdditionResponseTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkAdditionResponseTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkAdditionResponseTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponse_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationResponse_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationResponse_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCHInformationListIE_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationItem_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCHInformationListIE_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationItem_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_InformationListIE_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DSCH_InformationListIE_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DSCH_InformationListIE_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_USCH_InformationListIE_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_USCH_InformationListIE_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_USCH_InformationListIE_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_LCR_InformationResponse_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_LCR_InformationResponse_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_LCR_InformationResponse_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_USCH_LCR_InformationListIEs_RL_AdditionRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_USCH_LCR_InformationListIEs_RL_AdditionRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_USCH_LCR_InformationListIEs_RL_AdditionRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponse_RL_AdditionRspTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationResponse_RL_AdditionRspTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationResponse_RL_AdditionRspTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCHInformationListIE_RL_AdditionRspTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationItem_RL_AdditionRspTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCHInformationListIE_RL_AdditionRspTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationItem_RL_AdditionRspTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkAdditionFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkAdditionFailureFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkAdditionFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CauseLevel_RL_AdditionFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CauseLevel_RL_AdditionFailureFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_CauseLevel_RL_AdditionFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SuccessfulRL_InformationResponse_RL_AdditionFailureFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SuccessfulRL_InformationResponse_RL_AdditionFailureFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_SuccessfulRL_InformationResponse_RL_AdditionFailureFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkAdditionFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkAdditionFailureTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkAdditionFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CauseLevel_RL_AdditionFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CauseLevel_RL_AdditionFailureTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_CauseLevel_RL_AdditionFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkDeletionRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkDeletionRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkDeletionRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationList_RL_DeletionRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationList_RL_DeletionRqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationList_RL_DeletionRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_RL_DeletionRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Information_RL_DeletionRqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Information_RL_DeletionRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkDeletionResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkDeletionResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkDeletionResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationPrepareFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkReconfigurationPrepareFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkReconfigurationPrepareFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_Information_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_Information_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_Information_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_Information_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_Information_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_Information_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_Power_Information_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_Power_Information_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_Power_Information_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_DeleteList_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DCH_DeleteList_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DCH_DeleteList_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationList_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationList_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationList_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Information_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Information_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Information_RL_Reconf_Prep_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_RL_Reconf_Prep_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Prep(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Prep_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_F_DPCH_Information_RL_ReconfPrepFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_F_DPCH_Information_RL_ReconfPrepFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_F_DPCH_Information_RL_ReconfPrepFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationPrepareTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkReconfigurationPrepareTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkReconfigurationPrepareTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_AddInformation_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_AddInformation_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_AddInformation_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_DeleteList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DCH_DeleteList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DCH_DeleteList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_ModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DSCH_ModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DSCH_ModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCH_DeleteList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DSCH_DeleteList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DSCH_DeleteList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_USCH_ModifyList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_USCH_ModifyList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_USCH_ModifyList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_USCH_DeleteList_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_USCH_DeleteList_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_USCH_DeleteList_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_RL_ReconfPrepTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Information_RL_ReconfPrepTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Information_RL_ReconfPrepTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationReadyFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkReconfigurationReadyFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkReconfigurationReadyFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseList_RL_ReconfReadyFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationResponseList_RL_ReconfReadyFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationResponseList_RL_ReconfReadyFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseItem_RL_ReconfReadyFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationResponseItem_RL_ReconfReadyFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationResponseItem_RL_ReconfReadyFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_RL_Reconf_Response_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_HS_Cell_RL_Reconf_Response(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_HS_Cell_RL_Reconf_Response_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkReconfigurationReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkReconfigurationReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCHInformationListIE_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCHInformationListIE_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCHInformationListIE_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationAddList_RL_ReconfReadyTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_InformationAddList_RL_ReconfReadyTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_InformationAddList_RL_ReconfReadyTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCHInformationListIE_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCHInformationListIE_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCHInformationListIE_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationAddList_RL_ReconfReadyTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_InformationAddList_RL_ReconfReadyTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_InformationAddList_RL_ReconfReadyTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_RL_InformationResponse_RL_ReconfReadyTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multiple_RL_InformationResponse_RL_ReconfReadyTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multiple_RL_InformationResponse_RL_ReconfReadyTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationCommit_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkReconfigurationCommit(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkReconfigurationCommit_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkReconfigurationFailure(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkReconfigurationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CauseLevel_RL_ReconfFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CauseLevel_RL_ReconfFailure(tvb, offset, &asn1_ctx, tree, hf_rnsap_CauseLevel_RL_ReconfFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_ReconfigurationFailure_RL_ReconfFail_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_ReconfigurationFailure_RL_ReconfFail(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_ReconfigurationFailure_RL_ReconfFail_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationCancel_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkReconfigurationCancel(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkReconfigurationCancel_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationRequestFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkReconfigurationRequestFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkReconfigurationRequestFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_Information_RL_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_Information_RL_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_Information_RL_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_Information_RL_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_Information_RL_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_Information_RL_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_DeleteList_RL_ReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DCH_DeleteList_RL_ReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DCH_DeleteList_RL_ReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Information_RL_Reconf_Req_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_RL_Reconf_Req_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Req(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Req_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_ReconfigurationRequestFDD_RL_InformationList_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_ReconfigurationRequestFDD_RL_InformationList(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_ReconfigurationRequestFDD_RL_InformationList_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_ReconfigurationRequestFDD_RL_Information_IEs_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_ReconfigurationRequestFDD_RL_Information_IEs(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_ReconfigurationRequestFDD_RL_Information_IEs_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkReconfigurationRequestTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkReconfigurationRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_DeleteList_RL_ReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DCH_DeleteList_RL_ReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DCH_DeleteList_RL_ReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_RL_ReconfigurationRequestTDD_RL_Information_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multiple_RL_ReconfigurationRequestTDD_RL_Information(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multiple_RL_ReconfigurationRequestTDD_RL_Information_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationResponseFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkReconfigurationResponseFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkReconfigurationResponseFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseList_RL_ReconfRspFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationResponseList_RL_ReconfRspFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationResponseList_RL_ReconfRspFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponseItem_RL_ReconfRspFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationResponseItem_RL_ReconfRspFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationResponseItem_RL_ReconfRspFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkReconfigurationResponseTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkReconfigurationResponseTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkReconfigurationResponseTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationResponse_RL_ReconfRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationResponse_RL_ReconfRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationResponse_RL_ReconfRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationList_RL_ReconfRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationList_RL_ReconfRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationList_RL_ReconfRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_RL_InformationResponse_RL_ReconfRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multiple_RL_InformationResponse_RL_ReconfRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multiple_RL_InformationResponse_RL_ReconfRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkFailureIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkFailureIndication(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkFailureIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Reporting_Object_RL_FailureInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Reporting_Object_RL_FailureInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_Reporting_Object_RL_FailureInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_RL_FailureInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Information_RL_FailureInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Information_RL_FailureInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_Information_RL_FailureInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Set_Information_RL_FailureInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Set_Information_RL_FailureInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CCTrCH_InformationItem_RL_FailureInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CCTrCH_InformationItem_RL_FailureInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_CCTrCH_InformationItem_RL_FailureInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkPreemptionRequiredIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkPreemptionRequiredIndication(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkPreemptionRequiredIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationList_RL_PreemptRequiredInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationList_RL_PreemptRequiredInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationList_RL_PreemptRequiredInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_RL_PreemptRequiredInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationItem_RL_PreemptRequiredInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationItem_RL_PreemptRequiredInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkRestoreIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkRestoreIndication(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkRestoreIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Reporting_Object_RL_RestoreInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Reporting_Object_RL_RestoreInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_Reporting_Object_RL_RestoreInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_RL_RestoreInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Information_RL_RestoreInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Information_RL_RestoreInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_Information_RL_RestoreInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Set_Information_RL_RestoreInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Set_Information_RL_RestoreInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CCTrCH_InformationItem_RL_RestoreInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CCTrCH_InformationItem_RL_RestoreInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_CCTrCH_InformationItem_RL_RestoreInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_PowerControlRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_PowerControlRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_PowerControlRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_ReferencePowerInformationList_DL_PC_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_ReferencePowerInformationList_DL_PC_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_ReferencePowerInformationList_DL_PC_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_ReferencePowerInformation_DL_PC_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_ReferencePowerInformation_DL_PC_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_ReferencePowerInformation_DL_PC_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_PowerTimeslotControlRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_PowerTimeslotControlRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_PowerTimeslotControlRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PhysicalChannelReconfigurationRequestFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PhysicalChannelReconfigurationRequestFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_PhysicalChannelReconfigurationRequestFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_PhyChReconfRqstFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Information_PhyChReconfRqstFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Information_PhyChReconfRqstFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PhysicalChannelReconfigurationRequestTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PhysicalChannelReconfigurationRequestTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_PhysicalChannelReconfigurationRequestTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Information_PhyChReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Information_PhyChReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Information_PhyChReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_DPCH_InformationItem_PhyChReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_DPCH_InformationItem_PhyChReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_DPCH_InformationItem_PhyChReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UL_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_DPCH_InformationItem_PhyChReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_DPCH_InformationItem_PhyChReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_DPCH_InformationItem_PhyChReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DL_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PhysicalChannelReconfigurationCommand_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PhysicalChannelReconfigurationCommand(tvb, offset, &asn1_ctx, tree, hf_rnsap_PhysicalChannelReconfigurationCommand_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PhysicalChannelReconfigurationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PhysicalChannelReconfigurationFailure(tvb, offset, &asn1_ctx, tree, hf_rnsap_PhysicalChannelReconfigurationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkCongestionIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkCongestionIndication(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkCongestionIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationList_RL_CongestInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationList_RL_CongestInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationList_RL_CongestInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_RL_CongestInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationItem_RL_CongestInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationItem_RL_CongestInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DCH_Rate_InformationItem_RL_CongestInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DCH_Rate_InformationItem_RL_CongestInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_DCH_Rate_InformationItem_RL_CongestInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_MacdFlowSpecificInformationList_RL_CongestInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDCH_MacdFlowSpecificInformationList_RL_CongestInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDCH_MacdFlowSpecificInformationList_RL_CongestInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EDCH_MacdFlowSpecificInformationItem_RL_CongestInd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_CongestInd(tvb, offset, &asn1_ctx, tree, hf_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_CongestInd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UplinkSignallingTransferIndicationFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UplinkSignallingTransferIndicationFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UplinkSignallingTransferIndicationFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UplinkSignallingTransferIndicationTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UplinkSignallingTransferIndicationTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_UplinkSignallingTransferIndicationTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DownlinkSignallingTransferRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DownlinkSignallingTransferRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_DownlinkSignallingTransferRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RelocationCommit_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RelocationCommit(tvb, offset, &asn1_ctx, tree, hf_rnsap_RelocationCommit_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PagingRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PagingRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_PagingRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PagingArea_PagingRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PagingArea_PagingRqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_PagingArea_PagingRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CNOriginatedPage_PagingRqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CNOriginatedPage_PagingRqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_CNOriginatedPage_PagingRqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementInitiationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DedicatedMeasurementInitiationRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_DedicatedMeasurementInitiationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementObjectType_DM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DedicatedMeasurementObjectType_DM_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_DedicatedMeasurementObjectType_DM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_DM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationItem_DM_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationItem_DM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSSICH_Info_DM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSSICH_Info_DM_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSSICH_Info_DM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_HSSICH_Info_DM_Rqst_Extension_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_HSSICH_Info_DM_Rqst_Extension(tvb, offset, &asn1_ctx, tree, hf_rnsap_HSSICH_Info_DM_Rqst_Extension_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_InformationItem_DM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Set_InformationItem_DM_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Set_InformationItem_DM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementInitiationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DedicatedMeasurementInitiationResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_DedicatedMeasurementInitiationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementObjectType_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DedicatedMeasurementObjectType_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_DedicatedMeasurementObjectType_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationItem_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationItem_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_InformationItem_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Set_InformationItem_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Set_InformationItem_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementInitiationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DedicatedMeasurementInitiationFailure(tvb, offset, &asn1_ctx, tree, hf_rnsap_DedicatedMeasurementInitiationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementObjectType_DM_Fail_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DedicatedMeasurementObjectType_DM_Fail(tvb, offset, &asn1_ctx, tree, hf_rnsap_DedicatedMeasurementObjectType_DM_Fail_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Unsuccessful_InformationItem_DM_Fail_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Unsuccessful_InformationItem_DM_Fail(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Unsuccessful_InformationItem_DM_Fail_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Successful_InformationItem_DM_Fail_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Successful_InformationItem_DM_Fail(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Successful_InformationItem_DM_Fail_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_Unsuccessful_InformationItem_DM_Fail_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_Successful_InformationItem_DM_Fail_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Set_Successful_InformationItem_DM_Fail(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Set_Successful_InformationItem_DM_Fail_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementReport_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DedicatedMeasurementReport(tvb, offset, &asn1_ctx, tree, hf_rnsap_DedicatedMeasurementReport_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementObjectType_DM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DedicatedMeasurementObjectType_DM_Rprt(tvb, offset, &asn1_ctx, tree, hf_rnsap_DedicatedMeasurementObjectType_DM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_InformationItem_DM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_InformationItem_DM_Rprt(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_InformationItem_DM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_InformationItem_DM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Set_InformationItem_DM_Rprt(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Set_InformationItem_DM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementTerminationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DedicatedMeasurementTerminationRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_DedicatedMeasurementTerminationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementFailureIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DedicatedMeasurementFailureIndication(tvb, offset, &asn1_ctx, tree, hf_rnsap_DedicatedMeasurementFailureIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DedicatedMeasurementObjectType_DM_Fail_Ind_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DedicatedMeasurementObjectType_DM_Fail_Ind(tvb, offset, &asn1_ctx, tree, hf_rnsap_DedicatedMeasurementObjectType_DM_Fail_Ind_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Unsuccessful_InformationItem_DM_Fail_Ind_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Unsuccessful_InformationItem_DM_Fail_Ind(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Unsuccessful_InformationItem_DM_Fail_Ind_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelResourcesReleaseRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonTransportChannelResourcesReleaseRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonTransportChannelResourcesReleaseRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelResourcesRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonTransportChannelResourcesRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonTransportChannelResourcesRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelResourcesResponseFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonTransportChannelResourcesResponseFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonTransportChannelResourcesResponseFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelResourcesResponseTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonTransportChannelResourcesResponseTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonTransportChannelResourcesResponseTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonTransportChannelResourcesFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonTransportChannelResourcesFailure(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonTransportChannelResourcesFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CompressedModeCommand_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CompressedModeCommand(tvb, offset, &asn1_ctx, tree, hf_rnsap_CompressedModeCommand_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ErrorIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ErrorIndication(tvb, offset, &asn1_ctx, tree, hf_rnsap_ErrorIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementInitiationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonMeasurementInitiationRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonMeasurementInitiationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementObjectType_CM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonMeasurementObjectType_CM_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonMeasurementObjectType_CM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GSM_Cell_CM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GSM_Cell_CM_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_GSM_Cell_CM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extension_CommonMeasurementObjectType_CM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Extension_CommonMeasurementObjectType_CM_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_Extension_CommonMeasurementObjectType_CM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GsmCellList_CM_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GsmCellList_CM_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_GsmCellList_CM_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementInitiationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonMeasurementInitiationResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonMeasurementInitiationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementObjectType_CM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonMeasurementObjectType_CM_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonMeasurementObjectType_CM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extension_CommonMeasurementObjectType_CM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Extension_CommonMeasurementObjectType_CM_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_Extension_CommonMeasurementObjectType_CM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GsmCellList_CM_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GsmCellList_CM_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_GsmCellList_CM_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementInitiationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonMeasurementInitiationFailure(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonMeasurementInitiationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extension_FailureMeasurementList_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Extension_FailureMeasurementList(tvb, offset, &asn1_ctx, tree, hf_rnsap_Extension_FailureMeasurementList_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementReport_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonMeasurementReport(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonMeasurementReport_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementObjectType_CM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonMeasurementObjectType_CM_Rprt(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonMeasurementObjectType_CM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extension_CommonMeasurementObjectType_CM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Extension_CommonMeasurementObjectType_CM_Rprt(tvb, offset, &asn1_ctx, tree, hf_rnsap_Extension_CommonMeasurementObjectType_CM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GsmCellList_CM_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GsmCellList_CM_Rprt(tvb, offset, &asn1_ctx, tree, hf_rnsap_GsmCellList_CM_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementTerminationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonMeasurementTerminationRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonMeasurementTerminationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extension_TerminationMeasurementList_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Extension_TerminationMeasurementList(tvb, offset, &asn1_ctx, tree, hf_rnsap_Extension_TerminationMeasurementList_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_CommonMeasurementFailureIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_CommonMeasurementFailureIndication(tvb, offset, &asn1_ctx, tree, hf_rnsap_CommonMeasurementFailureIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Extension_FailureIndicationMeasurementList_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Extension_FailureIndicationMeasurementList(tvb, offset, &asn1_ctx, tree, hf_rnsap_Extension_FailureIndicationMeasurementList_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeInitiationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InformationExchangeInitiationRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_InformationExchangeInitiationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeObjectType_InfEx_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InformationExchangeObjectType_InfEx_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_InformationExchangeObjectType_InfEx_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GSM_Cell_InfEx_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GSM_Cell_InfEx_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_GSM_Cell_InfEx_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMS_Cell_InfEx_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBMS_Cell_InfEx_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBMS_Cell_InfEx_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ANR_Cell_InfEx_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ANR_Cell_InfEx_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_ANR_Cell_InfEx_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_E_RGCH_Cell_InfEx_Rqst_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Common_E_RGCH_Cell_InfEx_Rqst(tvb, offset, &asn1_ctx, tree, hf_rnsap_Common_E_RGCH_Cell_InfEx_Rqst_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeInitiationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InformationExchangeInitiationResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_InformationExchangeInitiationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeObjectType_InfEx_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InformationExchangeObjectType_InfEx_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_InformationExchangeObjectType_InfEx_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMS_Bearer_Service_List_InfEx_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBMS_Bearer_Service_List_InfEx_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBMS_Bearer_Service_List_InfEx_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMS_Cell_InfEx_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBMS_Cell_InfEx_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBMS_Cell_InfEx_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ANR_Cell_InfEx_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ANR_Cell_InfEx_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_ANR_Cell_InfEx_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_E_RGCH_Cell_InfEx_Rsp_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Common_E_RGCH_Cell_InfEx_Rsp(tvb, offset, &asn1_ctx, tree, hf_rnsap_Common_E_RGCH_Cell_InfEx_Rsp_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeInitiationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InformationExchangeInitiationFailure(tvb, offset, &asn1_ctx, tree, hf_rnsap_InformationExchangeInitiationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationReport_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InformationReport(tvb, offset, &asn1_ctx, tree, hf_rnsap_InformationReport_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeObjectType_InfEx_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InformationExchangeObjectType_InfEx_Rprt(tvb, offset, &asn1_ctx, tree, hf_rnsap_InformationExchangeObjectType_InfEx_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMS_Cell_InfEx_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBMS_Cell_InfEx_Rprt(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBMS_Cell_InfEx_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Common_E_RGCH_Cell_InfEx_Rprt_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Common_E_RGCH_Cell_InfEx_Rprt(tvb, offset, &asn1_ctx, tree, hf_rnsap_Common_E_RGCH_Cell_InfEx_Rprt_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeTerminationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InformationExchangeTerminationRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_InformationExchangeTerminationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationExchangeFailureIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InformationExchangeFailureIndication(tvb, offset, &asn1_ctx, tree, hf_rnsap_InformationExchangeFailureIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ResetRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ResetRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_ResetRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ResetIndicator_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ResetIndicator(tvb, offset, &asn1_ctx, tree, hf_rnsap_ResetIndicator_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ContextInfoItem_Reset_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ContextInfoItem_Reset(tvb, offset, &asn1_ctx, tree, hf_rnsap_ContextInfoItem_Reset_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ContextGroupInfoItem_Reset_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ContextGroupInfoItem_Reset(tvb, offset, &asn1_ctx, tree, hf_rnsap_ContextGroupInfoItem_Reset_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ResetResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ResetResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_ResetResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkActivationCommandFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkActivationCommandFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkActivationCommandFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DelayedActivationInformationList_RL_ActivationCmdFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DelayedActivationInformationList_RL_ActivationCmdFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DelayedActivationInformationList_RL_ActivationCmdFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DelayedActivationInformation_RL_ActivationCmdFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DelayedActivationInformation_RL_ActivationCmdFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DelayedActivationInformation_RL_ActivationCmdFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkActivationCommandTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkActivationCommandTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkActivationCommandTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DelayedActivationInformationList_RL_ActivationCmdTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DelayedActivationInformationList_RL_ActivationCmdTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DelayedActivationInformationList_RL_ActivationCmdTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DelayedActivationInformation_RL_ActivationCmdTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DelayedActivationInformation_RL_ActivationCmdTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_DelayedActivationInformation_RL_ActivationCmdTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_GERANUplinkSignallingTransferIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_GERANUplinkSignallingTransferIndication(tvb, offset, &asn1_ctx, tree, hf_rnsap_GERANUplinkSignallingTransferIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkParameterUpdateIndicationFDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkParameterUpdateIndicationFDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkParameterUpdateIndicationFDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_ParameterUpdateIndicationFDD_RL_InformationList_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_ParameterUpdateIndicationFDD_RL_InformationList(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_ParameterUpdateIndicationFDD_RL_InformationList_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RL_ParameterUpdateIndicationFDD_RL_Information_Item_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RL_ParameterUpdateIndicationFDD_RL_Information_Item(tvb, offset, &asn1_ctx, tree, hf_rnsap_RL_ParameterUpdateIndicationFDD_RL_Information_Item_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UE_Measurement_Forwarding_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UE_Measurement_Forwarding(tvb, offset, &asn1_ctx, tree, hf_rnsap_UE_Measurement_Forwarding_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_HS_Cell_Information_RL_Param_Upd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_HS_Cell_Information_RL_Param_Upd(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_HS_Cell_Information_RL_Param_Upd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Additional_EDCH_Cell_Information_RL_Param_Upd_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd(tvb, offset, &asn1_ctx, tree, hf_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RadioLinkParameterUpdateIndicationTDD_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RadioLinkParameterUpdateIndicationTDD(tvb, offset, &asn1_ctx, tree, hf_rnsap_RadioLinkParameterUpdateIndicationTDD_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementInitiationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementInitiationRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementInitiationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementInitiationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementInitiationResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementInitiationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementInitiationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementInitiationFailure(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementInitiationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementReport_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementReport(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementReport_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementTerminationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementTerminationRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementTerminationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_UEMeasurementFailureIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_UEMeasurementFailureIndication(tvb, offset, &asn1_ctx, tree, hf_rnsap_UEMeasurementFailureIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IurInvokeTrace_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_IurInvokeTrace(tvb, offset, &asn1_ctx, tree, hf_rnsap_IurInvokeTrace_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_ListOfInterfacesToTrace_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_ListOfInterfacesToTrace(tvb, offset, &asn1_ctx, tree, hf_rnsap_ListOfInterfacesToTrace_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InterfacesToTraceItem_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InterfacesToTraceItem(tvb, offset, &asn1_ctx, tree, hf_rnsap_InterfacesToTraceItem_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_IurDeactivateTrace_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_IurDeactivateTrace(tvb, offset, &asn1_ctx, tree, hf_rnsap_IurDeactivateTrace_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMSAttachCommand_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBMSAttachCommand(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBMSAttachCommand_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBMSDetachCommand_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBMSDetachCommand(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBMSDetachCommand_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_DirectInformationTransfer_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_DirectInformationTransfer(tvb, offset, &asn1_ctx, tree, hf_rnsap_DirectInformationTransfer_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EnhancedRelocationRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EnhancedRelocationRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_EnhancedRelocationRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EnhancedRelocationResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EnhancedRelocationResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_EnhancedRelocationResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EnhancedRelocationFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EnhancedRelocationFailure(tvb, offset, &asn1_ctx, tree, hf_rnsap_EnhancedRelocationFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EnhancedRelocationCancel_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EnhancedRelocationCancel(tvb, offset, &asn1_ctx, tree, hf_rnsap_EnhancedRelocationCancel_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EnhancedRelocationSignallingTransfer_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EnhancedRelocationSignallingTransfer(tvb, offset, &asn1_ctx, tree, hf_rnsap_EnhancedRelocationSignallingTransfer_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EnhancedRelocationRelease_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EnhancedRelocationRelease(tvb, offset, &asn1_ctx, tree, hf_rnsap_EnhancedRelocationRelease_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_MBSFNMCCHInformation_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_MBSFNMCCHInformation(tvb, offset, &asn1_ctx, tree, hf_rnsap_MBSFNMCCHInformation_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SecondaryULFrequencyReport_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SecondaryULFrequencyReport(tvb, offset, &asn1_ctx, tree, hf_rnsap_SecondaryULFrequencyReport_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_SecondaryULFrequencyUpdateIndication_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_SecondaryULFrequencyUpdateIndication(tvb, offset, &asn1_ctx, tree, hf_rnsap_SecondaryULFrequencyUpdateIndication_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EnhancedRelocationResourceRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EnhancedRelocationResourceRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_EnhancedRelocationResourceRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EnhancedRelocationResourceResponse_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EnhancedRelocationResourceResponse(tvb, offset, &asn1_ctx, tree, hf_rnsap_EnhancedRelocationResourceResponse_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EnhancedRelocationResourceFailure_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EnhancedRelocationResourceFailure(tvb, offset, &asn1_ctx, tree, hf_rnsap_EnhancedRelocationResourceFailure_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EnhancedRelocationResourceReleaseCommand_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EnhancedRelocationResourceReleaseCommand(tvb, offset, &asn1_ctx, tree, hf_rnsap_EnhancedRelocationResourceReleaseCommand_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_EnhancedRelocationResourceReleaseComplete_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_EnhancedRelocationResourceReleaseComplete(tvb, offset, &asn1_ctx, tree, hf_rnsap_EnhancedRelocationResourceReleaseComplete_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_PrivateMessage_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_PrivateMessage(tvb, offset, &asn1_ctx, tree, hf_rnsap_PrivateMessage_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_InformationTransferControlRequest_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_InformationTransferControlRequest(tvb, offset, &asn1_ctx, tree, hf_rnsap_InformationTransferControlRequest_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_Control_Type_InformationTransferControlReq_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_Control_Type_InformationTransferControlReq(tvb, offset, &asn1_ctx, tree, hf_rnsap_Control_Type_InformationTransferControlReq_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_RNSAP_PDU_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_RNSAP_PDU(tvb, offset, &asn1_ctx, tree, hf_rnsap_RNSAP_PDU_PDU);
  offset += 7; offset >>= 3;
  return offset;
}
static int dissect_NULL_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_, void *data _U_) {
  int offset = 0;
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, TRUE, pinfo);
  offset = dissect_rnsap_NULL(tvb, offset, &asn1_ctx, tree, hf_rnsap_NULL_PDU);
  offset += 7; offset >>= 3;
  return offset;
}


static int dissect_ProtocolIEFieldValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  rnsap_private_data_t *pdata = rnsap_get_private_data(pinfo);
  return (dissector_try_uint(rnsap_ies_dissector_table, pdata->ProtocolIE_ID, tvb, pinfo, tree)) ? tvb_captured_length(tvb) : 0;
}

static int dissect_ProtocolExtensionFieldExtensionValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  rnsap_private_data_t *pdata = rnsap_get_private_data(pinfo);
  return (dissector_try_uint(rnsap_extension_dissector_table, pdata->ProtocolIE_ID, tvb, pinfo, tree)) ? tvb_captured_length(tvb) : 0;
}

static int dissect_PrivateIEFieldValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  rnsap_private_data_t *pdata = rnsap_get_private_data(pinfo);
  return (call_ber_oid_callback(pdata->obj_id, tvb, 0, pinfo, tree, NULL)) ? tvb_captured_length(tvb) : 0;
}

static int dissect_InitiatingMessageValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  rnsap_private_data_t *pdata = rnsap_get_private_data(pinfo);
  if (!pdata->ProcedureID) return 0;
  return (dissector_try_string(rnsap_proc_imsg_dissector_table, pdata->ProcedureID, tvb, pinfo, tree, NULL)) ? tvb_captured_length(tvb) : 0;
}

static int dissect_SuccessfulOutcomeValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  rnsap_private_data_t *pdata = rnsap_get_private_data(pinfo);
  if (!pdata->ProcedureID) return 0;
  return (dissector_try_string(rnsap_proc_sout_dissector_table, pdata->ProcedureID, tvb, pinfo, tree, NULL)) ? tvb_captured_length(tvb) : 0;
}

static int dissect_UnsuccessfulOutcomeValue(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  rnsap_private_data_t *pdata = rnsap_get_private_data(pinfo);
  if (!pdata->ProcedureID) return 0;
  return (dissector_try_string(rnsap_proc_uout_dissector_table, pdata->ProcedureID, tvb, pinfo, tree, NULL)) ? tvb_captured_length(tvb) : 0;
}

static int
dissect_rnsap(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data)
{
	proto_item	*rnsap_item = NULL;
	proto_tree	*rnsap_tree = NULL;

	/* make entry in the Protocol column on summary display */
	col_set_str(pinfo->cinfo, COL_PROTOCOL, "RNSAP");

	/* create the rnsap protocol tree */
	rnsap_item = proto_tree_add_item(tree, proto_rnsap, tvb, 0, -1, ENC_NA);
	rnsap_tree = proto_item_add_subtree(rnsap_item, ett_rnsap);

	/* remove any rnsap_private_data_t state from previous PDUs in this packet. */
	p_remove_proto_data(pinfo->pool, pinfo, proto_rnsap, 0);

	return dissect_RNSAP_PDU_PDU(tvb, pinfo, rnsap_tree, data);
}

/* Highest ProcedureCode value, used in heuristics */
#define RNSAP_MAX_PC 61 /* id-enhancedRelocationResourceRelease = 61*/
#define RNSAP_MSG_MIN_LENGTH 7
static gboolean
dissect_sccp_rnsap_heur(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data)
{
  guint8 pdu_type;
  guint8 procedure_id;
  guint8 dd_mode;
  guint8 criticality;
  guint8 transaction_id_type;
  guint length;
  int length_field_offset;

  #define PDU_TYPE_OFFSET 0
  #define PROC_CODE_OFFSET 1
  #define DD_CRIT_OFFSET 2
  if (tvb_captured_length(tvb) < RNSAP_MSG_MIN_LENGTH) {
    return FALSE;
  }

  pdu_type = tvb_get_guint8(tvb, PDU_TYPE_OFFSET);
  if (pdu_type & 0x1F) {
    /* pdu_type is not 0x00 (initiatingMessage), 0x20 (succesfulOutcome),
       0x40 (unsuccesfulOutcome) or 0x60 (outcome), ignore extension bit (0x80) */
    return FALSE;
  }

  procedure_id = tvb_get_guint8(tvb, PROC_CODE_OFFSET);
  if (procedure_id > RNSAP_MAX_PC) {
      return FALSE;
  }

  dd_mode = tvb_get_guint8(tvb, DD_CRIT_OFFSET) >> 5;
  if (dd_mode >= 0x03) {
    /* dd_mode is not 0x00 (tdd), 0x01 (fdd) or 0x02 (common) */
    return FALSE;
  }

  criticality = (tvb_get_guint8(tvb, DD_CRIT_OFFSET) & 0x18) >> 3;
  if (criticality == 0x03) {
    /* criticality is not 0x00 (reject), 0x01 (ignore) or 0x02 (notify) */
    return FALSE;
  }

  /* Finding the offset for the length field - depends on wether the transaction id is long or short */
  transaction_id_type = (tvb_get_guint8(tvb, DD_CRIT_OFFSET) & 0x04) >> 2;
  if(transaction_id_type == 0x00) { /* Short transaction id - 1 byte*/
    length_field_offset = 4;
  }
  else { /* Long transaction id - 2 bytes*/
    length_field_offset = 5;
  }

  /* compute aligned PER length determinant without calling dissect_per_length_determinant()
     to avoid exceptions and info added to tree, info column and expert info */
  length = tvb_get_guint8(tvb, length_field_offset);
  length_field_offset += 1;
  if (length & 0x80) {
    if ((length & 0xc0) == 0x80) {
      length &= 0x3f;
      length <<= 8;
      length += tvb_get_guint8(tvb, length_field_offset);
      length_field_offset += 1;
    } else {
      length = 0;
    }
  }
  if (length!= (tvb_reported_length(tvb) - length_field_offset)){
    return FALSE;
  }

  dissect_rnsap(tvb, pinfo, tree, data);

  return TRUE;
}


/*--- proto_register_rnsap -------------------------------------------*/
void proto_register_rnsap(void) {

  /* List of fields */
  static hf_register_info hf[] = {
    { &hf_rnsap_transportLayerAddress_ipv4,
      { "transportLayerAddress IPv4", "rnsap.transportLayerAddress_ipv4",
      FT_IPv4, BASE_NONE, NULL, 0,
    NULL, HFILL }},
    { &hf_rnsap_transportLayerAddress_ipv6,
      { "transportLayerAddress IPv6", "rnsap.transportLayerAddress_ipv6",
      FT_IPv6, BASE_NONE, NULL, 0,
      NULL, HFILL }},
    { &hf_rnsap_transportLayerAddress_nsap,
      { "transportLayerAddress NSAP", "rnsap.transportLayerAddress_NSAP",
      FT_BYTES, BASE_NONE, NULL, 0,
      NULL, HFILL }},
    { &hf_rnsap_ActivationInformation_PDU,
      { "ActivationInformation", "rnsap.ActivationInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Active_MBMS_Bearer_Service_ListFDD_PDU,
      { "Active-MBMS-Bearer-Service-ListFDD", "rnsap.Active_MBMS_Bearer_Service_ListFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Active_MBMS_Bearer_Service_ListFDD_PFL_PDU,
      { "Active-MBMS-Bearer-Service-ListFDD-PFL", "rnsap.Active_MBMS_Bearer_Service_ListFDD_PFL",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Active_MBMS_Bearer_Service_ListTDD_PDU,
      { "Active-MBMS-Bearer-Service-ListTDD", "rnsap.Active_MBMS_Bearer_Service_ListTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Active_MBMS_Bearer_Service_ListTDD_PFL_PDU,
      { "Active-MBMS-Bearer-Service-ListTDD-PFL", "rnsap.Active_MBMS_Bearer_Service_ListTDD_PFL",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Active_Pattern_Sequence_Information_PDU,
      { "Active-Pattern-Sequence-Information", "rnsap.Active_Pattern_Sequence_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_Response_RLAddList_PDU,
      { "Additional-EDCH-Cell-Information-Response-RLAddList", "rnsap.Additional_EDCH_Cell_Information_Response_RLAddList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Setup_Info_PDU,
      { "Additional-EDCH-Setup-Info", "rnsap.Additional_EDCH_Setup_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ANRReportIndication_PDU,
      { "ANRReportIndication", "rnsap.ANRReportIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DTX_Information2_PDU,
      { "DTX-Information2", "rnsap.DTX_Information2_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Implicit_Grant_Handling_PDU,
      { "Implicit-Grant-Handling", "rnsap.Implicit_Grant_Handling",
        FT_UINT32, BASE_DEC, VALS(rnsap_Implicit_Grant_Handling_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Minimum_TEBS_threshold_PDU,
      { "Minimum-TEBS-threshold", "rnsap.Minimum_TEBS_threshold",
        FT_UINT32, BASE_DEC, VALS(rnsap_Minimum_TEBS_threshold_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_Response_List_PDU,
      { "Additional-EDCH-Cell-Information-Response-List", "rnsap.Additional_EDCH_Cell_Information_Response_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_Response_RLReconf_List_PDU,
      { "Additional-EDCH-Cell-Information-Response-RLReconf-List", "rnsap.Additional_EDCH_Cell_Information_Response_RLReconf_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_AdjustmentPeriod_PDU,
      { "AdjustmentPeriod", "rnsap.AdjustmentPeriod",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Allowed_Rate_Information_PDU,
      { "Allowed-Rate-Information", "rnsap.Allowed_Rate_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_AllowedQueuingTime_PDU,
      { "AllowedQueuingTime", "rnsap.AllowedQueuingTime",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_AlternativeFormatReportingIndicator_PDU,
      { "AlternativeFormatReportingIndicator", "rnsap.AlternativeFormatReportingIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_AlternativeFormatReportingIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Angle_Of_Arrival_Value_LCR_PDU,
      { "Angle-Of-Arrival-Value-LCR", "rnsap.Angle_Of_Arrival_Value_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ANR_Cell_Information_PDU,
      { "ANR-Cell-Information", "rnsap.ANR_Cell_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_AntennaColocationIndicator_PDU,
      { "AntennaColocationIndicator", "rnsap.AntennaColocationIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_AntennaColocationIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_BDS_IonosphericGridModelReq_PDU,
      { "BDS-IonosphericGridModelReq", "rnsap.BDS_IonosphericGridModelReq",
        FT_UINT32, BASE_DEC, VALS(rnsap_BDS_IonosphericGridModelReq_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_BDS_Ionospheric_Grid_Model_PDU,
      { "BDS-Ionospheric-Grid-Model", "rnsap.BDS_Ionospheric_Grid_Model_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_BindingID_PDU,
      { "BindingID", "rnsap.BindingID",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Cause_PDU,
      { "Cause", "rnsap.Cause",
        FT_UINT32, BASE_DEC, VALS(rnsap_Cause_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_CellCapabilityContainer_FDD_PDU,
      { "CellCapabilityContainer-FDD", "rnsap.CellCapabilityContainer_FDD",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CellCapabilityContainerExtension_FDD_PDU,
      { "CellCapabilityContainerExtension-FDD", "rnsap.CellCapabilityContainerExtension_FDD",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CellCapabilityContainer_TDD_PDU,
      { "CellCapabilityContainer-TDD", "rnsap.CellCapabilityContainer_TDD",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CellCapabilityContainer_TDD_LCR_PDU,
      { "CellCapabilityContainer-TDD-LCR", "rnsap.CellCapabilityContainer_TDD_LCR",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CellCapabilityContainerExtension_TDD_LCR_PDU,
      { "CellCapabilityContainerExtension-TDD-LCR", "rnsap.CellCapabilityContainerExtension_TDD_LCR",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CellCapabilityContainer_TDD768_PDU,
      { "CellCapabilityContainer-TDD768", "rnsap.CellCapabilityContainer_TDD768",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_C_ID_PDU,
      { "C-ID", "rnsap.C_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Cell_Capacity_Class_Value_PDU,
      { "Cell-Capacity-Class-Value", "rnsap.Cell_Capacity_Class_Value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CellListValidityIndicator_PDU,
      { "CellListValidityIndicator", "rnsap.CellListValidityIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_CellListValidityIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_CellPortionID_PDU,
      { "CellPortionID", "rnsap.CellPortionID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CellPortionLCRID_PDU,
      { "CellPortionLCRID", "rnsap.CellPortionLCRID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CFN_PDU,
      { "CFN", "rnsap.CFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ChipOffset_PDU,
      { "ChipOffset", "rnsap.ChipOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ClassmarkInformation2_PDU,
      { "ClassmarkInformation2", "rnsap.ClassmarkInformation2",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ClassmarkInformation3_PDU,
      { "ClassmarkInformation3", "rnsap.ClassmarkInformation3",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ClosedLoopMode1_SupportIndicator_PDU,
      { "ClosedLoopMode1-SupportIndicator", "rnsap.ClosedLoopMode1_SupportIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_ClosedLoopMode1_SupportIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonMeasurementAccuracy_PDU,
      { "CommonMeasurementAccuracy", "rnsap.CommonMeasurementAccuracy",
        FT_UINT32, BASE_DEC, VALS(rnsap_CommonMeasurementAccuracy_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonMeasurementType_PDU,
      { "CommonMeasurementType", "rnsap.CommonMeasurementType",
        FT_UINT32, BASE_DEC, VALS(rnsap_CommonMeasurementType_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_CongestionCause_PDU,
      { "CongestionCause", "rnsap.CongestionCause",
        FT_UINT32, BASE_DEC, VALS(rnsap_CongestionCause_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonTransportChannelResourcesInitialisationNotRequired_PDU,
      { "CommonTransportChannelResourcesInitialisationNotRequired", "rnsap.CommonTransportChannelResourcesInitialisationNotRequired",
        FT_UINT32, BASE_DEC, VALS(rnsap_CommonTransportChannelResourcesInitialisationNotRequired_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDD_PDU,
      { "Common-EDCH-MAC-d-Flow-Specific-InformationFDD", "rnsap.Common_EDCH_MAC_d_Flow_Specific_InformationFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationLCR_PDU,
      { "Common-EDCH-MAC-d-Flow-Specific-InformationLCR", "rnsap.Common_EDCH_MAC_d_Flow_Specific_InformationLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Common_EDCH_Support_Indicator_PDU,
      { "Common-EDCH-Support-Indicator", "rnsap.Common_EDCH_Support_Indicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Common_E_RGCH_Cell_Information_PDU,
      { "Common-E-RGCH-Cell-Information", "rnsap.Common_E_RGCH_Cell_Information",
        FT_UINT32, BASE_DEC, VALS(rnsap_Common_E_RGCH_Cell_Information_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_CompleteAlmanacProvided_PDU,
      { "CompleteAlmanacProvided", "rnsap.CompleteAlmanacProvided",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information_PDU,
      { "Continuous-Packet-Connectivity-DTX-DRX-Information", "rnsap.Continuous_Packet_Connectivity_DTX_DRX_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_PDU,
      { "Continuous-Packet-Connectivity-HS-SCCH-Less-Information", "rnsap.Continuous_Packet_Connectivity_HS_SCCH_Less_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response_PDU,
      { "Continuous-Packet-Connectivity-HS-SCCH-Less-Information-Response", "rnsap.Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CoverageIndicator_PDU,
      { "CoverageIndicator", "rnsap.CoverageIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_CoverageIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_CPC_Information_PDU,
      { "CPC-Information", "rnsap.CPC_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CPC_RecoveryReport_PDU,
      { "CPC-RecoveryReport", "rnsap.CPC_RecoveryReport",
        FT_UINT32, BASE_DEC, VALS(rnsap_CPC_RecoveryReport_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Continuous_Packet_Connectivity_HS_SCCH_less_Deactivate_Indicator_PDU,
      { "Continuous-Packet-Connectivity-HS-SCCH-less-Deactivate-Indicator", "rnsap.Continuous_Packet_Connectivity_HS_SCCH_less_Deactivate_Indicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Counting_Information_PDU,
      { "Counting-Information", "rnsap.Counting_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CriticalityDiagnostics_PDU,
      { "CriticalityDiagnostics", "rnsap.CriticalityDiagnostics_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MessageStructure_PDU,
      { "MessageStructure", "rnsap.MessageStructure",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CN_CS_DomainIdentifier_PDU,
      { "CN-CS-DomainIdentifier", "rnsap.CN_CS_DomainIdentifier_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CN_PS_DomainIdentifier_PDU,
      { "CN-PS-DomainIdentifier", "rnsap.CN_PS_DomainIdentifier_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CQI_Cycle_Switch_Timer_PDU,
      { "CQI-Cycle-Switch-Timer", "rnsap.CQI_Cycle_Switch_Timer",
        FT_UINT32, BASE_DEC, VALS(rnsap_CQI_Cycle_Switch_Timer_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_ControlGAP_PDU,
      { "ControlGAP", "rnsap.ControlGAP",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CQI_Feedback_Cycle2_PDU,
      { "CQI-Feedback-Cycle2", "rnsap.CQI_Feedback_Cycle2",
        FT_UINT32, BASE_DEC, VALS(rnsap_CQI_Feedback_Cycle2_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_C_RNTI_PDU,
      { "C-RNTI", "rnsap.C_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CPC_InformationLCR_PDU,
      { "CPC-InformationLCR", "rnsap.CPC_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ContinuousPacketConnectivity_DRX_InformationLCR_PDU,
      { "ContinuousPacketConnectivity-DRX-InformationLCR", "rnsap.ContinuousPacketConnectivity_DRX_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext_PDU,
      { "Inactivity-Threshold-for-UE-DRX-Cycle-LCR-Ext", "rnsap.Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext",
        FT_UINT32, BASE_DEC, VALS(rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_E_AGCH_UE_Inactivity_Monitor_Threshold_PDU,
      { "E-AGCH-UE-Inactivity-Monitor-Threshold", "rnsap.E_AGCH_UE_Inactivity_Monitor_Threshold",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_AGCH_UE_Inactivity_Monitor_Threshold_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR_PDU,
      { "ContinuousPacketConnectivity-DRX-Information-ResponseLCR", "rnsap.ContinuousPacketConnectivity_DRX_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DBDS_CorrectionsReq_PDU,
      { "DBDS-CorrectionsReq", "rnsap.DBDS_CorrectionsReq_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DBDS_Corrections_PDU,
      { "DBDS-Corrections", "rnsap.DBDS_Corrections_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_ENH_Information_PDU,
      { "DCH-ENH-Information", "rnsap.DCH_ENH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_ENH_Information_Reconf_PDU,
      { "DCH-ENH-Information-Reconf", "rnsap.DCH_ENH_Information_Reconf_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_FDD_Information_PDU,
      { "DCH-FDD-Information", "rnsap.DCH_FDD_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_MeasurementOccasion_Information_PDU,
      { "DCH-MeasurementOccasion-Information", "rnsap.DCH_MeasurementOccasion_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_MeasurementType_Indicator_PDU,
      { "DCH-MeasurementType-Indicator", "rnsap.DCH_MeasurementType_Indicator",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_Indicator_For_E_DCH_HSDPA_Operation_PDU,
      { "DCH-Indicator-For-E-DCH-HSDPA-Operation", "rnsap.DCH_Indicator_For_E_DCH_HSDPA_Operation",
        FT_UINT32, BASE_DEC, VALS(rnsap_DCH_Indicator_For_E_DCH_HSDPA_Operation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_InformationResponse_PDU,
      { "DCH-InformationResponse", "rnsap.DCH_InformationResponse",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_TDD_Information_PDU,
      { "DCH-TDD-Information", "rnsap.DCH_TDD_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DedicatedMeasurementType_PDU,
      { "DedicatedMeasurementType", "rnsap.DedicatedMeasurementType",
        FT_UINT32, BASE_DEC, VALS(rnsap_DedicatedMeasurementType_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DelayedActivation_PDU,
      { "DelayedActivation", "rnsap.DelayedActivation",
        FT_UINT32, BASE_DEC, VALS(rnsap_DelayedActivation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DGANSS_Corrections_Req_PDU,
      { "DGANSS-Corrections-Req", "rnsap.DGANSS_Corrections_Req_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DGNSS_ValidityPeriod_PDU,
      { "DGNSS-ValidityPeriod", "rnsap.DGNSS_ValidityPeriod_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DiversityMode_PDU,
      { "DiversityMode", "rnsap.DiversityMode",
        FT_UINT32, BASE_DEC, VALS(rnsap_DiversityMode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_TBS_PDU,
      { "DL-TBS", "rnsap.DL_TBS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_TimingAdjustment_PDU,
      { "DL-DPCH-TimingAdjustment", "rnsap.DL_DPCH_TimingAdjustment",
        FT_UINT32, BASE_DEC, VALS(rnsap_DL_DPCH_TimingAdjustment_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_Power_PDU,
      { "DL-Power", "rnsap.DL_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_PowerBalancing_Information_PDU,
      { "DL-PowerBalancing-Information", "rnsap.DL_PowerBalancing_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_PowerBalancing_ActivationIndicator_PDU,
      { "DL-PowerBalancing-ActivationIndicator", "rnsap.DL_PowerBalancing_ActivationIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_DL_PowerBalancing_ActivationIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_PowerBalancing_UpdatedIndicator_PDU,
      { "DL-PowerBalancing-UpdatedIndicator", "rnsap.DL_PowerBalancing_UpdatedIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_DL_PowerBalancing_UpdatedIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_ReferencePowerInformation_PDU,
      { "DL-ReferencePowerInformation", "rnsap.DL_ReferencePowerInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_D_RNTI_PDU,
      { "D-RNTI", "rnsap.D_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_D_RNTI_ReleaseIndication_PDU,
      { "D-RNTI-ReleaseIndication", "rnsap.D_RNTI_ReleaseIndication",
        FT_UINT32, BASE_DEC, VALS(rnsap_D_RNTI_ReleaseIndication_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_TimeSlot_ISCP_Info_PDU,
      { "DL-TimeSlot-ISCP-Info", "rnsap.DL_TimeSlot_ISCP_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_TimeSlot_ISCP_LCR_Information_PDU,
      { "DL-TimeSlot-ISCP-LCR-Information", "rnsap.DL_TimeSlot_ISCP_LCR_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Downlink_TPC_enhancements_Information_PDU,
      { "Downlink-TPC-enhancements-Information", "rnsap.Downlink_TPC_enhancements_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Downlink_TPC_enhancements_Reconf_PDU,
      { "Downlink-TPC-enhancements-Reconf", "rnsap.Downlink_TPC_enhancements_Reconf_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DPC_Mode_PDU,
      { "DPC-Mode", "rnsap.DPC_Mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_DPC_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DPC_Mode_Change_SupportIndicator_PDU,
      { "DPC-Mode-Change-SupportIndicator", "rnsap.DPC_Mode_Change_SupportIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_DPC_Mode_Change_SupportIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DPCH_ID768_PDU,
      { "DPCH-ID768", "rnsap.DPCH_ID768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DRXCycleLengthCoefficient_PDU,
      { "DRXCycleLengthCoefficient", "rnsap.DRXCycleLengthCoefficient",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_RNTI_PDU,
      { "DSCH-RNTI", "rnsap.DSCH_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_InitialWindowSize_PDU,
      { "DSCH-InitialWindowSize", "rnsap.DSCH_InitialWindowSize",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_TDD_Information_PDU,
      { "DSCH-TDD-Information", "rnsap.DSCH_TDD_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EARFCN_Extended_PDU,
      { "EARFCN-Extended", "rnsap.EARFCN_Extended",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EARFCN_FDD_Extended_PDU,
      { "EARFCN-FDD-Extended", "rnsap.EARFCN_FDD_Extended_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_AGCH_Table_Choice_PDU,
      { "E-AGCH-Table-Choice", "rnsap.E_AGCH_Table_Choice",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_AGCH_Table_Choice_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_Decoupling_Indication_PDU,
      { "E-DCH-Decoupling-Indication", "rnsap.E_DCH_Decoupling_Indication",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_DCH_Decoupling_Indication_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_FDD_DL_ControlChannelInformation_PDU,
      { "EDCH-FDD-DL-ControlChannelInformation", "rnsap.EDCH_FDD_DL_ControlChannelInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_RGCH_E_HICH_ChannelisationCodeValidityIndicator_PDU,
      { "E-RGCH-E-HICH-ChannelisationCodeValidityIndicator", "rnsap.E_RGCH_E_HICH_ChannelisationCodeValidityIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_RGCH_E_HICH_ChannelisationCodeValidityIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_FDD_Information_PDU,
      { "EDCH-FDD-Information", "rnsap.EDCH_FDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UPH_Filtering_Measurement_Forwarding_Request_PDU,
      { "UPH-Filtering-Measurement-Forwarding-Request", "rnsap.UPH_Filtering_Measurement_Forwarding_Request",
        FT_UINT32, BASE_DEC, VALS(rnsap_UPH_Filtering_Measurement_Forwarding_Request_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_FDD_InformationResponse_PDU,
      { "EDCH-FDD-InformationResponse", "rnsap.EDCH_FDD_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_FDD_Information_To_Modify_PDU,
      { "EDCH-FDD-Information-To-Modify", "rnsap.EDCH_FDD_Information_To_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_FDD_Update_Information_PDU,
      { "E-DCH-FDD-Update-Information", "rnsap.E_DCH_FDD_Update_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_DL_Control_Channel_Change_Information_PDU,
      { "E-DCH-DL-Control-Channel-Change-Information", "rnsap.E_DCH_DL_Control_Channel_Change_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_DL_Control_Channel_Grant_Information_PDU,
      { "E-DCH-DL-Control-Channel-Grant-Information", "rnsap.E_DCH_DL_Control_Channel_Grant_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_PowerOffset_for_SchedulingInfo_PDU,
      { "E-DCH-PowerOffset-for-SchedulingInfo", "rnsap.E_DCH_PowerOffset_for_SchedulingInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_MACdPDUSizeFormat_PDU,
      { "E-DCH-MACdPDUSizeFormat", "rnsap.E_DCH_MACdPDUSizeFormat",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_DCH_MACdPDUSizeFormat_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_MACdFlows_Information_PDU,
      { "EDCH-MACdFlows-Information", "rnsap.EDCH_MACdFlows_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_MACdFlows_To_Delete_PDU,
      { "EDCH-MACdFlows-To-Delete", "rnsap.EDCH_MACdFlows_To_Delete",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_RL_Indication_PDU,
      { "EDCH-RL-Indication", "rnsap.EDCH_RL_Indication",
        FT_UINT32, BASE_DEC, VALS(rnsap_EDCH_RL_Indication_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_Serving_cell_change_informationResponse_PDU,
      { "E-DCH-Serving-cell-change-informationResponse", "rnsap.E_DCH_Serving_cell_change_informationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDPCH_Information_FDD_PDU,
      { "EDPCH-Information-FDD", "rnsap.EDPCH_Information_FDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDPCH_Information_RLReconfRequest_FDD_PDU,
      { "EDPCH-Information-RLReconfRequest-FDD", "rnsap.EDPCH_Information_RLReconfRequest_FDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DPDCH_PowerInterpolation_PDU,
      { "E-DPDCH-PowerInterpolation", "rnsap.E_DPDCH_PowerInterpolation",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_Serving_Grant_Value_PDU,
      { "E-Serving-Grant-Value", "rnsap.E_Serving_Grant_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_Serving_RL_PDU,
      { "EDCH-Serving-RL", "rnsap.EDCH_Serving_RL",
        FT_UINT32, BASE_DEC, VALS(rnsap_EDCH_Serving_RL_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Enhanced_FACH_Information_ResponseFDD_PDU,
      { "Enhanced-FACH-Information-ResponseFDD", "rnsap.Enhanced_FACH_Information_ResponseFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Enhanced_FACH_Information_ResponseLCR_PDU,
      { "Enhanced-FACH-Information-ResponseLCR", "rnsap.Enhanced_FACH_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Enhanced_FACH_Support_Indicator_PDU,
      { "Enhanced-FACH-Support-Indicator", "rnsap.Enhanced_FACH_Support_Indicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EnhancedHSServingCC_Abort_PDU,
      { "EnhancedHSServingCC-Abort", "rnsap.EnhancedHSServingCC_Abort",
        FT_UINT32, BASE_DEC, VALS(rnsap_EnhancedHSServingCC_Abort_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Enhanced_PCH_Capability_PDU,
      { "Enhanced-PCH-Capability", "rnsap.Enhanced_PCH_Capability",
        FT_UINT32, BASE_DEC, VALS(rnsap_Enhanced_PCH_Capability_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_E_RNTI_PDU,
      { "E-RNTI", "rnsap.E_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_TFCI_Boost_Information_PDU,
      { "E-TFCI-Boost-Information", "rnsap.E_TFCI_Boost_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_Minimum_Set_E_TFCIValidityIndicator_PDU,
      { "E-DCH-Minimum-Set-E-TFCIValidityIndicator", "rnsap.E_DCH_Minimum_Set_E_TFCIValidityIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_DCH_Minimum_Set_E_TFCIValidityIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Enhanced_PrimaryCPICH_EcNo_PDU,
      { "Enhanced-PrimaryCPICH-EcNo", "rnsap.Enhanced_PrimaryCPICH_EcNo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EventH_PDU,
      { "EventH", "rnsap.EventH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ExtendedGSMCellIndividualOffset_PDU,
      { "ExtendedGSMCellIndividualOffset", "rnsap.ExtendedGSMCellIndividualOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_Information_PDU,
      { "E-DCH-Information", "rnsap.E_DCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_MACdFlow_Retransmission_Timer_LCR_PDU,
      { "E-DCH-MACdFlow-Retransmission-Timer-LCR", "rnsap.E_DCH_MACdFlow_Retransmission_Timer_LCR",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_E_DCH_MACdFlow_Retransmission_Timer_LCR_vals_ext, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_Information_Reconfig_PDU,
      { "E-DCH-Information-Reconfig", "rnsap.E_DCH_Information_Reconfig_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_Information_Response_PDU,
      { "E-DCH-Information-Response", "rnsap.E_DCH_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ActivationDelay_PDU,
      { "ActivationDelay", "rnsap.ActivationDelay",
        FT_UINT32, BASE_DEC, VALS(rnsap_ActivationDelay_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Fast_TTI_switching_Mode_synchronized_PDU,
      { "Fast-TTI-switching-Mode-synchronized", "rnsap.Fast_TTI_switching_Mode_synchronized",
        FT_UINT32, BASE_DEC, VALS(rnsap_Fast_TTI_switching_Mode_synchronized_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Fast_TTI_switching_Mode_unsynchronized_PDU,
      { "Fast-TTI-switching-Mode-unsynchronized", "rnsap.Fast_TTI_switching_Mode_unsynchronized",
        FT_UINT32, BASE_DEC, VALS(rnsap_Fast_TTI_switching_Mode_unsynchronized_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Fast_TTI_switching_Mode_Supported_PDU,
      { "Fast-TTI-switching-Mode-Supported", "rnsap.Fast_TTI_switching_Mode_Supported",
        FT_UINT32, BASE_DEC, VALS(rnsap_Fast_TTI_switching_Mode_Supported_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_768_Information_PDU,
      { "E-DCH-768-Information", "rnsap.E_DCH_768_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_768_Information_Reconfig_PDU,
      { "E-DCH-768-Information-Reconfig", "rnsap.E_DCH_768_Information_Reconfig_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_768_Information_Response_PDU,
      { "E-DCH-768-Information-Response", "rnsap.E_DCH_768_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_LCR_Information_PDU,
      { "E-DCH-LCR-Information", "rnsap.E_DCH_LCR_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Extended_E_DCH_LCRTDD_PhysicalLayerCategory_PDU,
      { "Extended-E-DCH-LCRTDD-PhysicalLayerCategory", "rnsap.Extended_E_DCH_LCRTDD_PhysicalLayerCategory",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multi_Carrier_E_DCH_LCRTDD_PhysicalLayerCategory_PDU,
      { "Multi-Carrier-E-DCH-LCRTDD-PhysicalLayerCategory", "rnsap.Multi_Carrier_E_DCH_LCRTDD_PhysicalLayerCategory",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_LCR_Information_Reconfig_PDU,
      { "E-DCH-LCR-Information-Reconfig", "rnsap.E_DCH_LCR_Information_Reconfig_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_LCR_Information_Response_PDU,
      { "E-DCH-LCR-Information-Response", "rnsap.E_DCH_LCR_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Ext_Reference_E_TFCI_PO_PDU,
      { "Ext-Reference-E-TFCI-PO", "rnsap.Ext_Reference_E_TFCI_PO",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ExtendedPropagationDelay_PDU,
      { "ExtendedPropagationDelay", "rnsap.ExtendedPropagationDelay",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Radio_Links_without_DPCH_FDPCH_Indication_PDU,
      { "Radio-Links-without-DPCH-FDPCH-Indication", "rnsap.Radio_Links_without_DPCH_FDPCH_Indication",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Extended_RNC_ID_PDU,
      { "Extended-RNC-ID", "rnsap.Extended_RNC_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Extended_RNTI_PDU,
      { "Extended-RNTI", "rnsap.Extended_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Extended_Round_Trip_Time_Value_PDU,
      { "Extended-Round-Trip-Time-Value", "rnsap.Extended_Round_Trip_Time_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Extended_S_RNTI_Group_PDU,
      { "Extended-S-RNTI-Group", "rnsap.Extended_S_RNTI_Group_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ExtendedAffectedUEInformationForMBMS_PDU,
      { "ExtendedAffectedUEInformationForMBMS", "rnsap.ExtendedAffectedUEInformationForMBMS",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Ext_Max_Bits_MACe_PDU_non_scheduled_PDU,
      { "Ext-Max-Bits-MACe-PDU-non-scheduled", "rnsap.Ext_Max_Bits_MACe_PDU_non_scheduled",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_Semi_PersistentScheduling_Information_LCR_PDU,
      { "E-DCH-Semi-PersistentScheduling-Information-LCR", "rnsap.E_DCH_Semi_PersistentScheduling_Information_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU,
      { "E-DCH-Semi-PersistentScheduling-Information-ResponseLCR", "rnsap.E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_FACH_FlowControlInformation_PDU,
      { "FACH-FlowControlInformation", "rnsap.FACH_FlowControlInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Fast_Reconfiguration_Mode_PDU,
      { "Fast-Reconfiguration-Mode", "rnsap.Fast_Reconfiguration_Mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_Fast_Reconfiguration_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Fast_Reconfiguration_Permission_PDU,
      { "Fast-Reconfiguration-Permission", "rnsap.Fast_Reconfiguration_Permission",
        FT_UINT32, BASE_DEC, VALS(rnsap_Fast_Reconfiguration_Permission_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_FDD_DCHs_to_Modify_PDU,
      { "FDD-DCHs-to-Modify", "rnsap.FDD_DCHs_to_Modify",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_FDD_DL_CodeInformation_PDU,
      { "FDD-DL-CodeInformation", "rnsap.FDD_DL_CodeInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_F_DPCH_SlotFormat_PDU,
      { "F-DPCH-SlotFormat", "rnsap.F_DPCH_SlotFormat",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_F_DPCH_SlotFormatSupportRequest_PDU,
      { "F-DPCH-SlotFormatSupportRequest", "rnsap.F_DPCH_SlotFormatSupportRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_FNReportingIndicator_PDU,
      { "FNReportingIndicator", "rnsap.FNReportingIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_FNReportingIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_FTPICH_Information_PDU,
      { "FTPICH-Information", "rnsap.FTPICH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_FTPICH_Information_Reconf_PDU,
      { "FTPICH-Information-Reconf", "rnsap.FTPICH_Information_Reconf_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_FTPICH_Reconfiguration_Information_PDU,
      { "FTPICH-Reconfiguration-Information", "rnsap.FTPICH_Reconfiguration_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_FTPICH_Information_Response_PDU,
      { "FTPICH-Information-Response", "rnsap.FTPICH_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_FrameOffset_PDU,
      { "FrameOffset", "rnsap.FrameOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_FrequencyBandIndicator_PDU,
      { "FrequencyBandIndicator", "rnsap.FrequencyBandIndicator",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_FrequencyBandIndicator_vals_ext, 0,
        NULL, HFILL }},
    { &hf_rnsap_GA_Cell_PDU,
      { "GA-Cell", "rnsap.GA_Cell",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GA_CellAdditionalShapes_PDU,
      { "GA-CellAdditionalShapes", "rnsap.GA_CellAdditionalShapes",
        FT_UINT32, BASE_DEC, VALS(rnsap_GA_CellAdditionalShapes_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Gainfactors_10ms_mode_PDU,
      { "Gainfactors-10ms-mode", "rnsap.Gainfactors_10ms_mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_Gainfactors_10ms_mode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_AddIonoModelReq_PDU,
      { "GANSS-AddIonoModelReq", "rnsap.GANSS_AddIonoModelReq",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_AddNavigationModelsReq_PDU,
      { "GANSS-AddNavigationModelsReq", "rnsap.GANSS_AddNavigationModelsReq",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_AddUTCModelsReq_PDU,
      { "GANSS-AddUTCModelsReq", "rnsap.GANSS_AddUTCModelsReq",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_Additional_Ionospheric_Model_PDU,
      { "GANSS-Additional-Ionospheric-Model", "rnsap.GANSS_Additional_Ionospheric_Model_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_Additional_Navigation_Models_PDU,
      { "GANSS-Additional-Navigation-Models", "rnsap.GANSS_Additional_Navigation_Models_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_Additional_Time_Models_PDU,
      { "GANSS-Additional-Time-Models", "rnsap.GANSS_Additional_Time_Models",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_Additional_UTC_Models_PDU,
      { "GANSS-Additional-UTC-Models", "rnsap.GANSS_Additional_UTC_Models",
        FT_UINT32, BASE_DEC, VALS(rnsap_GANSS_Additional_UTC_Models_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_alm_keplerianNAVAlmanac_PDU,
      { "GANSS-alm-keplerianNAVAlmanac", "rnsap.GANSS_alm_keplerianNAVAlmanac_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_alm_KEPLERIANBDSALMANAC_PDU,
      { "GANSS-alm-KEPLERIANBDSALMANAC", "rnsap.GANSS_alm_KEPLERIANBDSALMANAC_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_alm_keplerianReducedAlmanac_PDU,
      { "GANSS-alm-keplerianReducedAlmanac", "rnsap.GANSS_alm_keplerianReducedAlmanac_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_alm_keplerianMidiAlmanac_PDU,
      { "GANSS-alm-keplerianMidiAlmanac", "rnsap.GANSS_alm_keplerianMidiAlmanac_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_alm_keplerianGLONASS_PDU,
      { "GANSS-alm-keplerianGLONASS", "rnsap.GANSS_alm_keplerianGLONASS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_alm_ecefSBASAlmanac_PDU,
      { "GANSS-alm-ecefSBASAlmanac", "rnsap.GANSS_alm_ecefSBASAlmanac_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_Auxiliary_Information_PDU,
      { "GANSS-Auxiliary-Information", "rnsap.GANSS_Auxiliary_Information",
        FT_UINT32, BASE_DEC, VALS(rnsap_GANSS_Auxiliary_Information_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_AuxInfoReq_PDU,
      { "GANSS-AuxInfoReq", "rnsap.GANSS_AuxInfoReq",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_BDSclockModel_PDU,
      { "GANSS-BDSclockModel", "rnsap.GANSS_BDSclockModel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_Common_Data_PDU,
      { "GANSS-Common-Data", "rnsap.GANSS_Common_Data_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_Delta_T_PDU,
      { "GANSS-Delta-T", "rnsap.GANSS_Delta_T",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_Earth_Orientation_Parameters_PDU,
      { "GANSS-Earth-Orientation-Parameters", "rnsap.GANSS_Earth_Orientation_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_EarthOrientParaReq_PDU,
      { "GANSS-EarthOrientParaReq", "rnsap.GANSS_EarthOrientParaReq",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_Generic_Data_PDU,
      { "GANSS-Generic-Data", "rnsap.GANSS_Generic_Data",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_ID_PDU,
      { "GANSS-ID", "rnsap.GANSS_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_Information_PDU,
      { "GANSS-Information", "rnsap.GANSS_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_NavModel_BDSKeplerianSet_PDU,
      { "GANSS-NavModel-BDSKeplerianSet", "rnsap.GANSS_NavModel_BDSKeplerianSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_SBAS_ID_PDU,
      { "GANSS-SBAS-ID", "rnsap.GANSS_SBAS_ID",
        FT_UINT32, BASE_DEC, VALS(rnsap_GANSS_SBAS_ID_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_Time_ID_PDU,
      { "GANSS-Time-ID", "rnsap.GANSS_Time_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GANSS_UTCmodelSet4_PDU,
      { "GANSS-UTCmodelSet4", "rnsap.GANSS_UTCmodelSet4_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GERAN_Cell_Capability_PDU,
      { "GERAN-Cell-Capability", "rnsap.GERAN_Cell_Capability",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GERAN_Classmark_PDU,
      { "GERAN-Classmark", "rnsap.GERAN_Classmark",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Satellite_Almanac_Information_ExtItem_PDU,
      { "Satellite-Almanac-Information-ExtItem", "rnsap.Satellite_Almanac_Information_ExtItem",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Guaranteed_Rate_Information_PDU,
      { "Guaranteed-Rate-Information", "rnsap.Guaranteed_Rate_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HARQ_MemoryPartitioningInfoExtForMIMO_PDU,
      { "HARQ-MemoryPartitioningInfoExtForMIMO", "rnsap.HARQ_MemoryPartitioningInfoExtForMIMO",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HARQ_Preamble_Mode_PDU,
      { "HARQ-Preamble-Mode", "rnsap.HARQ_Preamble_Mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_HARQ_Preamble_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_HARQ_Preamble_Mode_Activation_Indicator_PDU,
      { "HARQ-Preamble-Mode-Activation-Indicator", "rnsap.HARQ_Preamble_Mode_Activation_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_HARQ_Preamble_Mode_Activation_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_HCS_Prio_PDU,
      { "HCS-Prio", "rnsap.HCS_Prio",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_Configured_Indicator_PDU,
      { "HSDSCH-Configured-Indicator", "rnsap.HSDSCH_Configured_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_HSDSCH_Configured_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_FDD_Information_PDU,
      { "HSDSCH-FDD-Information", "rnsap.HSDSCH_FDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_FDD_Information_Response_PDU,
      { "HSDSCH-FDD-Information-Response", "rnsap.HSDSCH_FDD_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_Information_to_Modify_PDU,
      { "HSDSCH-Information-to-Modify", "rnsap.HSDSCH_Information_to_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_Information_to_Modify_Unsynchronised_PDU,
      { "HSDSCH-Information-to-Modify-Unsynchronised", "rnsap.HSDSCH_Information_to_Modify_Unsynchronised_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_MACdPDUSizeFormat_PDU,
      { "HSDSCH-MACdPDUSizeFormat", "rnsap.HSDSCH_MACdPDUSizeFormat",
        FT_UINT32, BASE_DEC, VALS(rnsap_HSDSCH_MACdPDUSizeFormat_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_MACdFlows_Information_PDU,
      { "HSDSCH-MACdFlows-Information", "rnsap.HSDSCH_MACdFlows_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_MACdFlows_to_Delete_PDU,
      { "HSDSCH-MACdFlows-to-Delete", "rnsap.HSDSCH_MACdFlows_to_Delete",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_PreconfigurationInfo_PDU,
      { "HSDSCH-PreconfigurationInfo", "rnsap.HSDSCH_PreconfigurationInfo_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Preconfiguration_Information_PDU,
      { "Additional-EDCH-Preconfiguration-Information", "rnsap.Additional_EDCH_Preconfiguration_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_PreconfigurationSetup_PDU,
      { "HSDSCH-PreconfigurationSetup", "rnsap.HSDSCH_PreconfigurationSetup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_RNTI_PDU,
      { "HSDSCH-RNTI", "rnsap.HSDSCH_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HS_DSCH_serving_cell_change_information_PDU,
      { "HS-DSCH-serving-cell-change-information", "rnsap.HS_DSCH_serving_cell_change_information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HS_DSCH_serving_cell_change_informationResponse_PDU,
      { "HS-DSCH-serving-cell-change-informationResponse", "rnsap.HS_DSCH_serving_cell_change_informationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_TBSizeTableIndicator_PDU,
      { "HSDSCH-TBSizeTableIndicator", "rnsap.HSDSCH_TBSizeTableIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_HSDSCH_TBSizeTableIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_TDD_Information_PDU,
      { "HSDSCH-TDD-Information", "rnsap.HSDSCH_TDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_TDD_Information_Response_PDU,
      { "HSDSCH-TDD-Information-Response", "rnsap.HSDSCH_TDD_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multicarrier_Number_PDU,
      { "Multicarrier-Number", "rnsap.Multicarrier_Number",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSPDSCH_TDD_Specific_InfoList_Response768_PDU,
      { "HSPDSCH-TDD-Specific-InfoList-Response768", "rnsap.HSPDSCH_TDD_Specific_InfoList_Response768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSSCCH_TDD_Specific_InfoList_Response768_PDU,
      { "HSSCCH-TDD-Specific-InfoList-Response768", "rnsap.HSSCCH_TDD_Specific_InfoList_Response768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HS_SICH_Reception_Quality_Value_PDU,
      { "HS-SICH-Reception-Quality-Value", "rnsap.HS_SICH_Reception_Quality_Value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HS_SICH_Reception_Quality_Measurement_Value_PDU,
      { "HS-SICH-Reception-Quality-Measurement-Value", "rnsap.HS_SICH_Reception_Quality_Measurement_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HS_SICH_ID_PDU,
      { "HS-SICH-ID", "rnsap.HS_SICH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HS_SICH_ID_Extension_PDU,
      { "HS-SICH-ID-Extension", "rnsap.HS_SICH_ID_Extension",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HS_PDSCH_Code_Change_Indicator_PDU,
      { "HS-PDSCH-Code-Change-Indicator", "rnsap.HS_PDSCH_Code_Change_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_HS_PDSCH_Code_Change_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_HS_PDSCH_Code_Change_Grant_PDU,
      { "HS-PDSCH-Code-Change-Grant", "rnsap.HS_PDSCH_Code_Change_Grant",
        FT_UINT32, BASE_DEC, VALS(rnsap_HS_PDSCH_Code_Change_Grant_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_FDD_Update_Information_PDU,
      { "HSDSCH-FDD-Update-Information", "rnsap.HSDSCH_FDD_Update_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_TDD_Update_Information_PDU,
      { "HSDSCH-TDD-Update-Information", "rnsap.HSDSCH_TDD_Update_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MIMO_ReferenceSignal_InformationListLCR_PDU,
      { "MIMO-ReferenceSignal-InformationListLCR", "rnsap.MIMO_ReferenceSignal_InformationListLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_LCR_PDU,
      { "HS-DSCH-Semi-PersistentScheduling-Information-LCR", "rnsap.HS_DSCH_Semi_PersistentScheduling_Information_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_Physical_Layer_Category_PDU,
      { "HSDSCH-Physical-Layer-Category", "rnsap.HSDSCH_Physical_Layer_Category",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TS0_HS_PDSCH_Indication_LCR_PDU,
      { "TS0-HS-PDSCH-Indication-LCR", "rnsap.TS0_HS_PDSCH_Indication_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU,
      { "HS-DSCH-Semi-PersistentScheduling-Information-ResponseLCR", "rnsap.HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_IMSI_PDU,
      { "IMSI", "rnsap.IMSI",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Improved_Synchronized_Indicator_PDU,
      { "Improved-Synchronized-Indicator", "rnsap.Improved_Synchronized_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_Improved_Synchronized_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_PDU,
      { "Inactivity-Threshold-for-UE-DRX-Cycle", "rnsap.Inactivity_Threshold_for_UE_DRX_Cycle",
        FT_UINT32, BASE_DEC, VALS(rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_InformationExchangeID_PDU,
      { "InformationExchangeID", "rnsap.InformationExchangeID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_InformationReportCharacteristics_PDU,
      { "InformationReportCharacteristics", "rnsap.InformationReportCharacteristics",
        FT_UINT32, BASE_DEC, VALS(rnsap_InformationReportCharacteristics_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_InformationType_PDU,
      { "InformationType", "rnsap.InformationType_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Initial_DL_DPCH_TimingAdjustment_Allowed_PDU,
      { "Initial-DL-DPCH-TimingAdjustment-Allowed", "rnsap.Initial_DL_DPCH_TimingAdjustment_Allowed",
        FT_UINT32, BASE_DEC, VALS(rnsap_Initial_DL_DPCH_TimingAdjustment_Allowed_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_InnerLoopDLPCStatus_PDU,
      { "InnerLoopDLPCStatus", "rnsap.InnerLoopDLPCStatus",
        FT_UINT32, BASE_DEC, VALS(rnsap_InnerLoopDLPCStatus_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Inter_Frequency_Cell_List_PDU,
      { "Inter-Frequency-Cell-List", "rnsap.Inter_Frequency_Cell_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Inter_Frequency_Cell_Information_PDU,
      { "Inter-Frequency-Cell-Information", "rnsap.Inter_Frequency_Cell_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Inter_Frequency_Cell_Information_SIB11ter_PDU,
      { "Inter-Frequency-Cell-Information-SIB11ter", "rnsap.Inter_Frequency_Cell_Information_SIB11ter",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_IPDL_TDD_ParametersLCR_PDU,
      { "IPDL-TDD-ParametersLCR", "rnsap.IPDL_TDD_ParametersLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_IdleIntervalInformation_PDU,
      { "IdleIntervalInformation", "rnsap.IdleIntervalInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_L3_Information_PDU,
      { "L3-Information", "rnsap.L3_Information",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Load_Value_IncrDecrThres_PDU,
      { "Load-Value-IncrDecrThres", "rnsap.Load_Value_IncrDecrThres",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Load_Value_PDU,
      { "Load-Value", "rnsap.Load_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_LoadValue_PDU,
      { "LoadValue", "rnsap.LoadValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_LCRTDD_Uplink_Physical_Channel_Capability_PDU,
      { "LCRTDD-Uplink-Physical-Channel-Capability", "rnsap.LCRTDD_Uplink_Physical_Channel_Capability_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MDT_Configuration_PDU,
      { "MDT-Configuration", "rnsap.MDT_Configuration_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MAC_PDU_SizeExtended_PDU,
      { "MAC-PDU-SizeExtended", "rnsap.MAC_PDU_SizeExtended",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MACes_Maximum_Bitrate_LCR_PDU,
      { "MACes-Maximum-Bitrate-LCR", "rnsap.MACes_Maximum_Bitrate_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MaxNrDLPhysicalchannels768_PDU,
      { "MaxNrDLPhysicalchannels768", "rnsap.MaxNrDLPhysicalchannels768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MaxNrDLPhysicalchannelsTS_PDU,
      { "MaxNrDLPhysicalchannelsTS", "rnsap.MaxNrDLPhysicalchannelsTS",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MaxNrDLPhysicalchannelsTS768_PDU,
      { "MaxNrDLPhysicalchannelsTS768", "rnsap.MaxNrDLPhysicalchannelsTS768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MaxNr_Retransmissions_EDCH_PDU,
      { "MaxNr-Retransmissions-EDCH", "rnsap.MaxNr_Retransmissions_EDCH",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Max_UE_DTX_Cycle_PDU,
      { "Max-UE-DTX-Cycle", "rnsap.Max_UE_DTX_Cycle",
        FT_UINT32, BASE_DEC, VALS(rnsap_Max_UE_DTX_Cycle_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_Full_Address_PDU,
      { "MBMS-Bearer-Service-Full-Address", "rnsap.MBMS_Bearer_Service_Full_Address_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_List_PDU,
      { "MBMS-Bearer-Service-List", "rnsap.MBMS_Bearer_Service_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Neighbouring_Cell_Information_PDU,
      { "MBMS-Neighbouring-Cell-Information", "rnsap.MBMS_Neighbouring_Cell_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_RLC_Sequence_Number_Information_PDU,
      { "MBMS-RLC-Sequence-Number-Information", "rnsap.MBMS_RLC_Sequence_Number_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBSFN_Cluster_Identity_PDU,
      { "MBSFN-Cluster-Identity", "rnsap.MBSFN_Cluster_Identity",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MCCH_Message_List_PDU,
      { "MCCH-Message-List", "rnsap.MCCH_Message_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MCCH_Configuration_Info_PDU,
      { "MCCH-Configuration-Info", "rnsap.MCCH_Configuration_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Info_List_PDU,
      { "MBSFN-Scheduling-Transmission-Time-Interval-Info-List", "rnsap.MBSFN_Scheduling_Transmission_Time_Interval_Info_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MeasurementFilterCoefficient_PDU,
      { "MeasurementFilterCoefficient", "rnsap.MeasurementFilterCoefficient",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_MeasurementFilterCoefficient_vals_ext, 0,
        NULL, HFILL }},
    { &hf_rnsap_MeasurementID_PDU,
      { "MeasurementID", "rnsap.MeasurementID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Measurement_Power_Offset_PDU,
      { "Measurement-Power-Offset", "rnsap.Measurement_Power_Offset",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MinimumSpreadingFactor768_PDU,
      { "MinimumSpreadingFactor768", "rnsap.MinimumSpreadingFactor768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MaxAdjustmentStep_PDU,
      { "MaxAdjustmentStep", "rnsap.MaxAdjustmentStep",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MeasurementRecoveryBehavior_PDU,
      { "MeasurementRecoveryBehavior", "rnsap.MeasurementRecoveryBehavior_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MeasurementRecoveryReportingIndicator_PDU,
      { "MeasurementRecoveryReportingIndicator", "rnsap.MeasurementRecoveryReportingIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MeasurementRecoverySupportIndicator_PDU,
      { "MeasurementRecoverySupportIndicator", "rnsap.MeasurementRecoverySupportIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MIMO_ActivationIndicator_PDU,
      { "MIMO-ActivationIndicator", "rnsap.MIMO_ActivationIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MIMO_InformationResponse_PDU,
      { "MIMO-InformationResponse", "rnsap.MIMO_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MIMO_Mode_Indicator_PDU,
      { "MIMO-Mode-Indicator", "rnsap.MIMO_Mode_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_MIMO_Mode_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_MIMO_N_M_Ratio_PDU,
      { "MIMO-N-M-Ratio", "rnsap.MIMO_N_M_Ratio",
        FT_UINT32, BASE_DEC, VALS(rnsap_MIMO_N_M_Ratio_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_MIMO_withfourtransmitantennas_ActivationIndicator_PDU,
      { "MIMO-withfourtransmitantennas-ActivationIndicator", "rnsap.MIMO_withfourtransmitantennas_ActivationIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MIMO_withfourtransmitantennas_ModeIndicator_PDU,
      { "MIMO-withfourtransmitantennas-ModeIndicator", "rnsap.MIMO_withfourtransmitantennas_ModeIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_MIMO_withfourtransmitantennas_ModeIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DualStream_MIMO_withfourtransmitantennas_ActivationIndicator_PDU,
      { "DualStream-MIMO-withfourtransmitantennas-ActivationIndicator", "rnsap.DualStream_MIMO_withfourtransmitantennas_ActivationIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DualStream_MIMO_withfourtransmitantennas_ModeIndicator_PDU,
      { "DualStream-MIMO-withfourtransmitantennas-ModeIndicator", "rnsap.DualStream_MIMO_withfourtransmitantennas_ModeIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_DualStream_MIMO_withfourtransmitantennas_ModeIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_Associated_Secondary_CPICH_PDU,
      { "Additional-Associated-Secondary-CPICH", "rnsap.Additional_Associated_Secondary_CPICH",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MinimumReducedE_DPDCH_GainFactor_PDU,
      { "MinimumReducedE-DPDCH-GainFactor", "rnsap.MinimumReducedE_DPDCH_GainFactor",
        FT_UINT32, BASE_DEC, VALS(rnsap_MinimumReducedE_DPDCH_GainFactor_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_MulticellEDCH_InformationItemIEs_PDU,
      { "MulticellEDCH-InformationItemIEs", "rnsap.MulticellEDCH_InformationItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MulticellEDCH_RL_SpecificInformationItemIEs_PDU,
      { "MulticellEDCH-RL-SpecificInformationItemIEs", "rnsap.MulticellEDCH_RL_SpecificInformationItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multicell_EDCH_Restriction_PDU,
      { "Multicell-EDCH-Restriction", "rnsap.Multicell_EDCH_Restriction",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_PLMN_List_PDU,
      { "Multiple-PLMN-List", "rnsap.Multiple_PLMN_List_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MAChs_ResetIndicator_PDU,
      { "MAChs-ResetIndicator", "rnsap.MAChs_ResetIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_MAChs_ResetIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_PDU,
      { "MultipleFreq-HSPDSCH-InformationList-ResponseTDDLCR", "rnsap.MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MIMO_SFMode_For_HSPDSCHDualStream_PDU,
      { "MIMO-SFMode-For-HSPDSCHDualStream", "rnsap.MIMO_SFMode_For_HSPDSCHDualStream",
        FT_UINT32, BASE_DEC, VALS(rnsap_MIMO_SFMode_For_HSPDSCHDualStream_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Multi_Carrier_EDCH_Info_PDU,
      { "Multi-Carrier-EDCH-Info", "rnsap.Multi_Carrier_EDCH_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SNPL_Carrier_Group_Indicator_PDU,
      { "SNPL-Carrier-Group-Indicator", "rnsap.SNPL_Carrier_Group_Indicator",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multi_Carrier_EDCH_Reconfigure_PDU,
      { "Multi-Carrier-EDCH-Reconfigure", "rnsap.Multi_Carrier_EDCH_Reconfigure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multi_Carrier_EDCH_Information_Response_PDU,
      { "Multi-Carrier-EDCH-Information-Response", "rnsap.Multi_Carrier_EDCH_Information_Response",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiflow_Reconfiguration_PDU,
      { "Multiflow-Reconfiguration", "rnsap.Multiflow_Reconfiguration",
        FT_UINT32, BASE_DEC, VALS(rnsap_Multiflow_Reconfiguration_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiflow_Information_PDU,
      { "Multiflow-Information", "rnsap.Multiflow_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiflow_RepetitionFactors_PDU,
      { "Multiflow-RepetitionFactors", "rnsap.Multiflow_RepetitionFactors_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiflow_OrdinalNumberOfFrequency_PDU,
      { "Multiflow-OrdinalNumberOfFrequency", "rnsap.Multiflow_OrdinalNumberOfFrequency",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MU_MIMO_InformationLCR_PDU,
      { "MU-MIMO-InformationLCR", "rnsap.MU_MIMO_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MU_MIMO_Indicator_PDU,
      { "MU-MIMO-Indicator", "rnsap.MU_MIMO_Indicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_NACC_Related_Data_PDU,
      { "NACC-Related-Data", "rnsap.NACC_Related_Data_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Neighbouring_UMTS_CellInformation_Ext_PDU,
      { "Neighbouring-UMTS-CellInformation-Ext", "rnsap.Neighbouring_UMTS_CellInformation_Ext",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Neighbouring_UMTS_CellInformationExtensionItem_PDU,
      { "Neighbouring-UMTS-CellInformationExtensionItem", "rnsap.Neighbouring_UMTS_CellInformationExtensionItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Neighbouring_UMTS_CellInformationItem_PDU,
      { "Neighbouring-UMTS-CellInformationItem", "rnsap.Neighbouring_UMTS_CellInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MultipleFrequencyBandsIndicatorList_PDU,
      { "MultipleFrequencyBandsIndicatorList", "rnsap.MultipleFrequencyBandsIndicatorList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Neighbouring_GSM_CellInformationIEs_PDU,
      { "Neighbouring-GSM-CellInformationIEs", "rnsap.Neighbouring_GSM_CellInformationIEs",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_NeighbouringTDDCellMeasurementInformationLCR_PDU,
      { "NeighbouringTDDCellMeasurementInformationLCR", "rnsap.NeighbouringTDDCellMeasurementInformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_NeighbouringTDDCellMeasurementInformation768_PDU,
      { "NeighbouringTDDCellMeasurementInformation768", "rnsap.NeighbouringTDDCellMeasurementInformation768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Neighbouring_LCR_TDD_CellInformation_PDU,
      { "Neighbouring-LCR-TDD-CellInformation", "rnsap.Neighbouring_LCR_TDD_CellInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Neighbouring_E_UTRA_CellInformation_PDU,
      { "Neighbouring-E-UTRA-CellInformation", "rnsap.Neighbouring_E_UTRA_CellInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MultipleFrequencyBandsIndicatorList_EUTRA_PDU,
      { "MultipleFrequencyBandsIndicatorList-EUTRA", "rnsap.MultipleFrequencyBandsIndicatorList_EUTRA",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_NonCellSpecificTxDiversity_PDU,
      { "NonCellSpecificTxDiversity", "rnsap.NonCellSpecificTxDiversity",
        FT_UINT32, BASE_DEC, VALS(rnsap_NonCellSpecificTxDiversity_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Non_rectangular_resource_allocation_indicator_PDU,
      { "Non-rectangular-resource-allocation-indicator", "rnsap.Non_rectangular_resource_allocation_indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_Non_rectangular_resource_allocation_indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Non_rectangular_resource_timeslot_set_PDU,
      { "Non-rectangular-resource-timeslot-set", "rnsap.Non_rectangular_resource_timeslot_set",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_NRT_Load_Information_Value_IncrDecrThres_PDU,
      { "NRT-Load-Information-Value-IncrDecrThres", "rnsap.NRT_Load_Information_Value_IncrDecrThres",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_NRT_Load_Information_Value_PDU,
      { "NRT-Load-Information-Value", "rnsap.NRT_Load_Information_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_NRTLoadInformationValue_PDU,
      { "NRTLoadInformationValue", "rnsap.NRTLoadInformationValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Number_Of_Supported_Carriers_PDU,
      { "Number-Of-Supported-Carriers", "rnsap.Number_Of_Supported_Carriers",
        FT_UINT32, BASE_DEC, VALS(rnsap_Number_Of_Supported_Carriers_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_NoOfTargetCellHS_SCCH_Order_PDU,
      { "NoOfTargetCellHS-SCCH-Order", "rnsap.NoOfTargetCellHS_SCCH_Order",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Non_Serving_RL_Preconfig_Setup_PDU,
      { "Non-Serving-RL-Preconfig-Setup", "rnsap.Non_Serving_RL_Preconfig_Setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup_PDU,
      { "Additional-E-DCH-Non-Serving-RL-Preconfiguration-Setup", "rnsap.Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_PDU,
      { "Additional-E-DCH-New-non-serving-RL-E-DCH-FDD-DL-Control-Channel-InfoList", "rnsap.Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_NeedforIdleInterval_PDU,
      { "NeedforIdleInterval", "rnsap.NeedforIdleInterval",
        FT_UINT32, BASE_DEC, VALS(rnsap_NeedforIdleInterval_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_OnModification_PDU,
      { "OnModification", "rnsap.OnModification_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_OrdinalNumberOfFrequency_PDU,
      { "OrdinalNumberOfFrequency", "rnsap.OrdinalNumberOfFrequency",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Out_of_Sychronization_Window_PDU,
      { "Out-of-Sychronization-Window", "rnsap.Out_of_Sychronization_Window",
        FT_UINT32, BASE_DEC, VALS(rnsap_Out_of_Sychronization_Window_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_PartialReportingIndicator_PDU,
      { "PartialReportingIndicator", "rnsap.PartialReportingIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_PartialReportingIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_PCI_PDU,
      { "PCI", "rnsap.PCI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Permanent_NAS_UE_Identity_PDU,
      { "Permanent-NAS-UE-Identity", "rnsap.Permanent_NAS_UE_Identity",
        FT_UINT32, BASE_DEC, VALS(rnsap_Permanent_NAS_UE_Identity_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Phase_Reference_Update_Indicator_PDU,
      { "Phase-Reference-Update-Indicator", "rnsap.Phase_Reference_Update_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_Phase_Reference_Update_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_PLMN_Identity_PDU,
      { "PLMN-Identity", "rnsap.PLMN_Identity",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PLMN_List_PDU,
      { "PLMN-List", "rnsap.PLMN_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PowerAdjustmentType_PDU,
      { "PowerAdjustmentType", "rnsap.PowerAdjustmentType",
        FT_UINT32, BASE_DEC, VALS(rnsap_PowerAdjustmentType_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_PowerOffsetForSecondaryCPICHforMIMO_PDU,
      { "PowerOffsetForSecondaryCPICHforMIMO", "rnsap.PowerOffsetForSecondaryCPICHforMIMO",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PowerOffsetForSecondaryCPICHforMIMORequestIndicator_PDU,
      { "PowerOffsetForSecondaryCPICHforMIMORequestIndicator", "rnsap.PowerOffsetForSecondaryCPICHforMIMORequestIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas_PDU,
      { "PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas", "rnsap.PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennasRequestIndicator_PDU,
      { "PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennasRequestIndicator", "rnsap.PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennasRequestIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Precoder_Weight_Set_Restriction_PDU,
      { "Precoder-Weight-Set-Restriction", "rnsap.Precoder_Weight_Set_Restriction",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Primary_CPICH_Usage_For_Channel_Estimation_PDU,
      { "Primary-CPICH-Usage-For-Channel-Estimation", "rnsap.Primary_CPICH_Usage_For_Channel_Estimation",
        FT_UINT32, BASE_DEC, VALS(rnsap_Primary_CPICH_Usage_For_Channel_Estimation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_PrimaryCCPCH_RSCP_PDU,
      { "PrimaryCCPCH-RSCP", "rnsap.PrimaryCCPCH_RSCP",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PrimaryCCPCH_RSCP_Delta_PDU,
      { "PrimaryCCPCH-RSCP-Delta", "rnsap.PrimaryCCPCH_RSCP_Delta",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PropagationDelay_PDU,
      { "PropagationDelay", "rnsap.PropagationDelay",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ProvidedInformation_PDU,
      { "ProvidedInformation", "rnsap.ProvidedInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UpPCH_InformationList_LCRTDD_PDU,
      { "UpPCH-InformationList-LCRTDD", "rnsap.UpPCH_InformationList_LCRTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UpPCH_InformationItem_LCRTDD_PDU,
      { "UpPCH-InformationItem-LCRTDD", "rnsap.UpPCH_InformationItem_LCRTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RANAP_EnhancedRelocationInformationRequest_PDU,
      { "RANAP-EnhancedRelocationInformationRequest", "rnsap.RANAP_EnhancedRelocationInformationRequest",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RANAP_EnhancedRelocationInformationResponse_PDU,
      { "RANAP-EnhancedRelocationInformationResponse", "rnsap.RANAP_EnhancedRelocationInformationResponse",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RANAP_RelocationInformation_PDU,
      { "RANAP-RelocationInformation", "rnsap.RANAP_RelocationInformation",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Received_Total_Wideband_Power_Value_PDU,
      { "Received-Total-Wideband-Power-Value", "rnsap.Received_Total_Wideband_Power_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Received_Total_Wideband_Power_Value_IncrDecrThres_PDU,
      { "Received-Total-Wideband-Power-Value-IncrDecrThres", "rnsap.Received_Total_Wideband_Power_Value_IncrDecrThres",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Released_CN_Domain_PDU,
      { "Released-CN-Domain", "rnsap.Released_CN_Domain",
        FT_UINT32, BASE_DEC, VALS(rnsap_Released_CN_Domain_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_ReportCharacteristics_PDU,
      { "ReportCharacteristics", "rnsap.ReportCharacteristics",
        FT_UINT32, BASE_DEC, VALS(rnsap_ReportCharacteristics_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RestrictionStateIndicator_PDU,
      { "RestrictionStateIndicator", "rnsap.RestrictionStateIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_RestrictionStateIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_ID_PDU,
      { "RL-ID", "rnsap.RL_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_ID_PDU,
      { "RL-Set-ID", "rnsap.RL_Set_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Specific_DCH_Info_PDU,
      { "RL-Specific-DCH-Info", "rnsap.RL_Specific_DCH_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Specific_EDCH_Information_PDU,
      { "RL-Specific-EDCH-Information", "rnsap.RL_Specific_EDCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_RLC_PDU_Size_Format_PDU,
      { "DL-RLC-PDU-Size-Format", "rnsap.DL_RLC_PDU_Size_Format",
        FT_UINT32, BASE_DEC, VALS(rnsap_DL_RLC_PDU_Size_Format_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RLC_Sequence_Number_PDU,
      { "RLC-Sequence-Number", "rnsap.RLC_Sequence_Number",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RNC_ID_PDU,
      { "RNC-ID", "rnsap.RNC_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RNTI_Allocation_Indicator_PDU,
      { "RNTI-Allocation-Indicator", "rnsap.RNTI_Allocation_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_RNTI_Allocation_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RT_Load_Value_IncrDecrThres_PDU,
      { "RT-Load-Value-IncrDecrThres", "rnsap.RT_Load_Value_IncrDecrThres",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RT_Load_Value_PDU,
      { "RT-Load-Value", "rnsap.RT_Load_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RTLoadValue_PDU,
      { "RTLoadValue", "rnsap.RTLoadValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RxTimingDeviationForTA_PDU,
      { "RxTimingDeviationForTA", "rnsap.RxTimingDeviationForTA",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RxTimingDeviationForTAext_PDU,
      { "RxTimingDeviationForTAext", "rnsap.RxTimingDeviationForTAext",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RxTimingDeviationForTA768_PDU,
      { "RxTimingDeviationForTA768", "rnsap.RxTimingDeviationForTA768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Rx_Timing_Deviation_Value_ext_PDU,
      { "Rx-Timing-Deviation-Value-ext", "rnsap.Rx_Timing_Deviation_Value_ext",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Rx_Timing_Deviation_Value_LCR_PDU,
      { "Rx-Timing-Deviation-Value-LCR", "rnsap.Rx_Timing_Deviation_Value_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Rx_Timing_Deviation_Value_768_PDU,
      { "Rx-Timing-Deviation-Value-768", "rnsap.Rx_Timing_Deviation_Value_768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SAI_PDU,
      { "SAI", "rnsap.SAI_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ScaledAdjustmentRatio_PDU,
      { "ScaledAdjustmentRatio", "rnsap.ScaledAdjustmentRatio",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Secondary_CPICH_Information_PDU,
      { "Secondary-CPICH-Information", "rnsap.Secondary_CPICH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Secondary_CPICH_Information_Change_PDU,
      { "Secondary-CPICH-Information-Change", "rnsap.Secondary_CPICH_Information_Change",
        FT_UINT32, BASE_DEC, VALS(rnsap_Secondary_CPICH_Information_Change_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Secondary_LCR_CCPCH_Info_TDD_PDU,
      { "Secondary-LCR-CCPCH-Info-TDD", "rnsap.Secondary_LCR_CCPCH_Info_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Secondary_CCPCH_Info_TDD768_PDU,
      { "Secondary-CCPCH-Info-TDD768", "rnsap.Secondary_CCPCH_Info_TDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Secondary_Serving_Cell_List_PDU,
      { "Secondary-Serving-Cell-List", "rnsap.Secondary_Serving_Cell_List_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SFN_PDU,
      { "SFN", "rnsap.SFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SFNSFNMeasurementThresholdInformation_PDU,
      { "SFNSFNMeasurementThresholdInformation", "rnsap.SFNSFNMeasurementThresholdInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Single_Stream_MIMO_ActivationIndicator_PDU,
      { "Single-Stream-MIMO-ActivationIndicator", "rnsap.Single_Stream_MIMO_ActivationIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Single_Stream_MIMO_Mode_Indicator_PDU,
      { "Single-Stream-MIMO-Mode-Indicator", "rnsap.Single_Stream_MIMO_Mode_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_Single_Stream_MIMO_Mode_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_SixteenQAM_UL_Operation_Indicator_PDU,
      { "SixteenQAM-UL-Operation-Indicator", "rnsap.SixteenQAM_UL_Operation_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SixteenQAM_UL_Operation_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_SixtyfourQAM_UL_Operation_Indicator_PDU,
      { "SixtyfourQAM-UL-Operation-Indicator", "rnsap.SixtyfourQAM_UL_Operation_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SixtyfourQAM_UL_Operation_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_SixtyfourQAM_UsageAllowedIndicator_PDU,
      { "SixtyfourQAM-UsageAllowedIndicator", "rnsap.SixtyfourQAM_UsageAllowedIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SixtyfourQAM_UsageAllowedIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_SixtyfourQAM_DL_SupportIndicator_PDU,
      { "SixtyfourQAM-DL-SupportIndicator", "rnsap.SixtyfourQAM_DL_SupportIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SixtyfourQAM_DL_SupportIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_SixtyfourQAM_DL_UsageIndicator_PDU,
      { "SixtyfourQAM-DL-UsageIndicator", "rnsap.SixtyfourQAM_DL_UsageIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SixtyfourQAM_DL_UsageIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_SNA_Information_PDU,
      { "SNA-Information", "rnsap.SNA_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SourceID_PDU,
      { "SourceID", "rnsap.SourceID_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SpeechVersion_PDU,
      { "SpeechVersion", "rnsap.SpeechVersion",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_S_RNTI_PDU,
      { "S-RNTI", "rnsap.S_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_STTD_SupportIndicator_PDU,
      { "STTD-SupportIndicator", "rnsap.STTD_SupportIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_STTD_SupportIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Support_8PSK_PDU,
      { "Support-8PSK", "rnsap.Support_8PSK",
        FT_UINT32, BASE_DEC, VALS(rnsap_Support_8PSK_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Support_PLCCH_PDU,
      { "Support-PLCCH", "rnsap.Support_PLCCH",
        FT_UINT32, BASE_DEC, VALS(rnsap_Support_PLCCH_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order_PDU,
      { "Support-of-Dynamic-DTXDRX-Related-HS-SCCH-Order", "rnsap.Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order",
        FT_UINT32, BASE_DEC, VALS(rnsap_Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TAC_PDU,
      { "TAC", "rnsap.TAC",
        FT_UINT16, BASE_DEC_HEX, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TargetID_PDU,
      { "TargetID", "rnsap.TargetID_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TDD_DCHs_to_Modify_PDU,
      { "TDD-DCHs-to-Modify", "rnsap.TDD_DCHs_to_Modify",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR_PDU,
      { "TDD-DL-DPCH-TimeSlotFormat-LCR", "rnsap.TDD_DL_DPCH_TimeSlotFormat_LCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TDD_TPC_DownlinkStepSize_PDU,
      { "TDD-TPC-DownlinkStepSize", "rnsap.TDD_TPC_DownlinkStepSize",
        FT_UINT32, BASE_DEC, VALS(rnsap_TDD_TPC_DownlinkStepSize_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TDD_TPC_UplinkStepSize_LCR_PDU,
      { "TDD-TPC-UplinkStepSize-LCR", "rnsap.TDD_TPC_UplinkStepSize_LCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_TDD_TPC_UplinkStepSize_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TDD_UL_DPCH_TimeSlotFormat_LCR_PDU,
      { "TDD-UL-DPCH-TimeSlotFormat-LCR", "rnsap.TDD_UL_DPCH_TimeSlotFormat_LCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_TDD_UL_DPCH_TimeSlotFormat_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TimeSlot_PDU,
      { "TimeSlot", "rnsap.TimeSlot",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SynchronisationIndicator_PDU,
      { "SynchronisationIndicator", "rnsap.SynchronisationIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SynchronisationIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TnlQos_PDU,
      { "TnlQos", "rnsap.TnlQos",
        FT_UINT32, BASE_DEC, VALS(rnsap_TnlQos_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TraceDepth_PDU,
      { "TraceDepth", "rnsap.TraceDepth",
        FT_UINT32, BASE_DEC, VALS(rnsap_TraceDepth_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TraceRecordingSessionReference_PDU,
      { "TraceRecordingSessionReference", "rnsap.TraceRecordingSessionReference",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TraceReference_PDU,
      { "TraceReference", "rnsap.TraceReference",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TrafficClass_PDU,
      { "TrafficClass", "rnsap.TrafficClass",
        FT_UINT32, BASE_DEC, VALS(rnsap_TrafficClass_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Transmission_Gap_Pattern_Sequence_Information_PDU,
      { "Transmission-Gap-Pattern-Sequence-Information", "rnsap.Transmission_Gap_Pattern_Sequence_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Affected_HSDSCH_Serving_Cell_List_PDU,
      { "Affected-HSDSCH-Serving-Cell-List", "rnsap.Affected_HSDSCH_Serving_Cell_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Transmission_Mode_Information_PDU,
      { "Transmission-Mode-Information", "rnsap.Transmission_Mode_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TransmitDiversityIndicator_PDU,
      { "TransmitDiversityIndicator", "rnsap.TransmitDiversityIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransmitDiversityIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Transmitted_Carrier_Power_Value_PDU,
      { "Transmitted-Carrier-Power-Value", "rnsap.Transmitted_Carrier_Power_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Transmitted_Carrier_Power_Value_IncrDecrThres_PDU,
      { "Transmitted-Carrier-Power-Value-IncrDecrThres", "rnsap.Transmitted_Carrier_Power_Value_IncrDecrThres",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TUTRANGANSSMeasurementThresholdInformation_PDU,
      { "TUTRANGANSSMeasurementThresholdInformation", "rnsap.TUTRANGANSSMeasurementThresholdInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TUTRANGANSSMeasurementValueInformation_PDU,
      { "TUTRANGANSSMeasurementValueInformation", "rnsap.TUTRANGANSSMeasurementValueInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TUTRANGPSMeasurementThresholdInformation_PDU,
      { "TUTRANGPSMeasurementThresholdInformation", "rnsap.TUTRANGPSMeasurementThresholdInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TransportBearerID_PDU,
      { "TransportBearerID", "rnsap.TransportBearerID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TransportBearerRequestIndicator_PDU,
      { "TransportBearerRequestIndicator", "rnsap.TransportBearerRequestIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransportBearerRequestIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TransportBearerNotRequestedIndicator_PDU,
      { "TransportBearerNotRequestedIndicator", "rnsap.TransportBearerNotRequestedIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransportBearerNotRequestedIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TransportBearerNotSetupIndicator_PDU,
      { "TransportBearerNotSetupIndicator", "rnsap.TransportBearerNotSetupIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransportBearerNotSetupIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TPC_slot_position_PDU,
      { "TPC-slot-position", "rnsap.TPC_slot_position_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TransportLayerAddress_PDU,
      { "TransportLayerAddress", "rnsap.TransportLayerAddress",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TrCH_SrcStatisticsDescr_PDU,
      { "TrCH-SrcStatisticsDescr", "rnsap.TrCH_SrcStatisticsDescr",
        FT_UINT32, BASE_DEC, VALS(rnsap_TrCH_SrcStatisticsDescr_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TSN_Length_PDU,
      { "TSN-Length", "rnsap.TSN_Length",
        FT_UINT32, BASE_DEC, VALS(rnsap_TSN_Length_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TSTD_Support_Indicator_PDU,
      { "TSTD-Support-Indicator", "rnsap.TSTD_Support_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_TSTD_Support_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TTI_Update_Indicator_PDU,
      { "TTI-Update-Indicator", "rnsap.TTI_Update_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_TTI_Update_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TypeOfError_PDU,
      { "TypeOfError", "rnsap.TypeOfError",
        FT_UINT32, BASE_DEC, VALS(rnsap_TypeOfError_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UARFCN_PDU,
      { "UARFCN", "rnsap.UARFCN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UE_AggregateMaximumBitRate_PDU,
      { "UE-AggregateMaximumBitRate", "rnsap.UE_AggregateMaximumBitRate_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UE_AggregateMaximumBitRate_Enforcement_Indicator_PDU,
      { "UE-AggregateMaximumBitRate-Enforcement-Indicator", "rnsap.UE_AggregateMaximumBitRate_Enforcement_Indicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UE_Capabilities_Info_PDU,
      { "UE-Capabilities-Info", "rnsap.UE_Capabilities_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UE_Support_of_non_rectangular_resource_allocation_PDU,
      { "UE-Support-of-non-rectangular-resource-allocation", "rnsap.UE_Support_of_non_rectangular_resource_allocation",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_Support_of_non_rectangular_resource_allocation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UE_TS0_CapabilityLCR_PDU,
      { "UE-TS0-CapabilityLCR", "rnsap.UE_TS0_CapabilityLCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_TS0_CapabilityLCR_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UE_RF_Band_CapabilityLCR_PDU,
      { "UE-RF-Band-CapabilityLCR", "rnsap.UE_RF_Band_CapabilityLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_LCRTDD_HSDSCH_Physical_Layer_Category_PDU,
      { "LCRTDD-HSDSCH-Physical-Layer-Category", "rnsap.LCRTDD_HSDSCH_Physical_Layer_Category",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UE_DRX_Cycle_PDU,
      { "UE-DRX-Cycle", "rnsap.UE_DRX_Cycle",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_DRX_Cycle_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UEIdentity_PDU,
      { "UEIdentity", "rnsap.UEIdentity",
        FT_UINT32, BASE_DEC, VALS(rnsap_UEIdentity_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementParameterModAllow_PDU,
      { "UEMeasurementParameterModAllow", "rnsap.UEMeasurementParameterModAllow",
        FT_UINT32, BASE_DEC, VALS(rnsap_UEMeasurementParameterModAllow_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementReportCharacteristics_PDU,
      { "UEMeasurementReportCharacteristics", "rnsap.UEMeasurementReportCharacteristics",
        FT_UINT32, BASE_DEC, VALS(rnsap_UEMeasurementReportCharacteristics_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementTimeslotInfoHCR_PDU,
      { "UEMeasurementTimeslotInfoHCR", "rnsap.UEMeasurementTimeslotInfoHCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementTimeslotInfoLCR_PDU,
      { "UEMeasurementTimeslotInfoLCR", "rnsap.UEMeasurementTimeslotInfoLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementTimeslotInfo768_PDU,
      { "UEMeasurementTimeslotInfo768", "rnsap.UEMeasurementTimeslotInfo768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementType_PDU,
      { "UEMeasurementType", "rnsap.UEMeasurementType",
        FT_UINT32, BASE_DEC, VALS(rnsap_UEMeasurementType_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementValueTransmittedPowerList768_PDU,
      { "UEMeasurementValueTransmittedPowerList768", "rnsap.UEMeasurementValueTransmittedPowerList768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementValueTimeslotISCPList768_PDU,
      { "UEMeasurementValueTimeslotISCPList768", "rnsap.UEMeasurementValueTimeslotISCPList768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementValueInformation_PDU,
      { "UEMeasurementValueInformation", "rnsap.UEMeasurementValueInformation",
        FT_UINT32, BASE_DEC, VALS(rnsap_UEMeasurementValueInformation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UE_SupportIndicatorExtension_PDU,
      { "UE-SupportIndicatorExtension", "rnsap.UE_SupportIndicatorExtension",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UE_State_PDU,
      { "UE-State", "rnsap.UE_State",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_State_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UE_transmission_power_headroom_Value_PDU,
      { "UE-transmission-power-headroom-Value", "rnsap.UE_transmission_power_headroom_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPDCHIndicatorEDCH_PDU,
      { "UL-DPDCHIndicatorEDCH", "rnsap.UL_DPDCHIndicatorEDCH",
        FT_UINT32, BASE_DEC, VALS(rnsap_UL_DPDCHIndicatorEDCH_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_PLCCHinformation_PDU,
      { "PLCCHinformation", "rnsap.PLCCHinformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_Timeslot_ISCP_Value_PDU,
      { "UL-Timeslot-ISCP-Value", "rnsap.UL_Timeslot_ISCP_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_Timeslot_ISCP_Value_IncrDecrThres_PDU,
      { "UL-Timeslot-ISCP-Value-IncrDecrThres", "rnsap.UL_Timeslot_ISCP_Value_IncrDecrThres",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_TimingAdvanceCtrl_LCR_PDU,
      { "UL-TimingAdvanceCtrl-LCR", "rnsap.UL_TimingAdvanceCtrl_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_SIR_PDU,
      { "UL-SIR", "rnsap.UL_SIR",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UC_ID_PDU,
      { "UC-ID", "rnsap.UC_ID_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_Synchronisation_Parameters_LCR_PDU,
      { "UL-Synchronisation-Parameters-LCR", "rnsap.UL_Synchronisation_Parameters_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UPPCHPositionLCR_PDU,
      { "UPPCHPositionLCR", "rnsap.UPPCHPositionLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UpPTSInterferenceValue_PDU,
      { "UpPTSInterferenceValue", "rnsap.UpPTSInterferenceValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Unidirectional_DCH_Indicator_PDU,
      { "Unidirectional-DCH-Indicator", "rnsap.Unidirectional_DCH_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_Unidirectional_DCH_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_URA_ID_PDU,
      { "URA-ID", "rnsap.URA_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_URA_Information_PDU,
      { "URA-Information", "rnsap.URA_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_USCH_Information_PDU,
      { "USCH-Information", "rnsap.USCH_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Usefulness_Of_Battery_Optimization_PDU,
      { "Usefulness-Of-Battery-Optimization", "rnsap.Usefulness_Of_Battery_Optimization",
        FT_UINT32, BASE_DEC, VALS(rnsap_Usefulness_Of_Battery_Optimization_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_User_Plane_Congestion_Fields_Inclusion_PDU,
      { "User-Plane-Congestion-Fields-Inclusion", "rnsap.User_Plane_Congestion_Fields_Inclusion",
        FT_UINT32, BASE_DEC, VALS(rnsap_User_Plane_Congestion_Fields_Inclusion_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UMTS_Cell_Info_PDU,
      { "UMTS-Cell-Info", "rnsap.UMTS_Cell_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CLTD_Information_PDU,
      { "UL-CLTD-Information", "rnsap.UL_CLTD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CLTD_Information_Reconf_PDU,
      { "UL-CLTD-Information-Reconf", "rnsap.UL_CLTD_Information_Reconf_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CLTD_State_Update_Information_PDU,
      { "UL-CLTD-State-Update-Information", "rnsap.UL_CLTD_State_Update_Information",
        FT_UINT32, BASE_DEC, VALS(rnsap_UL_CLTD_State_Update_Information_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCCH2_Information_PDU,
      { "UL-DPCCH2-Information", "rnsap.UL_DPCCH2_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCCH2_Information_Reconf_PDU,
      { "UL-DPCCH2-Information-Reconf", "rnsap.UL_DPCCH2_Information_Reconf_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_MIMO_Information_PDU,
      { "UL-MIMO-Information", "rnsap.UL_MIMO_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_MIMO_Reconfiguration_PDU,
      { "UL-MIMO-Reconfiguration", "rnsap.UL_MIMO_Reconfiguration",
        FT_UINT32, BASE_DEC, VALS(rnsap_UL_MIMO_Reconfiguration_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_MIMO_DL_Control_Channel_Information_PDU,
      { "UL-MIMO-DL-Control-Channel-Information", "rnsap.UL_MIMO_DL_Control_Channel_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkSetupRequestFDD_PDU,
      { "RadioLinkSetupRequestFDD", "rnsap.RadioLinkSetupRequestFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_Information_RL_SetupRqstFDD_PDU,
      { "UL-DPCH-Information-RL-SetupRqstFDD", "rnsap.UL_DPCH_Information_RL_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_Information_RL_SetupRqstFDD_PDU,
      { "DL-DPCH-Information-RL-SetupRqstFDD", "rnsap.DL_DPCH_Information_RL_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_RL_SetupRqstFDD_PDU,
      { "RL-InformationList-RL-SetupRqstFDD", "rnsap.RL_InformationList_RL_SetupRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationItem_RL_SetupRqstFDD_PDU,
      { "RL-InformationItem-RL-SetupRqstFDD", "rnsap.RL_InformationItem_RL_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Information_RL_Setup_List_PDU,
      { "Additional-HS-Cell-Information-RL-Setup-List", "rnsap.Additional_HS_Cell_Information_RL_Setup_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_F_DPCH_Information_RL_SetupRqstFDD_PDU,
      { "F-DPCH-Information-RL-SetupRqstFDD", "rnsap.F_DPCH_Information_RL_SetupRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkSetupRequestTDD_PDU,
      { "RadioLinkSetupRequestTDD", "rnsap.RadioLinkSetupRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_Physical_Channel_Information_RL_SetupRqstTDD_PDU,
      { "UL-Physical-Channel-Information-RL-SetupRqstTDD", "rnsap.UL_Physical_Channel_Information_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_Physical_Channel_Information_RL_SetupRqstTDD_PDU,
      { "DL-Physical-Channel-Information-RL-SetupRqstTDD", "rnsap.DL_Physical_Channel_Information_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU,
      { "UL-CCTrCH-InformationList-RL-SetupRqstTDD", "rnsap.UL_CCTrCH_InformationList_RL_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU,
      { "UL-CCTrCH-InformationItem-RL-SetupRqstTDD", "rnsap.UL_CCTrCH_InformationItem_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU,
      { "DL-CCTrCH-InformationList-RL-SetupRqstTDD", "rnsap.DL_CCTrCH_InformationList_RL_SetupRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU,
      { "DL-CCTrCH-InformationItem-RL-SetupRqstTDD", "rnsap.DL_CCTrCH_InformationItem_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Information_RL_SetupRqstTDD_PDU,
      { "RL-Information-RL-SetupRqstTDD", "rnsap.RL_Information_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkSetupResponseFDD_PDU,
      { "RadioLinkSetupResponseFDD", "rnsap.RadioLinkSetupResponseFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponseList_RL_SetupRspFDD_PDU,
      { "RL-InformationResponseList-RL-SetupRspFDD", "rnsap.RL_InformationResponseList_RL_SetupRspFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponseItem_RL_SetupRspFDD_PDU,
      { "RL-InformationResponseItem-RL-SetupRspFDD", "rnsap.RL_InformationResponseItem_RL_SetupRspFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Information_Response_List_PDU,
      { "Additional-HS-Cell-Information-Response-List", "rnsap.Additional_HS_Cell_Information_Response_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkSetupResponseTDD_PDU,
      { "RadioLinkSetupResponseTDD", "rnsap.RadioLinkSetupResponseTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponse_RL_SetupRspTDD_PDU,
      { "RL-InformationResponse-RL-SetupRspTDD", "rnsap.RL_InformationResponse_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD_PDU,
      { "UL-CCTrCHInformationListIE-RL-SetupRspTDD", "rnsap.UL_CCTrCHInformationListIE_RL_SetupRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD_PDU,
      { "UL-DPCH-InformationItem-RL-SetupRspTDD", "rnsap.UL_DPCH_InformationItem_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD_PDU,
      { "DL-CCTrCHInformationListIE-RL-SetupRspTDD", "rnsap.DL_CCTrCHInformationListIE_RL_SetupRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD_PDU,
      { "DL-DPCH-InformationItem-RL-SetupRspTDD", "rnsap.DL_DPCH_InformationItem_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_InformationListIEs_RL_SetupRspTDD_PDU,
      { "DSCH-InformationListIEs-RL-SetupRspTDD", "rnsap.DSCH_InformationListIEs_RL_SetupRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_USCH_InformationListIEs_RL_SetupRspTDD_PDU,
      { "USCH-InformationListIEs-RL-SetupRspTDD", "rnsap.USCH_InformationListIEs_RL_SetupRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_LCR_InformationResponse_RL_SetupRspTDD_PDU,
      { "RL-LCR-InformationResponse-RL-SetupRspTDD", "rnsap.RL_LCR_InformationResponse_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD_PDU,
      { "UL-LCR-CCTrCHInformationListIE-RL-SetupRspTDD", "rnsap.UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_LCR_InformationItem_RL_SetupRspTDD_PDU,
      { "UL-DPCH-LCR-InformationItem-RL-SetupRspTDD", "rnsap.UL_DPCH_LCR_InformationItem_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD_PDU,
      { "DL-CCTrCH-LCR-InformationListIE-RL-SetupRspTDD", "rnsap.DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_LCR_InformationItem_RL_SetupRspTDD_PDU,
      { "DL-DPCH-LCR-InformationItem-RL-SetupRspTDD", "rnsap.DL_DPCH_LCR_InformationItem_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_LCR_InformationListIEs_RL_SetupRspTDD_PDU,
      { "DSCH-LCR-InformationListIEs-RL-SetupRspTDD", "rnsap.DSCH_LCR_InformationListIEs_RL_SetupRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_USCH_LCR_InformationListIEs_RL_SetupRspTDD_PDU,
      { "USCH-LCR-InformationListIEs-RL-SetupRspTDD", "rnsap.USCH_LCR_InformationListIEs_RL_SetupRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponse_RL_SetupRspTDD768_PDU,
      { "RL-InformationResponse-RL-SetupRspTDD768", "rnsap.RL_InformationResponse_RL_SetupRspTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD768_PDU,
      { "UL-CCTrCHInformationListIE-RL-SetupRspTDD768", "rnsap.UL_CCTrCHInformationListIE_RL_SetupRspTDD768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD768_PDU,
      { "UL-DPCH-InformationItem-RL-SetupRspTDD768", "rnsap.UL_DPCH_InformationItem_RL_SetupRspTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD768_PDU,
      { "DL-CCTrCHInformationListIE-RL-SetupRspTDD768", "rnsap.DL_CCTrCHInformationListIE_RL_SetupRspTDD768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD768_PDU,
      { "DL-DPCH-InformationItem-RL-SetupRspTDD768", "rnsap.DL_DPCH_InformationItem_RL_SetupRspTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkSetupFailureFDD_PDU,
      { "RadioLinkSetupFailureFDD", "rnsap.RadioLinkSetupFailureFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CauseLevel_RL_SetupFailureFDD_PDU,
      { "CauseLevel-RL-SetupFailureFDD", "rnsap.CauseLevel_RL_SetupFailureFDD",
        FT_UINT32, BASE_DEC, VALS(rnsap_CauseLevel_RL_SetupFailureFDD_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD_PDU,
      { "UnsuccessfulRL-InformationResponse-RL-SetupFailureFDD", "rnsap.UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SuccessfulRL_InformationResponse_RL_SetupFailureFDD_PDU,
      { "SuccessfulRL-InformationResponse-RL-SetupFailureFDD", "rnsap.SuccessfulRL_InformationResponse_RL_SetupFailureFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkSetupFailureTDD_PDU,
      { "RadioLinkSetupFailureTDD", "rnsap.RadioLinkSetupFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CauseLevel_RL_SetupFailureTDD_PDU,
      { "CauseLevel-RL-SetupFailureTDD", "rnsap.CauseLevel_RL_SetupFailureTDD",
        FT_UINT32, BASE_DEC, VALS(rnsap_CauseLevel_RL_SetupFailureTDD_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD_PDU,
      { "UnsuccessfulRL-InformationResponse-RL-SetupFailureTDD", "rnsap.UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkAdditionRequestFDD_PDU,
      { "RadioLinkAdditionRequestFDD", "rnsap.RadioLinkAdditionRequestFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_RL_AdditionRqstFDD_PDU,
      { "RL-InformationList-RL-AdditionRqstFDD", "rnsap.RL_InformationList_RL_AdditionRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Information_RL_AdditionRqstFDD_PDU,
      { "RL-Information-RL-AdditionRqstFDD", "rnsap.RL_Information_RL_AdditionRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Information_RL_Addition_List_PDU,
      { "Additional-HS-Cell-Information-RL-Addition-List", "rnsap.Additional_HS_Cell_Information_RL_Addition_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_RL_Add_Req_PDU,
      { "Additional-EDCH-Cell-Information-RL-Add-Req", "rnsap.Additional_EDCH_Cell_Information_RL_Add_Req_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDPCH_Information_RLAdditionReq_FDD_PDU,
      { "EDPCH-Information-RLAdditionReq-FDD", "rnsap.EDPCH_Information_RLAdditionReq_FDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkAdditionRequestTDD_PDU,
      { "RadioLinkAdditionRequestTDD", "rnsap.RadioLinkAdditionRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Information_RL_AdditionRqstTDD_PDU,
      { "RL-Information-RL-AdditionRqstTDD", "rnsap.RL_Information_RL_AdditionRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU,
      { "UL-CCTrCH-InformationList-RL-AdditionRqstTDD", "rnsap.UL_CCTrCH_InformationList_RL_AdditionRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD_PDU,
      { "UL-CCTrCH-InformationItem-RL-AdditionRqstTDD", "rnsap.UL_CCTrCH_InformationItem_RL_AdditionRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU,
      { "DL-CCTrCH-InformationList-RL-AdditionRqstTDD", "rnsap.DL_CCTrCH_InformationList_RL_AdditionRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD_PDU,
      { "DL-CCTrCH-InformationItem-RL-AdditionRqstTDD", "rnsap.DL_CCTrCH_InformationItem_RL_AdditionRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkAdditionResponseFDD_PDU,
      { "RadioLinkAdditionResponseFDD", "rnsap.RadioLinkAdditionResponseFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponseList_RL_AdditionRspFDD_PDU,
      { "RL-InformationResponseList-RL-AdditionRspFDD", "rnsap.RL_InformationResponseList_RL_AdditionRspFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponseItem_RL_AdditionRspFDD_PDU,
      { "RL-InformationResponseItem-RL-AdditionRspFDD", "rnsap.RL_InformationResponseItem_RL_AdditionRspFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Change_Information_Response_List_PDU,
      { "Additional-HS-Cell-Change-Information-Response-List", "rnsap.Additional_HS_Cell_Change_Information_Response_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkAdditionResponseTDD_PDU,
      { "RadioLinkAdditionResponseTDD", "rnsap.RadioLinkAdditionResponseTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponse_RL_AdditionRspTDD_PDU,
      { "RL-InformationResponse-RL-AdditionRspTDD", "rnsap.RL_InformationResponse_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD_PDU,
      { "UL-CCTrCHInformationListIE-RL-AdditionRspTDD", "rnsap.UL_CCTrCHInformationListIE_RL_AdditionRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD_PDU,
      { "UL-DPCH-InformationItem-RL-AdditionRspTDD", "rnsap.UL_DPCH_InformationItem_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD_PDU,
      { "DL-CCTrCHInformationListIE-RL-AdditionRspTDD", "rnsap.DL_CCTrCHInformationListIE_RL_AdditionRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD_PDU,
      { "DL-DPCH-InformationItem-RL-AdditionRspTDD", "rnsap.DL_DPCH_InformationItem_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_InformationListIE_RL_AdditionRspTDD_PDU,
      { "DSCH-InformationListIE-RL-AdditionRspTDD", "rnsap.DSCH_InformationListIE_RL_AdditionRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_USCH_InformationListIE_RL_AdditionRspTDD_PDU,
      { "USCH-InformationListIE-RL-AdditionRspTDD", "rnsap.USCH_InformationListIE_RL_AdditionRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_LCR_InformationResponse_RL_AdditionRspTDD_PDU,
      { "RL-LCR-InformationResponse-RL-AdditionRspTDD", "rnsap.RL_LCR_InformationResponse_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_PDU,
      { "UL-CCTrCH-LCR-InformationListIE-RL-AdditionRspTDD", "rnsap.UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_PDU,
      { "UL-DPCH-LCR-InformationItem-RL-AdditionRspTDD", "rnsap.UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_PDU,
      { "DL-CCTrCH-LCR-InformationListIE-RL-AdditionRspTDD", "rnsap.DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_PDU,
      { "DL-DPCH-LCR-InformationItem-RL-AdditionRspTDD", "rnsap.DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD_PDU,
      { "DSCH-LCR-InformationListIEs-RL-AdditionRspTDD", "rnsap.DSCH_LCR_InformationListIEs_RL_AdditionRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_USCH_LCR_InformationListIEs_RL_AdditionRspTDD_PDU,
      { "USCH-LCR-InformationListIEs-RL-AdditionRspTDD", "rnsap.USCH_LCR_InformationListIEs_RL_AdditionRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponse_RL_AdditionRspTDD768_PDU,
      { "RL-InformationResponse-RL-AdditionRspTDD768", "rnsap.RL_InformationResponse_RL_AdditionRspTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD768_PDU,
      { "UL-CCTrCHInformationListIE-RL-AdditionRspTDD768", "rnsap.UL_CCTrCHInformationListIE_RL_AdditionRspTDD768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD768_PDU,
      { "UL-DPCH-InformationItem-RL-AdditionRspTDD768", "rnsap.UL_DPCH_InformationItem_RL_AdditionRspTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD768_PDU,
      { "DL-CCTrCHInformationListIE-RL-AdditionRspTDD768", "rnsap.DL_CCTrCHInformationListIE_RL_AdditionRspTDD768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD768_PDU,
      { "DL-DPCH-InformationItem-RL-AdditionRspTDD768", "rnsap.DL_DPCH_InformationItem_RL_AdditionRspTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkAdditionFailureFDD_PDU,
      { "RadioLinkAdditionFailureFDD", "rnsap.RadioLinkAdditionFailureFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CauseLevel_RL_AdditionFailureFDD_PDU,
      { "CauseLevel-RL-AdditionFailureFDD", "rnsap.CauseLevel_RL_AdditionFailureFDD",
        FT_UINT32, BASE_DEC, VALS(rnsap_CauseLevel_RL_AdditionFailureFDD_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD_PDU,
      { "UnsuccessfulRL-InformationResponse-RL-AdditionFailureFDD", "rnsap.UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SuccessfulRL_InformationResponse_RL_AdditionFailureFDD_PDU,
      { "SuccessfulRL-InformationResponse-RL-AdditionFailureFDD", "rnsap.SuccessfulRL_InformationResponse_RL_AdditionFailureFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkAdditionFailureTDD_PDU,
      { "RadioLinkAdditionFailureTDD", "rnsap.RadioLinkAdditionFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CauseLevel_RL_AdditionFailureTDD_PDU,
      { "CauseLevel-RL-AdditionFailureTDD", "rnsap.CauseLevel_RL_AdditionFailureTDD",
        FT_UINT32, BASE_DEC, VALS(rnsap_CauseLevel_RL_AdditionFailureTDD_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD_PDU,
      { "UnsuccessfulRL-InformationResponse-RL-AdditionFailureTDD", "rnsap.UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkDeletionRequest_PDU,
      { "RadioLinkDeletionRequest", "rnsap.RadioLinkDeletionRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_RL_DeletionRqst_PDU,
      { "RL-InformationList-RL-DeletionRqst", "rnsap.RL_InformationList_RL_DeletionRqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Information_RL_DeletionRqst_PDU,
      { "RL-Information-RL-DeletionRqst", "rnsap.RL_Information_RL_DeletionRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkDeletionResponse_PDU,
      { "RadioLinkDeletionResponse", "rnsap.RadioLinkDeletionResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkReconfigurationPrepareFDD_PDU,
      { "RadioLinkReconfigurationPrepareFDD", "rnsap.RadioLinkReconfigurationPrepareFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_Information_RL_ReconfPrepFDD_PDU,
      { "UL-DPCH-Information-RL-ReconfPrepFDD", "rnsap.UL_DPCH_Information_RL_ReconfPrepFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_Information_RL_ReconfPrepFDD_PDU,
      { "DL-DPCH-Information-RL-ReconfPrepFDD", "rnsap.DL_DPCH_Information_RL_ReconfPrepFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_Power_Information_RL_ReconfPrepFDD_PDU,
      { "DL-DPCH-Power-Information-RL-ReconfPrepFDD", "rnsap.DL_DPCH_Power_Information_RL_ReconfPrepFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_DeleteList_RL_ReconfPrepFDD_PDU,
      { "DCH-DeleteList-RL-ReconfPrepFDD", "rnsap.DCH_DeleteList_RL_ReconfPrepFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_RL_ReconfPrepFDD_PDU,
      { "RL-InformationList-RL-ReconfPrepFDD", "rnsap.RL_InformationList_RL_ReconfPrepFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Information_RL_ReconfPrepFDD_PDU,
      { "RL-Information-RL-ReconfPrepFDD", "rnsap.RL_Information_RL_ReconfPrepFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep_PDU,
      { "Additional-HS-Cell-Information-RL-Reconf-Prep", "rnsap.Additional_HS_Cell_Information_RL_Reconf_Prep",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Prep_PDU,
      { "Additional-EDCH-Cell-Information-RL-Reconf-Prep", "rnsap.Additional_EDCH_Cell_Information_RL_Reconf_Prep_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_F_DPCH_Information_RL_ReconfPrepFDD_PDU,
      { "F-DPCH-Information-RL-ReconfPrepFDD", "rnsap.F_DPCH_Information_RL_ReconfPrepFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkReconfigurationPrepareTDD_PDU,
      { "RadioLinkReconfigurationPrepareTDD", "rnsap.RadioLinkReconfigurationPrepareTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU,
      { "UL-CCTrCH-InformationAddList-RL-ReconfPrepTDD", "rnsap.UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_AddInformation_RL_ReconfPrepTDD_PDU,
      { "UL-CCTrCH-AddInformation-RL-ReconfPrepTDD", "rnsap.UL_CCTrCH_AddInformation_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU,
      { "UL-CCTrCH-InformationModifyList-RL-ReconfPrepTDD", "rnsap.UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD_PDU,
      { "UL-CCTrCH-ModifyInformation-RL-ReconfPrepTDD", "rnsap.UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU,
      { "UL-CCTrCH-InformationDeleteList-RL-ReconfPrepTDD", "rnsap.UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD_PDU,
      { "UL-CCTrCH-DeleteInformation-RL-ReconfPrepTDD", "rnsap.UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU,
      { "DL-CCTrCH-InformationAddList-RL-ReconfPrepTDD", "rnsap.DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_PDU,
      { "DL-CCTrCH-InformationAddItem-RL-ReconfPrepTDD", "rnsap.DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU,
      { "DL-CCTrCH-InformationModifyList-RL-ReconfPrepTDD", "rnsap.DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_PDU,
      { "DL-CCTrCH-InformationModifyItem-RL-ReconfPrepTDD", "rnsap.DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU,
      { "DL-CCTrCH-InformationDeleteList-RL-ReconfPrepTDD", "rnsap.DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_PDU,
      { "DL-CCTrCH-InformationDeleteItem-RL-ReconfPrepTDD", "rnsap.DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_DeleteList_RL_ReconfPrepTDD_PDU,
      { "DCH-DeleteList-RL-ReconfPrepTDD", "rnsap.DCH_DeleteList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_ModifyList_RL_ReconfPrepTDD_PDU,
      { "DSCH-ModifyList-RL-ReconfPrepTDD", "rnsap.DSCH_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_DeleteList_RL_ReconfPrepTDD_PDU,
      { "DSCH-DeleteList-RL-ReconfPrepTDD", "rnsap.DSCH_DeleteList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_USCH_ModifyList_RL_ReconfPrepTDD_PDU,
      { "USCH-ModifyList-RL-ReconfPrepTDD", "rnsap.USCH_ModifyList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_USCH_DeleteList_RL_ReconfPrepTDD_PDU,
      { "USCH-DeleteList-RL-ReconfPrepTDD", "rnsap.USCH_DeleteList_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Information_RL_ReconfPrepTDD_PDU,
      { "RL-Information-RL-ReconfPrepTDD", "rnsap.RL_Information_RL_ReconfPrepTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkReconfigurationReadyFDD_PDU,
      { "RadioLinkReconfigurationReadyFDD", "rnsap.RadioLinkReconfigurationReadyFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponseList_RL_ReconfReadyFDD_PDU,
      { "RL-InformationResponseList-RL-ReconfReadyFDD", "rnsap.RL_InformationResponseList_RL_ReconfReadyFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponseItem_RL_ReconfReadyFDD_PDU,
      { "RL-InformationResponseItem-RL-ReconfReadyFDD", "rnsap.RL_InformationResponseItem_RL_ReconfReadyFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_RL_Reconf_Response_PDU,
      { "Additional-HS-Cell-RL-Reconf-Response", "rnsap.Additional_HS_Cell_RL_Reconf_Response",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkReconfigurationReadyTDD_PDU,
      { "RadioLinkReconfigurationReadyTDD", "rnsap.RadioLinkReconfigurationReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCHInformationListIE_RL_ReconfReadyTDD_PDU,
      { "UL-CCTrCHInformationListIE-RL-ReconfReadyTDD", "rnsap.UL_CCTrCHInformationListIE_RL_ReconfReadyTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_PDU,
      { "UL-DPCH-LCR-InformationAddList-RL-ReconfReadyTDD", "rnsap.UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_PDU,
      { "UL-DPCH-InformationAddListIE-RL-ReconfReadyTDD", "rnsap.UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_InformationAddList_RL_ReconfReadyTDD768_PDU,
      { "UL-DPCH-InformationAddList-RL-ReconfReadyTDD768", "rnsap.UL_DPCH_InformationAddList_RL_ReconfReadyTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_PDU,
      { "UL-DPCH-InformationModifyListIE-RL-ReconfReadyTDD", "rnsap.UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_PDU,
      { "UL-TimeslotLCR-InformationModifyList-RL-ReconfReadyTDD", "rnsap.UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_PDU,
      { "UL-Timeslot-InformationModifyList-RL-ReconfReadyTDD768", "rnsap.UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_PDU,
      { "UL-DPCH-InformationDeleteListIE-RL-ReconfReadyTDD", "rnsap.UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCHInformationListIE_RL_ReconfReadyTDD_PDU,
      { "DL-CCTrCHInformationListIE-RL-ReconfReadyTDD", "rnsap.DL_CCTrCHInformationListIE_RL_ReconfReadyTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_PDU,
      { "DL-DPCH-LCR-InformationAddList-RL-ReconfReadyTDD", "rnsap.DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_PDU,
      { "DL-DPCH-InformationAddListIE-RL-ReconfReadyTDD", "rnsap.DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_InformationAddList_RL_ReconfReadyTDD768_PDU,
      { "DL-DPCH-InformationAddList-RL-ReconfReadyTDD768", "rnsap.DL_DPCH_InformationAddList_RL_ReconfReadyTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_PDU,
      { "DL-DPCH-InformationModifyListIE-RL-ReconfReadyTDD", "rnsap.DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_PDU,
      { "DL-TimeslotLCR-InformationModifyList-RL-ReconfReadyTDD", "rnsap.DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_PDU,
      { "DL-Timeslot-InformationModifyList-RL-ReconfReadyTDD768", "rnsap.DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_PDU,
      { "DL-DPCH-InformationDeleteListIE-RL-ReconfReadyTDD", "rnsap.DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD_PDU,
      { "DL-DPCH-InformationDeleteList768-RL-ReconfReadyTDD", "rnsap.DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_PDU,
      { "DSCHToBeAddedOrModifiedList-RL-ReconfReadyTDD", "rnsap.DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_PDU,
      { "USCHToBeAddedOrModifiedList-RL-ReconfReadyTDD", "rnsap.USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_RL_InformationResponse_RL_ReconfReadyTDD_PDU,
      { "Multiple-RL-InformationResponse-RL-ReconfReadyTDD", "rnsap.Multiple_RL_InformationResponse_RL_ReconfReadyTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkReconfigurationCommit_PDU,
      { "RadioLinkReconfigurationCommit", "rnsap.RadioLinkReconfigurationCommit_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkReconfigurationFailure_PDU,
      { "RadioLinkReconfigurationFailure", "rnsap.RadioLinkReconfigurationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CauseLevel_RL_ReconfFailure_PDU,
      { "CauseLevel-RL-ReconfFailure", "rnsap.CauseLevel_RL_ReconfFailure",
        FT_UINT32, BASE_DEC, VALS(rnsap_CauseLevel_RL_ReconfFailure_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_ReconfigurationFailure_RL_ReconfFail_PDU,
      { "RL-ReconfigurationFailure-RL-ReconfFail", "rnsap.RL_ReconfigurationFailure_RL_ReconfFail_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkReconfigurationCancel_PDU,
      { "RadioLinkReconfigurationCancel", "rnsap.RadioLinkReconfigurationCancel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkReconfigurationRequestFDD_PDU,
      { "RadioLinkReconfigurationRequestFDD", "rnsap.RadioLinkReconfigurationRequestFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_Information_RL_ReconfRqstFDD_PDU,
      { "UL-DPCH-Information-RL-ReconfRqstFDD", "rnsap.UL_DPCH_Information_RL_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_Information_RL_ReconfRqstFDD_PDU,
      { "DL-DPCH-Information-RL-ReconfRqstFDD", "rnsap.DL_DPCH_Information_RL_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_DeleteList_RL_ReconfRqstFDD_PDU,
      { "DCH-DeleteList-RL-ReconfRqstFDD", "rnsap.DCH_DeleteList_RL_ReconfRqstFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req_PDU,
      { "Additional-HS-Cell-Information-RL-Reconf-Req", "rnsap.Additional_HS_Cell_Information_RL_Reconf_Req",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Req_PDU,
      { "Additional-EDCH-Cell-Information-RL-Reconf-Req", "rnsap.Additional_EDCH_Cell_Information_RL_Reconf_Req_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_ReconfigurationRequestFDD_RL_InformationList_PDU,
      { "RL-ReconfigurationRequestFDD-RL-InformationList", "rnsap.RL_ReconfigurationRequestFDD_RL_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_ReconfigurationRequestFDD_RL_Information_IEs_PDU,
      { "RL-ReconfigurationRequestFDD-RL-Information-IEs", "rnsap.RL_ReconfigurationRequestFDD_RL_Information_IEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkReconfigurationRequestTDD_PDU,
      { "RadioLinkReconfigurationRequestTDD", "rnsap.RadioLinkReconfigurationRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU,
      { "UL-CCTrCH-InformationModifyList-RL-ReconfRqstTDD", "rnsap.UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU,
      { "UL-CCTrCH-InformationModifyItem-RL-ReconfRqstTDD", "rnsap.UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU,
      { "UL-CCTrCH-InformationDeleteList-RL-ReconfRqstTDD", "rnsap.UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU,
      { "UL-CCTrCH-InformationDeleteItem-RL-ReconfRqstTDD", "rnsap.UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU,
      { "DL-CCTrCH-InformationModifyList-RL-ReconfRqstTDD", "rnsap.DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU,
      { "DL-CCTrCH-InformationModifyItem-RL-ReconfRqstTDD", "rnsap.DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU,
      { "DL-CCTrCH-InformationDeleteList-RL-ReconfRqstTDD", "rnsap.DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU,
      { "DL-CCTrCH-InformationDeleteItem-RL-ReconfRqstTDD", "rnsap.DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_DeleteList_RL_ReconfRqstTDD_PDU,
      { "DCH-DeleteList-RL-ReconfRqstTDD", "rnsap.DCH_DeleteList_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_RL_ReconfigurationRequestTDD_RL_Information_PDU,
      { "Multiple-RL-ReconfigurationRequestTDD-RL-Information", "rnsap.Multiple_RL_ReconfigurationRequestTDD_RL_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkReconfigurationResponseFDD_PDU,
      { "RadioLinkReconfigurationResponseFDD", "rnsap.RadioLinkReconfigurationResponseFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponseList_RL_ReconfRspFDD_PDU,
      { "RL-InformationResponseList-RL-ReconfRspFDD", "rnsap.RL_InformationResponseList_RL_ReconfRspFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponseItem_RL_ReconfRspFDD_PDU,
      { "RL-InformationResponseItem-RL-ReconfRspFDD", "rnsap.RL_InformationResponseItem_RL_ReconfRspFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkReconfigurationResponseTDD_PDU,
      { "RadioLinkReconfigurationResponseTDD", "rnsap.RadioLinkReconfigurationResponseTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponse_RL_ReconfRspTDD_PDU,
      { "RL-InformationResponse-RL-ReconfRspTDD", "rnsap.RL_InformationResponse_RL_ReconfRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationList_RL_ReconfRspTDD_PDU,
      { "DL-CCTrCH-InformationList-RL-ReconfRspTDD", "rnsap.DL_CCTrCH_InformationList_RL_ReconfRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD_PDU,
      { "DL-DPCH-InformationModifyItem-LCR-RL-ReconfRspTDD", "rnsap.DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_RL_InformationResponse_RL_ReconfRspTDD_PDU,
      { "Multiple-RL-InformationResponse-RL-ReconfRspTDD", "rnsap.Multiple_RL_InformationResponse_RL_ReconfRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkFailureIndication_PDU,
      { "RadioLinkFailureIndication", "rnsap.RadioLinkFailureIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Reporting_Object_RL_FailureInd_PDU,
      { "Reporting-Object-RL-FailureInd", "rnsap.Reporting_Object_RL_FailureInd",
        FT_UINT32, BASE_DEC, VALS(rnsap_Reporting_Object_RL_FailureInd_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Information_RL_FailureInd_PDU,
      { "RL-Information-RL-FailureInd", "rnsap.RL_Information_RL_FailureInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_Information_RL_FailureInd_PDU,
      { "RL-Set-Information-RL-FailureInd", "rnsap.RL_Set_Information_RL_FailureInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CCTrCH_InformationItem_RL_FailureInd_PDU,
      { "CCTrCH-InformationItem-RL-FailureInd", "rnsap.CCTrCH_InformationItem_RL_FailureInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkPreemptionRequiredIndication_PDU,
      { "RadioLinkPreemptionRequiredIndication", "rnsap.RadioLinkPreemptionRequiredIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_RL_PreemptRequiredInd_PDU,
      { "RL-InformationList-RL-PreemptRequiredInd", "rnsap.RL_InformationList_RL_PreemptRequiredInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationItem_RL_PreemptRequiredInd_PDU,
      { "RL-InformationItem-RL-PreemptRequiredInd", "rnsap.RL_InformationItem_RL_PreemptRequiredInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd_PDU,
      { "HSDSCHMacdFlowSpecificInformationList-RL-PreemptRequiredInd", "rnsap.HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd_PDU,
      { "HSDSCHMacdFlowSpecificInformationItem-RL-PreemptRequiredInd", "rnsap.HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd_PDU,
      { "EDCH-MacdFlowSpecificInformationList-RL-PreemptRequiredInd", "rnsap.EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd_PDU,
      { "EDCH-MacdFlowSpecificInformationItem-RL-PreemptRequiredInd", "rnsap.EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkRestoreIndication_PDU,
      { "RadioLinkRestoreIndication", "rnsap.RadioLinkRestoreIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Reporting_Object_RL_RestoreInd_PDU,
      { "Reporting-Object-RL-RestoreInd", "rnsap.Reporting_Object_RL_RestoreInd",
        FT_UINT32, BASE_DEC, VALS(rnsap_Reporting_Object_RL_RestoreInd_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Information_RL_RestoreInd_PDU,
      { "RL-Information-RL-RestoreInd", "rnsap.RL_Information_RL_RestoreInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_Information_RL_RestoreInd_PDU,
      { "RL-Set-Information-RL-RestoreInd", "rnsap.RL_Set_Information_RL_RestoreInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CCTrCH_InformationItem_RL_RestoreInd_PDU,
      { "CCTrCH-InformationItem-RL-RestoreInd", "rnsap.CCTrCH_InformationItem_RL_RestoreInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_PowerControlRequest_PDU,
      { "DL-PowerControlRequest", "rnsap.DL_PowerControlRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_ReferencePowerInformationList_DL_PC_Rqst_PDU,
      { "DL-ReferencePowerInformationList-DL-PC-Rqst", "rnsap.DL_ReferencePowerInformationList_DL_PC_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_ReferencePowerInformation_DL_PC_Rqst_PDU,
      { "DL-ReferencePowerInformation-DL-PC-Rqst", "rnsap.DL_ReferencePowerInformation_DL_PC_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_PowerTimeslotControlRequest_PDU,
      { "DL-PowerTimeslotControlRequest", "rnsap.DL_PowerTimeslotControlRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PhysicalChannelReconfigurationRequestFDD_PDU,
      { "PhysicalChannelReconfigurationRequestFDD", "rnsap.PhysicalChannelReconfigurationRequestFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Information_PhyChReconfRqstFDD_PDU,
      { "RL-Information-PhyChReconfRqstFDD", "rnsap.RL_Information_PhyChReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PhysicalChannelReconfigurationRequestTDD_PDU,
      { "PhysicalChannelReconfigurationRequestTDD", "rnsap.PhysicalChannelReconfigurationRequestTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Information_PhyChReconfRqstTDD_PDU,
      { "RL-Information-PhyChReconfRqstTDD", "rnsap.RL_Information_PhyChReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_PDU,
      { "UL-CCTrCH-InformationListIE-PhyChReconfRqstTDD", "rnsap.UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_InformationItem_PhyChReconfRqstTDD_PDU,
      { "UL-DPCH-InformationItem-PhyChReconfRqstTDD", "rnsap.UL_DPCH_InformationItem_PhyChReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_PDU,
      { "UL-TimeslotLCR-InformationList-PhyChReconfRqstTDD", "rnsap.UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU,
      { "UL-Timeslot-InformationList-PhyChReconfRqstTDD768", "rnsap.UL_Timeslot_InformationList_PhyChReconfRqstTDD768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_PDU,
      { "DL-CCTrCH-InformationListIE-PhyChReconfRqstTDD", "rnsap.DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_InformationItem_PhyChReconfRqstTDD_PDU,
      { "DL-DPCH-InformationItem-PhyChReconfRqstTDD", "rnsap.DL_DPCH_InformationItem_PhyChReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_PDU,
      { "DL-TimeslotLCR-InformationList-PhyChReconfRqstTDD", "rnsap.DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU,
      { "DL-Timeslot-InformationList-PhyChReconfRqstTDD768", "rnsap.DL_Timeslot_InformationList_PhyChReconfRqstTDD768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD_PDU,
      { "HSPDSCH-Timeslot-InformationList-PhyChReconfRqstTDD", "rnsap.HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD_PDU,
      { "HSPDSCH-Timeslot-InformationListLCR-PhyChReconfRqstTDD", "rnsap.HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU,
      { "HSPDSCH-Timeslot-InformationList-PhyChReconfRqstTDD768", "rnsap.HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PhysicalChannelReconfigurationCommand_PDU,
      { "PhysicalChannelReconfigurationCommand", "rnsap.PhysicalChannelReconfigurationCommand_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PhysicalChannelReconfigurationFailure_PDU,
      { "PhysicalChannelReconfigurationFailure", "rnsap.PhysicalChannelReconfigurationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkCongestionIndication_PDU,
      { "RadioLinkCongestionIndication", "rnsap.RadioLinkCongestionIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_RL_CongestInd_PDU,
      { "RL-InformationList-RL-CongestInd", "rnsap.RL_InformationList_RL_CongestInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationItem_RL_CongestInd_PDU,
      { "RL-InformationItem-RL-CongestInd", "rnsap.RL_InformationItem_RL_CongestInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_Rate_InformationItem_RL_CongestInd_PDU,
      { "DCH-Rate-InformationItem-RL-CongestInd", "rnsap.DCH_Rate_InformationItem_RL_CongestInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_MacdFlowSpecificInformationList_RL_CongestInd_PDU,
      { "EDCH-MacdFlowSpecificInformationList-RL-CongestInd", "rnsap.EDCH_MacdFlowSpecificInformationList_RL_CongestInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_CongestInd_PDU,
      { "EDCH-MacdFlowSpecificInformationItem-RL-CongestInd", "rnsap.EDCH_MacdFlowSpecificInformationItem_RL_CongestInd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UplinkSignallingTransferIndicationFDD_PDU,
      { "UplinkSignallingTransferIndicationFDD", "rnsap.UplinkSignallingTransferIndicationFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UplinkSignallingTransferIndicationTDD_PDU,
      { "UplinkSignallingTransferIndicationTDD", "rnsap.UplinkSignallingTransferIndicationTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DownlinkSignallingTransferRequest_PDU,
      { "DownlinkSignallingTransferRequest", "rnsap.DownlinkSignallingTransferRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RelocationCommit_PDU,
      { "RelocationCommit", "rnsap.RelocationCommit_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PagingRequest_PDU,
      { "PagingRequest", "rnsap.PagingRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PagingArea_PagingRqst_PDU,
      { "PagingArea-PagingRqst", "rnsap.PagingArea_PagingRqst",
        FT_UINT32, BASE_DEC, VALS(rnsap_PagingArea_PagingRqst_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_CNOriginatedPage_PagingRqst_PDU,
      { "CNOriginatedPage-PagingRqst", "rnsap.CNOriginatedPage_PagingRqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DedicatedMeasurementInitiationRequest_PDU,
      { "DedicatedMeasurementInitiationRequest", "rnsap.DedicatedMeasurementInitiationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DedicatedMeasurementObjectType_DM_Rqst_PDU,
      { "DedicatedMeasurementObjectType-DM-Rqst", "rnsap.DedicatedMeasurementObjectType_DM_Rqst",
        FT_UINT32, BASE_DEC, VALS(rnsap_DedicatedMeasurementObjectType_DM_Rqst_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationItem_DM_Rqst_PDU,
      { "RL-InformationItem-DM-Rqst", "rnsap.RL_InformationItem_DM_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSSICH_Info_DM_Rqst_PDU,
      { "HSSICH-Info-DM-Rqst", "rnsap.HSSICH_Info_DM_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSSICH_Info_DM_Rqst_Extension_PDU,
      { "HSSICH-Info-DM-Rqst-Extension", "rnsap.HSSICH_Info_DM_Rqst_Extension",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_InformationItem_DM_Rqst_PDU,
      { "RL-Set-InformationItem-DM-Rqst", "rnsap.RL_Set_InformationItem_DM_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DedicatedMeasurementInitiationResponse_PDU,
      { "DedicatedMeasurementInitiationResponse", "rnsap.DedicatedMeasurementInitiationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DedicatedMeasurementObjectType_DM_Rsp_PDU,
      { "DedicatedMeasurementObjectType-DM-Rsp", "rnsap.DedicatedMeasurementObjectType_DM_Rsp",
        FT_UINT32, BASE_DEC, VALS(rnsap_DedicatedMeasurementObjectType_DM_Rsp_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationItem_DM_Rsp_PDU,
      { "RL-InformationItem-DM-Rsp", "rnsap.RL_InformationItem_DM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_InformationItem_DM_Rsp_PDU,
      { "RL-Set-InformationItem-DM-Rsp", "rnsap.RL_Set_InformationItem_DM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_PDU,
      { "Multiple-DedicatedMeasurementValueList-TDD-DM-Rsp", "rnsap.Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_PDU,
      { "Multiple-DedicatedMeasurementValueList-LCR-TDD-DM-Rsp", "rnsap.Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_PDU,
      { "Multiple-HSSICHMeasurementValueList-TDD-DM-Rsp", "rnsap.Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp_PDU,
      { "Multiple-DedicatedMeasurementValueList-TDD768-DM-Rsp", "rnsap.Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DedicatedMeasurementInitiationFailure_PDU,
      { "DedicatedMeasurementInitiationFailure", "rnsap.DedicatedMeasurementInitiationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DedicatedMeasurementObjectType_DM_Fail_PDU,
      { "DedicatedMeasurementObjectType-DM-Fail", "rnsap.DedicatedMeasurementObjectType_DM_Fail",
        FT_UINT32, BASE_DEC, VALS(rnsap_DedicatedMeasurementObjectType_DM_Fail_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Unsuccessful_InformationItem_DM_Fail_PDU,
      { "RL-Unsuccessful-InformationItem-DM-Fail", "rnsap.RL_Unsuccessful_InformationItem_DM_Fail_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Successful_InformationItem_DM_Fail_PDU,
      { "RL-Successful-InformationItem-DM-Fail", "rnsap.RL_Successful_InformationItem_DM_Fail_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail_PDU,
      { "RL-Set-Unsuccessful-InformationItem-DM-Fail", "rnsap.RL_Set_Unsuccessful_InformationItem_DM_Fail_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_Successful_InformationItem_DM_Fail_PDU,
      { "RL-Set-Successful-InformationItem-DM-Fail", "rnsap.RL_Set_Successful_InformationItem_DM_Fail_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DedicatedMeasurementReport_PDU,
      { "DedicatedMeasurementReport", "rnsap.DedicatedMeasurementReport_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DedicatedMeasurementObjectType_DM_Rprt_PDU,
      { "DedicatedMeasurementObjectType-DM-Rprt", "rnsap.DedicatedMeasurementObjectType_DM_Rprt",
        FT_UINT32, BASE_DEC, VALS(rnsap_DedicatedMeasurementObjectType_DM_Rprt_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationItem_DM_Rprt_PDU,
      { "RL-InformationItem-DM-Rprt", "rnsap.RL_InformationItem_DM_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_InformationItem_DM_Rprt_PDU,
      { "RL-Set-InformationItem-DM-Rprt", "rnsap.RL_Set_InformationItem_DM_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DedicatedMeasurementTerminationRequest_PDU,
      { "DedicatedMeasurementTerminationRequest", "rnsap.DedicatedMeasurementTerminationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DedicatedMeasurementFailureIndication_PDU,
      { "DedicatedMeasurementFailureIndication", "rnsap.DedicatedMeasurementFailureIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DedicatedMeasurementObjectType_DM_Fail_Ind_PDU,
      { "DedicatedMeasurementObjectType-DM-Fail-Ind", "rnsap.DedicatedMeasurementObjectType_DM_Fail_Ind",
        FT_UINT32, BASE_DEC, VALS(rnsap_DedicatedMeasurementObjectType_DM_Fail_Ind_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Unsuccessful_InformationItem_DM_Fail_Ind_PDU,
      { "RL-Unsuccessful-InformationItem-DM-Fail-Ind", "rnsap.RL_Unsuccessful_InformationItem_DM_Fail_Ind_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind_PDU,
      { "RL-Set-Unsuccessful-InformationItem-DM-Fail-Ind", "rnsap.RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonTransportChannelResourcesReleaseRequest_PDU,
      { "CommonTransportChannelResourcesReleaseRequest", "rnsap.CommonTransportChannelResourcesReleaseRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonTransportChannelResourcesRequest_PDU,
      { "CommonTransportChannelResourcesRequest", "rnsap.CommonTransportChannelResourcesRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonTransportChannelResourcesResponseFDD_PDU,
      { "CommonTransportChannelResourcesResponseFDD", "rnsap.CommonTransportChannelResourcesResponseFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD_PDU,
      { "FACH-InfoForUESelectedS-CCPCH-CTCH-ResourceRspFDD", "rnsap.FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonTransportChannelResourcesResponseTDD_PDU,
      { "CommonTransportChannelResourcesResponseTDD", "rnsap.CommonTransportChannelResourcesResponseTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD_PDU,
      { "FACH-InfoForUESelectedS-CCPCH-CTCH-ResourceRspTDD", "rnsap.FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonTransportChannelResourcesFailure_PDU,
      { "CommonTransportChannelResourcesFailure", "rnsap.CommonTransportChannelResourcesFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CompressedModeCommand_PDU,
      { "CompressedModeCommand", "rnsap.CompressedModeCommand_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ErrorIndication_PDU,
      { "ErrorIndication", "rnsap.ErrorIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonMeasurementInitiationRequest_PDU,
      { "CommonMeasurementInitiationRequest", "rnsap.CommonMeasurementInitiationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonMeasurementObjectType_CM_Rqst_PDU,
      { "CommonMeasurementObjectType-CM-Rqst", "rnsap.CommonMeasurementObjectType_CM_Rqst",
        FT_UINT32, BASE_DEC, VALS(rnsap_CommonMeasurementObjectType_CM_Rqst_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_GSM_Cell_CM_Rqst_PDU,
      { "GSM-Cell-CM-Rqst", "rnsap.GSM_Cell_CM_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Extension_CommonMeasurementObjectType_CM_Rqst_PDU,
      { "Extension-CommonMeasurementObjectType-CM-Rqst", "rnsap.Extension_CommonMeasurementObjectType_CM_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GsmCellList_CM_Rqst_PDU,
      { "GsmCellList-CM-Rqst", "rnsap.GsmCellList_CM_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonMeasurementInitiationResponse_PDU,
      { "CommonMeasurementInitiationResponse", "rnsap.CommonMeasurementInitiationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonMeasurementObjectType_CM_Rsp_PDU,
      { "CommonMeasurementObjectType-CM-Rsp", "rnsap.CommonMeasurementObjectType_CM_Rsp",
        FT_UINT32, BASE_DEC, VALS(rnsap_CommonMeasurementObjectType_CM_Rsp_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Extension_CommonMeasurementObjectType_CM_Rsp_PDU,
      { "Extension-CommonMeasurementObjectType-CM-Rsp", "rnsap.Extension_CommonMeasurementObjectType_CM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GsmCellList_CM_Rsp_PDU,
      { "GsmCellList-CM-Rsp", "rnsap.GsmCellList_CM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonMeasurementInitiationFailure_PDU,
      { "CommonMeasurementInitiationFailure", "rnsap.CommonMeasurementInitiationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Extension_FailureMeasurementList_PDU,
      { "Extension-FailureMeasurementList", "rnsap.Extension_FailureMeasurementList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonMeasurementReport_PDU,
      { "CommonMeasurementReport", "rnsap.CommonMeasurementReport_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonMeasurementObjectType_CM_Rprt_PDU,
      { "CommonMeasurementObjectType-CM-Rprt", "rnsap.CommonMeasurementObjectType_CM_Rprt",
        FT_UINT32, BASE_DEC, VALS(rnsap_CommonMeasurementObjectType_CM_Rprt_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Extension_CommonMeasurementObjectType_CM_Rprt_PDU,
      { "Extension-CommonMeasurementObjectType-CM-Rprt", "rnsap.Extension_CommonMeasurementObjectType_CM_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GsmCellList_CM_Rprt_PDU,
      { "GsmCellList-CM-Rprt", "rnsap.GsmCellList_CM_Rprt",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonMeasurementTerminationRequest_PDU,
      { "CommonMeasurementTerminationRequest", "rnsap.CommonMeasurementTerminationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Extension_TerminationMeasurementList_PDU,
      { "Extension-TerminationMeasurementList", "rnsap.Extension_TerminationMeasurementList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CommonMeasurementFailureIndication_PDU,
      { "CommonMeasurementFailureIndication", "rnsap.CommonMeasurementFailureIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Extension_FailureIndicationMeasurementList_PDU,
      { "Extension-FailureIndicationMeasurementList", "rnsap.Extension_FailureIndicationMeasurementList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_InformationExchangeInitiationRequest_PDU,
      { "InformationExchangeInitiationRequest", "rnsap.InformationExchangeInitiationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_InformationExchangeObjectType_InfEx_Rqst_PDU,
      { "InformationExchangeObjectType-InfEx-Rqst", "rnsap.InformationExchangeObjectType_InfEx_Rqst",
        FT_UINT32, BASE_DEC, VALS(rnsap_InformationExchangeObjectType_InfEx_Rqst_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_GSM_Cell_InfEx_Rqst_PDU,
      { "GSM-Cell-InfEx-Rqst", "rnsap.GSM_Cell_InfEx_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst_PDU,
      { "MBMS-Bearer-Service-in-MBMS-Cell-InfEx-Rqst", "rnsap.MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Cell_InfEx_Rqst_PDU,
      { "MBMS-Cell-InfEx-Rqst", "rnsap.MBMS_Cell_InfEx_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ANR_Cell_InfEx_Rqst_PDU,
      { "ANR-Cell-InfEx-Rqst", "rnsap.ANR_Cell_InfEx_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Common_E_RGCH_Cell_InfEx_Rqst_PDU,
      { "Common-E-RGCH-Cell-InfEx-Rqst", "rnsap.Common_E_RGCH_Cell_InfEx_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_InformationExchangeInitiationResponse_PDU,
      { "InformationExchangeInitiationResponse", "rnsap.InformationExchangeInitiationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_InformationExchangeObjectType_InfEx_Rsp_PDU,
      { "InformationExchangeObjectType-InfEx-Rsp", "rnsap.InformationExchangeObjectType_InfEx_Rsp",
        FT_UINT32, BASE_DEC, VALS(rnsap_InformationExchangeObjectType_InfEx_Rsp_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_List_InfEx_Rsp_PDU,
      { "MBMS-Bearer-Service-List-InfEx-Rsp", "rnsap.MBMS_Bearer_Service_List_InfEx_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp_PDU,
      { "MBMS-Bearer-Service-in-MBMS-Cell-InfEx-Rsp", "rnsap.MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Cell_InfEx_Rsp_PDU,
      { "MBMS-Cell-InfEx-Rsp", "rnsap.MBMS_Cell_InfEx_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ANR_Cell_InfEx_Rsp_PDU,
      { "ANR-Cell-InfEx-Rsp", "rnsap.ANR_Cell_InfEx_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Common_E_RGCH_Cell_InfEx_Rsp_PDU,
      { "Common-E-RGCH-Cell-InfEx-Rsp", "rnsap.Common_E_RGCH_Cell_InfEx_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_InformationExchangeInitiationFailure_PDU,
      { "InformationExchangeInitiationFailure", "rnsap.InformationExchangeInitiationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_InformationReport_PDU,
      { "InformationReport", "rnsap.InformationReport_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_InformationExchangeObjectType_InfEx_Rprt_PDU,
      { "InformationExchangeObjectType-InfEx-Rprt", "rnsap.InformationExchangeObjectType_InfEx_Rprt",
        FT_UINT32, BASE_DEC, VALS(rnsap_InformationExchangeObjectType_InfEx_Rprt_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt_PDU,
      { "MBMS-Bearer-Service-in-MBMS-Cell-InfEx-Rprt", "rnsap.MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Cell_InfEx_Rprt_PDU,
      { "MBMS-Cell-InfEx-Rprt", "rnsap.MBMS_Cell_InfEx_Rprt",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Common_E_RGCH_Cell_InfEx_Rprt_PDU,
      { "Common-E-RGCH-Cell-InfEx-Rprt", "rnsap.Common_E_RGCH_Cell_InfEx_Rprt",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_InformationExchangeTerminationRequest_PDU,
      { "InformationExchangeTerminationRequest", "rnsap.InformationExchangeTerminationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_InformationExchangeFailureIndication_PDU,
      { "InformationExchangeFailureIndication", "rnsap.InformationExchangeFailureIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ResetRequest_PDU,
      { "ResetRequest", "rnsap.ResetRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ResetIndicator_PDU,
      { "ResetIndicator", "rnsap.ResetIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_ResetIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_ContextInfoItem_Reset_PDU,
      { "ContextInfoItem-Reset", "rnsap.ContextInfoItem_Reset_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ContextGroupInfoItem_Reset_PDU,
      { "ContextGroupInfoItem-Reset", "rnsap.ContextGroupInfoItem_Reset_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ResetResponse_PDU,
      { "ResetResponse", "rnsap.ResetResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkActivationCommandFDD_PDU,
      { "RadioLinkActivationCommandFDD", "rnsap.RadioLinkActivationCommandFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DelayedActivationInformationList_RL_ActivationCmdFDD_PDU,
      { "DelayedActivationInformationList-RL-ActivationCmdFDD", "rnsap.DelayedActivationInformationList_RL_ActivationCmdFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DelayedActivationInformation_RL_ActivationCmdFDD_PDU,
      { "DelayedActivationInformation-RL-ActivationCmdFDD", "rnsap.DelayedActivationInformation_RL_ActivationCmdFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkActivationCommandTDD_PDU,
      { "RadioLinkActivationCommandTDD", "rnsap.RadioLinkActivationCommandTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DelayedActivationInformationList_RL_ActivationCmdTDD_PDU,
      { "DelayedActivationInformationList-RL-ActivationCmdTDD", "rnsap.DelayedActivationInformationList_RL_ActivationCmdTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DelayedActivationInformation_RL_ActivationCmdTDD_PDU,
      { "DelayedActivationInformation-RL-ActivationCmdTDD", "rnsap.DelayedActivationInformation_RL_ActivationCmdTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GERANUplinkSignallingTransferIndication_PDU,
      { "GERANUplinkSignallingTransferIndication", "rnsap.GERANUplinkSignallingTransferIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkParameterUpdateIndicationFDD_PDU,
      { "RadioLinkParameterUpdateIndicationFDD", "rnsap.RadioLinkParameterUpdateIndicationFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_ParameterUpdateIndicationFDD_RL_InformationList_PDU,
      { "RL-ParameterUpdateIndicationFDD-RL-InformationList", "rnsap.RL_ParameterUpdateIndicationFDD_RL_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_ParameterUpdateIndicationFDD_RL_Information_Item_PDU,
      { "RL-ParameterUpdateIndicationFDD-RL-Information-Item", "rnsap.RL_ParameterUpdateIndicationFDD_RL_Information_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UE_Measurement_Forwarding_PDU,
      { "UE-Measurement-Forwarding", "rnsap.UE_Measurement_Forwarding_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Information_RL_Param_Upd_PDU,
      { "Additional-HS-Cell-Information-RL-Param-Upd", "rnsap.Additional_HS_Cell_Information_RL_Param_Upd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd_PDU,
      { "Additional-EDCH-Cell-Information-RL-Param-Upd", "rnsap.Additional_EDCH_Cell_Information_RL_Param_Upd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RadioLinkParameterUpdateIndicationTDD_PDU,
      { "RadioLinkParameterUpdateIndicationTDD", "rnsap.RadioLinkParameterUpdateIndicationTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementInitiationRequest_PDU,
      { "UEMeasurementInitiationRequest", "rnsap.UEMeasurementInitiationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementInitiationResponse_PDU,
      { "UEMeasurementInitiationResponse", "rnsap.UEMeasurementInitiationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementInitiationFailure_PDU,
      { "UEMeasurementInitiationFailure", "rnsap.UEMeasurementInitiationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementReport_PDU,
      { "UEMeasurementReport", "rnsap.UEMeasurementReport_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementTerminationRequest_PDU,
      { "UEMeasurementTerminationRequest", "rnsap.UEMeasurementTerminationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementFailureIndication_PDU,
      { "UEMeasurementFailureIndication", "rnsap.UEMeasurementFailureIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_IurInvokeTrace_PDU,
      { "IurInvokeTrace", "rnsap.IurInvokeTrace_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ListOfInterfacesToTrace_PDU,
      { "ListOfInterfacesToTrace", "rnsap.ListOfInterfacesToTrace",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_InterfacesToTraceItem_PDU,
      { "InterfacesToTraceItem", "rnsap.InterfacesToTraceItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_IurDeactivateTrace_PDU,
      { "IurDeactivateTrace", "rnsap.IurDeactivateTrace_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMSAttachCommand_PDU,
      { "MBMSAttachCommand", "rnsap.MBMSAttachCommand_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMSDetachCommand_PDU,
      { "MBMSDetachCommand", "rnsap.MBMSDetachCommand_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DirectInformationTransfer_PDU,
      { "DirectInformationTransfer", "rnsap.DirectInformationTransfer_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EnhancedRelocationRequest_PDU,
      { "EnhancedRelocationRequest", "rnsap.EnhancedRelocationRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EnhancedRelocationResponse_PDU,
      { "EnhancedRelocationResponse", "rnsap.EnhancedRelocationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EnhancedRelocationFailure_PDU,
      { "EnhancedRelocationFailure", "rnsap.EnhancedRelocationFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EnhancedRelocationCancel_PDU,
      { "EnhancedRelocationCancel", "rnsap.EnhancedRelocationCancel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EnhancedRelocationSignallingTransfer_PDU,
      { "EnhancedRelocationSignallingTransfer", "rnsap.EnhancedRelocationSignallingTransfer_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EnhancedRelocationRelease_PDU,
      { "EnhancedRelocationRelease", "rnsap.EnhancedRelocationRelease_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBSFNMCCHInformation_PDU,
      { "MBSFNMCCHInformation", "rnsap.MBSFNMCCHInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SecondaryULFrequencyReport_PDU,
      { "SecondaryULFrequencyReport", "rnsap.SecondaryULFrequencyReport_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SecondaryULFrequencyUpdateIndication_PDU,
      { "SecondaryULFrequencyUpdateIndication", "rnsap.SecondaryULFrequencyUpdateIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EnhancedRelocationResourceRequest_PDU,
      { "EnhancedRelocationResourceRequest", "rnsap.EnhancedRelocationResourceRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EnhancedRelocationResourceResponse_PDU,
      { "EnhancedRelocationResourceResponse", "rnsap.EnhancedRelocationResourceResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EnhancedRelocationResourceFailure_PDU,
      { "EnhancedRelocationResourceFailure", "rnsap.EnhancedRelocationResourceFailure_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EnhancedRelocationResourceReleaseCommand_PDU,
      { "EnhancedRelocationResourceReleaseCommand", "rnsap.EnhancedRelocationResourceReleaseCommand_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EnhancedRelocationResourceReleaseComplete_PDU,
      { "EnhancedRelocationResourceReleaseComplete", "rnsap.EnhancedRelocationResourceReleaseComplete_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PrivateMessage_PDU,
      { "PrivateMessage", "rnsap.PrivateMessage_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_InformationTransferControlRequest_PDU,
      { "InformationTransferControlRequest", "rnsap.InformationTransferControlRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Control_Type_InformationTransferControlReq_PDU,
      { "Control-Type-InformationTransferControlReq", "rnsap.Control_Type_InformationTransferControlReq",
        FT_UINT32, BASE_DEC, VALS(rnsap_Control_Type_InformationTransferControlReq_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RNSAP_PDU_PDU,
      { "RNSAP-PDU", "rnsap.RNSAP_PDU",
        FT_UINT32, BASE_DEC, VALS(rnsap_RNSAP_PDU_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_NULL_PDU,
      { "NULL", "rnsap.NULL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_local,
      { "local", "rnsap.local",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_maxPrivateIEs", HFILL }},
    { &hf_rnsap_global,
      { "global", "rnsap.global",
        FT_OID, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_procedureCode,
      { "procedureCode", "rnsap.procedureCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_ProcedureCode_vals_ext, 0,
        NULL, HFILL }},
    { &hf_rnsap_ddMode,
      { "ddMode", "rnsap.ddMode",
        FT_UINT32, BASE_DEC, VALS(rnsap_DdMode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_shortTransActionId,
      { "shortTransActionId", "rnsap.shortTransActionId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_127", HFILL }},
    { &hf_rnsap_longTransActionId,
      { "longTransActionId", "rnsap.longTransActionId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_32767", HFILL }},
    { &hf_rnsap_ProtocolIE_Container_item,
      { "ProtocolIE-Field", "rnsap.ProtocolIE_Field_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_id,
      { "id", "rnsap.id",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_ProtocolIE_ID_vals_ext, 0,
        "ProtocolIE_ID", HFILL }},
    { &hf_rnsap_criticality,
      { "criticality", "rnsap.criticality",
        FT_UINT32, BASE_DEC, VALS(rnsap_Criticality_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_value,
      { "value", "rnsap.value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "ProtocolIE_Field_value", HFILL }},
    { &hf_rnsap_ProtocolExtensionContainer_item,
      { "ProtocolExtensionField", "rnsap.ProtocolExtensionField_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_extensionValue,
      { "extensionValue", "rnsap.extensionValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PrivateIE_Container_item,
      { "PrivateIE-Field", "rnsap.PrivateIE_Field_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_id_01,
      { "id", "rnsap.id",
        FT_UINT32, BASE_DEC, VALS(rnsap_PrivateIE_ID_vals), 0,
        "PrivateIE_ID", HFILL }},
    { &hf_rnsap_value_01,
      { "value", "rnsap.value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PrivateIE_Field_value", HFILL }},
    { &hf_rnsap_ActivationInformation_item,
      { "ActivationInformationItem", "rnsap.ActivationInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uU_ActivationState,
      { "uU-ActivationState", "rnsap.uU_ActivationState",
        FT_UINT32, BASE_DEC, VALS(rnsap_Uu_ActivationState_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_iE_Extensions,
      { "iE-Extensions", "rnsap.iE_Extensions",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ProtocolExtensionContainer", HFILL }},
    { &hf_rnsap_Active_MBMS_Bearer_Service_ListFDD_item,
      { "MBMS-Bearer-ServiceItemFDD", "rnsap.MBMS_Bearer_ServiceItemFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Active_MBMS_Bearer_Service_ListFDD_PFL_item,
      { "MBMS-Bearer-ServiceItemFDD-PFL", "rnsap.MBMS_Bearer_ServiceItemFDD_PFL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Active_MBMS_Bearer_Service_ListTDD_item,
      { "MBMS-Bearer-ServiceItemTDD", "rnsap.MBMS_Bearer_ServiceItemTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Active_MBMS_Bearer_Service_ListTDD_PFL_item,
      { "MBMS-Bearer-ServiceItemTDD-PFL", "rnsap.MBMS_Bearer_ServiceItemTDD_PFL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cMConfigurationChangeCFN,
      { "cMConfigurationChangeCFN", "rnsap.cMConfigurationChangeCFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CFN", HFILL }},
    { &hf_rnsap_transmission_Gap_Pattern_Sequence_Status,
      { "transmission-Gap-Pattern-Sequence-Status", "rnsap.transmission_Gap_Pattern_Sequence_Status",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Transmission_Gap_Pattern_Sequence_Status_List", HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_Response_RLAddList_item,
      { "Additional-EDCH-Cell-Information-Response-RLAdd-ItemIEs", "rnsap.Additional_EDCH_Cell_Information_Response_RLAdd_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_additional_EDCH_FDD_Information_Response,
      { "additional-EDCH-FDD-Information-Response", "rnsap.additional_EDCH_FDD_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Additional_EDCH_FDD_Information_Response_ItemIEs", HFILL }},
    { &hf_rnsap_additional_EDCH_Serving_Cell_Change_Information_Response_RLAdd,
      { "additional-EDCH-Serving-Cell-Change-Information-Response-RLAdd", "rnsap.additional_EDCH_Serving_Cell_Change_Information_Response_RLAdd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "E_DCH_Serving_cell_change_informationResponse", HFILL }},
    { &hf_rnsap_multicell_EDCH_Transport_Bearer_Mode,
      { "multicell-EDCH-Transport-Bearer-Mode", "rnsap.multicell_EDCH_Transport_Bearer_Mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_Multicell_EDCH_Transport_Bearer_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_additional_EDCH_Cell_Information_Setup,
      { "additional-EDCH-Cell-Information-Setup", "rnsap.additional_EDCH_Cell_Information_Setup",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_pLMN_Identity,
      { "pLMN-Identity", "rnsap.pLMN_Identity",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_l3_Information,
      { "l3-Information", "rnsap.l3_Information",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_Setup_item,
      { "Additional-EDCH-FDD-Setup-Cell-Information", "rnsap.Additional_EDCH_FDD_Setup_Cell_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_additional_EDCH_UL_DPCH_Information_Setup,
      { "additional-EDCH-UL-DPCH-Information-Setup", "rnsap.additional_EDCH_UL_DPCH_Information_Setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_additional_EDCH_RL_Specific_Information_To_Setup,
      { "additional-EDCH-RL-Specific-Information-To-Setup", "rnsap.additional_EDCH_RL_Specific_Information_To_Setup",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_RL_Specific_Information_To_Setup_List", HFILL }},
    { &hf_rnsap_additional_EDCH_FDD_Information,
      { "additional-EDCH-FDD-Information", "rnsap.additional_EDCH_FDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_additional_EDCH_F_DPCH_Information_Setup,
      { "additional-EDCH-F-DPCH-Information-Setup", "rnsap.additional_EDCH_F_DPCH_Information_Setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Additional_EDCH_F_DPCH_Information", HFILL }},
    { &hf_rnsap_multicellEDCH_Information,
      { "multicellEDCH-Information", "rnsap.multicellEDCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_ScramblingCode,
      { "ul-ScramblingCode", "rnsap.ul_ScramblingCode_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_SIR_Target,
      { "ul-SIR-Target", "rnsap.ul_SIR_Target",
        FT_INT32, BASE_DEC, NULL, 0,
        "UL_SIR", HFILL }},
    { &hf_rnsap_fdd_TPC_DownlinkStepSize,
      { "fdd-TPC-DownlinkStepSize", "rnsap.fdd_TPC_DownlinkStepSize",
        FT_UINT32, BASE_DEC, VALS(rnsap_FDD_TPC_DownlinkStepSize_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_limitedPowerIncrease,
      { "limitedPowerIncrease", "rnsap.limitedPowerIncrease",
        FT_UINT32, BASE_DEC, VALS(rnsap_LimitedPowerIncrease_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_innerLoopDLPCStatus,
      { "innerLoopDLPCStatus", "rnsap.innerLoopDLPCStatus",
        FT_UINT32, BASE_DEC, VALS(rnsap_InnerLoopDLPCStatus_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_f_DPCH_SlotFormatSupportRequest,
      { "f-DPCH-SlotFormatSupportRequest", "rnsap.f_DPCH_SlotFormatSupportRequest_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_RL_Specific_Information_To_Setup_List_item,
      { "Additional-EDCH-RL-Specific-Information-To-Setup-ItemIEs", "rnsap.Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eDCH_Additional_RL_ID,
      { "eDCH-Additional-RL-ID", "rnsap.eDCH_Additional_RL_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RL_ID", HFILL }},
    { &hf_rnsap_c_ID,
      { "c-ID", "rnsap.c_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_firstRLS_indicator,
      { "firstRLS-indicator", "rnsap.firstRLS_indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_FirstRLS_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_propagationDelay,
      { "propagationDelay", "rnsap.propagationDelay",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_initialDL_transmissionPower,
      { "initialDL-transmissionPower", "rnsap.initialDL_transmissionPower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_primaryCPICH_EcNo,
      { "primaryCPICH-EcNo", "rnsap.primaryCPICH_EcNo",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_AGCH_PowerOffset,
      { "e-AGCH-PowerOffset", "rnsap.e_AGCH_PowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_RGCH_PowerOffset,
      { "e-RGCH-PowerOffset", "rnsap.e_RGCH_PowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_HICH_PowerOffset,
      { "e-HICH-PowerOffset", "rnsap.e_HICH_PowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_additional_EDCH_MAC_d_Flows_Specific_Info_List,
      { "additional-EDCH-MAC-d-Flows-Specific-Info-List", "rnsap.additional_EDCH_MAC_d_Flows_Specific_Info_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_multicellEDCH_RL_SpecificInformation,
      { "multicellEDCH-RL-SpecificInformation", "rnsap.multicellEDCH_RL_SpecificInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_To_Add_List_item,
      { "Additional-EDCH-Cell-Information-To-Add-ItemIEs", "rnsap.Additional_EDCH_Cell_Information_To_Add_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_additional_EDCH_UL_DPCH_Information_Setup_01,
      { "additional-EDCH-UL-DPCH-Information-Setup", "rnsap.additional_EDCH_UL_DPCH_Information_Setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Additional_EDCH_UL_DPCH_Information_Addition", HFILL }},
    { &hf_rnsap_additional_EDCH_RL_Specific_Information_To_Add_List,
      { "additional-EDCH-RL-Specific-Information-To-Add-List", "rnsap.additional_EDCH_RL_Specific_Information_To_Add_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_List_item,
      { "Additional-EDCH-RL-Specific-Information-To-Add-ItemIEs", "rnsap.Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_RL_Specific_Information_To_Modify_List_item,
      { "Additional-EDCH-RL-Specific-Information-To-Modify-ItemIEs", "rnsap.Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hARQ_Process_Allocation_Scheduled_2ms_EDCH,
      { "hARQ-Process-Allocation-Scheduled-2ms-EDCH", "rnsap.hARQ_Process_Allocation_Scheduled_2ms_EDCH",
        FT_BYTES, BASE_NONE, NULL, 0,
        "HARQ_Process_Allocation_2ms_EDCH", HFILL }},
    { &hf_rnsap_e_DCH_Maximum_Bitrate,
      { "e-DCH-Maximum-Bitrate", "rnsap.e_DCH_Maximum_Bitrate",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_Processing_Overload_Level,
      { "e-DCH-Processing-Overload-Level", "rnsap.e_DCH_Processing_Overload_Level",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_Min_Set_E_TFCI,
      { "e-DCH-Min-Set-E-TFCI", "rnsap.e_DCH_Min_Set_E_TFCI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_TFCI", HFILL }},
    { &hf_rnsap_uE_DTX_Cycle1,
      { "uE-DTX-Cycle1", "rnsap.uE_DTX_Cycle1",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_DTX_Cycle1_2ms_vals), 0,
        "UE_DTX_Cycle1_2ms", HFILL }},
    { &hf_rnsap_uE_DTX_Cycle2,
      { "uE-DTX-Cycle2", "rnsap.uE_DTX_Cycle2",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_DTX_Cycle2_ext_2ms_vals), 0,
        "UE_DTX_Cycle2_ext_2ms", HFILL }},
    { &hf_rnsap_inactivity_Threshold_for_UE_DTX_Cycle2,
      { "inactivity-Threshold-for-UE-DTX-Cycle2", "rnsap.inactivity_Threshold_for_UE_DTX_Cycle2",
        FT_UINT32, BASE_DEC, VALS(rnsap_Inactivity_Threshold_for_UE_DTX_Cycle2_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_List_item,
      { "Additional-EDCH-MAC-d-Flows-Specific-Info", "rnsap.Additional_EDCH_MAC_d_Flows_Specific_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_MACdFlow_ID,
      { "e-DCH-MACdFlow-ID", "rnsap.e_DCH_MACdFlow_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EDCH_MACdFlow_ID", HFILL }},
    { &hf_rnsap_bindingID,
      { "bindingID", "rnsap.bindingID",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_transportLayerAddress,
      { "transportLayerAddress", "rnsap.transportLayerAddress",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List_item,
      { "Additional-EDCH-MAC-d-Flows-Specific-Info-Response", "rnsap.Additional_EDCH_MAC_d_Flows_Specific_Info_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_Response_List_item,
      { "Additional-EDCH-FDD-Information-Response-ItemIEs", "rnsap.Additional_EDCH_FDD_Information_Response_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eDCH_Additional_RL_Specific_Information_Response,
      { "eDCH-Additional-RL-Specific-Information-Response", "rnsap.eDCH_Additional_RL_Specific_Information_Response",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EDCH_Additional_RL_Specific_Information_Response_List", HFILL }},
    { &hf_rnsap_EDCH_Additional_RL_Specific_Information_Response_List_item,
      { "EDCH-Additional-RL-Specific-Information-Response-ItemIEs", "rnsap.EDCH_Additional_RL_Specific_Information_Response_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_received_total_wide_band_power,
      { "received-total-wide-band-power", "rnsap.received_total_wide_band_power",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dL_PowerBalancing_ActivationIndicator,
      { "dL-PowerBalancing-ActivationIndicator", "rnsap.dL_PowerBalancing_ActivationIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_DL_PowerBalancing_ActivationIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_Set_ID,
      { "rL-Set-ID", "rnsap.rL_Set_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_RL_Set_ID,
      { "e-DCH-RL-Set-ID", "rnsap.e_DCH_RL_Set_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RL_Set_ID", HFILL }},
    { &hf_rnsap_eDCH_FDD_DL_ControlChannelInformation,
      { "eDCH-FDD-DL-ControlChannelInformation", "rnsap.eDCH_FDD_DL_ControlChannelInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_CodeInformation,
      { "dl-CodeInformation", "rnsap.dl_CodeInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FDD_DL_CodeInformation", HFILL }},
    { &hf_rnsap_additional_EDCH_MAC_d_Flow_Specific_Information_Response_List,
      { "additional-EDCH-MAC-d-Flow-Specific-Information-Response-List", "rnsap.additional_EDCH_MAC_d_Flow_Specific_Information_Response_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_maxUL_SIR,
      { "maxUL-SIR", "rnsap.maxUL_SIR",
        FT_INT32, BASE_DEC, NULL, 0,
        "UL_SIR", HFILL }},
    { &hf_rnsap_minUL_SIR,
      { "minUL-SIR", "rnsap.minUL_SIR",
        FT_INT32, BASE_DEC, NULL, 0,
        "UL_SIR", HFILL }},
    { &hf_rnsap_maximumAllowedULTxPower,
      { "maximumAllowedULTxPower", "rnsap.maximumAllowedULTxPower",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_maximumDL_power,
      { "maximumDL-power", "rnsap.maximumDL_power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_minimumDL_power,
      { "minimumDL-power", "rnsap.minimumDL_power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_primaryScramblingCode,
      { "primaryScramblingCode", "rnsap.primaryScramblingCode",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_UARFCN,
      { "uL-UARFCN", "rnsap.uL_UARFCN",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UARFCN", HFILL }},
    { &hf_rnsap_dL_UARFCN,
      { "dL-UARFCN", "rnsap.dL_UARFCN",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UARFCN", HFILL }},
    { &hf_rnsap_primaryCPICH_Power,
      { "primaryCPICH-Power", "rnsap.primaryCPICH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_pC_Preamble,
      { "pC-Preamble", "rnsap.pC_Preamble",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_primary_CPICH_Usage_For_Channel_Estimation,
      { "primary-CPICH-Usage-For-Channel-Estimation", "rnsap.primary_CPICH_Usage_For_Channel_Estimation",
        FT_UINT32, BASE_DEC, VALS(rnsap_Primary_CPICH_Usage_For_Channel_Estimation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_secondary_CPICH_Information,
      { "secondary-CPICH-Information", "rnsap.secondary_CPICH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_f_DPCH_SlotFormat,
      { "f-DPCH-SlotFormat", "rnsap.f_DPCH_SlotFormat",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_Response_RLReconf_List_item,
      { "Additional-EDCH-FDD-Information-Response-RLReconf-Items", "rnsap.Additional_EDCH_FDD_Information_Response_RLReconf_Items_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eDCH_Additional_RL_Specific_Modified_Information_Response,
      { "eDCH-Additional-RL-Specific-Modified-Information-Response", "rnsap.eDCH_Additional_RL_Specific_Modified_Information_Response",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EDCH_Additional_RL_Specific_Modified_Information_Response_List", HFILL }},
    { &hf_rnsap_EDCH_Additional_RL_Specific_Modified_Information_Response_List_item,
      { "EDCH-Additional-RL-Specific-Modified-Information-Response-ItemIEs", "rnsap.EDCH_Additional_RL_Specific_Modified_Information_Response_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dL_PowerBalancing_UpdatedIndicator,
      { "dL-PowerBalancing-UpdatedIndicator", "rnsap.dL_PowerBalancing_UpdatedIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_DL_PowerBalancing_UpdatedIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_secondary_CPICH_Information_Change,
      { "secondary-CPICH-Information-Change", "rnsap.secondary_CPICH_Information_Change",
        FT_UINT32, BASE_DEC, VALS(rnsap_Secondary_CPICH_Information_Change_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_ConfigurationChange_List_item,
      { "Additional-EDCH-ConfigurationChange-Info-ItemIEs", "rnsap.Additional_EDCH_ConfigurationChange_Info_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_additional_EDCH_UL_DPCH_Information_Modify,
      { "additional-EDCH-UL-DPCH-Information-Modify", "rnsap.additional_EDCH_UL_DPCH_Information_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_additional_EDCH_RL_Specific_Information_To_Add,
      { "additional-EDCH-RL-Specific-Information-To-Add", "rnsap.additional_EDCH_RL_Specific_Information_To_Add",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_RL_Specific_Information_To_Add_List", HFILL }},
    { &hf_rnsap_additional_EDCH_RL_Specific_Information_To_Modify,
      { "additional-EDCH-RL-Specific-Information-To-Modify", "rnsap.additional_EDCH_RL_Specific_Information_To_Modify",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_RL_Specific_Information_To_Modify_List", HFILL }},
    { &hf_rnsap_additional_EDCH_FDD_Information_To_Modify,
      { "additional-EDCH-FDD-Information-To-Modify", "rnsap.additional_EDCH_FDD_Information_To_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Additional_EDCH_FDD_Information", HFILL }},
    { &hf_rnsap_additional_EDCH_F_DPCH_Information_Modify,
      { "additional-EDCH-F-DPCH-Information-Modify", "rnsap.additional_EDCH_F_DPCH_Information_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Additional_EDCH_F_DPCH_Information", HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_Removal_List_item,
      { "Additional-EDCH-Cell-Information-Removal-Info-ItemIEs", "rnsap.Additional_EDCH_Cell_Information_Removal_Info_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_on_Secondary_UL_Frequency,
      { "rL-on-Secondary-UL-Frequency", "rnsap.rL_on_Secondary_UL_Frequency",
        FT_UINT32, BASE_DEC, VALS(rnsap_RL_on_Secondary_UL_Frequency_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_additional_EDCH_DL_Control_Channel_Change_Information,
      { "additional-EDCH-DL-Control-Channel-Change-Information", "rnsap.additional_EDCH_DL_Control_Channel_Change_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_DL_Control_Channel_Change_Information_List", HFILL }},
    { &hf_rnsap_Additional_EDCH_DL_Control_Channel_Change_Information_List_item,
      { "Additional-EDCH-DL-Control-Channel-Change-Info-ItemIEs", "rnsap.Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_AdditionalPreferredFrequency_item,
      { "AdditionalPreferredFrequencyItem", "rnsap.AdditionalPreferredFrequencyItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_correspondingCells,
      { "correspondingCells", "rnsap.correspondingCells",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_AffectedUEInformationForMBMS_item,
      { "S-RNTI", "rnsap.S_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_priorityLevel,
      { "priorityLevel", "rnsap.priorityLevel",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_pre_emptionCapability,
      { "pre-emptionCapability", "rnsap.pre_emptionCapability",
        FT_UINT32, BASE_DEC, VALS(rnsap_Pre_emptionCapability_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_pre_emptionVulnerability,
      { "pre-emptionVulnerability", "rnsap.pre_emptionVulnerability",
        FT_UINT32, BASE_DEC, VALS(rnsap_Pre_emptionVulnerability_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_allowed_UL_Rate,
      { "allowed-UL-Rate", "rnsap.allowed_UL_Rate",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Allowed_Rate", HFILL }},
    { &hf_rnsap_allowed_DL_Rate,
      { "allowed-DL-Rate", "rnsap.allowed_DL_Rate",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Allowed_Rate", HFILL }},
    { &hf_rnsap_aOA_LCR,
      { "aOA-LCR", "rnsap.aOA_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_aOA_LCR_Accuracy_Class,
      { "aOA-LCR-Accuracy-Class", "rnsap.aOA_LCR_Accuracy_Class",
        FT_UINT32, BASE_DEC, VALS(rnsap_AOA_LCR_Accuracy_Class_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_rNC_ID,
      { "rNC-ID", "rnsap.rNC_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_lAC,
      { "lAC", "rnsap.lAC",
        FT_UINT16, BASE_DEC_HEX, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rAC,
      { "rAC", "rnsap.rAC",
        FT_UINT8, BASE_DEC_HEX, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_aNR_FDD_CellInformation,
      { "aNR-FDD-CellInformation", "rnsap.aNR_FDD_CellInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_aNR_TDD_CellInformation,
      { "aNR-TDD-CellInformation", "rnsap.aNR_TDD_CellInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_aNR_LCR_TDD_CellInformation,
      { "aNR-LCR-TDD-CellInformation", "rnsap.aNR_LCR_TDD_CellInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_extended_RNC_ID,
      { "extended-RNC-ID", "rnsap.extended_RNC_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_txDiversityIndicator,
      { "txDiversityIndicator", "rnsap.txDiversityIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_TxDiversityIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_sTTD_SupportIndicator,
      { "sTTD-SupportIndicator", "rnsap.sTTD_SupportIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_STTD_SupportIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_closedLoopMode1_SupportIndicator,
      { "closedLoopMode1-SupportIndicator", "rnsap.closedLoopMode1_SupportIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_ClosedLoopMode1_SupportIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_restrictionStateIndicator,
      { "restrictionStateIndicator", "rnsap.restrictionStateIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_RestrictionStateIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dpc_ModeChangeSupportIndicator,
      { "dpc-ModeChangeSupportIndicator", "rnsap.dpc_ModeChangeSupportIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_DPC_Mode_Change_SupportIndicator_vals), 0,
        "DPC_Mode_Change_SupportIndicator", HFILL }},
    { &hf_rnsap_cellCapabilityContainerFDD,
      { "cellCapabilityContainerFDD", "rnsap.cellCapabilityContainerFDD",
        FT_BYTES, BASE_NONE, NULL, 0,
        "CellCapabilityContainer_FDD", HFILL }},
    { &hf_rnsap_sNA_Information,
      { "sNA-Information", "rnsap.sNA_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_frequencyBandIndicator,
      { "frequencyBandIndicator", "rnsap.frequencyBandIndicator",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_FrequencyBandIndicator_vals_ext, 0,
        NULL, HFILL }},
    { &hf_rnsap_max_UE_DTX_Cycle,
      { "max-UE-DTX-Cycle", "rnsap.max_UE_DTX_Cycle",
        FT_UINT32, BASE_DEC, VALS(rnsap_Max_UE_DTX_Cycle_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_aNR_Multiple_PLMN_List,
      { "aNR-Multiple-PLMN-List", "rnsap.aNR_Multiple_PLMN_List_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_secondaryServingCellList,
      { "secondaryServingCellList", "rnsap.secondaryServingCellList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Secondary_Serving_Cell_List", HFILL }},
    { &hf_rnsap_dualBandSecondaryServingCellList,
      { "dualBandSecondaryServingCellList", "rnsap.dualBandSecondaryServingCellList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Secondary_Serving_Cell_List", HFILL }},
    { &hf_rnsap_cellCapabilityContainerExtensionFDD,
      { "cellCapabilityContainerExtensionFDD", "rnsap.cellCapabilityContainerExtensionFDD",
        FT_BYTES, BASE_NONE, NULL, 0,
        "CellCapabilityContainerExtension_FDD", HFILL }},
    { &hf_rnsap_syncCase,
      { "syncCase", "rnsap.syncCase",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_timeSlot,
      { "timeSlot", "rnsap.timeSlot",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sCH_TimeSlot,
      { "sCH-TimeSlot", "rnsap.sCH_TimeSlot",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sCTD_Indicator,
      { "sCTD-Indicator", "rnsap.sCTD_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SCTD_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dPCHConstantValue,
      { "dPCHConstantValue", "rnsap.dPCHConstantValue",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_pCCPCH_Power,
      { "pCCPCH-Power", "rnsap.pCCPCH_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cellCapabilityContainerTDD,
      { "cellCapabilityContainerTDD", "rnsap.cellCapabilityContainerTDD",
        FT_BYTES, BASE_NONE, NULL, 0,
        "CellCapabilityContainer_TDD", HFILL }},
    { &hf_rnsap_cellCapabilityContainerTDD768,
      { "cellCapabilityContainerTDD768", "rnsap.cellCapabilityContainerTDD768",
        FT_BYTES, BASE_NONE, NULL, 0,
        "CellCapabilityContainer_TDD768", HFILL }},
    { &hf_rnsap_cellCapabilityContainerTDD_LCR,
      { "cellCapabilityContainerTDD-LCR", "rnsap.cellCapabilityContainerTDD_LCR",
        FT_BYTES, BASE_NONE, NULL, 0,
        "CellCapabilityContainer_TDD_LCR", HFILL }},
    { &hf_rnsap_cellCapabilityContainerExternsionTDD_LCR,
      { "cellCapabilityContainerExternsionTDD-LCR", "rnsap.cellCapabilityContainerExternsionTDD_LCR",
        FT_BYTES, BASE_NONE, NULL, 0,
        "CellCapabilityContainerExtension_TDD_LCR", HFILL }},
    { &hf_rnsap_list_Of_PLMNs,
      { "list-Of-PLMNs", "rnsap.list_Of_PLMNs",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_badSatelliteInformation,
      { "badSatelliteInformation", "rnsap.badSatelliteInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_badSatelliteInformation_item,
      { "badSatelliteInformation item", "rnsap.badSatelliteInformation_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_badSAT_ID,
      { "badSAT-ID", "rnsap.badSAT_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SAT_ID", HFILL }},
    { &hf_rnsap_BDS_IGPInfoList_item,
      { "BDS-IGPInfo", "rnsap.BDS_IGPInfo_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_bds_IGPNumber,
      { "bds-IGPNumber", "rnsap.bds_IGPNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_320", HFILL }},
    { &hf_rnsap_bds_VerticalDelay,
      { "bds-VerticalDelay", "rnsap.bds_VerticalDelay",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_9", HFILL }},
    { &hf_rnsap_bds_GIVEI,
      { "bds-GIVEI", "rnsap.bds_GIVEI",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_rnsap_bDS_RefTime,
      { "bDS-RefTime", "rnsap.bDS_RefTime",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_119", HFILL }},
    { &hf_rnsap_bds_IGPInfoList,
      { "bds-IGPInfoList", "rnsap.bds_IGPInfoList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_nCC,
      { "nCC", "rnsap.nCC",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_bCC,
      { "bCC", "rnsap.bCC",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_burstStart,
      { "burstStart", "rnsap.burstStart",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_15", HFILL }},
    { &hf_rnsap_burstLength,
      { "burstLength", "rnsap.burstLength",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_10_25", HFILL }},
    { &hf_rnsap_burstFreq,
      { "burstFreq", "rnsap.burstFreq",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_16", HFILL }},
    { &hf_rnsap_radioNetwork,
      { "radioNetwork", "rnsap.radioNetwork",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_CauseRadioNetwork_vals_ext, 0,
        "CauseRadioNetwork", HFILL }},
    { &hf_rnsap_transport,
      { "transport", "rnsap.transport",
        FT_UINT32, BASE_DEC, VALS(rnsap_CauseTransport_vals), 0,
        "CauseTransport", HFILL }},
    { &hf_rnsap_protocol,
      { "protocol", "rnsap.protocol",
        FT_UINT32, BASE_DEC, VALS(rnsap_CauseProtocol_vals), 0,
        "CauseProtocol", HFILL }},
    { &hf_rnsap_misc,
      { "misc", "rnsap.misc",
        FT_UINT32, BASE_DEC, VALS(rnsap_CauseMisc_vals), 0,
        "CauseMisc", HFILL }},
    { &hf_rnsap_cellIdList,
      { "cellIdList", "rnsap.cellIdList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CellIdList_item,
      { "Cell-Id", "rnsap.Cell_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uplinkCellCapacityClassValue,
      { "uplinkCellCapacityClassValue", "rnsap.uplinkCellCapacityClassValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_100_", HFILL }},
    { &hf_rnsap_downlinkCellCapacityClassValue,
      { "downlinkCellCapacityClassValue", "rnsap.downlinkCellCapacityClassValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_100_", HFILL }},
    { &hf_rnsap_lAI,
      { "lAI", "rnsap.lAI_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cI,
      { "cI", "rnsap.cI",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tUTRANGPSMeasurementAccuracyClass,
      { "tUTRANGPSMeasurementAccuracyClass", "rnsap.tUTRANGPSMeasurementAccuracyClass",
        FT_UINT32, BASE_DEC, VALS(rnsap_TUTRANGPSAccuracyClass_vals), 0,
        "TUTRANGPSAccuracyClass", HFILL }},
    { &hf_rnsap_tUTRANGANSSMeasurementAccuracyClass,
      { "tUTRANGANSSMeasurementAccuracyClass", "rnsap.tUTRANGANSSMeasurementAccuracyClass",
        FT_UINT32, BASE_DEC, VALS(rnsap_TUTRANGANSSAccuracyClass_vals), 0,
        "TUTRANGANSSAccuracyClass", HFILL }},
    { &hf_rnsap_tUTRANGPSMeasurementValueInformation,
      { "tUTRANGPSMeasurementValueInformation", "rnsap.tUTRANGPSMeasurementValueInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sFNSFNMeasurementValueInformation,
      { "sFNSFNMeasurementValueInformation", "rnsap.sFNSFNMeasurementValueInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_loadValue,
      { "loadValue", "rnsap.loadValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_transmittedCarrierPowerValue,
      { "transmittedCarrierPowerValue", "rnsap.transmittedCarrierPowerValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_100", HFILL }},
    { &hf_rnsap_receivedTotalWideBandPowerValue,
      { "receivedTotalWideBandPowerValue", "rnsap.receivedTotalWideBandPowerValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_621", HFILL }},
    { &hf_rnsap_uplinkTimeslotISCPValue,
      { "uplinkTimeslotISCPValue", "rnsap.uplinkTimeslotISCPValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UL_TimeslotISCP", HFILL }},
    { &hf_rnsap_extension_CommonMeasurementValue,
      { "extension-CommonMeasurementValue", "rnsap.extension_CommonMeasurementValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_measurementAvailable,
      { "measurementAvailable", "rnsap.measurementAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "CommonMeasurementAvailable", HFILL }},
    { &hf_rnsap_measurementnotAvailable,
      { "measurementnotAvailable", "rnsap.measurementnotAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_commonMeasurementValue,
      { "commonMeasurementValue", "rnsap.commonMeasurementValue",
        FT_UINT32, BASE_DEC, VALS(rnsap_CommonMeasurementValue_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDD_item,
      { "Common-EDCH-MAC-d-Flow-Specific-InformationFDDItem", "rnsap.Common_EDCH_MAC_d_Flow_Specific_InformationFDDItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_common_EDCH_MACdFlow_ID,
      { "common-EDCH-MACdFlow-ID", "rnsap.common_EDCH_MACdFlow_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EDCH_MACdFlow_ID", HFILL }},
    { &hf_rnsap_maximum_Number_of_Retransmissions_For_E_DCH,
      { "maximum-Number-of-Retransmissions-For-E-DCH", "rnsap.maximum_Number_of_Retransmissions_For_E_DCH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MaxNr_Retransmissions_EDCH", HFILL }},
    { &hf_rnsap_eDCH_HARQ_PO_FDD,
      { "eDCH-HARQ-PO-FDD", "rnsap.eDCH_HARQ_PO_FDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_HARQ_PO_FDD", HFILL }},
    { &hf_rnsap_eDCH_MACdFlow_Multiplexing_List,
      { "eDCH-MACdFlow-Multiplexing-List", "rnsap.eDCH_MACdFlow_Multiplexing_List",
        FT_BYTES, BASE_NONE, NULL, 0,
        "E_DCH_MACdFlow_Multiplexing_List", HFILL }},
    { &hf_rnsap_common_E_DCHLogicalChannelInformation,
      { "common-E-DCHLogicalChannelInformation", "rnsap.common_E_DCHLogicalChannelInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Common_E_DCH_LogicalChannelInformation", HFILL }},
    { &hf_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationLCR_item,
      { "Common-EDCH-MAC-d-Flow-Specific-InformationItemLCR", "rnsap.Common_EDCH_MAC_d_Flow_Specific_InformationItemLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_common_EDCH_MACdFlow_ID_LCR,
      { "common-EDCH-MACdFlow-ID-LCR", "rnsap.common_EDCH_MACdFlow_ID_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EDCH_MACdFlow_ID_LCR", HFILL }},
    { &hf_rnsap_eDCH_HARQ_PO_TDD,
      { "eDCH-HARQ-PO-TDD", "rnsap.eDCH_HARQ_PO_TDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_HARQ_PO_TDD", HFILL }},
    { &hf_rnsap_Common_E_DCH_LogicalChannelInformation_item,
      { "Common-E-DCH-LogicalChannelInformationItem", "rnsap.Common_E_DCH_LogicalChannelInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_logicalChannelId,
      { "logicalChannelId", "rnsap.logicalChannelId",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_maximumMACdPDU_SizeExtended,
      { "maximumMACdPDU-SizeExtended", "rnsap.maximumMACdPDU_SizeExtended",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MAC_PDU_SizeExtended", HFILL }},
    { &hf_rnsap_deActivation,
      { "deActivation", "rnsap.deActivation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Common_E_RGCH_De_Activation", HFILL }},
    { &hf_rnsap_common_E_RGCH_Cell_Info,
      { "common-E-RGCH-Cell-Info", "rnsap.common_E_RGCH_Cell_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eRGCH_ChannelisationCode,
      { "eRGCH-ChannelisationCode", "rnsap.eRGCH_ChannelisationCode",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FDD_DL_ChannelisationCodeNumber", HFILL }},
    { &hf_rnsap_eRGCH_SignatureSequence,
      { "eRGCH-SignatureSequence", "rnsap.eRGCH_SignatureSequence",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_serving_Grant_Value,
      { "serving-Grant-Value", "rnsap.serving_Grant_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_Serving_Grant_Value", HFILL }},
    { &hf_rnsap_uE_DTX_DRX_Offset,
      { "uE-DTX-DRX-Offset", "rnsap.uE_DTX_DRX_Offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_enabling_Delay,
      { "enabling-Delay", "rnsap.enabling_Delay",
        FT_UINT32, BASE_DEC, VALS(rnsap_Enabling_Delay_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dTX_Information,
      { "dTX-Information", "rnsap.dTX_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dRX_Information,
      { "dRX-Information", "rnsap.dRX_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dTX_Information_to_Modify,
      { "dTX-Information-to-Modify", "rnsap.dTX_Information_to_Modify",
        FT_UINT32, BASE_DEC, VALS(rnsap_DTX_Information_to_Modify_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dRX_Information_to_Modify,
      { "dRX-Information-to-Modify", "rnsap.dRX_Information_to_Modify",
        FT_UINT32, BASE_DEC, VALS(rnsap_DRX_Information_to_Modify_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_item,
      { "Continuous-Packet-Connectivity-HS-SCCH-Less-InformationItem", "rnsap.Continuous_Packet_Connectivity_HS_SCCH_Less_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_transport_Block_Size_Index,
      { "transport-Block-Size-Index", "rnsap.transport_Block_Size_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSPDSCH_Second_Code_Support,
      { "hSPDSCH-Second-Code-Support", "rnsap.hSPDSCH_Second_Code_Support",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSPDSCH_First_Code_Index,
      { "hSPDSCH-First-Code-Index", "rnsap.hSPDSCH_First_Code_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSPDSCH_Second_Code_Index,
      { "hSPDSCH-Second-Code-Index", "rnsap.hSPDSCH_Second_Code_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CorrespondingCells_item,
      { "C-ID", "rnsap.C_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_continuous_Packet_Connectivity_DTX_DRX_Information,
      { "continuous-Packet-Connectivity-DTX-DRX-Information", "rnsap.continuous_Packet_Connectivity_DTX_DRX_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_continuous_Packet_Connectivity_DTX_DRX_Information_to_Modify,
      { "continuous-Packet-Connectivity-DTX-DRX-Information-to-Modify", "rnsap.continuous_Packet_Connectivity_DTX_DRX_Information_to_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_continuous_Packet_Connectivity_HS_SCCH_Less_Information,
      { "continuous-Packet-Connectivity-HS-SCCH-Less-Information", "rnsap.continuous_Packet_Connectivity_HS_SCCH_Less_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Counting_Information_item,
      { "Counting-Information-List", "rnsap.Counting_Information_List_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_counting_Result,
      { "counting-Result", "rnsap.counting_Result",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_procedureID,
      { "procedureID", "rnsap.procedureID_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_triggeringMessage,
      { "triggeringMessage", "rnsap.triggeringMessage",
        FT_UINT32, BASE_DEC, VALS(rnsap_TriggeringMessage_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_procedureCriticality,
      { "procedureCriticality", "rnsap.procedureCriticality",
        FT_UINT32, BASE_DEC, VALS(rnsap_Criticality_vals), 0,
        "Criticality", HFILL }},
    { &hf_rnsap_transactionID,
      { "transactionID", "rnsap.transactionID",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransactionID_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_iEsCriticalityDiagnostics,
      { "iEsCriticalityDiagnostics", "rnsap.iEsCriticalityDiagnostics",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CriticalityDiagnostics_IE_List", HFILL }},
    { &hf_rnsap_CriticalityDiagnostics_IE_List_item,
      { "CriticalityDiagnostics-IE-List item", "rnsap.CriticalityDiagnostics_IE_List_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_iECriticality,
      { "iECriticality", "rnsap.iECriticality",
        FT_UINT32, BASE_DEC, VALS(rnsap_Criticality_vals), 0,
        "Criticality", HFILL }},
    { &hf_rnsap_iE_ID,
      { "iE-ID", "rnsap.iE_ID",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_ProtocolIE_ID_vals_ext, 0,
        "ProtocolIE_ID", HFILL }},
    { &hf_rnsap_repetitionNumber,
      { "repetitionNumber", "rnsap.repetitionNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RepetitionNumber0", HFILL }},
    { &hf_rnsap_MessageStructure_item,
      { "MessageStructure item", "rnsap.MessageStructure_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_repetitionNumber_01,
      { "repetitionNumber", "rnsap.repetitionNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RepetitionNumber1", HFILL }},
    { &hf_rnsap_continuousPacketConnectivity_DRX_InformationLCR,
      { "continuousPacketConnectivity-DRX-InformationLCR", "rnsap.continuousPacketConnectivity_DRX_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_continuousPacketConnectivity_DRX_Information_to_Modify_LCR,
      { "continuousPacketConnectivity-DRX-Information-to-Modify-LCR", "rnsap.continuousPacketConnectivity_DRX_Information_to_Modify_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_DSCH_Semi_PersistentScheduling_Information_LCR,
      { "hS-DSCH-Semi-PersistentScheduling-Information-LCR", "rnsap.hS_DSCH_Semi_PersistentScheduling_Information_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR,
      { "hS-DSCH-Semi-PersistentScheduling-Information-to-Modify-LCR", "rnsap.hS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_DSCH_SPS_Deactivate_Indicator_LCR,
      { "hS-DSCH-SPS-Deactivate-Indicator-LCR", "rnsap.hS_DSCH_SPS_Deactivate_Indicator_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_Semi_PersistentScheduling_Information_LCR,
      { "e-DCH-Semi-PersistentScheduling-Information-LCR", "rnsap.e_DCH_Semi_PersistentScheduling_Information_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR,
      { "e-DCH-Semi-PersistentScheduling-Information-to-Modify-LCR", "rnsap.e_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_SPS_Deactivate_Indicator_LCR,
      { "e-DCH-SPS-Deactivate-Indicator-LCR", "rnsap.e_DCH_SPS_Deactivate_Indicator_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_SCCH_DRX_Information_LCR,
      { "hS-SCCH-DRX-Information-LCR", "rnsap.hS_SCCH_DRX_Information_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_AGCH_DRX_Information_LCR,
      { "e-AGCH-DRX-Information-LCR", "rnsap.e_AGCH_DRX_Information_LCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_AGCH_DRX_Information_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_SCCH_UE_DRX_Cycle_LCR,
      { "hS-SCCH-UE-DRX-Cycle-LCR", "rnsap.hS_SCCH_UE_DRX_Cycle_LCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_DRX_Cycle_LCR_vals), 0,
        "UE_DRX_Cycle_LCR", HFILL }},
    { &hf_rnsap_hS_SCCH_Inactivity_Threshold_for_UE_DRX_Cycle_LCR,
      { "hS-SCCH-Inactivity-Threshold-for-UE-DRX-Cycle-LCR", "rnsap.hS_SCCH_Inactivity_Threshold_for_UE_DRX_Cycle_LCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_vals), 0,
        "Inactivity_Threshold_for_UE_DRX_Cycle_LCR", HFILL }},
    { &hf_rnsap_hS_SCCH_UE_DRX_Offset_LCR,
      { "hS-SCCH-UE-DRX-Offset-LCR", "rnsap.hS_SCCH_UE_DRX_Offset_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UE_DRX_Offset_LCR", HFILL }},
    { &hf_rnsap_sameAsHS_SCCH,
      { "sameAsHS-SCCH", "rnsap.sameAsHS_SCCH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_AGCH_DRX_Parameters,
      { "e-AGCH-DRX-Parameters", "rnsap.e_AGCH_DRX_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_AGCH_UE_DRX_Cycle_LCR,
      { "e-AGCH-UE-DRX-Cycle-LCR", "rnsap.e_AGCH_UE_DRX_Cycle_LCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_DRX_Cycle_LCR_vals), 0,
        "UE_DRX_Cycle_LCR", HFILL }},
    { &hf_rnsap_e_AGCH_UE_Inactivity_Monitor_Threshold,
      { "e-AGCH-UE-Inactivity-Monitor-Threshold", "rnsap.e_AGCH_UE_Inactivity_Monitor_Threshold",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_AGCH_UE_Inactivity_Monitor_Threshold_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_e_AGCH_UE_DRX_Offset_LCR,
      { "e-AGCH-UE-DRX-Offset-LCR", "rnsap.e_AGCH_UE_DRX_Offset_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UE_DRX_Offset_LCR", HFILL }},
    { &hf_rnsap_dRX_Information_to_Modify_LCR,
      { "dRX-Information-to-Modify-LCR", "rnsap.dRX_Information_to_Modify_LCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_DRX_Information_to_Modify_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_modify,
      { "modify", "rnsap.modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DRX_Information_to_Modify_Items_LCR", HFILL }},
    { &hf_rnsap_deactivate,
      { "deactivate", "rnsap.deactivate_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_SCCH_DRX_Information_ResponseLCR,
      { "hS-SCCH-DRX-Information-ResponseLCR", "rnsap.hS_SCCH_DRX_Information_ResponseLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_AGCH_DRX_Information_ResponseLCR,
      { "e-AGCH-DRX-Information-ResponseLCR", "rnsap.e_AGCH_DRX_Information_ResponseLCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_AGCH_DRX_Information_ResponseLCR_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_e_AGCH_DRX_Parameters_Response,
      { "e-AGCH-DRX-Parameters-Response", "rnsap.e_AGCH_DRX_Parameters_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_multiple_PLMN_List,
      { "multiple-PLMN-List", "rnsap.multiple_PLMN_List_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_umts_cell_Info,
      { "umts-cell-Info", "rnsap.umts_cell_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UMTS_Cells_Info_List", HFILL }},
    { &hf_rnsap_dGANSS_Signal_ID,
      { "dGANSS-Signal-ID", "rnsap.dGANSS_Signal_ID",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_ie_Extensions,
      { "ie-Extensions", "rnsap.ie_Extensions",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ProtocolExtensionContainer", HFILL }},
    { &hf_rnsap_dBDS_InfoList,
      { "dBDS-InfoList", "rnsap.dBDS_InfoList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dBDS_SignalID,
      { "dBDS-SignalID", "rnsap.dBDS_SignalID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_Signal_ID", HFILL }},
    { &hf_rnsap_dBDS_SignalInfoList,
      { "dBDS-SignalInfoList", "rnsap.dBDS_SignalInfoList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DBDS_SignalInfoList_item,
      { "DBDS-SignalInfo", "rnsap.DBDS_SignalInfo_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_svID,
      { "svID", "rnsap.svID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_63", HFILL }},
    { &hf_rnsap_bds_UDREI,
      { "bds-UDREI", "rnsap.bds_UDREI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_15", HFILL }},
    { &hf_rnsap_bds_RURAI,
      { "bds-RURAI", "rnsap.bds_RURAI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_15", HFILL }},
    { &hf_rnsap_bds_ECC_DeltaT,
      { "bds-ECC-DeltaT", "rnsap.bds_ECC_DeltaT",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_rnsap_DBDS_InfoList_item,
      { "DBDS-Info", "rnsap.DBDS_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_pO2_SRB,
      { "pO2-SRB", "rnsap.pO2_SRB",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PowerOffset", HFILL }},
    { &hf_rnsap_dl_FET_Mode,
      { "dl-FET-Mode", "rnsap.dl_FET_Mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_DL_FET_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dCH_ENH_Concat,
      { "dCH-ENH-Concat", "rnsap.dCH_ENH_Concat",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_pO_SRB,
      { "pO-SRB", "rnsap.pO_SRB",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PowerOffset", HFILL }},
    { &hf_rnsap_DCH_ENH_Concat_item,
      { "DCH-ID", "rnsap.DCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH,
      { "setup-Or-ConfigurationChange-Or-Removal-Of-DCH-ENH", "rnsap.setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH",
        FT_UINT32, BASE_DEC, VALS(rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_FDD_Information_item,
      { "DCH-FDD-InformationItem", "rnsap.DCH_FDD_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_payloadCRC_PresenceIndicator,
      { "payloadCRC-PresenceIndicator", "rnsap.payloadCRC_PresenceIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_PayloadCRC_PresenceIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_FP_Mode,
      { "ul-FP-Mode", "rnsap.ul_FP_Mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_UL_FP_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_toAWS,
      { "toAWS", "rnsap.toAWS",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_toAWE,
      { "toAWE", "rnsap.toAWE",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dCH_SpecificInformationList,
      { "dCH-SpecificInformationList", "rnsap.dCH_SpecificInformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DCH_Specific_FDD_InformationList", HFILL }},
    { &hf_rnsap_DCH_MeasurementOccasion_Information_item,
      { "DchMeasurementOccasionInformation-Item", "rnsap.DchMeasurementOccasionInformation_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_pattern_Sequence_Identifier,
      { "pattern-Sequence-Identifier", "rnsap.pattern_Sequence_Identifier",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_status_Flag,
      { "status-Flag", "rnsap.status_Flag",
        FT_UINT32, BASE_DEC, VALS(rnsap_Status_Flag_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters,
      { "measurement-Occasion-Pattern-Sequence-parameters", "rnsap.measurement_Occasion_Pattern_Sequence_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters_k,
      { "measurement-Occasion-Pattern-Sequence-parameters-k", "rnsap.measurement_Occasion_Pattern_Sequence_parameters_k",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_9", HFILL }},
    { &hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters_offset,
      { "measurement-Occasion-Pattern-Sequence-parameters-offset", "rnsap.measurement_Occasion_Pattern_Sequence_parameters_offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_511", HFILL }},
    { &hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters_M_Length,
      { "measurement-Occasion-Pattern-Sequence-parameters-M-Length", "rnsap.measurement_Occasion_Pattern_Sequence_parameters_M_Length",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_512", HFILL }},
    { &hf_rnsap_measurement_Occasion_Pattern_Sequence_parameters_Timeslot_Bitmap,
      { "measurement-Occasion-Pattern-Sequence-parameters-Timeslot-Bitmap", "rnsap.measurement_Occasion_Pattern_Sequence_parameters_Timeslot_Bitmap",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_7", HFILL }},
    { &hf_rnsap_DCH_Specific_FDD_InformationList_item,
      { "DCH-Specific-FDD-Item", "rnsap.DCH_Specific_FDD_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dCH_ID,
      { "dCH-ID", "rnsap.dCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_trCH_SrcStatisticsDescr,
      { "trCH-SrcStatisticsDescr", "rnsap.trCH_SrcStatisticsDescr",
        FT_UINT32, BASE_DEC, VALS(rnsap_TrCH_SrcStatisticsDescr_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_transportFormatSet,
      { "ul-transportFormatSet", "rnsap.ul_transportFormatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TransportFormatSet", HFILL }},
    { &hf_rnsap_dl_transportFormatSet,
      { "dl-transportFormatSet", "rnsap.dl_transportFormatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TransportFormatSet", HFILL }},
    { &hf_rnsap_ul_BLER,
      { "ul-BLER", "rnsap.ul_BLER",
        FT_INT32, BASE_DEC, NULL, 0,
        "BLER", HFILL }},
    { &hf_rnsap_dl_BLER,
      { "dl-BLER", "rnsap.dl_BLER",
        FT_INT32, BASE_DEC, NULL, 0,
        "BLER", HFILL }},
    { &hf_rnsap_allocationRetentionPriority,
      { "allocationRetentionPriority", "rnsap.allocationRetentionPriority_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_frameHandlingPriority,
      { "frameHandlingPriority", "rnsap.frameHandlingPriority",
        FT_UINT32, BASE_DEC, VALS(rnsap_FrameHandlingPriority_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_qE_Selector,
      { "qE-Selector", "rnsap.qE_Selector",
        FT_UINT32, BASE_DEC, VALS(rnsap_QE_Selector_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dRACControl,
      { "dRACControl", "rnsap.dRACControl",
        FT_UINT32, BASE_DEC, VALS(rnsap_DRACControl_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_InformationResponse_item,
      { "DCH-InformationResponseItem", "rnsap.DCH_InformationResponseItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_TDD_Information_item,
      { "DCH-TDD-InformationItem", "rnsap.DCH_TDD_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dCH_SpecificInformationList_01,
      { "dCH-SpecificInformationList", "rnsap.dCH_SpecificInformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DCH_Specific_TDD_InformationList", HFILL }},
    { &hf_rnsap_DCH_Specific_TDD_InformationList_item,
      { "DCH-Specific-TDD-Item", "rnsap.DCH_Specific_TDD_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_cCTrCH_ID,
      { "ul-cCTrCH-ID", "rnsap.ul_cCTrCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_ID", HFILL }},
    { &hf_rnsap_dl_cCTrCH_ID,
      { "dl-cCTrCH-ID", "rnsap.dl_cCTrCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_ID", HFILL }},
    { &hf_rnsap_sIR_Value,
      { "sIR-Value", "rnsap.sIR_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sIR_ErrorValue,
      { "sIR-ErrorValue", "rnsap.sIR_ErrorValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SIR_Error_Value", HFILL }},
    { &hf_rnsap_transmittedCodePowerValue,
      { "transmittedCodePowerValue", "rnsap.transmittedCodePowerValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Transmitted_Code_Power_Value", HFILL }},
    { &hf_rnsap_rSCP,
      { "rSCP", "rnsap.rSCP",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RSCP_Value", HFILL }},
    { &hf_rnsap_rxTimingDeviationValue,
      { "rxTimingDeviationValue", "rnsap.rxTimingDeviationValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Rx_Timing_Deviation_Value", HFILL }},
    { &hf_rnsap_roundTripTime,
      { "roundTripTime", "rnsap.roundTripTime",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Round_Trip_Time_Value", HFILL }},
    { &hf_rnsap_extension_DedicatedMeasurementValue,
      { "extension-DedicatedMeasurementValue", "rnsap.extension_DedicatedMeasurementValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_measurementAvailable_01,
      { "measurementAvailable", "rnsap.measurementAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DedicatedMeasurementAvailable", HFILL }},
    { &hf_rnsap_measurementnotAvailable_01,
      { "measurementnotAvailable", "rnsap.measurementnotAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DedicatedMeasurementnotAvailable", HFILL }},
    { &hf_rnsap_dedicatedmeasurementValue,
      { "dedicatedmeasurementValue", "rnsap.dedicatedmeasurementValue",
        FT_UINT32, BASE_DEC, VALS(rnsap_DedicatedMeasurementValue_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_cFN,
      { "cFN", "rnsap.cFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cfn,
      { "cfn", "rnsap.cfn",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_separate_indication,
      { "separate-indication", "rnsap.separate_indication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_activate,
      { "activate", "rnsap.activate_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Activate_Info", HFILL }},
    { &hf_rnsap_deactivate_01,
      { "deactivate", "rnsap.deactivate_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Deactivate_Info", HFILL }},
    { &hf_rnsap_activation_type,
      { "activation-type", "rnsap.activation_type",
        FT_UINT32, BASE_DEC, VALS(rnsap_Execution_Type_vals), 0,
        "Execution_Type", HFILL }},
    { &hf_rnsap_initial_dl_tx_power,
      { "initial-dl-tx-power", "rnsap.initial_dl_tx_power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_firstRLS_Indicator,
      { "firstRLS-Indicator", "rnsap.firstRLS_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_FirstRLS_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_propagation_delay,
      { "propagation-delay", "rnsap.propagation_delay",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PropagationDelay", HFILL }},
    { &hf_rnsap_deactivation_type,
      { "deactivation-type", "rnsap.deactivation_type",
        FT_UINT32, BASE_DEC, VALS(rnsap_Execution_Type_vals), 0,
        "Execution_Type", HFILL }},
    { &hf_rnsap_synchronised,
      { "synchronised", "rnsap.synchronised",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CFN", HFILL }},
    { &hf_rnsap_unsynchronised,
      { "unsynchronised", "rnsap.unsynchronised_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dGANSS_ReferenceTime,
      { "dGANSS-ReferenceTime", "rnsap.dGANSS_ReferenceTime",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_119", HFILL }},
    { &hf_rnsap_dGANSS_Information,
      { "dGANSS-Information", "rnsap.dGANSS_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dGANSS_Information_item,
      { "dGANSS-Information item", "rnsap.dGANSS_Information_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gANSS_SignalId,
      { "gANSS-SignalId", "rnsap.gANSS_SignalId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_Signal_ID", HFILL }},
    { &hf_rnsap_gANSS_StatusHealth,
      { "gANSS-StatusHealth", "rnsap.gANSS_StatusHealth",
        FT_UINT32, BASE_DEC, VALS(rnsap_GANSS_StatusHealth_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dGANSS_SignalInformation,
      { "dGANSS-SignalInformation", "rnsap.dGANSS_SignalInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dGANSS_SignalInformation_item,
      { "dGANSS-SignalInformation item", "rnsap.dGANSS_SignalInformation_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_satId,
      { "satId", "rnsap.satId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_63", HFILL }},
    { &hf_rnsap_gANSS_iod,
      { "gANSS-iod", "rnsap.gANSS_iod",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_rnsap_udre,
      { "udre", "rnsap.udre",
        FT_UINT32, BASE_DEC, VALS(rnsap_UDRE_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_prc,
      { "ganss-prc", "rnsap.ganss_prc",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M2047_2047", HFILL }},
    { &hf_rnsap_ganss_rrc,
      { "ganss-rrc", "rnsap.ganss_rrc",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M127_127", HFILL }},
    { &hf_rnsap_pRCDeviation,
      { "pRCDeviation", "rnsap.pRCDeviation",
        FT_UINT32, BASE_DEC, VALS(rnsap_PRCDeviation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_udreGrowthRate,
      { "udreGrowthRate", "rnsap.udreGrowthRate",
        FT_UINT32, BASE_DEC, VALS(rnsap_UDREGrowthRate_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_udreValidityTime,
      { "udreValidityTime", "rnsap.udreValidityTime",
        FT_UINT32, BASE_DEC, VALS(rnsap_UDREValidityTime_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_gPSTOW,
      { "gPSTOW", "rnsap.gPSTOW",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gPS_Status_Health,
      { "gPS-Status-Health", "rnsap.gPS_Status_Health",
        FT_UINT32, BASE_DEC, VALS(rnsap_GPS_Status_Health_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_satellite_DGPSCorrections_Information,
      { "satellite-DGPSCorrections-Information", "rnsap.satellite_DGPSCorrections_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_satellite_DGPSCorrections_Information_item,
      { "satellite-DGPSCorrections-Information item", "rnsap.satellite_DGPSCorrections_Information_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sAT_ID,
      { "sAT-ID", "rnsap.sAT_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_iode_dgps,
      { "iode-dgps", "rnsap.iode_dgps",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_uDRE,
      { "uDRE", "rnsap.uDRE",
        FT_UINT32, BASE_DEC, VALS(rnsap_UDRE_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_pRC,
      { "pRC", "rnsap.pRC",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_range_Correction_Rate,
      { "range-Correction-Rate", "rnsap.range_Correction_Rate",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hs_DSCH_Cell_List,
      { "hs-DSCH-Cell-List", "rnsap.hs_DSCH_Cell_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HS_DSCH_Cell_List_item,
      { "HS-DSCH-Cell", "rnsap.HS_DSCH_Cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hs_DSCH_Cell_TBS,
      { "hs-DSCH-Cell-TBS", "rnsap.hs_DSCH_Cell_TBS",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_TBS_Value", HFILL }},
    { &hf_rnsap_powerAdjustmentType,
      { "powerAdjustmentType", "rnsap.powerAdjustmentType",
        FT_UINT32, BASE_DEC, VALS(rnsap_PowerAdjustmentType_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dLReferencePower,
      { "dLReferencePower", "rnsap.dLReferencePower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_dLReferencePowerList,
      { "dLReferencePowerList", "rnsap.dLReferencePowerList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_ReferencePowerInformationList", HFILL }},
    { &hf_rnsap_maxAdjustmentStep,
      { "maxAdjustmentStep", "rnsap.maxAdjustmentStep",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_adjustmentPeriod,
      { "adjustmentPeriod", "rnsap.adjustmentPeriod",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_adjustmentRatio,
      { "adjustmentRatio", "rnsap.adjustmentRatio",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ScaledAdjustmentRatio", HFILL }},
    { &hf_rnsap_DL_ReferencePowerInformationList_item,
      { "DL-ReferencePowerInformationItem", "rnsap.DL_ReferencePowerInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_ID,
      { "rL-ID", "rnsap.rL_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_Reference_Power,
      { "dl-Reference-Power", "rnsap.dl_Reference_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_common_DL_ReferencePowerInformation,
      { "common-DL-ReferencePowerInformation", "rnsap.common_DL_ReferencePowerInformation",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_individual_DL_ReferencePowerInformation,
      { "individual-DL-ReferencePowerInformation", "rnsap.individual_DL_ReferencePowerInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_ReferencePowerInformationList", HFILL }},
    { &hf_rnsap_DL_Timeslot_Information_item,
      { "DL-Timeslot-InformationItem", "rnsap.DL_Timeslot_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_midambleShiftAndBurstType,
      { "midambleShiftAndBurstType", "rnsap.midambleShiftAndBurstType",
        FT_UINT32, BASE_DEC, VALS(rnsap_MidambleShiftAndBurstType_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_tFCI_Presence,
      { "tFCI-Presence", "rnsap.tFCI_Presence",
        FT_UINT32, BASE_DEC, VALS(rnsap_TFCI_Presence_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dL_Code_Information,
      { "dL-Code-Information", "rnsap.dL_Code_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_DL_Code_Information", HFILL }},
    { &hf_rnsap_DL_TimeslotLCR_Information_item,
      { "DL-TimeslotLCR-InformationItem", "rnsap.DL_TimeslotLCR_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_timeSlotLCR,
      { "timeSlotLCR", "rnsap.timeSlotLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_midambleShiftLCR,
      { "midambleShiftLCR", "rnsap.midambleShiftLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dL_Code_LCR_Information,
      { "dL-Code-LCR-Information", "rnsap.dL_Code_LCR_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_DL_Code_LCR_Information", HFILL }},
    { &hf_rnsap_DL_Timeslot_Information768_item,
      { "DL-Timeslot-InformationItem768", "rnsap.DL_Timeslot_InformationItem768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_midambleShiftAndBurstType768,
      { "midambleShiftAndBurstType768", "rnsap.midambleShiftAndBurstType768",
        FT_UINT32, BASE_DEC, VALS(rnsap_MidambleShiftAndBurstType768_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dL_Code_Information768,
      { "dL-Code-Information768", "rnsap.dL_Code_Information768",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_DL_Code_Information768", HFILL }},
    { &hf_rnsap_DL_TimeSlot_ISCP_Info_item,
      { "DL-TimeSlot-ISCP-InfoItem", "rnsap.DL_TimeSlot_ISCP_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dL_TimeslotISCP,
      { "dL-TimeslotISCP", "rnsap.dL_TimeslotISCP",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_TimeSlot_ISCP_LCR_Information_item,
      { "DL-TimeSlot-ISCP-LCR-InfoItem", "rnsap.DL_TimeSlot_ISCP_LCR_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_decimationFactorforPrimaryFrequency,
      { "decimationFactorforPrimaryFrequency", "rnsap.decimationFactorforPrimaryFrequency",
        FT_UINT32, BASE_DEC, VALS(rnsap_DecimationFactor_vals), 0,
        "DecimationFactor", HFILL }},
    { &hf_rnsap_decimationFactorforSecFrequency,
      { "decimationFactorforSecFrequency", "rnsap.decimationFactorforSecFrequency",
        FT_UINT32, BASE_DEC, VALS(rnsap_DecimationFactor_vals), 0,
        "DecimationFactor", HFILL }},
    { &hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements,
      { "setup-Or-ConfigurationChange-Or-Removal-Of-Downlink-TPC-enhancements", "rnsap.setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements",
        FT_UINT32, BASE_DEC, VALS(rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_DRX_Cycle,
      { "uE-DRX-Cycle", "rnsap.uE_DRX_Cycle",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_DRX_Cycle_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_inactivity_Threshold_for_UE_DRX_Cycle,
      { "inactivity-Threshold-for-UE-DRX-Cycle", "rnsap.inactivity_Threshold_for_UE_DRX_Cycle",
        FT_UINT32, BASE_DEC, VALS(rnsap_Inactivity_Threshold_for_UE_DRX_Cycle_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_inactivity_Threshold_for_UE_Grant_Monitoring,
      { "inactivity-Threshold-for-UE-Grant-Monitoring", "rnsap.inactivity_Threshold_for_UE_Grant_Monitoring",
        FT_UINT32, BASE_DEC, VALS(rnsap_Inactivity_Threshold_for_UE_Grant_Monitoring_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_DRX_Grant_Monitoring,
      { "uE-DRX-Grant-Monitoring", "rnsap.uE_DRX_Grant_Monitoring",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_modify_01,
      { "modify", "rnsap.modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DRX_Information_to_Modify_Items", HFILL }},
    { &hf_rnsap_DSCH_FlowControlInformation_item,
      { "DSCH-FlowControlItem", "rnsap.DSCH_FlowControlItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dSCH_SchedulingPriority,
      { "dSCH-SchedulingPriority", "rnsap.dSCH_SchedulingPriority",
        FT_UINT32, BASE_DEC, VALS(rnsap_SchedulingPriorityIndicator_vals), 0,
        "SchedulingPriorityIndicator", HFILL }},
    { &hf_rnsap_mAC_c_sh_SDU_Lengths,
      { "mAC-c-sh-SDU-Lengths", "rnsap.mAC_c_sh_SDU_Lengths",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MAC_c_sh_SDU_LengthList", HFILL }},
    { &hf_rnsap_DSCH_TDD_Information_item,
      { "DSCH-TDD-InformationItem", "rnsap.DSCH_TDD_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dSCH_ID,
      { "dSCH-ID", "rnsap.dSCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_ccTrCHID,
      { "dl-ccTrCHID", "rnsap.dl_ccTrCHID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_ID", HFILL }},
    { &hf_rnsap_trChSourceStatisticsDescriptor,
      { "trChSourceStatisticsDescriptor", "rnsap.trChSourceStatisticsDescriptor",
        FT_UINT32, BASE_DEC, VALS(rnsap_TrCH_SrcStatisticsDescr_vals), 0,
        "TrCH_SrcStatisticsDescr", HFILL }},
    { &hf_rnsap_transportFormatSet,
      { "transportFormatSet", "rnsap.transportFormatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_schedulingPriorityIndicator,
      { "schedulingPriorityIndicator", "rnsap.schedulingPriorityIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SchedulingPriorityIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_bLER,
      { "bLER", "rnsap.bLER",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_DTX_Cycle1_2ms,
      { "uE-DTX-Cycle1-2ms", "rnsap.uE_DTX_Cycle1_2ms",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_DTX_Cycle1_2ms_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_DTX_Cycle2_2ms,
      { "uE-DTX-Cycle2-2ms", "rnsap.uE_DTX_Cycle2_2ms",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_DTX_Cycle2_2ms_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_mAC_DTX_Cycle_2ms,
      { "mAC-DTX-Cycle-2ms", "rnsap.mAC_DTX_Cycle_2ms",
        FT_UINT32, BASE_DEC, VALS(rnsap_MAC_DTX_Cycle_2ms_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_DTX_Cycle1_10ms,
      { "uE-DTX-Cycle1-10ms", "rnsap.uE_DTX_Cycle1_10ms",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_DTX_Cycle1_10ms_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_DTX_Cycle2_10ms,
      { "uE-DTX-Cycle2-10ms", "rnsap.uE_DTX_Cycle2_10ms",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_DTX_Cycle2_10ms_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_mAC_DTX_Cycle_10ms,
      { "mAC-DTX-Cycle-10ms", "rnsap.mAC_DTX_Cycle_10ms",
        FT_UINT32, BASE_DEC, VALS(rnsap_MAC_DTX_Cycle_10ms_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_TTI_Length,
      { "e-DCH-TTI-Length", "rnsap.e_DCH_TTI_Length",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_DCH_TTI_Length_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_DTX_Long_Preamble,
      { "uE-DTX-Long-Preamble", "rnsap.uE_DTX_Long_Preamble",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_DTX_Long_Preamble_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_mAC_Inactivity_Threshold,
      { "mAC-Inactivity-Threshold", "rnsap.mAC_Inactivity_Threshold",
        FT_UINT32, BASE_DEC, VALS(rnsap_MAC_Inactivity_Threshold_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_cQI_DTX_Timer,
      { "cQI-DTX-Timer", "rnsap.cQI_DTX_Timer",
        FT_UINT32, BASE_DEC, VALS(rnsap_CQI_DTX_Timer_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_DPCCH_burst1,
      { "uE-DPCCH-burst1", "rnsap.uE_DPCCH_burst1",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_DPCCH_burst1_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_DPCCH_burst2,
      { "uE-DPCCH-burst2", "rnsap.uE_DPCCH_burst2",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_DPCCH_burst2_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_modify_02,
      { "modify", "rnsap.modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DTX_Information_to_Modify_Items", HFILL }},
    { &hf_rnsap_e_DCH_TTI_Length_to_Modify,
      { "e-DCH-TTI-Length-to-Modify", "rnsap.e_DCH_TTI_Length_to_Modify",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_DCH_TTI_Length_to_Modify_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_fDD,
      { "fDD", "rnsap.fDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "EARFCN_FDD", HFILL }},
    { &hf_rnsap_tDD,
      { "tDD", "rnsap.tDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EARFCN", HFILL }},
    { &hf_rnsap_extension_EARFCN_Information,
      { "extension-EARFCN-Information", "rnsap.extension_EARFCN_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_EARFCN,
      { "uL-EARFCN", "rnsap.uL_EARFCN",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EARFCN", HFILL }},
    { &hf_rnsap_dL_EARFCN,
      { "dL-EARFCN", "rnsap.dL_EARFCN",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EARFCN", HFILL }},
    { &hf_rnsap_uL_EARFCN_01,
      { "uL-EARFCN", "rnsap.uL_EARFCN",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EARFCN_Extended", HFILL }},
    { &hf_rnsap_dL_EARFCN_01,
      { "dL-EARFCN", "rnsap.dL_EARFCN",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EARFCN_Extended", HFILL }},
    { &hf_rnsap_e_UTRAN_Cell_ID,
      { "e-UTRAN-Cell-ID", "rnsap.e_UTRAN_Cell_ID",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_28", HFILL }},
    { &hf_rnsap_eAGCH_ERGCH_EHICH_FDD_ScramblingCode,
      { "eAGCH-ERGCH-EHICH-FDD-ScramblingCode", "rnsap.eAGCH_ERGCH_EHICH_FDD_ScramblingCode",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_ScramblingCode", HFILL }},
    { &hf_rnsap_eAGCH_ChannelisationCode,
      { "eAGCH-ChannelisationCode", "rnsap.eAGCH_ChannelisationCode",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FDD_DL_ChannelisationCodeNumber", HFILL }},
    { &hf_rnsap_primary_e_RNTI,
      { "primary-e-RNTI", "rnsap.primary_e_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_RNTI", HFILL }},
    { &hf_rnsap_secondary_e_RNTI,
      { "secondary-e-RNTI", "rnsap.secondary_e_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_RNTI", HFILL }},
    { &hf_rnsap_eRGCH_EHICH_ChannelisationCode,
      { "eRGCH-EHICH-ChannelisationCode", "rnsap.eRGCH_EHICH_ChannelisationCode",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FDD_DL_ChannelisationCodeNumber", HFILL }},
    { &hf_rnsap_eHICH_SignatureSequence,
      { "eHICH-SignatureSequence", "rnsap.eHICH_SignatureSequence",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_primary_Secondary_Grant_Selector,
      { "primary-Secondary-Grant-Selector", "rnsap.primary_Secondary_Grant_Selector",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_Primary_Secondary_Grant_Selector_vals), 0,
        "E_Primary_Secondary_Grant_Selector", HFILL }},
    { &hf_rnsap_e_RGCH_Release_Indicator,
      { "e-RGCH-Release-Indicator", "rnsap.e_RGCH_Release_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_RGCH_Release_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_eDCH_MACdFlows_Information,
      { "eDCH-MACdFlows-Information", "rnsap.eDCH_MACdFlows_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_Reference_Power_Offset,
      { "e-DCH-Reference-Power-Offset", "rnsap.e_DCH_Reference_Power_Offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eDCH_MACdFlow_Specific_InformationResponse,
      { "eDCH-MACdFlow-Specific-InformationResponse", "rnsap.eDCH_MACdFlow_Specific_InformationResponse",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_MACdFlow_Specific_InformationResponse_item,
      { "EDCH-MACdFlow-Specific-InformationResponseItem", "rnsap.EDCH_MACdFlow_Specific_InformationResponseItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eDCH_MACdFlow_ID,
      { "eDCH-MACdFlow-ID", "rnsap.eDCH_MACdFlow_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hARQ_Process_Allocation_NonSched_2ms_EDCH,
      { "hARQ-Process-Allocation-NonSched-2ms-EDCH", "rnsap.hARQ_Process_Allocation_NonSched_2ms_EDCH",
        FT_BYTES, BASE_NONE, NULL, 0,
        "HARQ_Process_Allocation_2ms_EDCH", HFILL }},
    { &hf_rnsap_eDCH_MACdFlow_Specific_Information,
      { "eDCH-MACdFlow-Specific-Information", "rnsap.eDCH_MACdFlow_Specific_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EDCH_MACdFlow_Specific_InfoToModifyList", HFILL }},
    { &hf_rnsap_mACeReset_Indicator,
      { "mACeReset-Indicator", "rnsap.mACeReset_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_MACeReset_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_MACdFlow_Specific_UpdateInformation,
      { "e-DCH-MACdFlow-Specific-UpdateInformation", "rnsap.e_DCH_MACdFlow_Specific_UpdateInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_MACdFlow_Specific_UpdateInformation_item,
      { "E-DCH-MACdFlow-Specific-UpdateInformation-Item", "rnsap.E_DCH_MACdFlow_Specific_UpdateInformation_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_DL_Control_Channel_Change_Information_item,
      { "E-DCH-DL-Control-Channel-Change-Information-Item", "rnsap.E_DCH_DL_Control_Channel_Change_Information_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_RL_ID,
      { "e-DCH-RL-ID", "rnsap.e_DCH_RL_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RL_ID", HFILL }},
    { &hf_rnsap_E_DCH_DL_Control_Channel_Grant_Information_item,
      { "E-DCH-DL-Control-Channel-Grant-Information-Item", "rnsap.E_DCH_DL_Control_Channel_Grant_Information_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_Non_Scheduled_Transmission_Grant,
      { "e-DCH-Non-Scheduled-Transmission-Grant", "rnsap.e_DCH_Non_Scheduled_Transmission_Grant_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "E_DCH_Non_Scheduled_Transmission_Grant_Items", HFILL }},
    { &hf_rnsap_e_DCH_Scheduled_Transmission_Grant,
      { "e-DCH-Scheduled-Transmission-Grant", "rnsap.e_DCH_Scheduled_Transmission_Grant_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_LogicalChannelInformation_item,
      { "E-DCH-LogicalChannelInformationItem", "rnsap.E_DCH_LogicalChannelInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_schedulingInformation,
      { "schedulingInformation", "rnsap.schedulingInformation",
        FT_UINT32, BASE_DEC, VALS(rnsap_SchedulingInformation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_mACes_GuaranteedBitRate,
      { "mACes-GuaranteedBitRate", "rnsap.mACes_GuaranteedBitRate",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MACes_Guaranteed_Bitrate", HFILL }},
    { &hf_rnsap_eDCH_DDI_Value,
      { "eDCH-DDI-Value", "rnsap.eDCH_DDI_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mACd_PDU_Size_List,
      { "mACd-PDU-Size-List", "rnsap.mACd_PDU_Size_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_MACdPDU_SizeList", HFILL }},
    { &hf_rnsap_E_DCH_MACdPDU_SizeList_item,
      { "E-DCH-MACdPDU-SizeListItem", "rnsap.E_DCH_MACdPDU_SizeListItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mACdPDU_Size,
      { "mACdPDU-Size", "rnsap.mACdPDU_Size",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_LogicalChannelToModify_item,
      { "E-DCH-LogicalChannelToModifyItem", "rnsap.E_DCH_LogicalChannelToModifyItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mACd_PDU_Size_List_01,
      { "mACd-PDU-Size-List", "rnsap.mACd_PDU_Size_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_MACdPDU_SizeToModifyList", HFILL }},
    { &hf_rnsap_E_DCH_MACdPDU_SizeToModifyList_item,
      { "E-DCH-MACdPDU-SizeListItem", "rnsap.E_DCH_MACdPDU_SizeListItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_LogicalChannelToDelete_item,
      { "E-DCH-LogicalChannelToDeleteItem", "rnsap.E_DCH_LogicalChannelToDeleteItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eDCH_MACdFlow_Specific_Information_01,
      { "eDCH-MACdFlow-Specific-Information", "rnsap.eDCH_MACdFlow_Specific_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EDCH_MACdFlow_Specific_InfoList", HFILL }},
    { &hf_rnsap_EDCH_MACdFlow_Specific_InfoList_item,
      { "EDCH-MACdFlow-Specific-InfoItem", "rnsap.EDCH_MACdFlow_Specific_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tnlQoS,
      { "tnlQoS", "rnsap.tnlQoS",
        FT_UINT32, BASE_DEC, VALS(rnsap_TnlQos_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_maxNr_Retransmissions_EDCH,
      { "maxNr-Retransmissions-EDCH", "rnsap.maxNr_Retransmissions_EDCH",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_trafficClass,
      { "trafficClass", "rnsap.trafficClass",
        FT_UINT32, BASE_DEC, VALS(rnsap_TrafficClass_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_eDCH_Grant_Type_Information,
      { "eDCH-Grant-Type-Information", "rnsap.eDCH_Grant_Type_Information",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_DCH_Grant_Type_Information_vals), 0,
        "E_DCH_Grant_Type_Information", HFILL }},
    { &hf_rnsap_bundlingModeIndicator,
      { "bundlingModeIndicator", "rnsap.bundlingModeIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_BundlingModeIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_eDCHLogicalChannelInformation,
      { "eDCHLogicalChannelInformation", "rnsap.eDCHLogicalChannelInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_LogicalChannelInformation", HFILL }},
    { &hf_rnsap_EDCH_MACdFlow_Specific_InfoToModifyList_item,
      { "EDCH-MACdFlow-Specific-InfoToModifyItem", "rnsap.EDCH_MACdFlow_Specific_InfoToModifyItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_transportBearerRequestIndicator,
      { "transportBearerRequestIndicator", "rnsap.transportBearerRequestIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransportBearerRequestIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_eDCH_LogicalChannelToAdd,
      { "eDCH-LogicalChannelToAdd", "rnsap.eDCH_LogicalChannelToAdd",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_LogicalChannelInformation", HFILL }},
    { &hf_rnsap_eDCH_LogicalChannelToModify,
      { "eDCH-LogicalChannelToModify", "rnsap.eDCH_LogicalChannelToModify",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_LogicalChannelToModify", HFILL }},
    { &hf_rnsap_eDCH_LogicalChannelToDelete,
      { "eDCH-LogicalChannelToDelete", "rnsap.eDCH_LogicalChannelToDelete",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_LogicalChannelToDelete", HFILL }},
    { &hf_rnsap_EDCH_MACdFlows_To_Delete_item,
      { "EDCH-MACdFlows-To-Delete-Item", "rnsap.EDCH_MACdFlows_To_Delete_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_maxBits_MACe_PDU_non_scheduled,
      { "maxBits-MACe-PDU-non-scheduled", "rnsap.maxBits_MACe_PDU_non_scheduled",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Max_Bits_MACe_PDU_non_scheduled", HFILL }},
    { &hf_rnsap_hARQ_Process_Allocation_NonSched_2ms,
      { "hARQ-Process-Allocation-NonSched-2ms", "rnsap.hARQ_Process_Allocation_NonSched_2ms",
        FT_BYTES, BASE_NONE, NULL, 0,
        "HARQ_Process_Allocation_2ms_EDCH", HFILL }},
    { &hf_rnsap_e_DCH_serving_cell_outcome_choice,
      { "e-DCH-serving-cell-outcome-choice", "rnsap.e_DCH_serving_cell_outcome_choice",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_DCH_serving_cell_change_choice_vals), 0,
        "E_DCH_serving_cell_change_choice", HFILL }},
    { &hf_rnsap_e_DCH_serving_cell_change_successful,
      { "e-DCH-serving-cell-change-successful", "rnsap.e_DCH_serving_cell_change_successful_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_serving_cell_change_unsuccessful,
      { "e-DCH-serving-cell-change-unsuccessful", "rnsap.e_DCH_serving_cell_change_unsuccessful_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_RL_InformationList_Rsp,
      { "e-DCH-RL-InformationList-Rsp", "rnsap.e_DCH_RL_InformationList_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_RL_InformationList_Rsp_item,
      { "E-DCH-RL-InformationList-Rsp-Item", "rnsap.E_DCH_RL_InformationList_Rsp_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_reconfigured_RL_Id,
      { "e-DCH-reconfigured-RL-Id", "rnsap.e_DCH_reconfigured_RL_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RL_ID", HFILL }},
    { &hf_rnsap_e_DCH_FDD_DL_Control_Channel_Info,
      { "e-DCH-FDD-DL-Control-Channel-Info", "rnsap.e_DCH_FDD_DL_Control_Channel_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "EDCH_FDD_DL_ControlChannelInformation", HFILL }},
    { &hf_rnsap_cause,
      { "cause", "rnsap.cause",
        FT_UINT32, BASE_DEC, VALS(rnsap_Cause_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_two_ms,
      { "two-ms", "rnsap.two_ms_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DTX_Cycle_2ms_Items", HFILL }},
    { &hf_rnsap_ten_ms,
      { "ten-ms", "rnsap.ten_ms_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DTX_Cycle_10ms_Items", HFILL }},
    { &hf_rnsap_two_ms_01,
      { "two-ms", "rnsap.two_ms_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DTX_Cycle_2ms_to_Modify_Items", HFILL }},
    { &hf_rnsap_ten_ms_01,
      { "ten-ms", "rnsap.ten_ms_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DTX_Cycle_10ms_to_Modify_Items", HFILL }},
    { &hf_rnsap_maxSet_E_DPDCHs,
      { "maxSet-E-DPDCHs", "rnsap.maxSet_E_DPDCHs",
        FT_UINT32, BASE_DEC, VALS(rnsap_Max_Set_E_DPDCHs_vals), 0,
        "Max_Set_E_DPDCHs", HFILL }},
    { &hf_rnsap_punctureLimit,
      { "punctureLimit", "rnsap.punctureLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_TFCS_Information,
      { "e-TFCS-Information", "rnsap.e_TFCS_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_TTI,
      { "e-TTI", "rnsap.e_TTI",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_TTI_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DPCCH_PO,
      { "e-DPCCH-PO", "rnsap.e_DPCCH_PO",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_RGCH_2_IndexStepThreshold,
      { "e-RGCH-2-IndexStepThreshold", "rnsap.e_RGCH_2_IndexStepThreshold",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_RGCH_3_IndexStepThreshold,
      { "e-RGCH-3-IndexStepThreshold", "rnsap.e_RGCH_3_IndexStepThreshold",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hARQ_Info_for_E_DCH,
      { "hARQ-Info-for-E-DCH", "rnsap.hARQ_Info_for_E_DCH",
        FT_UINT32, BASE_DEC, VALS(rnsap_HARQ_Info_for_E_DCH_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_hSDSCH_Configured_Indicator,
      { "hSDSCH-Configured-Indicator", "rnsap.hSDSCH_Configured_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_HSDSCH_Configured_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_Serving_RL_in_this_DRNS,
      { "e-DCH-Serving-RL-in-this-DRNS", "rnsap.e_DCH_Serving_RL_in_this_DRNS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "EDCH_Serving_RL_in_this_DRNS", HFILL }},
    { &hf_rnsap_e_DCH_Serving_RL_not_in_this_DRNS,
      { "e-DCH-Serving-RL-not-in-this-DRNS", "rnsap.e_DCH_Serving_RL_not_in_this_DRNS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_Serving_RL_Id,
      { "e-DCH-Serving-RL-Id", "rnsap.e_DCH_Serving_RL_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RL_ID", HFILL }},
    { &hf_rnsap_common_HS_DSCH_RNTI_priorityQueueInfo_EnhancedFACH,
      { "common-HS-DSCH-RNTI-priorityQueueInfo-EnhancedFACH", "rnsap.common_HS_DSCH_RNTI_priorityQueueInfo_EnhancedFACH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_InfoList_EnhancedFACH_PCH", HFILL }},
    { &hf_rnsap_dedicated_HS_DSCH_RNTI_priorityQueueInfo_EnhancedFACH,
      { "dedicated-HS-DSCH-RNTI-priorityQueueInfo-EnhancedFACH", "rnsap.dedicated_HS_DSCH_RNTI_priorityQueueInfo_EnhancedFACH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_InfoList_EnhancedFACH_PCH", HFILL }},
    { &hf_rnsap_priorityQueueInfo_EnhancedPCH,
      { "priorityQueueInfo-EnhancedPCH", "rnsap.priorityQueueInfo_EnhancedPCH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_InfoList_EnhancedFACH_PCH", HFILL }},
    { &hf_rnsap_hSDSCH_Initial_Capacity_Allocation,
      { "hSDSCH-Initial-Capacity-Allocation", "rnsap.hSDSCH_Initial_Capacity_Allocation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSDSCH_RNTI,
      { "hSDSCH-RNTI", "rnsap.hSDSCH_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_TFCI_BetaEC_Boost,
      { "e-TFCI-BetaEC-Boost", "rnsap.e_TFCI_BetaEC_Boost",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_Delta_T2TP,
      { "uL-Delta-T2TP", "rnsap.uL_Delta_T2TP",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_TFCI_Table_Index,
      { "e-DCH-TFCI-Table-Index", "rnsap.e_DCH_TFCI_Table_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_reference_E_TFCI_Information,
      { "reference-E-TFCI-Information", "rnsap.reference_E_TFCI_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_measurementTreshold,
      { "measurementTreshold", "rnsap.measurementTreshold",
        FT_UINT32, BASE_DEC, VALS(rnsap_MeasurementThreshold_vals), 0,
        "MeasurementThreshold", HFILL }},
    { &hf_rnsap_measurementHysteresisTime,
      { "measurementHysteresisTime", "rnsap.measurementHysteresisTime",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_measurementIncreaseDecreaseThreshold,
      { "measurementIncreaseDecreaseThreshold", "rnsap.measurementIncreaseDecreaseThreshold",
        FT_UINT32, BASE_DEC, VALS(rnsap_MeasurementIncreaseDecreaseThreshold_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_measurementChangeTime,
      { "measurementChangeTime", "rnsap.measurementChangeTime",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_measurementThreshold1,
      { "measurementThreshold1", "rnsap.measurementThreshold1",
        FT_UINT32, BASE_DEC, VALS(rnsap_MeasurementThreshold_vals), 0,
        "MeasurementThreshold", HFILL }},
    { &hf_rnsap_measurementThreshold2,
      { "measurementThreshold2", "rnsap.measurementThreshold2",
        FT_UINT32, BASE_DEC, VALS(rnsap_MeasurementThreshold_vals), 0,
        "MeasurementThreshold", HFILL }},
    { &hf_rnsap_reportPeriodicity,
      { "reportPeriodicity", "rnsap.reportPeriodicity",
        FT_UINT32, BASE_DEC, VALS(rnsap_ReportPeriodicity_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_measurementFluctuationRange,
      { "measurementFluctuationRange", "rnsap.measurementFluctuationRange",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_100", HFILL }},
    { &hf_rnsap_measurementQuantity,
      { "measurementQuantity", "rnsap.measurementQuantity",
        FT_UINT32, BASE_DEC, VALS(rnsap_MeasurementQuantity_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_threshold,
      { "threshold", "rnsap.threshold",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M120_165", HFILL }},
    { &hf_rnsap_threshold_01,
      { "threshold", "rnsap.threshold",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M120_M25", HFILL }},
    { &hf_rnsap_e_PUCH_Information,
      { "e-PUCH-Information", "rnsap.e_PUCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_TFCS_Information_TDD,
      { "e-TFCS-Information-TDD", "rnsap.e_TFCS_Information_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_MACdFlows_Information_TDD,
      { "e-DCH-MACdFlows-Information-TDD", "rnsap.e_DCH_MACdFlows_Information_TDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_TDD_Information,
      { "e-DCH-TDD-Information", "rnsap.e_DCH_TDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_minCR,
      { "minCR", "rnsap.minCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CodeRate", HFILL }},
    { &hf_rnsap_maxCR,
      { "maxCR", "rnsap.maxCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CodeRate", HFILL }},
    { &hf_rnsap_harqInfo,
      { "harqInfo", "rnsap.harqInfo",
        FT_UINT32, BASE_DEC, VALS(rnsap_HARQ_Info_for_E_DCH_vals), 0,
        "HARQ_Info_for_E_DCH", HFILL }},
    { &hf_rnsap_n_E_UCCH,
      { "n-E-UCCH", "rnsap.n_E_UCCH",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_QPSK_RefBetaInfo,
      { "e-DCH-QPSK-RefBetaInfo", "rnsap.e_DCH_QPSK_RefBetaInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_sixteenQAM_RefBetaInfo,
      { "e-DCH-sixteenQAM-RefBetaInfo", "rnsap.e_DCH_sixteenQAM_RefBetaInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_QPSK_RefBetaInfo_item,
      { "E-DCH-RefBeta-Item", "rnsap.E_DCH_RefBeta_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_sixteenQAM_RefBetaInfo_item,
      { "E-DCH-RefBeta-Item", "rnsap.E_DCH_RefBeta_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_refCodeRate,
      { "refCodeRate", "rnsap.refCodeRate",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CodeRate_short", HFILL }},
    { &hf_rnsap_refBeta,
      { "refBeta", "rnsap.refBeta",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_MACdFlows_Information_TDD_item,
      { "E-DCH-MACdFlow-InfoTDDItem", "rnsap.E_DCH_MACdFlow_InfoTDDItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tnlQos,
      { "tnlQos", "rnsap.tnlQos",
        FT_UINT32, BASE_DEC, VALS(rnsap_TnlQos_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_eDCH_Grant_TypeTDD,
      { "eDCH-Grant-TypeTDD", "rnsap.eDCH_Grant_TypeTDD",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_DCH_Grant_TypeTDD_vals), 0,
        "E_DCH_Grant_TypeTDD", HFILL }},
    { &hf_rnsap_e_DCH_TDD_Maximum_Bitrate,
      { "e-DCH-TDD-Maximum-Bitrate", "rnsap.e_DCH_TDD_Maximum_Bitrate",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_PowerOffset_for_SchedulingInfo,
      { "e-DCH-PowerOffset-for-SchedulingInfo", "rnsap.e_DCH_PowerOffset_for_SchedulingInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_MACdFlows_to_Add,
      { "e-DCH-MACdFlows-to-Add", "rnsap.e_DCH_MACdFlows_to_Add",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_MACdFlows_Information_TDD", HFILL }},
    { &hf_rnsap_e_DCH_MACdFlows_to_Delete,
      { "e-DCH-MACdFlows-to-Delete", "rnsap.e_DCH_MACdFlows_to_Delete",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EDCH_MACdFlows_To_Delete", HFILL }},
    { &hf_rnsap_e_DCH_Non_Scheduled_Grant_Info,
      { "e-DCH-Non-Scheduled-Grant-Info", "rnsap.e_DCH_Non_Scheduled_Grant_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_TDD_Information_to_Modify,
      { "e-DCH-TDD-Information-to-Modify", "rnsap.e_DCH_TDD_Information_to_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_TDD_Information_to_Modify_List,
      { "e-DCH-TDD-Information-to-Modify-List", "rnsap.e_DCH_TDD_Information_to_Modify_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_DCH_TDD_Information_to_Modify_List_item,
      { "E-DCH-MACdFlow-ModifyTDDItem", "rnsap.E_DCH_MACdFlow_ModifyTDDItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_LogicalChannelToAdd,
      { "e-DCH-LogicalChannelToAdd", "rnsap.e_DCH_LogicalChannelToAdd",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_LogicalChannelInformation", HFILL }},
    { &hf_rnsap_e_DCH_LogicalChannelToModify,
      { "e-DCH-LogicalChannelToModify", "rnsap.e_DCH_LogicalChannelToModify",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_LogicalChannelToDelete,
      { "e-DCH-LogicalChannelToDelete", "rnsap.e_DCH_LogicalChannelToDelete",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_TDD_MACdFlow_Specific_InformationResp,
      { "e-DCH-TDD-MACdFlow-Specific-InformationResp", "rnsap.e_DCH_TDD_MACdFlow_Specific_InformationResp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_AGCH_Specific_Information_ResponseTDD,
      { "e-AGCH-Specific-Information-ResponseTDD", "rnsap.e_AGCH_Specific_Information_ResponseTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_AGCH_Specific_InformationRespListTDD", HFILL }},
    { &hf_rnsap_e_HICH_Information_Response,
      { "e-HICH-Information-Response", "rnsap.e_HICH_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "E_HICH_InformationResp", HFILL }},
    { &hf_rnsap_e_RNTI,
      { "e-RNTI", "rnsap.e_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mode1,
      { "mode1", "rnsap.mode1_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Mode1_Ind", HFILL }},
    { &hf_rnsap_mode2,
      { "mode2", "rnsap.mode2",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CFN", HFILL }},
    { &hf_rnsap_mode1_01,
      { "mode1", "rnsap.mode1",
        FT_UINT32, BASE_DEC, VALS(rnsap_ActivationDelay_vals), 0,
        "ActivationDelay", HFILL }},
    { &hf_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp_item,
      { "E-DCH-TDD-MACdFlow-Specific-InformationResp-Item", "rnsap.E_DCH_TDD_MACdFlow_Specific_InformationResp_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_MacdFlow_Id,
      { "e-DCH-MacdFlow-Id", "rnsap.e_DCH_MacdFlow_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EDCH_MACdFlow_ID", HFILL }},
    { &hf_rnsap_E_AGCH_Specific_InformationRespListTDD_item,
      { "E-AGCH-Specific-InformationResp-ItemTDD", "rnsap.E_AGCH_Specific_InformationResp_ItemTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_timeslot,
      { "timeslot", "rnsap.timeslot",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tDD_ChannelisationCode,
      { "tDD-ChannelisationCode", "rnsap.tDD_ChannelisationCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_TDD_ChannelisationCode_vals_ext, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_HICH_TimeOffset,
      { "e-HICH-TimeOffset", "rnsap.e_HICH_TimeOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_timeslotResource,
      { "timeslotResource", "rnsap.timeslotResource",
        FT_BYTES, BASE_NONE, NULL, 0,
        "E_DCH_TimeslotResource", HFILL }},
    { &hf_rnsap_powerResource,
      { "powerResource", "rnsap.powerResource",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_DCH_PowerResource", HFILL }},
    { &hf_rnsap_repetitionPeriod,
      { "repetitionPeriod", "rnsap.repetitionPeriod",
        FT_UINT32, BASE_DEC, VALS(rnsap_RepetitionPeriod_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_repetitionLength,
      { "repetitionLength", "rnsap.repetitionLength",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tddE_PUCH_Offset,
      { "tddE-PUCH-Offset", "rnsap.tddE_PUCH_Offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tdd_ChannelisationCode,
      { "tdd-ChannelisationCode", "rnsap.tdd_ChannelisationCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_TDD_ChannelisationCode_vals_ext, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_TDD_Information768,
      { "e-DCH-TDD-Information768", "rnsap.e_DCH_TDD_Information768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_TDD_Maximum_Bitrate768,
      { "e-DCH-TDD-Maximum-Bitrate768", "rnsap.e_DCH_TDD_Maximum_Bitrate768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_Non_Scheduled_Grant_Info768,
      { "e-DCH-Non-Scheduled-Grant-Info768", "rnsap.e_DCH_Non_Scheduled_Grant_Info768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_AGCH_Specific_Information_Response768TDD,
      { "e-AGCH-Specific-Information-Response768TDD", "rnsap.e_AGCH_Specific_Information_Response768TDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_AGCH_Specific_InformationRespList768TDD", HFILL }},
    { &hf_rnsap_e_HICH_Information_Response768,
      { "e-HICH-Information-Response768", "rnsap.e_HICH_Information_Response768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "E_HICH_InformationResp768", HFILL }},
    { &hf_rnsap_E_AGCH_Specific_InformationRespList768TDD_item,
      { "E-AGCH-Specific-InformationResp-Item768TDD", "rnsap.E_AGCH_Specific_InformationResp_Item768TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tDD_ChannelisationCode768,
      { "tDD-ChannelisationCode768", "rnsap.tDD_ChannelisationCode768",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_TDD_ChannelisationCode768_vals_ext, 0,
        NULL, HFILL }},
    { &hf_rnsap_tdd_ChannelisationCode768,
      { "tdd-ChannelisationCode768", "rnsap.tdd_ChannelisationCode768",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_TDD_ChannelisationCode768_vals_ext, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_PUCH_LCR_Information,
      { "e-PUCH-LCR-Information", "rnsap.e_PUCH_LCR_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_LCR_TDD_Information,
      { "e-DCH-LCR-TDD-Information", "rnsap.e_DCH_LCR_TDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_pRxdesBase,
      { "pRxdesBase", "rnsap.pRxdesBase",
        FT_INT32, BASE_DEC, NULL, 0,
        "E_PUCH_PRXdesBase", HFILL }},
    { &hf_rnsap_e_PUCH_TPC_Step_Size,
      { "e-PUCH-TPC-Step-Size", "rnsap.e_PUCH_TPC_Step_Size",
        FT_UINT32, BASE_DEC, VALS(rnsap_TDD_TPC_UplinkStepSize_LCR_vals), 0,
        "TDD_TPC_UplinkStepSize_LCR", HFILL }},
    { &hf_rnsap_n_E_UCCH_LCR,
      { "n-E-UCCH-LCR", "rnsap.n_E_UCCH_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_Physical_Layer_Category_LCR,
      { "e-DCH-Physical-Layer-Category-LCR", "rnsap.e_DCH_Physical_Layer_Category_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_AGCH_Specific_Information_Response_LCR_TDD,
      { "e-AGCH-Specific-Information-Response-LCR-TDD", "rnsap.e_AGCH_Specific_Information_Response_LCR_TDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_AGCH_Specific_InformationRespList_LCR_TDD", HFILL }},
    { &hf_rnsap_e_HICH_Specific_Information_Response_LCR,
      { "e-HICH-Specific-Information-Response-LCR", "rnsap.e_HICH_Specific_Information_Response_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "E_HICH_Specific_InformationResp_LCR", HFILL }},
    { &hf_rnsap_e_DCH_Non_Scheduled_Grant_Info_LCR,
      { "e-DCH-Non-Scheduled-Grant-Info-LCR", "rnsap.e_DCH_Non_Scheduled_Grant_Info_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_AGCH_Specific_InformationRespList_LCR_TDD_item,
      { "E-AGCH-Specific-InformationResp-Item-LCR-TDD", "rnsap.E_AGCH_Specific_InformationResp_Item_LCR_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_HICH_Scheduled_InformationResp_LCR,
      { "e-HICH-Scheduled-InformationResp-LCR", "rnsap.e_HICH_Scheduled_InformationResp_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_HICH_Scheduled_InformationRespList_LCR_TDD", HFILL }},
    { &hf_rnsap_e_HICH_non_Scheduled_InformationResp_LCR,
      { "e-HICH-non-Scheduled-InformationResp-LCR", "rnsap.e_HICH_non_Scheduled_InformationResp_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "E_HICH_InformationResp_LCR", HFILL }},
    { &hf_rnsap_e_HICH_TimeOffset_lcr,
      { "e-HICH-TimeOffset-lcr", "rnsap.e_HICH_TimeOffset_lcr",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_E_HICH_Scheduled_InformationRespList_LCR_TDD_item,
      { "E-HICH-Scheduled-InformationResp-Item-LCR-TDD", "rnsap.E_HICH_Scheduled_InformationResp_Item_LCR_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_HICH_EI,
      { "e-HICH-EI", "rnsap.e_HICH_EI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_HICH_Scheduled_InformationResp_LCR_01,
      { "e-HICH-Scheduled-InformationResp-LCR", "rnsap.e_HICH_Scheduled_InformationResp_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "E_HICH_InformationResp_LCR", HFILL }},
    { &hf_rnsap_signatureSequenceGroupIndex,
      { "signatureSequenceGroupIndex", "rnsap.signatureSequenceGroupIndex",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_timeslotResource_LCR,
      { "timeslotResource-LCR", "rnsap.timeslotResource_LCR",
        FT_BYTES, BASE_NONE, NULL, 0,
        "E_DCH_TimeslotResource_LCR", HFILL }},
    { &hf_rnsap_subframenumber,
      { "subframenumber", "rnsap.subframenumber",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_DCH_SubframeNumber_LCR_vals), 0,
        "E_DCH_SubframeNumber_LCR", HFILL }},
    { &hf_rnsap_Radio_Links_without_DPCH_FDPCH_Indication_item,
      { "Radio-Links-without-DPCH-FDPCH-Information", "rnsap.Radio_Links_without_DPCH_FDPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_radio_Links_without_DPCH_FDPCH_Operation_Indicator,
      { "radio-Links-without-DPCH-FDPCH-Operation-Indicator", "rnsap.radio_Links_without_DPCH_FDPCH_Operation_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_radio_Links_without_DPCH_FDPCH_Operation_Indicator_vals), 0,
        "T_radio_Links_without_DPCH_FDPCH_Operation_Indicator", HFILL }},
    { &hf_rnsap_extended_sRNTI,
      { "extended-sRNTI", "rnsap.extended_sRNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Extended_RNTI", HFILL }},
    { &hf_rnsap_extended_sRNTI_BitMaskIndex,
      { "extended-sRNTI-BitMaskIndex", "rnsap.extended_sRNTI_BitMaskIndex",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_extended_sRNTI_BitMaskIndex_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_ExtendedAffectedUEInformationForMBMS_item,
      { "Extended-RNTI", "rnsap.Extended_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_repetition_Period_List_LCR,
      { "repetition-Period-List-LCR", "rnsap.repetition_Period_List_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_SPS_Indicator,
      { "e-DCH-SPS-Indicator", "rnsap.e_DCH_SPS_Indicator",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_SPS_Reservation_Indicator,
      { "e-DCH-SPS-Reservation-Indicator", "rnsap.e_DCH_SPS_Reservation_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SPS_Reservation_Indicator_vals), 0,
        "SPS_Reservation_Indicator", HFILL }},
    { &hf_rnsap_initial_E_DCH_SPS_resource,
      { "initial-E-DCH-SPS-resource", "rnsap.initial_E_DCH_SPS_resource_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_e_DCH_SPS_HICH_Information,
      { "e-DCH-SPS-HICH-Information", "rnsap.e_DCH_SPS_HICH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_timeslot_Resource_Related_Information,
      { "timeslot-Resource-Related-Information", "rnsap.timeslot_Resource_Related_Information",
        FT_BYTES, BASE_NONE, NULL, 0,
        "E_DCH_TimeslotResource_LCR", HFILL }},
    { &hf_rnsap_repetitionPeriodIndex,
      { "repetitionPeriodIndex", "rnsap.repetitionPeriodIndex",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_subframeNumber,
      { "subframeNumber", "rnsap.subframeNumber",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_subframeNumber_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_n_E_UCCHLCR,
      { "n-E-UCCHLCR", "rnsap.n_E_UCCHLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "N_E_UCCH_LCR", HFILL }},
    { &hf_rnsap_e_HICH_Configuration,
      { "e-HICH-Configuration", "rnsap.e_HICH_Configuration",
        FT_UINT32, BASE_DEC, VALS(rnsap_E_HICH_Configuration_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_same_As_Scheduled_E_HICH,
      { "same-As-Scheduled-E-HICH", "rnsap.same_As_Scheduled_E_HICH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_explicit,
      { "explicit", "rnsap.explicit_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "E_HICH_InformationResp_ExplicitConfiguration_LCR", HFILL }},
    { &hf_rnsap_FACH_FlowControlInformation_item,
      { "FACH-FlowControlInformationItem", "rnsap.FACH_FlowControlInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_fACH_SchedulingPriority,
      { "fACH-SchedulingPriority", "rnsap.fACH_SchedulingPriority",
        FT_UINT32, BASE_DEC, VALS(rnsap_SchedulingPriorityIndicator_vals), 0,
        "SchedulingPriorityIndicator", HFILL }},
    { &hf_rnsap_fACH_InitialWindowSize,
      { "fACH-InitialWindowSize", "rnsap.fACH_InitialWindowSize",
        FT_UINT32, BASE_DEC, VALS(rnsap_FACH_InitialWindowSize_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_FACH_InformationList_item,
      { "FACH-InformationItem", "rnsap.FACH_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_FDD_DCHs_to_Modify_item,
      { "FDD-DCHs-to-ModifyItem", "rnsap.FDD_DCHs_to_ModifyItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dCH_SpecificInformationList_02,
      { "dCH-SpecificInformationList", "rnsap.dCH_SpecificInformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FDD_DCHs_to_ModifySpecificInformationList", HFILL }},
    { &hf_rnsap_FDD_DCHs_to_ModifySpecificInformationList_item,
      { "FDD-DCHs-to-ModifySpecificItem", "rnsap.FDD_DCHs_to_ModifySpecificItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_TransportformatSet,
      { "ul-TransportformatSet", "rnsap.ul_TransportformatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TransportFormatSet", HFILL }},
    { &hf_rnsap_dl_TransportformatSet,
      { "dl-TransportformatSet", "rnsap.dl_TransportformatSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TransportFormatSet", HFILL }},
    { &hf_rnsap_not_Used_dRACControl,
      { "not-Used-dRACControl", "rnsap.not_Used_dRACControl_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_FDD_DL_CodeInformation_item,
      { "FDD-DL-CodeInformationItem", "rnsap.FDD_DL_CodeInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_ScramblingCode,
      { "dl-ScramblingCode", "rnsap.dl_ScramblingCode",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_fDD_DL_ChannelisationCodeNumber,
      { "fDD-DL-ChannelisationCodeNumber", "rnsap.fDD_DL_ChannelisationCodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_transmission_Gap_Pattern_Sequence_ScramblingCode_Information,
      { "transmission-Gap-Pattern-Sequence-ScramblingCode-Information", "rnsap.transmission_Gap_Pattern_Sequence_ScramblingCode_Information",
        FT_UINT32, BASE_DEC, VALS(rnsap_Transmission_Gap_Pattern_Sequence_ScramblingCode_Information_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_fdd_dl_ChannelisationCodeNumber,
      { "fdd-dl-ChannelisationCodeNumber", "rnsap.fdd_dl_ChannelisationCodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_extended_E_DPCCH_PO,
      { "extended-E-DPCCH-PO", "rnsap.extended_E_DPCCH_PO",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tDD_ChannelisationCodeLCR,
      { "tDD-ChannelisationCodeLCR", "rnsap.tDD_ChannelisationCodeLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_wT,
      { "wT", "rnsap.wT",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_4", HFILL }},
    { &hf_rnsap_fTPICH_Offset,
      { "fTPICH-Offset", "rnsap.fTPICH_Offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information,
      { "setup-Or-ConfigurationChange-Or-Removal-Of-FTPICH-Information", "rnsap.setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information",
        FT_UINT32, BASE_DEC, VALS(rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_fTPICH_SlotFormat,
      { "fTPICH-SlotFormat", "rnsap.fTPICH_SlotFormat",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_fTPICH_ChannelisationCodenumber,
      { "fTPICH-ChannelisationCodenumber", "rnsap.fTPICH_ChannelisationCodenumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FDD_DL_ChannelisationCodeNumber", HFILL }},
    { &hf_rnsap_GA_Cell_item,
      { "GA-Cell item", "rnsap.GA_Cell_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cell_GAIgeographicalCoordinate,
      { "cell-GAIgeographicalCoordinate", "rnsap.cell_GAIgeographicalCoordinate_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeographicalCoordinate", HFILL }},
    { &hf_rnsap_pointWithUncertainty,
      { "pointWithUncertainty", "rnsap.pointWithUncertainty_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GA_PointWithUnCertainty", HFILL }},
    { &hf_rnsap_pointWithUncertaintyEllipse,
      { "pointWithUncertaintyEllipse", "rnsap.pointWithUncertaintyEllipse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GA_PointWithUnCertaintyEllipse", HFILL }},
    { &hf_rnsap_pointWithAltitude,
      { "pointWithAltitude", "rnsap.pointWithAltitude_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GA_PointWithAltitude", HFILL }},
    { &hf_rnsap_pointWithAltitudeAndUncertaintyEllipsoid,
      { "pointWithAltitudeAndUncertaintyEllipsoid", "rnsap.pointWithAltitudeAndUncertaintyEllipsoid_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GA_PointWithAltitudeAndUncertaintyEllipsoid", HFILL }},
    { &hf_rnsap_ellipsoidArc,
      { "ellipsoidArc", "rnsap.ellipsoidArc_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GA_EllipsoidArc", HFILL }},
    { &hf_rnsap_directionOfAltitude,
      { "directionOfAltitude", "rnsap.directionOfAltitude",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_directionOfAltitude_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_altitude,
      { "altitude", "rnsap.altitude",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_32767", HFILL }},
    { &hf_rnsap_geographicalCoordinates,
      { "geographicalCoordinates", "rnsap.geographicalCoordinates_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeographicalCoordinate", HFILL }},
    { &hf_rnsap_innerRadius,
      { "innerRadius", "rnsap.innerRadius",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_65535", HFILL }},
    { &hf_rnsap_uncertaintyRadius,
      { "uncertaintyRadius", "rnsap.uncertaintyRadius",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_127", HFILL }},
    { &hf_rnsap_offsetAngle,
      { "offsetAngle", "rnsap.offsetAngle",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_179", HFILL }},
    { &hf_rnsap_includedAngle,
      { "includedAngle", "rnsap.includedAngle",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_179", HFILL }},
    { &hf_rnsap_confidence,
      { "confidence", "rnsap.confidence",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_127", HFILL }},
    { &hf_rnsap_signalledGainFactors10ms,
      { "signalledGainFactors10ms", "rnsap.signalledGainFactors10ms_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gain10ms_betaC,
      { "gain10ms-betaC", "rnsap.gain10ms_betaC",
        FT_UINT32, BASE_DEC, NULL, 0,
        "BetaCD", HFILL }},
    { &hf_rnsap_gain10ms_betaD,
      { "gain10ms-betaD", "rnsap.gain10ms_betaD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "BetaCD", HFILL }},
    { &hf_rnsap_gain10ms_refTFCNumber,
      { "gain10ms-refTFCNumber", "rnsap.gain10ms_refTFCNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RefTFCNumber", HFILL }},
    { &hf_rnsap_computedGainFactors10ms,
      { "computedGainFactors10ms", "rnsap.computedGainFactors10ms",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RefTFCNumber", HFILL }},
    { &hf_rnsap_navClockModel,
      { "navClockModel", "rnsap.navClockModel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_NAVclockModel", HFILL }},
    { &hf_rnsap_cnavClockModel,
      { "cnavClockModel", "rnsap.cnavClockModel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_CNAVclockModel", HFILL }},
    { &hf_rnsap_glonassClockModel,
      { "glonassClockModel", "rnsap.glonassClockModel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_GLONASSclockModel", HFILL }},
    { &hf_rnsap_sbasClockModel,
      { "sbasClockModel", "rnsap.sbasClockModel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_SBASclockModel", HFILL }},
    { &hf_rnsap_extension_GANSS_AddClockModels,
      { "extension-GANSS-AddClockModels", "rnsap.extension_GANSS_AddClockModels_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_navKeplerianSet,
      { "navKeplerianSet", "rnsap.navKeplerianSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_NavModel_NAVKeplerianSet", HFILL }},
    { &hf_rnsap_cnavKeplerianSet,
      { "cnavKeplerianSet", "rnsap.cnavKeplerianSet_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_NavModel_CNAVKeplerianSet", HFILL }},
    { &hf_rnsap_glonassECEF,
      { "glonassECEF", "rnsap.glonassECEF_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_NavModel_GLONASSecef", HFILL }},
    { &hf_rnsap_sbasECEF,
      { "sbasECEF", "rnsap.sbasECEF_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_NavModel_SBASecef", HFILL }},
    { &hf_rnsap_extension_GANSS_AddOrbitModels,
      { "extension-GANSS-AddOrbitModels", "rnsap.extension_GANSS_AddOrbitModels_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dataID,
      { "dataID", "rnsap.dataID",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_rnsap_alpha_beta_parameters,
      { "alpha-beta-parameters", "rnsap.alpha_beta_parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GPS_Ionospheric_Model", HFILL }},
    { &hf_rnsap_ganss_Transmission_Time,
      { "ganss-Transmission-Time", "rnsap.ganss_Transmission_Time_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_non_broadcastIndication,
      { "non-broadcastIndication", "rnsap.non_broadcastIndication",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_non_broadcastIndication_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_ganssSatInfoNavList,
      { "ganssSatInfoNavList", "rnsap.ganssSatInfoNavList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Ganss_Sat_Info_AddNavList", HFILL }},
    { &hf_rnsap_GANSS_Additional_Time_Models_item,
      { "GANSS-Time-Model", "rnsap.GANSS_Time_Model_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_utcModel1,
      { "utcModel1", "rnsap.utcModel1_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_UTCmodelSet1", HFILL }},
    { &hf_rnsap_utcModel2,
      { "utcModel2", "rnsap.utcModel2_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_UTCmodelSet2", HFILL }},
    { &hf_rnsap_utcModel3,
      { "utcModel3", "rnsap.utcModel3_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_UTCmodelSet3", HFILL }},
    { &hf_rnsap_extension_GANSS_Additional_UTC_Models,
      { "extension-GANSS-Additional-UTC-Models", "rnsap.extension_GANSS_Additional_UTC_Models_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_wk_number,
      { "ganss-wk-number", "rnsap.ganss_wk_number",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_255", HFILL }},
    { &hf_rnsap_gANSS_AlmanacModel,
      { "gANSS-AlmanacModel", "rnsap.gANSS_AlmanacModel",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_gANSS_AlmanacModel_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_gANSS_keplerianParameters,
      { "gANSS-keplerianParameters", "rnsap.gANSS_keplerianParameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_t_oa,
      { "t-oa", "rnsap.t_oa",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_1023", HFILL }},
    { &hf_rnsap_iod_a,
      { "iod-a", "rnsap.iod_a",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_15", HFILL }},
    { &hf_rnsap_gANSS_SatelliteInformationKP,
      { "gANSS-SatelliteInformationKP", "rnsap.gANSS_SatelliteInformationKP",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_extension_GANSS_AlmanacModel,
      { "extension-GANSS-AlmanacModel", "rnsap.extension_GANSS_AlmanacModel_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_t_oa_01,
      { "t-oa", "rnsap.t_oa",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_255", HFILL }},
    { &hf_rnsap_sat_info_NAVkpList,
      { "sat-info-NAVkpList", "rnsap.sat_info_NAVkpList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_SAT_Info_Almanac_NAVkpList", HFILL }},
    { &hf_rnsap_sat_info_BDSkpList,
      { "sat-info-BDSkpList", "rnsap.sat_info_BDSkpList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_SAT_Info_Almanac_BDSkpList", HFILL }},
    { &hf_rnsap_sat_info_REDkpList,
      { "sat-info-REDkpList", "rnsap.sat_info_REDkpList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_SAT_Info_Almanac_REDkpList", HFILL }},
    { &hf_rnsap_sat_info_MIDIkpList,
      { "sat-info-MIDIkpList", "rnsap.sat_info_MIDIkpList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_SAT_Info_Almanac_MIDIkpList", HFILL }},
    { &hf_rnsap_sat_info_GLOkpList,
      { "sat-info-GLOkpList", "rnsap.sat_info_GLOkpList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_SAT_Info_Almanac_GLOkpList", HFILL }},
    { &hf_rnsap_sat_info_SBASecefList,
      { "sat-info-SBASecefList", "rnsap.sat_info_SBASecefList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_SAT_Info_Almanac_SBASecefList", HFILL }},
    { &hf_rnsap_ganssID1,
      { "ganssID1", "rnsap.ganssID1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_AuxInfoGANSS_ID1", HFILL }},
    { &hf_rnsap_ganssID3,
      { "ganssID3", "rnsap.ganssID3",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_AuxInfoGANSS_ID3", HFILL }},
    { &hf_rnsap_GANSS_AuxInfoGANSS_ID1_item,
      { "GANSS-AuxInfoGANSS-ID1 item", "rnsap.GANSS_AuxInfoGANSS_ID1_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_signalsAvailable,
      { "signalsAvailable", "rnsap.signalsAvailable",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_GANSS_AuxInfoGANSS_ID3_item,
      { "GANSS-AuxInfoGANSS-ID3 item", "rnsap.GANSS_AuxInfoGANSS_ID3_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_channelNumber,
      { "channelNumber", "rnsap.channelNumber",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M7_13", HFILL }},
    { &hf_rnsap_bdsToc,
      { "bdsToc", "rnsap.bdsToc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_rnsap_bdsa0,
      { "bdsa0", "rnsap.bdsa0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_bdsa1,
      { "bdsa1", "rnsap.bdsa1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_22", HFILL }},
    { &hf_rnsap_bdsa2,
      { "bdsa2", "rnsap.bdsa2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_bdsTgd1,
      { "bdsTgd1", "rnsap.bdsTgd1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_rnsap_bdsAODC,
      { "bdsAODC", "rnsap.bdsAODC",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_rnsap_GANSS_Clock_Model_item,
      { "GANSS-Clock-Model item", "rnsap.GANSS_Clock_Model_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_t_oc,
      { "t-oc", "rnsap.t_oc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_14", HFILL }},
    { &hf_rnsap_a_i2,
      { "a-i2", "rnsap.a_i2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_6", HFILL }},
    { &hf_rnsap_a_i1,
      { "a-i1", "rnsap.a_i1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_21", HFILL }},
    { &hf_rnsap_a_i0,
      { "a-i0", "rnsap.a_i0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_31", HFILL }},
    { &hf_rnsap_t_gd,
      { "t-gd", "rnsap.t_gd",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_rnsap_sisa,
      { "sisa", "rnsap.sisa",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_model_id,
      { "model-id", "rnsap.model_id",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_1_", HFILL }},
    { &hf_rnsap_cnavToc,
      { "cnavToc", "rnsap.cnavToc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_cnavTop,
      { "cnavTop", "rnsap.cnavTop",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_cnavURA0,
      { "cnavURA0", "rnsap.cnavURA0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_rnsap_cnavURA1,
      { "cnavURA1", "rnsap.cnavURA1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_3", HFILL }},
    { &hf_rnsap_cnavURA2,
      { "cnavURA2", "rnsap.cnavURA2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_3", HFILL }},
    { &hf_rnsap_cnavAf2,
      { "cnavAf2", "rnsap.cnavAf2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_rnsap_cnavAf1,
      { "cnavAf1", "rnsap.cnavAf1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_20", HFILL }},
    { &hf_rnsap_cnavAf0,
      { "cnavAf0", "rnsap.cnavAf0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_26", HFILL }},
    { &hf_rnsap_cnavTgd,
      { "cnavTgd", "rnsap.cnavTgd",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_rnsap_cnavISCl1cp,
      { "cnavISCl1cp", "rnsap.cnavISCl1cp",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_rnsap_cnavISCl1cd,
      { "cnavISCl1cd", "rnsap.cnavISCl1cd",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_rnsap_cnavISCl1ca,
      { "cnavISCl1ca", "rnsap.cnavISCl1ca",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_rnsap_cnavISCl2c,
      { "cnavISCl2c", "rnsap.cnavISCl2c",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_rnsap_cnavISCl5i5,
      { "cnavISCl5i5", "rnsap.cnavISCl5i5",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_rnsap_cnavISCl5q5,
      { "cnavISCl5q5", "rnsap.cnavISCl5q5",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_rnsap_ganss_Ionospheric_Model,
      { "ganss-Ionospheric-Model", "rnsap.ganss_Ionospheric_Model_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_Rx_Pos,
      { "ganss-Rx-Pos", "rnsap.ganss_Rx_Pos_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ionospheric_Model,
      { "ionospheric-Model", "rnsap.ionospheric_Model",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_rnsap_ganssTod,
      { "ganssTod", "rnsap.ganssTod",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_59_", HFILL }},
    { &hf_rnsap_dataBitAssistancelist,
      { "dataBitAssistancelist", "rnsap.dataBitAssistancelist",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_DataBitAssistanceList", HFILL }},
    { &hf_rnsap_GANSS_DataBitAssistanceList_item,
      { "GANSS-DataBitAssistanceItem", "rnsap.GANSS_DataBitAssistanceItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dataBitAssistanceSgnList,
      { "dataBitAssistanceSgnList", "rnsap.dataBitAssistanceSgnList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_DataBitAssistanceSgnList", HFILL }},
    { &hf_rnsap_GANSS_DataBitAssistanceSgnList_item,
      { "GANSS-DataBitAssistanceSgnItem", "rnsap.GANSS_DataBitAssistanceSgnItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_SignalId,
      { "ganss-SignalId", "rnsap.ganss_SignalId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_Signal_ID", HFILL }},
    { &hf_rnsap_ganssDataBits,
      { "ganssDataBits", "rnsap.ganssDataBits",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1_1024", HFILL }},
    { &hf_rnsap_ganssTod_01,
      { "ganssTod", "rnsap.ganssTod",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_86399", HFILL }},
    { &hf_rnsap_ganss_Data_Bit_Assistance_ReqList,
      { "ganss-Data-Bit-Assistance-ReqList", "rnsap.ganss_Data_Bit_Assistance_ReqList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_DataBitInterval,
      { "ganss-DataBitInterval", "rnsap.ganss_DataBitInterval",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_15", HFILL }},
    { &hf_rnsap_ganss_SatelliteInfo,
      { "ganss-SatelliteInfo", "rnsap.ganss_SatelliteInfo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_SatelliteInfo_item,
      { "ganss-SatelliteInfo item", "rnsap.ganss_SatelliteInfo_item",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_63", HFILL }},
    { &hf_rnsap_b1,
      { "b1", "rnsap.b1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_b2,
      { "b2", "rnsap.b2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_rnsap_teop,
      { "teop", "rnsap.teop",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_pmX,
      { "pmX", "rnsap.pmX",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_21", HFILL }},
    { &hf_rnsap_pmXdot,
      { "pmXdot", "rnsap.pmXdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_15", HFILL }},
    { &hf_rnsap_pmY,
      { "pmY", "rnsap.pmY",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_21", HFILL }},
    { &hf_rnsap_pmYdot,
      { "pmYdot", "rnsap.pmYdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_15", HFILL }},
    { &hf_rnsap_deltaUT1,
      { "deltaUT1", "rnsap.deltaUT1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_31", HFILL }},
    { &hf_rnsap_deltaUT1dot,
      { "deltaUT1dot", "rnsap.deltaUT1dot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_19", HFILL }},
    { &hf_rnsap_GANSS_GenericDataInfoReqList_item,
      { "GANSS-GenericDataInfoReqItem", "rnsap.GANSS_GenericDataInfoReqItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_Id,
      { "ganss-Id", "rnsap.ganss_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_Navigation_Model_And_Time_Recovery,
      { "ganss-Navigation-Model-And-Time-Recovery", "rnsap.ganss_Navigation_Model_And_Time_Recovery",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_rnsap_ganss_Time_Model_GNSS_GNSS,
      { "ganss-Time-Model-GNSS-GNSS", "rnsap.ganss_Time_Model_GNSS_GNSS",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_9", HFILL }},
    { &hf_rnsap_ganss_UTC_Model,
      { "ganss-UTC-Model", "rnsap.ganss_UTC_Model",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_rnsap_ganss_Almanac,
      { "ganss-Almanac", "rnsap.ganss_Almanac",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_rnsap_ganss_Real_Time_Integrity,
      { "ganss-Real-Time-Integrity", "rnsap.ganss_Real_Time_Integrity",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_rnsap_ganss_Data_Bit_Assistance_Req,
      { "ganss-Data-Bit-Assistance-Req", "rnsap.ganss_Data_Bit_Assistance_Req_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_Data_Bit_Assistance_ReqItem", HFILL }},
    { &hf_rnsap_GANSS_Generic_Data_item,
      { "GANSS-Generic-DataItem", "rnsap.GANSS_Generic_DataItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dganss_Correction,
      { "dganss-Correction", "rnsap.dganss_Correction_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DGANSSCorrections", HFILL }},
    { &hf_rnsap_ganss_Navigation_Model_And_Time_Recovery_01,
      { "ganss-Navigation-Model-And-Time-Recovery", "rnsap.ganss_Navigation_Model_And_Time_Recovery_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_Time_Model,
      { "ganss-Time-Model", "rnsap.ganss_Time_Model_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_UTC_TIME,
      { "ganss-UTC-TIME", "rnsap.ganss_UTC_TIME_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_UTC_Model", HFILL }},
    { &hf_rnsap_ganss_Almanac_01,
      { "ganss-Almanac", "rnsap.ganss_Almanac_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_Real_Time_Integrity_01,
      { "ganss-Real-Time-Integrity", "rnsap.ganss_Real_Time_Integrity",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_Data_Bit_Assistance,
      { "ganss-Data-Bit-Assistance", "rnsap.ganss_Data_Bit_Assistance_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gloTau,
      { "gloTau", "rnsap.gloTau",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_22", HFILL }},
    { &hf_rnsap_gloGamma,
      { "gloGamma", "rnsap.gloGamma",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_gloDeltaTau,
      { "gloDeltaTau", "rnsap.gloDeltaTau",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_rnsap_gANSS_CommonDataInfoReq,
      { "gANSS-CommonDataInfoReq", "rnsap.gANSS_CommonDataInfoReq_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gANSS_GenericDataInfoReqList,
      { "gANSS-GenericDataInfoReqList", "rnsap.gANSS_GenericDataInfoReqList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_alpha_zero_ionos,
      { "alpha-zero-ionos", "rnsap.alpha_zero_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_alpha_one_ionos,
      { "alpha-one-ionos", "rnsap.alpha_one_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_alpha_two_ionos,
      { "alpha-two-ionos", "rnsap.alpha_two_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_14", HFILL }},
    { &hf_rnsap_gANSS_IonosphereRegionalStormFlags,
      { "gANSS-IonosphereRegionalStormFlags", "rnsap.gANSS_IonosphereRegionalStormFlags_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_storm_flag_one,
      { "storm-flag-one", "rnsap.storm_flag_one",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_rnsap_storm_flag_two,
      { "storm-flag-two", "rnsap.storm_flag_two",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_rnsap_storm_flag_three,
      { "storm-flag-three", "rnsap.storm_flag_three",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_rnsap_storm_flag_four,
      { "storm-flag-four", "rnsap.storm_flag_four",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_rnsap_storm_flag_five,
      { "storm-flag-five", "rnsap.storm_flag_five",
        FT_BOOLEAN, BASE_NONE, NULL, 0,
        "BOOLEAN", HFILL }},
    { &hf_rnsap_navToc,
      { "navToc", "rnsap.navToc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_navaf2,
      { "navaf2", "rnsap.navaf2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_navaf1,
      { "navaf1", "rnsap.navaf1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_navaf0,
      { "navaf0", "rnsap.navaf0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_22", HFILL }},
    { &hf_rnsap_navTgd,
      { "navTgd", "rnsap.navTgd",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_non_broadcastIndication_01,
      { "non-broadcastIndication", "rnsap.non_broadcastIndication",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_non_broadcastIndication_01_vals), 0,
        "T_non_broadcastIndication_01", HFILL }},
    { &hf_rnsap_ganssSatInfoNav,
      { "ganssSatInfoNav", "rnsap.ganssSatInfoNav",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_Sat_Info_Nav", HFILL }},
    { &hf_rnsap_bdsURAI,
      { "bdsURAI", "rnsap.bdsURAI",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_rnsap_bdsToe,
      { "bdsToe", "rnsap.bdsToe",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_rnsap_bdsAPowerHalf,
      { "bdsAPowerHalf", "rnsap.bdsAPowerHalf",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_bdsE,
      { "bdsE", "rnsap.bdsE",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_bdsW,
      { "bdsW", "rnsap.bdsW",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_bdsDeltaN,
      { "bdsDeltaN", "rnsap.bdsDeltaN",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_bdsM0,
      { "bdsM0", "rnsap.bdsM0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_bdsOmega0,
      { "bdsOmega0", "rnsap.bdsOmega0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_bdsOmegaDot,
      { "bdsOmegaDot", "rnsap.bdsOmegaDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_bdsI0,
      { "bdsI0", "rnsap.bdsI0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_bdsIDot,
      { "bdsIDot", "rnsap.bdsIDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_14", HFILL }},
    { &hf_rnsap_bdsCuc,
      { "bdsCuc", "rnsap.bdsCuc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_18", HFILL }},
    { &hf_rnsap_bdsCus,
      { "bdsCus", "rnsap.bdsCus",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_18", HFILL }},
    { &hf_rnsap_bdsCrc,
      { "bdsCrc", "rnsap.bdsCrc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_18", HFILL }},
    { &hf_rnsap_bdsCrs,
      { "bdsCrs", "rnsap.bdsCrs",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_18", HFILL }},
    { &hf_rnsap_bdsCic,
      { "bdsCic", "rnsap.bdsCic",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_18", HFILL }},
    { &hf_rnsap_bdsCis,
      { "bdsCis", "rnsap.bdsCis",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_18", HFILL }},
    { &hf_rnsap_bdsAODE,
      { "bdsAODE", "rnsap.bdsAODE",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_rnsap_cnavURAindex,
      { "cnavURAindex", "rnsap.cnavURAindex",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_rnsap_cnavDeltaA,
      { "cnavDeltaA", "rnsap.cnavDeltaA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_26", HFILL }},
    { &hf_rnsap_cnavAdot,
      { "cnavAdot", "rnsap.cnavAdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_25", HFILL }},
    { &hf_rnsap_cnavDeltaNo,
      { "cnavDeltaNo", "rnsap.cnavDeltaNo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_rnsap_cnavDeltaNoDot,
      { "cnavDeltaNoDot", "rnsap.cnavDeltaNoDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_23", HFILL }},
    { &hf_rnsap_cnavMo,
      { "cnavMo", "rnsap.cnavMo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_33", HFILL }},
    { &hf_rnsap_cnavE,
      { "cnavE", "rnsap.cnavE",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_33", HFILL }},
    { &hf_rnsap_cnavOmega,
      { "cnavOmega", "rnsap.cnavOmega",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_33", HFILL }},
    { &hf_rnsap_cnavOMEGA0,
      { "cnavOMEGA0", "rnsap.cnavOMEGA0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_33", HFILL }},
    { &hf_rnsap_cnavDeltaOmegaDot,
      { "cnavDeltaOmegaDot", "rnsap.cnavDeltaOmegaDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_rnsap_cnavIo,
      { "cnavIo", "rnsap.cnavIo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_33", HFILL }},
    { &hf_rnsap_cnavIoDot,
      { "cnavIoDot", "rnsap.cnavIoDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_15", HFILL }},
    { &hf_rnsap_cnavCis,
      { "cnavCis", "rnsap.cnavCis",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_cnavCic,
      { "cnavCic", "rnsap.cnavCic",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_cnavCrs,
      { "cnavCrs", "rnsap.cnavCrs",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_cnavCrc,
      { "cnavCrc", "rnsap.cnavCrc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_cnavCus,
      { "cnavCus", "rnsap.cnavCus",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_21", HFILL }},
    { &hf_rnsap_cnavCuc,
      { "cnavCuc", "rnsap.cnavCuc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_21", HFILL }},
    { &hf_rnsap_gloEn,
      { "gloEn", "rnsap.gloEn",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_rnsap_gloP1,
      { "gloP1", "rnsap.gloP1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_rnsap_gloP2,
      { "gloP2", "rnsap.gloP2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_rnsap_gloM,
      { "gloM", "rnsap.gloM",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_rnsap_gloX,
      { "gloX", "rnsap.gloX",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_27", HFILL }},
    { &hf_rnsap_gloXdot,
      { "gloXdot", "rnsap.gloXdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_gloXdotdot,
      { "gloXdotdot", "rnsap.gloXdotdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_rnsap_gloY,
      { "gloY", "rnsap.gloY",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_27", HFILL }},
    { &hf_rnsap_gloYdot,
      { "gloYdot", "rnsap.gloYdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_gloYdotdot,
      { "gloYdotdot", "rnsap.gloYdotdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_rnsap_gloZ,
      { "gloZ", "rnsap.gloZ",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_27", HFILL }},
    { &hf_rnsap_gloZdot,
      { "gloZdot", "rnsap.gloZdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_gloZdotdot,
      { "gloZdotdot", "rnsap.gloZdotdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_rnsap_navURA,
      { "navURA", "rnsap.navURA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_rnsap_navFitFlag,
      { "navFitFlag", "rnsap.navFitFlag",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_rnsap_navToe,
      { "navToe", "rnsap.navToe",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_navOmega,
      { "navOmega", "rnsap.navOmega",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_navDeltaN,
      { "navDeltaN", "rnsap.navDeltaN",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_navM0,
      { "navM0", "rnsap.navM0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_navOmegaADot,
      { "navOmegaADot", "rnsap.navOmegaADot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_navE,
      { "navE", "rnsap.navE",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_navIDot,
      { "navIDot", "rnsap.navIDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_14", HFILL }},
    { &hf_rnsap_navAPowerHalf,
      { "navAPowerHalf", "rnsap.navAPowerHalf",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_navI0,
      { "navI0", "rnsap.navI0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_navOmegaA0,
      { "navOmegaA0", "rnsap.navOmegaA0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_navCrs,
      { "navCrs", "rnsap.navCrs",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_navCis,
      { "navCis", "rnsap.navCis",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_navCus,
      { "navCus", "rnsap.navCus",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_navCrc,
      { "navCrc", "rnsap.navCrc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_navCic,
      { "navCic", "rnsap.navCic",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_navCuc,
      { "navCuc", "rnsap.navCuc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_sbasTo,
      { "sbasTo", "rnsap.sbasTo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_rnsap_sbasAccuracy,
      { "sbasAccuracy", "rnsap.sbasAccuracy",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_rnsap_sbasXg,
      { "sbasXg", "rnsap.sbasXg",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_30", HFILL }},
    { &hf_rnsap_sbasYg,
      { "sbasYg", "rnsap.sbasYg",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_30", HFILL }},
    { &hf_rnsap_sbasZg,
      { "sbasZg", "rnsap.sbasZg",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_25", HFILL }},
    { &hf_rnsap_sbasXgDot,
      { "sbasXgDot", "rnsap.sbasXgDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_rnsap_sbasYgDot,
      { "sbasYgDot", "rnsap.sbasYgDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_rnsap_sbasZgDot,
      { "sbasZgDot", "rnsap.sbasZgDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_18", HFILL }},
    { &hf_rnsap_sbasXgDotDot,
      { "sbasXgDotDot", "rnsap.sbasXgDotDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_rnsap_sbagYgDotDot,
      { "sbagYgDotDot", "rnsap.sbagYgDotDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_rnsap_sbasZgDotDot,
      { "sbasZgDotDot", "rnsap.sbasZgDotDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_rnsap_gANSS_keplerianParameters_01,
      { "gANSS-keplerianParameters", "rnsap.gANSS_keplerianParameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "T_gANSS_keplerianParameters_01", HFILL }},
    { &hf_rnsap_toe_nav,
      { "toe-nav", "rnsap.toe_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_14", HFILL }},
    { &hf_rnsap_ganss_omega_nav,
      { "ganss-omega-nav", "rnsap.ganss_omega_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_delta_n_nav,
      { "delta-n-nav", "rnsap.delta_n_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_m_zero_nav,
      { "m-zero-nav", "rnsap.m_zero_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_omegadot_nav,
      { "omegadot-nav", "rnsap.omegadot_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_ganss_e_nav,
      { "ganss-e-nav", "rnsap.ganss_e_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_idot_nav,
      { "idot-nav", "rnsap.idot_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_14", HFILL }},
    { &hf_rnsap_a_sqrt_nav,
      { "a-sqrt-nav", "rnsap.a_sqrt_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_i_zero_nav,
      { "i-zero-nav", "rnsap.i_zero_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_omega_zero_nav,
      { "omega-zero-nav", "rnsap.omega_zero_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_c_rs_nav,
      { "c-rs-nav", "rnsap.c_rs_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_c_is_nav,
      { "c-is-nav", "rnsap.c_is_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_c_us_nav,
      { "c-us-nav", "rnsap.c_us_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_c_rc_nav,
      { "c-rc-nav", "rnsap.c_rc_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_c_ic_nav,
      { "c-ic-nav", "rnsap.c_ic_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_c_uc_nav,
      { "c-uc-nav", "rnsap.c_uc_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_GANSS_Real_Time_Integrity_item,
      { "GANSS-Real-Time-Integrity item", "rnsap.GANSS_Real_Time_Integrity_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_bad_ganss_satId,
      { "bad-ganss-satId", "rnsap.bad_ganss_satId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_63", HFILL }},
    { &hf_rnsap_bad_ganss_signalId,
      { "bad-ganss-signalId", "rnsap.bad_ganss_signalId",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_latitudeSign,
      { "latitudeSign", "rnsap.latitudeSign",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_latitudeSign_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_degreesOfLatitude,
      { "degreesOfLatitude", "rnsap.degreesOfLatitude",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_2147483647", HFILL }},
    { &hf_rnsap_degreesOfLongitude,
      { "degreesOfLongitude", "rnsap.degreesOfLongitude",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M2147483648_2147483647", HFILL }},
    { &hf_rnsap_directionOfAltitude_01,
      { "directionOfAltitude", "rnsap.directionOfAltitude",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_directionOfAltitude_01_vals), 0,
        "T_directionOfAltitude_01", HFILL }},
    { &hf_rnsap_GANSS_SatelliteInformationKP_item,
      { "GANSS-SatelliteInformationKP item", "rnsap.GANSS_SatelliteInformationKP_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_e_alm,
      { "ganss-e-alm", "rnsap.ganss_e_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_ganss_delta_I_alm,
      { "ganss-delta-I-alm", "rnsap.ganss_delta_I_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_ganss_omegadot_alm,
      { "ganss-omegadot-alm", "rnsap.ganss_omegadot_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_ganss_svStatusINAV_alm,
      { "ganss-svStatusINAV-alm", "rnsap.ganss_svStatusINAV_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_rnsap_ganss_svStatusFNAV_alm,
      { "ganss-svStatusFNAV-alm", "rnsap.ganss_svStatusFNAV_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_rnsap_ganss_delta_a_sqrt_alm,
      { "ganss-delta-a-sqrt-alm", "rnsap.ganss_delta_a_sqrt_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_rnsap_ganss_omegazero_alm,
      { "ganss-omegazero-alm", "rnsap.ganss_omegazero_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_ganss_m_zero_alm,
      { "ganss-m-zero-alm", "rnsap.ganss_m_zero_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_ganss_omega_alm,
      { "ganss-omega-alm", "rnsap.ganss_omega_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_ganss_af_zero_alm,
      { "ganss-af-zero-alm", "rnsap.ganss_af_zero_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_ganss_af_one_alm,
      { "ganss-af-one-alm", "rnsap.ganss_af_one_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_rnsap_Ganss_Sat_Info_AddNavList_item,
      { "Ganss-Sat-Info-AddNavList item", "rnsap.Ganss_Sat_Info_AddNavList_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_svHealth,
      { "svHealth", "rnsap.svHealth",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_9", HFILL }},
    { &hf_rnsap_iod,
      { "iod", "rnsap.iod",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_ganssAddClockModels,
      { "ganssAddClockModels", "rnsap.ganssAddClockModels",
        FT_UINT32, BASE_DEC, VALS(rnsap_GANSS_AddClockModels_vals), 0,
        "GANSS_AddClockModels", HFILL }},
    { &hf_rnsap_ganssAddOrbitModels,
      { "ganssAddOrbitModels", "rnsap.ganssAddOrbitModels",
        FT_UINT32, BASE_DEC, VALS(rnsap_GANSS_AddOrbitModels_vals), 0,
        "GANSS_AddOrbitModels", HFILL }},
    { &hf_rnsap_GANSS_SAT_Info_Almanac_BDSkpList_item,
      { "GANSS-SAT-Info-Almanac-BDS", "rnsap.GANSS_SAT_Info_Almanac_BDS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_bdsAlmToa,
      { "bdsAlmToa", "rnsap.bdsAlmToa",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_bdsAlmSqrtA,
      { "bdsAlmSqrtA", "rnsap.bdsAlmSqrtA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_bdsAlmE,
      { "bdsAlmE", "rnsap.bdsAlmE",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_rnsap_bdsAlmW,
      { "bdsAlmW", "rnsap.bdsAlmW",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_bdsAlmM0,
      { "bdsAlmM0", "rnsap.bdsAlmM0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_bdsAlmOmega0,
      { "bdsAlmOmega0", "rnsap.bdsAlmOmega0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_bdsAlmOmegaDot,
      { "bdsAlmOmegaDot", "rnsap.bdsAlmOmegaDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_rnsap_bdsAlmDeltaI,
      { "bdsAlmDeltaI", "rnsap.bdsAlmDeltaI",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_bdsAlmA0,
      { "bdsAlmA0", "rnsap.bdsAlmA0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_bdsAlmA1,
      { "bdsAlmA1", "rnsap.bdsAlmA1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_bdsSvHealth,
      { "bdsSvHealth", "rnsap.bdsSvHealth",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_9", HFILL }},
    { &hf_rnsap_GANSS_SAT_Info_Almanac_GLOkpList_item,
      { "GANSS-SAT-Info-Almanac-GLOkpList item", "rnsap.GANSS_SAT_Info_Almanac_GLOkpList_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gloAlmNA,
      { "gloAlmNA", "rnsap.gloAlmNA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_gloAlmnA,
      { "gloAlmnA", "rnsap.gloAlmnA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_rnsap_gloAlmHA,
      { "gloAlmHA", "rnsap.gloAlmHA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_rnsap_gloAlmLambdaA,
      { "gloAlmLambdaA", "rnsap.gloAlmLambdaA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_21", HFILL }},
    { &hf_rnsap_gloAlmTlambdaA,
      { "gloAlmTlambdaA", "rnsap.gloAlmTlambdaA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_21", HFILL }},
    { &hf_rnsap_gloAlmDeltaIA,
      { "gloAlmDeltaIA", "rnsap.gloAlmDeltaIA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_18", HFILL }},
    { &hf_rnsap_gloAkmDeltaTA,
      { "gloAkmDeltaTA", "rnsap.gloAkmDeltaTA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_22", HFILL }},
    { &hf_rnsap_gloAlmDeltaTdotA,
      { "gloAlmDeltaTdotA", "rnsap.gloAlmDeltaTdotA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_7", HFILL }},
    { &hf_rnsap_gloAlmEpsilonA,
      { "gloAlmEpsilonA", "rnsap.gloAlmEpsilonA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_15", HFILL }},
    { &hf_rnsap_gloAlmOmegaA,
      { "gloAlmOmegaA", "rnsap.gloAlmOmegaA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_gloAlmTauA,
      { "gloAlmTauA", "rnsap.gloAlmTauA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_rnsap_gloAlmCA,
      { "gloAlmCA", "rnsap.gloAlmCA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_rnsap_gloAlmMA,
      { "gloAlmMA", "rnsap.gloAlmMA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_rnsap_GANSS_SAT_Info_Almanac_MIDIkpList_item,
      { "GANSS-SAT-Info-Almanac-MIDIkpList item", "rnsap.GANSS_SAT_Info_Almanac_MIDIkpList_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_midiAlmE,
      { "midiAlmE", "rnsap.midiAlmE",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_midiAlmDeltaI,
      { "midiAlmDeltaI", "rnsap.midiAlmDeltaI",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_midiAlmOmegaDot,
      { "midiAlmOmegaDot", "rnsap.midiAlmOmegaDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_midiAlmSqrtA,
      { "midiAlmSqrtA", "rnsap.midiAlmSqrtA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_17", HFILL }},
    { &hf_rnsap_midiAlmOmega0,
      { "midiAlmOmega0", "rnsap.midiAlmOmega0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_midiAlmOmega,
      { "midiAlmOmega", "rnsap.midiAlmOmega",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_midiAlmMo,
      { "midiAlmMo", "rnsap.midiAlmMo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_midiAlmaf0,
      { "midiAlmaf0", "rnsap.midiAlmaf0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_midiAlmaf1,
      { "midiAlmaf1", "rnsap.midiAlmaf1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_rnsap_midiAlmL1Health,
      { "midiAlmL1Health", "rnsap.midiAlmL1Health",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_rnsap_midiAlmL2Health,
      { "midiAlmL2Health", "rnsap.midiAlmL2Health",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_rnsap_midiAlmL5Health,
      { "midiAlmL5Health", "rnsap.midiAlmL5Health",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_rnsap_GANSS_SAT_Info_Almanac_NAVkpList_item,
      { "GANSS-SAT-Info-Almanac-NAVkpList item", "rnsap.GANSS_SAT_Info_Almanac_NAVkpList_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_navAlmE,
      { "navAlmE", "rnsap.navAlmE",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_navAlmDeltaI,
      { "navAlmDeltaI", "rnsap.navAlmDeltaI",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_navAlmOMEGADOT,
      { "navAlmOMEGADOT", "rnsap.navAlmOMEGADOT",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_navAlmSVHealth,
      { "navAlmSVHealth", "rnsap.navAlmSVHealth",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_navAlmSqrtA,
      { "navAlmSqrtA", "rnsap.navAlmSqrtA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_navAlmOMEGAo,
      { "navAlmOMEGAo", "rnsap.navAlmOMEGAo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_navAlmOmega,
      { "navAlmOmega", "rnsap.navAlmOmega",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_navAlmMo,
      { "navAlmMo", "rnsap.navAlmMo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_navAlmaf0,
      { "navAlmaf0", "rnsap.navAlmaf0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_navAlmaf1,
      { "navAlmaf1", "rnsap.navAlmaf1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_GANSS_SAT_Info_Almanac_REDkpList_item,
      { "GANSS-SAT-Info-Almanac-REDkpList item", "rnsap.GANSS_SAT_Info_Almanac_REDkpList_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_redAlmDeltaA,
      { "redAlmDeltaA", "rnsap.redAlmDeltaA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_redAlmOmega0,
      { "redAlmOmega0", "rnsap.redAlmOmega0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_7", HFILL }},
    { &hf_rnsap_redAlmPhi0,
      { "redAlmPhi0", "rnsap.redAlmPhi0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_7", HFILL }},
    { &hf_rnsap_redAlmL1Health,
      { "redAlmL1Health", "rnsap.redAlmL1Health",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_rnsap_redAlmL2Health,
      { "redAlmL2Health", "rnsap.redAlmL2Health",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_rnsap_redAlmL5Health,
      { "redAlmL5Health", "rnsap.redAlmL5Health",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_rnsap_GANSS_SAT_Info_Almanac_SBASecefList_item,
      { "GANSS-SAT-Info-Almanac-SBASecefList item", "rnsap.GANSS_SAT_Info_Almanac_SBASecefList_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sbasAlmDataID,
      { "sbasAlmDataID", "rnsap.sbasAlmDataID",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_rnsap_sbasAlmHealth,
      { "sbasAlmHealth", "rnsap.sbasAlmHealth",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_sbasAlmXg,
      { "sbasAlmXg", "rnsap.sbasAlmXg",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_15", HFILL }},
    { &hf_rnsap_sbasAlmYg,
      { "sbasAlmYg", "rnsap.sbasAlmYg",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_15", HFILL }},
    { &hf_rnsap_sbasAlmZg,
      { "sbasAlmZg", "rnsap.sbasAlmZg",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_9", HFILL }},
    { &hf_rnsap_sbasAlmXgdot,
      { "sbasAlmXgdot", "rnsap.sbasAlmXgdot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_3", HFILL }},
    { &hf_rnsap_sbasAlmYgDot,
      { "sbasAlmYgDot", "rnsap.sbasAlmYgDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_3", HFILL }},
    { &hf_rnsap_sbasAlmZgDot,
      { "sbasAlmZgDot", "rnsap.sbasAlmZgDot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_rnsap_sbasAlmTo,
      { "sbasAlmTo", "rnsap.sbasAlmTo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_GANSS_Sat_Info_Nav_item,
      { "GANSS-Sat-Info-Nav item", "rnsap.GANSS_Sat_Info_Nav_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_svHealth_01,
      { "svHealth", "rnsap.svHealth",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_rnsap_iod_01,
      { "iod", "rnsap.iod",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_rnsap_ganssClockModel,
      { "ganssClockModel", "rnsap.ganssClockModel",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GANSS_Clock_Model", HFILL }},
    { &hf_rnsap_ganssOrbitModel,
      { "ganssOrbitModel", "rnsap.ganssOrbitModel",
        FT_UINT32, BASE_DEC, VALS(rnsap_GANSS_Orbit_Model_vals), 0,
        "GANSS_Orbit_Model", HFILL }},
    { &hf_rnsap_sbasAgfo,
      { "sbasAgfo", "rnsap.sbasAgfo",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_12", HFILL }},
    { &hf_rnsap_sbasAgf1,
      { "sbasAgf1", "rnsap.sbasAgf1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_ganss_time_model_Ref_Time,
      { "ganss-time-model-Ref-Time", "rnsap.ganss_time_model_Ref_Time",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_37799", HFILL }},
    { &hf_rnsap_ganss_t_a0,
      { "ganss-t-a0", "rnsap.ganss_t_a0",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M2147483648_2147483647", HFILL }},
    { &hf_rnsap_ganss_t_a1,
      { "ganss-t-a1", "rnsap.ganss_t_a1",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M8388608_8388607", HFILL }},
    { &hf_rnsap_ganss_t_a2,
      { "ganss-t-a2", "rnsap.ganss_t_a2",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M64_63", HFILL }},
    { &hf_rnsap_gnss_to_id,
      { "gnss-to-id", "rnsap.gnss_to_id",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_gnss_to_id_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_ganss_wk_number_01,
      { "ganss-wk-number", "rnsap.ganss_wk_number",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_8191", HFILL }},
    { &hf_rnsap_ganssDay,
      { "ganssDay", "rnsap.ganssDay",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_8191", HFILL }},
    { &hf_rnsap_a_one_utc,
      { "a-one-utc", "rnsap.a_one_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_a_zero_utc,
      { "a-zero-utc", "rnsap.a_zero_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_t_ot_utc,
      { "t-ot-utc", "rnsap.t_ot_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_w_n_t_utc,
      { "w-n-t-utc", "rnsap.w_n_t_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_delta_t_ls_utc,
      { "delta-t-ls-utc", "rnsap.delta_t_ls_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_w_n_lsf_utc,
      { "w-n-lsf-utc", "rnsap.w_n_lsf_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_dn_utc,
      { "dn-utc", "rnsap.dn_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_delta_t_lsf_utc,
      { "delta-t-lsf-utc", "rnsap.delta_t_lsf_utc",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_utcA0,
      { "utcA0", "rnsap.utcA0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_utcA1,
      { "utcA1", "rnsap.utcA1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_rnsap_utcA2,
      { "utcA2", "rnsap.utcA2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_7", HFILL }},
    { &hf_rnsap_utcDeltaTls,
      { "utcDeltaTls", "rnsap.utcDeltaTls",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_utcTot,
      { "utcTot", "rnsap.utcTot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_utcWNot,
      { "utcWNot", "rnsap.utcWNot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_13", HFILL }},
    { &hf_rnsap_utcWNlsf,
      { "utcWNlsf", "rnsap.utcWNlsf",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_utcDN,
      { "utcDN", "rnsap.utcDN",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_rnsap_utcDeltaTlsf,
      { "utcDeltaTlsf", "rnsap.utcDeltaTlsf",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_nA,
      { "nA", "rnsap.nA",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_tauC,
      { "tauC", "rnsap.tauC",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_deltaUT1_01,
      { "deltaUT1", "rnsap.deltaUT1_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GANSS_DeltaUT1", HFILL }},
    { &hf_rnsap_kp,
      { "kp", "rnsap.kp",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_rnsap_utcA1wnt,
      { "utcA1wnt", "rnsap.utcA1wnt",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_utcA0wnt,
      { "utcA0wnt", "rnsap.utcA0wnt",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_utcTot_01,
      { "utcTot", "rnsap.utcTot",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_utcWNt,
      { "utcWNt", "rnsap.utcWNt",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_utcDN_01,
      { "utcDN", "rnsap.utcDN",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_utcStandardID,
      { "utcStandardID", "rnsap.utcStandardID",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_3", HFILL }},
    { &hf_rnsap_utcA0_01,
      { "utcA0", "rnsap.utcA0",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_utcA1_01,
      { "utcA1", "rnsap.utcA1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_altitudeAndDirection,
      { "altitudeAndDirection", "rnsap.altitudeAndDirection_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GA_AltitudeAndDirection", HFILL }},
    { &hf_rnsap_uncertaintyEllipse,
      { "uncertaintyEllipse", "rnsap.uncertaintyEllipse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GA_UncertaintyEllipse", HFILL }},
    { &hf_rnsap_uncertaintyAltitude,
      { "uncertaintyAltitude", "rnsap.uncertaintyAltitude",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_127", HFILL }},
    { &hf_rnsap_uncertaintySemi_major,
      { "uncertaintySemi-major", "rnsap.uncertaintySemi_major",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_127", HFILL }},
    { &hf_rnsap_uncertaintySemi_minor,
      { "uncertaintySemi-minor", "rnsap.uncertaintySemi_minor",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_127", HFILL }},
    { &hf_rnsap_orientationOfMajorAxis,
      { "orientationOfMajorAxis", "rnsap.orientationOfMajorAxis",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_179", HFILL }},
    { &hf_rnsap_uncertaintyCode,
      { "uncertaintyCode", "rnsap.uncertaintyCode",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_127", HFILL }},
    { &hf_rnsap_geographicalCoordinate,
      { "geographicalCoordinate", "rnsap.geographicalCoordinate_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_latitudeSign_01,
      { "latitudeSign", "rnsap.latitudeSign",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_latitudeSign_01_vals), 0,
        "T_latitudeSign_01", HFILL }},
    { &hf_rnsap_latitude,
      { "latitude", "rnsap.latitude",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_8388607", HFILL }},
    { &hf_rnsap_longitude,
      { "longitude", "rnsap.longitude",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M8388608_8388607", HFILL }},
    { &hf_rnsap_sI,
      { "sI", "rnsap.sI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GERAN_SystemInfo", HFILL }},
    { &hf_rnsap_pSI,
      { "pSI", "rnsap.pSI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GERAN_SystemInfo", HFILL }},
    { &hf_rnsap_GERAN_SystemInfo_item,
      { "GERAN-SystemInfo item", "rnsap.GERAN_SystemInfo_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gERAN_SI_block,
      { "gERAN-SI-block", "rnsap.gERAN_SI_block",
        FT_BYTES, BASE_NONE, NULL, 0,
        "OCTET_STRING_SIZE_1_23", HFILL }},
    { &hf_rnsap_wna_alm,
      { "wna-alm", "rnsap.wna_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_satellite_Almanac_Information,
      { "satellite-Almanac-Information", "rnsap.satellite_Almanac_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_satellite_Almanac_Information_item,
      { "satellite-Almanac-Information item", "rnsap.satellite_Almanac_Information_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dATA_ID,
      { "dATA-ID", "rnsap.dATA_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gps_e_alm,
      { "gps-e-alm", "rnsap.gps_e_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_gps_toa_alm,
      { "gps-toa-alm", "rnsap.gps_toa_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_gps_delta_I_alm,
      { "gps-delta-I-alm", "rnsap.gps_delta_I_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_omegadot_alm,
      { "omegadot-alm", "rnsap.omegadot_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_svhealth_alm,
      { "svhealth-alm", "rnsap.svhealth_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_gps_a_sqrt_alm,
      { "gps-a-sqrt-alm", "rnsap.gps_a_sqrt_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_omegazero_alm,
      { "omegazero-alm", "rnsap.omegazero_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_m_zero_alm,
      { "m-zero-alm", "rnsap.m_zero_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_gps_omega_alm,
      { "gps-omega-alm", "rnsap.gps_omega_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_24", HFILL }},
    { &hf_rnsap_gps_af_zero_alm,
      { "gps-af-zero-alm", "rnsap.gps_af_zero_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_gps_af_one_alm,
      { "gps-af-one-alm", "rnsap.gps_af_one_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_11", HFILL }},
    { &hf_rnsap_sVGlobalHealth_alm,
      { "sVGlobalHealth-alm", "rnsap.sVGlobalHealth_alm",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_364", HFILL }},
    { &hf_rnsap_Satellite_Almanac_Information_ExtItem_item,
      { "Satellite-Almanac-Information-ExtItem item", "rnsap.Satellite_Almanac_Information_ExtItem_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GPSInformation_item,
      { "GPSInformation item", "rnsap.GPSInformation_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gPSInformationItem,
      { "gPSInformationItem", "rnsap.gPSInformationItem",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_gPSInformationItem_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_alpha_zero_ionos_01,
      { "alpha-zero-ionos", "rnsap.alpha_zero_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_alpha_one_ionos_01,
      { "alpha-one-ionos", "rnsap.alpha_one_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_alpha_two_ionos_01,
      { "alpha-two-ionos", "rnsap.alpha_two_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_alpha_three_ionos,
      { "alpha-three-ionos", "rnsap.alpha_three_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_beta_zero_ionos,
      { "beta-zero-ionos", "rnsap.beta_zero_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_beta_one_ionos,
      { "beta-one-ionos", "rnsap.beta_one_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_beta_two_ionos,
      { "beta-two-ionos", "rnsap.beta_two_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_beta_three_ionos,
      { "beta-three-ionos", "rnsap.beta_three_ionos",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_GPS_NavigationModel_and_TimeRecovery_item,
      { "GPS-NavigationModel-and-TimeRecovery item", "rnsap.GPS_NavigationModel_and_TimeRecovery_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tx_tow_nav,
      { "tx-tow-nav", "rnsap.tx_tow_nav",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_1048575", HFILL }},
    { &hf_rnsap_tlm_message_nav,
      { "tlm-message-nav", "rnsap.tlm_message_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_14", HFILL }},
    { &hf_rnsap_tlm_revd_c_nav,
      { "tlm-revd-c-nav", "rnsap.tlm_revd_c_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_rnsap_ho_word_nav,
      { "ho-word-nav", "rnsap.ho_word_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_22", HFILL }},
    { &hf_rnsap_w_n_nav,
      { "w-n-nav", "rnsap.w_n_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_rnsap_ca_or_p_on_l2_nav,
      { "ca-or-p-on-l2-nav", "rnsap.ca_or_p_on_l2_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_2", HFILL }},
    { &hf_rnsap_user_range_accuracy_index_nav,
      { "user-range-accuracy-index-nav", "rnsap.user_range_accuracy_index_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_4", HFILL }},
    { &hf_rnsap_sv_health_nav,
      { "sv-health-nav", "rnsap.sv_health_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_6", HFILL }},
    { &hf_rnsap_iodc_nav,
      { "iodc-nav", "rnsap.iodc_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_10", HFILL }},
    { &hf_rnsap_l2_p_dataflag_nav,
      { "l2-p-dataflag-nav", "rnsap.l2_p_dataflag_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_rnsap_sf1_reserved_nav,
      { "sf1-reserved-nav", "rnsap.sf1_reserved_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_87", HFILL }},
    { &hf_rnsap_t_gd_nav,
      { "t-gd-nav", "rnsap.t_gd_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_t_oc_nav,
      { "t-oc-nav", "rnsap.t_oc_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_a_f_2_nav,
      { "a-f-2-nav", "rnsap.a_f_2_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_a_f_1_nav,
      { "a-f-1-nav", "rnsap.a_f_1_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_a_f_zero_nav,
      { "a-f-zero-nav", "rnsap.a_f_zero_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_22", HFILL }},
    { &hf_rnsap_gps_e_nav,
      { "gps-e-nav", "rnsap.gps_e_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_t_oe_nav,
      { "t-oe-nav", "rnsap.t_oe_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_16", HFILL }},
    { &hf_rnsap_fit_interval_flag_nav,
      { "fit-interval-flag-nav", "rnsap.fit_interval_flag_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_1", HFILL }},
    { &hf_rnsap_aodo_nav,
      { "aodo-nav", "rnsap.aodo_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_5", HFILL }},
    { &hf_rnsap_gps_omega_nav,
      { "gps-omega-nav", "rnsap.gps_omega_nav",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_32", HFILL }},
    { &hf_rnsap_spare_zero_fill,
      { "spare-zero-fill", "rnsap.spare_zero_fill",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_20", HFILL }},
    { &hf_rnsap_badSatellites,
      { "badSatellites", "rnsap.badSatellites_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_noBadSatellite,
      { "noBadSatellite", "rnsap.noBadSatellite_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_guaranteed_UL_Rate,
      { "guaranteed-UL-Rate", "rnsap.guaranteed_UL_Rate",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Guaranteed_Rate", HFILL }},
    { &hf_rnsap_guaranteed_DL_Rate,
      { "guaranteed-DL-Rate", "rnsap.guaranteed_DL_Rate",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Guaranteed_Rate", HFILL }},
    { &hf_rnsap_implicit,
      { "implicit", "rnsap.implicit_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "HARQ_MemoryPartitioning_Implicit", HFILL }},
    { &hf_rnsap_explicit_01,
      { "explicit", "rnsap.explicit_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "HARQ_MemoryPartitioning_Explicit", HFILL }},
    { &hf_rnsap_number_of_Processes,
      { "number-of-Processes", "rnsap.number_of_Processes",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_8_", HFILL }},
    { &hf_rnsap_hARQ_MemoryPartitioningList,
      { "hARQ-MemoryPartitioningList", "rnsap.hARQ_MemoryPartitioningList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HARQ_MemoryPartitioningList_item,
      { "HARQ-MemoryPartitioningItem", "rnsap.HARQ_MemoryPartitioningItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HARQ_MemoryPartitioningInfoExtForMIMO_item,
      { "HARQ-MemoryPartitioningItem", "rnsap.HARQ_MemoryPartitioningItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_process_Memory_Size,
      { "process-Memory-Size", "rnsap.process_Memory_Size",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_process_Memory_Size_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_hSDSCH_MACdFlows_Information,
      { "hSDSCH-MACdFlows-Information", "rnsap.hSDSCH_MACdFlows_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_Capabilities_Info,
      { "uE-Capabilities-Info", "rnsap.uE_Capabilities_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mAChs_Reordering_Buffer_Size_for_RLC_UM,
      { "mAChs-Reordering-Buffer-Size-for-RLC-UM", "rnsap.mAChs_Reordering_Buffer_Size_for_RLC_UM",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MAChsReorderingBufferSize_for_RLC_UM", HFILL }},
    { &hf_rnsap_cqiFeedback_CycleK,
      { "cqiFeedback-CycleK", "rnsap.cqiFeedback_CycleK",
        FT_UINT32, BASE_DEC, VALS(rnsap_CQI_Feedback_Cycle_vals), 0,
        "CQI_Feedback_Cycle", HFILL }},
    { &hf_rnsap_cqiRepetitionFactor,
      { "cqiRepetitionFactor", "rnsap.cqiRepetitionFactor",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CQI_RepetitionFactor", HFILL }},
    { &hf_rnsap_ackNackRepetitionFactor,
      { "ackNackRepetitionFactor", "rnsap.ackNackRepetitionFactor",
        FT_UINT32, BASE_DEC, NULL, 0,
        "AckNack_RepetitionFactor", HFILL }},
    { &hf_rnsap_cqiPowerOffset,
      { "cqiPowerOffset", "rnsap.cqiPowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CQI_Power_Offset", HFILL }},
    { &hf_rnsap_ackPowerOffset,
      { "ackPowerOffset", "rnsap.ackPowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Ack_Power_Offset", HFILL }},
    { &hf_rnsap_nackPowerOffset,
      { "nackPowerOffset", "rnsap.nackPowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Nack_Power_Offset", HFILL }},
    { &hf_rnsap_hsscch_PowerOffset,
      { "hsscch-PowerOffset", "rnsap.hsscch_PowerOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSDSCH_MACdFlow_Specific_InfoList_Response,
      { "hSDSCH-MACdFlow-Specific-InfoList-Response", "rnsap.hSDSCH_MACdFlow_Specific_InfoList_Response",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSSCCH_Specific_InfoList_Response,
      { "hSSCCH-Specific-InfoList-Response", "rnsap.hSSCCH_Specific_InfoList_Response",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HSSCCH_FDD_Specific_InfoList_Response", HFILL }},
    { &hf_rnsap_hSPDSCH_and_HSSCCH_ScramblingCode,
      { "hSPDSCH-and-HSSCCH-ScramblingCode", "rnsap.hSPDSCH_and_HSSCCH_ScramblingCode",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_ScramblingCode", HFILL }},
    { &hf_rnsap_measurement_Power_Offset,
      { "measurement-Power-Offset", "rnsap.measurement_Power_Offset",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hARQ_MemoryPartitioning,
      { "hARQ-MemoryPartitioning", "rnsap.hARQ_MemoryPartitioning",
        FT_UINT32, BASE_DEC, VALS(rnsap_HARQ_MemoryPartitioning_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_sixtyfourQAM_UsageAllowedIndicator,
      { "sixtyfourQAM-UsageAllowedIndicator", "rnsap.sixtyfourQAM_UsageAllowedIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SixtyfourQAM_UsageAllowedIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_sixtyfourQAM_DL_UsageIndicator,
      { "sixtyfourQAM-DL-UsageIndicator", "rnsap.sixtyfourQAM_DL_UsageIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SixtyfourQAM_DL_UsageIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_hSDSCH_TBSizeTableIndicator,
      { "hSDSCH-TBSizeTableIndicator", "rnsap.hSDSCH_TBSizeTableIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_HSDSCH_TBSizeTableIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_hSSCCH_CodeChangeGrant,
      { "hSSCCH-CodeChangeGrant", "rnsap.hSSCCH_CodeChangeGrant",
        FT_UINT32, BASE_DEC, VALS(rnsap_HSSCCH_Code_Change_Grant_vals), 0,
        "HSSCCH_Code_Change_Grant", HFILL }},
    { &hf_rnsap_hsSCCHCodeChangeIndicator,
      { "hsSCCHCodeChangeIndicator", "rnsap.hsSCCHCodeChangeIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_HSSCCH_CodeChangeIndicator_vals), 0,
        "HSSCCH_CodeChangeIndicator", HFILL }},
    { &hf_rnsap_hS_PDSCH_Code_Change_Indicator,
      { "hS-PDSCH-Code-Change-Indicator", "rnsap.hS_PDSCH_Code_Change_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_HS_PDSCH_Code_Change_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_DSCH_Secondary_Serving_cell_choice,
      { "hS-DSCH-Secondary-Serving-cell-choice", "rnsap.hS_DSCH_Secondary_Serving_cell_choice",
        FT_UINT32, BASE_DEC, VALS(rnsap_HS_DSCH_Secondary_Serving_cell_change_choice_vals), 0,
        "HS_DSCH_Secondary_Serving_cell_change_choice", HFILL }},
    { &hf_rnsap_hS_Secondary_Serving_cell_change_successful,
      { "hS-Secondary-Serving-cell-change-successful", "rnsap.hS_Secondary_Serving_cell_change_successful_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_Secondary_Serving_cell_change_unsuccessful,
      { "hS-Secondary-Serving-cell-change-unsuccessful", "rnsap.hS_Secondary_Serving_cell_change_unsuccessful_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information_Response,
      { "hS-DSCH-FDD-Secondary-Serving-Information-Response", "rnsap.hS_DSCH_FDD_Secondary_Serving_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSDSCH_MACdFlow_Specific_InfoList_to_Modify,
      { "hSDSCH-MACdFlow-Specific-InfoList-to-Modify", "rnsap.hSDSCH_MACdFlow_Specific_InfoList_to_Modify",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_priorityQueue_Info_to_Modify,
      { "priorityQueue-Info-to-Modify", "rnsap.priorityQueue_Info_to_Modify",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_InfoList_to_Modify", HFILL }},
    { &hf_rnsap_tDDAckNackPowerOffset,
      { "tDDAckNackPowerOffset", "rnsap.tDDAckNackPowerOffset",
        FT_INT32, BASE_DEC, NULL, 0,
        "TDD_AckNack_Power_Offset", HFILL }},
    { &hf_rnsap_priorityQueueInfotoModifyUnsynchronised,
      { "priorityQueueInfotoModifyUnsynchronised", "rnsap.priorityQueueInfotoModifyUnsynchronised",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_InfoList_to_Modify_Unsynchronised", HFILL }},
    { &hf_rnsap_HSDSCH_MACdFlow_Specific_InfoList_item,
      { "HSDSCH-MACdFlow-Specific-InfoItem", "rnsap.HSDSCH_MACdFlow_Specific_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSDSCH_MACdFlow_ID,
      { "hSDSCH-MACdFlow-ID", "rnsap.hSDSCH_MACdFlow_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_MACdFlow_Specific_InfoList_Response_item,
      { "HSDSCH-MACdFlow-Specific-InfoItem-Response", "rnsap.HSDSCH_MACdFlow_Specific_InfoItem_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify_item,
      { "HSDSCH-MACdFlow-Specific-InfoItem-to-Modify", "rnsap.HSDSCH_MACdFlow_Specific_InfoItem_to_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSDSCH_MACdFlow_Specific_Info,
      { "hSDSCH-MACdFlow-Specific-Info", "rnsap.hSDSCH_MACdFlow_Specific_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HSDSCH_MACdFlow_Specific_InfoList", HFILL }},
    { &hf_rnsap_priorityQueue_Info,
      { "priorityQueue-Info", "rnsap.priorityQueue_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_InfoList", HFILL }},
    { &hf_rnsap_HSDSCH_MACdFlows_to_Delete_item,
      { "HSDSCH-MACdFlows-to-Delete-Item", "rnsap.HSDSCH_MACdFlows_to_Delete_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hsDSCH_MACdFlow_ID,
      { "hsDSCH-MACdFlow-ID", "rnsap.hsDSCH_MACdFlow_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCH_Initial_Capacity_Allocation_item,
      { "HSDSCH-Initial-Capacity-AllocationItem", "rnsap.HSDSCH_Initial_Capacity_AllocationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_maximum_MACdPDU_Size,
      { "maximum-MACdPDU-Size", "rnsap.maximum_MACdPDU_Size",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MACdPDU_Size", HFILL }},
    { &hf_rnsap_hSDSCH_InitialWindowSize,
      { "hSDSCH-InitialWindowSize", "rnsap.hSDSCH_InitialWindowSize",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_setsOfHS_SCCH_Codes,
      { "setsOfHS-SCCH-Codes", "rnsap.setsOfHS_SCCH_Codes",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hARQ_Preamble_Mode_Activation_Indicator,
      { "hARQ-Preamble-Mode-Activation-Indicator", "rnsap.hARQ_Preamble_Mode_Activation_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_HARQ_Preamble_Mode_Activation_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_mIMO_N_M_Ratio,
      { "mIMO-N-M-Ratio", "rnsap.mIMO_N_M_Ratio_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "MIMO_InformationResponse", HFILL }},
    { &hf_rnsap_continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response,
      { "continuous-Packet-Connectivity-HS-SCCH-Less-Information-Response", "rnsap.continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Preconfiguration_Information_item,
      { "Additional-EDCH-Preconfiguration-Information-ItemIEs", "rnsap.Additional_EDCH_Preconfiguration_Information_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mAChsResetScheme,
      { "mAChsResetScheme", "rnsap.mAChsResetScheme",
        FT_UINT32, BASE_DEC, VALS(rnsap_MAChsResetScheme_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_hSDSCH_Physical_Layer_Category,
      { "hSDSCH-Physical-Layer-Category", "rnsap.hSDSCH_Physical_Layer_Category",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_64_", HFILL }},
    { &hf_rnsap_secondaryServingCells,
      { "secondaryServingCells", "rnsap.secondaryServingCells",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_numPrimaryHS_SCCH_Codes,
      { "numPrimaryHS-SCCH-Codes", "rnsap.numPrimaryHS_SCCH_Codes",
        FT_UINT32, BASE_DEC, NULL, 0,
        "NumHS_SCCH_Codes", HFILL }},
    { &hf_rnsap_hARQ_Preamble_Mode,
      { "hARQ-Preamble-Mode", "rnsap.hARQ_Preamble_Mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_HARQ_Preamble_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_mIMO_ActivationIndicator,
      { "mIMO-ActivationIndicator", "rnsap.mIMO_ActivationIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSDSCH_MACdPDUSizeFormat,
      { "hSDSCH-MACdPDUSizeFormat", "rnsap.hSDSCH_MACdPDUSizeFormat",
        FT_UINT32, BASE_DEC, VALS(rnsap_HSDSCH_MACdPDUSizeFormat_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_with_enhanced_HS_SCCH_support_indicator,
      { "uE-with-enhanced-HS-SCCH-support-indicator", "rnsap.uE_with_enhanced_HS_SCCH_support_indicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HS_SCCH_PreconfiguredCodes_item,
      { "HS-SCCH-PreconfiguredCodesItem", "rnsap.HS_SCCH_PreconfiguredCodesItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_SCCH_CodeNumber,
      { "hS-SCCH-CodeNumber", "rnsap.hS_SCCH_CodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_PDSCH_RLID,
      { "hS-PDSCH-RLID", "rnsap.hS_PDSCH_RLID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RL_ID", HFILL }},
    { &hf_rnsap_hSDSCH_FDD_Information,
      { "hSDSCH-FDD-Information", "rnsap.hSDSCH_FDD_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_DSCH_serving_cell_choice,
      { "hS-DSCH-serving-cell-choice", "rnsap.hS_DSCH_serving_cell_choice",
        FT_UINT32, BASE_DEC, VALS(rnsap_HS_DSCH_serving_cell_change_choice_vals), 0,
        "HS_DSCH_serving_cell_change_choice", HFILL }},
    { &hf_rnsap_hS_serving_cell_change_successful,
      { "hS-serving-cell-change-successful", "rnsap.hS_serving_cell_change_successful_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_serving_cell_change_unsuccessful,
      { "hS-serving-cell-change-unsuccessful", "rnsap.hS_serving_cell_change_unsuccessful_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSDSCH_FDD_Information_Response,
      { "hSDSCH-FDD-Information-Response", "rnsap.hSDSCH_FDD_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tDD_AckNack_Power_Offset,
      { "tDD-AckNack-Power-Offset", "rnsap.tDD_AckNack_Power_Offset",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSSCCH_TDD_Specific_InfoList_Response,
      { "hSSCCH-TDD-Specific-InfoList-Response", "rnsap.hSSCCH_TDD_Specific_InfoList_Response",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSSCCH_TDD_Specific_InfoList_Response_LCR,
      { "hSSCCH-TDD-Specific-InfoList-Response-LCR", "rnsap.hSSCCH_TDD_Specific_InfoList_Response_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSPDSCH_TDD_Specific_InfoList_Response,
      { "hSPDSCH-TDD-Specific-InfoList-Response", "rnsap.hSPDSCH_TDD_Specific_InfoList_Response",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSPDSCH_TDD_Specific_InfoList_Response_LCR,
      { "hSPDSCH-TDD-Specific-InfoList-Response-LCR", "rnsap.hSPDSCH_TDD_Specific_InfoList_Response_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSPDSCH_TDD_Specific_InfoList_Response_item,
      { "HSPDSCH-TDD-Specific-InfoItem-Response", "rnsap.HSPDSCH_TDD_Specific_InfoItem_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSPDSCH_TDD_Specific_InfoList_Response_LCR_item,
      { "HSPDSCH-TDD-Specific-InfoItem-Response-LCR", "rnsap.HSPDSCH_TDD_Specific_InfoItem_Response_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_timeslotLCR,
      { "timeslotLCR", "rnsap.timeslotLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSPDSCH_TDD_Specific_InfoList_Response768_item,
      { "HSPDSCH-TDD-Specific-InfoItem-Response768", "rnsap.HSPDSCH_TDD_Specific_InfoItem_Response768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSSCCH_FDD_Specific_InfoList_Response_item,
      { "HSSCCH-FDD-Specific-InfoItem-Response", "rnsap.HSSCCH_FDD_Specific_InfoItem_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_code_Number,
      { "code-Number", "rnsap.code_Number",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_127", HFILL }},
    { &hf_rnsap_HSSCCH_TDD_Specific_InfoList_Response_item,
      { "HSSCCH-TDD-Specific-InfoItem-Response", "rnsap.HSSCCH_TDD_Specific_InfoItem_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSSICH_Info,
      { "hSSICH-Info", "rnsap.hSSICH_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSSCCH_TDD_Specific_InfoList_Response_LCR_item,
      { "HSSCCH-TDD-Specific-InfoItem-Response-LCR", "rnsap.HSSCCH_TDD_Specific_InfoItem_Response_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_first_TDD_ChannelisationCode,
      { "first-TDD-ChannelisationCode", "rnsap.first_TDD_ChannelisationCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_TDD_ChannelisationCode_vals_ext, 0,
        "TDD_ChannelisationCode", HFILL }},
    { &hf_rnsap_second_TDD_ChannelisationCode,
      { "second-TDD-ChannelisationCode", "rnsap.second_TDD_ChannelisationCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_TDD_ChannelisationCode_vals_ext, 0,
        "TDD_ChannelisationCode", HFILL }},
    { &hf_rnsap_hSSICH_InfoLCR,
      { "hSSICH-InfoLCR", "rnsap.hSSICH_InfoLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSSCCH_TDD_Specific_InfoList_Response768_item,
      { "HSSCCH-TDD-Specific-InfoItem-Response768", "rnsap.HSSCCH_TDD_Specific_InfoItem_Response768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSSICH_Info768,
      { "hSSICH-Info768", "rnsap.hSSICH_Info768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hsSICH_ID,
      { "hsSICH-ID", "rnsap.hsSICH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HS_SICH_ID", HFILL }},
    { &hf_rnsap_failed_HS_SICH,
      { "failed-HS-SICH", "rnsap.failed_HS_SICH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HS_SICH_failed", HFILL }},
    { &hf_rnsap_missed_HS_SICH,
      { "missed-HS-SICH", "rnsap.missed_HS_SICH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HS_SICH_missed", HFILL }},
    { &hf_rnsap_total_HS_SICH,
      { "total-HS-SICH", "rnsap.total_HS_SICH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HS_SICH_total", HFILL }},
    { &hf_rnsap_MIMO_ReferenceSignal_InformationListLCR_item,
      { "HSSICH-ReferenceSignal-InformationLCR", "rnsap.HSSICH_ReferenceSignal_InformationLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_midambleConfigurationLCR,
      { "midambleConfigurationLCR", "rnsap.midambleConfigurationLCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_MidambleConfigurationLCR_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_midambleShift,
      { "midambleShift", "rnsap.midambleShift",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_15", HFILL }},
    { &hf_rnsap_transport_Block_Size_List,
      { "transport-Block-Size-List", "rnsap.transport_Block_Size_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Transport_Block_Size_List_LCR", HFILL }},
    { &hf_rnsap_hS_DSCH_SPS_Reservation_Indicator,
      { "hS-DSCH-SPS-Reservation-Indicator", "rnsap.hS_DSCH_SPS_Reservation_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SPS_Reservation_Indicator_vals), 0,
        "SPS_Reservation_Indicator", HFILL }},
    { &hf_rnsap_hS_DSCH_SPS_Operation_Indicator,
      { "hS-DSCH-SPS-Operation-Indicator", "rnsap.hS_DSCH_SPS_Operation_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_HS_DSCH_SPS_Operation_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Transport_Block_Size_List_LCR_item,
      { "Transport-Block-Size-Item-LCR", "rnsap.Transport_Block_Size_Item_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_transport_Block_Size_maping_Index_LCR,
      { "transport-Block-Size-maping-Index-LCR", "rnsap.transport_Block_Size_maping_Index_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_transport_Block_Size_Index_LCR,
      { "transport-Block-Size-Index-LCR", "rnsap.transport_Block_Size_Index_LCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Repetition_Period_List_LCR_item,
      { "Repetition-Period-Item-LCR", "rnsap.Repetition_Period_Item_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_logicalChannellevel,
      { "logicalChannellevel", "rnsap.logicalChannellevel",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_priorityQueuelevel,
      { "priorityQueuelevel", "rnsap.priorityQueuelevel",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_SICH_InformationList_for_HS_DSCH_SPS,
      { "hS-SICH-InformationList-for-HS-DSCH-SPS", "rnsap.hS_SICH_InformationList_for_HS_DSCH_SPS",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_initial_HS_PDSCH_SPS_Resource,
      { "initial-HS-PDSCH-SPS-Resource", "rnsap.initial_HS_PDSCH_SPS_Resource_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_buffer_Size_for_HS_DSCH_SPS,
      { "buffer-Size-for-HS-DSCH-SPS", "rnsap.buffer_Size_for_HS_DSCH_SPS",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_Process_Memory_Size_vals_ext, 0,
        "Process_Memory_Size", HFILL }},
    { &hf_rnsap_number_of_Processes_for_HS_DSCH_SPS,
      { "number-of-Processes-for-HS-DSCH-SPS", "rnsap.number_of_Processes_for_HS_DSCH_SPS",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HS_SICH_InformationList_for_HS_DSCH_SPS_item,
      { "HS-SICH-InformationItem-for-HS-DSCH-SPS", "rnsap.HS_SICH_InformationItem_for_HS_DSCH_SPS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_SICH_Mapping_Index,
      { "hS-SICH-Mapping-Index", "rnsap.hS_SICH_Mapping_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_SICH_Type,
      { "hS-SICH-Type", "rnsap.hS_SICH_Type",
        FT_UINT32, BASE_DEC, VALS(rnsap_HS_SICH_Type_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_SCCH_Associated_HS_SICH,
      { "hS-SCCH-Associated-HS-SICH", "rnsap.hS_SCCH_Associated_HS_SICH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_non_HS_SCCH_Associated_HS_SICH,
      { "non-HS-SCCH-Associated-HS-SICH", "rnsap.non_HS_SCCH_Associated_HS_SICH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_extended_HS_SICH_ID,
      { "extended-HS-SICH-ID", "rnsap.extended_HS_SICH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HS_SICH_ID_Extension", HFILL }},
    { &hf_rnsap_non_HS_SCCH_Aassociated_HS_SICH_ID,
      { "non-HS-SCCH-Aassociated-HS-SICH-ID", "rnsap.non_HS_SCCH_Aassociated_HS_SICH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_PDSCH_Offset,
      { "hS-PDSCH-Offset", "rnsap.hS_PDSCH_Offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_PhysicalChannelOffset", HFILL }},
    { &hf_rnsap_hS_PDSCH_Midamble_Configuation,
      { "hS-PDSCH-Midamble-Configuation", "rnsap.hS_PDSCH_Midamble_Configuation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "MidambleShiftLCR", HFILL }},
    { &hf_rnsap_timeslot_Resource_Related_Information_01,
      { "timeslot-Resource-Related-Information", "rnsap.timeslot_Resource_Related_Information",
        FT_BYTES, BASE_NONE, NULL, 0,
        "HS_DSCH_TimeslotResourceLCR", HFILL }},
    { &hf_rnsap_startCode,
      { "startCode", "rnsap.startCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_TDD_ChannelisationCode_vals_ext, 0,
        "TDD_ChannelisationCode", HFILL }},
    { &hf_rnsap_endCode,
      { "endCode", "rnsap.endCode",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_TDD_ChannelisationCode_vals_ext, 0,
        "TDD_ChannelisationCode", HFILL }},
    { &hf_rnsap_transport_Block_Size_Index_01,
      { "transport-Block-Size-Index", "rnsap.transport_Block_Size_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Transport_Block_Size_Index_LCR", HFILL }},
    { &hf_rnsap_modulationType,
      { "modulationType", "rnsap.modulationType",
        FT_UINT32, BASE_DEC, VALS(rnsap_ModulationSPS_LCR_vals), 0,
        "ModulationSPS_LCR", HFILL }},
    { &hf_rnsap_requestedDataValue,
      { "requestedDataValue", "rnsap.requestedDataValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_onDemand,
      { "onDemand", "rnsap.onDemand_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_periodic,
      { "periodic", "rnsap.periodic_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PeriodicInformation", HFILL }},
    { &hf_rnsap_onModification,
      { "onModification", "rnsap.onModification_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "OnModificationInformation", HFILL }},
    { &hf_rnsap_min,
      { "min", "rnsap.min",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_60_", HFILL }},
    { &hf_rnsap_hour,
      { "hour", "rnsap.hour",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_24_", HFILL }},
    { &hf_rnsap_dGPSThreshold,
      { "dGPSThreshold", "rnsap.dGPSThreshold_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dGANSSThreshold,
      { "dGANSSThreshold", "rnsap.dGANSSThreshold_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_informationTypeItem,
      { "informationTypeItem", "rnsap.informationTypeItem",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_informationTypeItem_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_gPSInformation,
      { "gPSInformation", "rnsap.gPSInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_iPDL_FDD_Parameters,
      { "iPDL-FDD-Parameters", "rnsap.iPDL_FDD_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_iPDL_TDD_Parameters,
      { "iPDL-TDD-Parameters", "rnsap.iPDL_TDD_Parameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_extension_IPDLParameters,
      { "extension-IPDLParameters", "rnsap.extension_IPDLParameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Inter_Frequency_Cell_List_item,
      { "Inter-Frequency-Cell", "rnsap.Inter_Frequency_Cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_inter_Frequency_Cell_Information_SIB11,
      { "inter-Frequency-Cell-Information-SIB11", "rnsap.inter_Frequency_Cell_Information_SIB11",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_inter_Frequency_Cell_Information_SIB12,
      { "inter-Frequency-Cell-Information-SIB12", "rnsap.inter_Frequency_Cell_Information_SIB12",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Inter_Frequency_Cell_Information_SIB11_item,
      { "Inter-Frequency-Cells-Information-SIB11-Per-Indication", "rnsap.Inter_Frequency_Cells_Information_SIB11_Per_Indication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Inter_Frequency_Cell_Information_SIB12_item,
      { "Inter-Frequency-Cells-Information-SIB12-Per-Indication", "rnsap.Inter_Frequency_Cells_Information_SIB12_Per_Indication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Inter_Frequency_Cell_Information_SIB11ter_item,
      { "Inter-Frequency-Cells-Information-SIB11ter-Per-Indication", "rnsap.Inter_Frequency_Cells_Information_SIB11ter_Per_Indication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_inter_Frequency_Cell_Indication_SIB11,
      { "inter-Frequency-Cell-Indication-SIB11", "rnsap.inter_Frequency_Cell_Indication_SIB11",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Inter_Frequency_Cell_Indication", HFILL }},
    { &hf_rnsap_inter_Frequency_Cell_List_SIB11,
      { "inter-Frequency-Cell-List-SIB11", "rnsap.inter_Frequency_Cell_List_SIB11",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Inter_Frequency_Cell_SIB11_or_SIB12_List", HFILL }},
    { &hf_rnsap_inter_Frequency_Cell_Indication_SIB12,
      { "inter-Frequency-Cell-Indication-SIB12", "rnsap.inter_Frequency_Cell_Indication_SIB12",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Inter_Frequency_Cell_Indication", HFILL }},
    { &hf_rnsap_inter_Frequency_Cell_List_SIB12,
      { "inter-Frequency-Cell-List-SIB12", "rnsap.inter_Frequency_Cell_List_SIB12",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Inter_Frequency_Cell_SIB11_or_SIB12_List", HFILL }},
    { &hf_rnsap_inter_Frequency_Cell_Indication_SIB11ter,
      { "inter-Frequency-Cell-Indication-SIB11ter", "rnsap.inter_Frequency_Cell_Indication_SIB11ter",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Inter_Frequency_Cell_Indication", HFILL }},
    { &hf_rnsap_inter_Frequency_Cell_List_SIB11ter,
      { "inter-Frequency-Cell-List-SIB11ter", "rnsap.inter_Frequency_Cell_List_SIB11ter",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Inter_Frequency_Cell_SIB11ter_List", HFILL }},
    { &hf_rnsap_Inter_Frequency_Cell_SIB11ter_List_item,
      { "Inter-Frequency-Cell-SIB11ter", "rnsap.Inter_Frequency_Cell_SIB11ter_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_interFrequencyCellID,
      { "interFrequencyCellID", "rnsap.interFrequencyCellID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "InterFrequencyCellID_Ext", HFILL }},
    { &hf_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12_List_item,
      { "Inter-Frequency-Cell-SIB11-or-SIB12", "rnsap.Inter_Frequency_Cell_SIB11_or_SIB12_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_interFrequencyCellID_01,
      { "interFrequencyCellID", "rnsap.interFrequencyCellID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_iPSpacingFDD,
      { "iPSpacingFDD", "rnsap.iPSpacingFDD",
        FT_UINT32, BASE_DEC, VALS(rnsap_IPSpacingFDD_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_iPLength,
      { "iPLength", "rnsap.iPLength",
        FT_UINT32, BASE_DEC, VALS(rnsap_IPLength_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_iPOffset,
      { "iPOffset", "rnsap.iPOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_seed,
      { "seed", "rnsap.seed",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_burstModeParameters,
      { "burstModeParameters", "rnsap.burstModeParameters_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_iPSpacingTDD,
      { "iPSpacingTDD", "rnsap.iPSpacingTDD",
        FT_UINT32, BASE_DEC, VALS(rnsap_IPSpacingTDD_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_iPStart,
      { "iPStart", "rnsap.iPStart",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_iPSlot,
      { "iPSlot", "rnsap.iPSlot",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_iP_P_CCPCH,
      { "iP-P-CCPCH", "rnsap.iP_P_CCPCH",
        FT_UINT32, BASE_DEC, VALS(rnsap_IP_P_CCPCH_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_iPSub,
      { "iPSub", "rnsap.iPSub",
        FT_UINT32, BASE_DEC, VALS(rnsap_IPSub_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_idleIntervalInfo_k,
      { "idleIntervalInfo-k", "rnsap.idleIntervalInfo_k",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_2_3", HFILL }},
    { &hf_rnsap_idleIntervalInfo_offset,
      { "idleIntervalInfo-offset", "rnsap.idleIntervalInfo_offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_7", HFILL }},
    { &hf_rnsap_laiList,
      { "laiList", "rnsap.laiList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "LAI_List", HFILL }},
    { &hf_rnsap_LAI_List_item,
      { "LAI", "rnsap.LAI_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_List_Of_PLMNs_item,
      { "PLMN-Identity", "rnsap.PLMN_Identity",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uplinkLoadValue,
      { "uplinkLoadValue", "rnsap.uplinkLoadValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_100", HFILL }},
    { &hf_rnsap_downlinkLoadValue,
      { "downlinkLoadValue", "rnsap.downlinkLoadValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_100", HFILL }},
    { &hf_rnsap_maxTimeslotsPerSubFrame,
      { "maxTimeslotsPerSubFrame", "rnsap.maxTimeslotsPerSubFrame",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_6", HFILL }},
    { &hf_rnsap_maxPhysChPerTimeslot,
      { "maxPhysChPerTimeslot", "rnsap.maxPhysChPerTimeslot",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_maxPhysChPerTimeslot_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_periodic_01,
      { "periodic", "rnsap.periodic_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "MDT_Report_Parameters", HFILL }},
    { &hf_rnsap_event1F,
      { "event1F", "rnsap.event1F_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Event1F_Parameters", HFILL }},
    { &hf_rnsap_event1I,
      { "event1I", "rnsap.event1I_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Event1I_Parameters", HFILL }},
    { &hf_rnsap_cellbased,
      { "cellbased", "rnsap.cellbased_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_labased,
      { "labased", "rnsap.labased_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rabased,
      { "rabased", "rnsap.rabased_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_plmn_area_based,
      { "plmn-area-based", "rnsap.plmn_area_based_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mdtActivation,
      { "mdtActivation", "rnsap.mdtActivation",
        FT_UINT32, BASE_DEC, VALS(rnsap_MDT_Activation_vals), 0,
        "MDT_Activation", HFILL }},
    { &hf_rnsap_mdtAreaScope,
      { "mdtAreaScope", "rnsap.mdtAreaScope",
        FT_UINT32, BASE_DEC, VALS(rnsap_MDTAreaScope_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_m1report,
      { "m1report", "rnsap.m1report",
        FT_UINT32, BASE_DEC, VALS(rnsap_M1Report_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_m2report,
      { "m2report", "rnsap.m2report",
        FT_UINT32, BASE_DEC, VALS(rnsap_M2Report_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_m3report,
      { "m3report", "rnsap.m3report_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_reportInterval,
      { "reportInterval", "rnsap.reportInterval",
        FT_UINT32, BASE_DEC, VALS(rnsap_ReportInterval_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_reportAmount,
      { "reportAmount", "rnsap.reportAmount",
        FT_UINT32, BASE_DEC, VALS(rnsap_ReportAmount_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_MAC_c_sh_SDU_LengthList_item,
      { "MAC-c-sh-SDU-Length", "rnsap.MAC_c_sh_SDU_Length",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MACdPDU_Size_IndexList_item,
      { "MACdPDU-Size-IndexItem", "rnsap.MACdPDU_Size_IndexItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sID,
      { "sID", "rnsap.sID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MACdPDU_Size_IndexList_to_Modify_item,
      { "MACdPDU-Size-IndexItem-to-Modify", "rnsap.MACdPDU_Size_IndexItem_to_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_accessPointName,
      { "accessPointName", "rnsap.accessPointName",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_iPMulticastAddress,
      { "iPMulticastAddress", "rnsap.iPMulticastAddress",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_List_item,
      { "TMGI", "rnsap.TMGI_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tmgi,
      { "tmgi", "rnsap.tmgi_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_transmissionMode,
      { "transmissionMode", "rnsap.transmissionMode",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransmissionMode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_preferredFrequencyLayer,
      { "preferredFrequencyLayer", "rnsap.preferredFrequencyLayer",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UARFCN", HFILL }},
    { &hf_rnsap_tMGI,
      { "tMGI", "rnsap.tMGI_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_pTM_Cell_List,
      { "pTM-Cell-List", "rnsap.pTM_Cell_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PTMCellList", HFILL }},
    { &hf_rnsap_pTP_Cell_List,
      { "pTP-Cell-List", "rnsap.pTP_Cell_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PTPCellList", HFILL }},
    { &hf_rnsap_not_Provided_Cell_List,
      { "not-Provided-Cell-List", "rnsap.not_Provided_Cell_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        "NotProvidedCellList", HFILL }},
    { &hf_rnsap_affectedUEInformationForMBMS,
      { "affectedUEInformationForMBMS", "rnsap.affectedUEInformationForMBMS",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_preferredFrequencyLayerInfo,
      { "preferredFrequencyLayerInfo", "rnsap.preferredFrequencyLayerInfo_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mBMS_ConcatenatedServiceList,
      { "mBMS-ConcatenatedServiceList", "rnsap.mBMS_ConcatenatedServiceList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_l3_Information_1,
      { "l3-Information-1", "rnsap.l3_Information_1",
        FT_BYTES, BASE_NONE, NULL, 0,
        "L3_Information", HFILL }},
    { &hf_rnsap_l3_Information_2,
      { "l3-Information-2", "rnsap.l3_Information_2",
        FT_BYTES, BASE_NONE, NULL, 0,
        "L3_Information", HFILL }},
    { &hf_rnsap_MBMS_ConcatenatedServiceList_item,
      { "TMGI", "rnsap.TMGI_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_RLC_Sequence_Number_Information_item,
      { "MBMS-RLC-Sequence-Number-Information-List", "rnsap.MBMS_RLC_Sequence_Number_Information_List_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mBMS_Bearer_Service_List_RLC,
      { "mBMS-Bearer-Service-List-RLC", "rnsap.mBMS_Bearer_Service_List_RLC",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_List_RLC_item,
      { "MBMS-Bearer-Service-List-RLCinfo", "rnsap.MBMS_Bearer_Service_List_RLCinfo_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_time_Stamp,
      { "time-Stamp", "rnsap.time_Stamp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MCCH_Message_List_item,
      { "L3-Information", "rnsap.L3_Information",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_secondaryCCPCHSystemInformationMBMS,
      { "secondaryCCPCHSystemInformationMBMS", "rnsap.secondaryCCPCHSystemInformationMBMS",
        FT_BYTES, BASE_NONE, NULL, 0,
        "Secondary_CCPCH_System_Information_MBMS", HFILL }},
    { &hf_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Info_List_item,
      { "MBSFN-Scheduling-Transmission-Time-Interval-Item", "rnsap.MBSFN_Scheduling_Transmission_Time_Interval_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mbsfnSchedulingTransmissionTimeInterval,
      { "mbsfnSchedulingTransmissionTimeInterval", "rnsap.mbsfnSchedulingTransmissionTimeInterval",
        FT_UINT32, BASE_DEC, VALS(rnsap_MbsfnSchedulingTransmissionTimeInterval_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_sir,
      { "sir", "rnsap.sir",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SIR_Value_IncrDecrThres", HFILL }},
    { &hf_rnsap_sir_error,
      { "sir-error", "rnsap.sir_error",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SIR_Error_Value_IncrDecrThres", HFILL }},
    { &hf_rnsap_transmitted_code_power,
      { "transmitted-code-power", "rnsap.transmitted_code_power",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Transmitted_Code_Power_Value_IncrDecrThres", HFILL }},
    { &hf_rnsap_rscp,
      { "rscp", "rnsap.rscp",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RSCP_Value_IncrDecrThres", HFILL }},
    { &hf_rnsap_round_trip_time,
      { "round-trip-time", "rnsap.round_trip_time",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Round_Trip_Time_IncrDecrThres", HFILL }},
    { &hf_rnsap_extension_MeasurementIncreaseDecreaseThreshold,
      { "extension-MeasurementIncreaseDecreaseThreshold", "rnsap.extension_MeasurementIncreaseDecreaseThreshold_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sir_01,
      { "sir", "rnsap.sir",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SIR_Value", HFILL }},
    { &hf_rnsap_sir_error_01,
      { "sir-error", "rnsap.sir_error",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SIR_Error_Value", HFILL }},
    { &hf_rnsap_transmitted_code_power_01,
      { "transmitted-code-power", "rnsap.transmitted_code_power",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Transmitted_Code_Power_Value", HFILL }},
    { &hf_rnsap_rscp_01,
      { "rscp", "rnsap.rscp",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RSCP_Value", HFILL }},
    { &hf_rnsap_rx_timing_deviation,
      { "rx-timing-deviation", "rnsap.rx_timing_deviation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Rx_Timing_Deviation_Value", HFILL }},
    { &hf_rnsap_round_trip_time_01,
      { "round-trip-time", "rnsap.round_trip_time",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Round_Trip_Time_Value", HFILL }},
    { &hf_rnsap_extension_MeasurementThreshold,
      { "extension-MeasurementThreshold", "rnsap.extension_MeasurementThreshold_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_type1,
      { "type1", "rnsap.type1_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_midambleConfigurationBurstType1And3,
      { "midambleConfigurationBurstType1And3", "rnsap.midambleConfigurationBurstType1And3",
        FT_UINT32, BASE_DEC, VALS(rnsap_MidambleConfigurationBurstType1And3_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_midambleAllocationMode,
      { "midambleAllocationMode", "rnsap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(rnsap_MidambleAllocationMode1_vals), 0,
        "MidambleAllocationMode1", HFILL }},
    { &hf_rnsap_defaultMidamble,
      { "defaultMidamble", "rnsap.defaultMidamble_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_commonMidamble,
      { "commonMidamble", "rnsap.commonMidamble_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ueSpecificMidamble,
      { "ueSpecificMidamble", "rnsap.ueSpecificMidamble",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MidambleShiftLong", HFILL }},
    { &hf_rnsap_type2,
      { "type2", "rnsap.type2_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_midambleConfigurationBurstType2,
      { "midambleConfigurationBurstType2", "rnsap.midambleConfigurationBurstType2",
        FT_UINT32, BASE_DEC, VALS(rnsap_MidambleConfigurationBurstType2_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_midambleAllocationMode_01,
      { "midambleAllocationMode", "rnsap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(rnsap_MidambleAllocationMode2_vals), 0,
        "MidambleAllocationMode2", HFILL }},
    { &hf_rnsap_ueSpecificMidamble_01,
      { "ueSpecificMidamble", "rnsap.ueSpecificMidamble",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MidambleShiftShort", HFILL }},
    { &hf_rnsap_type3,
      { "type3", "rnsap.type3_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_midambleAllocationMode_02,
      { "midambleAllocationMode", "rnsap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(rnsap_MidambleAllocationMode3_vals), 0,
        "MidambleAllocationMode3", HFILL }},
    { &hf_rnsap_midambleAllocationMode_03,
      { "midambleAllocationMode", "rnsap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(rnsap_MidambleAllocationMode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_midambleShift_01,
      { "midambleShift", "rnsap.midambleShift",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MidambleShiftLong", HFILL }},
    { &hf_rnsap_type1_01,
      { "type1", "rnsap.type1_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Type7681", HFILL }},
    { &hf_rnsap_midambleAllocationMode_04,
      { "midambleAllocationMode", "rnsap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(rnsap_MidambleAllocationMode7681_vals), 0,
        "MidambleAllocationMode7681", HFILL }},
    { &hf_rnsap_type2_01,
      { "type2", "rnsap.type2_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Type7682", HFILL }},
    { &hf_rnsap_midambleConfigurationBurstType2_768,
      { "midambleConfigurationBurstType2-768", "rnsap.midambleConfigurationBurstType2_768",
        FT_UINT32, BASE_DEC, VALS(rnsap_MidambleConfigurationBurstType2_768_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_midambleAllocationMode_05,
      { "midambleAllocationMode", "rnsap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(rnsap_MidambleAllocationMode7682_vals), 0,
        "MidambleAllocationMode7682", HFILL }},
    { &hf_rnsap_ueSpecificMidamble_02,
      { "ueSpecificMidamble", "rnsap.ueSpecificMidamble",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MidambleShiftShort768", HFILL }},
    { &hf_rnsap_type3_01,
      { "type3", "rnsap.type3_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Type7683", HFILL }},
    { &hf_rnsap_midambleAllocationMode_06,
      { "midambleAllocationMode", "rnsap.midambleAllocationMode",
        FT_UINT32, BASE_DEC, VALS(rnsap_MidambleAllocationMode7683_vals), 0,
        "MidambleAllocationMode7683", HFILL }},
    { &hf_rnsap_mIMO_PilotConfiguration,
      { "mIMO-PilotConfiguration", "rnsap.mIMO_PilotConfiguration",
        FT_UINT32, BASE_DEC, VALS(rnsap_MIMO_PilotConfiguration_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_mIMO_N_M_Ratio_01,
      { "mIMO-N-M-Ratio", "rnsap.mIMO_N_M_Ratio",
        FT_UINT32, BASE_DEC, VALS(rnsap_MIMO_N_M_Ratio_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_primary_and_secondary_CPICH,
      { "primary-and-secondary-CPICH", "rnsap.primary_and_secondary_CPICH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MIMO_S_CPICH_Channelisation_Code", HFILL }},
    { &hf_rnsap_normal_and_diversity_primary_CPICH,
      { "normal-and-diversity-primary-CPICH", "rnsap.normal_and_diversity_primary_CPICH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_Associated_Secondary_CPICH_item,
      { "Additional-Secondary-CPICH", "rnsap.Additional_Secondary_CPICH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_additional_secondary_CPICH,
      { "additional-secondary-CPICH", "rnsap.additional_secondary_CPICH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MIMO_S_CPICH_Channelisation_Code", HFILL }},
    { &hf_rnsap_additional_D_CPICH,
      { "additional-D-CPICH", "rnsap.additional_D_CPICH",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MIMO_S_CPICH_Channelisation_Code", HFILL }},
    { &hf_rnsap_addPriorityQueue,
      { "addPriorityQueue", "rnsap.addPriorityQueue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PriorityQueue_InfoItem_to_Add", HFILL }},
    { &hf_rnsap_modifyPriorityQueue,
      { "modifyPriorityQueue", "rnsap.modifyPriorityQueue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PriorityQueue_InfoItem_to_Modify", HFILL }},
    { &hf_rnsap_deletePriorityQueue,
      { "deletePriorityQueue", "rnsap.deletePriorityQueue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_Id", HFILL }},
    { &hf_rnsap_dL_PowerBalancing_Information,
      { "dL-PowerBalancing-Information", "rnsap.dL_PowerBalancing_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_minimumReducedE_DPDCH_GainFactor,
      { "minimumReducedE-DPDCH-GainFactor", "rnsap.minimumReducedE_DPDCH_GainFactor",
        FT_UINT32, BASE_DEC, VALS(rnsap_MinimumReducedE_DPDCH_GainFactor_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_secondary_UL_Frequency_Activation_State,
      { "secondary-UL-Frequency-Activation-State", "rnsap.secondary_UL_Frequency_Activation_State",
        FT_UINT32, BASE_DEC, VALS(rnsap_Secondary_UL_Frequency_Activation_State_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_extendedPropagationDelay,
      { "extendedPropagationDelay", "rnsap.extendedPropagationDelay",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_enhanced_PrimaryCPICH_EcNo,
      { "enhanced-PrimaryCPICH-EcNo", "rnsap.enhanced_PrimaryCPICH_EcNo",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_phase_Reference_Update_Indicator,
      { "phase-Reference-Update-Indicator", "rnsap.phase_Reference_Update_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_Phase_Reference_Update_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_additional_e_DCH_DL_Control_Channel_Grant,
      { "additional-e-DCH-DL-Control-Channel-Grant", "rnsap.additional_e_DCH_DL_Control_Channel_Grant_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_item,
      { "MultipleFreq-HSPDSCH-InformationItem-ResponseTDDLCR", "rnsap.MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uARFCN,
      { "uARFCN", "rnsap.uARFCN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_multicarrier_EDCH_Transport_Bearer_Mode,
      { "multicarrier-EDCH-Transport-Bearer-Mode", "rnsap.multicarrier_EDCH_Transport_Bearer_Mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_Multicarrier_EDCH_Transport_Bearer_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_multi_carrier_EDCH_Information,
      { "multi-carrier-EDCH-Information", "rnsap.multi_carrier_EDCH_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multi_Carrier_EDCH_Information_item,
      { "Multi-Carrier-EDCH-LCR-InformationItem", "rnsap.Multi_Carrier_EDCH_LCR_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uARFCNforNt,
      { "uARFCNforNt", "rnsap.uARFCNforNt",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UARFCN", HFILL }},
    { &hf_rnsap_sNPL_carrier_group_indicator,
      { "sNPL-carrier-group-indicator", "rnsap.sNPL_carrier_group_indicator",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_multi_Carrier_EDCH_MACdFlows_Information_TDD,
      { "multi-Carrier-EDCH-MACdFlows-Information-TDD", "rnsap.multi_Carrier_EDCH_MACdFlows_Information_TDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multi_Carrier_EDCH_MACdFlows_Information_TDD_item,
      { "Multi-Carrier-EDCH-MACdFlows-Specific-Info", "rnsap.Multi_Carrier_EDCH_MACdFlows_Specific_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_continue_setup_change_Of_Multi_Carrier_EDCH,
      { "continue-setup-change-Of-Multi-Carrier-EDCH", "rnsap.continue_setup_change_Of_Multi_Carrier_EDCH",
        FT_UINT32, BASE_DEC, VALS(rnsap_Continue_Setup_Change_Multi_Carrier_EDCH_vals), 0,
        "Continue_Setup_Change_Multi_Carrier_EDCH", HFILL }},
    { &hf_rnsap_continue,
      { "continue", "rnsap.continue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_setup,
      { "setup", "rnsap.setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Multi_Carrier_EDCH_Info", HFILL }},
    { &hf_rnsap_change,
      { "change", "rnsap.change_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Multi_Carrier_EDCH_Change_Info", HFILL }},
    { &hf_rnsap_multi_Carrier_EDCH_Information_Removal_List,
      { "multi-Carrier-EDCH-Information-Removal-List", "rnsap.multi_Carrier_EDCH_Information_Removal_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multi_Carrier_EDCH_Information_Removal_List_item,
      { "Multi-Carrier-EDCH-Information-Removal-Info-ItemIEs", "rnsap.Multi_Carrier_EDCH_Information_Removal_Info_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multi_Carrier_EDCH_Information_Response_item,
      { "Multi-Carrier-EDCH-LCR-Information-ResponseItem", "rnsap.Multi_Carrier_EDCH_LCR_Information_ResponseItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_setup_01,
      { "setup", "rnsap.setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Multiflow_Information", HFILL }},
    { &hf_rnsap_configurationChange,
      { "configurationChange", "rnsap.configurationChange_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Multiflow_Information_To_Modify", HFILL }},
    { &hf_rnsap_stop,
      { "stop", "rnsap.stop",
        FT_UINT32, BASE_DEC, VALS(rnsap_Multiflow_Stop_vals), 0,
        "Multiflow_Stop", HFILL }},
    { &hf_rnsap_total_Number_of_HS_DSCH_Cells,
      { "total-Number-of-HS-DSCH-Cells", "rnsap.total_Number_of_HS_DSCH_Cells",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_2_32_", HFILL }},
    { &hf_rnsap_mode,
      { "mode", "rnsap.mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_Multiflow_Mode_vals), 0,
        "Multiflow_Mode", HFILL }},
    { &hf_rnsap_mimo,
      { "mimo", "rnsap.mimo",
        FT_UINT32, BASE_DEC, VALS(rnsap_Multiflow_MIMO_vals), 0,
        "Multiflow_MIMO", HFILL }},
    { &hf_rnsap_timing,
      { "timing", "rnsap.timing",
        FT_UINT32, BASE_DEC, VALS(rnsap_Multiflow_Timing_vals), 0,
        "Multiflow_Timing", HFILL }},
    { &hf_rnsap_max_Number_of_HS_SCCH_Sets_per_NodeB,
      { "max-Number-of-HS-SCCH-Sets-per-NodeB", "rnsap.max_Number_of_HS_SCCH_Sets_per_NodeB",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_16_", HFILL }},
    { &hf_rnsap_assisting_CQI_RepetitionFactor,
      { "assisting-CQI-RepetitionFactor", "rnsap.assisting_CQI_RepetitionFactor",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CQI_RepetitionFactor", HFILL }},
    { &hf_rnsap_assisting_AckNack_RepetitionFactor,
      { "assisting-AckNack-RepetitionFactor", "rnsap.assisting_AckNack_RepetitionFactor",
        FT_UINT32, BASE_DEC, NULL, 0,
        "AckNack_RepetitionFactor", HFILL }},
    { &hf_rnsap_time_Reference,
      { "time-Reference", "rnsap.time_Reference_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_non_Time_Reference,
      { "non-Time-Reference", "rnsap.non_Time_Reference",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_30_", HFILL }},
    { &hf_rnsap_mU_MIMO_Operation,
      { "mU-MIMO-Operation", "rnsap.mU_MIMO_Operation",
        FT_UINT32, BASE_DEC, VALS(rnsap_MU_MIMO_Operation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_standalone_Midamble_Channel_Information,
      { "standalone-Midamble-Channel-Information", "rnsap.standalone_Midamble_Channel_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_standalone_Midamble_Configuration,
      { "standalone-Midamble-Configuration", "rnsap.standalone_Midamble_Configuration",
        FT_UINT32, BASE_DEC, VALS(rnsap_Standalone_Midamble_Configuration_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_standalone_MidambleShift,
      { "standalone-MidambleShift", "rnsap.standalone_MidambleShift",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_repetitionPeriod_01,
      { "repetitionPeriod", "rnsap.repetitionPeriod",
        FT_UINT32, BASE_DEC, VALS(rnsap_Standalone_Midamble_RepetitionPeriod_vals), 0,
        "Standalone_Midamble_RepetitionPeriod", HFILL }},
    { &hf_rnsap_offset,
      { "offset", "rnsap.offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Standalone_Midamble_Offset", HFILL }},
    { &hf_rnsap_referenceBeta,
      { "referenceBeta", "rnsap.referenceBeta",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mU_MIMO_Usage_Indicator,
      { "mU-MIMO-Usage-Indicator", "rnsap.mU_MIMO_Usage_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_MU_MIMO_Usage_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_standalone_Midamble_Channel_Indicator,
      { "standalone-Midamble-Channel-Indicator", "rnsap.standalone_Midamble_Channel_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_Standalone_Midamble_Channel_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_gERAN_SI_Type,
      { "gERAN-SI-Type", "rnsap.gERAN_SI_Type",
        FT_UINT32, BASE_DEC, VALS(rnsap_GERAN_SI_Type_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Neighbouring_UMTS_CellInformation_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Neighbouring_UMTS_CellInformation_Ext_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cN_PS_DomainIdentifier,
      { "cN-PS-DomainIdentifier", "rnsap.cN_PS_DomainIdentifier_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cN_CS_DomainIdentifier,
      { "cN-CS-DomainIdentifier", "rnsap.cN_CS_DomainIdentifier_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_neighbouring_FDD_CellInformation,
      { "neighbouring-FDD-CellInformation", "rnsap.neighbouring_FDD_CellInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_neighbouring_TDD_CellInformation,
      { "neighbouring-TDD-CellInformation", "rnsap.neighbouring_TDD_CellInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_neighbouring_LCR_TDD_CellInformation,
      { "neighbouring-LCR-TDD-CellInformation", "rnsap.neighbouring_LCR_TDD_CellInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Neighbouring_FDD_CellInformation_item,
      { "Neighbouring-FDD-CellInformationItem", "rnsap.Neighbouring_FDD_CellInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uARFCNforNu,
      { "uARFCNforNu", "rnsap.uARFCNforNu",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UARFCN", HFILL }},
    { &hf_rnsap_uARFCNforNd,
      { "uARFCNforNd", "rnsap.uARFCNforNd",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UARFCN", HFILL }},
    { &hf_rnsap_frameOffset,
      { "frameOffset", "rnsap.frameOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cellIndividualOffset,
      { "cellIndividualOffset", "rnsap.cellIndividualOffset",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_not_used_closedLoopMode2_SupportIndicator,
      { "not-used-closedLoopMode2-SupportIndicator", "rnsap.not_used_closedLoopMode2_SupportIndicator_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MultipleFrequencyBandsIndicatorList_item,
      { "FrequencyBandIndicator", "rnsap.FrequencyBandIndicator",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_FrequencyBandIndicator_vals_ext, 0,
        NULL, HFILL }},
    { &hf_rnsap_uC_ID,
      { "uC-ID", "rnsap.uC_ID_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Neighbouring_GSM_CellInformationIEs_item,
      { "Neighbouring-GSM-CellInformationItem", "rnsap.Neighbouring_GSM_CellInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cGI,
      { "cGI", "rnsap.cGI_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_bSIC,
      { "bSIC", "rnsap.bSIC_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_band_Indicator,
      { "band-Indicator", "rnsap.band_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_Band_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_bCCH_ARFCN,
      { "bCCH-ARFCN", "rnsap.bCCH_ARFCN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Neighbouring_TDD_CellInformation_item,
      { "Neighbouring-TDD-CellInformationItem", "rnsap.Neighbouring_TDD_CellInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cellParameterID,
      { "cellParameterID", "rnsap.cellParameterID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Neighbouring_LCR_TDD_CellInformation_item,
      { "Neighbouring-LCR-TDD-CellInformationItem", "rnsap.Neighbouring_LCR_TDD_CellInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Neighbouring_E_UTRA_CellInformation_item,
      { "Neighbouring-E-UTRA-CellInformationItem", "rnsap.Neighbouring_E_UTRA_CellInformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eCGI,
      { "eCGI", "rnsap.eCGI_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eARFCN_Information,
      { "eARFCN-Information", "rnsap.eARFCN_Information",
        FT_UINT32, BASE_DEC, VALS(rnsap_EARFCN_Information_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_MultipleFrequencyBandsIndicatorList_EUTRA_item,
      { "RadioFrequencyBandEUTRA", "rnsap.RadioFrequencyBandEUTRA",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_NotProvidedCellList_item,
      { "MBMSChannelTypeCellList", "rnsap.MBMSChannelTypeCellList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uplinkNRTLoadInformationValue,
      { "uplinkNRTLoadInformationValue", "rnsap.uplinkNRTLoadInformationValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_3", HFILL }},
    { &hf_rnsap_downlinkNRTLoadInformationValue,
      { "downlinkNRTLoadInformationValue", "rnsap.downlinkNRTLoadInformationValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_3", HFILL }},
    { &hf_rnsap_new_non_serving_RL_selection,
      { "new-non-serving-RL-selection", "rnsap.new_non_serving_RL_selection",
        FT_UINT32, BASE_DEC, VALS(rnsap_New_non_serving_RL_setup_selection_vals), 0,
        "New_non_serving_RL_setup_selection", HFILL }},
    { &hf_rnsap_new_Serving_RL_in_DRNS,
      { "new-Serving-RL-in-DRNS", "rnsap.new_Serving_RL_in_DRNS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_new_Serving_RL_Not_in_DRNS,
      { "new-Serving-RL-Not-in-DRNS", "rnsap.new_Serving_RL_Not_in_DRNS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_new_Serving_RL_in_or_Not_in_DRNS,
      { "new-Serving-RL-in-or-Not-in-DRNS", "rnsap.new_Serving_RL_in_or_Not_in_DRNS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_item,
      { "Additional-E-DCH-New-non-serving-RL-E-DCH-FDD-DL-Control-Channel-InfoList item", "rnsap.Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_new_non_serving_RL_E_DCH_FDD_DL_Control_Channel_Information_A,
      { "new-non-serving-RL-E-DCH-FDD-DL-Control-Channel-Information-A", "rnsap.new_non_serving_RL_E_DCH_FDD_DL_Control_Channel_Information_A_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "EDCH_FDD_DL_ControlChannelInformation", HFILL }},
    { &hf_rnsap_new_non_serving_RL_E_DCH_FDD_DL_Control_Channel_Information_B,
      { "new-non-serving-RL-E-DCH-FDD-DL-Control-Channel-Information-B", "rnsap.new_non_serving_RL_E_DCH_FDD_DL_Control_Channel_Information_B_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "EDCH_FDD_DL_ControlChannelInformation", HFILL }},
    { &hf_rnsap_new_non_serving_RL_E_DCH_FDD_DL_Control_Channel_Information_C,
      { "new-non-serving-RL-E-DCH-FDD-DL-Control-Channel-Information-C", "rnsap.new_non_serving_RL_E_DCH_FDD_DL_Control_Channel_Information_C_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "EDCH_FDD_DL_ControlChannelInformation", HFILL }},
    { &hf_rnsap_measurementThreshold,
      { "measurementThreshold", "rnsap.measurementThreshold",
        FT_UINT32, BASE_DEC, VALS(rnsap_MeasurementThreshold_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_informationThreshold,
      { "informationThreshold", "rnsap.informationThreshold",
        FT_UINT32, BASE_DEC, VALS(rnsap_InformationThreshold_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_PCH_InformationList_item,
      { "PCH-InformationItem", "rnsap.PCH_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_informationReportPeriodicity,
      { "informationReportPeriodicity", "rnsap.informationReportPeriodicity",
        FT_UINT32, BASE_DEC, VALS(rnsap_InformationReportPeriodicity_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_imsi,
      { "imsi", "rnsap.imsi",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PLMN_List_item,
      { "PLMN-Identity", "rnsap.PLMN_Identity",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas_item,
      { "Associated-Secondary-CPICH", "rnsap.Associated_Secondary_CPICH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_associated_secondary_CPICH_channelId,
      { "associated-secondary-CPICH-channelId", "rnsap.associated_secondary_CPICH_channelId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CommonPhysicalChannelID", HFILL }},
    { &hf_rnsap_associated_secondary_CPICH,
      { "associated-secondary-CPICH", "rnsap.associated_secondary_CPICH",
        FT_INT32, BASE_DEC, NULL, 0,
        "PowerOffsetForSCPICH_DCPICHforMIMOwithfourtransmitantennas", HFILL }},
    { &hf_rnsap_associated_D_CPICH_channelId,
      { "associated-D-CPICH-channelId", "rnsap.associated_D_CPICH_channelId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CommonPhysicalChannelID", HFILL }},
    { &hf_rnsap_associated_D_CPICH,
      { "associated-D-CPICH", "rnsap.associated_D_CPICH",
        FT_INT32, BASE_DEC, NULL, 0,
        "PowerOffsetForSCPICH_DCPICHforMIMOwithfourtransmitantennas", HFILL }},
    { &hf_rnsap_defaultPreferredFrequency,
      { "defaultPreferredFrequency", "rnsap.defaultPreferredFrequency",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UARFCN", HFILL }},
    { &hf_rnsap_additionalPreferredFrequency,
      { "additionalPreferredFrequency", "rnsap.additionalPreferredFrequency",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PriorityQueue_InfoList_item,
      { "PriorityQueue-InfoItem", "rnsap.PriorityQueue_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_priorityQueue_Id,
      { "priorityQueue-Id", "rnsap.priorityQueue_Id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_associatedHSDSCH_MACdFlow,
      { "associatedHSDSCH-MACdFlow", "rnsap.associatedHSDSCH_MACdFlow",
        FT_UINT32, BASE_DEC, NULL, 0,
        "HSDSCH_MACdFlow_ID", HFILL }},
    { &hf_rnsap_t1,
      { "t1", "rnsap.t1",
        FT_UINT32, BASE_DEC, VALS(rnsap_T1_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_discardTimer,
      { "discardTimer", "rnsap.discardTimer",
        FT_UINT32, BASE_DEC|BASE_EXT_STRING, &rnsap_DiscardTimer_vals_ext, 0,
        NULL, HFILL }},
    { &hf_rnsap_mAC_hsWindowSize,
      { "mAC-hsWindowSize", "rnsap.mAC_hsWindowSize",
        FT_UINT32, BASE_DEC, VALS(rnsap_MAC_hsWindowSize_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_mAChsGuaranteedBitRate,
      { "mAChsGuaranteedBitRate", "rnsap.mAChsGuaranteedBitRate",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mACdPDU_Size_Index,
      { "mACdPDU-Size-Index", "rnsap.mACdPDU_Size_Index",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MACdPDU_Size_IndexList", HFILL }},
    { &hf_rnsap_rLC_Mode,
      { "rLC-Mode", "rnsap.rLC_Mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_RLC_Mode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_PriorityQueue_InfoList_EnhancedFACH_PCH_item,
      { "PriorityQueue-InfoItem-EnhancedFACH-PCH", "rnsap.PriorityQueue_InfoItem_EnhancedFACH_PCH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mAC_ehs_Reset_Timer,
      { "mAC-ehs-Reset-Timer", "rnsap.mAC_ehs_Reset_Timer",
        FT_UINT32, BASE_DEC, VALS(rnsap_MAC_ehs_Reset_Timer_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_maximum_MACdPDU_Size_01,
      { "maximum-MACdPDU-Size", "rnsap.maximum_MACdPDU_Size",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MAC_PDU_SizeExtended", HFILL }},
    { &hf_rnsap_PriorityQueue_InfoList_to_Modify_item,
      { "ModifyPriorityQueue", "rnsap.ModifyPriorityQueue",
        FT_UINT32, BASE_DEC, VALS(rnsap_ModifyPriorityQueue_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_mACdPDU_Size_Index_to_Modify,
      { "mACdPDU-Size-Index-to-Modify", "rnsap.mACdPDU_Size_Index_to_Modify",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MACdPDU_Size_IndexList_to_Modify", HFILL }},
    { &hf_rnsap_PriorityQueue_InfoList_to_Modify_Unsynchronised_item,
      { "PriorityQueue-InfoItem-to-Modify-Unsynchronised", "rnsap.PriorityQueue_InfoItem_to_Modify_Unsynchronised_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_priorityQueueId,
      { "priorityQueueId", "rnsap.priorityQueueId",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PriorityQueue_Id", HFILL }},
    { &hf_rnsap_mBMSChannelTypeInfo,
      { "mBMSChannelTypeInfo", "rnsap.mBMSChannelTypeInfo_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mBMSPreferredFreqLayerInfo,
      { "mBMSPreferredFreqLayerInfo", "rnsap.mBMSPreferredFreqLayerInfo_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UpPCH_InformationList_LCRTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uPPCHPositionLCR,
      { "uPPCHPositionLCR", "rnsap.uPPCHPositionLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PTMCellList_item,
      { "MBMSChannelTypeCellList", "rnsap.MBMSChannelTypeCellList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_PTPCellList_item,
      { "MBMSChannelTypeCellList", "rnsap.MBMSChannelTypeCellList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_raiList,
      { "raiList", "rnsap.raiList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RAI_List", HFILL }},
    { &hf_rnsap_RAI_List_item,
      { "RAI", "rnsap.RAI_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RB_Info_item,
      { "RB-Identity", "rnsap.RB_Identity",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Reference_E_TFCI_Information_item,
      { "Reference-E-TFCI-Information-Item", "rnsap.Reference_E_TFCI_Information_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_reference_E_TFCI,
      { "reference-E-TFCI", "rnsap.reference_E_TFCI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_TFCI", HFILL }},
    { &hf_rnsap_reference_E_TFCI_PO,
      { "reference-E-TFCI-PO", "rnsap.reference_E_TFCI_PO",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_pSDomain,
      { "pSDomain", "rnsap.pSDomain_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cSDomain,
      { "cSDomain", "rnsap.cSDomain_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_pS_CSDomain,
      { "pS-CSDomain", "rnsap.pS_CSDomain_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_periodic_02,
      { "periodic", "rnsap.periodic_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eventA,
      { "eventA", "rnsap.eventA_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eventB,
      { "eventB", "rnsap.eventB_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eventC,
      { "eventC", "rnsap.eventC_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eventD,
      { "eventD", "rnsap.eventD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eventE,
      { "eventE", "rnsap.eventE_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_eventF,
      { "eventF", "rnsap.eventF_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_extension_ReportCharacteristics,
      { "extension-ReportCharacteristics", "rnsap.extension_ReportCharacteristics_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ten_msec,
      { "ten-msec", "rnsap.ten_msec",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_6000_", HFILL }},
    { &hf_rnsap_gA_AccessPointPositionwithAltitude,
      { "gA-AccessPointPositionwithAltitude", "rnsap.gA_AccessPointPositionwithAltitude_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GA_AccessPointPositionwithOptionalAltitude", HFILL }},
    { &hf_rnsap_iPDLParameters,
      { "iPDLParameters", "rnsap.iPDLParameters",
        FT_UINT32, BASE_DEC, VALS(rnsap_IPDLParameters_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dGPSCorrections,
      { "dGPSCorrections", "rnsap.dGPSCorrections_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gPS_NavigationModel_and_TimeRecovery,
      { "gPS-NavigationModel-and-TimeRecovery", "rnsap.gPS_NavigationModel_and_TimeRecovery",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gPS_Ionospheric_Model,
      { "gPS-Ionospheric-Model", "rnsap.gPS_Ionospheric_Model_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gPS_UTC_Model,
      { "gPS-UTC-Model", "rnsap.gPS_UTC_Model_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gPS_Almanac,
      { "gPS-Almanac", "rnsap.gPS_Almanac_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gPS_RealTime_Integrity,
      { "gPS-RealTime-Integrity", "rnsap.gPS_RealTime_Integrity",
        FT_UINT32, BASE_DEC, VALS(rnsap_GPS_RealTime_Integrity_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_gPS_RX_POS,
      { "gPS-RX-POS", "rnsap.gPS_RX_POS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sFNSFN_GA_AccessPointPosition,
      { "sFNSFN-GA-AccessPointPosition", "rnsap.sFNSFN_GA_AccessPointPosition_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GA_AccessPointPositionwithOptionalAltitude", HFILL }},
    { &hf_rnsap_informationAvailable,
      { "informationAvailable", "rnsap.informationAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_informationNotAvailable,
      { "informationNotAvailable", "rnsap.informationNotAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Specific_DCH_Info_item,
      { "RL-Specific-DCH-Info-Item", "rnsap.RL_Specific_DCH_Info_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dCH_id,
      { "dCH-id", "rnsap.dCH_id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_Specific_EDCH_Info,
      { "rL-Specific-EDCH-Info", "rnsap.rL_Specific_EDCH_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Specific_EDCH_Info_item,
      { "RL-Specific-EDCH-InfoItem", "rnsap.RL_Specific_EDCH_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uplinkRTLoadValue,
      { "uplinkRTLoadValue", "rnsap.uplinkRTLoadValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_100", HFILL }},
    { &hf_rnsap_downlinkRTLoadValue,
      { "downlinkRTLoadValue", "rnsap.downlinkRTLoadValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_100", HFILL }},
    { &hf_rnsap_sAC,
      { "sAC", "rnsap.sAC",
        FT_UINT16, BASE_DEC_HEX, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SecondaryServingCells_item,
      { "SecondaryServingCellsItem", "rnsap.SecondaryServingCellsItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_secondaryC_ID,
      { "secondaryC-ID", "rnsap.secondaryC_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "C_ID", HFILL }},
    { &hf_rnsap_numSecondaryHS_SCCH_Codes,
      { "numSecondaryHS-SCCH-Codes", "rnsap.numSecondaryHS_SCCH_Codes",
        FT_UINT32, BASE_DEC, NULL, 0,
        "NumHS_SCCH_Codes", HFILL }},
    { &hf_rnsap_dl_TFCS,
      { "dl-TFCS", "rnsap.dl_TFCS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TFCS", HFILL }},
    { &hf_rnsap_tFCI_Coding,
      { "tFCI-Coding", "rnsap.tFCI_Coding",
        FT_UINT32, BASE_DEC, VALS(rnsap_TFCI_Coding_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_secondary_CCPCH_TDD_InformationList,
      { "secondary-CCPCH-TDD-InformationList", "rnsap.secondary_CCPCH_TDD_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_fACH_InformationList,
      { "fACH-InformationList", "rnsap.fACH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_pCH_InformationList,
      { "pCH-InformationList", "rnsap.pCH_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_new_secondary_CPICH,
      { "new-secondary-CPICH", "rnsap.new_secondary_CPICH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Secondary_CPICH_Information", HFILL }},
    { &hf_rnsap_secondary_CPICH_shall_not_be_used,
      { "secondary-CPICH-shall-not-be-used", "rnsap.secondary_CPICH_shall_not_be_used_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_secondary_LCR_CCPCH_TDD_InformationList,
      { "secondary-LCR-CCPCH-TDD-InformationList", "rnsap.secondary_LCR_CCPCH_TDD_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_secondary_CCPCH_TDD_InformationList768,
      { "secondary-CCPCH-TDD-InformationList768", "rnsap.secondary_CCPCH_TDD_InformationList768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Secondary_CCPCH_TDD_InformationList_item,
      { "Secondary-CCPCH-TDD-InformationItem", "rnsap.Secondary_CCPCH_TDD_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_secondary_CCPCH_TDD_Code_Information,
      { "secondary-CCPCH-TDD-Code-Information", "rnsap.secondary_CCPCH_TDD_Code_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tDD_PhysicalChannelOffset,
      { "tDD-PhysicalChannelOffset", "rnsap.tDD_PhysicalChannelOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Secondary_LCR_CCPCH_TDD_InformationList_item,
      { "Secondary-LCR-CCPCH-TDD-InformationItem", "rnsap.Secondary_LCR_CCPCH_TDD_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_secondary_LCR_CCPCH_TDD_Code_Information,
      { "secondary-LCR-CCPCH-TDD-Code-Information", "rnsap.secondary_LCR_CCPCH_TDD_Code_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Secondary_CCPCH_TDD_InformationList768_item,
      { "Secondary-CCPCH-TDD-InformationItem768", "rnsap.Secondary_CCPCH_TDD_InformationItem768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_secondary_CCPCH_TDD_Code_Information768,
      { "secondary-CCPCH-TDD-Code-Information768", "rnsap.secondary_CCPCH_TDD_Code_Information768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Secondary_CCPCH_TDD_Code_Information_item,
      { "Secondary-CCPCH-TDD-Code-InformationItem", "rnsap.Secondary_CCPCH_TDD_Code_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Secondary_LCR_CCPCH_TDD_Code_Information_item,
      { "Secondary-LCR-CCPCH-TDD-Code-InformationItem", "rnsap.Secondary_LCR_CCPCH_TDD_Code_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_s_CCPCH_TimeSlotFormat_LCR,
      { "s-CCPCH-TimeSlotFormat-LCR", "rnsap.s_CCPCH_TimeSlotFormat_LCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR_vals), 0,
        "TDD_DL_DPCH_TimeSlotFormat_LCR", HFILL }},
    { &hf_rnsap_Secondary_CCPCH_TDD_Code_Information768_item,
      { "Secondary-CCPCH-TDD-Code-InformationItem768", "rnsap.Secondary_CCPCH_TDD_Code_InformationItem768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_possible_Secondary_Serving_Cell_List,
      { "possible-Secondary-Serving-Cell-List", "rnsap.possible_Secondary_Serving_Cell_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Possible_Secondary_Serving_Cell_List_item,
      { "Possible-Secondary-Serving-Cell", "rnsap.Possible_Secondary_Serving_Cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SetsOfHS_SCCH_Codes_item,
      { "SetsOfHS-SCCH-CodesItem", "rnsap.SetsOfHS_SCCH_CodesItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_SCCH_PreconfiguredCodes,
      { "hS-SCCH-PreconfiguredCodes", "rnsap.hS_SCCH_PreconfiguredCodes",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sixtyfourQAM_DL_SupportIndicator,
      { "sixtyfourQAM-DL-SupportIndicator", "rnsap.sixtyfourQAM_DL_SupportIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SixtyfourQAM_DL_SupportIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_setup_02,
      { "setup", "rnsap.setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Additional_EDCH_Setup_Info", HFILL }},
    { &hf_rnsap_configurationChange_01,
      { "configurationChange", "rnsap.configurationChange",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_Cell_Information_ConfigurationChange_List", HFILL }},
    { &hf_rnsap_removal,
      { "removal", "rnsap.removal",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_Cell_Information_Removal_List", HFILL }},
    { &hf_rnsap_setup_03,
      { "setup", "rnsap.setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Downlink_TPC_enhancements_Information", HFILL }},
    { &hf_rnsap_configurationChange_02,
      { "configurationChange", "rnsap.configurationChange_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Downlink_TPC_enhancements_Information", HFILL }},
    { &hf_rnsap_removal_01,
      { "removal", "rnsap.removal",
        FT_UINT32, BASE_DEC, VALS(rnsap_Downlink_TPC_enhancements_Information_Removal_vals), 0,
        "Downlink_TPC_enhancements_Information_Removal", HFILL }},
    { &hf_rnsap_setup_04,
      { "setup", "rnsap.setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_CLTD_Information", HFILL }},
    { &hf_rnsap_configurationChange_03,
      { "configurationChange", "rnsap.configurationChange_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_CLTD_Information_To_Modify", HFILL }},
    { &hf_rnsap_removal_02,
      { "removal", "rnsap.removal",
        FT_UINT32, BASE_DEC, VALS(rnsap_UL_CLTD_Information_Removal_vals), 0,
        "UL_CLTD_Information_Removal", HFILL }},
    { &hf_rnsap_setup_05,
      { "setup", "rnsap.setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "FTPICH_Information", HFILL }},
    { &hf_rnsap_configurationChange_04,
      { "configurationChange", "rnsap.configurationChange_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "FTPICH_Information_To_Modify", HFILL }},
    { &hf_rnsap_removal_03,
      { "removal", "rnsap.removal",
        FT_UINT32, BASE_DEC, VALS(rnsap_FTPICH_Information_Removal_vals), 0,
        "FTPICH_Information_Removal", HFILL }},
    { &hf_rnsap_setup_06,
      { "setup", "rnsap.setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DCH_ENH_Information", HFILL }},
    { &hf_rnsap_configurationChange_05,
      { "configurationChange", "rnsap.configurationChange_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DCH_ENH_Information_to_Modify", HFILL }},
    { &hf_rnsap_removal_04,
      { "removal", "rnsap.removal",
        FT_UINT32, BASE_DEC, VALS(rnsap_DCH_ENH_Information_Removal_vals), 0,
        "DCH_ENH_Information_Removal", HFILL }},
    { &hf_rnsap_setup_07,
      { "setup", "rnsap.setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCCH2_Information", HFILL }},
    { &hf_rnsap_configurationChange_06,
      { "configurationChange", "rnsap.configurationChange_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCCH2_Information_To_Modify", HFILL }},
    { &hf_rnsap_removal_05,
      { "removal", "rnsap.removal",
        FT_UINT32, BASE_DEC, VALS(rnsap_UL_DPCCH2_Information_Removal_vals), 0,
        "UL_DPCCH2_Information_Removal", HFILL }},
    { &hf_rnsap_sFNSFNChangeLimit,
      { "sFNSFNChangeLimit", "rnsap.sFNSFNChangeLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_predictedSFNSFNDeviationLimit,
      { "predictedSFNSFNDeviationLimit", "rnsap.predictedSFNSFNDeviationLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation,
      { "successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation", "rnsap.successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item,
      { "successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation item", "rnsap.successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sFNSFNValue,
      { "sFNSFNValue", "rnsap.sFNSFNValue",
        FT_UINT32, BASE_DEC, VALS(rnsap_SFNSFNValue_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_sFNSFNQuality,
      { "sFNSFNQuality", "rnsap.sFNSFNQuality",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sFNSFNDriftRate,
      { "sFNSFNDriftRate", "rnsap.sFNSFNDriftRate",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sFNSFNDriftRateQuality,
      { "sFNSFNDriftRateQuality", "rnsap.sFNSFNDriftRateQuality",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sFNSFNTimeStampInformation,
      { "sFNSFNTimeStampInformation", "rnsap.sFNSFNTimeStampInformation",
        FT_UINT32, BASE_DEC, VALS(rnsap_SFNSFNTimeStampInformation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation,
      { "unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation", "rnsap.unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item,
      { "unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation item", "rnsap.unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sFNSFNTimeStamp_FDD,
      { "sFNSFNTimeStamp-FDD", "rnsap.sFNSFNTimeStamp_FDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SFN", HFILL }},
    { &hf_rnsap_sFNSFNTimeStamp_TDD,
      { "sFNSFNTimeStamp-TDD", "rnsap.sFNSFNTimeStamp_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sFN,
      { "sFN", "rnsap.sFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sFNSFN_FDD,
      { "sFNSFN-FDD", "rnsap.sFNSFN_FDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sFNSFN_TDD,
      { "sFNSFN-TDD", "rnsap.sFNSFN_TDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sFNSFN_TDD768,
      { "sFNSFN-TDD768", "rnsap.sFNSFN_TDD768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_listOfSNAs,
      { "listOfSNAs", "rnsap.listOfSNAs",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ListOfSNAs_item,
      { "SNACode", "rnsap.SNACode",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sRNTI,
      { "sRNTI", "rnsap.sRNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "S_RNTI", HFILL }},
    { &hf_rnsap_sRNTI_BitMaskIndex,
      { "sRNTI-BitMaskIndex", "rnsap.sRNTI_BitMaskIndex",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_sRNTI_BitMaskIndex_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_n_INSYNC_IND,
      { "n-INSYNC-IND", "rnsap.n_INSYNC_IND",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_256", HFILL }},
    { &hf_rnsap_n_OUTSYNC_IND,
      { "n-OUTSYNC-IND", "rnsap.n_OUTSYNC_IND",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_256", HFILL }},
    { &hf_rnsap_t_RLFAILURE,
      { "t-RLFAILURE", "rnsap.t_RLFAILURE",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_255", HFILL }},
    { &hf_rnsap_maxSYNC_UL_transmissions,
      { "maxSYNC-UL-transmissions", "rnsap.maxSYNC_UL_transmissions",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_maxSYNC_UL_transmissions_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_powerRampStep,
      { "powerRampStep", "rnsap.powerRampStep",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_3_", HFILL }},
    { &hf_rnsap_modulation,
      { "modulation", "rnsap.modulation",
        FT_UINT32, BASE_DEC, VALS(rnsap_Modulation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TDD_DCHs_to_Modify_item,
      { "TDD-DCHs-to-ModifyItem", "rnsap.TDD_DCHs_to_ModifyItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dCH_SpecificInformationList_03,
      { "dCH-SpecificInformationList", "rnsap.dCH_SpecificInformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_DCHs_to_ModifySpecificInformationList", HFILL }},
    { &hf_rnsap_TDD_DCHs_to_ModifySpecificInformationList_item,
      { "TDD-DCHs-to-ModifySpecificItem", "rnsap.TDD_DCHs_to_ModifySpecificItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_CCTrCH_ID,
      { "ul-CCTrCH-ID", "rnsap.ul_CCTrCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_ID", HFILL }},
    { &hf_rnsap_dl_CCTrCH_ID,
      { "dl-CCTrCH-ID", "rnsap.dl_CCTrCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_ID", HFILL }},
    { &hf_rnsap_TDD_DL_Code_Information_item,
      { "TDD-DL-Code-InformationItem", "rnsap.TDD_DL_Code_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dPCH_ID,
      { "dPCH-ID", "rnsap.dPCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TDD_DL_Code_LCR_Information_item,
      { "TDD-DL-Code-LCR-InformationItem", "rnsap.TDD_DL_Code_LCR_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tdd_ChannelisationCodeLCR,
      { "tdd-ChannelisationCodeLCR", "rnsap.tdd_ChannelisationCodeLCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tdd_DL_DPCH_TimeSlotFormat_LCR,
      { "tdd-DL-DPCH-TimeSlotFormat-LCR", "rnsap.tdd_DL_DPCH_TimeSlotFormat_LCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TDD_DL_Code_Information768_item,
      { "TDD-DL-Code-InformationItem768", "rnsap.TDD_DL_Code_InformationItem768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_qPSK,
      { "qPSK", "rnsap.qPSK",
        FT_UINT32, BASE_DEC, NULL, 0,
        "QPSK_DL_DPCH_TimeSlotFormatTDD_LCR", HFILL }},
    { &hf_rnsap_eightPSK,
      { "eightPSK", "rnsap.eightPSK",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EightPSK_DL_DPCH_TimeSlotFormatTDD_LCR", HFILL }},
    { &hf_rnsap_initialOffset,
      { "initialOffset", "rnsap.initialOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_255", HFILL }},
    { &hf_rnsap_noinitialOffset,
      { "noinitialOffset", "rnsap.noinitialOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_63", HFILL }},
    { &hf_rnsap_TDD_UL_Code_Information_item,
      { "TDD-UL-Code-InformationItem", "rnsap.TDD_UL_Code_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TDD_UL_Code_LCR_Information_item,
      { "TDD-UL-Code-LCR-InformationItem", "rnsap.TDD_UL_Code_LCR_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tdd_UL_DPCH_TimeSlotFormat_LCR,
      { "tdd-UL-DPCH-TimeSlotFormat-LCR", "rnsap.tdd_UL_DPCH_TimeSlotFormat_LCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_TDD_UL_DPCH_TimeSlotFormat_LCR_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_TDD_UL_Code_Information768_item,
      { "TDD-UL-Code-InformationItem768", "rnsap.TDD_UL_Code_InformationItem768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_qPSK_01,
      { "qPSK", "rnsap.qPSK",
        FT_UINT32, BASE_DEC, NULL, 0,
        "QPSK_UL_DPCH_TimeSlotFormatTDD_LCR", HFILL }},
    { &hf_rnsap_eightPSK_01,
      { "eightPSK", "rnsap.eightPSK",
        FT_UINT32, BASE_DEC, NULL, 0,
        "EightPSK_UL_DPCH_TimeSlotFormatTDD_LCR", HFILL }},
    { &hf_rnsap_plmn_id,
      { "plmn-id", "rnsap.plmn_id",
        FT_BYTES, BASE_NONE, NULL, 0,
        "PLMN_Identity", HFILL }},
    { &hf_rnsap_service_id,
      { "service-id", "rnsap.service_id",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dsField,
      { "dsField", "rnsap.dsField",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_genericTrafficCategory,
      { "genericTrafficCategory", "rnsap.genericTrafficCategory",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Transmission_Gap_Pattern_Sequence_Information_item,
      { "Transmission-Gap-Pattern-Sequence-Information item", "rnsap.Transmission_Gap_Pattern_Sequence_Information_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tGPSID,
      { "tGPSID", "rnsap.tGPSID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tGSN,
      { "tGSN", "rnsap.tGSN",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tGL1,
      { "tGL1", "rnsap.tGL1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GapLength", HFILL }},
    { &hf_rnsap_tGL2,
      { "tGL2", "rnsap.tGL2",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GapLength", HFILL }},
    { &hf_rnsap_tGD,
      { "tGD", "rnsap.tGD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tGPL1,
      { "tGPL1", "rnsap.tGPL1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GapDuration", HFILL }},
    { &hf_rnsap_not_to_be_used_1,
      { "not-to-be-used-1", "rnsap.not_to_be_used_1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "GapDuration", HFILL }},
    { &hf_rnsap_uL_DL_mode,
      { "uL-DL-mode", "rnsap.uL_DL_mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_UL_DL_mode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_downlink_Compressed_Mode_Method,
      { "downlink-Compressed-Mode-Method", "rnsap.downlink_Compressed_Mode_Method",
        FT_UINT32, BASE_DEC, VALS(rnsap_Downlink_Compressed_Mode_Method_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uplink_Compressed_Mode_Method,
      { "uplink-Compressed-Mode-Method", "rnsap.uplink_Compressed_Mode_Method",
        FT_UINT32, BASE_DEC, VALS(rnsap_Uplink_Compressed_Mode_Method_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dL_FrameType,
      { "dL-FrameType", "rnsap.dL_FrameType",
        FT_UINT32, BASE_DEC, VALS(rnsap_DL_FrameType_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_delta_SIR1,
      { "delta-SIR1", "rnsap.delta_SIR1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DeltaSIR", HFILL }},
    { &hf_rnsap_delta_SIR_after1,
      { "delta-SIR-after1", "rnsap.delta_SIR_after1",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DeltaSIR", HFILL }},
    { &hf_rnsap_delta_SIR2,
      { "delta-SIR2", "rnsap.delta_SIR2",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DeltaSIR", HFILL }},
    { &hf_rnsap_delta_SIR_after2,
      { "delta-SIR-after2", "rnsap.delta_SIR_after2",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DeltaSIR", HFILL }},
    { &hf_rnsap_Transmission_Gap_Pattern_Sequence_Status_List_item,
      { "Transmission-Gap-Pattern-Sequence-Status-List item", "rnsap.Transmission_Gap_Pattern_Sequence_Status_List_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tGPRC,
      { "tGPRC", "rnsap.tGPRC",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tGCFN,
      { "tGCFN", "rnsap.tGCFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CFN", HFILL }},
    { &hf_rnsap_Affected_HSDSCH_Serving_Cell_List_item,
      { "C-ID", "rnsap.C_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Transmission_Mode_Information_item,
      { "Transmission-Mode-Information-List", "rnsap.Transmission_Mode_Information_List_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mS,
      { "mS", "rnsap.mS",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_16383", HFILL }},
    { &hf_rnsap_lS,
      { "lS", "rnsap.lS",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_4294967295", HFILL }},
    { &hf_rnsap_tUTRANGANSSChangeLimit,
      { "tUTRANGANSSChangeLimit", "rnsap.tUTRANGANSSChangeLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_256", HFILL }},
    { &hf_rnsap_predictedTUTRANGANSSDeviationLimit,
      { "predictedTUTRANGANSSDeviationLimit", "rnsap.predictedTUTRANGANSSDeviationLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_256", HFILL }},
    { &hf_rnsap_tUTRANGANSS,
      { "tUTRANGANSS", "rnsap.tUTRANGANSS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tUTRANGANSSQuality,
      { "tUTRANGANSSQuality", "rnsap.tUTRANGANSSQuality",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_255", HFILL }},
    { &hf_rnsap_tUTRANGANSSDriftRate,
      { "tUTRANGANSSDriftRate", "rnsap.tUTRANGANSSDriftRate",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M50_50", HFILL }},
    { &hf_rnsap_tUTRANGANSSDriftRateQuality,
      { "tUTRANGANSSDriftRateQuality", "rnsap.tUTRANGANSSDriftRateQuality",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_50", HFILL }},
    { &hf_rnsap_ms_part,
      { "ms-part", "rnsap.ms_part",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_16383", HFILL }},
    { &hf_rnsap_ls_part,
      { "ls-part", "rnsap.ls_part",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_4294967295", HFILL }},
    { &hf_rnsap_tUTRANGPSChangeLimit,
      { "tUTRANGPSChangeLimit", "rnsap.tUTRANGPSChangeLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_predictedTUTRANGPSDeviationLimit,
      { "predictedTUTRANGPSDeviationLimit", "rnsap.predictedTUTRANGPSDeviationLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tUTRANGPS,
      { "tUTRANGPS", "rnsap.tUTRANGPS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tUTRANGPSQuality,
      { "tUTRANGPSQuality", "rnsap.tUTRANGPSQuality",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tUTRANGPSDriftRate,
      { "tUTRANGPSDriftRate", "rnsap.tUTRANGPSDriftRate",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tUTRANGPSDriftRateQuality,
      { "tUTRANGPSDriftRateQuality", "rnsap.tUTRANGPSDriftRateQuality",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_iEe_Extensions,
      { "iEe-Extensions", "rnsap.iEe_Extensions",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ProtocolExtensionContainer", HFILL }},
    { &hf_rnsap_signalledGainFactors,
      { "signalledGainFactors", "rnsap.signalledGainFactors_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_betaC,
      { "betaC", "rnsap.betaC",
        FT_UINT32, BASE_DEC, NULL, 0,
        "BetaCD", HFILL }},
    { &hf_rnsap_betaD,
      { "betaD", "rnsap.betaD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "BetaCD", HFILL }},
    { &hf_rnsap_refTFCNumber,
      { "refTFCNumber", "rnsap.refTFCNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tFCSvalues,
      { "tFCSvalues", "rnsap.tFCSvalues",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_tFCSvalues_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_no_Split_in_TFCI,
      { "no-Split-in-TFCI", "rnsap.no_Split_in_TFCI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TFCS_TFCSList", HFILL }},
    { &hf_rnsap_not_Used_split_in_TFCI,
      { "not-Used-split-in-TFCI", "rnsap.not_Used_split_in_TFCI_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TFCS_TFCSList_item,
      { "TFCS-TFCSList item", "rnsap.TFCS_TFCSList_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cTFC,
      { "cTFC", "rnsap.cTFC",
        FT_UINT32, BASE_DEC, VALS(rnsap_TFCS_CTFC_vals), 0,
        "TFCS_CTFC", HFILL }},
    { &hf_rnsap_tFC_Beta,
      { "tFC-Beta", "rnsap.tFC_Beta",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransportFormatCombination_Beta_vals), 0,
        "TransportFormatCombination_Beta", HFILL }},
    { &hf_rnsap_ctfc2bit,
      { "ctfc2bit", "rnsap.ctfc2bit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_3", HFILL }},
    { &hf_rnsap_ctfc4bit,
      { "ctfc4bit", "rnsap.ctfc4bit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_15", HFILL }},
    { &hf_rnsap_ctfc6bit,
      { "ctfc6bit", "rnsap.ctfc6bit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_63", HFILL }},
    { &hf_rnsap_ctfc8bit,
      { "ctfc8bit", "rnsap.ctfc8bit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_255", HFILL }},
    { &hf_rnsap_ctfc12bit,
      { "ctfc12bit", "rnsap.ctfc12bit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_4095", HFILL }},
    { &hf_rnsap_ctfc16bit,
      { "ctfc16bit", "rnsap.ctfc16bit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_65535", HFILL }},
    { &hf_rnsap_ctfcmaxbit,
      { "ctfcmaxbit", "rnsap.ctfcmaxbit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_maxCTFC", HFILL }},
    { &hf_rnsap_slot_position,
      { "slot-position", "rnsap.slot_position",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_4", HFILL }},
    { &hf_rnsap_dynamicParts,
      { "dynamicParts", "rnsap.dynamicParts",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TransportFormatSet_DynamicPartList", HFILL }},
    { &hf_rnsap_semi_staticPart,
      { "semi-staticPart", "rnsap.semi_staticPart_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TransportFormatSet_Semi_staticPart", HFILL }},
    { &hf_rnsap_TransportFormatSet_DynamicPartList_item,
      { "TransportFormatSet-DynamicPartList item", "rnsap.TransportFormatSet_DynamicPartList_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_nrOfTransportBlocks,
      { "nrOfTransportBlocks", "rnsap.nrOfTransportBlocks",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_transportBlockSize,
      { "transportBlockSize", "rnsap.transportBlockSize",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mode_01,
      { "mode", "rnsap.mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransportFormatSet_ModeDP_vals), 0,
        "TransportFormatSet_ModeDP", HFILL }},
    { &hf_rnsap_tdd,
      { "tdd", "rnsap.tdd_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TDD_TransportFormatSet_ModeDP", HFILL }},
    { &hf_rnsap_notApplicable,
      { "notApplicable", "rnsap.notApplicable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_transmissionTimeIntervalInformation,
      { "transmissionTimeIntervalInformation", "rnsap.transmissionTimeIntervalInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_TransmissionTimeIntervalInformation_item,
      { "TransmissionTimeIntervalInformation item", "rnsap.TransmissionTimeIntervalInformation_item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_transmissionTimeInterval,
      { "transmissionTimeInterval", "rnsap.transmissionTimeInterval",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransmissionTimeIntervalDynamic_vals), 0,
        "TransmissionTimeIntervalDynamic", HFILL }},
    { &hf_rnsap_transmissionTime,
      { "transmissionTime", "rnsap.transmissionTime",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransmissionTimeIntervalSemiStatic_vals), 0,
        "TransmissionTimeIntervalSemiStatic", HFILL }},
    { &hf_rnsap_channelCoding,
      { "channelCoding", "rnsap.channelCoding",
        FT_UINT32, BASE_DEC, VALS(rnsap_ChannelCodingType_vals), 0,
        "ChannelCodingType", HFILL }},
    { &hf_rnsap_codingRate,
      { "codingRate", "rnsap.codingRate",
        FT_UINT32, BASE_DEC, VALS(rnsap_CodingRate_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_rateMatcingAttribute,
      { "rateMatcingAttribute", "rnsap.rateMatcingAttribute",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RateMatchingAttribute", HFILL }},
    { &hf_rnsap_cRC_Size,
      { "cRC-Size", "rnsap.cRC_Size",
        FT_UINT32, BASE_DEC, VALS(rnsap_CRC_Size_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_mode_02,
      { "mode", "rnsap.mode",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransportFormatSet_ModeSSP_vals), 0,
        "TransportFormatSet_ModeSSP", HFILL }},
    { &hf_rnsap_tdd_01,
      { "tdd", "rnsap.tdd",
        FT_UINT32, BASE_DEC, VALS(rnsap_SecondInterleavingMode_vals), 0,
        "SecondInterleavingMode", HFILL }},
    { &hf_rnsap_tTI_Update_CFN,
      { "tTI-Update-CFN", "rnsap.tTI_Update_CFN",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CFN", HFILL }},
    { &hf_rnsap_tTI_Update_Ind,
      { "tTI-Update-Ind", "rnsap.tTI_Update_Ind_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_AggregateMaximumBitRateDownlink,
      { "uE-AggregateMaximumBitRateDownlink", "rnsap.uE_AggregateMaximumBitRateDownlink",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_AggregateMaximumBitRateUplink,
      { "uE-AggregateMaximumBitRateUplink", "rnsap.uE_AggregateMaximumBitRateUplink",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UE_RF_Band_CapabilityLCR_item,
      { "Radio-Frequency-BandItem", "rnsap.Radio_Frequency_BandItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_radio_Frequency_Band,
      { "radio-Frequency-Band", "rnsap.radio_Frequency_Band",
        FT_UINT32, BASE_DEC, VALS(rnsap_Radio_Frequency_Band_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_imei,
      { "imei", "rnsap.imei",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_imeisv,
      { "imeisv", "rnsap.imeisv",
        FT_BYTES, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_periodic_03,
      { "periodic", "rnsap.periodic_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UEMeasurementReportCharacteristicsPeriodic", HFILL }},
    { &hf_rnsap_event1h,
      { "event1h", "rnsap.event1h_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UEMeasurementReportCharacteristicsEvent1h", HFILL }},
    { &hf_rnsap_event1i,
      { "event1i", "rnsap.event1i_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UEMeasurementReportCharacteristicsEvent1i", HFILL }},
    { &hf_rnsap_event6a,
      { "event6a", "rnsap.event6a_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UEMeasurementReportCharacteristicsEvent6a", HFILL }},
    { &hf_rnsap_event6b,
      { "event6b", "rnsap.event6b_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UEMeasurementReportCharacteristicsEvent6b", HFILL }},
    { &hf_rnsap_event6c,
      { "event6c", "rnsap.event6c_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UEMeasurementReportCharacteristicsEvent6c", HFILL }},
    { &hf_rnsap_event6d,
      { "event6d", "rnsap.event6d_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UEMeasurementReportCharacteristicsEvent6d", HFILL }},
    { &hf_rnsap_extension_ReportCharacteristics_01,
      { "extension-ReportCharacteristics", "rnsap.extension_ReportCharacteristics_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UEMeasurementReportCharacteristics_Extension", HFILL }},
    { &hf_rnsap_uEMeasurementTreshold,
      { "uEMeasurementTreshold", "rnsap.uEMeasurementTreshold",
        FT_UINT32, BASE_DEC, VALS(rnsap_UEMeasurementThreshold_vals), 0,
        "UEMeasurementThreshold", HFILL }},
    { &hf_rnsap_uEMeasurementTimeToTrigger,
      { "uEMeasurementTimeToTrigger", "rnsap.uEMeasurementTimeToTrigger",
        FT_UINT32, BASE_DEC, VALS(rnsap_UEMeasurementTimeToTrigger_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uEMeasurementHysteresisTime,
      { "uEMeasurementHysteresisTime", "rnsap.uEMeasurementHysteresisTime",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_amountofReporting,
      { "amountofReporting", "rnsap.amountofReporting",
        FT_UINT32, BASE_DEC, VALS(rnsap_UEMeasurementReportCharacteristicsPeriodicAmountofReporting_vals), 0,
        "UEMeasurementReportCharacteristicsPeriodicAmountofReporting", HFILL }},
    { &hf_rnsap_reportingInterval,
      { "reportingInterval", "rnsap.reportingInterval",
        FT_UINT32, BASE_DEC, VALS(rnsap_UEMeasurementReportCharacteristicsPeriodicReportingInterval_vals), 0,
        "UEMeasurementReportCharacteristicsPeriodicReportingInterval", HFILL }},
    { &hf_rnsap_timeslotISCP,
      { "timeslotISCP", "rnsap.timeslotISCP",
        FT_INT32, BASE_DEC, NULL, 0,
        "UEMeasurementThresholdDLTimeslotISCP", HFILL }},
    { &hf_rnsap_uETransmitPower,
      { "uETransmitPower", "rnsap.uETransmitPower",
        FT_INT32, BASE_DEC, NULL, 0,
        "UEMeasurementThresholdUETransmitPower", HFILL }},
    { &hf_rnsap_extension_UEMeasurementThreshold,
      { "extension-UEMeasurementThreshold", "rnsap.extension_UEMeasurementThreshold_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UEMeasurementThreshold_Extension", HFILL }},
    { &hf_rnsap_UEMeasurementTimeslotInfoHCR_item,
      { "UEMeasurementTimeslotInfoHCR-IEs", "rnsap.UEMeasurementTimeslotInfoHCR_IEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_burstType,
      { "burstType", "rnsap.burstType",
        FT_UINT32, BASE_DEC, VALS(rnsap_UEMeasurementTimeslotInfoHCRBurstType_vals), 0,
        "UEMeasurementTimeslotInfoHCRBurstType", HFILL }},
    { &hf_rnsap_UEMeasurementTimeslotInfoLCR_item,
      { "UEMeasurementTimeslotInfoLCR-IEs", "rnsap.UEMeasurementTimeslotInfoLCR_IEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementTimeslotInfo768_item,
      { "UEMeasurementTimeslotInfo768-IEs", "rnsap.UEMeasurementTimeslotInfo768_IEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_burstType_01,
      { "burstType", "rnsap.burstType",
        FT_UINT32, BASE_DEC, VALS(rnsap_UEMeasurementTimeslotInfo768BurstType_vals), 0,
        "UEMeasurementTimeslotInfo768BurstType", HFILL }},
    { &hf_rnsap_uE_Transmitted_Power,
      { "uE-Transmitted-Power", "rnsap.uE_Transmitted_Power_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UE_MeasurementValue_UE_Transmitted_Power", HFILL }},
    { &hf_rnsap_primary_CCPCH_RSCP,
      { "primary-CCPCH-RSCP", "rnsap.primary_CCPCH_RSCP_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UE_MeasurementValue_Primary_CCPCH_RSCP", HFILL }},
    { &hf_rnsap_dL_Timeslot_ISCP,
      { "dL-Timeslot-ISCP", "rnsap.dL_Timeslot_ISCP_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UE_MeasurementValue_DL_Timeslot_ISCP", HFILL }},
    { &hf_rnsap_extension_UEMeasurementValue,
      { "extension-UEMeasurementValue", "rnsap.extension_UEMeasurementValue_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UEMeasurementValue_Extension", HFILL }},
    { &hf_rnsap_uEMeasurementTransmittedPowerListHCR,
      { "uEMeasurementTransmittedPowerListHCR", "rnsap.uEMeasurementTransmittedPowerListHCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UEMeasurementValueTransmittedPowerListHCR", HFILL }},
    { &hf_rnsap_uEMeasurementTransmittedPowerListLCR,
      { "uEMeasurementTransmittedPowerListLCR", "rnsap.uEMeasurementTransmittedPowerListLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UEMeasurementValueTransmittedPowerListLCR", HFILL }},
    { &hf_rnsap_UEMeasurementValueTransmittedPowerListHCR_item,
      { "UEMeasurementValueTransmittedPowerListHCR-IEs", "rnsap.UEMeasurementValueTransmittedPowerListHCR_IEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uETransmitPower_01,
      { "uETransmitPower", "rnsap.uETransmitPower",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_104", HFILL }},
    { &hf_rnsap_UEMeasurementValueTransmittedPowerListLCR_item,
      { "UEMeasurementValueTransmittedPowerListLCR-IEs", "rnsap.UEMeasurementValueTransmittedPowerListLCR_IEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementValueTransmittedPowerList768_item,
      { "UEMeasurementValueTransmittedPowerList768-IEs", "rnsap.UEMeasurementValueTransmittedPowerList768_IEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_primaryCCPCH_RSCP,
      { "primaryCCPCH-RSCP", "rnsap.primaryCCPCH_RSCP",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_primaryCCPCH_RSCP_Delta,
      { "primaryCCPCH-RSCP-Delta", "rnsap.primaryCCPCH_RSCP_Delta",
        FT_INT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uEMeasurementTimeslotISCPListHCR,
      { "uEMeasurementTimeslotISCPListHCR", "rnsap.uEMeasurementTimeslotISCPListHCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UEMeasurementValueTimeslotISCPListHCR", HFILL }},
    { &hf_rnsap_uEMeasurementTimeslotISCPListLCR,
      { "uEMeasurementTimeslotISCPListLCR", "rnsap.uEMeasurementTimeslotISCPListLCR",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UEMeasurementValueTimeslotISCPListLCR", HFILL }},
    { &hf_rnsap_UEMeasurementValueTimeslotISCPListHCR_item,
      { "UEMeasurementValueTimeslotISCPListHCR-IEs", "rnsap.UEMeasurementValueTimeslotISCPListHCR_IEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementValueTimeslotISCPListLCR_item,
      { "UEMeasurementValueTimeslotISCPListLCR-IEs", "rnsap.UEMeasurementValueTimeslotISCPListLCR_IEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UEMeasurementValueTimeslotISCPList768_item,
      { "UEMeasurementValueTimeslotISCPList768-IEs", "rnsap.UEMeasurementValueTimeslotISCPList768_IEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_measurementAvailable_02,
      { "measurementAvailable", "rnsap.measurementAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UEMeasurementValueInformationAvailable", HFILL }},
    { &hf_rnsap_measurementnotAvailable_02,
      { "measurementnotAvailable", "rnsap.measurementnotAvailable_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UEMeasurementValueInformationnotAvailable", HFILL }},
    { &hf_rnsap_uEmeasurementValue,
      { "uEmeasurementValue", "rnsap.uEmeasurementValue",
        FT_UINT32, BASE_DEC, VALS(rnsap_UEMeasurementValue_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uPHFiltering_Value,
      { "uPHFiltering-Value", "rnsap.uPHFiltering_Value",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_extension_UE_Measurement_Value,
      { "extension-UE-Measurement-Value", "rnsap.extension_UE_Measurement_Value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cell_fach_pch,
      { "cell-fach-pch", "rnsap.cell_fach_pch_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_Fach_Pch_State", HFILL }},
    { &hf_rnsap_ura_pch,
      { "ura-pch", "rnsap.ura_pch_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Ura_Pch_State", HFILL }},
    { &hf_rnsap_d_RNTI,
      { "d-RNTI", "rnsap.d_RNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_srnc_id,
      { "srnc-id", "rnsap.srnc_id",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RNC_ID", HFILL }},
    { &hf_rnsap_ura_id,
      { "ura-id", "rnsap.ura_id",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_Timeslot_Information_item,
      { "UL-Timeslot-InformationItem", "rnsap.UL_Timeslot_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_Code_Information,
      { "uL-Code-Information", "rnsap.uL_Code_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_UL_Code_Information", HFILL }},
    { &hf_rnsap_UL_TimeslotLCR_Information_item,
      { "UL-TimeslotLCR-InformationItem", "rnsap.UL_TimeslotLCR_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_Code_LCR_InformationList,
      { "uL-Code-LCR-InformationList", "rnsap.uL_Code_LCR_InformationList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_UL_Code_LCR_Information", HFILL }},
    { &hf_rnsap_sequenceNumber,
      { "sequenceNumber", "rnsap.sequenceNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PLCCHsequenceNumber", HFILL }},
    { &hf_rnsap_UL_Timeslot_Information768_item,
      { "UL-Timeslot-InformationItem768", "rnsap.UL_Timeslot_InformationItem768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_Code_Information768,
      { "uL-Code-Information768", "rnsap.uL_Code_Information768",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_UL_Code_Information768", HFILL }},
    { &hf_rnsap_UL_TimeSlot_ISCP_Info_item,
      { "UL-TimeSlot-ISCP-InfoItem", "rnsap.UL_TimeSlot_ISCP_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_TimeslotISCP,
      { "uL-TimeslotISCP", "rnsap.uL_TimeslotISCP",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_TimeSlot_ISCP_LCR_Info_item,
      { "UL-TimeSlot-ISCP-LCR-InfoItem", "rnsap.UL_TimeSlot_ISCP_LCR_InfoItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_iSCP,
      { "iSCP", "rnsap.iSCP",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UL_Timeslot_ISCP_Value", HFILL }},
    { &hf_rnsap_sync_UL_codes_bitmap,
      { "sync-UL-codes-bitmap", "rnsap.sync_UL_codes_bitmap",
        FT_BYTES, BASE_NONE, NULL, 0,
        "BIT_STRING_SIZE_8", HFILL }},
    { &hf_rnsap_fPACH_info,
      { "fPACH-info", "rnsap.fPACH_info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "FPACH_Information", HFILL }},
    { &hf_rnsap_prxUpPCHdes,
      { "prxUpPCHdes", "rnsap.prxUpPCHdes",
        FT_INT32, BASE_DEC, NULL, 0,
        "INTEGER_M120_M58_", HFILL }},
    { &hf_rnsap_syncUL_procParameter,
      { "syncUL-procParameter", "rnsap.syncUL_procParameter_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "SYNC_UL_ProcParameters", HFILL }},
    { &hf_rnsap_mMax,
      { "mMax", "rnsap.mMax",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_1_32", HFILL }},
    { &hf_rnsap_ul_ScramblingCodeNumber,
      { "ul-ScramblingCodeNumber", "rnsap.ul_ScramblingCodeNumber",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_ScramblingCodeLength,
      { "ul-ScramblingCodeLength", "rnsap.ul_ScramblingCodeLength",
        FT_UINT32, BASE_DEC, VALS(rnsap_UL_ScramblingCodeLength_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_Synchronisation_StepSize,
      { "uL-Synchronisation-StepSize", "rnsap.uL_Synchronisation_StepSize",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_Synchronisation_Frequency,
      { "uL-Synchronisation-Frequency", "rnsap.uL_Synchronisation_Frequency",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uRA_ID,
      { "uRA-ID", "rnsap.uRA_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_multipleURAsIndicator,
      { "multipleURAsIndicator", "rnsap.multipleURAsIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_MultipleURAsIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_rNCsWithCellsInTheAccessedURA_List,
      { "rNCsWithCellsInTheAccessedURA-List", "rnsap.rNCsWithCellsInTheAccessedURA_List",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RNCsWithCellsInTheAccessedURA_List_item,
      { "RNCsWithCellsInTheAccessedURA-Item", "rnsap.RNCsWithCellsInTheAccessedURA_Item_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_USCH_Information_item,
      { "USCH-InformationItem", "rnsap.USCH_InformationItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uSCH_ID,
      { "uSCH-ID", "rnsap.uSCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rb_Info,
      { "rb-Info", "rnsap.rb_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UMTS_Cells_Info_List_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sDPCCH_PowerOffsetInformation,
      { "sDPCCH-PowerOffsetInformation", "rnsap.sDPCCH_PowerOffsetInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_CLTD_Activation_Information,
      { "uL-CLTD-Activation-Information", "rnsap.uL_CLTD_Activation_Information",
        FT_UINT32, BASE_DEC, VALS(rnsap_UL_CLTD_Activation_Information_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD,
      { "setup-Or-ConfigurationChange-Or-Removal-Of-UL-CLTD", "rnsap.setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD",
        FT_UINT32, BASE_DEC, VALS(rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_f_DPCH_Info,
      { "f-DPCH-Info", "rnsap.f_DPCH_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2,
      { "setup-Or-ConfigurationChange-Or-Removal-Of-UL-DPCCH2", "rnsap.setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2",
        FT_UINT32, BASE_DEC, VALS(rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_f_DPCH_Info_01,
      { "f-DPCH-Info", "rnsap.f_DPCH_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "F_DPCH_Info_To_Modify", HFILL }},
    { &hf_rnsap_e_roch_power_offset,
      { "e-roch-power-offset", "rnsap.e_roch_power_offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_ROCH_PowerOffset", HFILL }},
    { &hf_rnsap_s_e_dpcch_power_offset,
      { "s-e-dpcch-power-offset", "rnsap.s_e_dpcch_power_offset",
        FT_UINT32, BASE_DEC, NULL, 0,
        "S_E_DPCCH_PowerOffset", HFILL }},
    { &hf_rnsap_interstream_compensation,
      { "interstream-compensation", "rnsap.interstream_compensation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "InterStream_Interference_Compensation", HFILL }},
    { &hf_rnsap_minimum_E_TFCI_rank2,
      { "minimum-E-TFCI-rank2", "rnsap.minimum_E_TFCI_rank2",
        FT_UINT32, BASE_DEC, NULL, 0,
        "INTEGER_0_127", HFILL }},
    { &hf_rnsap_setup_08,
      { "setup", "rnsap.setup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_MIMO_Information", HFILL }},
    { &hf_rnsap_configurationChange_07,
      { "configurationChange", "rnsap.configurationChange_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_MIMO_Information_To_Modify", HFILL }},
    { &hf_rnsap_removal_06,
      { "removal", "rnsap.removal",
        FT_UINT32, BASE_DEC, VALS(rnsap_UL_MIMO_Removal_vals), 0,
        "UL_MIMO_Removal", HFILL }},
    { &hf_rnsap_e_roch_channelization_code,
      { "e-roch-channelization-code", "rnsap.e_roch_channelization_code",
        FT_UINT32, BASE_DEC, NULL, 0,
        "FDD_DL_ChannelisationCodeNumber", HFILL }},
    { &hf_rnsap_s_e_rnti,
      { "s-e-rnti", "rnsap.s_e_rnti",
        FT_UINT32, BASE_DEC, NULL, 0,
        "E_RNTI", HFILL }},
    { &hf_rnsap_s_signature_sequence,
      { "s-signature-sequence", "rnsap.s_signature_sequence",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ERGCH_SignatureSequence", HFILL }},
    { &hf_rnsap_s_e_roch_release_indicator,
      { "s-e-roch-release-indicator", "rnsap.s_e_roch_release_indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_S_E_ROCH_Release_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_protocolIEs,
      { "protocolIEs", "rnsap.protocolIEs",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ProtocolIE_Container", HFILL }},
    { &hf_rnsap_protocolExtensions,
      { "protocolExtensions", "rnsap.protocolExtensions",
        FT_UINT32, BASE_DEC, NULL, 0,
        "ProtocolExtensionContainer", HFILL }},
    { &hf_rnsap_minUL_ChannelisationCodeLength,
      { "minUL-ChannelisationCodeLength", "rnsap.minUL_ChannelisationCodeLength",
        FT_UINT32, BASE_DEC, VALS(rnsap_MinUL_ChannelisationCodeLength_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_maxNrOfUL_DPCHs,
      { "maxNrOfUL-DPCHs", "rnsap.maxNrOfUL_DPCHs",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_PunctureLimit,
      { "ul-PunctureLimit", "rnsap.ul_PunctureLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PunctureLimit", HFILL }},
    { &hf_rnsap_ul_TFCS,
      { "ul-TFCS", "rnsap.ul_TFCS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "TFCS", HFILL }},
    { &hf_rnsap_ul_DPCCH_SlotFormat,
      { "ul-DPCCH-SlotFormat", "rnsap.ul_DPCCH_SlotFormat",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_SIRTarget,
      { "ul-SIRTarget", "rnsap.ul_SIRTarget",
        FT_INT32, BASE_DEC, NULL, 0,
        "UL_SIR", HFILL }},
    { &hf_rnsap_diversityMode,
      { "diversityMode", "rnsap.diversityMode",
        FT_UINT32, BASE_DEC, VALS(rnsap_DiversityMode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_not_Used_sSDT_CellIdLength,
      { "not-Used-sSDT-CellIdLength", "rnsap.not_Used_sSDT_CellIdLength_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_not_Used_s_FieldLength,
      { "not-Used-s-FieldLength", "rnsap.not_Used_s_FieldLength_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tFCS,
      { "tFCS", "rnsap.tFCS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_DPCH_SlotFormat,
      { "dl-DPCH-SlotFormat", "rnsap.dl_DPCH_SlotFormat",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_nrOfDLchannelisationcodes,
      { "nrOfDLchannelisationcodes", "rnsap.nrOfDLchannelisationcodes",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tFCI_SignallingMode,
      { "tFCI-SignallingMode", "rnsap.tFCI_SignallingMode",
        FT_UINT32, BASE_DEC, VALS(rnsap_TFCI_SignallingMode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_multiplexingPosition,
      { "multiplexingPosition", "rnsap.multiplexingPosition",
        FT_UINT32, BASE_DEC, VALS(rnsap_MultiplexingPosition_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_powerOffsetInformation,
      { "powerOffsetInformation", "rnsap.powerOffsetInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PowerOffsetInformation_RL_SetupRqstFDD", HFILL }},
    { &hf_rnsap_fdd_dl_TPC_DownlinkStepSize,
      { "fdd-dl-TPC-DownlinkStepSize", "rnsap.fdd_dl_TPC_DownlinkStepSize",
        FT_UINT32, BASE_DEC, VALS(rnsap_FDD_TPC_DownlinkStepSize_vals), 0,
        "FDD_TPC_DownlinkStepSize", HFILL }},
    { &hf_rnsap_po1_ForTFCI_Bits,
      { "po1-ForTFCI-Bits", "rnsap.po1_ForTFCI_Bits",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PowerOffset", HFILL }},
    { &hf_rnsap_po2_ForTPC_Bits,
      { "po2-ForTPC-Bits", "rnsap.po2_ForTPC_Bits",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PowerOffset", HFILL }},
    { &hf_rnsap_po3_ForPilotBits,
      { "po3-ForPilotBits", "rnsap.po3_ForPilotBits",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PowerOffset", HFILL }},
    { &hf_rnsap_RL_InformationList_RL_SetupRqstFDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_chipOffset,
      { "chipOffset", "rnsap.chipOffset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_diversityControlField,
      { "diversityControlField", "rnsap.diversityControlField",
        FT_UINT32, BASE_DEC, VALS(rnsap_DiversityControlField_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_InitialTX_Power,
      { "dl-InitialTX-Power", "rnsap.dl_InitialTX_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_not_Used_sSDT_CellID,
      { "not-Used-sSDT-CellID", "rnsap.not_Used_sSDT_CellID_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_transmitDiversityIndicator,
      { "transmitDiversityIndicator", "rnsap.transmitDiversityIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransmitDiversityIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Information_RL_Setup_List_item,
      { "Additional-HS-Cell-Information-RL-Setup-ItemIEs", "rnsap.Additional_HS_Cell_Information_RL_Setup_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hSPDSCH_RL_ID,
      { "hSPDSCH-RL-ID", "rnsap.hSPDSCH_RL_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "RL_ID", HFILL }},
    { &hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information,
      { "hS-DSCH-FDD-Secondary-Serving-Information", "rnsap.hS_DSCH_FDD_Secondary_Serving_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_powerOffsetInformation_01,
      { "powerOffsetInformation", "rnsap.powerOffsetInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD", HFILL }},
    { &hf_rnsap_maxNrTimeslots_UL,
      { "maxNrTimeslots-UL", "rnsap.maxNrTimeslots_UL",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MaxNrTimeslots", HFILL }},
    { &hf_rnsap_minimumSpreadingFactor_UL,
      { "minimumSpreadingFactor-UL", "rnsap.minimumSpreadingFactor_UL",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MinimumSpreadingFactor", HFILL }},
    { &hf_rnsap_maxNrULPhysicalchannels,
      { "maxNrULPhysicalchannels", "rnsap.maxNrULPhysicalchannels",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_maxNrTimeslots_DL,
      { "maxNrTimeslots-DL", "rnsap.maxNrTimeslots_DL",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MaxNrTimeslots", HFILL }},
    { &hf_rnsap_minimumSpreadingFactor_DL,
      { "minimumSpreadingFactor-DL", "rnsap.minimumSpreadingFactor_DL",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MinimumSpreadingFactor", HFILL }},
    { &hf_rnsap_maxNrDLPhysicalchannels,
      { "maxNrDLPhysicalchannels", "rnsap.maxNrDLPhysicalchannels",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationList_RL_SetupRqstTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cCTrCH_ID,
      { "cCTrCH-ID", "rnsap.cCTrCH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationList_RL_SetupRqstTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_PunctureLimit,
      { "dl-PunctureLimit", "rnsap.dl_PunctureLimit",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PunctureLimit", HFILL }},
    { &hf_rnsap_tdd_TPC_DownlinkStepSize,
      { "tdd-TPC-DownlinkStepSize", "rnsap.tdd_TPC_DownlinkStepSize",
        FT_UINT32, BASE_DEC, VALS(rnsap_TDD_TPC_DownlinkStepSize_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_cCTrCH_TPCList,
      { "cCTrCH-TPCList", "rnsap.cCTrCH_TPCList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_TPCList_RL_SetupRqstTDD", HFILL }},
    { &hf_rnsap_CCTrCH_TPCList_RL_SetupRqstTDD_item,
      { "CCTrCH-TPCItem-RL-SetupRqstTDD", "rnsap.CCTrCH_TPCItem_RL_SetupRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_specialBurstScheduling,
      { "specialBurstScheduling", "rnsap.specialBurstScheduling",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dL_TimeSlot_ISCP,
      { "dL-TimeSlot-ISCP", "rnsap.dL_TimeSlot_ISCP",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_TimeSlot_ISCP_Info", HFILL }},
    { &hf_rnsap_RL_InformationResponseList_RL_SetupRspFDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uRA_Information,
      { "uRA-Information", "rnsap.uRA_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sAI,
      { "sAI", "rnsap.sAI_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gA_Cell,
      { "gA-Cell", "rnsap.gA_Cell",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gA_AccessPointPosition,
      { "gA-AccessPointPosition", "rnsap.gA_AccessPointPosition_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_not_Used_secondary_CCPCH_Info,
      { "not-Used-secondary-CCPCH-Info", "rnsap.not_Used_secondary_CCPCH_Info_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_diversityIndication,
      { "diversityIndication", "rnsap.diversityIndication",
        FT_UINT32, BASE_DEC, VALS(rnsap_DiversityIndication_RL_SetupRspFDD_vals), 0,
        "DiversityIndication_RL_SetupRspFDD", HFILL }},
    { &hf_rnsap_sSDT_SupportIndicator,
      { "sSDT-SupportIndicator", "rnsap.sSDT_SupportIndicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_SSDT_SupportIndicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_closedlooptimingadjustmentmode,
      { "closedlooptimingadjustmentmode", "rnsap.closedlooptimingadjustmentmode",
        FT_UINT32, BASE_DEC, VALS(rnsap_Closedlooptimingadjustmentmode_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_maximumDLTxPower,
      { "maximumDLTxPower", "rnsap.maximumDLTxPower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_minimumDLTxPower,
      { "minimumDLTxPower", "rnsap.minimumDLTxPower",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_not_Used_dSCHInformationResponse,
      { "not-Used-dSCHInformationResponse", "rnsap.not_Used_dSCHInformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_neighbouring_UMTS_CellInformation,
      { "neighbouring-UMTS-CellInformation", "rnsap.neighbouring_UMTS_CellInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_neighbouring_GSM_CellInformation,
      { "neighbouring-GSM-CellInformation", "rnsap.neighbouring_GSM_CellInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_sRB_Delay,
      { "sRB-Delay", "rnsap.sRB_Delay",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_combining,
      { "combining", "rnsap.combining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Combining_RL_SetupRspFDD", HFILL }},
    { &hf_rnsap_nonCombiningOrFirstRL,
      { "nonCombiningOrFirstRL", "rnsap.nonCombiningOrFirstRL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "NonCombiningOrFirstRL_RL_SetupRspFDD", HFILL }},
    { &hf_rnsap_dCH_InformationResponse,
      { "dCH-InformationResponse", "rnsap.dCH_InformationResponse",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Information_Response_List_item,
      { "Additional-HS-Cell-Information-Response-ItemIEs", "rnsap.Additional_HS_Cell_Information_Response_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_TimeSlot_ISCP_Info,
      { "ul-TimeSlot-ISCP-Info", "rnsap.ul_TimeSlot_ISCP_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_timingAdvanceApplied,
      { "timingAdvanceApplied", "rnsap.timingAdvanceApplied",
        FT_UINT32, BASE_DEC, VALS(rnsap_TimingAdvanceApplied_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_alphaValue,
      { "alphaValue", "rnsap.alphaValue",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_PhysCH_SF_Variation,
      { "ul-PhysCH-SF-Variation", "rnsap.ul_PhysCH_SF_Variation",
        FT_UINT32, BASE_DEC, VALS(rnsap_UL_PhysCH_SF_Variation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_synchronisationConfiguration,
      { "synchronisationConfiguration", "rnsap.synchronisationConfiguration_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_secondary_CCPCH_Info_TDD,
      { "secondary-CCPCH-Info-TDD", "rnsap.secondary_CCPCH_Info_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_CCTrCHInformation,
      { "ul-CCTrCHInformation", "rnsap.ul_CCTrCHInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_CCTrCHInformationList_RL_SetupRspTDD", HFILL }},
    { &hf_rnsap_dl_CCTrCHInformation,
      { "dl-CCTrCHInformation", "rnsap.dl_CCTrCHInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_CCTrCHInformationList_RL_SetupRspTDD", HFILL }},
    { &hf_rnsap_dCH_InformationResponse_01,
      { "dCH-InformationResponse", "rnsap.dCH_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DCH_InformationResponseList_RL_SetupRspTDD", HFILL }},
    { &hf_rnsap_dsch_InformationResponse,
      { "dsch-InformationResponse", "rnsap.dsch_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DSCH_InformationResponse_RL_SetupRspTDD", HFILL }},
    { &hf_rnsap_usch_InformationResponse,
      { "usch-InformationResponse", "rnsap.usch_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "USCH_InformationResponse_RL_SetupRspTDD", HFILL }},
    { &hf_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD_item,
      { "UL-CCTrCHInformationItem-RL-SetupRspTDD", "rnsap.UL_CCTrCHInformationItem_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_DPCH_Information,
      { "ul-DPCH-Information", "rnsap.ul_DPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_InformationList_RL_SetupRspTDD", HFILL }},
    { &hf_rnsap_tDD_DPCHOffset,
      { "tDD-DPCHOffset", "rnsap.tDD_DPCHOffset",
        FT_UINT32, BASE_DEC, VALS(rnsap_TDD_DPCHOffset_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_Timeslot_Information,
      { "uL-Timeslot-Information", "rnsap.uL_Timeslot_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD_item,
      { "DL-CCTrCHInformationItem-RL-SetupRspTDD", "rnsap.DL_CCTrCHInformationItem_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_DPCH_Information,
      { "dl-DPCH-Information", "rnsap.dl_DPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationList_RL_SetupRspTDD", HFILL }},
    { &hf_rnsap_dL_Timeslot_Information,
      { "dL-Timeslot-Information", "rnsap.dL_Timeslot_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_InformationListIEs_RL_SetupRspTDD_item,
      { "DSCHInformationItem-RL-SetupRspTDD", "rnsap.DSCHInformationItem_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dsch_ID,
      { "dsch-ID", "rnsap.dsch_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dSCH_FlowControlInformation,
      { "dSCH-FlowControlInformation", "rnsap.dSCH_FlowControlInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_transportFormatManagement,
      { "transportFormatManagement", "rnsap.transportFormatManagement",
        FT_UINT32, BASE_DEC, VALS(rnsap_TransportFormatManagement_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_USCH_InformationListIEs_RL_SetupRspTDD_item,
      { "USCHInformationItem-RL-SetupRspTDD", "rnsap.USCHInformationItem_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_usch_ID,
      { "usch-ID", "rnsap.usch_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_TimeSlot_ISCP_LCR_Info,
      { "ul-TimeSlot-ISCP-LCR-Info", "rnsap.ul_TimeSlot_ISCP_LCR_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_secondary_LCR_CCPCH_Info_TDD,
      { "secondary-LCR-CCPCH-Info-TDD", "rnsap.secondary_LCR_CCPCH_Info_TDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_LCR_CCTrCHInformation,
      { "ul-LCR-CCTrCHInformation", "rnsap.ul_LCR_CCTrCHInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_LCR_CCTrCHInformationList_RL_SetupRspTDD", HFILL }},
    { &hf_rnsap_dl_LCR_CCTrCHInformation,
      { "dl-LCR-CCTrCHInformation", "rnsap.dl_LCR_CCTrCHInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_LCR_CCTrCHInformationList_RL_SetupRspTDD", HFILL }},
    { &hf_rnsap_dsch_LCR_InformationResponse,
      { "dsch-LCR-InformationResponse", "rnsap.dsch_LCR_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DSCH_LCR_InformationResponse_RL_SetupRspTDD", HFILL }},
    { &hf_rnsap_usch_LCR_InformationResponse,
      { "usch-LCR-InformationResponse", "rnsap.usch_LCR_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "USCH_LCR_InformationResponse_RL_SetupRspTDD", HFILL }},
    { &hf_rnsap_UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD_item,
      { "UL-LCR-CCTrCHInformationItem-RL-SetupRspTDD", "rnsap.UL_LCR_CCTrCHInformationItem_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_DPCH_LCR_Information,
      { "ul-DPCH-LCR-Information", "rnsap.ul_DPCH_LCR_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_LCR_InformationList_RL_SetupRspTDD", HFILL }},
    { &hf_rnsap_uL_TimeslotLCR_Information,
      { "uL-TimeslotLCR-Information", "rnsap.uL_TimeslotLCR_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD_item,
      { "DL-CCTrCH-LCR-InformationItem-RL-SetupRspTDD", "rnsap.DL_CCTrCH_LCR_InformationItem_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_DPCH_LCR_Information,
      { "dl-DPCH-LCR-Information", "rnsap.dl_DPCH_LCR_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_LCR_InformationList_RL_SetupRspTDD", HFILL }},
    { &hf_rnsap_dL_Timeslot_LCR_Information,
      { "dL-Timeslot-LCR-Information", "rnsap.dL_Timeslot_LCR_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_TimeslotLCR_Information", HFILL }},
    { &hf_rnsap_tSTD_Indicator,
      { "tSTD-Indicator", "rnsap.tSTD_Indicator",
        FT_UINT32, BASE_DEC, VALS(rnsap_TSTD_Indicator_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_LCR_InformationListIEs_RL_SetupRspTDD_item,
      { "DSCH-LCR-InformationItem-RL-SetupRspTDD", "rnsap.DSCH_LCR_InformationItem_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_USCH_LCR_InformationListIEs_RL_SetupRspTDD_item,
      { "USCH-LCR-InformationItem-RL-SetupRspTDD", "rnsap.USCH_LCR_InformationItem_RL_SetupRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_secondary_CCPCH_Info_TDD768,
      { "secondary-CCPCH-Info-TDD768", "rnsap.secondary_CCPCH_Info_TDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_CCTrCHInformation768,
      { "ul-CCTrCHInformation768", "rnsap.ul_CCTrCHInformation768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_CCTrCHInformationList_RL_SetupRspTDD768", HFILL }},
    { &hf_rnsap_dl_CCTrCHInformation768,
      { "dl-CCTrCHInformation768", "rnsap.dl_CCTrCHInformation768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_CCTrCHInformationList_RL_SetupRspTDD768", HFILL }},
    { &hf_rnsap_gA_CellAdditionalShapes,
      { "gA-CellAdditionalShapes", "rnsap.gA_CellAdditionalShapes",
        FT_UINT32, BASE_DEC, VALS(rnsap_GA_CellAdditionalShapes_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_hCS_Prio,
      { "hCS-Prio", "rnsap.hCS_Prio",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_timeSlot_RL_SetupRspTDD,
      { "timeSlot-RL-SetupRspTDD", "rnsap.timeSlot_RL_SetupRspTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TimeSlot", HFILL }},
    { &hf_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD768_item,
      { "UL-CCTrCHInformationItem-RL-SetupRspTDD768", "rnsap.UL_CCTrCHInformationItem_RL_SetupRspTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_DPCH_Information768,
      { "ul-DPCH-Information768", "rnsap.ul_DPCH_Information768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_InformationList_RL_SetupRspTDD768", HFILL }},
    { &hf_rnsap_uL_SIR_Target_CCTrCH_InformationItem_RL_SetupRspTDD768,
      { "uL-SIR-Target-CCTrCH-InformationItem-RL-SetupRspTDD768", "rnsap.uL_SIR_Target_CCTrCH_InformationItem_RL_SetupRspTDD768",
        FT_INT32, BASE_DEC, NULL, 0,
        "UL_SIR", HFILL }},
    { &hf_rnsap_uL_Timeslot_Information768,
      { "uL-Timeslot-Information768", "rnsap.uL_Timeslot_Information768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD768_item,
      { "DL-CCTrCHInformationItem-RL-SetupRspTDD768", "rnsap.DL_CCTrCHInformationItem_RL_SetupRspTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_DPCH_Information768,
      { "dl-DPCH-Information768", "rnsap.dl_DPCH_Information768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationList_RL_SetupRspTDD768", HFILL }},
    { &hf_rnsap_cCTrCH_Maximum_DL_Power,
      { "cCTrCH-Maximum-DL-Power", "rnsap.cCTrCH_Maximum_DL_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_cCTrCH_Minimum_DL_Power,
      { "cCTrCH-Minimum-DL-Power", "rnsap.cCTrCH_Minimum_DL_Power",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_dL_Timeslot_Information768,
      { "dL-Timeslot-Information768", "rnsap.dL_Timeslot_Information768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_generalCause,
      { "generalCause", "rnsap.generalCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeneralCauseList_RL_SetupFailureFDD", HFILL }},
    { &hf_rnsap_rLSpecificCause,
      { "rLSpecificCause", "rnsap.rLSpecificCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RLSpecificCauseList_RL_SetupFailureFDD", HFILL }},
    { &hf_rnsap_unsuccessful_RL_InformationRespList_RL_SetupFailureFDD,
      { "unsuccessful-RL-InformationRespList-RL-SetupFailureFDD", "rnsap.unsuccessful_RL_InformationRespList_RL_SetupFailureFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UnsuccessfulRL_InformationResponseList_RL_SetupFailureFDD", HFILL }},
    { &hf_rnsap_successful_RL_InformationRespList_RL_SetupFailureFDD,
      { "successful-RL-InformationRespList-RL-SetupFailureFDD", "rnsap.successful_RL_InformationRespList_RL_SetupFailureFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SuccessfulRL_InformationResponseList_RL_SetupFailureFDD", HFILL }},
    { &hf_rnsap_UnsuccessfulRL_InformationResponseList_RL_SetupFailureFDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SuccessfulRL_InformationResponseList_RL_SetupFailureFDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_diversityIndication_01,
      { "diversityIndication", "rnsap.diversityIndication",
        FT_UINT32, BASE_DEC, VALS(rnsap_DiversityIndication_RL_SetupFailureFDD_vals), 0,
        "DiversityIndication_RL_SetupFailureFDD", HFILL }},
    { &hf_rnsap_not_Used_dSCH_InformationResponse_RL_SetupFailureFDD,
      { "not-Used-dSCH-InformationResponse-RL-SetupFailureFDD", "rnsap.not_Used_dSCH_InformationResponse_RL_SetupFailureFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_combining_01,
      { "combining", "rnsap.combining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Combining_RL_SetupFailureFDD", HFILL }},
    { &hf_rnsap_nonCombiningOrFirstRL_01,
      { "nonCombiningOrFirstRL", "rnsap.nonCombiningOrFirstRL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "NonCombiningOrFirstRL_RL_SetupFailureFDD", HFILL }},
    { &hf_rnsap_generalCause_01,
      { "generalCause", "rnsap.generalCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeneralCauseList_RL_SetupFailureTDD", HFILL }},
    { &hf_rnsap_rLSpecificCause_01,
      { "rLSpecificCause", "rnsap.rLSpecificCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RLSpecificCauseList_RL_SetupFailureTDD", HFILL }},
    { &hf_rnsap_unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD,
      { "unsuccessful-RL-InformationRespItem-RL-SetupFailureTDD", "rnsap.unsuccessful_RL_InformationRespItem_RL_SetupFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_RL_AdditionRqstFDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Information_RL_Addition_List_item,
      { "Additional-HS-Cell-Information-RL-Addition-ItemIEs", "rnsap.Additional_HS_Cell_Information_RL_Addition_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency,
      { "setup-Or-Addition-Of-EDCH-On-secondary-UL-Frequency", "rnsap.setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency",
        FT_UINT32, BASE_DEC, VALS(rnsap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_addition,
      { "addition", "rnsap.addition",
        FT_UINT32, BASE_DEC, NULL, 0,
        "Additional_EDCH_Cell_Information_To_Add_List", HFILL }},
    { &hf_rnsap_dL_TimeSlot_ISCP_Info,
      { "dL-TimeSlot-ISCP-Info", "rnsap.dL_TimeSlot_ISCP_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uplinkStepSizeLCR,
      { "uplinkStepSizeLCR", "rnsap.uplinkStepSizeLCR",
        FT_UINT32, BASE_DEC, VALS(rnsap_TDD_TPC_UplinkStepSize_LCR_vals), 0,
        "TDD_TPC_UplinkStepSize_LCR", HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_downlinkStepSize,
      { "downlinkStepSize", "rnsap.downlinkStepSize",
        FT_UINT32, BASE_DEC, VALS(rnsap_TDD_TPC_DownlinkStepSize_vals), 0,
        "TDD_TPC_DownlinkStepSize", HFILL }},
    { &hf_rnsap_RL_InformationResponseList_RL_AdditionRspFDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_CodeInformation_01,
      { "dl-CodeInformation", "rnsap.dl_CodeInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_CodeInformationList_RL_AdditionRspFDD", HFILL }},
    { &hf_rnsap_diversityIndication_02,
      { "diversityIndication", "rnsap.diversityIndication",
        FT_UINT32, BASE_DEC, VALS(rnsap_DiversityIndication_RL_AdditionRspFDD_vals), 0,
        "DiversityIndication_RL_AdditionRspFDD", HFILL }},
    { &hf_rnsap_combining_02,
      { "combining", "rnsap.combining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Combining_RL_AdditionRspFDD", HFILL }},
    { &hf_rnsap_nonCombining,
      { "nonCombining", "rnsap.nonCombining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "NonCombining_RL_AdditionRspFDD", HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Change_Information_Response_List_item,
      { "Additional-HS-Cell-Change-Information-Response-ItemIEs", "rnsap.Additional_HS_Cell_Change_Information_Response_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_DSCH_Secondary_Serving_Cell_Change_Information_Response,
      { "hS-DSCH-Secondary-Serving-Cell-Change-Information-Response", "rnsap.hS_DSCH_Secondary_Serving_Cell_Change_Information_Response_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_CCTrCHInformation_01,
      { "ul-CCTrCHInformation", "rnsap.ul_CCTrCHInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_CCTrCHInformationList_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_dl_CCTrCHInformation_01,
      { "dl-CCTrCHInformation", "rnsap.dl_CCTrCHInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_CCTrCHInformationList_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_dCH_Information,
      { "dCH-Information", "rnsap.dCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DCH_Information_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_dSCH_InformationResponse,
      { "dSCH-InformationResponse", "rnsap.dSCH_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DSCH_InformationResponse_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_uSCH_InformationResponse,
      { "uSCH-InformationResponse", "rnsap.uSCH_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "USCH_InformationResponse_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD_item,
      { "UL-CCTrCHInformationItem-RL-AdditionRspTDD", "rnsap.UL_CCTrCHInformationItem_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_DPCH_Information_01,
      { "ul-DPCH-Information", "rnsap.ul_DPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_InformationList_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD_item,
      { "DL-CCTrCHInformationItem-RL-AdditionRspTDD", "rnsap.DL_CCTrCHInformationItem_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_DPCH_Information_01,
      { "dl-DPCH-Information", "rnsap.dl_DPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationList_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_diversityIndication_03,
      { "diversityIndication", "rnsap.diversityIndication",
        FT_UINT32, BASE_DEC, VALS(rnsap_DiversityIndication_RL_AdditionRspTDD_vals), 0,
        "DiversityIndication_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_combining_03,
      { "combining", "rnsap.combining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Combining_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_nonCombining_01,
      { "nonCombining", "rnsap.nonCombining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "NonCombining_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_DSCH_InformationListIE_RL_AdditionRspTDD_item,
      { "DSCHInformationItem-RL-AdditionRspTDD", "rnsap.DSCHInformationItem_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_diversityIndication_04,
      { "diversityIndication", "rnsap.diversityIndication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DiversityIndication_RL_AdditionRspTDD2", HFILL }},
    { &hf_rnsap_USCH_InformationListIE_RL_AdditionRspTDD_item,
      { "USCHInformationItem-RL-AdditionRspTDD", "rnsap.USCHInformationItem_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_CCTrCH_LCR_Information,
      { "ul-CCTrCH-LCR-Information", "rnsap.ul_CCTrCH_LCR_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_CCTrCH_LCR_InformationList_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_dl_CCTrCH_LCR_Information,
      { "dl-CCTrCH-LCR-Information", "rnsap.dl_CCTrCH_LCR_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_CCTrCH_LCR_InformationList_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_dCH_InformationResponse_02,
      { "dCH-InformationResponse", "rnsap.dCH_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DCH_InformationResponseList_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_dsch_LCR_InformationResponse_01,
      { "dsch-LCR-InformationResponse", "rnsap.dsch_LCR_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DSCH_LCR_InformationResponse_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_usch_LCR_InformationResponse_01,
      { "usch-LCR-InformationResponse", "rnsap.usch_LCR_InformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "USCH_LCR_InformationResponse_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_item,
      { "UL-CCTrCH-LCR-InformationItem-RL-AdditionRspTDD", "rnsap.UL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_DPCH_LCR_Information_01,
      { "ul-DPCH-LCR-Information", "rnsap.ul_DPCH_LCR_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_LCR_InformationList_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_item,
      { "DL-CCTrCH-LCR-InformationItem-RL-AdditionRspTDD", "rnsap.DL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_DPCH_LCR_Information_01,
      { "dl-DPCH-LCR-Information", "rnsap.dl_DPCH_LCR_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_LCR_InformationList_RL_AdditionRspTDD", HFILL }},
    { &hf_rnsap_dL_TimeslotLCR_Information,
      { "dL-TimeslotLCR-Information", "rnsap.dL_TimeslotLCR_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD_item,
      { "DSCH-LCR-InformationItem-RL-AdditionRspTDD", "rnsap.DSCH_LCR_InformationItem_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_USCH_LCR_InformationListIEs_RL_AdditionRspTDD_item,
      { "USCH-LCR-InformationItem-RL-AdditionRspTDD", "rnsap.USCH_LCR_InformationItem_RL_AdditionRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_CCTrCHInformation768_01,
      { "ul-CCTrCHInformation768", "rnsap.ul_CCTrCHInformation768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_CCTrCHInformationList_RL_AdditionRspTDD768", HFILL }},
    { &hf_rnsap_dl_CCTrCHInformation768_01,
      { "dl-CCTrCHInformation768", "rnsap.dl_CCTrCHInformation768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_CCTrCHInformationList_RL_AdditionRspTDD768", HFILL }},
    { &hf_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD768_item,
      { "UL-CCTrCHInformationItem-RL-AdditionRspTDD768", "rnsap.UL_CCTrCHInformationItem_RL_AdditionRspTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_DPCH_Information768_01,
      { "ul-DPCH-Information768", "rnsap.ul_DPCH_Information768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_InformationList_RL_AdditionRspTDD768", HFILL }},
    { &hf_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD768_item,
      { "DL-CCTrCHInformationItem-RL-AdditionRspTDD768", "rnsap.DL_CCTrCHInformationItem_RL_AdditionRspTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_DPCH_Information768_01,
      { "dl-DPCH-Information768", "rnsap.dl_DPCH_Information768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationList_RL_AdditionRspTDD768", HFILL }},
    { &hf_rnsap_generalCause_02,
      { "generalCause", "rnsap.generalCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeneralCauseList_RL_AdditionFailureFDD", HFILL }},
    { &hf_rnsap_rLSpecificCause_02,
      { "rLSpecificCause", "rnsap.rLSpecificCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RLSpecificCauseList_RL_AdditionFailureFDD", HFILL }},
    { &hf_rnsap_unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD,
      { "unsuccessful-RL-InformationRespList-RL-AdditionFailureFDD", "rnsap.unsuccessful_RL_InformationRespList_RL_AdditionFailureFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UnsuccessfulRL_InformationResponseList_RL_AdditionFailureFDD", HFILL }},
    { &hf_rnsap_successful_RL_InformationRespList_RL_AdditionFailureFDD,
      { "successful-RL-InformationRespList-RL-AdditionFailureFDD", "rnsap.successful_RL_InformationRespList_RL_AdditionFailureFDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "SuccessfulRL_InformationResponseList_RL_AdditionFailureFDD", HFILL }},
    { &hf_rnsap_UnsuccessfulRL_InformationResponseList_RL_AdditionFailureFDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_SuccessfulRL_InformationResponseList_RL_AdditionFailureFDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_CodeInformation_02,
      { "dl-CodeInformation", "rnsap.dl_CodeInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_CodeInformationList_RL_AdditionFailureFDD", HFILL }},
    { &hf_rnsap_diversityIndication_05,
      { "diversityIndication", "rnsap.diversityIndication",
        FT_UINT32, BASE_DEC, VALS(rnsap_DiversityIndication_RL_AdditionFailureFDD_vals), 0,
        "DiversityIndication_RL_AdditionFailureFDD", HFILL }},
    { &hf_rnsap_combining_04,
      { "combining", "rnsap.combining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Combining_RL_AdditionFailureFDD", HFILL }},
    { &hf_rnsap_nonCombining_02,
      { "nonCombining", "rnsap.nonCombining_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "NonCombining_RL_AdditionFailureFDD", HFILL }},
    { &hf_rnsap_generalCause_03,
      { "generalCause", "rnsap.generalCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeneralCauseList_RL_AdditionFailureTDD", HFILL }},
    { &hf_rnsap_rLSpecificCause_03,
      { "rLSpecificCause", "rnsap.rLSpecificCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RLSpecificCauseList_RL_AdditionFailureTDD", HFILL }},
    { &hf_rnsap_unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD,
      { "unsuccessful-RL-InformationRespItem-RL-AdditionFailureTDD", "rnsap.unsuccessful_RL_InformationRespItem_RL_AdditionFailureTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_RL_DeletionRqst_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_maxNrOfUL_DPDCHs,
      { "maxNrOfUL-DPDCHs", "rnsap.maxNrOfUL_DPDCHs",
        FT_UINT32, BASE_DEC, NULL, 0,
        "MaxNrOfUL_DPCHs", HFILL }},
    { &hf_rnsap_not_Used_sSDT_CellIDLength,
      { "not-Used-sSDT-CellIDLength", "rnsap.not_Used_sSDT_CellIDLength_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_powerOffsetInformation_02,
      { "powerOffsetInformation", "rnsap.powerOffsetInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PowerOffsetInformation_RL_ReconfPrepFDD", HFILL }},
    { &hf_rnsap_pO1_ForTFCI_Bits,
      { "pO1-ForTFCI-Bits", "rnsap.pO1_ForTFCI_Bits",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PowerOffset", HFILL }},
    { &hf_rnsap_pO2_ForTPC_Bits,
      { "pO2-ForTPC-Bits", "rnsap.pO2_ForTPC_Bits",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PowerOffset", HFILL }},
    { &hf_rnsap_pO3_ForPilotBits,
      { "pO3-ForPilotBits", "rnsap.pO3_ForPilotBits",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PowerOffset", HFILL }},
    { &hf_rnsap_DCH_DeleteList_RL_ReconfPrepFDD_item,
      { "DCH-DeleteItem-RL-ReconfPrepFDD", "rnsap.DCH_DeleteItem_RL_ReconfPrepFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_RL_ReconfPrepFDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_not_Used_sSDT_Indication,
      { "not-Used-sSDT-Indication", "rnsap.not_Used_sSDT_Indication_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_not_Used_sSDT_CellIdentity,
      { "not-Used-sSDT-CellIdentity", "rnsap.not_Used_sSDT_CellIdentity_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep_item,
      { "Additional-HS-Cell-Information-RL-Reconf-Prep-ItemIEs", "rnsap.Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_DSCH_Secondary_Serving_Information_To_Modify,
      { "hS-DSCH-Secondary-Serving-Information-To-Modify", "rnsap.hS_DSCH_Secondary_Serving_Information_To_Modify_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_HS_DSCH_Secondary_Serving_Remove,
      { "hS-HS-DSCH-Secondary-Serving-Remove", "rnsap.hS_HS_DSCH_Secondary_Serving_Remove_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "HS_DSCH_Secondary_Serving_Remove", HFILL }},
    { &hf_rnsap_setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency,
      { "setup-Or-ConfigurationChange-Or-Removal-Of-EDCH-On-secondary-UL-Frequency", "rnsap.setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency",
        FT_UINT32, BASE_DEC, VALS(rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_powerOffsetInformation_03,
      { "powerOffsetInformation", "rnsap.powerOffsetInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD", HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cCTrCH_TPCList_01,
      { "cCTrCH-TPCList", "rnsap.cCTrCH_TPCList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_TPCAddList_RL_ReconfPrepTDD", HFILL }},
    { &hf_rnsap_CCTrCH_TPCAddList_RL_ReconfPrepTDD_item,
      { "CCTrCH-TPCAddItem-RL-ReconfPrepTDD", "rnsap.CCTrCH_TPCAddItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cCTrCH_TPCList_02,
      { "cCTrCH-TPCList", "rnsap.cCTrCH_TPCList",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_TPCModifyList_RL_ReconfPrepTDD", HFILL }},
    { &hf_rnsap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD_item,
      { "CCTrCH-TPCModifyItem-RL-ReconfPrepTDD", "rnsap.CCTrCH_TPCModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_DeleteList_RL_ReconfPrepTDD_item,
      { "DCH-DeleteItem-RL-ReconfPrepTDD", "rnsap.DCH_DeleteItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_ModifyList_RL_ReconfPrepTDD_item,
      { "DSCH-ModifyItem-RL-ReconfPrepTDD", "rnsap.DSCH_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCH_DeleteList_RL_ReconfPrepTDD_item,
      { "DSCH-DeleteItem-RL-ReconfPrepTDD", "rnsap.DSCH_DeleteItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_USCH_ModifyList_RL_ReconfPrepTDD_item,
      { "USCH-ModifyItem-RL-ReconfPrepTDD", "rnsap.USCH_ModifyItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_ccTrCHID,
      { "ul-ccTrCHID", "rnsap.ul_ccTrCHID",
        FT_UINT32, BASE_DEC, NULL, 0,
        "CCTrCH_ID", HFILL }},
    { &hf_rnsap_USCH_DeleteList_RL_ReconfPrepTDD_item,
      { "USCH-DeleteItem-RL-ReconfPrepTDD", "rnsap.USCH_DeleteItem_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Information_RL_ReconfPrepTDD_item,
      { "RL-InformationIE-RL-ReconfPrepTDD", "rnsap.RL_InformationIE_RL_ReconfPrepTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_Specific_DCH_Info,
      { "rL-Specific-DCH-Info", "rnsap.rL_Specific_DCH_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponseList_RL_ReconfReadyFDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_max_UL_SIR,
      { "max-UL-SIR", "rnsap.max_UL_SIR",
        FT_INT32, BASE_DEC, NULL, 0,
        "UL_SIR", HFILL }},
    { &hf_rnsap_min_UL_SIR,
      { "min-UL-SIR", "rnsap.min_UL_SIR",
        FT_INT32, BASE_DEC, NULL, 0,
        "UL_SIR", HFILL }},
    { &hf_rnsap_dl_CodeInformationList,
      { "dl-CodeInformationList", "rnsap.dl_CodeInformationList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_CodeInformationList_RL_ReconfReadyFDD", HFILL }},
    { &hf_rnsap_dCHInformationResponse,
      { "dCHInformationResponse", "rnsap.dCHInformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DCH_InformationResponseList_RL_ReconfReadyFDD", HFILL }},
    { &hf_rnsap_not_Used_dSCHsToBeAddedOrModified,
      { "not-Used-dSCHsToBeAddedOrModified", "rnsap.not_Used_dSCHsToBeAddedOrModified_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_RL_Reconf_Response_item,
      { "Additional-HS-Cell-RL-Reconf-Response-ItemIEs", "rnsap.Additional_HS_Cell_RL_Reconf_Response_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_CCTrCH_Information,
      { "ul-CCTrCH-Information", "rnsap.ul_CCTrCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_CCTrCH_InformationList_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_dl_CCTrCH_Information,
      { "dl-CCTrCH-Information", "rnsap.dl_CCTrCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_CCTrCH_InformationList_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_dCHInformationResponse_01,
      { "dCHInformationResponse", "rnsap.dCHInformationResponse_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DCH_InformationResponseList_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_dSCHsToBeAddedOrModified,
      { "dSCHsToBeAddedOrModified", "rnsap.dSCHsToBeAddedOrModified_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DSCHToBeAddedOrModified_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_uSCHsToBeAddedOrModified,
      { "uSCHsToBeAddedOrModified", "rnsap.uSCHsToBeAddedOrModified_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "USCHToBeAddedOrModified_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_UL_CCTrCHInformationListIE_RL_ReconfReadyTDD_item,
      { "UL-CCTrCH-InformationItem-RL-ReconfReadyTDD", "rnsap.UL_CCTrCH_InformationItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_DPCH_AddInformation,
      { "ul-DPCH-AddInformation", "rnsap.ul_DPCH_AddInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_InformationAddList_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_ul_DPCH_ModifyInformation,
      { "ul-DPCH-ModifyInformation", "rnsap.ul_DPCH_ModifyInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_InformationModifyList_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_ul_DPCH_DeleteInformation,
      { "ul-DPCH-DeleteInformation", "rnsap.ul_DPCH_DeleteInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_InformationDeleteList_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_uL_TimeslotLCR_Info,
      { "uL-TimeslotLCR-Info", "rnsap.uL_TimeslotLCR_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        "UL_TimeslotLCR_Information", HFILL }},
    { &hf_rnsap_rxTimingDeviationForTA,
      { "rxTimingDeviationForTA", "rnsap.rxTimingDeviationForTA",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rxTimingDeviationForTA768,
      { "rxTimingDeviationForTA768", "rnsap.rxTimingDeviationForTA768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_Timeslot_InformationModifyList_RL_ReconfReadyTDD,
      { "uL-Timeslot-InformationModifyList-RL-ReconfReadyTDD", "rnsap.uL_Timeslot_InformationModifyList_RL_ReconfReadyTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_item,
      { "UL-TimeslotLCR-InformationModifyItem-RL-ReconfReadyTDD", "rnsap.UL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tDD_uL_Code_LCR_Information,
      { "tDD-uL-Code-LCR-Information", "rnsap.tDD_uL_Code_LCR_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_UL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_TDD_UL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD_item,
      { "TDD-UL-Code-LCR-InformationModifyItem-RL-ReconfReadyTDD", "rnsap.TDD_UL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD_item,
      { "UL-Timeslot-InformationModifyItem-RL-ReconfReadyTDD", "rnsap.UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_Code_Information_01,
      { "uL-Code-Information", "rnsap.uL_Code_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD_item,
      { "TDD-UL-Code-InformationModifyItem-RL-ReconfReadyTDD", "rnsap.TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_item,
      { "UL-Timeslot-InformationModifyItem-RL-ReconfReadyTDD768", "rnsap.UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_Code_Information768_01,
      { "uL-Code-Information768", "rnsap.uL_Code_Information768",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD768", HFILL }},
    { &hf_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD768_item,
      { "TDD-UL-Code-InformationModifyItem-RL-ReconfReadyTDD768", "rnsap.TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_item,
      { "UL-DPCH-InformationDeleteItem-RL-ReconfReadyTDD", "rnsap.UL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCHInformationListIE_RL_ReconfReadyTDD_item,
      { "DL-CCTrCH-InformationItem-RL-ReconfReadyTDD", "rnsap.DL_CCTrCH_InformationItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_DPCH_AddInformation,
      { "dl-DPCH-AddInformation", "rnsap.dl_DPCH_AddInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationAddList_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_dl_DPCH_ModifyInformation,
      { "dl-DPCH-ModifyInformation", "rnsap.dl_DPCH_ModifyInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationModifyList_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_dl_DPCH_DeleteInformation,
      { "dl-DPCH-DeleteInformation", "rnsap.dl_DPCH_DeleteInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationDeleteList_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_dL_TimeslotLCR_Info,
      { "dL-TimeslotLCR-Info", "rnsap.dL_TimeslotLCR_Info",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_TimeslotLCR_Information", HFILL }},
    { &hf_rnsap_dL_Timeslot_InformationModifyList_RL_ReconfReadyTDD,
      { "dL-Timeslot-InformationModifyList-RL-ReconfReadyTDD", "rnsap.dL_Timeslot_InformationModifyList_RL_ReconfReadyTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_item,
      { "DL-TimeslotLCR-InformationModifyItem-RL-ReconfReadyTDD", "rnsap.DL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_tDD_dL_Code_LCR_Information,
      { "tDD-dL-Code-LCR-Information", "rnsap.tDD_dL_Code_LCR_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_DL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_TDD_DL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD_item,
      { "TDD-DL-Code-LCR-InformationModifyItem-RL-ReconfReadyTDD", "rnsap.TDD_DL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD_item,
      { "DL-Timeslot-InformationModifyItem-RL-ReconfReadyTDD", "rnsap.DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dL_Code_Information_01,
      { "dL-Code-Information", "rnsap.dL_Code_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD", HFILL }},
    { &hf_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD_item,
      { "TDD-DL-Code-InformationModifyItem-RL-ReconfReadyTDD", "rnsap.TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_item,
      { "DL-Timeslot-InformationModifyItem-RL-ReconfReadyTDD768", "rnsap.DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dL_Code_Information768_01,
      { "dL-Code-Information768", "rnsap.dL_Code_Information768",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD768", HFILL }},
    { &hf_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD768_item,
      { "TDD-DL-Code-InformationModifyItem-RL-ReconfReadyTDD768", "rnsap.TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dPCH_ID768,
      { "dPCH-ID768", "rnsap.dPCH_ID768",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_item,
      { "DL-DPCH-InformationDeleteItem-RL-ReconfReadyTDD", "rnsap.DL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD_item,
      { "DL-DPCH-InformationDeleteItem768-RL-ReconfReadyTDD", "rnsap.DL_DPCH_InformationDeleteItem768_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_item,
      { "DSCHToBeAddedOrModifiedItem-RL-ReconfReadyTDD", "rnsap.DSCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_item,
      { "USCHToBeAddedOrModifiedItem-RL-ReconfReadyTDD", "rnsap.USCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_RL_InformationResponse_RL_ReconfReadyTDD_item,
      { "RL-InformationResponse-RL-ReconfReadyTDD", "rnsap.RL_InformationResponse_RL_ReconfReadyTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_generalCause_04,
      { "generalCause", "rnsap.generalCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GeneralCauseList_RL_ReconfFailure", HFILL }},
    { &hf_rnsap_rLSpecificCause_04,
      { "rLSpecificCause", "rnsap.rLSpecificCause_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RLSpecificCauseList_RL_ReconfFailure", HFILL }},
    { &hf_rnsap_rL_ReconfigurationFailureList_RL_ReconfFailure,
      { "rL-ReconfigurationFailureList-RL-ReconfFailure", "rnsap.rL_ReconfigurationFailureList_RL_ReconfFailure",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_ReconfigurationFailureList_RL_ReconfFailure_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_DeleteList_RL_ReconfRqstFDD_item,
      { "DCH-DeleteItem-RL-ReconfRqstFDD", "rnsap.DCH_DeleteItem_RL_ReconfRqstFDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req_item,
      { "Additional-HS-Cell-Information-RL-Reconf-Req-ItemIEs", "rnsap.Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised,
      { "hS-DSCH-FDD-Secondary-Serving-Information-To-Modify-Unsynchronised", "rnsap.hS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_DSCH_Secondary_Serving_Remove,
      { "hS-DSCH-Secondary-Serving-Remove", "rnsap.hS_DSCH_Secondary_Serving_Remove_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_ReconfigurationRequestFDD_RL_InformationList_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DCH_DeleteList_RL_ReconfRqstTDD_item,
      { "DCH-DeleteItem-RL-ReconfRqstTDD", "rnsap.DCH_DeleteItem_RL_ReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_RL_ReconfigurationRequestTDD_RL_Information_item,
      { "RL-ReconfigurationRequestTDD-RL-Information", "rnsap.RL_ReconfigurationRequestTDD_RL_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationResponseList_RL_ReconfRspFDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dCHsInformationResponseList,
      { "dCHsInformationResponseList", "rnsap.dCHsInformationResponseList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DCH_InformationResponseList_RL_ReconfRspFDD", HFILL }},
    { &hf_rnsap_dL_CodeInformationList_RL_ReconfResp,
      { "dL-CodeInformationList-RL-ReconfResp", "rnsap.dL_CodeInformationList_RL_ReconfResp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_CodeInformationList_RL_ReconfRspFDD", HFILL }},
    { &hf_rnsap_dCHsInformationResponseList_01,
      { "dCHsInformationResponseList", "rnsap.dCHsInformationResponseList_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DCH_InformationResponseList_RL_ReconfRspTDD", HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationList_RL_ReconfRspTDD_item,
      { "DL-CCTrCH-InformationItem-RL-ReconfRspTDD", "rnsap.DL_CCTrCH_InformationItem_RL_ReconfRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_DPCH_ModifyInformation_LCR,
      { "dl-DPCH-ModifyInformation-LCR", "rnsap.dl_DPCH_ModifyInformation_LCR_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationModifyList_LCR_RL_ReconfRspTDD", HFILL }},
    { &hf_rnsap_dL_Timeslot_LCR_InformationModifyList_RL_ReconfRqstTDD,
      { "dL-Timeslot-LCR-InformationModifyList-RL-ReconfRqstTDD", "rnsap.dL_Timeslot_LCR_InformationModifyList_RL_ReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DL_Timeslot_LCR_InformationModifyList_RL_ReconfRspTDD", HFILL }},
    { &hf_rnsap_DL_Timeslot_LCR_InformationModifyList_RL_ReconfRspTDD_item,
      { "DL-Timeslot-LCR-InformationModifyItem-RL-ReconfRspTDD", "rnsap.DL_Timeslot_LCR_InformationModifyItem_RL_ReconfRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_maxPowerLCR,
      { "maxPowerLCR", "rnsap.maxPowerLCR",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_minPowerLCR,
      { "minPowerLCR", "rnsap.minPowerLCR",
        FT_INT32, BASE_DEC, NULL, 0,
        "DL_Power", HFILL }},
    { &hf_rnsap_Multiple_RL_InformationResponse_RL_ReconfRspTDD_item,
      { "RL-InformationResponse-RL-ReconfRspTDD", "rnsap.RL_InformationResponse_RL_ReconfRspTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL,
      { "rL", "rnsap.rL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_RL_FailureInd", HFILL }},
    { &hf_rnsap_rL_Set,
      { "rL-Set", "rnsap.rL_Set_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_RL_FailureInd", HFILL }},
    { &hf_rnsap_cCTrCH,
      { "cCTrCH", "rnsap.cCTrCH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "CCTrCH_RL_FailureInd", HFILL }},
    { &hf_rnsap_rL_InformationList_RL_FailureInd,
      { "rL-InformationList-RL-FailureInd", "rnsap.rL_InformationList_RL_FailureInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_RL_FailureInd_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_Set_InformationList_RL_FailureInd,
      { "rL-Set-InformationList-RL-FailureInd", "rnsap.rL_Set_InformationList_RL_FailureInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_InformationList_RL_FailureInd_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cCTrCH_InformationList_RL_FailureInd,
      { "cCTrCH-InformationList-RL-FailureInd", "rnsap.cCTrCH_InformationList_RL_FailureInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CCTrCH_InformationList_RL_FailureInd_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_RL_PreemptRequiredInd_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_01,
      { "rL", "rnsap.rL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_RL_RestoreInd", HFILL }},
    { &hf_rnsap_rL_Set_01,
      { "rL-Set", "rnsap.rL_Set_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_RL_RestoreInd", HFILL }},
    { &hf_rnsap_cCTrCH_01,
      { "cCTrCH", "rnsap.cCTrCH_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "CCTrCH_RL_RestoreInd", HFILL }},
    { &hf_rnsap_rL_InformationList_RL_RestoreInd,
      { "rL-InformationList-RL-RestoreInd", "rnsap.rL_InformationList_RL_RestoreInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_RL_RestoreInd_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_Set_InformationList_RL_RestoreInd,
      { "rL-Set-InformationList-RL-RestoreInd", "rnsap.rL_Set_InformationList_RL_RestoreInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_InformationList_RL_RestoreInd_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cCTrCH_InformationList_RL_RestoreInd,
      { "cCTrCH-InformationList-RL-RestoreInd", "rnsap.cCTrCH_InformationList_RL_RestoreInd",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_CCTrCH_InformationList_RL_RestoreInd_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_ReferencePowerInformationList_DL_PC_Rqst_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_CodeInformation_03,
      { "dl-CodeInformation", "rnsap.dl_CodeInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_CodeInformationList_PhyChReconfRqstFDD", HFILL }},
    { &hf_rnsap_ul_CCTrCH_Information_01,
      { "ul-CCTrCH-Information", "rnsap.ul_CCTrCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_CCTrCH_InformationList_PhyChReconfRqstTDD", HFILL }},
    { &hf_rnsap_dl_CCTrCH_Information_01,
      { "dl-CCTrCH-Information", "rnsap.dl_CCTrCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_CCTrCH_InformationList_PhyChReconfRqstTDD", HFILL }},
    { &hf_rnsap_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_item,
      { "UL-CCTrCH-InformationItem-PhyChReconfRqstTDD", "rnsap.UL_CCTrCH_InformationItem_PhyChReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ul_DPCH_Information_02,
      { "ul-DPCH-Information", "rnsap.ul_DPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UL_DPCH_InformationList_PhyChReconfRqstTDD", HFILL }},
    { &hf_rnsap_uL_Timeslot_InformationList_PhyChReconfRqstTDD,
      { "uL-Timeslot-InformationList-PhyChReconfRqstTDD", "rnsap.uL_Timeslot_InformationList_PhyChReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_item,
      { "UL-TimeslotLCR-InformationItem-PhyChReconfRqstTDD", "rnsap.UL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uL_Code_LCR_Information,
      { "uL-Code-LCR-Information", "rnsap.uL_Code_LCR_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "TDD_UL_Code_LCR_Information", HFILL }},
    { &hf_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD_item,
      { "UL-Timeslot-InformationItem-PhyChReconfRqstTDD", "rnsap.UL_Timeslot_InformationItem_PhyChReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD768_item,
      { "UL-Timeslot-InformationItem-PhyChReconfRqstTDD768", "rnsap.UL_Timeslot_InformationItem_PhyChReconfRqstTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_item,
      { "DL-CCTrCH-InformationItem-PhyChReconfRqstTDD", "rnsap.DL_CCTrCH_InformationItem_PhyChReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dl_DPCH_Information_02,
      { "dl-DPCH-Information", "rnsap.dl_DPCH_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "DL_DPCH_InformationList_PhyChReconfRqstTDD", HFILL }},
    { &hf_rnsap_dL_Timeslot_InformationList_PhyChReconfRqstTDD,
      { "dL-Timeslot-InformationList-PhyChReconfRqstTDD", "rnsap.dL_Timeslot_InformationList_PhyChReconfRqstTDD",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_item,
      { "DL-TimeslotLCR-InformationItem-PhyChReconfRqstTDD", "rnsap.DL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD_item,
      { "DL-Timeslot-InformationItem-PhyChReconfRqstTDD", "rnsap.DL_Timeslot_InformationItem_PhyChReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD768_item,
      { "DL-Timeslot-InformationItem-PhyChReconfRqstTDD768", "rnsap.DL_Timeslot_InformationItem_PhyChReconfRqstTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD_item,
      { "HSPDSCH-Timeslot-InformationItem-PhyChReconfRqstTDD", "rnsap.HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD_item,
      { "HSPDSCH-Timeslot-InformationItemLCR-PhyChReconfRqstTDD", "rnsap.HSPDSCH_Timeslot_InformationItemLCR_PhyChReconfRqstTDD_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768_item,
      { "HSPDSCH-Timeslot-InformationItem-PhyChReconfRqstTDD768", "rnsap.HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_RL_CongestInd_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dCH_Rate_Information,
      { "dCH-Rate-Information", "rnsap.dCH_Rate_Information",
        FT_UINT32, BASE_DEC, NULL, 0,
        "DCH_Rate_Information_RL_CongestInd", HFILL }},
    { &hf_rnsap_DCH_Rate_Information_RL_CongestInd_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_allowed_Rate_Information,
      { "allowed-Rate-Information", "rnsap.allowed_Rate_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_EDCH_MacdFlowSpecificInformationList_RL_CongestInd_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uRA,
      { "uRA", "rnsap.uRA_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "URA_PagingRqst", HFILL }},
    { &hf_rnsap_cell,
      { "cell", "rnsap.cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_PagingRqst", HFILL }},
    { &hf_rnsap_pagingCause,
      { "pagingCause", "rnsap.pagingCause",
        FT_UINT32, BASE_DEC, VALS(rnsap_PagingCause_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_cNDomainType,
      { "cNDomainType", "rnsap.cNDomainType",
        FT_UINT32, BASE_DEC, VALS(rnsap_CNDomainType_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_pagingRecordType,
      { "pagingRecordType", "rnsap.pagingRecordType",
        FT_UINT32, BASE_DEC, VALS(rnsap_PagingRecordType_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_02,
      { "rL", "rnsap.rL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Rqst", HFILL }},
    { &hf_rnsap_rLS,
      { "rLS", "rnsap.rLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Rqst", HFILL }},
    { &hf_rnsap_allRL,
      { "allRL", "rnsap.allRL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "All_RL_DM_Rqst", HFILL }},
    { &hf_rnsap_allRLS,
      { "allRLS", "rnsap.allRLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "All_RL_Set_DM_Rqst", HFILL }},
    { &hf_rnsap_rL_InformationList_DM_Rqst,
      { "rL-InformationList-DM-Rqst", "rnsap.rL_InformationList_DM_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_DM_Rqst_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSSICH_Info_DM_Rqst_item,
      { "HS-SICH-ID", "rnsap.HS_SICH_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_HSSICH_Info_DM_Rqst_Extension_item,
      { "HS-SICH-ID-Extension", "rnsap.HS_SICH_ID_Extension",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_Set_InformationList_DM_Rqst,
      { "rL-Set-InformationList-DM-Rqst", "rnsap.rL_Set_InformationList_DM_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_InformationList_DM_Rqst_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rLs,
      { "rLs", "rnsap.rLs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Rsp", HFILL }},
    { &hf_rnsap_rLS_01,
      { "rLS", "rnsap.rLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Rsp", HFILL }},
    { &hf_rnsap_allRL_01,
      { "allRL", "rnsap.allRL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Rsp", HFILL }},
    { &hf_rnsap_allRLS_01,
      { "allRLS", "rnsap.allRLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Rsp", HFILL }},
    { &hf_rnsap_rL_InformationList_DM_Rsp,
      { "rL-InformationList-DM-Rsp", "rnsap.rL_InformationList_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_Set_InformationList_DM_Rsp,
      { "rL-Set-InformationList-DM-Rsp", "rnsap.rL_Set_InformationList_DM_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_DM_Rsp_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dedicatedMeasurementValue,
      { "dedicatedMeasurementValue", "rnsap.dedicatedMeasurementValue",
        FT_UINT32, BASE_DEC, VALS(rnsap_DedicatedMeasurementValue_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_InformationList_DM_Rsp_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_item,
      { "Multiple-DedicatedMeasurementValueItem-TDD-DM-Rsp", "rnsap.Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_item,
      { "Multiple-DedicatedMeasurementValueItem-LCR-TDD-DM-Rsp", "rnsap.Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_item,
      { "Multiple-HSSICHMeasurementValueItem-TDD-DM-Rsp", "rnsap.Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp_item,
      { "Multiple-DedicatedMeasurementValueItem-TDD768-DM-Rsp", "rnsap.Multiple_DedicatedMeasurementValueItem_TDD768_DM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_03,
      { "rL", "rnsap.rL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Fail", HFILL }},
    { &hf_rnsap_rLS_02,
      { "rLS", "rnsap.rLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Fail", HFILL }},
    { &hf_rnsap_allRL_02,
      { "allRL", "rnsap.allRL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Fail", HFILL }},
    { &hf_rnsap_allRLS_02,
      { "allRLS", "rnsap.allRLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Fail", HFILL }},
    { &hf_rnsap_rL_unsuccessful_InformationRespList_DM_Fail,
      { "rL-unsuccessful-InformationRespList-DM-Fail", "rnsap.rL_unsuccessful_InformationRespList_DM_Fail",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_successful_InformationRespList_DM_Fail,
      { "rL-successful-InformationRespList-DM-Fail", "rnsap.rL_successful_InformationRespList_DM_Fail",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_Set_unsuccessful_InformationRespList_DM_Fail,
      { "rL-Set-unsuccessful-InformationRespList-DM-Fail", "rnsap.rL_Set_unsuccessful_InformationRespList_DM_Fail",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_Set_successful_InformationRespList_DM_Fail,
      { "rL-Set-successful-InformationRespList-DM-Fail", "rnsap.rL_Set_successful_InformationRespList_DM_Fail",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_individualcause,
      { "individualcause", "rnsap.individualcause",
        FT_UINT32, BASE_DEC, VALS(rnsap_Cause_vals), 0,
        "Cause", HFILL }},
    { &hf_rnsap_RL_Successful_InformationRespList_DM_Fail_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_Successful_InformationRespList_DM_Fail_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rLs_01,
      { "rLs", "rnsap.rLs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Rprt", HFILL }},
    { &hf_rnsap_rLS_03,
      { "rLS", "rnsap.rLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Rprt", HFILL }},
    { &hf_rnsap_allRL_03,
      { "allRL", "rnsap.allRL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Rprt", HFILL }},
    { &hf_rnsap_allRLS_03,
      { "allRLS", "rnsap.allRLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Rprt", HFILL }},
    { &hf_rnsap_rL_InformationList_DM_Rprt,
      { "rL-InformationList-DM-Rprt", "rnsap.rL_InformationList_DM_Rprt",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_Set_InformationList_DM_Rprt,
      { "rL-Set-InformationList-DM-Rprt", "rnsap.rL_Set_InformationList_DM_Rprt",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_InformationList_DM_Rprt_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_dedicatedMeasurementValueInformation,
      { "dedicatedMeasurementValueInformation", "rnsap.dedicatedMeasurementValueInformation",
        FT_UINT32, BASE_DEC, VALS(rnsap_DedicatedMeasurementValueInformation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_InformationList_DM_Rprt_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_04,
      { "rL", "rnsap.rL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Fail_Ind", HFILL }},
    { &hf_rnsap_rLS_04,
      { "rLS", "rnsap.rLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Fail_Ind", HFILL }},
    { &hf_rnsap_allRL_04,
      { "allRL", "rnsap.allRL_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_DM_Fail_Ind", HFILL }},
    { &hf_rnsap_allRLS_04,
      { "allRLS", "rnsap.allRLS_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "RL_Set_DM_Fail_Ind", HFILL }},
    { &hf_rnsap_rL_unsuccessful_InformationRespList_DM_Fail_Ind,
      { "rL-unsuccessful-InformationRespList-DM-Fail-Ind", "rnsap.rL_unsuccessful_InformationRespList_DM_Fail_Ind",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_rL_Set_unsuccessful_InformationRespList_DM_Fail_Ind,
      { "rL-Set-unsuccessful-InformationRespList-DM-Fail-Ind", "rnsap.rL_Set_unsuccessful_InformationRespList_DM_Fail_Ind",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail_Ind_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail_Ind_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_fACH_FlowControlInformation,
      { "fACH-FlowControlInformation", "rnsap.fACH_FlowControlInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "FACH_FlowControlInformation_CTCH_ResourceRspFDD", HFILL }},
    { &hf_rnsap_fACH_FlowControlInformation_01,
      { "fACH-FlowControlInformation", "rnsap.fACH_FlowControlInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "FACH_FlowControlInformation_CTCH_ResourceRspTDD", HFILL }},
    { &hf_rnsap_cell_01,
      { "cell", "rnsap.cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_CM_Rqst", HFILL }},
    { &hf_rnsap_additional_CommonMeasurementObjectType_CM_Rqst,
      { "additional-CommonMeasurementObjectType-CM-Rqst", "rnsap.additional_CommonMeasurementObjectType_CM_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_neighbouringCellMeasurementInformation,
      { "neighbouringCellMeasurementInformation", "rnsap.neighbouringCellMeasurementInformation",
        FT_UINT32, BASE_DEC, NULL, 0,
        "NeighbouringCellMeasurementInfo", HFILL }},
    { &hf_rnsap_NeighbouringCellMeasurementInfo_item,
      { "NeighbouringCellMeasurementInfo item", "rnsap.NeighbouringCellMeasurementInfo_item",
        FT_UINT32, BASE_DEC, VALS(rnsap_NeighbouringCellMeasurementInfo_item_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_neighbouringFDDCellMeasurementInformation,
      { "neighbouringFDDCellMeasurementInformation", "rnsap.neighbouringFDDCellMeasurementInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_neighbouringTDDCellMeasurementInformation,
      { "neighbouringTDDCellMeasurementInformation", "rnsap.neighbouringTDDCellMeasurementInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_extension_neighbouringCellMeasurementInformation,
      { "extension-neighbouringCellMeasurementInformation", "rnsap.extension_neighbouringCellMeasurementInformation_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_extension_neighbouringCellMeasurementInformation768,
      { "extension-neighbouringCellMeasurementInformation768", "rnsap.extension_neighbouringCellMeasurementInformation768_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_GsmCellList_CM_Rqst_item,
      { "GsmCellItem-CM-Rqst", "rnsap.GsmCellItem_CM_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_measurementID,
      { "measurementID", "rnsap.measurementID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_gsmCell,
      { "gsmCell", "rnsap.gsmCell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "GSM_Cell_CM_Rqst", HFILL }},
    { &hf_rnsap_cell_02,
      { "cell", "rnsap.cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_CM_Rsp", HFILL }},
    { &hf_rnsap_GsmCellList_CM_Rsp_item,
      { "GsmCellItem-CM-Rsp", "rnsap.GsmCellItem_CM_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Extension_FailureMeasurementList_item,
      { "Extension-FailureMeasurementItem", "rnsap.Extension_FailureMeasurementItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cell_03,
      { "cell", "rnsap.cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_CM_Rprt", HFILL }},
    { &hf_rnsap_commonMeasurementValueInformation,
      { "commonMeasurementValueInformation", "rnsap.commonMeasurementValueInformation",
        FT_UINT32, BASE_DEC, VALS(rnsap_CommonMeasurementValueInformation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_GsmCellList_CM_Rprt_item,
      { "GsmCellItem-CM-Rprt", "rnsap.GsmCellItem_CM_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Extension_TerminationMeasurementList_item,
      { "Extension-TerminationMeasurementItem", "rnsap.Extension_TerminationMeasurementItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Extension_FailureIndicationMeasurementList_item,
      { "Extension-FailureIndicationMeasurementItem", "rnsap.Extension_FailureIndicationMeasurementItem_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cell_04,
      { "cell", "rnsap.cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_InfEx_Rqst", HFILL }},
    { &hf_rnsap_extension_InformationExchangeObjectType_InfEx_Rqst,
      { "extension-InformationExchangeObjectType-InfEx-Rqst", "rnsap.extension_InformationExchangeObjectType_InfEx_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst_item,
      { "MBMS-Bearer-Service-in-MBMS-Cell-Item-InfEx-Rqst", "rnsap.MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rqst_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mBMS_Bearer_Service_List_InfEx_Rqst,
      { "mBMS-Bearer-Service-List-InfEx-Rqst", "rnsap.mBMS_Bearer_Service_List_InfEx_Rqst",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_List_InfEx_Rqst_item,
      { "TMGI", "rnsap.TMGI_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Cell_InfEx_Rqst_item,
      { "C-ID", "rnsap.C_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ANR_Cell_InfEx_Rqst_item,
      { "C-ID", "rnsap.C_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Common_E_RGCH_Cell_InfEx_Rqst_item,
      { "C-ID", "rnsap.C_ID",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cell_05,
      { "cell", "rnsap.cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_InfEx_Rsp", HFILL }},
    { &hf_rnsap_extension_InformationExchangeObjectType_InfEx_Rsp,
      { "extension-InformationExchangeObjectType-InfEx-Rsp", "rnsap.extension_InformationExchangeObjectType_InfEx_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_List_InfEx_Rsp_item,
      { "MBMS-Bearer-ServiceItemIEs-InfEx-Rsp", "rnsap.MBMS_Bearer_ServiceItemIEs_InfEx_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp_item,
      { "MBMS-Bearer-Service-in-MBMS-Cell-Item-InfEx-Rsp", "rnsap.MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mBMS_Bearer_Service_List_InfEx_Rsp,
      { "mBMS-Bearer-Service-List-InfEx-Rsp", "rnsap.mBMS_Bearer_Service_List_InfEx_Rsp",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Cell_InfEx_Rsp_item,
      { "MBMS-Cell-Item-InfEx-Rsp", "rnsap.MBMS_Cell_Item_InfEx_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ANR_Cell_InfEx_Rsp_item,
      { "ANR-Cell-ItemIEs-InfEx-Rsp", "rnsap.ANR_Cell_ItemIEs_InfEx_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Common_E_RGCH_Cell_InfEx_Rsp_item,
      { "Common-E-RGCH-Cell-ItemIEs-InfEx-Rsp", "rnsap.Common_E_RGCH_Cell_ItemIEs_InfEx_Rsp_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_cell_06,
      { "cell", "rnsap.cell_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Cell_InfEx_Rprt", HFILL }},
    { &hf_rnsap_extension_InformationExchangeObjectType_InfEx_Rprt,
      { "extension-InformationExchangeObjectType-InfEx-Rprt", "rnsap.extension_InformationExchangeObjectType_InfEx_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_requestedDataValueInformation,
      { "requestedDataValueInformation", "rnsap.requestedDataValueInformation",
        FT_UINT32, BASE_DEC, VALS(rnsap_RequestedDataValueInformation_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt_item,
      { "MBMS-Bearer-Service-in-MBMS-Cell-Item-InfEx-Rprt", "rnsap.MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_mBMS_Bearer_Service_List_InfEx_Rprt,
      { "mBMS-Bearer-Service-List-InfEx-Rprt", "rnsap.mBMS_Bearer_Service_List_InfEx_Rprt",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Bearer_Service_List_InfEx_Rprt_item,
      { "MBMS-Bearer-Service-List-Item-InfEx-Rprt", "rnsap.MBMS_Bearer_Service_List_Item_InfEx_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_MBMS_Cell_InfEx_Rprt_item,
      { "MBMS-Cell-Item-InfEx-Rprt", "rnsap.MBMS_Cell_Item_InfEx_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Common_E_RGCH_Cell_InfEx_Rprt_item,
      { "Common-E-RGCH-Cell-Item-InfEx-Rprt", "rnsap.Common_E_RGCH_Cell_Item_InfEx_Rprt_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_context,
      { "context", "rnsap.context_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "ContextList_Reset", HFILL }},
    { &hf_rnsap_all_contexts,
      { "all-contexts", "rnsap.all_contexts_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_contextGroup,
      { "contextGroup", "rnsap.contextGroup_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "ContextGroupList_Reset", HFILL }},
    { &hf_rnsap_contextInfoList_Reset,
      { "contextInfoList-Reset", "rnsap.contextInfoList_Reset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ContextInfoList_Reset_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_contextType_Reset,
      { "contextType-Reset", "rnsap.contextType_Reset",
        FT_UINT32, BASE_DEC, VALS(rnsap_ContextType_Reset_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_dRNTI,
      { "dRNTI", "rnsap.dRNTI",
        FT_UINT32, BASE_DEC, NULL, 0,
        "D_RNTI", HFILL }},
    { &hf_rnsap_extension_ContextType_Reset,
      { "extension-ContextType-Reset", "rnsap.extension_ContextType_Reset_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_contextGroupInfoList_Reset,
      { "contextGroupInfoList-Reset", "rnsap.contextGroupInfoList_Reset",
        FT_UINT32, BASE_DEC, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ContextGroupInfoList_Reset_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_s_RNTI_Group,
      { "s-RNTI-Group", "rnsap.s_RNTI_Group_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_DelayedActivationInformationList_RL_ActivationCmdFDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_delayed_activation_update,
      { "delayed-activation-update", "rnsap.delayed_activation_update",
        FT_UINT32, BASE_DEC, VALS(rnsap_DelayedActivationUpdate_vals), 0,
        "DelayedActivationUpdate", HFILL }},
    { &hf_rnsap_DelayedActivationInformationList_RL_ActivationCmdTDD_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_RL_ParameterUpdateIndicationFDD_RL_InformationList_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_uE_Measurement_Value,
      { "uE-Measurement-Value", "rnsap.uE_Measurement_Value",
        FT_UINT32, BASE_DEC, VALS(rnsap_UE_Measurement_Value_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_HS_Cell_Information_RL_Param_Upd_item,
      { "Additional-HS-Cell-Information-RL-Param-Upd-ItemIEs", "rnsap.Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_hS_DSCH_FDD_Secondary_Serving_Update_Information,
      { "hS-DSCH-FDD-Secondary-Serving-Update-Information", "rnsap.hS_DSCH_FDD_Secondary_Serving_Update_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd_item,
      { "Additional-EDCH-Cell-Information-RL-Param-Upd-ItemIEs", "rnsap.Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_additional_EDCH_FDD_Update_Information,
      { "additional-EDCH-FDD-Update-Information", "rnsap.additional_EDCH_FDD_Update_Information_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_ListOfInterfacesToTrace_item,
      { "ProtocolIE-Single-Container", "rnsap.ProtocolIE_Single_Container_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_interface,
      { "interface", "rnsap.interface",
        FT_UINT32, BASE_DEC, VALS(rnsap_T_interface_vals), 0,
        NULL, HFILL }},
    { &hf_rnsap_privateIEs,
      { "privateIEs", "rnsap.privateIEs",
        FT_UINT32, BASE_DEC, NULL, 0,
        "PrivateIE_Container", HFILL }},
    { &hf_rnsap_suspension,
      { "suspension", "rnsap.suspension_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Suspension_Control_Type", HFILL }},
    { &hf_rnsap_resume,
      { "resume", "rnsap.resume_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Resume_Control_Type", HFILL }},
    { &hf_rnsap_controlled_Object_Scope,
      { "controlled-Object-Scope", "rnsap.controlled_Object_Scope_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_initiatingMessage,
      { "initiatingMessage", "rnsap.initiatingMessage_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_successfulOutcome,
      { "successfulOutcome", "rnsap.successfulOutcome_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_unsuccessfulOutcome,
      { "unsuccessfulOutcome", "rnsap.unsuccessfulOutcome_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_outcome,
      { "outcome", "rnsap.outcome_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},
    { &hf_rnsap_value_02,
      { "value", "rnsap.value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "InitiatingMessage_value", HFILL }},
    { &hf_rnsap_value_03,
      { "value", "rnsap.value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "SuccessfulOutcome_value", HFILL }},
    { &hf_rnsap_value_04,
      { "value", "rnsap.value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "UnsuccessfulOutcome_value", HFILL }},
    { &hf_rnsap_value_05,
      { "value", "rnsap.value_element",
        FT_NONE, BASE_NONE, NULL, 0,
        "Outcome_value", HFILL }},
  };

  /* List of subtrees */
  static gint *ett[] = {
    &ett_rnsap,
    &ett_rnsap_transportLayerAddress,
    &ett_rnsap_transportLayerAddress_nsap,
    &ett_rnsap_PrivateIE_ID,
    &ett_rnsap_ProcedureID,
    &ett_rnsap_TransactionID,
    &ett_rnsap_ProtocolIE_Container,
    &ett_rnsap_ProtocolIE_Field,
    &ett_rnsap_ProtocolExtensionContainer,
    &ett_rnsap_ProtocolExtensionField,
    &ett_rnsap_PrivateIE_Container,
    &ett_rnsap_PrivateIE_Field,
    &ett_rnsap_ActivationInformation,
    &ett_rnsap_ActivationInformationItem,
    &ett_rnsap_Active_MBMS_Bearer_Service_ListFDD,
    &ett_rnsap_Active_MBMS_Bearer_Service_ListFDD_PFL,
    &ett_rnsap_Active_MBMS_Bearer_Service_ListTDD,
    &ett_rnsap_Active_MBMS_Bearer_Service_ListTDD_PFL,
    &ett_rnsap_Active_Pattern_Sequence_Information,
    &ett_rnsap_Additional_EDCH_Cell_Information_Response_RLAddList,
    &ett_rnsap_Additional_EDCH_Cell_Information_Response_RLAdd_ItemIEs,
    &ett_rnsap_Additional_EDCH_Setup_Info,
    &ett_rnsap_ANRReportIndication,
    &ett_rnsap_Additional_EDCH_Cell_Information_Setup,
    &ett_rnsap_Additional_EDCH_FDD_Setup_Cell_Information,
    &ett_rnsap_Additional_EDCH_UL_DPCH_Information_Setup,
    &ett_rnsap_Additional_EDCH_F_DPCH_Information,
    &ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Setup_List,
    &ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Setup_ItemIEs,
    &ett_rnsap_Additional_EDCH_Cell_Information_To_Add_List,
    &ett_rnsap_Additional_EDCH_Cell_Information_To_Add_ItemIEs,
    &ett_rnsap_Additional_EDCH_UL_DPCH_Information_Addition,
    &ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_List,
    &ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Add_ItemIEs,
    &ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Modify_List,
    &ett_rnsap_Additional_EDCH_RL_Specific_Information_To_Modify_ItemIEs,
    &ett_rnsap_Additional_EDCH_FDD_Information,
    &ett_rnsap_DTX_Information2,
    &ett_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_List,
    &ett_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info,
    &ett_rnsap_Additional_EDCH_MAC_d_Flow_Specific_Information_Response_List,
    &ett_rnsap_Additional_EDCH_MAC_d_Flows_Specific_Info_Response,
    &ett_rnsap_Additional_EDCH_Cell_Information_Response_List,
    &ett_rnsap_Additional_EDCH_FDD_Information_Response_ItemIEs,
    &ett_rnsap_EDCH_Additional_RL_Specific_Information_Response_List,
    &ett_rnsap_EDCH_Additional_RL_Specific_Information_Response_ItemIEs,
    &ett_rnsap_Additional_EDCH_Cell_Information_Response_RLReconf_List,
    &ett_rnsap_Additional_EDCH_FDD_Information_Response_RLReconf_Items,
    &ett_rnsap_EDCH_Additional_RL_Specific_Modified_Information_Response_List,
    &ett_rnsap_EDCH_Additional_RL_Specific_Modified_Information_Response_ItemIEs,
    &ett_rnsap_Additional_EDCH_Cell_Information_ConfigurationChange_List,
    &ett_rnsap_Additional_EDCH_ConfigurationChange_Info_ItemIEs,
    &ett_rnsap_Additional_EDCH_UL_DPCH_Information_Modify,
    &ett_rnsap_Additional_EDCH_Cell_Information_Removal_List,
    &ett_rnsap_Additional_EDCH_Cell_Information_Removal_Info_ItemIEs,
    &ett_rnsap_Additional_EDCH_FDD_Update_Information,
    &ett_rnsap_Additional_EDCH_DL_Control_Channel_Change_Information_List,
    &ett_rnsap_Additional_EDCH_DL_Control_Channel_Change_Info_ItemIEs,
    &ett_rnsap_AdditionalPreferredFrequency,
    &ett_rnsap_AdditionalPreferredFrequencyItem,
    &ett_rnsap_AffectedUEInformationForMBMS,
    &ett_rnsap_AllocationRetentionPriority,
    &ett_rnsap_Allowed_Rate_Information,
    &ett_rnsap_Angle_Of_Arrival_Value_LCR,
    &ett_rnsap_ANR_Cell_Information,
    &ett_rnsap_ANR_FDD_CellInformation,
    &ett_rnsap_ANR_TDD_CellInformation,
    &ett_rnsap_ANR_LCR_TDD_CellInformation,
    &ett_rnsap_ANR_Multiple_PLMN_List,
    &ett_rnsap_BadSatellites,
    &ett_rnsap_T_badSatelliteInformation,
    &ett_rnsap_T_badSatelliteInformation_item,
    &ett_rnsap_BDS_IGPInfoList,
    &ett_rnsap_BDS_IGPInfo,
    &ett_rnsap_BDS_Ionospheric_Grid_Model,
    &ett_rnsap_BSIC,
    &ett_rnsap_BurstModeParameters,
    &ett_rnsap_Cause,
    &ett_rnsap_CellBased,
    &ett_rnsap_CellIdList,
    &ett_rnsap_Cell_Capacity_Class_Value,
    &ett_rnsap_CGI,
    &ett_rnsap_CommonMeasurementAccuracy,
    &ett_rnsap_CommonMeasurementValue,
    &ett_rnsap_CommonMeasurementValueInformation,
    &ett_rnsap_CommonMeasurementAvailable,
    &ett_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDD,
    &ett_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationFDDItem,
    &ett_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationLCR,
    &ett_rnsap_Common_EDCH_MAC_d_Flow_Specific_InformationItemLCR,
    &ett_rnsap_Common_E_DCH_LogicalChannelInformation,
    &ett_rnsap_Common_E_DCH_LogicalChannelInformationItem,
    &ett_rnsap_Common_E_RGCH_Cell_Information,
    &ett_rnsap_Common_E_RGCH_Cell_Info,
    &ett_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information,
    &ett_rnsap_Continuous_Packet_Connectivity_DTX_DRX_Information_to_Modify,
    &ett_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information,
    &ett_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_InformationItem,
    &ett_rnsap_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response,
    &ett_rnsap_CorrespondingCells,
    &ett_rnsap_CPC_Information,
    &ett_rnsap_Counting_Information,
    &ett_rnsap_Counting_Information_List,
    &ett_rnsap_CriticalityDiagnostics,
    &ett_rnsap_CriticalityDiagnostics_IE_List,
    &ett_rnsap_CriticalityDiagnostics_IE_List_item,
    &ett_rnsap_MessageStructure,
    &ett_rnsap_MessageStructure_item,
    &ett_rnsap_CN_CS_DomainIdentifier,
    &ett_rnsap_CN_PS_DomainIdentifier,
    &ett_rnsap_CPC_InformationLCR,
    &ett_rnsap_ContinuousPacketConnectivity_DRX_InformationLCR,
    &ett_rnsap_HS_SCCH_DRX_Information_LCR,
    &ett_rnsap_E_AGCH_DRX_Information_LCR,
    &ett_rnsap_E_AGCH_DRX_Parameters,
    &ett_rnsap_ContinuousPacketConnectivity_DRX_Information_to_Modify_LCR,
    &ett_rnsap_DRX_Information_to_Modify_LCR,
    &ett_rnsap_DRX_Information_to_Modify_Items_LCR,
    &ett_rnsap_ContinuousPacketConnectivity_DRX_Information_ResponseLCR,
    &ett_rnsap_HS_SCCH_DRX_Information_ResponseLCR,
    &ett_rnsap_E_AGCH_DRX_Information_ResponseLCR,
    &ett_rnsap_E_AGCH_DRX_Parameters_Response,
    &ett_rnsap_Controlled_Object_Scope,
    &ett_rnsap_DBDS_CorrectionsReq,
    &ett_rnsap_DBDS_Corrections,
    &ett_rnsap_DBDS_Info,
    &ett_rnsap_DBDS_SignalInfoList,
    &ett_rnsap_DBDS_SignalInfo,
    &ett_rnsap_DBDS_InfoList,
    &ett_rnsap_DCH_ENH_Information,
    &ett_rnsap_DCH_ENH_Information_to_Modify,
    &ett_rnsap_DCH_ENH_Concat,
    &ett_rnsap_DCH_ENH_Information_Reconf,
    &ett_rnsap_DCH_FDD_Information,
    &ett_rnsap_DCH_FDD_InformationItem,
    &ett_rnsap_DCH_MeasurementOccasion_Information,
    &ett_rnsap_DchMeasurementOccasionInformation_Item,
    &ett_rnsap_Measurement_Occasion_Pattern_Sequence_parameters,
    &ett_rnsap_DCH_Specific_FDD_InformationList,
    &ett_rnsap_DCH_Specific_FDD_Item,
    &ett_rnsap_DCH_InformationResponse,
    &ett_rnsap_DCH_InformationResponseItem,
    &ett_rnsap_DCH_TDD_Information,
    &ett_rnsap_DCH_TDD_InformationItem,
    &ett_rnsap_DCH_Specific_TDD_InformationList,
    &ett_rnsap_DCH_Specific_TDD_Item,
    &ett_rnsap_DedicatedMeasurementValue,
    &ett_rnsap_DedicatedMeasurementValueInformation,
    &ett_rnsap_DedicatedMeasurementAvailable,
    &ett_rnsap_DelayedActivation,
    &ett_rnsap_DelayedActivationUpdate,
    &ett_rnsap_Activate_Info,
    &ett_rnsap_Deactivate_Info,
    &ett_rnsap_Execution_Type,
    &ett_rnsap_DGANSSCorrections,
    &ett_rnsap_T_dGANSS_Information,
    &ett_rnsap_T_dGANSS_Information_item,
    &ett_rnsap_T_dGANSS_SignalInformation,
    &ett_rnsap_T_dGANSS_SignalInformation_item,
    &ett_rnsap_DGANSS_Corrections_Req,
    &ett_rnsap_DGANSSThreshold,
    &ett_rnsap_DGNSS_ValidityPeriod,
    &ett_rnsap_DGPSCorrections,
    &ett_rnsap_T_satellite_DGPSCorrections_Information,
    &ett_rnsap_T_satellite_DGPSCorrections_Information_item,
    &ett_rnsap_DGPSThreshold,
    &ett_rnsap_DL_TBS,
    &ett_rnsap_HS_DSCH_Cell_List,
    &ett_rnsap_HS_DSCH_Cell,
    &ett_rnsap_DL_PowerBalancing_Information,
    &ett_rnsap_DL_ReferencePowerInformationList,
    &ett_rnsap_DL_ReferencePowerInformationItem,
    &ett_rnsap_DL_ReferencePowerInformation,
    &ett_rnsap_DL_Timeslot_Information,
    &ett_rnsap_DL_Timeslot_InformationItem,
    &ett_rnsap_DL_TimeslotLCR_Information,
    &ett_rnsap_DL_TimeslotLCR_InformationItem,
    &ett_rnsap_DL_Timeslot_Information768,
    &ett_rnsap_DL_Timeslot_InformationItem768,
    &ett_rnsap_DL_TimeSlot_ISCP_Info,
    &ett_rnsap_DL_TimeSlot_ISCP_InfoItem,
    &ett_rnsap_DL_TimeSlot_ISCP_LCR_Information,
    &ett_rnsap_DL_TimeSlot_ISCP_LCR_InfoItem,
    &ett_rnsap_Downlink_TPC_enhancements_Information,
    &ett_rnsap_Downlink_TPC_enhancements_Reconf,
    &ett_rnsap_DRX_Information,
    &ett_rnsap_DRX_Information_to_Modify,
    &ett_rnsap_DRX_Information_to_Modify_Items,
    &ett_rnsap_DSCH_FlowControlInformation,
    &ett_rnsap_DSCH_FlowControlItem,
    &ett_rnsap_DSCH_TDD_Information,
    &ett_rnsap_DSCH_TDD_InformationItem,
    &ett_rnsap_DTX_Cycle_2ms_Items,
    &ett_rnsap_DTX_Cycle_2ms_to_Modify_Items,
    &ett_rnsap_DTX_Cycle_10ms_Items,
    &ett_rnsap_DTX_Cycle_10ms_to_Modify_Items,
    &ett_rnsap_DTX_Information,
    &ett_rnsap_DTX_Information_to_Modify,
    &ett_rnsap_DTX_Information_to_Modify_Items,
    &ett_rnsap_EARFCN_Information,
    &ett_rnsap_EARFCN_FDD,
    &ett_rnsap_EARFCN_FDD_Extended,
    &ett_rnsap_ECGI,
    &ett_rnsap_EDCH_FDD_DL_ControlChannelInformation,
    &ett_rnsap_EDCH_FDD_Information,
    &ett_rnsap_EDCH_FDD_InformationResponse,
    &ett_rnsap_EDCH_MACdFlow_Specific_InformationResponse,
    &ett_rnsap_EDCH_MACdFlow_Specific_InformationResponseItem,
    &ett_rnsap_EDCH_FDD_Information_To_Modify,
    &ett_rnsap_E_DCH_FDD_Update_Information,
    &ett_rnsap_E_DCH_MACdFlow_Specific_UpdateInformation,
    &ett_rnsap_E_DCH_MACdFlow_Specific_UpdateInformation_Item,
    &ett_rnsap_E_DCH_DL_Control_Channel_Change_Information,
    &ett_rnsap_E_DCH_DL_Control_Channel_Change_Information_Item,
    &ett_rnsap_E_DCH_DL_Control_Channel_Grant_Information,
    &ett_rnsap_E_DCH_DL_Control_Channel_Grant_Information_Item,
    &ett_rnsap_E_DCH_Grant_Type_Information,
    &ett_rnsap_E_DCH_LogicalChannelInformation,
    &ett_rnsap_E_DCH_LogicalChannelInformationItem,
    &ett_rnsap_E_DCH_MACdPDU_SizeList,
    &ett_rnsap_E_DCH_MACdPDU_SizeListItem,
    &ett_rnsap_E_DCH_LogicalChannelToModify,
    &ett_rnsap_E_DCH_LogicalChannelToModifyItem,
    &ett_rnsap_E_DCH_MACdPDU_SizeToModifyList,
    &ett_rnsap_E_DCH_LogicalChannelToDelete,
    &ett_rnsap_E_DCH_LogicalChannelToDeleteItem,
    &ett_rnsap_EDCH_MACdFlows_Information,
    &ett_rnsap_EDCH_MACdFlow_Specific_InfoList,
    &ett_rnsap_EDCH_MACdFlow_Specific_InfoItem,
    &ett_rnsap_EDCH_MACdFlow_Specific_InfoToModifyList,
    &ett_rnsap_EDCH_MACdFlow_Specific_InfoToModifyItem,
    &ett_rnsap_EDCH_MACdFlows_To_Delete,
    &ett_rnsap_EDCH_MACdFlows_To_Delete_Item,
    &ett_rnsap_E_DCH_Non_Scheduled_Transmission_Grant_Items,
    &ett_rnsap_E_DCH_Serving_cell_change_informationResponse,
    &ett_rnsap_E_DCH_serving_cell_change_choice,
    &ett_rnsap_E_DCH_serving_cell_change_successful,
    &ett_rnsap_E_DCH_RL_InformationList_Rsp,
    &ett_rnsap_E_DCH_RL_InformationList_Rsp_Item,
    &ett_rnsap_E_DCH_serving_cell_change_unsuccessful,
    &ett_rnsap_E_DCH_TTI_Length,
    &ett_rnsap_E_DCH_TTI_Length_to_Modify,
    &ett_rnsap_EDPCH_Information_FDD,
    &ett_rnsap_EDPCH_Information_RLReconfRequest_FDD,
    &ett_rnsap_EDCH_Serving_RL,
    &ett_rnsap_EDCH_Serving_RL_in_this_DRNS,
    &ett_rnsap_Enhanced_FACH_Information_ResponseFDD,
    &ett_rnsap_Enhanced_FACH_Information_ResponseLCR,
    &ett_rnsap_E_TFCI_Boost_Information,
    &ett_rnsap_E_TFCS_Information,
    &ett_rnsap_EventA,
    &ett_rnsap_EventB,
    &ett_rnsap_EventC,
    &ett_rnsap_EventD,
    &ett_rnsap_EventE,
    &ett_rnsap_EventF,
    &ett_rnsap_EventH,
    &ett_rnsap_Event1F_Parameters,
    &ett_rnsap_Event1I_Parameters,
    &ett_rnsap_E_DCH_Information,
    &ett_rnsap_E_PUCH_Information,
    &ett_rnsap_E_TFCS_Information_TDD,
    &ett_rnsap_E_DCH_QPSK_RefBetaInfo,
    &ett_rnsap_E_DCH_sixteenQAM_RefBetaInfo,
    &ett_rnsap_E_DCH_RefBeta_Item,
    &ett_rnsap_E_DCH_MACdFlows_Information_TDD,
    &ett_rnsap_E_DCH_MACdFlow_InfoTDDItem,
    &ett_rnsap_E_DCH_TDD_Information,
    &ett_rnsap_E_DCH_Information_Reconfig,
    &ett_rnsap_E_DCH_TDD_Information_to_Modify,
    &ett_rnsap_E_DCH_TDD_Information_to_Modify_List,
    &ett_rnsap_E_DCH_MACdFlow_ModifyTDDItem,
    &ett_rnsap_E_DCH_Information_Response,
    &ett_rnsap_Fast_TTI_switching_Mode_synchronized,
    &ett_rnsap_Fast_TTI_switching_Mode_unsynchronized,
    &ett_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp,
    &ett_rnsap_E_DCH_TDD_MACdFlow_Specific_InformationResp_Item,
    &ett_rnsap_E_AGCH_Specific_InformationRespListTDD,
    &ett_rnsap_E_AGCH_Specific_InformationResp_ItemTDD,
    &ett_rnsap_E_HICH_InformationResp,
    &ett_rnsap_E_DCH_Non_Scheduled_Grant_Info,
    &ett_rnsap_E_DCH_768_Information,
    &ett_rnsap_E_DCH_TDD_Information768,
    &ett_rnsap_E_DCH_768_Information_Reconfig,
    &ett_rnsap_E_DCH_768_Information_Response,
    &ett_rnsap_E_AGCH_Specific_InformationRespList768TDD,
    &ett_rnsap_E_AGCH_Specific_InformationResp_Item768TDD,
    &ett_rnsap_E_HICH_InformationResp768,
    &ett_rnsap_E_DCH_Non_Scheduled_Grant_Info768,
    &ett_rnsap_E_DCH_LCR_Information,
    &ett_rnsap_E_PUCH_LCR_Information,
    &ett_rnsap_E_DCH_LCR_TDD_Information,
    &ett_rnsap_E_DCH_LCR_Information_Reconfig,
    &ett_rnsap_E_DCH_LCR_Information_Response,
    &ett_rnsap_E_AGCH_Specific_InformationRespList_LCR_TDD,
    &ett_rnsap_E_AGCH_Specific_InformationResp_Item_LCR_TDD,
    &ett_rnsap_E_HICH_Specific_InformationResp_LCR,
    &ett_rnsap_E_HICH_Scheduled_InformationRespList_LCR_TDD,
    &ett_rnsap_E_HICH_Scheduled_InformationResp_Item_LCR_TDD,
    &ett_rnsap_E_HICH_InformationResp_LCR,
    &ett_rnsap_E_DCH_Non_Scheduled_Grant_Info_LCR,
    &ett_rnsap_Radio_Links_without_DPCH_FDPCH_Indication,
    &ett_rnsap_Radio_Links_without_DPCH_FDPCH_Information,
    &ett_rnsap_Extended_S_RNTI_Group,
    &ett_rnsap_ExtendedAffectedUEInformationForMBMS,
    &ett_rnsap_E_DCH_Semi_PersistentScheduling_Information_LCR,
    &ett_rnsap_E_DCH_Semi_PersistentScheduling_Information_to_Modify_LCR,
    &ett_rnsap_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR,
    &ett_rnsap_Initial_E_DCH_SPS_resource,
    &ett_rnsap_E_DCH_SPS_HICH_Information,
    &ett_rnsap_E_HICH_Configuration,
    &ett_rnsap_Same_As_Scheduled_E_HICH,
    &ett_rnsap_E_HICH_InformationResp_ExplicitConfiguration_LCR,
    &ett_rnsap_FACH_FlowControlInformation,
    &ett_rnsap_FACH_FlowControlInformationItem,
    &ett_rnsap_FACH_InformationList,
    &ett_rnsap_FACH_InformationItem,
    &ett_rnsap_FDD_DCHs_to_Modify,
    &ett_rnsap_FDD_DCHs_to_ModifyItem,
    &ett_rnsap_FDD_DCHs_to_ModifySpecificInformationList,
    &ett_rnsap_FDD_DCHs_to_ModifySpecificItem,
    &ett_rnsap_FDD_DL_CodeInformation,
    &ett_rnsap_FDD_DL_CodeInformationItem,
    &ett_rnsap_F_DPCH_Info,
    &ett_rnsap_F_DPCH_Info_To_Modify,
    &ett_rnsap_FPACH_Information,
    &ett_rnsap_FTPICH_Information,
    &ett_rnsap_FTPICH_Information_To_Modify,
    &ett_rnsap_FTPICH_Information_Reconf,
    &ett_rnsap_FTPICH_Reconfiguration_Information,
    &ett_rnsap_FTPICH_Information_Response,
    &ett_rnsap_GA_Cell,
    &ett_rnsap_GA_Cell_item,
    &ett_rnsap_GA_CellAdditionalShapes,
    &ett_rnsap_GA_AltitudeAndDirection,
    &ett_rnsap_GA_EllipsoidArc,
    &ett_rnsap_Gainfactors_10ms_mode,
    &ett_rnsap_T_signalledGainFactors10ms,
    &ett_rnsap_GANSS_AddClockModels,
    &ett_rnsap_GANSS_AddOrbitModels,
    &ett_rnsap_GANSS_Additional_Ionospheric_Model,
    &ett_rnsap_GANSS_Additional_Navigation_Models,
    &ett_rnsap_GANSS_Additional_Time_Models,
    &ett_rnsap_GANSS_Additional_UTC_Models,
    &ett_rnsap_GANSS_Almanac,
    &ett_rnsap_T_gANSS_AlmanacModel,
    &ett_rnsap_T_gANSS_keplerianParameters,
    &ett_rnsap_GANSS_alm_keplerianNAVAlmanac,
    &ett_rnsap_GANSS_alm_KEPLERIANBDSALMANAC,
    &ett_rnsap_GANSS_alm_keplerianReducedAlmanac,
    &ett_rnsap_GANSS_alm_keplerianMidiAlmanac,
    &ett_rnsap_GANSS_alm_keplerianGLONASS,
    &ett_rnsap_GANSS_alm_ecefSBASAlmanac,
    &ett_rnsap_GANSS_Auxiliary_Information,
    &ett_rnsap_GANSS_AuxInfoGANSS_ID1,
    &ett_rnsap_GANSS_AuxInfoGANSS_ID1_item,
    &ett_rnsap_GANSS_AuxInfoGANSS_ID3,
    &ett_rnsap_GANSS_AuxInfoGANSS_ID3_item,
    &ett_rnsap_GANSS_BDSclockModel,
    &ett_rnsap_GANSS_Clock_Model,
    &ett_rnsap_GANSS_Clock_Model_item,
    &ett_rnsap_GANSS_CNAVclockModel,
    &ett_rnsap_GANSS_Common_Data,
    &ett_rnsap_GANSS_CommonDataInfoReq,
    &ett_rnsap_GANSS_Data_Bit_Assistance,
    &ett_rnsap_GANSS_DataBitAssistanceList,
    &ett_rnsap_GANSS_DataBitAssistanceItem,
    &ett_rnsap_GANSS_DataBitAssistanceSgnList,
    &ett_rnsap_GANSS_DataBitAssistanceSgnItem,
    &ett_rnsap_GANSS_Data_Bit_Assistance_ReqItem,
    &ett_rnsap_GANSS_Data_Bit_Assistance_ReqList,
    &ett_rnsap_T_ganss_SatelliteInfo,
    &ett_rnsap_GANSS_DeltaUT1,
    &ett_rnsap_GANSS_Earth_Orientation_Parameters,
    &ett_rnsap_GANSS_GenericDataInfoReqList,
    &ett_rnsap_GANSS_GenericDataInfoReqItem,
    &ett_rnsap_GANSS_Generic_Data,
    &ett_rnsap_GANSS_Generic_DataItem,
    &ett_rnsap_GANSS_GLONASSclockModel,
    &ett_rnsap_GANSS_Information,
    &ett_rnsap_GANSS_Ionospheric_Model,
    &ett_rnsap_GANSS_IonosphereRegionalStormFlags,
    &ett_rnsap_GANSS_NAVclockModel,
    &ett_rnsap_GANSS_Navigation_Model_And_Time_Recovery,
    &ett_rnsap_GANSS_NavModel_BDSKeplerianSet,
    &ett_rnsap_GANSS_NavModel_CNAVKeplerianSet,
    &ett_rnsap_GANSS_NavModel_GLONASSecef,
    &ett_rnsap_GANSS_NavModel_NAVKeplerianSet,
    &ett_rnsap_GANSS_NavModel_SBASecef,
    &ett_rnsap_GANSS_Orbit_Model,
    &ett_rnsap_T_gANSS_keplerianParameters_01,
    &ett_rnsap_GANSS_Real_Time_Integrity,
    &ett_rnsap_GANSS_Real_Time_Integrity_item,
    &ett_rnsap_GANSS_RX_Pos,
    &ett_rnsap_GANSS_SatelliteInformationKP,
    &ett_rnsap_GANSS_SatelliteInformationKP_item,
    &ett_rnsap_Ganss_Sat_Info_AddNavList,
    &ett_rnsap_Ganss_Sat_Info_AddNavList_item,
    &ett_rnsap_GANSS_SAT_Info_Almanac_BDSkpList,
    &ett_rnsap_GANSS_SAT_Info_Almanac_BDS,
    &ett_rnsap_GANSS_SAT_Info_Almanac_GLOkpList,
    &ett_rnsap_GANSS_SAT_Info_Almanac_GLOkpList_item,
    &ett_rnsap_GANSS_SAT_Info_Almanac_MIDIkpList,
    &ett_rnsap_GANSS_SAT_Info_Almanac_MIDIkpList_item,
    &ett_rnsap_GANSS_SAT_Info_Almanac_NAVkpList,
    &ett_rnsap_GANSS_SAT_Info_Almanac_NAVkpList_item,
    &ett_rnsap_GANSS_SAT_Info_Almanac_REDkpList,
    &ett_rnsap_GANSS_SAT_Info_Almanac_REDkpList_item,
    &ett_rnsap_GANSS_SAT_Info_Almanac_SBASecefList,
    &ett_rnsap_GANSS_SAT_Info_Almanac_SBASecefList_item,
    &ett_rnsap_GANSS_Sat_Info_Nav,
    &ett_rnsap_GANSS_Sat_Info_Nav_item,
    &ett_rnsap_GANSS_SBASclockModel,
    &ett_rnsap_GANSS_Time_Model,
    &ett_rnsap_GANSS_Transmission_Time,
    &ett_rnsap_GANSS_UTC_Model,
    &ett_rnsap_GANSS_UTCmodelSet1,
    &ett_rnsap_GANSS_UTCmodelSet2,
    &ett_rnsap_GANSS_UTCmodelSet3,
    &ett_rnsap_GANSS_UTCmodelSet4,
    &ett_rnsap_GA_PointWithAltitude,
    &ett_rnsap_GA_PointWithAltitudeAndUncertaintyEllipsoid,
    &ett_rnsap_GA_PointWithUnCertaintyEllipse,
    &ett_rnsap_GA_UncertaintyEllipse,
    &ett_rnsap_GA_PointWithUnCertainty,
    &ett_rnsap_GA_AccessPointPosition,
    &ett_rnsap_GeographicalCoordinate,
    &ett_rnsap_GERAN_SI_Type,
    &ett_rnsap_GERAN_SystemInfo,
    &ett_rnsap_GERAN_SystemInfo_item,
    &ett_rnsap_GPS_Almanac,
    &ett_rnsap_T_satellite_Almanac_Information,
    &ett_rnsap_T_satellite_Almanac_Information_item,
    &ett_rnsap_Satellite_Almanac_Information_ExtItem,
    &ett_rnsap_Satellite_Almanac_Information_ExtItem_item,
    &ett_rnsap_GPSInformation,
    &ett_rnsap_GPSInformation_item,
    &ett_rnsap_GPS_Ionospheric_Model,
    &ett_rnsap_GPS_NavigationModel_and_TimeRecovery,
    &ett_rnsap_GPS_NavigationModel_and_TimeRecovery_item,
    &ett_rnsap_GPS_RealTime_Integrity,
    &ett_rnsap_GPS_RX_POS,
    &ett_rnsap_GPS_UTC_Model,
    &ett_rnsap_Guaranteed_Rate_Information,
    &ett_rnsap_HARQ_MemoryPartitioning,
    &ett_rnsap_HARQ_MemoryPartitioning_Implicit,
    &ett_rnsap_HARQ_MemoryPartitioning_Explicit,
    &ett_rnsap_HARQ_MemoryPartitioningList,
    &ett_rnsap_HARQ_MemoryPartitioningInfoExtForMIMO,
    &ett_rnsap_HARQ_MemoryPartitioningItem,
    &ett_rnsap_HSDSCH_FDD_Information,
    &ett_rnsap_HSDSCH_FDD_Information_Response,
    &ett_rnsap_HS_DSCH_FDD_Secondary_Serving_Information,
    &ett_rnsap_HS_DSCH_FDD_Secondary_Serving_Information_Response,
    &ett_rnsap_HS_DSCH_Secondary_Serving_Information_To_Modify,
    &ett_rnsap_HS_DSCH_FDD_Secondary_Serving_Information_To_Modify_Unsynchronised,
    &ett_rnsap_HS_DSCH_FDD_Secondary_Serving_Update_Information,
    &ett_rnsap_HS_DSCH_Secondary_Serving_Cell_Change_Information_Response,
    &ett_rnsap_HS_DSCH_Secondary_Serving_cell_change_choice,
    &ett_rnsap_HS_Secondary_Serving_cell_change_successful,
    &ett_rnsap_HS_Secondary_Serving_cell_change_unsuccessful,
    &ett_rnsap_HSDSCH_Information_to_Modify,
    &ett_rnsap_HSDSCH_Information_to_Modify_Unsynchronised,
    &ett_rnsap_HSDSCH_MACdFlow_Specific_InfoList,
    &ett_rnsap_HSDSCH_MACdFlow_Specific_InfoItem,
    &ett_rnsap_HSDSCH_MACdFlow_Specific_InfoList_Response,
    &ett_rnsap_HSDSCH_MACdFlow_Specific_InfoItem_Response,
    &ett_rnsap_HSDSCH_MACdFlow_Specific_InfoList_to_Modify,
    &ett_rnsap_HSDSCH_MACdFlow_Specific_InfoItem_to_Modify,
    &ett_rnsap_HSDSCH_MACdFlows_Information,
    &ett_rnsap_HSDSCH_MACdFlows_to_Delete,
    &ett_rnsap_HSDSCH_MACdFlows_to_Delete_Item,
    &ett_rnsap_HSDSCH_Initial_Capacity_Allocation,
    &ett_rnsap_HSDSCH_Initial_Capacity_AllocationItem,
    &ett_rnsap_HSDSCH_PreconfigurationInfo,
    &ett_rnsap_Additional_EDCH_Preconfiguration_Information,
    &ett_rnsap_Additional_EDCH_Preconfiguration_Information_ItemIEs,
    &ett_rnsap_HSDSCH_PreconfigurationSetup,
    &ett_rnsap_HS_SCCH_PreconfiguredCodes,
    &ett_rnsap_HS_SCCH_PreconfiguredCodesItem,
    &ett_rnsap_HS_DSCH_serving_cell_change_information,
    &ett_rnsap_HS_DSCH_serving_cell_change_informationResponse,
    &ett_rnsap_HS_DSCH_serving_cell_change_choice,
    &ett_rnsap_HS_serving_cell_change_successful,
    &ett_rnsap_HS_serving_cell_change_unsuccessful,
    &ett_rnsap_HSDSCH_TDD_Information,
    &ett_rnsap_HSDSCH_TDD_Information_Response,
    &ett_rnsap_HSPDSCH_TDD_Specific_InfoList_Response,
    &ett_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response,
    &ett_rnsap_HSPDSCH_TDD_Specific_InfoList_Response_LCR,
    &ett_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response_LCR,
    &ett_rnsap_HSPDSCH_TDD_Specific_InfoList_Response768,
    &ett_rnsap_HSPDSCH_TDD_Specific_InfoItem_Response768,
    &ett_rnsap_HSSCCH_FDD_Specific_InfoList_Response,
    &ett_rnsap_HSSCCH_FDD_Specific_InfoItem_Response,
    &ett_rnsap_HSSCCH_TDD_Specific_InfoList_Response,
    &ett_rnsap_HSSCCH_TDD_Specific_InfoItem_Response,
    &ett_rnsap_HSSCCH_TDD_Specific_InfoList_Response_LCR,
    &ett_rnsap_HSSCCH_TDD_Specific_InfoItem_Response_LCR,
    &ett_rnsap_HSSCCH_TDD_Specific_InfoList_Response768,
    &ett_rnsap_HSSCCH_TDD_Specific_InfoItem_Response768,
    &ett_rnsap_HSSICH_Info,
    &ett_rnsap_HSSICH_InfoLCR,
    &ett_rnsap_HSSICH_Info768,
    &ett_rnsap_HS_SICH_Reception_Quality_Value,
    &ett_rnsap_HSDSCH_FDD_Update_Information,
    &ett_rnsap_HSDSCH_TDD_Update_Information,
    &ett_rnsap_MIMO_ReferenceSignal_InformationListLCR,
    &ett_rnsap_HSSICH_ReferenceSignal_InformationLCR,
    &ett_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_LCR,
    &ett_rnsap_Transport_Block_Size_List_LCR,
    &ett_rnsap_Transport_Block_Size_Item_LCR,
    &ett_rnsap_Repetition_Period_List_LCR,
    &ett_rnsap_Repetition_Period_Item_LCR,
    &ett_rnsap_HS_DSCH_SPS_Operation_Indicator,
    &ett_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_to_Modify_LCR,
    &ett_rnsap_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR,
    &ett_rnsap_HS_SICH_InformationList_for_HS_DSCH_SPS,
    &ett_rnsap_HS_SICH_InformationItem_for_HS_DSCH_SPS,
    &ett_rnsap_HS_SICH_Type,
    &ett_rnsap_HS_SCCH_Associated_HS_SICH,
    &ett_rnsap_Non_HS_SCCH_Associated_HS_SICH,
    &ett_rnsap_Initial_HS_PDSCH_SPS_Resource,
    &ett_rnsap_InformationAvailable,
    &ett_rnsap_InformationReportCharacteristics,
    &ett_rnsap_InformationReportPeriodicity,
    &ett_rnsap_InformationThreshold,
    &ett_rnsap_InformationType,
    &ett_rnsap_IPDLParameters,
    &ett_rnsap_Inter_Frequency_Cell_List,
    &ett_rnsap_Inter_Frequency_Cell,
    &ett_rnsap_Inter_Frequency_Cell_Information,
    &ett_rnsap_Inter_Frequency_Cell_Information_SIB11,
    &ett_rnsap_Inter_Frequency_Cell_Information_SIB12,
    &ett_rnsap_Inter_Frequency_Cell_Information_SIB11ter,
    &ett_rnsap_Inter_Frequency_Cells_Information_SIB11_Per_Indication,
    &ett_rnsap_Inter_Frequency_Cells_Information_SIB12_Per_Indication,
    &ett_rnsap_Inter_Frequency_Cells_Information_SIB11ter_Per_Indication,
    &ett_rnsap_Inter_Frequency_Cell_SIB11ter_List,
    &ett_rnsap_Inter_Frequency_Cell_SIB11ter,
    &ett_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12_List,
    &ett_rnsap_Inter_Frequency_Cell_SIB11_or_SIB12,
    &ett_rnsap_IPDL_FDD_Parameters,
    &ett_rnsap_IPDL_TDD_Parameters,
    &ett_rnsap_IPDL_TDD_ParametersLCR,
    &ett_rnsap_IdleIntervalInformation,
    &ett_rnsap_LABased,
    &ett_rnsap_LAI_List,
    &ett_rnsap_LAI,
    &ett_rnsap_List_Of_PLMNs,
    &ett_rnsap_LoadValue,
    &ett_rnsap_LCRTDD_Uplink_Physical_Channel_Capability,
    &ett_rnsap_M1Report,
    &ett_rnsap_M2Report,
    &ett_rnsap_MDTAreaScope,
    &ett_rnsap_MDT_Configuration,
    &ett_rnsap_MDT_Report_Parameters,
    &ett_rnsap_MAC_c_sh_SDU_LengthList,
    &ett_rnsap_MACdPDU_Size_IndexList,
    &ett_rnsap_MACdPDU_Size_IndexItem,
    &ett_rnsap_MACdPDU_Size_IndexList_to_Modify,
    &ett_rnsap_MACdPDU_Size_IndexItem_to_Modify,
    &ett_rnsap_MBMS_Bearer_Service_Full_Address,
    &ett_rnsap_MBMS_Bearer_Service_List,
    &ett_rnsap_MBMS_Bearer_ServiceItemFDD,
    &ett_rnsap_MBMS_Bearer_ServiceItemFDD_PFL,
    &ett_rnsap_MBMS_Bearer_ServiceItemTDD,
    &ett_rnsap_MBMS_Bearer_ServiceItemTDD_PFL,
    &ett_rnsap_MBMSChannelTypeInfo,
    &ett_rnsap_MBMSChannelTypeCellList,
    &ett_rnsap_MBMSPreferredFreqLayerInfo,
    &ett_rnsap_MBMS_Neighbouring_Cell_Information,
    &ett_rnsap_MBMS_ConcatenatedServiceList,
    &ett_rnsap_MBMS_RLC_Sequence_Number_Information,
    &ett_rnsap_MBMS_RLC_Sequence_Number_Information_List,
    &ett_rnsap_MBMS_Bearer_Service_List_RLC,
    &ett_rnsap_MBMS_Bearer_Service_List_RLCinfo,
    &ett_rnsap_MCCH_Message_List,
    &ett_rnsap_MCCH_Configuration_Info,
    &ett_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Info_List,
    &ett_rnsap_MBSFN_Scheduling_Transmission_Time_Interval_Item,
    &ett_rnsap_MeasurementIncreaseDecreaseThreshold,
    &ett_rnsap_MeasurementThreshold,
    &ett_rnsap_MidambleShiftAndBurstType,
    &ett_rnsap_Type1,
    &ett_rnsap_MidambleAllocationMode1,
    &ett_rnsap_Type2,
    &ett_rnsap_MidambleAllocationMode2,
    &ett_rnsap_Type3,
    &ett_rnsap_MidambleAllocationMode3,
    &ett_rnsap_MidambleShiftLCR,
    &ett_rnsap_MidambleShiftAndBurstType768,
    &ett_rnsap_Type7681,
    &ett_rnsap_MidambleAllocationMode7681,
    &ett_rnsap_Type7682,
    &ett_rnsap_MidambleAllocationMode7682,
    &ett_rnsap_Type7683,
    &ett_rnsap_MidambleAllocationMode7683,
    &ett_rnsap_MIMO_InformationResponse,
    &ett_rnsap_MIMO_PilotConfiguration,
    &ett_rnsap_Additional_Associated_Secondary_CPICH,
    &ett_rnsap_Additional_Secondary_CPICH,
    &ett_rnsap_ModifyPriorityQueue,
    &ett_rnsap_MulticellEDCH_InformationItemIEs,
    &ett_rnsap_MulticellEDCH_RL_SpecificInformationItemIEs,
    &ett_rnsap_Multiple_PLMN_List,
    &ett_rnsap_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR,
    &ett_rnsap_MultipleFreq_HSPDSCH_InformationItem_ResponseTDDLCR,
    &ett_rnsap_Multi_Carrier_EDCH_Info,
    &ett_rnsap_Multi_Carrier_EDCH_Information,
    &ett_rnsap_Multi_Carrier_EDCH_LCR_InformationItem,
    &ett_rnsap_Multi_Carrier_EDCH_MACdFlows_Information_TDD,
    &ett_rnsap_Multi_Carrier_EDCH_MACdFlows_Specific_Info,
    &ett_rnsap_Multi_Carrier_EDCH_Reconfigure,
    &ett_rnsap_Continue_Setup_Change_Multi_Carrier_EDCH,
    &ett_rnsap_Multi_Carrier_EDCH_Change_Info,
    &ett_rnsap_Multi_Carrier_EDCH_Information_Removal_List,
    &ett_rnsap_Multi_Carrier_EDCH_Information_Removal_Info_ItemIEs,
    &ett_rnsap_Multi_Carrier_EDCH_Information_Response,
    &ett_rnsap_Multi_Carrier_EDCH_LCR_Information_ResponseItem,
    &ett_rnsap_Multiflow_Reconfiguration,
    &ett_rnsap_Multiflow_Information,
    &ett_rnsap_Multiflow_Information_To_Modify,
    &ett_rnsap_Multiflow_RepetitionFactors,
    &ett_rnsap_Multiflow_Timing,
    &ett_rnsap_MU_MIMO_InformationLCR,
    &ett_rnsap_Standalone_Midamble_Channel_Information,
    &ett_rnsap_MU_MIMO_Indicator,
    &ett_rnsap_NACC_Related_Data,
    &ett_rnsap_Neighbouring_UMTS_CellInformation,
    &ett_rnsap_Neighbouring_UMTS_CellInformation_Ext,
    &ett_rnsap_Neighbouring_UMTS_CellInformationExtensionItem,
    &ett_rnsap_Neighbouring_UMTS_CellInformationItem,
    &ett_rnsap_Neighbouring_FDD_CellInformation,
    &ett_rnsap_Neighbouring_FDD_CellInformationItem,
    &ett_rnsap_MultipleFrequencyBandsIndicatorList,
    &ett_rnsap_NeighbouringFDDCellMeasurementInformation,
    &ett_rnsap_Neighbouring_GSM_CellInformationIEs,
    &ett_rnsap_Neighbouring_GSM_CellInformationItem,
    &ett_rnsap_Neighbouring_TDD_CellInformation,
    &ett_rnsap_Neighbouring_TDD_CellInformationItem,
    &ett_rnsap_NeighbouringTDDCellMeasurementInformation,
    &ett_rnsap_NeighbouringTDDCellMeasurementInformationLCR,
    &ett_rnsap_NeighbouringTDDCellMeasurementInformation768,
    &ett_rnsap_Neighbouring_LCR_TDD_CellInformation,
    &ett_rnsap_Neighbouring_LCR_TDD_CellInformationItem,
    &ett_rnsap_Neighbouring_E_UTRA_CellInformation,
    &ett_rnsap_Neighbouring_E_UTRA_CellInformationItem,
    &ett_rnsap_MultipleFrequencyBandsIndicatorList_EUTRA,
    &ett_rnsap_NotProvidedCellList,
    &ett_rnsap_NRTLoadInformationValue,
    &ett_rnsap_Non_Serving_RL_Preconfig_Setup,
    &ett_rnsap_New_non_serving_RL_setup_selection,
    &ett_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList,
    &ett_rnsap_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_item,
    &ett_rnsap_OnModification,
    &ett_rnsap_OnModificationInformation,
    &ett_rnsap_PCH_InformationList,
    &ett_rnsap_PCH_InformationItem,
    &ett_rnsap_Periodic,
    &ett_rnsap_PeriodicInformation,
    &ett_rnsap_Permanent_NAS_UE_Identity,
    &ett_rnsap_PLMN_List,
    &ett_rnsap_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas,
    &ett_rnsap_Associated_Secondary_CPICH,
    &ett_rnsap_PreferredFrequencyLayerInfo,
    &ett_rnsap_PriorityQueue_InfoList,
    &ett_rnsap_PriorityQueue_InfoItem,
    &ett_rnsap_PriorityQueue_InfoList_EnhancedFACH_PCH,
    &ett_rnsap_PriorityQueue_InfoItem_EnhancedFACH_PCH,
    &ett_rnsap_PriorityQueue_InfoList_to_Modify,
    &ett_rnsap_PriorityQueue_InfoItem_to_Add,
    &ett_rnsap_PriorityQueue_InfoItem_to_Modify,
    &ett_rnsap_PriorityQueue_InfoList_to_Modify_Unsynchronised,
    &ett_rnsap_PriorityQueue_InfoItem_to_Modify_Unsynchronised,
    &ett_rnsap_ProvidedInformation,
    &ett_rnsap_UpPCH_InformationList_LCRTDD,
    &ett_rnsap_UpPCH_InformationItem_LCRTDD,
    &ett_rnsap_PTMCellList,
    &ett_rnsap_PTPCellList,
    &ett_rnsap_RABased,
    &ett_rnsap_RAI_List,
    &ett_rnsap_RAI,
    &ett_rnsap_RB_Info,
    &ett_rnsap_Reference_E_TFCI_Information,
    &ett_rnsap_Reference_E_TFCI_Information_Item,
    &ett_rnsap_Released_CN_Domain,
    &ett_rnsap_ReportCharacteristics,
    &ett_rnsap_ReportPeriodicity,
    &ett_rnsap_RequestedDataValue,
    &ett_rnsap_RequestedDataValueInformation,
    &ett_rnsap_RL_Specific_DCH_Info,
    &ett_rnsap_RL_Specific_DCH_Info_Item,
    &ett_rnsap_RL_Specific_EDCH_Information,
    &ett_rnsap_RL_Specific_EDCH_Info,
    &ett_rnsap_RL_Specific_EDCH_InfoItem,
    &ett_rnsap_RTLoadValue,
    &ett_rnsap_SAI,
    &ett_rnsap_SecondaryServingCells,
    &ett_rnsap_SecondaryServingCellsItem,
    &ett_rnsap_Secondary_CCPCH_Info_TDD,
    &ett_rnsap_Secondary_CPICH_Information,
    &ett_rnsap_Secondary_CPICH_Information_Change,
    &ett_rnsap_Secondary_LCR_CCPCH_Info_TDD,
    &ett_rnsap_Secondary_CCPCH_Info_TDD768,
    &ett_rnsap_Secondary_CCPCH_TDD_InformationList,
    &ett_rnsap_Secondary_CCPCH_TDD_InformationItem,
    &ett_rnsap_Secondary_LCR_CCPCH_TDD_InformationList,
    &ett_rnsap_Secondary_LCR_CCPCH_TDD_InformationItem,
    &ett_rnsap_Secondary_CCPCH_TDD_InformationList768,
    &ett_rnsap_Secondary_CCPCH_TDD_InformationItem768,
    &ett_rnsap_Secondary_CCPCH_TDD_Code_Information,
    &ett_rnsap_Secondary_CCPCH_TDD_Code_InformationItem,
    &ett_rnsap_Secondary_LCR_CCPCH_TDD_Code_Information,
    &ett_rnsap_Secondary_LCR_CCPCH_TDD_Code_InformationItem,
    &ett_rnsap_Secondary_CCPCH_TDD_Code_Information768,
    &ett_rnsap_Secondary_CCPCH_TDD_Code_InformationItem768,
    &ett_rnsap_Secondary_Serving_Cell_List,
    &ett_rnsap_Possible_Secondary_Serving_Cell_List,
    &ett_rnsap_Possible_Secondary_Serving_Cell,
    &ett_rnsap_SetsOfHS_SCCH_Codes,
    &ett_rnsap_SetsOfHS_SCCH_CodesItem,
    &ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_EDCH_On_secondary_UL_Frequency,
    &ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_Downlink_TPC_enhancements,
    &ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_CLTD,
    &ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_FTPICH_Information,
    &ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_DCH_ENH,
    &ett_rnsap_Setup_Or_ConfigurationChange_Or_Removal_Of_UL_DPCCH2,
    &ett_rnsap_GA_AccessPointPositionwithOptionalAltitude,
    &ett_rnsap_SFNSFNMeasurementThresholdInformation,
    &ett_rnsap_SFNSFNMeasurementValueInformation,
    &ett_rnsap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation,
    &ett_rnsap_T_successfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item,
    &ett_rnsap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation,
    &ett_rnsap_T_unsuccessfullNeighbouringCellSFNSFNObservedTimeDifferenceMeasurementInformation_item,
    &ett_rnsap_SFNSFNTimeStampInformation,
    &ett_rnsap_SFNSFNTimeStamp_TDD,
    &ett_rnsap_SFNSFNValue,
    &ett_rnsap_SNA_Information,
    &ett_rnsap_ListOfSNAs,
    &ett_rnsap_S_RNTI_Group,
    &ett_rnsap_SynchronisationConfiguration,
    &ett_rnsap_SYNC_UL_ProcParameters,
    &ett_rnsap_TDD_ChannelisationCodeLCR,
    &ett_rnsap_TDD_DCHs_to_Modify,
    &ett_rnsap_TDD_DCHs_to_ModifyItem,
    &ett_rnsap_TDD_DCHs_to_ModifySpecificInformationList,
    &ett_rnsap_TDD_DCHs_to_ModifySpecificItem,
    &ett_rnsap_TDD_DL_Code_Information,
    &ett_rnsap_TDD_DL_Code_InformationItem,
    &ett_rnsap_TDD_DL_Code_LCR_Information,
    &ett_rnsap_TDD_DL_Code_LCR_InformationItem,
    &ett_rnsap_TDD_DL_Code_Information768,
    &ett_rnsap_TDD_DL_Code_InformationItem768,
    &ett_rnsap_TDD_DL_DPCH_TimeSlotFormat_LCR,
    &ett_rnsap_TDD_DPCHOffset,
    &ett_rnsap_TDD_UL_Code_Information,
    &ett_rnsap_TDD_UL_Code_InformationItem,
    &ett_rnsap_TDD_UL_Code_LCR_Information,
    &ett_rnsap_TDD_UL_Code_LCR_InformationItem,
    &ett_rnsap_TDD_UL_Code_Information768,
    &ett_rnsap_TDD_UL_Code_InformationItem768,
    &ett_rnsap_TDD_UL_DPCH_TimeSlotFormat_LCR,
    &ett_rnsap_TMGI,
    &ett_rnsap_TnlQos,
    &ett_rnsap_Transmission_Gap_Pattern_Sequence_Information,
    &ett_rnsap_Transmission_Gap_Pattern_Sequence_Information_item,
    &ett_rnsap_Transmission_Gap_Pattern_Sequence_Status_List,
    &ett_rnsap_Transmission_Gap_Pattern_Sequence_Status_List_item,
    &ett_rnsap_Affected_HSDSCH_Serving_Cell_List,
    &ett_rnsap_Transmission_Mode_Information,
    &ett_rnsap_Transmission_Mode_Information_List,
    &ett_rnsap_TUTRANGANSS,
    &ett_rnsap_TUTRANGANSSMeasurementThresholdInformation,
    &ett_rnsap_TUTRANGANSSMeasurementValueInformation,
    &ett_rnsap_TUTRANGPS,
    &ett_rnsap_TUTRANGPSMeasurementThresholdInformation,
    &ett_rnsap_TUTRANGPSMeasurementValueInformation,
    &ett_rnsap_TransportFormatCombination_Beta,
    &ett_rnsap_T_signalledGainFactors,
    &ett_rnsap_TFCS,
    &ett_rnsap_T_tFCSvalues,
    &ett_rnsap_TFCS_TFCSList,
    &ett_rnsap_TFCS_TFCSList_item,
    &ett_rnsap_TFCS_CTFC,
    &ett_rnsap_TPC_slot_position,
    &ett_rnsap_TransportFormatSet,
    &ett_rnsap_TransportFormatSet_DynamicPartList,
    &ett_rnsap_TransportFormatSet_DynamicPartList_item,
    &ett_rnsap_TransportFormatSet_ModeDP,
    &ett_rnsap_TDD_TransportFormatSet_ModeDP,
    &ett_rnsap_TransmissionTimeIntervalInformation,
    &ett_rnsap_TransmissionTimeIntervalInformation_item,
    &ett_rnsap_TransportFormatSet_Semi_staticPart,
    &ett_rnsap_TransportFormatSet_ModeSSP,
    &ett_rnsap_TTI_Update_Indicator,
    &ett_rnsap_UE_AggregateMaximumBitRate,
    &ett_rnsap_UE_Capabilities_Info,
    &ett_rnsap_UE_RF_Band_CapabilityLCR,
    &ett_rnsap_Radio_Frequency_BandItem,
    &ett_rnsap_UEIdentity,
    &ett_rnsap_UEMeasurementReportCharacteristics,
    &ett_rnsap_UEMeasurementReportCharacteristicsEvent1h,
    &ett_rnsap_UEMeasurementReportCharacteristicsEvent1i,
    &ett_rnsap_UEMeasurementReportCharacteristicsEvent6a,
    &ett_rnsap_UEMeasurementReportCharacteristicsEvent6b,
    &ett_rnsap_UEMeasurementReportCharacteristicsEvent6c,
    &ett_rnsap_UEMeasurementReportCharacteristicsEvent6d,
    &ett_rnsap_UEMeasurementReportCharacteristicsPeriodic,
    &ett_rnsap_UEMeasurementThreshold,
    &ett_rnsap_UEMeasurementTimeslotInfoHCR,
    &ett_rnsap_UEMeasurementTimeslotInfoHCR_IEs,
    &ett_rnsap_UEMeasurementTimeslotInfoLCR,
    &ett_rnsap_UEMeasurementTimeslotInfoLCR_IEs,
    &ett_rnsap_UEMeasurementTimeslotInfo768,
    &ett_rnsap_UEMeasurementTimeslotInfo768_IEs,
    &ett_rnsap_UEMeasurementValue,
    &ett_rnsap_UE_MeasurementValue_UE_Transmitted_Power,
    &ett_rnsap_UEMeasurementValueTransmittedPowerListHCR,
    &ett_rnsap_UEMeasurementValueTransmittedPowerListHCR_IEs,
    &ett_rnsap_UEMeasurementValueTransmittedPowerListLCR,
    &ett_rnsap_UEMeasurementValueTransmittedPowerListLCR_IEs,
    &ett_rnsap_UEMeasurementValueTransmittedPowerList768,
    &ett_rnsap_UEMeasurementValueTransmittedPowerList768_IEs,
    &ett_rnsap_UE_MeasurementValue_Primary_CCPCH_RSCP,
    &ett_rnsap_UE_MeasurementValue_DL_Timeslot_ISCP,
    &ett_rnsap_UEMeasurementValueTimeslotISCPListHCR,
    &ett_rnsap_UEMeasurementValueTimeslotISCPListHCR_IEs,
    &ett_rnsap_UEMeasurementValueTimeslotISCPListLCR,
    &ett_rnsap_UEMeasurementValueTimeslotISCPListLCR_IEs,
    &ett_rnsap_UEMeasurementValueTimeslotISCPList768,
    &ett_rnsap_UEMeasurementValueTimeslotISCPList768_IEs,
    &ett_rnsap_UEMeasurementValueInformation,
    &ett_rnsap_UEMeasurementValueInformationAvailable,
    &ett_rnsap_UE_Measurement_Value,
    &ett_rnsap_UE_State,
    &ett_rnsap_Cell_Fach_Pch_State,
    &ett_rnsap_Ura_Pch_State,
    &ett_rnsap_UL_Timeslot_Information,
    &ett_rnsap_UL_Timeslot_InformationItem,
    &ett_rnsap_UL_TimeslotLCR_Information,
    &ett_rnsap_UL_TimeslotLCR_InformationItem,
    &ett_rnsap_PLCCHinformation,
    &ett_rnsap_UL_Timeslot_Information768,
    &ett_rnsap_UL_Timeslot_InformationItem768,
    &ett_rnsap_UL_TimeSlot_ISCP_Info,
    &ett_rnsap_UL_TimeSlot_ISCP_InfoItem,
    &ett_rnsap_UL_TimeSlot_ISCP_LCR_Info,
    &ett_rnsap_UL_TimeSlot_ISCP_LCR_InfoItem,
    &ett_rnsap_UL_TimingAdvanceCtrl_LCR,
    &ett_rnsap_UC_ID,
    &ett_rnsap_UL_ScramblingCode,
    &ett_rnsap_UL_Synchronisation_Parameters_LCR,
    &ett_rnsap_URA_Information,
    &ett_rnsap_RNCsWithCellsInTheAccessedURA_List,
    &ett_rnsap_RNCsWithCellsInTheAccessedURA_Item,
    &ett_rnsap_USCH_Information,
    &ett_rnsap_USCH_InformationItem,
    &ett_rnsap_UMTS_Cells_Info_List,
    &ett_rnsap_UMTS_Cell_Info,
    &ett_rnsap_UL_CLTD_Information,
    &ett_rnsap_UL_CLTD_Information_Reconf,
    &ett_rnsap_UL_CLTD_Information_To_Modify,
    &ett_rnsap_UL_DPCCH2_Information,
    &ett_rnsap_UL_DPCCH2_Information_Reconf,
    &ett_rnsap_UL_DPCCH2_Information_To_Modify,
    &ett_rnsap_UL_MIMO_Information,
    &ett_rnsap_UL_MIMO_Information_To_Modify,
    &ett_rnsap_UL_MIMO_Reconfiguration,
    &ett_rnsap_UL_MIMO_DL_Control_Channel_Information,
    &ett_rnsap_RadioLinkSetupRequestFDD,
    &ett_rnsap_UL_DPCH_Information_RL_SetupRqstFDD,
    &ett_rnsap_DL_DPCH_Information_RL_SetupRqstFDD,
    &ett_rnsap_PowerOffsetInformation_RL_SetupRqstFDD,
    &ett_rnsap_RL_InformationList_RL_SetupRqstFDD,
    &ett_rnsap_RL_InformationItem_RL_SetupRqstFDD,
    &ett_rnsap_Additional_HS_Cell_Information_RL_Setup_List,
    &ett_rnsap_Additional_HS_Cell_Information_RL_Setup_ItemIEs,
    &ett_rnsap_F_DPCH_Information_RL_SetupRqstFDD,
    &ett_rnsap_PowerOffsetInformation_F_DPCH_RL_SetupRqstFDD,
    &ett_rnsap_RadioLinkSetupRequestTDD,
    &ett_rnsap_UL_Physical_Channel_Information_RL_SetupRqstTDD,
    &ett_rnsap_DL_Physical_Channel_Information_RL_SetupRqstTDD,
    &ett_rnsap_UL_CCTrCH_InformationList_RL_SetupRqstTDD,
    &ett_rnsap_UL_CCTrCH_InformationItem_RL_SetupRqstTDD,
    &ett_rnsap_DL_CCTrCH_InformationList_RL_SetupRqstTDD,
    &ett_rnsap_DL_CCTrCH_InformationItem_RL_SetupRqstTDD,
    &ett_rnsap_CCTrCH_TPCList_RL_SetupRqstTDD,
    &ett_rnsap_CCTrCH_TPCItem_RL_SetupRqstTDD,
    &ett_rnsap_RL_Information_RL_SetupRqstTDD,
    &ett_rnsap_RadioLinkSetupResponseFDD,
    &ett_rnsap_RL_InformationResponseList_RL_SetupRspFDD,
    &ett_rnsap_RL_InformationResponseItem_RL_SetupRspFDD,
    &ett_rnsap_DiversityIndication_RL_SetupRspFDD,
    &ett_rnsap_Combining_RL_SetupRspFDD,
    &ett_rnsap_NonCombiningOrFirstRL_RL_SetupRspFDD,
    &ett_rnsap_Additional_HS_Cell_Information_Response_List,
    &ett_rnsap_Additional_HS_Cell_Information_Response_ItemIEs,
    &ett_rnsap_RadioLinkSetupResponseTDD,
    &ett_rnsap_RL_InformationResponse_RL_SetupRspTDD,
    &ett_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD,
    &ett_rnsap_UL_CCTrCHInformationItem_RL_SetupRspTDD,
    &ett_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD,
    &ett_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD,
    &ett_rnsap_DL_CCTrCHInformationItem_RL_SetupRspTDD,
    &ett_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD,
    &ett_rnsap_DSCH_InformationListIEs_RL_SetupRspTDD,
    &ett_rnsap_DSCHInformationItem_RL_SetupRspTDD,
    &ett_rnsap_USCH_InformationListIEs_RL_SetupRspTDD,
    &ett_rnsap_USCHInformationItem_RL_SetupRspTDD,
    &ett_rnsap_RL_LCR_InformationResponse_RL_SetupRspTDD,
    &ett_rnsap_UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD,
    &ett_rnsap_UL_LCR_CCTrCHInformationItem_RL_SetupRspTDD,
    &ett_rnsap_UL_DPCH_LCR_InformationItem_RL_SetupRspTDD,
    &ett_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD,
    &ett_rnsap_DL_CCTrCH_LCR_InformationItem_RL_SetupRspTDD,
    &ett_rnsap_DL_DPCH_LCR_InformationItem_RL_SetupRspTDD,
    &ett_rnsap_DSCH_LCR_InformationListIEs_RL_SetupRspTDD,
    &ett_rnsap_DSCH_LCR_InformationItem_RL_SetupRspTDD,
    &ett_rnsap_USCH_LCR_InformationListIEs_RL_SetupRspTDD,
    &ett_rnsap_USCH_LCR_InformationItem_RL_SetupRspTDD,
    &ett_rnsap_RL_InformationResponse_RL_SetupRspTDD768,
    &ett_rnsap_UL_CCTrCHInformationListIE_RL_SetupRspTDD768,
    &ett_rnsap_UL_CCTrCHInformationItem_RL_SetupRspTDD768,
    &ett_rnsap_UL_DPCH_InformationItem_RL_SetupRspTDD768,
    &ett_rnsap_DL_CCTrCHInformationListIE_RL_SetupRspTDD768,
    &ett_rnsap_DL_CCTrCHInformationItem_RL_SetupRspTDD768,
    &ett_rnsap_DL_DPCH_InformationItem_RL_SetupRspTDD768,
    &ett_rnsap_RadioLinkSetupFailureFDD,
    &ett_rnsap_CauseLevel_RL_SetupFailureFDD,
    &ett_rnsap_GeneralCauseList_RL_SetupFailureFDD,
    &ett_rnsap_RLSpecificCauseList_RL_SetupFailureFDD,
    &ett_rnsap_UnsuccessfulRL_InformationResponseList_RL_SetupFailureFDD,
    &ett_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD,
    &ett_rnsap_SuccessfulRL_InformationResponseList_RL_SetupFailureFDD,
    &ett_rnsap_SuccessfulRL_InformationResponse_RL_SetupFailureFDD,
    &ett_rnsap_DiversityIndication_RL_SetupFailureFDD,
    &ett_rnsap_Combining_RL_SetupFailureFDD,
    &ett_rnsap_NonCombiningOrFirstRL_RL_SetupFailureFDD,
    &ett_rnsap_RadioLinkSetupFailureTDD,
    &ett_rnsap_CauseLevel_RL_SetupFailureTDD,
    &ett_rnsap_GeneralCauseList_RL_SetupFailureTDD,
    &ett_rnsap_RLSpecificCauseList_RL_SetupFailureTDD,
    &ett_rnsap_UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD,
    &ett_rnsap_RadioLinkAdditionRequestFDD,
    &ett_rnsap_RL_InformationList_RL_AdditionRqstFDD,
    &ett_rnsap_RL_Information_RL_AdditionRqstFDD,
    &ett_rnsap_Additional_HS_Cell_Information_RL_Addition_List,
    &ett_rnsap_Additional_HS_Cell_Information_RL_Addition_ItemIEs,
    &ett_rnsap_Additional_EDCH_Cell_Information_RL_Add_Req,
    &ett_rnsap_Setup_Or_Addition_Of_EDCH_On_secondary_UL_Frequency,
    &ett_rnsap_EDPCH_Information_RLAdditionReq_FDD,
    &ett_rnsap_RadioLinkAdditionRequestTDD,
    &ett_rnsap_RL_Information_RL_AdditionRqstTDD,
    &ett_rnsap_UL_CCTrCH_InformationList_RL_AdditionRqstTDD,
    &ett_rnsap_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD,
    &ett_rnsap_DL_CCTrCH_InformationList_RL_AdditionRqstTDD,
    &ett_rnsap_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD,
    &ett_rnsap_RadioLinkAdditionResponseFDD,
    &ett_rnsap_RL_InformationResponseList_RL_AdditionRspFDD,
    &ett_rnsap_RL_InformationResponseItem_RL_AdditionRspFDD,
    &ett_rnsap_DiversityIndication_RL_AdditionRspFDD,
    &ett_rnsap_Combining_RL_AdditionRspFDD,
    &ett_rnsap_NonCombining_RL_AdditionRspFDD,
    &ett_rnsap_Additional_HS_Cell_Change_Information_Response_List,
    &ett_rnsap_Additional_HS_Cell_Change_Information_Response_ItemIEs,
    &ett_rnsap_RadioLinkAdditionResponseTDD,
    &ett_rnsap_RL_InformationResponse_RL_AdditionRspTDD,
    &ett_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD,
    &ett_rnsap_UL_CCTrCHInformationItem_RL_AdditionRspTDD,
    &ett_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD,
    &ett_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD,
    &ett_rnsap_DL_CCTrCHInformationItem_RL_AdditionRspTDD,
    &ett_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD,
    &ett_rnsap_DCH_Information_RL_AdditionRspTDD,
    &ett_rnsap_DiversityIndication_RL_AdditionRspTDD,
    &ett_rnsap_Combining_RL_AdditionRspTDD,
    &ett_rnsap_NonCombining_RL_AdditionRspTDD,
    &ett_rnsap_DSCH_InformationListIE_RL_AdditionRspTDD,
    &ett_rnsap_DSCHInformationItem_RL_AdditionRspTDD,
    &ett_rnsap_DiversityIndication_RL_AdditionRspTDD2,
    &ett_rnsap_USCH_InformationListIE_RL_AdditionRspTDD,
    &ett_rnsap_USCHInformationItem_RL_AdditionRspTDD,
    &ett_rnsap_RL_LCR_InformationResponse_RL_AdditionRspTDD,
    &ett_rnsap_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD,
    &ett_rnsap_UL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD,
    &ett_rnsap_UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD,
    &ett_rnsap_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD,
    &ett_rnsap_DL_CCTrCH_LCR_InformationItem_RL_AdditionRspTDD,
    &ett_rnsap_DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD,
    &ett_rnsap_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD,
    &ett_rnsap_DSCH_LCR_InformationItem_RL_AdditionRspTDD,
    &ett_rnsap_USCH_LCR_InformationListIEs_RL_AdditionRspTDD,
    &ett_rnsap_USCH_LCR_InformationItem_RL_AdditionRspTDD,
    &ett_rnsap_RL_InformationResponse_RL_AdditionRspTDD768,
    &ett_rnsap_UL_CCTrCHInformationListIE_RL_AdditionRspTDD768,
    &ett_rnsap_UL_CCTrCHInformationItem_RL_AdditionRspTDD768,
    &ett_rnsap_UL_DPCH_InformationItem_RL_AdditionRspTDD768,
    &ett_rnsap_DL_CCTrCHInformationListIE_RL_AdditionRspTDD768,
    &ett_rnsap_DL_CCTrCHInformationItem_RL_AdditionRspTDD768,
    &ett_rnsap_DL_DPCH_InformationItem_RL_AdditionRspTDD768,
    &ett_rnsap_RadioLinkAdditionFailureFDD,
    &ett_rnsap_CauseLevel_RL_AdditionFailureFDD,
    &ett_rnsap_GeneralCauseList_RL_AdditionFailureFDD,
    &ett_rnsap_RLSpecificCauseList_RL_AdditionFailureFDD,
    &ett_rnsap_UnsuccessfulRL_InformationResponseList_RL_AdditionFailureFDD,
    &ett_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD,
    &ett_rnsap_SuccessfulRL_InformationResponseList_RL_AdditionFailureFDD,
    &ett_rnsap_SuccessfulRL_InformationResponse_RL_AdditionFailureFDD,
    &ett_rnsap_DiversityIndication_RL_AdditionFailureFDD,
    &ett_rnsap_Combining_RL_AdditionFailureFDD,
    &ett_rnsap_NonCombining_RL_AdditionFailureFDD,
    &ett_rnsap_RadioLinkAdditionFailureTDD,
    &ett_rnsap_CauseLevel_RL_AdditionFailureTDD,
    &ett_rnsap_GeneralCauseList_RL_AdditionFailureTDD,
    &ett_rnsap_RLSpecificCauseList_RL_AdditionFailureTDD,
    &ett_rnsap_UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD,
    &ett_rnsap_RadioLinkDeletionRequest,
    &ett_rnsap_RL_InformationList_RL_DeletionRqst,
    &ett_rnsap_RL_Information_RL_DeletionRqst,
    &ett_rnsap_RadioLinkDeletionResponse,
    &ett_rnsap_RadioLinkReconfigurationPrepareFDD,
    &ett_rnsap_UL_DPCH_Information_RL_ReconfPrepFDD,
    &ett_rnsap_DL_DPCH_Information_RL_ReconfPrepFDD,
    &ett_rnsap_DL_DPCH_Power_Information_RL_ReconfPrepFDD,
    &ett_rnsap_PowerOffsetInformation_RL_ReconfPrepFDD,
    &ett_rnsap_DCH_DeleteList_RL_ReconfPrepFDD,
    &ett_rnsap_DCH_DeleteItem_RL_ReconfPrepFDD,
    &ett_rnsap_RL_InformationList_RL_ReconfPrepFDD,
    &ett_rnsap_RL_Information_RL_ReconfPrepFDD,
    &ett_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep,
    &ett_rnsap_Additional_HS_Cell_Information_RL_Reconf_Prep_ItemIEs,
    &ett_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Prep,
    &ett_rnsap_F_DPCH_Information_RL_ReconfPrepFDD,
    &ett_rnsap_PowerOffsetInformation_F_DPCH_RL_ReconfPrepFDD,
    &ett_rnsap_RadioLinkReconfigurationPrepareTDD,
    &ett_rnsap_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD,
    &ett_rnsap_UL_CCTrCH_AddInformation_RL_ReconfPrepTDD,
    &ett_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD,
    &ett_rnsap_UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD,
    &ett_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD,
    &ett_rnsap_UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD,
    &ett_rnsap_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD,
    &ett_rnsap_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD,
    &ett_rnsap_CCTrCH_TPCAddList_RL_ReconfPrepTDD,
    &ett_rnsap_CCTrCH_TPCAddItem_RL_ReconfPrepTDD,
    &ett_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD,
    &ett_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD,
    &ett_rnsap_CCTrCH_TPCModifyList_RL_ReconfPrepTDD,
    &ett_rnsap_CCTrCH_TPCModifyItem_RL_ReconfPrepTDD,
    &ett_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD,
    &ett_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD,
    &ett_rnsap_DCH_DeleteList_RL_ReconfPrepTDD,
    &ett_rnsap_DCH_DeleteItem_RL_ReconfPrepTDD,
    &ett_rnsap_DSCH_ModifyList_RL_ReconfPrepTDD,
    &ett_rnsap_DSCH_ModifyItem_RL_ReconfPrepTDD,
    &ett_rnsap_DSCH_DeleteList_RL_ReconfPrepTDD,
    &ett_rnsap_DSCH_DeleteItem_RL_ReconfPrepTDD,
    &ett_rnsap_USCH_ModifyList_RL_ReconfPrepTDD,
    &ett_rnsap_USCH_ModifyItem_RL_ReconfPrepTDD,
    &ett_rnsap_USCH_DeleteList_RL_ReconfPrepTDD,
    &ett_rnsap_USCH_DeleteItem_RL_ReconfPrepTDD,
    &ett_rnsap_RL_Information_RL_ReconfPrepTDD,
    &ett_rnsap_RL_InformationIE_RL_ReconfPrepTDD,
    &ett_rnsap_RadioLinkReconfigurationReadyFDD,
    &ett_rnsap_RL_InformationResponseList_RL_ReconfReadyFDD,
    &ett_rnsap_RL_InformationResponseItem_RL_ReconfReadyFDD,
    &ett_rnsap_Additional_HS_Cell_RL_Reconf_Response,
    &ett_rnsap_Additional_HS_Cell_RL_Reconf_Response_ItemIEs,
    &ett_rnsap_RadioLinkReconfigurationReadyTDD,
    &ett_rnsap_RL_InformationResponse_RL_ReconfReadyTDD,
    &ett_rnsap_UL_CCTrCHInformationListIE_RL_ReconfReadyTDD,
    &ett_rnsap_UL_CCTrCH_InformationItem_RL_ReconfReadyTDD,
    &ett_rnsap_UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD,
    &ett_rnsap_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD,
    &ett_rnsap_UL_DPCH_InformationAddList_RL_ReconfReadyTDD768,
    &ett_rnsap_UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD,
    &ett_rnsap_UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD,
    &ett_rnsap_UL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD,
    &ett_rnsap_TDD_UL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD,
    &ett_rnsap_TDD_UL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD,
    &ett_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD,
    &ett_rnsap_UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD,
    &ett_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD,
    &ett_rnsap_TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD,
    &ett_rnsap_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768,
    &ett_rnsap_UL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768,
    &ett_rnsap_TDD_UL_Code_InformationModifyList_RL_ReconfReadyTDD768,
    &ett_rnsap_TDD_UL_Code_InformationModifyItem_RL_ReconfReadyTDD768,
    &ett_rnsap_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD,
    &ett_rnsap_UL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD,
    &ett_rnsap_DL_CCTrCHInformationListIE_RL_ReconfReadyTDD,
    &ett_rnsap_DL_CCTrCH_InformationItem_RL_ReconfReadyTDD,
    &ett_rnsap_DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD,
    &ett_rnsap_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD,
    &ett_rnsap_DL_DPCH_InformationAddList_RL_ReconfReadyTDD768,
    &ett_rnsap_DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD,
    &ett_rnsap_DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD,
    &ett_rnsap_DL_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD,
    &ett_rnsap_TDD_DL_Code_LCR_InformationModifyList_RL_ReconfReadyTDD,
    &ett_rnsap_TDD_DL_Code_LCR_InformationModifyItem_RL_ReconfReadyTDD,
    &ett_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD,
    &ett_rnsap_DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD,
    &ett_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD,
    &ett_rnsap_TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD,
    &ett_rnsap_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768,
    &ett_rnsap_DL_Timeslot_InformationModifyItem_RL_ReconfReadyTDD768,
    &ett_rnsap_TDD_DL_Code_InformationModifyList_RL_ReconfReadyTDD768,
    &ett_rnsap_TDD_DL_Code_InformationModifyItem_RL_ReconfReadyTDD768,
    &ett_rnsap_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD,
    &ett_rnsap_DL_DPCH_InformationDeleteItem_RL_ReconfReadyTDD,
    &ett_rnsap_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD,
    &ett_rnsap_DL_DPCH_InformationDeleteItem768_RL_ReconfReadyTDD,
    &ett_rnsap_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD,
    &ett_rnsap_DSCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD,
    &ett_rnsap_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD,
    &ett_rnsap_USCHToBeAddedOrModifiedItem_RL_ReconfReadyTDD,
    &ett_rnsap_Multiple_RL_InformationResponse_RL_ReconfReadyTDD,
    &ett_rnsap_RadioLinkReconfigurationCommit,
    &ett_rnsap_RadioLinkReconfigurationFailure,
    &ett_rnsap_CauseLevel_RL_ReconfFailure,
    &ett_rnsap_GeneralCauseList_RL_ReconfFailure,
    &ett_rnsap_RLSpecificCauseList_RL_ReconfFailure,
    &ett_rnsap_RL_ReconfigurationFailureList_RL_ReconfFailure,
    &ett_rnsap_RL_ReconfigurationFailure_RL_ReconfFail,
    &ett_rnsap_RadioLinkReconfigurationCancel,
    &ett_rnsap_RadioLinkReconfigurationRequestFDD,
    &ett_rnsap_UL_DPCH_Information_RL_ReconfRqstFDD,
    &ett_rnsap_DL_DPCH_Information_RL_ReconfRqstFDD,
    &ett_rnsap_DCH_DeleteList_RL_ReconfRqstFDD,
    &ett_rnsap_DCH_DeleteItem_RL_ReconfRqstFDD,
    &ett_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req,
    &ett_rnsap_Additional_HS_Cell_Information_RL_Reconf_Req_ItemIEs,
    &ett_rnsap_Additional_EDCH_Cell_Information_RL_Reconf_Req,
    &ett_rnsap_RL_ReconfigurationRequestFDD_RL_InformationList,
    &ett_rnsap_RL_ReconfigurationRequestFDD_RL_Information_IEs,
    &ett_rnsap_RadioLinkReconfigurationRequestTDD,
    &ett_rnsap_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD,
    &ett_rnsap_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD,
    &ett_rnsap_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD,
    &ett_rnsap_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD,
    &ett_rnsap_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD,
    &ett_rnsap_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD,
    &ett_rnsap_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD,
    &ett_rnsap_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD,
    &ett_rnsap_DCH_DeleteList_RL_ReconfRqstTDD,
    &ett_rnsap_DCH_DeleteItem_RL_ReconfRqstTDD,
    &ett_rnsap_Multiple_RL_ReconfigurationRequestTDD_RL_Information,
    &ett_rnsap_RL_ReconfigurationRequestTDD_RL_Information,
    &ett_rnsap_RadioLinkReconfigurationResponseFDD,
    &ett_rnsap_RL_InformationResponseList_RL_ReconfRspFDD,
    &ett_rnsap_RL_InformationResponseItem_RL_ReconfRspFDD,
    &ett_rnsap_RadioLinkReconfigurationResponseTDD,
    &ett_rnsap_RL_InformationResponse_RL_ReconfRspTDD,
    &ett_rnsap_DL_CCTrCH_InformationList_RL_ReconfRspTDD,
    &ett_rnsap_DL_CCTrCH_InformationItem_RL_ReconfRspTDD,
    &ett_rnsap_DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD,
    &ett_rnsap_DL_Timeslot_LCR_InformationModifyList_RL_ReconfRspTDD,
    &ett_rnsap_DL_Timeslot_LCR_InformationModifyItem_RL_ReconfRspTDD,
    &ett_rnsap_Multiple_RL_InformationResponse_RL_ReconfRspTDD,
    &ett_rnsap_RadioLinkFailureIndication,
    &ett_rnsap_Reporting_Object_RL_FailureInd,
    &ett_rnsap_RL_RL_FailureInd,
    &ett_rnsap_RL_InformationList_RL_FailureInd,
    &ett_rnsap_RL_Information_RL_FailureInd,
    &ett_rnsap_RL_Set_RL_FailureInd,
    &ett_rnsap_RL_Set_InformationList_RL_FailureInd,
    &ett_rnsap_RL_Set_Information_RL_FailureInd,
    &ett_rnsap_CCTrCH_RL_FailureInd,
    &ett_rnsap_CCTrCH_InformationList_RL_FailureInd,
    &ett_rnsap_CCTrCH_InformationItem_RL_FailureInd,
    &ett_rnsap_RadioLinkPreemptionRequiredIndication,
    &ett_rnsap_RL_InformationList_RL_PreemptRequiredInd,
    &ett_rnsap_RL_InformationItem_RL_PreemptRequiredInd,
    &ett_rnsap_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd,
    &ett_rnsap_HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd,
    &ett_rnsap_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd,
    &ett_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd,
    &ett_rnsap_RadioLinkRestoreIndication,
    &ett_rnsap_Reporting_Object_RL_RestoreInd,
    &ett_rnsap_RL_RL_RestoreInd,
    &ett_rnsap_RL_InformationList_RL_RestoreInd,
    &ett_rnsap_RL_Information_RL_RestoreInd,
    &ett_rnsap_RL_Set_RL_RestoreInd,
    &ett_rnsap_RL_Set_InformationList_RL_RestoreInd,
    &ett_rnsap_RL_Set_Information_RL_RestoreInd,
    &ett_rnsap_CCTrCH_RL_RestoreInd,
    &ett_rnsap_CCTrCH_InformationList_RL_RestoreInd,
    &ett_rnsap_CCTrCH_InformationItem_RL_RestoreInd,
    &ett_rnsap_DL_PowerControlRequest,
    &ett_rnsap_DL_ReferencePowerInformationList_DL_PC_Rqst,
    &ett_rnsap_DL_ReferencePowerInformation_DL_PC_Rqst,
    &ett_rnsap_DL_PowerTimeslotControlRequest,
    &ett_rnsap_PhysicalChannelReconfigurationRequestFDD,
    &ett_rnsap_RL_Information_PhyChReconfRqstFDD,
    &ett_rnsap_PhysicalChannelReconfigurationRequestTDD,
    &ett_rnsap_RL_Information_PhyChReconfRqstTDD,
    &ett_rnsap_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD,
    &ett_rnsap_UL_CCTrCH_InformationItem_PhyChReconfRqstTDD,
    &ett_rnsap_UL_DPCH_InformationItem_PhyChReconfRqstTDD,
    &ett_rnsap_UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD,
    &ett_rnsap_UL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD,
    &ett_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD,
    &ett_rnsap_UL_Timeslot_InformationItem_PhyChReconfRqstTDD,
    &ett_rnsap_UL_Timeslot_InformationList_PhyChReconfRqstTDD768,
    &ett_rnsap_UL_Timeslot_InformationItem_PhyChReconfRqstTDD768,
    &ett_rnsap_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD,
    &ett_rnsap_DL_CCTrCH_InformationItem_PhyChReconfRqstTDD,
    &ett_rnsap_DL_DPCH_InformationItem_PhyChReconfRqstTDD,
    &ett_rnsap_DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD,
    &ett_rnsap_DL_TimeslotLCR_InformationItem_PhyChReconfRqstTDD,
    &ett_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD,
    &ett_rnsap_DL_Timeslot_InformationItem_PhyChReconfRqstTDD,
    &ett_rnsap_DL_Timeslot_InformationList_PhyChReconfRqstTDD768,
    &ett_rnsap_DL_Timeslot_InformationItem_PhyChReconfRqstTDD768,
    &ett_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD,
    &ett_rnsap_HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD,
    &ett_rnsap_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD,
    &ett_rnsap_HSPDSCH_Timeslot_InformationItemLCR_PhyChReconfRqstTDD,
    &ett_rnsap_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768,
    &ett_rnsap_HSPDSCH_Timeslot_InformationItem_PhyChReconfRqstTDD768,
    &ett_rnsap_PhysicalChannelReconfigurationCommand,
    &ett_rnsap_PhysicalChannelReconfigurationFailure,
    &ett_rnsap_RadioLinkCongestionIndication,
    &ett_rnsap_RL_InformationList_RL_CongestInd,
    &ett_rnsap_RL_InformationItem_RL_CongestInd,
    &ett_rnsap_DCH_Rate_Information_RL_CongestInd,
    &ett_rnsap_DCH_Rate_InformationItem_RL_CongestInd,
    &ett_rnsap_EDCH_MacdFlowSpecificInformationList_RL_CongestInd,
    &ett_rnsap_EDCH_MacdFlowSpecificInformationItem_RL_CongestInd,
    &ett_rnsap_UplinkSignallingTransferIndicationFDD,
    &ett_rnsap_UplinkSignallingTransferIndicationTDD,
    &ett_rnsap_DownlinkSignallingTransferRequest,
    &ett_rnsap_RelocationCommit,
    &ett_rnsap_PagingRequest,
    &ett_rnsap_PagingArea_PagingRqst,
    &ett_rnsap_URA_PagingRqst,
    &ett_rnsap_Cell_PagingRqst,
    &ett_rnsap_CNOriginatedPage_PagingRqst,
    &ett_rnsap_DedicatedMeasurementInitiationRequest,
    &ett_rnsap_DedicatedMeasurementObjectType_DM_Rqst,
    &ett_rnsap_RL_DM_Rqst,
    &ett_rnsap_RL_InformationList_DM_Rqst,
    &ett_rnsap_RL_InformationItem_DM_Rqst,
    &ett_rnsap_HSSICH_Info_DM_Rqst,
    &ett_rnsap_HSSICH_Info_DM_Rqst_Extension,
    &ett_rnsap_RL_Set_DM_Rqst,
    &ett_rnsap_RL_Set_InformationList_DM_Rqst,
    &ett_rnsap_RL_Set_InformationItem_DM_Rqst,
    &ett_rnsap_DedicatedMeasurementInitiationResponse,
    &ett_rnsap_DedicatedMeasurementObjectType_DM_Rsp,
    &ett_rnsap_RL_DM_Rsp,
    &ett_rnsap_RL_Set_DM_Rsp,
    &ett_rnsap_RL_InformationList_DM_Rsp,
    &ett_rnsap_RL_InformationItem_DM_Rsp,
    &ett_rnsap_RL_Set_InformationList_DM_Rsp,
    &ett_rnsap_RL_Set_InformationItem_DM_Rsp,
    &ett_rnsap_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp,
    &ett_rnsap_Multiple_DedicatedMeasurementValueItem_TDD_DM_Rsp,
    &ett_rnsap_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp,
    &ett_rnsap_Multiple_DedicatedMeasurementValueItem_LCR_TDD_DM_Rsp,
    &ett_rnsap_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp,
    &ett_rnsap_Multiple_HSSICHMeasurementValueItem_TDD_DM_Rsp,
    &ett_rnsap_Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp,
    &ett_rnsap_Multiple_DedicatedMeasurementValueItem_TDD768_DM_Rsp,
    &ett_rnsap_DedicatedMeasurementInitiationFailure,
    &ett_rnsap_DedicatedMeasurementObjectType_DM_Fail,
    &ett_rnsap_RL_DM_Fail,
    &ett_rnsap_RL_Set_DM_Fail,
    &ett_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail,
    &ett_rnsap_RL_Unsuccessful_InformationItem_DM_Fail,
    &ett_rnsap_RL_Successful_InformationRespList_DM_Fail,
    &ett_rnsap_RL_Successful_InformationItem_DM_Fail,
    &ett_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail,
    &ett_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail,
    &ett_rnsap_RL_Set_Successful_InformationRespList_DM_Fail,
    &ett_rnsap_RL_Set_Successful_InformationItem_DM_Fail,
    &ett_rnsap_DedicatedMeasurementReport,
    &ett_rnsap_DedicatedMeasurementObjectType_DM_Rprt,
    &ett_rnsap_RL_DM_Rprt,
    &ett_rnsap_RL_Set_DM_Rprt,
    &ett_rnsap_RL_InformationList_DM_Rprt,
    &ett_rnsap_RL_InformationItem_DM_Rprt,
    &ett_rnsap_RL_Set_InformationList_DM_Rprt,
    &ett_rnsap_RL_Set_InformationItem_DM_Rprt,
    &ett_rnsap_DedicatedMeasurementTerminationRequest,
    &ett_rnsap_DedicatedMeasurementFailureIndication,
    &ett_rnsap_DedicatedMeasurementObjectType_DM_Fail_Ind,
    &ett_rnsap_RL_DM_Fail_Ind,
    &ett_rnsap_RL_Set_DM_Fail_Ind,
    &ett_rnsap_RL_Unsuccessful_InformationRespList_DM_Fail_Ind,
    &ett_rnsap_RL_Unsuccessful_InformationItem_DM_Fail_Ind,
    &ett_rnsap_RL_Set_Unsuccessful_InformationRespList_DM_Fail_Ind,
    &ett_rnsap_RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind,
    &ett_rnsap_CommonTransportChannelResourcesReleaseRequest,
    &ett_rnsap_CommonTransportChannelResourcesRequest,
    &ett_rnsap_CommonTransportChannelResourcesResponseFDD,
    &ett_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD,
    &ett_rnsap_CommonTransportChannelResourcesResponseTDD,
    &ett_rnsap_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD,
    &ett_rnsap_CommonTransportChannelResourcesFailure,
    &ett_rnsap_CompressedModeCommand,
    &ett_rnsap_ErrorIndication,
    &ett_rnsap_CommonMeasurementInitiationRequest,
    &ett_rnsap_CommonMeasurementObjectType_CM_Rqst,
    &ett_rnsap_Cell_CM_Rqst,
    &ett_rnsap_NeighbouringCellMeasurementInfo,
    &ett_rnsap_NeighbouringCellMeasurementInfo_item,
    &ett_rnsap_GSM_Cell_CM_Rqst,
    &ett_rnsap_GsmCellList_CM_Rqst,
    &ett_rnsap_GsmCellItem_CM_Rqst,
    &ett_rnsap_CommonMeasurementInitiationResponse,
    &ett_rnsap_CommonMeasurementObjectType_CM_Rsp,
    &ett_rnsap_Cell_CM_Rsp,
    &ett_rnsap_GsmCellList_CM_Rsp,
    &ett_rnsap_GsmCellItem_CM_Rsp,
    &ett_rnsap_CommonMeasurementInitiationFailure,
    &ett_rnsap_Extension_FailureMeasurementList,
    &ett_rnsap_Extension_FailureMeasurementItem,
    &ett_rnsap_CommonMeasurementReport,
    &ett_rnsap_CommonMeasurementObjectType_CM_Rprt,
    &ett_rnsap_Cell_CM_Rprt,
    &ett_rnsap_GsmCellList_CM_Rprt,
    &ett_rnsap_GsmCellItem_CM_Rprt,
    &ett_rnsap_CommonMeasurementTerminationRequest,
    &ett_rnsap_Extension_TerminationMeasurementList,
    &ett_rnsap_Extension_TerminationMeasurementItem,
    &ett_rnsap_CommonMeasurementFailureIndication,
    &ett_rnsap_Extension_FailureIndicationMeasurementList,
    &ett_rnsap_Extension_FailureIndicationMeasurementItem,
    &ett_rnsap_InformationExchangeInitiationRequest,
    &ett_rnsap_InformationExchangeObjectType_InfEx_Rqst,
    &ett_rnsap_Cell_InfEx_Rqst,
    &ett_rnsap_GSM_Cell_InfEx_Rqst,
    &ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst,
    &ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rqst,
    &ett_rnsap_MBMS_Bearer_Service_List_InfEx_Rqst,
    &ett_rnsap_MBMS_Cell_InfEx_Rqst,
    &ett_rnsap_ANR_Cell_InfEx_Rqst,
    &ett_rnsap_Common_E_RGCH_Cell_InfEx_Rqst,
    &ett_rnsap_InformationExchangeInitiationResponse,
    &ett_rnsap_InformationExchangeObjectType_InfEx_Rsp,
    &ett_rnsap_Cell_InfEx_Rsp,
    &ett_rnsap_MBMS_Bearer_Service_List_InfEx_Rsp,
    &ett_rnsap_MBMS_Bearer_ServiceItemIEs_InfEx_Rsp,
    &ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp,
    &ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rsp,
    &ett_rnsap_MBMS_Cell_InfEx_Rsp,
    &ett_rnsap_MBMS_Cell_Item_InfEx_Rsp,
    &ett_rnsap_ANR_Cell_InfEx_Rsp,
    &ett_rnsap_ANR_Cell_ItemIEs_InfEx_Rsp,
    &ett_rnsap_Common_E_RGCH_Cell_InfEx_Rsp,
    &ett_rnsap_Common_E_RGCH_Cell_ItemIEs_InfEx_Rsp,
    &ett_rnsap_InformationExchangeInitiationFailure,
    &ett_rnsap_InformationReport,
    &ett_rnsap_InformationExchangeObjectType_InfEx_Rprt,
    &ett_rnsap_Cell_InfEx_Rprt,
    &ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt,
    &ett_rnsap_MBMS_Bearer_Service_in_MBMS_Cell_Item_InfEx_Rprt,
    &ett_rnsap_MBMS_Bearer_Service_List_InfEx_Rprt,
    &ett_rnsap_MBMS_Bearer_Service_List_Item_InfEx_Rprt,
    &ett_rnsap_MBMS_Cell_InfEx_Rprt,
    &ett_rnsap_MBMS_Cell_Item_InfEx_Rprt,
    &ett_rnsap_Common_E_RGCH_Cell_InfEx_Rprt,
    &ett_rnsap_Common_E_RGCH_Cell_Item_InfEx_Rprt,
    &ett_rnsap_InformationExchangeTerminationRequest,
    &ett_rnsap_InformationExchangeFailureIndication,
    &ett_rnsap_ResetRequest,
    &ett_rnsap_ResetIndicator,
    &ett_rnsap_ContextList_Reset,
    &ett_rnsap_ContextInfoList_Reset,
    &ett_rnsap_ContextInfoItem_Reset,
    &ett_rnsap_ContextType_Reset,
    &ett_rnsap_ContextGroupList_Reset,
    &ett_rnsap_ContextGroupInfoList_Reset,
    &ett_rnsap_ContextGroupInfoItem_Reset,
    &ett_rnsap_ResetResponse,
    &ett_rnsap_RadioLinkActivationCommandFDD,
    &ett_rnsap_DelayedActivationInformationList_RL_ActivationCmdFDD,
    &ett_rnsap_DelayedActivationInformation_RL_ActivationCmdFDD,
    &ett_rnsap_RadioLinkActivationCommandTDD,
    &ett_rnsap_DelayedActivationInformationList_RL_ActivationCmdTDD,
    &ett_rnsap_DelayedActivationInformation_RL_ActivationCmdTDD,
    &ett_rnsap_GERANUplinkSignallingTransferIndication,
    &ett_rnsap_RadioLinkParameterUpdateIndicationFDD,
    &ett_rnsap_RL_ParameterUpdateIndicationFDD_RL_InformationList,
    &ett_rnsap_RL_ParameterUpdateIndicationFDD_RL_Information_Item,
    &ett_rnsap_UE_Measurement_Forwarding,
    &ett_rnsap_Additional_HS_Cell_Information_RL_Param_Upd,
    &ett_rnsap_Additional_HS_Cell_Information_RL_Param_Upd_ItemIEs,
    &ett_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd,
    &ett_rnsap_Additional_EDCH_Cell_Information_RL_Param_Upd_ItemIEs,
    &ett_rnsap_RadioLinkParameterUpdateIndicationTDD,
    &ett_rnsap_UEMeasurementInitiationRequest,
    &ett_rnsap_UEMeasurementInitiationResponse,
    &ett_rnsap_UEMeasurementInitiationFailure,
    &ett_rnsap_UEMeasurementReport,
    &ett_rnsap_UEMeasurementTerminationRequest,
    &ett_rnsap_UEMeasurementFailureIndication,
    &ett_rnsap_IurInvokeTrace,
    &ett_rnsap_ListOfInterfacesToTrace,
    &ett_rnsap_InterfacesToTraceItem,
    &ett_rnsap_IurDeactivateTrace,
    &ett_rnsap_MBMSAttachCommand,
    &ett_rnsap_MBMSDetachCommand,
    &ett_rnsap_DirectInformationTransfer,
    &ett_rnsap_EnhancedRelocationRequest,
    &ett_rnsap_EnhancedRelocationResponse,
    &ett_rnsap_EnhancedRelocationFailure,
    &ett_rnsap_EnhancedRelocationCancel,
    &ett_rnsap_EnhancedRelocationSignallingTransfer,
    &ett_rnsap_EnhancedRelocationRelease,
    &ett_rnsap_MBSFNMCCHInformation,
    &ett_rnsap_SecondaryULFrequencyReport,
    &ett_rnsap_SecondaryULFrequencyUpdateIndication,
    &ett_rnsap_EnhancedRelocationResourceRequest,
    &ett_rnsap_EnhancedRelocationResourceResponse,
    &ett_rnsap_EnhancedRelocationResourceFailure,
    &ett_rnsap_EnhancedRelocationResourceReleaseCommand,
    &ett_rnsap_EnhancedRelocationResourceReleaseComplete,
    &ett_rnsap_PrivateMessage,
    &ett_rnsap_InformationTransferControlRequest,
    &ett_rnsap_Control_Type_InformationTransferControlReq,
    &ett_rnsap_Suspension_Control_Type,
    &ett_rnsap_Resume_Control_Type,
    &ett_rnsap_RNSAP_PDU,
    &ett_rnsap_InitiatingMessage,
    &ett_rnsap_SuccessfulOutcome,
    &ett_rnsap_UnsuccessfulOutcome,
    &ett_rnsap_Outcome,
  };


  /* Register protocol */
  proto_rnsap = proto_register_protocol(PNAME, PSNAME, PFNAME);
  /* Register fields and subtrees */
  proto_register_field_array(proto_rnsap, hf, array_length(hf));
  proto_register_subtree_array(ett, array_length(ett));

  /* Register dissector */
  rnsap_handle = register_dissector("rnsap", dissect_rnsap, proto_rnsap);

  /* Register dissector tables */
  rnsap_ies_dissector_table = register_dissector_table("rnsap.ies", "RNSAP-PROTOCOL-IES", proto_rnsap, FT_UINT32, BASE_DEC);
  rnsap_extension_dissector_table = register_dissector_table("rnsap.extension", "RNSAP-PROTOCOL-EXTENSION", proto_rnsap, FT_UINT32, BASE_DEC);
  rnsap_proc_imsg_dissector_table = register_dissector_table("rnsap.proc.imsg", "RNSAP-ELEMENTARY-PROCEDURE InitiatingMessage", proto_rnsap, FT_STRING, STRING_CASE_SENSITIVE);
  rnsap_proc_sout_dissector_table = register_dissector_table("rnsap.proc.sout", "RNSAP-ELEMENTARY-PROCEDURE SuccessfulOutcome", proto_rnsap, FT_STRING, STRING_CASE_SENSITIVE);
  rnsap_proc_uout_dissector_table = register_dissector_table("rnsap.proc.uout", "RNSAP-ELEMENTARY-PROCEDURE UnsuccessfulOutcome", proto_rnsap, FT_STRING, STRING_CASE_SENSITIVE);

}


/*--- proto_reg_handoff_rnsap ---------------------------------------*/
void
proto_reg_handoff_rnsap(void)
{
	ranap_handle = find_dissector("ranap");
	rrc_dl_ccch_handle = find_dissector_add_dependency("rrc.dl.ccch", proto_rnsap);
	rrc_ul_ccch_handle = find_dissector_add_dependency("rrc.ul.ccch", proto_rnsap);

	dissector_add_uint("sccp.ssn", SCCP_SSN_RNSAP, rnsap_handle);
	heur_dissector_add("sccp", dissect_sccp_rnsap_heur, "RNSAP over SCCP", "rnsap_sccp", proto_rnsap, HEURISTIC_ENABLE);

  dissector_add_uint("rnsap.ies", id_RTLoadValue, create_dissector_handle(dissect_RTLoadValue_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_NRTLoadInformationValue, create_dissector_handle(dissect_NRTLoadInformationValue_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UpPTSInterferenceValue, create_dissector_handle(dissect_UpPTSInterferenceValue_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_TUTRANGANSSMeasurementValueInformation, create_dissector_handle(dissect_TUTRANGANSSMeasurementValueInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Rx_Timing_Deviation_Value_LCR, create_dissector_handle(dissect_Rx_Timing_Deviation_Value_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Angle_Of_Arrival_Value_LCR, create_dissector_handle(dissect_Angle_Of_Arrival_Value_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_HS_SICH_Reception_Quality, create_dissector_handle(dissect_HS_SICH_Reception_Quality_Value_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Rx_Timing_Deviation_Value_768, create_dissector_handle(dissect_Rx_Timing_Deviation_Value_768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Rx_Timing_Deviation_Value_ext, create_dissector_handle(dissect_Rx_Timing_Deviation_Value_ext_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Extended_Round_Trip_Time_Value, create_dissector_handle(dissect_Extended_Round_Trip_Time_Value_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_IPDL_TDD_ParametersLCR, create_dissector_handle(dissect_IPDL_TDD_ParametersLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Load_Value_IncrDecrThres, create_dissector_handle(dissect_Load_Value_IncrDecrThres_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Transmitted_Carrier_Power_Value_IncrDecrThres, create_dissector_handle(dissect_Transmitted_Carrier_Power_Value_IncrDecrThres_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Received_Total_Wideband_Power_Value_IncrDecrThres, create_dissector_handle(dissect_Received_Total_Wideband_Power_Value_IncrDecrThres_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_Timeslot_ISCP_Value_IncrDecrThres, create_dissector_handle(dissect_UL_Timeslot_ISCP_Value_IncrDecrThres_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RT_Load_Value_IncrDecrThres, create_dissector_handle(dissect_RT_Load_Value_IncrDecrThres_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_NRT_Load_Information_Value_IncrDecrThres, create_dissector_handle(dissect_NRT_Load_Information_Value_IncrDecrThres_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_TUTRANGPSMeasurementThresholdInformation, create_dissector_handle(dissect_TUTRANGPSMeasurementThresholdInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_SFNSFNMeasurementThresholdInformation, create_dissector_handle(dissect_SFNSFNMeasurementThresholdInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Load_Value, create_dissector_handle(dissect_Load_Value_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Transmitted_Carrier_Power_Value, create_dissector_handle(dissect_Transmitted_Carrier_Power_Value_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Received_Total_Wideband_Power_Value, create_dissector_handle(dissect_Received_Total_Wideband_Power_Value_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_Timeslot_ISCP_Value, create_dissector_handle(dissect_UL_Timeslot_ISCP_Value_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RT_Load_Value, create_dissector_handle(dissect_RT_Load_Value_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_NRT_Load_Information_Value, create_dissector_handle(dissect_NRT_Load_Information_Value_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_HS_SICH_Reception_Quality_Measurement_Value, create_dissector_handle(dissect_HS_SICH_Reception_Quality_Measurement_Value_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_TUTRANGANSSMeasurementThresholdInformation, create_dissector_handle(dissect_TUTRANGANSSMeasurementThresholdInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Neighbouring_UMTS_CellInformationItem, create_dissector_handle(dissect_Neighbouring_UMTS_CellInformationItem_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Neighbouring_GSM_CellInformation, create_dissector_handle(dissect_Neighbouring_GSM_CellInformationIEs_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_OnModification, create_dissector_handle(dissect_OnModification_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_SRNC_ID, create_dissector_handle(dissect_RNC_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_S_RNTI, create_dissector_handle(dissect_S_RNTI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_D_RNTI, create_dissector_handle(dissect_D_RNTI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_AllowedQueuingTime, create_dissector_handle(dissect_AllowedQueuingTime_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_DPCH_Information_RL_SetupRqstFDD, create_dissector_handle(dissect_UL_DPCH_Information_RL_SetupRqstFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_Information_RL_SetupRqstFDD, create_dissector_handle(dissect_DL_DPCH_Information_RL_SetupRqstFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DCH_FDD_Information, create_dissector_handle(dissect_DCH_FDD_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Information_RL_SetupRqstFDD, create_dissector_handle(dissect_RL_InformationList_RL_SetupRqstFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Transmission_Gap_Pattern_Sequence_Information, create_dissector_handle(dissect_Transmission_Gap_Pattern_Sequence_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Active_Pattern_Sequence_Information, create_dissector_handle(dissect_Active_Pattern_Sequence_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationItem_RL_SetupRqstFDD, create_dissector_handle(dissect_RL_InformationItem_RL_SetupRqstFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_Physical_Channel_Information_RL_SetupRqstTDD, create_dissector_handle(dissect_UL_Physical_Channel_Information_RL_SetupRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_Physical_Channel_Information_RL_SetupRqstTDD, create_dissector_handle(dissect_DL_Physical_Channel_Information_RL_SetupRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationList_RL_SetupRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationList_RL_SetupRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationList_RL_SetupRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DCH_TDD_Information, create_dissector_handle(dissect_DCH_TDD_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DSCH_TDD_Information, create_dissector_handle(dissect_DSCH_TDD_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_USCH_Information, create_dissector_handle(dissect_USCH_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Information_RL_SetupRqstTDD, create_dissector_handle(dissect_RL_Information_RL_SetupRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationItem_RL_SetupRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationItem_RL_SetupRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationItem_RL_SetupRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CN_PS_DomainIdentifier, create_dissector_handle(dissect_CN_PS_DomainIdentifier_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CN_CS_DomainIdentifier, create_dissector_handle(dissect_CN_CS_DomainIdentifier_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationResponseList_RL_SetupRspFDD, create_dissector_handle(dissect_RL_InformationResponseList_RL_SetupRspFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_SIRTarget, create_dissector_handle(dissect_UL_SIR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CriticalityDiagnostics, create_dissector_handle(dissect_CriticalityDiagnostics_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationResponseItem_RL_SetupRspFDD, create_dissector_handle(dissect_RL_InformationResponseItem_RL_SetupRspFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationResponse_RL_SetupRspTDD, create_dissector_handle(dissect_RL_InformationResponse_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationListIE_RL_SetupRspTDD, create_dissector_handle(dissect_UL_CCTrCHInformationListIE_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_DPCH_InformationItem_RL_SetupRspTDD, create_dissector_handle(dissect_UL_DPCH_InformationItem_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationListIE_RL_SetupRspTDD, create_dissector_handle(dissect_DL_CCTrCHInformationListIE_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_InformationItem_RL_SetupRspTDD, create_dissector_handle(dissect_DL_DPCH_InformationItem_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DCH_InformationResponse, create_dissector_handle(dissect_DCH_InformationResponse_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DSCH_InformationListIEs_RL_SetupRspTDD, create_dissector_handle(dissect_DSCH_InformationListIEs_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_USCH_InformationListIEs_RL_SetupRspTDD, create_dissector_handle(dissect_USCH_InformationListIEs_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD, create_dissector_handle(dissect_UL_LCR_CCTrCHInformationListIE_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_DPCH_LCR_InformationItem_RL_SetupRspTDD, create_dissector_handle(dissect_UL_DPCH_LCR_InformationItem_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD, create_dissector_handle(dissect_DL_CCTrCH_LCR_InformationListIE_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_LCR_InformationItem_RL_SetupRspTDD, create_dissector_handle(dissect_DL_DPCH_LCR_InformationItem_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DSCH_LCR_InformationListIEs_RL_SetupRspTDD, create_dissector_handle(dissect_DSCH_LCR_InformationListIEs_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_USCH_LCR_InformationListIEs_RL_SetupRspTDD, create_dissector_handle(dissect_USCH_LCR_InformationListIEs_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationListIE_RL_SetupRspTDD768, create_dissector_handle(dissect_UL_CCTrCHInformationListIE_RL_SetupRspTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_DPCH_InformationItem_RL_SetupRspTDD768, create_dissector_handle(dissect_UL_DPCH_InformationItem_RL_SetupRspTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationListIE_RL_SetupRspTDD768, create_dissector_handle(dissect_DL_CCTrCHInformationListIE_RL_SetupRspTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_InformationItem_RL_SetupRspTDD768, create_dissector_handle(dissect_DL_DPCH_InformationItem_RL_SetupRspTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CauseLevel_RL_SetupFailureFDD, create_dissector_handle(dissect_CauseLevel_RL_SetupFailureFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD, create_dissector_handle(dissect_UnsuccessfulRL_InformationResponse_RL_SetupFailureFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_SuccessfulRL_InformationResponse_RL_SetupFailureFDD, create_dissector_handle(dissect_SuccessfulRL_InformationResponse_RL_SetupFailureFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CauseLevel_RL_SetupFailureTDD, create_dissector_handle(dissect_CauseLevel_RL_SetupFailureTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD, create_dissector_handle(dissect_UnsuccessfulRL_InformationResponse_RL_SetupFailureTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationList_RL_AdditionRqstFDD, create_dissector_handle(dissect_RL_InformationList_RL_AdditionRqstFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Information_RL_AdditionRqstFDD, create_dissector_handle(dissect_RL_Information_RL_AdditionRqstFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Information_RL_AdditionRqstTDD, create_dissector_handle(dissect_RL_Information_RL_AdditionRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationItem_RL_AdditionRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationItem_RL_AdditionRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationResponseList_RL_AdditionRspFDD, create_dissector_handle(dissect_RL_InformationResponseList_RL_AdditionRspFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationResponseItem_RL_AdditionRspFDD, create_dissector_handle(dissect_RL_InformationResponseItem_RL_AdditionRspFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_FDD_DL_CodeInformation, create_dissector_handle(dissect_FDD_DL_CodeInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationResponse_RL_AdditionRspTDD, create_dissector_handle(dissect_RL_InformationResponse_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationListIE_RL_AdditionRspTDD, create_dissector_handle(dissect_UL_CCTrCHInformationListIE_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_DPCH_InformationItem_RL_AdditionRspTDD, create_dissector_handle(dissect_UL_DPCH_InformationItem_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationListIE_RL_AdditionRspTDD, create_dissector_handle(dissect_DL_CCTrCHInformationListIE_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_InformationItem_RL_AdditionRspTDD, create_dissector_handle(dissect_DL_DPCH_InformationItem_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DSCH_InformationListIE_RL_AdditionRspTDD, create_dissector_handle(dissect_DSCH_InformationListIE_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_USCH_InformationListIE_RL_AdditionRspTDD, create_dissector_handle(dissect_USCH_InformationListIE_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD, create_dissector_handle(dissect_UL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD, create_dissector_handle(dissect_UL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD, create_dissector_handle(dissect_DL_CCTrCH_LCR_InformationListIE_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD, create_dissector_handle(dissect_DL_DPCH_LCR_InformationItem_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD, create_dissector_handle(dissect_DSCH_LCR_InformationListIEs_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_USCH_LCR_InformationListIEs_RL_AdditionRspTDD, create_dissector_handle(dissect_USCH_LCR_InformationListIEs_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationListIE_RL_AdditionRspTDD768, create_dissector_handle(dissect_UL_CCTrCHInformationListIE_RL_AdditionRspTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_DPCH_InformationItem_RL_AdditionRspTDD768, create_dissector_handle(dissect_UL_DPCH_InformationItem_RL_AdditionRspTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationListIE_RL_AdditionRspTDD768, create_dissector_handle(dissect_DL_CCTrCHInformationListIE_RL_AdditionRspTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_InformationItem_RL_AdditionRspTDD768, create_dissector_handle(dissect_DL_DPCH_InformationItem_RL_AdditionRspTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CauseLevel_RL_AdditionFailureFDD, create_dissector_handle(dissect_CauseLevel_RL_AdditionFailureFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD, create_dissector_handle(dissect_UnsuccessfulRL_InformationResponse_RL_AdditionFailureFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_SuccessfulRL_InformationResponse_RL_AdditionFailureFDD, create_dissector_handle(dissect_SuccessfulRL_InformationResponse_RL_AdditionFailureFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CauseLevel_RL_AdditionFailureTDD, create_dissector_handle(dissect_CauseLevel_RL_AdditionFailureTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD, create_dissector_handle(dissect_UnsuccessfulRL_InformationResponse_RL_AdditionFailureTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationList_RL_DeletionRqst, create_dissector_handle(dissect_RL_InformationList_RL_DeletionRqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Information_RL_DeletionRqst, create_dissector_handle(dissect_RL_Information_RL_DeletionRqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_DPCH_Information_RL_ReconfPrepFDD, create_dissector_handle(dissect_UL_DPCH_Information_RL_ReconfPrepFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_Information_RL_ReconfPrepFDD, create_dissector_handle(dissect_DL_DPCH_Information_RL_ReconfPrepFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_FDD_DCHs_to_Modify, create_dissector_handle(dissect_FDD_DCHs_to_Modify_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DCHs_to_Add_FDD, create_dissector_handle(dissect_DCH_FDD_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DCH_DeleteList_RL_ReconfPrepFDD, create_dissector_handle(dissect_DCH_DeleteList_RL_ReconfPrepFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationList_RL_ReconfPrepFDD, create_dissector_handle(dissect_RL_InformationList_RL_ReconfPrepFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Information_RL_ReconfPrepFDD, create_dissector_handle(dissect_RL_Information_RL_ReconfPrepFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationAddList_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationModifyList_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationDeleteList_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_TDD_DCHs_to_Modify, create_dissector_handle(dissect_TDD_DCHs_to_Modify_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DCHs_to_Add_TDD, create_dissector_handle(dissect_DCH_TDD_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DCH_DeleteList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DCH_DeleteList_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DSCH_ModifyList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DSCH_ModifyList_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DSCHs_to_Add_TDD, create_dissector_handle(dissect_DSCH_TDD_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DSCH_DeleteList_RL_ReconfPrepTDD, create_dissector_handle(dissect_DSCH_DeleteList_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_USCH_ModifyList_RL_ReconfPrepTDD, create_dissector_handle(dissect_USCH_ModifyList_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_USCHs_to_Add, create_dissector_handle(dissect_USCH_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_USCH_DeleteList_RL_ReconfPrepTDD, create_dissector_handle(dissect_USCH_DeleteList_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_AddInformation_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_CCTrCH_AddInformation_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_CCTrCH_ModifyInformation_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD, create_dissector_handle(dissect_UL_CCTrCH_DeleteInformation_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationAddItem_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationModifyItem_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationDeleteItem_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationResponseList_RL_ReconfReadyFDD, create_dissector_handle(dissect_RL_InformationResponseList_RL_ReconfReadyFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationResponseItem_RL_ReconfReadyFDD, create_dissector_handle(dissect_RL_InformationResponseItem_RL_ReconfReadyFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationListIE_RL_ReconfReadyTDD, create_dissector_handle(dissect_UL_CCTrCHInformationListIE_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD, create_dissector_handle(dissect_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD, create_dissector_handle(dissect_UL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD, create_dissector_handle(dissect_UL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationListIE_RL_ReconfReadyTDD, create_dissector_handle(dissect_DL_CCTrCHInformationListIE_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD, create_dissector_handle(dissect_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD, create_dissector_handle(dissect_DL_DPCH_InformationModifyListIE_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD, create_dissector_handle(dissect_DL_DPCH_InformationDeleteListIE_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD, create_dissector_handle(dissect_DSCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD, create_dissector_handle(dissect_USCHToBeAddedOrModifiedList_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CFN, create_dissector_handle(dissect_CFN_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CauseLevel_RL_ReconfFailure, create_dissector_handle(dissect_CauseLevel_RL_ReconfFailure_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_ReconfigurationFailure_RL_ReconfFail, create_dissector_handle(dissect_RL_ReconfigurationFailure_RL_ReconfFail_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_DPCH_Information_RL_ReconfRqstFDD, create_dissector_handle(dissect_UL_DPCH_Information_RL_ReconfRqstFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_Information_RL_ReconfRqstFDD, create_dissector_handle(dissect_DL_DPCH_Information_RL_ReconfRqstFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DCH_DeleteList_RL_ReconfRqstFDD, create_dissector_handle(dissect_DCH_DeleteList_RL_ReconfRqstFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_ReconfigurationRequestFDD_RL_Information_IEs, create_dissector_handle(dissect_RL_ReconfigurationRequestFDD_RL_Information_IEs_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationModifyList_RL_ReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationDeleteList_RL_ReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DCH_DeleteList_RL_ReconfRqstTDD, create_dissector_handle(dissect_DCH_DeleteList_RL_ReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationModifyItem_RL_ReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationDeleteItem_RL_ReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationResponseList_RL_ReconfRspFDD, create_dissector_handle(dissect_RL_InformationResponseList_RL_ReconfRspFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationResponseItem_RL_ReconfRspFDD, create_dissector_handle(dissect_RL_InformationResponseItem_RL_ReconfRspFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationResponse_RL_ReconfRspTDD, create_dissector_handle(dissect_RL_InformationResponse_RL_ReconfRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD, create_dissector_handle(dissect_DL_DPCH_InformationModifyItem_LCR_RL_ReconfRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Reporting_Object_RL_FailureInd, create_dissector_handle(dissect_Reporting_Object_RL_FailureInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Information_RL_FailureInd, create_dissector_handle(dissect_RL_Information_RL_FailureInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Set_Information_RL_FailureInd, create_dissector_handle(dissect_RL_Set_Information_RL_FailureInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CCTrCH_InformationItem_RL_FailureInd, create_dissector_handle(dissect_CCTrCH_InformationItem_RL_FailureInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationList_RL_PreemptRequiredInd, create_dissector_handle(dissect_RL_InformationList_RL_PreemptRequiredInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationItem_RL_PreemptRequiredInd, create_dissector_handle(dissect_RL_InformationItem_RL_PreemptRequiredInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd, create_dissector_handle(dissect_HSDSCHMacdFlowSpecificInformationItem_RL_PreemptRequiredInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd, create_dissector_handle(dissect_EDCH_MacdFlowSpecificInformationItem_RL_PreemptRequiredInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Reporing_Object_RL_RestoreInd, create_dissector_handle(dissect_Reporting_Object_RL_RestoreInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Information_RL_RestoreInd, create_dissector_handle(dissect_RL_Information_RL_RestoreInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Set_Information_RL_RestoreInd, create_dissector_handle(dissect_RL_Set_Information_RL_RestoreInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CCTrCH_InformationItem_RL_RestoreInd, create_dissector_handle(dissect_CCTrCH_InformationItem_RL_RestoreInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_PowerAdjustmentType, create_dissector_handle(dissect_PowerAdjustmentType_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DLReferencePower, create_dissector_handle(dissect_DL_Power_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_InnerLoopDLPCStatus, create_dissector_handle(dissect_InnerLoopDLPCStatus_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DLReferencePowerList_DL_PC_Rqst, create_dissector_handle(dissect_DL_ReferencePowerInformationList_DL_PC_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MaxAdjustmentStep, create_dissector_handle(dissect_MaxAdjustmentStep_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_AdjustmentPeriod, create_dissector_handle(dissect_AdjustmentPeriod_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_AdjustmentRatio, create_dissector_handle(dissect_ScaledAdjustmentRatio_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_ReferencePowerInformation_DL_PC_Rqst, create_dissector_handle(dissect_DL_ReferencePowerInformation_DL_PC_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_timeSlot_ISCP, create_dissector_handle(dissect_DL_TimeSlot_ISCP_Info_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Information_PhyChReconfRqstFDD, create_dissector_handle(dissect_RL_Information_PhyChReconfRqstFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Information_PhyChReconfRqstTDD, create_dissector_handle(dissect_RL_Information_PhyChReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UL_DPCH_InformationItem_PhyChReconfRqstTDD, create_dissector_handle(dissect_UL_DPCH_InformationItem_PhyChReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationListIE_PhyChReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_DPCH_InformationItem_PhyChReconfRqstTDD, create_dissector_handle(dissect_DL_DPCH_InformationItem_PhyChReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Cause, create_dissector_handle(dissect_Cause_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CongestionCause, create_dissector_handle(dissect_CongestionCause_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationList_RL_CongestInd, create_dissector_handle(dissect_RL_InformationList_RL_CongestInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationItem_RL_CongestInd, create_dissector_handle(dissect_RL_InformationItem_RL_CongestInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DCH_Rate_InformationItem_RL_CongestInd, create_dissector_handle(dissect_DCH_Rate_InformationItem_RL_CongestInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_EDCH_MacdFlowSpecificInformationItem_RL_CongestInd, create_dissector_handle(dissect_EDCH_MacdFlowSpecificInformationItem_RL_CongestInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UC_ID, create_dissector_handle(dissect_UC_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_SAI, create_dissector_handle(dissect_SAI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GA_Cell, create_dissector_handle(dissect_GA_Cell_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_C_RNTI, create_dissector_handle(dissect_C_RNTI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_PropagationDelay, create_dissector_handle(dissect_PropagationDelay_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_STTD_SupportIndicator, create_dissector_handle(dissect_STTD_SupportIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_ClosedLoopMode1_SupportIndicator, create_dissector_handle(dissect_ClosedLoopMode1_SupportIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_L3_Information, create_dissector_handle(dissect_L3_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_URA_Information, create_dissector_handle(dissect_URA_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RxTimingDeviationForTA, create_dissector_handle(dissect_RxTimingDeviationForTA_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_C_ID, create_dissector_handle(dissect_C_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_D_RNTI_ReleaseIndication, create_dissector_handle(dissect_D_RNTI_ReleaseIndication_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RANAP_RelocationInformation, create_dissector_handle(dissect_RANAP_RelocationInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_PagingArea_PagingRqst, create_dissector_handle(dissect_PagingArea_PagingRqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_IMSI, create_dissector_handle(dissect_IMSI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DRXCycleLengthCoefficient, create_dissector_handle(dissect_DRXCycleLengthCoefficient_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CNOriginatedPage_PagingRqst, create_dissector_handle(dissect_CNOriginatedPage_PagingRqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MeasurementID, create_dissector_handle(dissect_MeasurementID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DedicatedMeasurementObjectType_DM_Rqst, create_dissector_handle(dissect_DedicatedMeasurementObjectType_DM_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DedicatedMeasurementType, create_dissector_handle(dissect_DedicatedMeasurementType_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MeasurementFilterCoefficient, create_dissector_handle(dissect_MeasurementFilterCoefficient_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_ReportCharacteristics, create_dissector_handle(dissect_ReportCharacteristics_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CFNReportingIndicator, create_dissector_handle(dissect_FNReportingIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationItem_DM_Rqst, create_dissector_handle(dissect_RL_InformationItem_DM_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Set_InformationItem_DM_Rqst, create_dissector_handle(dissect_RL_Set_InformationItem_DM_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DedicatedMeasurementObjectType_DM_Rsp, create_dissector_handle(dissect_DedicatedMeasurementObjectType_DM_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationItem_DM_Rsp, create_dissector_handle(dissect_RL_InformationItem_DM_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Set_InformationItem_DM_Rsp, create_dissector_handle(dissect_RL_Set_InformationItem_DM_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Unsuccessful_InformationItem_DM_Fail, create_dissector_handle(dissect_RL_Unsuccessful_InformationItem_DM_Fail_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Successful_InformationItem_DM_Fail, create_dissector_handle(dissect_RL_Successful_InformationItem_DM_Fail_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Set_Unsuccessful_InformationItem_DM_Fail, create_dissector_handle(dissect_RL_Set_Unsuccessful_InformationItem_DM_Fail_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Set_Successful_InformationItem_DM_Fail, create_dissector_handle(dissect_RL_Set_Successful_InformationItem_DM_Fail_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DedicatedMeasurementObjectType_DM_Rprt, create_dissector_handle(dissect_DedicatedMeasurementObjectType_DM_Rprt_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_InformationItem_DM_Rprt, create_dissector_handle(dissect_RL_InformationItem_DM_Rprt_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Set_InformationItem_DM_Rprt, create_dissector_handle(dissect_RL_Set_InformationItem_DM_Rprt_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Unsuccessful_InformationItem_DM_Fail_Ind, create_dissector_handle(dissect_RL_Unsuccessful_InformationItem_DM_Fail_Ind_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind, create_dissector_handle(dissect_RL_Set_Unsuccessful_InformationItem_DM_Fail_Ind_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_TransportBearerRequestIndicator, create_dissector_handle(dissect_TransportBearerRequestIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_TransportBearerID, create_dissector_handle(dissect_TransportBearerID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD, create_dissector_handle(dissect_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_TransportLayerAddress, create_dissector_handle(dissect_TransportLayerAddress_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_BindingID, create_dissector_handle(dissect_BindingID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_FACH_FlowControlInformation, create_dissector_handle(dissect_FACH_FlowControlInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD, create_dissector_handle(dissect_FACH_InfoForUESelectedS_CCPCH_CTCH_ResourceRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CommonMeasurementObjectType_CM_Rqst, create_dissector_handle(dissect_CommonMeasurementObjectType_CM_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CommonMeasurementType, create_dissector_handle(dissect_CommonMeasurementType_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_SFNReportingIndicator, create_dissector_handle(dissect_FNReportingIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_SFN, create_dissector_handle(dissect_SFN_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CommonMeasurementAccuracy, create_dissector_handle(dissect_CommonMeasurementAccuracy_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_neighbouringTDDCellMeasurementInformationLCR, create_dissector_handle(dissect_NeighbouringTDDCellMeasurementInformationLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_neighbouringTDDCellMeasurementInformation768, create_dissector_handle(dissect_NeighbouringTDDCellMeasurementInformation768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CommonMeasurementObjectType_CM_Rsp, create_dissector_handle(dissect_CommonMeasurementObjectType_CM_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_CommonMeasurementObjectType_CM_Rprt, create_dissector_handle(dissect_CommonMeasurementObjectType_CM_Rprt_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_InformationExchangeID, create_dissector_handle(dissect_InformationExchangeID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_InformationExchangeObjectType_InfEx_Rqst, create_dissector_handle(dissect_InformationExchangeObjectType_InfEx_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_InformationType, create_dissector_handle(dissect_InformationType_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_InformationReportCharacteristics, create_dissector_handle(dissect_InformationReportCharacteristics_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GSM_Cell_InfEx_Rqst, create_dissector_handle(dissect_GSM_Cell_InfEx_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MBMS_Bearer_Service_List, create_dissector_handle(dissect_MBMS_Bearer_Service_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_InformationExchangeObjectType_InfEx_Rsp, create_dissector_handle(dissect_InformationExchangeObjectType_InfEx_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MBMS_Bearer_Service_List_InfEx_Rsp, create_dissector_handle(dissect_MBMS_Bearer_Service_List_InfEx_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_InformationExchangeObjectType_InfEx_Rprt, create_dissector_handle(dissect_InformationExchangeObjectType_InfEx_Rprt_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RNC_ID, create_dissector_handle(dissect_RNC_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_ResetIndicator, create_dissector_handle(dissect_ResetIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_ContextInfoItem_Reset, create_dissector_handle(dissect_ContextInfoItem_Reset_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_ContextGroupInfoItem_Reset, create_dissector_handle(dissect_ContextGroupInfoItem_Reset_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DelayedActivationList_RL_ActivationCmdFDD, create_dissector_handle(dissect_DelayedActivationInformationList_RL_ActivationCmdFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DelayedActivationInformation_RL_ActivationCmdFDD, create_dissector_handle(dissect_DelayedActivationInformation_RL_ActivationCmdFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DelayedActivationList_RL_ActivationCmdTDD, create_dissector_handle(dissect_DelayedActivationInformationList_RL_ActivationCmdTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DelayedActivationInformation_RL_ActivationCmdTDD, create_dissector_handle(dissect_DelayedActivationInformation_RL_ActivationCmdTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_HSDSCH_FDD_Update_Information, create_dissector_handle(dissect_HSDSCH_FDD_Update_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_ParameterUpdateIndicationFDD_RL_InformationList, create_dissector_handle(dissect_RL_ParameterUpdateIndicationFDD_RL_InformationList_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RL_ParameterUpdateIndicationFDD_RL_Information_Item, create_dissector_handle(dissect_RL_ParameterUpdateIndicationFDD_RL_Information_Item_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_HSDSCH_TDD_Update_Information, create_dissector_handle(dissect_HSDSCH_TDD_Update_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UEMeasurementType, create_dissector_handle(dissect_UEMeasurementType_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UEMeasurementTimeslotInfoHCR, create_dissector_handle(dissect_UEMeasurementTimeslotInfoHCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UEMeasurementTimeslotInfoLCR, create_dissector_handle(dissect_UEMeasurementTimeslotInfoLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UEMeasurementReportCharacteristics, create_dissector_handle(dissect_UEMeasurementReportCharacteristics_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UEMeasurementParameterModAllow, create_dissector_handle(dissect_UEMeasurementParameterModAllow_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UEMeasurementValueInformation, create_dissector_handle(dissect_UEMeasurementValueInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_TraceReference, create_dissector_handle(dissect_TraceReference_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UEIdentity, create_dissector_handle(dissect_UEIdentity_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_TraceRecordingSessionReference, create_dissector_handle(dissect_TraceRecordingSessionReference_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_ListOfInterfacesToTrace, create_dissector_handle(dissect_ListOfInterfacesToTrace_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_TraceDepth, create_dissector_handle(dissect_TraceDepth_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_InterfacesToTraceItem, create_dissector_handle(dissect_InterfacesToTraceItem_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UE_State, create_dissector_handle(dissect_UE_State_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_ProvidedInformation, create_dissector_handle(dissect_ProvidedInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Neighbouring_UMTS_CellInformationExtensionItem, create_dissector_handle(dissect_Neighbouring_UMTS_CellInformationExtensionItem_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Control_Type_InformationTransferControlReq, create_dissector_handle(dissect_Control_Type_InformationTransferControlReq_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UMTS_Cells_Info, create_dissector_handle(dissect_UMTS_Cell_Info_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_ANR_Cell_InfEx_Rqst, create_dissector_handle(dissect_ANR_Cell_InfEx_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_ANR_Cell_InfEx_Rsp, create_dissector_handle(dissect_ANR_Cell_InfEx_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_ANR_Cell_Information, create_dissector_handle(dissect_ANR_Cell_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_UE_transmission_power_headroom, create_dissector_handle(dissect_UE_transmission_power_headroom_Value_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Common_E_RGCH_Cell_InfEx_Rqst, create_dissector_handle(dissect_Common_E_RGCH_Cell_InfEx_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Common_E_RGCH_Cell_InfEx_Rsp, create_dissector_handle(dissect_Common_E_RGCH_Cell_InfEx_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MessageStructure, create_dissector_handle(dissect_MessageStructure_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TypeOfError, create_dissector_handle(dissect_TypeOfError_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TnlQos, create_dissector_handle(dissect_TnlQos_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Guaranteed_Rate_Information, create_dissector_handle(dissect_Guaranteed_Rate_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TrafficClass, create_dissector_handle(dissect_TrafficClass_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Unidirectional_DCH_Indicator, create_dissector_handle(dissect_Unidirectional_DCH_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Allowed_Rate_Information, create_dissector_handle(dissect_Allowed_Rate_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_ExtendedPropagationDelay, create_dissector_handle(dissect_ExtendedPropagationDelay_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Maximum_DL_Power_TimeslotLCR_InformationItem, create_dissector_handle(dissect_DL_Power_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Minimum_DL_Power_TimeslotLCR_InformationItem, create_dissector_handle(dissect_DL_Power_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DSCH_InitialWindowSize, create_dissector_handle(dissect_DSCH_InitialWindowSize_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_BindingID, create_dissector_handle(dissect_BindingID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TransportLayerAddress, create_dissector_handle(dissect_TransportLayerAddress_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_RGCH_E_HICH_ChannelisationCodeValidityIndicator, create_dissector_handle(dissect_E_RGCH_E_HICH_ChannelisationCodeValidityIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_PowerOffset_for_SchedulingInfo, create_dissector_handle(dissect_E_DCH_PowerOffset_for_SchedulingInfo_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TrCH_SrcStatisticsDescr, create_dissector_handle(dissect_TrCH_SrcStatisticsDescr_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_Minimum_Set_E_TFCIValidityIndicator, create_dissector_handle(dissect_E_DCH_Minimum_Set_E_TFCIValidityIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_eDCH_MACdFlow_Retransmission_Timer_LCR, create_dissector_handle(dissect_E_DCH_MACdFlow_Retransmission_Timer_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Satellite_Almanac_Information_ExtItem, create_dissector_handle(dissect_Satellite_Almanac_Information_ExtItem_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HARQ_Preamble_Mode, create_dissector_handle(dissect_HARQ_Preamble_Mode_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_MACdPDUSizeFormat, create_dissector_handle(dissect_HSDSCH_MACdPDUSizeFormat_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_User_Plane_Congestion_Fields_Inclusion, create_dissector_handle(dissect_User_Plane_Congestion_Fields_Inclusion_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HARQ_Preamble_Mode_Activation_Indicator, create_dissector_handle(dissect_HARQ_Preamble_Mode_Activation_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MIMO_InformationResponse, create_dissector_handle(dissect_MIMO_InformationResponse_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HS_PDSCH_Code_Change_Grant, create_dissector_handle(dissect_HS_PDSCH_Code_Change_Grant_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MaximumMACdPDU_SizeExtended, create_dissector_handle(dissect_MAC_PDU_SizeExtended_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Continuous_Packet_Connectivity_HS_SCCH_Less_Information, create_dissector_handle(dissect_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response, create_dissector_handle(dissect_Continuous_Packet_Connectivity_HS_SCCH_Less_Information_Response_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_hSSCCH_TDD_Specific_InfoList_Response768, create_dissector_handle(dissect_HSSCCH_TDD_Specific_InfoList_Response768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_hSPDSCH_TDD_Specific_InfoList_Response768, create_dissector_handle(dissect_HSPDSCH_TDD_Specific_InfoList_Response768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HS_PDSCH_Code_Change_Indicator, create_dissector_handle(dissect_HS_PDSCH_Code_Change_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_Information, create_dissector_handle(dissect_GANSS_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_neighbouring_LCR_TDD_CellInformation, create_dissector_handle(dissect_Neighbouring_LCR_TDD_CellInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Extended_RNC_ID, create_dissector_handle(dissect_Extended_RNC_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RestrictionStateIndicator, create_dissector_handle(dissect_RestrictionStateIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DPC_Mode_Change_SupportIndicator, create_dissector_handle(dissect_DPC_Mode_Change_SupportIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CoverageIndicator, create_dissector_handle(dissect_CoverageIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_AntennaColocationIndicator, create_dissector_handle(dissect_AntennaColocationIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HCS_Prio, create_dissector_handle(dissect_HCS_Prio_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CellCapabilityContainer_FDD, create_dissector_handle(dissect_CellCapabilityContainer_FDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_SNA_Information, create_dissector_handle(dissect_SNA_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_FrequencyBandIndicator, create_dissector_handle(dissect_FrequencyBandIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GERAN_Cell_Capability, create_dissector_handle(dissect_GERAN_Cell_Capability_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GERAN_Classmark, create_dissector_handle(dissect_GERAN_Classmark_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_ExtendedGSMCellIndividualOffset, create_dissector_handle(dissect_ExtendedGSMCellIndividualOffset_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CellCapabilityContainer_TDD, create_dissector_handle(dissect_CellCapabilityContainer_TDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CellCapabilityContainer_TDD768, create_dissector_handle(dissect_CellCapabilityContainer_TDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CellCapabilityContainer_TDD_LCR, create_dissector_handle(dissect_CellCapabilityContainer_TDD_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Cell_Capacity_Class_Value, create_dissector_handle(dissect_Cell_Capacity_Class_Value_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_NACC_Related_Data, create_dissector_handle(dissect_NACC_Related_Data_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MBMS_Bearer_Service_Full_Address, create_dissector_handle(dissect_MBMS_Bearer_Service_Full_Address_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Inter_Frequency_Cell_Information, create_dissector_handle(dissect_Inter_Frequency_Cell_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_Common_Data, create_dissector_handle(dissect_GANSS_Common_Data_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_Generic_Data, create_dissector_handle(dissect_GANSS_Generic_Data_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UEMeasurementValueTransmittedPowerList768, create_dissector_handle(dissect_UEMeasurementValueTransmittedPowerList768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UEMeasurementValueTimeslotISCPList768, create_dissector_handle(dissect_UEMeasurementValueTimeslotISCPList768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Extended_SRNC_ID, create_dissector_handle(dissect_Extended_RNC_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_PLCCH_Information_UL_TimeslotLCR_Info, create_dissector_handle(dissect_PLCCHinformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DPC_Mode, create_dissector_handle(dissect_DPC_Mode_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_DPDCHIndicatorEDCH, create_dissector_handle(dissect_UL_DPDCHIndicatorEDCH_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Enhanced_PrimaryCPICH_EcNo, create_dissector_handle(dissect_Enhanced_PrimaryCPICH_EcNo_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RL_Specific_DCH_Info, create_dissector_handle(dissect_RL_Specific_DCH_Info_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DelayedActivation, create_dissector_handle(dissect_DelayedActivation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CellPortionID, create_dissector_handle(dissect_CellPortionID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RL_Specific_EDCH_Information, create_dissector_handle(dissect_RL_Specific_EDCH_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDCH_RL_Indication, create_dissector_handle(dissect_EDCH_RL_Indication_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Permanent_NAS_UE_Identity, create_dissector_handle(dissect_Permanent_NAS_UE_Identity_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_PowerBalancing_Information, create_dissector_handle(dissect_DL_PowerBalancing_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_FDD_Information, create_dissector_handle(dissect_HSDSCH_FDD_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSPDSCH_RL_ID, create_dissector_handle(dissect_RL_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MBMS_Bearer_Service_List, create_dissector_handle(dissect_MBMS_Bearer_Service_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDPCH_Information, create_dissector_handle(dissect_EDPCH_Information_FDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDCH_FDD_Information, create_dissector_handle(dissect_EDCH_FDD_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Serving_EDCHRL_Id, create_dissector_handle(dissect_EDCH_Serving_RL_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_F_DPCH_Information_RL_SetupRqstFDD, create_dissector_handle(dissect_F_DPCH_Information_RL_SetupRqstFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Initial_DL_DPCH_TimingAdjustment_Allowed, create_dissector_handle(dissect_Initial_DL_DPCH_TimingAdjustment_Allowed_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DCH_Indicator_For_E_DCH_HSDPA_Operation, create_dissector_handle(dissect_DCH_Indicator_For_E_DCH_HSDPA_Operation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Serving_cell_change_CFN, create_dissector_handle(dissect_CFN_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Continuous_Packet_Connectivity_DTX_DRX_Information, create_dissector_handle(dissect_Continuous_Packet_Connectivity_DTX_DRX_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_F_DPCH_SlotFormatSupportRequest, create_dissector_handle(dissect_F_DPCH_SlotFormatSupportRequest_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD_Support_8PSK, create_dissector_handle(dissect_Support_8PSK_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD768_minimumSpreadingFactor_UL, create_dissector_handle(dissect_MinimumSpreadingFactor768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD_maxNrDLPhysicalchannels, create_dissector_handle(dissect_MaxNrDLPhysicalchannelsTS_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD_Support_PLCCH, create_dissector_handle(dissect_Support_PLCCH_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD768_minimumSpreadingFactor_DL, create_dissector_handle(dissect_MinimumSpreadingFactor768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD768_maxNrDLPhysicalchannels, create_dissector_handle(dissect_MaxNrDLPhysicalchannels768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD768_maxNrDLPhysicalchannelsTS, create_dissector_handle(dissect_MaxNrDLPhysicalchannelsTS768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD_TPC_UplinkStepSize_LCR_RL_SetupRqstTDD, create_dissector_handle(dissect_TDD_TPC_UplinkStepSize_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_Timeslot_ISCP_LCR_Information_RL_SetupRqstTDD, create_dissector_handle(dissect_DL_TimeSlot_ISCP_LCR_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TSTD_Support_Indicator_RL_SetupRqstTDD, create_dissector_handle(dissect_TSTD_Support_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_Synchronisation_Parameters_LCR, create_dissector_handle(dissect_UL_Synchronisation_Parameters_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_PrimaryCCPCH_RSCP_Delta, create_dissector_handle(dissect_PrimaryCCPCH_RSCP_Delta_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_TDD_Information, create_dissector_handle(dissect_HSDSCH_TDD_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_PDSCH_RL_ID, create_dissector_handle(dissect_RL_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_Information, create_dissector_handle(dissect_E_DCH_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_Serving_RL_ID, create_dissector_handle(dissect_RL_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_768_Information, create_dissector_handle(dissect_E_DCH_768_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_LCR_Information, create_dissector_handle(dissect_E_DCH_LCR_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GA_CellAdditionalShapes, create_dissector_handle(dissect_GA_CellAdditionalShapes_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_PowerBalancing_ActivationIndicator, create_dissector_handle(dissect_DL_PowerBalancing_ActivationIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Primary_CPICH_Usage_For_Channel_Estimation, create_dissector_handle(dissect_Primary_CPICH_Usage_For_Channel_Estimation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Secondary_CPICH_Information, create_dissector_handle(dissect_Secondary_CPICH_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Active_MBMS_Bearer_ServiceFDD_PFL, create_dissector_handle(dissect_Active_MBMS_Bearer_Service_ListFDD_PFL_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDCH_RLSet_Id, create_dissector_handle(dissect_RL_Set_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDCH_FDD_DL_ControlChannelInformation, create_dissector_handle(dissect_EDCH_FDD_DL_ControlChannelInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Initial_DL_DPCH_TimingAdjustment, create_dissector_handle(dissect_DL_DPCH_TimingAdjustment_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_F_DPCH_SlotFormat, create_dissector_handle(dissect_F_DPCH_SlotFormat_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DCH_InformationResponse, create_dissector_handle(dissect_DCH_InformationResponse_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDCH_FDD_InformationResponse, create_dissector_handle(dissect_EDCH_FDD_InformationResponse_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_RNTI, create_dissector_handle(dissect_HSDSCH_RNTI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_FDD_Information_Response, create_dissector_handle(dissect_HSDSCH_FDD_Information_Response_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_SixtyfourQAM_DL_SupportIndicator, create_dissector_handle(dissect_SixtyfourQAM_DL_SupportIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TimeSlot_RL_SetupRspTDD, create_dissector_handle(dissect_TimeSlot_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_SIR_Target_CCTrCH_InformationItem_RL_SetupRspTDD, create_dissector_handle(dissect_UL_SIR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CCTrCH_Maximum_DL_Power_RL_SetupRspTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CCTrCH_Minimum_DL_Power_RL_SetupRspTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RL_LCR_InformationResponse_RL_SetupRspTDD, create_dissector_handle(dissect_RL_LCR_InformationResponse_RL_SetupRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_TDD_Information_Response, create_dissector_handle(dissect_HSDSCH_TDD_Information_Response_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DSCH_RNTI, create_dissector_handle(dissect_DSCH_RNTI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Active_MBMS_Bearer_ServiceTDD_PFL, create_dissector_handle(dissect_Active_MBMS_Bearer_Service_ListTDD_PFL_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RL_InformationResponse_RL_SetupRspTDD768, create_dissector_handle(dissect_RL_InformationResponse_RL_SetupRspTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_Information_Response, create_dissector_handle(dissect_E_DCH_Information_Response_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_768_Information_Response, create_dissector_handle(dissect_E_DCH_768_Information_Response_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_LCR_Information_Response, create_dissector_handle(dissect_E_DCH_LCR_Information_Response_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_TimingAdvanceCtrl_LCR, create_dissector_handle(dissect_UL_TimingAdvanceCtrl_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_SIR_Target_CCTrCH_LCR_InformationItem_RL_SetupRspTDD, create_dissector_handle(dissect_UL_SIR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Max_UE_DTX_Cycle, create_dissector_handle(dissect_Max_UE_DTX_Cycle_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DLReferencePower, create_dissector_handle(dissect_DL_Power_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_SynchronisationIndicator, create_dissector_handle(dissect_SynchronisationIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HS_DSCH_serving_cell_change_information, create_dissector_handle(dissect_HS_DSCH_serving_cell_change_information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDPCH_Information_RLAdditionReq_FDD, create_dissector_handle(dissect_EDPCH_Information_RLAdditionReq_FDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_Configured_Indicator, create_dissector_handle(dissect_HSDSCH_Configured_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_Timeslot_ISCP_LCR_Information_RL_AdditionRqstTDD, create_dissector_handle(dissect_DL_TimeSlot_ISCP_LCR_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_CCTrCH_InformationList_RL_AdditionRqstTDD, create_dissector_handle(dissect_UL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_CCTrCH_InformationList_RL_AdditionRqstTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationList_RL_AdditionRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HS_DSCH_serving_cell_change_informationResponse, create_dissector_handle(dissect_HS_DSCH_serving_cell_change_informationResponse_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_Serving_cell_change_informationResponse, create_dissector_handle(dissect_E_DCH_Serving_cell_change_informationResponse_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MAChs_ResetIndicator, create_dissector_handle(dissect_MAChs_ResetIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CCTrCH_Maximum_DL_Power_RL_AdditionRspTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CCTrCH_Minimum_DL_Power_RL_AdditionRspTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RL_LCR_InformationResponse_RL_AdditionRspTDD, create_dissector_handle(dissect_RL_LCR_InformationResponse_RL_AdditionRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RL_InformationResponse_RL_AdditionRspTDD768, create_dissector_handle(dissect_RL_InformationResponse_RL_AdditionRspTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_DPCH_Power_Information_RL_ReconfPrepFDD, create_dissector_handle(dissect_DL_DPCH_Power_Information_RL_ReconfPrepFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_DPCH_TimingAdjustment, create_dissector_handle(dissect_DL_DPCH_TimingAdjustment_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Phase_Reference_Update_Indicator, create_dissector_handle(dissect_Phase_Reference_Update_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_Information_to_Modify, create_dissector_handle(dissect_HSDSCH_Information_to_Modify_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_MACdFlows_to_Add, create_dissector_handle(dissect_HSDSCH_MACdFlows_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_MACdFlows_to_Delete, create_dissector_handle(dissect_HSDSCH_MACdFlows_to_Delete_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDCH_FDD_Information_To_Modify, create_dissector_handle(dissect_EDCH_FDD_Information_To_Modify_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDCH_MACdFlows_To_Add, create_dissector_handle(dissect_EDCH_MACdFlows_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDCH_MACdFlows_To_Delete, create_dissector_handle(dissect_EDCH_MACdFlows_To_Delete_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDPCH_Information_RLReconfRequest_FDD, create_dissector_handle(dissect_EDPCH_Information_RLReconfRequest_FDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_F_DPCH_Information_RL_ReconfPrepFDD, create_dissector_handle(dissect_F_DPCH_Information_RL_ReconfPrepFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Fast_Reconfiguration_Mode, create_dissector_handle(dissect_Fast_Reconfiguration_Mode_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CPC_Information, create_dissector_handle(dissect_CPC_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_SIRTarget, create_dissector_handle(dissect_UL_SIR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD_TPC_UplinkStepSize_InformationAdd_LCR_RL_ReconfPrepTDD, create_dissector_handle(dissect_TDD_TPC_UplinkStepSize_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD_TPC_UplinkStepSize_InformationModify_LCR_RL_ReconfPrepTDD, create_dissector_handle(dissect_TDD_TPC_UplinkStepSize_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD_TPC_DownlinkStepSize_InformationAdd_RL_ReconfPrepTDD, create_dissector_handle(dissect_TDD_TPC_DownlinkStepSize_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD_TPC_DownlinkStepSize_InformationModify_RL_ReconfPrepTDD, create_dissector_handle(dissect_TDD_TPC_DownlinkStepSize_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_PrimaryCCPCH_RSCP_RL_ReconfPrepTDD, create_dissector_handle(dissect_PrimaryCCPCH_RSCP_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_TimeSlot_ISCP_Info_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_TimeSlot_ISCP_Info_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_Timeslot_ISCP_LCR_Information_RL_ReconfPrepTDD, create_dissector_handle(dissect_DL_TimeSlot_ISCP_LCR_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RL_Information_RL_ReconfPrepTDD, create_dissector_handle(dissect_RL_Information_RL_ReconfPrepTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_Information_Reconfig, create_dissector_handle(dissect_E_DCH_Information_Reconfig_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_768_Information_Reconfig, create_dissector_handle(dissect_E_DCH_768_Information_Reconfig_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_LCR_Information_Reconfig, create_dissector_handle(dissect_E_DCH_LCR_Information_Reconfig_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_PowerBalancing_UpdatedIndicator, create_dissector_handle(dissect_DL_PowerBalancing_UpdatedIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Secondary_CPICH_Information_Change, create_dissector_handle(dissect_Secondary_CPICH_Information_Change_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Fast_Reconfiguration_Permission, create_dissector_handle(dissect_Fast_Reconfiguration_Permission_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_secondary_LCR_CCPCH_Info_TDD, create_dissector_handle(dissect_Secondary_LCR_CCPCH_Info_TDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_secondary_CCPCH_Info_RL_ReconfReadyTDD768, create_dissector_handle(dissect_Secondary_CCPCH_Info_TDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_DPCH_LCR_InformationAddListIE_RL_ReconfReadyTDD, create_dissector_handle(dissect_UL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_DPCH_InformationAddListIE_RL_ReconfReadyTDD768, create_dissector_handle(dissect_UL_DPCH_InformationAddList_RL_ReconfReadyTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RxTimingDeviationForTAext, create_dissector_handle(dissect_RxTimingDeviationForTAext_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_Timeslot_LCR_InformationModifyList_RL_ReconfReadyTDD, create_dissector_handle(dissect_UL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768, create_dissector_handle(dissect_UL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD_UL_DPCH_TimeSlotFormatModifyItem_LCR_RL_ReconfReadyTDD, create_dissector_handle(dissect_TDD_UL_DPCH_TimeSlotFormat_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_DPCH_LCR_InformationAddListIE_RL_ReconfReadyTDD, create_dissector_handle(dissect_DL_DPCH_LCR_InformationAddList_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CCTrCH_Maximum_DL_Power_RL_ReconfReadyTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CCTrCH_Minimum_DL_Power_RL_ReconfReadyTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_DPCH_InformationAddListIE_RL_ReconfReadyTDD768, create_dissector_handle(dissect_DL_DPCH_InformationAddList_RL_ReconfReadyTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD, create_dissector_handle(dissect_DL_DPCH_InformationDeleteList768_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_Timeslot_LCR_InformationModifyList_RL_ReconfReadyTDD, create_dissector_handle(dissect_DL_TimeslotLCR_InformationModifyList_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768, create_dissector_handle(dissect_DL_Timeslot_InformationModifyList_RL_ReconfReadyTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TDD_DL_DPCH_TimeSlotFormatModifyItem_LCR_RL_ReconfReadyTDD, create_dissector_handle(dissect_TDD_DL_DPCH_TimeSlotFormat_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Maximum_DL_Power_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Minimum_DL_Power_TimeslotLCR_InformationModifyItem_RL_ReconfReadyTDD, create_dissector_handle(dissect_DL_Power_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Multiple_RL_InformationResponse_RL_ReconfReadyTDD, create_dissector_handle(dissect_Multiple_RL_InformationResponse_RL_ReconfReadyTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RL_ReconfigurationRequestFDD_RL_InformationList, create_dissector_handle(dissect_RL_ReconfigurationRequestFDD_RL_InformationList_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_ReferencePowerInformation, create_dissector_handle(dissect_DL_ReferencePowerInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_Information_to_Modify_Unsynchronised, create_dissector_handle(dissect_HSDSCH_Information_to_Modify_Unsynchronised_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RL_ReconfigurationRequestTDD_RL_Information, create_dissector_handle(dissect_Multiple_RL_ReconfigurationRequestTDD_RL_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_CCTrCH_InformationList_RL_ReconfRspTDD, create_dissector_handle(dissect_DL_CCTrCH_InformationList_RL_ReconfRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RL_ReconfigurationResponseTDD_RL_Information, create_dissector_handle(dissect_Multiple_RL_InformationResponse_RL_ReconfRspTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd, create_dissector_handle(dissect_EDCH_MacdFlowSpecificInformationList_RL_PreemptRequiredInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd, create_dissector_handle(dissect_HSDSCHMacdFlowSpecificInformationList_RL_PreemptRequiredInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_timeSlot_ISCP_LCR_List_DL_PC_Rqst_TDD, create_dissector_handle(dissect_DL_TimeSlot_ISCP_LCR_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_PrimCCPCH_RSCP_DL_PC_RqstTDD, create_dissector_handle(dissect_PrimaryCCPCH_RSCP_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD, create_dissector_handle(dissect_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD, create_dissector_handle(dissect_HSPDSCH_Timeslot_InformationListLCR_PhyChReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768, create_dissector_handle(dissect_HSPDSCH_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_Timeslot_LCR_InformationList_PhyChReconfRqstTDD, create_dissector_handle(dissect_UL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_Timeslot_InformationList_PhyChReconfRqstTDD768, create_dissector_handle(dissect_UL_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_PLCCH_Information_PhyChReconfRqstTDD, create_dissector_handle(dissect_PLCCHinformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_Timeslot_LCR_InformationList_PhyChReconfRqstTDD, create_dissector_handle(dissect_DL_TimeslotLCR_InformationList_PhyChReconfRqstTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_Timeslot_InformationList_PhyChReconfRqstTDD768, create_dissector_handle(dissect_DL_Timeslot_InformationList_PhyChReconfRqstTDD768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDCH_MacdFlowSpecificInformationList_RL_CongestInd, create_dissector_handle(dissect_EDCH_MacdFlowSpecificInformationList_RL_CongestInd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CommonTransportChannelResourcesInitialisationNotRequired, create_dissector_handle(dissect_CommonTransportChannelResourcesInitialisationNotRequired_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Active_MBMS_Bearer_ServiceFDD, create_dissector_handle(dissect_Active_MBMS_Bearer_Service_ListFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Inter_Frequency_Cell_List, create_dissector_handle(dissect_Inter_Frequency_Cell_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Active_MBMS_Bearer_ServiceTDD, create_dissector_handle(dissect_Active_MBMS_Bearer_Service_ListTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RxTimingDeviationForTA768, create_dissector_handle(dissect_RxTimingDeviationForTA768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_URA_ID, create_dissector_handle(dissect_URA_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Old_URA_ID, create_dissector_handle(dissect_URA_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_SRNC_ID, create_dissector_handle(dissect_RNC_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSSICH_Info_DM_Rqst, create_dissector_handle(dissect_HSSICH_Info_DM_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DPCH_ID768_DM_Rqst, create_dissector_handle(dissect_DPCH_ID768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_PartialReportingIndicator, create_dissector_handle(dissect_PartialReportingIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MeasurementRecoveryBehavior, create_dissector_handle(dissect_MeasurementRecoveryBehavior_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_AlternativeFormatReportingIndicator, create_dissector_handle(dissect_AlternativeFormatReportingIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSSICH_Info_DM, create_dissector_handle(dissect_HS_SICH_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_multiple_DedicatedMeasurementValueList_TDD_DM_Rsp, create_dissector_handle(dissect_Multiple_DedicatedMeasurementValueList_TDD_DM_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp, create_dissector_handle(dissect_Multiple_DedicatedMeasurementValueList_LCR_TDD_DM_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_multiple_HSSICHMeasurementValueList_TDD_DM_Rsp, create_dissector_handle(dissect_Multiple_HSSICHMeasurementValueList_TDD_DM_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp, create_dissector_handle(dissect_Multiple_DedicatedMeasurementValueList_TDD768_DM_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DPCH_ID768_DM_Rsp, create_dissector_handle(dissect_DPCH_ID768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MeasurementRecoverySupportIndicator, create_dissector_handle(dissect_MeasurementRecoverySupportIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DedicatedMeasurementObjectType_DM_Fail, create_dissector_handle(dissect_DedicatedMeasurementObjectType_DM_Fail_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSSICH_Info_DM_Rprt, create_dissector_handle(dissect_HS_SICH_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DPCH_ID768_DM_Rprt, create_dissector_handle(dissect_DPCH_ID768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MeasurementRecoveryReportingIndicator, create_dissector_handle(dissect_MeasurementRecoveryReportingIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DedicatedMeasurementObjectType_DM_Fail_Ind, create_dissector_handle(dissect_DedicatedMeasurementObjectType_DM_Fail_Ind_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Enhanced_FACH_Support_Indicator, create_dissector_handle(dissect_Enhanced_FACH_Support_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_C_ID, create_dissector_handle(dissect_C_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Enhanced_FACH_Information_ResponseFDD, create_dissector_handle(dissect_Enhanced_FACH_Information_ResponseFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_S_RNTI, create_dissector_handle(dissect_S_RNTI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_D_RNTI, create_dissector_handle(dissect_D_RNTI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_FDD_Update_Information, create_dissector_handle(dissect_E_DCH_FDD_Update_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UEMeasurementTimeslotInfo768, create_dissector_handle(dissect_UEMeasurementTimeslotInfo768_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Ext_Reference_E_TFCI_PO, create_dissector_handle(dissect_Ext_Reference_E_TFCI_PO_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Ext_Max_Bits_MACe_PDU_non_scheduled, create_dissector_handle(dissect_Ext_Max_Bits_MACe_PDU_non_scheduled_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HARQ_MemoryPartitioningInfoExtForMIMO, create_dissector_handle(dissect_HARQ_MemoryPartitioningInfoExtForMIMO_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MIMO_ActivationIndicator, create_dissector_handle(dissect_MIMO_ActivationIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MIMO_Mode_Indicator, create_dissector_handle(dissect_MIMO_Mode_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MIMO_N_M_Ratio, create_dissector_handle(dissect_MIMO_N_M_Ratio_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TransportBearerNotSetupIndicator, create_dissector_handle(dissect_TransportBearerNotSetupIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TransportBearerNotRequestedIndicator, create_dissector_handle(dissect_TransportBearerNotRequestedIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UARFCNforNt, create_dissector_handle(dissect_UARFCN_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_LCRTDD_uplink_Physical_Channel_Capability, create_dissector_handle(dissect_LCRTDD_Uplink_Physical_Channel_Capability_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_number_Of_Supported_Carriers, create_dissector_handle(dissect_Number_Of_Supported_Carriers_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSSICH_SIRTarget, create_dissector_handle(dissect_UL_SIR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSSICH_TPC_StepSize, create_dissector_handle(dissect_TDD_TPC_UplinkStepSize_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_tSN_Length, create_dissector_handle(dissect_TSN_Length_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HS_SICH_ID_Extension, create_dissector_handle(dissect_HS_SICH_ID_Extension_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSSICH_Info_DM_Rqst_Extension, create_dissector_handle(dissect_HSSICH_Info_DM_Rqst_Extension_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_multipleFreq_HSPDSCH_InformationList_ResponseTDDLCR, create_dissector_handle(dissect_MultipleFreq_HSPDSCH_InformationList_ResponseTDDLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_multicarrier_number, create_dissector_handle(dissect_Multicarrier_Number_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UPPCHPositionLCR, create_dissector_handle(dissect_UPPCHPositionLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UpPCH_InformationList_LCRTDD, create_dissector_handle(dissect_UpPCH_InformationList_LCRTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UpPCH_InformationItem_LCRTDD, create_dissector_handle(dissect_UpPCH_InformationItem_LCRTDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Multiple_PLMN_List, create_dissector_handle(dissect_Multiple_PLMN_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UE_Capabilities_Info, create_dissector_handle(dissect_UE_Capabilities_Info_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_FrameOffset, create_dissector_handle(dissect_FrameOffset_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_ChipOffset, create_dissector_handle(dissect_ChipOffset_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Enhanced_PCH_Capability, create_dissector_handle(dissect_Enhanced_PCH_Capability_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_SixteenQAM_UL_Operation_Indicator, create_dissector_handle(dissect_SixteenQAM_UL_Operation_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_TFCI_Boost_Information, create_dissector_handle(dissect_E_TFCI_Boost_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_SixtyfourQAM_UsageAllowedIndicator, create_dissector_handle(dissect_SixtyfourQAM_UsageAllowedIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_SixtyfourQAM_DL_UsageIndicator, create_dissector_handle(dissect_SixtyfourQAM_DL_UsageIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Default_Serving_Grant_in_DTX_Cycle2, create_dissector_handle(dissect_E_Serving_Grant_Value_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DPDCH_PowerInterpolation, create_dissector_handle(dissect_E_DPDCH_PowerInterpolation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Extended_E_DCH_LCRTDD_PhysicalLayerCategory, create_dissector_handle(dissect_Extended_E_DCH_LCRTDD_PhysicalLayerCategory_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_MACdPDUSizeFormat, create_dissector_handle(dissect_E_DCH_MACdPDUSizeFormat_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Continuous_Packet_Connectivity_HS_SCCH_Less_Deactivate_Indicator, create_dissector_handle(dissect_Continuous_Packet_Connectivity_HS_SCCH_less_Deactivate_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_DL_Control_Channel_Change_Information, create_dissector_handle(dissect_E_DCH_DL_Control_Channel_Change_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_DL_Control_Channel_Grant_Information, create_dissector_handle(dissect_E_DCH_DL_Control_Channel_Grant_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MaximumNumber_Of_Retransmission_For_SchedulingInfo_LCRTDD, create_dissector_handle(dissect_MaxNr_Retransmissions_EDCH_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_RetransmissionTimer_For_SchedulingInfo_LCRTDD, create_dissector_handle(dissect_E_DCH_MACdFlow_Retransmission_Timer_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_PUCH_PowerControlGAP, create_dissector_handle(dissect_ControlGAP_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UE_with_enhanced_HS_SCCH_support_indicator, create_dissector_handle(dissect_NULL_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_TBSizeTableIndicator, create_dissector_handle(dissect_HSDSCH_TBSizeTableIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DGANSS_Corrections_Req, create_dissector_handle(dissect_DGANSS_Corrections_Req_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_AGCH_Table_Choice, create_dissector_handle(dissect_E_AGCH_Table_Choice_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RANAP_EnhancedRelocationInformationRequest, create_dissector_handle(dissect_RANAP_EnhancedRelocationInformationRequest_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_RANAP_EnhancedRelocationInformationResponse, create_dissector_handle(dissect_RANAP_EnhancedRelocationInformationResponse_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Common_EDCH_MAC_d_Flow_Specific_InformationFDD, create_dissector_handle(dissect_Common_EDCH_MAC_d_Flow_Specific_InformationFDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Common_EDCH_Support_Indicator, create_dissector_handle(dissect_Common_EDCH_Support_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_RNTI, create_dissector_handle(dissect_E_RNTI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Released_CN_Domain, create_dissector_handle(dissect_Released_CN_Domain_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst, create_dissector_handle(dissect_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp, create_dissector_handle(dissect_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt, create_dissector_handle(dissect_MBMS_Bearer_Service_in_MBMS_Cell_InfEx_Rprt_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MBMS_Cell_InfEx_Rqst, create_dissector_handle(dissect_MBMS_Cell_InfEx_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MBMS_Cell_InfEx_Rsp, create_dissector_handle(dissect_MBMS_Cell_InfEx_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MBMS_Cell_InfEx_Rprt, create_dissector_handle(dissect_MBMS_Cell_InfEx_Rprt_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Counting_Information, create_dissector_handle(dissect_Counting_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Transmission_Mode_Information, create_dissector_handle(dissect_Transmission_Mode_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MBMS_Neighbouring_Cell_Information, create_dissector_handle(dissect_MBMS_Neighbouring_Cell_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MBMS_RLC_Sequence_Number_Information, create_dissector_handle(dissect_MBMS_RLC_Sequence_Number_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RLC_Sequence_Number, create_dissector_handle(dissect_RLC_Sequence_Number_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Neighbouring_E_UTRA_CellInformation, create_dissector_handle(dissect_Neighbouring_E_UTRA_CellInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MBSFN_Cluster_Identity, create_dissector_handle(dissect_MBSFN_Cluster_Identity_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MCCH_Configuration_Info, create_dissector_handle(dissect_MCCH_Configuration_Info_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MCCH_Message_List, create_dissector_handle(dissect_MCCH_Message_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MBSFN_Scheduling_Transmission_Time_Interval_Info_List, create_dissector_handle(dissect_MBSFN_Scheduling_Transmission_Time_Interval_Info_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_Time_ID, create_dissector_handle(dissect_GANSS_Time_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_AddIonoModelReq, create_dissector_handle(dissect_GANSS_AddIonoModelReq_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_EarthOrientParaReq, create_dissector_handle(dissect_GANSS_EarthOrientParaReq_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_AddNavigationModelsReq, create_dissector_handle(dissect_GANSS_AddNavigationModelsReq_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_AddUTCModelsReq, create_dissector_handle(dissect_GANSS_AddUTCModelsReq_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_AuxInfoReq, create_dissector_handle(dissect_GANSS_AuxInfoReq_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_SBAS_ID, create_dissector_handle(dissect_GANSS_SBAS_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_ID, create_dissector_handle(dissect_GANSS_ID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_Additional_Ionospheric_Model, create_dissector_handle(dissect_GANSS_Additional_Ionospheric_Model_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_Earth_Orientation_Parameters, create_dissector_handle(dissect_GANSS_Earth_Orientation_Parameters_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_Additional_Time_Models, create_dissector_handle(dissect_GANSS_Additional_Time_Models_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_Additional_Navigation_Models, create_dissector_handle(dissect_GANSS_Additional_Navigation_Models_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_Additional_UTC_Models, create_dissector_handle(dissect_GANSS_Additional_UTC_Models_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_GANSS_Auxiliary_Information, create_dissector_handle(dissect_GANSS_Auxiliary_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MinimumReducedE_DPDCH_GainFactor, create_dissector_handle(dissect_MinimumReducedE_DPDCH_GainFactor_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Enhanced_FACH_Information_ResponseLCR, create_dissector_handle(dissect_Enhanced_FACH_Information_ResponseLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Common_EDCH_MAC_d_Flow_Specific_InformationLCR, create_dissector_handle(dissect_Common_EDCH_MAC_d_Flow_Specific_InformationLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_PreconfigurationSetup, create_dissector_handle(dissect_HSDSCH_PreconfigurationSetup_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_PreconfigurationInfo, create_dissector_handle(dissect_HSDSCH_PreconfigurationInfo_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_NoOfTargetCellHS_SCCH_Order, create_dissector_handle(dissect_NoOfTargetCellHS_SCCH_Order_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EnhancedHSServingCC_Abort, create_dissector_handle(dissect_EnhancedHSServingCC_Abort_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_HS_Cell_Information_RL_Setup, create_dissector_handle(dissect_Additional_HS_Cell_Information_RL_Setup_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_HS_Cell_Information_Response, create_dissector_handle(dissect_Additional_HS_Cell_Information_Response_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_HS_Cell_Information_RL_Addition, create_dissector_handle(dissect_Additional_HS_Cell_Information_RL_Addition_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_HS_Cell_Change_Information_Response, create_dissector_handle(dissect_Additional_HS_Cell_Change_Information_Response_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_HS_Cell_Information_RL_Reconf_Prep, create_dissector_handle(dissect_Additional_HS_Cell_Information_RL_Reconf_Prep_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_HS_Cell_Information_RL_Reconf_Req, create_dissector_handle(dissect_Additional_HS_Cell_Information_RL_Reconf_Req_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_HS_Cell_RL_Reconf_Response, create_dissector_handle(dissect_Additional_HS_Cell_RL_Reconf_Response_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_HS_Cell_Information_RL_Param_Upd, create_dissector_handle(dissect_Additional_HS_Cell_Information_RL_Param_Upd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Secondary_Serving_Cell_List, create_dissector_handle(dissect_Secondary_Serving_Cell_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MultiCarrier_HSDSCH_Physical_Layer_Category, create_dissector_handle(dissect_LCRTDD_HSDSCH_Physical_Layer_Category_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_IdleIntervalInformation, create_dissector_handle(dissect_IdleIntervalInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_NeedforIdleInterval, create_dissector_handle(dissect_NeedforIdleInterval_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_IdleIntervalConfigurationIndicator, create_dissector_handle(dissect_NULL_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_ContinuousPacketConnectivity_DRX_InformationLCR, create_dissector_handle(dissect_ContinuousPacketConnectivity_DRX_InformationLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_ContinuousPacketConnectivity_DRX_Information_ResponseLCR, create_dissector_handle(dissect_ContinuousPacketConnectivity_DRX_Information_ResponseLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_AGCH_UE_Inactivity_Monitor_Threshold, create_dissector_handle(dissect_E_AGCH_UE_Inactivity_Monitor_Threshold_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CPC_InformationLCR, create_dissector_handle(dissect_CPC_InformationLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_Semi_PersistentScheduling_Information_LCR, create_dissector_handle(dissect_E_DCH_Semi_PersistentScheduling_Information_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HS_DSCH_Semi_PersistentScheduling_Information_LCR, create_dissector_handle(dissect_HS_DSCH_Semi_PersistentScheduling_Information_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR, create_dissector_handle(dissect_HS_DSCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR, create_dissector_handle(dissect_E_DCH_Semi_PersistentScheduling_Information_ResponseLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MIMO_SFMode_For_HSPDSCHDualStream, create_dissector_handle(dissect_MIMO_SFMode_For_HSPDSCHDualStream_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MIMO_SFMode_Supported_For_HSPDSCHDualStream, create_dissector_handle(dissect_MIMO_SFMode_For_HSPDSCHDualStream_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MIMO_ReferenceSignal_InformationListLCR, create_dissector_handle(dissect_MIMO_ReferenceSignal_InformationListLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GANSS_alm_keplerianNAVAlmanac, create_dissector_handle(dissect_GANSS_alm_keplerianNAVAlmanac_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GANSS_alm_keplerianReducedAlmanac, create_dissector_handle(dissect_GANSS_alm_keplerianReducedAlmanac_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GANSS_alm_keplerianMidiAlmanac, create_dissector_handle(dissect_GANSS_alm_keplerianMidiAlmanac_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GANSS_alm_keplerianGLONASS, create_dissector_handle(dissect_GANSS_alm_keplerianGLONASS_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GANSS_alm_ecefSBASAlmanac, create_dissector_handle(dissect_GANSS_alm_ecefSBASAlmanac_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DL_RLC_PDU_Size_Format, create_dissector_handle(dissect_DL_RLC_PDU_Size_Format_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MACes_Maximum_Bitrate_LCR, create_dissector_handle(dissect_MACes_Maximum_Bitrate_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Single_Stream_MIMO_ActivationIndicator, create_dissector_handle(dissect_Single_Stream_MIMO_ActivationIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Single_Stream_MIMO_Mode_Indicator, create_dissector_handle(dissect_Single_Stream_MIMO_Mode_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Dual_Band_Secondary_Serving_Cell_List, create_dissector_handle(dissect_Secondary_Serving_Cell_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UE_AggregateMaximumBitRate, create_dissector_handle(dissect_UE_AggregateMaximumBitRate_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_power_offset_for_S_CPICH_for_MIMO, create_dissector_handle(dissect_PowerOffsetForSecondaryCPICHforMIMO_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_power_offset_for_S_CPICH_for_MIMO_Request_Indicator, create_dissector_handle(dissect_PowerOffsetForSecondaryCPICHforMIMORequestIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UE_SupportIndicatorExtension, create_dissector_handle(dissect_UE_SupportIndicatorExtension_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_ActivationInformation, create_dissector_handle(dissect_ActivationInformation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CellPortionLCRID, create_dissector_handle(dissect_CellPortionLCRID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_EDCH_Cell_Information_RL_Setup_Req, create_dissector_handle(dissect_Additional_EDCH_Setup_Info_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_EDCH_Cell_Information_Response, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_Response_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_EDCH_Cell_Information_RL_Add_Req, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_RL_Add_Req_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_EDCH_Cell_Information_Response_RLAdd, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_Response_RLAddList_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_EDCH_Cell_Information_RL_Reconf_Prep, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_RL_Reconf_Prep_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_EDCH_Cell_Information_RL_Reconf_Req, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_RL_Reconf_Req_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_EDCH_Cell_Information_RL_Param_Upd, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_RL_Param_Upd_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_EDCH_Preconfiguration_Information, create_dissector_handle(dissect_Additional_EDCH_Preconfiguration_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MulticellEDCH_Information, create_dissector_handle(dissect_MulticellEDCH_InformationItemIEs_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_EDCH_Cell_Information_ResponseRLReconf, create_dissector_handle(dissect_Additional_EDCH_Cell_Information_Response_RLReconf_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EDCH_Indicator, create_dissector_handle(dissect_NULL_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DiversityMode, create_dissector_handle(dissect_DiversityMode_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TransmitDiversityIndicator, create_dissector_handle(dissect_TransmitDiversityIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_NonCellSpecificTxDiversity, create_dissector_handle(dissect_NonCellSpecificTxDiversity_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CellCapabilityContainerExtension_FDD, create_dissector_handle(dissect_CellCapabilityContainerExtension_FDD_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HSDSCH_Physical_Layer_Category, create_dissector_handle(dissect_HSDSCH_Physical_Layer_Category_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_RNTI_For_FACH, create_dissector_handle(dissect_E_RNTI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_H_RNTI_For_FACH, create_dissector_handle(dissect_HSDSCH_RNTI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_RNTI_Allocation_Indicator, create_dissector_handle(dissect_RNTI_Allocation_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UE_AggregateMaximumBitRate_Enforcement_Indicator, create_dissector_handle(dissect_UE_AggregateMaximumBitRate_Enforcement_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DCH_MeasurementOccasion_Information, create_dissector_handle(dissect_DCH_MeasurementOccasion_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DCH_MeasurementType_Indicator, create_dissector_handle(dissect_DCH_MeasurementType_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Out_of_Sychronization_Window, create_dissector_handle(dissect_Out_of_Sychronization_Window_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_MulticellEDCH_RL_SpecificInformation, create_dissector_handle(dissect_MulticellEDCH_RL_SpecificInformationItemIEs_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DGNSS_ValidityPeriod, create_dissector_handle(dissect_DGNSS_ValidityPeriod_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TS0_HS_PDSCH_Indication_LCR, create_dissector_handle(dissect_TS0_HS_PDSCH_Indication_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UE_TS0_CapabilityLCR, create_dissector_handle(dissect_UE_TS0_CapabilityLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Non_Serving_RL_Preconfig_Setup, create_dissector_handle(dissect_Non_Serving_RL_Preconfig_Setup_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup, create_dissector_handle(dissect_Additional_E_DCH_Non_Serving_RL_Preconfiguration_Setup_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList, create_dissector_handle(dissect_Additional_E_DCH_New_non_serving_RL_E_DCH_FDD_DL_Control_Channel_InfoList_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CellListValidityIndicator, create_dissector_handle(dissect_CellListValidityIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_completeAlmanacProvided, create_dissector_handle(dissect_CompleteAlmanacProvided_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_ganss_Delta_T, create_dissector_handle(dissect_GANSS_Delta_T_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_OrdinalNumberOfFrequency, create_dissector_handle(dissect_OrdinalNumberOfFrequency_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Multicell_EDCH_Restriction, create_dissector_handle(dissect_Multicell_EDCH_Restriction_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CellCapabilityContainerExtension_TDD_LCR, create_dissector_handle(dissect_CellCapabilityContainerExtension_TDD_LCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Multi_Carrier_EDCH_Setup, create_dissector_handle(dissect_Multi_Carrier_EDCH_Info_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Multi_Carrier_EDCH_Reconfigure, create_dissector_handle(dissect_Multi_Carrier_EDCH_Reconfigure_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Multi_Carrier_EDCH_Response, create_dissector_handle(dissect_Multi_Carrier_EDCH_Information_Response_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_SNPL_Carrier_Group_Indicator, create_dissector_handle(dissect_SNPL_Carrier_Group_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MU_MIMO_InformationLCR, create_dissector_handle(dissect_MU_MIMO_InformationLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MU_MIMO_Indicator, create_dissector_handle(dissect_MU_MIMO_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_HS_SCCH_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext, create_dissector_handle(dissect_Inactivity_Threshold_for_UE_DRX_Cycle_LCR_Ext_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Usefulness_Of_Battery_Optimization, create_dissector_handle(dissect_Usefulness_Of_Battery_Optimization_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Multi_Carrier_E_DCH_LCRTDD_PhysicalLayerCategory, create_dissector_handle(dissect_Multi_Carrier_E_DCH_LCRTDD_PhysicalLayerCategory_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Measurement_Power_Offset, create_dissector_handle(dissect_Measurement_Power_Offset_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MDT_Configuration, create_dissector_handle(dissect_MDT_Configuration_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Neighbouring_UMTS_CellInformation_Ext, create_dissector_handle(dissect_Neighbouring_UMTS_CellInformation_Ext_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_ANRReportIndication, create_dissector_handle(dissect_ANRReportIndication_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Trace_Collection_Entity_IP_Address, create_dissector_handle(dissect_TransportLayerAddress_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Affected_HSDSCH_Serving_Cell_List, create_dissector_handle(dissect_Affected_HSDSCH_Serving_Cell_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_CLTD_Information, create_dissector_handle(dissect_UL_CLTD_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_CLTD_Information_Reconf, create_dissector_handle(dissect_UL_CLTD_Information_Reconf_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_CLTD_State_Update_Information, create_dissector_handle(dissect_UL_CLTD_State_Update_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order, create_dissector_handle(dissect_Support_of_Dynamic_DTXDRX_Related_HS_SCCH_Order_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CPC_RecoveryReport, create_dissector_handle(dissect_CPC_RecoveryReport_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_FTPICH_Information, create_dissector_handle(dissect_FTPICH_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_FTPICH_Information_Reconf, create_dissector_handle(dissect_FTPICH_Information_Reconf_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UE_RF_Band_CapabilityLCR, create_dissector_handle(dissect_UE_RF_Band_CapabilityLCR_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Extended_S_RNTI, create_dissector_handle(dissect_Extended_RNTI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_ExtendedAffectedUEInformationForMBMS, create_dissector_handle(dissect_ExtendedAffectedUEInformationForMBMS_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Extended_S_RNTI_Group, create_dissector_handle(dissect_Extended_S_RNTI_Group_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_FTPICH_Information_Response, create_dissector_handle(dissect_FTPICH_Information_Response_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_FTPICH_Reconfiguration_Information, create_dissector_handle(dissect_FTPICH_Reconfiguration_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MIMO_withfourtransmitantennas_ActivationIndicator, create_dissector_handle(dissect_MIMO_withfourtransmitantennas_ActivationIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MIMO_withfourtransmitantennas_ModeIndicator, create_dissector_handle(dissect_MIMO_withfourtransmitantennas_ModeIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_power_offset_for_S_CPICH_for_MIMO_withfourtransmitantennas, create_dissector_handle(dissect_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennas_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_power_offset_for_S_CPICH_for_MIMO_withfourtransmitantennas_Request_Indicator, create_dissector_handle(dissect_PowerOffsetForSecondaryCPICHforMIMOwithfourtransmitantennasRequestIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DualStream_MIMO_withfourtransmitantennas_ActivationIndicator, create_dissector_handle(dissect_DualStream_MIMO_withfourtransmitantennas_ActivationIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DualStream_MIMO_withfourtransmitantennas_ModeIndicator, create_dissector_handle(dissect_DualStream_MIMO_withfourtransmitantennas_ModeIndicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Additional_Associated_Secondary_CPICH, create_dissector_handle(dissect_Additional_Associated_Secondary_CPICH_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Multiflow_Information, create_dissector_handle(dissect_Multiflow_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Multiflow_Reconfiguration, create_dissector_handle(dissect_Multiflow_Reconfiguration_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Multiflow_OrdinalNumberOfFrequency, create_dissector_handle(dissect_Multiflow_OrdinalNumberOfFrequency_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_MIMO_Information, create_dissector_handle(dissect_UL_MIMO_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_MIMO_Reconfiguration, create_dissector_handle(dissect_UL_MIMO_Reconfiguration_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_MIMO_DL_Control_Channel_Information, create_dissector_handle(dissect_UL_MIMO_DL_Control_Channel_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_SixtyfourQAM_UL_Operation_Indicator, create_dissector_handle(dissect_SixtyfourQAM_UL_Operation_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_Common_E_RGCH_Cell_InfEx_Rprt, create_dissector_handle(dissect_Common_E_RGCH_Cell_InfEx_Rprt_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Common_E_RGCH_Cell_Information, create_dissector_handle(dissect_Common_E_RGCH_Cell_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_PCI, create_dissector_handle(dissect_PCI_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TAC, create_dissector_handle(dissect_TAC_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_PLMN_List, create_dissector_handle(dissect_PLMN_List_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_PrecoderWeightSetRestriction, create_dissector_handle(dissect_Precoder_Weight_Set_Restriction_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_EARFCN_FDD_Extended, create_dissector_handle(dissect_EARFCN_FDD_Extended_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_EARFCN_TDD_Extended, create_dissector_handle(dissect_EARFCN_Extended_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Non_rectangular_resource_allocation_indicator, create_dissector_handle(dissect_Non_rectangular_resource_allocation_indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Non_rectangular_resource_timeslot_set, create_dissector_handle(dissect_Non_rectangular_resource_timeslot_set_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UE_Support_of_non_rectangular_resource_allocation, create_dissector_handle(dissect_UE_Support_of_non_rectangular_resource_allocation_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DBDS_CorrectionsReq, create_dissector_handle(dissect_DBDS_CorrectionsReq_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DBDS_Corrections, create_dissector_handle(dissect_DBDS_Corrections_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_BDS_IonosphericGridModelReq, create_dissector_handle(dissect_BDS_IonosphericGridModelReq_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_BDS_Ionospheric_Grid_Model, create_dissector_handle(dissect_BDS_Ionospheric_Grid_Model_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GANSS_alm_keplerianBDSAlmanac, create_dissector_handle(dissect_GANSS_alm_KEPLERIANBDSALMANAC_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Assisting_RepetitionFactors, create_dissector_handle(dissect_Multiflow_RepetitionFactors_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_E_DCH_Decoupling_Indication, create_dissector_handle(dissect_E_DCH_Decoupling_Indication_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UE_Measurement_Forwarding, create_dissector_handle(dissect_UE_Measurement_Forwarding_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UPH_Filtering_Measurement_Forwarding_Request, create_dissector_handle(dissect_UPH_Filtering_Measurement_Forwarding_Request_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TTI_Update_Indicator, create_dissector_handle(dissect_TTI_Update_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CQI_Feedback_Cycle2_k, create_dissector_handle(dissect_CQI_Feedback_Cycle2_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_CQI_Cycle_Switch_Timer, create_dissector_handle(dissect_CQI_Cycle_Switch_Timer_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UE_DRX_Cycle2, create_dissector_handle(dissect_UE_DRX_Cycle_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Inactivity_Threshold_for_UE_DRX_Cycle2, create_dissector_handle(dissect_Inactivity_Threshold_for_UE_DRX_Cycle_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DTX_Information2, create_dissector_handle(dissect_DTX_Information2_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DCH_ENH_Information, create_dissector_handle(dissect_DCH_ENH_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_DCH_ENH_Information_Reconf, create_dissector_handle(dissect_DCH_ENH_Information_Reconf_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Gainfactors_10ms_mode, create_dissector_handle(dissect_Gainfactors_10ms_mode_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Radio_Links_without_DPCH_FDPCH_Indication, create_dissector_handle(dissect_Radio_Links_without_DPCH_FDPCH_Indication_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_DPCCH2_Information, create_dissector_handle(dissect_UL_DPCCH2_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_UL_DPCCH2_Information_Reconf, create_dissector_handle(dissect_UL_DPCCH2_Information_Reconf_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_ImplicitGrantHandling, create_dissector_handle(dissect_Implicit_Grant_Handling_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MinimumTEBSthreshold, create_dissector_handle(dissect_Minimum_TEBS_threshold_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MultipleFrequencyBandsIndicatorList_EUTRA, create_dissector_handle(dissect_MultipleFrequencyBandsIndicatorList_EUTRA_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_MultipleFrequencyBandsIndicatorList, create_dissector_handle(dissect_MultipleFrequencyBandsIndicatorList_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_ActivationDelay, create_dissector_handle(dissect_ActivationDelay_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Fast_TTI_switching_Mode_synchronized, create_dissector_handle(dissect_Fast_TTI_switching_Mode_synchronized_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Fast_TTI_switching_Mode_unsynchronized, create_dissector_handle(dissect_Fast_TTI_switching_Mode_unsynchronized_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Fast_TTI_switching_Mode_Supported, create_dissector_handle(dissect_Fast_TTI_switching_Mode_Supported_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GANSS_BDSclockModel, create_dissector_handle(dissect_GANSS_BDSclockModel_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GANSS_NavModel_BDSKeplerianSet, create_dissector_handle(dissect_GANSS_NavModel_BDSKeplerianSet_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GANSS_UTCmodelSet4, create_dissector_handle(dissect_GANSS_UTCmodelSet4_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Inter_Frequency_Cell_Information_SIB11ter, create_dissector_handle(dissect_Inter_Frequency_Cell_Information_SIB11ter_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Improved_Synchronized_Indicator, create_dissector_handle(dissect_Improved_Synchronized_Indicator_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Downlink_TPC_enhancements_Information, create_dissector_handle(dissect_Downlink_TPC_enhancements_Information_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Downlink_TPC_enhancements_Reconf, create_dissector_handle(dissect_Downlink_TPC_enhancements_Reconf_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_TPC_slot_position, create_dissector_handle(dissect_TPC_slot_position_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_DL_TBS, create_dissector_handle(dissect_DL_TBS_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Selected_PLMN_Identity, create_dissector_handle(dissect_PLMN_Identity_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_SpeechVersion, create_dissector_handle(dissect_SpeechVersion_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_SourceID, create_dissector_handle(dissect_SourceID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_TargetID, create_dissector_handle(dissect_TargetID_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_ClassmarkInformation2, create_dissector_handle(dissect_ClassmarkInformation2_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_ClassmarkInformation3, create_dissector_handle(dissect_ClassmarkInformation3_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GSM_Cell_CM_Rqst, create_dissector_handle(dissect_GSM_Cell_CM_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Extension_CommonMeasurementObjectType_CM_Rprt, create_dissector_handle(dissect_Extension_CommonMeasurementObjectType_CM_Rprt_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Extension_CommonMeasurementObjectType_CM_Rqst, create_dissector_handle(dissect_Extension_CommonMeasurementObjectType_CM_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Extension_CommonMeasurementObjectType_CM_Rsp, create_dissector_handle(dissect_Extension_CommonMeasurementObjectType_CM_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Extension_FailureIndicationMeasurementList, create_dissector_handle(dissect_Extension_FailureIndicationMeasurementList_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Extension_FailureMeasurementList, create_dissector_handle(dissect_Extension_FailureMeasurementList_PDU, proto_rnsap));
  dissector_add_uint("rnsap.extension", id_Extension_TerminationMeasurementList, create_dissector_handle(dissect_Extension_TerminationMeasurementList_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GsmCellList_CM_Rprt, create_dissector_handle(dissect_GsmCellList_CM_Rprt_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GsmCellList_CM_Rqst, create_dissector_handle(dissect_GsmCellList_CM_Rqst_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_GsmCellList_CM_Rsp, create_dissector_handle(dissect_GsmCellList_CM_Rsp_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_LoadValue, create_dissector_handle(dissect_LoadValue_PDU, proto_rnsap));
  dissector_add_uint("rnsap.ies", id_EventH, create_dissector_handle(dissect_EventH_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-radioLinkSetup/fdd", create_dissector_handle(dissect_RadioLinkSetupRequestFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-radioLinkSetup/fdd", create_dissector_handle(dissect_RadioLinkSetupResponseFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-radioLinkSetup/fdd", create_dissector_handle(dissect_RadioLinkSetupFailureFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-radioLinkSetup/tdd", create_dissector_handle(dissect_RadioLinkSetupRequestTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-radioLinkSetup/tdd", create_dissector_handle(dissect_RadioLinkSetupResponseTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-radioLinkSetup/tdd", create_dissector_handle(dissect_RadioLinkSetupFailureTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-radioLinkAddition/fdd", create_dissector_handle(dissect_RadioLinkAdditionRequestFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-radioLinkAddition/fdd", create_dissector_handle(dissect_RadioLinkAdditionResponseFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-radioLinkAddition/fdd", create_dissector_handle(dissect_RadioLinkAdditionFailureFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-radioLinkAddition/tdd", create_dissector_handle(dissect_RadioLinkAdditionRequestTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-radioLinkAddition/tdd", create_dissector_handle(dissect_RadioLinkAdditionResponseTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-radioLinkAddition/tdd", create_dissector_handle(dissect_RadioLinkAdditionFailureTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-radioLinkDeletion/common", create_dissector_handle(dissect_RadioLinkDeletionRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-radioLinkDeletion/common", create_dissector_handle(dissect_RadioLinkDeletionResponse_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-synchronisedRadioLinkReconfigurationPreparation/fdd", create_dissector_handle(dissect_RadioLinkReconfigurationPrepareFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-synchronisedRadioLinkReconfigurationPreparation/fdd", create_dissector_handle(dissect_RadioLinkReconfigurationReadyFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-synchronisedRadioLinkReconfigurationPreparation/fdd", create_dissector_handle(dissect_RadioLinkReconfigurationFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-synchronisedRadioLinkReconfigurationPreparation/tdd", create_dissector_handle(dissect_RadioLinkReconfigurationPrepareTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-synchronisedRadioLinkReconfigurationPreparation/tdd", create_dissector_handle(dissect_RadioLinkReconfigurationReadyTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-synchronisedRadioLinkReconfigurationPreparation/tdd", create_dissector_handle(dissect_RadioLinkReconfigurationFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-unSynchronisedRadioLinkReconfiguration/fdd", create_dissector_handle(dissect_RadioLinkReconfigurationRequestFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-unSynchronisedRadioLinkReconfiguration/fdd", create_dissector_handle(dissect_RadioLinkReconfigurationResponseFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-unSynchronisedRadioLinkReconfiguration/fdd", create_dissector_handle(dissect_RadioLinkReconfigurationFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-unSynchronisedRadioLinkReconfiguration/tdd", create_dissector_handle(dissect_RadioLinkReconfigurationRequestTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-unSynchronisedRadioLinkReconfiguration/tdd", create_dissector_handle(dissect_RadioLinkReconfigurationResponseTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-unSynchronisedRadioLinkReconfiguration/tdd", create_dissector_handle(dissect_RadioLinkReconfigurationFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-physicalChannelReconfiguration/fdd", create_dissector_handle(dissect_PhysicalChannelReconfigurationRequestFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-physicalChannelReconfiguration/fdd", create_dissector_handle(dissect_PhysicalChannelReconfigurationCommand_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-physicalChannelReconfiguration/fdd", create_dissector_handle(dissect_PhysicalChannelReconfigurationFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-physicalChannelReconfiguration/tdd", create_dissector_handle(dissect_PhysicalChannelReconfigurationRequestTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-physicalChannelReconfiguration/tdd", create_dissector_handle(dissect_PhysicalChannelReconfigurationCommand_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-physicalChannelReconfiguration/tdd", create_dissector_handle(dissect_PhysicalChannelReconfigurationFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-dedicatedMeasurementInitiation/common", create_dissector_handle(dissect_DedicatedMeasurementInitiationRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-dedicatedMeasurementInitiation/common", create_dissector_handle(dissect_DedicatedMeasurementInitiationResponse_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-dedicatedMeasurementInitiation/common", create_dissector_handle(dissect_DedicatedMeasurementInitiationFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-commonTransportChannelResourcesInitialisation/fdd", create_dissector_handle(dissect_CommonTransportChannelResourcesRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-commonTransportChannelResourcesInitialisation/fdd", create_dissector_handle(dissect_CommonTransportChannelResourcesResponseFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-commonTransportChannelResourcesInitialisation/fdd", create_dissector_handle(dissect_CommonTransportChannelResourcesFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-commonTransportChannelResourcesInitialisation/tdd", create_dissector_handle(dissect_CommonTransportChannelResourcesRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-commonTransportChannelResourcesInitialisation/tdd", create_dissector_handle(dissect_CommonTransportChannelResourcesResponseTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-commonTransportChannelResourcesInitialisation/tdd", create_dissector_handle(dissect_CommonTransportChannelResourcesFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-uplinkSignallingTransfer/fdd", create_dissector_handle(dissect_UplinkSignallingTransferIndicationFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-uplinkSignallingTransfer/tdd", create_dissector_handle(dissect_UplinkSignallingTransferIndicationTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-downlinkSignallingTransfer/common", create_dissector_handle(dissect_DownlinkSignallingTransferRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-relocationCommit/common", create_dissector_handle(dissect_RelocationCommit_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-paging/common", create_dissector_handle(dissect_PagingRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-synchronisedRadioLinkReconfigurationCommit/common", create_dissector_handle(dissect_RadioLinkReconfigurationCommit_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-synchronisedRadioLinkReconfigurationCancellation/common", create_dissector_handle(dissect_RadioLinkReconfigurationCancel_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-radioLinkFailure/common", create_dissector_handle(dissect_RadioLinkFailureIndication_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-radioLinkPreemption/common", create_dissector_handle(dissect_RadioLinkPreemptionRequiredIndication_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-radioLinkRestoration/common", create_dissector_handle(dissect_RadioLinkRestoreIndication_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-dedicatedMeasurementReporting/common", create_dissector_handle(dissect_DedicatedMeasurementReport_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-dedicatedMeasurementTermination/common", create_dissector_handle(dissect_DedicatedMeasurementTerminationRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-dedicatedMeasurementFailure/common", create_dissector_handle(dissect_DedicatedMeasurementFailureIndication_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-radioLinkCongestion/common", create_dissector_handle(dissect_RadioLinkCongestionIndication_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-downlinkPowerControl/fdd", create_dissector_handle(dissect_DL_PowerControlRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-downlinkPowerTimeslotControl/tdd", create_dissector_handle(dissect_DL_PowerTimeslotControlRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-compressedModeCommand/fdd", create_dissector_handle(dissect_CompressedModeCommand_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-commonTransportChannelResourcesRelease/common", create_dissector_handle(dissect_CommonTransportChannelResourcesReleaseRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-errorIndication/common", create_dissector_handle(dissect_ErrorIndication_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-commonMeasurementInitiation/common", create_dissector_handle(dissect_CommonMeasurementInitiationRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-commonMeasurementInitiation/common", create_dissector_handle(dissect_CommonMeasurementInitiationResponse_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-commonMeasurementInitiation/common", create_dissector_handle(dissect_CommonMeasurementInitiationFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-commonMeasurementReporting/common", create_dissector_handle(dissect_CommonMeasurementReport_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-commonMeasurementTermination/common", create_dissector_handle(dissect_CommonMeasurementTerminationRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-commonMeasurementFailure/common", create_dissector_handle(dissect_CommonMeasurementFailureIndication_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-informationExchangeInitiation/common", create_dissector_handle(dissect_InformationExchangeInitiationRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-informationExchangeInitiation/common", create_dissector_handle(dissect_InformationExchangeInitiationResponse_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-informationExchangeInitiation/common", create_dissector_handle(dissect_InformationExchangeInitiationFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-informationReporting/common", create_dissector_handle(dissect_InformationReport_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-informationExchangeTermination/common", create_dissector_handle(dissect_InformationExchangeTerminationRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-informationExchangeFailure/common", create_dissector_handle(dissect_InformationExchangeFailureIndication_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-privateMessage/common", create_dissector_handle(dissect_PrivateMessage_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-reset/common", create_dissector_handle(dissect_ResetRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-reset/common", create_dissector_handle(dissect_ResetResponse_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-radioLinkActivation/fdd", create_dissector_handle(dissect_RadioLinkActivationCommandFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-radioLinkActivation/tdd", create_dissector_handle(dissect_RadioLinkActivationCommandTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-gERANuplinkSignallingTransfer/common", create_dissector_handle(dissect_GERANUplinkSignallingTransferIndication_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-radioLinkParameterUpdate/fdd", create_dissector_handle(dissect_RadioLinkParameterUpdateIndicationFDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-radioLinkParameterUpdate/tdd", create_dissector_handle(dissect_RadioLinkParameterUpdateIndicationTDD_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-uEMeasurementInitiation/tdd", create_dissector_handle(dissect_UEMeasurementInitiationRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-uEMeasurementInitiation/tdd", create_dissector_handle(dissect_UEMeasurementInitiationResponse_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-uEMeasurementInitiation/tdd", create_dissector_handle(dissect_UEMeasurementInitiationFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-uEMeasurementReporting/tdd", create_dissector_handle(dissect_UEMeasurementReport_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-uEMeasurementTermination/tdd", create_dissector_handle(dissect_UEMeasurementTerminationRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-uEMeasurementFailure/tdd", create_dissector_handle(dissect_UEMeasurementFailureIndication_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-iurInvokeTrace/common", create_dissector_handle(dissect_IurInvokeTrace_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-iurDeactivateTrace/common", create_dissector_handle(dissect_IurDeactivateTrace_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-mBMSAttach/common", create_dissector_handle(dissect_MBMSAttachCommand_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-mBMSDetach/common", create_dissector_handle(dissect_MBMSDetachCommand_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-directInformationTransfer/common", create_dissector_handle(dissect_DirectInformationTransfer_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-enhancedRelocation/common", create_dissector_handle(dissect_EnhancedRelocationRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-enhancedRelocation/common", create_dissector_handle(dissect_EnhancedRelocationResponse_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-enhancedRelocation/common", create_dissector_handle(dissect_EnhancedRelocationFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-enhancedRelocationCancel/common", create_dissector_handle(dissect_EnhancedRelocationCancel_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-enhancedRelocationSignallingTransfer/common", create_dissector_handle(dissect_EnhancedRelocationSignallingTransfer_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-enhancedRelocationRelease/common", create_dissector_handle(dissect_EnhancedRelocationRelease_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-mBSFNMCCHInformation/common", create_dissector_handle(dissect_MBSFNMCCHInformation_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-secondaryULFrequencyReporting/fdd", create_dissector_handle(dissect_SecondaryULFrequencyReport_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-secondaryULFrequencyUpdate/fdd", create_dissector_handle(dissect_SecondaryULFrequencyUpdateIndication_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-enhancedRelocationResourceAllocation/tdd", create_dissector_handle(dissect_EnhancedRelocationResourceRequest_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-enhancedRelocationResourceAllocation/tdd", create_dissector_handle(dissect_EnhancedRelocationResourceResponse_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.uout", "id-enhancedRelocationResourceAllocation/tdd", create_dissector_handle(dissect_EnhancedRelocationResourceFailure_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-enhancedRelocationResourceRelease/tdd", create_dissector_handle(dissect_EnhancedRelocationResourceReleaseCommand_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.sout", "id-enhancedRelocationResourceRelease/tdd", create_dissector_handle(dissect_EnhancedRelocationResourceReleaseComplete_PDU, proto_rnsap));
  dissector_add_string("rnsap.proc.imsg", "id-informationTransferControl/common", create_dissector_handle(dissect_InformationTransferControlRequest_PDU, proto_rnsap));

}


